;;; phps-mode-parser.el --- Exported Emacs Parser Generator -*- lexical-binding: t -*-

;; Copyright (C) 2018-2023  Free Software Foundation, Inc.

;; This file is not part of GNU Emacs.


;;; Commentary:


;;; Code:


(defvar-local
 phps-mode-parser-tokens
 nil
 "Tokens for parser.")

(define-error
 'phps-parser-error
 "PHPs Parser Error")


;;; Variables:


(defvar
  phps-mode-parser--action-tables
  #s(hash-table size 28048 test equal rehash-size 1.5 rehash-threshold 0.8125 data (0 0 1 1 2 2 3 3 4 4 5 2 6 2 7 2 8 8 9 2 10 10 11 11 12 2 13 13 14 2 15 15 16 16 17 17 18 2 19 16 20 20 21 15 22 22 23 11 24 24 25 25 26 26 27 2 28 2 29 11 30 16 31 11 32 32 33 33 34 34 35 35 36 11 37 11 38 38 39 39 40 40 41 16 42 11 43 11 44 22 45 2 46 2 47 47 48 16 49 2 50 11 51 51 52 11 53 53 54 11 55 55 56 56 57 57 58 58 59 59 60 60 61 61 62 2 63 2 64 2 65 2 66 15 67 67 68 68 69 69 70 2 71 11 72 2 73 16 74 74 75 75 76 11 77 2 78 78 79 79 80 11 81 81 82 2 83 83 84 84 85 85 86 86 87 87 88 88 89 89 90 90 91 11 92 92 93 93 94 94 95 95 96 96 97 97 98 98 99 99 100 100 101 101 102 102 103 103 104 104 105 105 106 106 107 107 108 108 109 109 110 110 111 111 112 112 113 113 114 114 115 115 116 116 117 117 118 118 119 119 120 120 121 121 122 97 123 123 124 2 125 13 126 126 127 127 128 128 129 129 130 130 131 131 132 102 133 133 134 134 135 135 136 136 137 11 138 138 139 13 140 140 141 141 142 142 143 143 144 144 145 145 146 146 147 147 148 11 149 149 150 150 151 151 152 75 153 123 154 154 155 155 156 15 157 16 158 15 159 11 160 26 161 2 162 16 163 11 164 11 165 40 166 16 167 11 168 11 169 169 170 16 171 15 172 68 173 69 174 11 175 16 176 75 177 11 178 11 179 179 180 86 181 181 182 90 183 183 184 96 185 185 186 101 187 104 188 188 189 189 190 190 191 191 192 192 193 193 194 194 195 123 196 196 197 197 198 154 199 199 200 26 201 11 202 2 203 2 204 3 205 4 206 2 207 2 208 2 209 2 210 11 211 2 212 2 213 213 214 2 215 215 216 22 217 217 218 218 219 2 220 11 221 11 222 222 223 223 224 224 225 22 226 2 227 2 228 2 229 11 230 230 231 11 232 232 233 11 234 234 235 235 236 236 237 237 238 60 239 239 240 2 241 2 242 2 243 2 244 67 245 127 246 246 247 2 248 2 249 249 250 2 251 251 252 252 253 2 254 83 255 84 256 87 257 130 258 11 259 259 260 260 261 97 262 262 263 263 264 264 265 102 266 102 267 267 268 268 269 269 270 270 271 271 272 272 273 273 274 274 275 275 276 276 277 97 278 2 279 279 280 2 281 2 282 2 283 2 284 2 285 2 286 2 287 2 288 288 289 2 290 2 291 2 292 2 293 2 294 294 295 2 296 2 297 2 298 2 299 2 300 2 301 2 302 2 303 2 304 2 305 2 306 2 307 2 308 2 309 2 310 310 311 311 312 312 313 313 314 314 315 315 316 316 317 317 318 318 319 319 320 320 321 321 322 322 323 323 324 324 325 4 326 2 327 327 328 328 329 329 330 330 331 331 332 332 333 333 334 334 335 335 336 336 337 337 338 338 339 338 340 40 341 341 342 2 343 2 344 3 345 4 346 2 347 2 348 2 349 2 350 11 351 2 352 2 353 353 354 2 355 355 356 22 357 357 358 358 359 2 360 11 361 11 362 362 363 363 364 364 365 22 366 2 367 2 368 2 369 11 370 370 371 11 372 372 373 11 374 374 375 375 376 376 377 377 378 60 379 379 380 2 381 2 382 2 383 2 384 67 385 127 386 386 387 2 388 2 389 389 390 2 391 391 392 392 393 2 394 83 395 84 396 87 397 130 398 11 399 399 400 400 401 97 402 402 403 403 404 404 405 102 406 102 407 407 408 408 409 409 410 410 411 411 412 412 413 413 414 414 415 415 416 416 417 97 418 2 419 419 420 2 421 2 422 2 423 2 424 2 425 2 426 2 427 2 428 288 429 2 430 2 431 2 432 2 433 2 434 294 435 2 436 2 437 2 438 2 439 2 440 2 441 2 442 2 443 2 444 2 445 2 446 2 447 2 448 2 449 2 450 450 451 451 452 452 453 453 454 454 455 455 456 456 457 457 458 458 459 459 460 460 461 461 462 462 463 463 464 464 465 4 466 2 467 467 468 468 469 469 470 470 471 471 472 472 473 473 474 474 475 475 476 476 477 477 478 338 479 338 480 40 481 341 482 2 483 404 484 484 485 2 486 3 487 4 488 2 489 2 490 2 491 2 492 11 493 2 494 2 495 495 496 2 497 497 498 22 499 499 500 500 501 2 502 11 503 11 504 504 505 505 506 506 507 22 508 2 509 2 510 2 511 11 512 512 513 11 514 514 515 11 516 516 517 517 518 518 519 519 520 60 521 521 522 2 523 2 524 2 525 2 526 67 527 127 528 528 529 2 530 2 531 531 532 2 533 533 534 534 535 2 536 83 537 84 538 87 539 130 540 11 541 541 542 542 543 97 544 544 545 545 546 546 547 102 548 102 549 549 550 550 551 551 552 552 553 553 554 554 555 555 556 556 557 557 558 558 559 559 560 97 561 2 562 562 563 2 564 2 565 2 566 2 567 2 568 2 569 2 570 2 571 288 572 2 573 2 574 2 575 2 576 2 577 294 578 2 579 2 580 2 581 2 582 2 583 2 584 2 585 2 586 2 587 2 588 2 589 2 590 2 591 2 592 2 593 593 594 594 595 595 596 596 597 597 598 598 599 599 600 600 601 601 602 602 603 603 604 604 605 605 606 606 607 607 608 4 609 2 610 610 611 611 612 612 613 613 614 614 615 615 616 616 617 617 618 618 619 619 620 620 621 338 622 338 623 40 624 341 625 2 626 404 627 627 628 555 629 629 630 630 631 631 632 632 633 633 634 2 635 404 636 636 637 637 638 40 639 639 640 264 641 641 642 642 643 2 644 404 645 645 646 646 647 647 648 648 649 649 650 650 651 2 652 3 653 4 654 2 655 2 656 2 657 2 658 2 659 11 660 2 661 2 662 662 663 2 664 664 665 22 666 666 667 667 668 2 669 11 670 11 671 671 672 672 673 673 674 22 675 2 676 2 677 2 678 11 679 679 680 11 681 681 682 11 683 683 684 684 685 685 686 686 687 60 688 688 689 2 690 2 691 2 692 2 693 67 694 127 695 695 696 2 697 2 698 698 699 2 700 700 701 701 702 2 703 83 704 84 705 87 706 130 707 11 708 708 709 709 710 97 711 711 712 712 713 713 714 102 715 102 716 716 717 717 718 718 719 719 720 720 721 721 722 722 723 723 724 724 725 725 726 97 727 2 728 728 729 2 730 2 731 2 732 2 733 2 734 2 735 2 736 2 737 288 738 2 739 2 740 2 741 2 742 2 743 294 744 2 745 2 746 2 747 2 748 2 749 2 750 2 751 2 752 2 753 2 754 2 755 2 756 2 757 2 758 2 759 759 760 760 761 761 762 762 763 763 764 764 765 765 766 766 767 767 768 768 769 769 770 770 771 771 772 772 773 773 774 4 775 2 776 776 777 777 778 778 779 779 780 780 781 781 782 782 783 783 784 784 785 785 786 786 787 338 788 338 789 40 790 341 791 2 792 404 793 793 794 555 795 795 796 796 797 797 798 798 799 799 800 2 801 404 802 802 803 803 804 40 805 805 806 264 807 807 808 808 809 2 810 404 811 811 812 812 813 813 814 814 815 815 816 816 817 2 818 713 819 2 820 820 821 821 822 822 823 823 824 824 825 825 826 826 827 827 828 828 829 829 830 830 831 831 832 832 833 833 834 834 835 835 836 836 837 837 838 838 839 839 840 840 841 841 842 842 843 843 844 844 845 845 846 846 847 847 848 848 849 849 850 850 851 851 852 852 853 853 854 854 855 855 856 856 857 857 858 858 859 859 860 860 861 861 862 862 863 863 864 864 865 865 866 866 867 867 868 868 869 869 870 870 871 871 872 872 873 873 874 874 875 875 876 876 877 877 878 878 879 879 880 880 881 881 882 882 883 883 884 884 885 885 886 886 887 887 888 888 889 889 890 890 891 891 892 892 893 893 894 894 895 895 896 896 897 897 898 898 899 899 900 900 901 901 902 902 903 903 904 904 905 905 906 906 907 907 908 908 909 909 910 11 911 911 912 912 913 913 914 2 915 404 916 916 917 917 918 918 919 2 920 3 921 4 922 2 923 923 924 2 925 2 926 2 927 927 928 928 929 2 930 930 931 2 932 932 933 933 934 934 935 935 936 936 937 937 938 938 939 939 940 940 941 941 942 22 943 943 944 944 945 945 946 946 947 947 948 2 949 949 950 950 951 951 952 952 953 953 954 954 955 955 956 956 957 957 958 958 959 959 960 960 961 961 962 962 963 963 964 964 965 965 966 966 967 967 968 968 969 969 970 970 971 971 972 972 973 973 974 974 975 975 976 22 977 977 978 978 979 979 980 980 981 981 982 2 983 983 984 984 985 985 986 986 987 987 988 988 989 989 990 990 991 991 992 992 993 993 994 994 995 995 996 996 997 997 998 2 999 999 1000 1000 1001 1001 1002 1002 1003 1003 1004 1004 1005 1005 1006 1006 1007 67 1008 1008 1009 1009 1010 2 1011 1011 1012 1012 1013 1013 1014 1014 1015 1015 1016 1016 1017 2 1018 1018 1019 1019 1020 1020 1021 1021 1022 1022 1023 2 1024 83 1025 84 1026 1026 1027 87 1028 130 1029 11 1030 1030 1031 1031 1032 97 1033 1033 1034 1034 1035 1035 1036 102 1037 102 1038 1038 1039 1039 1040 1040 1041 1041 1042 1042 1043 1043 1044 1044 1045 143 1046 1046 1047 1047 1048 1048 1049 1049 1050 1050 1051 1051 1052 97 1053 2 1054 11 1055 1055 1056 2 1057 1057 1058 11 1059 11 1060 1060 1061 1061 1062 1062 1063 2 1064 2 1065 11 1066 1066 1067 11 1068 11 1069 1069 1070 60 1071 1071 1072 2 1073 2 1074 2 1075 127 1076 1076 1077 2 1078 1078 1079 1079 1080 1080 1081 2 1082 2 1083 2 1084 2 1085 2 1086 2 1087 2 1088 2 1089 288 1090 2 1091 2 1092 2 1093 2 1094 2 1095 294 1096 2 1097 2 1098 2 1099 2 1100 2 1101 2 1102 2 1103 2 1104 2 1105 2 1106 2 1107 2 1108 2 1109 2 1110 2 1111 1111 1112 1112 1113 1113 1114 1114 1115 1115 1116 1116 1117 1117 1118 1118 1119 1119 1120 1120 1121 1121 1122 1122 1123 1123 1124 1124 1125 1125 1126 4 1127 2 1128 1128 1129 1129 1130 1130 1131 1131 1132 1132 1133 1133 1134 1134 1135 1135 1136 1136 1137 1137 1138 1138 1139 338 1140 338 1141 40 1142 341 1143 2 1144 404 1145 1145 1146 555 1147 1147 1148 1148 1149 1149 1150 1150 1151 1151 1152 2 1153 404 1154 1154 1155 1155 1156 40 1157 1157 1158 264 1159 1159 1160 1160 1161 2 1162 404 1163 1163 1164 1164 1165 1165 1166 1166 1167 1167 1168 1168 1169 2 1170 713 1171 2 1172 1172 1173 1173 1174 1174 1175 1175 1176 1176 1177 1177 1178 1178 1179 1179 1180 1180 1181 1181 1182 2 1183 3 1184 4 1185 2 1186 2 1187 2 1188 2 1189 11 1190 2 1191 2 1192 1192 1193 2 1194 1194 1195 22 1196 1196 1197 1197 1198 2 1199 11 1200 11 1201 1201 1202 1202 1203 1203 1204 22 1205 2 1206 2 1207 2 1208 11 1209 1209 1210 11 1211 1211 1212 11 1213 1213 1214 1214 1215 1215 1216 1216 1217 60 1218 1218 1219 2 1220 2 1221 2 1222 2 1223 67 1224 127 1225 1225 1226 2 1227 2 1228 1228 1229 2 1230 1230 1231 1231 1232 2 1233 83 1234 84 1235 87 1236 130 1237 11 1238 1238 1239 1239 1240 97 1241 1241 1242 1242 1243 1243 1244 102 1245 102 1246 1246 1247 1247 1248 1248 1249 1249 1250 1250 1251 1251 1252 1252 1253 1253 1254 1254 1255 1255 1256 97 1257 2 1258 1258 1259 2 1260 2 1261 2 1262 2 1263 2 1264 2 1265 2 1266 2 1267 288 1268 2 1269 2 1270 2 1271 2 1272 2 1273 294 1274 2 1275 2 1276 2 1277 2 1278 2 1279 2 1280 2 1281 2 1282 2 1283 2 1284 2 1285 2 1286 2 1287 2 1288 2 1289 1289 1290 1290 1291 1291 1292 1292 1293 1293 1294 1294 1295 1295 1296 1296 1297 1297 1298 1298 1299 1299 1300 1300 1301 1301 1302 1302 1303 1303 1304 4 1305 2 1306 1306 1307 1307 1308 1308 1309 1309 1310 1310 1311 1311 1312 1312 1313 1313 1314 1314 1315 1315 1316 1316 1317 338 1318 338 1319 40 1320 341 1321 2 1322 404 1323 1323 1324 555 1325 1325 1326 1326 1327 1327 1328 1328 1329 1329 1330 2 1331 404 1332 1332 1333 1333 1334 40 1335 1335 1336 264 1337 1337 1338 1338 1339 2 1340 404 1341 1341 1342 1342 1343 1343 1344 1344 1345 1345 1346 1346 1347 2 1348 713 1349 2 1350 1350 1351 1351 1352 1352 1353 1353 1354 1354 1355 1355 1356 1356 1357 1357 1358 1358 1359 1359 1360 830 1361 1361 1362 1362 1363 1363 1364 1364 1365 1365 1366 1366 1367 1367 1368 1368 1369 1369 1370 1370 1371 1371 1372 1372 1373 1373 1374 1374 1375 1375 1376 1376 1377 1377 1378 1378 1379 1379 1380 1380 1381 1381 1382 1382 1383 1383 1384 1384 1385 1385 1386 1386 1387 1387 1388 1388 1389 1389 1390 1390 1391 1391 1392 1392 1393 1393 1394 1394 1395 1395 1396 1396 1397 1397 1398 1398 1399 1399 1400 1400 1401 1401 1402 1402 1403 1403 1404 1404 1405 1405 1406 1406 1407 1407 1408 1408 1409 1409 1410 1410 1411 1411 1412 1412 1413 1413 1414 1414 1415 1415 1416 1416 1417 1417 1418 1418 1419 1419 1420 1420 1421 1421 1422 1422 1423 1423 1424 1424 1425 1425 1426 1426 1427 1427 1428 1428 1429 1429 1430 1430 1431 1431 1432 1432 1433 1433 1434 1434 1435 1435 1436 1436 1437 1437 1438 1438 1439 1439 1440 11 1441 1441 1442 1442 1443 1443 1444 917 1445 1445 1446 1446 1447 950 1448 143 1449 1449 1450 146 1451 1451 1452 2 1453 1453 1454 1454 1455 1455 1456 1456 1457 2 1458 2 1459 2 1460 1460 1461 2 1462 1462 1463 2 1464 2 1465 2 1466 2 1467 2 1468 2 1469 2 1470 2 1471 2 1472 1472 1473 1473 1474 1474 1475 1475 1476 1476 1477 1477 1478 1478 1479 1479 1480 1480 1481 1481 1482 1482 1483 1483 1484 1484 1485 1485 1486 1486 1487 22 1488 1488 1489 3 1490 4 1491 2 1492 11 1493 1493 1494 1494 1495 1495 1496 1496 1497 1497 1498 1498 1499 1499 1500 1500 1501 1501 1502 1502 1503 1503 1504 1504 1505 1505 1506 1506 1507 1507 1508 83 1509 87 1510 11 1511 1511 1512 1512 1513 97 1514 1514 1515 1515 1516 1516 1517 1517 1518 1518 1519 1519 1520 1520 1521 97 1522 830 1523 1523 1524 1524 1525 1525 1526 1526 1527 1527 1528 1528 1529 1529 1530 1530 1531 1531 1532 1532 1533 1533 1534 1534 1535 1535 1536 1536 1537 1537 1538 1538 1539 1539 1540 1540 1541 1541 1542 1542 1543 1543 1544 1544 1545 1545 1546 1546 1547 1547 1548 1548 1549 1549 1550 1550 1551 1551 1552 1552 1553 1553 1554 1554 1555 1555 1556 1556 1557 1557 1558 1558 1559 1559 1560 1560 1561 1561 1562 1562 1563 1563 1564 1564 1565 1565 1566 1566 1567 1567 1568 1568 1569 1569 1570 1570 1571 1571 1572 1572 1573 1573 1574 1574 1575 1575 1576 1576 1577 1577 1578 1578 1579 1579 1580 1580 1581 1581 1582 1582 1583 1583 1584 1584 1585 1585 1586 1586 1587 1587 1588 1588 1589 1589 1590 1590 1591 1591 1592 1592 1593 1593 1594 1594 1595 1595 1596 1596 1597 1597 1598 1598 1599 1599 1600 1600 1601 1601 1602 11 1603 1603 1604 1604 1605 1605 1606 917 1607 1607 1608 1608 1609 950 1610 143 1611 146 1612 1612 1613 1613 1614 1614 1615 830 1616 1616 1617 11 1618 1618 1619 1619 1620 1620 1621 338 1622 338 1623 341 1624 2 1625 404 1626 1626 1627 555 1628 1628 1629 1629 1630 1630 1631 1631 1632 2 1633 404 1634 1634 1635 1635 1636 1636 1637 1637 1638 2 1639 3 1640 4 1641 2 1642 2 1643 2 1644 2 1645 11 1646 2 1647 2 1648 1648 1649 2 1650 1650 1651 22 1652 1652 1653 1653 1654 2 1655 2 1656 11 1657 11 1658 1658 1659 1659 1660 1660 1661 22 1662 2 1663 2 1664 2 1665 11 1666 1666 1667 11 1668 1668 1669 11 1670 1670 1671 1671 1672 1672 1673 1673 1674 60 1675 1675 1676 2 1677 2 1678 2 1679 2 1680 67 1681 127 1682 1682 1683 2 1684 2 1685 1685 1686 2 1687 1687 1688 1688 1689 2 1690 83 1691 84 1692 22 1693 87 1694 1694 1695 555 1696 130 1697 11 1698 1698 1699 1699 1700 97 1701 1701 1702 1702 1703 1703 1704 102 1705 102 1706 1706 1707 1707 1708 1708 1709 1709 1710 1710 1711 1711 1712 1712 1713 1713 1714 1714 1715 1715 1716 1716 1717 1717 1718 97 1719 2 1720 11 1721 1721 1722 2 1723 2 1724 2 1725 2 1726 2 1727 2 1728 2 1729 2 1730 288 1731 2 1732 2 1733 2 1734 2 1735 2 1736 294 1737 2 1738 2 1739 2 1740 2 1741 2 1742 2 1743 2 1744 2 1745 2 1746 2 1747 2 1748 2 1749 2 1750 2 1751 2 1752 1752 1753 1753 1754 1754 1755 1755 1756 1756 1757 1757 1758 1758 1759 1759 1760 1760 1761 1761 1762 1762 1763 1763 1764 1764 1765 1765 1766 1766 1767 4 1768 2 1769 1769 1770 1770 1771 1771 1772 1772 1773 1773 1774 1774 1775 1775 1776 1776 1777 1777 1778 1778 1779 1779 1780 338 1781 338 1782 40 1783 341 1784 2 1785 404 1786 1786 1787 555 1788 1788 1789 1789 1790 1790 1791 1791 1792 1792 1793 2 1794 404 1795 1795 1796 1796 1797 40 1798 1798 1799 264 1800 1800 1801 1801 1802 2 1803 404 1804 1804 1805 1805 1806 1806 1807 1807 1808 1808 1809 1809 1810 2 1811 713 1812 2 1813 1813 1814 1814 1815 1815 1816 1816 1817 1817 1818 1818 1819 1819 1820 1820 1821 1821 1822 1822 1823 1823 1824 2 1825 3 1826 4 1827 2 1828 2 1829 2 1830 2 1831 11 1832 2 1833 2 1834 1192 1835 2 1836 1194 1837 22 1838 1196 1839 1197 1840 2 1841 2 1842 11 1843 11 1844 1201 1845 1202 1846 1203 1847 22 1848 2 1849 2 1850 2 1851 11 1852 1209 1853 11 1854 1211 1855 11 1856 1213 1857 1214 1858 1215 1859 1216 1860 60 1861 1218 1862 2 1863 2 1864 2 1865 2 1866 67 1867 127 1868 1225 1869 2 1870 2 1871 1228 1872 2 1873 1230 1874 1231 1875 2 1876 83 1877 84 1878 22 1879 87 1880 1880 1881 146 1882 130 1883 11 1884 1238 1885 1239 1886 97 1887 1241 1888 1242 1889 1889 1890 102 1891 102 1892 1246 1893 1247 1894 1248 1895 1249 1896 1250 1897 1251 1898 1898 1899 1899 1900 1252 1901 1253 1902 1254 1903 1255 1904 97 1905 2 1906 11 1907 1258 1908 2 1909 2 1910 2 1911 2 1912 2 1913 2 1914 2 1915 2 1916 288 1917 2 1918 2 1919 2 1920 2 1921 2 1922 294 1923 2 1924 2 1925 2 1926 2 1927 2 1928 2 1929 2 1930 2 1931 2 1932 2 1933 2 1934 2 1935 2 1936 2 1937 2 1938 1289 1939 1290 1940 1291 1941 1292 1942 1293 1943 1294 1944 1295 1945 1296 1946 1297 1947 1298 1948 1299 1949 1300 1950 1301 1951 1302 1952 1303 1953 4 1954 2 1955 1306 1956 1307 1957 1308 1958 1309 1959 1310 1960 1311 1961 1312 1962 1313 1963 1314 1964 1315 1965 1316 1966 338 1967 338 1968 40 1969 341 1970 2 1971 404 1972 1323 1973 555 1974 1325 1975 1326 1976 1327 1977 1328 1978 1329 1979 2 1980 404 1981 1332 1982 1333 1983 40 1984 1335 1985 264 1986 1337 1987 1338 1988 2 1989 404 1990 1341 1991 1342 1992 1343 1993 1344 1994 1345 1995 1346 1996 2 1997 713 1998 2 1999 1350 2000 1351 2001 1352 2002 1353 2003 1354 2004 1355 2005 1356 2006 1357 2007 1358 2008 1359 2009 1823 2010 146 2011 2011 2012 2 2013 2013 2014 830 2015 1361 2016 1362 2017 1363 2018 1364 2019 1365 2020 1366 2021 1367 2022 1368 2023 1369 2024 1370 2025 1371 2026 1372 2027 1373 2028 1374 2029 1375 2030 1376 2031 1377 2032 1378 2033 1379 2034 1380 2035 1381 2036 1382 2037 1383 2038 1384 2039 1385 2040 1386 2041 1387 2042 1388 2043 1389 2044 1390 2045 1391 2046 1392 2047 1393 2048 1394 2049 1395 2050 1396 2051 1397 2052 1398 2053 1399 2054 1400 2055 1401 2056 1402 2057 1403 2058 1404 2059 1405 2060 1406 2061 1407 2062 1408 2063 1409 2064 1410 2065 1411 2066 1412 2067 1413 2068 1414 2069 1415 2070 1416 2071 1417 2072 1418 2073 1419 2074 1420 2075 1421 2076 1422 2077 1423 2078 1424 2079 1425 2080 1426 2081 1427 2082 1428 2083 1429 2084 1430 2085 1431 2086 1432 2087 1433 2088 1434 2089 1435 2090 1436 2091 1437 2092 1438 2093 1439 2094 11 2095 1441 2096 1442 2097 1443 2098 917 2099 1445 2100 1446 2101 950 2102 143 2103 146 2104 1451 2105 1455 2106 1456 2107 2 2108 2 2109 2 2110 1460 2111 2 2112 1462 2113 2 2114 2 2115 2 2116 2 2117 2 2118 2 2119 2 2120 2 2121 2 2122 1472 2123 1473 2124 1474 2125 1475 2126 1476 2127 1477 2128 1478 2129 1479 2130 1480 2131 1481 2132 1482 2133 1483 2134 1484 2135 22 2136 1488 2137 4 2138 1507 2139 87 2140 11 2141 1511 2142 97 2143 1516 2144 1517 2145 1518 2146 1519 2147 1520 2148 97 2149 830 2150 11 2151 1605 2152 917 2153 1607 2154 1608 2155 950 2156 143 2157 146 2158 1612 2159 1613 2160 1614 2161 830 2162 11 2163 1618 2164 1619 2165 1620 2166 338 2167 338 2168 341 2169 2 2170 404 2171 1626 2172 555 2173 1628 2174 1629 2175 1630 2176 1631 2177 2 2178 404 2179 1634 2180 1635 2181 1636 2182 1637 2183 2183 2184 2 2185 404 2186 2186 2187 1823 2188 2188 2189 2189 2190 136 2191 11 2192 138 2193 146 2194 147 2195 149 2196 151 2197 75 2198 123 2199 154 2200 2200 2201 2201 2202 136 2203 11 2204 138 2205 146 2206 2206 2207 150 2208 2208 2209 2209 2210 2210 2211 2 2212 2212 2213 2213 2214 2214 2215 2215 2216 2216 2217 2217 2218 2218 2219 2219 2220 2220 2221 2221 2222 2222 2223 2223 2224 2224 2225 2225 2226 2226 2227 2227 2228 2214 2229 2229 2230 2230 2231 2231 2232 2232 2233 2233 2234 2234 2235 2235 2236 2236 2237 2237 2238 2238 2239 2214 2240 2214 2241 2241 2242 2242 2243 2243 2244 2244 2245 2245 2246 2246 2247 2247 2248 2248 2249 2249 2250 2250 2251 2251 2252 2214 2253 2253 2254 2254 2255 2255 2256 2256 2257 2257 2258 2258 2259 2259 2260 2260 2261 2261 2262 2262 2263 2263 2264 1456 2265 2 2266 3 2267 4 2268 2 2269 2 2270 2 2271 2 2272 11 2273 2 2274 2 2275 1055 2276 2 2277 940 2278 22 2279 1057 2280 946 2281 2 2282 11 2283 11 2284 1060 2285 1061 2286 1062 2287 22 2288 2 2289 2 2290 2 2291 11 2292 1066 2293 11 2294 986 2295 11 2296 1069 2297 993 2298 994 2299 995 2300 60 2301 1071 2302 2 2303 2 2304 2 2305 2 2306 67 2307 127 2308 1076 2309 2 2310 2 2311 1078 2312 2 2313 1020 2314 1079 2315 2 2316 83 2317 84 2318 22 2319 87 2320 130 2321 11 2322 1030 2323 1031 2324 97 2325 1033 2326 1034 2327 2327 2328 102 2329 102 2330 1038 2331 1040 2332 1041 2333 1042 2334 1043 2335 1044 2336 1047 2337 1049 2338 1050 2339 1051 2340 97 2341 2 2342 11 2343 1080 2344 2 2345 2 2346 2 2347 2 2348 2 2349 2 2350 2 2351 2 2352 288 2353 2 2354 2 2355 2 2356 2 2357 2 2358 294 2359 2 2360 2 2361 2 2362 2 2363 2 2364 2 2365 2 2366 2 2367 2 2368 2 2369 2 2370 2 2371 2 2372 2 2373 2 2374 1111 2375 1112 2376 1113 2377 1114 2378 1115 2379 1116 2380 1117 2381 1118 2382 1119 2383 1120 2384 1121 2385 1122 2386 1123 2387 1124 2388 1125 2389 4 2390 2 2391 1128 2392 1129 2393 1130 2394 1131 2395 1132 2396 1133 2397 1134 2398 1135 2399 1136 2400 1137 2401 1138 2402 338 2403 338 2404 40 2405 341 2406 2 2407 404 2408 1145 2409 555 2410 1147 2411 1148 2412 1149 2413 1150 2414 1151 2415 2 2416 404 2417 1154 2418 1155 2419 40 2420 1157 2421 264 2422 1159 2423 1160 2424 2 2425 404 2426 1163 2427 1164 2428 1165 2429 1166 2430 1167 2431 1168 2432 2 2433 713 2434 2 2435 1172 2436 1173 2437 1174 2438 1175 2439 1176 2440 1177 2441 1178 2442 1179 2443 1180 2444 1181 2445 1823 2446 146 2447 2011 2448 2 2449 2449 2450 830 2451 2451 2452 2452 2453 2453 2454 2454 2455 2455 2456 2456 2457 2457 2458 2458 2459 2459 2460 2460 2461 2461 2462 2462 2463 2463 2464 2464 2465 2465 2466 2466 2467 2467 2468 2468 2469 2469 2470 2470 2471 2471 2472 2472 2473 2473 2474 2474 2475 2475 2476 2476 2477 2477 2478 2478 2479 2479 2480 2480 2481 2481 2482 2482 2483 2483 2484 2484 2485 2485 2486 2486 2487 2487 2488 2488 2489 2489 2490 2490 2491 2491 2492 2492 2493 2493 2494 2494 2495 2495 2496 2496 2497 2497 2498 2498 2499 2499 2500 2500 2501 2501 2502 2502 2503 2503 2504 2504 2505 2505 2506 2506 2507 2507 2508 2508 2509 2509 2510 2510 2511 2511 2512 2512 2513 2513 2514 2514 2515 2515 2516 2516 2517 2517 2518 2518 2519 2519 2520 2520 2521 2521 2522 2522 2523 2523 2524 2524 2525 2525 2526 2526 2527 2527 2528 2528 2529 2529 2530 11 2531 2531 2532 2532 2533 2533 2534 917 2535 2535 2536 2536 2537 950 2538 143 2539 146 2540 2540 2541 2541 2542 1456 2543 2 2544 2 2545 2 2546 2546 2547 2 2548 2548 2549 2 2550 2 2551 2 2552 2 2553 2 2554 2 2555 2 2556 2 2557 2 2558 2558 2559 2559 2560 2560 2561 2561 2562 2562 2563 2563 2564 2564 2565 2565 2566 2566 2567 2567 2568 2568 2569 2569 2570 2570 2571 22 2572 2572 2573 4 2574 2574 2575 87 2576 11 2577 2577 2578 97 2579 2579 2580 1517 2581 2581 2582 2582 2583 2583 2584 97 2585 830 2586 11 2587 2587 2588 917 2589 2589 2590 2590 2591 950 2592 143 2593 146 2594 2594 2595 2595 2596 2596 2597 830 2598 11 2599 2599 2600 2600 2601 2601 2602 338 2603 338 2604 341 2605 2 2606 404 2607 2607 2608 555 2609 2609 2610 2610 2611 2611 2612 2612 2613 2 2614 404 2615 2615 2616 2616 2617 2617 2618 2618 2619 2619 2620 2 2621 404 2622 2622 2623 2623 2624 136 2625 11 2626 138 2627 146 2628 147 2629 149 2630 151 2631 75 2632 123 2633 154 2634 2634 2635 2635 2636 136 2637 11 2638 138 2639 146 2640 2206 2641 2208 2642 2209 2643 2210 2644 2 2645 2645 2646 2646 2647 830 2648 2648 2649 11 2650 2650 2651 2651 2652 2652 2653 2653 2654 2654 2655 2655 2656 2656 2657 338 2658 338 2659 341 2660 2 2661 404 2662 2662 2663 555 2664 2664 2665 2665 2666 2666 2667 2667 2668 2 2669 404 2670 2670 2671 2671 2672 2672 2673 2673 2674 4 2675 2675 2676 87 2677 11 2678 2678 2679 97 2680 2680 2681 1517 2682 2682 2683 2683 2684 2684 2685 97 2686 830 2687 11 2688 2688 2689 917 2690 2690 2691 2691 2692 950 2693 143 2694 146 2695 2695 2696 2696 2697 2697 2698 830 2699 11 2700 2700 2701 2701 2702 2702 2703 338 2704 338 2705 341 2706 2 2707 404 2708 2708 2709 555 2710 2710 2711 2711 2712 2712 2713 2713 2714 2 2715 404 2716 2716 2717 2717 2718 2718 2719 2719 2720 2720 2721 2 2722 404 2723 2723 2724 22 2725 2725 2726 2726 2727 2727 2728 2728 2729 2729 2730 2730 2731 2731 2732 2732 2733 2733 2734 16 2735 16 2736 2736 2737 2737 2738 2738 2739 2739 2740 2740 2741 555 2742 2742 2743 2743 2744 2744 2745 2745 2746 2746 2747 2747 2748 404 2749 2749 2750 2 2751 2751 2752 2752 2753 2753 2754 2754 2755 4 2756 2756 2757 87 2758 11 2759 2759 2760 97 2761 2761 2762 1517 2763 2763 2764 2764 2765 2765 2766 97 2767 830 2768 11 2769 2769 2770 917 2771 2771 2772 2772 2773 950 2774 143 2775 146 2776 2776 2777 2777 2778 2778 2779 2779 2780 830 2781 11 2782 2782 2783 2783 2784 2784 2785 338 2786 338 2787 341 2788 2 2789 404 2790 2790 2791 555 2792 2792 2793 2793 2794 2794 2795 2795 2796 2 2797 404 2798 2798 2799 2799 2800 2800 2801 2801 2802 2802 2803 2 2804 404 2805 2805 2806 555 2807 2807 2808 2 2809 2809 2810 2810 2811 1243 2812 2 2813 2813 2814 2814 2815 2815 2816 2816 2817 2817 2818 22 2819 2725 2820 2820 2821 2821 2822 2822 2823 67 2824 2824 2825 2825 2826 2826 2827 2827 2828 16 2829 16 2830 2736 2831 555 2832 2832 2833 2833 2834 2834 2835 2835 2836 2836 2837 2747 2838 404 2839 2839 2840 2 2841 2841 2842 2752 2843 2753 2844 2844 2845 2765 2846 2846 2847 2847 2848 2848 2849 2849 2850 2850 2851 4 2852 2 2853 13 2854 2854 2855 2855 2856 2856 2857 2857 2858 2858 2859 2859 2860 2574 2861 2861 2862 2862 2863 2863 2864 2864 2865 2865 2866 2866 2867 2867 2868 2868 2869 338 2870 338 2871 40 2872 341 2873 2 2874 404 2875 2875 2876 555 2877 2877 2878 2878 2879 2610 2880 2880 2881 2612 2882 2 2883 404 2884 2615 2885 2885 2886 917 2887 2887 2888 2888 2889 2889 2890 950 2891 143 2892 146 2893 2893 2894 2894 2895 40 2896 2896 2897 2897 2898 2898 2899 917 2900 2900 2901 2901 2902 13 2903 2903 2904 13 2905 2905 2906 75 2907 2907 2908 2908 2909 2909 2910 13 2911 2911 2912 2912 2913 2913 2914 2914 2915 2915 2916 2916 2917 2917 2918 13 2919 2919 2920 2920 2921 2921 2922 2922 2923 2923 2924 2924 2925 2925 2926 2926 2927 2927 2928 2928 2929 2929 2930 2930 2931 2931 2932 2932 2933 2933 2934 2934 2935 2935 2936 2936 2937 2937 2938 2938 2939 2939 2940 2940 2941 2941 2942 2942 2943 2930 2944 2944 2945 2945 2946 2946 2947 2947 2948 2948 2949 2949 2950 2950 2951 2951 2952 2930 2953 2930 2954 2954 2955 2955 2956 2956 2957 2957 2958 2958 2959 2959 2960 2960 2961 2961 2962 2962 2963 2963 2964 2964 2965 2965 2966 2939 2967 2967 2968 2968 2969 2 2970 2970 2971 2 2972 3 2973 4 2974 2 2975 2 2976 2 2977 2 2978 11 2979 2 2980 2 2981 2981 2982 2 2983 2983 2984 22 2985 2985 2986 2986 2987 2 2988 11 2989 11 2990 2990 2991 2991 2992 2992 2993 22 2994 2 2995 2 2996 2 2997 11 2998 2998 2999 11 3000 3000 3001 11 3002 3002 3003 3003 3004 3004 3005 3005 3006 60 3007 3007 3008 2 3009 2 3010 2 3011 2 3012 67 3013 127 3014 3014 3015 2 3016 2 3017 3017 3018 2 3019 3019 3020 3020 3021 2 3022 83 3023 84 3024 87 3025 130 3026 11 3027 3027 3028 3028 3029 97 3030 3030 3031 3031 3032 3032 3033 102 3034 102 3035 3035 3036 3036 3037 3037 3038 3038 3039 3039 3040 3040 3041 3041 3042 3042 3043 3043 3044 3044 3045 97 3046 2 3047 3047 3048 2 3049 2 3050 2 3051 2 3052 2 3053 2 3054 2 3055 2 3056 288 3057 2 3058 2 3059 2 3060 2 3061 2 3062 294 3063 2 3064 2 3065 2 3066 2 3067 2 3068 2 3069 2 3070 2 3071 2 3072 2 3073 2 3074 2 3075 2 3076 2 3077 2 3078 3078 3079 3079 3080 3080 3081 3081 3082 3082 3083 3083 3084 3084 3085 3085 3086 3086 3087 3087 3088 3088 3089 3089 3090 3090 3091 3091 3092 3092 3093 4 3094 2 3095 3095 3096 3096 3097 3097 3098 3098 3099 3099 3100 3100 3101 3101 3102 3102 3103 3103 3104 3104 3105 3105 3106 338 3107 338 3108 40 3109 341 3110 2 3111 404 3112 3112 3113 555 3114 3114 3115 3115 3116 3116 3117 3117 3118 3118 3119 2 3120 404 3121 3121 3122 3122 3123 40 3124 3124 3125 264 3126 3126 3127 3127 3128 2 3129 404 3130 3130 3131 3131 3132 3132 3133 3133 3134 3134 3135 3135 3136 2 3137 713 3138 2 3139 3139 3140 3140 3141 3141 3142 3142 3143 3143 3144 3144 3145 3145 3146 3146 3147 3147 3148 3148 3149 830 3150 3150 3151 3151 3152 3152 3153 3153 3154 3154 3155 3155 3156 3156 3157 3157 3158 3158 3159 3159 3160 3160 3161 3161 3162 3162 3163 3163 3164 3164 3165 3165 3166 3166 3167 3167 3168 3168 3169 3169 3170 3170 3171 3171 3172 3172 3173 3173 3174 3174 3175 3175 3176 3176 3177 3177 3178 3178 3179 3179 3180 3180 3181 3181 3182 3182 3183 3183 3184 3184 3185 3185 3186 3186 3187 3187 3188 3188 3189 3189 3190 3190 3191 3191 3192 3192 3193 3193 3194 3194 3195 3195 3196 3196 3197 3197 3198 3198 3199 3199 3200 3200 3201 3201 3202 3202 3203 3203 3204 3204 3205 3205 3206 3206 3207 3207 3208 3208 3209 3209 3210 3210 3211 3211 3212 3212 3213 3213 3214 3214 3215 3215 3216 3216 3217 3217 3218 3218 3219 3219 3220 3220 3221 3221 3222 3222 3223 3223 3224 3224 3225 3225 3226 3226 3227 3227 3228 3228 3229 11 3230 3230 3231 3231 3232 3232 3233 917 3234 3234 3235 3235 3236 950 3237 143 3238 146 3239 3239 3240 3240 3241 1456 3242 2 3243 2 3244 2 3245 3245 3246 2 3247 3247 3248 2 3249 2 3250 2 3251 2 3252 2 3253 2 3254 2 3255 2 3256 2 3257 3257 3258 3258 3259 3259 3260 3260 3261 3261 3262 3262 3263 3263 3264 3264 3265 3265 3266 3266 3267 3267 3268 3268 3269 3269 3270 22 3271 3271 3272 4 3273 3273 3274 87 3275 11 3276 3276 3277 97 3278 3278 3279 1517 3280 3280 3281 3281 3282 3282 3283 97 3284 830 3285 11 3286 3286 3287 917 3288 3288 3289 3289 3290 950 3291 143 3292 146 3293 3293 3294 3294 3295 3295 3296 830 3297 11 3298 3298 3299 3299 3300 3300 3301 338 3302 338 3303 341 3304 2 3305 404 3306 3306 3307 555 3308 3308 3309 3309 3310 3310 3311 3311 3312 2 3313 404 3314 3314 3315 3315 3316 3316 3317 3317 3318 3318 3319 2 3320 404 3321 3321 3322 3322 3323 136 3324 11 3325 138 3326 146 3327 147 3328 149 3329 151 3330 75 3331 123 3332 154 3333 3333 3334 3334 3335 136 3336 11 3337 138 3338 146 3339 2206 3340 2208 3341 2209 3342 2210 3343 2 3344 3344 3345 3345 3346 3346 3347 830 3348 3348 3349 11 3350 3350 3351 3351 3352 3352 3353 2653 3354 3354 3355 3355 3356 338 3357 338 3358 341 3359 2 3360 404 3361 3361 3362 555 3363 3363 3364 3364 3365 3365 3366 3366 3367 2 3368 404 3369 3369 3370 3370 3371 3371 3372 3372 3373 2728 3374 3374 3375 555 3376 3376 3377 2 3378 3378 3379 3379 3380 2 3381 3 3382 4 3383 2 3384 2 3385 2 3386 2 3387 11 3388 2 3389 2 3390 3390 3391 2 3392 3392 3393 22 3394 3394 3395 3395 3396 2 3397 11 3398 11 3399 3399 3400 3400 3401 3401 3402 22 3403 2 3404 2 3405 2 3406 11 3407 3407 3408 11 3409 3409 3410 11 3411 3411 3412 3412 3413 3413 3414 3414 3415 60 3416 3416 3417 2 3418 2 3419 2 3420 2 3421 67 3422 127 3423 3423 3424 2 3425 2 3426 3426 3427 2 3428 3428 3429 3429 3430 2 3431 83 3432 84 3433 87 3434 130 3435 11 3436 3436 3437 3437 3438 97 3439 3439 3440 3440 3441 3441 3442 102 3443 102 3444 3444 3445 3445 3446 3446 3447 3447 3448 3448 3449 3449 3450 3450 3451 3451 3452 3452 3453 3453 3454 97 3455 2 3456 3456 3457 2 3458 2 3459 2 3460 2 3461 2 3462 2 3463 2 3464 2 3465 288 3466 2 3467 2 3468 2 3469 2 3470 2 3471 294 3472 2 3473 2 3474 2 3475 2 3476 2 3477 2 3478 2 3479 2 3480 2 3481 2 3482 2 3483 2 3484 2 3485 2 3486 2 3487 3487 3488 3488 3489 3489 3490 3490 3491 3491 3492 3492 3493 3493 3494 3494 3495 3495 3496 3496 3497 3497 3498 3498 3499 3499 3500 3500 3501 3501 3502 4 3503 2 3504 3504 3505 3505 3506 3506 3507 3507 3508 3508 3509 3509 3510 3510 3511 3511 3512 3512 3513 3513 3514 3514 3515 338 3516 338 3517 40 3518 341 3519 2 3520 404 3521 3521 3522 555 3523 3523 3524 3524 3525 3525 3526 3526 3527 3527 3528 2 3529 404 3530 3530 3531 3531 3532 40 3533 3533 3534 264 3535 3535 3536 3536 3537 2 3538 404 3539 3539 3540 3540 3541 3541 3542 3542 3543 3543 3544 3544 3545 2 3546 713 3547 2 3548 3548 3549 3549 3550 3550 3551 3551 3552 3552 3553 3553 3554 3554 3555 3555 3556 3556 3557 3557 3558 830 3559 3559 3560 3560 3561 3561 3562 3562 3563 3563 3564 3564 3565 3565 3566 3566 3567 3567 3568 3568 3569 3569 3570 3570 3571 3571 3572 3572 3573 3573 3574 3574 3575 3575 3576 3576 3577 3577 3578 3578 3579 3579 3580 3580 3581 3581 3582 3582 3583 3583 3584 3584 3585 3585 3586 3586 3587 3587 3588 3588 3589 3589 3590 3590 3591 3591 3592 3592 3593 3593 3594 3594 3595 3595 3596 3596 3597 3597 3598 3598 3599 3599 3600 3600 3601 3601 3602 3602 3603 3603 3604 3604 3605 3605 3606 3606 3607 3607 3608 3608 3609 3609 3610 3610 3611 3611 3612 3612 3613 3613 3614 3614 3615 3615 3616 3616 3617 3617 3618 3618 3619 3619 3620 3620 3621 3621 3622 3622 3623 3623 3624 3624 3625 3625 3626 3626 3627 3627 3628 3628 3629 3629 3630 3630 3631 3631 3632 3632 3633 3633 3634 3634 3635 3635 3636 3636 3637 3637 3638 11 3639 3639 3640 3640 3641 3641 3642 917 3643 3643 3644 3644 3645 950 3646 143 3647 146 3648 3648 3649 3649 3650 1456 3651 2 3652 2 3653 2 3654 3654 3655 2 3656 3656 3657 2 3658 2 3659 2 3660 2 3661 2 3662 2 3663 2 3664 2 3665 2 3666 3666 3667 3667 3668 3668 3669 3669 3670 3670 3671 3671 3672 3672 3673 3673 3674 3674 3675 3675 3676 3676 3677 3677 3678 3678 3679 22 3680 3680 3681 4 3682 3682 3683 87 3684 11 3685 3685 3686 97 3687 3687 3688 1517 3689 3689 3690 3690 3691 3691 3692 97 3693 830 3694 11 3695 3695 3696 917 3697 3697 3698 3698 3699 950 3700 143 3701 146 3702 3702 3703 3703 3704 3704 3705 830 3706 11 3707 3707 3708 3708 3709 3709 3710 338 3711 338 3712 341 3713 2 3714 404 3715 3715 3716 555 3717 3717 3718 3718 3719 3719 3720 3720 3721 2 3722 404 3723 3723 3724 3724 3725 3725 3726 3726 3727 3727 3728 2 3729 404 3730 3730 3731 3731 3732 136 3733 11 3734 138 3735 146 3736 147 3737 149 3738 151 3739 75 3740 123 3741 154 3742 3742 3743 3743 3744 136 3745 11 3746 138 3747 146 3748 2206 3749 2208 3750 2209 3751 2210 3752 2 3753 3753 3754 3754 3755 2 3756 3756 3757 830 3758 3758 3759 11 3760 3760 3761 3761 3762 3762 3763 2653 3764 3764 3765 3765 3766 338 3767 338 3768 341 3769 2 3770 404 3771 3771 3772 555 3773 3773 3774 3774 3775 3775 3776 3776 3777 2 3778 404 3779 3779 3780 3780 3781 3781 3782 3782 3783 2728 3784 3784 3785 555 3786 3786 3787 2 3788 3788 3789 3789 3790 3441 3791 2 3792 3792 3793 3793 3794 3794 3795 3795 3796 3796 3797 2820 3798 3798 3799 67 3800 3800 3801 3801 3802 3802 3803 3803 3804 3804 3805 3805 3806 2 3807 2854 3808 3808 3809 3809 3810 3810 3811 3811 3812 3812 3813 3813 3814 2863 3815 3815 3816 3816 3817 3817 3818 3818 3819 3819 3820 338 3821 338 3822 40 3823 341 3824 2 3825 404 3826 3826 3827 555 3828 3828 3829 3829 3830 3830 3831 3831 3832 917 3833 3833 3834 3834 3835 3835 3836 950 3837 143 3838 146 3839 3839 3840 3840 3841 40 3842 3842 3843 3843 3844 2901 3845 2903 3846 2905 3847 75 3848 2907 3849 2908 3850 3850 3851 264 3852 3852 3853 2 3854 264 3855 75 3856 3856 3857 2 3858 3 3859 4 3860 2 3861 2 3862 2 3863 2 3864 11 3865 2 3866 2 3867 3867 3868 2 3869 3869 3870 22 3871 3871 3872 3872 3873 3873 3874 2 3875 11 3876 11 3877 3877 3878 3878 3879 3879 3880 22 3881 2 3882 2 3883 2 3884 11 3885 3885 3886 11 3887 3887 3888 11 3889 3889 3890 3890 3891 3891 3892 3892 3893 60 3894 3894 3895 2 3896 2 3897 2 3898 2 3899 67 3900 127 3901 3901 3902 2 3903 2 3904 3904 3905 2 3906 3906 3907 3907 3908 2 3909 83 3910 84 3911 87 3912 130 3913 11 3914 3914 3915 3915 3916 97 3917 3917 3918 3918 3919 3919 3920 102 3921 102 3922 3922 3923 3923 3924 3924 3925 3925 3926 3926 3927 3871 3928 2753 3929 3929 3930 3930 3931 3931 3932 3932 3933 3933 3934 3934 3935 3935 3936 97 3937 2 3938 3938 3939 2 3940 2 3941 2 3942 2 3943 2 3944 2 3945 2 3946 2 3947 288 3948 2 3949 2 3950 2 3951 2 3952 2 3953 294 3954 2 3955 2 3956 2 3957 2 3958 2 3959 2 3960 2 3961 2 3962 2 3963 2 3964 2 3965 2 3966 2 3967 2 3968 2 3969 3969 3970 3970 3971 3971 3972 3972 3973 3973 3974 3974 3975 3975 3976 3976 3977 3977 3978 3978 3979 3979 3980 3980 3981 3981 3982 3982 3983 3983 3984 4 3985 2 3986 3986 3987 3987 3988 3988 3989 3989 3990 3990 3991 3991 3992 3992 3993 3993 3994 3994 3995 3995 3996 3996 3997 338 3998 338 3999 40 4000 341 4001 2 4002 404 4003 4003 4004 555 4005 4005 4006 4006 4007 4007 4008 4008 4009 4009 4010 2 4011 404 4012 4012 4013 4013 4014 40 4015 4015 4016 264 4017 4017 4018 4018 4019 2 4020 404 4021 4021 4022 4022 4023 4023 4024 4024 4025 4025 4026 4026 4027 2 4028 713 4029 2 4030 4030 4031 4031 4032 4032 4033 4033 4034 4034 4035 4035 4036 4036 4037 4037 4038 4038 4039 4039 4040 830 4041 4041 4042 4042 4043 4043 4044 4044 4045 4045 4046 4046 4047 4047 4048 4048 4049 4049 4050 4050 4051 4051 4052 4052 4053 4053 4054 4054 4055 4055 4056 4056 4057 4057 4058 4058 4059 4059 4060 4060 4061 4061 4062 4062 4063 4063 4064 4064 4065 4065 4066 4066 4067 4067 4068 4068 4069 4069 4070 4070 4071 4071 4072 4072 4073 4073 4074 4074 4075 4075 4076 4076 4077 4077 4078 4078 4079 4079 4080 4080 4081 4081 4082 4082 4083 4083 4084 4084 4085 4085 4086 4086 4087 4087 4088 4088 4089 4089 4090 4090 4091 4091 4092 4092 4093 4093 4094 4094 4095 4095 4096 4096 4097 4097 4098 4098 4099 4099 4100 4100 4101 4101 4102 4102 4103 4103 4104 4104 4105 4105 4106 4106 4107 4107 4108 4108 4109 4109 4110 4110 4111 4111 4112 4112 4113 4113 4114 4114 4115 4115 4116 4116 4117 4117 4118 4118 4119 4119 4120 11 4121 4121 4122 4122 4123 4123 4124 917 4125 4125 4126 4126 4127 950 4128 143 4129 146 4130 4130 4131 4131 4132 1456 4133 2 4134 2 4135 2 4136 4136 4137 2 4138 4138 4139 2 4140 2 4141 2 4142 2 4143 2 4144 2 4145 2 4146 2 4147 2 4148 4148 4149 4149 4150 4150 4151 4151 4152 4152 4153 4153 4154 4154 4155 4155 4156 4156 4157 4157 4158 4158 4159 4159 4160 4160 4161 22 4162 4162 4163 4 4164 4164 4165 87 4166 11 4167 4167 4168 97 4169 4169 4170 1517 4171 4171 4172 4172 4173 4173 4174 97 4175 830 4176 11 4177 4177 4178 917 4179 4179 4180 4180 4181 950 4182 143 4183 146 4184 4184 4185 4185 4186 4186 4187 830 4188 11 4189 4189 4190 4190 4191 4191 4192 338 4193 338 4194 341 4195 2 4196 404 4197 4197 4198 555 4199 4199 4200 4200 4201 4201 4202 4202 4203 2 4204 404 4205 4205 4206 4206 4207 4207 4208 4208 4209 4209 4210 2 4211 404 4212 4212 4213 4213 4214 4214 4215 4215 4216 4216 4217 4217 4218 4218 4219 2208 4220 2 4221 2 4222 3 4223 4 4224 2 4225 2 4226 2 4227 2 4228 11 4229 2 4230 2 4231 4231 4232 2 4233 4233 4234 22 4235 4235 4236 4236 4237 2 4238 11 4239 11 4240 4240 4241 4241 4242 4242 4243 22 4244 2 4245 2 4246 2 4247 11 4248 4248 4249 11 4250 4250 4251 11 4252 4252 4253 4253 4254 4254 4255 4255 4256 60 4257 4257 4258 2 4259 2 4260 2 4261 2 4262 67 4263 127 4264 4264 4265 2 4266 2 4267 4267 4268 2 4269 4269 4270 4270 4271 2 4272 83 4273 84 4274 87 4275 130 4276 11 4277 4277 4278 4278 4279 97 4280 4280 4281 4281 4282 4282 4283 102 4284 102 4285 4285 4286 4286 4287 4287 4288 4288 4289 4289 4290 4290 4291 4291 4292 4292 4293 4293 4294 4294 4295 97 4296 2 4297 4297 4298 2 4299 2 4300 2 4301 2 4302 2 4303 2 4304 2 4305 2 4306 288 4307 2 4308 2 4309 2 4310 2 4311 2 4312 294 4313 2 4314 2 4315 2 4316 2 4317 2 4318 2 4319 2 4320 2 4321 2 4322 2 4323 2 4324 2 4325 2 4326 2 4327 2 4328 4328 4329 4329 4330 4330 4331 4331 4332 4332 4333 4333 4334 4334 4335 4335 4336 4336 4337 4337 4338 4338 4339 4339 4340 4340 4341 4341 4342 4342 4343 4 4344 2 4345 4345 4346 4346 4347 4347 4348 4348 4349 4349 4350 4350 4351 4351 4352 4352 4353 4353 4354 4354 4355 4355 4356 338 4357 338 4358 40 4359 341 4360 2 4361 404 4362 4362 4363 555 4364 4364 4365 4365 4366 4366 4367 4367 4368 4368 4369 2 4370 404 4371 4371 4372 4372 4373 40 4374 4374 4375 264 4376 4376 4377 4377 4378 2 4379 404 4380 4380 4381 4381 4382 4382 4383 4383 4384 4384 4385 4385 4386 2 4387 713 4388 2 4389 4389 4390 4390 4391 4391 4392 4392 4393 4393 4394 4394 4395 4395 4396 4396 4397 4397 4398 4398 4399 830 4400 4400 4401 4401 4402 4402 4403 4403 4404 4404 4405 4405 4406 4406 4407 4407 4408 4408 4409 4409 4410 4410 4411 4411 4412 4412 4413 4413 4414 4414 4415 4415 4416 4416 4417 4417 4418 4418 4419 4419 4420 4420 4421 4421 4422 4422 4423 4423 4424 4424 4425 4425 4426 4426 4427 4427 4428 4428 4429 4429 4430 4430 4431 4431 4432 4432 4433 4433 4434 4434 4435 4435 4436 4436 4437 4437 4438 4438 4439 4439 4440 4440 4441 4441 4442 4442 4443 4443 4444 4444 4445 4445 4446 4446 4447 4447 4448 4448 4449 4449 4450 4450 4451 4451 4452 4452 4453 4453 4454 4454 4455 4455 4456 4456 4457 4457 4458 4458 4459 4459 4460 4460 4461 4461 4462 4462 4463 4463 4464 4464 4465 4465 4466 4466 4467 4467 4468 4468 4469 4469 4470 4470 4471 4471 4472 4472 4473 4473 4474 4474 4475 4475 4476 4476 4477 4477 4478 4478 4479 11 4480 4480 4481 4481 4482 4482 4483 917 4484 4484 4485 4485 4486 950 4487 143 4488 146 4489 4489 4490 4490 4491 1456 4492 2 4493 2 4494 2 4495 4495 4496 2 4497 4497 4498 2 4499 2 4500 2 4501 2 4502 2 4503 2 4504 2 4505 2 4506 2 4507 4507 4508 4508 4509 4509 4510 4510 4511 4511 4512 4512 4513 4513 4514 4514 4515 4515 4516 4516 4517 4517 4518 4518 4519 4519 4520 22 4521 4521 4522 4 4523 4523 4524 87 4525 11 4526 4526 4527 97 4528 4528 4529 1517 4530 4530 4531 4531 4532 4532 4533 97 4534 830 4535 11 4536 4536 4537 917 4538 4538 4539 4539 4540 950 4541 143 4542 146 4543 4543 4544 4544 4545 4545 4546 830 4547 11 4548 4548 4549 4549 4550 4550 4551 338 4552 338 4553 341 4554 2 4555 404 4556 4556 4557 555 4558 4558 4559 4559 4560 4560 4561 4561 4562 2 4563 404 4564 4564 4565 4565 4566 4566 4567 4567 4568 4568 4569 2 4570 404 4571 4571 4572 4572 4573 136 4574 11 4575 138 4576 146 4577 147 4578 149 4579 151 4580 75 4581 123 4582 154 4583 4583 4584 4584 4585 136 4586 11 4587 138 4588 146 4589 2206 4590 2208 4591 2209 4592 2210 4593 2 4594 4594 4595 4595 4596 830 4597 4597 4598 11 4599 4599 4600 4600 4601 4601 4602 2653 4603 4603 4604 4604 4605 338 4606 338 4607 341 4608 2 4609 404 4610 4610 4611 555 4612 4612 4613 4613 4614 4614 4615 4615 4616 2 4617 404 4618 4618 4619 4619 4620 4620 4621 4621 4622 2728 4623 4623 4624 555 4625 4625 4626 2 4627 4627 4628 4628 4629 2 4630 3 4631 4 4632 2 4633 2 4634 2 4635 2 4636 11 4637 2 4638 2 4639 4639 4640 2 4641 4641 4642 22 4643 4643 4644 4644 4645 2 4646 11 4647 11 4648 4648 4649 4649 4650 4650 4651 22 4652 2 4653 2 4654 2 4655 11 4656 4656 4657 11 4658 4658 4659 11 4660 4660 4661 4661 4662 4662 4663 4663 4664 60 4665 4665 4666 2 4667 2 4668 2 4669 2 4670 67 4671 127 4672 4672 4673 2 4674 2 4675 4675 4676 2 4677 4677 4678 4678 4679 2 4680 83 4681 84 4682 87 4683 130 4684 11 4685 4685 4686 4686 4687 97 4688 4688 4689 4689 4690 4690 4691 102 4692 102 4693 4693 4694 4694 4695 4695 4696 4696 4697 4697 4698 4698 4699 4699 4700 4700 4701 4701 4702 4702 4703 97 4704 2 4705 4705 4706 2 4707 2 4708 2 4709 2 4710 2 4711 2 4712 2 4713 2 4714 288 4715 2 4716 2 4717 2 4718 2 4719 2 4720 294 4721 2 4722 2 4723 2 4724 2 4725 2 4726 2 4727 2 4728 2 4729 2 4730 2 4731 2 4732 2 4733 2 4734 2 4735 2 4736 4736 4737 4737 4738 4738 4739 4739 4740 4740 4741 4741 4742 4742 4743 4743 4744 4744 4745 4745 4746 4746 4747 4747 4748 4748 4749 4749 4750 4750 4751 4 4752 2 4753 4753 4754 4754 4755 4755 4756 4756 4757 4757 4758 4758 4759 4759 4760 4760 4761 4761 4762 4762 4763 4763 4764 338 4765 338 4766 40 4767 341 4768 2 4769 404 4770 4770 4771 555 4772 4772 4773 4773 4774 4774 4775 4775 4776 4776 4777 2 4778 404 4779 4779 4780 4780 4781 40 4782 4782 4783 264 4784 4784 4785 4785 4786 2 4787 404 4788 4788 4789 4789 4790 4790 4791 4791 4792 4792 4793 4793 4794 2 4795 713 4796 2 4797 4797 4798 4798 4799 4799 4800 4800 4801 4801 4802 4802 4803 4803 4804 4804 4805 4805 4806 4806 4807 830 4808 4808 4809 4809 4810 4810 4811 4811 4812 4812 4813 4813 4814 4814 4815 4815 4816 4816 4817 4817 4818 4818 4819 4819 4820 4820 4821 4821 4822 4822 4823 4823 4824 4824 4825 4825 4826 4826 4827 4827 4828 4828 4829 4829 4830 4830 4831 4831 4832 4832 4833 4833 4834 4834 4835 4835 4836 4836 4837 4837 4838 4838 4839 4839 4840 4840 4841 4841 4842 4842 4843 4843 4844 4844 4845 4845 4846 4846 4847 4847 4848 4848 4849 4849 4850 4850 4851 4851 4852 4852 4853 4853 4854 4854 4855 4855 4856 4856 4857 4857 4858 4858 4859 4859 4860 4860 4861 4861 4862 4862 4863 4863 4864 4864 4865 4865 4866 4866 4867 4867 4868 4868 4869 4869 4870 4870 4871 4871 4872 4872 4873 4873 4874 4874 4875 4875 4876 4876 4877 4877 4878 4878 4879 4879 4880 4880 4881 4881 4882 4882 4883 4883 4884 4884 4885 4885 4886 4886 4887 11 4888 4888 4889 4889 4890 4890 4891 917 4892 4892 4893 4893 4894 950 4895 143 4896 146 4897 4897 4898 4898 4899 1456 4900 2 4901 2 4902 2 4903 4903 4904 2 4905 4905 4906 2 4907 2 4908 2 4909 2 4910 2 4911 2 4912 2 4913 2 4914 2 4915 4915 4916 4916 4917 4917 4918 4918 4919 4919 4920 4920 4921 4921 4922 4922 4923 4923 4924 4924 4925 4925 4926 4926 4927 4927 4928 22 4929 4929 4930 4 4931 4931 4932 87 4933 11 4934 4934 4935 97 4936 4936 4937 1517 4938 4938 4939 4939 4940 4940 4941 97 4942 830 4943 11 4944 4944 4945 917 4946 4946 4947 4947 4948 950 4949 143 4950 146 4951 4951 4952 4952 4953 4953 4954 830 4955 11 4956 4956 4957 4957 4958 4958 4959 338 4960 338 4961 341 4962 2 4963 404 4964 4964 4965 555 4966 4966 4967 4967 4968 4968 4969 4969 4970 2 4971 404 4972 4972 4973 4973 4974 4974 4975 4975 4976 4976 4977 2 4978 404 4979 4979 4980 4980 4981 136 4982 11 4983 138 4984 146 4985 147 4986 149 4987 151 4988 75 4989 123 4990 154 4991 4991 4992 4992 4993 136 4994 11 4995 138 4996 146 4997 2206 4998 2208 4999 2209 5000 2210 5001 2 5002 5002 5003 5003 5004 2 5005 5005 5006 830 5007 5007 5008 11 5009 5009 5010 5010 5011 5011 5012 2653 5013 5013 5014 5014 5015 338 5016 338 5017 341 5018 2 5019 404 5020 5020 5021 555 5022 5022 5023 5023 5024 5024 5025 5025 5026 2 5027 404 5028 5028 5029 5029 5030 5030 5031 5031 5032 2728 5033 5033 5034 555 5035 5035 5036 2 5037 5037 5038 5038 5039 4690 5040 2 5041 5041 5042 5042 5043 5043 5044 5044 5045 5045 5046 2820 5047 5047 5048 67 5049 5049 5050 5050 5051 5051 5052 5052 5053 5053 5054 5054 5055 2 5056 2854 5057 5057 5058 5058 5059 5059 5060 5060 5061 5061 5062 5062 5063 2863 5064 5064 5065 5065 5066 5066 5067 5067 5068 5068 5069 338 5070 338 5071 40 5072 341 5073 2 5074 404 5075 5075 5076 555 5077 5077 5078 5078 5079 5079 5080 5080 5081 917 5082 5082 5083 5083 5084 5084 5085 950 5086 143 5087 146 5088 5088 5089 5089 5090 40 5091 5091 5092 5092 5093 2901 5094 2903 5095 2905 5096 75 5097 2907 5098 2908 5099 5099 5100 264 5101 5101 5102 2 5103 264 5104 75 5105 3856 5106 2753 5107 5107 5108 1823 5109 146 5110 2011 5111 2 5112 5112 5113 2 5114 5114 5115 5115 5116 143 5117 5117 5118 146 5119 5119 5120 5120 5121 5121 5122 5122 5123 5123 5124 5124 5125 5125 5126 5126 5127 5127 5128 146 5129 5129 5130 2 5131 264 5132 5132 5133 2 5134 264 5135 5135 5136 5136 5137 5137 5138 5138 5139 5139 5140 5140 5141 1823 5142 146 5143 5143 5144 5144 5145 5145 5146 5146 5147 264 5148 5148 5149 5149 5150 2 5151 404 5152 5152 5153 22 5154 2725 5155 5155 5156 5156 5157 5157 5158 5158 5159 5159 5160 5160 5161 5161 5162 16 5163 16 5164 2736 5165 555 5166 5166 5167 5167 5168 5168 5169 5169 5170 2747 5171 404 5172 5172 5173 2 5174 5174 5175 2752 5176 2753 5177 5177 5178 2765 5179 5179 5180 5180 5181 5181 5182 5182 5183 5183 5184 5184 5185 2820 5186 5186 5187 67 5188 5188 5189 5189 5190 5190 5191 5191 5192 5192 5193 5193 5194 2 5195 2854 5196 5196 5197 5197 5198 5198 5199 5199 5200 5200 5201 5201 5202 2863 5203 5203 5204 5204 5205 5205 5206 5206 5207 5207 5208 338 5209 338 5210 40 5211 341 5212 2 5213 404 5214 5214 5215 555 5216 5216 5217 5217 5218 5218 5219 5219 5220 917 5221 5221 5222 5222 5223 5223 5224 950 5225 143 5226 146 5227 5227 5228 5228 5229 40 5230 5230 5231 5231 5232 2901 5233 2903 5234 2905 5235 75 5236 2907 5237 2908 5238 5238 5239 264 5240 5240 5241 2 5242 264 5243 75 5244 3856 5245 2753 5246 5246 5247 1823 5248 146 5249 2011 5250 2 5251 5251 5252 2 5253 143 5254 146 5255 5255 5256 5256 5257 5257 5258 5258 5259 5259 5260 5125 5261 5261 5262 146 5263 5263 5264 2 5265 264 5266 5266 5267 2 5268 264 5269 5269 5270 5270 5271 5271 5272 5272 5273 5273 5274 5274 5275 1823 5276 146 5277 5277 5278 5278 5279 5279 5280 5280 5281 264 5282 5282 5283 5283 5284 2 5285 404 5286 5286 5287 5157 5288 5288 5289 5289 5290 5290 5291 136 5292 11 5293 138 5294 146 5295 147 5296 149 5297 151 5298 75 5299 123 5300 154 5301 5301 5302 5302 5303 136 5304 11 5305 138 5306 146 5307 2206 5308 2208 5309 2209 5310 2210 5311 2 5312 5312 5313 5313 5314 830 5315 5315 5316 11 5317 5317 5318 5318 5319 5319 5320 2653 5321 5321 5322 5322 5323 338 5324 338 5325 341 5326 2 5327 404 5328 5328 5329 555 5330 5330 5331 5331 5332 5332 5333 5333 5334 2 5335 404 5336 5336 5337 5337 5338 5338 5339 5339 5340 2728 5341 5341 5342 555 5343 5343 5344 2 5345 5345 5346 5346 5347 5347 5348 2 5349 5349 5350 5350 5351 5351 5352 5352 5353 5353 5354 2820 5355 5355 5356 67 5357 5357 5358 5358 5359 5359 5360 5360 5361 5361 5362 5362 5363 2 5364 2854 5365 5365 5366 5366 5367 5367 5368 5368 5369 5369 5370 5370 5371 2863 5372 5372 5373 5373 5374 5374 5375 5375 5376 5376 5377 338 5378 338 5379 40 5380 341 5381 2 5382 404 5383 5383 5384 555 5385 5385 5386 5386 5387 5387 5388 5388 5389 917 5390 5390 5391 5391 5392 5392 5393 950 5394 143 5395 146 5396 5396 5397 5397 5398 40 5399 5399 5400 5400 5401 2901 5402 2903 5403 2905 5404 75 5405 2907 5406 2908 5407 5407 5408 264 5409 5409 5410 2 5411 264 5412 75 5413 3856 5414 2753 5415 5415 5416 1823 5417 146 5418 2011 5419 2 5420 5420 5421 2 5422 143 5423 146 5424 5424 5425 5425 5426 5426 5427 5427 5428 5428 5429 5125 5430 5430 5431 146 5432 5432 5433 2 5434 264 5435 5435 5436 2 5437 264 5438 5438 5439 5439 5440 5440 5441 2208 5442 2 5443 5443 5444 5444 5445 5445 5446 5446 5447 5447 5448 1823 5449 146 5450 5450 5451 5451 5452 5452 5453 5453 5454 264 5455 5455 5456 5456 5457 2 5458 404 5459 5459 5460 5157 5461 5461 5462 5462 5463 1823 5464 146 5465 2011 5466 2 5467 5467 5468 2 5469 143 5470 146 5471 5471 5472 5472 5473 5473 5474 5474 5475 5475 5476 5125 5477 5477 5478 146 5479 5479 5480 2 5481 264 5482 5482 5483 2 5484 264 5485 5485 5486 5486 5487 5487 5488 5488 5489 5489 5490 5490 5491 1823 5492 146 5493 5493 5494 5494 5495 5495 5496 5496 5497 264 5498 5498 5499 5499 5500 2 5501 404 5502 5502 5503 5157 5504 5504 5505 5505 5506 5506 5507 5507 5508 5508 5509 5509 5510 2820 5511 5511 5512 67 5513 5513 5514 5514 5515 5515 5516 5516 5517 5517 5518 5518 5519 2 5520 2854 5521 5521 5522 5522 5523 5523 5524 5524 5525 5525 5526 5526 5527 2863 5528 5528 5529 5529 5530 5530 5531 5531 5532 5532 5533 338 5534 338 5535 40 5536 341 5537 2 5538 404 5539 5539 5540 555 5541 5541 5542 5542 5543 5543 5544 5544 5545 917 5546 5546 5547 5547 5548 5548 5549 950 5550 143 5551 146 5552 5552 5553 5553 5554 40 5555 5555 5556 5556 5557 2901 5558 2903 5559 2905 5560 75 5561 2907 5562 2908 5563 5563 5564 264 5565 5565 5566 2 5567 264 5568 75 5569 3856 5570 2753 5571 5571 5572 1823 5573 146 5574 2011 5575 2 5576 5576 5577 2 5578 143 5579 146 5580 5580 5581 5581 5582 5582 5583 5583 5584 5584 5585 5125 5586 5586 5587 146 5588 5588 5589 2 5590 264 5591 5591 5592 2 5593 264 5594 5594 5595 5595 5596 5596 5597 5597 5598 5598 5599 5599 5600 1823 5601 146 5602 5602 5603 5603 5604 5604 5605 5605 5606 264 5607 5607 5608 5608 5609 2 5610 404 5611 5611 5612 5157 5613 5613 5614 5614 5615 2930 5616 5616 5617 5617 5618 5618 5619 5619 5620 5620 5621 5621 5622 5622 5623 5623 5624 5624 5625 5625 5626 5626 5627 5627 5628 5628 5629 5629 5630 5630 5631 5631 5632 5626 5633 5633 5634 5634 5635 5635 5636 5636 5637 5637 5638 5638 5639 5639 5640 5640 5641 5641 5642 5642 5643 5643 5644 5644 5645 5645 5646 5646 5647 5647 5648 5648 5649 5649 5650 5650 5651 5651 5652 5652 5653 5653 5654 5654 5655 5655 5656 5656 5657 5657 5658 5658 5659 5659 5660 5660 5661 5661 5662 5662 5663 5663 5664 5664 5665 5665 5666 5666 5667 5667 5668 5668 5669 5669 5670 5670 5671 5671 5672 5672 5673 5673 5674 5674 5675 5675 5676 5676 5677 5677 5678 5678 5679 5679 5680 5680 5681 5681 5682 5682 5683 5683 5684 5684 5685 5685 5686 5686 5687 5687 5688 5688 5689 5689 5690 5690 5691 5691 5692 5692 5693 5693 5694 5694 5695 5695 5696 5696 5697 5697 5698 5698 5699 5699 5700 5700 5701 5701 5702 5702 5703 5703 5704 5704 5705 5705 5706 5706 5707 5707 5708 5708 5709 5709 5710 5710 5711 136 5712 5712 5713 5713 5714 11 5715 138 5716 146 5717 5717 5718 150 5719 5719 5720 5720 5721 5721 5722 5722 5723 123 5724 154 5725 5725 5726 5726 5727 2214 5728 5728 5729 5729 5730 5730 5731 5731 5732 5732 5733 5733 5734 5734 5735 5735 5736 5736 5737 5737 5738 5738 5739 5739 5740 5740 5741 2214 5742 5742 5743 5743 5744 5744 5745 5745 5746 5746 5747 5747 5748 5748 5749 5749 5750 5750 5751 5751 5752 2214 5753 2214 5754 5754 5755 5755 5756 5756 5757 5757 5758 5758 5759 5759 5760 5760 5761 5761 5762 5762 5763 5763 5764 5764 5765 2214 5766 5766 5767 5767 5768 5768 5769 5769 5770 5770 5771 5771 5772 5772 5773 5773 5774 5774 5775 5775 5776 5776 5777 5777 5778 5778 5779 5779 5780 5780 5781 5781 5782 5782 5783 5783 5784 5784 5785 5785 5786 5786 5787 5787 5788 5788 5789 5789 5790 5790 5791 5791 5792 5792 5793 5793 5794 5794 5795 5795 5796 5796 5797 5797 5798 5798 5799 5799 5800 5800 5801 5801 5802 5802 5803 5803 5804 5804 5805 5805 5806 5806 5807 5807 5808 5808 5809 5809 5810 5810 5811 5811 5812 5812 5813 5813 5814 5814 5815 5815 5816 5816 5817 5817 5818 5818 5819 5819 5820 5820 5821 5821 5822 5822 5823 5823 5824 5824 5825 5825 5826 5826 5827 5827 5828 5828 5829 5829 5830 5830 5831 5831 5832 5832 5833 5833 5834 5834 5835 5835 5836 5836 5837 5837 5838 5838 5839 5839 5840 5840 5841 5841 5842 5842 5843 5843 5844 5844 5845 5845 5846 5846 5847 5847 5848 5848 5849 5849 5850 5850 5851 5851 5852 5852 5853 5853 5854 5854 5855 5855 5856 2965 5857 2011 5858 5858 5859 5859 5860 2 5861 3032 5862 5862 5863 5626 5864 5864 5865 5865 5866 5866 5867 5867 5868 5868 5869 5869 5870 5870 5871 5871 5872 5872 5873 5873 5874 5874 5875 5875 5876 13 5877 5877 5878 5878 5879 5879 5880 5880 5881 5881 5882 5882 5883 5883 5884 5884 5885 5885 5886 5886 5887 5887 5888 5888 5889 5889 5890 5890 5891 5891 5892 5892 5893 5893 5894 5894 5895 5895 5896 5896 5897 5897 5898 5898 5899 5899 5900 5900 5901 5901 5902 5902 5903 5903 5904 5904 5905 5905 5906 5906 5907 5907 5908 5908 5909 5909 5910 5910 5911 5911 5912 5912 5913 5913 5914 5914 5915 5915 5916 5916 5917 5917 5918 5918 5919 5919 5920 5920 5921 5921 5922 5922 5923 5923 5924 5924 5925 5925 5926 5926 5927 5927 5928 5928 5929 5929 5930 5930 5931 5931 5932 5932 5933 5933 5934 5934 5935 5935 5936 5936 5937 5937 5938 5938 5939 5939 5940 5940 5941 5941 5942 5942 5943 5943 5944 5944 5945 5945 5946 5946 5947 5947 5948 5948 5949 5949 5950 5950 5951 5951 5952 5952 5953 5953 5954 5954 5955 5955 5956 5956 5957 5957 5958 5958 5959 5959 5960 5960 5961 5961 5962 97 5963 5963 5964 5964 5965 5965 5966 5966 5967 5967 5968 5879 5969 5969 5970 5970 5971 5969 5972 5972 5973 5973 5974 5626 5975 5975 5976 5976 5977 5977 5978 5978 5979 5979 5980 5980 5981 5981 5982 5982 5983 5983 5984 5984 5985 5985 5986 5986 5987 5987 5988 5988 5989 5989 5990 5990 5991 5991 5992 5992 5993 5993 5994 5994 5995 5995 5996 5996 5997 5997 5998 5998 5999 5999 6000 6000 6001 6001 6002 6002 6003 6003 6004 6004 6005 6005 6006 6006 6007 6007 6008 6008 6009 6009 6010 6010 6011 6011 6012 6012 6013 6013 6014 6014 6015 6015 6016 6016 6017 6017 6018 6018 6019 6019 6020 6020 6021 6021 6022 6022 6023 6023 6024 6024 6025 6025 6026 6026 6027 6027 6028 6028 6029 6029 6030 6030 6031 6031 6032 6032 6033 6033 6034 6034 6035 6035 6036 6036 6037 6037 6038 6038 6039 6039 6040 6040 6041 6041 6042 6042 6043 6043 6044 6044 6045 6045 6046 6046 6047 6047 6048 6048 6049 6049 6050 6050 6051 6051 6052 6052 6053 6053 6054 6054 6055 6055 6056 6056 6057 6057 6058 6058 6059 6059 6060 6060 6061 6061 6062 6062 6063 6063 6064 6064 6065 6065 6066 6066 6067 6067 6068 6068 6069 5626 6070 6070 6071 6071 6072 6072 6073 6073 6074 6074 6075 6075 6076 6076 6077 6077 6078 6078 6079 6079 6080 6080 6081 6081 6082 6082 6083 6083 6084 6084 6085 6085 6086 6086 6087 6087 6088 6088 6089 6089 6090 6090 6091 6091 6092 6092 6093 6093 6094 6094 6095 6095 6096 6096 6097 6097 6098 6098 6099 6099 6100 6100 6101 6101 6102 6102 6103 6103 6104 6104 6105 6105 6106 6106 6107 6107 6108 6108 6109 6109 6110 6110 6111 6111 6112 6112 6113 6113 6114 6114 6115 6115 6116 6116 6117 6117 6118 6118 6119 6119 6120 6120 6121 6121 6122 6122 6123 6123 6124 6124 6125 6125 6126 6126 6127 6127 6128 6128 6129 6129 6130 6130 6131 6131 6132 6132 6133 6133 6134 6134 6135 6135 6136 6136 6137 6137 6138 6138 6139 6139 6140 6140 6141 6141 6142 6142 6143 6143 6144 6144 6145 6145 6146 6146 6147 6147 6148 6148 6149 6149 6150 6150 6151 13 6152 6152 6153 6153 6154 6154 6155 6155 6156 6156 6157 6157 6158 6158 6159 6159 6160 13 6161 6161 6162 6162 6163 5626 6164 6164 6165 6165 6166 6166 6167 6167 6168 6168 6169 6169 6170 6170 6171 6171 6172 6172 6173 6173 6174 6174 6175 6175 6176 6176 6177 6177 6178 6178 6179 6179 6180 6180 6181 6181 6182 6182 6183 6183 6184 6184 6185 6185 6186 6186 6187 6187 6188 6188 6189 6189 6190 6190 6191 6191 6192 6192 6193 6193 6194 6194 6195 6195 6196 6196 6197 6197 6198 6198 6199 6199 6200 6200 6201 6201 6202 6202 6203 6203 6204 6204 6205 6205 6206 6206 6207 6207 6208 6208 6209 6209 6210 6210 6211 6211 6212 6212 6213 6213 6214 6214 6215 6215 6216 6216 6217 6217 6218 6218 6219 6219 6220 6220 6221 6221 6222 6222 6223 6223 6224 6224 6225 6225 6226 6226 6227 6227 6228 6228 6229 6229 6230 6230 6231 6231 6232 6232 6233 6233 6234 6234 6235 6235 6236 6236 6237 6237 6238 6238 6239 6239 6240 6240 6241 6241 6242 6242 6243 6243 6244 6244 6245 2 6246 5969 6247 6247 6248 6248 6249 2 6250 2 6251 2 6252 2 6253 2 6254 2 6255 2 6256 2 6257 288 6258 2 6259 2 6260 2 6261 2 6262 2 6263 294 6264 2 6265 2 6266 2 6267 2 6268 2 6269 2 6270 2 6271 2 6272 2 6273 2 6274 2 6275 2 6276 2 6277 2 6278 2 6279 6279 6280 6280 6281 6281 6282 6282 6283 6283 6284 6284 6285 6285 6286 6286 6287 6287 6288 6288 6289 6289 6290 6290 6291 6291 6292 6292 6293 6293 6294 4 6295 2 6296 6296 6297 6297 6298 6298 6299 6299 6300 6300 6301 6301 6302 6302 6303 6303 6304 6304 6305 6305 6306 6306 6307 338 6308 338 6309 40 6310 341 6311 2 6312 404 6313 6313 6314 555 6315 6315 6316 6316 6317 6317 6318 6318 6319 6319 6320 2 6321 404 6322 6322 6323 6323 6324 40 6325 6325 6326 264 6327 6327 6328 6328 6329 2 6330 404 6331 6331 6332 6332 6333 6333 6334 6334 6335 6335 6336 6336 6337 2 6338 713 6339 2 6340 6340 6341 6341 6342 6342 6343 6343 6344 6344 6345 6345 6346 6346 6347 6347 6348 6348 6349 6349 6350 6350 6351 6351 6352 6352 6353 6353 6354 6354 6355 6355 6356 6356 6357 6357 6358 6358 6359 917 6360 6360 6361 6361 6362 950 6363 143 6364 146 6365 6365 6366 6366 6367 6367 6368 555 6369 6369 6370 6370 6371 6371 6372 6372 6373 6373 6374 6374 6375 6375 6376 6376 6377 6377 6378 6378 6379 13 6380 6380 6381 6381 6382 6382 6383 6383 6384 6384 6385 6385 6386 6386 6387 6387 6388 6388 6389 950 6390 143 6391 146 6392 6392 6393 6393 6394 6356 6395 555 6396 6396 6397 264 6398 6398 6399 2619 6400 2 6401 404 6402 2622 6403 2 6404 264 6405 75 6406 3856 6407 2753 6408 6408 6409 1823 6410 146 6411 6411 6412 2 6413 143 6414 146 6415 6415 6416 6416 6417 6417 6418 6418 6419 6419 6420 5125 6421 6421 6422 146 6423 6423 6424 2 6425 264 6426 6426 6427 2 6428 264 6429 6429 6430 6430 6431 6431 6432 6432 6433 6433 6434 6434 6435 1823 6436 146 6437 6437 6438 6438 6439 6439 6440 6440 6441 264 6442 6442 6443 6443 6444 2 6445 404 6446 6446 6447 5157 6448 6448 6449 6449 6450 830 6451 6451 6452 11 6453 6453 6454 6454 6455 6455 6456 2653 6457 6457 6458 6458 6459 2011 6460 2 6461 6461 6462 338 6463 338 6464 341 6465 2 6466 404 6467 6467 6468 555 6469 6469 6470 6470 6471 6471 6472 6472 6473 2 6474 404 6475 6475 6476 6476 6477 6477 6478 6478 6479 6479 6480 2728 6481 6481 6482 555 6483 6483 6484 2 6485 6485 6486 6486 6487 1243 6488 2 6489 6489 6490 6490 6491 6491 6492 6492 6493 6493 6494 2820 6495 6495 6496 67 6497 6497 6498 6498 6499 6499 6500 6500 6501 6501 6502 6502 6503 2 6504 2854 6505 6505 6506 6506 6507 6507 6508 6508 6509 6509 6510 6510 6511 2863 6512 6512 6513 6513 6514 6514 6515 6515 6516 6516 6517 338 6518 338 6519 40 6520 341 6521 2 6522 404 6523 6523 6524 555 6525 6525 6526 6526 6527 6527 6528 6528 6529 917 6530 6530 6531 6531 6532 6532 6533 950 6534 143 6535 146 6536 6536 6537 6537 6538 40 6539 6539 6540 6540 6541 2901 6542 2903 6543 2905 6544 75 6545 2907 6546 2908 6547 6547 6548 264 6549 6549 6550 2 6551 264 6552 75 6553 3856 6554 2753 6555 6555 6556 1823 6557 146 6558 6558 6559 2 6560 143 6561 146 6562 6562 6563 6563 6564 6564 6565 6565 6566 6566 6567 5125 6568 6568 6569 146 6570 6570 6571 2 6572 264 6573 6573 6574 2 6575 264 6576 6576 6577 6577 6578 6578 6579 6579 6580 6580 6581 6581 6582 6582 6583 1823 6584 146 6585 6585 6586 6586 6587 6587 6588 6588 6589 264 6590 6590 6591 6591 6592 2 6593 404 6594 6594 6595 5157 6596 6596 6597 6597 6598 830 6599 6599 6600 6600 6601 6601 6602 6602 6603 6603 6604 6604 6605 6605 6606 6606 6607 6607 6608 6608 6609 6609 6610 6610 6611 6611 6612 6612 6613 6613 6614 6614 6615 6615 6616 6616 6617 6617 6618 6618 6619 6619 6620 6620 6621 6621 6622 6622 6623 6623 6624 6624 6625 6625 6626 6626 6627 6627 6628 6628 6629 6629 6630 6630 6631 6631 6632 6632 6633 6633 6634 6634 6635 6635 6636 6636 6637 6637 6638 6638 6639 6639 6640 6640 6641 6641 6642 6642 6643 6643 6644 6644 6645 6645 6646 6646 6647 6647 6648 6648 6649 6649 6650 6650 6651 6651 6652 6652 6653 6653 6654 6654 6655 6655 6656 6656 6657 6657 6658 6658 6659 6659 6660 6660 6661 6661 6662 6662 6663 6663 6664 6664 6665 6665 6666 6666 6667 6667 6668 6668 6669 6669 6670 6670 6671 6671 6672 6672 6673 6673 6674 6674 6675 6675 6676 6676 6677 6677 6678 11 6679 6679 6680 6680 6681 6681 6682 917 6683 6683 6684 6684 6685 950 6686 143 6687 146 6688 6688 6689 6689 6690 1456 6691 2 6692 2 6693 2 6694 6694 6695 2 6696 6696 6697 2 6698 2 6699 2 6700 2 6701 2 6702 2 6703 2 6704 2 6705 2 6706 6706 6707 6707 6708 6708 6709 6709 6710 6710 6711 6711 6712 6712 6713 6713 6714 6714 6715 6715 6716 6716 6717 6717 6718 6718 6719 22 6720 6720 6721 4 6722 6722 6723 87 6724 11 6725 6725 6726 97 6727 6727 6728 1517 6729 6729 6730 6730 6731 6731 6732 97 6733 830 6734 11 6735 6735 6736 917 6737 6737 6738 6738 6739 950 6740 143 6741 146 6742 6742 6743 6743 6744 6744 6745 830 6746 11 6747 6747 6748 6748 6749 6749 6750 338 6751 338 6752 341 6753 2 6754 404 6755 6755 6756 555 6757 6757 6758 6758 6759 6759 6760 6760 6761 2 6762 404 6763 6763 6764 6764 6765 6765 6766 6766 6767 6767 6768 2 6769 404 6770 6770 6771 83 6772 6772 6773 6773 6774 136 6775 11 6776 138 6777 146 6778 147 6779 149 6780 151 6781 75 6782 123 6783 154 6784 6784 6785 6785 6786 136 6787 11 6788 138 6789 146 6790 2206 6791 2208 6792 2209 6793 2210 6794 2 6795 2 6796 3 6797 4 6798 2 6799 2 6800 2 6801 2 6802 11 6803 2 6804 2 6805 1648 6806 2 6807 1650 6808 22 6809 1652 6810 1653 6811 2 6812 11 6813 11 6814 1658 6815 1659 6816 1660 6817 22 6818 2 6819 2 6820 2 6821 11 6822 1666 6823 11 6824 1668 6825 11 6826 1670 6827 1671 6828 1672 6829 1673 6830 60 6831 1675 6832 2 6833 2 6834 2 6835 2 6836 67 6837 127 6838 1682 6839 2 6840 2 6841 1685 6842 2 6843 1687 6844 1688 6845 2 6846 83 6847 84 6848 87 6849 130 6850 11 6851 1698 6852 1699 6853 97 6854 1701 6855 1702 6856 6856 6857 102 6858 102 6859 1706 6860 1707 6861 1708 6862 1709 6863 1710 6864 1711 6865 1714 6866 1715 6867 1716 6868 1717 6869 97 6870 2 6871 1721 6872 2 6873 2 6874 2 6875 2 6876 2 6877 2 6878 2 6879 2 6880 288 6881 2 6882 2 6883 2 6884 2 6885 2 6886 294 6887 2 6888 2 6889 2 6890 2 6891 2 6892 2 6893 2 6894 2 6895 2 6896 2 6897 2 6898 2 6899 2 6900 2 6901 2 6902 1752 6903 1753 6904 1754 6905 1755 6906 1756 6907 1757 6908 1758 6909 1759 6910 1760 6911 1761 6912 1762 6913 1763 6914 1764 6915 1765 6916 1766 6917 4 6918 2 6919 1769 6920 1770 6921 1771 6922 1772 6923 1773 6924 1774 6925 1775 6926 1776 6927 1777 6928 1778 6929 1779 6930 338 6931 338 6932 40 6933 341 6934 2 6935 404 6936 1786 6937 555 6938 1788 6939 1789 6940 1790 6941 1791 6942 1792 6943 2 6944 404 6945 1795 6946 1796 6947 40 6948 1798 6949 264 6950 1800 6951 1801 6952 2 6953 404 6954 1804 6955 1805 6956 1806 6957 1807 6958 1808 6959 1809 6960 2 6961 713 6962 2 6963 1813 6964 1814 6965 1815 6966 1816 6967 1817 6968 1818 6969 1819 6970 1820 6971 1821 6972 1822 6973 830 6974 6599 6975 6600 6976 6601 6977 6602 6978 6603 6979 6604 6980 6605 6981 6606 6982 6607 6983 6608 6984 6609 6985 6610 6986 6611 6987 6612 6988 6613 6989 6614 6990 6615 6991 6616 6992 6617 6993 6618 6994 6619 6995 6620 6996 6621 6997 6622 6998 6623 6999 6624 7000 6625 7001 6626 7002 6627 7003 6628 7004 6629 7005 6630 7006 6631 7007 6632 7008 6633 7009 6634 7010 6635 7011 6636 7012 6637 7013 6638 7014 6639 7015 6640 7016 6641 7017 6642 7018 6643 7019 6644 7020 6645 7021 6646 7022 6647 7023 6648 7024 6649 7025 6650 7026 6651 7027 6652 7028 6653 7029 6654 7030 6655 7031 6656 7032 6657 7033 6658 7034 6659 7035 6660 7036 6661 7037 6662 7038 6663 7039 6664 7040 6665 7041 6666 7042 6667 7043 6668 7044 6669 7045 6670 7046 6671 7047 6672 7048 6673 7049 6674 7050 6675 7051 6676 7052 6677 7053 11 7054 6679 7055 6680 7056 6681 7057 917 7058 6683 7059 6684 7060 950 7061 143 7062 146 7063 6688 7064 6689 7065 1456 7066 2 7067 2 7068 2 7069 6694 7070 2 7071 6696 7072 2 7073 2 7074 2 7075 2 7076 2 7077 2 7078 2 7079 2 7080 2 7081 6706 7082 6707 7083 6708 7084 6709 7085 6710 7086 6711 7087 6712 7088 6713 7089 6714 7090 6715 7091 6716 7092 6717 7093 6718 7094 22 7095 6720 7096 4 7097 6722 7098 87 7099 11 7100 6725 7101 97 7102 6727 7103 1517 7104 6729 7105 6730 7106 6731 7107 97 7108 830 7109 11 7110 6735 7111 917 7112 6737 7113 6738 7114 950 7115 143 7116 146 7117 6742 7118 6743 7119 6744 7120 830 7121 11 7122 6747 7123 6748 7124 6749 7125 338 7126 338 7127 341 7128 2 7129 404 7130 6755 7131 555 7132 6757 7133 6758 7134 6759 7135 6760 7136 2 7137 404 7138 6763 7139 6764 7140 6765 7141 6766 7142 6767 7143 2 7144 404 7145 6770 7146 6773 7147 136 7148 11 7149 138 7150 146 7151 147 7152 149 7153 151 7154 75 7155 123 7156 154 7157 6784 7158 6785 7159 136 7160 11 7161 138 7162 146 7163 2206 7164 2208 7165 2209 7166 2210 7167 2 7168 6856 7169 7169 7170 7169 7171 830 7172 7172 7173 11 7174 7174 7175 7175 7176 7176 7177 2653 7178 7178 7179 7179 7180 338 7181 338 7182 341 7183 2 7184 404 7185 7185 7186 555 7187 7187 7188 7188 7189 7189 7190 7190 7191 2 7192 404 7193 7193 7194 7194 7195 7195 7196 7196 7197 2728 7198 7198 7199 555 7200 7200 7201 2 7202 7202 7203 7203 7204 7204 7205 2 7206 7206 7207 7207 7208 7208 7209 7209 7210 7210 7211 2820 7212 7212 7213 67 7214 7214 7215 7215 7216 7216 7217 7217 7218 7218 7219 7219 7220 2 7221 2854 7222 7222 7223 7223 7224 7224 7225 7225 7226 7226 7227 7227 7228 2863 7229 7229 7230 7230 7231 7231 7232 7232 7233 7233 7234 338 7235 338 7236 40 7237 341 7238 2 7239 404 7240 7240 7241 555 7242 7242 7243 7243 7244 7244 7245 7245 7246 917 7247 7247 7248 7248 7249 7249 7250 950 7251 143 7252 146 7253 7253 7254 7254 7255 40 7256 7256 7257 7257 7258 2901 7259 2903 7260 2905 7261 75 7262 2907 7263 2908 7264 7264 7265 264 7266 7266 7267 2 7268 264 7269 75 7270 3856 7271 2753 7272 7272 7273 1823 7274 146 7275 2011 7276 2 7277 6461 7278 2 7279 143 7280 146 7281 7281 7282 7282 7283 7283 7284 7284 7285 7285 7286 5125 7287 7287 7288 146 7289 7289 7290 2 7291 264 7292 7292 7293 2 7294 264 7295 7295 7296 7296 7297 7297 7298 7298 7299 7299 7300 7300 7301 1823 7302 146 7303 7303 7304 7304 7305 7305 7306 7306 7307 264 7308 7308 7309 7309 7310 2 7311 404 7312 7312 7313 5157 7314 7314 7315 7315 7316 1456 7317 2 7318 3 7319 4 7320 2 7321 2 7322 2 7323 2 7324 11 7325 2 7326 2 7327 7327 7328 2 7329 7329 7330 22 7331 7331 7332 7332 7333 2 7334 11 7335 11 7336 7336 7337 7337 7338 7338 7339 22 7340 2 7341 2 7342 2 7343 11 7344 7344 7345 11 7346 7346 7347 11 7348 7348 7349 7349 7350 7350 7351 7351 7352 60 7353 7353 7354 2 7355 2 7356 2 7357 2 7358 67 7359 127 7360 7360 7361 2 7362 2 7363 7363 7364 2 7365 7365 7366 7366 7367 2 7368 83 7369 84 7370 22 7371 87 7372 130 7373 11 7374 7374 7375 7375 7376 97 7377 7377 7378 7378 7379 7379 7380 102 7381 102 7382 7382 7383 7383 7384 7384 7385 7385 7386 7386 7387 7387 7388 7388 7389 7389 7390 7390 7391 7391 7392 97 7393 2 7394 11 7395 7395 7396 2 7397 2 7398 2 7399 2 7400 2 7401 2 7402 2 7403 2 7404 288 7405 2 7406 2 7407 2 7408 2 7409 2 7410 294 7411 2 7412 2 7413 2 7414 2 7415 2 7416 2 7417 2 7418 2 7419 2 7420 2 7421 2 7422 2 7423 2 7424 2 7425 2 7426 7426 7427 7427 7428 7428 7429 7429 7430 7430 7431 7431 7432 7432 7433 7433 7434 7434 7435 7435 7436 7436 7437 7437 7438 7438 7439 7439 7440 7440 7441 4 7442 2 7443 7443 7444 7444 7445 7445 7446 7446 7447 7447 7448 7448 7449 7449 7450 7450 7451 7451 7452 7452 7453 7453 7454 338 7455 338 7456 40 7457 341 7458 2 7459 404 7460 7460 7461 555 7462 7462 7463 7463 7464 7464 7465 7465 7466 7466 7467 2 7468 404 7469 7469 7470 7470 7471 40 7472 7472 7473 264 7474 7474 7475 7475 7476 2 7477 404 7478 7478 7479 7479 7480 7480 7481 7481 7482 7482 7483 7483 7484 2 7485 713 7486 2 7487 7487 7488 7488 7489 7489 7490 7490 7491 7491 7492 7492 7493 7493 7494 7494 7495 7495 7496 7496 7497 1823 7498 146 7499 2011 7500 2 7501 7501 7502 830 7503 7503 7504 7504 7505 7505 7506 7506 7507 7507 7508 7508 7509 7509 7510 7510 7511 7511 7512 7512 7513 7513 7514 7514 7515 7515 7516 7516 7517 7517 7518 7518 7519 7519 7520 7520 7521 7521 7522 7522 7523 7523 7524 7524 7525 7525 7526 7526 7527 7527 7528 7528 7529 7529 7530 7530 7531 7531 7532 7532 7533 7533 7534 7534 7535 7535 7536 7536 7537 7537 7538 7538 7539 7539 7540 7540 7541 7541 7542 7542 7543 7543 7544 7544 7545 7545 7546 7546 7547 7547 7548 7548 7549 7549 7550 7550 7551 7551 7552 7552 7553 7553 7554 7554 7555 7555 7556 7556 7557 7557 7558 7558 7559 7559 7560 7560 7561 7561 7562 7562 7563 7563 7564 7564 7565 7565 7566 7566 7567 7567 7568 7568 7569 7569 7570 7570 7571 7571 7572 7572 7573 7573 7574 7574 7575 7575 7576 7576 7577 7577 7578 7578 7579 7579 7580 7580 7581 7581 7582 11 7583 7583 7584 7584 7585 7585 7586 917 7587 7587 7588 7588 7589 950 7590 143 7591 146 7592 7592 7593 7593 7594 1456 7595 2 7596 2 7597 2 7598 7598 7599 2 7600 7600 7601 2 7602 2 7603 2 7604 2 7605 2 7606 2 7607 2 7608 2 7609 2 7610 7610 7611 7611 7612 7612 7613 7613 7614 7614 7615 7615 7616 7616 7617 7617 7618 7618 7619 7619 7620 7620 7621 7621 7622 7622 7623 22 7624 7624 7625 4 7626 7626 7627 87 7628 11 7629 7629 7630 97 7631 7631 7632 1517 7633 7633 7634 7634 7635 7635 7636 97 7637 830 7638 11 7639 7639 7640 917 7641 7641 7642 7642 7643 950 7644 143 7645 146 7646 7646 7647 7647 7648 7648 7649 830 7650 11 7651 7651 7652 7652 7653 7653 7654 338 7655 338 7656 341 7657 2 7658 404 7659 7659 7660 555 7661 7661 7662 7662 7663 7663 7664 7664 7665 2 7666 404 7667 7667 7668 7668 7669 7669 7670 7670 7671 7671 7672 2 7673 404 7674 7674 7675 7675 7676 136 7677 11 7678 138 7679 146 7680 147 7681 149 7682 151 7683 75 7684 123 7685 154 7686 7686 7687 7687 7688 136 7689 11 7690 138 7691 146 7692 2206 7693 2208 7694 2209 7695 2210 7696 2 7697 2 7698 3 7699 4 7700 2 7701 2 7702 2 7703 2 7704 11 7705 2 7706 2 7707 7327 7708 2 7709 7329 7710 22 7711 7331 7712 7332 7713 2 7714 11 7715 11 7716 7336 7717 7337 7718 7338 7719 22 7720 2 7721 2 7722 2 7723 11 7724 7344 7725 11 7726 7346 7727 11 7728 7348 7729 7349 7730 7350 7731 7351 7732 60 7733 7353 7734 2 7735 2 7736 2 7737 2 7738 67 7739 127 7740 7360 7741 2 7742 2 7743 7363 7744 2 7745 7365 7746 7366 7747 2 7748 83 7749 84 7750 87 7751 130 7752 11 7753 7374 7754 7375 7755 97 7756 7377 7757 7378 7758 7758 7759 102 7760 102 7761 7382 7762 7383 7763 7384 7764 7385 7765 7386 7766 7387 7767 7388 7768 7389 7769 7390 7770 7391 7771 97 7772 2 7773 7395 7774 2 7775 2 7776 2 7777 2 7778 2 7779 2 7780 2 7781 2 7782 288 7783 2 7784 2 7785 2 7786 2 7787 2 7788 294 7789 2 7790 2 7791 2 7792 2 7793 2 7794 2 7795 2 7796 2 7797 2 7798 2 7799 2 7800 2 7801 2 7802 2 7803 2 7804 7426 7805 7427 7806 7428 7807 7429 7808 7430 7809 7431 7810 7432 7811 7433 7812 7434 7813 7435 7814 7436 7815 7437 7816 7438 7817 7439 7818 7440 7819 4 7820 2 7821 7443 7822 7444 7823 7445 7824 7446 7825 7447 7826 7448 7827 7449 7828 7450 7829 7451 7830 7452 7831 7453 7832 338 7833 338 7834 40 7835 341 7836 2 7837 404 7838 7460 7839 555 7840 7462 7841 7463 7842 7464 7843 7465 7844 7466 7845 2 7846 404 7847 7469 7848 7470 7849 40 7850 7472 7851 264 7852 7474 7853 7475 7854 2 7855 404 7856 7478 7857 7479 7858 7480 7859 7481 7860 7482 7861 7483 7862 2 7863 713 7864 2 7865 7487 7866 7488 7867 7489 7868 7490 7869 7491 7870 7492 7871 7493 7872 7494 7873 7495 7874 7496 7875 830 7876 7503 7877 7504 7878 7505 7879 7506 7880 7507 7881 7508 7882 7509 7883 7510 7884 7511 7885 7512 7886 7513 7887 7514 7888 7515 7889 7516 7890 7517 7891 7518 7892 7519 7893 7520 7894 7521 7895 7522 7896 7523 7897 7524 7898 7525 7899 7526 7900 7527 7901 7528 7902 7529 7903 7530 7904 7531 7905 7532 7906 7533 7907 7534 7908 7535 7909 7536 7910 7537 7911 7538 7912 7539 7913 7540 7914 7541 7915 7542 7916 7543 7917 7544 7918 7545 7919 7546 7920 7547 7921 7548 7922 7549 7923 7550 7924 7551 7925 7552 7926 7553 7927 7554 7928 7555 7929 7556 7930 7557 7931 7558 7932 7559 7933 7560 7934 7561 7935 7562 7936 7563 7937 7564 7938 7565 7939 7566 7940 7567 7941 7568 7942 7569 7943 7570 7944 7571 7945 7572 7946 7573 7947 7574 7948 7575 7949 7576 7950 7577 7951 7578 7952 7579 7953 7580 7954 7581 7955 11 7956 7583 7957 7584 7958 7585 7959 917 7960 7587 7961 7588 7962 950 7963 143 7964 146 7965 7592 7966 7593 7967 1456 7968 2 7969 2 7970 2 7971 7598 7972 2 7973 7600 7974 2 7975 2 7976 2 7977 2 7978 2 7979 2 7980 2 7981 2 7982 2 7983 7610 7984 7611 7985 7612 7986 7613 7987 7614 7988 7615 7989 7616 7990 7617 7991 7618 7992 7619 7993 7620 7994 7621 7995 7622 7996 22 7997 7624 7998 4 7999 7626 8000 87 8001 11 8002 7629 8003 97 8004 7631 8005 1517 8006 7633 8007 7634 8008 7635 8009 97 8010 830 8011 11 8012 7639 8013 917 8014 7641 8015 7642 8016 950 8017 143 8018 146 8019 7646 8020 7647 8021 7648 8022 830 8023 11 8024 7651 8025 7652 8026 7653 8027 338 8028 338 8029 341 8030 2 8031 404 8032 7659 8033 555 8034 7661 8035 7662 8036 7663 8037 7664 8038 2 8039 404 8040 7667 8041 7668 8042 7669 8043 7670 8044 7671 8045 2 8046 404 8047 7674 8048 7675 8049 136 8050 11 8051 138 8052 146 8053 147 8054 149 8055 151 8056 75 8057 123 8058 154 8059 7686 8060 7687 8061 136 8062 11 8063 138 8064 146 8065 2206 8066 2208 8067 2209 8068 2210 8069 2 8070 7758 8071 8071 8072 8071 8073 830 8074 8074 8075 11 8076 8076 8077 8077 8078 8078 8079 2653 8080 8080 8081 8081 8082 338 8083 338 8084 341 8085 2 8086 404 8087 8087 8088 555 8089 8089 8090 8090 8091 8091 8092 8092 8093 2 8094 404 8095 8095 8096 8096 8097 8097 8098 8098 8099 2728 8100 8100 8101 555 8102 8102 8103 2 8104 8104 8105 8105 8106 7204 8107 2 8108 8108 8109 8109 8110 8110 8111 8111 8112 8112 8113 2820 8114 8114 8115 67 8116 8116 8117 8117 8118 8118 8119 8119 8120 8120 8121 8121 8122 2 8123 2854 8124 8124 8125 8125 8126 8126 8127 8127 8128 8128 8129 8129 8130 2863 8131 8131 8132 8132 8133 8133 8134 8134 8135 8135 8136 338 8137 338 8138 40 8139 341 8140 2 8141 404 8142 8142 8143 555 8144 8144 8145 8145 8146 8146 8147 8147 8148 917 8149 8149 8150 8150 8151 8151 8152 950 8153 143 8154 146 8155 8155 8156 8156 8157 40 8158 8158 8159 8159 8160 2901 8161 2903 8162 2905 8163 75 8164 2907 8165 2908 8166 8166 8167 264 8168 8168 8169 2 8170 264 8171 75 8172 3856 8173 2753 8174 8174 8175 1823 8176 146 8177 2011 8178 2 8179 7501 8180 2 8181 143 8182 146 8183 8183 8184 8184 8185 8185 8186 8186 8187 8187 8188 5125 8189 8189 8190 146 8191 8191 8192 2 8193 264 8194 8194 8195 2 8196 264 8197 8197 8198 8198 8199 8199 8200 8200 8201 8201 8202 8202 8203 1823 8204 146 8205 8205 8206 8206 8207 8207 8208 8208 8209 264 8210 8210 8211 8211 8212 2 8213 404 8214 8214 8215 5157 8216 8216 8217 8217 8218 830 8219 8074 8220 11 8221 8076 8222 8077 8223 8078 8224 2653 8225 8080 8226 8081 8227 338 8228 338 8229 341 8230 2 8231 404 8232 8087 8233 555 8234 8089 8235 8090 8236 8091 8237 8092 8238 2 8239 404 8240 8095 8241 8096 8242 8097 8243 8098 8244 4 8245 8245 8246 87 8247 11 8248 8248 8249 97 8250 8250 8251 1517 8252 8252 8253 8253 8254 8254 8255 97 8256 830 8257 11 8258 8258 8259 917 8260 8260 8261 8261 8262 950 8263 143 8264 146 8265 8265 8266 8266 8267 8267 8268 830 8269 11 8270 8270 8271 8271 8272 8272 8273 338 8274 338 8275 341 8276 2 8277 404 8278 8278 8279 555 8280 8280 8281 8281 8282 8282 8283 8283 8284 2 8285 404 8286 8286 8287 8287 8288 8288 8289 8289 8290 8290 8291 2 8292 404 8293 8293 8294 2728 8295 8100 8296 555 8297 8102 8298 2 8299 8104 8300 8105 8301 7204 8302 2 8303 8108 8304 8109 8305 8110 8306 8111 8307 8112 8308 2820 8309 8114 8310 67 8311 8116 8312 8117 8313 8118 8314 8119 8315 8120 8316 8121 8317 2854 8318 8129 8319 2863 8320 8132 8321 917 8322 8149 8323 8150 8324 8151 8325 950 8326 143 8327 146 8328 8155 8329 8156 8330 8159 8331 2901 8332 2903 8333 2905 8334 75 8335 2907 8336 2908 8337 8166 8338 2 8339 264 8340 75 8341 3856 8342 2753 8343 8174 8344 1823 8345 146 8346 8346 8347 2 8348 143 8349 146 8350 8183 8351 8184 8352 8185 8353 8186 8354 8187 8355 5125 8356 8189 8357 146 8358 8191 8359 2 8360 264 8361 8194 8362 2 8363 264 8364 8197 8365 8198 8366 8199 8367 8200 8368 8201 8369 8202 8370 1823 8371 146 8372 8205 8373 8206 8374 8207 8375 8208 8376 264 8377 8210 8378 8211 8379 2 8380 404 8381 8214 8382 5157 8383 8216 8384 8217 8385 830 8386 7172 8387 11 8388 7174 8389 7175 8390 7176 8391 2653 8392 7178 8393 7179 8394 8394 8395 338 8396 338 8397 341 8398 2 8399 404 8400 7185 8401 555 8402 7187 8403 7188 8404 7189 8405 7190 8406 2 8407 404 8408 7193 8409 7194 8410 7195 8411 7196 8412 8412 8413 2728 8414 7198 8415 555 8416 7200 8417 2 8418 7202 8419 7203 8420 7204 8421 2 8422 7206 8423 7207 8424 7208 8425 7209 8426 7210 8427 2820 8428 7212 8429 67 8430 7214 8431 7215 8432 7216 8433 7217 8434 7218 8435 7219 8436 2854 8437 7227 8438 2863 8439 7230 8440 917 8441 7247 8442 7248 8443 7249 8444 950 8445 143 8446 146 8447 7253 8448 7254 8449 7257 8450 2901 8451 2903 8452 2905 8453 75 8454 2907 8455 2908 8456 7264 8457 2 8458 264 8459 75 8460 3856 8461 2753 8462 7272 8463 1823 8464 146 8465 8465 8466 2 8467 143 8468 146 8469 7281 8470 7282 8471 7283 8472 7284 8473 7285 8474 5125 8475 7287 8476 146 8477 7289 8478 2 8479 264 8480 7292 8481 2 8482 264 8483 7295 8484 8484 8485 7296 8486 7297 8487 7298 8488 7299 8489 7300 8490 1823 8491 146 8492 7303 8493 7304 8494 7305 8495 7306 8496 264 8497 7308 8498 7309 8499 2 8500 404 8501 7312 8502 5157 8503 7314 8504 7315 8505 1823 8506 146 8507 8507 8508 264 8509 8509 8510 2183 8511 2 8512 404 8513 2186 8514 5157 8515 8515 8516 2189 8517 136 8518 11 8519 138 8520 146 8521 147 8522 149 8523 151 8524 75 8525 123 8526 154 8527 2200 8528 2201 8529 136 8530 11 8531 138 8532 146 8533 2206 8534 2208 8535 2209 8536 2210 8537 2 8538 2212 8539 2213 8540 2 8541 2813 8542 830 8543 6451 8544 11 8545 6453 8546 6454 8547 6455 8548 2653 8549 6457 8550 6458 8551 338 8552 338 8553 341 8554 2 8555 404 8556 6467 8557 555 8558 6469 8559 6470 8560 6471 8561 6472 8562 2 8563 404 8564 6475 8565 6476 8566 6477 8567 6478 8568 2728 8569 6481 8570 555 8571 6483 8572 2 8573 6485 8574 6486 8575 1243 8576 2 8577 6489 8578 6490 8579 6491 8580 6492 8581 6493 8582 2820 8583 6495 8584 67 8585 6497 8586 6498 8587 6499 8588 6500 8589 6501 8590 6502 8591 2854 8592 6510 8593 2863 8594 6513 8595 917 8596 6530 8597 6531 8598 6532 8599 950 8600 143 8601 146 8602 6536 8603 6537 8604 6540 8605 2901 8606 2903 8607 2905 8608 75 8609 2907 8610 2908 8611 6547 8612 2 8613 264 8614 75 8615 3856 8616 2753 8617 6555 8618 1823 8619 146 8620 2011 8621 2 8622 2013 8623 2 8624 143 8625 146 8626 6562 8627 6563 8628 6564 8629 6565 8630 6566 8631 5125 8632 6568 8633 146 8634 6570 8635 2 8636 264 8637 6573 8638 2 8639 264 8640 6576 8641 6578 8642 6579 8643 6580 8644 6581 8645 6582 8646 1823 8647 146 8648 6585 8649 6586 8650 6587 8651 6588 8652 264 8653 6590 8654 6591 8655 2 8656 404 8657 6594 8658 5157 8659 6596 8660 6597 8661 2815 8662 2817 8663 2847 8664 2848 8665 2849 8666 2854 8667 2861 8668 2863 8669 2865 8670 917 8671 2887 8672 2888 8673 2889 8674 950 8675 143 8676 146 8677 2893 8678 2894 8679 2897 8680 2901 8681 2903 8682 2905 8683 75 8684 2907 8685 2908 8686 2929 8687 2 8688 264 8689 75 8690 3856 8691 2753 8692 6408 8693 1823 8694 146 8695 2011 8696 2 8697 2449 8698 2 8699 143 8700 146 8701 6415 8702 6417 8703 6418 8704 5125 8705 6421 8706 146 8707 6423 8708 2 8709 264 8710 6426 8711 2 8712 264 8713 6429 8714 6432 8715 1823 8716 146 8717 6437 8718 830 8719 2451 8720 2452 8721 2453 8722 2454 8723 2455 8724 2456 8725 2457 8726 2458 8727 2459 8728 2460 8729 2461 8730 2462 8731 2463 8732 2464 8733 2465 8734 2466 8735 2467 8736 2468 8737 2469 8738 2470 8739 2471 8740 2472 8741 2473 8742 2474 8743 2475 8744 2476 8745 2477 8746 2478 8747 2479 8748 2480 8749 2481 8750 2482 8751 2483 8752 2484 8753 2485 8754 2486 8755 2487 8756 2488 8757 2489 8758 2490 8759 2491 8760 2492 8761 2493 8762 2494 8763 2495 8764 2496 8765 2497 8766 2498 8767 2499 8768 2500 8769 2501 8770 2502 8771 2503 8772 2504 8773 2505 8774 2506 8775 2507 8776 2508 8777 2509 8778 2510 8779 2511 8780 2512 8781 2513 8782 2514 8783 2515 8784 2516 8785 2517 8786 2518 8787 2519 8788 2520 8789 2521 8790 2522 8791 2523 8792 2524 8793 2525 8794 2526 8795 2527 8796 2528 8797 2529 8798 11 8799 2531 8800 2532 8801 2533 8802 917 8803 2535 8804 2536 8805 950 8806 143 8807 146 8808 2540 8809 2541 8810 1456 8811 2 8812 2 8813 2 8814 2546 8815 2 8816 2548 8817 2 8818 2 8819 2 8820 2 8821 2 8822 2 8823 2 8824 2 8825 2 8826 2558 8827 2559 8828 2560 8829 2561 8830 2562 8831 2563 8832 2564 8833 2565 8834 2566 8835 2567 8836 2568 8837 2569 8838 2570 8839 22 8840 2572 8841 87 8842 11 8843 2577 8844 97 8845 2579 8846 1517 8847 2581 8848 2582 8849 2583 8850 97 8851 830 8852 11 8853 2587 8854 917 8855 2589 8856 2590 8857 950 8858 143 8859 146 8860 2594 8861 2595 8862 2596 8863 830 8864 11 8865 2599 8866 2600 8867 2601 8868 338 8869 338 8870 341 8871 2 8872 404 8873 2607 8874 555 8875 2609 8876 2611 8877 2616 8878 2617 8879 2618 8880 146 8881 8881 8882 2623 8883 2 8884 8884 8885 136 8886 11 8887 138 8888 146 8889 147 8890 149 8891 151 8892 75 8893 123 8894 154 8895 2634 8896 2635 8897 136 8898 11 8899 138 8900 146 8901 2206 8902 2208 8903 2209 8904 2210 8905 2 8906 2645 8907 2646 8908 830 8909 2648 8910 11 8911 2650 8912 2651 8913 2652 8914 2653 8915 2655 8916 2656 8917 338 8918 338 8919 341 8920 2 8921 404 8922 2662 8923 555 8924 2664 8925 2665 8926 2666 8927 2667 8928 2 8929 404 8930 2670 8931 2671 8932 2672 8933 2673 8934 2728 8935 2733 8936 555 8937 2807 8938 2 8939 2809 8940 2810 8941 2814 8942 2816 8943 2820 8944 2821 8945 67 8946 2826 8947 2835 8948 2850 8949 6416 8950 6419 8951 8951 8952 6430 8953 6431 8954 6433 8955 6434 8956 6438 8957 6439 8958 6440 8959 264 8960 6442 8961 6443 8962 2 8963 404 8964 6446 8965 5157 8966 6448 8967 6449 8968 1456 8969 2 8970 2 8971 2 8972 8972 8973 2 8974 8974 8975 2 8976 2 8977 2 8978 2 8979 2 8980 2 8981 2 8982 2 8983 2 8984 8984 8985 8985 8986 8986 8987 8987 8988 8988 8989 8989 8990 8990 8991 8991 8992 8992 8993 8993 8994 8994 8995 8995 8996 8996 8997 22 8998 8998 8999 4 9000 9000 9001 87 9002 11 9003 9003 9004 97 9005 9005 9006 1517 9007 9007 9008 9008 9009 9009 9010 97 9011 830 9012 11 9013 9013 9014 917 9015 9015 9016 9016 9017 950 9018 143 9019 146 9020 9020 9021 9021 9022 9022 9023 830 9024 11 9025 9025 9026 9026 9027 9027 9028 338 9029 338 9030 341 9031 2 9032 404 9033 9033 9034 555 9035 9035 9036 9036 9037 9037 9038 9038 9039 2 9040 404 9041 9041 9042 9042 9043 9043 9044 9044 9045 9045 9046 2 9047 404 9048 9048 9049 9049 9050 136 9051 11 9052 138 9053 146 9054 147 9055 149 9056 151 9057 75 9058 123 9059 154 9060 9060 9061 9061 9062 136 9063 11 9064 138 9065 146 9066 2206 9067 2208 9068 2209 9069 2210 9070 2 9071 2 9072 3 9073 4 9074 2 9075 2 9076 2 9077 2 9078 11 9079 2 9080 2 9081 662 9082 2 9083 664 9084 22 9085 666 9086 667 9087 2 9088 11 9089 11 9090 671 9091 672 9092 673 9093 22 9094 2 9095 2 9096 2 9097 11 9098 679 9099 11 9100 681 9101 11 9102 683 9103 684 9104 685 9105 686 9106 60 9107 688 9108 2 9109 2 9110 2 9111 2 9112 67 9113 127 9114 695 9115 2 9116 2 9117 698 9118 2 9119 700 9120 701 9121 2 9122 83 9123 84 9124 87 9125 130 9126 11 9127 708 9128 709 9129 97 9130 711 9131 712 9132 9132 9133 102 9134 102 9135 716 9136 717 9137 718 9138 719 9139 720 9140 721 9141 722 9142 723 9143 724 9144 725 9145 97 9146 2 9147 728 9148 2 9149 2 9150 2 9151 2 9152 2 9153 2 9154 2 9155 2 9156 288 9157 2 9158 2 9159 2 9160 2 9161 2 9162 294 9163 2 9164 2 9165 2 9166 2 9167 2 9168 2 9169 2 9170 2 9171 2 9172 2 9173 2 9174 2 9175 2 9176 2 9177 2 9178 759 9179 760 9180 761 9181 762 9182 763 9183 764 9184 765 9185 766 9186 767 9187 768 9188 769 9189 770 9190 771 9191 772 9192 773 9193 4 9194 2 9195 776 9196 777 9197 778 9198 779 9199 780 9200 781 9201 782 9202 783 9203 784 9204 785 9205 786 9206 338 9207 338 9208 40 9209 341 9210 2 9211 404 9212 793 9213 555 9214 795 9215 796 9216 797 9217 798 9218 799 9219 2 9220 404 9221 802 9222 803 9223 40 9224 805 9225 264 9226 807 9227 808 9228 2 9229 404 9230 811 9231 812 9232 813 9233 814 9234 815 9235 816 9236 2 9237 713 9238 2 9239 820 9240 821 9241 822 9242 823 9243 824 9244 825 9245 826 9246 827 9247 828 9248 829 9249 830 9250 831 9251 832 9252 833 9253 834 9254 835 9255 836 9256 837 9257 838 9258 839 9259 840 9260 841 9261 842 9262 843 9263 844 9264 845 9265 846 9266 847 9267 848 9268 849 9269 850 9270 851 9271 852 9272 853 9273 854 9274 855 9275 856 9276 857 9277 858 9278 859 9279 860 9280 861 9281 862 9282 863 9283 864 9284 865 9285 866 9286 867 9287 868 9288 869 9289 870 9290 871 9291 872 9292 873 9293 874 9294 875 9295 876 9296 877 9297 878 9298 879 9299 880 9300 881 9301 882 9302 883 9303 884 9304 885 9305 886 9306 887 9307 888 9308 889 9309 890 9310 891 9311 892 9312 893 9313 894 9314 895 9315 896 9316 897 9317 898 9318 899 9319 900 9320 901 9321 902 9322 903 9323 904 9324 905 9325 906 9326 907 9327 908 9328 909 9329 11 9330 911 9331 912 9332 913 9333 917 9334 918 9335 923 9336 950 9337 143 9338 146 9339 8881 9340 8951 9341 1456 9342 2 9343 2 9344 2 9345 8972 9346 2 9347 8974 9348 2 9349 2 9350 2 9351 2 9352 2 9353 2 9354 2 9355 2 9356 2 9357 8984 9358 8985 9359 8986 9360 8987 9361 8988 9362 8989 9363 8990 9364 8991 9365 8992 9366 8993 9367 8994 9368 8995 9369 8996 9370 22 9371 8998 9372 4 9373 9000 9374 87 9375 11 9376 9003 9377 97 9378 9005 9379 1517 9380 9007 9381 9008 9382 9009 9383 97 9384 830 9385 11 9386 9013 9387 917 9388 9015 9389 9016 9390 950 9391 143 9392 146 9393 9020 9394 9021 9395 9022 9396 830 9397 11 9398 9025 9399 9026 9400 9027 9401 338 9402 338 9403 341 9404 2 9405 404 9406 9033 9407 555 9408 9035 9409 9036 9410 9037 9411 9038 9412 2 9413 404 9414 9041 9415 9042 9416 9043 9417 9044 9418 9045 9419 2 9420 404 9421 9048 9422 9049 9423 136 9424 11 9425 138 9426 146 9427 147 9428 149 9429 151 9430 75 9431 123 9432 154 9433 9060 9434 9061 9435 136 9436 11 9437 138 9438 146 9439 2206 9440 2208 9441 2209 9442 2210 9443 2 9444 9132 9445 9445 9446 9445 9447 830 9448 9448 9449 11 9450 9450 9451 9451 9452 9452 9453 2653 9454 9454 9455 9455 9456 338 9457 338 9458 341 9459 2 9460 404 9461 9461 9462 555 9463 9463 9464 9464 9465 9465 9466 9466 9467 2 9468 404 9469 9469 9470 9470 9471 9471 9472 9472 9473 2728 9474 9474 9475 555 9476 9476 9477 2 9478 9478 9479 9479 9480 2 9481 3 9482 4 9483 2 9484 2 9485 2 9486 2 9487 11 9488 2 9489 2 9490 9490 9491 2 9492 9492 9493 22 9494 9494 9495 9495 9496 2 9497 11 9498 11 9499 9499 9500 9500 9501 9501 9502 22 9503 2 9504 2 9505 2 9506 11 9507 9507 9508 11 9509 9509 9510 11 9511 9511 9512 9512 9513 9513 9514 9514 9515 60 9516 9516 9517 2 9518 2 9519 2 9520 2 9521 67 9522 127 9523 9523 9524 2 9525 2 9526 9526 9527 2 9528 9528 9529 9529 9530 2 9531 83 9532 84 9533 87 9534 130 9535 11 9536 9536 9537 9537 9538 97 9539 9539 9540 9540 9541 9541 9542 102 9543 102 9544 9544 9545 9545 9546 9546 9547 9547 9548 9548 9549 9549 9550 9550 9551 9551 9552 9552 9553 9553 9554 97 9555 2 9556 9556 9557 2 9558 2 9559 2 9560 2 9561 2 9562 2 9563 2 9564 2 9565 288 9566 2 9567 2 9568 2 9569 2 9570 2 9571 294 9572 2 9573 2 9574 2 9575 2 9576 2 9577 2 9578 2 9579 2 9580 2 9581 2 9582 2 9583 2 9584 2 9585 2 9586 2 9587 9587 9588 9588 9589 9589 9590 9590 9591 9591 9592 9592 9593 9593 9594 9594 9595 9595 9596 9596 9597 9597 9598 9598 9599 9599 9600 9600 9601 9601 9602 4 9603 2 9604 9604 9605 9605 9606 9606 9607 9607 9608 9608 9609 9609 9610 9610 9611 9611 9612 9612 9613 9613 9614 9614 9615 338 9616 338 9617 40 9618 341 9619 2 9620 404 9621 9621 9622 555 9623 9623 9624 9624 9625 9625 9626 9626 9627 9627 9628 2 9629 404 9630 9630 9631 9631 9632 40 9633 9633 9634 264 9635 9635 9636 9636 9637 2 9638 404 9639 9639 9640 9640 9641 9641 9642 9642 9643 9643 9644 9644 9645 2 9646 713 9647 2 9648 9648 9649 9649 9650 9650 9651 9651 9652 9652 9653 9653 9654 9654 9655 9655 9656 9656 9657 9657 9658 830 9659 9659 9660 9660 9661 9661 9662 9662 9663 9663 9664 9664 9665 9665 9666 9666 9667 9667 9668 9668 9669 9669 9670 9670 9671 9671 9672 9672 9673 9673 9674 9674 9675 9675 9676 9676 9677 9677 9678 9678 9679 9679 9680 9680 9681 9681 9682 9682 9683 9683 9684 9684 9685 9685 9686 9686 9687 9687 9688 9688 9689 9689 9690 9690 9691 9691 9692 9692 9693 9693 9694 9694 9695 9695 9696 9696 9697 9697 9698 9698 9699 9699 9700 9700 9701 9701 9702 9702 9703 9703 9704 9704 9705 9705 9706 9706 9707 9707 9708 9708 9709 9709 9710 9710 9711 9711 9712 9712 9713 9713 9714 9714 9715 9715 9716 9716 9717 9717 9718 9718 9719 9719 9720 9720 9721 9721 9722 9722 9723 9723 9724 9724 9725 9725 9726 9726 9727 9727 9728 9728 9729 9729 9730 9730 9731 9731 9732 9732 9733 9733 9734 9734 9735 9735 9736 9736 9737 9737 9738 11 9739 9739 9740 9740 9741 9741 9742 917 9743 9743 9744 9744 9745 950 9746 143 9747 146 9748 9748 9749 9749 9750 1456 9751 2 9752 2 9753 2 9754 9754 9755 2 9756 9756 9757 2 9758 2 9759 2 9760 2 9761 2 9762 2 9763 2 9764 2 9765 2 9766 9766 9767 9767 9768 9768 9769 9769 9770 9770 9771 9771 9772 9772 9773 9773 9774 9774 9775 9775 9776 9776 9777 9777 9778 9778 9779 22 9780 9780 9781 4 9782 9782 9783 87 9784 11 9785 9785 9786 97 9787 9787 9788 1517 9789 9789 9790 9790 9791 9791 9792 97 9793 830 9794 11 9795 9795 9796 917 9797 9797 9798 9798 9799 950 9800 143 9801 146 9802 9802 9803 9803 9804 9804 9805 830 9806 11 9807 9807 9808 9808 9809 9809 9810 338 9811 338 9812 341 9813 2 9814 404 9815 9815 9816 555 9817 9817 9818 9818 9819 9819 9820 9820 9821 2 9822 404 9823 9823 9824 9824 9825 9825 9826 9826 9827 9827 9828 2 9829 404 9830 9830 9831 9831 9832 136 9833 11 9834 138 9835 146 9836 147 9837 149 9838 151 9839 75 9840 123 9841 154 9842 9842 9843 9843 9844 136 9845 11 9846 138 9847 146 9848 2206 9849 2208 9850 2209 9851 2210 9852 2 9853 9853 9854 9854 9855 2 9856 9856 9857 830 9858 9858 9859 11 9860 9860 9861 9861 9862 9862 9863 2653 9864 9864 9865 9865 9866 338 9867 338 9868 341 9869 2 9870 404 9871 9871 9872 555 9873 9873 9874 9874 9875 9875 9876 9876 9877 2 9878 404 9879 9879 9880 9880 9881 9881 9882 9882 9883 2728 9884 9884 9885 555 9886 9886 9887 2 9888 9888 9889 9889 9890 9541 9891 2 9892 9892 9893 9893 9894 9894 9895 9895 9896 9896 9897 2820 9898 9898 9899 67 9900 9900 9901 9901 9902 9902 9903 9903 9904 9904 9905 9905 9906 2 9907 2854 9908 9908 9909 9909 9910 9910 9911 9911 9912 9912 9913 9913 9914 2863 9915 9915 9916 9916 9917 9917 9918 9918 9919 9919 9920 338 9921 338 9922 40 9923 341 9924 2 9925 404 9926 9926 9927 555 9928 9928 9929 9929 9930 9930 9931 9931 9932 917 9933 9933 9934 9934 9935 9935 9936 950 9937 143 9938 146 9939 9939 9940 9940 9941 40 9942 9942 9943 9943 9944 2901 9945 2903 9946 2905 9947 75 9948 2907 9949 2908 9950 9950 9951 264 9952 9952 9953 2 9954 264 9955 75 9956 3856 9957 2753 9958 9958 9959 1823 9960 146 9961 2011 9962 2 9963 9963 9964 2 9965 143 9966 146 9967 9967 9968 9968 9969 9969 9970 9970 9971 9971 9972 5125 9973 9973 9974 146 9975 9975 9976 2 9977 264 9978 9978 9979 2 9980 264 9981 9981 9982 9982 9983 9983 9984 9984 9985 9985 9986 9986 9987 1823 9988 146 9989 9989 9990 9990 9991 9991 9992 9992 9993 264 9994 9994 9995 9995 9996 2 9997 404 9998 9998 9999 5157 10000 10000 10001 10001 10002 10002 10003 10003 10004 10004 10005 10005 10006 2820 10007 10007 10008 67 10009 10009 10010 10010 10011 10011 10012 10012 10013 10013 10014 10014 10015 2 10016 2854 10017 10017 10018 10018 10019 10019 10020 10020 10021 10021 10022 10022 10023 2863 10024 10024 10025 10025 10026 10026 10027 10027 10028 10028 10029 338 10030 338 10031 40 10032 341 10033 2 10034 404 10035 10035 10036 555 10037 10037 10038 10038 10039 10039 10040 10040 10041 917 10042 10042 10043 10043 10044 10044 10045 950 10046 143 10047 146 10048 10048 10049 10049 10050 40 10051 10051 10052 10052 10053 2901 10054 2903 10055 2905 10056 75 10057 2907 10058 2908 10059 10059 10060 264 10061 10061 10062 2 10063 264 10064 75 10065 3856 10066 2753 10067 10067 10068 1823 10069 146 10070 2011 10071 2 10072 10072 10073 2 10074 143 10075 146 10076 10076 10077 10077 10078 10078 10079 10079 10080 10080 10081 5125 10082 10082 10083 146 10084 10084 10085 2 10086 264 10087 10087 10088 2 10089 264 10090 10090 10091 10091 10092 10092 10093 10093 10094 10094 10095 10095 10096 1823 10097 146 10098 10098 10099 10099 10100 10100 10101 10101 10102 264 10103 10103 10104 10104 10105 2 10106 404 10107 10107 10108 5157 10109 10109 10110 10110 10111 2 10112 10112 10113 830 10114 9448 10115 11 10116 9450 10117 9451 10118 9452 10119 2653 10120 9454 10121 9455 10122 338 10123 338 10124 341 10125 2 10126 404 10127 9461 10128 555 10129 9463 10130 9464 10131 9465 10132 9466 10133 2 10134 404 10135 9469 10136 9470 10137 9471 10138 9472 10139 2728 10140 9474 10141 555 10142 9476 10143 2 10144 9478 10145 9479 10146 2 10147 3 10148 4 10149 2 10150 2 10151 2 10152 2 10153 11 10154 2 10155 2 10156 9490 10157 2 10158 9492 10159 22 10160 9494 10161 9495 10162 2 10163 11 10164 11 10165 9499 10166 9500 10167 9501 10168 22 10169 2 10170 2 10171 2 10172 11 10173 9507 10174 11 10175 9509 10176 11 10177 9511 10178 9512 10179 9513 10180 9514 10181 60 10182 9516 10183 2 10184 2 10185 2 10186 2 10187 67 10188 127 10189 9523 10190 2 10191 2 10192 9526 10193 2 10194 9528 10195 9529 10196 2 10197 83 10198 84 10199 87 10200 130 10201 11 10202 9536 10203 9537 10204 97 10205 9539 10206 9540 10207 9541 10208 102 10209 102 10210 9544 10211 9545 10212 9546 10213 9547 10214 9548 10215 9549 10216 9550 10217 9551 10218 9552 10219 9553 10220 97 10221 2 10222 9556 10223 2 10224 2 10225 2 10226 2 10227 2 10228 2 10229 2 10230 2 10231 288 10232 2 10233 2 10234 2 10235 2 10236 2 10237 294 10238 2 10239 2 10240 2 10241 2 10242 2 10243 2 10244 2 10245 2 10246 2 10247 2 10248 2 10249 2 10250 2 10251 2 10252 2 10253 9587 10254 9588 10255 9589 10256 9590 10257 9591 10258 9592 10259 9593 10260 9594 10261 9595 10262 9596 10263 9597 10264 9598 10265 9599 10266 9600 10267 9601 10268 4 10269 2 10270 9604 10271 9605 10272 9606 10273 9607 10274 9608 10275 9609 10276 9610 10277 9611 10278 9612 10279 9613 10280 9614 10281 338 10282 338 10283 40 10284 341 10285 2 10286 404 10287 9621 10288 555 10289 9623 10290 9624 10291 9625 10292 9626 10293 9627 10294 2 10295 404 10296 9630 10297 9631 10298 40 10299 9633 10300 264 10301 9635 10302 9636 10303 2 10304 404 10305 9639 10306 9640 10307 9641 10308 9642 10309 9643 10310 9644 10311 2 10312 713 10313 2 10314 9648 10315 9649 10316 9650 10317 9651 10318 9652 10319 9653 10320 9654 10321 9655 10322 9656 10323 9657 10324 830 10325 9659 10326 9660 10327 9661 10328 9662 10329 9663 10330 9664 10331 9665 10332 9666 10333 9667 10334 9668 10335 9669 10336 9670 10337 9671 10338 9672 10339 9673 10340 9674 10341 9675 10342 9676 10343 9677 10344 9678 10345 9679 10346 9680 10347 9681 10348 9682 10349 9683 10350 9684 10351 9685 10352 9686 10353 9687 10354 9688 10355 9689 10356 9690 10357 9691 10358 9692 10359 9693 10360 9694 10361 9695 10362 9696 10363 9697 10364 9698 10365 9699 10366 9700 10367 9701 10368 9702 10369 9703 10370 9704 10371 9705 10372 9706 10373 9707 10374 9708 10375 9709 10376 9710 10377 9711 10378 9712 10379 9713 10380 9714 10381 9715 10382 9716 10383 9717 10384 9718 10385 9719 10386 9720 10387 9721 10388 9722 10389 9723 10390 9724 10391 9725 10392 9726 10393 9727 10394 9728 10395 9729 10396 9730 10397 9731 10398 9732 10399 9733 10400 9734 10401 9735 10402 9736 10403 9737 10404 11 10405 9739 10406 9740 10407 9741 10408 917 10409 9743 10410 9744 10411 950 10412 143 10413 146 10414 9748 10415 9749 10416 1456 10417 2 10418 2 10419 2 10420 9754 10421 2 10422 9756 10423 2 10424 2 10425 2 10426 2 10427 2 10428 2 10429 2 10430 2 10431 2 10432 9766 10433 9767 10434 9768 10435 9769 10436 9770 10437 9771 10438 9772 10439 9773 10440 9774 10441 9775 10442 9776 10443 9777 10444 9778 10445 22 10446 9780 10447 4 10448 9782 10449 87 10450 11 10451 9785 10452 97 10453 9787 10454 1517 10455 9789 10456 9790 10457 9791 10458 97 10459 830 10460 11 10461 9795 10462 917 10463 9797 10464 9798 10465 950 10466 143 10467 146 10468 9802 10469 9803 10470 9804 10471 830 10472 11 10473 9807 10474 9808 10475 9809 10476 338 10477 338 10478 341 10479 2 10480 404 10481 9815 10482 555 10483 9817 10484 9818 10485 9819 10486 9820 10487 2 10488 404 10489 9823 10490 9824 10491 9825 10492 9826 10493 9827 10494 2 10495 404 10496 9830 10497 9831 10498 136 10499 11 10500 138 10501 146 10502 147 10503 149 10504 151 10505 75 10506 123 10507 154 10508 9842 10509 9843 10510 136 10511 11 10512 138 10513 146 10514 2206 10515 2208 10516 2209 10517 2210 10518 2 10519 9853 10520 9854 10521 2 10522 9856 10523 830 10524 9858 10525 11 10526 9860 10527 9861 10528 9862 10529 2653 10530 9864 10531 9865 10532 338 10533 338 10534 341 10535 2 10536 404 10537 9871 10538 555 10539 9873 10540 9874 10541 9875 10542 9876 10543 2 10544 404 10545 9879 10546 9880 10547 9881 10548 9882 10549 2728 10550 9884 10551 555 10552 9886 10553 2 10554 9888 10555 9889 10556 9541 10557 2 10558 9892 10559 9893 10560 9894 10561 9895 10562 9896 10563 2820 10564 9898 10565 67 10566 9900 10567 9901 10568 9902 10569 9903 10570 9904 10571 9905 10572 2854 10573 9913 10574 2863 10575 9916 10576 917 10577 9933 10578 9934 10579 9935 10580 950 10581 143 10582 146 10583 9939 10584 9940 10585 9943 10586 2901 10587 2903 10588 2905 10589 75 10590 2907 10591 2908 10592 9950 10593 2 10594 264 10595 75 10596 3856 10597 2753 10598 9958 10599 1823 10600 146 10601 2011 10602 2 10603 9963 10604 2 10605 143 10606 146 10607 9967 10608 9968 10609 9969 10610 9970 10611 9971 10612 5125 10613 9973 10614 146 10615 9975 10616 2 10617 264 10618 9978 10619 2 10620 264 10621 9981 10622 9982 10623 9983 10624 9984 10625 9985 10626 9986 10627 1823 10628 146 10629 9989 10630 9990 10631 9991 10632 9992 10633 264 10634 9994 10635 9995 10636 2 10637 404 10638 9998 10639 5157 10640 10000 10641 10001 10642 10002 10643 10003 10644 10004 10645 10005 10646 2820 10647 10007 10648 67 10649 10009 10650 10010 10651 10011 10652 10012 10653 10013 10654 10014 10655 2854 10656 10022 10657 2863 10658 10025 10659 917 10660 10042 10661 10043 10662 10044 10663 950 10664 143 10665 146 10666 10048 10667 10049 10668 10052 10669 2901 10670 2903 10671 2905 10672 75 10673 2907 10674 2908 10675 10059 10676 2 10677 264 10678 75 10679 3856 10680 2753 10681 10067 10682 1823 10683 146 10684 2011 10685 2 10686 10072 10687 2 10688 143 10689 146 10690 10076 10691 10077 10692 10078 10693 10079 10694 10080 10695 5125 10696 10082 10697 146 10698 10084 10699 2 10700 264 10701 10087 10702 2 10703 264 10704 10090 10705 10091 10706 10092 10707 10093 10708 10094 10709 10095 10710 1823 10711 146 10712 10098 10713 10099 10714 10714 10715 10100 10716 10101 10717 264 10718 10103 10719 10104 10720 2 10721 404 10722 10107 10723 5157 10724 10109 10725 10110 10726 10726 10727 10727 10728 10728 10729 10729 10730 10730 10731 10731 10732 10732 10733 10733 10734 830 10735 10735 10736 10736 10737 10737 10738 10738 10739 10739 10740 10740 10741 10741 10742 10742 10743 10743 10744 10744 10745 10745 10746 10746 10747 10747 10748 10748 10749 10749 10750 10750 10751 10751 10752 10752 10753 10753 10754 10754 10755 10755 10756 10756 10757 10757 10758 10758 10759 10759 10760 10760 10761 10761 10762 10762 10763 10763 10764 10764 10765 10765 10766 10766 10767 10767 10768 10768 10769 10769 10770 10770 10771 10771 10772 10772 10773 10773 10774 10774 10775 10775 10776 10776 10777 10777 10778 10778 10779 10779 10780 10780 10781 10781 10782 10782 10783 10783 10784 10784 10785 10785 10786 10786 10787 10787 10788 10788 10789 10789 10790 10790 10791 10791 10792 10792 10793 10793 10794 10794 10795 10795 10796 10796 10797 10797 10798 10798 10799 10799 10800 10800 10801 10801 10802 10802 10803 10803 10804 10804 10805 10805 10806 10806 10807 10807 10808 10808 10809 10809 10810 10810 10811 10811 10812 10812 10813 10813 10814 11 10815 10815 10816 10816 10817 10817 10818 917 10819 10819 10820 10820 10821 950 10822 143 10823 146 10824 10824 10825 10825 10826 1456 10827 2 10828 2 10829 2 10830 10830 10831 2 10832 10832 10833 2 10834 2 10835 2 10836 2 10837 2 10838 2 10839 2 10840 2 10841 2 10842 10842 10843 10843 10844 10844 10845 10845 10846 10846 10847 10847 10848 10848 10849 10849 10850 10850 10851 10851 10852 10852 10853 10853 10854 10854 10855 22 10856 10856 10857 4 10858 10858 10859 87 10860 11 10861 10861 10862 97 10863 10863 10864 1517 10865 10865 10866 10866 10867 10867 10868 97 10869 830 10870 11 10871 10871 10872 917 10873 10873 10874 10874 10875 950 10876 143 10877 146 10878 10878 10879 10879 10880 10880 10881 830 10882 11 10883 10883 10884 10884 10885 10885 10886 338 10887 338 10888 341 10889 2 10890 404 10891 10891 10892 555 10893 10893 10894 10894 10895 10895 10896 10896 10897 2 10898 404 10899 10899 10900 10900 10901 10901 10902 10902 10903 10903 10904 2 10905 404 10906 10906 10907 10907 10908 10908 10909 136 10910 11 10911 138 10912 146 10913 147 10914 149 10915 151 10916 75 10917 123 10918 154 10919 10919 10920 10920 10921 136 10922 11 10923 138 10924 146 10925 2206 10926 2208 10927 2209 10928 2210 10929 2 10930 2 10931 3 10932 4 10933 2 10934 2 10935 2 10936 2 10937 11 10938 2 10939 2 10940 495 10941 2 10942 497 10943 22 10944 499 10945 500 10946 2 10947 11 10948 11 10949 504 10950 505 10951 506 10952 22 10953 2 10954 2 10955 2 10956 11 10957 512 10958 11 10959 514 10960 11 10961 516 10962 517 10963 518 10964 519 10965 60 10966 521 10967 2 10968 2 10969 2 10970 2 10971 67 10972 127 10973 528 10974 2 10975 2 10976 531 10977 2 10978 533 10979 534 10980 2 10981 83 10982 84 10983 87 10984 130 10985 11 10986 541 10987 542 10988 97 10989 544 10990 545 10991 10991 10992 102 10993 102 10994 549 10995 550 10996 551 10997 552 10998 553 10999 554 11000 556 11001 557 11002 558 11003 559 11004 97 11005 2 11006 562 11007 2 11008 2 11009 2 11010 2 11011 2 11012 2 11013 2 11014 2 11015 288 11016 2 11017 2 11018 2 11019 2 11020 2 11021 294 11022 2 11023 2 11024 2 11025 2 11026 2 11027 2 11028 2 11029 2 11030 2 11031 2 11032 2 11033 2 11034 2 11035 2 11036 2 11037 593 11038 594 11039 595 11040 596 11041 597 11042 598 11043 599 11044 600 11045 601 11046 602 11047 603 11048 604 11049 605 11050 606 11051 607 11052 4 11053 2 11054 610 11055 611 11056 612 11057 613 11058 614 11059 615 11060 616 11061 617 11062 618 11063 619 11064 620 11065 338 11066 338 11067 40 11068 341 11069 2 11070 404 11071 627 11072 555 11073 629 11074 630 11075 631 11076 632 11077 633 11078 2 11079 404 11080 636 11081 637 11082 40 11083 639 11084 264 11085 641 11086 642 11087 2 11088 404 11089 645 11090 646 11091 647 11092 648 11093 649 11094 650 11095 2 11096 713 11097 2 11098 10112 11099 10714 11100 10726 11101 10727 11102 10728 11103 10729 11104 10730 11105 10731 11106 10732 11107 10733 11108 830 11109 10735 11110 10736 11111 10737 11112 10738 11113 10739 11114 10740 11115 10741 11116 10742 11117 10743 11118 10744 11119 10745 11120 10746 11121 10747 11122 10748 11123 10749 11124 10750 11125 10751 11126 10752 11127 10753 11128 10754 11129 10755 11130 10756 11131 10757 11132 10758 11133 10759 11134 10760 11135 10761 11136 10762 11137 10763 11138 10764 11139 10765 11140 10766 11141 10767 11142 10768 11143 10769 11144 10770 11145 10771 11146 10772 11147 10773 11148 10774 11149 10775 11150 10776 11151 10777 11152 10778 11153 10779 11154 10780 11155 10781 11156 10782 11157 10783 11158 10784 11159 10785 11160 10786 11161 10787 11162 10788 11163 10789 11164 10790 11165 10791 11166 10792 11167 10793 11168 10794 11169 10795 11170 10796 11171 10797 11172 10798 11173 10799 11174 10800 11175 10801 11176 10802 11177 10803 11178 10804 11179 10805 11180 10806 11181 10807 11182 10808 11183 10809 11184 10810 11185 10811 11186 10812 11187 10813 11188 11 11189 10815 11190 10816 11191 10817 11192 917 11193 10819 11194 10820 11195 950 11196 143 11197 146 11198 10824 11199 10825 11200 1456 11201 2 11202 2 11203 2 11204 10830 11205 2 11206 10832 11207 2 11208 2 11209 2 11210 2 11211 2 11212 2 11213 2 11214 2 11215 2 11216 10842 11217 10843 11218 10844 11219 10845 11220 10846 11221 10847 11222 10848 11223 10849 11224 10850 11225 10851 11226 10852 11227 10853 11228 10854 11229 22 11230 10856 11231 4 11232 10858 11233 87 11234 11 11235 10861 11236 97 11237 10863 11238 1517 11239 10865 11240 10866 11241 10867 11242 97 11243 830 11244 11 11245 10871 11246 917 11247 10873 11248 10874 11249 950 11250 143 11251 146 11252 10878 11253 10879 11254 10880 11255 830 11256 11 11257 10883 11258 10884 11259 10885 11260 338 11261 338 11262 341 11263 2 11264 404 11265 10891 11266 555 11267 10893 11268 10894 11269 10895 11270 10896 11271 2 11272 404 11273 10899 11274 10900 11275 10901 11276 10902 11277 10903 11278 2 11279 404 11280 10906 11281 10908 11282 136 11283 11 11284 138 11285 146 11286 147 11287 149 11288 151 11289 75 11290 123 11291 154 11292 10919 11293 10920 11294 136 11295 11 11296 138 11297 146 11298 2206 11299 2208 11300 2209 11301 2210 11302 2 11303 10991 11304 11304 11305 11304 11306 830 11307 11307 11308 11 11309 11309 11310 11310 11311 11311 11312 2653 11313 11313 11314 11314 11315 338 11316 338 11317 341 11318 2 11319 404 11320 11320 11321 555 11322 11322 11323 11323 11324 11324 11325 11325 11326 2 11327 404 11328 11328 11329 11329 11330 11330 11331 11331 11332 2728 11333 11333 11334 555 11335 11335 11336 2 11337 11337 11338 11338 11339 2 11340 3 11341 4 11342 2 11343 2 11344 2 11345 2 11346 11 11347 2 11348 2 11349 11349 11350 2 11351 11351 11352 22 11353 11353 11354 11354 11355 2 11356 11 11357 11 11358 11358 11359 11359 11360 11360 11361 22 11362 2 11363 2 11364 2 11365 11 11366 11366 11367 11 11368 11368 11369 11 11370 11370 11371 11371 11372 11372 11373 11373 11374 60 11375 11375 11376 2 11377 2 11378 2 11379 2 11380 67 11381 127 11382 11382 11383 2 11384 2 11385 11385 11386 2 11387 11387 11388 11388 11389 2 11390 83 11391 84 11392 87 11393 130 11394 11 11395 11395 11396 11396 11397 97 11398 11398 11399 11399 11400 11400 11401 102 11402 102 11403 11403 11404 11404 11405 11405 11406 11406 11407 11407 11408 11408 11409 11409 11410 11410 11411 11411 11412 11412 11413 97 11414 2 11415 11415 11416 2 11417 2 11418 2 11419 2 11420 2 11421 2 11422 2 11423 2 11424 288 11425 2 11426 2 11427 2 11428 2 11429 2 11430 294 11431 2 11432 2 11433 2 11434 2 11435 2 11436 2 11437 2 11438 2 11439 2 11440 2 11441 2 11442 2 11443 2 11444 2 11445 2 11446 11446 11447 11447 11448 11448 11449 11449 11450 11450 11451 11451 11452 11452 11453 11453 11454 11454 11455 11455 11456 11456 11457 11457 11458 11458 11459 11459 11460 11460 11461 4 11462 2 11463 11463 11464 11464 11465 11465 11466 11466 11467 11467 11468 11468 11469 11469 11470 11470 11471 11471 11472 11472 11473 11473 11474 338 11475 338 11476 40 11477 341 11478 2 11479 404 11480 11480 11481 555 11482 11482 11483 11483 11484 11484 11485 11485 11486 11486 11487 2 11488 404 11489 11489 11490 11490 11491 40 11492 11492 11493 264 11494 11494 11495 11495 11496 2 11497 404 11498 11498 11499 11499 11500 11500 11501 11501 11502 11502 11503 11503 11504 2 11505 713 11506 2 11507 11507 11508 11508 11509 11509 11510 11510 11511 11511 11512 11512 11513 11513 11514 11514 11515 11515 11516 11516 11517 830 11518 11518 11519 11519 11520 11520 11521 11521 11522 11522 11523 11523 11524 11524 11525 11525 11526 11526 11527 11527 11528 11528 11529 11529 11530 11530 11531 11531 11532 11532 11533 11533 11534 11534 11535 11535 11536 11536 11537 11537 11538 11538 11539 11539 11540 11540 11541 11541 11542 11542 11543 11543 11544 11544 11545 11545 11546 11546 11547 11547 11548 11548 11549 11549 11550 11550 11551 11551 11552 11552 11553 11553 11554 11554 11555 11555 11556 11556 11557 11557 11558 11558 11559 11559 11560 11560 11561 11561 11562 11562 11563 11563 11564 11564 11565 11565 11566 11566 11567 11567 11568 11568 11569 11569 11570 11570 11571 11571 11572 11572 11573 11573 11574 11574 11575 11575 11576 11576 11577 11577 11578 11578 11579 11579 11580 11580 11581 11581 11582 11582 11583 11583 11584 11584 11585 11585 11586 11586 11587 11587 11588 11588 11589 11589 11590 11590 11591 11591 11592 11592 11593 11593 11594 11594 11595 11595 11596 11596 11597 11 11598 11598 11599 11599 11600 11600 11601 917 11602 11602 11603 11603 11604 950 11605 143 11606 146 11607 11607 11608 11608 11609 1456 11610 2 11611 2 11612 2 11613 11613 11614 2 11615 11615 11616 2 11617 2 11618 2 11619 2 11620 2 11621 2 11622 2 11623 2 11624 2 11625 11625 11626 11626 11627 11627 11628 11628 11629 11629 11630 11630 11631 11631 11632 11632 11633 11633 11634 11634 11635 11635 11636 11636 11637 11637 11638 22 11639 11639 11640 4 11641 11641 11642 87 11643 11 11644 11644 11645 97 11646 11646 11647 1517 11648 11648 11649 11649 11650 11650 11651 97 11652 830 11653 11 11654 11654 11655 917 11656 11656 11657 11657 11658 950 11659 143 11660 146 11661 11661 11662 11662 11663 11663 11664 830 11665 11 11666 11666 11667 11667 11668 11668 11669 338 11670 338 11671 341 11672 2 11673 404 11674 11674 11675 555 11676 11676 11677 11677 11678 11678 11679 11679 11680 2 11681 404 11682 11682 11683 11683 11684 11684 11685 11685 11686 11686 11687 2 11688 404 11689 11689 11690 11690 11691 136 11692 11 11693 138 11694 146 11695 147 11696 149 11697 151 11698 75 11699 123 11700 154 11701 11701 11702 11702 11703 136 11704 11 11705 138 11706 146 11707 2206 11708 2208 11709 2209 11710 2210 11711 2 11712 11712 11713 11713 11714 2 11715 11715 11716 830 11717 11717 11718 11 11719 11719 11720 11720 11721 11721 11722 2653 11723 11723 11724 11724 11725 338 11726 338 11727 341 11728 2 11729 404 11730 11730 11731 555 11732 11732 11733 11733 11734 11734 11735 11735 11736 2 11737 404 11738 11738 11739 11739 11740 11740 11741 11741 11742 2728 11743 11743 11744 555 11745 11745 11746 2 11747 11747 11748 11748 11749 11400 11750 2 11751 11751 11752 11752 11753 11753 11754 11754 11755 11755 11756 2820 11757 11757 11758 67 11759 11759 11760 11760 11761 11761 11762 11762 11763 11763 11764 11764 11765 2 11766 2854 11767 11767 11768 11768 11769 11769 11770 11770 11771 11771 11772 11772 11773 2863 11774 11774 11775 11775 11776 11776 11777 11777 11778 11778 11779 338 11780 338 11781 40 11782 341 11783 2 11784 404 11785 11785 11786 555 11787 11787 11788 11788 11789 11789 11790 11790 11791 917 11792 11792 11793 11793 11794 11794 11795 950 11796 143 11797 146 11798 11798 11799 11799 11800 40 11801 11801 11802 11802 11803 2901 11804 2903 11805 2905 11806 75 11807 2907 11808 2908 11809 11809 11810 264 11811 11811 11812 2 11813 264 11814 75 11815 3856 11816 2753 11817 11817 11818 1823 11819 146 11820 2011 11821 2 11822 11822 11823 2 11824 143 11825 146 11826 11826 11827 11827 11828 11828 11829 11829 11830 11830 11831 5125 11832 11832 11833 146 11834 11834 11835 2 11836 264 11837 11837 11838 2 11839 264 11840 11840 11841 11841 11842 11842 11843 11843 11844 11844 11845 11845 11846 1823 11847 146 11848 11848 11849 11849 11850 11850 11851 11851 11852 264 11853 11853 11854 11854 11855 2 11856 404 11857 11857 11858 5157 11859 11859 11860 11860 11861 11861 11862 11862 11863 11863 11864 11864 11865 2820 11866 11866 11867 67 11868 11868 11869 11869 11870 11870 11871 11871 11872 11872 11873 11873 11874 2 11875 2854 11876 11876 11877 11877 11878 11878 11879 11879 11880 11880 11881 11881 11882 2863 11883 11883 11884 11884 11885 11885 11886 11886 11887 11887 11888 338 11889 338 11890 40 11891 341 11892 2 11893 404 11894 11894 11895 555 11896 11896 11897 11897 11898 11898 11899 11899 11900 917 11901 11901 11902 11902 11903 11903 11904 950 11905 143 11906 146 11907 11907 11908 11908 11909 40 11910 11910 11911 11911 11912 2901 11913 2903 11914 2905 11915 75 11916 2907 11917 2908 11918 11918 11919 264 11920 11920 11921 2 11922 264 11923 75 11924 3856 11925 2753 11926 11926 11927 1823 11928 146 11929 2011 11930 2 11931 11931 11932 2 11933 143 11934 146 11935 11935 11936 11936 11937 11937 11938 11938 11939 11939 11940 5125 11941 11941 11942 146 11943 11943 11944 2 11945 264 11946 11946 11947 2 11948 264 11949 11949 11950 11950 11951 11951 11952 11952 11953 11953 11954 11954 11955 1823 11956 146 11957 11957 11958 11958 11959 11959 11960 11960 11961 264 11962 11962 11963 11963 11964 2 11965 404 11966 11966 11967 5157 11968 11968 11969 11969 11970 830 11971 11307 11972 11 11973 11309 11974 11310 11975 11311 11976 2653 11977 11313 11978 11314 11979 338 11980 338 11981 341 11982 2 11983 404 11984 11320 11985 555 11986 11322 11987 11323 11988 11324 11989 11325 11990 2 11991 404 11992 11328 11993 11329 11994 11330 11995 11331 11996 2728 11997 11333 11998 555 11999 11335 12000 2 12001 11337 12002 11338 12003 2 12004 3 12005 4 12006 2 12007 2 12008 2 12009 2 12010 11 12011 2 12012 2 12013 11349 12014 2 12015 11351 12016 22 12017 11353 12018 11354 12019 2 12020 11 12021 11 12022 11358 12023 11359 12024 11360 12025 22 12026 2 12027 2 12028 2 12029 11 12030 11366 12031 11 12032 11368 12033 11 12034 11370 12035 11371 12036 11372 12037 11373 12038 60 12039 11375 12040 2 12041 2 12042 2 12043 2 12044 67 12045 127 12046 11382 12047 2 12048 2 12049 11385 12050 2 12051 11387 12052 11388 12053 2 12054 83 12055 84 12056 87 12057 130 12058 11 12059 11395 12060 11396 12061 97 12062 11398 12063 11399 12064 11400 12065 102 12066 102 12067 11403 12068 11404 12069 11405 12070 11406 12071 11407 12072 11408 12073 11409 12074 11410 12075 11411 12076 11412 12077 97 12078 2 12079 11415 12080 2 12081 2 12082 2 12083 2 12084 2 12085 2 12086 2 12087 2 12088 288 12089 2 12090 2 12091 2 12092 2 12093 2 12094 294 12095 2 12096 2 12097 2 12098 2 12099 2 12100 2 12101 2 12102 2 12103 2 12104 2 12105 2 12106 2 12107 2 12108 2 12109 2 12110 11446 12111 11447 12112 11448 12113 11449 12114 11450 12115 11451 12116 11452 12117 11453 12118 11454 12119 11455 12120 11456 12121 11457 12122 11458 12123 11459 12124 11460 12125 4 12126 2 12127 11463 12128 11464 12129 11465 12130 11466 12131 11467 12132 11468 12133 11469 12134 11470 12135 11471 12136 11472 12137 11473 12138 338 12139 338 12140 40 12141 341 12142 2 12143 404 12144 11480 12145 555 12146 11482 12147 11483 12148 11484 12149 11485 12150 11486 12151 2 12152 404 12153 11489 12154 11490 12155 40 12156 11492 12157 264 12158 11494 12159 11495 12160 2 12161 404 12162 11498 12163 11499 12164 11500 12165 11501 12166 11502 12167 11503 12168 2 12169 713 12170 2 12171 11507 12172 11508 12173 11509 12174 11510 12175 11511 12176 11512 12177 11513 12178 11514 12179 11515 12180 11516 12181 830 12182 11518 12183 11519 12184 11520 12185 11521 12186 11522 12187 11523 12188 11524 12189 11525 12190 11526 12191 11527 12192 11528 12193 11529 12194 11530 12195 11531 12196 11532 12197 11533 12198 11534 12199 11535 12200 11536 12201 11537 12202 11538 12203 11539 12204 11540 12205 11541 12206 11542 12207 11543 12208 11544 12209 11545 12210 11546 12211 11547 12212 11548 12213 11549 12214 11550 12215 11551 12216 11552 12217 11553 12218 11554 12219 11555 12220 11556 12221 11557 12222 11558 12223 11559 12224 11560 12225 11561 12226 11562 12227 11563 12228 11564 12229 11565 12230 11566 12231 11567 12232 11568 12233 11569 12234 11570 12235 11571 12236 11572 12237 11573 12238 11574 12239 11575 12240 11576 12241 11577 12242 11578 12243 11579 12244 11580 12245 11581 12246 11582 12247 11583 12248 11584 12249 11585 12250 11586 12251 11587 12252 11588 12253 11589 12254 11590 12255 11591 12256 11592 12257 11593 12258 11594 12259 11595 12260 11596 12261 11 12262 11598 12263 11599 12264 11600 12265 917 12266 11602 12267 11603 12268 950 12269 143 12270 146 12271 11607 12272 11608 12273 1456 12274 2 12275 2 12276 2 12277 11613 12278 2 12279 11615 12280 2 12281 2 12282 2 12283 2 12284 2 12285 2 12286 2 12287 2 12288 2 12289 11625 12290 11626 12291 11627 12292 11628 12293 11629 12294 11630 12295 11631 12296 11632 12297 11633 12298 11634 12299 11635 12300 11636 12301 11637 12302 22 12303 11639 12304 4 12305 11641 12306 87 12307 11 12308 11644 12309 97 12310 11646 12311 1517 12312 11648 12313 11649 12314 11650 12315 97 12316 830 12317 11 12318 11654 12319 917 12320 11656 12321 11657 12322 950 12323 143 12324 146 12325 11661 12326 11662 12327 11663 12328 830 12329 11 12330 11666 12331 11667 12332 11668 12333 338 12334 338 12335 341 12336 2 12337 404 12338 11674 12339 555 12340 11676 12341 11677 12342 11678 12343 11679 12344 2 12345 404 12346 11682 12347 11683 12348 11684 12349 11685 12350 11686 12351 2 12352 404 12353 11689 12354 11690 12355 136 12356 11 12357 138 12358 146 12359 147 12360 149 12361 151 12362 75 12363 123 12364 154 12365 11701 12366 11702 12367 136 12368 11 12369 138 12370 146 12371 2206 12372 2208 12373 2209 12374 2210 12375 2 12376 11712 12377 11713 12378 2 12379 11715 12380 830 12381 11717 12382 11 12383 11719 12384 11720 12385 11721 12386 2653 12387 11723 12388 11724 12389 338 12390 338 12391 341 12392 2 12393 404 12394 11730 12395 555 12396 11732 12397 11733 12398 11734 12399 11735 12400 2 12401 404 12402 11738 12403 11739 12404 11740 12405 11741 12406 2728 12407 11743 12408 555 12409 11745 12410 2 12411 11747 12412 11748 12413 11400 12414 2 12415 11751 12416 11752 12417 11753 12418 11754 12419 11755 12420 2820 12421 11757 12422 67 12423 11759 12424 11760 12425 11761 12426 11762 12427 11763 12428 11764 12429 2854 12430 11772 12431 2863 12432 11775 12433 917 12434 11792 12435 11793 12436 11794 12437 950 12438 143 12439 146 12440 11798 12441 11799 12442 11802 12443 2901 12444 2903 12445 2905 12446 75 12447 2907 12448 2908 12449 11809 12450 2 12451 264 12452 75 12453 3856 12454 2753 12455 11817 12456 1823 12457 146 12458 2011 12459 2 12460 11822 12461 2 12462 143 12463 146 12464 11826 12465 11827 12466 11828 12467 11829 12468 11830 12469 5125 12470 11832 12471 146 12472 11834 12473 2 12474 264 12475 11837 12476 2 12477 264 12478 11840 12479 11841 12480 11842 12481 11843 12482 11844 12483 11845 12484 1823 12485 146 12486 11848 12487 11849 12488 11850 12489 11851 12490 264 12491 11853 12492 11854 12493 2 12494 404 12495 11857 12496 5157 12497 11859 12498 11860 12499 11861 12500 11862 12501 11863 12502 11864 12503 2820 12504 11866 12505 67 12506 11868 12507 11869 12508 11870 12509 11871 12510 11872 12511 11873 12512 2854 12513 11881 12514 2863 12515 11884 12516 917 12517 11901 12518 11902 12519 11903 12520 950 12521 143 12522 146 12523 11907 12524 11908 12525 11911 12526 2901 12527 2903 12528 2905 12529 75 12530 2907 12531 2908 12532 11918 12533 2 12534 264 12535 75 12536 3856 12537 2753 12538 11926 12539 1823 12540 146 12541 2011 12542 2 12543 11931 12544 2 12545 143 12546 146 12547 11935 12548 11936 12549 11937 12550 11938 12551 11939 12552 5125 12553 11941 12554 146 12555 11943 12556 2 12557 264 12558 11946 12559 2 12560 264 12561 11949 12562 11950 12563 11951 12564 11952 12565 11953 12566 11954 12567 1823 12568 146 12569 11957 12570 11958 12571 11959 12572 11960 12573 264 12574 11962 12575 11963 12576 2 12577 404 12578 11966 12579 5157 12580 11968 12581 11969 12582 12582 12583 12583 12584 12584 12585 12585 12586 2 12587 404 12588 12588 12589 12589 12590 40 12591 12591 12592 264 12593 12593 12594 12594 12595 2 12596 404 12597 12597 12598 12598 12599 12599 12600 12600 12601 12601 12602 12602 12603 2 12604 713 12605 2 12606 12606 12607 12607 12608 12608 12609 12609 12610 12610 12611 12611 12612 12612 12613 12613 12614 12614 12615 12615 12616 830 12617 12617 12618 12618 12619 12619 12620 12620 12621 12621 12622 12622 12623 12623 12624 12624 12625 12625 12626 12626 12627 12627 12628 12628 12629 12629 12630 12630 12631 12631 12632 12632 12633 12633 12634 12634 12635 12635 12636 12636 12637 12637 12638 12638 12639 12639 12640 12640 12641 12641 12642 12642 12643 12643 12644 12644 12645 12645 12646 12646 12647 12647 12648 12648 12649 12649 12650 12650 12651 12651 12652 12652 12653 12653 12654 12654 12655 12655 12656 12656 12657 12657 12658 12658 12659 12659 12660 12660 12661 12661 12662 12662 12663 12663 12664 12664 12665 12665 12666 12666 12667 12667 12668 12668 12669 12669 12670 12670 12671 12671 12672 12672 12673 12673 12674 12674 12675 12675 12676 12676 12677 12677 12678 12678 12679 12679 12680 12680 12681 12681 12682 12682 12683 12683 12684 12684 12685 12685 12686 12686 12687 12687 12688 12688 12689 12689 12690 12690 12691 12691 12692 12692 12693 12693 12694 12694 12695 12695 12696 11 12697 12697 12698 12698 12699 12699 12700 917 12701 12701 12702 12702 12703 950 12704 143 12705 146 12706 12706 12707 12707 12708 1456 12709 2 12710 2 12711 2 12712 12712 12713 2 12714 12714 12715 2 12716 2 12717 2 12718 2 12719 2 12720 2 12721 2 12722 2 12723 2 12724 12724 12725 12725 12726 12726 12727 12727 12728 12728 12729 12729 12730 12730 12731 12731 12732 12732 12733 12733 12734 12734 12735 12735 12736 12736 12737 22 12738 12738 12739 4 12740 12740 12741 87 12742 11 12743 12743 12744 97 12745 12745 12746 1517 12747 12747 12748 12748 12749 12749 12750 97 12751 830 12752 11 12753 12753 12754 917 12755 12755 12756 12756 12757 950 12758 143 12759 146 12760 12760 12761 12761 12762 12762 12763 830 12764 11 12765 12765 12766 12766 12767 12767 12768 338 12769 338 12770 341 12771 2 12772 404 12773 12773 12774 555 12775 12775 12776 12776 12777 12777 12778 12778 12779 2 12780 404 12781 12781 12782 12782 12783 12783 12784 12784 12785 12785 12786 2 12787 404 12788 12788 12789 12789 12790 136 12791 11 12792 138 12793 146 12794 147 12795 149 12796 151 12797 75 12798 123 12799 154 12800 12800 12801 12801 12802 136 12803 11 12804 138 12805 146 12806 2206 12807 2208 12808 2209 12809 2210 12810 2 12811 2 12812 3 12813 4 12814 2 12815 2 12816 2 12817 2 12818 11 12819 2 12820 2 12821 353 12822 2 12823 355 12824 22 12825 357 12826 358 12827 2 12828 11 12829 11 12830 362 12831 363 12832 364 12833 22 12834 2 12835 2 12836 2 12837 11 12838 370 12839 11 12840 372 12841 11 12842 374 12843 375 12844 376 12845 377 12846 60 12847 379 12848 2 12849 2 12850 2 12851 2 12852 67 12853 127 12854 386 12855 2 12856 2 12857 389 12858 2 12859 391 12860 392 12861 2 12862 83 12863 84 12864 87 12865 130 12866 11 12867 399 12868 400 12869 97 12870 402 12871 403 12872 12872 12873 102 12874 102 12875 407 12876 408 12877 409 12878 410 12879 411 12880 412 12881 413 12882 414 12883 415 12884 416 12885 97 12886 2 12887 419 12888 2 12889 2 12890 2 12891 2 12892 2 12893 2 12894 2 12895 2 12896 288 12897 2 12898 2 12899 2 12900 2 12901 2 12902 294 12903 2 12904 2 12905 2 12906 2 12907 2 12908 2 12909 2 12910 2 12911 2 12912 2 12913 2 12914 2 12915 2 12916 2 12917 2 12918 450 12919 451 12920 452 12921 453 12922 454 12923 455 12924 456 12925 457 12926 458 12927 459 12928 460 12929 461 12930 462 12931 463 12932 464 12933 4 12934 2 12935 467 12936 468 12937 469 12938 470 12939 471 12940 472 12941 473 12942 474 12943 475 12944 476 12945 477 12946 338 12947 338 12948 40 12949 341 12950 2 12951 404 12952 484 12953 555 12954 10907 12955 12582 12956 12583 12957 12584 12958 12585 12959 2 12960 404 12961 12588 12962 12589 12963 40 12964 12591 12965 264 12966 12593 12967 12594 12968 2 12969 404 12970 12597 12971 12598 12972 12599 12973 12600 12974 12601 12975 12602 12976 2 12977 713 12978 2 12979 12606 12980 12607 12981 12608 12982 12609 12983 12610 12984 12611 12985 12612 12986 12613 12987 12614 12988 12615 12989 830 12990 12617 12991 12618 12992 12619 12993 12620 12994 12621 12995 12622 12996 12623 12997 12624 12998 12625 12999 12626 13000 12627 13001 12628 13002 12629 13003 12630 13004 12631 13005 12632 13006 12633 13007 12634 13008 12635 13009 12636 13010 12637 13011 12638 13012 12639 13013 12640 13014 12641 13015 12642 13016 12643 13017 12644 13018 12645 13019 12646 13020 12647 13021 12648 13022 12649 13023 12650 13024 12651 13025 12652 13026 12653 13027 12654 13028 12655 13029 12656 13030 12657 13031 12658 13032 12659 13033 12660 13034 12661 13035 12662 13036 12663 13037 12664 13038 12665 13039 12666 13040 12667 13041 12668 13042 12669 13043 12670 13044 12671 13045 12672 13046 12673 13047 12674 13048 12675 13049 12676 13050 12677 13051 12678 13052 12679 13053 12680 13054 12681 13055 12682 13056 12683 13057 12684 13058 12685 13059 12686 13060 12687 13061 12688 13062 12689 13063 12690 13064 12691 13065 12692 13066 12693 13067 12694 13068 12695 13069 11 13070 12697 13071 12698 13072 12699 13073 917 13074 12701 13075 12702 13076 950 13077 143 13078 146 13079 12706 13080 12707 13081 1456 13082 2 13083 2 13084 2 13085 12712 13086 2 13087 12714 13088 2 13089 2 13090 2 13091 2 13092 2 13093 2 13094 2 13095 2 13096 2 13097 12724 13098 12725 13099 12726 13100 12727 13101 12728 13102 12729 13103 12730 13104 12731 13105 12732 13106 12733 13107 12734 13108 12735 13109 12736 13110 22 13111 12738 13112 4 13113 12740 13114 87 13115 11 13116 12743 13117 97 13118 12745 13119 1517 13120 12747 13121 12748 13122 12749 13123 97 13124 830 13125 11 13126 12753 13127 917 13128 12755 13129 12756 13130 950 13131 143 13132 146 13133 12760 13134 12761 13135 12762 13136 830 13137 11 13138 12765 13139 12766 13140 12767 13141 338 13142 338 13143 341 13144 2 13145 404 13146 12773 13147 555 13148 12775 13149 12776 13150 12777 13151 12778 13152 2 13153 404 13154 12781 13155 12782 13156 12783 13157 12784 13158 12785 13159 2 13160 404 13161 12788 13162 12789 13163 136 13164 11 13165 138 13166 146 13167 147 13168 149 13169 151 13170 75 13171 123 13172 154 13173 12800 13174 12801 13175 136 13176 11 13177 138 13178 146 13179 2206 13180 2208 13181 2209 13182 2210 13183 2 13184 12872 13185 13185 13186 13185 13187 830 13188 13188 13189 11 13190 13190 13191 13191 13192 13192 13193 2653 13194 13194 13195 13195 13196 338 13197 338 13198 341 13199 2 13200 404 13201 13201 13202 555 13203 13203 13204 13204 13205 13205 13206 13206 13207 2 13208 404 13209 13209 13210 13210 13211 13211 13212 13212 13213 2728 13214 13214 13215 555 13216 13216 13217 2 13218 13218 13219 13219 13220 2 13221 3 13222 4 13223 2 13224 2 13225 2 13226 2 13227 11 13228 2 13229 2 13230 13230 13231 2 13232 13232 13233 22 13234 13234 13235 13235 13236 2 13237 11 13238 11 13239 13239 13240 13240 13241 13241 13242 22 13243 2 13244 2 13245 2 13246 11 13247 13247 13248 11 13249 13249 13250 11 13251 13251 13252 13252 13253 13253 13254 13254 13255 60 13256 13256 13257 2 13258 2 13259 2 13260 2 13261 67 13262 127 13263 13263 13264 2 13265 2 13266 13266 13267 2 13268 13268 13269 13269 13270 2 13271 83 13272 84 13273 87 13274 130 13275 11 13276 13276 13277 13277 13278 97 13279 13279 13280 13280 13281 13281 13282 102 13283 102 13284 13284 13285 13285 13286 13286 13287 13287 13288 13288 13289 13289 13290 13290 13291 13291 13292 13292 13293 13293 13294 97 13295 2 13296 13296 13297 2 13298 2 13299 2 13300 2 13301 2 13302 2 13303 2 13304 2 13305 288 13306 2 13307 2 13308 2 13309 2 13310 2 13311 294 13312 2 13313 2 13314 2 13315 2 13316 2 13317 2 13318 2 13319 2 13320 2 13321 2 13322 2 13323 2 13324 2 13325 2 13326 2 13327 13327 13328 13328 13329 13329 13330 13330 13331 13331 13332 13332 13333 13333 13334 13334 13335 13335 13336 13336 13337 13337 13338 13338 13339 13339 13340 13340 13341 13341 13342 4 13343 2 13344 13344 13345 13345 13346 13346 13347 13347 13348 13348 13349 13349 13350 13350 13351 13351 13352 13352 13353 13353 13354 13354 13355 338 13356 338 13357 40 13358 341 13359 2 13360 404 13361 13361 13362 555 13363 13363 13364 13364 13365 13365 13366 13366 13367 13367 13368 2 13369 404 13370 13370 13371 13371 13372 40 13373 13373 13374 264 13375 13375 13376 13376 13377 2 13378 404 13379 13379 13380 13380 13381 13381 13382 13382 13383 13383 13384 13384 13385 2 13386 713 13387 2 13388 13388 13389 13389 13390 13390 13391 13391 13392 13392 13393 13393 13394 13394 13395 13395 13396 13396 13397 13397 13398 830 13399 13399 13400 13400 13401 13401 13402 13402 13403 13403 13404 13404 13405 13405 13406 13406 13407 13407 13408 13408 13409 13409 13410 13410 13411 13411 13412 13412 13413 13413 13414 13414 13415 13415 13416 13416 13417 13417 13418 13418 13419 13419 13420 13420 13421 13421 13422 13422 13423 13423 13424 13424 13425 13425 13426 13426 13427 13427 13428 13428 13429 13429 13430 13430 13431 13431 13432 13432 13433 13433 13434 13434 13435 13435 13436 13436 13437 13437 13438 13438 13439 13439 13440 13440 13441 13441 13442 13442 13443 13443 13444 13444 13445 13445 13446 13446 13447 13447 13448 13448 13449 13449 13450 13450 13451 13451 13452 13452 13453 13453 13454 13454 13455 13455 13456 13456 13457 13457 13458 13458 13459 13459 13460 13460 13461 13461 13462 13462 13463 13463 13464 13464 13465 13465 13466 13466 13467 13467 13468 13468 13469 13469 13470 13470 13471 13471 13472 13472 13473 13473 13474 13474 13475 13475 13476 13476 13477 13477 13478 11 13479 13479 13480 13480 13481 13481 13482 917 13483 13483 13484 13484 13485 950 13486 143 13487 146 13488 13488 13489 13489 13490 1456 13491 2 13492 2 13493 2 13494 13494 13495 2 13496 13496 13497 2 13498 2 13499 2 13500 2 13501 2 13502 2 13503 2 13504 2 13505 2 13506 13506 13507 13507 13508 13508 13509 13509 13510 13510 13511 13511 13512 13512 13513 13513 13514 13514 13515 13515 13516 13516 13517 13517 13518 13518 13519 22 13520 13520 13521 4 13522 13522 13523 87 13524 11 13525 13525 13526 97 13527 13527 13528 1517 13529 13529 13530 13530 13531 13531 13532 97 13533 830 13534 11 13535 13535 13536 917 13537 13537 13538 13538 13539 950 13540 143 13541 146 13542 13542 13543 13543 13544 13544 13545 830 13546 11 13547 13547 13548 13548 13549 13549 13550 338 13551 338 13552 341 13553 2 13554 404 13555 13555 13556 555 13557 13557 13558 13558 13559 13559 13560 13560 13561 2 13562 404 13563 13563 13564 13564 13565 13565 13566 13566 13567 13567 13568 2 13569 404 13570 13570 13571 13571 13572 136 13573 11 13574 138 13575 146 13576 147 13577 149 13578 151 13579 75 13580 123 13581 154 13582 13582 13583 13583 13584 136 13585 11 13586 138 13587 146 13588 2206 13589 2208 13590 2209 13591 2210 13592 2 13593 13593 13594 13594 13595 2 13596 13596 13597 830 13598 13598 13599 11 13600 13600 13601 13601 13602 13602 13603 2653 13604 13604 13605 13605 13606 338 13607 338 13608 341 13609 2 13610 404 13611 13611 13612 555 13613 13613 13614 13614 13615 13615 13616 13616 13617 2 13618 404 13619 13619 13620 13620 13621 13621 13622 13622 13623 2728 13624 13624 13625 555 13626 13626 13627 2 13628 13628 13629 13629 13630 13281 13631 2 13632 13632 13633 13633 13634 13634 13635 13635 13636 13636 13637 2820 13638 13638 13639 67 13640 13640 13641 13641 13642 13642 13643 13643 13644 13644 13645 13645 13646 2 13647 2854 13648 13648 13649 13649 13650 13650 13651 13651 13652 13652 13653 13653 13654 2863 13655 13655 13656 13656 13657 13657 13658 13658 13659 13659 13660 338 13661 338 13662 40 13663 341 13664 2 13665 404 13666 13666 13667 555 13668 13668 13669 13669 13670 13670 13671 13671 13672 917 13673 13673 13674 13674 13675 13675 13676 950 13677 143 13678 146 13679 13679 13680 13680 13681 40 13682 13682 13683 13683 13684 2901 13685 2903 13686 2905 13687 75 13688 2907 13689 2908 13690 13690 13691 264 13692 13692 13693 2 13694 264 13695 75 13696 3856 13697 2753 13698 13698 13699 1823 13700 146 13701 2011 13702 2 13703 13703 13704 2 13705 143 13706 146 13707 13707 13708 13708 13709 13709 13710 13710 13711 13711 13712 5125 13713 13713 13714 146 13715 13715 13716 2 13717 264 13718 13718 13719 2 13720 264 13721 13721 13722 13722 13723 13723 13724 13724 13725 13725 13726 13726 13727 1823 13728 146 13729 13729 13730 13730 13731 13731 13732 13732 13733 264 13734 13734 13735 13735 13736 2 13737 404 13738 13738 13739 5157 13740 13740 13741 13741 13742 13742 13743 13743 13744 13744 13745 13745 13746 2820 13747 13747 13748 67 13749 13749 13750 13750 13751 13751 13752 13752 13753 13753 13754 13754 13755 2 13756 2854 13757 13757 13758 13758 13759 13759 13760 13760 13761 13761 13762 13762 13763 2863 13764 13764 13765 13765 13766 13766 13767 13767 13768 13768 13769 338 13770 338 13771 40 13772 341 13773 2 13774 404 13775 13775 13776 555 13777 13777 13778 13778 13779 13779 13780 13780 13781 917 13782 13782 13783 13783 13784 13784 13785 950 13786 143 13787 146 13788 13788 13789 13789 13790 40 13791 13791 13792 13792 13793 2901 13794 2903 13795 2905 13796 75 13797 2907 13798 2908 13799 13799 13800 264 13801 13801 13802 2 13803 264 13804 75 13805 3856 13806 2753 13807 13807 13808 1823 13809 146 13810 2011 13811 2 13812 13812 13813 2 13814 143 13815 146 13816 13816 13817 13817 13818 13818 13819 13819 13820 13820 13821 5125 13822 13822 13823 146 13824 13824 13825 2 13826 264 13827 13827 13828 2 13829 264 13830 13830 13831 13831 13832 13832 13833 13833 13834 13834 13835 13835 13836 1823 13837 146 13838 13838 13839 13839 13840 13840 13841 13841 13842 264 13843 13843 13844 13844 13845 2 13846 404 13847 13847 13848 5157 13849 13849 13850 13850 13851 13851 13852 830 13853 13188 13854 11 13855 13190 13856 13191 13857 13192 13858 2653 13859 13194 13860 13195 13861 338 13862 338 13863 341 13864 2 13865 404 13866 13201 13867 555 13868 13203 13869 13204 13870 13205 13871 13206 13872 2 13873 404 13874 13209 13875 13210 13876 13211 13877 13212 13878 2728 13879 13214 13880 555 13881 13216 13882 2 13883 13218 13884 13219 13885 2 13886 3 13887 4 13888 2 13889 2 13890 2 13891 2 13892 11 13893 2 13894 2 13895 13230 13896 2 13897 13232 13898 22 13899 13234 13900 13235 13901 2 13902 11 13903 11 13904 13239 13905 13240 13906 13241 13907 22 13908 2 13909 2 13910 2 13911 11 13912 13247 13913 11 13914 13249 13915 11 13916 13251 13917 13252 13918 13253 13919 13254 13920 60 13921 13256 13922 2 13923 2 13924 2 13925 2 13926 67 13927 127 13928 13263 13929 2 13930 2 13931 13266 13932 2 13933 13268 13934 13269 13935 2 13936 83 13937 84 13938 87 13939 130 13940 11 13941 13276 13942 13277 13943 97 13944 13279 13945 13280 13946 13281 13947 102 13948 102 13949 13284 13950 13285 13951 13286 13952 13287 13953 13288 13954 13289 13955 13290 13956 13291 13957 13292 13958 13293 13959 97 13960 2 13961 13296 13962 2 13963 2 13964 2 13965 2 13966 2 13967 2 13968 2 13969 2 13970 288 13971 2 13972 2 13973 2 13974 2 13975 2 13976 294 13977 2 13978 2 13979 2 13980 2 13981 2 13982 2 13983 2 13984 2 13985 2 13986 2 13987 2 13988 2 13989 2 13990 2 13991 2 13992 13327 13993 13328 13994 13329 13995 13330 13996 13331 13997 13332 13998 13333 13999 13334 14000 13335 14001 13336 14002 13337 14003 13338 14004 13339 14005 13340 14006 13341 14007 4 14008 2 14009 13344 14010 13345 14011 13346 14012 13347 14013 13348 14014 13349 14015 13350 14016 13351 14017 13352 14018 13353 14019 13354 14020 338 14021 338 14022 40 14023 341 14024 2 14025 404 14026 13361 14027 555 14028 13363 14029 13364 14030 13365 14031 13366 14032 13367 14033 2 14034 404 14035 13370 14036 13371 14037 40 14038 13373 14039 264 14040 13375 14041 13376 14042 2 14043 404 14044 13379 14045 13380 14046 13381 14047 13382 14048 13383 14049 13384 14050 2 14051 713 14052 2 14053 13388 14054 13389 14055 13390 14056 13391 14057 13392 14058 13393 14059 13394 14060 13395 14061 13396 14062 13397 14063 830 14064 13399 14065 13400 14066 13401 14067 13402 14068 13403 14069 13404 14070 13405 14071 13406 14072 13407 14073 13408 14074 13409 14075 13410 14076 13411 14077 13412 14078 13413 14079 13414 14080 13415 14081 13416 14082 13417 14083 13418 14084 13419 14085 13420 14086 13421 14087 13422 14088 13423 14089 13424 14090 13425 14091 13426 14092 13427 14093 13428 14094 13429 14095 13430 14096 13431 14097 13432 14098 13433 14099 13434 14100 13435 14101 13436 14102 13437 14103 13438 14104 13439 14105 13440 14106 13441 14107 13442 14108 13443 14109 13444 14110 13445 14111 13446 14112 13447 14113 13448 14114 13449 14115 13450 14116 13451 14117 13452 14118 13453 14119 13454 14120 13455 14121 13456 14122 13457 14123 13458 14124 13459 14125 13460 14126 13461 14127 13462 14128 13463 14129 13464 14130 13465 14131 13466 14132 13467 14133 13468 14134 13469 14135 13470 14136 13471 14137 13472 14138 13473 14139 13474 14140 13475 14141 13476 14142 13477 14143 11 14144 13479 14145 13480 14146 13481 14147 917 14148 13483 14149 13484 14150 950 14151 143 14152 146 14153 13488 14154 13489 14155 1456 14156 2 14157 2 14158 2 14159 13494 14160 2 14161 13496 14162 2 14163 2 14164 2 14165 2 14166 2 14167 2 14168 2 14169 2 14170 2 14171 13506 14172 13507 14173 13508 14174 13509 14175 13510 14176 13511 14177 13512 14178 13513 14179 13514 14180 13515 14181 13516 14182 13517 14183 13518 14184 22 14185 13520 14186 4 14187 13522 14188 87 14189 11 14190 13525 14191 97 14192 13527 14193 1517 14194 13529 14195 13530 14196 13531 14197 97 14198 830 14199 11 14200 13535 14201 917 14202 13537 14203 13538 14204 950 14205 143 14206 146 14207 13542 14208 13543 14209 13544 14210 830 14211 11 14212 13547 14213 13548 14214 13549 14215 338 14216 338 14217 341 14218 2 14219 404 14220 13555 14221 555 14222 13557 14223 13558 14224 13559 14225 13560 14226 2 14227 404 14228 13563 14229 13564 14230 13565 14231 13566 14232 13567 14233 2 14234 404 14235 13570 14236 13571 14237 136 14238 11 14239 138 14240 146 14241 147 14242 149 14243 151 14244 75 14245 123 14246 154 14247 13582 14248 13583 14249 136 14250 11 14251 138 14252 146 14253 2206 14254 2208 14255 2209 14256 2210 14257 2 14258 13593 14259 13594 14260 2 14261 13596 14262 830 14263 13598 14264 11 14265 13600 14266 13601 14267 13602 14268 2653 14269 13604 14270 13605 14271 338 14272 338 14273 341 14274 2 14275 404 14276 13611 14277 555 14278 13613 14279 13614 14280 13615 14281 13616 14282 2 14283 404 14284 13619 14285 13620 14286 13621 14287 13622 14288 2728 14289 13624 14290 555 14291 13626 14292 2 14293 13628 14294 13629 14295 13281 14296 2 14297 13632 14298 13633 14299 13634 14300 13635 14301 13636 14302 2820 14303 13638 14304 67 14305 13640 14306 13641 14307 13642 14308 13643 14309 13644 14310 13645 14311 2854 14312 13653 14313 2863 14314 13656 14315 917 14316 13673 14317 13674 14318 13675 14319 950 14320 143 14321 146 14322 13679 14323 13680 14324 13683 14325 2901 14326 2903 14327 2905 14328 75 14329 2907 14330 2908 14331 13690 14332 2 14333 264 14334 75 14335 3856 14336 2753 14337 13698 14338 1823 14339 146 14340 2011 14341 2 14342 13703 14343 2 14344 143 14345 146 14346 13707 14347 13708 14348 13709 14349 13710 14350 13711 14351 5125 14352 13713 14353 146 14354 13715 14355 2 14356 264 14357 13718 14358 2 14359 264 14360 13721 14361 13722 14362 13723 14363 13724 14364 13725 14365 13726 14366 1823 14367 146 14368 13729 14369 13730 14370 13731 14371 13732 14372 264 14373 13734 14374 13735 14375 2 14376 404 14377 13738 14378 5157 14379 13740 14380 13741 14381 13742 14382 13743 14383 13744 14384 13745 14385 2820 14386 13747 14387 67 14388 13749 14389 13750 14390 13751 14391 13752 14392 13753 14393 13754 14394 2854 14395 13762 14396 2863 14397 13765 14398 917 14399 13782 14400 13783 14401 13784 14402 950 14403 143 14404 146 14405 13788 14406 13789 14407 13792 14408 2901 14409 2903 14410 2905 14411 75 14412 2907 14413 2908 14414 13799 14415 2 14416 264 14417 75 14418 3856 14419 2753 14420 13807 14421 1823 14422 146 14423 2011 14424 2 14425 13812 14426 2 14427 143 14428 146 14429 13816 14430 13817 14431 13818 14432 13819 14433 13820 14434 5125 14435 13822 14436 146 14437 13824 14438 2 14439 264 14440 13827 14441 2 14442 264 14443 13830 14444 13831 14445 13832 14446 13833 14447 13834 14448 13835 14449 1823 14450 146 14451 13838 14452 13839 14453 13840 14454 13841 14455 264 14456 13843 14457 13844 14458 2 14459 404 14460 13847 14461 5157 14462 13849 14463 13850 14464 555 14465 14465 14466 14466 14467 14467 14468 14468 14469 14469 14470 2 14471 404 14472 14472 14473 14473 14474 40 14475 14475 14476 264 14477 14477 14478 14478 14479 2 14480 404 14481 14481 14482 14482 14483 14483 14484 14484 14485 14485 14486 14486 14487 2 14488 713 14489 2 14490 14490 14491 14491 14492 14492 14493 14493 14494 14494 14495 14495 14496 14496 14497 14497 14498 14498 14499 14499 14500 830 14501 14501 14502 14502 14503 14503 14504 14504 14505 14505 14506 14506 14507 14507 14508 14508 14509 14509 14510 14510 14511 14511 14512 14512 14513 14513 14514 14514 14515 14515 14516 14516 14517 14517 14518 14518 14519 14519 14520 14520 14521 14521 14522 14522 14523 14523 14524 14524 14525 14525 14526 14526 14527 14527 14528 14528 14529 14529 14530 14530 14531 14531 14532 14532 14533 14533 14534 14534 14535 14535 14536 14536 14537 14537 14538 14538 14539 14539 14540 14540 14541 14541 14542 14542 14543 14543 14544 14544 14545 14545 14546 14546 14547 14547 14548 14548 14549 14549 14550 14550 14551 14551 14552 14552 14553 14553 14554 14554 14555 14555 14556 14556 14557 14557 14558 14558 14559 14559 14560 14560 14561 14561 14562 14562 14563 14563 14564 14564 14565 14565 14566 14566 14567 14567 14568 14568 14569 14569 14570 14570 14571 14571 14572 14572 14573 14573 14574 14574 14575 14575 14576 14576 14577 14577 14578 14578 14579 14579 14580 11 14581 14581 14582 14582 14583 14583 14584 917 14585 14585 14586 14586 14587 950 14588 143 14589 146 14590 14590 14591 14591 14592 1456 14593 2 14594 2 14595 2 14596 14596 14597 2 14598 14598 14599 2 14600 2 14601 2 14602 2 14603 2 14604 2 14605 2 14606 2 14607 2 14608 14608 14609 14609 14610 14610 14611 14611 14612 14612 14613 14613 14614 14614 14615 14615 14616 14616 14617 14617 14618 14618 14619 14619 14620 14620 14621 22 14622 14622 14623 4 14624 14624 14625 87 14626 11 14627 14627 14628 97 14629 14629 14630 1517 14631 14631 14632 14632 14633 14633 14634 97 14635 830 14636 11 14637 14637 14638 917 14639 14639 14640 14640 14641 950 14642 143 14643 146 14644 14644 14645 14645 14646 14646 14647 830 14648 11 14649 14649 14650 14650 14651 14651 14652 338 14653 338 14654 341 14655 2 14656 404 14657 14657 14658 555 14659 14659 14660 14660 14661 14661 14662 14662 14663 2 14664 404 14665 14665 14666 14666 14667 14667 14668 14668 14669 14669 14670 2 14671 404 14672 14672 14673 14673 14674 136 14675 11 14676 138 14677 146 14678 147 14679 149 14680 151 14681 75 14682 123 14683 154 14684 14684 14685 14685 14686 136 14687 11 14688 138 14689 146 14690 2206 14691 2208 14692 2209 14693 2210 14694 2 14695 2 14696 3 14697 4 14698 2 14699 2 14700 2 14701 2 14702 11 14703 2 14704 2 14705 213 14706 2 14707 215 14708 22 14709 217 14710 218 14711 2 14712 11 14713 11 14714 222 14715 223 14716 224 14717 22 14718 2 14719 2 14720 2 14721 11 14722 230 14723 11 14724 232 14725 11 14726 234 14727 235 14728 236 14729 237 14730 60 14731 239 14732 2 14733 2 14734 2 14735 2 14736 67 14737 127 14738 246 14739 2 14740 2 14741 249 14742 2 14743 251 14744 252 14745 2 14746 83 14747 84 14748 87 14749 130 14750 11 14751 259 14752 260 14753 97 14754 262 14755 263 14756 14756 14757 102 14758 102 14759 267 14760 268 14761 269 14762 270 14763 271 14764 272 14765 273 14766 274 14767 275 14768 276 14769 97 14770 2 14771 279 14772 2 14773 2 14774 2 14775 2 14776 2 14777 2 14778 2 14779 2 14780 288 14781 2 14782 2 14783 2 14784 2 14785 2 14786 294 14787 2 14788 2 14789 2 14790 2 14791 2 14792 2 14793 2 14794 2 14795 2 14796 2 14797 2 14798 2 14799 2 14800 2 14801 2 14802 310 14803 311 14804 312 14805 313 14806 314 14807 315 14808 316 14809 317 14810 318 14811 319 14812 320 14813 321 14814 322 14815 323 14816 324 14817 4 14818 2 14819 327 14820 328 14821 329 14822 330 14823 331 14824 332 14825 333 14826 334 14827 335 14828 336 14829 337 14830 338 14831 338 14832 40 14833 341 14834 2 14835 404 14836 13851 14837 555 14838 14465 14839 14466 14840 14467 14841 14468 14842 14469 14843 2 14844 404 14845 14472 14846 14473 14847 40 14848 14475 14849 264 14850 14477 14851 14478 14852 2 14853 404 14854 14481 14855 14482 14856 14483 14857 14484 14858 14485 14859 14486 14860 2 14861 713 14862 2 14863 14490 14864 14491 14865 14492 14866 14493 14867 14494 14868 14495 14869 14496 14870 14497 14871 14498 14872 14499 14873 830 14874 14501 14875 14502 14876 14503 14877 14504 14878 14505 14879 14506 14880 14507 14881 14508 14882 14509 14883 14510 14884 14511 14885 14512 14886 14513 14887 14514 14888 14515 14889 14516 14890 14517 14891 14518 14892 14519 14893 14520 14894 14521 14895 14522 14896 14523 14897 14524 14898 14525 14899 14526 14900 14527 14901 14528 14902 14529 14903 14530 14904 14531 14905 14532 14906 14533 14907 14534 14908 14535 14909 14536 14910 14537 14911 14538 14912 14539 14913 14540 14914 14541 14915 14542 14916 14543 14917 14544 14918 14545 14919 14546 14920 14547 14921 14548 14922 14549 14923 14550 14924 14551 14925 14552 14926 14553 14927 14554 14928 14555 14929 14556 14930 14557 14931 14558 14932 14559 14933 14560 14934 14561 14935 14562 14936 14563 14937 14564 14938 14565 14939 14566 14940 14567 14941 14568 14942 14569 14943 14570 14944 14571 14945 14572 14946 14573 14947 14574 14948 14575 14949 14576 14950 14577 14951 14578 14952 14579 14953 11 14954 14581 14955 14582 14956 14583 14957 917 14958 14585 14959 14586 14960 950 14961 143 14962 146 14963 14590 14964 14591 14965 1456 14966 2 14967 2 14968 2 14969 14596 14970 2 14971 14598 14972 2 14973 2 14974 2 14975 2 14976 2 14977 2 14978 2 14979 2 14980 2 14981 14608 14982 14609 14983 14610 14984 14611 14985 14612 14986 14613 14987 14614 14988 14615 14989 14616 14990 14617 14991 14618 14992 14619 14993 14620 14994 22 14995 14622 14996 4 14997 14624 14998 87 14999 11 15000 14627 15001 97 15002 14629 15003 1517 15004 14631 15005 14632 15006 14633 15007 97 15008 830 15009 11 15010 14637 15011 917 15012 14639 15013 14640 15014 950 15015 143 15016 146 15017 14644 15018 14645 15019 14646 15020 830 15021 11 15022 14649 15023 14650 15024 14651 15025 338 15026 338 15027 341 15028 2 15029 404 15030 14657 15031 555 15032 14659 15033 14660 15034 14661 15035 14662 15036 2 15037 404 15038 14665 15039 14666 15040 14667 15041 14668 15042 14669 15043 2 15044 404 15045 14672 15046 14673 15047 136 15048 11 15049 138 15050 146 15051 147 15052 149 15053 151 15054 75 15055 123 15056 154 15057 14684 15058 14685 15059 136 15060 11 15061 138 15062 146 15063 2206 15064 2208 15065 2209 15066 2210 15067 2 15068 14756 15069 15069 15070 15069 15071 830 15072 15072 15073 11 15074 15074 15075 15075 15076 15076 15077 2653 15078 15078 15079 15079 15080 338 15081 338 15082 341 15083 2 15084 404 15085 15085 15086 555 15087 15087 15088 15088 15089 15089 15090 15090 15091 2 15092 404 15093 15093 15094 15094 15095 15095 15096 15096 15097 2728 15098 15098 15099 555 15100 15100 15101 2 15102 15102 15103 15103 15104 2 15105 3 15106 4 15107 2 15108 2 15109 2 15110 2 15111 11 15112 2 15113 2 15114 15114 15115 2 15116 15116 15117 22 15118 15118 15119 15119 15120 2 15121 11 15122 11 15123 15123 15124 15124 15125 15125 15126 22 15127 2 15128 2 15129 2 15130 11 15131 15131 15132 11 15133 15133 15134 11 15135 15135 15136 15136 15137 15137 15138 15138 15139 60 15140 15140 15141 2 15142 2 15143 2 15144 2 15145 67 15146 127 15147 15147 15148 2 15149 2 15150 15150 15151 2 15152 15152 15153 15153 15154 2 15155 83 15156 84 15157 87 15158 130 15159 11 15160 15160 15161 15161 15162 97 15163 15163 15164 15164 15165 15165 15166 102 15167 102 15168 15168 15169 15169 15170 15170 15171 15171 15172 15172 15173 15173 15174 15174 15175 15175 15176 15176 15177 15177 15178 97 15179 2 15180 15180 15181 2 15182 2 15183 2 15184 2 15185 2 15186 2 15187 2 15188 2 15189 288 15190 2 15191 2 15192 2 15193 2 15194 2 15195 294 15196 2 15197 2 15198 2 15199 2 15200 2 15201 2 15202 2 15203 2 15204 2 15205 2 15206 2 15207 2 15208 2 15209 2 15210 2 15211 15211 15212 15212 15213 15213 15214 15214 15215 15215 15216 15216 15217 15217 15218 15218 15219 15219 15220 15220 15221 15221 15222 15222 15223 15223 15224 15224 15225 15225 15226 4 15227 2 15228 15228 15229 15229 15230 15230 15231 15231 15232 15232 15233 15233 15234 15234 15235 15235 15236 15236 15237 15237 15238 15238 15239 338 15240 338 15241 40 15242 341 15243 2 15244 404 15245 15245 15246 555 15247 15247 15248 15248 15249 15249 15250 15250 15251 15251 15252 2 15253 404 15254 15254 15255 15255 15256 40 15257 15257 15258 264 15259 15259 15260 15260 15261 2 15262 404 15263 15263 15264 15264 15265 15265 15266 15266 15267 15267 15268 15268 15269 2 15270 713 15271 2 15272 15272 15273 15273 15274 15274 15275 15275 15276 15276 15277 15277 15278 15278 15279 15279 15280 15280 15281 15281 15282 830 15283 15283 15284 15284 15285 15285 15286 15286 15287 15287 15288 15288 15289 15289 15290 15290 15291 15291 15292 15292 15293 15293 15294 15294 15295 15295 15296 15296 15297 15297 15298 15298 15299 15299 15300 15300 15301 15301 15302 15302 15303 15303 15304 15304 15305 15305 15306 15306 15307 15307 15308 15308 15309 15309 15310 15310 15311 15311 15312 15312 15313 15313 15314 15314 15315 15315 15316 15316 15317 15317 15318 15318 15319 15319 15320 15320 15321 15321 15322 15322 15323 15323 15324 15324 15325 15325 15326 15326 15327 15327 15328 15328 15329 15329 15330 15330 15331 15331 15332 15332 15333 15333 15334 15334 15335 15335 15336 15336 15337 15337 15338 15338 15339 15339 15340 15340 15341 15341 15342 15342 15343 15343 15344 15344 15345 15345 15346 15346 15347 15347 15348 15348 15349 15349 15350 15350 15351 15351 15352 15352 15353 15353 15354 15354 15355 15355 15356 15356 15357 15357 15358 15358 15359 15359 15360 15360 15361 15361 15362 11 15363 15363 15364 15364 15365 15365 15366 917 15367 15367 15368 15368 15369 950 15370 143 15371 146 15372 15372 15373 15373 15374 1456 15375 2 15376 2 15377 2 15378 15378 15379 2 15380 15380 15381 2 15382 2 15383 2 15384 2 15385 2 15386 2 15387 2 15388 2 15389 2 15390 15390 15391 15391 15392 15392 15393 15393 15394 15394 15395 15395 15396 15396 15397 15397 15398 15398 15399 15399 15400 15400 15401 15401 15402 15402 15403 22 15404 15404 15405 4 15406 15406 15407 87 15408 11 15409 15409 15410 97 15411 15411 15412 1517 15413 15413 15414 15414 15415 15415 15416 97 15417 830 15418 11 15419 15419 15420 917 15421 15421 15422 15422 15423 950 15424 143 15425 146 15426 15426 15427 15427 15428 15428 15429 830 15430 11 15431 15431 15432 15432 15433 15433 15434 338 15435 338 15436 341 15437 2 15438 404 15439 15439 15440 555 15441 15441 15442 15442 15443 15443 15444 15444 15445 2 15446 404 15447 15447 15448 15448 15449 15449 15450 15450 15451 15451 15452 2 15453 404 15454 15454 15455 15455 15456 136 15457 11 15458 138 15459 146 15460 147 15461 149 15462 151 15463 75 15464 123 15465 154 15466 15466 15467 15467 15468 136 15469 11 15470 138 15471 146 15472 2206 15473 2208 15474 2209 15475 2210 15476 2 15477 15477 15478 15478 15479 2 15480 15480 15481 830 15482 15482 15483 11 15484 15484 15485 15485 15486 15486 15487 2653 15488 15488 15489 15489 15490 338 15491 338 15492 341 15493 2 15494 404 15495 15495 15496 555 15497 15497 15498 15498 15499 15499 15500 15500 15501 2 15502 404 15503 15503 15504 15504 15505 15505 15506 15506 15507 2728 15508 15508 15509 555 15510 15510 15511 2 15512 15512 15513 15513 15514 15165 15515 2 15516 15516 15517 15517 15518 15518 15519 15519 15520 15520 15521 2820 15522 15522 15523 67 15524 15524 15525 15525 15526 15526 15527 15527 15528 15528 15529 15529 15530 2 15531 2854 15532 15532 15533 15533 15534 15534 15535 15535 15536 15536 15537 15537 15538 2863 15539 15539 15540 15540 15541 15541 15542 15542 15543 15543 15544 338 15545 338 15546 40 15547 341 15548 2 15549 404 15550 15550 15551 555 15552 15552 15553 15553 15554 15554 15555 15555 15556 917 15557 15557 15558 15558 15559 15559 15560 950 15561 143 15562 146 15563 15563 15564 15564 15565 40 15566 15566 15567 15567 15568 2901 15569 2903 15570 2905 15571 75 15572 2907 15573 2908 15574 15574 15575 264 15576 15576 15577 2 15578 264 15579 75 15580 3856 15581 2753 15582 15582 15583 1823 15584 146 15585 2011 15586 2 15587 15587 15588 2 15589 143 15590 146 15591 15591 15592 15592 15593 15593 15594 15594 15595 15595 15596 5125 15597 15597 15598 146 15599 15599 15600 2 15601 264 15602 15602 15603 2 15604 264 15605 15605 15606 15606 15607 15607 15608 15608 15609 15609 15610 15610 15611 1823 15612 146 15613 15613 15614 15614 15615 15615 15616 15616 15617 264 15618 15618 15619 15619 15620 2 15621 404 15622 15622 15623 5157 15624 15624 15625 15625 15626 15626 15627 15627 15628 15628 15629 15629 15630 2820 15631 15631 15632 67 15633 15633 15634 15634 15635 15635 15636 15636 15637 15637 15638 15638 15639 2 15640 2854 15641 15641 15642 15642 15643 15643 15644 15644 15645 15645 15646 15646 15647 2863 15648 15648 15649 15649 15650 15650 15651 15651 15652 15652 15653 338 15654 338 15655 40 15656 341 15657 2 15658 404 15659 15659 15660 555 15661 15661 15662 15662 15663 15663 15664 15664 15665 917 15666 15666 15667 15667 15668 15668 15669 950 15670 143 15671 146 15672 15672 15673 15673 15674 40 15675 15675 15676 15676 15677 2901 15678 2903 15679 2905 15680 75 15681 2907 15682 2908 15683 15683 15684 264 15685 15685 15686 2 15687 264 15688 75 15689 3856 15690 2753 15691 15691 15692 1823 15693 146 15694 2011 15695 2 15696 15696 15697 2 15698 143 15699 146 15700 15700 15701 15701 15702 15702 15703 15703 15704 15704 15705 5125 15706 15706 15707 146 15708 15708 15709 2 15710 264 15711 15711 15712 2 15713 264 15714 15714 15715 15715 15716 15716 15717 15717 15718 15718 15719 15719 15720 1823 15721 146 15722 15722 15723 15723 15724 15724 15725 15725 15726 264 15727 15727 15728 15728 15729 2 15730 404 15731 15731 15732 5157 15733 15733 15734 15734 15735 26 15736 15736 15737 15 15738 15 15739 11 15740 26 15741 2 15742 11 15743 11 15744 40 15745 16 15746 15746 15747 15 15748 68 15749 69 15750 11 15751 75 15752 11 15753 11 15754 15754 15755 86 15756 101 15757 15757 15758 190 15759 15759 15760 123 15761 154 15762 15762 15763 26 15764 15764 15765 15765 15766 1039 15767 11 15768 5969 15769 15769 15770 2 15771 264 15772 1039 15773 15773 15774 15774 15775 15775 15776 15 15777 16 15778 15 15779 11 15780 26 15781 2 15782 16 15783 11 15784 11 15785 40 15786 16 15787 11 15788 11 15789 15789 15790 16 15791 15 15792 68 15793 69 15794 11 15795 16 15796 75 15797 11 15798 11 15799 15799 15800 86 15801 15801 15802 90 15803 15803 15804 96 15805 15805 15806 101 15807 104 15808 15808 15809 15809 15810 15810 15811 15811 15812 15812 15813 15813 15814 15814 15815 123 15816 154 15817 15817 15818 26 15819 11 15820 2 15821 264 15822 26 15823 15823 15824 15824 15825 15825 15826 15826 15827 15827 15828 15828 15829 136 15830 11 15831 138 15832 146 15833 149 15834 151 15835 75 15836 123 15837 154 15838 15838 15839 15839 15840 15840 15841 16 15842 15842 15843 2901 15844 2903 15845 2905 15846 75 15847 2907 15848 2908 15849 15849 15850 2653 15851 15851 15852 15852 15853 15853 15854 15854 15855 1039 15856 5969 15857 15857 15858 15858 15859 15859 15860 15860 15861 15 15862 16 15863 15 15864 11 15865 26 15866 2 15867 5969 15868 16 15869 11 15870 11 15871 40 15872 16 15873 11 15874 15874 15875 16 15876 15 15877 68 15878 69 15879 11 15880 16 15881 75 15882 11 15883 11 15884 15884 15885 86 15886 15886 15887 90 15888 15888 15889 96 15890 15890 15891 101 15892 104 15893 15893 15894 15894 15895 15810 15896 15896 15897 15897 15898 15898 15899 15899 15900 123 15901 154 15902 15902 15903 26 15904 15904 15905 15828 15906 136 15907 11 15908 138 15909 146 15910 149 15911 151 15912 75 15913 123 15914 154 15915 15915 15916 15916 15917 15917 15918 16 15919 2901 15920 2903 15921 2905 15922 75 15923 2907 15924 2908 15925 15925 15926 15926 15927 1039 15928 5969 15929 15929 15930 15858 15931 15859 15932 5969 15933 15933 15934 2 15935 264 15936 15936 15937 15937 15938 15938 15939 15939 15940 15940 15941 15941 15942 15 15943 16 15944 15 15945 11 15946 26 15947 2 15948 5969 15949 16 15950 11 15951 11 15952 40 15953 16 15954 11 15955 11 15956 15956 15957 16 15958 15 15959 68 15960 69 15961 11 15962 16 15963 75 15964 11 15965 11 15966 15966 15967 86 15968 15968 15969 90 15970 15970 15971 96 15972 15972 15973 101 15974 104 15975 15975 15976 15976 15977 15977 15978 15978 15979 15979 15980 15980 15981 15981 15982 123 15983 154 15984 15984 15985 26 15986 11 15987 2 15988 264 15989 26 15990 15990 15991 15 15992 15 15993 11 15994 26 15995 2 15996 11 15997 11 15998 40 15999 16 16000 16000 16001 15 16002 68 16003 69 16004 11 16005 75 16006 11 16007 11 16008 16008 16009 86 16010 101 16011 16011 16012 15977 16013 16013 16014 123 16015 154 16016 16016 16017 26 16018 16018 16019 16019 16020 1039 16021 5969 16022 16022 16023 15858 16024 15859 16025 5969 16026 16026 16027 2 16028 264 16029 15936 16030 15937 16031 16031 16032 16032 16033 15940 16034 5969 16035 16035 16036 22 16037 4 16038 2675 16039 87 16040 11 16041 2678 16042 97 16043 2680 16044 1517 16045 2682 16046 2683 16047 16047 16048 143 16049 16049 16050 97 16051 830 16052 11 16053 2688 16054 917 16055 2690 16056 2691 16057 950 16058 143 16059 146 16060 2695 16061 2696 16062 16062 16063 146 16064 5969 16065 16065 16066 16066 16067 2697 16068 830 16069 11 16070 2700 16071 2701 16072 2702 16073 338 16074 338 16075 341 16076 2 16077 404 16078 2708 16079 555 16080 2710 16081 2711 16082 2712 16083 2713 16084 2 16085 404 16086 2716 16087 2717 16088 2718 16089 2719 16090 2720 16091 2 16092 404 16093 2723 16094 123 16095 154 16096 16096 16097 16097 16098 11 16099 75 16100 16100 16101 123 16102 154 16103 16103 16104 13 16105 16105 16106 16106 16107 16107 16108 16108 16109 16109 16110 16110 16111 16111 16112 16112 16113 16113 16114 146 16115 13 16116 16116 16117 75 16118 123 16119 154 16120 16120 16121 2 16122 264 16123 16123 16124 16124 16125 16125 16126 16126 16127 16127 16128 16128 16129 2 16130 16130 16131 16131 16132 16132 16133 16133 16134 16134 16135 16135 16136 16136 16137 15 16138 16 16139 15 16140 11 16141 26 16142 2 16143 16 16144 11 16145 11 16146 40 16147 16 16148 11 16149 11 16150 16150 16151 16 16152 15 16153 68 16154 69 16155 11 16156 16 16157 75 16158 11 16159 11 16160 16160 16161 86 16162 16162 16163 90 16164 16164 16165 96 16166 16166 16167 101 16168 104 16169 16169 16170 16170 16171 16171 16172 16172 16173 16173 16174 16174 16175 16175 16176 123 16177 154 16178 16178 16179 26 16180 11 16181 2 16182 264 16183 26 16184 16184 16185 15 16186 15 16187 11 16188 26 16189 2 16190 11 16191 11 16192 40 16193 16 16194 16194 16195 15 16196 68 16197 69 16198 11 16199 75 16200 11 16201 11 16202 16202 16203 86 16204 101 16205 16205 16206 16171 16207 16207 16208 123 16209 154 16210 16210 16211 26 16212 16212 16213 16213 16214 1039 16215 5969 16216 16216 16217 15858 16218 15859 16219 5969 16220 16220 16221 2 16222 264 16223 15936 16224 15937 16225 16225 16226 16226 16227 15940 16228 5969 16229 16229 16230 22 16231 143 16232 146 16233 5969 16234 16234 16235 123 16236 154 16237 16237 16238 16238 16239 11 16240 75 16241 16241 16242 123 16243 154 16244 16244 16245 13 16246 16110 16247 146 16248 75 16249 123 16250 154 16251 16251 16252 2 16253 264 16254 16123 16255 16124 16256 16256 16257 16126 16258 16127 16259 16128 16260 16260 16261 16128 16262 16262 16263 16263 16264 16264 16265 2 16266 16130 16267 5969 16268 16268 16269 16269 16270 16270 16271 16271 16272 16272 16273 16273 16274 15 16275 16 16276 15 16277 11 16278 26 16279 2 16280 16 16281 11 16282 11 16283 40 16284 16 16285 11 16286 11 16287 16287 16288 16 16289 15 16290 68 16291 69 16292 11 16293 16 16294 75 16295 11 16296 11 16297 16297 16298 86 16299 16299 16300 90 16301 16301 16302 96 16303 16303 16304 101 16305 104 16306 16306 16307 16307 16308 16308 16309 16309 16310 16310 16311 16311 16312 16312 16313 123 16314 154 16315 16315 16316 26 16317 11 16318 2 16319 264 16320 26 16321 16321 16322 15 16323 15 16324 11 16325 26 16326 2 16327 11 16328 11 16329 40 16330 16 16331 16331 16332 15 16333 68 16334 69 16335 11 16336 75 16337 11 16338 11 16339 16339 16340 86 16341 101 16342 16342 16343 16308 16344 16344 16345 123 16346 154 16347 16347 16348 26 16349 16349 16350 16350 16351 1039 16352 5969 16353 16353 16354 15858 16355 15859 16356 5969 16357 16357 16358 2 16359 264 16360 15936 16361 15937 16362 16362 16363 16363 16364 15940 16365 5969 16366 16366 16367 22 16368 143 16369 146 16370 5969 16371 16371 16372 123 16373 154 16374 16374 16375 16375 16376 11 16377 75 16378 16378 16379 123 16380 154 16381 16381 16382 13 16383 16110 16384 146 16385 75 16386 123 16387 154 16388 16388 16389 2 16390 264 16391 16123 16392 16124 16393 16393 16394 16126 16395 16127 16396 16128 16397 16397 16398 16128 16399 16399 16400 16263 16401 16264 16402 5969 16403 16403 16404 16264 16405 5969 16406 16406 16407 16407 16408 16408 16409 16409 16410 16410 16411 2965 16412 2939 16413 16413 16414 16414 16415 2 16416 2 16417 3 16418 4 16419 2 16420 2 16421 2 16422 2 16423 11 16424 2 16425 2 16426 2981 16427 2 16428 2983 16429 22 16430 2985 16431 2986 16432 2 16433 11 16434 11 16435 2990 16436 2991 16437 2992 16438 22 16439 2 16440 2 16441 2 16442 11 16443 2998 16444 11 16445 3000 16446 11 16447 3002 16448 3003 16449 3004 16450 3005 16451 60 16452 3007 16453 2 16454 2 16455 2 16456 2 16457 67 16458 127 16459 3014 16460 2 16461 2 16462 3017 16463 2 16464 3019 16465 3020 16466 2 16467 83 16468 84 16469 87 16470 130 16471 11 16472 3027 16473 3028 16474 97 16475 3030 16476 3031 16477 16477 16478 102 16479 102 16480 3035 16481 3036 16482 3037 16483 3038 16484 3039 16485 3040 16486 3041 16487 3042 16488 3043 16489 3044 16490 97 16491 2 16492 3047 16493 2 16494 2 16495 2 16496 2 16497 2 16498 2 16499 2 16500 2 16501 288 16502 2 16503 2 16504 2 16505 2 16506 2 16507 294 16508 2 16509 2 16510 2 16511 2 16512 2 16513 2 16514 2 16515 2 16516 2 16517 2 16518 2 16519 2 16520 2 16521 2 16522 2 16523 3078 16524 3079 16525 3080 16526 3081 16527 3082 16528 3083 16529 3084 16530 3085 16531 3086 16532 3087 16533 3088 16534 3089 16535 3090 16536 3091 16537 3092 16538 4 16539 2 16540 3095 16541 3096 16542 3097 16543 3098 16544 3099 16545 3100 16546 3101 16547 3102 16548 3103 16549 3104 16550 3105 16551 338 16552 338 16553 40 16554 341 16555 2 16556 404 16557 3112 16558 555 16559 3114 16560 3115 16561 3116 16562 3117 16563 3118 16564 2 16565 404 16566 3121 16567 3122 16568 40 16569 3124 16570 264 16571 3126 16572 3127 16573 2 16574 404 16575 3130 16576 3131 16577 3132 16578 3133 16579 3134 16580 3135 16581 2 16582 713 16583 2 16584 3139 16585 3140 16586 3141 16587 3142 16588 3143 16589 3144 16590 3145 16591 3146 16592 3147 16593 3148 16594 830 16595 3150 16596 3151 16597 3152 16598 3153 16599 3154 16600 3155 16601 3156 16602 3157 16603 3158 16604 3159 16605 3160 16606 3161 16607 3162 16608 3163 16609 3164 16610 3165 16611 3166 16612 3167 16613 3168 16614 3169 16615 3170 16616 3171 16617 3172 16618 3173 16619 3174 16620 3175 16621 3176 16622 3177 16623 3178 16624 3179 16625 3180 16626 3181 16627 3182 16628 3183 16629 3184 16630 3185 16631 3186 16632 3187 16633 3188 16634 3189 16635 3190 16636 3191 16637 3192 16638 3193 16639 3194 16640 3195 16641 3196 16642 3197 16643 3198 16644 3199 16645 3200 16646 3201 16647 3202 16648 3203 16649 3204 16650 3205 16651 3206 16652 3207 16653 3208 16654 3209 16655 3210 16656 3211 16657 3212 16658 3213 16659 3214 16660 3215 16661 3216 16662 3217 16663 3218 16664 3219 16665 3220 16666 3221 16667 3222 16668 3223 16669 3224 16670 3225 16671 3226 16672 3227 16673 3228 16674 11 16675 3230 16676 3231 16677 3232 16678 917 16679 3234 16680 3235 16681 950 16682 143 16683 146 16684 3239 16685 3240 16686 1456 16687 2 16688 2 16689 2 16690 3245 16691 2 16692 3247 16693 2 16694 2 16695 2 16696 2 16697 2 16698 2 16699 2 16700 2 16701 2 16702 3257 16703 3258 16704 3259 16705 3260 16706 3261 16707 3262 16708 3263 16709 3264 16710 3265 16711 3266 16712 3267 16713 3268 16714 3269 16715 22 16716 3271 16717 4 16718 3273 16719 87 16720 11 16721 3276 16722 97 16723 3278 16724 1517 16725 3280 16726 3281 16727 3282 16728 97 16729 830 16730 11 16731 3286 16732 917 16733 3288 16734 3289 16735 950 16736 143 16737 146 16738 3293 16739 3294 16740 3295 16741 830 16742 11 16743 3298 16744 3299 16745 3300 16746 338 16747 338 16748 341 16749 2 16750 404 16751 3306 16752 555 16753 3308 16754 3309 16755 3310 16756 3311 16757 2 16758 404 16759 3314 16760 3315 16761 3316 16762 3317 16763 3318 16764 2 16765 404 16766 3321 16767 3322 16768 136 16769 11 16770 138 16771 146 16772 147 16773 149 16774 151 16775 75 16776 123 16777 154 16778 3333 16779 3334 16780 136 16781 11 16782 138 16783 146 16784 2206 16785 2208 16786 2209 16787 2210 16788 2 16789 3344 16790 3345 16791 830 16792 3348 16793 11 16794 3350 16795 3351 16796 3352 16797 2653 16798 3354 16799 3355 16800 338 16801 338 16802 341 16803 2 16804 404 16805 3361 16806 555 16807 3363 16808 3364 16809 3365 16810 3366 16811 2 16812 404 16813 3369 16814 3370 16815 3371 16816 3372 16817 2728 16818 3374 16819 555 16820 3376 16821 2 16822 3378 16823 3379 16824 2 16825 3 16826 4 16827 2 16828 2 16829 2 16830 2 16831 11 16832 2 16833 2 16834 3390 16835 2 16836 3392 16837 22 16838 3394 16839 3395 16840 2 16841 11 16842 11 16843 3399 16844 3400 16845 3401 16846 22 16847 2 16848 2 16849 2 16850 11 16851 3407 16852 11 16853 3409 16854 11 16855 3411 16856 3412 16857 3413 16858 3414 16859 60 16860 3416 16861 2 16862 2 16863 2 16864 2 16865 67 16866 127 16867 3423 16868 2 16869 2 16870 3426 16871 2 16872 3428 16873 3429 16874 2 16875 83 16876 84 16877 87 16878 130 16879 11 16880 3436 16881 3437 16882 97 16883 3439 16884 3440 16885 3441 16886 102 16887 102 16888 3444 16889 3445 16890 3446 16891 3447 16892 3448 16893 3449 16894 3450 16895 3451 16896 3452 16897 3453 16898 97 16899 2 16900 3456 16901 2 16902 2 16903 2 16904 2 16905 2 16906 2 16907 2 16908 2 16909 288 16910 2 16911 2 16912 2 16913 2 16914 2 16915 294 16916 2 16917 2 16918 2 16919 2 16920 2 16921 2 16922 2 16923 2 16924 2 16925 2 16926 2 16927 2 16928 2 16929 2 16930 2 16931 3487 16932 3488 16933 3489 16934 3490 16935 3491 16936 3492 16937 3493 16938 3494 16939 3495 16940 3496 16941 3497 16942 3498 16943 3499 16944 3500 16945 3501 16946 4 16947 2 16948 3504 16949 3505 16950 3506 16951 3507 16952 3508 16953 3509 16954 3510 16955 3511 16956 3512 16957 3513 16958 3514 16959 338 16960 338 16961 40 16962 341 16963 2 16964 404 16965 3521 16966 555 16967 3523 16968 3524 16969 3525 16970 3526 16971 3527 16972 2 16973 404 16974 3530 16975 3531 16976 40 16977 3533 16978 264 16979 3535 16980 3536 16981 2 16982 404 16983 3539 16984 3540 16985 3541 16986 3542 16987 3543 16988 3544 16989 2 16990 713 16991 2 16992 3548 16993 3549 16994 3550 16995 3551 16996 3552 16997 3553 16998 3554 16999 3555 17000 3556 17001 3557 17002 830 17003 3559 17004 3560 17005 3561 17006 3562 17007 3563 17008 3564 17009 3565 17010 3566 17011 3567 17012 3568 17013 3569 17014 3570 17015 3571 17016 3572 17017 3573 17018 3574 17019 3575 17020 3576 17021 3577 17022 3578 17023 3579 17024 3580 17025 3581 17026 3582 17027 3583 17028 3584 17029 3585 17030 3586 17031 3587 17032 3588 17033 3589 17034 3590 17035 3591 17036 3592 17037 3593 17038 3594 17039 3595 17040 3596 17041 3597 17042 3598 17043 3599 17044 3600 17045 3601 17046 3602 17047 3603 17048 3604 17049 3605 17050 3606 17051 3607 17052 3608 17053 3609 17054 3610 17055 3611 17056 3612 17057 3613 17058 3614 17059 3615 17060 3616 17061 3617 17062 3618 17063 3619 17064 3620 17065 3621 17066 3622 17067 3623 17068 3624 17069 3625 17070 3626 17071 3627 17072 3628 17073 3629 17074 3630 17075 3631 17076 3632 17077 3633 17078 3634 17079 3635 17080 3636 17081 3637 17082 11 17083 3639 17084 3640 17085 3641 17086 917 17087 3643 17088 3644 17089 950 17090 143 17091 146 17092 3648 17093 3649 17094 1456 17095 2 17096 2 17097 2 17098 3654 17099 2 17100 3656 17101 2 17102 2 17103 2 17104 2 17105 2 17106 2 17107 2 17108 2 17109 2 17110 3666 17111 3667 17112 3668 17113 3669 17114 3670 17115 3671 17116 3672 17117 3673 17118 3674 17119 3675 17120 3676 17121 3677 17122 3678 17123 22 17124 3680 17125 4 17126 3682 17127 87 17128 11 17129 3685 17130 97 17131 3687 17132 1517 17133 3689 17134 3690 17135 3691 17136 97 17137 830 17138 11 17139 3695 17140 917 17141 3697 17142 3698 17143 950 17144 143 17145 146 17146 3702 17147 3703 17148 3704 17149 830 17150 11 17151 3707 17152 3708 17153 3709 17154 338 17155 338 17156 341 17157 2 17158 404 17159 3715 17160 555 17161 3717 17162 3718 17163 3719 17164 3720 17165 2 17166 404 17167 3723 17168 3724 17169 3725 17170 3726 17171 3727 17172 2 17173 404 17174 3730 17175 3731 17176 136 17177 11 17178 138 17179 146 17180 147 17181 149 17182 151 17183 75 17184 123 17185 154 17186 3742 17187 3743 17188 136 17189 11 17190 138 17191 146 17192 2206 17193 2208 17194 2209 17195 2210 17196 2 17197 3753 17198 3754 17199 2 17200 3756 17201 830 17202 3758 17203 11 17204 3760 17205 3761 17206 3762 17207 2653 17208 3764 17209 3765 17210 338 17211 338 17212 341 17213 2 17214 404 17215 3771 17216 555 17217 3773 17218 3774 17219 3775 17220 3776 17221 2 17222 404 17223 3779 17224 3780 17225 3781 17226 3782 17227 2728 17228 3784 17229 555 17230 3786 17231 2 17232 3788 17233 3789 17234 3441 17235 2 17236 3792 17237 3793 17238 3794 17239 3795 17240 3796 17241 2820 17242 3798 17243 67 17244 3800 17245 3801 17246 3802 17247 3803 17248 3804 17249 3805 17250 2854 17251 3813 17252 2863 17253 3816 17254 917 17255 3833 17256 3834 17257 3835 17258 950 17259 143 17260 146 17261 3839 17262 3840 17263 3843 17264 2901 17265 2903 17266 2905 17267 75 17268 2907 17269 2908 17270 3850 17271 2 17272 264 17273 75 17274 3856 17275 2753 17276 4217 17277 1823 17278 146 17279 2011 17280 2 17281 5467 17282 2 17283 143 17284 146 17285 5471 17286 5472 17287 5473 17288 5474 17289 5475 17290 5125 17291 5477 17292 146 17293 5479 17294 2 17295 264 17296 5482 17297 2 17298 264 17299 5485 17300 5486 17301 5487 17302 5488 17303 5489 17304 5490 17305 1823 17306 146 17307 5493 17308 5494 17309 5495 17310 5496 17311 264 17312 5498 17313 5499 17314 2 17315 404 17316 5502 17317 5157 17318 5504 17319 5505 17320 5506 17321 5507 17322 5508 17323 5509 17324 2820 17325 5511 17326 67 17327 5513 17328 5514 17329 5515 17330 5516 17331 5517 17332 5518 17333 2854 17334 5526 17335 2863 17336 5529 17337 917 17338 5546 17339 5547 17340 5548 17341 950 17342 143 17343 146 17344 5552 17345 5553 17346 5556 17347 2901 17348 2903 17349 2905 17350 75 17351 2907 17352 2908 17353 5563 17354 2 17355 264 17356 75 17357 3856 17358 2753 17359 5571 17360 1823 17361 146 17362 2011 17363 2 17364 5576 17365 2 17366 143 17367 146 17368 5580 17369 5581 17370 5582 17371 5583 17372 5584 17373 5125 17374 5586 17375 146 17376 5588 17377 2 17378 264 17379 5591 17380 2 17381 264 17382 5594 17383 5595 17384 5596 17385 5597 17386 5598 17387 5599 17388 1823 17389 146 17390 5602 17391 5603 17392 5604 17393 5605 17394 264 17395 5607 17396 5608 17397 2 17398 404 17399 5611 17400 5157 17401 5613 17402 5614 17403 17403 17404 5969 17405 17405 17406 5969 17407 17407 17408 4 17409 17409 17410 17410 17411 2965 17412 17412 17413 40 17414 17414 17415 17415 17416 17416 17417 2 17418 404 17419 17419 17420 2 17421 2 17422 3 17423 4 17424 2 17425 2 17426 2 17427 2 17428 11 17429 2 17430 2 17431 17431 17432 2 17433 17433 17434 22 17435 17435 17436 17436 17437 2 17438 11 17439 11 17440 17440 17441 17441 17442 17442 17443 22 17444 2 17445 2 17446 2 17447 11 17448 17448 17449 11 17450 17450 17451 11 17452 17452 17453 17453 17454 17454 17455 17455 17456 60 17457 17457 17458 2 17459 2 17460 2 17461 2 17462 67 17463 127 17464 17464 17465 2 17466 2 17467 17467 17468 2 17469 17469 17470 17470 17471 2 17472 83 17473 84 17474 87 17475 130 17476 11 17477 17477 17478 17478 17479 97 17480 17480 17481 17481 17482 17482 17483 102 17484 102 17485 17485 17486 17486 17487 17487 17488 17488 17489 17489 17490 17490 17491 17491 17492 17492 17493 17493 17494 17494 17495 97 17496 2 17497 17497 17498 2 17499 2 17500 2 17501 2 17502 2 17503 2 17504 2 17505 2 17506 288 17507 2 17508 2 17509 2 17510 2 17511 2 17512 294 17513 2 17514 2 17515 2 17516 2 17517 2 17518 2 17519 2 17520 2 17521 2 17522 2 17523 2 17524 2 17525 2 17526 2 17527 2 17528 17528 17529 17529 17530 17530 17531 17531 17532 17532 17533 17533 17534 17534 17535 17535 17536 17536 17537 17537 17538 17538 17539 17539 17540 17540 17541 17541 17542 17542 17543 4 17544 2 17545 17545 17546 17546 17547 17547 17548 17548 17549 17549 17550 17550 17551 17551 17552 17552 17553 17553 17554 17554 17555 17555 17556 338 17557 338 17558 40 17559 341 17560 2 17561 404 17562 17562 17563 555 17564 17564 17565 17565 17566 17566 17567 17567 17568 17568 17569 2 17570 404 17571 17571 17572 17572 17573 40 17574 17574 17575 264 17576 17576 17577 17577 17578 2 17579 404 17580 17580 17581 17581 17582 17582 17583 17583 17584 17584 17585 17585 17586 2 17587 713 17588 2 17589 17589 17590 17590 17591 17591 17592 17592 17593 17593 17594 17594 17595 17595 17596 17596 17597 17597 17598 17598 17599 830 17600 17600 17601 17601 17602 17602 17603 17603 17604 17604 17605 17605 17606 17606 17607 17607 17608 17608 17609 17609 17610 17610 17611 17611 17612 17612 17613 17613 17614 17614 17615 17615 17616 17616 17617 17617 17618 17618 17619 17619 17620 17620 17621 17621 17622 17622 17623 17623 17624 17624 17625 17625 17626 17626 17627 17627 17628 17628 17629 17629 17630 17630 17631 17631 17632 17632 17633 17633 17634 17634 17635 17635 17636 17636 17637 17637 17638 17638 17639 17639 17640 17640 17641 17641 17642 17642 17643 17643 17644 17644 17645 17645 17646 17646 17647 17647 17648 17648 17649 17649 17650 17650 17651 17651 17652 17652 17653 17653 17654 17654 17655 17655 17656 17656 17657 17657 17658 17658 17659 17659 17660 17660 17661 17661 17662 17662 17663 17663 17664 17664 17665 17665 17666 17666 17667 17667 17668 17668 17669 17669 17670 17670 17671 17671 17672 17672 17673 17673 17674 17674 17675 17675 17676 17676 17677 17677 17678 17678 17679 11 17680 17680 17681 17681 17682 17682 17683 917 17684 17684 17685 17685 17686 950 17687 143 17688 146 17689 17689 17690 17690 17691 1456 17692 2 17693 2 17694 2 17695 17695 17696 2 17697 17697 17698 2 17699 2 17700 2 17701 2 17702 2 17703 2 17704 2 17705 2 17706 2 17707 17707 17708 17708 17709 17709 17710 17710 17711 17711 17712 17712 17713 17713 17714 17714 17715 17715 17716 17716 17717 17717 17718 17718 17719 17719 17720 22 17721 17721 17722 4 17723 17723 17724 87 17725 11 17726 17726 17727 97 17728 17728 17729 1517 17730 17730 17731 17731 17732 17732 17733 97 17734 830 17735 11 17736 17736 17737 917 17738 17738 17739 17739 17740 950 17741 143 17742 146 17743 17743 17744 17744 17745 17745 17746 830 17747 11 17748 17748 17749 17749 17750 17750 17751 338 17752 338 17753 341 17754 2 17755 404 17756 17756 17757 555 17758 17758 17759 17759 17760 17760 17761 17761 17762 2 17763 404 17764 17764 17765 17765 17766 17766 17767 17767 17768 17768 17769 2 17770 404 17771 17771 17772 17772 17773 136 17774 11 17775 138 17776 146 17777 147 17778 149 17779 151 17780 75 17781 123 17782 154 17783 17783 17784 17784 17785 136 17786 11 17787 138 17788 146 17789 2206 17790 2208 17791 2209 17792 2210 17793 2 17794 2 17795 3 17796 4 17797 2 17798 2 17799 2 17800 2 17801 11 17802 2 17803 2 17804 17431 17805 2 17806 17433 17807 22 17808 17435 17809 17436 17810 2 17811 11 17812 11 17813 17440 17814 17441 17815 17442 17816 22 17817 2 17818 2 17819 2 17820 11 17821 17448 17822 11 17823 17450 17824 11 17825 17452 17826 17453 17827 17454 17828 17455 17829 60 17830 17457 17831 2 17832 2 17833 2 17834 2 17835 67 17836 127 17837 17464 17838 2 17839 2 17840 17467 17841 2 17842 17469 17843 17470 17844 2 17845 83 17846 84 17847 87 17848 130 17849 11 17850 17477 17851 17478 17852 97 17853 17480 17854 17481 17855 17855 17856 102 17857 102 17858 17485 17859 17486 17860 17487 17861 17488 17862 17489 17863 17490 17864 17491 17865 17492 17866 17493 17867 17494 17868 97 17869 2 17870 17497 17871 2 17872 2 17873 2 17874 2 17875 2 17876 2 17877 2 17878 2 17879 288 17880 2 17881 2 17882 2 17883 2 17884 2 17885 294 17886 2 17887 2 17888 2 17889 2 17890 2 17891 2 17892 2 17893 2 17894 2 17895 2 17896 2 17897 2 17898 2 17899 2 17900 2 17901 17528 17902 17529 17903 17530 17904 17531 17905 17532 17906 17533 17907 17534 17908 17535 17909 17536 17910 17537 17911 17538 17912 17539 17913 17540 17914 17541 17915 17542 17916 4 17917 2 17918 17545 17919 17546 17920 17547 17921 17548 17922 17549 17923 17550 17924 17551 17925 17552 17926 17553 17927 17554 17928 17555 17929 338 17930 338 17931 40 17932 341 17933 2 17934 404 17935 17562 17936 555 17937 17564 17938 17565 17939 17566 17940 17567 17941 17568 17942 2 17943 404 17944 17571 17945 17572 17946 40 17947 17574 17948 264 17949 17576 17950 17577 17951 2 17952 404 17953 17580 17954 17581 17955 17582 17956 17583 17957 17584 17958 17585 17959 2 17960 713 17961 2 17962 17589 17963 17590 17964 17591 17965 17592 17966 17593 17967 17594 17968 17595 17969 17596 17970 17597 17971 17598 17972 830 17973 17600 17974 17601 17975 17602 17976 17603 17977 17604 17978 17605 17979 17606 17980 17607 17981 17608 17982 17609 17983 17610 17984 17611 17985 17612 17986 17613 17987 17614 17988 17615 17989 17616 17990 17617 17991 17618 17992 17619 17993 17620 17994 17621 17995 17622 17996 17623 17997 17624 17998 17625 17999 17626 18000 17627 18001 17628 18002 17629 18003 17630 18004 17631 18005 17632 18006 17633 18007 17634 18008 17635 18009 17636 18010 17637 18011 17638 18012 17639 18013 17640 18014 17641 18015 17642 18016 17643 18017 17644 18018 17645 18019 17646 18020 17647 18021 17648 18022 17649 18023 17650 18024 17651 18025 17652 18026 17653 18027 17654 18028 17655 18029 17656 18030 17657 18031 17658 18032 17659 18033 17660 18034 17661 18035 17662 18036 17663 18037 17664 18038 17665 18039 17666 18040 17667 18041 17668 18042 17669 18043 17670 18044 17671 18045 17672 18046 17673 18047 17674 18048 17675 18049 17676 18050 17677 18051 17678 18052 11 18053 17680 18054 17681 18055 17682 18056 917 18057 17684 18058 17685 18059 950 18060 143 18061 146 18062 17689 18063 17690 18064 1456 18065 2 18066 2 18067 2 18068 17695 18069 2 18070 17697 18071 2 18072 2 18073 2 18074 2 18075 2 18076 2 18077 2 18078 2 18079 2 18080 17707 18081 17708 18082 17709 18083 17710 18084 17711 18085 17712 18086 17713 18087 17714 18088 17715 18089 17716 18090 17717 18091 17718 18092 17719 18093 22 18094 17721 18095 4 18096 17723 18097 87 18098 11 18099 17726 18100 97 18101 17728 18102 1517 18103 17730 18104 17731 18105 17732 18106 97 18107 830 18108 11 18109 17736 18110 917 18111 17738 18112 17739 18113 950 18114 143 18115 146 18116 17743 18117 17744 18118 17745 18119 830 18120 11 18121 17748 18122 17749 18123 17750 18124 338 18125 338 18126 341 18127 2 18128 404 18129 17756 18130 555 18131 17758 18132 17759 18133 17760 18134 17761 18135 2 18136 404 18137 17764 18138 17765 18139 17766 18140 17767 18141 17768 18142 2 18143 404 18144 17771 18145 17772 18146 136 18147 11 18148 138 18149 146 18150 147 18151 149 18152 151 18153 75 18154 123 18155 154 18156 17783 18157 17784 18158 136 18159 11 18160 138 18161 146 18162 2206 18163 2208 18164 2209 18165 2210 18166 2 18167 17855 18168 18168 18169 18168 18170 830 18171 18171 18172 11 18173 18173 18174 18174 18175 18175 18176 2653 18177 18177 18178 18178 18179 338 18180 338 18181 341 18182 2 18183 404 18184 18184 18185 555 18186 18186 18187 18187 18188 18188 18189 18189 18190 2 18191 404 18192 18192 18193 18193 18194 18194 18195 18195 18196 2728 18197 18197 18198 555 18199 18199 18200 2 18201 18201 18202 18202 18203 2 18204 3 18205 4 18206 2 18207 2 18208 2 18209 2 18210 11 18211 2 18212 2 18213 18213 18214 2 18215 18215 18216 22 18217 18217 18218 18218 18219 2 18220 11 18221 11 18222 18222 18223 18223 18224 18224 18225 22 18226 2 18227 2 18228 2 18229 11 18230 18230 18231 11 18232 18232 18233 11 18234 18234 18235 18235 18236 18236 18237 18237 18238 60 18239 18239 18240 2 18241 2 18242 2 18243 2 18244 67 18245 127 18246 18246 18247 2 18248 2 18249 18249 18250 2 18251 18251 18252 18252 18253 2 18254 83 18255 84 18256 87 18257 130 18258 11 18259 18259 18260 18260 18261 97 18262 18262 18263 18263 18264 18264 18265 102 18266 102 18267 18267 18268 18268 18269 18269 18270 18270 18271 18271 18272 18272 18273 18273 18274 18274 18275 18275 18276 18276 18277 97 18278 2 18279 18279 18280 2 18281 2 18282 2 18283 2 18284 2 18285 2 18286 2 18287 2 18288 288 18289 2 18290 2 18291 2 18292 2 18293 2 18294 294 18295 2 18296 2 18297 2 18298 2 18299 2 18300 2 18301 2 18302 2 18303 2 18304 2 18305 2 18306 2 18307 2 18308 2 18309 2 18310 18310 18311 18311 18312 18312 18313 18313 18314 18314 18315 18315 18316 18316 18317 18317 18318 18318 18319 18319 18320 18320 18321 18321 18322 18322 18323 18323 18324 18324 18325 4 18326 2 18327 18327 18328 18328 18329 18329 18330 18330 18331 18331 18332 18332 18333 18333 18334 18334 18335 18335 18336 18336 18337 18337 18338 338 18339 338 18340 40 18341 341 18342 2 18343 404 18344 18344 18345 555 18346 18346 18347 18347 18348 18348 18349 18349 18350 18350 18351 2 18352 404 18353 18353 18354 18354 18355 40 18356 18356 18357 264 18358 18358 18359 18359 18360 2 18361 404 18362 18362 18363 18363 18364 18364 18365 18365 18366 18366 18367 18367 18368 2 18369 713 18370 2 18371 18371 18372 18372 18373 18373 18374 18374 18375 18375 18376 18376 18377 18377 18378 18378 18379 18379 18380 18380 18381 830 18382 18382 18383 18383 18384 18384 18385 18385 18386 18386 18387 18387 18388 18388 18389 18389 18390 18390 18391 18391 18392 18392 18393 18393 18394 18394 18395 18395 18396 18396 18397 18397 18398 18398 18399 18399 18400 18400 18401 18401 18402 18402 18403 18403 18404 18404 18405 18405 18406 18406 18407 18407 18408 18408 18409 18409 18410 18410 18411 18411 18412 18412 18413 18413 18414 18414 18415 18415 18416 18416 18417 18417 18418 18418 18419 18419 18420 18420 18421 18421 18422 18422 18423 18423 18424 18424 18425 18425 18426 18426 18427 18427 18428 18428 18429 18429 18430 18430 18431 18431 18432 18432 18433 18433 18434 18434 18435 18435 18436 18436 18437 18437 18438 18438 18439 18439 18440 18440 18441 18441 18442 18442 18443 18443 18444 18444 18445 18445 18446 18446 18447 18447 18448 18448 18449 18449 18450 18450 18451 18451 18452 18452 18453 18453 18454 18454 18455 18455 18456 18456 18457 18457 18458 18458 18459 18459 18460 18460 18461 11 18462 18462 18463 18463 18464 18464 18465 917 18466 18466 18467 18467 18468 950 18469 143 18470 146 18471 18471 18472 18472 18473 1456 18474 2 18475 2 18476 2 18477 18477 18478 2 18479 18479 18480 2 18481 2 18482 2 18483 2 18484 2 18485 2 18486 2 18487 2 18488 2 18489 18489 18490 18490 18491 18491 18492 18492 18493 18493 18494 18494 18495 18495 18496 18496 18497 18497 18498 18498 18499 18499 18500 18500 18501 18501 18502 22 18503 18503 18504 4 18505 18505 18506 87 18507 11 18508 18508 18509 97 18510 18510 18511 1517 18512 18512 18513 18513 18514 18514 18515 97 18516 830 18517 11 18518 18518 18519 917 18520 18520 18521 18521 18522 950 18523 143 18524 146 18525 18525 18526 18526 18527 18527 18528 830 18529 11 18530 18530 18531 18531 18532 18532 18533 338 18534 338 18535 341 18536 2 18537 404 18538 18538 18539 555 18540 18540 18541 18541 18542 18542 18543 18543 18544 2 18545 404 18546 18546 18547 18547 18548 18548 18549 18549 18550 18550 18551 2 18552 404 18553 18553 18554 18554 18555 136 18556 11 18557 138 18558 146 18559 147 18560 149 18561 151 18562 75 18563 123 18564 154 18565 18565 18566 18566 18567 136 18568 11 18569 138 18570 146 18571 2206 18572 2208 18573 2209 18574 2210 18575 2 18576 18576 18577 18577 18578 2 18579 18579 18580 830 18581 18581 18582 11 18583 18583 18584 18584 18585 18585 18586 2653 18587 18587 18588 18588 18589 338 18590 338 18591 341 18592 2 18593 404 18594 18594 18595 555 18596 18596 18597 18597 18598 18598 18599 18599 18600 2 18601 404 18602 18602 18603 18603 18604 18604 18605 18605 18606 2728 18607 18607 18608 555 18609 18609 18610 2 18611 18611 18612 18612 18613 18264 18614 2 18615 18615 18616 18616 18617 18617 18618 18618 18619 18619 18620 2820 18621 18621 18622 67 18623 18623 18624 18624 18625 18625 18626 18626 18627 18627 18628 18628 18629 2 18630 2854 18631 18631 18632 18632 18633 18633 18634 18634 18635 18635 18636 18636 18637 2863 18638 18638 18639 18639 18640 18640 18641 18641 18642 18642 18643 338 18644 338 18645 40 18646 341 18647 2 18648 404 18649 18649 18650 555 18651 18651 18652 18652 18653 18653 18654 18654 18655 917 18656 18656 18657 18657 18658 18658 18659 950 18660 143 18661 146 18662 18662 18663 18663 18664 40 18665 18665 18666 18666 18667 2901 18668 2903 18669 2905 18670 75 18671 2907 18672 2908 18673 18673 18674 264 18675 18675 18676 2 18677 264 18678 75 18679 3856 18680 2753 18681 18681 18682 1823 18683 146 18684 2011 18685 2 18686 18686 18687 2 18688 143 18689 146 18690 18690 18691 18691 18692 18692 18693 18693 18694 18694 18695 5125 18696 18696 18697 146 18698 18698 18699 2 18700 264 18701 18701 18702 2 18703 264 18704 18704 18705 18705 18706 18706 18707 18707 18708 18708 18709 18709 18710 1823 18711 146 18712 18712 18713 18713 18714 18714 18715 18715 18716 264 18717 18717 18718 18718 18719 2 18720 404 18721 18721 18722 5157 18723 18723 18724 18724 18725 18725 18726 18726 18727 18727 18728 18728 18729 2820 18730 18730 18731 67 18732 18732 18733 18733 18734 18734 18735 18735 18736 18736 18737 18737 18738 2 18739 2854 18740 18740 18741 18741 18742 18742 18743 18743 18744 18744 18745 18745 18746 2863 18747 18747 18748 18748 18749 18749 18750 18750 18751 18751 18752 338 18753 338 18754 40 18755 341 18756 2 18757 404 18758 18758 18759 555 18760 18760 18761 18761 18762 18762 18763 18763 18764 917 18765 18765 18766 18766 18767 18767 18768 950 18769 143 18770 146 18771 18771 18772 18772 18773 40 18774 18774 18775 18775 18776 2901 18777 2903 18778 2905 18779 75 18780 2907 18781 2908 18782 18782 18783 264 18784 18784 18785 2 18786 264 18787 75 18788 3856 18789 2753 18790 18790 18791 1823 18792 146 18793 2011 18794 2 18795 18795 18796 2 18797 143 18798 146 18799 18799 18800 18800 18801 18801 18802 18802 18803 18803 18804 5125 18805 18805 18806 146 18807 18807 18808 2 18809 264 18810 18810 18811 2 18812 264 18813 18813 18814 18814 18815 18815 18816 18816 18817 18817 18818 18818 18819 1823 18820 146 18821 18821 18822 18822 18823 18823 18824 18824 18825 264 18826 18826 18827 18827 18828 2 18829 404 18830 18830 18831 5157 18832 18832 18833 18833 18834 18834 18835 4 18836 11 18837 18837 18838 83 18839 22 18840 87 18841 11 18842 18842 18843 97 18844 18844 18845 18845 18846 1517 18847 18847 18848 18848 18849 18849 18850 97 18851 830 18852 11 18853 18853 18854 917 18855 18855 18856 18856 18857 950 18858 143 18859 146 18860 18860 18861 18861 18862 18862 18863 15936 18864 18834 18865 4 18866 11 18867 18867 18868 83 18869 22 18870 87 18871 11 18872 18872 18873 97 18874 146 18875 18875 18876 1517 18877 18877 18878 18878 18879 18879 18880 97 18881 830 18882 11 18883 18883 18884 917 18885 18885 18886 18886 18887 950 18888 143 18889 146 18890 18890 18891 18891 18892 18892 18893 15936 18894 18894 18895 18895 18896 18896 18897 18897 18898 18898 18899 15 18900 16 18901 15 18902 11 18903 26 18904 2 18905 5969 18906 16 18907 11 18908 11 18909 40 18910 16 18911 11 18912 11 18913 18913 18914 16 18915 15 18916 68 18917 69 18918 11 18919 16 18920 75 18921 11 18922 11 18923 18923 18924 86 18925 18925 18926 90 18927 18927 18928 96 18929 18929 18930 101 18931 104 18932 18932 18933 18933 18934 18934 18935 18935 18936 18936 18937 18937 18938 18938 18939 123 18940 154 18941 18941 18942 26 18943 11 18944 2 18945 264 18946 26 18947 18947 18948 15 18949 15 18950 11 18951 26 18952 2 18953 11 18954 11 18955 40 18956 16 18957 18957 18958 15 18959 68 18960 69 18961 11 18962 75 18963 11 18964 11 18965 18965 18966 86 18967 101 18968 18968 18969 18934 18970 18970 18971 123 18972 154 18973 18973 18974 26 18975 18975 18976 18976 18977 1039 18978 5969 18979 18979 18980 15858 18981 15859 18982 5969 18983 18983 18984 2 18985 264 18986 15936 18987 15937 18988 18988 18989 18989 18990 15940 18991 5969 18992 18992 18993 22 18994 143 18995 146 18996 5969 18997 18997 18998 123 18999 154 19000 19000 19001 19001 19002 11 19003 75 19004 19004 19005 123 19006 154 19007 19007 19008 13 19009 16110 19010 146 19011 75 19012 123 19013 154 19014 19014 19015 2 19016 264 19017 16123 19018 16124 19019 19019 19020 16126 19021 16127 19022 16128 19023 19023 19024 16128 19025 19025 19026 16263 19027 16264 19028 5969 19029 19029 19030 16264 19031 5969 19032 19032 19033 19033 19034 2965 19035 19035 19036 5969 19037 19037 19038 5969 19039 19039 19040 2965 19041 19041 19042 2 19043 17482 19044 18834 19045 18844 19046 15936 19047 18834 19048 146 19049 15936 19050 18894 19051 19051 19052 19052 19053 18897 19054 5969 19055 19055 19056 19056 19057 19057 19058 19058 19059 5969 19060 19060 19061 2 19062 19062 19063 19057 19064 5969 19065 19065 19066 19066 19067 146 19068 19068 19069 2 19070 19070 19071 15936 19072 19072 19073 19073 19074 19074 19075 19075 19076 19076 19077 15 19078 16 19079 15 19080 11 19081 26 19082 2 19083 5969 19084 16 19085 11 19086 11 19087 40 19088 16 19089 11 19090 11 19091 19091 19092 16 19093 15 19094 68 19095 69 19096 11 19097 16 19098 75 19099 11 19100 11 19101 19101 19102 86 19103 19103 19104 90 19105 19105 19106 96 19107 19107 19108 101 19109 104 19110 19110 19111 19111 19112 19112 19113 19113 19114 19114 19115 19115 19116 19116 19117 123 19118 154 19119 19119 19120 26 19121 11 19122 2 19123 264 19124 26 19125 19125 19126 15 19127 15 19128 11 19129 26 19130 2 19131 11 19132 11 19133 40 19134 16 19135 19135 19136 15 19137 68 19138 69 19139 11 19140 75 19141 11 19142 11 19143 19143 19144 86 19145 101 19146 19146 19147 19112 19148 19148 19149 123 19150 154 19151 19151 19152 26 19153 19153 19154 19154 19155 1039 19156 5969 19157 19157 19158 15858 19159 15859 19160 5969 19161 19161 19162 2 19163 264 19164 15936 19165 15937 19166 19166 19167 19167 19168 15940 19169 5969 19170 19170 19171 22 19172 143 19173 146 19174 5969 19175 19175 19176 123 19177 154 19178 19178 19179 19179 19180 11 19181 75 19182 19182 19183 123 19184 154 19185 19185 19186 13 19187 16110 19188 146 19189 75 19190 123 19191 154 19192 19192 19193 2 19194 264 19195 16123 19196 16124 19197 19197 19198 16126 19199 16127 19200 16128 19201 19201 19202 16128 19203 19203 19204 16263 19205 16264 19206 5969 19207 19207 19208 16264 19209 5969 19210 19210 19211 19211 19212 2965 19213 19213 19214 5969 19215 19215 19216 5969 19217 19217 19218 2965 19219 19219 19220 2 19221 17482 19222 18834 19223 18844 19224 15936 19225 18834 19226 146 19227 15936 19228 18894 19229 19229 19230 19230 19231 18897 19232 5969 19233 19233 19234 19234 19235 19057 19236 5969 19237 19057 19238 5969 19239 19065 19240 146 19241 15936 19242 19072 19243 19243 19244 19244 19245 19075 19246 5969 19247 19247 19248 19248 19249 2965 19250 19250 19251 2 19252 19252 19253 19253 19254 19254 19255 15 19256 15 19257 11 19258 26 19259 2 19260 11 19261 11 19262 40 19263 16 19264 11 19265 19265 19266 15 19267 68 19268 69 19269 11 19270 75 19271 11 19272 11 19273 19273 19274 86 19275 101 19276 19276 19277 19277 19278 19278 19279 123 19280 154 19281 19281 19282 11 19283 2 19284 264 19285 5969 19286 19286 19287 26 19288 11 19289 2 19290 264 19291 26 19292 19292 19293 15 19294 15 19295 11 19296 26 19297 2 19298 11 19299 11 19300 40 19301 16 19302 19302 19303 15 19304 68 19305 69 19306 11 19307 75 19308 11 19309 11 19310 19310 19311 86 19312 101 19313 19313 19314 19277 19315 19315 19316 123 19317 154 19318 19318 19319 26 19320 19320 19321 19321 19322 1039 19323 5969 19324 19324 19325 15858 19326 15859 19327 5969 19328 19328 19329 2 19330 264 19331 15936 19332 15937 19333 19333 19334 19334 19335 15940 19336 5969 19337 19337 19338 22 19339 143 19340 146 19341 5969 19342 19342 19343 123 19344 154 19345 19345 19346 19346 19347 11 19348 75 19349 19349 19350 123 19351 154 19352 19352 19353 13 19354 16110 19355 146 19356 75 19357 123 19358 154 19359 19359 19360 2 19361 264 19362 16123 19363 16124 19364 19364 19365 16126 19366 16127 19367 16128 19368 19368 19369 16128 19370 19370 19371 16263 19372 16264 19373 5969 19374 19374 19375 16264 19376 5969 19377 19377 19378 19378 19379 2965 19380 19380 19381 5969 19382 19382 19383 5969 19384 19384 19385 2965 19386 19386 19387 2 19388 17482 19389 18834 19390 18844 19391 15936 19392 18834 19393 146 19394 15936 19395 18894 19396 19396 19397 19397 19398 18897 19399 5969 19400 19400 19401 19401 19402 19057 19403 5969 19404 19057 19405 5969 19406 19065 19407 146 19408 15936 19409 19072 19410 19410 19411 19411 19412 19075 19413 5969 19414 19414 19415 2965 19416 19416 19417 19278 19418 11 19419 2 19420 264 19421 5969 19422 19422 19423 16 19424 2011 19425 19425 19426 19426 19427 15936 19428 16 19429 19429 19430 19430 19431 19431 19432 19432 19433 19433 19434 19434 19435 15 19436 16 19437 15 19438 11 19439 26 19440 2 19441 5969 19442 16 19443 11 19444 11 19445 40 19446 16 19447 11 19448 11 19449 19449 19450 16 19451 15 19452 68 19453 69 19454 11 19455 16 19456 75 19457 11 19458 11 19459 19459 19460 86 19461 19461 19462 90 19463 19463 19464 96 19465 19465 19466 101 19467 104 19468 19468 19469 19469 19470 19470 19471 19471 19472 19472 19473 19473 19474 19474 19475 123 19476 154 19477 19477 19478 26 19479 11 19480 2 19481 264 19482 26 19483 19483 19484 15 19485 15 19486 11 19487 26 19488 2 19489 11 19490 11 19491 40 19492 16 19493 19493 19494 15 19495 68 19496 69 19497 11 19498 75 19499 11 19500 11 19501 19501 19502 86 19503 101 19504 19504 19505 19470 19506 19506 19507 123 19508 154 19509 19509 19510 26 19511 19511 19512 19512 19513 1039 19514 5969 19515 19515 19516 15858 19517 15859 19518 5969 19519 19519 19520 2 19521 264 19522 15936 19523 15937 19524 19524 19525 19525 19526 15940 19527 5969 19528 19528 19529 22 19530 143 19531 146 19532 5969 19533 19533 19534 123 19535 154 19536 19536 19537 19537 19538 11 19539 75 19540 19540 19541 123 19542 154 19543 19543 19544 13 19545 16110 19546 146 19547 75 19548 123 19549 154 19550 19550 19551 2 19552 264 19553 16123 19554 16124 19555 19555 19556 16126 19557 16127 19558 16128 19559 19559 19560 16128 19561 19561 19562 16263 19563 16264 19564 5969 19565 19565 19566 16264 19567 5969 19568 19568 19569 19569 19570 2965 19571 19571 19572 5969 19573 19573 19574 5969 19575 19575 19576 2965 19577 19577 19578 2 19579 17482 19580 18834 19581 18844 19582 15936 19583 18834 19584 146 19585 15936 19586 18894 19587 19587 19588 19588 19589 18897 19590 5969 19591 19591 19592 19592 19593 19057 19594 5969 19595 19057 19596 5969 19597 19065 19598 146 19599 15936 19600 19072 19601 19601 19602 19602 19603 19075 19604 5969 19605 19605 19606 2965 19607 19607 19608 19278 19609 11 19610 2 19611 264 19612 5969 19613 19613 19614 16 19615 19426 19616 15936 19617 19430 19618 19618 19619 19619 19620 19433 19621 5969 19622 19622 19623 5969 19624 19624 19625 5969 19626 19626 19627 19627 19628 15828 19629 136 19630 11 19631 138 19632 146 19633 149 19634 151 19635 75 19636 123 19637 154 19638 19638 19639 19639 19640 19640 19641 16 19642 2901 19643 2903 19644 2905 19645 75 19646 2907 19647 2908 19648 19648 19649 19649 19650 1039 19651 5969 19652 19652 19653 15858 19654 15859 19655 5969 19656 19656 19657 2 19658 264 19659 15936 19660 15937 19661 19661 19662 19662 19663 15940 19664 5969 19665 19665 19666 22 19667 143 19668 146 19669 5969 19670 19670 19671 123 19672 154 19673 19673 19674 19674 19675 11 19676 75 19677 19677 19678 123 19679 154 19680 19680 19681 13 19682 16110 19683 146 19684 75 19685 123 19686 154 19687 19687 19688 2905 19689 75 19690 2907 19691 2908 19692 19692 19693 2 19694 264 19695 16123 19696 16124 19697 19697 19698 16126 19699 16127 19700 16128 19701 19701 19702 16128 19703 19703 19704 16263 19705 16264 19706 5969 19707 19707 19708 16264 19709 5969 19710 19710 19711 19711 19712 2965 19713 19713 19714 5969 19715 19715 19716 19716 19717 13 19718 2905 19719 75 19720 2907 19721 2908 19722 19722 19723 19723 19724 2 19725 264 19726 15936 19727 15773 19728 19728 19729 19729 19730 146 19731 5969 19732 19732 19733 5969 19734 19734 19735 2965 19736 19736 19737 2 19738 17482 19739 18834 19740 18844 19741 15936 19742 18834 19743 146 19744 15936 19745 18894 19746 19746 19747 19747 19748 18897 19749 5969 19750 19750 19751 19751 19752 19057 19753 5969 19754 19057 19755 5969 19756 19065 19757 146 19758 15936 19759 19072 19760 19760 19761 19761 19762 19075 19763 5969 19764 19764 19765 19765 19766 150 19767 2903 19768 2905 19769 75 19770 2907 19771 2908 19772 19772 19773 2214 19774 19774 19775 19775 19776 19776 19777 19777 19778 19778 19779 19779 19780 19780 19781 19781 19782 19782 19783 19783 19784 19784 19785 19785 19786 19786 19787 2214 19788 19788 19789 19789 19790 19790 19791 19791 19792 19792 19793 19793 19794 19794 19795 19795 19796 19796 19797 19797 19798 2214 19799 2214 19800 19800 19801 19801 19802 19802 19803 19803 19804 19804 19805 19805 19806 19806 19807 19807 19808 19808 19809 19809 19810 19810 19811 2214 19812 19812 19813 19813 19814 19814 19815 19815 19816 19816 19817 19817 19818 19818 19819 19819 19820 2965 19821 19821 19822 19278 19823 11 19824 2 19825 264 19826 5969 19827 19827 19828 16 19829 19426 19830 15936 19831 19430 19832 19832 19833 19833 19834 19433 19835 5969 19836 19836 19837 5969 19838 19838 19839 2901 19840 2903 19841 2905 19842 75 19843 2907 19844 2908 19845 19845 19846 5969 19847 19847 19848 2 19849 19849 19850 19850 19851 5969 19852 19852 19853 5969 19854 19854 19855 19855 19856 19856 19857 1039 19858 5969 19859 19859 19860 15858 19861 15859 19862 5969 19863 19863 19864 2 19865 264 19866 15936 19867 15937 19868 19868 19869 19869 19870 15940 19871 5969 19872 19872 19873 22 19874 143 19875 146 19876 5969 19877 19877 19878 123 19879 154 19880 19880 19881 19881 19882 11 19883 75 19884 19884 19885 123 19886 154 19887 19887 19888 13 19889 16110 19890 146 19891 75 19892 123 19893 154 19894 19894 19895 2 19896 264 19897 16123 19898 16124 19899 19899 19900 16126 19901 16127 19902 16128 19903 19903 19904 16128 19905 19905 19906 16263 19907 16264 19908 5969 19909 19909 19910 16264 19911 5969 19912 19912 19913 19913 19914 2965 19915 19915 19916 5969 19917 19917 19918 2 19919 264 19920 15936 19921 19921 19922 5969 19923 19923 19924 2965 19925 19925 19926 2 19927 17482 19928 18834 19929 18844 19930 15936 19931 18834 19932 146 19933 15936 19934 18894 19935 19935 19936 19936 19937 18897 19938 5969 19939 19939 19940 19940 19941 19057 19942 5969 19943 19057 19944 5969 19945 19065 19946 146 19947 15936 19948 19072 19949 19949 19950 19950 19951 19075 19952 5969 19953 19953 19954 2965 19955 19955 19956 19278 19957 11 19958 2 19959 264 19960 5969 19961 19961 19962 16 19963 19426 19964 15936 19965 19430 19966 19966 19967 19967 19968 19433 19969 5969 19970 19970 19971 5969 19972 19972 19973 5969 19974 19974 19975 16 19976 19426 19977 15936 19978 19430 19979 19979 19980 19980 19981 19433 19982 5969 19983 19983 19984 5969 19985 19985 19986 5969 19987 19987 19988 19988 19989 15828 19990 136 19991 11 19992 138 19993 146 19994 149 19995 151 19996 75 19997 123 19998 154 19999 19999 20000 20000 20001 20001 20002 16 20003 2901 20004 2903 20005 2905 20006 75 20007 2907 20008 2908 20009 20009 20010 20010 20011 1039 20012 5969 20013 20013 20014 15858 20015 15859 20016 5969 20017 20017 20018 2 20019 264 20020 15936 20021 15937 20022 20022 20023 20023 20024 15940 20025 5969 20026 20026 20027 22 20028 143 20029 146 20030 5969 20031 20031 20032 123 20033 154 20034 20034 20035 20035 20036 11 20037 75 20038 20038 20039 123 20040 154 20041 20041 20042 13 20043 16110 20044 146 20045 75 20046 123 20047 154 20048 20048 20049 2905 20050 75 20051 2907 20052 2908 20053 20053 20054 2 20055 264 20056 16123 20057 16124 20058 20058 20059 16126 20060 16127 20061 16128 20062 20062 20063 16128 20064 20064 20065 16263 20066 16264 20067 5969 20068 20068 20069 16264 20070 5969 20071 20071 20072 20072 20073 2965 20074 20074 20075 5969 20076 20076 20077 19716 20078 2905 20079 75 20080 2907 20081 2908 20082 20082 20083 2 20084 264 20085 15936 20086 20086 20087 146 20088 5969 20089 20089 20090 5969 20091 20091 20092 2965 20093 20093 20094 2 20095 17482 20096 18834 20097 18844 20098 15936 20099 18834 20100 146 20101 15936 20102 18894 20103 20103 20104 20104 20105 18897 20106 5969 20107 20107 20108 20108 20109 19057 20110 5969 20111 19057 20112 5969 20113 19065 20114 146 20115 15936 20116 19072 20117 20117 20118 20118 20119 19075 20120 5969 20121 20121 20122 19765 20123 2903 20124 2905 20125 75 20126 2907 20127 2908 20128 20128 20129 20129 20130 2965 20131 20131 20132 19278 20133 11 20134 2 20135 264 20136 5969 20137 20137 20138 16 20139 19426 20140 15936 20141 19430 20142 20142 20143 20143 20144 19433 20145 5969 20146 20146 20147 5969 20148 20148 20149 2901 20150 2903 20151 2905 20152 75 20153 2907 20154 2908 20155 20155 20156 5969 20157 20157 20158 2965 20159 20159 20160 19278 20161 11 20162 2 20163 264 20164 5969 20165 20165 20166 16 20167 19426 20168 15936 20169 19430 20170 20170 20171 20171 20172 19433 20173 5969 20174 20174 20175 5969 20176 20176 20177 5969 20178 20178 20179 20179 20180 15828 20181 136 20182 11 20183 138 20184 146 20185 149 20186 151 20187 75 20188 123 20189 154 20190 20190 20191 20191 20192 20192 20193 16 20194 2901 20195 2903 20196 2905 20197 75 20198 2907 20199 2908 20200 20200 20201 20201 20202 1039 20203 5969 20204 20204 20205 15858 20206 15859 20207 5969 20208 20208 20209 2 20210 264 20211 15936 20212 15937 20213 20213 20214 20214 20215 15940 20216 5969 20217 20217 20218 22 20219 143 20220 146 20221 5969 20222 20222 20223 123 20224 154 20225 20225 20226 20226 20227 11 20228 75 20229 20229 20230 123 20231 154 20232 20232 20233 13 20234 16110 20235 146 20236 75 20237 123 20238 154 20239 20239 20240 2905 20241 75 20242 2907 20243 2908 20244 20244 20245 2 20246 264 20247 16123 20248 16124 20249 20249 20250 16126 20251 16127 20252 16128 20253 20253 20254 16128 20255 20255 20256 16263 20257 16264 20258 5969 20259 20259 20260 16264 20261 5969 20262 20262 20263 20263 20264 2965 20265 20265 20266 5969 20267 20267 20268 19716 20269 2905 20270 75 20271 2907 20272 2908 20273 20273 20274 2 20275 264 20276 15936 20277 20277 20278 146 20279 5969 20280 20280 20281 5969 20282 20282 20283 2965 20284 20284 20285 2 20286 17482 20287 18834 20288 18844 20289 15936 20290 18834 20291 146 20292 15936 20293 18894 20294 20294 20295 20295 20296 18897 20297 5969 20298 20298 20299 20299 20300 19057 20301 5969 20302 19057 20303 5969 20304 19065 20305 146 20306 15936 20307 19072 20308 20308 20309 20309 20310 19075 20311 5969 20312 20312 20313 19765 20314 2903 20315 2905 20316 75 20317 2907 20318 2908 20319 20319 20320 20320 20321 2965 20322 20322 20323 19278 20324 11 20325 2 20326 264 20327 5969 20328 20328 20329 16 20330 19426 20331 15936 20332 19430 20333 20333 20334 20334 20335 19433 20336 5969 20337 20337 20338 5969 20339 20339 20340 2901 20341 2903 20342 2905 20343 75 20344 2907 20345 2908 20346 20346 20347 5969 20348 20348 20349 830 20350 11 20351 20351 20352 20352 20353 20353 20354 338 20355 338 20356 341 20357 2 20358 404 20359 20359 20360 555 20361 20361 20362 20362 20363 20363 20364 20364 20365 2 20366 404 20367 20367 20368 20368 20369 20369 20370 20370 20371 20371 20372 555 20373 20373 20374 1823 20375 146 20376 20376 20377 20377 20378 2 20379 404 20380 20380 20381 20381 20382 830 20383 11 20384 20384 20385 20385 20386 20386 20387 338 20388 338 20389 341 20390 2 20391 404 20392 20392 20393 555 20394 20394 20395 20395 20396 20396 20397 20397 20398 2 20399 404 20400 20400 20401 20401 20402 20402 20403 20403 20404 20404 20405 555 20406 20406 20407 1823 20408 146 20409 20409 20410 20410 20411 2 20412 404 20413 20413 20414 830 20415 18171 20416 11 20417 18173 20418 18174 20419 18175 20420 2653 20421 18177 20422 18178 20423 338 20424 338 20425 341 20426 2 20427 404 20428 18184 20429 555 20430 18186 20431 18187 20432 18188 20433 18189 20434 2 20435 404 20436 18192 20437 18193 20438 18194 20439 18195 20440 2728 20441 18197 20442 555 20443 18199 20444 2 20445 18201 20446 18202 20447 2 20448 3 20449 4 20450 2 20451 2 20452 2 20453 2 20454 11 20455 2 20456 2 20457 18213 20458 2 20459 18215 20460 22 20461 18217 20462 18218 20463 2 20464 11 20465 11 20466 18222 20467 18223 20468 18224 20469 22 20470 2 20471 2 20472 2 20473 11 20474 18230 20475 11 20476 18232 20477 11 20478 18234 20479 18235 20480 18236 20481 18237 20482 60 20483 18239 20484 2 20485 2 20486 2 20487 2 20488 67 20489 127 20490 18246 20491 2 20492 2 20493 18249 20494 2 20495 18251 20496 18252 20497 2 20498 83 20499 84 20500 87 20501 130 20502 11 20503 18259 20504 18260 20505 97 20506 18262 20507 18263 20508 18264 20509 102 20510 102 20511 18267 20512 18268 20513 18269 20514 18270 20515 18271 20516 18272 20517 18273 20518 18274 20519 18275 20520 18276 20521 97 20522 2 20523 18279 20524 2 20525 2 20526 2 20527 2 20528 2 20529 2 20530 2 20531 2 20532 288 20533 2 20534 2 20535 2 20536 2 20537 2 20538 294 20539 2 20540 2 20541 2 20542 2 20543 2 20544 2 20545 2 20546 2 20547 2 20548 2 20549 2 20550 2 20551 2 20552 2 20553 2 20554 18310 20555 18311 20556 18312 20557 18313 20558 18314 20559 18315 20560 18316 20561 18317 20562 18318 20563 18319 20564 18320 20565 18321 20566 18322 20567 18323 20568 18324 20569 4 20570 2 20571 18327 20572 18328 20573 18329 20574 18330 20575 18331 20576 18332 20577 18333 20578 18334 20579 18335 20580 18336 20581 18337 20582 338 20583 338 20584 40 20585 341 20586 2 20587 404 20588 18344 20589 555 20590 18346 20591 18347 20592 18348 20593 18349 20594 18350 20595 2 20596 404 20597 18353 20598 18354 20599 40 20600 18356 20601 264 20602 18358 20603 18359 20604 2 20605 404 20606 18362 20607 18363 20608 18364 20609 18365 20610 18366 20611 18367 20612 2 20613 713 20614 2 20615 18371 20616 18372 20617 18373 20618 18374 20619 18375 20620 18376 20621 18377 20622 18378 20623 18379 20624 18380 20625 830 20626 18382 20627 18383 20628 18384 20629 18385 20630 18386 20631 18387 20632 18388 20633 18389 20634 18390 20635 18391 20636 18392 20637 18393 20638 18394 20639 18395 20640 18396 20641 18397 20642 18398 20643 18399 20644 18400 20645 18401 20646 18402 20647 18403 20648 18404 20649 18405 20650 18406 20651 18407 20652 18408 20653 18409 20654 18410 20655 18411 20656 18412 20657 18413 20658 18414 20659 18415 20660 18416 20661 18417 20662 18418 20663 18419 20664 18420 20665 18421 20666 18422 20667 18423 20668 18424 20669 18425 20670 18426 20671 18427 20672 18428 20673 18429 20674 18430 20675 18431 20676 18432 20677 18433 20678 18434 20679 18435 20680 18436 20681 18437 20682 18438 20683 18439 20684 18440 20685 18441 20686 18442 20687 18443 20688 18444 20689 18445 20690 18446 20691 18447 20692 18448 20693 18449 20694 18450 20695 18451 20696 18452 20697 18453 20698 18454 20699 18455 20700 18456 20701 18457 20702 18458 20703 18459 20704 18460 20705 11 20706 18462 20707 18463 20708 18464 20709 917 20710 18466 20711 18467 20712 950 20713 143 20714 146 20715 18471 20716 18472 20717 1456 20718 2 20719 2 20720 2 20721 18477 20722 2 20723 18479 20724 2 20725 2 20726 2 20727 2 20728 2 20729 2 20730 2 20731 2 20732 2 20733 18489 20734 18490 20735 18491 20736 18492 20737 18493 20738 18494 20739 18495 20740 18496 20741 18497 20742 18498 20743 18499 20744 18500 20745 18501 20746 22 20747 18503 20748 4 20749 18505 20750 87 20751 11 20752 18508 20753 97 20754 18510 20755 1517 20756 18512 20757 18513 20758 18514 20759 97 20760 830 20761 11 20762 18518 20763 917 20764 18520 20765 18521 20766 950 20767 143 20768 146 20769 18525 20770 18526 20771 18527 20772 830 20773 11 20774 18530 20775 18531 20776 18532 20777 338 20778 338 20779 341 20780 2 20781 404 20782 18538 20783 555 20784 18540 20785 18541 20786 18542 20787 18543 20788 2 20789 404 20790 18546 20791 18547 20792 18548 20793 18549 20794 18550 20795 2 20796 404 20797 18553 20798 18554 20799 136 20800 11 20801 138 20802 146 20803 147 20804 149 20805 151 20806 75 20807 123 20808 154 20809 18565 20810 18566 20811 136 20812 11 20813 138 20814 146 20815 2206 20816 2208 20817 2209 20818 2210 20819 2 20820 18576 20821 18577 20822 2 20823 18579 20824 830 20825 18581 20826 11 20827 18583 20828 18584 20829 18585 20830 2653 20831 18587 20832 18588 20833 338 20834 338 20835 341 20836 2 20837 404 20838 18594 20839 555 20840 18596 20841 18597 20842 18598 20843 18599 20844 2 20845 404 20846 18602 20847 18603 20848 18604 20849 18605 20850 2728 20851 18607 20852 555 20853 18609 20854 2 20855 18611 20856 18612 20857 18264 20858 2 20859 18615 20860 18616 20861 18617 20862 18618 20863 18619 20864 2820 20865 18621 20866 67 20867 18623 20868 18624 20869 18625 20870 18626 20871 18627 20872 18628 20873 2854 20874 18636 20875 2863 20876 18639 20877 917 20878 18656 20879 18657 20880 18658 20881 950 20882 143 20883 146 20884 18662 20885 18663 20886 18666 20887 2901 20888 2903 20889 2905 20890 75 20891 2907 20892 2908 20893 18673 20894 2 20895 264 20896 75 20897 3856 20898 2753 20899 18681 20900 1823 20901 146 20902 2011 20903 2 20904 18686 20905 2 20906 143 20907 146 20908 18690 20909 18691 20910 18692 20911 18693 20912 18694 20913 5125 20914 18696 20915 146 20916 18698 20917 2 20918 264 20919 18701 20920 2 20921 264 20922 18704 20923 18705 20924 18706 20925 18707 20926 18708 20927 18709 20928 1823 20929 146 20930 18712 20931 18713 20932 18714 20933 18715 20934 264 20935 18717 20936 18718 20937 2 20938 404 20939 18721 20940 5157 20941 18723 20942 18724 20943 18725 20944 18726 20945 18727 20946 18728 20947 2820 20948 18730 20949 67 20950 18732 20951 18733 20952 18734 20953 18735 20954 18736 20955 18737 20956 2854 20957 18745 20958 2863 20959 18748 20960 917 20961 18765 20962 18766 20963 18767 20964 950 20965 143 20966 146 20967 18771 20968 18772 20969 18775 20970 2901 20971 2903 20972 2905 20973 75 20974 2907 20975 2908 20976 18782 20977 2 20978 264 20979 75 20980 3856 20981 2753 20982 18790 20983 1823 20984 146 20985 2011 20986 2 20987 18795 20988 2 20989 143 20990 146 20991 18799 20992 18800 20993 18801 20994 18802 20995 18803 20996 5125 20997 18805 20998 146 20999 18807 21000 2 21001 264 21002 18810 21003 2 21004 264 21005 18813 21006 18814 21007 18815 21008 18816 21009 18817 21010 18818 21011 1823 21012 146 21013 18821 21014 18822 21015 18823 21016 18824 21017 264 21018 18826 21019 18827 21020 2 21021 404 21022 18830 21023 5157 21024 18832 21025 18833 21026 19057 21027 5969 21028 19057 21029 5969 21030 19065 21031 146 21032 15936 21033 19072 21034 21034 21035 21035 21036 19075 21037 5969 21038 21038 21039 2965 21040 21040 21041 19278 21042 11 21043 2 21044 264 21045 5969 21046 21046 21047 16 21048 19426 21049 15936 21050 19430 21051 21051 21052 21052 21053 19433 21054 5969 21055 21055 21056 5969 21057 21057 21058 5969 21059 21059 21060 21060 21061 15828 21062 136 21063 11 21064 138 21065 146 21066 149 21067 151 21068 75 21069 123 21070 154 21071 21071 21072 21072 21073 21073 21074 16 21075 2901 21076 2903 21077 2905 21078 75 21079 2907 21080 2908 21081 21081 21082 21082 21083 1039 21084 5969 21085 21085 21086 15858 21087 15859 21088 5969 21089 21089 21090 2 21091 264 21092 15936 21093 15937 21094 21094 21095 21095 21096 15940 21097 5969 21098 21098 21099 22 21100 143 21101 146 21102 5969 21103 21103 21104 123 21105 154 21106 21106 21107 21107 21108 11 21109 75 21110 21110 21111 123 21112 154 21113 21113 21114 13 21115 16110 21116 146 21117 75 21118 123 21119 154 21120 21120 21121 2905 21122 75 21123 2907 21124 2908 21125 21125 21126 2 21127 264 21128 16123 21129 16124 21130 21130 21131 16126 21132 16127 21133 16128 21134 21134 21135 16128 21136 21136 21137 16263 21138 16264 21139 5969 21140 21140 21141 16264 21142 5969 21143 21143 21144 21144 21145 2965 21146 21146 21147 5969 21148 21148 21149 19716 21150 2905 21151 75 21152 2907 21153 2908 21154 21154 21155 2 21156 264 21157 15936 21158 21158 21159 146 21160 5969 21161 21161 21162 5969 21163 21163 21164 2965 21165 21165 21166 2 21167 17482 21168 18834 21169 18844 21170 15936 21171 18834 21172 146 21173 15936 21174 18894 21175 21175 21176 21176 21177 18897 21178 5969 21179 21179 21180 21180 21181 19057 21182 5969 21183 19057 21184 5969 21185 19065 21186 146 21187 15936 21188 19072 21189 21189 21190 21190 21191 19075 21192 5969 21193 21193 21194 19765 21195 2903 21196 2905 21197 75 21198 2907 21199 2908 21200 21200 21201 2965 21202 21202 21203 19278 21204 11 21205 2 21206 264 21207 5969 21208 21208 21209 16 21210 19426 21211 15936 21212 19430 21213 21213 21214 21214 21215 19433 21216 5969 21217 21217 21218 5969 21219 21219 21220 2901 21221 2903 21222 2905 21223 75 21224 2907 21225 2908 21226 21226 21227 5969 21228 21228 21229 2 21230 3 21231 4 21232 2 21233 2 21234 2 21235 2 21236 11 21237 2 21238 2 21239 21239 21240 2 21241 21241 21242 22 21243 21243 21244 21244 21245 2 21246 11 21247 11 21248 21248 21249 21249 21250 21250 21251 22 21252 2 21253 2 21254 2 21255 11 21256 21256 21257 11 21258 21258 21259 11 21260 21260 21261 21261 21262 21262 21263 21263 21264 60 21265 21265 21266 2 21267 2 21268 2 21269 2 21270 67 21271 127 21272 21272 21273 2 21274 2 21275 21275 21276 2 21277 21277 21278 21278 21279 2 21280 83 21281 84 21282 87 21283 130 21284 11 21285 21285 21286 21286 21287 97 21288 21288 21289 21289 21290 21290 21291 102 21292 102 21293 21293 21294 21294 21295 21295 21296 21296 21297 21297 21298 21298 21299 21299 21300 21300 21301 21301 21302 21302 21303 97 21304 2 21305 21305 21306 2 21307 2 21308 2 21309 2 21310 2 21311 2 21312 2 21313 2 21314 288 21315 2 21316 2 21317 2 21318 2 21319 2 21320 294 21321 2 21322 2 21323 2 21324 2 21325 2 21326 2 21327 2 21328 2 21329 2 21330 2 21331 2 21332 2 21333 2 21334 2 21335 2 21336 21336 21337 21337 21338 21338 21339 21339 21340 21340 21341 21341 21342 21342 21343 21343 21344 21344 21345 21345 21346 21346 21347 21347 21348 21348 21349 21349 21350 21350 21351 4 21352 2 21353 21353 21354 21354 21355 21355 21356 21356 21357 21357 21358 21358 21359 21359 21360 21360 21361 21361 21362 21362 21363 21363 21364 338 21365 338 21366 40 21367 341 21368 2 21369 404 21370 21370 21371 555 21372 21372 21373 21373 21374 21374 21375 21375 21376 21376 21377 2 21378 404 21379 21379 21380 21380 21381 40 21382 21382 21383 264 21384 21384 21385 21385 21386 2 21387 404 21388 21388 21389 21389 21390 21390 21391 21391 21392 21392 21393 21393 21394 2 21395 713 21396 2 21397 21397 21398 21398 21399 21399 21400 21400 21401 21401 21402 21402 21403 21403 21404 21404 21405 21405 21406 21406 21407 830 21408 21408 21409 21409 21410 21410 21411 21411 21412 21412 21413 21413 21414 21414 21415 21415 21416 21416 21417 21417 21418 21418 21419 21419 21420 21420 21421 21421 21422 21422 21423 21423 21424 21424 21425 21425 21426 21426 21427 21427 21428 21428 21429 21429 21430 21430 21431 21431 21432 21432 21433 21433 21434 21434 21435 21435 21436 21436 21437 21437 21438 21438 21439 21439 21440 21440 21441 21441 21442 21442 21443 21443 21444 21444 21445 21445 21446 21446 21447 21447 21448 21448 21449 21449 21450 21450 21451 21451 21452 21452 21453 21453 21454 21454 21455 21455 21456 21456 21457 21457 21458 21458 21459 21459 21460 21460 21461 21461 21462 21462 21463 21463 21464 21464 21465 21465 21466 21466 21467 21467 21468 21468 21469 21469 21470 21470 21471 21471 21472 21472 21473 21473 21474 21474 21475 21475 21476 21476 21477 21477 21478 21478 21479 21479 21480 21480 21481 21481 21482 21482 21483 21483 21484 21484 21485 21485 21486 21486 21487 11 21488 21488 21489 21489 21490 21490 21491 917 21492 21492 21493 21493 21494 950 21495 143 21496 146 21497 21497 21498 21498 21499 1456 21500 2 21501 2 21502 2 21503 21503 21504 2 21505 21505 21506 2 21507 2 21508 2 21509 2 21510 2 21511 2 21512 2 21513 2 21514 2 21515 21515 21516 21516 21517 21517 21518 21518 21519 21519 21520 21520 21521 21521 21522 21522 21523 21523 21524 21524 21525 21525 21526 21526 21527 21527 21528 22 21529 21529 21530 4 21531 21531 21532 87 21533 11 21534 21534 21535 97 21536 21536 21537 1517 21538 21538 21539 21539 21540 21540 21541 97 21542 830 21543 11 21544 21544 21545 917 21546 21546 21547 21547 21548 950 21549 143 21550 146 21551 21551 21552 21552 21553 21553 21554 830 21555 11 21556 21556 21557 21557 21558 21558 21559 338 21560 338 21561 341 21562 2 21563 404 21564 21564 21565 555 21566 21566 21567 21567 21568 21568 21569 21569 21570 2 21571 404 21572 21572 21573 21573 21574 21574 21575 21575 21576 21576 21577 2 21578 404 21579 21579 21580 21580 21581 136 21582 11 21583 138 21584 146 21585 147 21586 149 21587 151 21588 75 21589 123 21590 154 21591 21591 21592 21592 21593 136 21594 11 21595 138 21596 146 21597 2206 21598 2208 21599 2209 21600 2210 21601 2 21602 2 21603 3 21604 4 21605 2 21606 2 21607 2 21608 2 21609 11 21610 2 21611 2 21612 21239 21613 2 21614 21241 21615 22 21616 21243 21617 21244 21618 2 21619 11 21620 11 21621 21248 21622 21249 21623 21250 21624 22 21625 2 21626 2 21627 2 21628 11 21629 21256 21630 11 21631 21258 21632 11 21633 21260 21634 21261 21635 21262 21636 21263 21637 60 21638 21265 21639 2 21640 2 21641 2 21642 2 21643 67 21644 127 21645 21272 21646 2 21647 2 21648 21275 21649 2 21650 21277 21651 21278 21652 2 21653 83 21654 84 21655 87 21656 130 21657 11 21658 21285 21659 21286 21660 97 21661 21288 21662 21289 21663 21663 21664 102 21665 102 21666 21293 21667 21294 21668 21295 21669 21296 21670 21297 21671 21298 21672 21299 21673 21300 21674 21301 21675 21302 21676 97 21677 2 21678 21305 21679 2 21680 2 21681 2 21682 2 21683 2 21684 2 21685 2 21686 2 21687 288 21688 2 21689 2 21690 2 21691 2 21692 2 21693 294 21694 2 21695 2 21696 2 21697 2 21698 2 21699 2 21700 2 21701 2 21702 2 21703 2 21704 2 21705 2 21706 2 21707 2 21708 2 21709 21336 21710 21337 21711 21338 21712 21339 21713 21340 21714 21341 21715 21342 21716 21343 21717 21344 21718 21345 21719 21346 21720 21347 21721 21348 21722 21349 21723 21350 21724 4 21725 2 21726 21353 21727 21354 21728 21355 21729 21356 21730 21357 21731 21358 21732 21359 21733 21360 21734 21361 21735 21362 21736 21363 21737 338 21738 338 21739 40 21740 341 21741 2 21742 404 21743 21370 21744 555 21745 21372 21746 21373 21747 21374 21748 21375 21749 21376 21750 2 21751 404 21752 21379 21753 21380 21754 40 21755 21382 21756 264 21757 21384 21758 21385 21759 2 21760 404 21761 21388 21762 21389 21763 21390 21764 21391 21765 21392 21766 21393 21767 2 21768 713 21769 2 21770 21397 21771 21398 21772 21399 21773 21400 21774 21401 21775 21402 21776 21403 21777 21404 21778 21405 21779 21406 21780 830 21781 21408 21782 21409 21783 21410 21784 21411 21785 21412 21786 21413 21787 21414 21788 21415 21789 21416 21790 21417 21791 21418 21792 21419 21793 21420 21794 21421 21795 21422 21796 21423 21797 21424 21798 21425 21799 21426 21800 21427 21801 21428 21802 21429 21803 21430 21804 21431 21805 21432 21806 21433 21807 21434 21808 21435 21809 21436 21810 21437 21811 21438 21812 21439 21813 21440 21814 21441 21815 21442 21816 21443 21817 21444 21818 21445 21819 21446 21820 21447 21821 21448 21822 21449 21823 21450 21824 21451 21825 21452 21826 21453 21827 21454 21828 21455 21829 21456 21830 21457 21831 21458 21832 21459 21833 21460 21834 21461 21835 21462 21836 21463 21837 21464 21838 21465 21839 21466 21840 21467 21841 21468 21842 21469 21843 21470 21844 21471 21845 21472 21846 21473 21847 21474 21848 21475 21849 21476 21850 21477 21851 21478 21852 21479 21853 21480 21854 21481 21855 21482 21856 21483 21857 21484 21858 21485 21859 21486 21860 11 21861 21488 21862 21489 21863 21490 21864 917 21865 21492 21866 21493 21867 950 21868 143 21869 146 21870 21497 21871 21498 21872 1456 21873 2 21874 2 21875 2 21876 21503 21877 2 21878 21505 21879 2 21880 2 21881 2 21882 2 21883 2 21884 2 21885 2 21886 2 21887 2 21888 21515 21889 21516 21890 21517 21891 21518 21892 21519 21893 21520 21894 21521 21895 21522 21896 21523 21897 21524 21898 21525 21899 21526 21900 21527 21901 22 21902 21529 21903 4 21904 21531 21905 87 21906 11 21907 21534 21908 97 21909 21536 21910 1517 21911 21538 21912 21539 21913 21540 21914 97 21915 830 21916 11 21917 21544 21918 917 21919 21546 21920 21547 21921 950 21922 143 21923 146 21924 21551 21925 21552 21926 21553 21927 830 21928 11 21929 21556 21930 21557 21931 21558 21932 338 21933 338 21934 341 21935 2 21936 404 21937 21564 21938 555 21939 21566 21940 21567 21941 21568 21942 21569 21943 2 21944 404 21945 21572 21946 21573 21947 21574 21948 21575 21949 21576 21950 2 21951 404 21952 21579 21953 21580 21954 136 21955 11 21956 138 21957 146 21958 147 21959 149 21960 151 21961 75 21962 123 21963 154 21964 21591 21965 21592 21966 136 21967 11 21968 138 21969 146 21970 2206 21971 2208 21972 2209 21973 2210 21974 2 21975 21663 21976 21976 21977 21976 21978 830 21979 21979 21980 11 21981 21981 21982 21982 21983 21983 21984 2653 21985 21985 21986 21986 21987 338 21988 338 21989 341 21990 2 21991 404 21992 21992 21993 555 21994 21994 21995 21995 21996 21996 21997 21997 21998 2 21999 404 22000 22000 22001 22001 22002 22002 22003 22003 22004 2728 22005 22005 22006 555 22007 22007 22008 2 22009 22009 22010 22010 22011 2 22012 3 22013 4 22014 2 22015 2 22016 2 22017 2 22018 11 22019 2 22020 2 22021 22021 22022 2 22023 22023 22024 22 22025 22025 22026 22026 22027 2 22028 11 22029 11 22030 22030 22031 22031 22032 22032 22033 22 22034 2 22035 2 22036 2 22037 11 22038 22038 22039 11 22040 22040 22041 11 22042 22042 22043 22043 22044 22044 22045 22045 22046 60 22047 22047 22048 2 22049 2 22050 2 22051 2 22052 67 22053 127 22054 22054 22055 2 22056 2 22057 22057 22058 2 22059 22059 22060 22060 22061 2 22062 83 22063 84 22064 87 22065 130 22066 11 22067 22067 22068 22068 22069 97 22070 22070 22071 22071 22072 22072 22073 102 22074 102 22075 22075 22076 22076 22077 22077 22078 22078 22079 22079 22080 22080 22081 22081 22082 22082 22083 22083 22084 22084 22085 97 22086 2 22087 22087 22088 2 22089 2 22090 2 22091 2 22092 2 22093 2 22094 2 22095 2 22096 288 22097 2 22098 2 22099 2 22100 2 22101 2 22102 294 22103 2 22104 2 22105 2 22106 2 22107 2 22108 2 22109 2 22110 2 22111 2 22112 2 22113 2 22114 2 22115 2 22116 2 22117 2 22118 22118 22119 22119 22120 22120 22121 22121 22122 22122 22123 22123 22124 22124 22125 22125 22126 22126 22127 22127 22128 22128 22129 22129 22130 22130 22131 22131 22132 22132 22133 4 22134 2 22135 22135 22136 22136 22137 22137 22138 22138 22139 22139 22140 22140 22141 22141 22142 22142 22143 22143 22144 22144 22145 22145 22146 338 22147 338 22148 40 22149 341 22150 2 22151 404 22152 22152 22153 555 22154 22154 22155 22155 22156 22156 22157 22157 22158 22158 22159 2 22160 404 22161 22161 22162 22162 22163 40 22164 22164 22165 264 22166 22166 22167 22167 22168 2 22169 404 22170 22170 22171 22171 22172 22172 22173 22173 22174 22174 22175 22175 22176 2 22177 713 22178 2 22179 22179 22180 22180 22181 22181 22182 22182 22183 22183 22184 22184 22185 22185 22186 22186 22187 22187 22188 22188 22189 830 22190 22190 22191 22191 22192 22192 22193 22193 22194 22194 22195 22195 22196 22196 22197 22197 22198 22198 22199 22199 22200 22200 22201 22201 22202 22202 22203 22203 22204 22204 22205 22205 22206 22206 22207 22207 22208 22208 22209 22209 22210 22210 22211 22211 22212 22212 22213 22213 22214 22214 22215 22215 22216 22216 22217 22217 22218 22218 22219 22219 22220 22220 22221 22221 22222 22222 22223 22223 22224 22224 22225 22225 22226 22226 22227 22227 22228 22228 22229 22229 22230 22230 22231 22231 22232 22232 22233 22233 22234 22234 22235 22235 22236 22236 22237 22237 22238 22238 22239 22239 22240 22240 22241 22241 22242 22242 22243 22243 22244 22244 22245 22245 22246 22246 22247 22247 22248 22248 22249 22249 22250 22250 22251 22251 22252 22252 22253 22253 22254 22254 22255 22255 22256 22256 22257 22257 22258 22258 22259 22259 22260 22260 22261 22261 22262 22262 22263 22263 22264 22264 22265 22265 22266 22266 22267 22267 22268 22268 22269 11 22270 22270 22271 22271 22272 22272 22273 917 22274 22274 22275 22275 22276 950 22277 143 22278 146 22279 22279 22280 22280 22281 1456 22282 2 22283 2 22284 2 22285 22285 22286 2 22287 22287 22288 2 22289 2 22290 2 22291 2 22292 2 22293 2 22294 2 22295 2 22296 2 22297 22297 22298 22298 22299 22299 22300 22300 22301 22301 22302 22302 22303 22303 22304 22304 22305 22305 22306 22306 22307 22307 22308 22308 22309 22309 22310 22 22311 22311 22312 4 22313 22313 22314 87 22315 11 22316 22316 22317 97 22318 22318 22319 1517 22320 22320 22321 22321 22322 22322 22323 97 22324 830 22325 11 22326 22326 22327 917 22328 22328 22329 22329 22330 950 22331 143 22332 146 22333 22333 22334 22334 22335 22335 22336 830 22337 11 22338 22338 22339 22339 22340 22340 22341 338 22342 338 22343 341 22344 2 22345 404 22346 22346 22347 555 22348 22348 22349 22349 22350 22350 22351 22351 22352 2 22353 404 22354 22354 22355 22355 22356 22356 22357 22357 22358 22358 22359 2 22360 404 22361 22361 22362 22362 22363 136 22364 11 22365 138 22366 146 22367 147 22368 149 22369 151 22370 75 22371 123 22372 154 22373 22373 22374 22374 22375 136 22376 11 22377 138 22378 146 22379 2206 22380 2208 22381 2209 22382 2210 22383 2 22384 22384 22385 22385 22386 2 22387 22387 22388 830 22389 22389 22390 11 22391 22391 22392 22392 22393 22393 22394 2653 22395 22395 22396 22396 22397 338 22398 338 22399 341 22400 2 22401 404 22402 22402 22403 555 22404 22404 22405 22405 22406 22406 22407 22407 22408 2 22409 404 22410 22410 22411 22411 22412 22412 22413 22413 22414 2728 22415 22415 22416 555 22417 22417 22418 2 22419 22419 22420 22420 22421 22072 22422 2 22423 22423 22424 22424 22425 22425 22426 22426 22427 22427 22428 2820 22429 22429 22430 67 22431 22431 22432 22432 22433 22433 22434 22434 22435 22435 22436 22436 22437 2 22438 2854 22439 22439 22440 22440 22441 22441 22442 22442 22443 22443 22444 22444 22445 2863 22446 22446 22447 22447 22448 22448 22449 22449 22450 22450 22451 338 22452 338 22453 40 22454 341 22455 2 22456 404 22457 22457 22458 555 22459 22459 22460 22460 22461 22461 22462 22462 22463 917 22464 22464 22465 22465 22466 22466 22467 950 22468 143 22469 146 22470 22470 22471 22471 22472 40 22473 22473 22474 22474 22475 2901 22476 2903 22477 2905 22478 75 22479 2907 22480 2908 22481 22481 22482 264 22483 22483 22484 2 22485 264 22486 75 22487 3856 22488 2753 22489 22489 22490 1823 22491 146 22492 2011 22493 2 22494 22494 22495 2 22496 143 22497 146 22498 22498 22499 22499 22500 22500 22501 22501 22502 22502 22503 5125 22504 22504 22505 146 22506 22506 22507 2 22508 264 22509 22509 22510 2 22511 264 22512 22512 22513 22513 22514 22514 22515 22515 22516 22516 22517 22517 22518 1823 22519 146 22520 22520 22521 22521 22522 22522 22523 22523 22524 264 22525 22525 22526 22526 22527 2 22528 404 22529 22529 22530 5157 22531 22531 22532 22532 22533 22533 22534 22534 22535 22535 22536 22536 22537 2820 22538 22538 22539 67 22540 22540 22541 22541 22542 22542 22543 22543 22544 22544 22545 22545 22546 2 22547 2854 22548 22548 22549 22549 22550 22550 22551 22551 22552 22552 22553 22553 22554 2863 22555 22555 22556 22556 22557 22557 22558 22558 22559 22559 22560 338 22561 338 22562 40 22563 341 22564 2 22565 404 22566 22566 22567 555 22568 22568 22569 22569 22570 22570 22571 22571 22572 917 22573 22573 22574 22574 22575 22575 22576 950 22577 143 22578 146 22579 22579 22580 22580 22581 40 22582 22582 22583 22583 22584 2901 22585 2903 22586 2905 22587 75 22588 2907 22589 2908 22590 22590 22591 264 22592 22592 22593 2 22594 264 22595 75 22596 3856 22597 2753 22598 22598 22599 1823 22600 146 22601 2011 22602 2 22603 22603 22604 2 22605 143 22606 146 22607 22607 22608 22608 22609 22609 22610 22610 22611 22611 22612 5125 22613 22613 22614 146 22615 22615 22616 2 22617 264 22618 22618 22619 2 22620 264 22621 22621 22622 22622 22623 22623 22624 22624 22625 22625 22626 22626 22627 1823 22628 146 22629 22629 22630 22630 22631 22631 22632 22632 22633 264 22634 22634 22635 22635 22636 2 22637 404 22638 22638 22639 5157 22640 22640 22641 22641 22642 16271 22643 22643 22644 830 22645 21979 22646 11 22647 21981 22648 21982 22649 21983 22650 2653 22651 21985 22652 21986 22653 338 22654 338 22655 341 22656 2 22657 404 22658 21992 22659 555 22660 21994 22661 21995 22662 21996 22663 21997 22664 2 22665 404 22666 22000 22667 22001 22668 22002 22669 22003 22670 2728 22671 22005 22672 555 22673 22007 22674 2 22675 22009 22676 22010 22677 2 22678 3 22679 4 22680 2 22681 2 22682 2 22683 2 22684 11 22685 2 22686 2 22687 22021 22688 2 22689 22023 22690 22 22691 22025 22692 22026 22693 2 22694 11 22695 11 22696 22030 22697 22031 22698 22032 22699 22 22700 2 22701 2 22702 2 22703 11 22704 22038 22705 11 22706 22040 22707 11 22708 22042 22709 22043 22710 22044 22711 22045 22712 60 22713 22047 22714 2 22715 2 22716 2 22717 2 22718 67 22719 127 22720 22054 22721 2 22722 2 22723 22057 22724 2 22725 22059 22726 22060 22727 2 22728 83 22729 84 22730 87 22731 130 22732 11 22733 22067 22734 22068 22735 97 22736 22070 22737 22071 22738 22072 22739 102 22740 102 22741 22075 22742 22076 22743 22077 22744 22078 22745 22079 22746 22080 22747 22081 22748 22082 22749 22083 22750 22084 22751 97 22752 2 22753 22087 22754 2 22755 2 22756 2 22757 2 22758 2 22759 2 22760 2 22761 2 22762 288 22763 2 22764 2 22765 2 22766 2 22767 2 22768 294 22769 2 22770 2 22771 2 22772 2 22773 2 22774 2 22775 2 22776 2 22777 2 22778 2 22779 2 22780 2 22781 2 22782 2 22783 2 22784 22118 22785 22119 22786 22120 22787 22121 22788 22122 22789 22123 22790 22124 22791 22125 22792 22126 22793 22127 22794 22128 22795 22129 22796 22130 22797 22131 22798 22132 22799 4 22800 2 22801 22135 22802 22136 22803 22137 22804 22138 22805 22139 22806 22140 22807 22141 22808 22142 22809 22143 22810 22144 22811 22145 22812 338 22813 338 22814 40 22815 341 22816 2 22817 404 22818 22152 22819 555 22820 22154 22821 22155 22822 22156 22823 22157 22824 22158 22825 2 22826 404 22827 22161 22828 22162 22829 40 22830 22164 22831 264 22832 22166 22833 22167 22834 2 22835 404 22836 22170 22837 22171 22838 22172 22839 22173 22840 22174 22841 22175 22842 2 22843 713 22844 2 22845 22179 22846 22180 22847 22181 22848 22182 22849 22183 22850 22184 22851 22185 22852 22186 22853 22187 22854 22188 22855 830 22856 22190 22857 22191 22858 22192 22859 22193 22860 22194 22861 22195 22862 22196 22863 22197 22864 22198 22865 22199 22866 22200 22867 22201 22868 22202 22869 22203 22870 22204 22871 22205 22872 22206 22873 22207 22874 22208 22875 22209 22876 22210 22877 22211 22878 22212 22879 22213 22880 22214 22881 22215 22882 22216 22883 22217 22884 22218 22885 22219 22886 22220 22887 22221 22888 22222 22889 22223 22890 22224 22891 22225 22892 22226 22893 22227 22894 22228 22895 22229 22896 22230 22897 22231 22898 22232 22899 22233 22900 22234 22901 22235 22902 22236 22903 22237 22904 22238 22905 22239 22906 22240 22907 22241 22908 22242 22909 22243 22910 22244 22911 22245 22912 22246 22913 22247 22914 22248 22915 22249 22916 22250 22917 22251 22918 22252 22919 22253 22920 22254 22921 22255 22922 22256 22923 22257 22924 22258 22925 22259 22926 22260 22927 22261 22928 22262 22929 22263 22930 22264 22931 22265 22932 22266 22933 22267 22934 22268 22935 11 22936 22270 22937 22271 22938 22272 22939 917 22940 22274 22941 22275 22942 950 22943 143 22944 146 22945 22279 22946 22280 22947 1456 22948 2 22949 2 22950 2 22951 22285 22952 2 22953 22287 22954 2 22955 2 22956 2 22957 2 22958 2 22959 2 22960 2 22961 2 22962 2 22963 22297 22964 22298 22965 22299 22966 22300 22967 22301 22968 22302 22969 22303 22970 22304 22971 22305 22972 22306 22973 22307 22974 22308 22975 22309 22976 22 22977 22311 22978 4 22979 22313 22980 87 22981 11 22982 22316 22983 97 22984 22318 22985 1517 22986 22320 22987 22321 22988 22322 22989 97 22990 830 22991 11 22992 22326 22993 917 22994 22328 22995 22329 22996 950 22997 143 22998 146 22999 22333 23000 22334 23001 22335 23002 830 23003 11 23004 22338 23005 22339 23006 22340 23007 338 23008 338 23009 341 23010 2 23011 404 23012 22346 23013 555 23014 22348 23015 22349 23016 22350 23017 22351 23018 2 23019 404 23020 22354 23021 22355 23022 22356 23023 22357 23024 22358 23025 2 23026 404 23027 22361 23028 22362 23029 136 23030 11 23031 138 23032 146 23033 147 23034 149 23035 151 23036 75 23037 123 23038 154 23039 22373 23040 22374 23041 136 23042 11 23043 138 23044 146 23045 2206 23046 2208 23047 2209 23048 2210 23049 2 23050 22384 23051 22385 23052 2 23053 22387 23054 830 23055 22389 23056 11 23057 22391 23058 22392 23059 22393 23060 2653 23061 22395 23062 22396 23063 338 23064 338 23065 341 23066 2 23067 404 23068 22402 23069 555 23070 22404 23071 22405 23072 22406 23073 22407 23074 2 23075 404 23076 22410 23077 22411 23078 22412 23079 22413 23080 2728 23081 22415 23082 555 23083 22417 23084 2 23085 22419 23086 22420 23087 22072 23088 2 23089 22423 23090 22424 23091 22425 23092 22426 23093 22427 23094 2820 23095 22429 23096 67 23097 22431 23098 22432 23099 22433 23100 22434 23101 22435 23102 22436 23103 2854 23104 22444 23105 2863 23106 22447 23107 917 23108 22464 23109 22465 23110 22466 23111 950 23112 143 23113 146 23114 22470 23115 22471 23116 22474 23117 2901 23118 2903 23119 2905 23120 75 23121 2907 23122 2908 23123 22481 23124 2 23125 264 23126 75 23127 3856 23128 2753 23129 22489 23130 1823 23131 146 23132 2011 23133 2 23134 22494 23135 2 23136 143 23137 146 23138 22498 23139 22499 23140 22500 23141 22501 23142 22502 23143 5125 23144 22504 23145 146 23146 22506 23147 2 23148 264 23149 22509 23150 2 23151 264 23152 22512 23153 22513 23154 22514 23155 22515 23156 22516 23157 22517 23158 1823 23159 146 23160 22520 23161 22521 23162 22522 23163 22523 23164 264 23165 22525 23166 22526 23167 2 23168 404 23169 22529 23170 5157 23171 22531 23172 22532 23173 22533 23174 22534 23175 22535 23176 22536 23177 2820 23178 22538 23179 67 23180 22540 23181 22541 23182 22542 23183 22543 23184 22544 23185 22545 23186 2854 23187 22553 23188 2863 23189 22556 23190 917 23191 22573 23192 22574 23193 22575 23194 950 23195 143 23196 146 23197 22579 23198 22580 23199 22583 23200 2901 23201 2903 23202 2905 23203 75 23204 2907 23205 2908 23206 22590 23207 2 23208 264 23209 75 23210 3856 23211 2753 23212 22598 23213 1823 23214 146 23215 2011 23216 2 23217 22603 23218 2 23219 143 23220 146 23221 22607 23222 22608 23223 22609 23224 22610 23225 22611 23226 5125 23227 22613 23228 146 23229 22615 23230 2 23231 264 23232 22618 23233 2 23234 264 23235 22621 23236 22622 23237 22623 23238 22624 23239 22625 23240 22626 23241 1823 23242 146 23243 22629 23244 22630 23245 22631 23246 22632 23247 264 23248 22634 23249 22635 23250 2 23251 404 23252 22638 23253 5157 23254 22640 23255 22641 23256 16264 23257 5969 23258 23258 23259 23259 23260 2965 23261 23261 23262 5969 23263 23263 23264 5969 23265 23265 23266 2965 23267 23267 23268 2 23269 17482 23270 18834 23271 18844 23272 15936 23273 18834 23274 146 23275 15936 23276 18894 23277 23277 23278 23278 23279 18897 23280 5969 23281 23281 23282 23282 23283 19057 23284 5969 23285 19057 23286 5969 23287 19065 23288 146 23289 15936 23290 19072 23291 23291 23292 23292 23293 19075 23294 5969 23295 23295 23296 2965 23297 23297 23298 19278 23299 11 23300 2 23301 264 23302 5969 23303 23303 23304 16 23305 19426 23306 15936 23307 19430 23308 23308 23309 23309 23310 19433 23311 5969 23312 23312 23313 5969 23314 23314 23315 5969 23316 23316 23317 23317 23318 15828 23319 136 23320 11 23321 138 23322 146 23323 149 23324 151 23325 75 23326 123 23327 154 23328 23328 23329 23329 23330 23330 23331 16 23332 2901 23333 2903 23334 2905 23335 75 23336 2907 23337 2908 23338 23338 23339 23339 23340 1039 23341 5969 23342 23342 23343 15858 23344 15859 23345 5969 23346 23346 23347 2 23348 264 23349 15936 23350 15937 23351 23351 23352 23352 23353 15940 23354 5969 23355 23355 23356 22 23357 143 23358 146 23359 5969 23360 23360 23361 123 23362 154 23363 23363 23364 23364 23365 11 23366 75 23367 23367 23368 123 23369 154 23370 23370 23371 13 23372 16110 23373 146 23374 75 23375 123 23376 154 23377 23377 23378 2905 23379 75 23380 2907 23381 2908 23382 23382 23383 2 23384 264 23385 16123 23386 16124 23387 23387 23388 16126 23389 16127 23390 16128 23391 23391 23392 16128 23393 23393 23394 16263 23395 16264 23396 5969 23397 23397 23398 16264 23399 5969 23400 23400 23401 23401 23402 2965 23403 23403 23404 5969 23405 23405 23406 19716 23407 2905 23408 75 23409 2907 23410 2908 23411 23411 23412 2 23413 264 23414 15936 23415 23415 23416 146 23417 5969 23418 23418 23419 5969 23420 23420 23421 2965 23422 23422 23423 2 23424 17482 23425 18834 23426 18844 23427 15936 23428 18834 23429 146 23430 15936 23431 18894 23432 23432 23433 23433 23434 18897 23435 5969 23436 23436 23437 23437 23438 19057 23439 5969 23440 19057 23441 5969 23442 19065 23443 146 23444 15936 23445 19072 23446 23446 23447 23447 23448 19075 23449 5969 23450 23450 23451 19765 23452 2903 23453 2905 23454 75 23455 2907 23456 2908 23457 23457 23458 2965 23459 23459 23460 19278 23461 11 23462 2 23463 264 23464 5969 23465 23465 23466 16 23467 19426 23468 15936 23469 19430 23470 23470 23471 23471 23472 19433 23473 5969 23474 23474 23475 5969 23476 23476 23477 2901 23478 2903 23479 2905 23480 75 23481 2907 23482 2908 23483 23483 23484 5969 23485 23485 23486 21290 23487 16134 23488 23488 23489 16128 23490 23490 23491 16263 23492 16264 23493 5969 23494 23494 23495 16264 23496 5969 23497 23497 23498 23498 23499 2965 23500 23500 23501 5969 23502 23502 23503 5969 23504 23504 23505 2965 23506 23506 23507 2 23508 17482 23509 18834 23510 18844 23511 15936 23512 18834 23513 146 23514 15936 23515 18894 23516 23516 23517 23517 23518 18897 23519 5969 23520 23520 23521 23521 23522 19057 23523 5969 23524 19057 23525 5969 23526 19065 23527 146 23528 15936 23529 19072 23530 23530 23531 23531 23532 19075 23533 5969 23534 23534 23535 2965 23536 23536 23537 19278 23538 11 23539 2 23540 264 23541 5969 23542 23542 23543 16 23544 19426 23545 15936 23546 19430 23547 23547 23548 23548 23549 19433 23550 5969 23551 23551 23552 5969 23553 23553 23554 5969 23555 23555 23556 23556 23557 15828 23558 136 23559 11 23560 138 23561 146 23562 149 23563 151 23564 75 23565 123 23566 154 23567 23567 23568 23568 23569 23569 23570 16 23571 2901 23572 2903 23573 2905 23574 75 23575 2907 23576 2908 23577 23577 23578 23578 23579 1039 23580 5969 23581 23581 23582 15858 23583 15859 23584 5969 23585 23585 23586 2 23587 264 23588 15936 23589 15937 23590 23590 23591 23591 23592 15940 23593 5969 23594 23594 23595 22 23596 143 23597 146 23598 5969 23599 23599 23600 123 23601 154 23602 23602 23603 23603 23604 11 23605 75 23606 23606 23607 123 23608 154 23609 23609 23610 13 23611 16110 23612 146 23613 75 23614 123 23615 154 23616 23616 23617 2905 23618 75 23619 2907 23620 2908 23621 23621 23622 2 23623 264 23624 16123 23625 16124 23626 23626 23627 16126 23628 16127 23629 16128 23630 23630 23631 16128 23632 23632 23633 16263 23634 16264 23635 5969 23636 23636 23637 16264 23638 5969 23639 23639 23640 23640 23641 2965 23642 23642 23643 5969 23644 23644 23645 19716 23646 2905 23647 75 23648 2907 23649 2908 23650 23650 23651 2 23652 264 23653 15936 23654 23654 23655 146 23656 5969 23657 23657 23658 5969 23659 23659 23660 2965 23661 23661 23662 2 23663 17482 23664 18834 23665 18844 23666 15936 23667 18834 23668 146 23669 15936 23670 18894 23671 23671 23672 23672 23673 18897 23674 5969 23675 23675 23676 23676 23677 19057 23678 5969 23679 19057 23680 5969 23681 19065 23682 146 23683 15936 23684 19072 23685 23685 23686 23686 23687 19075 23688 5969 23689 23689 23690 19765 23691 2903 23692 2905 23693 75 23694 2907 23695 2908 23696 23696 23697 23697 23698 2965 23699 23699 23700 19278 23701 11 23702 2 23703 264 23704 5969 23705 23705 23706 16 23707 19426 23708 15936 23709 19430 23710 23710 23711 23711 23712 19433 23713 5969 23714 23714 23715 5969 23716 23716 23717 2901 23718 2903 23719 2905 23720 75 23721 2907 23722 2908 23723 23723 23724 5969 23725 23725 23726 22 23727 143 23728 146 23729 5969 23730 23730 23731 123 23732 154 23733 23733 23734 23734 23735 11 23736 75 23737 23737 23738 123 23739 154 23740 23740 23741 13 23742 16110 23743 146 23744 75 23745 123 23746 154 23747 23747 23748 2905 23749 75 23750 2907 23751 2908 23752 23752 23753 2 23754 264 23755 16123 23756 16124 23757 23757 23758 16126 23759 16127 23760 16128 23761 23761 23762 16128 23763 23763 23764 16263 23765 16264 23766 5969 23767 23767 23768 16264 23769 5969 23770 23770 23771 23771 23772 2965 23773 23773 23774 5969 23775 23775 23776 19716 23777 2905 23778 75 23779 2907 23780 2908 23781 23781 23782 146 23783 5969 23784 23784 23785 5969 23786 23786 23787 2965 23788 23788 23789 2 23790 17482 23791 18834 23792 18844 23793 15936 23794 18834 23795 146 23796 15936 23797 18894 23798 23798 23799 23799 23800 18897 23801 5969 23802 23802 23803 23803 23804 19057 23805 5969 23806 19057 23807 5969 23808 19065 23809 146 23810 15936 23811 19072 23812 23812 23813 23813 23814 19075 23815 5969 23816 23816 23817 19765 23818 2903 23819 2905 23820 75 23821 2907 23822 2908 23823 23823 23824 23824 23825 2965 23826 23826 23827 19278 23828 11 23829 2 23830 264 23831 5969 23832 23832 23833 16 23834 19426 23835 15936 23836 19430 23837 23837 23838 23838 23839 19433 23840 5969 23841 23841 23842 5969 23843 23843 23844 2901 23845 2903 23846 2905 23847 75 23848 2907 23849 2908 23850 23850 23851 5969 23852 23852 23853 2 23854 264 23855 15936 23856 15937 23857 23857 23858 23858 23859 15940 23860 5969 23861 23861 23862 22 23863 143 23864 146 23865 5969 23866 23866 23867 123 23868 154 23869 23869 23870 23870 23871 11 23872 75 23873 23873 23874 123 23875 154 23876 23876 23877 13 23878 16110 23879 146 23880 75 23881 123 23882 154 23883 23883 23884 2905 23885 75 23886 2907 23887 2908 23888 23888 23889 2 23890 264 23891 16123 23892 16124 23893 23893 23894 16126 23895 16127 23896 16128 23897 23897 23898 16128 23899 23899 23900 16263 23901 16264 23902 5969 23903 23903 23904 16264 23905 5969 23906 23906 23907 23907 23908 2965 23909 23909 23910 5969 23911 23911 23912 19716 23913 2905 23914 75 23915 2907 23916 2908 23917 23917 23918 2 23919 264 23920 15936 23921 23921 23922 146 23923 5969 23924 23924 23925 5969 23926 23926 23927 2965 23928 23928 23929 2 23930 17482 23931 18834 23932 18844 23933 15936 23934 18834 23935 146 23936 15936 23937 18894 23938 23938 23939 23939 23940 18897 23941 5969 23942 23942 23943 23943 23944 19057 23945 5969 23946 19057 23947 5969 23948 19065 23949 146 23950 15936 23951 19072 23952 23952 23953 23953 23954 19075 23955 5969 23956 23956 23957 19765 23958 2903 23959 2905 23960 75 23961 2907 23962 2908 23963 23963 23964 2965 23965 23965 23966 19278 23967 11 23968 2 23969 264 23970 5969 23971 23971 23972 16 23973 19426 23974 15936 23975 19430 23976 23976 23977 23977 23978 19433 23979 5969 23980 23980 23981 5969 23982 23982 23983 2901 23984 2903 23985 2905 23986 75 23987 2907 23988 2908 23989 23989 23990 5969 23991 23991 23992 15858 23993 15859 23994 5969 23995 23995 23996 2 23997 264 23998 15936 23999 15937 24000 24000 24001 24001 24002 15940 24003 5969 24004 24004 24005 22 24006 143 24007 146 24008 5969 24009 24009 24010 123 24011 154 24012 24012 24013 24013 24014 11 24015 75 24016 24016 24017 123 24018 154 24019 24019 24020 13 24021 16110 24022 146 24023 75 24024 123 24025 154 24026 24026 24027 2 24028 264 24029 16123 24030 16124 24031 24031 24032 16126 24033 16127 24034 16128 24035 24035 24036 16128 24037 24037 24038 16263 24039 16264 24040 5969 24041 24041 24042 16264 24043 5969 24044 24044 24045 24045 24046 2965 24047 24047 24048 5969 24049 24049 24050 5969 24051 24051 24052 2965 24053 24053 24054 2 24055 17482 24056 18834 24057 18844 24058 15936 24059 18834 24060 146 24061 15936 24062 18894 24063 24063 24064 24064 24065 18897 24066 5969 24067 24067 24068 24068 24069 19057 24070 5969 24071 19057 24072 5969 24073 19065 24074 146 24075 15936 24076 19072 24077 24077 24078 24078 24079 19075 24080 5969 24081 24081 24082 2965 24083 24083 24084 19278 24085 11 24086 2 24087 264 24088 5969 24089 24089 24090 16 24091 19426 24092 15936 24093 19430 24094 24094 24095 24095 24096 19433 24097 5969 24098 24098 24099 5969 24100 24100 24101 5969 24102 24102 24103 830 24104 15072 24105 11 24106 15074 24107 15075 24108 15076 24109 2653 24110 15078 24111 15079 24112 338 24113 338 24114 341 24115 2 24116 404 24117 15085 24118 555 24119 15087 24120 15088 24121 15089 24122 15090 24123 2 24124 404 24125 15093 24126 15094 24127 15095 24128 15096 24129 2728 24130 15098 24131 555 24132 15100 24133 2 24134 15102 24135 15103 24136 2 24137 3 24138 4 24139 2 24140 2 24141 2 24142 2 24143 11 24144 2 24145 2 24146 15114 24147 2 24148 15116 24149 22 24150 15118 24151 15119 24152 2 24153 11 24154 11 24155 15123 24156 15124 24157 15125 24158 22 24159 2 24160 2 24161 2 24162 11 24163 15131 24164 11 24165 15133 24166 11 24167 15135 24168 15136 24169 15137 24170 15138 24171 60 24172 15140 24173 2 24174 2 24175 2 24176 2 24177 67 24178 127 24179 15147 24180 2 24181 2 24182 15150 24183 2 24184 15152 24185 15153 24186 2 24187 83 24188 84 24189 87 24190 130 24191 11 24192 15160 24193 15161 24194 97 24195 15163 24196 15164 24197 15165 24198 102 24199 102 24200 15168 24201 15169 24202 15170 24203 15171 24204 15172 24205 15173 24206 15174 24207 15175 24208 15176 24209 15177 24210 97 24211 2 24212 15180 24213 2 24214 2 24215 2 24216 2 24217 2 24218 2 24219 2 24220 2 24221 288 24222 2 24223 2 24224 2 24225 2 24226 2 24227 294 24228 2 24229 2 24230 2 24231 2 24232 2 24233 2 24234 2 24235 2 24236 2 24237 2 24238 2 24239 2 24240 2 24241 2 24242 2 24243 15211 24244 15212 24245 15213 24246 15214 24247 15215 24248 15216 24249 15217 24250 15218 24251 15219 24252 15220 24253 15221 24254 15222 24255 15223 24256 15224 24257 15225 24258 4 24259 2 24260 15228 24261 15229 24262 15230 24263 15231 24264 15232 24265 15233 24266 15234 24267 15235 24268 15236 24269 15237 24270 15238 24271 338 24272 338 24273 40 24274 341 24275 2 24276 404 24277 15245 24278 555 24279 15247 24280 15248 24281 15249 24282 15250 24283 15251 24284 2 24285 404 24286 15254 24287 15255 24288 40 24289 15257 24290 264 24291 15259 24292 15260 24293 2 24294 404 24295 15263 24296 15264 24297 15265 24298 15266 24299 15267 24300 15268 24301 2 24302 713 24303 2 24304 15272 24305 15273 24306 15274 24307 15275 24308 15276 24309 15277 24310 15278 24311 15279 24312 15280 24313 15281 24314 830 24315 15283 24316 15284 24317 15285 24318 15286 24319 15287 24320 15288 24321 15289 24322 15290 24323 15291 24324 15292 24325 15293 24326 15294 24327 15295 24328 15296 24329 15297 24330 15298 24331 15299 24332 15300 24333 15301 24334 15302 24335 15303 24336 15304 24337 15305 24338 15306 24339 15307 24340 15308 24341 15309 24342 15310 24343 15311 24344 15312 24345 15313 24346 15314 24347 15315 24348 15316 24349 15317 24350 15318 24351 15319 24352 15320 24353 15321 24354 15322 24355 15323 24356 15324 24357 15325 24358 15326 24359 15327 24360 15328 24361 15329 24362 15330 24363 15331 24364 15332 24365 15333 24366 15334 24367 15335 24368 15336 24369 15337 24370 15338 24371 15339 24372 15340 24373 15341 24374 15342 24375 15343 24376 15344 24377 15345 24378 15346 24379 15347 24380 15348 24381 15349 24382 15350 24383 15351 24384 15352 24385 15353 24386 15354 24387 15355 24388 15356 24389 15357 24390 15358 24391 15359 24392 15360 24393 15361 24394 11 24395 15363 24396 15364 24397 15365 24398 917 24399 15367 24400 15368 24401 950 24402 143 24403 146 24404 15372 24405 15373 24406 1456 24407 2 24408 2 24409 2 24410 15378 24411 2 24412 15380 24413 2 24414 2 24415 2 24416 2 24417 2 24418 2 24419 2 24420 2 24421 2 24422 15390 24423 15391 24424 15392 24425 15393 24426 15394 24427 15395 24428 15396 24429 15397 24430 15398 24431 15399 24432 15400 24433 15401 24434 15402 24435 22 24436 15404 24437 4 24438 15406 24439 87 24440 11 24441 15409 24442 97 24443 15411 24444 1517 24445 15413 24446 15414 24447 15415 24448 97 24449 830 24450 11 24451 15419 24452 917 24453 15421 24454 15422 24455 950 24456 143 24457 146 24458 15426 24459 15427 24460 15428 24461 830 24462 11 24463 15431 24464 15432 24465 15433 24466 338 24467 338 24468 341 24469 2 24470 404 24471 15439 24472 555 24473 15441 24474 15442 24475 15443 24476 15444 24477 2 24478 404 24479 15447 24480 15448 24481 15449 24482 15450 24483 15451 24484 2 24485 404 24486 15454 24487 15455 24488 136 24489 11 24490 138 24491 146 24492 147 24493 149 24494 151 24495 75 24496 123 24497 154 24498 15466 24499 15467 24500 136 24501 11 24502 138 24503 146 24504 2206 24505 2208 24506 2209 24507 2210 24508 2 24509 15477 24510 15478 24511 2 24512 15480 24513 830 24514 15482 24515 11 24516 15484 24517 15485 24518 15486 24519 2653 24520 15488 24521 15489 24522 338 24523 338 24524 341 24525 2 24526 404 24527 15495 24528 555 24529 15497 24530 15498 24531 15499 24532 15500 24533 2 24534 404 24535 15503 24536 15504 24537 15505 24538 15506 24539 2728 24540 15508 24541 555 24542 15510 24543 2 24544 15512 24545 15513 24546 15165 24547 2 24548 15516 24549 15517 24550 15518 24551 15519 24552 15520 24553 2820 24554 15522 24555 67 24556 15524 24557 15525 24558 15526 24559 15527 24560 15528 24561 15529 24562 2854 24563 15537 24564 2863 24565 15540 24566 917 24567 15557 24568 15558 24569 15559 24570 950 24571 143 24572 146 24573 15563 24574 15564 24575 15567 24576 2901 24577 2903 24578 2905 24579 75 24580 2907 24581 2908 24582 15574 24583 2 24584 264 24585 75 24586 3856 24587 2753 24588 15582 24589 1823 24590 146 24591 2011 24592 2 24593 15587 24594 2 24595 143 24596 146 24597 15591 24598 15592 24599 15593 24600 15594 24601 15595 24602 5125 24603 15597 24604 146 24605 15599 24606 2 24607 264 24608 15602 24609 2 24610 264 24611 15605 24612 15606 24613 15607 24614 15608 24615 15609 24616 15610 24617 1823 24618 146 24619 15613 24620 15614 24621 15615 24622 15616 24623 264 24624 15618 24625 15619 24626 2 24627 404 24628 15622 24629 5157 24630 15624 24631 15625 24632 15626 24633 15627 24634 15628 24635 15629 24636 2820 24637 15631 24638 67 24639 15633 24640 15634 24641 15635 24642 15636 24643 15637 24644 15638 24645 2854 24646 15646 24647 2863 24648 15649 24649 917 24650 15666 24651 15667 24652 15668 24653 950 24654 143 24655 146 24656 15672 24657 15673 24658 15676 24659 2901 24660 2903 24661 2905 24662 75 24663 2907 24664 2908 24665 15683 24666 2 24667 264 24668 75 24669 3856 24670 2753 24671 15691 24672 1823 24673 146 24674 2011 24675 2 24676 15696 24677 2 24678 143 24679 146 24680 15700 24681 15701 24682 15702 24683 15703 24684 15704 24685 5125 24686 15706 24687 146 24688 15708 24689 2 24690 264 24691 15711 24692 2 24693 264 24694 15714 24695 15715 24696 15716 24697 15717 24698 15718 24699 15719 24700 1823 24701 146 24702 15722 24703 15723 24704 15724 24705 15725 24706 264 24707 15727 24708 15728 24709 2 24710 404 24711 15731 24712 5157 24713 15733 24714 15734 24715 24715 24716 15828 24717 136 24718 11 24719 138 24720 146 24721 149 24722 151 24723 75 24724 123 24725 154 24726 24726 24727 24727 24728 24728 24729 16 24730 2901 24731 2903 24732 2905 24733 75 24734 2907 24735 2908 24736 24736 24737 24737 24738 1039 24739 5969 24740 24740 24741 15858 24742 15859 24743 5969 24744 24744 24745 2 24746 264 24747 15936 24748 15937 24749 24749 24750 24750 24751 15940 24752 5969 24753 24753 24754 22 24755 143 24756 146 24757 5969 24758 24758 24759 123 24760 154 24761 24761 24762 24762 24763 11 24764 75 24765 24765 24766 123 24767 154 24768 24768 24769 13 24770 16110 24771 146 24772 75 24773 123 24774 154 24775 24775 24776 2905 24777 75 24778 2907 24779 2908 24780 24780 24781 2 24782 264 24783 16123 24784 16124 24785 24785 24786 16126 24787 16127 24788 16128 24789 24789 24790 16128 24791 24791 24792 16263 24793 16264 24794 5969 24795 24795 24796 16264 24797 5969 24798 24798 24799 24799 24800 2965 24801 24801 24802 5969 24803 24803 24804 19716 24805 2905 24806 75 24807 2907 24808 2908 24809 24809 24810 2 24811 264 24812 15936 24813 24813 24814 146 24815 5969 24816 24816 24817 5969 24818 24818 24819 2965 24820 24820 24821 2 24822 17482 24823 18834 24824 18844 24825 15936 24826 18834 24827 146 24828 15936 24829 18894 24830 24830 24831 24831 24832 18897 24833 5969 24834 24834 24835 24835 24836 19057 24837 5969 24838 19057 24839 5969 24840 19065 24841 146 24842 15936 24843 19072 24844 24844 24845 24845 24846 19075 24847 5969 24848 24848 24849 19765 24850 2903 24851 2905 24852 75 24853 2907 24854 2908 24855 24855 24856 2965 24857 24857 24858 19278 24859 11 24860 2 24861 264 24862 5969 24863 24863 24864 16 24865 19426 24866 15936 24867 19430 24868 24868 24869 24869 24870 19433 24871 5969 24872 24872 24873 5969 24874 24874 24875 2901 24876 2903 24877 2905 24878 75 24879 2907 24880 2908 24881 24881 24882 5969 24883 24883 24884 2214 24885 24885 24886 24886 24887 24887 24888 24888 24889 24889 24890 24890 24891 24891 24892 24892 24893 24893 24894 24894 24895 24895 24896 24896 24897 24897 24898 2214 24899 24899 24900 24900 24901 24901 24902 24902 24903 24903 24904 24904 24905 24905 24906 24906 24907 24907 24908 24908 24909 2214 24910 2214 24911 24911 24912 24912 24913 24913 24914 24914 24915 24915 24916 24916 24917 24917 24918 24918 24919 24919 24920 24920 24921 24921 24922 2214 24923 24923 24924 24924 24925 24925 24926 24926 24927 24927 24928 24928 24929 24929 24930 24930 24931 24931 24932 24932 24933 24933 24934 24934 24935 24935 24936 24936 24937 24937 24938 2939 24939 24939 24940 143 24941 24941 24942 146 24943 24943 24944 24944 24945 24945 24946 2930 24947 24947 24948 24948 24949 24949 24950 24950 24951 24951 24952 24952 24953 24953 24954 24954 24955 24955 24956 24956 24957 24957 24958 24958 24959 24959 24960 24960 24961 24961 24962 24962 24963 24963 24964 2930 24965 24965 24966 24966 24967 24967 24968 24968 24969 24969 24970 24970 24971 24971 24972 24972 24973 2930 24974 2930 24975 24975 24976 24976 24977 24977 24978 24978 24979 24979 24980 24980 24981 24981 24982 24982 24983 24983 24984 24984 24985 24985 24986 24986 24987 2939 24988 24988 24989 24989 24990 2 24991 24991 24992 2930 24993 24993 24994 24994 24995 24995 24996 24996 24997 24997 24998 24998 24999 24999 25000 25000 25001 25001 25002 25002 25003 25003 25004 25004 25005 25005 25006 25006 25007 6356 25008 555 25009 25009 25010 6356 25011 555 25012 25012 25013 154 25014 25014 25015 830 25016 25016 25017 25017 25018 25018 25019 25019 25020 25020 25021 25021 25022 25022 25023 25023 25024 25024 25025 25025 25026 25026 25027 25027 25028 25028 25029 25029 25030 25030 25031 25031 25032 25032 25033 25033 25034 25034 25035 25035 25036 25036 25037 25037 25038 25038 25039 25039 25040 25040 25041 25041 25042 25042 25043 25043 25044 25044 25045 25045 25046 25046 25047 25047 25048 25048 25049 25049 25050 25050 25051 25051 25052 25052 25053 25053 25054 25054 25055 25055 25056 25056 25057 25057 25058 25058 25059 25059 25060 25060 25061 25061 25062 25062 25063 25063 25064 25064 25065 25065 25066 25066 25067 25067 25068 25068 25069 25069 25070 25070 25071 25071 25072 25072 25073 25073 25074 25074 25075 25075 25076 25076 25077 25077 25078 25078 25079 25079 25080 25080 25081 25081 25082 25082 25083 25083 25084 25084 25085 25085 25086 25086 25087 25087 25088 25088 25089 25089 25090 25090 25091 25091 25092 25092 25093 25093 25094 25094 25095 11 25096 25096 25097 25097 25098 25098 25099 917 25100 25100 25101 25101 25102 950 25103 143 25104 146 25105 25105 25106 25106 25107 1456 25108 2 25109 2 25110 2 25111 25111 25112 2 25113 25113 25114 2 25115 2 25116 2 25117 2 25118 2 25119 2 25120 2 25121 2 25122 2 25123 25123 25124 25124 25125 25125 25126 25126 25127 25127 25128 25128 25129 25129 25130 25130 25131 25131 25132 25132 25133 25133 25134 25134 25135 25135 25136 22 25137 25137 25138 4 25139 25139 25140 87 25141 11 25142 25142 25143 97 25144 25144 25145 1517 25146 25146 25147 25147 25148 25148 25149 97 25150 830 25151 11 25152 25152 25153 917 25154 25154 25155 25155 25156 950 25157 143 25158 146 25159 25159 25160 25160 25161 25161 25162 830 25163 11 25164 25164 25165 25165 25166 25166 25167 338 25168 338 25169 341 25170 2 25171 404 25172 25172 25173 555 25174 25174 25175 25175 25176 25176 25177 25177 25178 2 25179 404 25180 25180 25181 25181 25182 25182 25183 25183 25184 25184 25185 2 25186 404 25187 25187 25188 25188 25189 26 25190 11 25191 2 25192 264 25193 26 25194 25194 25195 15 25196 15 25197 11 25198 26 25199 2 25200 11 25201 11 25202 40 25203 16 25204 25204 25205 15 25206 68 25207 69 25208 11 25209 75 25210 11 25211 11 25212 25212 25213 86 25214 101 25215 25215 25216 108 25217 25217 25218 123 25219 154 25220 25220 25221 26 25222 25222 25223 25223 25224 1039 25225 5969 25226 25226 25227 15858 25228 15859 25229 5969 25230 25230 25231 2 25232 264 25233 15936 25234 15937 25235 25235 25236 25236 25237 15940 25238 5969 25239 25239 25240 22 25241 143 25242 146 25243 5969 25244 25244 25245 123 25246 154 25247 25247 25248 25248 25249 11 25250 75 25251 25251 25252 123 25253 154 25254 25254 25255 13 25256 16110 25257 146 25258 75 25259 123 25260 154 25261 25261 25262 2 25263 264 25264 16123 25265 16124 25266 25266 25267 16126 25268 16127 25269 16128 25270 25270 25271 16128 25272 25272 25273 16263 25274 16264 25275 5969 25276 25276 25277 16264 25278 5969 25279 25279 25280 25280 25281 2965 25282 25282 25283 5969 25284 25284 25285 5969 25286 25286 25287 2965 25288 25288 25289 2 25290 17482 25291 18834 25292 18844 25293 15936 25294 18834 25295 146 25296 15936 25297 18894 25298 25298 25299 25299 25300 18897 25301 5969 25302 25302 25303 25303 25304 19057 25305 5969 25306 19057 25307 5969 25308 19065 25309 146 25310 15936 25311 19072 25312 25312 25313 25313 25314 19075 25315 5969 25316 25316 25317 2965 25318 25318 25319 19278 25320 11 25321 2 25322 264 25323 5969 25324 25324 25325 16 25326 19426 25327 15936 25328 19430 25329 25329 25330 25330 25331 19433 25332 5969 25333 25333 25334 5969 25335 25335 25336 5969 25337 25337 25338 25338 25339 15828 25340 136 25341 11 25342 138 25343 146 25344 149 25345 151 25346 75 25347 123 25348 154 25349 25349 25350 25350 25351 136 25352 11 25353 138 25354 146 25355 2206 25356 2208 25357 2209 25358 2210 25359 2 25360 2 25361 3 25362 4 25363 2 25364 2 25365 2 25366 2 25367 11 25368 2 25369 2 25370 17 25371 2 25372 20 25373 22 25374 24 25375 25 25376 2 25377 11 25378 11 25379 32 25380 33 25381 39 25382 22 25383 2 25384 2 25385 2 25386 11 25387 51 25388 11 25389 53 25390 11 25391 55 25392 57 25393 58 25394 59 25395 60 25396 61 25397 2 25398 2 25399 2 25400 2 25401 67 25402 127 25403 128 25404 2 25405 2 25406 74 25407 2 25408 79 25409 81 25410 2 25411 83 25412 84 25413 87 25414 130 25415 11 25416 92 25417 93 25418 97 25419 98 25420 99 25421 25421 25422 102 25423 102 25424 105 25425 109 25426 111 25427 112 25428 113 25429 114 25430 115 25431 116 25432 118 25433 121 25434 97 25435 2 25436 131 25437 2 25438 2 25439 2 25440 2 25441 2 25442 2 25443 2 25444 2 25445 288 25446 2 25447 2 25448 2 25449 2 25450 2 25451 294 25452 2 25453 2 25454 2 25455 2 25456 2 25457 2 25458 2 25459 2 25460 2 25461 2 25462 2 25463 2 25464 2 25465 2 25466 2 25467 6279 25468 6280 25469 6281 25470 6282 25471 6283 25472 6284 25473 6285 25474 6286 25475 6287 25476 6288 25477 6289 25478 6290 25479 6291 25480 6292 25481 6293 25482 4 25483 2 25484 6296 25485 6297 25486 6298 25487 6299 25488 6300 25489 6301 25490 6302 25491 6303 25492 6304 25493 6305 25494 6306 25495 338 25496 338 25497 40 25498 341 25499 2 25500 404 25501 6313 25502 555 25503 6315 25504 6316 25505 6317 25506 6318 25507 6319 25508 2 25509 404 25510 6322 25511 6323 25512 40 25513 6325 25514 264 25515 6327 25516 6328 25517 2 25518 404 25519 6331 25520 6332 25521 6333 25522 6334 25523 6335 25524 6336 25525 2 25526 713 25527 2 25528 6340 25529 6341 25530 6342 25531 6343 25532 6344 25533 6345 25534 6346 25535 6347 25536 6348 25537 6349 25538 830 25539 25016 25540 25017 25541 25018 25542 25019 25543 25020 25544 25021 25545 25022 25546 25023 25547 25024 25548 25025 25549 25026 25550 25027 25551 25028 25552 25029 25553 25030 25554 25031 25555 25032 25556 25033 25557 25034 25558 25035 25559 25036 25560 25037 25561 25038 25562 25039 25563 25040 25564 25041 25565 25042 25566 25043 25567 25044 25568 25045 25569 25046 25570 25047 25571 25048 25572 25049 25573 25050 25574 25051 25575 25052 25576 25053 25577 25054 25578 25055 25579 25056 25580 25057 25581 25058 25582 25059 25583 25060 25584 25061 25585 25062 25586 25063 25587 25064 25588 25065 25589 25066 25590 25067 25591 25068 25592 25069 25593 25070 25594 25071 25595 25072 25596 25073 25597 25074 25598 25075 25599 25076 25600 25077 25601 25078 25602 25079 25603 25080 25604 25081 25605 25082 25606 25083 25607 25084 25608 25085 25609 25086 25610 25087 25611 25088 25612 25089 25613 25090 25614 25091 25615 25092 25616 25093 25617 25094 25618 11 25619 25096 25620 25097 25621 25098 25622 917 25623 25100 25624 25101 25625 950 25626 143 25627 146 25628 25105 25629 25106 25630 1456 25631 2 25632 2 25633 2 25634 25111 25635 2 25636 25113 25637 2 25638 2 25639 2 25640 2 25641 2 25642 2 25643 2 25644 2 25645 2 25646 25123 25647 25124 25648 25125 25649 25126 25650 25127 25651 25128 25652 25129 25653 25130 25654 25131 25655 25132 25656 25133 25657 25134 25658 25135 25659 22 25660 25137 25661 4 25662 25139 25663 87 25664 11 25665 25142 25666 97 25667 25144 25668 1517 25669 25146 25670 25147 25671 25148 25672 97 25673 830 25674 11 25675 25152 25676 917 25677 25154 25678 25155 25679 950 25680 143 25681 146 25682 25159 25683 25160 25684 25161 25685 830 25686 11 25687 25164 25688 25165 25689 25166 25690 338 25691 338 25692 341 25693 2 25694 404 25695 25172 25696 555 25697 25174 25698 25175 25699 25176 25700 25177 25701 2 25702 404 25703 25180 25704 25181 25705 25182 25706 25183 25707 25184 25708 2 25709 404 25710 25187 25711 25188 25712 136 25713 11 25714 138 25715 146 25716 147 25717 149 25718 151 25719 75 25720 123 25721 154 25722 196 25723 197 25724 136 25725 11 25726 138 25727 146 25728 2206 25729 2208 25730 2209 25731 2210 25732 2 25733 25421 25734 25734 25735 25734 25736 830 25737 25737 25738 11 25739 25739 25740 25740 25741 25741 25742 2653 25743 15853 25744 15854 25745 338 25746 338 25747 341 25748 2 25749 404 25750 25750 25751 555 25752 25752 25753 25753 25754 25754 25755 25755 25756 2 25757 404 25758 25758 25759 25759 25760 25760 25761 25761 25762 2728 25763 25763 25764 555 25765 25765 25766 2 25767 25767 25768 25768 25769 2 25770 3 25771 4 25772 2 25773 2 25774 2 25775 2 25776 11 25777 2 25778 2 25779 25779 25780 2 25781 25781 25782 22 25783 25783 25784 25784 25785 2 25786 11 25787 11 25788 25788 25789 25789 25790 25790 25791 22 25792 2 25793 2 25794 2 25795 11 25796 25796 25797 11 25798 25798 25799 11 25800 25800 25801 25801 25802 25802 25803 25803 25804 60 25805 25805 25806 2 25807 2 25808 2 25809 2 25810 67 25811 127 25812 25812 25813 2 25814 2 25815 25815 25816 2 25817 25817 25818 25818 25819 2 25820 83 25821 84 25822 87 25823 130 25824 11 25825 25825 25826 25826 25827 97 25828 25828 25829 25829 25830 25830 25831 102 25832 102 25833 25833 25834 25834 25835 25835 25836 25836 25837 25837 25838 25838 25839 25839 25840 25840 25841 25841 25842 25842 25843 97 25844 2 25845 25845 25846 2 25847 2 25848 2 25849 2 25850 2 25851 2 25852 2 25853 2 25854 288 25855 2 25856 2 25857 2 25858 2 25859 2 25860 294 25861 2 25862 2 25863 2 25864 2 25865 2 25866 2 25867 2 25868 2 25869 2 25870 2 25871 2 25872 2 25873 2 25874 2 25875 2 25876 25876 25877 25877 25878 25878 25879 25879 25880 25880 25881 25881 25882 25882 25883 25883 25884 25884 25885 25885 25886 25886 25887 25887 25888 25888 25889 25889 25890 25890 25891 4 25892 2 25893 25893 25894 25894 25895 25895 25896 25896 25897 25897 25898 25898 25899 25899 25900 25900 25901 25901 25902 25902 25903 25903 25904 338 25905 338 25906 40 25907 341 25908 2 25909 404 25910 25910 25911 555 25912 25912 25913 25913 25914 25914 25915 25915 25916 25916 25917 2 25918 404 25919 25919 25920 25920 25921 40 25922 25922 25923 264 25924 25924 25925 25925 25926 2 25927 404 25928 25928 25929 25929 25930 25930 25931 25931 25932 25932 25933 25933 25934 2 25935 713 25936 2 25937 25937 25938 25938 25939 25939 25940 25940 25941 25941 25942 25942 25943 25943 25944 25944 25945 25945 25946 25946 25947 830 25948 25948 25949 25949 25950 25950 25951 25951 25952 25952 25953 25953 25954 25954 25955 25955 25956 25956 25957 25957 25958 25958 25959 25959 25960 25960 25961 25961 25962 25962 25963 25963 25964 25964 25965 25965 25966 25966 25967 25967 25968 25968 25969 25969 25970 25970 25971 25971 25972 25972 25973 25973 25974 25974 25975 25975 25976 25976 25977 25977 25978 25978 25979 25979 25980 25980 25981 25981 25982 25982 25983 25983 25984 25984 25985 25985 25986 25986 25987 25987 25988 25988 25989 25989 25990 25990 25991 25991 25992 25992 25993 25993 25994 25994 25995 25995 25996 25996 25997 25997 25998 25998 25999 25999 26000 26000 26001 26001 26002 26002 26003 26003 26004 26004 26005 26005 26006 26006 26007 26007 26008 26008 26009 26009 26010 26010 26011 26011 26012 26012 26013 26013 26014 26014 26015 26015 26016 26016 26017 26017 26018 26018 26019 26019 26020 26020 26021 26021 26022 26022 26023 26023 26024 26024 26025 26025 26026 26026 26027 11 26028 26028 26029 26029 26030 26030 26031 917 26032 26032 26033 26033 26034 950 26035 143 26036 146 26037 26037 26038 26038 26039 1456 26040 2 26041 2 26042 2 26043 26043 26044 2 26045 26045 26046 2 26047 2 26048 2 26049 2 26050 2 26051 2 26052 2 26053 2 26054 2 26055 26055 26056 26056 26057 26057 26058 26058 26059 26059 26060 26060 26061 26061 26062 26062 26063 26063 26064 26064 26065 26065 26066 26066 26067 26067 26068 22 26069 26069 26070 4 26071 26071 26072 87 26073 11 26074 26074 26075 97 26076 26076 26077 1517 26078 26078 26079 26079 26080 26080 26081 97 26082 830 26083 11 26084 26084 26085 917 26086 26086 26087 26087 26088 950 26089 143 26090 146 26091 26091 26092 26092 26093 26093 26094 830 26095 11 26096 26096 26097 26097 26098 26098 26099 338 26100 338 26101 341 26102 2 26103 404 26104 26104 26105 555 26106 26106 26107 26107 26108 26108 26109 26109 26110 2 26111 404 26112 26112 26113 26113 26114 26114 26115 26115 26116 26116 26117 2 26118 404 26119 26119 26120 26120 26121 136 26122 11 26123 138 26124 146 26125 147 26126 149 26127 151 26128 75 26129 123 26130 154 26131 26131 26132 26132 26133 136 26134 11 26135 138 26136 146 26137 2206 26138 2208 26139 2209 26140 2210 26141 2 26142 26142 26143 26143 26144 2 26145 26145 26146 830 26147 26147 26148 11 26149 26149 26150 26150 26151 26151 26152 2653 26153 26153 26154 26154 26155 338 26156 338 26157 341 26158 2 26159 404 26160 26160 26161 555 26162 26162 26163 26163 26164 26164 26165 26165 26166 2 26167 404 26168 26168 26169 26169 26170 26170 26171 26171 26172 2728 26173 26173 26174 555 26175 26175 26176 2 26177 26177 26178 26178 26179 25830 26180 2 26181 26181 26182 26182 26183 26183 26184 26184 26185 26185 26186 2820 26187 26187 26188 67 26189 26189 26190 26190 26191 26191 26192 26192 26193 26193 26194 26194 26195 2 26196 2854 26197 26197 26198 26198 26199 26199 26200 26200 26201 26201 26202 26202 26203 2863 26204 26204 26205 26205 26206 26206 26207 26207 26208 26208 26209 338 26210 338 26211 40 26212 341 26213 2 26214 404 26215 26215 26216 555 26217 26217 26218 26218 26219 26219 26220 26220 26221 917 26222 26222 26223 26223 26224 26224 26225 950 26226 143 26227 146 26228 26228 26229 26229 26230 40 26231 26231 26232 26232 26233 2901 26234 2903 26235 2905 26236 75 26237 2907 26238 2908 26239 26239 26240 264 26241 26241 26242 2 26243 264 26244 75 26245 3856 26246 2753 26247 26247 26248 1823 26249 146 26250 2011 26251 2 26252 26252 26253 2 26254 143 26255 146 26256 26256 26257 26257 26258 26258 26259 26259 26260 26260 26261 5125 26262 26262 26263 146 26264 26264 26265 2 26266 264 26267 26267 26268 2 26269 264 26270 26270 26271 26271 26272 26272 26273 26273 26274 26274 26275 26275 26276 1823 26277 146 26278 26278 26279 26279 26280 26280 26281 26281 26282 264 26283 26283 26284 26284 26285 2 26286 404 26287 26287 26288 5157 26289 26289 26290 26290 26291 26291 26292 26292 26293 26293 26294 16409 26295 2820 26296 26296 26297 67 26298 26298 26299 26299 26300 26300 26301 26301 26302 26302 26303 26303 26304 2 26305 2854 26306 26306 26307 26307 26308 26308 26309 26309 26310 26310 26311 26311 26312 2863 26313 26313 26314 26314 26315 26315 26316 26316 26317 26317 26318 338 26319 338 26320 40 26321 341 26322 2 26323 404 26324 26324 26325 555 26326 26326 26327 26327 26328 26328 26329 26329 26330 917 26331 26331 26332 26332 26333 26333 26334 950 26335 143 26336 146 26337 26337 26338 26338 26339 40 26340 26340 26341 26341 26342 2901 26343 2903 26344 2905 26345 75 26346 2907 26347 2908 26348 26348 26349 264 26350 26350 26351 2 26352 264 26353 75 26354 3856 26355 2753 26356 26356 26357 1823 26358 146 26359 2011 26360 2 26361 26361 26362 2 26363 143 26364 146 26365 26365 26366 26366 26367 26367 26368 26368 26369 26369 26370 5125 26371 26371 26372 146 26373 26373 26374 2 26375 264 26376 26376 26377 2 26378 264 26379 26379 26380 26380 26381 26381 26382 26382 26383 26383 26384 26384 26385 1823 26386 146 26387 26387 26388 26388 26389 26389 26390 26390 26391 264 26392 26392 26393 26393 26394 2 26395 404 26396 26396 26397 5157 26398 26398 26399 26399 26400 26400 26401 830 26402 25737 26403 11 26404 25739 26405 25740 26406 16 26407 2901 26408 2903 26409 2905 26410 75 26411 2907 26412 2908 26413 26413 26414 25741 26415 16 26416 16 26417 16 26418 16 26419 26419 26420 26420 26421 96 26422 26422 26423 104 26424 26424 26425 26425 26426 26426 26427 15828 26428 136 26429 11 26430 138 26431 146 26432 149 26433 151 26434 75 26435 123 26436 154 26437 26437 26438 26438 26439 16 26440 2901 26441 2903 26442 2905 26443 75 26444 2907 26445 2908 26446 26446 26447 2905 26448 75 26449 2907 26450 2908 26451 26451 26452 19716 26453 2905 26454 75 26455 2907 26456 2908 26457 26457 26458 19765 26459 2903 26460 2905 26461 75 26462 2907 26463 2908 26464 26464 26465 2901 26466 2903 26467 2905 26468 75 26469 2907 26470 2908 26471 26471 26472 338 26473 338 26474 341 26475 2 26476 404 26477 25750 26478 555 26479 25752 26480 25753 26481 25754 26482 25755 26483 2 26484 404 26485 25758 26486 25759 26487 25760 26488 25761 26489 1039 26490 5969 26491 26491 26492 15858 26493 15859 26494 5969 26495 26495 26496 2728 26497 25763 26498 555 26499 25765 26500 2 26501 25767 26502 25768 26503 2 26504 3 26505 4 26506 2 26507 2 26508 2 26509 2 26510 11 26511 2 26512 2 26513 25779 26514 2 26515 25781 26516 22 26517 25783 26518 25784 26519 2 26520 11 26521 11 26522 25788 26523 25789 26524 25790 26525 22 26526 2 26527 2 26528 2 26529 11 26530 25796 26531 11 26532 25798 26533 11 26534 25800 26535 25801 26536 25802 26537 25803 26538 60 26539 25805 26540 2 26541 2 26542 2 26543 2 26544 67 26545 127 26546 25812 26547 2 26548 2 26549 25815 26550 2 26551 25817 26552 25818 26553 2 26554 83 26555 84 26556 87 26557 130 26558 11 26559 25825 26560 25826 26561 97 26562 25828 26563 25829 26564 25830 26565 102 26566 102 26567 25833 26568 25834 26569 25835 26570 25836 26571 25837 26572 25838 26573 25839 26574 25840 26575 25841 26576 25842 26577 97 26578 2 26579 25845 26580 2 26581 2 26582 2 26583 2 26584 2 26585 2 26586 2 26587 2 26588 288 26589 2 26590 2 26591 2 26592 2 26593 2 26594 294 26595 2 26596 2 26597 2 26598 2 26599 2 26600 2 26601 2 26602 2 26603 2 26604 2 26605 2 26606 2 26607 2 26608 2 26609 2 26610 25876 26611 25877 26612 25878 26613 25879 26614 25880 26615 25881 26616 25882 26617 25883 26618 25884 26619 25885 26620 25886 26621 25887 26622 25888 26623 25889 26624 25890 26625 4 26626 2 26627 25893 26628 25894 26629 25895 26630 25896 26631 25897 26632 25898 26633 25899 26634 25900 26635 25901 26636 25902 26637 25903 26638 338 26639 338 26640 40 26641 341 26642 2 26643 404 26644 25910 26645 555 26646 25912 26647 25913 26648 25914 26649 25915 26650 25916 26651 2 26652 404 26653 25919 26654 25920 26655 40 26656 25922 26657 264 26658 25924 26659 25925 26660 2 26661 404 26662 25928 26663 25929 26664 25930 26665 25931 26666 25932 26667 25933 26668 2 26669 713 26670 2 26671 25937 26672 25938 26673 25939 26674 25940 26675 25941 26676 25942 26677 25943 26678 25944 26679 25945 26680 25946 26681 830 26682 25948 26683 25949 26684 25950 26685 25951 26686 25952 26687 25953 26688 25954 26689 25955 26690 25956 26691 25957 26692 25958 26693 25959 26694 25960 26695 25961 26696 25962 26697 25963 26698 25964 26699 25965 26700 25966 26701 25967 26702 25968 26703 25969 26704 25970 26705 25971 26706 25972 26707 25973 26708 25974 26709 25975 26710 25976 26711 25977 26712 25978 26713 25979 26714 25980 26715 25981 26716 25982 26717 25983 26718 25984 26719 25985 26720 25986 26721 25987 26722 25988 26723 25989 26724 25990 26725 25991 26726 25992 26727 25993 26728 25994 26729 25995 26730 25996 26731 25997 26732 25998 26733 25999 26734 26000 26735 26001 26736 26002 26737 26003 26738 26004 26739 26005 26740 26006 26741 26007 26742 26008 26743 26009 26744 26010 26745 26011 26746 26012 26747 26013 26748 26014 26749 26015 26750 26016 26751 26017 26752 26018 26753 26019 26754 26020 26755 26021 26756 26022 26757 26023 26758 26024 26759 26025 26760 26026 26761 11 26762 26028 26763 26029 26764 26030 26765 917 26766 26032 26767 26033 26768 950 26769 143 26770 146 26771 26037 26772 26038 26773 1456 26774 2 26775 2 26776 2 26777 26043 26778 2 26779 26045 26780 2 26781 2 26782 2 26783 2 26784 2 26785 2 26786 2 26787 2 26788 2 26789 26055 26790 26056 26791 26057 26792 26058 26793 26059 26794 26060 26795 26061 26796 26062 26797 26063 26798 26064 26799 26065 26800 26066 26801 26067 26802 22 26803 26069 26804 4 26805 26071 26806 87 26807 11 26808 26074 26809 97 26810 26076 26811 1517 26812 26078 26813 26079 26814 26080 26815 97 26816 830 26817 11 26818 26084 26819 917 26820 26086 26821 26087 26822 950 26823 143 26824 146 26825 26091 26826 26092 26827 26093 26828 830 26829 11 26830 26096 26831 26097 26832 26098 26833 338 26834 338 26835 341 26836 2 26837 404 26838 26104 26839 555 26840 26106 26841 26107 26842 26108 26843 26109 26844 2 26845 404 26846 26112 26847 26113 26848 26114 26849 26115 26850 26116 26851 2 26852 404 26853 26119 26854 26120 26855 136 26856 11 26857 138 26858 146 26859 147 26860 149 26861 151 26862 75 26863 123 26864 154 26865 26131 26866 26132 26867 136 26868 11 26869 138 26870 146 26871 2206 26872 2208 26873 2209 26874 2210 26875 2 26876 26142 26877 26143 26878 2 26879 26145 26880 830 26881 26147 26882 11 26883 26149 26884 26150 26885 26151 26886 2653 26887 26153 26888 26154 26889 338 26890 338 26891 341 26892 2 26893 404 26894 26160 26895 555 26896 26162 26897 26163 26898 26164 26899 26165 26900 2 26901 404 26902 26168 26903 26169 26904 26170 26905 26171 26906 2728 26907 26173 26908 555 26909 26175 26910 2 26911 26177 26912 26178 26913 25830 26914 2 26915 26181 26916 26182 26917 26183 26918 26184 26919 26185 26920 2820 26921 26187 26922 67 26923 26189 26924 26190 26925 26191 26926 26192 26927 26193 26928 26194 26929 2854 26930 26202 26931 2863 26932 26205 26933 917 26934 26222 26935 26223 26936 26224 26937 950 26938 143 26939 146 26940 26228 26941 26229 26942 26232 26943 2901 26944 2903 26945 2905 26946 75 26947 2907 26948 2908 26949 26239 26950 2 26951 264 26952 75 26953 3856 26954 2753 26955 26247 26956 1823 26957 146 26958 2011 26959 2 26960 26252 26961 2 26962 143 26963 146 26964 26256 26965 26257 26966 26258 26967 26259 26968 26260 26969 5125 26970 26262 26971 146 26972 26264 26973 2 26974 264 26975 26267 26976 2 26977 264 26978 26270 26979 26271 26980 26272 26981 26273 26982 26274 26983 26275 26984 1823 26985 146 26986 26278 26987 26279 26988 26280 26989 26281 26990 264 26991 26283 26992 26284 26993 2 26994 404 26995 26287 26996 5157 26997 26289 26998 26290 26999 2 27000 264 27001 15936 27002 15937 27003 27003 27004 27004 27005 15940 27006 5969 27007 27007 27008 27008 27009 27009 27010 27010 27011 27011 27012 27012 27013 27013 27014 5969 27015 27015 27016 27016 27017 2965 27018 27018 27019 5969 27020 27013 27021 2965 27022 27018 27023 27023 27024 27024 27025 27025 27026 27026 27027 27027 27028 27028 27029 27029 27030 16 27031 27031 27032 75 27033 27033 27034 27034 27035 27035 27036 27036 27037 27037 27038 3931 27039 27039 27040 2753 27041 27041 27042 27042 27043 16 27044 27044 27045 27045 27046 27046 27047 27047 27048 75 27049 27049 27050 27050 27051 27051 27052 27052 27053 3931 27054 27054 27055 27033 27056 27056 27057 27057 27058 2753 27059 27059 27060 27060 27061 26291 27062 22 27063 143 27064 146 27065 5969 27066 27066 27067 123 27068 154 27069 27069 27070 27070 27071 11 27072 75 27073 27073 27074 123 27075 154 27076 27076 27077 13 27078 16110 27079 146 27080 75 27081 123 27082 154 27083 27083 27084 2905 27085 75 27086 2907 27087 2908 27088 27088 27089 26292 27090 2 27091 264 27092 16123 27093 16124 27094 27094 27095 16126 27096 16127 27097 16128 27098 27098 27099 16128 27100 27100 27101 16263 27102 16264 27103 5969 27104 27104 27105 16264 27106 5969 27107 27107 27108 26293 27109 27109 27110 2965 27111 27111 27112 2820 27113 26296 27114 67 27115 26298 27116 26299 27117 5969 27118 27118 27119 26300 27120 26301 27121 26302 27122 26303 27123 2854 27124 26311 27125 2863 27126 26314 27127 917 27128 26331 27129 26332 27130 26333 27131 950 27132 143 27133 146 27134 26337 27135 26338 27136 26341 27137 2901 27138 2903 27139 2905 27140 75 27141 2907 27142 2908 27143 26348 27144 27144 27145 27145 27146 27146 27147 27147 27148 27148 27149 27149 27150 27150 27151 27151 27152 27152 27153 27153 27154 27154 27155 27155 27156 27156 27157 27157 27158 27158 27159 27159 27160 27160 27161 27161 27162 27162 27163 27163 27164 27164 27165 27165 27166 27166 27167 27167 27168 27168 27169 27169 27170 27170 27171 27171 27172 27172 27173 27173 27174 27174 27175 27175 27176 27176 27177 27177 27178 27178 27179 27179 27180 27180 27181 27181 27182 27182 27183 27183 27184 27184 27185 27185 27186 27186 27187 27187 27188 27188 27189 27189 27190 27190 27191 27191 27192 27192 27193 27193 27194 27194 27195 27195 27196 27196 27197 27197 27198 27198 27199 27199 27200 27200 27201 27201 27202 27202 27203 27203 27204 27204 27205 27205 27206 27206 27207 27207 27208 27208 27209 27209 27210 27210 27211 27211 27212 27212 27213 27213 27214 27214 27215 27215 27216 27216 27217 27217 27218 27218 27219 27219 27220 27220 27221 27221 27222 27222 27223 27223 27224 5720 27225 27225 27226 27226 27227 27227 27228 27228 27229 27229 27230 15 27231 16 27232 16 27233 15 27234 11 27235 26 27236 2 27237 16 27238 11 27239 11 27240 40 27241 16 27242 11 27243 11 27244 27244 27245 16 27246 56 27247 15 27248 68 27249 69 27250 11 27251 16 27252 75 27253 11 27254 78 27255 11 27256 27256 27257 86 27258 27258 27259 90 27260 27260 27261 96 27262 27262 27263 101 27264 104 27265 27265 27266 27266 27267 27267 27268 27268 27269 27269 27270 27270 27271 27271 27272 123 27273 27273 27274 154 27275 27275 27276 26 27277 11 27278 2 27279 264 27280 26 27281 27281 27282 15 27283 15 27284 11 27285 26 27286 2 27287 11 27288 11 27289 40 27290 16 27291 27291 27292 15 27293 68 27294 69 27295 11 27296 75 27297 11 27298 11 27299 27299 27300 86 27301 101 27302 27302 27303 27267 27304 27304 27305 123 27306 154 27307 27307 27308 26 27309 27309 27310 27310 27311 1039 27312 5969 27313 27313 27314 15858 27315 15859 27316 5969 27317 27317 27318 2 27319 264 27320 15936 27321 15937 27322 27322 27323 27323 27324 15940 27325 5969 27326 27326 27327 22 27328 143 27329 146 27330 5969 27331 27331 27332 123 27333 154 27334 27334 27335 27335 27336 11 27337 75 27338 27338 27339 123 27340 154 27341 27341 27342 13 27343 16110 27344 146 27345 75 27346 123 27347 154 27348 27348 27349 2 27350 264 27351 16123 27352 16124 27353 27353 27354 16126 27355 16127 27356 16128 27357 27357 27358 16128 27359 27359 27360 16263 27361 16264 27362 5969 27363 27363 27364 16264 27365 5969 27366 27366 27367 27367 27368 2965 27369 27369 27370 5969 27371 27371 27372 5969 27373 27373 27374 2965 27375 27375 27376 2 27377 17482 27378 18834 27379 18844 27380 15936 27381 18834 27382 146 27383 15936 27384 18894 27385 27385 27386 27386 27387 18897 27388 5969 27389 27389 27390 27390 27391 19057 27392 5969 27393 19057 27394 5969 27395 19065 27396 146 27397 15936 27398 19072 27399 27399 27400 27400 27401 19075 27402 5969 27403 27403 27404 2965 27405 27405 27406 19278 27407 11 27408 2 27409 264 27410 5969 27411 27411 27412 16 27413 19426 27414 15936 27415 19430 27416 27416 27417 27417 27418 19433 27419 5969 27420 27420 27421 5969 27422 27422 27423 5969 27424 27424 27425 27425 27426 15828 27427 136 27428 11 27429 138 27430 146 27431 149 27432 151 27433 75 27434 123 27435 154 27436 27436 27437 27437 27438 27438 27439 16 27440 2901 27441 2903 27442 2905 27443 75 27444 2907 27445 2908 27446 27446 27447 27447 27448 1039 27449 5969 27450 27450 27451 15858 27452 15859 27453 5969 27454 27454 27455 2 27456 264 27457 15936 27458 15937 27459 27459 27460 27460 27461 15940 27462 5969 27463 27463 27464 5969 27465 2965 27466 27018 27467 5969 27468 2965 27469 27469 27470 27470 27471 27471 27472 27472 27473 22 27474 143 27475 146 27476 5969 27477 27477 27478 123 27479 154 27480 27480 27481 27481 27482 11 27483 75 27484 27484 27485 123 27486 154 27487 27487 27488 13 27489 16110 27490 146 27491 75 27492 123 27493 154 27494 27494 27495 2905 27496 75 27497 2907 27498 2908 27499 27499 27500 2 27501 264 27502 16123 27503 16124 27504 27504 27505 16126 27506 16127 27507 16128 27508 27508 27509 16128 27510 27510 27511 16263 27512 16264 27513 5969 27514 27514 27515 16264 27516 5969 27517 27517 27518 27518 27519 2965 27520 27520 27521 5969 27522 27522 27523 5720 27524 27227 27525 27228 27526 27526 27527 27527 27528 27227 27529 27228 27530 27530 27531 19716 27532 2905 27533 75 27534 2907 27535 2908 27536 27536 27537 2 27538 264 27539 15936 27540 27540 27541 146 27542 5969 27543 27543 27544 5969 27545 27545 27546 2965 27547 27547 27548 2 27549 17482 27550 18834 27551 18844 27552 15936 27553 18834 27554 146 27555 15936 27556 18894 27557 27557 27558 27558 27559 18897 27560 5969 27561 27561 27562 27562 27563 19057 27564 5969 27565 19057 27566 5969 27567 19065 27568 146 27569 15936 27570 19072 27571 27571 27572 27572 27573 19075 27574 5969 27575 27575 27576 19765 27577 2903 27578 2905 27579 75 27580 2907 27581 2908 27582 27582 27583 2965 27584 27584 27585 19278 27586 11 27587 2 27588 264 27589 5969 27590 27590 27591 16 27592 19426 27593 15936 27594 19430 27595 27595 27596 27596 27597 19433 27598 5969 27599 27599 27600 5969 27601 27601 27602 2011 27603 27603 27604 2965 27605 16 27606 27606 27607 27607 27608 2 27609 3032 27610 27610 27611 2901 27612 2903 27613 2905 27614 75 27615 2907 27616 2908 27617 27617 27618 5969 27619 27619 27620 27620 27621 27227 27622 27228 27623 27623 27624 2 27625 264 27626 75 27627 3856 27628 2753 27629 26356 27630 1823 27631 146 27632 2011 27633 2 27634 26361 27635 2 27636 143 27637 146 27638 26365 27639 26366 27640 19716 27641 2905 27642 75 27643 2907 27644 2908 27645 27645 27646 26367 27647 26368 27648 26369 27649 2 27650 264 27651 15936 27652 27652 27653 15 27654 15 27655 11 27656 26 27657 2 27658 11 27659 11 27660 40 27661 16 27662 11 27663 27663 27664 15 27665 68 27666 69 27667 11 27668 75 27669 11 27670 11 27671 27671 27672 86 27673 101 27674 27674 27675 27675 27676 27676 27677 123 27678 154 27679 27679 27680 26 27681 11 27682 2 27683 264 27684 26 27685 27685 27686 27686 27687 27687 27688 1039 27689 5969 27690 27690 27691 15858 27692 15859 27693 5969 27694 27694 27695 2 27696 264 27697 15936 27698 15937 27699 27699 27700 27700 27701 15940 27702 5969 27703 27703 27704 22 27705 143 27706 146 27707 5969 27708 27708 27709 123 27710 154 27711 27711 27712 27712 27713 11 27714 75 27715 27715 27716 123 27717 154 27718 27718 27719 13 27720 16110 27721 146 27722 75 27723 123 27724 154 27725 27725 27726 2 27727 264 27728 16123 27729 16124 27730 27730 27731 16126 27732 16127 27733 16128 27734 27734 27735 16128 27736 27736 27737 16263 27738 16264 27739 5969 27740 27740 27741 16264 27742 5969 27743 27743 27744 27744 27745 2965 27746 27746 27747 5969 27748 27748 27749 2 27750 264 27751 15936 27752 27752 27753 5969 27754 27754 27755 2965 27756 27756 27757 2 27758 17482 27759 18834 27760 18844 27761 15936 27762 18834 27763 146 27764 15936 27765 18894 27766 27766 27767 27767 27768 18897 27769 5969 27770 27770 27771 27771 27772 19057 27773 5969 27774 19057 27775 5969 27776 19065 27777 146 27778 15936 27779 19072 27780 27780 27781 27781 27782 19075 27783 5969 27784 27784 27785 2965 27786 27786 27787 19278 27788 11 27789 2 27790 264 27791 5969 27792 27792 27793 16 27794 19426 27795 15936 27796 19430 27797 27797 27798 27798 27799 19433 27800 5969 27801 27801 27802 5969 27803 27803 27804 5969 27805 27805 27806 146 27807 5969 27808 27808 27809 5969 27810 27810 27811 2965 27812 27812 27813 2 27814 17482 27815 18834 27816 18844 27817 15936 27818 18834 27819 146 27820 15936 27821 18894 27822 27822 27823 27823 27824 18897 27825 5969 27826 27826 27827 27827 27828 19057 27829 5969 27830 19057 27831 5969 27832 19065 27833 146 27834 15936 27835 19072 27836 27836 27837 27837 27838 19075 27839 5969 27840 27840 27841 5125 27842 26371 27843 146 27844 26373 27845 2 27846 264 27847 26376 27848 19765 27849 2903 27850 2905 27851 75 27852 2907 27853 2908 27854 27854 27855 2 27856 264 27857 26379 27858 2965 27859 27859 27860 26380 27861 19278 27862 11 27863 2 27864 264 27865 5969 27866 27866 27867 16 27868 19426 27869 15936 27870 19430 27871 27871 27872 27872 27873 19433 27874 5969 27875 27875 27876 26381 27877 5969 27878 27878 27879 2965 27880 27880 27881 26382 27882 2901 27883 2903 27884 2905 27885 75 27886 2907 27887 2908 27888 27888 27889 5969 27890 27890 27891 26383 27892 6356 27893 555 27894 27894 27895 26384 27896 1823 27897 146 27898 26387 27899 26388 27900 26389 27901 26390 27902 264 27903 26392 27904 26393 27905 2 27906 404 27907 26396 27908 5157 27909 26398 27910 26399))
  "The generated action-tables.")

(defvar
  phps-mode-parser--distinct-action-tables
  #s(hash-table size 12466 test equal rehash-size 1.5 rehash-threshold 0.8125 data (0 ((("!") reduce 84) (("\"") reduce 84) (($) reduce 84) (("$") reduce 84) (("(") reduce 84) (("+") reduce 84) (("-") reduce 84) ((";") reduce 84) (("@") reduce 84) ((T_ABSTRACT) reduce 84) ((T_ARRAY) reduce 84) ((T_ARRAY_CAST) reduce 84) ((T_ATTRIBUTE) reduce 84) ((T_BOOL_CAST) reduce 84) ((T_BREAK) reduce 84) ((T_CLASS) reduce 84) ((T_CLASS_C) reduce 84) ((T_CLONE) reduce 84) ((T_CONST) reduce 84) ((T_CONSTANT_ENCAPSED_STRING) reduce 84) ((T_CONTINUE) reduce 84) ((T_DEC) reduce 84) ((T_DECLARE) reduce 84) ((T_DIR) reduce 84) ((T_DNUMBER) reduce 84) ((T_DO) reduce 84) ((T_DOUBLE_CAST) reduce 84) ((T_ECHO) reduce 84) ((T_EMPTY) reduce 84) ((T_ENUM) reduce 84) ((T_EVAL) reduce 84) ((T_EXIT) reduce 84) ((T_FILE) reduce 84) ((T_FINAL) reduce 84) ((T_FN) reduce 84) ((T_FOR) reduce 84) ((T_FOREACH) reduce 84) ((T_FUNCTION) reduce 84) ((T_FUNC_C) reduce 84) ((T_GLOBAL) reduce 84) ((T_GOTO) reduce 84) ((T_HALT_COMPILER) reduce 84) ((T_IF) reduce 84) ((T_INC) reduce 84) ((T_INCLUDE) reduce 84) ((T_INCLUDE_ONCE) reduce 84) ((T_INLINE_HTML) reduce 84) ((T_INTERFACE) reduce 84) ((T_INT_CAST) reduce 84) ((T_ISSET) reduce 84) ((T_LINE) reduce 84) ((T_LIST) reduce 84) ((T_LNUMBER) reduce 84) ((T_MATCH) reduce 84) ((T_METHOD_C) reduce 84) ((T_NAMESPACE) reduce 84) ((T_NAME_FULLY_QUALIFIED) reduce 84) ((T_NAME_QUALIFIED) reduce 84) ((T_NAME_RELATIVE) reduce 84) ((T_NEW) reduce 84) ((T_NS_C) reduce 84) ((T_OBJECT_CAST) reduce 84) ((T_PRINT) reduce 84) ((T_REQUIRE) reduce 84) ((T_REQUIRE_ONCE) reduce 84) ((T_RETURN) reduce 84) ((T_START_HEREDOC) reduce 84) ((T_STATIC) reduce 84) ((T_STRING) reduce 84) ((T_STRING_CAST) reduce 84) ((T_SWITCH) reduce 84) ((T_THROW) reduce 84) ((T_TRAIT) reduce 84) ((T_TRAIT_C) reduce 84) ((T_TRY) reduce 84) ((T_UNSET) reduce 84) ((T_UNSET_CAST) reduce 84) ((T_USE) reduce 84) ((T_VARIABLE) reduce 84) ((T_WHILE) reduce 84) ((T_YIELD) reduce 84) ((T_YIELD_FROM) reduce 84) (("[") reduce 84) (("`") reduce 84) (("{") reduce 84) (("~") reduce 84)) 1 ((("!") shift) (("\"") shift) (($) accept) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_USE) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 2 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 3 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_VARIABLE) shift)) 4 ((("$") shift) ((T_VARIABLE) shift) (("{") shift)) 8 ((("!") reduce 164) (("\"") reduce 164) (($) reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 10 (((T_ABSTRACT) reduce 188) ((T_CLASS) reduce 188) ((T_FINAL) reduce 188)) 11 ((("(") shift)) 13 (((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift)) 15 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") reduce 493) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 16 (((T_STRING) shift)) 17 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 20 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 22 ((("\"") shift) (("$") shift) (("(") shift) ((T_ARRAY) shift) ((T_CLASS_C) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DIR) shift) ((T_FILE) shift) ((T_FUNC_C) shift) ((T_LINE) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NS_C) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_TRAIT_C) shift) ((T_VARIABLE) shift) (("[") shift)) 24 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 25 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 26 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 32 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) ((";") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 33 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 34 (((T_ABSTRACT) reduce 189) ((T_CLASS) reduce 189) ((T_FINAL) reduce 189)) 35 ((("(") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441)) 38 ((("(") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_STRING) reduce 442)) 39 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 40 ((("$") shift) ((T_VARIABLE) shift)) 47 ((("!") reduce 158) (("\"") reduce 158) (($) reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 51 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) ((";") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 53 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 55 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 56 (((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_QUALIFIED) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift) (("{") shift)) 57 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 58 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 59 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 60 ((("$") shift) (("(") shift) ((T_ATTRIBUTE) shift) ((T_CLASS) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_VARIABLE) shift)) 61 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 67 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_END_HEREDOC) shift) ((T_VARIABLE) shift)) 68 (((T_FN) shift) ((T_FUNCTION) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_VARIABLE) shift)) 69 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") shift) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 74 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 75 ((("{") shift)) 78 (((T_CONST) shift) ((T_FUNCTION) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_STRING) shift)) 79 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 81 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 83 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) ((",") reduce 534) (("-") shift) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_ELLIPSIS) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 534) (("`") shift) (("~") shift)) 84 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_VARIABLE) shift) (("`") reduce 465)) 85 ((("!") reduce 147) (("\"") reduce 147) (($) reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 86 (((T_ELSE) shift) ((T_ELSEIF) shift) ((T_ENDIF) shift)) 87 (((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) (("[") shift) (("{") shift)) 88 (((T_ABSTRACT) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_CLASS) reduce 100) ((T_ENUM) reduce 100) ((T_FINAL) reduce 100) ((T_FN) reduce 100) ((T_FUNCTION) reduce 100) ((T_INTERFACE) reduce 100) ((T_STATIC) reduce 100) ((T_TRAIT) reduce 100)) 89 ((("!") reduce 108) (("\"") reduce 108) (("$") reduce 108) (($) reduce 108) (("(") reduce 108) (("+") reduce 108) (("-") reduce 108) ((";") reduce 108) (("@") reduce 108) ((T_ABSTRACT) reduce 108) ((T_ARRAY) reduce 108) ((T_ARRAY_CAST) reduce 108) ((T_ATTRIBUTE) reduce 108) ((T_BOOL_CAST) reduce 108) ((T_BREAK) reduce 108) ((T_CLASS) reduce 108) ((T_CLASS_C) reduce 108) ((T_CLONE) reduce 108) ((T_CONST) reduce 108) ((T_CONSTANT_ENCAPSED_STRING) reduce 108) ((T_CONTINUE) reduce 108) ((T_DEC) reduce 108) ((T_DECLARE) reduce 108) ((T_DIR) reduce 108) ((T_DNUMBER) reduce 108) ((T_DO) reduce 108) ((T_DOUBLE_CAST) reduce 108) ((T_ECHO) reduce 108) ((T_EMPTY) reduce 108) ((T_ENUM) reduce 108) ((T_EVAL) reduce 108) ((T_EXIT) reduce 108) ((T_FILE) reduce 108) ((T_FINAL) reduce 108) ((T_FN) reduce 108) ((T_FOR) reduce 108) ((T_FOREACH) reduce 108) ((T_FUNCTION) reduce 108) ((T_FUNC_C) reduce 108) ((T_GLOBAL) reduce 108) ((T_GOTO) reduce 108) ((T_HALT_COMPILER) reduce 108) ((T_IF) reduce 108) ((T_INC) reduce 108) ((T_INCLUDE) reduce 108) ((T_INCLUDE_ONCE) reduce 108) ((T_INLINE_HTML) reduce 108) ((T_INTERFACE) reduce 108) ((T_INT_CAST) reduce 108) ((T_ISSET) reduce 108) ((T_LINE) reduce 108) ((T_LIST) reduce 108) ((T_LNUMBER) reduce 108) ((T_MATCH) reduce 108) ((T_METHOD_C) reduce 108) ((T_NAMESPACE) reduce 108) ((T_NAME_FULLY_QUALIFIED) reduce 108) ((T_NAME_QUALIFIED) reduce 108) ((T_NAME_RELATIVE) reduce 108) ((T_NEW) reduce 108) ((T_NS_C) reduce 108) ((T_OBJECT_CAST) reduce 108) ((T_PRINT) reduce 108) ((T_REQUIRE) reduce 108) ((T_REQUIRE_ONCE) reduce 108) ((T_RETURN) reduce 108) ((T_START_HEREDOC) reduce 108) ((T_STATIC) reduce 108) ((T_STRING) reduce 108) ((T_STRING_CAST) reduce 108) ((T_SWITCH) reduce 108) ((T_THROW) reduce 108) ((T_TRAIT) reduce 108) ((T_TRAIT_C) reduce 108) ((T_TRY) reduce 108) ((T_UNSET) reduce 108) ((T_UNSET_CAST) reduce 108) ((T_USE) reduce 108) ((T_VARIABLE) reduce 108) ((T_WHILE) reduce 108) ((T_YIELD) reduce 108) ((T_YIELD_FROM) reduce 108) (("[") reduce 108) (("`") reduce 108) (("{") reduce 108) (("~") reduce 108)) 90 (((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CLASS) shift) ((T_ENUM) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_INTERFACE) shift) ((T_STATIC) shift) ((T_TRAIT) shift)) 92 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 93 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) ((";") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 94 ((("!") reduce 103) (("\"") reduce 103) (($) reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONST) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAMESPACE) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_USE) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 95 (((T_ABSTRACT) reduce 186) ((T_CLASS) reduce 186) ((T_FINAL) reduce 186)) 96 (((T_ABSTRACT) shift) ((T_CLASS) shift) ((T_FINAL) shift)) 97 (((T_PAAMAYIM_NEKUDOTAYIM) shift)) 98 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) ((";") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 99 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 100 ((("!") reduce 106) (("\"") reduce 106) (($) reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONST) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAMESPACE) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_USE) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 101 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 102 ((("(") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift)) 103 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 500) ((T_OBJECT_OPERATOR) reduce 500) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 495) (("[") reduce 500) (("{") reduce 500)) 104 ((("(") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_STRING) reduce 446)) 105 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 106 ((("!") reduce 102) (("\"") reduce 102) (($) reduce 102) (("$") reduce 102) (("(") reduce 102) (("+") reduce 102) (("-") reduce 102) ((";") reduce 102) (("@") reduce 102) ((T_ABSTRACT) reduce 102) ((T_ARRAY) reduce 102) ((T_ARRAY_CAST) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_BOOL_CAST) reduce 102) ((T_BREAK) reduce 102) ((T_CLASS) reduce 102) ((T_CLASS_C) reduce 102) ((T_CLONE) reduce 102) ((T_CONST) reduce 102) ((T_CONSTANT_ENCAPSED_STRING) reduce 102) ((T_CONTINUE) reduce 102) ((T_DEC) reduce 102) ((T_DECLARE) reduce 102) ((T_DIR) reduce 102) ((T_DNUMBER) reduce 102) ((T_DO) reduce 102) ((T_DOUBLE_CAST) reduce 102) ((T_ECHO) reduce 102) ((T_EMPTY) reduce 102) ((T_ENUM) reduce 102) ((T_EVAL) reduce 102) ((T_EXIT) reduce 102) ((T_FILE) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FOR) reduce 102) ((T_FOREACH) reduce 102) ((T_FUNCTION) reduce 102) ((T_FUNC_C) reduce 102) ((T_GLOBAL) reduce 102) ((T_GOTO) reduce 102) ((T_HALT_COMPILER) reduce 102) ((T_IF) reduce 102) ((T_INC) reduce 102) ((T_INCLUDE) reduce 102) ((T_INCLUDE_ONCE) reduce 102) ((T_INLINE_HTML) reduce 102) ((T_INTERFACE) reduce 102) ((T_INT_CAST) reduce 102) ((T_ISSET) reduce 102) ((T_LINE) reduce 102) ((T_LIST) reduce 102) ((T_LNUMBER) reduce 102) ((T_MATCH) reduce 102) ((T_METHOD_C) reduce 102) ((T_NAMESPACE) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_NEW) reduce 102) ((T_NS_C) reduce 102) ((T_OBJECT_CAST) reduce 102) ((T_PRINT) reduce 102) ((T_REQUIRE) reduce 102) ((T_REQUIRE_ONCE) reduce 102) ((T_RETURN) reduce 102) ((T_START_HEREDOC) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_STRING_CAST) reduce 102) ((T_SWITCH) reduce 102) ((T_THROW) reduce 102) ((T_TRAIT) reduce 102) ((T_TRAIT_C) reduce 102) ((T_TRY) reduce 102) ((T_UNSET) reduce 102) ((T_UNSET_CAST) reduce 102) ((T_USE) reduce 102) ((T_VARIABLE) reduce 102) ((T_WHILE) reduce 102) ((T_YIELD) reduce 102) ((T_YIELD_FROM) reduce 102) (("[") reduce 102) (("`") reduce 102) (("{") reduce 102) (("~") reduce 102)) 107 ((("!") reduce 146) (("\"") reduce 146) (($) reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONST) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAMESPACE) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_USE) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 108 ((("!") reduce 236) (("\"") reduce 236) (("$") reduce 236) (($) reduce 236) (("(") reduce 236) (("+") reduce 236) (("-") reduce 236) ((";") reduce 236) (("@") reduce 236) ((T_ABSTRACT) reduce 236) ((T_ARRAY) reduce 236) ((T_ARRAY_CAST) reduce 236) ((T_ATTRIBUTE) reduce 236) ((T_BOOL_CAST) reduce 236) ((T_BREAK) reduce 236) ((T_CLASS) reduce 236) ((T_CLASS_C) reduce 236) ((T_CLONE) reduce 236) ((T_CONST) reduce 236) ((T_CONSTANT_ENCAPSED_STRING) reduce 236) ((T_CONTINUE) reduce 236) ((T_DEC) reduce 236) ((T_DECLARE) reduce 236) ((T_DIR) reduce 236) ((T_DNUMBER) reduce 236) ((T_DO) reduce 236) ((T_DOUBLE_CAST) reduce 236) ((T_ECHO) reduce 236) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 236) ((T_ENUM) reduce 236) ((T_EVAL) reduce 236) ((T_EXIT) reduce 236) ((T_FILE) reduce 236) ((T_FINAL) reduce 236) ((T_FN) reduce 236) ((T_FOR) reduce 236) ((T_FOREACH) reduce 236) ((T_FUNCTION) reduce 236) ((T_FUNC_C) reduce 236) ((T_GLOBAL) reduce 236) ((T_GOTO) reduce 236) ((T_HALT_COMPILER) reduce 236) ((T_IF) reduce 236) ((T_INC) reduce 236) ((T_INCLUDE) reduce 236) ((T_INCLUDE_ONCE) reduce 236) ((T_INLINE_HTML) reduce 236) ((T_INTERFACE) reduce 236) ((T_INT_CAST) reduce 236) ((T_ISSET) reduce 236) ((T_LINE) reduce 236) ((T_LIST) reduce 236) ((T_LNUMBER) reduce 236) ((T_MATCH) reduce 236) ((T_METHOD_C) reduce 236) ((T_NAMESPACE) reduce 236) ((T_NAME_FULLY_QUALIFIED) reduce 236) ((T_NAME_QUALIFIED) reduce 236) ((T_NAME_RELATIVE) reduce 236) ((T_NEW) reduce 236) ((T_NS_C) reduce 236) ((T_OBJECT_CAST) reduce 236) ((T_PRINT) reduce 236) ((T_REQUIRE) reduce 236) ((T_REQUIRE_ONCE) reduce 236) ((T_RETURN) reduce 236) ((T_START_HEREDOC) reduce 236) ((T_STATIC) reduce 236) ((T_STRING) reduce 236) ((T_STRING_CAST) reduce 236) ((T_SWITCH) reduce 236) ((T_THROW) reduce 236) ((T_TRAIT) reduce 236) ((T_TRAIT_C) reduce 236) ((T_TRY) reduce 236) ((T_UNSET) reduce 236) ((T_UNSET_CAST) reduce 236) ((T_USE) reduce 236) ((T_VARIABLE) reduce 236) ((T_WHILE) reduce 236) ((T_YIELD) reduce 236) ((T_YIELD_FROM) reduce 236) (("[") reduce 236) (("`") reduce 236) (("{") reduce 236) (("~") reduce 236)) 109 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) ((";") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 110 ((("!") reduce 105) (("\"") reduce 105) (($) reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONST) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAMESPACE) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_USE) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 111 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) ((";") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 112 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 113 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) ((";") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 114 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) ((";") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 115 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 116 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 117 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (($) reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONST) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAMESPACE) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_USE) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 118 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 119 ((("!") reduce 83) (("\"") reduce 83) (($) reduce 83) (("$") reduce 83) (("(") reduce 83) (("+") reduce 83) (("-") reduce 83) ((";") reduce 83) (("@") reduce 83) ((T_ABSTRACT) reduce 83) ((T_ARRAY) reduce 83) ((T_ARRAY_CAST) reduce 83) ((T_ATTRIBUTE) reduce 83) ((T_BOOL_CAST) reduce 83) ((T_BREAK) reduce 83) ((T_CLASS) reduce 83) ((T_CLASS_C) reduce 83) ((T_CLONE) reduce 83) ((T_CONST) reduce 83) ((T_CONSTANT_ENCAPSED_STRING) reduce 83) ((T_CONTINUE) reduce 83) ((T_DEC) reduce 83) ((T_DECLARE) reduce 83) ((T_DIR) reduce 83) ((T_DNUMBER) reduce 83) ((T_DO) reduce 83) ((T_DOUBLE_CAST) reduce 83) ((T_ECHO) reduce 83) ((T_EMPTY) reduce 83) ((T_ENUM) reduce 83) ((T_EVAL) reduce 83) ((T_EXIT) reduce 83) ((T_FILE) reduce 83) ((T_FINAL) reduce 83) ((T_FN) reduce 83) ((T_FOR) reduce 83) ((T_FOREACH) reduce 83) ((T_FUNCTION) reduce 83) ((T_FUNC_C) reduce 83) ((T_GLOBAL) reduce 83) ((T_GOTO) reduce 83) ((T_HALT_COMPILER) reduce 83) ((T_IF) reduce 83) ((T_INC) reduce 83) ((T_INCLUDE) reduce 83) ((T_INCLUDE_ONCE) reduce 83) ((T_INLINE_HTML) reduce 83) ((T_INTERFACE) reduce 83) ((T_INT_CAST) reduce 83) ((T_ISSET) reduce 83) ((T_LINE) reduce 83) ((T_LIST) reduce 83) ((T_LNUMBER) reduce 83) ((T_MATCH) reduce 83) ((T_METHOD_C) reduce 83) ((T_NAMESPACE) reduce 83) ((T_NAME_FULLY_QUALIFIED) reduce 83) ((T_NAME_QUALIFIED) reduce 83) ((T_NAME_RELATIVE) reduce 83) ((T_NEW) reduce 83) ((T_NS_C) reduce 83) ((T_OBJECT_CAST) reduce 83) ((T_PRINT) reduce 83) ((T_REQUIRE) reduce 83) ((T_REQUIRE_ONCE) reduce 83) ((T_RETURN) reduce 83) ((T_START_HEREDOC) reduce 83) ((T_STATIC) reduce 83) ((T_STRING) reduce 83) ((T_STRING_CAST) reduce 83) ((T_SWITCH) reduce 83) ((T_THROW) reduce 83) ((T_TRAIT) reduce 83) ((T_TRAIT_C) reduce 83) ((T_TRY) reduce 83) ((T_UNSET) reduce 83) ((T_UNSET_CAST) reduce 83) ((T_USE) reduce 83) ((T_VARIABLE) reduce 83) ((T_WHILE) reduce 83) ((T_YIELD) reduce 83) ((T_YIELD_FROM) reduce 83) (("[") reduce 83) (("`") reduce 83) (("{") reduce 83) (("~") reduce 83)) 120 ((("!") reduce 104) (("\"") reduce 104) (($) reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONST) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAMESPACE) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_USE) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 121 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) ((";") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 123 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("}") reduce 140) (("~") reduce 140)) 126 ((("(") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442)) 127 (((T_FN) shift) ((T_FUNCTION) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458)) 128 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 129 (((T_ATTRIBUTE) reduce 100) ((T_FN) reduce 100) ((T_FUNCTION) reduce 100) ((T_STATIC) reduce 100)) 130 (((T_ATTRIBUTE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_STATIC) shift)) 131 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 133 ((("(") reduce 79)) 134 ((("(") reduce 80)) 135 ((("(") reduce 447)) 136 ((("(") reduce 443)) 138 (((")") reduce 243) (("?") reduce 248) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 248) ((T_ARRAY) reduce 248) ((T_ATTRIBUTE) shift) ((T_CALLABLE) reduce 248) ((T_ELLIPSIS) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_PRIVATE) reduce 248) ((T_PROTECTED) reduce 248) ((T_PUBLIC) reduce 248) ((T_READONLY) reduce 248) ((T_STRING) reduce 248) ((T_VARIABLE) reduce 248)) 140 ((("?") reduce 100) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 100) ((T_ARRAY) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_CALLABLE) reduce 100) ((T_ELLIPSIS) reduce 100) ((T_NAME_FULLY_QUALIFIED) reduce 100) ((T_NAME_QUALIFIED) reduce 100) ((T_NAME_RELATIVE) reduce 100) ((T_PRIVATE) reduce 100) ((T_PROTECTED) reduce 100) ((T_PUBLIC) reduce 100) ((T_READONLY) reduce 100) ((T_STRING) reduce 100) ((T_VARIABLE) reduce 100)) 141 (((")") reduce 244) ((",") reduce 244)) 142 ((("?") reduce 248) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 248) ((T_ARRAY) reduce 248) ((T_ATTRIBUTE) shift) ((T_CALLABLE) reduce 248) ((T_ELLIPSIS) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_PRIVATE) reduce 248) ((T_PROTECTED) reduce 248) ((T_PUBLIC) reduce 248) ((T_READONLY) reduce 248) ((T_STRING) reduce 248) ((T_VARIABLE) reduce 248)) 143 (((")") reduce 123) ((",") shift)) 144 ((("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 256) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_ELLIPSIS) reduce 256) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_PRIVATE) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_STRING) shift) ((T_VARIABLE) reduce 256)) 145 (((")") reduce 247) ((",") reduce 247)) 146 (((")") shift)) 147 (((":") reduce 448) ((T_USE) shift) (("{") reduce 448)) 149 (((":") shift) (("{") reduce 279)) 150 ((("?") shift) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift)) 151 ((("{") reduce 444)) 154 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") shift) (("~") shift)) 155 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 169 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 179 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 181 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("}") reduce 142) (("~") reduce 142)) 183 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("}") reduce 103) (("~") reduce 103)) 185 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("}") reduce 106) (("~") reduce 106)) 188 ((("!") reduce 102) (("\"") reduce 102) (("$") reduce 102) (("(") reduce 102) (("+") reduce 102) (("-") reduce 102) ((";") reduce 102) (("@") reduce 102) ((T_ABSTRACT) reduce 102) ((T_ARRAY) reduce 102) ((T_ARRAY_CAST) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_BOOL_CAST) reduce 102) ((T_BREAK) reduce 102) ((T_CLASS) reduce 102) ((T_CLASS_C) reduce 102) ((T_CLONE) reduce 102) ((T_CONSTANT_ENCAPSED_STRING) reduce 102) ((T_CONTINUE) reduce 102) ((T_DEC) reduce 102) ((T_DECLARE) reduce 102) ((T_DIR) reduce 102) ((T_DNUMBER) reduce 102) ((T_DO) reduce 102) ((T_DOUBLE_CAST) reduce 102) ((T_ECHO) reduce 102) ((T_EMPTY) reduce 102) ((T_ENUM) reduce 102) ((T_EVAL) reduce 102) ((T_EXIT) reduce 102) ((T_FILE) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FOR) reduce 102) ((T_FOREACH) reduce 102) ((T_FUNCTION) reduce 102) ((T_FUNC_C) reduce 102) ((T_GLOBAL) reduce 102) ((T_GOTO) reduce 102) ((T_HALT_COMPILER) reduce 102) ((T_IF) reduce 102) ((T_INC) reduce 102) ((T_INCLUDE) reduce 102) ((T_INCLUDE_ONCE) reduce 102) ((T_INLINE_HTML) reduce 102) ((T_INTERFACE) reduce 102) ((T_INT_CAST) reduce 102) ((T_ISSET) reduce 102) ((T_LINE) reduce 102) ((T_LIST) reduce 102) ((T_LNUMBER) reduce 102) ((T_MATCH) reduce 102) ((T_METHOD_C) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_NEW) reduce 102) ((T_NS_C) reduce 102) ((T_OBJECT_CAST) reduce 102) ((T_PRINT) reduce 102) ((T_REQUIRE) reduce 102) ((T_REQUIRE_ONCE) reduce 102) ((T_RETURN) reduce 102) ((T_START_HEREDOC) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_STRING_CAST) reduce 102) ((T_SWITCH) reduce 102) ((T_THROW) reduce 102) ((T_TRAIT) reduce 102) ((T_TRAIT_C) reduce 102) ((T_TRY) reduce 102) ((T_UNSET) reduce 102) ((T_UNSET_CAST) reduce 102) ((T_VARIABLE) reduce 102) ((T_WHILE) reduce 102) ((T_YIELD) reduce 102) ((T_YIELD_FROM) reduce 102) (("[") reduce 102) (("`") reduce 102) (("{") reduce 102) (("}") reduce 102) (("~") reduce 102)) 189 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 190 ((("!") reduce 236) (("\"") reduce 236) (("$") reduce 236) (("(") reduce 236) (("+") reduce 236) (("-") reduce 236) ((";") reduce 236) (("@") reduce 236) ((T_ABSTRACT) reduce 236) ((T_ARRAY) reduce 236) ((T_ARRAY_CAST) reduce 236) ((T_ATTRIBUTE) reduce 236) ((T_BOOL_CAST) reduce 236) ((T_BREAK) reduce 236) ((T_CLASS) reduce 236) ((T_CLASS_C) reduce 236) ((T_CLONE) reduce 236) ((T_CONSTANT_ENCAPSED_STRING) reduce 236) ((T_CONTINUE) reduce 236) ((T_DEC) reduce 236) ((T_DECLARE) reduce 236) ((T_DIR) reduce 236) ((T_DNUMBER) reduce 236) ((T_DO) reduce 236) ((T_DOUBLE_CAST) reduce 236) ((T_ECHO) reduce 236) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 236) ((T_ENUM) reduce 236) ((T_EVAL) reduce 236) ((T_EXIT) reduce 236) ((T_FILE) reduce 236) ((T_FINAL) reduce 236) ((T_FN) reduce 236) ((T_FOR) reduce 236) ((T_FOREACH) reduce 236) ((T_FUNCTION) reduce 236) ((T_FUNC_C) reduce 236) ((T_GLOBAL) reduce 236) ((T_GOTO) reduce 236) ((T_HALT_COMPILER) reduce 236) ((T_IF) reduce 236) ((T_INC) reduce 236) ((T_INCLUDE) reduce 236) ((T_INCLUDE_ONCE) reduce 236) ((T_INLINE_HTML) reduce 236) ((T_INTERFACE) reduce 236) ((T_INT_CAST) reduce 236) ((T_ISSET) reduce 236) ((T_LINE) reduce 236) ((T_LIST) reduce 236) ((T_LNUMBER) reduce 236) ((T_MATCH) reduce 236) ((T_METHOD_C) reduce 236) ((T_NAME_FULLY_QUALIFIED) reduce 236) ((T_NAME_QUALIFIED) reduce 236) ((T_NAME_RELATIVE) reduce 236) ((T_NEW) reduce 236) ((T_NS_C) reduce 236) ((T_OBJECT_CAST) reduce 236) ((T_PRINT) reduce 236) ((T_REQUIRE) reduce 236) ((T_REQUIRE_ONCE) reduce 236) ((T_RETURN) reduce 236) ((T_START_HEREDOC) reduce 236) ((T_STATIC) reduce 236) ((T_STRING) reduce 236) ((T_STRING_CAST) reduce 236) ((T_SWITCH) reduce 236) ((T_THROW) reduce 236) ((T_TRAIT) reduce 236) ((T_TRAIT_C) reduce 236) ((T_TRY) reduce 236) ((T_UNSET) reduce 236) ((T_UNSET_CAST) reduce 236) ((T_VARIABLE) reduce 236) ((T_WHILE) reduce 236) ((T_YIELD) reduce 236) ((T_YIELD_FROM) reduce 236) (("[") reduce 236) (("`") reduce 236) (("{") reduce 236) (("}") reduce 236) (("~") reduce 236)) 191 ((("!") reduce 139) (("\"") reduce 139) (("$") reduce 139) (("(") reduce 139) (("+") reduce 139) (("-") reduce 139) ((";") reduce 139) (("@") reduce 139) ((T_ABSTRACT) reduce 139) ((T_ARRAY) reduce 139) ((T_ARRAY_CAST) reduce 139) ((T_ATTRIBUTE) reduce 139) ((T_BOOL_CAST) reduce 139) ((T_BREAK) reduce 139) ((T_CLASS) reduce 139) ((T_CLASS_C) reduce 139) ((T_CLONE) reduce 139) ((T_CONSTANT_ENCAPSED_STRING) reduce 139) ((T_CONTINUE) reduce 139) ((T_DEC) reduce 139) ((T_DECLARE) reduce 139) ((T_DIR) reduce 139) ((T_DNUMBER) reduce 139) ((T_DO) reduce 139) ((T_DOUBLE_CAST) reduce 139) ((T_ECHO) reduce 139) ((T_EMPTY) reduce 139) ((T_ENUM) reduce 139) ((T_EVAL) reduce 139) ((T_EXIT) reduce 139) ((T_FILE) reduce 139) ((T_FINAL) reduce 139) ((T_FN) reduce 139) ((T_FOR) reduce 139) ((T_FOREACH) reduce 139) ((T_FUNCTION) reduce 139) ((T_FUNC_C) reduce 139) ((T_GLOBAL) reduce 139) ((T_GOTO) reduce 139) ((T_HALT_COMPILER) reduce 139) ((T_IF) reduce 139) ((T_INC) reduce 139) ((T_INCLUDE) reduce 139) ((T_INCLUDE_ONCE) reduce 139) ((T_INLINE_HTML) reduce 139) ((T_INTERFACE) reduce 139) ((T_INT_CAST) reduce 139) ((T_ISSET) reduce 139) ((T_LINE) reduce 139) ((T_LIST) reduce 139) ((T_LNUMBER) reduce 139) ((T_MATCH) reduce 139) ((T_METHOD_C) reduce 139) ((T_NAME_FULLY_QUALIFIED) reduce 139) ((T_NAME_QUALIFIED) reduce 139) ((T_NAME_RELATIVE) reduce 139) ((T_NEW) reduce 139) ((T_NS_C) reduce 139) ((T_OBJECT_CAST) reduce 139) ((T_PRINT) reduce 139) ((T_REQUIRE) reduce 139) ((T_REQUIRE_ONCE) reduce 139) ((T_RETURN) reduce 139) ((T_START_HEREDOC) reduce 139) ((T_STATIC) reduce 139) ((T_STRING) reduce 139) ((T_STRING_CAST) reduce 139) ((T_SWITCH) reduce 139) ((T_THROW) reduce 139) ((T_TRAIT) reduce 139) ((T_TRAIT_C) reduce 139) ((T_TRY) reduce 139) ((T_UNSET) reduce 139) ((T_UNSET_CAST) reduce 139) ((T_VARIABLE) reduce 139) ((T_WHILE) reduce 139) ((T_YIELD) reduce 139) ((T_YIELD_FROM) reduce 139) (("[") reduce 139) (("`") reduce 139) (("{") reduce 139) (("}") reduce 139) (("~") reduce 139)) 192 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("}") reduce 105) (("~") reduce 105)) 193 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("}") reduce 141) (("~") reduce 141)) 194 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("}") reduce 104) (("~") reduce 104)) 196 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 197 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 199 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("}") reduce 145) (("~") reduce 145)) 213 ((("%") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 215 ((("%") reduce 472) (("(") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 217 ((("%") reduce 485) ((")") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 218 ((("%") reduce 475) ((")") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 222 ((("%") reduce 463) (("(") shift) ((")") reduce 463) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 223 ((("%") reduce 484) ((")") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 224 ((("%") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 230 ((("%") reduce 483) ((")") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 232 ((("%") reduce 474) ((")") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 234 ((("%") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 235 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 236 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 237 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 239 ((("%") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 246 ((("%") reduce 91) (("(") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 249 ((("%") reduce 486) ((")") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 251 ((("%") reduce 515) (("(") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 252 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) ((")") reduce 429) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 259 ((("%") reduce 511) (("(") reduce 502) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 260 ((("%") reduce 481) ((")") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 262 ((("%") reduce 480) ((")") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 263 ((("%") reduce 479) (("(") reduce 504) ((")") reduce 479) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 264 ((("%") shift) ((")") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 267 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 268 ((("%") reduce 434) ((")") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 269 ((("%") reduce 416) ((")") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 270 ((("%") reduce 438) ((")") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 271 ((("%") reduce 482) (("(") shift) ((")") reduce 482) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 272 ((("%") reduce 412) ((")") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 273 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 274 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 275 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 276 ((("%") reduce 356) ((")") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 279 ((("%") reduce 400) ((")") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 288 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 294 ((("$") shift) (("(") shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_VARIABLE) shift)) 310 ((("%") shift) ((")") reduce 384) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 311 ((("%") shift) ((")") reduce 387) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 312 ((("%") shift) ((")") reduce 396) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 313 ((("%") shift) ((")") reduce 409) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 314 ((("%") shift) ((")") reduce 395) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 315 ((("%") reduce 392) ((")") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 316 ((("%") shift) ((")") reduce 383) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 317 ((("%") shift) ((")") reduce 381) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 318 ((("%") shift) ((")") reduce 382) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 319 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 320 ((("%") shift) ((")") reduce 402) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 321 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 322 ((("%") shift) ((")") reduce 401) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 323 ((("%") shift) ((")") reduce 408) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 324 ((("%") shift) ((")") reduce 403) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 327 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 328 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 329 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 330 ((("%") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 331 ((("%") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 332 ((("%") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 333 ((("%") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 334 ((("%") reduce 410) ((")") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 335 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 336 ((("%") reduce 461) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 337 ((("%") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 338 ((("$") shift) ((T_STRING) shift) ((T_VARIABLE) shift) (("{") shift)) 341 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 493) (("`") shift) (("~") shift)) 353 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490) (("}") reduce 490)) 355 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472) (("}") reduce 472)) 357 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485) (("}") reduce 485)) 358 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475) (("}") reduce 475)) 362 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463) (("}") reduce 463)) 363 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484) (("}") reduce 484)) 364 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488) (("}") reduce 488)) 370 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483) (("}") reduce 483)) 372 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474) (("}") reduce 474)) 374 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487) (("}") reduce 487)) 375 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93) (("}") reduce 93)) 376 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92) (("}") reduce 92)) 377 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94) (("}") reduce 94)) 379 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489) (("}") reduce 489)) 386 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91) (("}") reduce 91)) 389 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486) (("}") reduce 486)) 391 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515) (("}") reduce 515)) 392 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("}") reduce 429) (("~") shift)) 399 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511) (("}") reduce 511)) 400 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481) (("}") reduce 481)) 402 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480) (("}") reduce 480)) 403 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479) (("}") reduce 479)) 404 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") shift)) 407 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510) (("}") reduce 510)) 408 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434) (("}") reduce 434)) 409 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416) (("}") reduce 416)) 410 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438) (("}") reduce 438)) 411 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482) (("}") reduce 482)) 412 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412) (("}") reduce 412)) 413 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426) (("}") reduce 426)) 414 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505) (("}") reduce 505)) 415 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512) (("}") reduce 512)) 416 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356) (("}") reduce 356)) 419 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400) (("}") reduce 400)) 450 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384) (("}") reduce 384)) 451 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387) (("}") reduce 387)) 452 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396) (("}") reduce 396)) 453 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409) (("}") reduce 409)) 454 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395) (("}") reduce 395)) 455 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392) (("}") reduce 392)) 456 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 383)) 457 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 381)) 458 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 382)) 459 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406) (("}") reduce 406)) 460 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402) (("}") reduce 402)) 461 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404) (("}") reduce 404)) 462 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401) (("}") reduce 401)) 463 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408) (("}") reduce 408)) 464 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403) (("}") reduce 403)) 467 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 468 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 469 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 470 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 471 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91) (("}") reduce 91)) 472 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515) (("}") reduce 515)) 473 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 474 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410) (("}") reduce 410)) 475 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 476 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461) (("}") reduce 461)) 477 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520) (("}") reduce 520)) 484 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522) (("}") reduce 522)) 495 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("]") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 497 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("]") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 499 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("]") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 500 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("]") reduce 475) (("^") reduce 475) (("|") reduce 475)) 504 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("]") reduce 463) (("^") reduce 463) (("|") reduce 463)) 505 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("]") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 506 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("]") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 512 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("]") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 514 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("]") reduce 474) (("^") reduce 474) (("|") reduce 474)) 516 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("]") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 517 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("]") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 518 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("]") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 519 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("]") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 521 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("]") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 528 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("]") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 531 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("]") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 533 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 534 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 429) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 541 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("]") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 542 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("]") reduce 481) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 544 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("]") reduce 480) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 545 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("]") reduce 479) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 546 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 494) (("^") shift) (("|") shift)) 549 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("]") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 550 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("]") reduce 434) (("^") reduce 434) (("|") reduce 434)) 551 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("]") reduce 416) (("^") reduce 416) (("|") reduce 416)) 552 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("]") reduce 438) (("^") reduce 438) (("|") reduce 438)) 553 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("]") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 554 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("]") reduce 412) (("^") reduce 412) (("|") reduce 412)) 555 ((("]") shift)) 556 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("]") reduce 426) (("^") reduce 426) (("|") reduce 426)) 557 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 558 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("]") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 559 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("]") reduce 356) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 562 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("]") reduce 400) (("^") reduce 400) (("|") reduce 400)) 593 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 384) (("^") shift) (("|") reduce 384)) 594 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 387) (("^") reduce 387) (("|") reduce 387)) 595 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("]") reduce 396) (("^") reduce 396) (("|") reduce 396)) 596 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 409) (("^") reduce 409) (("|") reduce 409)) 597 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("]") reduce 395) (("^") reduce 395) (("|") reduce 395)) 598 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("]") reduce 392) (("^") reduce 392) (("|") reduce 392)) 599 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 383) (("^") shift) (("|") shift)) 600 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 381) (("^") shift) (("|") shift)) 601 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 382) (("^") shift) (("|") shift)) 602 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("]") reduce 406) (("^") reduce 406) (("|") reduce 406)) 603 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 402) (("^") reduce 402) (("|") reduce 402)) 604 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 404) (("^") reduce 404) (("|") reduce 404)) 605 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 401) (("^") reduce 401) (("|") reduce 401)) 606 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("]") reduce 408) (("^") reduce 408) (("|") reduce 408)) 607 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 403) (("^") reduce 403) (("|") reduce 403)) 610 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 611 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 612 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 613 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 614 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("]") reduce 91) (("^") reduce 91) (("|") reduce 91)) 615 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 616 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 617 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("]") reduce 410) (("^") reduce 410) (("|") reduce 410)) 618 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 619 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("]") reduce 461) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 620 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 627 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 629 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 630 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("]") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 631 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("]") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 632 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("]") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 633 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 636 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 637 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("]") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 639 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("]") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 641 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("]") reduce 462) (("^") reduce 462) (("|") reduce 462)) 642 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 645 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 646 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 415) (("^") shift) (("|") shift)) 647 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 379) (("^") shift) (("|") shift)) 648 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 380) (("^") shift) (("|") shift)) 649 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 385) (("^") reduce 385) (("|") reduce 385)) 650 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 386) (("^") reduce 386) (("|") reduce 386)) 662 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 664 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((":") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 666 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((":") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 667 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((":") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 671 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) ((":") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 672 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((":") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 673 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 679 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) ((":") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 681 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((":") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 683 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 684 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 685 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 686 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 688 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 695 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 698 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((":") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 700 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 701 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) ((":") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 708 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 709 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) ((":") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 711 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) ((":") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 712 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) ((":") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 713 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 716 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 717 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) ((":") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 718 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) ((":") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 719 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) ((":") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 720 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) ((":") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 721 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) ((":") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 722 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((":") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 723 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 724 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 725 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) ((":") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 728 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((":") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 759 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 384) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 760 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 387) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 761 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) ((":") reduce 396) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 762 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 409) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 763 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) ((":") reduce 395) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 764 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((":") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 765 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 383) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 766 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 381) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 767 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 382) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 768 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 406) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 769 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 402) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 770 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 404) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 771 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 401) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 772 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 408) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 773 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 403) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 776 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 777 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 778 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 779 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((":") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 780 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 781 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 782 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 783 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((":") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 784 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 785 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((":") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 786 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 793 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 795 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 796 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 797 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 798 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 799 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 802 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 803 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 805 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 807 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 808 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 811 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 812 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 415) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 813 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 379) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 814 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 380) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 815 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 385) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 816 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 386) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 820 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 413) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 821 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 414) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 822 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 407) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 823 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 405) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 824 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((":") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 825 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) ((":") reduce 388) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 826 ((("%") shift) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) ((":") reduce 390) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 827 ((("%") shift) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) ((":") reduce 389) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 828 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) ((":") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 829 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) ((":") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 830 ((("$") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) (("{") shift)) 831 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((":") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 832 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((":") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 833 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((":") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 834 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((":") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 835 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((":") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 836 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((":") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 837 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((":") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 838 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((":") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 839 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((":") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 840 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((":") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 841 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((":") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 842 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((":") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 843 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((":") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 844 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((":") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 845 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((":") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 846 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((":") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 847 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((":") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 848 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((":") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 849 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((":") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 850 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((":") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 851 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((":") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 852 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((":") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 853 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((":") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 854 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((":") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 855 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((":") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 856 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((":") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 857 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((":") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 858 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((":") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 859 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((":") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 860 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((":") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 861 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((":") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 862 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((":") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 863 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((":") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 864 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((":") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 865 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((":") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 866 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((":") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 867 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((":") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 868 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((":") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 869 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((":") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 870 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((":") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 871 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((":") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 872 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((":") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 873 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((":") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 874 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((":") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 875 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((":") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 876 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((":") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 877 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((":") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 878 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((":") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 879 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((":") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 880 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((":") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 881 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((":") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 882 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((":") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 883 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((":") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 884 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((":") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 885 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((":") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 886 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((":") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 887 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((":") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 888 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((":") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 889 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((":") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 890 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((":") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 891 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((":") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 892 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((":") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 893 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((":") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 894 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((":") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 895 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((":") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 896 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((":") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 897 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((":") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 898 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) ((":") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 899 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((":") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 900 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((":") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 901 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((":") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 902 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((":") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 903 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((":") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 904 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((":") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 905 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((":") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 906 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((":") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 907 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((":") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 908 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((":") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 909 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((":") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 911 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((":") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 912 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((":") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 913 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 916 ((("(") reduce 528)) 917 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_AS) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_ELLIPSIS) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_USE) shift) ((T_VAR) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 918 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((":") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 923 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((":") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 927 (((":") reduce 73)) 928 ((("(") shift) ((":") reduce 52)) 930 (((":") reduce 27)) 932 (((":") reduce 51)) 933 (((":") reduce 53)) 934 (((":") reduce 49)) 935 (((":") reduce 29)) 936 (((":") reduce 59)) 937 ((("%") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 60) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 938 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 11) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 939 (((":") reduce 42)) 940 ((("%") reduce 472) (("(") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 941 (((":") reduce 39)) 943 (((":") reduce 25)) 944 (((":") reduce 50)) 945 ((("%") reduce 485) ((")") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((":") reduce 66) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 946 ((("%") reduce 475) ((")") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 947 (((":") reduce 18)) 949 (((":") reduce 17)) 950 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 951 (((":") reduce 15)) 952 (((":") reduce 14)) 953 ((("(") shift) ((":") reduce 38)) 954 (((":") reduce 26)) 955 (((":") reduce 22)) 956 (((":") reduce 24)) 957 (((":") reduce 16)) 958 (((":") reduce 48)) 959 (((":") reduce 20)) 960 (((":") reduce 70)) 961 ((("(") shift) ((":") reduce 3)) 962 ((("%") reduce 463) (("(") shift) ((")") reduce 463) (("*") reduce 463) (("+") reduce 463) ((",") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) ((":") reduce 12) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 963 (((":") reduce 54)) 964 ((("%") reduce 484) ((")") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((":") reduce 65) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 965 (((":") reduce 74)) 966 (((":") reduce 30)) 967 ((("(") reduce 441) ((":") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441)) 968 (((":") reduce 21)) 969 (((":") reduce 23)) 970 ((("(") reduce 442) ((":") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442)) 971 ((("%") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 62) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 972 (((":") reduce 34)) 973 (((":") reduce 40)) 974 (((":") reduce 13)) 975 (((":") reduce 55)) 977 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 1) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 978 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 2) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 979 (((":") reduce 9)) 980 (((":") reduce 33)) 981 (((":") reduce 58)) 983 ((("(") shift) ((":") reduce 37)) 984 ((("%") reduce 483) ((")") reduce 483) (("*") reduce 483) (("+") reduce 483) ((",") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) ((":") reduce 64) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 985 ((("(") shift) ((":") reduce 46)) 986 ((("%") reduce 474) ((")") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 987 (((":") reduce 8)) 988 (((":") reduce 6)) 989 (((":") reduce 7)) 990 ((("(") shift) ((":") reduce 69)) 991 ((("%") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 63) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 992 (((":") reduce 56)) 993 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 994 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 995 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 996 ((("$") shift) (("(") shift) ((":") reduce 10) ((T_ATTRIBUTE) shift) ((T_CLASS) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_VARIABLE) shift)) 997 ((("%") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 67) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 999 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 44) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1000 (((":") reduce 75)) 1001 (((":") reduce 76)) 1002 (((":") reduce 77)) 1003 (((":") reduce 78)) 1004 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 4) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1005 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 5) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1006 (((":") reduce 43)) 1008 (((":") reduce 72) ((T_FN) shift) ((T_FUNCTION) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458)) 1009 ((("%") reduce 91) (("(") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 81) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 1011 (((":") reduce 47)) 1012 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 31) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1013 (((":") reduce 57)) 1014 ((("%") reduce 486) ((")") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((":") reduce 61) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 1015 (((":") reduce 28)) 1016 (((":") reduce 36)) 1018 (((":") reduce 32)) 1019 (((":") reduce 35)) 1020 ((("%") reduce 515) (("(") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 1021 (((":") reduce 19)) 1022 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) ((")") reduce 429) (("*") reduce 429) (("+") shift) ((",") reduce 429) (("-") shift) ((".") reduce 429) (("/") reduce 429) ((":") reduce 45) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 1026 (((")") reduce 284) ((",") reduce 284)) 1030 ((("%") reduce 511) (("(") reduce 502) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 1031 ((("%") reduce 481) ((")") reduce 481) (("*") reduce 481) (("+") reduce 481) ((",") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 1033 ((("%") reduce 480) ((")") reduce 480) (("*") reduce 480) (("+") reduce 480) ((",") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 1034 ((("%") reduce 479) (("(") reduce 504) ((")") reduce 479) (("*") reduce 479) (("+") reduce 479) ((",") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 1035 ((("%") shift) ((")") reduce 286) (("*") shift) (("+") shift) ((",") reduce 286) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1038 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 1039 (((":") shift)) 1040 ((("%") reduce 434) ((")") reduce 434) (("*") reduce 434) (("+") reduce 434) ((",") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 1041 ((("%") reduce 416) ((")") reduce 416) (("*") reduce 416) (("+") reduce 416) ((",") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 1042 ((("%") reduce 438) ((")") reduce 438) (("*") reduce 438) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 1043 ((("%") reduce 482) (("(") shift) ((")") reduce 482) (("*") reduce 482) (("+") reduce 482) ((",") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 1044 ((("%") reduce 412) ((")") reduce 412) (("*") reduce 412) (("+") reduce 412) ((",") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 1046 (((":") reduce 71)) 1047 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 1048 (((":") reduce 82)) 1049 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 1050 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 1051 ((("%") reduce 356) ((")") reduce 356) (("*") reduce 356) (("+") reduce 356) ((",") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 1055 ((("%") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 1057 ((("%") reduce 485) ((")") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 1060 ((("%") reduce 463) (("(") shift) ((")") reduce 463) (("*") reduce 463) (("+") reduce 463) ((",") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 1061 ((("%") reduce 484) ((")") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 1062 ((("%") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 1066 ((("%") reduce 483) ((")") reduce 483) (("*") reduce 483) (("+") reduce 483) ((",") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 1069 ((("%") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 1071 ((("%") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 1076 ((("%") reduce 91) (("(") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 1078 ((("%") reduce 486) ((")") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 1079 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) ((")") reduce 429) (("*") reduce 429) (("+") shift) ((",") reduce 429) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 1080 ((("%") reduce 400) ((")") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 1111 ((("%") shift) ((")") reduce 384) (("*") shift) (("+") shift) ((",") reduce 384) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 1112 ((("%") shift) ((")") reduce 387) (("*") shift) (("+") shift) ((",") reduce 387) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 1113 ((("%") shift) ((")") reduce 396) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 1114 ((("%") shift) ((")") reduce 409) (("*") shift) (("+") shift) ((",") reduce 409) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 1115 ((("%") shift) ((")") reduce 395) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 1116 ((("%") reduce 392) ((")") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 1117 ((("%") shift) ((")") reduce 383) (("*") shift) (("+") shift) ((",") reduce 383) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1118 ((("%") shift) ((")") reduce 381) (("*") shift) (("+") shift) ((",") reduce 381) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1119 ((("%") shift) ((")") reduce 382) (("*") shift) (("+") shift) ((",") reduce 382) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1120 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 1121 ((("%") shift) ((")") reduce 402) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 1122 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 1123 ((("%") shift) ((")") reduce 401) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 1124 ((("%") shift) ((")") reduce 408) (("*") shift) (("+") shift) ((",") reduce 408) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 1125 ((("%") shift) ((")") reduce 403) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 1128 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 1129 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 1130 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 1131 ((("%") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 1132 ((("%") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 1133 ((("%") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 1134 ((("%") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 1135 ((("%") reduce 410) ((")") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 1136 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 1137 ((("%") reduce 461) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 1138 ((("%") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 1145 ((("%") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 1147 ((("%") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 1148 ((("%") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 1149 ((("%") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 1150 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 1151 ((("%") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 1154 ((("%") reduce 531) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 1155 ((("%") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 1157 ((("%") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 1159 ((("%") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 1160 ((("%") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 1163 ((("%") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 1164 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1165 ((("%") shift) ((")") reduce 379) (("*") shift) (("+") shift) ((",") reduce 379) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1166 ((("%") shift) ((")") reduce 380) (("*") shift) (("+") shift) ((",") reduce 380) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1167 ((("%") shift) ((")") reduce 385) (("*") shift) (("+") shift) ((",") reduce 385) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 1168 ((("%") shift) ((")") reduce 386) (("*") shift) (("+") shift) ((",") reduce 386) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 1172 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1173 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1174 ((("%") shift) ((")") reduce 407) (("*") shift) (("+") shift) ((",") reduce 407) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 1175 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 1176 ((("%") reduce 393) ((")") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 1177 ((("%") shift) ((")") reduce 388) (("*") shift) (("+") shift) ((",") reduce 388) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 1178 ((("%") shift) ((")") reduce 390) (("*") shift) (("+") reduce 390) ((",") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 1179 ((("%") shift) ((")") reduce 389) (("*") shift) (("+") reduce 389) ((",") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 1180 ((("%") reduce 391) ((")") reduce 391) (("*") reduce 391) (("+") reduce 391) ((",") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 1181 ((("%") reduce 394) ((")") reduce 394) (("*") reduce 394) (("+") reduce 394) ((",") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 1192 ((("%") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 1194 ((("%") reduce 472) (("(") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 1196 ((("%") reduce 485) ((")") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 1197 ((("%") reduce 475) ((")") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 1201 ((("%") reduce 463) (("(") shift) ((")") reduce 463) (("*") reduce 463) (("+") reduce 463) ((",") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 1202 ((("%") reduce 484) ((")") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 1203 ((("%") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 1209 ((("%") reduce 483) ((")") reduce 483) (("*") reduce 483) (("+") reduce 483) ((",") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 1211 ((("%") reduce 474) ((")") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 1213 ((("%") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 1214 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 1215 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 1216 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 1218 ((("%") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 1225 ((("%") reduce 91) (("(") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 1228 ((("%") reduce 486) ((")") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 1230 ((("%") reduce 515) (("(") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 1231 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) ((")") reduce 429) (("*") reduce 429) (("+") shift) ((",") reduce 429) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 1238 ((("%") reduce 511) (("(") reduce 502) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 1239 ((("%") reduce 481) ((")") reduce 481) (("*") reduce 481) (("+") reduce 481) ((",") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 1241 ((("%") reduce 480) ((")") reduce 480) (("*") reduce 480) (("+") reduce 480) ((",") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_DOUBLE_ARROW) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 1242 ((("%") reduce 479) (("(") reduce 504) ((")") reduce 479) (("*") reduce 479) (("+") reduce 479) ((",") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_DOUBLE_ARROW) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 1243 ((("%") shift) ((")") reduce 430) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1246 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 1247 ((("%") reduce 434) ((")") reduce 434) (("*") reduce 434) (("+") reduce 434) ((",") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 1248 ((("%") reduce 416) ((")") reduce 416) (("*") reduce 416) (("+") reduce 416) ((",") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 1249 ((("%") reduce 438) ((")") reduce 438) (("*") reduce 438) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 1250 ((("%") reduce 482) (("(") shift) ((")") reduce 482) (("*") reduce 482) (("+") reduce 482) ((",") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 1251 ((("%") reduce 412) ((")") reduce 412) (("*") reduce 412) (("+") reduce 412) ((",") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 1252 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 1253 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 1254 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 1255 ((("%") reduce 356) ((")") reduce 356) (("*") reduce 356) (("+") reduce 356) ((",") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 356) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 1258 ((("%") reduce 400) ((")") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 1289 ((("%") shift) ((")") reduce 384) (("*") shift) (("+") shift) ((",") reduce 384) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 1290 ((("%") shift) ((")") reduce 387) (("*") shift) (("+") shift) ((",") reduce 387) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 1291 ((("%") shift) ((")") reduce 396) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 1292 ((("%") shift) ((")") reduce 409) (("*") shift) (("+") shift) ((",") reduce 409) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 1293 ((("%") shift) ((")") reduce 395) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 1294 ((("%") reduce 392) ((")") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 1295 ((("%") shift) ((")") reduce 383) (("*") shift) (("+") shift) ((",") reduce 383) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1296 ((("%") shift) ((")") reduce 381) (("*") shift) (("+") shift) ((",") reduce 381) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 381) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1297 ((("%") shift) ((")") reduce 382) (("*") shift) (("+") shift) ((",") reduce 382) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 382) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1298 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 1299 ((("%") shift) ((")") reduce 402) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 1300 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 1301 ((("%") shift) ((")") reduce 401) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 1302 ((("%") shift) ((")") reduce 408) (("*") shift) (("+") shift) ((",") reduce 408) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 1303 ((("%") shift) ((")") reduce 403) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 1306 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 1307 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 1308 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 1309 ((("%") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 1310 ((("%") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 1311 ((("%") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 1312 ((("%") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 1313 ((("%") reduce 410) ((")") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 1314 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 1315 ((("%") reduce 461) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 1316 ((("%") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 1323 ((("%") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 1325 ((("%") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 1326 ((("%") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 1327 ((("%") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 1328 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 1329 ((("%") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 1332 ((("%") reduce 531) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 1333 ((("%") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 1335 ((("%") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 1337 ((("%") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 1338 ((("%") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 1341 ((("%") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 1342 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1343 ((("%") shift) ((")") reduce 379) (("*") shift) (("+") shift) ((",") reduce 379) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_DOUBLE_ARROW) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1344 ((("%") shift) ((")") reduce 380) (("*") shift) (("+") shift) ((",") reduce 380) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_DOUBLE_ARROW) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1345 ((("%") shift) ((")") reduce 385) (("*") shift) (("+") shift) ((",") reduce 385) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 1346 ((("%") shift) ((")") reduce 386) (("*") shift) (("+") shift) ((",") reduce 386) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 1350 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1351 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1352 ((("%") shift) ((")") reduce 407) (("*") shift) (("+") shift) ((",") reduce 407) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 1353 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 1354 ((("%") reduce 393) ((")") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 1355 ((("%") shift) ((")") reduce 388) (("*") shift) (("+") shift) ((",") reduce 388) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_DOUBLE_ARROW) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 1356 ((("%") shift) ((")") reduce 390) (("*") shift) (("+") reduce 390) ((",") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 1357 ((("%") shift) ((")") reduce 389) (("*") shift) (("+") reduce 389) ((",") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 1358 ((("%") reduce 391) ((")") reduce 391) (("*") reduce 391) (("+") reduce 391) ((",") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 1359 ((("%") reduce 394) ((")") reduce 394) (("*") reduce 394) (("+") reduce 394) ((",") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 1361 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 1362 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 1363 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 1364 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 1365 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 1366 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 1367 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 1368 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 1369 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 1370 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 1371 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 1372 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 1373 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 1374 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 1375 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 1376 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 1377 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 1378 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 1379 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 1380 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 1381 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 1382 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 1383 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 1384 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 1385 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 1386 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 1387 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 1388 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 1389 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 1390 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 1391 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 1392 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 1393 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 1394 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 1395 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 1396 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 1397 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 1398 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 1399 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 1400 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 1401 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 1402 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 1403 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 1404 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 1405 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 1406 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 1407 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 1408 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 1409 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 1410 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 1411 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 1412 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 1413 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 1414 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 1415 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 1416 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 1417 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 1418 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 1419 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 1420 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 1421 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 1422 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 1423 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 1424 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 1425 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 1426 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 1427 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 1428 ((("%") reduce 81) (("(") reduce 81) ((")") reduce 81) (("*") reduce 81) (("+") reduce 81) ((",") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_DOUBLE_ARROW) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 1429 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 1430 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 1431 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 1432 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 1433 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 1434 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 1435 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 1436 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 1437 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 1438 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 1439 ((("%") reduce 492) (("(") reduce 527) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 1441 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 1442 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 1443 ((("%") reduce 519) (("(") reduce 529) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 1445 ((("%") reduce 456) (("(") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 1446 ((("%") reduce 281) (("(") reduce 281) ((")") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 1449 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 124) (("+") shift) (("-") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_AS) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_ELLIPSIS) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_USE) shift) ((T_VAR) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1451 ((("%") reduce 282) (("(") reduce 282) ((")") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 1453 (((")") reduce 285) ((",") reduce 285)) 1454 ((("%") shift) ((")") reduce 288) (("*") shift) (("+") shift) ((",") reduce 288) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1455 ((("%") reduce 283) (("(") reduce 283) ((")") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 1456 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1460 ((("%") reduce 377) ((")") reduce 377) (("*") reduce 377) (("+") reduce 377) ((",") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_DOUBLE_ARROW) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 1462 ((("%") reduce 375) ((")") reduce 375) (("*") reduce 375) (("+") reduce 375) ((",") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_DOUBLE_ARROW) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 1472 ((("%") shift) ((")") reduce 371) (("*") shift) (("+") shift) ((",") reduce 371) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 371) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1473 ((("%") shift) ((")") reduce 373) (("*") shift) (("+") shift) ((",") reduce 373) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 373) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1474 ((("%") shift) ((")") reduce 372) (("*") shift) (("+") shift) ((",") reduce 372) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 372) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1475 ((("%") shift) ((")") reduce 365) (("*") shift) (("+") shift) ((",") reduce 365) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 365) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1476 ((("%") shift) ((")") reduce 362) (("*") shift) (("+") shift) ((",") reduce 362) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 362) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1477 ((("%") shift) ((")") reduce 370) (("*") shift) (("+") shift) ((",") reduce 370) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 370) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1478 ((("%") shift) ((")") reduce 364) (("*") shift) (("+") shift) ((",") reduce 364) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 364) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1479 ((("%") shift) ((")") reduce 368) (("*") shift) (("+") shift) ((",") reduce 368) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 368) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1480 ((("%") shift) ((")") reduce 363) (("*") shift) (("+") shift) ((",") reduce 363) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 363) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1481 ((("%") shift) ((")") reduce 366) (("*") shift) (("+") shift) ((",") reduce 366) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 366) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1482 ((("%") shift) ((")") reduce 367) (("*") shift) (("+") shift) ((",") reduce 367) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 367) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1483 ((("%") shift) ((")") reduce 374) (("*") shift) (("+") shift) ((",") reduce 374) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 374) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1484 ((("%") shift) ((")") reduce 369) (("*") shift) (("+") shift) ((",") reduce 369) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 369) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1485 ((("\"") reduce 79) (("$") reduce 79) (("(") reduce 79) ((T_ARRAY) reduce 79) ((T_CLASS_C) reduce 79) ((T_CONSTANT_ENCAPSED_STRING) reduce 79) ((T_DIR) reduce 79) ((T_FILE) reduce 79) ((T_FUNC_C) reduce 79) ((T_LINE) reduce 79) ((T_METHOD_C) reduce 79) ((T_NAME_FULLY_QUALIFIED) reduce 79) ((T_NAME_QUALIFIED) reduce 79) ((T_NAME_RELATIVE) reduce 79) ((T_NS_C) reduce 79) ((T_STATIC) reduce 79) ((T_STRING) reduce 79) ((T_TRAIT_C) reduce 79) ((T_VARIABLE) reduce 79) (("[") reduce 79)) 1486 ((("\"") reduce 80) (("$") reduce 80) (("(") reduce 80) ((T_ARRAY) reduce 80) ((T_CLASS_C) reduce 80) ((T_CONSTANT_ENCAPSED_STRING) reduce 80) ((T_DIR) reduce 80) ((T_FILE) reduce 80) ((T_FUNC_C) reduce 80) ((T_LINE) reduce 80) ((T_METHOD_C) reduce 80) ((T_NAME_FULLY_QUALIFIED) reduce 80) ((T_NAME_QUALIFIED) reduce 80) ((T_NAME_RELATIVE) reduce 80) ((T_NS_C) reduce 80) ((T_STATIC) reduce 80) ((T_STRING) reduce 80) ((T_TRAIT_C) reduce 80) ((T_VARIABLE) reduce 80) (("[") reduce 80)) 1488 ((("%") shift) ((")") reduce 359) (("*") shift) (("+") shift) ((",") reduce 359) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 359) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1493 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) (("[") reduce 490) (("{") reduce 490)) 1494 ((("(") reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) (("[") reduce 472) (("{") reduce 472)) 1495 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) (("[") reduce 485) (("{") reduce 485)) 1496 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) (("[") reduce 484) (("{") reduce 484)) 1497 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) (("[") reduce 488) (("{") reduce 488)) 1498 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) (("[") reduce 483) (("{") reduce 483)) 1499 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) (("[") reduce 487) (("{") reduce 487)) 1500 ((("(") reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) (("[") reduce 93) (("{") reduce 93)) 1501 ((("(") reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) (("[") reduce 92) (("{") reduce 92)) 1502 ((("(") reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) (("[") reduce 94) (("{") reduce 94)) 1503 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) (("[") reduce 489) (("{") reduce 489)) 1504 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 458)) 1505 ((("(") reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) (("[") reduce 91) (("{") reduce 91)) 1506 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) (("[") reduce 486) (("{") reduce 486)) 1507 ((("%") reduce 515) (("(") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 1511 ((("%") reduce 511) (("(") reduce 502) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 1512 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) (("[") reduce 499) (("{") reduce 499)) 1514 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) (("[") reduce 501) (("{") reduce 501)) 1515 ((("(") reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) (("[") reduce 498) (("{") reduce 498)) 1516 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 1517 ((("(") shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) (("[") reduce 482) (("{") reduce 482)) 1518 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 1519 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 1520 ((("%") reduce 360) ((")") reduce 360) (("*") reduce 360) (("+") reduce 360) ((",") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_DOUBLE_ARROW) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 1523 ((("(") reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) (("[") reduce 73) (("{") reduce 73)) 1524 ((("(") reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) (("[") reduce 52) (("{") reduce 52)) 1525 ((("(") reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) (("[") reduce 27) (("{") reduce 27)) 1526 ((("(") reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) (("[") reduce 51) (("{") reduce 51)) 1527 ((("(") reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) (("[") reduce 53) (("{") reduce 53)) 1528 ((("(") reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) (("[") reduce 49) (("{") reduce 49)) 1529 ((("(") reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) (("[") reduce 29) (("{") reduce 29)) 1530 ((("(") reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) (("[") reduce 59) (("{") reduce 59)) 1531 ((("(") reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) (("[") reduce 60) (("{") reduce 60)) 1532 ((("(") reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) (("[") reduce 11) (("{") reduce 11)) 1533 ((("(") reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) (("[") reduce 42) (("{") reduce 42)) 1534 ((("(") reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) (("[") reduce 39) (("{") reduce 39)) 1535 ((("(") reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) (("[") reduce 25) (("{") reduce 25)) 1536 ((("(") reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) (("[") reduce 50) (("{") reduce 50)) 1537 ((("(") reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) (("[") reduce 66) (("{") reduce 66)) 1538 ((("(") reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) (("[") reduce 18) (("{") reduce 18)) 1539 ((("(") reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) (("[") reduce 17) (("{") reduce 17)) 1540 ((("(") reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) (("[") reduce 15) (("{") reduce 15)) 1541 ((("(") reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) (("[") reduce 14) (("{") reduce 14)) 1542 ((("(") reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) (("[") reduce 38) (("{") reduce 38)) 1543 ((("(") reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) (("[") reduce 26) (("{") reduce 26)) 1544 ((("(") reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) (("[") reduce 22) (("{") reduce 22)) 1545 ((("(") reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) (("[") reduce 24) (("{") reduce 24)) 1546 ((("(") reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) (("[") reduce 16) (("{") reduce 16)) 1547 ((("(") reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) (("[") reduce 48) (("{") reduce 48)) 1548 ((("(") reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) (("[") reduce 20) (("{") reduce 20)) 1549 ((("(") reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) (("[") reduce 70) (("{") reduce 70)) 1550 ((("(") reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) (("[") reduce 3) (("{") reduce 3)) 1551 ((("(") reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) (("[") reduce 12) (("{") reduce 12)) 1552 ((("(") reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) (("[") reduce 54) (("{") reduce 54)) 1553 ((("(") reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) (("[") reduce 65) (("{") reduce 65)) 1554 ((("(") reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) (("[") reduce 74) (("{") reduce 74)) 1555 ((("(") reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) (("[") reduce 30) (("{") reduce 30)) 1556 ((("(") reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) (("[") reduce 68) (("{") reduce 68)) 1557 ((("(") reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) (("[") reduce 21) (("{") reduce 21)) 1558 ((("(") reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) (("[") reduce 23) (("{") reduce 23)) 1559 ((("(") reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) (("[") reduce 41) (("{") reduce 41)) 1560 ((("(") reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) (("[") reduce 62) (("{") reduce 62)) 1561 ((("(") reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) (("[") reduce 34) (("{") reduce 34)) 1562 ((("(") reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) (("[") reduce 40) (("{") reduce 40)) 1563 ((("(") reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) (("[") reduce 13) (("{") reduce 13)) 1564 ((("(") reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) (("[") reduce 55) (("{") reduce 55)) 1565 ((("(") reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) (("[") reduce 1) (("{") reduce 1)) 1566 ((("(") reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) (("[") reduce 2) (("{") reduce 2)) 1567 ((("(") reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) (("[") reduce 9) (("{") reduce 9)) 1568 ((("(") reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) (("[") reduce 33) (("{") reduce 33)) 1569 ((("(") reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) (("[") reduce 58) (("{") reduce 58)) 1570 ((("(") reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) (("[") reduce 37) (("{") reduce 37)) 1571 ((("(") reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) (("[") reduce 64) (("{") reduce 64)) 1572 ((("(") reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) (("[") reduce 46) (("{") reduce 46)) 1573 ((("(") reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) (("[") reduce 8) (("{") reduce 8)) 1574 ((("(") reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) (("[") reduce 6) (("{") reduce 6)) 1575 ((("(") reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) (("[") reduce 7) (("{") reduce 7)) 1576 ((("(") reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) (("[") reduce 69) (("{") reduce 69)) 1577 ((("(") reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) (("[") reduce 63) (("{") reduce 63)) 1578 ((("(") reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) (("[") reduce 56) (("{") reduce 56)) 1579 ((("(") reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) (("[") reduce 10) (("{") reduce 10)) 1580 ((("(") reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) (("[") reduce 67) (("{") reduce 67)) 1581 ((("(") reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) (("[") reduce 44) (("{") reduce 44)) 1582 ((("(") reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) (("[") reduce 75) (("{") reduce 75)) 1583 ((("(") reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) (("[") reduce 76) (("{") reduce 76)) 1584 ((("(") reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) (("[") reduce 77) (("{") reduce 77)) 1585 ((("(") reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) (("[") reduce 78) (("{") reduce 78)) 1586 ((("(") reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) (("[") reduce 4) (("{") reduce 4)) 1587 ((("(") reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) (("[") reduce 5) (("{") reduce 5)) 1588 ((("(") reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) (("[") reduce 43) (("{") reduce 43)) 1589 ((("(") reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) (("[") reduce 72) (("{") reduce 72)) 1590 ((("(") reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) (("[") reduce 81) (("{") reduce 81)) 1591 ((("(") reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) (("[") reduce 47) (("{") reduce 47)) 1592 ((("(") reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) (("[") reduce 31) (("{") reduce 31)) 1593 ((("(") reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) (("[") reduce 57) (("{") reduce 57)) 1594 ((("(") reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) (("[") reduce 61) (("{") reduce 61)) 1595 ((("(") reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) (("[") reduce 28) (("{") reduce 28)) 1596 ((("(") reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) (("[") reduce 36) (("{") reduce 36)) 1597 ((("(") reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) (("[") reduce 32) (("{") reduce 32)) 1598 ((("(") reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) (("[") reduce 35) (("{") reduce 35)) 1599 ((("(") reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) (("[") reduce 19) (("{") reduce 19)) 1600 ((("(") reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) (("[") reduce 45) (("{") reduce 45)) 1601 ((("(") reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) (("[") reduce 492) (("{") reduce 492)) 1603 ((("(") reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) (("[") reduce 71) (("{") reduce 71)) 1604 ((("(") reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) (("[") reduce 82) (("{") reduce 82)) 1605 ((("%") reduce 519) (("(") reduce 529) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 1607 ((("%") reduce 456) (("(") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 1608 ((("%") reduce 281) (("(") reduce 281) ((")") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 1612 ((("%") reduce 282) (("(") reduce 282) ((")") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 1613 ((("%") reduce 283) (("(") reduce 283) ((")") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 1614 ((("%") reduce 454) (("(") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 1616 ((("(") reduce 527) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) (("[") reduce 491) (("{") reduce 491)) 1618 ((("%") reduce 518) (("(") reduce 529) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 1619 ((("%") reduce 455) (("(") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 1620 ((("%") reduce 457) (("(") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 1626 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 1628 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 1629 ((("%") reduce 530) (("(") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 1630 ((("%") reduce 513) (("(") shift) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 1631 ((("%") reduce 532) (("(") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 1634 ((("%") reduce 531) (("(") reduce 531) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 1635 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 1636 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 1637 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 1648 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("]") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 1650 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("]") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 1652 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("]") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 1653 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("]") reduce 475) (("^") reduce 475) (("|") reduce 475)) 1658 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) ((",") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("]") reduce 463) (("^") reduce 463) (("|") reduce 463)) 1659 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("]") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 1660 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("]") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 1666 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) ((",") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("]") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 1668 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("]") reduce 474) (("^") reduce 474) (("|") reduce 474)) 1670 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("]") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 1671 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("]") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 1672 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("]") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 1673 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("]") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 1675 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("]") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 1682 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("]") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 1685 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("]") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 1687 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 1688 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) ((",") reduce 429) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 429) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 1694 (((",") reduce 535) (("]") reduce 535)) 1698 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("]") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 1699 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) ((",") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("]") reduce 481) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 1701 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) ((",") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_DOUBLE_ARROW) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("]") reduce 480) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 1702 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) ((",") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_DOUBLE_ARROW) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("]") reduce 479) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 1703 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 539) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 539) (("^") shift) (("|") shift)) 1706 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("]") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 1707 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) ((",") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("]") reduce 434) (("^") reduce 434) (("|") reduce 434)) 1708 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) ((",") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("]") reduce 416) (("^") reduce 416) (("|") reduce 416)) 1709 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("]") reduce 438) (("^") reduce 438) (("|") reduce 438)) 1710 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) ((",") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("]") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 1711 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) ((",") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("]") reduce 412) (("^") reduce 412) (("|") reduce 412)) 1712 (((",") shift) (("]") reduce 533)) 1713 (((",") reduce 537) (("]") reduce 537)) 1714 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("]") reduce 426) (("^") reduce 426) (("|") reduce 426)) 1715 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 1716 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("]") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 1717 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) ((",") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 356) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("]") reduce 356) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 1721 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("]") reduce 400) (("^") reduce 400) (("|") reduce 400)) 1752 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 384) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 384) (("^") shift) (("|") reduce 384)) 1753 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 387) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 387) (("^") reduce 387) (("|") reduce 387)) 1754 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("]") reduce 396) (("^") reduce 396) (("|") reduce 396)) 1755 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 409) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 409) (("^") reduce 409) (("|") reduce 409)) 1756 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("]") reduce 395) (("^") reduce 395) (("|") reduce 395)) 1757 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("]") reduce 392) (("^") reduce 392) (("|") reduce 392)) 1758 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 383) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 383) (("^") shift) (("|") shift)) 1759 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 381) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 381) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 381) (("^") shift) (("|") shift)) 1760 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 382) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 382) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 382) (("^") shift) (("|") shift)) 1761 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("]") reduce 406) (("^") reduce 406) (("|") reduce 406)) 1762 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 402) (("^") reduce 402) (("|") reduce 402)) 1763 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 404) (("^") reduce 404) (("|") reduce 404)) 1764 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 401) (("^") reduce 401) (("|") reduce 401)) 1765 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 408) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("]") reduce 408) (("^") reduce 408) (("|") reduce 408)) 1766 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 403) (("^") reduce 403) (("|") reduce 403)) 1769 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 1770 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 1771 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 1772 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 1773 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("]") reduce 91) (("^") reduce 91) (("|") reduce 91)) 1774 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 1775 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 1776 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("]") reduce 410) (("^") reduce 410) (("|") reduce 410)) 1777 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 1778 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("]") reduce 461) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 1779 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 1786 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 1788 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 1789 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("]") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 1790 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("]") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 1791 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("]") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 1792 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 1795 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 1796 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("]") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 1798 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("]") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 1800 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("]") reduce 462) (("^") reduce 462) (("|") reduce 462)) 1801 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 1804 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 1805 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 415) (("^") shift) (("|") shift)) 1806 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 379) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_DOUBLE_ARROW) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 379) (("^") shift) (("|") shift)) 1807 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 380) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_DOUBLE_ARROW) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 380) (("^") shift) (("|") shift)) 1808 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 385) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 385) (("^") reduce 385) (("|") reduce 385)) 1809 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 386) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 386) (("^") reduce 386) (("|") reduce 386)) 1813 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 413) (("^") shift) (("|") shift)) 1814 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 414) (("^") shift) (("|") shift)) 1815 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 407) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("]") reduce 407) (("^") reduce 407) (("|") reduce 407)) 1816 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("]") reduce 405) (("^") reduce 405) (("|") reduce 405)) 1817 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("]") reduce 393) (("^") reduce 393) (("|") reduce 393)) 1818 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 388) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_DOUBLE_ARROW) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("]") reduce 388) (("^") reduce 388) (("|") reduce 388)) 1819 ((("%") shift) (("*") shift) (("+") reduce 390) ((",") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("]") reduce 390) (("^") reduce 390) (("|") reduce 390)) 1820 ((("%") shift) (("*") shift) (("+") reduce 389) ((",") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("]") reduce 389) (("^") reduce 389) (("|") reduce 389)) 1821 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) ((",") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("]") reduce 391) (("^") reduce 391) (("|") reduce 391)) 1822 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) ((",") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("]") reduce 394) (("^") reduce 394) (("|") reduce 394)) 1823 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 534) (("+") shift) ((",") reduce 534) (("-") shift) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_ELLIPSIS) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1880 (((")") reduce 535) ((",") reduce 535)) 1889 ((("%") shift) ((")") reduce 539) (("*") shift) (("+") shift) ((",") reduce 539) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1898 (((")") reduce 533) ((",") shift)) 1899 (((")") reduce 537) ((",") reduce 537)) 2011 ((("=") shift)) 2013 ((("%") shift) ((")") reduce 357) (("*") shift) (("+") shift) ((",") reduce 357) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 357) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2183 ((("%") reduce 517) (("(") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 2186 ((("%") reduce 516) (("(") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 2188 (((")") reduce 536) ((",") reduce 536)) 2189 ((("%") reduce 454) (("(") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 2200 ((("%") reduce 444) ((")") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 2201 ((("%") reduce 439) ((")") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 2206 (((":") shift) ((T_DOUBLE_ARROW) reduce 279)) 2208 (((T_DOUBLE_ARROW) shift)) 2209 ((("!") reduce 444) (("\"") reduce 444) (("$") reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) (("@") reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_DEC) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_EMPTY) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FN) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT_C) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_VARIABLE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("~") reduce 444)) 2210 ((("!") reduce 445) (("\"") reduce 445) (("$") reduce 445) (("(") reduce 445) (("+") reduce 445) (("-") reduce 445) (("@") reduce 445) ((T_ARRAY) reduce 445) ((T_ARRAY_CAST) reduce 445) ((T_ATTRIBUTE) reduce 445) ((T_BOOL_CAST) reduce 445) ((T_CLASS_C) reduce 445) ((T_CLONE) reduce 445) ((T_CONSTANT_ENCAPSED_STRING) reduce 445) ((T_DEC) reduce 445) ((T_DIR) reduce 445) ((T_DNUMBER) reduce 445) ((T_DOUBLE_CAST) reduce 445) ((T_EMPTY) reduce 445) ((T_EVAL) reduce 445) ((T_EXIT) reduce 445) ((T_FILE) reduce 445) ((T_FN) reduce 445) ((T_FUNCTION) reduce 445) ((T_FUNC_C) reduce 445) ((T_INC) reduce 445) ((T_INCLUDE) reduce 445) ((T_INCLUDE_ONCE) reduce 445) ((T_INT_CAST) reduce 445) ((T_ISSET) reduce 445) ((T_LINE) reduce 445) ((T_LIST) reduce 445) ((T_LNUMBER) reduce 445) ((T_MATCH) reduce 445) ((T_METHOD_C) reduce 445) ((T_NAME_FULLY_QUALIFIED) reduce 445) ((T_NAME_QUALIFIED) reduce 445) ((T_NAME_RELATIVE) reduce 445) ((T_NEW) reduce 445) ((T_NS_C) reduce 445) ((T_OBJECT_CAST) reduce 445) ((T_PRINT) reduce 445) ((T_REQUIRE) reduce 445) ((T_REQUIRE_ONCE) reduce 445) ((T_START_HEREDOC) reduce 445) ((T_STATIC) reduce 445) ((T_STRING) reduce 445) ((T_STRING_CAST) reduce 445) ((T_THROW) reduce 445) ((T_TRAIT_C) reduce 445) ((T_UNSET_CAST) reduce 445) ((T_VARIABLE) reduce 445) ((T_YIELD) reduce 445) ((T_YIELD_FROM) reduce 445) (("[") reduce 445) (("`") reduce 445) (("~") reduce 445)) 2212 ((("%") shift) ((")") reduce 444) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2213 ((("%") reduce 440) ((")") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 2214 (((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift)) 2215 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_DOUBLE_ARROW) reduce 272) (("|") reduce 272)) 2216 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_DOUBLE_ARROW) reduce 273) (("|") reduce 273)) 2217 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_DOUBLE_ARROW) reduce 93) (("|") reduce 93)) 2218 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_DOUBLE_ARROW) reduce 92) (("|") reduce 92)) 2219 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_DOUBLE_ARROW) reduce 94) (("|") reduce 94)) 2220 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 263) ((T_DOUBLE_ARROW) reduce 263) (("|") reduce 263)) 2221 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_DOUBLE_ARROW) reduce 91) (("|") reduce 91)) 2222 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_DOUBLE_ARROW) reduce 261)) 2223 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_DOUBLE_ARROW) reduce 274) (("|") reduce 274)) 2224 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_DOUBLE_ARROW) reduce 258) (("|") shift)) 2225 (((T_DOUBLE_ARROW) reduce 280)) 2226 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 262) ((T_DOUBLE_ARROW) reduce 262) (("|") reduce 262)) 2227 (((T_DOUBLE_ARROW) reduce 260) (("|") shift)) 2229 (((T_DOUBLE_ARROW) reduce 272) (("|") reduce 272)) 2230 (((T_DOUBLE_ARROW) reduce 273) (("|") reduce 273)) 2231 (((T_DOUBLE_ARROW) reduce 93) (("|") reduce 93)) 2232 (((T_DOUBLE_ARROW) reduce 92) (("|") reduce 92)) 2233 (((T_DOUBLE_ARROW) reduce 94) (("|") reduce 94)) 2234 (((T_DOUBLE_ARROW) reduce 263) (("|") reduce 263)) 2235 (((T_DOUBLE_ARROW) reduce 91) (("|") reduce 91)) 2236 (((T_DOUBLE_ARROW) reduce 274) (("|") reduce 274)) 2237 (((T_DOUBLE_ARROW) reduce 265) (("|") reduce 265)) 2238 (((T_DOUBLE_ARROW) reduce 262) (("|") reduce 262)) 2241 (((T_DOUBLE_ARROW) reduce 264) (("|") reduce 264)) 2242 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_DOUBLE_ARROW) reduce 272)) 2243 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_DOUBLE_ARROW) reduce 273)) 2244 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_DOUBLE_ARROW) reduce 93)) 2245 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_DOUBLE_ARROW) reduce 92)) 2246 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_DOUBLE_ARROW) reduce 94)) 2247 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 263) ((T_DOUBLE_ARROW) reduce 263)) 2248 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_DOUBLE_ARROW) reduce 91)) 2249 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_DOUBLE_ARROW) reduce 274)) 2250 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) ((T_DOUBLE_ARROW) reduce 266)) 2251 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 262) ((T_DOUBLE_ARROW) reduce 262)) 2253 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) ((T_DOUBLE_ARROW) reduce 267)) 2254 (((T_DOUBLE_ARROW) reduce 272)) 2255 (((T_DOUBLE_ARROW) reduce 273)) 2256 (((T_DOUBLE_ARROW) reduce 93)) 2257 (((T_DOUBLE_ARROW) reduce 92)) 2258 (((T_DOUBLE_ARROW) reduce 94)) 2259 (((T_DOUBLE_ARROW) reduce 263)) 2260 (((T_DOUBLE_ARROW) reduce 91)) 2261 (((T_DOUBLE_ARROW) reduce 274)) 2262 (((T_DOUBLE_ARROW) reduce 259)) 2263 (((T_DOUBLE_ARROW) reduce 262)) 2327 ((("%") shift) ((")") reduce 538) (("*") shift) (("+") shift) ((",") reduce 538) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2449 ((("%") shift) ((")") reduce 357) (("*") shift) (("+") shift) ((",") reduce 357) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2451 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 2452 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 2453 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 2454 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 2455 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 2456 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 2457 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 2458 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 2459 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 2460 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 2461 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 2462 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 2463 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 2464 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 2465 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 2466 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 2467 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 2468 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 2469 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 2470 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 2471 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 2472 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 2473 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 2474 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 2475 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 2476 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 2477 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 2478 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 2479 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 2480 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 2481 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 2482 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 2483 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 2484 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 2485 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 2486 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 2487 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 2488 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 2489 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 2490 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 2491 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 2492 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 2493 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 2494 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 2495 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 2496 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 2497 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 2498 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 2499 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 2500 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 2501 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 2502 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 2503 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 2504 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 2505 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 2506 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 2507 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 2508 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 2509 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 2510 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 2511 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 2512 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 2513 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 2514 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 2515 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 2516 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 2517 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 2518 ((("%") reduce 81) (("(") reduce 81) ((")") reduce 81) (("*") reduce 81) (("+") reduce 81) ((",") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 2519 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 2520 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 2521 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 2522 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 2523 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 2524 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 2525 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 2526 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 2527 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 2528 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 2529 ((("%") reduce 492) (("(") reduce 527) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 2531 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 2532 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 2533 ((("%") reduce 519) (("(") reduce 529) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 2535 ((("%") reduce 456) (("(") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 2536 ((("%") reduce 281) (("(") reduce 281) ((")") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 2540 ((("%") reduce 282) (("(") reduce 282) ((")") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 2541 ((("%") reduce 283) (("(") reduce 283) ((")") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 2546 ((("%") reduce 377) ((")") reduce 377) (("*") reduce 377) (("+") reduce 377) ((",") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 2548 ((("%") reduce 375) ((")") reduce 375) (("*") reduce 375) (("+") reduce 375) ((",") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 2558 ((("%") shift) ((")") reduce 371) (("*") shift) (("+") shift) ((",") reduce 371) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2559 ((("%") shift) ((")") reduce 373) (("*") shift) (("+") shift) ((",") reduce 373) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2560 ((("%") shift) ((")") reduce 372) (("*") shift) (("+") shift) ((",") reduce 372) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2561 ((("%") shift) ((")") reduce 365) (("*") shift) (("+") shift) ((",") reduce 365) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2562 ((("%") shift) ((")") reduce 362) (("*") shift) (("+") shift) ((",") reduce 362) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2563 ((("%") shift) ((")") reduce 370) (("*") shift) (("+") shift) ((",") reduce 370) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2564 ((("%") shift) ((")") reduce 364) (("*") shift) (("+") shift) ((",") reduce 364) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2565 ((("%") shift) ((")") reduce 368) (("*") shift) (("+") shift) ((",") reduce 368) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2566 ((("%") shift) ((")") reduce 363) (("*") shift) (("+") shift) ((",") reduce 363) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2567 ((("%") shift) ((")") reduce 366) (("*") shift) (("+") shift) ((",") reduce 366) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2568 ((("%") shift) ((")") reduce 367) (("*") shift) (("+") shift) ((",") reduce 367) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2569 ((("%") shift) ((")") reduce 374) (("*") shift) (("+") shift) ((",") reduce 374) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2570 ((("%") shift) ((")") reduce 369) (("*") shift) (("+") shift) ((",") reduce 369) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2572 ((("%") shift) ((")") reduce 359) (("*") shift) (("+") shift) ((",") reduce 359) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2574 ((("%") reduce 515) (("(") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 2577 ((("%") reduce 511) (("(") reduce 502) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 2579 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 2581 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 2582 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 2583 ((("%") reduce 360) ((")") reduce 360) (("*") reduce 360) (("+") reduce 360) ((",") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 2587 ((("%") reduce 519) (("(") reduce 529) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 2589 ((("%") reduce 456) (("(") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 2590 ((("%") reduce 281) (("(") reduce 281) ((")") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 2594 ((("%") reduce 282) (("(") reduce 282) ((")") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 2595 ((("%") reduce 283) (("(") reduce 283) ((")") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 2596 ((("%") reduce 454) (("(") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 2599 ((("%") reduce 518) (("(") reduce 529) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 2600 ((("%") reduce 455) (("(") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 2601 ((("%") reduce 457) (("(") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 2607 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 2609 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 2610 ((("%") reduce 530) (("(") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 2611 ((("%") reduce 513) (("(") shift) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 2612 ((("%") reduce 532) (("(") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 2615 ((("%") reduce 531) (("(") reduce 531) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 2616 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 2617 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 2618 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 2619 ((("%") reduce 517) (("(") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 2622 ((("%") reduce 516) (("(") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 2623 ((("%") reduce 454) (("(") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 2634 ((("%") reduce 444) ((")") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 2635 ((("%") reduce 439) ((")") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 2645 ((("%") shift) ((")") reduce 444) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2646 ((("%") reduce 440) ((")") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 2648 ((("%") reduce 491) (("(") reduce 527) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 2650 ((("%") reduce 518) (("(") reduce 529) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 2651 ((("%") reduce 455) (("(") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 2652 ((("%") reduce 457) (("(") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 2653 (((T_FN) shift) ((T_FUNCTION) shift)) 2654 (((T_ATTRIBUTE) reduce 101) ((T_FN) reduce 101) ((T_FUNCTION) reduce 101) ((T_STATIC) reduce 101)) 2655 ((("%") reduce 435) ((")") reduce 435) (("*") reduce 435) (("+") reduce 435) ((",") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 2656 ((("%") reduce 437) ((")") reduce 437) (("*") reduce 437) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 2662 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 2664 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 2665 ((("%") reduce 530) (("(") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 2666 ((("%") reduce 513) (("(") shift) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 2667 ((("%") reduce 532) (("(") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 2670 ((("%") reduce 531) (("(") reduce 531) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 2671 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 2672 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 2673 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 2675 ((("(") reduce 515) ((")") reduce 515) ((",") reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) (("[") reduce 515) (("{") reduce 515)) 2678 ((("(") reduce 502) ((")") reduce 511) ((",") reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) (("[") reduce 511) (("{") reduce 511)) 2680 ((("(") reduce 510) ((")") reduce 510) ((",") reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) (("[") reduce 510) (("{") reduce 510)) 2682 ((("(") reduce 505) ((")") reduce 505) ((",") reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) (("[") reduce 505) (("{") reduce 505)) 2683 (((")") reduce 512) ((",") reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) (("[") reduce 512) (("{") reduce 512)) 2684 (((")") reduce 540) ((",") reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) (("[") reduce 496) (("{") reduce 496)) 2688 ((("(") reduce 529) ((")") reduce 519) ((",") reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) (("[") reduce 519) (("{") reduce 519)) 2690 ((("(") reduce 456) ((")") reduce 456) ((",") reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) (("[") reduce 456) (("{") reduce 456)) 2691 ((("(") reduce 281) ((")") reduce 281) ((",") reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) (("[") reduce 281) (("{") reduce 281)) 2695 ((("(") reduce 282) ((")") reduce 282) ((",") reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) (("[") reduce 282) (("{") reduce 282)) 2696 ((("(") reduce 283) ((")") reduce 283) ((",") reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) (("[") reduce 283) (("{") reduce 283)) 2697 ((("(") reduce 454) ((")") reduce 454) ((",") reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) (("[") reduce 454) (("{") reduce 454)) 2700 ((("(") reduce 529) ((")") reduce 518) ((",") reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) (("[") reduce 518) (("{") reduce 518)) 2701 ((("(") reduce 455) ((")") reduce 455) ((",") reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) (("[") reduce 455) (("{") reduce 455)) 2702 ((("(") reduce 457) ((")") reduce 457) ((",") reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) (("[") reduce 457) (("{") reduce 457)) 2708 ((("(") reduce 507) ((")") reduce 507) ((",") reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) (("[") reduce 507) (("{") reduce 507)) 2710 ((("(") reduce 506) ((")") reduce 506) ((",") reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) (("[") reduce 506) (("{") reduce 506)) 2711 ((("(") reduce 530) ((")") reduce 530) ((",") reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) (("[") reduce 530) (("{") reduce 530)) 2712 ((("(") shift) ((")") reduce 513) ((",") reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) (("[") reduce 513) (("{") reduce 513)) 2713 ((("(") reduce 532) ((")") reduce 532) ((",") reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) (("[") reduce 532) (("{") reduce 532)) 2716 ((("(") reduce 531) ((")") reduce 531) ((",") reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531) (("{") reduce 531)) 2717 ((("(") reduce 508) ((")") reduce 508) ((",") reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) (("[") reduce 508) (("{") reduce 508)) 2718 ((("(") shift) ((")") reduce 514) ((",") reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) (("[") reduce 514) (("{") reduce 514)) 2719 ((("(") reduce 509) ((")") reduce 509) ((",") reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) (("[") reduce 509) (("{") reduce 509)) 2720 ((("(") reduce 517) ((")") reduce 517) ((",") reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) (("[") reduce 517) (("{") reduce 517)) 2723 ((("(") reduce 516) ((")") reduce 516) ((",") reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) (("[") reduce 516) (("{") reduce 516)) 2725 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_STRING_VARNAME) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 2726 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_VARIABLE) shift) (("`") reduce 466)) 2727 (((T_CURLY_OPEN) reduce 549) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 549) ((T_ENCAPSED_AND_WHITESPACE) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_VARIABLE) reduce 549) (("[") shift) (("`") reduce 549)) 2728 ((("`") shift)) 2729 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_VARIABLE) shift) (("`") reduce 467)) 2730 (((T_CURLY_OPEN) reduce 547) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 547) ((T_ENCAPSED_AND_WHITESPACE) reduce 547) ((T_VARIABLE) reduce 547) (("`") reduce 547)) 2731 (((T_CURLY_OPEN) reduce 546) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 546) ((T_ENCAPSED_AND_WHITESPACE) reduce 546) ((T_VARIABLE) reduce 546) (("`") reduce 546)) 2732 (((T_CURLY_OPEN) reduce 545) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 545) ((T_ENCAPSED_AND_WHITESPACE) reduce 545) ((T_VARIABLE) reduce 545) (("`") reduce 545)) 2733 ((("%") reduce 427) ((")") reduce 427) (("*") reduce 427) (("+") reduce 427) ((",") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 2736 ((("-") shift) ((T_NUM_STRING) shift) ((T_STRING) shift) ((T_VARIABLE) shift)) 2737 (((T_NUM_STRING) shift)) 2738 ((("]") reduce 558)) 2739 ((("]") reduce 557)) 2740 ((("]") reduce 560)) 2742 (((T_CURLY_OPEN) reduce 550) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 550) ((T_ENCAPSED_AND_WHITESPACE) reduce 550) ((T_VARIABLE) reduce 550) (("`") reduce 550)) 2743 ((("]") reduce 559)) 2744 (((T_CURLY_OPEN) reduce 551) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 551) ((T_ENCAPSED_AND_WHITESPACE) reduce 551) ((T_VARIABLE) reduce 551) (("`") reduce 551)) 2745 (((T_CURLY_OPEN) reduce 552) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 552) ((T_ENCAPSED_AND_WHITESPACE) reduce 552) ((T_VARIABLE) reduce 552) (("`") reduce 552)) 2746 (((T_CURLY_OPEN) reduce 548) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 548) ((T_ENCAPSED_AND_WHITESPACE) reduce 548) ((T_VARIABLE) reduce 548) (("`") reduce 548)) 2747 ((("[") shift) (("}") shift)) 2749 (((T_CURLY_OPEN) reduce 553) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 553) ((T_ENCAPSED_AND_WHITESPACE) reduce 553) ((T_VARIABLE) reduce 553) (("`") reduce 553)) 2751 (((T_CURLY_OPEN) reduce 554) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 554) ((T_ENCAPSED_AND_WHITESPACE) reduce 554) ((T_VARIABLE) reduce 554) (("`") reduce 554)) 2752 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") shift) (("^") shift) (("|") shift)) 2753 ((("}") shift)) 2754 (((T_CURLY_OPEN) reduce 555) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 555) ((T_ENCAPSED_AND_WHITESPACE) reduce 555) ((T_VARIABLE) reduce 555) (("`") reduce 555)) 2756 ((("(") reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) (("[") reduce 515) (("{") reduce 515) (("}") reduce 515)) 2759 ((("(") reduce 502) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) (("[") reduce 511) (("{") reduce 511) (("}") reduce 511)) 2761 ((("(") reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) (("[") reduce 510) (("{") reduce 510) (("}") reduce 510)) 2763 ((("(") reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) (("[") reduce 505) (("{") reduce 505) (("}") reduce 505)) 2764 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) (("[") reduce 512) (("{") reduce 512) (("}") reduce 512)) 2765 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) (("[") reduce 496) (("{") reduce 496) (("}") shift)) 2769 ((("(") reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) (("[") reduce 519) (("{") reduce 519) (("}") reduce 519)) 2771 ((("(") reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) (("[") reduce 456) (("{") reduce 456) (("}") reduce 456)) 2772 ((("(") reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) (("[") reduce 281) (("{") reduce 281) (("}") reduce 281)) 2776 ((("(") reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) (("[") reduce 282) (("{") reduce 282) (("}") reduce 282)) 2777 ((("(") reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) (("[") reduce 283) (("{") reduce 283) (("}") reduce 283)) 2778 (((T_CURLY_OPEN) reduce 556) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 556) ((T_ENCAPSED_AND_WHITESPACE) reduce 556) ((T_VARIABLE) reduce 556) (("`") reduce 556)) 2779 ((("(") reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) (("[") reduce 454) (("{") reduce 454) (("}") reduce 454)) 2782 ((("(") reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) (("[") reduce 518) (("{") reduce 518) (("}") reduce 518)) 2783 ((("(") reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) (("[") reduce 455) (("{") reduce 455) (("}") reduce 455)) 2784 ((("(") reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) (("[") reduce 457) (("{") reduce 457) (("}") reduce 457)) 2790 ((("(") reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) (("[") reduce 507) (("{") reduce 507) (("}") reduce 507)) 2792 ((("(") reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) (("[") reduce 506) (("{") reduce 506) (("}") reduce 506)) 2793 ((("(") reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) (("[") reduce 530) (("{") reduce 530) (("}") reduce 530)) 2794 ((("(") shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) (("[") reduce 513) (("{") reduce 513) (("}") reduce 513)) 2795 ((("(") reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) (("[") reduce 532) (("{") reduce 532) (("}") reduce 532)) 2798 ((("(") reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531) (("{") reduce 531) (("}") reduce 531)) 2799 ((("(") reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) (("[") reduce 508) (("{") reduce 508) (("}") reduce 508)) 2800 ((("(") shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) (("[") reduce 514) (("{") reduce 514) (("}") reduce 514)) 2801 ((("(") reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) (("[") reduce 509) (("{") reduce 509) (("}") reduce 509)) 2802 ((("(") reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) (("[") reduce 517) (("{") reduce 517) (("}") reduce 517)) 2805 ((("(") reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) (("[") reduce 516) (("{") reduce 516) (("}") reduce 516)) 2807 ((("%") reduce 471) (("(") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 2809 ((("%") shift) ((")") reduce 358) (("*") shift) (("+") shift) ((",") reduce 358) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2810 ((("%") shift) ((")") reduce 432) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2813 ((("%") shift) ((")") reduce 431) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2814 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 2815 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2816 ((("%") reduce 419) ((")") reduce 419) (("*") reduce 419) (("+") reduce 419) ((",") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 2817 ((("%") reduce 436) ((")") reduce 436) (("*") reduce 436) (("+") reduce 436) ((",") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 2820 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_END_HEREDOC) shift) ((T_VARIABLE) shift)) 2821 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 2822 (((T_CURLY_OPEN) reduce 549) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 549) ((T_ENCAPSED_AND_WHITESPACE) reduce 549) ((T_END_HEREDOC) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_VARIABLE) reduce 549) (("[") shift)) 2824 (((T_CURLY_OPEN) reduce 547) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 547) ((T_ENCAPSED_AND_WHITESPACE) reduce 547) ((T_END_HEREDOC) reduce 547) ((T_VARIABLE) reduce 547)) 2825 (((T_CURLY_OPEN) reduce 546) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 546) ((T_ENCAPSED_AND_WHITESPACE) reduce 546) ((T_END_HEREDOC) reduce 546) ((T_VARIABLE) reduce 546)) 2826 ((("%") reduce 478) ((")") reduce 478) (("*") reduce 478) (("+") reduce 478) ((",") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 2827 (((T_CURLY_OPEN) reduce 545) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 545) ((T_ENCAPSED_AND_WHITESPACE) reduce 545) ((T_END_HEREDOC) reduce 545) ((T_VARIABLE) reduce 545)) 2832 (((T_CURLY_OPEN) reduce 550) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 550) ((T_ENCAPSED_AND_WHITESPACE) reduce 550) ((T_END_HEREDOC) reduce 550) ((T_VARIABLE) reduce 550)) 2833 (((T_CURLY_OPEN) reduce 551) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 551) ((T_ENCAPSED_AND_WHITESPACE) reduce 551) ((T_END_HEREDOC) reduce 551) ((T_VARIABLE) reduce 551)) 2834 (((T_CURLY_OPEN) reduce 552) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 552) ((T_ENCAPSED_AND_WHITESPACE) reduce 552) ((T_END_HEREDOC) reduce 552) ((T_VARIABLE) reduce 552)) 2835 ((("%") reduce 476) ((")") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 2836 (((T_CURLY_OPEN) reduce 548) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 548) ((T_ENCAPSED_AND_WHITESPACE) reduce 548) ((T_END_HEREDOC) reduce 548) ((T_VARIABLE) reduce 548)) 2839 (((T_CURLY_OPEN) reduce 553) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 553) ((T_ENCAPSED_AND_WHITESPACE) reduce 553) ((T_END_HEREDOC) reduce 553) ((T_VARIABLE) reduce 553)) 2841 (((T_CURLY_OPEN) reduce 554) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 554) ((T_ENCAPSED_AND_WHITESPACE) reduce 554) ((T_END_HEREDOC) reduce 554) ((T_VARIABLE) reduce 554)) 2844 (((T_CURLY_OPEN) reduce 555) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 555) ((T_ENCAPSED_AND_WHITESPACE) reduce 555) ((T_END_HEREDOC) reduce 555) ((T_VARIABLE) reduce 555)) 2846 (((T_CURLY_OPEN) reduce 556) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 556) ((T_ENCAPSED_AND_WHITESPACE) reduce 556) ((T_END_HEREDOC) reduce 556) ((T_VARIABLE) reduce 556)) 2847 ((("%") shift) ((")") reduce 567) (("*") shift) (("+") shift) ((",") reduce 567) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2848 ((("%") shift) ((")") reduce 566) (("*") shift) (("+") shift) ((",") reduce 566) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2849 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2850 ((("%") reduce 421) ((")") reduce 421) (("*") reduce 421) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 2854 ((("(") shift) ((T_EXTENDS) reduce 468) ((T_IMPLEMENTS) reduce 468) (("{") reduce 468)) 2855 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 2856 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 2857 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 2858 ((("%") reduce 458) (("(") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 2859 ((("%") reduce 91) (("(") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 2861 ((("%") reduce 354) ((")") reduce 354) (("*") reduce 354) (("+") reduce 354) ((",") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 2862 (((T_ATTRIBUTE) reduce 100) ((T_CLASS) reduce 100)) 2863 (((T_ATTRIBUTE) shift) ((T_CLASS) shift)) 2864 ((("%") reduce 460) (("(") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 2865 ((("%") reduce 468) (("(") shift) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 2866 ((("%") reduce 459) (("(") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 2867 ((("%") reduce 461) (("(") reduce 461) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 2868 ((("%") reduce 520) (("(") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 2875 ((("%") reduce 522) (("(") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 2877 ((("%") reduce 521) (("(") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 2878 ((("%") reduce 526) (("(") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 2880 ((("%") reduce 523) (("(") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 2885 ((("%") reduce 524) (("(") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 2887 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 2888 ((("%") reduce 353) ((")") reduce 353) (("*") reduce 353) (("+") reduce 353) ((",") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 2889 ((("%") reduce 281) ((")") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 2893 ((("%") reduce 282) ((")") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 2894 ((("%") reduce 283) ((")") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 2896 ((("%") reduce 525) (("(") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 2897 ((("%") reduce 355) ((")") reduce 355) (("*") reduce 355) (("+") reduce 355) ((",") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 2898 (((T_ATTRIBUTE) reduce 101) ((T_CLASS) reduce 101)) 2900 (((T_EXTENDS) reduce 469) ((T_IMPLEMENTS) reduce 469) (("{") reduce 469)) 2901 (((T_EXTENDS) shift) ((T_IMPLEMENTS) reduce 198) (("{") reduce 198)) 2903 (((T_IMPLEMENTS) shift) (("{") reduce 202)) 2905 ((("{") reduce 443)) 2907 (((T_ABSTRACT) reduce 297) ((T_ATTRIBUTE) reduce 297) ((T_CASE) reduce 297) ((T_CONST) reduce 297) ((T_FINAL) reduce 297) ((T_FUNCTION) reduce 297) ((T_PRIVATE) reduce 297) ((T_PROTECTED) reduce 297) ((T_PUBLIC) reduce 297) ((T_READONLY) reduce 297) ((T_STATIC) reduce 297) ((T_USE) reduce 297) ((T_VAR) reduce 297) (("}") reduce 297)) 2908 (((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CASE) shift) ((T_CONST) reduce 326) ((T_FINAL) shift) ((T_FUNCTION) reduce 326) ((T_PRIVATE) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_USE) shift) ((T_VAR) shift) (("}") shift)) 2909 ((("?") reduce 334) ((T_ABSTRACT) reduce 334) ((T_ARRAY) reduce 334) ((T_CALLABLE) reduce 334) ((T_CONST) reduce 334) ((T_FINAL) reduce 334) ((T_FUNCTION) reduce 334) ((T_NAME_FULLY_QUALIFIED) reduce 334) ((T_NAME_QUALIFIED) reduce 334) ((T_NAME_RELATIVE) reduce 334) ((T_PRIVATE) reduce 334) ((T_PROTECTED) reduce 334) ((T_PUBLIC) reduce 334) ((T_READONLY) reduce 334) ((T_STATIC) reduce 334) ((T_STRING) reduce 334) ((T_VARIABLE) reduce 334)) 2911 (((T_ABSTRACT) reduce 443) ((T_ARRAY) reduce 443) ((T_AS) reduce 443) ((T_BREAK) reduce 443) ((T_CALLABLE) reduce 443) ((T_CASE) reduce 443) ((T_CATCH) reduce 443) ((T_CLASS) reduce 443) ((T_CLASS_C) reduce 443) ((T_CLONE) reduce 443) ((T_CONST) reduce 443) ((T_CONTINUE) reduce 443) ((T_DECLARE) reduce 443) ((T_DEFAULT) reduce 443) ((T_DIR) reduce 443) ((T_DO) reduce 443) ((T_ECHO) reduce 443) ((T_ELSE) reduce 443) ((T_ELSEIF) reduce 443) ((T_EMPTY) reduce 443) ((T_ENDDECLARE) reduce 443) ((T_ENDFOR) reduce 443) ((T_ENDFOREACH) reduce 443) ((T_ENDIF) reduce 443) ((T_ENDSWITCH) reduce 443) ((T_ENDWHILE) reduce 443) ((T_ENUM) reduce 443) ((T_EVAL) reduce 443) ((T_EXIT) reduce 443) ((T_EXTENDS) reduce 443) ((T_FILE) reduce 443) ((T_FINAL) reduce 443) ((T_FINALLY) reduce 443) ((T_FN) reduce 443) ((T_FOR) reduce 443) ((T_FOREACH) reduce 443) ((T_FUNCTION) reduce 443) ((T_FUNC_C) reduce 443) ((T_GLOBAL) reduce 443) ((T_GOTO) reduce 443) ((T_IF) reduce 443) ((T_IMPLEMENTS) reduce 443) ((T_INCLUDE) reduce 443) ((T_INCLUDE_ONCE) reduce 443) ((T_INSTANCEOF) reduce 443) ((T_INSTEADOF) reduce 443) ((T_INTERFACE) reduce 443) ((T_ISSET) reduce 443) ((T_LINE) reduce 443) ((T_LIST) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_MATCH) reduce 443) ((T_METHOD_C) reduce 443) ((T_NAMESPACE) reduce 443) ((T_NEW) reduce 443) ((T_NS_C) reduce 443) ((T_PRINT) reduce 443) ((T_PRIVATE) reduce 443) ((T_PROTECTED) reduce 443) ((T_PUBLIC) reduce 443) ((T_READONLY) reduce 443) ((T_REQUIRE) reduce 443) ((T_REQUIRE_ONCE) reduce 443) ((T_RETURN) reduce 443) ((T_STATIC) reduce 443) ((T_STRING) reduce 443) ((T_SWITCH) reduce 443) ((T_THROW) reduce 443) ((T_TRAIT) reduce 443) ((T_TRAIT_C) reduce 443) ((T_TRY) reduce 443) ((T_UNSET) reduce 443) ((T_USE) reduce 443) ((T_VAR) reduce 443) ((T_WHILE) reduce 443) ((T_YIELD) reduce 443)) 2912 ((("?") reduce 335) ((T_ABSTRACT) reduce 335) ((T_ARRAY) reduce 335) ((T_CALLABLE) reduce 335) ((T_CONST) reduce 335) ((T_FINAL) reduce 335) ((T_FUNCTION) reduce 335) ((T_NAME_FULLY_QUALIFIED) reduce 335) ((T_NAME_QUALIFIED) reduce 335) ((T_NAME_RELATIVE) reduce 335) ((T_PRIVATE) reduce 335) ((T_PROTECTED) reduce 335) ((T_PUBLIC) reduce 335) ((T_READONLY) reduce 335) ((T_STATIC) reduce 335) ((T_STRING) reduce 335) ((T_VARIABLE) reduce 335)) 2913 ((("?") reduce 332) ((T_ABSTRACT) reduce 332) ((T_ARRAY) reduce 332) ((T_CALLABLE) reduce 332) ((T_CONST) reduce 332) ((T_FINAL) reduce 332) ((T_FUNCTION) reduce 332) ((T_NAME_FULLY_QUALIFIED) reduce 332) ((T_NAME_QUALIFIED) reduce 332) ((T_NAME_RELATIVE) reduce 332) ((T_PRIVATE) reduce 332) ((T_PROTECTED) reduce 332) ((T_PUBLIC) reduce 332) ((T_READONLY) reduce 332) ((T_STATIC) reduce 332) ((T_STRING) reduce 332) ((T_VARIABLE) reduce 332)) 2914 ((("?") reduce 331) ((T_ABSTRACT) reduce 331) ((T_ARRAY) reduce 331) ((T_CALLABLE) reduce 331) ((T_CONST) reduce 331) ((T_FINAL) reduce 331) ((T_FUNCTION) reduce 331) ((T_NAME_FULLY_QUALIFIED) reduce 331) ((T_NAME_QUALIFIED) reduce 331) ((T_NAME_RELATIVE) reduce 331) ((T_PRIVATE) reduce 331) ((T_PROTECTED) reduce 331) ((T_PUBLIC) reduce 331) ((T_READONLY) reduce 331) ((T_STATIC) reduce 331) ((T_STRING) reduce 331) ((T_VARIABLE) reduce 331)) 2915 ((("?") reduce 330) ((T_ABSTRACT) reduce 330) ((T_ARRAY) reduce 330) ((T_CALLABLE) reduce 330) ((T_CONST) reduce 330) ((T_FINAL) reduce 330) ((T_FUNCTION) reduce 330) ((T_NAME_FULLY_QUALIFIED) reduce 330) ((T_NAME_QUALIFIED) reduce 330) ((T_NAME_RELATIVE) reduce 330) ((T_PRIVATE) reduce 330) ((T_PROTECTED) reduce 330) ((T_PUBLIC) reduce 330) ((T_READONLY) reduce 330) ((T_STATIC) reduce 330) ((T_STRING) reduce 330) ((T_VARIABLE) reduce 330)) 2916 ((("?") reduce 336) ((T_ABSTRACT) reduce 336) ((T_ARRAY) reduce 336) ((T_CALLABLE) reduce 336) ((T_CONST) reduce 336) ((T_FINAL) reduce 336) ((T_FUNCTION) reduce 336) ((T_NAME_FULLY_QUALIFIED) reduce 336) ((T_NAME_QUALIFIED) reduce 336) ((T_NAME_RELATIVE) reduce 336) ((T_PRIVATE) reduce 336) ((T_PROTECTED) reduce 336) ((T_PUBLIC) reduce 336) ((T_READONLY) reduce 336) ((T_STATIC) reduce 336) ((T_STRING) reduce 336) ((T_VARIABLE) reduce 336)) 2917 ((("?") reduce 333) ((T_ABSTRACT) reduce 333) ((T_ARRAY) reduce 333) ((T_CALLABLE) reduce 333) ((T_CONST) reduce 333) ((T_FINAL) reduce 333) ((T_FUNCTION) reduce 333) ((T_NAME_FULLY_QUALIFIED) reduce 333) ((T_NAME_QUALIFIED) reduce 333) ((T_NAME_RELATIVE) reduce 333) ((T_PRIVATE) reduce 333) ((T_PROTECTED) reduce 333) ((T_PUBLIC) reduce 333) ((T_READONLY) reduce 333) ((T_STATIC) reduce 333) ((T_STRING) reduce 333) ((T_VARIABLE) reduce 333)) 2919 ((("?") reduce 325) ((T_ARRAY) reduce 325) ((T_CALLABLE) reduce 325) ((T_NAME_FULLY_QUALIFIED) reduce 325) ((T_NAME_QUALIFIED) reduce 325) ((T_NAME_RELATIVE) reduce 325) ((T_STRING) reduce 325) ((T_VARIABLE) reduce 325)) 2920 (((T_ABSTRACT) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_CASE) reduce 100) ((T_CONST) reduce 100) ((T_FINAL) reduce 100) ((T_FUNCTION) reduce 100) ((T_PRIVATE) reduce 100) ((T_PROTECTED) reduce 100) ((T_PUBLIC) reduce 100) ((T_READONLY) reduce 100) ((T_STATIC) reduce 100) ((T_VAR) reduce 100)) 2921 (((T_ABSTRACT) reduce 302) ((T_ATTRIBUTE) reduce 302) ((T_CASE) reduce 302) ((T_CONST) reduce 302) ((T_FINAL) reduce 302) ((T_FUNCTION) reduce 302) ((T_PRIVATE) reduce 302) ((T_PROTECTED) reduce 302) ((T_PUBLIC) reduce 302) ((T_READONLY) reduce 302) ((T_STATIC) reduce 302) ((T_USE) reduce 302) ((T_VAR) reduce 302) (("}") reduce 302)) 2922 (((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CASE) shift) ((T_CONST) reduce 326) ((T_FINAL) shift) ((T_FUNCTION) reduce 326) ((T_PRIVATE) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_VAR) shift)) 2923 (((T_ABSTRACT) reduce 296) ((T_ATTRIBUTE) reduce 296) ((T_CASE) reduce 296) ((T_CONST) reduce 296) ((T_FINAL) reduce 296) ((T_FUNCTION) reduce 296) ((T_PRIVATE) reduce 296) ((T_PROTECTED) reduce 296) ((T_PUBLIC) reduce 296) ((T_READONLY) reduce 296) ((T_STATIC) reduce 296) ((T_USE) reduce 296) ((T_VAR) reduce 296) (("}") reduce 296)) 2924 (((T_ABSTRACT) reduce 301) ((T_ATTRIBUTE) reduce 301) ((T_CASE) reduce 301) ((T_CONST) reduce 301) ((T_FINAL) reduce 301) ((T_FUNCTION) reduce 301) ((T_PRIVATE) reduce 301) ((T_PROTECTED) reduce 301) ((T_PUBLIC) reduce 301) ((T_READONLY) reduce 301) ((T_STATIC) reduce 301) ((T_USE) reduce 301) ((T_VAR) reduce 301) (("}") reduce 301)) 2925 ((("?") reduce 328) ((T_ABSTRACT) reduce 328) ((T_ARRAY) reduce 328) ((T_CALLABLE) reduce 328) ((T_CONST) reduce 328) ((T_FINAL) reduce 328) ((T_FUNCTION) reduce 328) ((T_NAME_FULLY_QUALIFIED) reduce 328) ((T_NAME_QUALIFIED) reduce 328) ((T_NAME_RELATIVE) reduce 328) ((T_PRIVATE) reduce 328) ((T_PROTECTED) reduce 328) ((T_PUBLIC) reduce 328) ((T_READONLY) reduce 328) ((T_STATIC) reduce 328) ((T_STRING) reduce 328) ((T_VARIABLE) reduce 328)) 2926 (((T_CONST) shift) ((T_FUNCTION) shift)) 2927 ((("?") reduce 324) ((T_ABSTRACT) shift) ((T_ARRAY) reduce 324) ((T_CALLABLE) reduce 324) ((T_CONST) reduce 327) ((T_FINAL) shift) ((T_FUNCTION) reduce 327) ((T_NAME_FULLY_QUALIFIED) reduce 324) ((T_NAME_QUALIFIED) reduce 324) ((T_NAME_RELATIVE) reduce 324) ((T_PRIVATE) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 324) ((T_VARIABLE) reduce 324)) 2928 ((("?") shift) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STRING) shift) ((T_VARIABLE) reduce 256)) 2929 ((("%") reduce 352) ((")") reduce 352) (("*") reduce 352) (("+") reduce 352) ((",") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 2930 (((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STRING) shift)) 2931 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_VARIABLE) reduce 272) (("|") reduce 272)) 2932 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_VARIABLE) reduce 273) (("|") reduce 273)) 2933 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_VARIABLE) reduce 93) (("|") reduce 93)) 2934 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_VARIABLE) reduce 92) (("|") reduce 92)) 2935 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_VARIABLE) reduce 94) (("|") reduce 94)) 2936 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_VARIABLE) reduce 91) (("|") reduce 91)) 2937 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_VARIABLE) reduce 271)) 2938 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_VARIABLE) reduce 274) (("|") reduce 274)) 2939 (((T_VARIABLE) shift)) 2940 (((T_VARIABLE) reduce 257)) 2941 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_VARIABLE) reduce 268) (("|") shift)) 2942 (((T_VARIABLE) reduce 270) (("|") shift)) 2944 (((T_VARIABLE) reduce 272) (("|") reduce 272)) 2945 (((T_VARIABLE) reduce 273) (("|") reduce 273)) 2946 (((T_VARIABLE) reduce 93) (("|") reduce 93)) 2947 (((T_VARIABLE) reduce 92) (("|") reduce 92)) 2948 (((T_VARIABLE) reduce 94) (("|") reduce 94)) 2949 (((T_VARIABLE) reduce 91) (("|") reduce 91)) 2950 (((T_VARIABLE) reduce 274) (("|") reduce 274)) 2951 (((T_VARIABLE) reduce 276) (("|") reduce 276)) 2954 (((T_VARIABLE) reduce 275) (("|") reduce 275)) 2955 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_VARIABLE) reduce 272)) 2956 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_VARIABLE) reduce 273)) 2957 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_VARIABLE) reduce 93)) 2958 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_VARIABLE) reduce 92)) 2959 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_VARIABLE) reduce 94)) 2960 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_VARIABLE) reduce 91)) 2961 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_VARIABLE) reduce 274)) 2962 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 277) ((T_VARIABLE) reduce 277)) 2963 (((",") reduce 443) ((";") reduce 443) (("=") shift)) 2964 (((",") reduce 338) ((";") reduce 338)) 2965 (((",") shift) ((";") shift)) 2967 (((T_ABSTRACT) reduce 298) ((T_ATTRIBUTE) reduce 298) ((T_CASE) reduce 298) ((T_CONST) reduce 298) ((T_FINAL) reduce 298) ((T_FUNCTION) reduce 298) ((T_PRIVATE) reduce 298) ((T_PROTECTED) reduce 298) ((T_PUBLIC) reduce 298) ((T_READONLY) reduce 298) ((T_STATIC) reduce 298) ((T_USE) reduce 298) ((T_VAR) reduce 298) (("}") reduce 298)) 2968 (((",") reduce 337) ((";") reduce 337)) 2970 (((",") reduce 339) ((";") reduce 339)) 2981 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 2983 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 2985 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 2986 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 2990 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) ((",") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) ((";") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 2991 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 2992 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 2998 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) ((",") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) ((";") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 3000 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 3002 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 3003 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 3004 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 3005 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 3007 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 3014 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 3017 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 3019 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 3020 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) ((",") reduce 429) (("-") shift) ((".") reduce 429) (("/") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 3027 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 3028 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) ((",") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) ((";") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 3030 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) ((",") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) ((";") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 3031 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) ((",") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 3032 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 443) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3035 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 3036 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) ((",") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) ((";") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 3037 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) ((",") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) ((";") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 3038 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 3039 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) ((",") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) ((";") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 3040 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) ((",") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) ((";") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 3041 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 3042 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 3043 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 3044 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) ((",") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) ((";") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 3047 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 3078 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 384) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 384) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 3079 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 387) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 387) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 3080 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") reduce 396) (("/") shift) ((";") reduce 396) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 3081 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 409) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 409) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 3082 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") reduce 395) (("/") shift) ((";") reduce 395) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 3083 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 3084 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 383) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 383) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3085 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 381) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 381) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3086 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 382) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 382) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3087 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 3088 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 402) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 3089 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 3090 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 401) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 3091 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 408) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 408) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 3092 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 403) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 3095 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 3096 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 3097 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 3098 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 3099 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 3100 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 3101 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 3102 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 3103 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 3104 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 3105 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 3112 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 3114 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 3115 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 3116 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 3117 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 3118 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 3121 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 3122 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 3124 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 3126 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 3127 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 3130 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 3131 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3132 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 379) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 379) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3133 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 380) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 380) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3134 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 385) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 385) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 3135 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 386) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 386) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 3139 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3140 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3141 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 407) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 407) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 3142 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 3143 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 3144 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 388) (("-") shift) ((".") reduce 388) (("/") shift) ((";") reduce 388) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 3145 ((("%") shift) (("*") shift) (("+") reduce 390) ((",") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) ((";") reduce 390) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 3146 ((("%") shift) (("*") shift) (("+") reduce 389) ((",") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) ((";") reduce 389) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 3147 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) ((",") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) ((";") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 3148 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) ((",") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) ((";") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 3150 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 3151 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 3152 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 3153 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 3154 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 3155 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 3156 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 3157 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 3158 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 3159 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 3160 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 3161 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 3162 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 3163 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 3164 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 3165 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 3166 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 3167 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 3168 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 3169 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 3170 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 3171 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 3172 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 3173 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 3174 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 3175 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 3176 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 3177 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 3178 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 3179 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 3180 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 3181 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 3182 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 3183 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 3184 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 3185 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 3186 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 3187 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 3188 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 3189 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 3190 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 3191 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 3192 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 3193 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 3194 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 3195 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 3196 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 3197 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 3198 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 3199 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 3200 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 3201 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 3202 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 3203 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 3204 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 3205 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 3206 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 3207 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 3208 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 3209 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 3210 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 3211 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 3212 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 3213 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 3214 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 3215 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 3216 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 3217 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) ((",") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) ((";") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 3218 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 3219 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 3220 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 3221 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 3222 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 3223 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 3224 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 3225 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 3226 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 3227 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 3228 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 3230 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 3231 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 3232 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 3234 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 3235 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((";") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 3239 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((";") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 3240 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((";") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 3245 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) ((",") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) ((";") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 3247 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) ((",") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) ((";") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 3257 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 371) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 371) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3258 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 373) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 373) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3259 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 372) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 372) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3260 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 365) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 365) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3261 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 362) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 362) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3262 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 370) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 370) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3263 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 364) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 364) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3264 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 368) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 368) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3265 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 363) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 363) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3266 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 366) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 366) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3267 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 367) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 367) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3268 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 374) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 374) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3269 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 369) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 369) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3271 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 359) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 359) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3273 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 3276 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 3278 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 3280 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 3281 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 3282 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) ((",") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) ((";") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 3286 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 3288 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 3289 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((";") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 3293 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((";") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 3294 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((";") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 3295 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 3298 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 3299 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 3300 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 3306 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 3308 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 3309 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 3310 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 3311 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 3314 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 3315 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 3316 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 3317 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 3318 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 3321 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 3322 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 3333 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 3334 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 3344 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 444) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3345 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 3346 (((",") reduce 340) ((";") reduce 340)) 3348 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 3350 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 3351 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 3352 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 3354 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) ((",") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) ((";") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 3355 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 3361 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 3363 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 3364 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 3365 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 3366 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 3369 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 3370 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 3371 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 3372 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 3374 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) ((",") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 3376 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 3378 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 358) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 358) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3379 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 432) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3390 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 3392 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 3394 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 3395 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 3399 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) ((",") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) ((";") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 3400 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 3401 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 3407 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) ((",") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) ((";") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 3409 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 3411 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 3412 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 3413 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 3414 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 3416 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 3423 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 3426 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 3428 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 3429 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) ((",") reduce 429) (("-") shift) ((".") reduce 429) (("/") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 3436 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 3437 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) ((",") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) ((";") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 3439 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) ((",") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) ((";") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_DOUBLE_ARROW) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 3440 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) ((",") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_DOUBLE_ARROW) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 3441 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3444 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 3445 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) ((",") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) ((";") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 3446 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) ((",") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) ((";") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 3447 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 3448 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) ((",") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) ((";") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 3449 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) ((",") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) ((";") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 3450 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 3451 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 3452 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 3453 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) ((",") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) ((";") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 356) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 3456 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 3487 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 384) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 384) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 3488 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 387) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 387) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 3489 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") reduce 396) (("/") shift) ((";") reduce 396) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 3490 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 409) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 409) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 3491 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") reduce 395) (("/") shift) ((";") reduce 395) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 3492 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 3493 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 383) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 383) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3494 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 381) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 381) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 381) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3495 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 382) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 382) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 382) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3496 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 3497 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 402) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 3498 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 3499 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 401) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 3500 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 408) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 408) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 3501 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 403) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 3504 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 3505 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 3506 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 3507 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 3508 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 3509 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 3510 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 3511 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 3512 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 3513 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 3514 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 3521 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 3523 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 3524 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 3525 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 3526 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 3527 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 3530 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 3531 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 3533 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 3535 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 3536 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 3539 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 3540 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3541 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 379) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 379) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_DOUBLE_ARROW) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3542 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 380) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 380) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_DOUBLE_ARROW) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3543 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 385) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 385) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 3544 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 386) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 386) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 3548 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3549 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3550 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 407) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 407) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 3551 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 3552 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 3553 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 388) (("-") shift) ((".") reduce 388) (("/") shift) ((";") reduce 388) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_DOUBLE_ARROW) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 3554 ((("%") shift) (("*") shift) (("+") reduce 390) ((",") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) ((";") reduce 390) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 3555 ((("%") shift) (("*") shift) (("+") reduce 389) ((",") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) ((";") reduce 389) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 3556 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) ((",") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) ((";") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 3557 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) ((",") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) ((";") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 3559 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 3560 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 3561 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 3562 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 3563 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 3564 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 3565 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 3566 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 3567 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 3568 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 3569 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 3570 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 3571 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 3572 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 3573 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 3574 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 3575 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 3576 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 3577 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 3578 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 3579 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 3580 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 3581 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 3582 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 3583 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 3584 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 3585 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 3586 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 3587 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 3588 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 3589 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 3590 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 3591 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 3592 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 3593 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 3594 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 3595 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 3596 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 3597 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 3598 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 3599 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 3600 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 3601 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 3602 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 3603 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 3604 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 3605 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 3606 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 3607 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 3608 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 3609 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 3610 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 3611 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 3612 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 3613 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 3614 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 3615 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 3616 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 3617 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 3618 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 3619 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 3620 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 3621 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 3622 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 3623 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 3624 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 3625 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 3626 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) ((",") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) ((";") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_DOUBLE_ARROW) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 3627 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 3628 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 3629 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 3630 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 3631 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 3632 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 3633 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 3634 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 3635 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 3636 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 3637 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 3639 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 3640 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 3641 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 3643 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 3644 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((";") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 3648 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((";") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 3649 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((";") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 3654 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) ((",") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) ((";") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_DOUBLE_ARROW) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 3656 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) ((",") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) ((";") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_DOUBLE_ARROW) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 3666 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 371) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 371) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 371) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3667 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 373) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 373) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 373) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3668 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 372) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 372) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 372) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3669 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 365) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 365) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 365) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3670 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 362) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 362) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 362) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3671 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 370) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 370) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 370) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3672 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 364) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 364) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 364) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3673 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 368) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 368) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 368) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3674 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 363) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 363) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 363) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3675 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 366) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 366) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 366) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3676 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 367) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 367) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 367) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3677 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 374) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 374) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 374) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3678 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 369) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 369) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 369) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3680 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 359) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 359) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 359) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3682 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 3685 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 3687 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 3689 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 3690 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 3691 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) ((",") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) ((";") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_DOUBLE_ARROW) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 3695 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 3697 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 3698 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((";") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 3702 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((";") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 3703 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((";") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 3704 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 3707 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 3708 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 3709 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 3715 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 3717 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 3718 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 3719 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 3720 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 3723 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 3724 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 3725 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 3726 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 3727 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 3730 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 3731 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 3742 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 3743 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 3753 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 444) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3754 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 3756 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3758 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 3760 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 3761 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 3762 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 3764 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) ((",") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) ((";") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 3765 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 3771 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 3773 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 3774 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 3775 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 3776 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 3779 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 3780 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 3781 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 3782 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 3784 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) ((",") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 3786 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 3788 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 358) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 358) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 358) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3789 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 432) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3792 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3793 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 3794 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3795 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) ((",") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) ((";") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 3796 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) ((",") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 3798 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 3800 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) ((",") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) ((";") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 3801 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 3802 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 567) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 567) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3803 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 566) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 566) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3804 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 428) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3805 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 3808 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 3809 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 3810 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 3811 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 3812 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 3813 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) ((",") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) ((";") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_DOUBLE_ARROW) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 3815 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 3816 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 3817 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 3818 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 3819 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 3826 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 3828 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 3829 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 3830 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 3831 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 3833 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 3834 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) ((",") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) ((";") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_DOUBLE_ARROW) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 3835 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((";") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 3839 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((";") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 3840 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((";") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 3842 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 3843 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) ((",") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) ((";") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_DOUBLE_ARROW) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 3850 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) ((",") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) ((";") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_DOUBLE_ARROW) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 3852 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 3856 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("}") reduce 224) (("~") shift)) 3867 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 3869 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 3871 (((",") shift) ((T_DOUBLE_ARROW) reduce 123)) 3872 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 3873 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 3877 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) ((",") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 3878 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 3879 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 3885 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) ((",") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 3887 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 3889 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 3890 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 3891 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 3892 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 3894 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 3901 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 3904 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 3906 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 3907 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) ((",") reduce 429) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 3914 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 3915 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) ((",") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 3917 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) ((",") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_DOUBLE_ARROW) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 3918 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) ((",") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_DOUBLE_ARROW) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 3919 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 230) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 230) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3922 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 3923 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) ((",") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 3924 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) ((",") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 3925 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 3926 (((",") reduce 226) (("}") reduce 226)) 3929 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) ((",") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 3930 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) ((",") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 3931 (((",") shift) (("}") reduce 123)) 3932 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 3933 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 3934 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 3935 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) ((",") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 356) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 3938 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 3969 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 384) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 3970 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 387) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 3971 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 3972 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 409) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 3973 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 3974 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 3975 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 383) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3976 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 381) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 381) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3977 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 382) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 382) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3978 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 3979 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 3980 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 3981 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 3982 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 408) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 3983 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 3986 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 3987 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 3988 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 3989 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 3990 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 3991 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 3992 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 3993 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 3994 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 3995 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 3996 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 4003 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 4005 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 4006 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 4007 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 4008 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 4009 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 4012 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 4013 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 4015 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 4017 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 4018 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 4021 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 4022 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4023 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 379) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_DOUBLE_ARROW) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4024 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 380) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_DOUBLE_ARROW) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4025 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 385) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 4026 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 386) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 4030 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4031 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4032 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 407) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 4033 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 4034 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 4035 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 388) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_DOUBLE_ARROW) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 4036 ((("%") shift) (("*") shift) (("+") reduce 390) ((",") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 4037 ((("%") shift) (("*") shift) (("+") reduce 389) ((",") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 4038 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) ((",") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 4039 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) ((",") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 4041 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 4042 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 4043 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 4044 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 4045 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 4046 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 4047 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 4048 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 4049 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 4050 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 4051 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 4052 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 4053 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 4054 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 4055 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 4056 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 4057 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 4058 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 4059 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 4060 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 4061 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 4062 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 4063 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 4064 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 4065 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 4066 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 4067 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 4068 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 4069 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 4070 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 4071 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 4072 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 4073 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 4074 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 4075 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 4076 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 4077 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 4078 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 4079 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 4080 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 4081 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 4082 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 4083 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 4084 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 4085 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 4086 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 4087 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 4088 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 4089 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 4090 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 4091 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 4092 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 4093 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 4094 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 4095 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 4096 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 4097 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 4098 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 4099 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 4100 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 4101 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 4102 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 4103 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 4104 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 4105 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 4106 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 4107 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 4108 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) ((",") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_DOUBLE_ARROW) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 4109 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 4110 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 4111 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 4112 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 4113 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 4114 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 4115 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 4116 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 4117 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 4118 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 4119 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 4121 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 4122 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 4123 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 4125 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 4126 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 4130 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 4131 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 4136 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) ((",") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_DOUBLE_ARROW) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 4138 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) ((",") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_DOUBLE_ARROW) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 4148 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 371) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 371) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4149 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 373) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 373) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4150 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 372) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 372) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4151 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 365) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 365) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4152 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 362) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 362) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4153 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 370) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 370) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4154 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 364) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 364) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4155 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 368) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 368) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4156 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 363) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 363) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4157 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 366) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 366) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4158 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 367) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 367) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4159 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 374) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 374) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4160 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 369) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 369) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4162 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 359) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 359) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4164 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 4167 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 4169 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 4171 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 4172 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 4173 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) ((",") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_DOUBLE_ARROW) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 4177 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 4179 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 4180 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 4184 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 4185 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 4186 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 4189 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 4190 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 4191 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 4197 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 4199 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 4200 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 4201 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 4202 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 4205 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 4206 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 4207 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 4208 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 4209 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 4212 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 4213 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("}") reduce 124) (("~") shift)) 4214 ((("}") reduce 225)) 4215 (((",") reduce 227) (("}") reduce 227)) 4216 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 4217 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) ((",") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) ((";") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_DOUBLE_ARROW) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 4218 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 124) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 4231 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490) (("}") reduce 490)) 4233 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472) (("}") reduce 472)) 4235 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485) (("}") reduce 485)) 4236 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475) (("}") reduce 475)) 4240 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) ((",") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463) (("}") reduce 463)) 4241 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484) (("}") reduce 484)) 4242 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488) (("}") reduce 488)) 4248 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) ((",") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483) (("}") reduce 483)) 4250 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474) (("}") reduce 474)) 4252 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487) (("}") reduce 487)) 4253 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93) (("}") reduce 93)) 4254 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92) (("}") reduce 92)) 4255 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94) (("}") reduce 94)) 4257 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489) (("}") reduce 489)) 4264 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91) (("}") reduce 91)) 4267 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486) (("}") reduce 486)) 4269 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515) (("}") reduce 515)) 4270 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) ((",") reduce 429) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("}") reduce 429) (("~") shift)) 4277 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511) (("}") reduce 511)) 4278 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) ((",") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481) (("}") reduce 481)) 4280 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) ((",") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480) (("}") reduce 480)) 4281 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) ((",") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479) (("}") reduce 479)) 4282 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 228) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 228)) 4285 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510) (("}") reduce 510)) 4286 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) ((",") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434) (("}") reduce 434)) 4287 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) ((",") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416) (("}") reduce 416)) 4288 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438) (("}") reduce 438)) 4289 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) ((",") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482) (("}") reduce 482)) 4290 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) ((",") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412) (("}") reduce 412)) 4291 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426) (("}") reduce 426)) 4292 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505) (("}") reduce 505)) 4293 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512) (("}") reduce 512)) 4294 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) ((",") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356) (("}") reduce 356)) 4297 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400) (("}") reduce 400)) 4328 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 384) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384) (("}") reduce 384)) 4329 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 387) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387) (("}") reduce 387)) 4330 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396) (("}") reduce 396)) 4331 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 409) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409) (("}") reduce 409)) 4332 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395) (("}") reduce 395)) 4333 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392) (("}") reduce 392)) 4334 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 383) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 383)) 4335 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 381) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 381)) 4336 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 382) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 382)) 4337 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406) (("}") reduce 406)) 4338 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402) (("}") reduce 402)) 4339 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404) (("}") reduce 404)) 4340 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401) (("}") reduce 401)) 4341 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 408) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408) (("}") reduce 408)) 4342 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403) (("}") reduce 403)) 4345 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 4346 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 4347 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 4348 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 4349 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91) (("}") reduce 91)) 4350 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515) (("}") reduce 515)) 4351 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 4352 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410) (("}") reduce 410)) 4353 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 4354 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461) (("}") reduce 461)) 4355 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520) (("}") reduce 520)) 4362 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522) (("}") reduce 522)) 4364 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521) (("}") reduce 521)) 4365 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526) (("}") reduce 526)) 4366 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530) (("}") reduce 530)) 4367 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523) (("}") reduce 523)) 4368 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532) (("}") reduce 532)) 4371 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531) (("}") reduce 531)) 4372 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524) (("}") reduce 524)) 4374 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525) (("}") reduce 525)) 4376 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462) (("}") reduce 462)) 4377 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517) (("}") reduce 517)) 4380 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516) (("}") reduce 516)) 4381 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 415)) 4382 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 379) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 379)) 4383 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 380) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 380)) 4384 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 385) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385) (("}") reduce 385)) 4385 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 386) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386) (("}") reduce 386)) 4389 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 413)) 4390 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 414)) 4391 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 407) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407) (("}") reduce 407)) 4392 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405) (("}") reduce 405)) 4393 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393) (("}") reduce 393)) 4394 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 388) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388) (("}") reduce 388)) 4395 ((("%") shift) (("*") shift) (("+") reduce 390) ((",") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390) (("}") reduce 390)) 4396 ((("%") shift) (("*") shift) (("+") reduce 389) ((",") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389) (("}") reduce 389)) 4397 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) ((",") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391) (("}") reduce 391)) 4398 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) ((",") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394) (("}") reduce 394)) 4400 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73) (("}") reduce 73)) 4401 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52) (("}") reduce 52)) 4402 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27) (("}") reduce 27)) 4403 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51) (("}") reduce 51)) 4404 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53) (("}") reduce 53)) 4405 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49) (("}") reduce 49)) 4406 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29) (("}") reduce 29)) 4407 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59) (("}") reduce 59)) 4408 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60) (("}") reduce 60)) 4409 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11) (("}") reduce 11)) 4410 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42) (("}") reduce 42)) 4411 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39) (("}") reduce 39)) 4412 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25) (("}") reduce 25)) 4413 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50) (("}") reduce 50)) 4414 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66) (("}") reduce 66)) 4415 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18) (("}") reduce 18)) 4416 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17) (("}") reduce 17)) 4417 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15) (("}") reduce 15)) 4418 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14) (("}") reduce 14)) 4419 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38) (("}") reduce 38)) 4420 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26) (("}") reduce 26)) 4421 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22) (("}") reduce 22)) 4422 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24) (("}") reduce 24)) 4423 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16) (("}") reduce 16)) 4424 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48) (("}") reduce 48)) 4425 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20) (("}") reduce 20)) 4426 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70) (("}") reduce 70)) 4427 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3) (("}") reduce 3)) 4428 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12) (("}") reduce 12)) 4429 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54) (("}") reduce 54)) 4430 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65) (("}") reduce 65)) 4431 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74) (("}") reduce 74)) 4432 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30) (("}") reduce 30)) 4433 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68) (("}") reduce 68)) 4434 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21) (("}") reduce 21)) 4435 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23) (("}") reduce 23)) 4436 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41) (("}") reduce 41)) 4437 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62) (("}") reduce 62)) 4438 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34) (("}") reduce 34)) 4439 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40) (("}") reduce 40)) 4440 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13) (("}") reduce 13)) 4441 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55) (("}") reduce 55)) 4442 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1) (("}") reduce 1)) 4443 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2) (("}") reduce 2)) 4444 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9) (("}") reduce 9)) 4445 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33) (("}") reduce 33)) 4446 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58) (("}") reduce 58)) 4447 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37) (("}") reduce 37)) 4448 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64) (("}") reduce 64)) 4449 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46) (("}") reduce 46)) 4450 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8) (("}") reduce 8)) 4451 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6) (("}") reduce 6)) 4452 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7) (("}") reduce 7)) 4453 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69) (("}") reduce 69)) 4454 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63) (("}") reduce 63)) 4455 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56) (("}") reduce 56)) 4456 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10) (("}") reduce 10)) 4457 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67) (("}") reduce 67)) 4458 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44) (("}") reduce 44)) 4459 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75) (("}") reduce 75)) 4460 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76) (("}") reduce 76)) 4461 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77) (("}") reduce 77)) 4462 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78) (("}") reduce 78)) 4463 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4) (("}") reduce 4)) 4464 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5) (("}") reduce 5)) 4465 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43) (("}") reduce 43)) 4466 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72) (("}") reduce 72)) 4467 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) ((",") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81) (("}") reduce 81)) 4468 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47) (("}") reduce 47)) 4469 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31) (("}") reduce 31)) 4470 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57) (("}") reduce 57)) 4471 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61) (("}") reduce 61)) 4472 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28) (("}") reduce 28)) 4473 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36) (("}") reduce 36)) 4474 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32) (("}") reduce 32)) 4475 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35) (("}") reduce 35)) 4476 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19) (("}") reduce 19)) 4477 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45) (("}") reduce 45)) 4478 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492) (("}") reduce 492)) 4480 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71) (("}") reduce 71)) 4481 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82) (("}") reduce 82)) 4482 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519) (("}") reduce 519)) 4484 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456) (("}") reduce 456)) 4485 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281) (("}") reduce 281)) 4489 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282) (("}") reduce 282)) 4490 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283) (("}") reduce 283)) 4495 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) ((",") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377) (("}") reduce 377)) 4497 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) ((",") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375) (("}") reduce 375)) 4507 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 371) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 371)) 4508 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 373) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 373)) 4509 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 372) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 372)) 4510 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 365) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 365)) 4511 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 362) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 362)) 4512 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 370) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 370)) 4513 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 364) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 364)) 4514 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 368) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 368)) 4515 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 363) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 363)) 4516 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 366) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 366)) 4517 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 367) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 367)) 4518 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 374) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 374)) 4519 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 369) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 369)) 4521 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 359) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 359)) 4523 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515) (("}") reduce 515)) 4526 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511) (("}") reduce 511)) 4528 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510) (("}") reduce 510)) 4530 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505) (("}") reduce 505)) 4531 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512) (("}") reduce 512)) 4532 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) ((",") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360) (("}") reduce 360)) 4536 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519) (("}") reduce 519)) 4538 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456) (("}") reduce 456)) 4539 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281) (("}") reduce 281)) 4543 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282) (("}") reduce 282)) 4544 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283) (("}") reduce 283)) 4545 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454) (("}") reduce 454)) 4548 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518) (("}") reduce 518)) 4549 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455) (("}") reduce 455)) 4550 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457) (("}") reduce 457)) 4556 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507) (("}") reduce 507)) 4558 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506) (("}") reduce 506)) 4559 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530) (("}") reduce 530)) 4560 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513) (("}") reduce 513)) 4561 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532) (("}") reduce 532)) 4564 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531) (("}") reduce 531)) 4565 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508) (("}") reduce 508)) 4566 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514) (("}") reduce 514)) 4567 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509) (("}") reduce 509)) 4568 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517) (("}") reduce 517)) 4571 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516) (("}") reduce 516)) 4572 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454) (("}") reduce 454)) 4583 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444) (("}") reduce 444)) 4584 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439) (("}") reduce 439)) 4594 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 444)) 4595 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440) (("}") reduce 440)) 4597 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491) (("}") reduce 491)) 4599 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518) (("}") reduce 518)) 4600 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455) (("}") reduce 455)) 4601 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457) (("}") reduce 457)) 4603 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) ((",") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435) (("}") reduce 435)) 4604 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437) (("}") reduce 437)) 4610 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507) (("}") reduce 507)) 4612 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506) (("}") reduce 506)) 4613 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530) (("}") reduce 530)) 4614 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513) (("}") reduce 513)) 4615 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532) (("}") reduce 532)) 4618 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531) (("}") reduce 531)) 4619 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508) (("}") reduce 508)) 4620 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514) (("}") reduce 514)) 4621 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509) (("}") reduce 509)) 4623 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) ((",") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427) (("}") reduce 427)) 4625 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471) (("}") reduce 471)) 4627 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 358) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 358)) 4628 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 432)) 4639 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490) (("}") reduce 490)) 4641 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472) (("}") reduce 472)) 4643 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485) (("}") reduce 485)) 4644 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475) (("}") reduce 475)) 4648 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) ((",") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463) (("}") reduce 463)) 4649 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484) (("}") reduce 484)) 4650 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488) (("}") reduce 488)) 4656 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) ((",") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483) (("}") reduce 483)) 4658 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474) (("}") reduce 474)) 4660 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487) (("}") reduce 487)) 4661 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93) (("}") reduce 93)) 4662 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92) (("}") reduce 92)) 4663 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94) (("}") reduce 94)) 4665 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489) (("}") reduce 489)) 4672 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91) (("}") reduce 91)) 4675 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486) (("}") reduce 486)) 4677 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515) (("}") reduce 515)) 4678 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) ((",") reduce 429) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("}") reduce 429) (("~") shift)) 4685 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511) (("}") reduce 511)) 4686 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) ((",") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481) (("}") reduce 481)) 4688 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) ((",") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_DOUBLE_ARROW) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480) (("}") reduce 480)) 4689 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) ((",") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_DOUBLE_ARROW) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479) (("}") reduce 479)) 4690 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 430)) 4693 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510) (("}") reduce 510)) 4694 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) ((",") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434) (("}") reduce 434)) 4695 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) ((",") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416) (("}") reduce 416)) 4696 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438) (("}") reduce 438)) 4697 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) ((",") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482) (("}") reduce 482)) 4698 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) ((",") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412) (("}") reduce 412)) 4699 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426) (("}") reduce 426)) 4700 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505) (("}") reduce 505)) 4701 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512) (("}") reduce 512)) 4702 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) ((",") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 356) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356) (("}") reduce 356)) 4705 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400) (("}") reduce 400)) 4736 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 384) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384) (("}") reduce 384)) 4737 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 387) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387) (("}") reduce 387)) 4738 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396) (("}") reduce 396)) 4739 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 409) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409) (("}") reduce 409)) 4740 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395) (("}") reduce 395)) 4741 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392) (("}") reduce 392)) 4742 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 383) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 383)) 4743 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 381) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 381) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 381)) 4744 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 382) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 382) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 382)) 4745 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406) (("}") reduce 406)) 4746 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402) (("}") reduce 402)) 4747 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404) (("}") reduce 404)) 4748 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401) (("}") reduce 401)) 4749 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 408) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408) (("}") reduce 408)) 4750 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403) (("}") reduce 403)) 4753 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 4754 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 4755 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 4756 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 4757 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91) (("}") reduce 91)) 4758 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515) (("}") reduce 515)) 4759 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 4760 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410) (("}") reduce 410)) 4761 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 4762 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461) (("}") reduce 461)) 4763 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520) (("}") reduce 520)) 4770 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522) (("}") reduce 522)) 4772 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521) (("}") reduce 521)) 4773 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526) (("}") reduce 526)) 4774 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530) (("}") reduce 530)) 4775 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523) (("}") reduce 523)) 4776 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532) (("}") reduce 532)) 4779 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531) (("}") reduce 531)) 4780 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524) (("}") reduce 524)) 4782 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525) (("}") reduce 525)) 4784 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462) (("}") reduce 462)) 4785 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517) (("}") reduce 517)) 4788 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516) (("}") reduce 516)) 4789 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 415)) 4790 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 379) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_DOUBLE_ARROW) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 379)) 4791 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 380) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_DOUBLE_ARROW) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 380)) 4792 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 385) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385) (("}") reduce 385)) 4793 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 386) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386) (("}") reduce 386)) 4797 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 413)) 4798 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 414)) 4799 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 407) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407) (("}") reduce 407)) 4800 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405) (("}") reduce 405)) 4801 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393) (("}") reduce 393)) 4802 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 388) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_DOUBLE_ARROW) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388) (("}") reduce 388)) 4803 ((("%") shift) (("*") shift) (("+") reduce 390) ((",") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390) (("}") reduce 390)) 4804 ((("%") shift) (("*") shift) (("+") reduce 389) ((",") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389) (("}") reduce 389)) 4805 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) ((",") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391) (("}") reduce 391)) 4806 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) ((",") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394) (("}") reduce 394)) 4808 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73) (("}") reduce 73)) 4809 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52) (("}") reduce 52)) 4810 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27) (("}") reduce 27)) 4811 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51) (("}") reduce 51)) 4812 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53) (("}") reduce 53)) 4813 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49) (("}") reduce 49)) 4814 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29) (("}") reduce 29)) 4815 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59) (("}") reduce 59)) 4816 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60) (("}") reduce 60)) 4817 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11) (("}") reduce 11)) 4818 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42) (("}") reduce 42)) 4819 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39) (("}") reduce 39)) 4820 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25) (("}") reduce 25)) 4821 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50) (("}") reduce 50)) 4822 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66) (("}") reduce 66)) 4823 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18) (("}") reduce 18)) 4824 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17) (("}") reduce 17)) 4825 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15) (("}") reduce 15)) 4826 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14) (("}") reduce 14)) 4827 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38) (("}") reduce 38)) 4828 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26) (("}") reduce 26)) 4829 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22) (("}") reduce 22)) 4830 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24) (("}") reduce 24)) 4831 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16) (("}") reduce 16)) 4832 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48) (("}") reduce 48)) 4833 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20) (("}") reduce 20)) 4834 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70) (("}") reduce 70)) 4835 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3) (("}") reduce 3)) 4836 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12) (("}") reduce 12)) 4837 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54) (("}") reduce 54)) 4838 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65) (("}") reduce 65)) 4839 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74) (("}") reduce 74)) 4840 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30) (("}") reduce 30)) 4841 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68) (("}") reduce 68)) 4842 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21) (("}") reduce 21)) 4843 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23) (("}") reduce 23)) 4844 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41) (("}") reduce 41)) 4845 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62) (("}") reduce 62)) 4846 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34) (("}") reduce 34)) 4847 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40) (("}") reduce 40)) 4848 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13) (("}") reduce 13)) 4849 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55) (("}") reduce 55)) 4850 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1) (("}") reduce 1)) 4851 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2) (("}") reduce 2)) 4852 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9) (("}") reduce 9)) 4853 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33) (("}") reduce 33)) 4854 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58) (("}") reduce 58)) 4855 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37) (("}") reduce 37)) 4856 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64) (("}") reduce 64)) 4857 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46) (("}") reduce 46)) 4858 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8) (("}") reduce 8)) 4859 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6) (("}") reduce 6)) 4860 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7) (("}") reduce 7)) 4861 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69) (("}") reduce 69)) 4862 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63) (("}") reduce 63)) 4863 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56) (("}") reduce 56)) 4864 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10) (("}") reduce 10)) 4865 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67) (("}") reduce 67)) 4866 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44) (("}") reduce 44)) 4867 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75) (("}") reduce 75)) 4868 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76) (("}") reduce 76)) 4869 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77) (("}") reduce 77)) 4870 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78) (("}") reduce 78)) 4871 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4) (("}") reduce 4)) 4872 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5) (("}") reduce 5)) 4873 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43) (("}") reduce 43)) 4874 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72) (("}") reduce 72)) 4875 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) ((",") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_DOUBLE_ARROW) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81) (("}") reduce 81)) 4876 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47) (("}") reduce 47)) 4877 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31) (("}") reduce 31)) 4878 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57) (("}") reduce 57)) 4879 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61) (("}") reduce 61)) 4880 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28) (("}") reduce 28)) 4881 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36) (("}") reduce 36)) 4882 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32) (("}") reduce 32)) 4883 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35) (("}") reduce 35)) 4884 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19) (("}") reduce 19)) 4885 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45) (("}") reduce 45)) 4886 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492) (("}") reduce 492)) 4888 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71) (("}") reduce 71)) 4889 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82) (("}") reduce 82)) 4890 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519) (("}") reduce 519)) 4892 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456) (("}") reduce 456)) 4893 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281) (("}") reduce 281)) 4897 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282) (("}") reduce 282)) 4898 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283) (("}") reduce 283)) 4903 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) ((",") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_DOUBLE_ARROW) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377) (("}") reduce 377)) 4905 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) ((",") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_DOUBLE_ARROW) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375) (("}") reduce 375)) 4915 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 371) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 371) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 371)) 4916 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 373) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 373) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 373)) 4917 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 372) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 372) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 372)) 4918 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 365) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 365) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 365)) 4919 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 362) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 362) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 362)) 4920 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 370) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 370) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 370)) 4921 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 364) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 364) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 364)) 4922 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 368) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 368) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 368)) 4923 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 363) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 363) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 363)) 4924 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 366) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 366) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 366)) 4925 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 367) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 367) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 367)) 4926 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 374) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 374) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 374)) 4927 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 369) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 369) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 369)) 4929 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 359) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 359) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 359)) 4931 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515) (("}") reduce 515)) 4934 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511) (("}") reduce 511)) 4936 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510) (("}") reduce 510)) 4938 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505) (("}") reduce 505)) 4939 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512) (("}") reduce 512)) 4940 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) ((",") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_DOUBLE_ARROW) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360) (("}") reduce 360)) 4944 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519) (("}") reduce 519)) 4946 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456) (("}") reduce 456)) 4947 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281) (("}") reduce 281)) 4951 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282) (("}") reduce 282)) 4952 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283) (("}") reduce 283)) 4953 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454) (("}") reduce 454)) 4956 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518) (("}") reduce 518)) 4957 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455) (("}") reduce 455)) 4958 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457) (("}") reduce 457)) 4964 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507) (("}") reduce 507)) 4966 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506) (("}") reduce 506)) 4967 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530) (("}") reduce 530)) 4968 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513) (("}") reduce 513)) 4969 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532) (("}") reduce 532)) 4972 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531) (("}") reduce 531)) 4973 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508) (("}") reduce 508)) 4974 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514) (("}") reduce 514)) 4975 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509) (("}") reduce 509)) 4976 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517) (("}") reduce 517)) 4979 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516) (("}") reduce 516)) 4980 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454) (("}") reduce 454)) 4991 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444) (("}") reduce 444)) 4992 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439) (("}") reduce 439)) 5002 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 444)) 5003 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440) (("}") reduce 440)) 5005 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 431)) 5007 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491) (("}") reduce 491)) 5009 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518) (("}") reduce 518)) 5010 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455) (("}") reduce 455)) 5011 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457) (("}") reduce 457)) 5013 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) ((",") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435) (("}") reduce 435)) 5014 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437) (("}") reduce 437)) 5020 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507) (("}") reduce 507)) 5022 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506) (("}") reduce 506)) 5023 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530) (("}") reduce 530)) 5024 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513) (("}") reduce 513)) 5025 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532) (("}") reduce 532)) 5028 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531) (("}") reduce 531)) 5029 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508) (("}") reduce 508)) 5030 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514) (("}") reduce 514)) 5031 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509) (("}") reduce 509)) 5033 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) ((",") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427) (("}") reduce 427)) 5035 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471) (("}") reduce 471)) 5037 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 358) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 358) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 358)) 5038 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 432)) 5041 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 431)) 5042 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423) (("}") reduce 423)) 5043 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 433)) 5044 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) ((",") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419) (("}") reduce 419)) 5045 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) ((",") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436) (("}") reduce 436)) 5047 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477) (("}") reduce 477)) 5049 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) ((",") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478) (("}") reduce 478)) 5050 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476) (("}") reduce 476)) 5051 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 567) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 567)) 5052 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 566) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 566)) 5053 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 428)) 5054 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421) (("}") reduce 421)) 5057 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 5058 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 5059 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 5060 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 5061 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91) (("}") reduce 91)) 5062 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) ((",") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_DOUBLE_ARROW) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354) (("}") reduce 354)) 5064 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 5065 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468) (("}") reduce 468)) 5066 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 5067 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461) (("}") reduce 461)) 5068 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520) (("}") reduce 520)) 5075 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522) (("}") reduce 522)) 5077 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521) (("}") reduce 521)) 5078 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526) (("}") reduce 526)) 5079 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523) (("}") reduce 523)) 5080 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524) (("}") reduce 524)) 5082 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469) (("}") reduce 469)) 5083 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) ((",") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_DOUBLE_ARROW) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353) (("}") reduce 353)) 5084 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281) (("}") reduce 281)) 5088 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282) (("}") reduce 282)) 5089 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283) (("}") reduce 283)) 5091 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525) (("}") reduce 525)) 5092 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) ((",") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_DOUBLE_ARROW) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355) (("}") reduce 355)) 5099 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) ((",") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_DOUBLE_ARROW) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352) (("}") reduce 352)) 5101 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462) (("}") reduce 462)) 5107 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) ((",") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_DOUBLE_ARROW) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223) (("}") reduce 223)) 5112 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 357) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 357) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 357)) 5114 ((("%") shift) ((")") reduce 570) (("*") shift) (("+") shift) ((",") reduce 570) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5115 (((")") reduce 568) ((",") reduce 568)) 5117 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 124) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 5119 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 5120 (((")") reduce 569) ((",") reduce 569)) 5121 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) ((",") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417) (("}") reduce 417)) 5122 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 564) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 564)) 5123 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 563) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 563)) 5124 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) ((",") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_DOUBLE_ARROW) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376) (("}") reduce 376)) 5125 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 493) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 5126 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424) (("}") reduce 424)) 5127 ((("%") shift) ((")") reduce 494) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5129 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) ((",") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464) (("}") reduce 464)) 5132 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 5135 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562) (("}") reduce 562)) 5136 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) ((",") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418) (("}") reduce 418)) 5137 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) ((",") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_DOUBLE_ARROW) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378) (("}") reduce 378)) 5138 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) ((",") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_DOUBLE_ARROW) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361) (("}") reduce 361)) 5139 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422) (("}") reduce 422)) 5140 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) ((",") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420) (("}") reduce 420)) 5143 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470) (("}") reduce 470)) 5144 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425) (("}") reduce 425)) 5145 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398) (("}") reduce 398)) 5146 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) ((",") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397) (("}") reduce 397)) 5148 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) ((",") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411) (("}") reduce 411)) 5149 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517) (("}") reduce 517)) 5152 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516) (("}") reduce 516)) 5155 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_VARIABLE) shift)) 5156 ((("\"") reduce 549) ((T_CURLY_OPEN) reduce 549) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 549) ((T_ENCAPSED_AND_WHITESPACE) reduce 549) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_VARIABLE) reduce 549) (("[") shift)) 5157 ((("\"") shift) ((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_VARIABLE) shift)) 5158 ((("\"") reduce 547) ((T_CURLY_OPEN) reduce 547) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 547) ((T_ENCAPSED_AND_WHITESPACE) reduce 547) ((T_VARIABLE) reduce 547)) 5159 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473) (("}") reduce 473)) 5160 ((("\"") reduce 546) ((T_CURLY_OPEN) reduce 546) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 546) ((T_ENCAPSED_AND_WHITESPACE) reduce 546) ((T_VARIABLE) reduce 546)) 5161 ((("\"") reduce 545) ((T_CURLY_OPEN) reduce 545) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 545) ((T_ENCAPSED_AND_WHITESPACE) reduce 545) ((T_VARIABLE) reduce 545)) 5166 ((("\"") reduce 550) ((T_CURLY_OPEN) reduce 550) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 550) ((T_ENCAPSED_AND_WHITESPACE) reduce 550) ((T_VARIABLE) reduce 550)) 5167 ((("\"") reduce 551) ((T_CURLY_OPEN) reduce 551) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 551) ((T_ENCAPSED_AND_WHITESPACE) reduce 551) ((T_VARIABLE) reduce 551)) 5168 ((("\"") reduce 552) ((T_CURLY_OPEN) reduce 552) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 552) ((T_ENCAPSED_AND_WHITESPACE) reduce 552) ((T_VARIABLE) reduce 552)) 5169 ((("\"") reduce 548) ((T_CURLY_OPEN) reduce 548) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 548) ((T_ENCAPSED_AND_WHITESPACE) reduce 548) ((T_VARIABLE) reduce 548)) 5172 ((("\"") reduce 553) ((T_CURLY_OPEN) reduce 553) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 553) ((T_ENCAPSED_AND_WHITESPACE) reduce 553) ((T_VARIABLE) reduce 553)) 5174 ((("\"") reduce 554) ((T_CURLY_OPEN) reduce 554) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 554) ((T_ENCAPSED_AND_WHITESPACE) reduce 554) ((T_VARIABLE) reduce 554)) 5177 ((("\"") reduce 555) ((T_CURLY_OPEN) reduce 555) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 555) ((T_ENCAPSED_AND_WHITESPACE) reduce 555) ((T_VARIABLE) reduce 555)) 5179 ((("\"") reduce 556) ((T_CURLY_OPEN) reduce 556) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 556) ((T_ENCAPSED_AND_WHITESPACE) reduce 556) ((T_VARIABLE) reduce 556)) 5180 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399) (("}") reduce 399)) 5181 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423) (("}") reduce 423)) 5182 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 433)) 5183 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) ((",") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419) (("}") reduce 419)) 5184 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) ((",") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436) (("}") reduce 436)) 5186 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477) (("}") reduce 477)) 5188 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) ((",") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478) (("}") reduce 478)) 5189 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476) (("}") reduce 476)) 5190 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 567) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 567)) 5191 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 566) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 566)) 5192 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 428)) 5193 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421) (("}") reduce 421)) 5196 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 5197 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 5198 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 5199 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 5200 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91) (("}") reduce 91)) 5201 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) ((",") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354) (("}") reduce 354)) 5203 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 5204 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468) (("}") reduce 468)) 5205 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 5206 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461) (("}") reduce 461)) 5207 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520) (("}") reduce 520)) 5214 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522) (("}") reduce 522)) 5216 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521) (("}") reduce 521)) 5217 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526) (("}") reduce 526)) 5218 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523) (("}") reduce 523)) 5219 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524) (("}") reduce 524)) 5221 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469) (("}") reduce 469)) 5222 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) ((",") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353) (("}") reduce 353)) 5223 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281) (("}") reduce 281)) 5227 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282) (("}") reduce 282)) 5228 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283) (("}") reduce 283)) 5230 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525) (("}") reduce 525)) 5231 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) ((",") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355) (("}") reduce 355)) 5238 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) ((",") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352) (("}") reduce 352)) 5240 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462) (("}") reduce 462)) 5246 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) ((",") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223) (("}") reduce 223)) 5251 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 357) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 357)) 5255 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 5256 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) ((",") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417) (("}") reduce 417)) 5257 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 564) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 564)) 5258 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 563) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 563)) 5259 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) ((",") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376) (("}") reduce 376)) 5261 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424) (("}") reduce 424)) 5263 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) ((",") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464) (("}") reduce 464)) 5266 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 5269 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562) (("}") reduce 562)) 5270 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) ((",") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418) (("}") reduce 418)) 5271 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) ((",") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378) (("}") reduce 378)) 5272 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) ((",") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361) (("}") reduce 361)) 5273 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422) (("}") reduce 422)) 5274 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) ((",") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420) (("}") reduce 420)) 5277 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470) (("}") reduce 470)) 5278 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425) (("}") reduce 425)) 5279 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398) (("}") reduce 398)) 5280 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) ((",") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397) (("}") reduce 397)) 5282 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) ((",") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411) (("}") reduce 411)) 5283 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517) (("}") reduce 517)) 5286 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516) (("}") reduce 516)) 5288 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473) (("}") reduce 473)) 5289 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399) (("}") reduce 399)) 5290 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 231) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 231) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5301 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 5302 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 5312 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5313 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 5315 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 5317 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 5318 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 5319 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 5321 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) ((",") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 5322 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 5328 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 5330 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 5331 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 5332 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 5333 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 5336 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 5337 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 5338 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 5339 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 5341 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) ((",") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 5343 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 5345 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 358) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 358) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5346 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5347 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5349 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5350 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 5351 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5352 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) ((",") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 5353 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) ((",") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 5355 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 5357 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) ((",") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 5358 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 5359 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 567) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5360 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 566) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5361 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5362 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 5365 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 5366 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 5367 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 5368 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 5369 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 5370 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) ((",") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_DOUBLE_ARROW) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 5372 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 5373 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 5374 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 5375 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 5376 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 5383 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 5385 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 5386 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 5387 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 5388 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 5390 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 5391 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) ((",") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_DOUBLE_ARROW) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 5392 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 5396 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 5397 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 5399 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 5400 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) ((",") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_DOUBLE_ARROW) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 5407 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) ((",") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_DOUBLE_ARROW) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 5409 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 5415 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) ((",") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_DOUBLE_ARROW) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 5420 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 357) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 357) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5424 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 5425 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) ((",") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 5426 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 564) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5427 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 563) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5428 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) ((",") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_DOUBLE_ARROW) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 5430 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 5432 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) ((",") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 5435 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 5438 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 5439 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) ((",") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 5440 (((T_DOUBLE_ARROW) reduce 124)) 5443 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 229) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 229)) 5444 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) ((",") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_DOUBLE_ARROW) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 5445 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) ((",") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_DOUBLE_ARROW) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 5446 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 5447 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) ((",") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 5450 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 5451 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 5452 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 5453 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) ((",") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 5455 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) ((",") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 5456 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 5459 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 5461 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 5462 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 5467 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 357) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 357) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 357) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5471 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 5472 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) ((",") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) ((";") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 5473 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 564) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 564) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5474 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 563) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 563) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5475 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) ((",") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) ((";") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_DOUBLE_ARROW) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 5477 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 5479 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) ((",") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) ((";") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 5482 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 5485 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 5486 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) ((",") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) ((";") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 5487 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) ((",") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) ((";") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_DOUBLE_ARROW) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 5488 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) ((",") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) ((";") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_DOUBLE_ARROW) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 5489 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 5490 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) ((",") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 5493 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 5494 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 5495 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 5496 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) ((",") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) ((";") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 5498 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) ((",") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) ((";") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 5499 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 5502 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 5504 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 5505 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 5506 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 5507 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5508 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) ((",") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) ((";") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 5509 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) ((",") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 5511 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 5513 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) ((",") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) ((";") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 5514 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 5515 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 567) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 567) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5516 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 566) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 566) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5517 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 428) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5518 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 5521 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 5522 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 5523 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 5524 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 5525 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 5526 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) ((",") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) ((";") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 5528 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 5529 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 5530 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 5531 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 5532 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 5539 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 5541 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 5542 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 5543 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 5544 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 5546 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 5547 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) ((",") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) ((";") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 5548 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((";") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 5552 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((";") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 5553 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((";") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 5555 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 5556 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) ((",") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) ((";") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 5563 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) ((",") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) ((";") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 5565 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 5571 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) ((",") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) ((";") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 5576 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 357) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 357) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5580 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 5581 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) ((",") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) ((";") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 5582 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 564) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 564) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5583 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 563) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 563) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5584 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) ((",") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) ((";") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 5586 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 5588 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) ((",") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) ((";") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 5591 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 5594 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 5595 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) ((",") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) ((";") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 5596 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) ((",") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) ((";") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 5597 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) ((",") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) ((";") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 5598 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 5599 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) ((",") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 5602 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 5603 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 5604 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 5605 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) ((",") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) ((";") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 5607 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) ((",") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) ((";") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 5608 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 5611 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 5613 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 5614 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 5616 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_VARIABLE) reduce 278)) 5617 (((T_VARIABLE) reduce 272)) 5618 (((T_VARIABLE) reduce 273)) 5619 (((T_VARIABLE) reduce 93)) 5620 (((T_VARIABLE) reduce 92)) 5621 (((T_VARIABLE) reduce 94)) 5622 (((T_VARIABLE) reduce 91)) 5623 (((T_VARIABLE) reduce 274)) 5624 (((T_VARIABLE) reduce 269)) 5625 ((("?") reduce 329) ((T_ABSTRACT) reduce 329) ((T_ARRAY) reduce 329) ((T_CALLABLE) reduce 329) ((T_CONST) reduce 329) ((T_FINAL) reduce 329) ((T_FUNCTION) reduce 329) ((T_NAME_FULLY_QUALIFIED) reduce 329) ((T_NAME_QUALIFIED) reduce 329) ((T_NAME_RELATIVE) reduce 329) ((T_PRIVATE) reduce 329) ((T_PROTECTED) reduce 329) ((T_PUBLIC) reduce 329) ((T_READONLY) reduce 329) ((T_STATIC) reduce 329) ((T_STRING) reduce 329) ((T_VARIABLE) reduce 329)) 5626 (((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift)) 5627 (((T_ABSTRACT) reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_ARRAY) reduce 442) ((T_AS) reduce 442) ((T_BREAK) reduce 442) ((T_CALLABLE) reduce 442) ((T_CASE) reduce 442) ((T_CATCH) reduce 442) ((T_CLASS) reduce 442) ((T_CLASS_C) reduce 442) ((T_CLONE) reduce 442) ((T_CONST) reduce 442) ((T_CONTINUE) reduce 442) ((T_DECLARE) reduce 442) ((T_DEFAULT) reduce 442) ((T_DIR) reduce 442) ((T_DO) reduce 442) ((T_ECHO) reduce 442) ((T_ELSE) reduce 442) ((T_ELSEIF) reduce 442) ((T_EMPTY) reduce 442) ((T_ENDDECLARE) reduce 442) ((T_ENDFOR) reduce 442) ((T_ENDFOREACH) reduce 442) ((T_ENDIF) reduce 442) ((T_ENDSWITCH) reduce 442) ((T_ENDWHILE) reduce 442) ((T_ENUM) reduce 442) ((T_EVAL) reduce 442) ((T_EXIT) reduce 442) ((T_EXTENDS) reduce 442) ((T_FILE) reduce 442) ((T_FINAL) reduce 442) ((T_FINALLY) reduce 442) ((T_FN) reduce 442) ((T_FOR) reduce 442) ((T_FOREACH) reduce 442) ((T_FUNCTION) reduce 442) ((T_FUNC_C) reduce 442) ((T_GLOBAL) reduce 442) ((T_GOTO) reduce 442) ((T_IF) reduce 442) ((T_IMPLEMENTS) reduce 442) ((T_INCLUDE) reduce 442) ((T_INCLUDE_ONCE) reduce 442) ((T_INSTANCEOF) reduce 442) ((T_INSTEADOF) reduce 442) ((T_INTERFACE) reduce 442) ((T_ISSET) reduce 442) ((T_LINE) reduce 442) ((T_LIST) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_MATCH) reduce 442) ((T_METHOD_C) reduce 442) ((T_NAMESPACE) reduce 442) ((T_NEW) reduce 442) ((T_NS_C) reduce 442) ((T_PRINT) reduce 442) ((T_PRIVATE) reduce 442) ((T_PROTECTED) reduce 442) ((T_PUBLIC) reduce 442) ((T_READONLY) reduce 442) ((T_REQUIRE) reduce 442) ((T_REQUIRE_ONCE) reduce 442) ((T_RETURN) reduce 442) ((T_STATIC) reduce 442) ((T_STRING) reduce 442) ((T_SWITCH) reduce 442) ((T_THROW) reduce 442) ((T_TRAIT) reduce 442) ((T_TRAIT_C) reduce 442) ((T_TRY) reduce 442) ((T_UNSET) reduce 442) ((T_USE) reduce 442) ((T_VAR) reduce 442) ((T_WHILE) reduce 442) ((T_YIELD) reduce 442)) 5628 (((T_ABSTRACT) reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) reduce 446) ((T_AS) reduce 446) ((T_BREAK) reduce 446) ((T_CALLABLE) reduce 446) ((T_CASE) reduce 446) ((T_CATCH) reduce 446) ((T_CLASS) reduce 446) ((T_CLASS_C) reduce 446) ((T_CLONE) reduce 446) ((T_CONST) reduce 446) ((T_CONTINUE) reduce 446) ((T_DECLARE) reduce 446) ((T_DEFAULT) reduce 446) ((T_DIR) reduce 446) ((T_DO) reduce 446) ((T_ECHO) reduce 446) ((T_ELSE) reduce 446) ((T_ELSEIF) reduce 446) ((T_EMPTY) reduce 446) ((T_ENDDECLARE) reduce 446) ((T_ENDFOR) reduce 446) ((T_ENDFOREACH) reduce 446) ((T_ENDIF) reduce 446) ((T_ENDSWITCH) reduce 446) ((T_ENDWHILE) reduce 446) ((T_ENUM) reduce 446) ((T_EVAL) reduce 446) ((T_EXIT) reduce 446) ((T_EXTENDS) reduce 446) ((T_FILE) reduce 446) ((T_FINAL) reduce 446) ((T_FINALLY) reduce 446) ((T_FN) reduce 446) ((T_FOR) reduce 446) ((T_FOREACH) reduce 446) ((T_FUNCTION) reduce 446) ((T_FUNC_C) reduce 446) ((T_GLOBAL) reduce 446) ((T_GOTO) reduce 446) ((T_IF) reduce 446) ((T_IMPLEMENTS) reduce 446) ((T_INCLUDE) reduce 446) ((T_INCLUDE_ONCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_INSTEADOF) reduce 446) ((T_INTERFACE) reduce 446) ((T_ISSET) reduce 446) ((T_LINE) reduce 446) ((T_LIST) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_MATCH) reduce 446) ((T_METHOD_C) reduce 446) ((T_NAMESPACE) reduce 446) ((T_NEW) reduce 446) ((T_NS_C) reduce 446) ((T_PRINT) reduce 446) ((T_PRIVATE) reduce 446) ((T_PROTECTED) reduce 446) ((T_PUBLIC) reduce 446) ((T_READONLY) reduce 446) ((T_REQUIRE) reduce 446) ((T_REQUIRE_ONCE) reduce 446) ((T_RETURN) reduce 446) ((T_STATIC) reduce 446) ((T_STRING) reduce 446) ((T_SWITCH) reduce 446) ((T_THROW) reduce 446) ((T_TRAIT) reduce 446) ((T_TRAIT_C) reduce 446) ((T_TRY) reduce 446) ((T_UNSET) reduce 446) ((T_USE) reduce 446) ((T_VAR) reduce 446) ((T_WHILE) reduce 446) ((T_YIELD) reduce 446)) 5629 (((T_ABSTRACT) reduce 79) ((T_ARRAY) reduce 79) ((T_AS) reduce 79) ((T_BREAK) reduce 79) ((T_CALLABLE) reduce 79) ((T_CASE) reduce 79) ((T_CATCH) reduce 79) ((T_CLASS) reduce 79) ((T_CLASS_C) reduce 79) ((T_CLONE) reduce 79) ((T_CONST) reduce 79) ((T_CONTINUE) reduce 79) ((T_DECLARE) reduce 79) ((T_DEFAULT) reduce 79) ((T_DIR) reduce 79) ((T_DO) reduce 79) ((T_ECHO) reduce 79) ((T_ELSE) reduce 79) ((T_ELSEIF) reduce 79) ((T_EMPTY) reduce 79) ((T_ENDDECLARE) reduce 79) ((T_ENDFOR) reduce 79) ((T_ENDFOREACH) reduce 79) ((T_ENDIF) reduce 79) ((T_ENDSWITCH) reduce 79) ((T_ENDWHILE) reduce 79) ((T_ENUM) reduce 79) ((T_EVAL) reduce 79) ((T_EXIT) reduce 79) ((T_EXTENDS) reduce 79) ((T_FILE) reduce 79) ((T_FINAL) reduce 79) ((T_FINALLY) reduce 79) ((T_FN) reduce 79) ((T_FOR) reduce 79) ((T_FOREACH) reduce 79) ((T_FUNCTION) reduce 79) ((T_FUNC_C) reduce 79) ((T_GLOBAL) reduce 79) ((T_GOTO) reduce 79) ((T_IF) reduce 79) ((T_IMPLEMENTS) reduce 79) ((T_INCLUDE) reduce 79) ((T_INCLUDE_ONCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_INSTEADOF) reduce 79) ((T_INTERFACE) reduce 79) ((T_ISSET) reduce 79) ((T_LINE) reduce 79) ((T_LIST) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_MATCH) reduce 79) ((T_METHOD_C) reduce 79) ((T_NAMESPACE) reduce 79) ((T_NEW) reduce 79) ((T_NS_C) reduce 79) ((T_PRINT) reduce 79) ((T_PRIVATE) reduce 79) ((T_PROTECTED) reduce 79) ((T_PUBLIC) reduce 79) ((T_READONLY) reduce 79) ((T_REQUIRE) reduce 79) ((T_REQUIRE_ONCE) reduce 79) ((T_RETURN) reduce 79) ((T_STATIC) reduce 79) ((T_STRING) reduce 79) ((T_SWITCH) reduce 79) ((T_THROW) reduce 79) ((T_TRAIT) reduce 79) ((T_TRAIT_C) reduce 79) ((T_TRY) reduce 79) ((T_UNSET) reduce 79) ((T_USE) reduce 79) ((T_VAR) reduce 79) ((T_WHILE) reduce 79) ((T_YIELD) reduce 79)) 5630 (((T_ABSTRACT) reduce 80) ((T_ARRAY) reduce 80) ((T_AS) reduce 80) ((T_BREAK) reduce 80) ((T_CALLABLE) reduce 80) ((T_CASE) reduce 80) ((T_CATCH) reduce 80) ((T_CLASS) reduce 80) ((T_CLASS_C) reduce 80) ((T_CLONE) reduce 80) ((T_CONST) reduce 80) ((T_CONTINUE) reduce 80) ((T_DECLARE) reduce 80) ((T_DEFAULT) reduce 80) ((T_DIR) reduce 80) ((T_DO) reduce 80) ((T_ECHO) reduce 80) ((T_ELSE) reduce 80) ((T_ELSEIF) reduce 80) ((T_EMPTY) reduce 80) ((T_ENDDECLARE) reduce 80) ((T_ENDFOR) reduce 80) ((T_ENDFOREACH) reduce 80) ((T_ENDIF) reduce 80) ((T_ENDSWITCH) reduce 80) ((T_ENDWHILE) reduce 80) ((T_ENUM) reduce 80) ((T_EVAL) reduce 80) ((T_EXIT) reduce 80) ((T_EXTENDS) reduce 80) ((T_FILE) reduce 80) ((T_FINAL) reduce 80) ((T_FINALLY) reduce 80) ((T_FN) reduce 80) ((T_FOR) reduce 80) ((T_FOREACH) reduce 80) ((T_FUNCTION) reduce 80) ((T_FUNC_C) reduce 80) ((T_GLOBAL) reduce 80) ((T_GOTO) reduce 80) ((T_IF) reduce 80) ((T_IMPLEMENTS) reduce 80) ((T_INCLUDE) reduce 80) ((T_INCLUDE_ONCE) reduce 80) ((T_INSTANCEOF) reduce 80) ((T_INSTEADOF) reduce 80) ((T_INTERFACE) reduce 80) ((T_ISSET) reduce 80) ((T_LINE) reduce 80) ((T_LIST) reduce 80) ((T_LOGICAL_AND) reduce 80) ((T_LOGICAL_OR) reduce 80) ((T_LOGICAL_XOR) reduce 80) ((T_MATCH) reduce 80) ((T_METHOD_C) reduce 80) ((T_NAMESPACE) reduce 80) ((T_NEW) reduce 80) ((T_NS_C) reduce 80) ((T_PRINT) reduce 80) ((T_PRIVATE) reduce 80) ((T_PROTECTED) reduce 80) ((T_PUBLIC) reduce 80) ((T_READONLY) reduce 80) ((T_REQUIRE) reduce 80) ((T_REQUIRE_ONCE) reduce 80) ((T_RETURN) reduce 80) ((T_STATIC) reduce 80) ((T_STRING) reduce 80) ((T_SWITCH) reduce 80) ((T_THROW) reduce 80) ((T_TRAIT) reduce 80) ((T_TRAIT_C) reduce 80) ((T_TRY) reduce 80) ((T_UNSET) reduce 80) ((T_USE) reduce 80) ((T_VAR) reduce 80) ((T_WHILE) reduce 80) ((T_YIELD) reduce 80)) 5631 (((T_ABSTRACT) reduce 447) ((T_ARRAY) reduce 447) ((T_AS) reduce 447) ((T_BREAK) reduce 447) ((T_CALLABLE) reduce 447) ((T_CASE) reduce 447) ((T_CATCH) reduce 447) ((T_CLASS) reduce 447) ((T_CLASS_C) reduce 447) ((T_CLONE) reduce 447) ((T_CONST) reduce 447) ((T_CONTINUE) reduce 447) ((T_DECLARE) reduce 447) ((T_DEFAULT) reduce 447) ((T_DIR) reduce 447) ((T_DO) reduce 447) ((T_ECHO) reduce 447) ((T_ELSE) reduce 447) ((T_ELSEIF) reduce 447) ((T_EMPTY) reduce 447) ((T_ENDDECLARE) reduce 447) ((T_ENDFOR) reduce 447) ((T_ENDFOREACH) reduce 447) ((T_ENDIF) reduce 447) ((T_ENDSWITCH) reduce 447) ((T_ENDWHILE) reduce 447) ((T_ENUM) reduce 447) ((T_EVAL) reduce 447) ((T_EXIT) reduce 447) ((T_EXTENDS) reduce 447) ((T_FILE) reduce 447) ((T_FINAL) reduce 447) ((T_FINALLY) reduce 447) ((T_FN) reduce 447) ((T_FOR) reduce 447) ((T_FOREACH) reduce 447) ((T_FUNCTION) reduce 447) ((T_FUNC_C) reduce 447) ((T_GLOBAL) reduce 447) ((T_GOTO) reduce 447) ((T_IF) reduce 447) ((T_IMPLEMENTS) reduce 447) ((T_INCLUDE) reduce 447) ((T_INCLUDE_ONCE) reduce 447) ((T_INSTANCEOF) reduce 447) ((T_INSTEADOF) reduce 447) ((T_INTERFACE) reduce 447) ((T_ISSET) reduce 447) ((T_LINE) reduce 447) ((T_LIST) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_MATCH) reduce 447) ((T_METHOD_C) reduce 447) ((T_NAMESPACE) reduce 447) ((T_NEW) reduce 447) ((T_NS_C) reduce 447) ((T_PRINT) reduce 447) ((T_PRIVATE) reduce 447) ((T_PROTECTED) reduce 447) ((T_PUBLIC) reduce 447) ((T_READONLY) reduce 447) ((T_REQUIRE) reduce 447) ((T_REQUIRE_ONCE) reduce 447) ((T_RETURN) reduce 447) ((T_STATIC) reduce 447) ((T_STRING) reduce 447) ((T_SWITCH) reduce 447) ((T_THROW) reduce 447) ((T_TRAIT) reduce 447) ((T_TRAIT_C) reduce 447) ((T_TRY) reduce 447) ((T_UNSET) reduce 447) ((T_USE) reduce 447) ((T_VAR) reduce 447) ((T_WHILE) reduce 447) ((T_YIELD) reduce 447)) 5633 ((("(") reduce 73)) 5634 ((("(") reduce 52)) 5635 ((("(") reduce 27)) 5636 ((("(") reduce 51)) 5637 ((("(") reduce 53)) 5638 ((("(") reduce 49)) 5639 ((("(") reduce 29)) 5640 ((("(") reduce 59)) 5641 ((("(") reduce 60)) 5642 ((("(") reduce 11)) 5643 ((("(") reduce 42)) 5644 ((("(") reduce 39)) 5645 ((("(") reduce 25)) 5646 ((("(") reduce 50)) 5647 ((("(") reduce 66)) 5648 ((("(") reduce 18)) 5649 ((("(") reduce 17)) 5650 ((("(") reduce 15)) 5651 ((("(") reduce 14)) 5652 ((("(") reduce 38)) 5653 ((("(") reduce 26)) 5654 ((("(") reduce 22)) 5655 ((("(") reduce 24)) 5656 ((("(") reduce 16)) 5657 ((("(") reduce 48)) 5658 ((("(") reduce 20)) 5659 ((("(") reduce 70)) 5660 ((("(") reduce 3)) 5661 ((("(") reduce 12)) 5662 ((("(") reduce 54)) 5663 ((("(") reduce 65)) 5664 ((("(") reduce 74)) 5665 ((("(") reduce 30)) 5666 ((("(") reduce 68)) 5667 ((("(") reduce 21)) 5668 ((("(") reduce 23)) 5669 ((("(") reduce 41)) 5670 ((("(") reduce 62)) 5671 ((("(") reduce 34)) 5672 ((("(") reduce 40)) 5673 ((("(") reduce 13)) 5674 ((("(") reduce 55)) 5675 ((("(") reduce 1)) 5676 ((("(") reduce 2)) 5677 ((("(") reduce 9)) 5678 ((("(") reduce 33)) 5679 ((("(") reduce 58)) 5680 ((("(") reduce 37)) 5681 ((("(") reduce 64)) 5682 ((("(") reduce 46)) 5683 ((("(") reduce 8)) 5684 ((("(") reduce 6)) 5685 ((("(") reduce 7)) 5686 ((("(") reduce 69)) 5687 ((("(") reduce 63)) 5688 ((("(") reduce 56)) 5689 ((("(") reduce 10)) 5690 ((("(") reduce 67)) 5691 ((("(") reduce 44)) 5692 ((("(") reduce 75)) 5693 ((("(") reduce 76)) 5694 ((("(") reduce 77)) 5695 ((("(") reduce 78)) 5696 ((("(") reduce 4)) 5697 ((("(") reduce 5)) 5698 ((("(") reduce 43)) 5699 ((("(") reduce 72)) 5700 ((("(") reduce 81)) 5701 ((("(") reduce 47)) 5702 ((("(") reduce 31)) 5703 ((("(") reduce 57)) 5704 ((("(") reduce 61)) 5705 ((("(") reduce 28)) 5706 ((("(") reduce 36)) 5707 ((("(") reduce 32)) 5708 ((("(") reduce 35)) 5709 ((("(") reduce 19)) 5710 ((("(") reduce 45)) 5712 ((("(") reduce 71)) 5713 ((("(") reduce 82)) 5717 (((":") shift) ((";") reduce 279) (("{") reduce 279)) 5719 (((";") reduce 444) (("{") reduce 444)) 5720 (((";") shift) (("{") shift)) 5721 (((T_ABSTRACT) reduce 322) ((T_ATTRIBUTE) reduce 322) ((T_CASE) reduce 322) ((T_CONST) reduce 322) ((T_FINAL) reduce 322) ((T_FUNCTION) reduce 322) ((T_PRIVATE) reduce 322) ((T_PROTECTED) reduce 322) ((T_PUBLIC) reduce 322) ((T_READONLY) reduce 322) ((T_STATIC) reduce 322) ((T_USE) reduce 322) ((T_VAR) reduce 322) (("}") reduce 322)) 5722 (((T_ABSTRACT) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_CASE) reduce 444) ((T_CONST) reduce 444) ((T_FINAL) reduce 444) ((T_FUNCTION) reduce 444) ((T_PRIVATE) reduce 444) ((T_PROTECTED) reduce 444) ((T_PUBLIC) reduce 444) ((T_READONLY) reduce 444) ((T_STATIC) reduce 444) ((T_USE) reduce 444) ((T_VAR) reduce 444) (("}") reduce 444)) 5725 (((T_ABSTRACT) reduce 323) ((T_ATTRIBUTE) reduce 323) ((T_CASE) reduce 323) ((T_CONST) reduce 323) ((T_FINAL) reduce 323) ((T_FUNCTION) reduce 323) ((T_PRIVATE) reduce 323) ((T_PROTECTED) reduce 323) ((T_PUBLIC) reduce 323) ((T_READONLY) reduce 323) ((T_STATIC) reduce 323) ((T_USE) reduce 323) ((T_VAR) reduce 323) (("}") reduce 323)) 5726 (((T_ABSTRACT) reduce 300) ((T_ATTRIBUTE) reduce 300) ((T_CASE) reduce 300) ((T_CONST) reduce 300) ((T_FINAL) reduce 300) ((T_FUNCTION) reduce 300) ((T_PRIVATE) reduce 300) ((T_PROTECTED) reduce 300) ((T_PUBLIC) reduce 300) ((T_READONLY) reduce 300) ((T_STATIC) reduce 300) ((T_USE) reduce 300) ((T_VAR) reduce 300) (("}") reduce 300)) 5728 (((";") reduce 272) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) (("{") reduce 272) (("|") reduce 272)) 5729 (((";") reduce 273) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) (("{") reduce 273) (("|") reduce 273)) 5730 (((";") reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) (("{") reduce 93) (("|") reduce 93)) 5731 (((";") reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) (("{") reduce 92) (("|") reduce 92)) 5732 (((";") reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) (("{") reduce 94) (("|") reduce 94)) 5733 (((";") reduce 263) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 263) (("{") reduce 263) (("|") reduce 263)) 5734 (((";") reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) (("{") reduce 91) (("|") reduce 91)) 5735 (((";") reduce 261) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) (("{") reduce 261)) 5736 (((";") reduce 274) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) (("{") reduce 274) (("|") reduce 274)) 5737 (((";") reduce 258) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) (("{") reduce 258) (("|") shift)) 5738 (((";") reduce 280) (("{") reduce 280)) 5739 (((";") reduce 262) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 262) (("{") reduce 262) (("|") reduce 262)) 5740 (((";") reduce 260) (("{") reduce 260) (("|") shift)) 5742 (((";") reduce 272) (("{") reduce 272) (("|") reduce 272)) 5743 (((";") reduce 273) (("{") reduce 273) (("|") reduce 273)) 5744 (((";") reduce 93) (("{") reduce 93) (("|") reduce 93)) 5745 (((";") reduce 92) (("{") reduce 92) (("|") reduce 92)) 5746 (((";") reduce 94) (("{") reduce 94) (("|") reduce 94)) 5747 (((";") reduce 263) (("{") reduce 263) (("|") reduce 263)) 5748 (((";") reduce 91) (("{") reduce 91) (("|") reduce 91)) 5749 (((";") reduce 274) (("{") reduce 274) (("|") reduce 274)) 5750 (((";") reduce 265) (("{") reduce 265) (("|") reduce 265)) 5751 (((";") reduce 262) (("{") reduce 262) (("|") reduce 262)) 5754 (((";") reduce 264) (("{") reduce 264) (("|") reduce 264)) 5755 (((";") reduce 272) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) (("{") reduce 272)) 5756 (((";") reduce 273) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) (("{") reduce 273)) 5757 (((";") reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) (("{") reduce 93)) 5758 (((";") reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) (("{") reduce 92)) 5759 (((";") reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) (("{") reduce 94)) 5760 (((";") reduce 263) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 263) (("{") reduce 263)) 5761 (((";") reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) (("{") reduce 91)) 5762 (((";") reduce 274) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) (("{") reduce 274)) 5763 (((";") reduce 266) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) (("{") reduce 266)) 5764 (((";") reduce 262) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 262) (("{") reduce 262)) 5766 (((";") reduce 267) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) (("{") reduce 267)) 5767 (((";") reduce 272) (("{") reduce 272)) 5768 (((";") reduce 273) (("{") reduce 273)) 5769 (((";") reduce 93) (("{") reduce 93)) 5770 (((";") reduce 92) (("{") reduce 92)) 5771 (((";") reduce 94) (("{") reduce 94)) 5772 (((";") reduce 263) (("{") reduce 263)) 5773 (((";") reduce 91) (("{") reduce 91)) 5774 (((";") reduce 274) (("{") reduce 274)) 5775 (((";") reduce 259) (("{") reduce 259)) 5776 (((";") reduce 262) (("{") reduce 262)) 5777 ((("=") reduce 73)) 5778 ((("=") reduce 52)) 5779 ((("=") reduce 27)) 5780 ((("=") reduce 51)) 5781 ((("=") reduce 53)) 5782 ((("=") reduce 49)) 5783 ((("=") reduce 29)) 5784 ((("=") reduce 59)) 5785 ((("=") reduce 60)) 5786 ((("=") reduce 11)) 5787 ((("=") reduce 42)) 5788 ((("=") reduce 39)) 5789 ((("=") reduce 25)) 5790 ((("=") reduce 50)) 5791 ((("=") reduce 66)) 5792 ((("=") reduce 18)) 5793 ((("=") reduce 17)) 5794 ((("=") reduce 15)) 5795 ((("=") reduce 14)) 5796 ((("=") reduce 38)) 5797 ((("=") reduce 26)) 5798 ((("=") reduce 22)) 5799 ((("=") reduce 24)) 5800 ((("=") reduce 16)) 5801 ((("=") reduce 48)) 5802 ((("=") reduce 20)) 5803 ((("=") reduce 70)) 5804 ((("=") reduce 3)) 5805 ((("=") reduce 12)) 5806 ((("=") reduce 54)) 5807 ((("=") reduce 65)) 5808 ((("=") reduce 74)) 5809 ((("=") reduce 30)) 5810 ((("=") reduce 68)) 5811 ((("=") reduce 21)) 5812 ((("=") reduce 23)) 5813 ((("=") reduce 41)) 5814 ((("=") reduce 62)) 5815 ((("=") reduce 34)) 5816 ((("=") reduce 40)) 5817 ((("=") reduce 13)) 5818 ((("=") reduce 55)) 5819 ((("=") reduce 1)) 5820 ((("=") reduce 2)) 5821 ((("=") reduce 9)) 5822 ((("=") reduce 33)) 5823 ((("=") reduce 58)) 5824 ((("=") reduce 37)) 5825 ((("=") reduce 64)) 5826 ((("=") reduce 46)) 5827 ((("=") reduce 8)) 5828 ((("=") reduce 6)) 5829 ((("=") reduce 7)) 5830 ((("=") reduce 69)) 5831 ((("=") reduce 63)) 5832 ((("=") reduce 56)) 5833 ((("=") reduce 10)) 5834 ((("=") reduce 67)) 5835 ((("=") reduce 44)) 5836 ((("=") reduce 75)) 5837 ((("=") reduce 76)) 5838 ((("=") reduce 77)) 5839 ((("=") reduce 78)) 5840 ((("=") reduce 4)) 5841 ((("=") reduce 5)) 5842 ((("=") reduce 43)) 5843 ((("=") reduce 72)) 5844 ((("=") reduce 81)) 5845 ((("=") reduce 47)) 5846 ((("=") reduce 31)) 5847 ((("=") reduce 57)) 5848 ((("=") reduce 61)) 5849 ((("=") reduce 28)) 5850 ((("=") reduce 36)) 5851 ((("=") reduce 32)) 5852 ((("=") reduce 35)) 5853 ((("=") reduce 19)) 5854 ((("=") reduce 45)) 5855 (((",") reduce 342) ((";") reduce 342)) 5858 ((("=") reduce 71)) 5859 ((("=") reduce 82)) 5862 (((",") reduce 343) ((";") reduce 343)) 5864 (((T_ABSTRACT) reduce 299) ((T_ATTRIBUTE) reduce 299) ((T_CASE) reduce 299) ((T_CONST) reduce 299) ((T_FINAL) reduce 299) ((T_FUNCTION) reduce 299) ((T_PRIVATE) reduce 299) ((T_PROTECTED) reduce 299) ((T_PUBLIC) reduce 299) ((T_READONLY) reduce 299) ((T_STATIC) reduce 299) ((T_USE) reduce 299) ((T_VAR) reduce 299) (("}") reduce 299)) 5865 (((",") reduce 341) ((";") reduce 341)) 5866 (((T_ABSTRACT) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_CASE) reduce 101) ((T_CONST) reduce 101) ((T_FINAL) reduce 101) ((T_FUNCTION) reduce 101) ((T_PRIVATE) reduce 101) ((T_PROTECTED) reduce 101) ((T_PUBLIC) reduce 101) ((T_READONLY) reduce 101) ((T_STATIC) reduce 101) ((T_VAR) reduce 101)) 5867 (((T_ABSTRACT) reduce 303) ((T_ATTRIBUTE) reduce 303) ((T_CASE) reduce 303) ((T_CONST) reduce 303) ((T_FINAL) reduce 303) ((T_FUNCTION) reduce 303) ((T_PRIVATE) reduce 303) ((T_PROTECTED) reduce 303) ((T_PUBLIC) reduce 303) ((T_READONLY) reduce 303) ((T_STATIC) reduce 303) ((T_USE) reduce 303) ((T_VAR) reduce 303) (("}") reduce 303)) 5868 (((",") reduce 93) ((";") reduce 93) (("{") reduce 93)) 5869 (((",") reduce 92) ((";") reduce 92) (("{") reduce 92)) 5870 (((",") reduce 94) ((";") reduce 94) (("{") reduce 94)) 5871 (((",") reduce 458) ((";") reduce 458) (("{") reduce 458)) 5872 (((",") reduce 91) ((";") reduce 91) (("{") reduce 91)) 5873 (((",") reduce 305) ((";") reduce 305) (("{") reduce 305)) 5874 (((",") shift) ((";") shift) (("{") shift)) 5875 (((",") reduce 459) ((";") reduce 459) (("{") reduce 459)) 5877 (((T_ABSTRACT) reduce 307) ((T_ATTRIBUTE) reduce 307) ((T_CASE) reduce 307) ((T_CONST) reduce 307) ((T_FINAL) reduce 307) ((T_FUNCTION) reduce 307) ((T_PRIVATE) reduce 307) ((T_PROTECTED) reduce 307) ((T_PUBLIC) reduce 307) ((T_READONLY) reduce 307) ((T_STATIC) reduce 307) ((T_USE) reduce 307) ((T_VAR) reduce 307) (("}") reduce 307)) 5878 (((T_ABSTRACT) reduce 304) ((T_ATTRIBUTE) reduce 304) ((T_CASE) reduce 304) ((T_CONST) reduce 304) ((T_FINAL) reduce 304) ((T_FUNCTION) reduce 304) ((T_PRIVATE) reduce 304) ((T_PROTECTED) reduce 304) ((T_PUBLIC) reduce 304) ((T_READONLY) reduce 304) ((T_STATIC) reduce 304) ((T_USE) reduce 304) ((T_VAR) reduce 304) (("}") reduce 304)) 5879 (((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift) (("}") shift)) 5880 (((T_AS) reduce 73)) 5881 (((T_AS) reduce 52)) 5882 (((T_AS) reduce 27)) 5883 (((T_AS) reduce 51)) 5884 (((T_AS) reduce 53)) 5885 (((T_AS) reduce 49)) 5886 (((T_AS) reduce 29)) 5887 (((T_AS) reduce 59)) 5888 (((T_AS) reduce 60)) 5889 (((T_AS) reduce 11)) 5890 (((T_AS) reduce 42)) 5891 (((T_AS) reduce 39)) 5892 (((T_AS) reduce 25)) 5893 (((T_AS) reduce 50)) 5894 (((T_AS) reduce 66)) 5895 (((T_AS) reduce 18)) 5896 (((T_AS) reduce 17)) 5897 (((T_AS) reduce 15)) 5898 (((T_AS) reduce 14)) 5899 (((T_AS) reduce 38)) 5900 (((T_AS) reduce 26)) 5901 (((T_AS) reduce 22)) 5902 (((T_AS) reduce 24)) 5903 (((T_AS) reduce 16)) 5904 (((T_AS) reduce 48)) 5905 (((T_AS) reduce 20)) 5906 (((T_AS) reduce 70)) 5907 (((T_AS) reduce 3)) 5908 (((T_AS) reduce 12)) 5909 (((T_AS) reduce 54)) 5910 (((T_AS) reduce 65)) 5911 (((T_AS) reduce 74)) 5912 (((T_AS) reduce 30)) 5913 (((T_AS) reduce 68)) 5914 (((T_AS) reduce 21)) 5915 (((T_AS) reduce 23)) 5916 (((T_AS) reduce 41)) 5917 (((T_AS) reduce 62)) 5918 (((T_AS) reduce 34)) 5919 (((T_AS) reduce 40)) 5920 (((T_AS) reduce 13)) 5921 (((T_AS) reduce 55)) 5922 (((T_AS) reduce 1)) 5923 (((T_AS) reduce 2)) 5924 (((T_AS) reduce 9)) 5925 (((T_AS) reduce 33)) 5926 (((T_AS) reduce 58)) 5927 (((T_AS) reduce 37)) 5928 (((T_AS) reduce 64)) 5929 (((T_AS) reduce 46)) 5930 (((T_AS) reduce 8)) 5931 (((T_AS) reduce 6)) 5932 (((T_AS) reduce 7)) 5933 (((T_AS) reduce 69)) 5934 (((T_AS) reduce 63)) 5935 (((T_AS) reduce 56)) 5936 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 93)) 5937 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 92)) 5938 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 94)) 5939 (((T_AS) reduce 10)) 5940 (((T_AS) reduce 67)) 5941 (((T_AS) reduce 44)) 5942 (((T_AS) reduce 75)) 5943 (((T_AS) reduce 76)) 5944 (((T_AS) reduce 77)) 5945 (((T_AS) reduce 78)) 5946 (((T_AS) reduce 4)) 5947 (((T_AS) reduce 5)) 5948 (((T_AS) reduce 43)) 5949 (((T_AS) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458)) 5950 (((T_AS) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91)) 5951 (((T_AS) reduce 47)) 5952 (((T_AS) reduce 31)) 5953 (((T_AS) reduce 57)) 5954 (((T_AS) reduce 61)) 5955 (((T_AS) reduce 28)) 5956 (((T_AS) reduce 36)) 5957 (((T_AS) reduce 32)) 5958 (((T_AS) reduce 35)) 5959 (((T_AS) reduce 19)) 5960 (((T_AS) reduce 45)) 5961 (((T_AS) reduce 320) ((T_INSTEADOF) shift)) 5963 (((T_AS) reduce 319)) 5964 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 459)) 5965 (((T_AS) reduce 71)) 5966 (((T_AS) reduce 82)) 5967 (((T_ABSTRACT) reduce 310) ((T_ARRAY) reduce 310) ((T_AS) reduce 310) ((T_BREAK) reduce 310) ((T_CALLABLE) reduce 310) ((T_CASE) reduce 310) ((T_CATCH) reduce 310) ((T_CLASS) reduce 310) ((T_CLASS_C) reduce 310) ((T_CLONE) reduce 310) ((T_CONST) reduce 310) ((T_CONTINUE) reduce 310) ((T_DECLARE) reduce 310) ((T_DEFAULT) reduce 310) ((T_DIR) reduce 310) ((T_DO) reduce 310) ((T_ECHO) reduce 310) ((T_ELSE) reduce 310) ((T_ELSEIF) reduce 310) ((T_EMPTY) reduce 310) ((T_ENDDECLARE) reduce 310) ((T_ENDFOR) reduce 310) ((T_ENDFOREACH) reduce 310) ((T_ENDIF) reduce 310) ((T_ENDSWITCH) reduce 310) ((T_ENDWHILE) reduce 310) ((T_ENUM) reduce 310) ((T_EVAL) reduce 310) ((T_EXIT) reduce 310) ((T_EXTENDS) reduce 310) ((T_FILE) reduce 310) ((T_FINAL) reduce 310) ((T_FINALLY) reduce 310) ((T_FN) reduce 310) ((T_FOR) reduce 310) ((T_FOREACH) reduce 310) ((T_FUNCTION) reduce 310) ((T_FUNC_C) reduce 310) ((T_GLOBAL) reduce 310) ((T_GOTO) reduce 310) ((T_IF) reduce 310) ((T_IMPLEMENTS) reduce 310) ((T_INCLUDE) reduce 310) ((T_INCLUDE_ONCE) reduce 310) ((T_INSTANCEOF) reduce 310) ((T_INSTEADOF) reduce 310) ((T_INTERFACE) reduce 310) ((T_ISSET) reduce 310) ((T_LINE) reduce 310) ((T_LIST) reduce 310) ((T_LOGICAL_AND) reduce 310) ((T_LOGICAL_OR) reduce 310) ((T_LOGICAL_XOR) reduce 310) ((T_MATCH) reduce 310) ((T_METHOD_C) reduce 310) ((T_NAMESPACE) reduce 310) ((T_NAME_FULLY_QUALIFIED) reduce 310) ((T_NAME_QUALIFIED) reduce 310) ((T_NAME_RELATIVE) reduce 310) ((T_NEW) reduce 310) ((T_NS_C) reduce 310) ((T_PRINT) reduce 310) ((T_PRIVATE) reduce 310) ((T_PROTECTED) reduce 310) ((T_PUBLIC) reduce 310) ((T_READONLY) reduce 310) ((T_REQUIRE) reduce 310) ((T_REQUIRE_ONCE) reduce 310) ((T_RETURN) reduce 310) ((T_STATIC) reduce 310) ((T_STRING) reduce 310) ((T_SWITCH) reduce 310) ((T_THROW) reduce 310) ((T_TRAIT) reduce 310) ((T_TRAIT_C) reduce 310) ((T_TRY) reduce 310) ((T_UNSET) reduce 310) ((T_USE) reduce 310) ((T_VAR) reduce 310) ((T_WHILE) reduce 310) ((T_YIELD) reduce 310) (("}") reduce 310)) 5969 (((";") shift)) 5970 (((T_AS) shift)) 5972 (((T_ABSTRACT) reduce 308) ((T_ATTRIBUTE) reduce 308) ((T_CASE) reduce 308) ((T_CONST) reduce 308) ((T_FINAL) reduce 308) ((T_FUNCTION) reduce 308) ((T_PRIVATE) reduce 308) ((T_PROTECTED) reduce 308) ((T_PUBLIC) reduce 308) ((T_READONLY) reduce 308) ((T_STATIC) reduce 308) ((T_USE) reduce 308) ((T_VAR) reduce 308) (("}") reduce 308)) 5973 (((T_ABSTRACT) reduce 312) ((T_ARRAY) reduce 312) ((T_AS) reduce 312) ((T_BREAK) reduce 312) ((T_CALLABLE) reduce 312) ((T_CASE) reduce 312) ((T_CATCH) reduce 312) ((T_CLASS) reduce 312) ((T_CLASS_C) reduce 312) ((T_CLONE) reduce 312) ((T_CONST) reduce 312) ((T_CONTINUE) reduce 312) ((T_DECLARE) reduce 312) ((T_DEFAULT) reduce 312) ((T_DIR) reduce 312) ((T_DO) reduce 312) ((T_ECHO) reduce 312) ((T_ELSE) reduce 312) ((T_ELSEIF) reduce 312) ((T_EMPTY) reduce 312) ((T_ENDDECLARE) reduce 312) ((T_ENDFOR) reduce 312) ((T_ENDFOREACH) reduce 312) ((T_ENDIF) reduce 312) ((T_ENDSWITCH) reduce 312) ((T_ENDWHILE) reduce 312) ((T_ENUM) reduce 312) ((T_EVAL) reduce 312) ((T_EXIT) reduce 312) ((T_EXTENDS) reduce 312) ((T_FILE) reduce 312) ((T_FINAL) reduce 312) ((T_FINALLY) reduce 312) ((T_FN) reduce 312) ((T_FOR) reduce 312) ((T_FOREACH) reduce 312) ((T_FUNCTION) reduce 312) ((T_FUNC_C) reduce 312) ((T_GLOBAL) reduce 312) ((T_GOTO) reduce 312) ((T_IF) reduce 312) ((T_IMPLEMENTS) reduce 312) ((T_INCLUDE) reduce 312) ((T_INCLUDE_ONCE) reduce 312) ((T_INSTANCEOF) reduce 312) ((T_INSTEADOF) reduce 312) ((T_INTERFACE) reduce 312) ((T_ISSET) reduce 312) ((T_LINE) reduce 312) ((T_LIST) reduce 312) ((T_LOGICAL_AND) reduce 312) ((T_LOGICAL_OR) reduce 312) ((T_LOGICAL_XOR) reduce 312) ((T_MATCH) reduce 312) ((T_METHOD_C) reduce 312) ((T_NAMESPACE) reduce 312) ((T_NAME_FULLY_QUALIFIED) reduce 312) ((T_NAME_QUALIFIED) reduce 312) ((T_NAME_RELATIVE) reduce 312) ((T_NEW) reduce 312) ((T_NS_C) reduce 312) ((T_PRINT) reduce 312) ((T_PRIVATE) reduce 312) ((T_PROTECTED) reduce 312) ((T_PUBLIC) reduce 312) ((T_READONLY) reduce 312) ((T_REQUIRE) reduce 312) ((T_REQUIRE_ONCE) reduce 312) ((T_RETURN) reduce 312) ((T_STATIC) reduce 312) ((T_STRING) reduce 312) ((T_SWITCH) reduce 312) ((T_THROW) reduce 312) ((T_TRAIT) reduce 312) ((T_TRAIT_C) reduce 312) ((T_TRY) reduce 312) ((T_UNSET) reduce 312) ((T_USE) reduce 312) ((T_VAR) reduce 312) ((T_WHILE) reduce 312) ((T_YIELD) reduce 312) (("}") reduce 312)) 5975 (((";") reduce 334) ((T_ABSTRACT) reduce 334) ((T_ARRAY) reduce 334) ((T_AS) reduce 334) ((T_BREAK) reduce 334) ((T_CALLABLE) reduce 334) ((T_CASE) reduce 334) ((T_CATCH) reduce 334) ((T_CLASS) reduce 334) ((T_CLASS_C) reduce 334) ((T_CLONE) reduce 334) ((T_CONST) reduce 334) ((T_CONTINUE) reduce 334) ((T_DECLARE) reduce 334) ((T_DEFAULT) reduce 334) ((T_DIR) reduce 334) ((T_DO) reduce 334) ((T_ECHO) reduce 334) ((T_ELSE) reduce 334) ((T_ELSEIF) reduce 334) ((T_EMPTY) reduce 334) ((T_ENDDECLARE) reduce 334) ((T_ENDFOR) reduce 334) ((T_ENDFOREACH) reduce 334) ((T_ENDIF) reduce 334) ((T_ENDSWITCH) reduce 334) ((T_ENDWHILE) reduce 334) ((T_ENUM) reduce 334) ((T_EVAL) reduce 334) ((T_EXIT) reduce 334) ((T_EXTENDS) reduce 334) ((T_FILE) reduce 334) ((T_FINAL) reduce 334) ((T_FINALLY) reduce 334) ((T_FN) reduce 334) ((T_FOR) reduce 334) ((T_FOREACH) reduce 334) ((T_FUNCTION) reduce 334) ((T_FUNC_C) reduce 334) ((T_GLOBAL) reduce 334) ((T_GOTO) reduce 334) ((T_IF) reduce 334) ((T_IMPLEMENTS) reduce 334) ((T_INCLUDE) reduce 334) ((T_INCLUDE_ONCE) reduce 334) ((T_INSTANCEOF) reduce 334) ((T_INSTEADOF) reduce 334) ((T_INTERFACE) reduce 334) ((T_ISSET) reduce 334) ((T_LINE) reduce 334) ((T_LIST) reduce 334) ((T_LOGICAL_AND) reduce 334) ((T_LOGICAL_OR) reduce 334) ((T_LOGICAL_XOR) reduce 334) ((T_MATCH) reduce 334) ((T_METHOD_C) reduce 334) ((T_NAMESPACE) reduce 334) ((T_NEW) reduce 334) ((T_NS_C) reduce 334) ((T_PRINT) reduce 334) ((T_PRIVATE) reduce 334) ((T_PROTECTED) reduce 334) ((T_PUBLIC) reduce 334) ((T_READONLY) reduce 334) ((T_REQUIRE) reduce 334) ((T_REQUIRE_ONCE) reduce 334) ((T_RETURN) reduce 334) ((T_STATIC) reduce 334) ((T_STRING) reduce 334) ((T_SWITCH) reduce 334) ((T_THROW) reduce 334) ((T_TRAIT) reduce 334) ((T_TRAIT_C) reduce 334) ((T_TRY) reduce 334) ((T_UNSET) reduce 334) ((T_USE) reduce 334) ((T_VAR) reduce 334) ((T_WHILE) reduce 334) ((T_YIELD) reduce 334)) 5976 (((";") reduce 52)) 5977 (((";") reduce 27)) 5978 (((";") reduce 51)) 5979 (((";") reduce 53)) 5980 (((";") reduce 49)) 5981 (((";") reduce 29)) 5982 (((";") reduce 59)) 5983 (((";") reduce 60)) 5984 (((";") reduce 11)) 5985 (((";") reduce 42)) 5986 (((";") reduce 39)) 5987 (((";") reduce 25)) 5988 (((";") reduce 50)) 5989 (((";") reduce 66)) 5990 (((";") reduce 18)) 5991 (((";") reduce 17)) 5992 (((";") reduce 15)) 5993 (((";") reduce 14)) 5994 (((";") reduce 38)) 5995 (((";") reduce 26)) 5996 (((";") reduce 22)) 5997 (((";") reduce 24)) 5998 (((";") reduce 16)) 5999 (((";") reduce 48)) 6000 (((";") reduce 20)) 6001 (((";") reduce 70)) 6002 (((";") reduce 3)) 6003 (((";") reduce 12)) 6004 (((";") reduce 54)) 6005 (((";") reduce 65)) 6006 (((";") reduce 335) ((T_ABSTRACT) reduce 335) ((T_ARRAY) reduce 335) ((T_AS) reduce 335) ((T_BREAK) reduce 335) ((T_CALLABLE) reduce 335) ((T_CASE) reduce 335) ((T_CATCH) reduce 335) ((T_CLASS) reduce 335) ((T_CLASS_C) reduce 335) ((T_CLONE) reduce 335) ((T_CONST) reduce 335) ((T_CONTINUE) reduce 335) ((T_DECLARE) reduce 335) ((T_DEFAULT) reduce 335) ((T_DIR) reduce 335) ((T_DO) reduce 335) ((T_ECHO) reduce 335) ((T_ELSE) reduce 335) ((T_ELSEIF) reduce 335) ((T_EMPTY) reduce 335) ((T_ENDDECLARE) reduce 335) ((T_ENDFOR) reduce 335) ((T_ENDFOREACH) reduce 335) ((T_ENDIF) reduce 335) ((T_ENDSWITCH) reduce 335) ((T_ENDWHILE) reduce 335) ((T_ENUM) reduce 335) ((T_EVAL) reduce 335) ((T_EXIT) reduce 335) ((T_EXTENDS) reduce 335) ((T_FILE) reduce 335) ((T_FINAL) reduce 335) ((T_FINALLY) reduce 335) ((T_FN) reduce 335) ((T_FOR) reduce 335) ((T_FOREACH) reduce 335) ((T_FUNCTION) reduce 335) ((T_FUNC_C) reduce 335) ((T_GLOBAL) reduce 335) ((T_GOTO) reduce 335) ((T_IF) reduce 335) ((T_IMPLEMENTS) reduce 335) ((T_INCLUDE) reduce 335) ((T_INCLUDE_ONCE) reduce 335) ((T_INSTANCEOF) reduce 335) ((T_INSTEADOF) reduce 335) ((T_INTERFACE) reduce 335) ((T_ISSET) reduce 335) ((T_LINE) reduce 335) ((T_LIST) reduce 335) ((T_LOGICAL_AND) reduce 335) ((T_LOGICAL_OR) reduce 335) ((T_LOGICAL_XOR) reduce 335) ((T_MATCH) reduce 335) ((T_METHOD_C) reduce 335) ((T_NAMESPACE) reduce 335) ((T_NEW) reduce 335) ((T_NS_C) reduce 335) ((T_PRINT) reduce 335) ((T_PRIVATE) reduce 335) ((T_PROTECTED) reduce 335) ((T_PUBLIC) reduce 335) ((T_READONLY) reduce 335) ((T_REQUIRE) reduce 335) ((T_REQUIRE_ONCE) reduce 335) ((T_RETURN) reduce 335) ((T_STATIC) reduce 335) ((T_STRING) reduce 335) ((T_SWITCH) reduce 335) ((T_THROW) reduce 335) ((T_TRAIT) reduce 335) ((T_TRAIT_C) reduce 335) ((T_TRY) reduce 335) ((T_UNSET) reduce 335) ((T_USE) reduce 335) ((T_VAR) reduce 335) ((T_WHILE) reduce 335) ((T_YIELD) reduce 335)) 6007 (((";") reduce 30)) 6008 (((";") reduce 68)) 6009 (((";") reduce 21)) 6010 (((";") reduce 23)) 6011 (((";") reduce 41)) 6012 (((";") reduce 62)) 6013 (((";") reduce 34)) 6014 (((";") reduce 40)) 6015 (((";") reduce 13)) 6016 (((";") reduce 55)) 6017 (((";") reduce 1)) 6018 (((";") reduce 2)) 6019 (((";") reduce 9)) 6020 (((";") reduce 33)) 6021 (((";") reduce 58)) 6022 (((";") reduce 37)) 6023 (((";") reduce 64)) 6024 (((";") reduce 46)) 6025 (((";") reduce 8)) 6026 (((";") reduce 6)) 6027 (((";") reduce 7)) 6028 (((";") reduce 69)) 6029 (((";") reduce 63)) 6030 (((";") reduce 56)) 6031 (((";") reduce 10)) 6032 (((";") reduce 67)) 6033 (((";") reduce 44)) 6034 (((";") reduce 332) ((T_ABSTRACT) reduce 332) ((T_ARRAY) reduce 332) ((T_AS) reduce 332) ((T_BREAK) reduce 332) ((T_CALLABLE) reduce 332) ((T_CASE) reduce 332) ((T_CATCH) reduce 332) ((T_CLASS) reduce 332) ((T_CLASS_C) reduce 332) ((T_CLONE) reduce 332) ((T_CONST) reduce 332) ((T_CONTINUE) reduce 332) ((T_DECLARE) reduce 332) ((T_DEFAULT) reduce 332) ((T_DIR) reduce 332) ((T_DO) reduce 332) ((T_ECHO) reduce 332) ((T_ELSE) reduce 332) ((T_ELSEIF) reduce 332) ((T_EMPTY) reduce 332) ((T_ENDDECLARE) reduce 332) ((T_ENDFOR) reduce 332) ((T_ENDFOREACH) reduce 332) ((T_ENDIF) reduce 332) ((T_ENDSWITCH) reduce 332) ((T_ENDWHILE) reduce 332) ((T_ENUM) reduce 332) ((T_EVAL) reduce 332) ((T_EXIT) reduce 332) ((T_EXTENDS) reduce 332) ((T_FILE) reduce 332) ((T_FINAL) reduce 332) ((T_FINALLY) reduce 332) ((T_FN) reduce 332) ((T_FOR) reduce 332) ((T_FOREACH) reduce 332) ((T_FUNCTION) reduce 332) ((T_FUNC_C) reduce 332) ((T_GLOBAL) reduce 332) ((T_GOTO) reduce 332) ((T_IF) reduce 332) ((T_IMPLEMENTS) reduce 332) ((T_INCLUDE) reduce 332) ((T_INCLUDE_ONCE) reduce 332) ((T_INSTANCEOF) reduce 332) ((T_INSTEADOF) reduce 332) ((T_INTERFACE) reduce 332) ((T_ISSET) reduce 332) ((T_LINE) reduce 332) ((T_LIST) reduce 332) ((T_LOGICAL_AND) reduce 332) ((T_LOGICAL_OR) reduce 332) ((T_LOGICAL_XOR) reduce 332) ((T_MATCH) reduce 332) ((T_METHOD_C) reduce 332) ((T_NAMESPACE) reduce 332) ((T_NEW) reduce 332) ((T_NS_C) reduce 332) ((T_PRINT) reduce 332) ((T_PRIVATE) reduce 332) ((T_PROTECTED) reduce 332) ((T_PUBLIC) reduce 332) ((T_READONLY) reduce 332) ((T_REQUIRE) reduce 332) ((T_REQUIRE_ONCE) reduce 332) ((T_RETURN) reduce 332) ((T_STATIC) reduce 332) ((T_STRING) reduce 332) ((T_SWITCH) reduce 332) ((T_THROW) reduce 332) ((T_TRAIT) reduce 332) ((T_TRAIT_C) reduce 332) ((T_TRY) reduce 332) ((T_UNSET) reduce 332) ((T_USE) reduce 332) ((T_VAR) reduce 332) ((T_WHILE) reduce 332) ((T_YIELD) reduce 332)) 6035 (((";") reduce 331) ((T_ABSTRACT) reduce 331) ((T_ARRAY) reduce 331) ((T_AS) reduce 331) ((T_BREAK) reduce 331) ((T_CALLABLE) reduce 331) ((T_CASE) reduce 331) ((T_CATCH) reduce 331) ((T_CLASS) reduce 331) ((T_CLASS_C) reduce 331) ((T_CLONE) reduce 331) ((T_CONST) reduce 331) ((T_CONTINUE) reduce 331) ((T_DECLARE) reduce 331) ((T_DEFAULT) reduce 331) ((T_DIR) reduce 331) ((T_DO) reduce 331) ((T_ECHO) reduce 331) ((T_ELSE) reduce 331) ((T_ELSEIF) reduce 331) ((T_EMPTY) reduce 331) ((T_ENDDECLARE) reduce 331) ((T_ENDFOR) reduce 331) ((T_ENDFOREACH) reduce 331) ((T_ENDIF) reduce 331) ((T_ENDSWITCH) reduce 331) ((T_ENDWHILE) reduce 331) ((T_ENUM) reduce 331) ((T_EVAL) reduce 331) ((T_EXIT) reduce 331) ((T_EXTENDS) reduce 331) ((T_FILE) reduce 331) ((T_FINAL) reduce 331) ((T_FINALLY) reduce 331) ((T_FN) reduce 331) ((T_FOR) reduce 331) ((T_FOREACH) reduce 331) ((T_FUNCTION) reduce 331) ((T_FUNC_C) reduce 331) ((T_GLOBAL) reduce 331) ((T_GOTO) reduce 331) ((T_IF) reduce 331) ((T_IMPLEMENTS) reduce 331) ((T_INCLUDE) reduce 331) ((T_INCLUDE_ONCE) reduce 331) ((T_INSTANCEOF) reduce 331) ((T_INSTEADOF) reduce 331) ((T_INTERFACE) reduce 331) ((T_ISSET) reduce 331) ((T_LINE) reduce 331) ((T_LIST) reduce 331) ((T_LOGICAL_AND) reduce 331) ((T_LOGICAL_OR) reduce 331) ((T_LOGICAL_XOR) reduce 331) ((T_MATCH) reduce 331) ((T_METHOD_C) reduce 331) ((T_NAMESPACE) reduce 331) ((T_NEW) reduce 331) ((T_NS_C) reduce 331) ((T_PRINT) reduce 331) ((T_PRIVATE) reduce 331) ((T_PROTECTED) reduce 331) ((T_PUBLIC) reduce 331) ((T_READONLY) reduce 331) ((T_REQUIRE) reduce 331) ((T_REQUIRE_ONCE) reduce 331) ((T_RETURN) reduce 331) ((T_STATIC) reduce 331) ((T_STRING) reduce 331) ((T_SWITCH) reduce 331) ((T_THROW) reduce 331) ((T_TRAIT) reduce 331) ((T_TRAIT_C) reduce 331) ((T_TRY) reduce 331) ((T_UNSET) reduce 331) ((T_USE) reduce 331) ((T_VAR) reduce 331) ((T_WHILE) reduce 331) ((T_YIELD) reduce 331)) 6036 (((";") reduce 330) ((T_ABSTRACT) reduce 330) ((T_ARRAY) reduce 330) ((T_AS) reduce 330) ((T_BREAK) reduce 330) ((T_CALLABLE) reduce 330) ((T_CASE) reduce 330) ((T_CATCH) reduce 330) ((T_CLASS) reduce 330) ((T_CLASS_C) reduce 330) ((T_CLONE) reduce 330) ((T_CONST) reduce 330) ((T_CONTINUE) reduce 330) ((T_DECLARE) reduce 330) ((T_DEFAULT) reduce 330) ((T_DIR) reduce 330) ((T_DO) reduce 330) ((T_ECHO) reduce 330) ((T_ELSE) reduce 330) ((T_ELSEIF) reduce 330) ((T_EMPTY) reduce 330) ((T_ENDDECLARE) reduce 330) ((T_ENDFOR) reduce 330) ((T_ENDFOREACH) reduce 330) ((T_ENDIF) reduce 330) ((T_ENDSWITCH) reduce 330) ((T_ENDWHILE) reduce 330) ((T_ENUM) reduce 330) ((T_EVAL) reduce 330) ((T_EXIT) reduce 330) ((T_EXTENDS) reduce 330) ((T_FILE) reduce 330) ((T_FINAL) reduce 330) ((T_FINALLY) reduce 330) ((T_FN) reduce 330) ((T_FOR) reduce 330) ((T_FOREACH) reduce 330) ((T_FUNCTION) reduce 330) ((T_FUNC_C) reduce 330) ((T_GLOBAL) reduce 330) ((T_GOTO) reduce 330) ((T_IF) reduce 330) ((T_IMPLEMENTS) reduce 330) ((T_INCLUDE) reduce 330) ((T_INCLUDE_ONCE) reduce 330) ((T_INSTANCEOF) reduce 330) ((T_INSTEADOF) reduce 330) ((T_INTERFACE) reduce 330) ((T_ISSET) reduce 330) ((T_LINE) reduce 330) ((T_LIST) reduce 330) ((T_LOGICAL_AND) reduce 330) ((T_LOGICAL_OR) reduce 330) ((T_LOGICAL_XOR) reduce 330) ((T_MATCH) reduce 330) ((T_METHOD_C) reduce 330) ((T_NAMESPACE) reduce 330) ((T_NEW) reduce 330) ((T_NS_C) reduce 330) ((T_PRINT) reduce 330) ((T_PRIVATE) reduce 330) ((T_PROTECTED) reduce 330) ((T_PUBLIC) reduce 330) ((T_READONLY) reduce 330) ((T_REQUIRE) reduce 330) ((T_REQUIRE_ONCE) reduce 330) ((T_RETURN) reduce 330) ((T_STATIC) reduce 330) ((T_STRING) reduce 330) ((T_SWITCH) reduce 330) ((T_THROW) reduce 330) ((T_TRAIT) reduce 330) ((T_TRAIT_C) reduce 330) ((T_TRY) reduce 330) ((T_UNSET) reduce 330) ((T_USE) reduce 330) ((T_VAR) reduce 330) ((T_WHILE) reduce 330) ((T_YIELD) reduce 330)) 6037 (((";") reduce 336) ((T_ABSTRACT) reduce 336) ((T_ARRAY) reduce 336) ((T_AS) reduce 336) ((T_BREAK) reduce 336) ((T_CALLABLE) reduce 336) ((T_CASE) reduce 336) ((T_CATCH) reduce 336) ((T_CLASS) reduce 336) ((T_CLASS_C) reduce 336) ((T_CLONE) reduce 336) ((T_CONST) reduce 336) ((T_CONTINUE) reduce 336) ((T_DECLARE) reduce 336) ((T_DEFAULT) reduce 336) ((T_DIR) reduce 336) ((T_DO) reduce 336) ((T_ECHO) reduce 336) ((T_ELSE) reduce 336) ((T_ELSEIF) reduce 336) ((T_EMPTY) reduce 336) ((T_ENDDECLARE) reduce 336) ((T_ENDFOR) reduce 336) ((T_ENDFOREACH) reduce 336) ((T_ENDIF) reduce 336) ((T_ENDSWITCH) reduce 336) ((T_ENDWHILE) reduce 336) ((T_ENUM) reduce 336) ((T_EVAL) reduce 336) ((T_EXIT) reduce 336) ((T_EXTENDS) reduce 336) ((T_FILE) reduce 336) ((T_FINAL) reduce 336) ((T_FINALLY) reduce 336) ((T_FN) reduce 336) ((T_FOR) reduce 336) ((T_FOREACH) reduce 336) ((T_FUNCTION) reduce 336) ((T_FUNC_C) reduce 336) ((T_GLOBAL) reduce 336) ((T_GOTO) reduce 336) ((T_IF) reduce 336) ((T_IMPLEMENTS) reduce 336) ((T_INCLUDE) reduce 336) ((T_INCLUDE_ONCE) reduce 336) ((T_INSTANCEOF) reduce 336) ((T_INSTEADOF) reduce 336) ((T_INTERFACE) reduce 336) ((T_ISSET) reduce 336) ((T_LINE) reduce 336) ((T_LIST) reduce 336) ((T_LOGICAL_AND) reduce 336) ((T_LOGICAL_OR) reduce 336) ((T_LOGICAL_XOR) reduce 336) ((T_MATCH) reduce 336) ((T_METHOD_C) reduce 336) ((T_NAMESPACE) reduce 336) ((T_NEW) reduce 336) ((T_NS_C) reduce 336) ((T_PRINT) reduce 336) ((T_PRIVATE) reduce 336) ((T_PROTECTED) reduce 336) ((T_PUBLIC) reduce 336) ((T_READONLY) reduce 336) ((T_REQUIRE) reduce 336) ((T_REQUIRE_ONCE) reduce 336) ((T_RETURN) reduce 336) ((T_STATIC) reduce 336) ((T_STRING) reduce 336) ((T_SWITCH) reduce 336) ((T_THROW) reduce 336) ((T_TRAIT) reduce 336) ((T_TRAIT_C) reduce 336) ((T_TRY) reduce 336) ((T_UNSET) reduce 336) ((T_USE) reduce 336) ((T_VAR) reduce 336) ((T_WHILE) reduce 336) ((T_YIELD) reduce 336)) 6038 (((";") reduce 4)) 6039 (((";") reduce 5)) 6040 (((";") reduce 43)) 6041 (((";") reduce 333) ((T_ABSTRACT) reduce 333) ((T_ARRAY) reduce 333) ((T_AS) reduce 333) ((T_BREAK) reduce 333) ((T_CALLABLE) reduce 333) ((T_CASE) reduce 333) ((T_CATCH) reduce 333) ((T_CLASS) reduce 333) ((T_CLASS_C) reduce 333) ((T_CLONE) reduce 333) ((T_CONST) reduce 333) ((T_CONTINUE) reduce 333) ((T_DECLARE) reduce 333) ((T_DEFAULT) reduce 333) ((T_DIR) reduce 333) ((T_DO) reduce 333) ((T_ECHO) reduce 333) ((T_ELSE) reduce 333) ((T_ELSEIF) reduce 333) ((T_EMPTY) reduce 333) ((T_ENDDECLARE) reduce 333) ((T_ENDFOR) reduce 333) ((T_ENDFOREACH) reduce 333) ((T_ENDIF) reduce 333) ((T_ENDSWITCH) reduce 333) ((T_ENDWHILE) reduce 333) ((T_ENUM) reduce 333) ((T_EVAL) reduce 333) ((T_EXIT) reduce 333) ((T_EXTENDS) reduce 333) ((T_FILE) reduce 333) ((T_FINAL) reduce 333) ((T_FINALLY) reduce 333) ((T_FN) reduce 333) ((T_FOR) reduce 333) ((T_FOREACH) reduce 333) ((T_FUNCTION) reduce 333) ((T_FUNC_C) reduce 333) ((T_GLOBAL) reduce 333) ((T_GOTO) reduce 333) ((T_IF) reduce 333) ((T_IMPLEMENTS) reduce 333) ((T_INCLUDE) reduce 333) ((T_INCLUDE_ONCE) reduce 333) ((T_INSTANCEOF) reduce 333) ((T_INSTEADOF) reduce 333) ((T_INTERFACE) reduce 333) ((T_ISSET) reduce 333) ((T_LINE) reduce 333) ((T_LIST) reduce 333) ((T_LOGICAL_AND) reduce 333) ((T_LOGICAL_OR) reduce 333) ((T_LOGICAL_XOR) reduce 333) ((T_MATCH) reduce 333) ((T_METHOD_C) reduce 333) ((T_NAMESPACE) reduce 333) ((T_NEW) reduce 333) ((T_NS_C) reduce 333) ((T_PRINT) reduce 333) ((T_PRIVATE) reduce 333) ((T_PROTECTED) reduce 333) ((T_PUBLIC) reduce 333) ((T_READONLY) reduce 333) ((T_REQUIRE) reduce 333) ((T_REQUIRE_ONCE) reduce 333) ((T_RETURN) reduce 333) ((T_STATIC) reduce 333) ((T_STRING) reduce 333) ((T_SWITCH) reduce 333) ((T_THROW) reduce 333) ((T_TRAIT) reduce 333) ((T_TRAIT_C) reduce 333) ((T_TRY) reduce 333) ((T_UNSET) reduce 333) ((T_USE) reduce 333) ((T_VAR) reduce 333) ((T_WHILE) reduce 333) ((T_YIELD) reduce 333)) 6042 (((";") reduce 315)) 6043 (((";") reduce 47)) 6044 (((";") reduce 31)) 6045 (((";") reduce 57)) 6046 (((";") reduce 61)) 6047 (((";") reduce 28)) 6048 (((";") reduce 36)) 6049 (((";") reduce 32)) 6050 (((";") reduce 35)) 6051 (((";") reduce 19)) 6052 (((";") reduce 45)) 6053 (((";") reduce 318) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift)) 6054 (((";") reduce 316)) 6055 (((";") reduce 73)) 6056 (((";") reduce 74)) 6057 (((";") reduce 75)) 6058 (((";") reduce 76)) 6059 (((";") reduce 77)) 6060 (((";") reduce 78)) 6061 (((";") reduce 72)) 6062 (((";") reduce 81)) 6063 (((";") reduce 317)) 6064 (((";") reduce 71)) 6065 (((";") reduce 82)) 6066 (((T_ABSTRACT) reduce 313) ((T_ARRAY) reduce 313) ((T_AS) reduce 313) ((T_BREAK) reduce 313) ((T_CALLABLE) reduce 313) ((T_CASE) reduce 313) ((T_CATCH) reduce 313) ((T_CLASS) reduce 313) ((T_CLASS_C) reduce 313) ((T_CLONE) reduce 313) ((T_CONST) reduce 313) ((T_CONTINUE) reduce 313) ((T_DECLARE) reduce 313) ((T_DEFAULT) reduce 313) ((T_DIR) reduce 313) ((T_DO) reduce 313) ((T_ECHO) reduce 313) ((T_ELSE) reduce 313) ((T_ELSEIF) reduce 313) ((T_EMPTY) reduce 313) ((T_ENDDECLARE) reduce 313) ((T_ENDFOR) reduce 313) ((T_ENDFOREACH) reduce 313) ((T_ENDIF) reduce 313) ((T_ENDSWITCH) reduce 313) ((T_ENDWHILE) reduce 313) ((T_ENUM) reduce 313) ((T_EVAL) reduce 313) ((T_EXIT) reduce 313) ((T_EXTENDS) reduce 313) ((T_FILE) reduce 313) ((T_FINAL) reduce 313) ((T_FINALLY) reduce 313) ((T_FN) reduce 313) ((T_FOR) reduce 313) ((T_FOREACH) reduce 313) ((T_FUNCTION) reduce 313) ((T_FUNC_C) reduce 313) ((T_GLOBAL) reduce 313) ((T_GOTO) reduce 313) ((T_IF) reduce 313) ((T_IMPLEMENTS) reduce 313) ((T_INCLUDE) reduce 313) ((T_INCLUDE_ONCE) reduce 313) ((T_INSTANCEOF) reduce 313) ((T_INSTEADOF) reduce 313) ((T_INTERFACE) reduce 313) ((T_ISSET) reduce 313) ((T_LINE) reduce 313) ((T_LIST) reduce 313) ((T_LOGICAL_AND) reduce 313) ((T_LOGICAL_OR) reduce 313) ((T_LOGICAL_XOR) reduce 313) ((T_MATCH) reduce 313) ((T_METHOD_C) reduce 313) ((T_NAMESPACE) reduce 313) ((T_NAME_FULLY_QUALIFIED) reduce 313) ((T_NAME_QUALIFIED) reduce 313) ((T_NAME_RELATIVE) reduce 313) ((T_NEW) reduce 313) ((T_NS_C) reduce 313) ((T_PRINT) reduce 313) ((T_PRIVATE) reduce 313) ((T_PROTECTED) reduce 313) ((T_PUBLIC) reduce 313) ((T_READONLY) reduce 313) ((T_REQUIRE) reduce 313) ((T_REQUIRE_ONCE) reduce 313) ((T_RETURN) reduce 313) ((T_STATIC) reduce 313) ((T_STRING) reduce 313) ((T_SWITCH) reduce 313) ((T_THROW) reduce 313) ((T_TRAIT) reduce 313) ((T_TRAIT_C) reduce 313) ((T_TRY) reduce 313) ((T_UNSET) reduce 313) ((T_USE) reduce 313) ((T_VAR) reduce 313) ((T_WHILE) reduce 313) ((T_YIELD) reduce 313) (("}") reduce 313)) 6067 (((T_ABSTRACT) reduce 311) ((T_ARRAY) reduce 311) ((T_AS) reduce 311) ((T_BREAK) reduce 311) ((T_CALLABLE) reduce 311) ((T_CASE) reduce 311) ((T_CATCH) reduce 311) ((T_CLASS) reduce 311) ((T_CLASS_C) reduce 311) ((T_CLONE) reduce 311) ((T_CONST) reduce 311) ((T_CONTINUE) reduce 311) ((T_DECLARE) reduce 311) ((T_DEFAULT) reduce 311) ((T_DIR) reduce 311) ((T_DO) reduce 311) ((T_ECHO) reduce 311) ((T_ELSE) reduce 311) ((T_ELSEIF) reduce 311) ((T_EMPTY) reduce 311) ((T_ENDDECLARE) reduce 311) ((T_ENDFOR) reduce 311) ((T_ENDFOREACH) reduce 311) ((T_ENDIF) reduce 311) ((T_ENDSWITCH) reduce 311) ((T_ENDWHILE) reduce 311) ((T_ENUM) reduce 311) ((T_EVAL) reduce 311) ((T_EXIT) reduce 311) ((T_EXTENDS) reduce 311) ((T_FILE) reduce 311) ((T_FINAL) reduce 311) ((T_FINALLY) reduce 311) ((T_FN) reduce 311) ((T_FOR) reduce 311) ((T_FOREACH) reduce 311) ((T_FUNCTION) reduce 311) ((T_FUNC_C) reduce 311) ((T_GLOBAL) reduce 311) ((T_GOTO) reduce 311) ((T_IF) reduce 311) ((T_IMPLEMENTS) reduce 311) ((T_INCLUDE) reduce 311) ((T_INCLUDE_ONCE) reduce 311) ((T_INSTANCEOF) reduce 311) ((T_INSTEADOF) reduce 311) ((T_INTERFACE) reduce 311) ((T_ISSET) reduce 311) ((T_LINE) reduce 311) ((T_LIST) reduce 311) ((T_LOGICAL_AND) reduce 311) ((T_LOGICAL_OR) reduce 311) ((T_LOGICAL_XOR) reduce 311) ((T_MATCH) reduce 311) ((T_METHOD_C) reduce 311) ((T_NAMESPACE) reduce 311) ((T_NAME_FULLY_QUALIFIED) reduce 311) ((T_NAME_QUALIFIED) reduce 311) ((T_NAME_RELATIVE) reduce 311) ((T_NEW) reduce 311) ((T_NS_C) reduce 311) ((T_PRINT) reduce 311) ((T_PRIVATE) reduce 311) ((T_PROTECTED) reduce 311) ((T_PUBLIC) reduce 311) ((T_READONLY) reduce 311) ((T_REQUIRE) reduce 311) ((T_REQUIRE_ONCE) reduce 311) ((T_RETURN) reduce 311) ((T_STATIC) reduce 311) ((T_STRING) reduce 311) ((T_SWITCH) reduce 311) ((T_THROW) reduce 311) ((T_TRAIT) reduce 311) ((T_TRAIT_C) reduce 311) ((T_TRY) reduce 311) ((T_UNSET) reduce 311) ((T_USE) reduce 311) ((T_VAR) reduce 311) ((T_WHILE) reduce 311) ((T_YIELD) reduce 311) (("}") reduce 311)) 6068 (((T_ABSTRACT) reduce 309) ((T_ATTRIBUTE) reduce 309) ((T_CASE) reduce 309) ((T_CONST) reduce 309) ((T_FINAL) reduce 309) ((T_FUNCTION) reduce 309) ((T_PRIVATE) reduce 309) ((T_PROTECTED) reduce 309) ((T_PUBLIC) reduce 309) ((T_READONLY) reduce 309) ((T_STATIC) reduce 309) ((T_USE) reduce 309) ((T_VAR) reduce 309) (("}") reduce 309)) 6070 (((T_AS) reduce 73) ((T_INSTEADOF) reduce 73)) 6071 (((T_AS) reduce 52) ((T_INSTEADOF) reduce 52)) 6072 (((T_AS) reduce 27) ((T_INSTEADOF) reduce 27)) 6073 (((T_AS) reduce 51) ((T_INSTEADOF) reduce 51)) 6074 (((T_AS) reduce 53) ((T_INSTEADOF) reduce 53)) 6075 (((T_AS) reduce 49) ((T_INSTEADOF) reduce 49)) 6076 (((T_AS) reduce 29) ((T_INSTEADOF) reduce 29)) 6077 (((T_AS) reduce 59) ((T_INSTEADOF) reduce 59)) 6078 (((T_AS) reduce 60) ((T_INSTEADOF) reduce 60)) 6079 (((T_AS) reduce 11) ((T_INSTEADOF) reduce 11)) 6080 (((T_AS) reduce 42) ((T_INSTEADOF) reduce 42)) 6081 (((T_AS) reduce 39) ((T_INSTEADOF) reduce 39)) 6082 (((T_AS) reduce 25) ((T_INSTEADOF) reduce 25)) 6083 (((T_AS) reduce 50) ((T_INSTEADOF) reduce 50)) 6084 (((T_AS) reduce 66) ((T_INSTEADOF) reduce 66)) 6085 (((T_AS) reduce 18) ((T_INSTEADOF) reduce 18)) 6086 (((T_AS) reduce 17) ((T_INSTEADOF) reduce 17)) 6087 (((T_AS) reduce 15) ((T_INSTEADOF) reduce 15)) 6088 (((T_AS) reduce 14) ((T_INSTEADOF) reduce 14)) 6089 (((T_AS) reduce 38) ((T_INSTEADOF) reduce 38)) 6090 (((T_AS) reduce 26) ((T_INSTEADOF) reduce 26)) 6091 (((T_AS) reduce 22) ((T_INSTEADOF) reduce 22)) 6092 (((T_AS) reduce 24) ((T_INSTEADOF) reduce 24)) 6093 (((T_AS) reduce 16) ((T_INSTEADOF) reduce 16)) 6094 (((T_AS) reduce 48) ((T_INSTEADOF) reduce 48)) 6095 (((T_AS) reduce 20) ((T_INSTEADOF) reduce 20)) 6096 (((T_AS) reduce 70) ((T_INSTEADOF) reduce 70)) 6097 (((T_AS) reduce 3) ((T_INSTEADOF) reduce 3)) 6098 (((T_AS) reduce 12) ((T_INSTEADOF) reduce 12)) 6099 (((T_AS) reduce 54) ((T_INSTEADOF) reduce 54)) 6100 (((T_AS) reduce 65) ((T_INSTEADOF) reduce 65)) 6101 (((T_AS) reduce 74) ((T_INSTEADOF) reduce 74)) 6102 (((T_AS) reduce 30) ((T_INSTEADOF) reduce 30)) 6103 (((T_AS) reduce 68) ((T_INSTEADOF) reduce 68)) 6104 (((T_AS) reduce 21) ((T_INSTEADOF) reduce 21)) 6105 (((T_AS) reduce 23) ((T_INSTEADOF) reduce 23)) 6106 (((T_AS) reduce 41) ((T_INSTEADOF) reduce 41)) 6107 (((T_AS) reduce 62) ((T_INSTEADOF) reduce 62)) 6108 (((T_AS) reduce 34) ((T_INSTEADOF) reduce 34)) 6109 (((T_AS) reduce 40) ((T_INSTEADOF) reduce 40)) 6110 (((T_AS) reduce 13) ((T_INSTEADOF) reduce 13)) 6111 (((T_AS) reduce 55) ((T_INSTEADOF) reduce 55)) 6112 (((T_AS) reduce 1) ((T_INSTEADOF) reduce 1)) 6113 (((T_AS) reduce 2) ((T_INSTEADOF) reduce 2)) 6114 (((T_AS) reduce 9) ((T_INSTEADOF) reduce 9)) 6115 (((T_AS) reduce 33) ((T_INSTEADOF) reduce 33)) 6116 (((T_AS) reduce 58) ((T_INSTEADOF) reduce 58)) 6117 (((T_AS) reduce 37) ((T_INSTEADOF) reduce 37)) 6118 (((T_AS) reduce 64) ((T_INSTEADOF) reduce 64)) 6119 (((T_AS) reduce 46) ((T_INSTEADOF) reduce 46)) 6120 (((T_AS) reduce 8) ((T_INSTEADOF) reduce 8)) 6121 (((T_AS) reduce 6) ((T_INSTEADOF) reduce 6)) 6122 (((T_AS) reduce 7) ((T_INSTEADOF) reduce 7)) 6123 (((T_AS) reduce 69) ((T_INSTEADOF) reduce 69)) 6124 (((T_AS) reduce 63) ((T_INSTEADOF) reduce 63)) 6125 (((T_AS) reduce 56) ((T_INSTEADOF) reduce 56)) 6126 (((T_AS) reduce 10) ((T_INSTEADOF) reduce 10)) 6127 (((T_AS) reduce 67) ((T_INSTEADOF) reduce 67)) 6128 (((T_AS) reduce 44) ((T_INSTEADOF) reduce 44)) 6129 (((T_AS) reduce 75) ((T_INSTEADOF) reduce 75)) 6130 (((T_AS) reduce 76) ((T_INSTEADOF) reduce 76)) 6131 (((T_AS) reduce 77) ((T_INSTEADOF) reduce 77)) 6132 (((T_AS) reduce 78) ((T_INSTEADOF) reduce 78)) 6133 (((T_AS) reduce 4) ((T_INSTEADOF) reduce 4)) 6134 (((T_AS) reduce 5) ((T_INSTEADOF) reduce 5)) 6135 (((T_AS) reduce 43) ((T_INSTEADOF) reduce 43)) 6136 (((T_AS) reduce 72) ((T_INSTEADOF) reduce 72)) 6137 (((T_AS) reduce 81) ((T_INSTEADOF) reduce 81)) 6138 (((T_AS) reduce 47) ((T_INSTEADOF) reduce 47)) 6139 (((T_AS) reduce 31) ((T_INSTEADOF) reduce 31)) 6140 (((T_AS) reduce 57) ((T_INSTEADOF) reduce 57)) 6141 (((T_AS) reduce 61) ((T_INSTEADOF) reduce 61)) 6142 (((T_AS) reduce 28) ((T_INSTEADOF) reduce 28)) 6143 (((T_AS) reduce 36) ((T_INSTEADOF) reduce 36)) 6144 (((T_AS) reduce 32) ((T_INSTEADOF) reduce 32)) 6145 (((T_AS) reduce 35) ((T_INSTEADOF) reduce 35)) 6146 (((T_AS) reduce 19) ((T_INSTEADOF) reduce 19)) 6147 (((T_AS) reduce 45) ((T_INSTEADOF) reduce 45)) 6148 (((T_AS) reduce 321) ((T_INSTEADOF) reduce 321)) 6149 (((T_AS) reduce 71) ((T_INSTEADOF) reduce 71)) 6150 (((T_AS) reduce 82) ((T_INSTEADOF) reduce 82)) 6152 (((",") reduce 93) ((";") reduce 93)) 6153 (((",") reduce 92) ((";") reduce 92)) 6154 (((",") reduce 94) ((";") reduce 94)) 6155 (((",") reduce 458) ((";") reduce 458)) 6156 (((",") reduce 91) ((";") reduce 91)) 6157 (((",") reduce 305) ((";") reduce 305)) 6158 (((",") shift) ((";") reduce 314)) 6159 (((",") reduce 459) ((";") reduce 459)) 6161 (((",") reduce 306) ((";") reduce 306)) 6162 (((",") reduce 306) ((";") reduce 306) (("{") reduce 306)) 6164 (((";") reduce 73) (("=") reduce 73)) 6165 (((";") reduce 52) (("=") reduce 52)) 6166 (((";") reduce 27) (("=") reduce 27)) 6167 (((";") reduce 51) (("=") reduce 51)) 6168 (((";") reduce 53) (("=") reduce 53)) 6169 (((";") reduce 49) (("=") reduce 49)) 6170 (((";") reduce 29) (("=") reduce 29)) 6171 (((";") reduce 59) (("=") reduce 59)) 6172 (((";") reduce 60) (("=") reduce 60)) 6173 (((";") reduce 11) (("=") reduce 11)) 6174 (((";") reduce 42) (("=") reduce 42)) 6175 (((";") reduce 39) (("=") reduce 39)) 6176 (((";") reduce 25) (("=") reduce 25)) 6177 (((";") reduce 50) (("=") reduce 50)) 6178 (((";") reduce 66) (("=") reduce 66)) 6179 (((";") reduce 18) (("=") reduce 18)) 6180 (((";") reduce 17) (("=") reduce 17)) 6181 (((";") reduce 15) (("=") reduce 15)) 6182 (((";") reduce 14) (("=") reduce 14)) 6183 (((";") reduce 38) (("=") reduce 38)) 6184 (((";") reduce 26) (("=") reduce 26)) 6185 (((";") reduce 22) (("=") reduce 22)) 6186 (((";") reduce 24) (("=") reduce 24)) 6187 (((";") reduce 16) (("=") reduce 16)) 6188 (((";") reduce 48) (("=") reduce 48)) 6189 (((";") reduce 20) (("=") reduce 20)) 6190 (((";") reduce 70) (("=") reduce 70)) 6191 (((";") reduce 3) (("=") reduce 3)) 6192 (((";") reduce 12) (("=") reduce 12)) 6193 (((";") reduce 54) (("=") reduce 54)) 6194 (((";") reduce 65) (("=") reduce 65)) 6195 (((";") reduce 74) (("=") reduce 74)) 6196 (((";") reduce 30) (("=") reduce 30)) 6197 (((";") reduce 68) (("=") reduce 68)) 6198 (((";") reduce 21) (("=") reduce 21)) 6199 (((";") reduce 23) (("=") reduce 23)) 6200 (((";") reduce 41) (("=") reduce 41)) 6201 (((";") reduce 62) (("=") reduce 62)) 6202 (((";") reduce 34) (("=") reduce 34)) 6203 (((";") reduce 40) (("=") reduce 40)) 6204 (((";") reduce 13) (("=") reduce 13)) 6205 (((";") reduce 55) (("=") reduce 55)) 6206 (((";") reduce 1) (("=") reduce 1)) 6207 (((";") reduce 2) (("=") reduce 2)) 6208 (((";") reduce 9) (("=") reduce 9)) 6209 (((";") reduce 33) (("=") reduce 33)) 6210 (((";") reduce 58) (("=") reduce 58)) 6211 (((";") reduce 37) (("=") reduce 37)) 6212 (((";") reduce 64) (("=") reduce 64)) 6213 (((";") reduce 46) (("=") reduce 46)) 6214 (((";") reduce 8) (("=") reduce 8)) 6215 (((";") reduce 6) (("=") reduce 6)) 6216 (((";") reduce 7) (("=") reduce 7)) 6217 (((";") reduce 69) (("=") reduce 69)) 6218 (((";") reduce 63) (("=") reduce 63)) 6219 (((";") reduce 56) (("=") reduce 56)) 6220 (((";") reduce 10) (("=") reduce 10)) 6221 (((";") reduce 67) (("=") reduce 67)) 6222 (((";") reduce 44) (("=") reduce 44)) 6223 (((";") reduce 75) (("=") reduce 75)) 6224 (((";") reduce 76) (("=") reduce 76)) 6225 (((";") reduce 77) (("=") reduce 77)) 6226 (((";") reduce 78) (("=") reduce 78)) 6227 (((";") reduce 4) (("=") reduce 4)) 6228 (((";") reduce 5) (("=") reduce 5)) 6229 (((";") reduce 43) (("=") reduce 43)) 6230 (((";") reduce 72) (("=") reduce 72)) 6231 (((";") reduce 81) (("=") reduce 81)) 6232 (((";") reduce 47) (("=") reduce 47)) 6233 (((";") reduce 31) (("=") reduce 31)) 6234 (((";") reduce 57) (("=") reduce 57)) 6235 (((";") reduce 61) (("=") reduce 61)) 6236 (((";") reduce 28) (("=") reduce 28)) 6237 (((";") reduce 36) (("=") reduce 36)) 6238 (((";") reduce 32) (("=") reduce 32)) 6239 (((";") reduce 35) (("=") reduce 35)) 6240 (((";") reduce 19) (("=") reduce 19)) 6241 (((";") reduce 45) (("=") reduce 45)) 6242 (((";") reduce 196) (("=") shift)) 6243 (((";") reduce 71) (("=") reduce 71)) 6244 (((";") reduce 82) (("=") reduce 82)) 6247 (((T_ABSTRACT) reduce 195) ((T_ATTRIBUTE) reduce 195) ((T_CASE) reduce 195) ((T_CONST) reduce 195) ((T_FINAL) reduce 195) ((T_FUNCTION) reduce 195) ((T_PRIVATE) reduce 195) ((T_PROTECTED) reduce 195) ((T_PUBLIC) reduce 195) ((T_READONLY) reduce 195) ((T_STATIC) reduce 195) ((T_USE) reduce 195) ((T_VAR) reduce 195) (("}") reduce 195)) 6248 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 197) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6279 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 384) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 6280 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 387) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 6281 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) ((";") reduce 396) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 6282 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 409) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 6283 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) ((";") reduce 395) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 6284 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 6285 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 383) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6286 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 381) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6287 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 382) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6288 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 6289 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 402) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 6290 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 6291 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 401) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 6292 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 408) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 6293 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 403) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 6296 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 6297 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 6298 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 6299 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 6300 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 6301 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 6302 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 6303 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 6304 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 6305 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 6306 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 6313 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 6315 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 6316 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 6317 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 6318 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 6319 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 6322 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 6323 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 6325 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 6327 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 6328 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 6331 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 6332 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6333 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 379) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6334 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 380) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6335 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 385) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 6336 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 386) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 6340 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6341 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6342 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 407) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 6343 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 6344 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 6345 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) ((";") reduce 388) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 6346 ((("%") shift) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) ((";") reduce 390) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 6347 ((("%") shift) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) ((";") reduce 389) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 6348 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) ((";") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 6349 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) ((";") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 6350 ((("(") reduce 93) ((",") reduce 93) (("]") reduce 93)) 6351 ((("(") reduce 92) ((",") reduce 92) (("]") reduce 92)) 6352 ((("(") reduce 94) ((",") reduce 94) (("]") reduce 94)) 6353 ((("(") reduce 458) ((",") reduce 458) (("]") reduce 458)) 6354 ((("(") reduce 91) ((",") reduce 91) (("]") reduce 91)) 6355 (((",") reduce 97) (("]") reduce 97)) 6356 (((",") shift) (("]") reduce 123)) 6357 ((("(") shift) ((",") reduce 95) (("]") reduce 95)) 6358 ((("(") reduce 459) ((",") reduce 459) (("]") reduce 459)) 6360 (((",") reduce 96) (("]") reduce 96)) 6361 (((",") reduce 281) (("]") reduce 281)) 6365 (((",") reduce 282) (("]") reduce 282)) 6366 (((",") reduce 283) (("]") reduce 283)) 6367 (((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift) (("]") reduce 124)) 6369 (((T_ABSTRACT) reduce 99) ((T_ATTRIBUTE) reduce 99) ((T_CASE) reduce 99) ((T_CONST) reduce 99) ((T_FINAL) reduce 99) ((T_FUNCTION) reduce 99) ((T_PRIVATE) reduce 99) ((T_PROTECTED) reduce 99) ((T_PUBLIC) reduce 99) ((T_READONLY) reduce 99) ((T_STATIC) reduce 99) ((T_VAR) reduce 99)) 6370 (((",") reduce 98) (("]") reduce 98)) 6371 (((",") reduce 93) (("{") reduce 93)) 6372 (((",") reduce 92) (("{") reduce 92)) 6373 (((",") reduce 94) (("{") reduce 94)) 6374 (((",") reduce 458) (("{") reduce 458)) 6375 (((",") reduce 91) (("{") reduce 91)) 6376 (((",") reduce 305) (("{") reduce 305)) 6377 (((",") shift) (("{") reduce 203)) 6378 (((",") reduce 459) (("{") reduce 459)) 6380 (((",") reduce 306) (("{") reduce 306)) 6381 (((T_IMPLEMENTS) reduce 93) (("{") reduce 93)) 6382 (((T_IMPLEMENTS) reduce 92) (("{") reduce 92)) 6383 (((T_IMPLEMENTS) reduce 94) (("{") reduce 94)) 6384 (((T_IMPLEMENTS) reduce 458) (("{") reduce 458)) 6385 (((T_IMPLEMENTS) reduce 91) (("{") reduce 91)) 6386 (((T_IMPLEMENTS) reduce 199) (("{") reduce 199)) 6387 (((T_IMPLEMENTS) reduce 459) (("{") reduce 459)) 6388 (((T_EXTENDS) reduce 281) ((T_IMPLEMENTS) reduce 281) (("{") reduce 281)) 6392 (((T_EXTENDS) reduce 282) ((T_IMPLEMENTS) reduce 282) (("{") reduce 282)) 6393 (((T_EXTENDS) reduce 283) ((T_IMPLEMENTS) reduce 283) (("{") reduce 283)) 6396 (((T_ATTRIBUTE) reduce 99) ((T_CLASS) reduce 99)) 6398 ((("%") reduce 462) (("(") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 6408 ((("%") reduce 223) ((")") reduce 223) (("*") reduce 223) (("+") reduce 223) ((",") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 6411 (((")") reduce 543) ((",") reduce 543) (("=") shift)) 6415 ((("%") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 6416 ((("%") reduce 417) ((")") reduce 417) (("*") reduce 417) (("+") reduce 417) ((",") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 6417 ((("%") shift) ((")") reduce 564) (("*") shift) (("+") shift) ((",") reduce 564) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6418 ((("%") shift) ((")") reduce 563) (("*") shift) (("+") shift) ((",") reduce 563) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6419 ((("%") reduce 376) ((")") reduce 376) (("*") reduce 376) (("+") reduce 376) ((",") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 6421 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 6423 ((("%") reduce 464) ((")") reduce 464) (("*") reduce 464) (("+") reduce 464) ((",") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 6426 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 6429 ((("%") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 6430 ((("%") reduce 418) ((")") reduce 418) (("*") reduce 418) (("+") reduce 418) ((",") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 6431 ((("%") reduce 378) ((")") reduce 378) (("*") reduce 378) (("+") reduce 378) ((",") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 6432 ((("%") reduce 361) ((")") reduce 361) (("*") reduce 361) (("+") reduce 361) ((",") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 6433 ((("%") reduce 422) ((")") reduce 422) (("*") reduce 422) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 6434 ((("%") reduce 420) ((")") reduce 420) (("*") reduce 420) (("+") reduce 420) ((",") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 6437 ((("%") reduce 470) (("(") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 6438 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 6439 ((("%") reduce 398) ((")") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 6440 ((("%") reduce 397) ((")") reduce 397) (("*") reduce 397) (("+") reduce 397) ((",") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 6442 ((("%") reduce 411) (("(") reduce 503) ((")") reduce 411) (("*") reduce 411) (("+") reduce 411) ((",") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 6443 ((("%") reduce 517) (("(") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 6446 ((("%") reduce 516) (("(") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 6448 ((("%") reduce 473) (("(") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 6449 ((("%") reduce 399) ((")") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 6451 ((("%") reduce 491) (("(") reduce 527) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 6453 ((("%") reduce 518) (("(") reduce 529) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 6454 ((("%") reduce 455) (("(") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 6455 ((("%") reduce 457) (("(") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 6457 ((("%") reduce 435) ((")") reduce 435) (("*") reduce 435) (("+") reduce 435) ((",") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 6458 ((("%") reduce 437) ((")") reduce 437) (("*") reduce 437) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 6461 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 357) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 357) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 357) (("^") shift) (("|") shift)) 6467 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 6469 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 6470 ((("%") reduce 530) (("(") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 6471 ((("%") reduce 513) (("(") shift) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 6472 ((("%") reduce 532) (("(") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 6475 ((("%") reduce 531) (("(") reduce 531) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 6476 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 6477 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 6478 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 6479 (((")") reduce 541) ((",") reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) (("[") reduce 496) (("{") reduce 496)) 6481 ((("%") reduce 427) ((")") reduce 427) (("*") reduce 427) (("+") reduce 427) ((",") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 6483 ((("%") reduce 471) (("(") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 6485 ((("%") shift) ((")") reduce 358) (("*") shift) (("+") shift) ((",") reduce 358) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 358) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6486 ((("%") shift) ((")") reduce 432) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6489 ((("%") shift) ((")") reduce 431) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6490 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 6491 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6492 ((("%") reduce 419) ((")") reduce 419) (("*") reduce 419) (("+") reduce 419) ((",") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 6493 ((("%") reduce 436) ((")") reduce 436) (("*") reduce 436) (("+") reduce 436) ((",") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 6495 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 6497 ((("%") reduce 478) ((")") reduce 478) (("*") reduce 478) (("+") reduce 478) ((",") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 6498 ((("%") reduce 476) ((")") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 6499 ((("%") shift) ((")") reduce 567) (("*") shift) (("+") shift) ((",") reduce 567) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6500 ((("%") shift) ((")") reduce 566) (("*") shift) (("+") shift) ((",") reduce 566) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6501 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6502 ((("%") reduce 421) ((")") reduce 421) (("*") reduce 421) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 6505 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 6506 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 6507 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 6508 ((("%") reduce 458) (("(") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 6509 ((("%") reduce 91) (("(") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 6510 ((("%") reduce 354) ((")") reduce 354) (("*") reduce 354) (("+") reduce 354) ((",") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_DOUBLE_ARROW) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 6512 ((("%") reduce 460) (("(") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 6513 ((("%") reduce 468) (("(") shift) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 6514 ((("%") reduce 459) (("(") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 6515 ((("%") reduce 461) (("(") reduce 461) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 6516 ((("%") reduce 520) (("(") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 6523 ((("%") reduce 522) (("(") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 6525 ((("%") reduce 521) (("(") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 6526 ((("%") reduce 526) (("(") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 6527 ((("%") reduce 523) (("(") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 6528 ((("%") reduce 524) (("(") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 6530 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 6531 ((("%") reduce 353) ((")") reduce 353) (("*") reduce 353) (("+") reduce 353) ((",") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_DOUBLE_ARROW) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 6532 ((("%") reduce 281) ((")") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 6536 ((("%") reduce 282) ((")") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 6537 ((("%") reduce 283) ((")") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 6539 ((("%") reduce 525) (("(") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 6540 ((("%") reduce 355) ((")") reduce 355) (("*") reduce 355) (("+") reduce 355) ((",") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_DOUBLE_ARROW) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 6547 ((("%") reduce 352) ((")") reduce 352) (("*") reduce 352) (("+") reduce 352) ((",") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_DOUBLE_ARROW) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 6549 ((("%") reduce 462) (("(") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 6555 ((("%") reduce 223) ((")") reduce 223) (("*") reduce 223) (("+") reduce 223) ((",") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_DOUBLE_ARROW) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 6558 (((")") reduce 544) ((",") reduce 544) (("=") shift)) 6562 ((("%") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 6563 ((("%") reduce 417) ((")") reduce 417) (("*") reduce 417) (("+") reduce 417) ((",") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 6564 ((("%") shift) ((")") reduce 564) (("*") shift) (("+") shift) ((",") reduce 564) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6565 ((("%") shift) ((")") reduce 563) (("*") shift) (("+") shift) ((",") reduce 563) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6566 ((("%") reduce 376) ((")") reduce 376) (("*") reduce 376) (("+") reduce 376) ((",") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_DOUBLE_ARROW) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 6568 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 6570 ((("%") reduce 464) ((")") reduce 464) (("*") reduce 464) (("+") reduce 464) ((",") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 6573 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 6576 ((("%") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 6577 ((("%") shift) ((")") reduce 542) (("*") shift) (("+") shift) ((",") reduce 542) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 6578 ((("%") reduce 418) ((")") reduce 418) (("*") reduce 418) (("+") reduce 418) ((",") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 6579 ((("%") reduce 378) ((")") reduce 378) (("*") reduce 378) (("+") reduce 378) ((",") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_DOUBLE_ARROW) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 6580 ((("%") reduce 361) ((")") reduce 361) (("*") reduce 361) (("+") reduce 361) ((",") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_DOUBLE_ARROW) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 6581 ((("%") reduce 422) ((")") reduce 422) (("*") reduce 422) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 6582 ((("%") reduce 420) ((")") reduce 420) (("*") reduce 420) (("+") reduce 420) ((",") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 6585 ((("%") reduce 470) (("(") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 6586 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 6587 ((("%") reduce 398) ((")") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 6588 ((("%") reduce 397) ((")") reduce 397) (("*") reduce 397) (("+") reduce 397) ((",") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 6590 ((("%") reduce 411) (("(") reduce 503) ((")") reduce 411) (("*") reduce 411) (("+") reduce 411) ((",") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 6591 ((("%") reduce 517) (("(") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 6594 ((("%") reduce 516) (("(") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 6596 ((("%") reduce 473) (("(") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 6597 ((("%") reduce 399) ((")") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 6599 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("]") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 6600 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("]") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 6601 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("]") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 6602 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("]") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 6603 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("]") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 6604 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("]") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 6605 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("]") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 6606 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("]") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 6607 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("]") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 6608 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("]") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 6609 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("]") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 6610 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("]") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 6611 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("]") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 6612 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("]") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 6613 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("]") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 6614 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("]") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 6615 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("]") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 6616 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("]") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 6617 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("]") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 6618 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("]") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 6619 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("]") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 6620 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("]") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 6621 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("]") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 6622 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("]") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 6623 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("]") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 6624 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("]") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 6625 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("]") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 6626 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("]") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 6627 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("]") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 6628 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("]") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 6629 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("]") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 6630 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("]") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 6631 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("]") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 6632 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("]") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 6633 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("]") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 6634 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("]") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 6635 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("]") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 6636 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("]") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 6637 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("]") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 6638 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("]") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 6639 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("]") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 6640 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("]") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 6641 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("]") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 6642 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("]") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 6643 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("]") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 6644 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("]") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 6645 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("]") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 6646 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("]") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 6647 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("]") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 6648 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("]") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 6649 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("]") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 6650 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("]") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 6651 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("]") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 6652 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("]") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 6653 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("]") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 6654 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("]") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 6655 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("]") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 6656 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("]") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 6657 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("]") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 6658 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("]") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 6659 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("]") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 6660 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("]") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 6661 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("]") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 6662 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("]") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 6663 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("]") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 6664 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("]") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 6665 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("]") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 6666 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) ((",") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_DOUBLE_ARROW) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("]") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 6667 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("]") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 6668 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("]") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 6669 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("]") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 6670 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("]") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 6671 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("]") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 6672 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("]") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 6673 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("]") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 6674 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("]") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 6675 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("]") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 6676 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("]") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 6677 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("]") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 6679 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("]") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 6680 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("]") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 6681 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 6683 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("]") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 6684 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("]") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 6688 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("]") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 6689 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("]") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 6694 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) ((",") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_DOUBLE_ARROW) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("]") reduce 377) (("^") reduce 377) (("|") reduce 377)) 6696 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) ((",") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_DOUBLE_ARROW) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("]") reduce 375) (("^") reduce 375) (("|") reduce 375)) 6706 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 371) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 371) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 371) (("^") shift) (("|") shift)) 6707 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 373) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 373) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 373) (("^") shift) (("|") shift)) 6708 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 372) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 372) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 372) (("^") shift) (("|") shift)) 6709 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 365) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 365) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 365) (("^") shift) (("|") shift)) 6710 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 362) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 362) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 362) (("^") shift) (("|") shift)) 6711 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 370) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 370) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 370) (("^") shift) (("|") shift)) 6712 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 364) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 364) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 364) (("^") shift) (("|") shift)) 6713 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 368) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 368) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 368) (("^") shift) (("|") shift)) 6714 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 363) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 363) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 363) (("^") shift) (("|") shift)) 6715 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 366) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 366) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 366) (("^") shift) (("|") shift)) 6716 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 367) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 367) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 367) (("^") shift) (("|") shift)) 6717 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 374) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 374) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 374) (("^") shift) (("|") shift)) 6718 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 369) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 369) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 369) (("^") shift) (("|") shift)) 6720 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 359) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 359) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 359) (("^") shift) (("|") shift)) 6722 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 6725 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("]") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 6727 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("]") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 6729 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 6730 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("]") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 6731 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) ((",") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_DOUBLE_ARROW) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("]") reduce 360) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 6735 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 6737 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("]") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 6738 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("]") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 6742 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("]") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 6743 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("]") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 6744 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("]") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 6747 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("]") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 6748 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("]") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 6749 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("]") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 6755 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 6757 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 6758 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("]") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 6759 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("]") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 6760 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 6763 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 6764 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 6765 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("]") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 6766 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("]") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 6767 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 6770 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 6772 (((",") reduce 536) (("]") reduce 536)) 6773 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("]") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 6784 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("]") reduce 444) (("^") reduce 444) (("|") reduce 444)) 6785 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("]") reduce 439) (("^") reduce 439) (("|") reduce 439)) 6856 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 444) (("^") shift) (("|") shift)) 7169 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("]") reduce 440) (("^") reduce 440) (("|") reduce 440)) 7172 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("]") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 7174 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("]") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 7175 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("]") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 7176 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("]") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 7178 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) ((",") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("]") reduce 435) (("^") reduce 435) (("|") reduce 435)) 7179 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("]") reduce 437) (("^") reduce 437) (("|") reduce 437)) 7185 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 7187 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 7188 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("]") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 7189 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("]") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 7190 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 7193 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 7194 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 7195 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("]") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 7196 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("]") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 7198 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) ((",") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("]") reduce 427) (("^") reduce 427) (("|") reduce 427)) 7200 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("]") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 7202 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 358) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 358) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 358) (("^") shift) (("|") shift)) 7203 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 432) (("^") shift) (("|") shift)) 7204 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 430) (("^") shift) (("|") shift)) 7206 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 431) (("^") shift) (("|") shift)) 7207 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("]") reduce 423) (("^") reduce 423) (("|") reduce 423)) 7208 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 433) (("^") shift) (("|") shift)) 7209 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) ((",") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("]") reduce 419) (("^") reduce 419) (("|") reduce 419)) 7210 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) ((",") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("]") reduce 436) (("^") reduce 436) (("|") reduce 436)) 7212 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("]") reduce 477) (("^") reduce 477) (("|") reduce 477)) 7214 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) ((",") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("]") reduce 478) (("^") reduce 478) (("|") reduce 478)) 7215 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("]") reduce 476) (("^") reduce 476) (("|") reduce 476)) 7216 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 567) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 567) (("^") shift) (("|") shift)) 7217 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 566) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 566) (("^") shift) (("|") shift)) 7218 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 428) (("^") shift) (("|") shift)) 7219 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("]") reduce 421) (("^") reduce 421) (("|") reduce 421)) 7222 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 7223 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 7224 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 7225 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 7226 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("]") reduce 91) (("^") reduce 91) (("|") reduce 91)) 7227 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) ((",") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_DOUBLE_ARROW) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("]") reduce 354) (("^") reduce 354) (("|") reduce 354)) 7229 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 7230 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("]") reduce 468) (("^") reduce 468) (("|") reduce 468)) 7231 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 7232 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("]") reduce 461) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 7233 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 7240 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 7242 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 7243 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("]") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 7244 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("]") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 7245 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("]") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 7247 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("]") reduce 469) (("^") reduce 469) (("|") reduce 469)) 7248 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) ((",") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_DOUBLE_ARROW) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("]") reduce 353) (("^") reduce 353) (("|") reduce 353)) 7249 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("]") reduce 281) (("^") reduce 281) (("|") reduce 281)) 7253 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("]") reduce 282) (("^") reduce 282) (("|") reduce 282)) 7254 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("]") reduce 283) (("^") reduce 283) (("|") reduce 283)) 7256 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("]") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 7257 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) ((",") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_DOUBLE_ARROW) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("]") reduce 355) (("^") reduce 355) (("|") reduce 355)) 7264 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) ((",") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_DOUBLE_ARROW) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("]") reduce 352) (("^") reduce 352) (("|") reduce 352)) 7266 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("]") reduce 462) (("^") reduce 462) (("|") reduce 462)) 7272 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) ((",") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_DOUBLE_ARROW) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("]") reduce 223) (("^") reduce 223) (("|") reduce 223)) 7281 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 7282 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) ((",") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("]") reduce 417) (("^") reduce 417) (("|") reduce 417)) 7283 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 564) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 564) (("^") shift) (("|") shift)) 7284 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 563) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 563) (("^") shift) (("|") shift)) 7285 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) ((",") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_DOUBLE_ARROW) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("]") reduce 376) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 7287 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("]") reduce 424) (("^") reduce 424) (("|") reduce 424)) 7289 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) ((",") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("]") reduce 464) (("^") reduce 464) (("|") reduce 464)) 7292 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 7295 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("]") reduce 562) (("^") reduce 562) (("|") reduce 562)) 7296 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) ((",") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("]") reduce 418) (("^") reduce 418) (("|") reduce 418)) 7297 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) ((",") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_DOUBLE_ARROW) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("]") reduce 378) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 7298 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) ((",") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_DOUBLE_ARROW) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("]") reduce 361) (("^") reduce 361) (("|") reduce 361)) 7299 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("]") reduce 422) (("^") reduce 422) (("|") reduce 422)) 7300 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) ((",") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("]") reduce 420) (("^") reduce 420) (("|") reduce 420)) 7303 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("]") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 7304 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("]") reduce 425) (("^") reduce 425) (("|") reduce 425)) 7305 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("]") reduce 398) (("^") reduce 398) (("|") reduce 398)) 7306 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) ((",") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("]") reduce 397) (("^") reduce 397) (("|") reduce 397)) 7308 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) ((",") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("]") reduce 411) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 7309 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 7312 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 7314 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("]") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 7315 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("]") reduce 399) (("^") reduce 399) (("|") reduce 399)) 7327 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("]") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 7329 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("]") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 7331 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("]") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 7332 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("]") reduce 475) (("^") reduce 475) (("|") reduce 475)) 7336 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) ((",") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("]") reduce 463) (("^") reduce 463) (("|") reduce 463)) 7337 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("]") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 7338 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("]") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 7344 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) ((",") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("]") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 7346 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("]") reduce 474) (("^") reduce 474) (("|") reduce 474)) 7348 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("]") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 7349 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("]") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 7350 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("]") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 7351 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("]") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 7353 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("]") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 7360 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("]") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 7363 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("]") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 7365 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 7366 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) ((",") reduce 429) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 429) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 7374 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("]") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 7375 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) ((",") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("]") reduce 481) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 7377 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) ((",") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("]") reduce 480) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 7378 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) ((",") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("]") reduce 479) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 7379 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 538) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 538) (("^") shift) (("|") shift)) 7382 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("]") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 7383 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) ((",") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("]") reduce 434) (("^") reduce 434) (("|") reduce 434)) 7384 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) ((",") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("]") reduce 416) (("^") reduce 416) (("|") reduce 416)) 7385 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("]") reduce 438) (("^") reduce 438) (("|") reduce 438)) 7386 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) ((",") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("]") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 7387 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) ((",") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("]") reduce 412) (("^") reduce 412) (("|") reduce 412)) 7388 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("]") reduce 426) (("^") reduce 426) (("|") reduce 426)) 7389 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 7390 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("]") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 7391 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) ((",") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("]") reduce 356) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 7395 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("]") reduce 400) (("^") reduce 400) (("|") reduce 400)) 7426 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 384) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 384) (("^") shift) (("|") reduce 384)) 7427 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 387) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 387) (("^") reduce 387) (("|") reduce 387)) 7428 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("]") reduce 396) (("^") reduce 396) (("|") reduce 396)) 7429 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 409) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 409) (("^") reduce 409) (("|") reduce 409)) 7430 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("]") reduce 395) (("^") reduce 395) (("|") reduce 395)) 7431 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) ((",") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("]") reduce 392) (("^") reduce 392) (("|") reduce 392)) 7432 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 383) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 383) (("^") shift) (("|") shift)) 7433 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 381) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 381) (("^") shift) (("|") shift)) 7434 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 382) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 382) (("^") shift) (("|") shift)) 7435 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("]") reduce 406) (("^") reduce 406) (("|") reduce 406)) 7436 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 402) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 402) (("^") reduce 402) (("|") reduce 402)) 7437 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 404) (("^") reduce 404) (("|") reduce 404)) 7438 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 401) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 401) (("^") reduce 401) (("|") reduce 401)) 7439 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 408) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("]") reduce 408) (("^") reduce 408) (("|") reduce 408)) 7440 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 403) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 403) (("^") reduce 403) (("|") reduce 403)) 7443 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 7444 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 7445 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 7446 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 7447 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("]") reduce 91) (("^") reduce 91) (("|") reduce 91)) 7448 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 7449 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 7450 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) ((",") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("]") reduce 410) (("^") reduce 410) (("|") reduce 410)) 7451 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 7452 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("]") reduce 461) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 7453 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 7460 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 7462 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 7463 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("]") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 7464 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("]") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 7465 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("]") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 7466 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 7469 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 7470 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("]") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 7472 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("]") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 7474 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("]") reduce 462) (("^") reduce 462) (("|") reduce 462)) 7475 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 7478 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 7479 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 415) (("^") shift) (("|") shift)) 7480 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 379) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 379) (("^") shift) (("|") shift)) 7481 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 380) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 380) (("^") shift) (("|") shift)) 7482 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 385) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 385) (("^") reduce 385) (("|") reduce 385)) 7483 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 386) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 386) (("^") reduce 386) (("|") reduce 386)) 7487 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 413) (("^") shift) (("|") shift)) 7488 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 414) (("^") shift) (("|") shift)) 7489 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 407) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("]") reduce 407) (("^") reduce 407) (("|") reduce 407)) 7490 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("]") reduce 405) (("^") reduce 405) (("|") reduce 405)) 7491 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) ((",") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("]") reduce 393) (("^") reduce 393) (("|") reduce 393)) 7492 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 388) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("]") reduce 388) (("^") reduce 388) (("|") reduce 388)) 7493 ((("%") shift) (("*") shift) (("+") reduce 390) ((",") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("]") reduce 390) (("^") reduce 390) (("|") reduce 390)) 7494 ((("%") shift) (("*") shift) (("+") reduce 389) ((",") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("]") reduce 389) (("^") reduce 389) (("|") reduce 389)) 7495 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) ((",") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("]") reduce 391) (("^") reduce 391) (("|") reduce 391)) 7496 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) ((",") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("]") reduce 394) (("^") reduce 394) (("|") reduce 394)) 7501 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 357) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 357) (("^") shift) (("|") shift)) 7503 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("]") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 7504 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("]") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 7505 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("]") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 7506 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("]") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 7507 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("]") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 7508 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("]") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 7509 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("]") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 7510 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("]") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 7511 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("]") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 7512 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("]") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 7513 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("]") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 7514 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("]") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 7515 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("]") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 7516 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("]") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 7517 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("]") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 7518 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("]") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 7519 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("]") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 7520 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("]") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 7521 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("]") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 7522 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("]") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 7523 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("]") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 7524 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("]") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 7525 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("]") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 7526 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("]") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 7527 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("]") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 7528 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("]") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 7529 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("]") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 7530 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("]") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 7531 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("]") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 7532 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("]") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 7533 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("]") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 7534 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("]") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 7535 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("]") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 7536 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("]") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 7537 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("]") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 7538 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("]") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 7539 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("]") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 7540 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("]") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 7541 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("]") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 7542 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("]") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 7543 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("]") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 7544 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("]") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 7545 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("]") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 7546 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("]") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 7547 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("]") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 7548 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("]") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 7549 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("]") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 7550 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("]") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 7551 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("]") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 7552 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("]") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 7553 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("]") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 7554 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("]") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 7555 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("]") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 7556 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("]") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 7557 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("]") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 7558 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("]") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 7559 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("]") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 7560 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("]") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 7561 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("]") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 7562 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("]") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 7563 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("]") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 7564 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("]") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 7565 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("]") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 7566 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("]") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 7567 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("]") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 7568 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("]") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 7569 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("]") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 7570 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) ((",") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("]") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 7571 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("]") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 7572 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("]") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 7573 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("]") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 7574 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("]") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 7575 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("]") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 7576 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("]") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 7577 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("]") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 7578 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("]") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 7579 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("]") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 7580 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("]") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 7581 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) ((",") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("]") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 7583 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("]") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 7584 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("]") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 7585 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 7587 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("]") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 7588 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("]") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 7592 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("]") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 7593 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("]") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 7598 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) ((",") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("]") reduce 377) (("^") reduce 377) (("|") reduce 377)) 7600 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) ((",") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("]") reduce 375) (("^") reduce 375) (("|") reduce 375)) 7610 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 371) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 371) (("^") shift) (("|") shift)) 7611 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 373) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 373) (("^") shift) (("|") shift)) 7612 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 372) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 372) (("^") shift) (("|") shift)) 7613 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 365) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 365) (("^") shift) (("|") shift)) 7614 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 362) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 362) (("^") shift) (("|") shift)) 7615 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 370) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 370) (("^") shift) (("|") shift)) 7616 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 364) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 364) (("^") shift) (("|") shift)) 7617 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 368) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 368) (("^") shift) (("|") shift)) 7618 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 363) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 363) (("^") shift) (("|") shift)) 7619 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 366) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 366) (("^") shift) (("|") shift)) 7620 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 367) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 367) (("^") shift) (("|") shift)) 7621 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 374) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 374) (("^") shift) (("|") shift)) 7622 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 369) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 369) (("^") shift) (("|") shift)) 7624 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 359) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 359) (("^") shift) (("|") shift)) 7626 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) ((",") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 7629 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("]") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 7631 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("]") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 7633 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 7634 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("]") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 7635 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) ((",") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("]") reduce 360) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 7639 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 7641 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) ((",") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("]") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 7642 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("]") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 7646 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("]") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 7647 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("]") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 7648 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("]") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 7651 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("]") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 7652 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("]") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 7653 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("]") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 7659 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 7661 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 7662 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("]") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 7663 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("]") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 7664 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 7667 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 7668 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 7669 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("]") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 7670 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("]") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 7671 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 7674 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 7675 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) ((",") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("]") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 7686 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) ((",") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("]") reduce 444) (("^") reduce 444) (("|") reduce 444)) 7687 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("]") reduce 439) (("^") reduce 439) (("|") reduce 439)) 7758 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 444) (("^") shift) (("|") shift)) 8071 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("]") reduce 440) (("^") reduce 440) (("|") reduce 440)) 8074 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) ((",") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("]") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 8076 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("]") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 8077 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) ((",") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("]") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 8078 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) ((",") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("]") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 8080 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) ((",") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("]") reduce 435) (("^") reduce 435) (("|") reduce 435)) 8081 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("]") reduce 437) (("^") reduce 437) (("|") reduce 437)) 8087 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 8089 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 8090 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("]") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 8091 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("]") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 8092 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 8095 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 8096 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 8097 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) ((",") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("]") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 8098 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("]") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 8100 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) ((",") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("]") reduce 427) (("^") reduce 427) (("|") reduce 427)) 8102 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("]") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 8104 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 358) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 358) (("^") shift) (("|") shift)) 8105 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 432) (("^") shift) (("|") shift)) 8108 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 431) (("^") shift) (("|") shift)) 8109 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("]") reduce 423) (("^") reduce 423) (("|") reduce 423)) 8110 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 433) (("^") shift) (("|") shift)) 8111 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) ((",") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("]") reduce 419) (("^") reduce 419) (("|") reduce 419)) 8112 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) ((",") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("]") reduce 436) (("^") reduce 436) (("|") reduce 436)) 8114 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("]") reduce 477) (("^") reduce 477) (("|") reduce 477)) 8116 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) ((",") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("]") reduce 478) (("^") reduce 478) (("|") reduce 478)) 8117 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("]") reduce 476) (("^") reduce 476) (("|") reduce 476)) 8118 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 567) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 567) (("^") shift) (("|") shift)) 8119 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 566) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 566) (("^") shift) (("|") shift)) 8120 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 428) (("^") shift) (("|") shift)) 8121 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) ((",") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("]") reduce 421) (("^") reduce 421) (("|") reduce 421)) 8124 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 8125 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 8126 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 8127 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) ((",") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 8128 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) ((",") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("]") reduce 91) (("^") reduce 91) (("|") reduce 91)) 8129 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) ((",") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("]") reduce 354) (("^") reduce 354) (("|") reduce 354)) 8131 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 8132 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("]") reduce 468) (("^") reduce 468) (("|") reduce 468)) 8133 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 8134 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("]") reduce 461) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 8135 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 8142 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 8144 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 8145 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("]") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 8146 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("]") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 8147 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("]") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 8149 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("]") reduce 469) (("^") reduce 469) (("|") reduce 469)) 8150 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) ((",") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("]") reduce 353) (("^") reduce 353) (("|") reduce 353)) 8151 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) ((",") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("]") reduce 281) (("^") reduce 281) (("|") reduce 281)) 8155 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) ((",") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("]") reduce 282) (("^") reduce 282) (("|") reduce 282)) 8156 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) ((",") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("]") reduce 283) (("^") reduce 283) (("|") reduce 283)) 8158 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("]") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 8159 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) ((",") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("]") reduce 355) (("^") reduce 355) (("|") reduce 355)) 8166 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) ((",") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("]") reduce 352) (("^") reduce 352) (("|") reduce 352)) 8168 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("]") reduce 462) (("^") reduce 462) (("|") reduce 462)) 8174 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) ((",") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("]") reduce 223) (("^") reduce 223) (("|") reduce 223)) 8183 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 8184 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) ((",") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("]") reduce 417) (("^") reduce 417) (("|") reduce 417)) 8185 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 564) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 564) (("^") shift) (("|") shift)) 8186 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 563) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 563) (("^") shift) (("|") shift)) 8187 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) ((",") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("]") reduce 376) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 8189 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("]") reduce 424) (("^") reduce 424) (("|") reduce 424)) 8191 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) ((",") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("]") reduce 464) (("^") reduce 464) (("|") reduce 464)) 8194 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 8197 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("]") reduce 562) (("^") reduce 562) (("|") reduce 562)) 8198 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) ((",") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("]") reduce 418) (("^") reduce 418) (("|") reduce 418)) 8199 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) ((",") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("]") reduce 378) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 8200 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) ((",") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("]") reduce 361) (("^") reduce 361) (("|") reduce 361)) 8201 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) ((",") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("]") reduce 422) (("^") reduce 422) (("|") reduce 422)) 8202 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) ((",") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("]") reduce 420) (("^") reduce 420) (("|") reduce 420)) 8205 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("]") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 8206 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("]") reduce 425) (("^") reduce 425) (("|") reduce 425)) 8207 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("]") reduce 398) (("^") reduce 398) (("|") reduce 398)) 8208 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) ((",") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("]") reduce 397) (("^") reduce 397) (("|") reduce 397)) 8210 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) ((",") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("]") reduce 411) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 8211 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 8214 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) ((",") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 8216 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("]") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 8217 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("]") reduce 399) (("^") reduce 399) (("|") reduce 399)) 8245 ((("(") reduce 515) ((",") reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) (("[") reduce 515) (("]") reduce 515) (("{") reduce 515)) 8248 ((("(") reduce 502) ((",") reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) (("[") reduce 511) (("]") reduce 511) (("{") reduce 511)) 8250 ((("(") reduce 510) ((",") reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) (("[") reduce 510) (("]") reduce 510) (("{") reduce 510)) 8252 ((("(") reduce 505) ((",") reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) (("[") reduce 505) (("]") reduce 505) (("{") reduce 505)) 8253 (((",") reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) (("[") reduce 512) (("]") reduce 512) (("{") reduce 512)) 8254 (((",") reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) (("[") reduce 496) (("]") reduce 540) (("{") reduce 496)) 8258 ((("(") reduce 529) ((",") reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) (("[") reduce 519) (("]") reduce 519) (("{") reduce 519)) 8260 ((("(") reduce 456) ((",") reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) (("[") reduce 456) (("]") reduce 456) (("{") reduce 456)) 8261 ((("(") reduce 281) ((",") reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) (("[") reduce 281) (("]") reduce 281) (("{") reduce 281)) 8265 ((("(") reduce 282) ((",") reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) (("[") reduce 282) (("]") reduce 282) (("{") reduce 282)) 8266 ((("(") reduce 283) ((",") reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) (("[") reduce 283) (("]") reduce 283) (("{") reduce 283)) 8267 ((("(") reduce 454) ((",") reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) (("[") reduce 454) (("]") reduce 454) (("{") reduce 454)) 8270 ((("(") reduce 529) ((",") reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) (("[") reduce 518) (("]") reduce 518) (("{") reduce 518)) 8271 ((("(") reduce 455) ((",") reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) (("[") reduce 455) (("]") reduce 455) (("{") reduce 455)) 8272 ((("(") reduce 457) ((",") reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) (("[") reduce 457) (("]") reduce 457) (("{") reduce 457)) 8278 ((("(") reduce 507) ((",") reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) (("[") reduce 507) (("]") reduce 507) (("{") reduce 507)) 8280 ((("(") reduce 506) ((",") reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) (("[") reduce 506) (("]") reduce 506) (("{") reduce 506)) 8281 ((("(") reduce 530) ((",") reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) (("[") reduce 530) (("]") reduce 530) (("{") reduce 530)) 8282 ((("(") shift) ((",") reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) (("[") reduce 513) (("]") reduce 513) (("{") reduce 513)) 8283 ((("(") reduce 532) ((",") reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) (("[") reduce 532) (("]") reduce 532) (("{") reduce 532)) 8286 ((("(") reduce 531) ((",") reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531) (("]") reduce 531) (("{") reduce 531)) 8287 ((("(") reduce 508) ((",") reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) (("[") reduce 508) (("]") reduce 508) (("{") reduce 508)) 8288 ((("(") shift) ((",") reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) (("[") reduce 514) (("]") reduce 514) (("{") reduce 514)) 8289 ((("(") reduce 509) ((",") reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) (("[") reduce 509) (("]") reduce 509) (("{") reduce 509)) 8290 ((("(") reduce 517) ((",") reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) (("[") reduce 517) (("]") reduce 517) (("{") reduce 517)) 8293 ((("(") reduce 516) ((",") reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) (("[") reduce 516) (("]") reduce 516) (("{") reduce 516)) 8346 (((",") reduce 543) (("=") shift) (("]") reduce 543)) 8394 ((("(") reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) (("[") reduce 471) (("{") reduce 471)) 8412 (((",") reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) (("[") reduce 496) (("]") reduce 541) (("{") reduce 496)) 8465 (((",") reduce 544) (("=") shift) (("]") reduce 544)) 8484 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 542) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 542) (("^") shift) (("|") shift)) 8507 ((("(") reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) (("[") reduce 470) (("{") reduce 470)) 8509 ((("(") reduce 503) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) (("[") reduce 497) (("{") reduce 497)) 8515 ((("(") reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) (("[") reduce 473) (("{") reduce 473)) 8881 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((":") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 8884 ((("%") shift) ((")") reduce 287) (("*") shift) (("+") shift) ((",") reduce 287) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8951 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((":") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 8972 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) ((":") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 8974 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) ((":") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 8984 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 371) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8985 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 373) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8986 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 372) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8987 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 365) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8988 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 362) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8989 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 370) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8990 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 364) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8991 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 368) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8992 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 363) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8993 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 366) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8994 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 367) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8995 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 374) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8996 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 369) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 8998 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 359) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9000 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 9003 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 9005 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 9007 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 9008 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 9009 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) ((":") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 9013 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 9015 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((":") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 9016 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((":") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 9020 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((":") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 9021 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((":") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 9022 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((":") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 9025 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 9026 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((":") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 9027 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((":") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 9033 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 9035 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 9036 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 9037 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 9038 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 9041 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 9042 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 9043 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((":") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 9044 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 9045 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 9048 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 9049 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((":") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 9060 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((":") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 9061 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((":") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 9132 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 444) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9445 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((":") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 9448 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((":") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 9450 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 9451 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((":") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 9452 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((":") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 9454 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) ((":") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 9455 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) ((":") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 9461 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 9463 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 9464 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 9465 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 9466 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 9469 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 9470 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 9471 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((":") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 9472 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 9474 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) ((":") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 9476 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((":") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 9478 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 358) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9479 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 432) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9490 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 9492 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((":") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 9494 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((":") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 9495 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((":") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 9499 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) ((":") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 9500 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((":") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 9501 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 9507 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) ((":") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 9509 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((":") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 9511 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 9512 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 9513 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 9514 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 9516 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 9523 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 9526 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((":") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 9528 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 9529 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) ((":") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 9536 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 9537 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) ((":") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 9539 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) ((":") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_DOUBLE_ARROW) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 9540 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) ((":") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_DOUBLE_ARROW) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 9541 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9544 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 9545 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) ((":") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 9546 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) ((":") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 9547 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) ((":") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 9548 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) ((":") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 9549 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) ((":") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 9550 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((":") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 9551 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 9552 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 9553 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) ((":") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 356) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 9556 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((":") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 9587 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 384) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 9588 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 387) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 9589 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) ((":") reduce 396) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 9590 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 409) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 9591 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) ((":") reduce 395) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 9592 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((":") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 9593 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 383) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9594 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 381) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 381) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9595 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 382) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 382) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9596 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 406) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 9597 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 402) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 9598 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 404) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 9599 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 401) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 9600 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 408) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 9601 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 403) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 9604 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 9605 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 9606 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 9607 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((":") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 9608 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 9609 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 9610 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 9611 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((":") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 9612 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 9613 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((":") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 9614 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 9621 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 9623 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 9624 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 9625 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 9626 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 9627 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 9630 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 9631 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 9633 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 9635 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 9636 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 9639 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 9640 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 415) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9641 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 379) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_DOUBLE_ARROW) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9642 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 380) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_DOUBLE_ARROW) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9643 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 385) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 9644 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 386) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 9648 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 413) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9649 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 414) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9650 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 407) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 9651 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 405) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 9652 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((":") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 9653 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) ((":") reduce 388) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_DOUBLE_ARROW) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 9654 ((("%") shift) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) ((":") reduce 390) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 9655 ((("%") shift) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) ((":") reduce 389) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 9656 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) ((":") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 9657 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) ((":") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 9659 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((":") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 9660 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((":") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 9661 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((":") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 9662 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((":") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 9663 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((":") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 9664 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((":") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 9665 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((":") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 9666 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((":") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 9667 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((":") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 9668 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((":") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 9669 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((":") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 9670 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((":") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 9671 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((":") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 9672 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((":") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 9673 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((":") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 9674 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((":") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 9675 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((":") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 9676 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((":") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 9677 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((":") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 9678 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((":") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 9679 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((":") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 9680 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((":") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 9681 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((":") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 9682 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((":") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 9683 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((":") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 9684 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((":") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 9685 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((":") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 9686 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((":") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 9687 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((":") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 9688 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((":") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 9689 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((":") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 9690 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((":") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 9691 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((":") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 9692 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((":") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 9693 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((":") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 9694 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((":") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 9695 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((":") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 9696 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((":") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 9697 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((":") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 9698 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((":") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 9699 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((":") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 9700 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((":") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 9701 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((":") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 9702 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((":") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 9703 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((":") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 9704 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((":") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 9705 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((":") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 9706 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((":") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 9707 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((":") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 9708 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((":") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 9709 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((":") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 9710 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((":") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 9711 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((":") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 9712 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((":") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 9713 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((":") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 9714 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((":") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 9715 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((":") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 9716 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((":") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 9717 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((":") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 9718 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((":") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 9719 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((":") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 9720 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((":") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 9721 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((":") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 9722 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((":") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 9723 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((":") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 9724 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((":") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 9725 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((":") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 9726 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) ((":") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_DOUBLE_ARROW) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 9727 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((":") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 9728 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((":") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 9729 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((":") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 9730 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((":") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 9731 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((":") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 9732 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((":") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 9733 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((":") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 9734 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((":") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 9735 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((":") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 9736 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((":") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 9737 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((":") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 9739 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((":") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 9740 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((":") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 9741 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 9743 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((":") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 9744 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((":") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 9748 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((":") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 9749 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((":") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 9754 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) ((":") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_DOUBLE_ARROW) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 9756 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) ((":") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_DOUBLE_ARROW) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 9766 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 371) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 371) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9767 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 373) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 373) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9768 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 372) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 372) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9769 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 365) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 365) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9770 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 362) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 362) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9771 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 370) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 370) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9772 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 364) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 364) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9773 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 368) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 368) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9774 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 363) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 363) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9775 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 366) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 366) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9776 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 367) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 367) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9777 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 374) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 374) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9778 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 369) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 369) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9780 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 359) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 359) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9782 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 9785 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 9787 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 9789 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 9790 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 9791 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) ((":") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_DOUBLE_ARROW) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 9795 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 9797 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((":") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 9798 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((":") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 9802 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((":") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 9803 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((":") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 9804 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((":") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 9807 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 9808 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((":") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 9809 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((":") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 9815 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 9817 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 9818 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 9819 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 9820 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 9823 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 9824 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 9825 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((":") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 9826 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 9827 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 9830 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 9831 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((":") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 9842 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((":") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 9843 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((":") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 9853 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 444) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9854 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((":") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 9856 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9858 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((":") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 9860 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 9861 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((":") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 9862 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((":") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 9864 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) ((":") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 9865 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) ((":") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 9871 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 9873 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 9874 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 9875 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 9876 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 9879 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 9880 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 9881 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((":") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 9882 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 9884 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) ((":") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 9886 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((":") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 9888 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 358) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 358) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9889 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 432) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9892 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9893 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((":") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 9894 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 433) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9895 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) ((":") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 9896 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) ((":") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 9898 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((":") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 9900 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) ((":") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 9901 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((":") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 9902 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 567) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9903 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 566) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9904 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 428) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9905 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) ((":") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 9908 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 9909 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 9910 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 9911 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((":") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 9912 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 9913 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) ((":") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_DOUBLE_ARROW) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 9915 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 9916 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((":") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 9917 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 9918 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((":") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 9919 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 9926 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 9928 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 9929 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 9930 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 9931 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 9933 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((":") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 9934 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) ((":") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_DOUBLE_ARROW) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 9935 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((":") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 9939 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((":") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 9940 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((":") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 9942 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 9943 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) ((":") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_DOUBLE_ARROW) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 9950 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) ((":") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_DOUBLE_ARROW) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 9952 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 9958 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) ((":") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_DOUBLE_ARROW) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 9963 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 357) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 357) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9967 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 9968 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) ((":") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 9969 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 564) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9970 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 563) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 9971 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) ((":") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_DOUBLE_ARROW) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 9973 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((":") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 9975 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) ((":") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 9978 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 9981 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((":") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 9982 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) ((":") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 9983 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) ((":") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_DOUBLE_ARROW) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 9984 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) ((":") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_DOUBLE_ARROW) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 9985 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) ((":") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 9986 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) ((":") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 9989 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((":") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 9990 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((":") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 9991 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((":") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 9992 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) ((":") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 9994 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) ((":") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 9995 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 9998 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 10000 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((":") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 10001 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((":") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 10002 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((":") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 10003 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 433) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10004 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) ((":") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 10005 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) ((":") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 10007 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((":") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 10009 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) ((":") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 10010 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((":") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 10011 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 567) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10012 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 566) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10013 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 428) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10014 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) ((":") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 10017 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 10018 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 10019 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 10020 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((":") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 10021 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 10022 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) ((":") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 10024 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 10025 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((":") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 10026 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 10027 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((":") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 10028 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 10035 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 10037 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 10038 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 10039 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 10040 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 10042 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((":") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 10043 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) ((":") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 10044 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((":") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 10048 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((":") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 10049 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((":") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 10051 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 10052 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) ((":") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 10059 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) ((":") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 10061 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 10067 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) ((":") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 10072 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 357) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10076 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 10077 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) ((":") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 10078 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 564) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10079 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 563) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10080 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) ((":") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 10082 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((":") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 10084 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) ((":") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 10087 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 10090 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((":") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 10091 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) ((":") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 10092 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) ((":") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 10093 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) ((":") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 10094 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) ((":") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 10095 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) ((":") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 10098 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((":") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 10099 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((":") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 10100 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((":") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 10101 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) ((":") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 10103 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) ((":") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 10104 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 10107 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 10109 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((":") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 10110 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((":") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 10112 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 413) (("^") shift) (("|") shift)) 10714 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 414) (("^") shift) (("|") shift)) 10726 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("]") reduce 407) (("^") reduce 407) (("|") reduce 407)) 10727 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("]") reduce 405) (("^") reduce 405) (("|") reduce 405)) 10728 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("]") reduce 393) (("^") reduce 393) (("|") reduce 393)) 10729 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("]") reduce 388) (("^") reduce 388) (("|") reduce 388)) 10730 ((("%") shift) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("]") reduce 390) (("^") reduce 390) (("|") reduce 390)) 10731 ((("%") shift) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("]") reduce 389) (("^") reduce 389) (("|") reduce 389)) 10732 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("]") reduce 391) (("^") reduce 391) (("|") reduce 391)) 10733 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("]") reduce 394) (("^") reduce 394) (("|") reduce 394)) 10735 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("]") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 10736 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("]") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 10737 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("]") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 10738 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("]") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 10739 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("]") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 10740 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("]") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 10741 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("]") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 10742 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("]") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 10743 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("]") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 10744 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("]") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 10745 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("]") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 10746 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("]") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 10747 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("]") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 10748 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("]") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 10749 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("]") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 10750 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("]") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 10751 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("]") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 10752 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("]") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 10753 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("]") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 10754 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("]") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 10755 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("]") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 10756 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("]") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 10757 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("]") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 10758 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("]") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 10759 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("]") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 10760 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("]") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 10761 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("]") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 10762 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("]") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 10763 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("]") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 10764 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("]") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 10765 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("]") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 10766 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("]") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 10767 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("]") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 10768 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("]") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 10769 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("]") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 10770 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("]") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 10771 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("]") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 10772 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("]") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 10773 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("]") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 10774 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("]") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 10775 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("]") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 10776 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("]") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 10777 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("]") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 10778 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("]") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 10779 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("]") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 10780 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("]") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 10781 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("]") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 10782 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("]") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 10783 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("]") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 10784 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("]") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 10785 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("]") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 10786 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("]") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 10787 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("]") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 10788 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("]") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 10789 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("]") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 10790 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("]") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 10791 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("]") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 10792 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("]") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 10793 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("]") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 10794 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("]") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 10795 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("]") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 10796 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("]") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 10797 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("]") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 10798 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("]") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 10799 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("]") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 10800 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("]") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 10801 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("]") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 10802 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("]") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 10803 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("]") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 10804 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("]") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 10805 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("]") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 10806 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("]") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 10807 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("]") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 10808 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("]") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 10809 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("]") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 10810 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("]") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 10811 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("]") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 10812 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("]") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 10813 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("]") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 10815 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("]") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 10816 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("]") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 10817 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 10819 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("]") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 10820 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("]") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 10824 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("]") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 10825 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("]") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 10830 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("]") reduce 377) (("^") reduce 377) (("|") reduce 377)) 10832 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("]") reduce 375) (("^") reduce 375) (("|") reduce 375)) 10842 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 371) (("^") shift) (("|") shift)) 10843 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 373) (("^") shift) (("|") shift)) 10844 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 372) (("^") shift) (("|") shift)) 10845 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 365) (("^") shift) (("|") shift)) 10846 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 362) (("^") shift) (("|") shift)) 10847 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 370) (("^") shift) (("|") shift)) 10848 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 364) (("^") shift) (("|") shift)) 10849 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 368) (("^") shift) (("|") shift)) 10850 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 363) (("^") shift) (("|") shift)) 10851 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 366) (("^") shift) (("|") shift)) 10852 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 367) (("^") shift) (("|") shift)) 10853 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 374) (("^") shift) (("|") shift)) 10854 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 369) (("^") shift) (("|") shift)) 10856 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 359) (("^") shift) (("|") shift)) 10858 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 10861 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("]") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 10863 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("]") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 10865 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 10866 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("]") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 10867 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("]") reduce 360) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 10871 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 10873 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("]") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 10874 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("]") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 10878 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("]") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 10879 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("]") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 10880 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("]") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 10883 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("]") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 10884 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("]") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 10885 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("]") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 10891 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 10893 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 10894 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("]") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 10895 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("]") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 10896 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 10899 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 10900 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 10901 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("]") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 10902 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("]") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 10903 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 10906 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 10907 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521) (("}") reduce 521)) 10908 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("]") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 10919 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("]") reduce 444) (("^") reduce 444) (("|") reduce 444)) 10920 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("]") reduce 439) (("^") reduce 439) (("|") reduce 439)) 10991 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 444) (("^") shift) (("|") shift)) 11304 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("]") reduce 440) (("^") reduce 440) (("|") reduce 440)) 11307 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("]") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 11309 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("]") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 11310 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("]") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 11311 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("]") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 11313 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("]") reduce 435) (("^") reduce 435) (("|") reduce 435)) 11314 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("]") reduce 437) (("^") reduce 437) (("|") reduce 437)) 11320 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 11322 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 11323 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("]") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 11324 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("]") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 11325 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 11328 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 11329 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 11330 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("]") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 11331 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("]") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 11333 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("]") reduce 427) (("^") reduce 427) (("|") reduce 427)) 11335 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("]") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 11337 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 358) (("^") shift) (("|") shift)) 11338 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 432) (("^") shift) (("|") shift)) 11349 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("]") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 11351 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("]") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 11353 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("]") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 11354 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("]") reduce 475) (("^") reduce 475) (("|") reduce 475)) 11358 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("]") reduce 463) (("^") reduce 463) (("|") reduce 463)) 11359 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("]") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 11360 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("]") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 11366 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("]") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 11368 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("]") reduce 474) (("^") reduce 474) (("|") reduce 474)) 11370 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("]") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 11371 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("]") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 11372 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("]") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 11373 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("]") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 11375 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("]") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 11382 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("]") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 11385 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("]") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 11387 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 11388 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 429) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 11395 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("]") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 11396 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("]") reduce 481) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 11398 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_DOUBLE_ARROW) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("]") reduce 480) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 11399 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_DOUBLE_ARROW) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("]") reduce 479) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 11400 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 430) (("^") shift) (("|") shift)) 11403 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("]") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 11404 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("]") reduce 434) (("^") reduce 434) (("|") reduce 434)) 11405 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("]") reduce 416) (("^") reduce 416) (("|") reduce 416)) 11406 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("]") reduce 438) (("^") reduce 438) (("|") reduce 438)) 11407 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("]") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 11408 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("]") reduce 412) (("^") reduce 412) (("|") reduce 412)) 11409 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("]") reduce 426) (("^") reduce 426) (("|") reduce 426)) 11410 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 11411 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("]") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 11412 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 356) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("]") reduce 356) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 11415 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("]") reduce 400) (("^") reduce 400) (("|") reduce 400)) 11446 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 384) (("^") shift) (("|") reduce 384)) 11447 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 387) (("^") reduce 387) (("|") reduce 387)) 11448 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("]") reduce 396) (("^") reduce 396) (("|") reduce 396)) 11449 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 409) (("^") reduce 409) (("|") reduce 409)) 11450 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("]") reduce 395) (("^") reduce 395) (("|") reduce 395)) 11451 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("]") reduce 392) (("^") reduce 392) (("|") reduce 392)) 11452 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 383) (("^") shift) (("|") shift)) 11453 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 381) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 381) (("^") shift) (("|") shift)) 11454 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 382) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 382) (("^") shift) (("|") shift)) 11455 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("]") reduce 406) (("^") reduce 406) (("|") reduce 406)) 11456 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 402) (("^") reduce 402) (("|") reduce 402)) 11457 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 404) (("^") reduce 404) (("|") reduce 404)) 11458 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 401) (("^") reduce 401) (("|") reduce 401)) 11459 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("]") reduce 408) (("^") reduce 408) (("|") reduce 408)) 11460 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 403) (("^") reduce 403) (("|") reduce 403)) 11463 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 11464 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 11465 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 11466 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 11467 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("]") reduce 91) (("^") reduce 91) (("|") reduce 91)) 11468 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 11469 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 11470 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("]") reduce 410) (("^") reduce 410) (("|") reduce 410)) 11471 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 11472 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("]") reduce 461) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 11473 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 11480 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 11482 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 11483 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("]") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 11484 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("]") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 11485 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("]") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 11486 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 11489 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 11490 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("]") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 11492 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("]") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 11494 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("]") reduce 462) (("^") reduce 462) (("|") reduce 462)) 11495 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 11498 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 11499 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 415) (("^") shift) (("|") shift)) 11500 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_DOUBLE_ARROW) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 379) (("^") shift) (("|") shift)) 11501 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_DOUBLE_ARROW) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 380) (("^") shift) (("|") shift)) 11502 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 385) (("^") reduce 385) (("|") reduce 385)) 11503 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 386) (("^") reduce 386) (("|") reduce 386)) 11507 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 413) (("^") shift) (("|") shift)) 11508 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 414) (("^") shift) (("|") shift)) 11509 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("]") reduce 407) (("^") reduce 407) (("|") reduce 407)) 11510 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("]") reduce 405) (("^") reduce 405) (("|") reduce 405)) 11511 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("]") reduce 393) (("^") reduce 393) (("|") reduce 393)) 11512 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_DOUBLE_ARROW) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("]") reduce 388) (("^") reduce 388) (("|") reduce 388)) 11513 ((("%") shift) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("]") reduce 390) (("^") reduce 390) (("|") reduce 390)) 11514 ((("%") shift) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("]") reduce 389) (("^") reduce 389) (("|") reduce 389)) 11515 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("]") reduce 391) (("^") reduce 391) (("|") reduce 391)) 11516 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("]") reduce 394) (("^") reduce 394) (("|") reduce 394)) 11518 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("]") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 11519 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("]") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 11520 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("]") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 11521 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("]") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 11522 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("]") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 11523 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("]") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 11524 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("]") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 11525 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("]") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 11526 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("]") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 11527 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("]") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 11528 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("]") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 11529 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("]") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 11530 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("]") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 11531 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("]") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 11532 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("]") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 11533 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("]") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 11534 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("]") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 11535 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("]") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 11536 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("]") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 11537 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("]") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 11538 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("]") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 11539 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("]") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 11540 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("]") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 11541 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("]") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 11542 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("]") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 11543 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("]") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 11544 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("]") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 11545 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("]") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 11546 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("]") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 11547 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("]") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 11548 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("]") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 11549 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("]") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 11550 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("]") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 11551 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("]") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 11552 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("]") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 11553 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("]") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 11554 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("]") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 11555 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("]") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 11556 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("]") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 11557 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("]") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 11558 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("]") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 11559 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("]") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 11560 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("]") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 11561 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("]") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 11562 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("]") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 11563 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("]") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 11564 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("]") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 11565 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("]") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 11566 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("]") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 11567 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("]") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 11568 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("]") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 11569 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("]") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 11570 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("]") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 11571 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("]") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 11572 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("]") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 11573 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("]") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 11574 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("]") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 11575 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("]") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 11576 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("]") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 11577 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("]") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 11578 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("]") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 11579 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("]") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 11580 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("]") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 11581 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("]") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 11582 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("]") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 11583 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("]") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 11584 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("]") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 11585 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_DOUBLE_ARROW) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("]") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 11586 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("]") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 11587 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("]") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 11588 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("]") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 11589 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("]") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 11590 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("]") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 11591 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("]") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 11592 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("]") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 11593 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("]") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 11594 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("]") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 11595 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("]") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 11596 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("]") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 11598 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("]") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 11599 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("]") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 11600 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 11602 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("]") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 11603 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("]") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 11607 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("]") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 11608 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("]") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 11613 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_DOUBLE_ARROW) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("]") reduce 377) (("^") reduce 377) (("|") reduce 377)) 11615 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_DOUBLE_ARROW) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("]") reduce 375) (("^") reduce 375) (("|") reduce 375)) 11625 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 371) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 371) (("^") shift) (("|") shift)) 11626 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 373) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 373) (("^") shift) (("|") shift)) 11627 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 372) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 372) (("^") shift) (("|") shift)) 11628 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 365) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 365) (("^") shift) (("|") shift)) 11629 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 362) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 362) (("^") shift) (("|") shift)) 11630 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 370) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 370) (("^") shift) (("|") shift)) 11631 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 364) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 364) (("^") shift) (("|") shift)) 11632 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 368) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 368) (("^") shift) (("|") shift)) 11633 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 363) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 363) (("^") shift) (("|") shift)) 11634 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 366) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 366) (("^") shift) (("|") shift)) 11635 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 367) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 367) (("^") shift) (("|") shift)) 11636 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 374) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 374) (("^") shift) (("|") shift)) 11637 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 369) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 369) (("^") shift) (("|") shift)) 11639 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 359) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 359) (("^") shift) (("|") shift)) 11641 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("]") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 11644 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("]") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 11646 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("]") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 11648 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 11649 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("]") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 11650 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_DOUBLE_ARROW) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("]") reduce 360) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 11654 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 11656 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("]") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 11657 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("]") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 11661 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("]") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 11662 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("]") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 11663 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("]") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 11666 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("]") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 11667 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("]") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 11668 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("]") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 11674 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 11676 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 11677 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("]") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 11678 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("]") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 11679 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 11682 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 11683 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 11684 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("]") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 11685 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("]") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 11686 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 11689 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 11690 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("]") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 11701 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("]") reduce 444) (("^") reduce 444) (("|") reduce 444)) 11702 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("]") reduce 439) (("^") reduce 439) (("|") reduce 439)) 11712 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 444) (("^") shift) (("|") shift)) 11713 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("]") reduce 440) (("^") reduce 440) (("|") reduce 440)) 11715 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 431) (("^") shift) (("|") shift)) 11717 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("]") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 11719 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("]") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 11720 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("]") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 11721 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("]") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 11723 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("]") reduce 435) (("^") reduce 435) (("|") reduce 435)) 11724 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("]") reduce 437) (("^") reduce 437) (("|") reduce 437)) 11730 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 11732 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 11733 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("]") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 11734 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("]") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 11735 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 11738 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 11739 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 11740 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("]") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 11741 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("]") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 11743 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("]") reduce 427) (("^") reduce 427) (("|") reduce 427)) 11745 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("]") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 11747 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 358) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 358) (("^") shift) (("|") shift)) 11748 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 432) (("^") shift) (("|") shift)) 11751 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 431) (("^") shift) (("|") shift)) 11752 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("]") reduce 423) (("^") reduce 423) (("|") reduce 423)) 11753 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 433) (("^") shift) (("|") shift)) 11754 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("]") reduce 419) (("^") reduce 419) (("|") reduce 419)) 11755 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("]") reduce 436) (("^") reduce 436) (("|") reduce 436)) 11757 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("]") reduce 477) (("^") reduce 477) (("|") reduce 477)) 11759 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("]") reduce 478) (("^") reduce 478) (("|") reduce 478)) 11760 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("]") reduce 476) (("^") reduce 476) (("|") reduce 476)) 11761 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 567) (("^") shift) (("|") shift)) 11762 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 566) (("^") shift) (("|") shift)) 11763 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 428) (("^") shift) (("|") shift)) 11764 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("]") reduce 421) (("^") reduce 421) (("|") reduce 421)) 11767 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 11768 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 11769 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 11770 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 11771 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("]") reduce 91) (("^") reduce 91) (("|") reduce 91)) 11772 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_DOUBLE_ARROW) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("]") reduce 354) (("^") reduce 354) (("|") reduce 354)) 11774 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 11775 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("]") reduce 468) (("^") reduce 468) (("|") reduce 468)) 11776 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 11777 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("]") reduce 461) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 11778 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 11785 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 11787 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 11788 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("]") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 11789 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("]") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 11790 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("]") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 11792 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("]") reduce 469) (("^") reduce 469) (("|") reduce 469)) 11793 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_DOUBLE_ARROW) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("]") reduce 353) (("^") reduce 353) (("|") reduce 353)) 11794 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("]") reduce 281) (("^") reduce 281) (("|") reduce 281)) 11798 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("]") reduce 282) (("^") reduce 282) (("|") reduce 282)) 11799 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("]") reduce 283) (("^") reduce 283) (("|") reduce 283)) 11801 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("]") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 11802 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_DOUBLE_ARROW) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("]") reduce 355) (("^") reduce 355) (("|") reduce 355)) 11809 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_DOUBLE_ARROW) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("]") reduce 352) (("^") reduce 352) (("|") reduce 352)) 11811 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("]") reduce 462) (("^") reduce 462) (("|") reduce 462)) 11817 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_DOUBLE_ARROW) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("]") reduce 223) (("^") reduce 223) (("|") reduce 223)) 11822 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 357) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 357) (("^") shift) (("|") shift)) 11826 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 11827 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("]") reduce 417) (("^") reduce 417) (("|") reduce 417)) 11828 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 564) (("^") shift) (("|") shift)) 11829 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 563) (("^") shift) (("|") shift)) 11830 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_DOUBLE_ARROW) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("]") reduce 376) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 11832 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("]") reduce 424) (("^") reduce 424) (("|") reduce 424)) 11834 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("]") reduce 464) (("^") reduce 464) (("|") reduce 464)) 11837 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 11840 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("]") reduce 562) (("^") reduce 562) (("|") reduce 562)) 11841 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("]") reduce 418) (("^") reduce 418) (("|") reduce 418)) 11842 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_DOUBLE_ARROW) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("]") reduce 378) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 11843 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_DOUBLE_ARROW) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("]") reduce 361) (("^") reduce 361) (("|") reduce 361)) 11844 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("]") reduce 422) (("^") reduce 422) (("|") reduce 422)) 11845 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("]") reduce 420) (("^") reduce 420) (("|") reduce 420)) 11848 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("]") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 11849 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("]") reduce 425) (("^") reduce 425) (("|") reduce 425)) 11850 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("]") reduce 398) (("^") reduce 398) (("|") reduce 398)) 11851 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("]") reduce 397) (("^") reduce 397) (("|") reduce 397)) 11853 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("]") reduce 411) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 11854 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 11857 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 11859 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("]") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 11860 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("]") reduce 399) (("^") reduce 399) (("|") reduce 399)) 11861 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("]") reduce 423) (("^") reduce 423) (("|") reduce 423)) 11862 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 433) (("^") shift) (("|") shift)) 11863 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("]") reduce 419) (("^") reduce 419) (("|") reduce 419)) 11864 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("]") reduce 436) (("^") reduce 436) (("|") reduce 436)) 11866 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("]") reduce 477) (("^") reduce 477) (("|") reduce 477)) 11868 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("]") reduce 478) (("^") reduce 478) (("|") reduce 478)) 11869 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("]") reduce 476) (("^") reduce 476) (("|") reduce 476)) 11870 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 567) (("^") shift) (("|") shift)) 11871 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 566) (("^") shift) (("|") shift)) 11872 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 428) (("^") shift) (("|") shift)) 11873 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("]") reduce 421) (("^") reduce 421) (("|") reduce 421)) 11876 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 11877 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 11878 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 11879 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 11880 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("]") reduce 91) (("^") reduce 91) (("|") reduce 91)) 11881 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("]") reduce 354) (("^") reduce 354) (("|") reduce 354)) 11883 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 11884 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("]") reduce 468) (("^") reduce 468) (("|") reduce 468)) 11885 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 11886 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("]") reduce 461) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 11887 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 11894 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 11896 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 11897 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("]") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 11898 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("]") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 11899 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("]") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 11901 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("]") reduce 469) (("^") reduce 469) (("|") reduce 469)) 11902 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("]") reduce 353) (("^") reduce 353) (("|") reduce 353)) 11903 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("]") reduce 281) (("^") reduce 281) (("|") reduce 281)) 11907 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("]") reduce 282) (("^") reduce 282) (("|") reduce 282)) 11908 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("]") reduce 283) (("^") reduce 283) (("|") reduce 283)) 11910 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("]") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 11911 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("]") reduce 355) (("^") reduce 355) (("|") reduce 355)) 11918 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("]") reduce 352) (("^") reduce 352) (("|") reduce 352)) 11920 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("]") reduce 462) (("^") reduce 462) (("|") reduce 462)) 11926 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("]") reduce 223) (("^") reduce 223) (("|") reduce 223)) 11931 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 357) (("^") shift) (("|") shift)) 11935 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 11936 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("]") reduce 417) (("^") reduce 417) (("|") reduce 417)) 11937 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 564) (("^") shift) (("|") shift)) 11938 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 563) (("^") shift) (("|") shift)) 11939 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("]") reduce 376) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 11941 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("]") reduce 424) (("^") reduce 424) (("|") reduce 424)) 11943 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("]") reduce 464) (("^") reduce 464) (("|") reduce 464)) 11946 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 11949 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("]") reduce 562) (("^") reduce 562) (("|") reduce 562)) 11950 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("]") reduce 418) (("^") reduce 418) (("|") reduce 418)) 11951 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("]") reduce 378) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 11952 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("]") reduce 361) (("^") reduce 361) (("|") reduce 361)) 11953 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("]") reduce 422) (("^") reduce 422) (("|") reduce 422)) 11954 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("]") reduce 420) (("^") reduce 420) (("|") reduce 420)) 11957 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("]") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 11958 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("]") reduce 425) (("^") reduce 425) (("|") reduce 425)) 11959 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("]") reduce 398) (("^") reduce 398) (("|") reduce 398)) 11960 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("]") reduce 397) (("^") reduce 397) (("|") reduce 397)) 11962 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("]") reduce 411) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 11963 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("]") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 11966 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("]") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 11968 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("]") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 11969 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("]") reduce 399) (("^") reduce 399) (("|") reduce 399)) 12582 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526) (("}") reduce 526)) 12583 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530) (("}") reduce 530)) 12584 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523) (("}") reduce 523)) 12585 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532) (("}") reduce 532)) 12588 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531) (("}") reduce 531)) 12589 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524) (("}") reduce 524)) 12591 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525) (("}") reduce 525)) 12593 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462) (("}") reduce 462)) 12594 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517) (("}") reduce 517)) 12597 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516) (("}") reduce 516)) 12598 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 415)) 12599 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 379)) 12600 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 380)) 12601 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385) (("}") reduce 385)) 12602 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386) (("}") reduce 386)) 12606 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 413)) 12607 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 414)) 12608 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407) (("}") reduce 407)) 12609 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405) (("}") reduce 405)) 12610 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393) (("}") reduce 393)) 12611 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388) (("}") reduce 388)) 12612 ((("%") shift) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390) (("}") reduce 390)) 12613 ((("%") shift) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389) (("}") reduce 389)) 12614 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391) (("}") reduce 391)) 12615 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394) (("}") reduce 394)) 12617 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73) (("}") reduce 73)) 12618 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52) (("}") reduce 52)) 12619 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27) (("}") reduce 27)) 12620 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51) (("}") reduce 51)) 12621 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53) (("}") reduce 53)) 12622 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49) (("}") reduce 49)) 12623 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29) (("}") reduce 29)) 12624 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59) (("}") reduce 59)) 12625 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60) (("}") reduce 60)) 12626 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11) (("}") reduce 11)) 12627 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42) (("}") reduce 42)) 12628 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39) (("}") reduce 39)) 12629 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25) (("}") reduce 25)) 12630 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50) (("}") reduce 50)) 12631 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66) (("}") reduce 66)) 12632 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18) (("}") reduce 18)) 12633 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17) (("}") reduce 17)) 12634 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15) (("}") reduce 15)) 12635 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14) (("}") reduce 14)) 12636 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38) (("}") reduce 38)) 12637 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26) (("}") reduce 26)) 12638 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22) (("}") reduce 22)) 12639 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24) (("}") reduce 24)) 12640 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16) (("}") reduce 16)) 12641 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48) (("}") reduce 48)) 12642 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20) (("}") reduce 20)) 12643 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70) (("}") reduce 70)) 12644 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3) (("}") reduce 3)) 12645 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12) (("}") reduce 12)) 12646 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54) (("}") reduce 54)) 12647 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65) (("}") reduce 65)) 12648 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74) (("}") reduce 74)) 12649 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30) (("}") reduce 30)) 12650 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68) (("}") reduce 68)) 12651 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21) (("}") reduce 21)) 12652 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23) (("}") reduce 23)) 12653 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41) (("}") reduce 41)) 12654 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62) (("}") reduce 62)) 12655 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34) (("}") reduce 34)) 12656 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40) (("}") reduce 40)) 12657 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13) (("}") reduce 13)) 12658 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55) (("}") reduce 55)) 12659 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1) (("}") reduce 1)) 12660 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2) (("}") reduce 2)) 12661 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9) (("}") reduce 9)) 12662 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33) (("}") reduce 33)) 12663 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58) (("}") reduce 58)) 12664 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37) (("}") reduce 37)) 12665 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64) (("}") reduce 64)) 12666 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46) (("}") reduce 46)) 12667 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8) (("}") reduce 8)) 12668 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6) (("}") reduce 6)) 12669 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7) (("}") reduce 7)) 12670 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69) (("}") reduce 69)) 12671 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63) (("}") reduce 63)) 12672 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56) (("}") reduce 56)) 12673 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10) (("}") reduce 10)) 12674 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67) (("}") reduce 67)) 12675 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44) (("}") reduce 44)) 12676 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75) (("}") reduce 75)) 12677 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76) (("}") reduce 76)) 12678 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77) (("}") reduce 77)) 12679 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78) (("}") reduce 78)) 12680 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4) (("}") reduce 4)) 12681 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5) (("}") reduce 5)) 12682 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43) (("}") reduce 43)) 12683 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72) (("}") reduce 72)) 12684 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81) (("}") reduce 81)) 12685 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47) (("}") reduce 47)) 12686 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31) (("}") reduce 31)) 12687 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57) (("}") reduce 57)) 12688 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61) (("}") reduce 61)) 12689 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28) (("}") reduce 28)) 12690 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36) (("}") reduce 36)) 12691 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32) (("}") reduce 32)) 12692 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35) (("}") reduce 35)) 12693 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19) (("}") reduce 19)) 12694 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45) (("}") reduce 45)) 12695 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492) (("}") reduce 492)) 12697 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71) (("}") reduce 71)) 12698 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82) (("}") reduce 82)) 12699 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519) (("}") reduce 519)) 12701 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456) (("}") reduce 456)) 12702 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281) (("}") reduce 281)) 12706 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282) (("}") reduce 282)) 12707 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283) (("}") reduce 283)) 12712 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377) (("}") reduce 377)) 12714 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375) (("}") reduce 375)) 12724 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 371)) 12725 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 373)) 12726 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 372)) 12727 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 365)) 12728 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 362)) 12729 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 370)) 12730 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 364)) 12731 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 368)) 12732 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 363)) 12733 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 366)) 12734 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 367)) 12735 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 374)) 12736 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 369)) 12738 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 359)) 12740 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515) (("}") reduce 515)) 12743 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511) (("}") reduce 511)) 12745 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510) (("}") reduce 510)) 12747 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505) (("}") reduce 505)) 12748 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512) (("}") reduce 512)) 12749 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360) (("}") reduce 360)) 12753 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519) (("}") reduce 519)) 12755 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456) (("}") reduce 456)) 12756 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281) (("}") reduce 281)) 12760 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282) (("}") reduce 282)) 12761 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283) (("}") reduce 283)) 12762 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454) (("}") reduce 454)) 12765 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518) (("}") reduce 518)) 12766 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455) (("}") reduce 455)) 12767 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457) (("}") reduce 457)) 12773 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507) (("}") reduce 507)) 12775 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506) (("}") reduce 506)) 12776 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530) (("}") reduce 530)) 12777 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513) (("}") reduce 513)) 12778 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532) (("}") reduce 532)) 12781 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531) (("}") reduce 531)) 12782 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508) (("}") reduce 508)) 12783 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514) (("}") reduce 514)) 12784 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509) (("}") reduce 509)) 12785 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517) (("}") reduce 517)) 12788 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516) (("}") reduce 516)) 12789 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454) (("}") reduce 454)) 12800 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444) (("}") reduce 444)) 12801 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439) (("}") reduce 439)) 12872 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 444)) 13185 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440) (("}") reduce 440)) 13188 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491) (("}") reduce 491)) 13190 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518) (("}") reduce 518)) 13191 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455) (("}") reduce 455)) 13192 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457) (("}") reduce 457)) 13194 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435) (("}") reduce 435)) 13195 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437) (("}") reduce 437)) 13201 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507) (("}") reduce 507)) 13203 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506) (("}") reduce 506)) 13204 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530) (("}") reduce 530)) 13205 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513) (("}") reduce 513)) 13206 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532) (("}") reduce 532)) 13209 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531) (("}") reduce 531)) 13210 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508) (("}") reduce 508)) 13211 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514) (("}") reduce 514)) 13212 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509) (("}") reduce 509)) 13214 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427) (("}") reduce 427)) 13216 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471) (("}") reduce 471)) 13218 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 358)) 13219 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 432)) 13230 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490) (("}") reduce 490)) 13232 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472) (("}") reduce 472)) 13234 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485) (("}") reduce 485)) 13235 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475) (("}") reduce 475)) 13239 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463) (("}") reduce 463)) 13240 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484) (("}") reduce 484)) 13241 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488) (("}") reduce 488)) 13247 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483) (("}") reduce 483)) 13249 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474) (("}") reduce 474)) 13251 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487) (("}") reduce 487)) 13252 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93) (("}") reduce 93)) 13253 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92) (("}") reduce 92)) 13254 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94) (("}") reduce 94)) 13256 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489) (("}") reduce 489)) 13263 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91) (("}") reduce 91)) 13266 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486) (("}") reduce 486)) 13268 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515) (("}") reduce 515)) 13269 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("}") reduce 429) (("~") shift)) 13276 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511) (("}") reduce 511)) 13277 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481) (("}") reduce 481)) 13279 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_DOUBLE_ARROW) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480) (("}") reduce 480)) 13280 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_DOUBLE_ARROW) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479) (("}") reduce 479)) 13281 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 430)) 13284 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510) (("}") reduce 510)) 13285 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434) (("}") reduce 434)) 13286 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416) (("}") reduce 416)) 13287 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438) (("}") reduce 438)) 13288 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482) (("}") reduce 482)) 13289 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412) (("}") reduce 412)) 13290 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426) (("}") reduce 426)) 13291 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505) (("}") reduce 505)) 13292 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512) (("}") reduce 512)) 13293 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 356) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356) (("}") reduce 356)) 13296 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400) (("}") reduce 400)) 13327 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384) (("}") reduce 384)) 13328 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387) (("}") reduce 387)) 13329 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396) (("}") reduce 396)) 13330 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409) (("}") reduce 409)) 13331 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395) (("}") reduce 395)) 13332 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392) (("}") reduce 392)) 13333 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 383)) 13334 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 381) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 381)) 13335 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 382) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 382)) 13336 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406) (("}") reduce 406)) 13337 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402) (("}") reduce 402)) 13338 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404) (("}") reduce 404)) 13339 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401) (("}") reduce 401)) 13340 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408) (("}") reduce 408)) 13341 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403) (("}") reduce 403)) 13344 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 13345 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 13346 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 13347 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 13348 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91) (("}") reduce 91)) 13349 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515) (("}") reduce 515)) 13350 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 13351 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410) (("}") reduce 410)) 13352 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 13353 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461) (("}") reduce 461)) 13354 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520) (("}") reduce 520)) 13361 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522) (("}") reduce 522)) 13363 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521) (("}") reduce 521)) 13364 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526) (("}") reduce 526)) 13365 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530) (("}") reduce 530)) 13366 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523) (("}") reduce 523)) 13367 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532) (("}") reduce 532)) 13370 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531) (("}") reduce 531)) 13371 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524) (("}") reduce 524)) 13373 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525) (("}") reduce 525)) 13375 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462) (("}") reduce 462)) 13376 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517) (("}") reduce 517)) 13379 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516) (("}") reduce 516)) 13380 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 415)) 13381 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_DOUBLE_ARROW) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 379)) 13382 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_DOUBLE_ARROW) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 380)) 13383 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385) (("}") reduce 385)) 13384 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386) (("}") reduce 386)) 13388 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 413)) 13389 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 414)) 13390 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407) (("}") reduce 407)) 13391 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405) (("}") reduce 405)) 13392 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393) (("}") reduce 393)) 13393 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_DOUBLE_ARROW) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388) (("}") reduce 388)) 13394 ((("%") shift) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390) (("}") reduce 390)) 13395 ((("%") shift) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389) (("}") reduce 389)) 13396 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391) (("}") reduce 391)) 13397 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394) (("}") reduce 394)) 13399 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73) (("}") reduce 73)) 13400 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52) (("}") reduce 52)) 13401 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27) (("}") reduce 27)) 13402 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51) (("}") reduce 51)) 13403 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53) (("}") reduce 53)) 13404 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49) (("}") reduce 49)) 13405 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29) (("}") reduce 29)) 13406 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59) (("}") reduce 59)) 13407 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60) (("}") reduce 60)) 13408 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11) (("}") reduce 11)) 13409 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42) (("}") reduce 42)) 13410 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39) (("}") reduce 39)) 13411 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25) (("}") reduce 25)) 13412 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50) (("}") reduce 50)) 13413 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66) (("}") reduce 66)) 13414 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18) (("}") reduce 18)) 13415 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17) (("}") reduce 17)) 13416 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15) (("}") reduce 15)) 13417 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14) (("}") reduce 14)) 13418 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38) (("}") reduce 38)) 13419 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26) (("}") reduce 26)) 13420 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22) (("}") reduce 22)) 13421 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24) (("}") reduce 24)) 13422 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16) (("}") reduce 16)) 13423 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48) (("}") reduce 48)) 13424 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20) (("}") reduce 20)) 13425 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70) (("}") reduce 70)) 13426 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3) (("}") reduce 3)) 13427 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12) (("}") reduce 12)) 13428 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54) (("}") reduce 54)) 13429 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65) (("}") reduce 65)) 13430 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74) (("}") reduce 74)) 13431 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30) (("}") reduce 30)) 13432 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68) (("}") reduce 68)) 13433 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21) (("}") reduce 21)) 13434 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23) (("}") reduce 23)) 13435 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41) (("}") reduce 41)) 13436 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62) (("}") reduce 62)) 13437 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34) (("}") reduce 34)) 13438 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40) (("}") reduce 40)) 13439 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13) (("}") reduce 13)) 13440 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55) (("}") reduce 55)) 13441 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1) (("}") reduce 1)) 13442 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2) (("}") reduce 2)) 13443 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9) (("}") reduce 9)) 13444 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33) (("}") reduce 33)) 13445 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58) (("}") reduce 58)) 13446 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37) (("}") reduce 37)) 13447 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64) (("}") reduce 64)) 13448 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46) (("}") reduce 46)) 13449 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8) (("}") reduce 8)) 13450 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6) (("}") reduce 6)) 13451 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7) (("}") reduce 7)) 13452 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69) (("}") reduce 69)) 13453 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63) (("}") reduce 63)) 13454 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56) (("}") reduce 56)) 13455 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10) (("}") reduce 10)) 13456 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67) (("}") reduce 67)) 13457 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44) (("}") reduce 44)) 13458 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75) (("}") reduce 75)) 13459 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76) (("}") reduce 76)) 13460 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77) (("}") reduce 77)) 13461 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78) (("}") reduce 78)) 13462 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4) (("}") reduce 4)) 13463 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5) (("}") reduce 5)) 13464 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43) (("}") reduce 43)) 13465 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72) (("}") reduce 72)) 13466 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_DOUBLE_ARROW) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81) (("}") reduce 81)) 13467 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47) (("}") reduce 47)) 13468 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31) (("}") reduce 31)) 13469 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57) (("}") reduce 57)) 13470 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61) (("}") reduce 61)) 13471 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28) (("}") reduce 28)) 13472 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36) (("}") reduce 36)) 13473 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32) (("}") reduce 32)) 13474 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35) (("}") reduce 35)) 13475 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19) (("}") reduce 19)) 13476 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45) (("}") reduce 45)) 13477 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492) (("}") reduce 492)) 13479 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71) (("}") reduce 71)) 13480 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82) (("}") reduce 82)) 13481 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519) (("}") reduce 519)) 13483 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456) (("}") reduce 456)) 13484 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281) (("}") reduce 281)) 13488 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282) (("}") reduce 282)) 13489 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283) (("}") reduce 283)) 13494 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_DOUBLE_ARROW) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377) (("}") reduce 377)) 13496 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_DOUBLE_ARROW) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375) (("}") reduce 375)) 13506 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 371) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 371)) 13507 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 373) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 373)) 13508 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 372) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 372)) 13509 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 365) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 365)) 13510 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 362) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 362)) 13511 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 370) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 370)) 13512 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 364) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 364)) 13513 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 368) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 368)) 13514 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 363) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 363)) 13515 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 366) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 366)) 13516 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 367) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 367)) 13517 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 374) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 374)) 13518 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 369) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 369)) 13520 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 359) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 359)) 13522 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515) (("}") reduce 515)) 13525 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511) (("}") reduce 511)) 13527 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510) (("}") reduce 510)) 13529 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505) (("}") reduce 505)) 13530 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512) (("}") reduce 512)) 13531 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_DOUBLE_ARROW) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360) (("}") reduce 360)) 13535 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519) (("}") reduce 519)) 13537 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456) (("}") reduce 456)) 13538 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281) (("}") reduce 281)) 13542 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282) (("}") reduce 282)) 13543 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283) (("}") reduce 283)) 13544 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454) (("}") reduce 454)) 13547 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518) (("}") reduce 518)) 13548 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455) (("}") reduce 455)) 13549 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457) (("}") reduce 457)) 13555 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507) (("}") reduce 507)) 13557 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506) (("}") reduce 506)) 13558 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530) (("}") reduce 530)) 13559 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513) (("}") reduce 513)) 13560 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532) (("}") reduce 532)) 13563 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531) (("}") reduce 531)) 13564 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508) (("}") reduce 508)) 13565 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514) (("}") reduce 514)) 13566 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509) (("}") reduce 509)) 13567 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517) (("}") reduce 517)) 13570 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516) (("}") reduce 516)) 13571 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454) (("}") reduce 454)) 13582 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444) (("}") reduce 444)) 13583 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439) (("}") reduce 439)) 13593 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 444)) 13594 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440) (("}") reduce 440)) 13596 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 431)) 13598 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491) (("}") reduce 491)) 13600 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518) (("}") reduce 518)) 13601 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455) (("}") reduce 455)) 13602 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457) (("}") reduce 457)) 13604 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435) (("}") reduce 435)) 13605 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437) (("}") reduce 437)) 13611 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507) (("}") reduce 507)) 13613 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506) (("}") reduce 506)) 13614 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530) (("}") reduce 530)) 13615 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513) (("}") reduce 513)) 13616 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532) (("}") reduce 532)) 13619 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531) (("}") reduce 531)) 13620 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508) (("}") reduce 508)) 13621 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514) (("}") reduce 514)) 13622 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509) (("}") reduce 509)) 13624 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427) (("}") reduce 427)) 13626 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471) (("}") reduce 471)) 13628 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 358) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 358)) 13629 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 432)) 13632 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 431)) 13633 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423) (("}") reduce 423)) 13634 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 433)) 13635 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419) (("}") reduce 419)) 13636 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436) (("}") reduce 436)) 13638 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477) (("}") reduce 477)) 13640 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478) (("}") reduce 478)) 13641 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476) (("}") reduce 476)) 13642 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 567)) 13643 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 566)) 13644 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 428)) 13645 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421) (("}") reduce 421)) 13648 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 13649 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 13650 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 13651 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 13652 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91) (("}") reduce 91)) 13653 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_DOUBLE_ARROW) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354) (("}") reduce 354)) 13655 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 13656 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468) (("}") reduce 468)) 13657 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 13658 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461) (("}") reduce 461)) 13659 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520) (("}") reduce 520)) 13666 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522) (("}") reduce 522)) 13668 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521) (("}") reduce 521)) 13669 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526) (("}") reduce 526)) 13670 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523) (("}") reduce 523)) 13671 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524) (("}") reduce 524)) 13673 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469) (("}") reduce 469)) 13674 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_DOUBLE_ARROW) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353) (("}") reduce 353)) 13675 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281) (("}") reduce 281)) 13679 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282) (("}") reduce 282)) 13680 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283) (("}") reduce 283)) 13682 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525) (("}") reduce 525)) 13683 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_DOUBLE_ARROW) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355) (("}") reduce 355)) 13690 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_DOUBLE_ARROW) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352) (("}") reduce 352)) 13692 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462) (("}") reduce 462)) 13698 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_DOUBLE_ARROW) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223) (("}") reduce 223)) 13703 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 357) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 357)) 13707 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 13708 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417) (("}") reduce 417)) 13709 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 564)) 13710 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 563)) 13711 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_DOUBLE_ARROW) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376) (("}") reduce 376)) 13713 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424) (("}") reduce 424)) 13715 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464) (("}") reduce 464)) 13718 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 13721 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562) (("}") reduce 562)) 13722 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418) (("}") reduce 418)) 13723 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_DOUBLE_ARROW) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378) (("}") reduce 378)) 13724 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_DOUBLE_ARROW) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361) (("}") reduce 361)) 13725 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422) (("}") reduce 422)) 13726 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420) (("}") reduce 420)) 13729 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470) (("}") reduce 470)) 13730 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425) (("}") reduce 425)) 13731 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398) (("}") reduce 398)) 13732 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397) (("}") reduce 397)) 13734 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411) (("}") reduce 411)) 13735 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517) (("}") reduce 517)) 13738 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516) (("}") reduce 516)) 13740 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473) (("}") reduce 473)) 13741 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399) (("}") reduce 399)) 13742 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423) (("}") reduce 423)) 13743 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 433)) 13744 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419) (("}") reduce 419)) 13745 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436) (("}") reduce 436)) 13747 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477) (("}") reduce 477)) 13749 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478) (("}") reduce 478)) 13750 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476) (("}") reduce 476)) 13751 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 567)) 13752 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 566)) 13753 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 428)) 13754 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421) (("}") reduce 421)) 13757 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 13758 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 13759 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 13760 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 13761 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91) (("}") reduce 91)) 13762 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354) (("}") reduce 354)) 13764 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 13765 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468) (("}") reduce 468)) 13766 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 13767 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461) (("}") reduce 461)) 13768 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520) (("}") reduce 520)) 13775 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522) (("}") reduce 522)) 13777 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521) (("}") reduce 521)) 13778 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526) (("}") reduce 526)) 13779 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523) (("}") reduce 523)) 13780 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524) (("}") reduce 524)) 13782 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469) (("}") reduce 469)) 13783 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353) (("}") reduce 353)) 13784 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281) (("}") reduce 281)) 13788 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282) (("}") reduce 282)) 13789 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283) (("}") reduce 283)) 13791 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525) (("}") reduce 525)) 13792 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355) (("}") reduce 355)) 13799 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352) (("}") reduce 352)) 13801 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462) (("}") reduce 462)) 13807 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223) (("}") reduce 223)) 13812 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 357)) 13816 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 13817 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417) (("}") reduce 417)) 13818 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 564)) 13819 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 563)) 13820 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376) (("}") reduce 376)) 13822 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424) (("}") reduce 424)) 13824 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464) (("}") reduce 464)) 13827 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 13830 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562) (("}") reduce 562)) 13831 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418) (("}") reduce 418)) 13832 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378) (("}") reduce 378)) 13833 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361) (("}") reduce 361)) 13834 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422) (("}") reduce 422)) 13835 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420) (("}") reduce 420)) 13838 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470) (("}") reduce 470)) 13839 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425) (("}") reduce 425)) 13840 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398) (("}") reduce 398)) 13841 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397) (("}") reduce 397)) 13843 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411) (("}") reduce 411)) 13844 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517) (("}") reduce 517)) 13847 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516) (("}") reduce 516)) 13849 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473) (("}") reduce 473)) 13850 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399) (("}") reduce 399)) 13851 ((("%") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 14465 ((("%") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 14466 ((("%") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 14467 ((("%") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 14468 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 14469 ((("%") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 14472 ((("%") reduce 531) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 14473 ((("%") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 14475 ((("%") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 14477 ((("%") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 14478 ((("%") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 14481 ((("%") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 14482 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14483 ((("%") shift) ((")") reduce 379) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14484 ((("%") shift) ((")") reduce 380) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14485 ((("%") shift) ((")") reduce 385) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 14486 ((("%") shift) ((")") reduce 386) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 14490 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14491 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14492 ((("%") shift) ((")") reduce 407) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 14493 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 14494 ((("%") reduce 393) ((")") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 14495 ((("%") shift) ((")") reduce 388) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 14496 ((("%") shift) ((")") reduce 390) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 14497 ((("%") shift) ((")") reduce 389) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 14498 ((("%") reduce 391) ((")") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 14499 ((("%") reduce 394) ((")") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 14501 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 14502 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 14503 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 14504 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 14505 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 14506 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 14507 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 14508 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 14509 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 14510 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 14511 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 14512 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 14513 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 14514 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 14515 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 14516 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 14517 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 14518 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 14519 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 14520 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 14521 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 14522 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 14523 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 14524 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 14525 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 14526 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 14527 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 14528 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 14529 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 14530 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 14531 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 14532 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 14533 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 14534 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 14535 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 14536 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 14537 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 14538 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 14539 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 14540 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 14541 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 14542 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 14543 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 14544 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 14545 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 14546 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 14547 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 14548 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 14549 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 14550 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 14551 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 14552 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 14553 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 14554 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 14555 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 14556 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 14557 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 14558 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 14559 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 14560 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 14561 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 14562 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 14563 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 14564 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 14565 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 14566 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 14567 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 14568 ((("%") reduce 81) (("(") reduce 81) ((")") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 14569 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 14570 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 14571 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 14572 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 14573 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 14574 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 14575 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 14576 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 14577 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 14578 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 14579 ((("%") reduce 492) (("(") reduce 527) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 14581 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 14582 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 14583 ((("%") reduce 519) (("(") reduce 529) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 14585 ((("%") reduce 456) (("(") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 14586 ((("%") reduce 281) (("(") reduce 281) ((")") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 14590 ((("%") reduce 282) (("(") reduce 282) ((")") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 14591 ((("%") reduce 283) (("(") reduce 283) ((")") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 14596 ((("%") reduce 377) ((")") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 14598 ((("%") reduce 375) ((")") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 14608 ((("%") shift) ((")") reduce 371) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14609 ((("%") shift) ((")") reduce 373) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14610 ((("%") shift) ((")") reduce 372) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14611 ((("%") shift) ((")") reduce 365) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14612 ((("%") shift) ((")") reduce 362) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14613 ((("%") shift) ((")") reduce 370) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14614 ((("%") shift) ((")") reduce 364) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14615 ((("%") shift) ((")") reduce 368) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14616 ((("%") shift) ((")") reduce 363) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14617 ((("%") shift) ((")") reduce 366) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14618 ((("%") shift) ((")") reduce 367) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14619 ((("%") shift) ((")") reduce 374) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14620 ((("%") shift) ((")") reduce 369) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14622 ((("%") shift) ((")") reduce 359) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 14624 ((("%") reduce 515) (("(") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 14627 ((("%") reduce 511) (("(") reduce 502) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 14629 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 14631 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 14632 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 14633 ((("%") reduce 360) ((")") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 14637 ((("%") reduce 519) (("(") reduce 529) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 14639 ((("%") reduce 456) (("(") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 14640 ((("%") reduce 281) (("(") reduce 281) ((")") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 14644 ((("%") reduce 282) (("(") reduce 282) ((")") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 14645 ((("%") reduce 283) (("(") reduce 283) ((")") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 14646 ((("%") reduce 454) (("(") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 14649 ((("%") reduce 518) (("(") reduce 529) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 14650 ((("%") reduce 455) (("(") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 14651 ((("%") reduce 457) (("(") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 14657 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 14659 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 14660 ((("%") reduce 530) (("(") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 14661 ((("%") reduce 513) (("(") shift) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 14662 ((("%") reduce 532) (("(") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 14665 ((("%") reduce 531) (("(") reduce 531) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 14666 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 14667 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 14668 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 14669 ((("%") reduce 517) (("(") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 14672 ((("%") reduce 516) (("(") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 14673 ((("%") reduce 454) (("(") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 14684 ((("%") reduce 444) ((")") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 14685 ((("%") reduce 439) ((")") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 14756 ((("%") shift) ((")") reduce 444) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15069 ((("%") reduce 440) ((")") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 15072 ((("%") reduce 491) (("(") reduce 527) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 15074 ((("%") reduce 518) (("(") reduce 529) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 15075 ((("%") reduce 455) (("(") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 15076 ((("%") reduce 457) (("(") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 15078 ((("%") reduce 435) ((")") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 15079 ((("%") reduce 437) ((")") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 15085 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 15087 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 15088 ((("%") reduce 530) (("(") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 15089 ((("%") reduce 513) (("(") shift) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 15090 ((("%") reduce 532) (("(") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 15093 ((("%") reduce 531) (("(") reduce 531) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 15094 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 15095 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 15096 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 15098 ((("%") reduce 427) ((")") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 15100 ((("%") reduce 471) (("(") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 15102 ((("%") shift) ((")") reduce 358) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15103 ((("%") shift) ((")") reduce 432) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15114 ((("%") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 15116 ((("%") reduce 472) (("(") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 15118 ((("%") reduce 485) ((")") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 15119 ((("%") reduce 475) ((")") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 15123 ((("%") reduce 463) (("(") shift) ((")") reduce 463) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 15124 ((("%") reduce 484) ((")") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 15125 ((("%") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 15131 ((("%") reduce 483) ((")") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 15133 ((("%") reduce 474) ((")") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 15135 ((("%") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 15136 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 15137 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 15138 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 15140 ((("%") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 15147 ((("%") reduce 91) (("(") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 15150 ((("%") reduce 486) ((")") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 15152 ((("%") reduce 515) (("(") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 15153 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) ((")") reduce 429) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 15160 ((("%") reduce 511) (("(") reduce 502) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 15161 ((("%") reduce 481) ((")") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 15163 ((("%") reduce 480) ((")") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_DOUBLE_ARROW) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 15164 ((("%") reduce 479) (("(") reduce 504) ((")") reduce 479) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_DOUBLE_ARROW) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 15165 ((("%") shift) ((")") reduce 430) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15168 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 15169 ((("%") reduce 434) ((")") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 15170 ((("%") reduce 416) ((")") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 15171 ((("%") reduce 438) ((")") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 15172 ((("%") reduce 482) (("(") shift) ((")") reduce 482) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 15173 ((("%") reduce 412) ((")") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 15174 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 15175 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 15176 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 15177 ((("%") reduce 356) ((")") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 356) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 15180 ((("%") reduce 400) ((")") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 15211 ((("%") shift) ((")") reduce 384) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 15212 ((("%") shift) ((")") reduce 387) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 15213 ((("%") shift) ((")") reduce 396) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 15214 ((("%") shift) ((")") reduce 409) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 15215 ((("%") shift) ((")") reduce 395) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 15216 ((("%") reduce 392) ((")") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 15217 ((("%") shift) ((")") reduce 383) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15218 ((("%") shift) ((")") reduce 381) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 381) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15219 ((("%") shift) ((")") reduce 382) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 382) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15220 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 15221 ((("%") shift) ((")") reduce 402) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 15222 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 15223 ((("%") shift) ((")") reduce 401) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 15224 ((("%") shift) ((")") reduce 408) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 15225 ((("%") shift) ((")") reduce 403) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 15228 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 15229 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 15230 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 15231 ((("%") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 15232 ((("%") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 15233 ((("%") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 15234 ((("%") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 15235 ((("%") reduce 410) ((")") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 15236 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 15237 ((("%") reduce 461) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 15238 ((("%") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 15245 ((("%") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 15247 ((("%") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 15248 ((("%") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 15249 ((("%") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 15250 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 15251 ((("%") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 15254 ((("%") reduce 531) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 15255 ((("%") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 15257 ((("%") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 15259 ((("%") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 15260 ((("%") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 15263 ((("%") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 15264 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15265 ((("%") shift) ((")") reduce 379) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_DOUBLE_ARROW) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15266 ((("%") shift) ((")") reduce 380) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_DOUBLE_ARROW) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15267 ((("%") shift) ((")") reduce 385) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 15268 ((("%") shift) ((")") reduce 386) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 15272 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15273 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15274 ((("%") shift) ((")") reduce 407) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 15275 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 15276 ((("%") reduce 393) ((")") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 15277 ((("%") shift) ((")") reduce 388) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_DOUBLE_ARROW) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 15278 ((("%") shift) ((")") reduce 390) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 15279 ((("%") shift) ((")") reduce 389) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 15280 ((("%") reduce 391) ((")") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 15281 ((("%") reduce 394) ((")") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 15283 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 15284 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 15285 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 15286 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 15287 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 15288 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 15289 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 15290 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 15291 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 15292 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 15293 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 15294 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 15295 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 15296 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 15297 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 15298 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 15299 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 15300 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 15301 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 15302 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 15303 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 15304 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 15305 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 15306 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 15307 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 15308 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 15309 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 15310 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 15311 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 15312 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 15313 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 15314 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 15315 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 15316 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 15317 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 15318 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 15319 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 15320 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 15321 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 15322 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 15323 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 15324 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 15325 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 15326 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 15327 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 15328 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 15329 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 15330 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 15331 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 15332 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 15333 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 15334 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 15335 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 15336 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 15337 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 15338 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 15339 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 15340 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 15341 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 15342 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 15343 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 15344 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 15345 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 15346 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 15347 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 15348 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 15349 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 15350 ((("%") reduce 81) (("(") reduce 81) ((")") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_DOUBLE_ARROW) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 15351 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 15352 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 15353 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 15354 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 15355 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 15356 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 15357 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 15358 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 15359 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 15360 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 15361 ((("%") reduce 492) (("(") reduce 527) ((")") reduce 492) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 15363 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 15364 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 15365 ((("%") reduce 519) (("(") reduce 529) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 15367 ((("%") reduce 456) (("(") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 15368 ((("%") reduce 281) (("(") reduce 281) ((")") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 15372 ((("%") reduce 282) (("(") reduce 282) ((")") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 15373 ((("%") reduce 283) (("(") reduce 283) ((")") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 15378 ((("%") reduce 377) ((")") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_DOUBLE_ARROW) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 15380 ((("%") reduce 375) ((")") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_DOUBLE_ARROW) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 15390 ((("%") shift) ((")") reduce 371) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 371) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15391 ((("%") shift) ((")") reduce 373) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 373) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15392 ((("%") shift) ((")") reduce 372) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 372) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15393 ((("%") shift) ((")") reduce 365) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 365) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15394 ((("%") shift) ((")") reduce 362) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 362) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15395 ((("%") shift) ((")") reduce 370) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 370) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15396 ((("%") shift) ((")") reduce 364) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 364) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15397 ((("%") shift) ((")") reduce 368) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 368) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15398 ((("%") shift) ((")") reduce 363) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 363) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15399 ((("%") shift) ((")") reduce 366) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 366) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15400 ((("%") shift) ((")") reduce 367) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 367) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15401 ((("%") shift) ((")") reduce 374) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 374) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15402 ((("%") shift) ((")") reduce 369) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 369) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15404 ((("%") shift) ((")") reduce 359) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 359) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15406 ((("%") reduce 515) (("(") reduce 515) ((")") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 15409 ((("%") reduce 511) (("(") reduce 502) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 15411 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 15413 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 15414 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 15415 ((("%") reduce 360) ((")") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_DOUBLE_ARROW) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 15419 ((("%") reduce 519) (("(") reduce 529) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 15421 ((("%") reduce 456) (("(") reduce 456) ((")") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 15422 ((("%") reduce 281) (("(") reduce 281) ((")") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 15426 ((("%") reduce 282) (("(") reduce 282) ((")") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 15427 ((("%") reduce 283) (("(") reduce 283) ((")") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 15428 ((("%") reduce 454) (("(") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 15431 ((("%") reduce 518) (("(") reduce 529) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 15432 ((("%") reduce 455) (("(") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 15433 ((("%") reduce 457) (("(") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 15439 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 15441 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 15442 ((("%") reduce 530) (("(") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 15443 ((("%") reduce 513) (("(") shift) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 15444 ((("%") reduce 532) (("(") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 15447 ((("%") reduce 531) (("(") reduce 531) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 15448 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 15449 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 15450 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 15451 ((("%") reduce 517) (("(") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 15454 ((("%") reduce 516) (("(") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 15455 ((("%") reduce 454) (("(") reduce 454) ((")") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 15466 ((("%") reduce 444) ((")") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 15467 ((("%") reduce 439) ((")") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 15477 ((("%") shift) ((")") reduce 444) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15478 ((("%") reduce 440) ((")") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 15480 ((("%") shift) ((")") reduce 431) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15482 ((("%") reduce 491) (("(") reduce 527) ((")") reduce 491) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 15484 ((("%") reduce 518) (("(") reduce 529) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 15485 ((("%") reduce 455) (("(") reduce 455) ((")") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 15486 ((("%") reduce 457) (("(") reduce 457) ((")") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 15488 ((("%") reduce 435) ((")") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 15489 ((("%") reduce 437) ((")") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 15495 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 15497 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 15498 ((("%") reduce 530) (("(") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 15499 ((("%") reduce 513) (("(") shift) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 15500 ((("%") reduce 532) (("(") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 15503 ((("%") reduce 531) (("(") reduce 531) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 15504 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 15505 ((("%") reduce 514) (("(") shift) ((")") reduce 514) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 15506 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 15508 ((("%") reduce 427) ((")") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 15510 ((("%") reduce 471) (("(") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 15512 ((("%") shift) ((")") reduce 358) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 358) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15513 ((("%") shift) ((")") reduce 432) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15516 ((("%") shift) ((")") reduce 431) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15517 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 15518 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15519 ((("%") reduce 419) ((")") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 15520 ((("%") reduce 436) ((")") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 15522 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 15524 ((("%") reduce 478) ((")") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 15525 ((("%") reduce 476) ((")") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 15526 ((("%") shift) ((")") reduce 567) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15527 ((("%") shift) ((")") reduce 566) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15528 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15529 ((("%") reduce 421) ((")") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 15532 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 15533 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 15534 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 15535 ((("%") reduce 458) (("(") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 15536 ((("%") reduce 91) (("(") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 15537 ((("%") reduce 354) ((")") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_DOUBLE_ARROW) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 15539 ((("%") reduce 460) (("(") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 15540 ((("%") reduce 468) (("(") shift) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 15541 ((("%") reduce 459) (("(") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 15542 ((("%") reduce 461) (("(") reduce 461) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 15543 ((("%") reduce 520) (("(") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 15550 ((("%") reduce 522) (("(") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 15552 ((("%") reduce 521) (("(") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 15553 ((("%") reduce 526) (("(") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 15554 ((("%") reduce 523) (("(") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 15555 ((("%") reduce 524) (("(") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 15557 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 15558 ((("%") reduce 353) ((")") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_DOUBLE_ARROW) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 15559 ((("%") reduce 281) ((")") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 15563 ((("%") reduce 282) ((")") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 15564 ((("%") reduce 283) ((")") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 15566 ((("%") reduce 525) (("(") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 15567 ((("%") reduce 355) ((")") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_DOUBLE_ARROW) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 15574 ((("%") reduce 352) ((")") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_DOUBLE_ARROW) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 15576 ((("%") reduce 462) (("(") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 15582 ((("%") reduce 223) ((")") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_DOUBLE_ARROW) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 15587 ((("%") shift) ((")") reduce 357) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 357) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15591 ((("%") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 15592 ((("%") reduce 417) ((")") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 15593 ((("%") shift) ((")") reduce 564) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15594 ((("%") shift) ((")") reduce 563) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15595 ((("%") reduce 376) ((")") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_DOUBLE_ARROW) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 15597 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 15599 ((("%") reduce 464) ((")") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 15602 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 15605 ((("%") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 15606 ((("%") reduce 418) ((")") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 15607 ((("%") reduce 378) ((")") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_DOUBLE_ARROW) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 15608 ((("%") reduce 361) ((")") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_DOUBLE_ARROW) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 15609 ((("%") reduce 422) ((")") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 15610 ((("%") reduce 420) ((")") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 15613 ((("%") reduce 470) (("(") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 15614 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 15615 ((("%") reduce 398) ((")") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 15616 ((("%") reduce 397) ((")") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 15618 ((("%") reduce 411) (("(") reduce 503) ((")") reduce 411) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 15619 ((("%") reduce 517) (("(") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 15622 ((("%") reduce 516) (("(") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 15624 ((("%") reduce 473) (("(") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 15625 ((("%") reduce 399) ((")") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 15626 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 15627 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15628 ((("%") reduce 419) ((")") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 15629 ((("%") reduce 436) ((")") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 15631 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 15633 ((("%") reduce 478) ((")") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 15634 ((("%") reduce 476) ((")") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 15635 ((("%") shift) ((")") reduce 567) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15636 ((("%") shift) ((")") reduce 566) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15637 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15638 ((("%") reduce 421) ((")") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 15641 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 15642 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 15643 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 15644 ((("%") reduce 458) (("(") reduce 458) ((")") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 15645 ((("%") reduce 91) (("(") reduce 91) ((")") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 15646 ((("%") reduce 354) ((")") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 15648 ((("%") reduce 460) (("(") reduce 460) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 15649 ((("%") reduce 468) (("(") shift) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 15650 ((("%") reduce 459) (("(") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 15651 ((("%") reduce 461) (("(") reduce 461) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 15652 ((("%") reduce 520) (("(") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 15659 ((("%") reduce 522) (("(") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 15661 ((("%") reduce 521) (("(") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 15662 ((("%") reduce 526) (("(") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 15663 ((("%") reduce 523) (("(") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 15664 ((("%") reduce 524) (("(") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 15666 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 15667 ((("%") reduce 353) ((")") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 15668 ((("%") reduce 281) ((")") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 15672 ((("%") reduce 282) ((")") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 15673 ((("%") reduce 283) ((")") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 15675 ((("%") reduce 525) (("(") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 15676 ((("%") reduce 355) ((")") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 15683 ((("%") reduce 352) ((")") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 15685 ((("%") reduce 462) (("(") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 15691 ((("%") reduce 223) ((")") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 15696 ((("%") shift) ((")") reduce 357) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15700 ((("%") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 15701 ((("%") reduce 417) ((")") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 15702 ((("%") shift) ((")") reduce 564) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15703 ((("%") shift) ((")") reduce 563) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 15704 ((("%") reduce 376) ((")") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 15706 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 15708 ((("%") reduce 464) ((")") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 15711 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 15714 ((("%") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 15715 ((("%") reduce 418) ((")") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 15716 ((("%") reduce 378) ((")") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 15717 ((("%") reduce 361) ((")") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 15718 ((("%") reduce 422) ((")") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 15719 ((("%") reduce 420) ((")") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 15722 ((("%") reduce 470) (("(") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 15723 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 15724 ((("%") reduce 398) ((")") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 15725 ((("%") reduce 397) ((")") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 15727 ((("%") reduce 411) (("(") reduce 503) ((")") reduce 411) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 15728 ((("%") reduce 517) (("(") reduce 517) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 15731 ((("%") reduce 516) (("(") reduce 516) ((")") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 15733 ((("%") reduce 473) (("(") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 15734 ((("%") reduce 399) ((")") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 15736 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 15746 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 15754 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 15757 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 15759 ((("!") reduce 235) (("\"") reduce 235) (("$") reduce 235) (("(") reduce 235) (("+") reduce 235) (("-") reduce 235) ((";") reduce 235) (("@") reduce 235) ((T_ABSTRACT) reduce 235) ((T_ARRAY) reduce 235) ((T_ARRAY_CAST) reduce 235) ((T_ATTRIBUTE) reduce 235) ((T_BOOL_CAST) reduce 235) ((T_BREAK) reduce 235) ((T_CLASS) reduce 235) ((T_CLASS_C) reduce 235) ((T_CLONE) reduce 235) ((T_CONSTANT_ENCAPSED_STRING) reduce 235) ((T_CONTINUE) reduce 235) ((T_DEC) reduce 235) ((T_DECLARE) reduce 235) ((T_DIR) reduce 235) ((T_DNUMBER) reduce 235) ((T_DO) reduce 235) ((T_DOUBLE_CAST) reduce 235) ((T_ECHO) reduce 235) ((T_ELSE) reduce 235) ((T_ELSEIF) reduce 235) ((T_EMPTY) reduce 235) ((T_ENUM) reduce 235) ((T_EVAL) reduce 235) ((T_EXIT) reduce 235) ((T_FILE) reduce 235) ((T_FINAL) reduce 235) ((T_FN) reduce 235) ((T_FOR) reduce 235) ((T_FOREACH) reduce 235) ((T_FUNCTION) reduce 235) ((T_FUNC_C) reduce 235) ((T_GLOBAL) reduce 235) ((T_GOTO) reduce 235) ((T_HALT_COMPILER) reduce 235) ((T_IF) reduce 235) ((T_INC) reduce 235) ((T_INCLUDE) reduce 235) ((T_INCLUDE_ONCE) reduce 235) ((T_INLINE_HTML) reduce 235) ((T_INTERFACE) reduce 235) ((T_INT_CAST) reduce 235) ((T_ISSET) reduce 235) ((T_LINE) reduce 235) ((T_LIST) reduce 235) ((T_LNUMBER) reduce 235) ((T_MATCH) reduce 235) ((T_METHOD_C) reduce 235) ((T_NAME_FULLY_QUALIFIED) reduce 235) ((T_NAME_QUALIFIED) reduce 235) ((T_NAME_RELATIVE) reduce 235) ((T_NEW) reduce 235) ((T_NS_C) reduce 235) ((T_OBJECT_CAST) reduce 235) ((T_PRINT) reduce 235) ((T_REQUIRE) reduce 235) ((T_REQUIRE_ONCE) reduce 235) ((T_RETURN) reduce 235) ((T_START_HEREDOC) reduce 235) ((T_STATIC) reduce 235) ((T_STRING) reduce 235) ((T_STRING_CAST) reduce 235) ((T_SWITCH) reduce 235) ((T_THROW) reduce 235) ((T_TRAIT) reduce 235) ((T_TRAIT_C) reduce 235) ((T_TRY) reduce 235) ((T_UNSET) reduce 235) ((T_UNSET_CAST) reduce 235) ((T_VARIABLE) reduce 235) ((T_WHILE) reduce 235) ((T_YIELD) reduce 235) ((T_YIELD_FROM) reduce 235) (("[") reduce 235) (("`") reduce 235) (("{") reduce 235) (("}") reduce 235) (("~") reduce 235)) 15762 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("}") reduce 145) (("~") reduce 145)) 15764 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_ELSE) reduce 237) ((T_ELSEIF) reduce 237) ((T_EMPTY) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("}") reduce 237) (("~") reduce 237)) 15765 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 15769 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("}") reduce 240) (("~") reduce 240)) 15773 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_ELSE) reduce 140) ((T_ELSEIF) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDIF) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 15774 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_ELSE) reduce 239) ((T_ELSEIF) reduce 239) ((T_EMPTY) shift) ((T_ENDIF) reduce 239) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 15775 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDIF) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 15789 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDIF) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 15799 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDIF) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 15801 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_ELSE) reduce 142) ((T_ELSEIF) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDIF) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 15803 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_ELSE) reduce 103) ((T_ELSEIF) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDIF) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 15805 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_ELSE) reduce 106) ((T_ELSEIF) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDIF) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 15808 ((("!") reduce 102) (("\"") reduce 102) (("$") reduce 102) (("(") reduce 102) (("+") reduce 102) (("-") reduce 102) ((";") reduce 102) (("@") reduce 102) ((T_ABSTRACT) reduce 102) ((T_ARRAY) reduce 102) ((T_ARRAY_CAST) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_BOOL_CAST) reduce 102) ((T_BREAK) reduce 102) ((T_CLASS) reduce 102) ((T_CLASS_C) reduce 102) ((T_CLONE) reduce 102) ((T_CONSTANT_ENCAPSED_STRING) reduce 102) ((T_CONTINUE) reduce 102) ((T_DEC) reduce 102) ((T_DECLARE) reduce 102) ((T_DIR) reduce 102) ((T_DNUMBER) reduce 102) ((T_DO) reduce 102) ((T_DOUBLE_CAST) reduce 102) ((T_ECHO) reduce 102) ((T_ELSE) reduce 102) ((T_ELSEIF) reduce 102) ((T_EMPTY) reduce 102) ((T_ENDIF) reduce 102) ((T_ENUM) reduce 102) ((T_EVAL) reduce 102) ((T_EXIT) reduce 102) ((T_FILE) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FOR) reduce 102) ((T_FOREACH) reduce 102) ((T_FUNCTION) reduce 102) ((T_FUNC_C) reduce 102) ((T_GLOBAL) reduce 102) ((T_GOTO) reduce 102) ((T_HALT_COMPILER) reduce 102) ((T_IF) reduce 102) ((T_INC) reduce 102) ((T_INCLUDE) reduce 102) ((T_INCLUDE_ONCE) reduce 102) ((T_INLINE_HTML) reduce 102) ((T_INTERFACE) reduce 102) ((T_INT_CAST) reduce 102) ((T_ISSET) reduce 102) ((T_LINE) reduce 102) ((T_LIST) reduce 102) ((T_LNUMBER) reduce 102) ((T_MATCH) reduce 102) ((T_METHOD_C) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_NEW) reduce 102) ((T_NS_C) reduce 102) ((T_OBJECT_CAST) reduce 102) ((T_PRINT) reduce 102) ((T_REQUIRE) reduce 102) ((T_REQUIRE_ONCE) reduce 102) ((T_RETURN) reduce 102) ((T_START_HEREDOC) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_STRING_CAST) reduce 102) ((T_SWITCH) reduce 102) ((T_THROW) reduce 102) ((T_TRAIT) reduce 102) ((T_TRAIT_C) reduce 102) ((T_TRY) reduce 102) ((T_UNSET) reduce 102) ((T_UNSET_CAST) reduce 102) ((T_VARIABLE) reduce 102) ((T_WHILE) reduce 102) ((T_YIELD) reduce 102) ((T_YIELD_FROM) reduce 102) (("[") reduce 102) (("`") reduce 102) (("{") reduce 102) (("~") reduce 102)) 15809 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDIF) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 15810 ((("!") reduce 236) (("\"") reduce 236) (("$") reduce 236) (("(") reduce 236) (("+") reduce 236) (("-") reduce 236) ((";") reduce 236) (("@") reduce 236) ((T_ABSTRACT) reduce 236) ((T_ARRAY) reduce 236) ((T_ARRAY_CAST) reduce 236) ((T_ATTRIBUTE) reduce 236) ((T_BOOL_CAST) reduce 236) ((T_BREAK) reduce 236) ((T_CLASS) reduce 236) ((T_CLASS_C) reduce 236) ((T_CLONE) reduce 236) ((T_CONSTANT_ENCAPSED_STRING) reduce 236) ((T_CONTINUE) reduce 236) ((T_DEC) reduce 236) ((T_DECLARE) reduce 236) ((T_DIR) reduce 236) ((T_DNUMBER) reduce 236) ((T_DO) reduce 236) ((T_DOUBLE_CAST) reduce 236) ((T_ECHO) reduce 236) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 236) ((T_ENDIF) reduce 236) ((T_ENUM) reduce 236) ((T_EVAL) reduce 236) ((T_EXIT) reduce 236) ((T_FILE) reduce 236) ((T_FINAL) reduce 236) ((T_FN) reduce 236) ((T_FOR) reduce 236) ((T_FOREACH) reduce 236) ((T_FUNCTION) reduce 236) ((T_FUNC_C) reduce 236) ((T_GLOBAL) reduce 236) ((T_GOTO) reduce 236) ((T_HALT_COMPILER) reduce 236) ((T_IF) reduce 236) ((T_INC) reduce 236) ((T_INCLUDE) reduce 236) ((T_INCLUDE_ONCE) reduce 236) ((T_INLINE_HTML) reduce 236) ((T_INTERFACE) reduce 236) ((T_INT_CAST) reduce 236) ((T_ISSET) reduce 236) ((T_LINE) reduce 236) ((T_LIST) reduce 236) ((T_LNUMBER) reduce 236) ((T_MATCH) reduce 236) ((T_METHOD_C) reduce 236) ((T_NAME_FULLY_QUALIFIED) reduce 236) ((T_NAME_QUALIFIED) reduce 236) ((T_NAME_RELATIVE) reduce 236) ((T_NEW) reduce 236) ((T_NS_C) reduce 236) ((T_OBJECT_CAST) reduce 236) ((T_PRINT) reduce 236) ((T_REQUIRE) reduce 236) ((T_REQUIRE_ONCE) reduce 236) ((T_RETURN) reduce 236) ((T_START_HEREDOC) reduce 236) ((T_STATIC) reduce 236) ((T_STRING) reduce 236) ((T_STRING_CAST) reduce 236) ((T_SWITCH) reduce 236) ((T_THROW) reduce 236) ((T_TRAIT) reduce 236) ((T_TRAIT_C) reduce 236) ((T_TRY) reduce 236) ((T_UNSET) reduce 236) ((T_UNSET_CAST) reduce 236) ((T_VARIABLE) reduce 236) ((T_WHILE) reduce 236) ((T_YIELD) reduce 236) ((T_YIELD_FROM) reduce 236) (("[") reduce 236) (("`") reduce 236) (("{") reduce 236) (("~") reduce 236)) 15811 ((("!") reduce 139) (("\"") reduce 139) (("$") reduce 139) (("(") reduce 139) (("+") reduce 139) (("-") reduce 139) ((";") reduce 139) (("@") reduce 139) ((T_ABSTRACT) reduce 139) ((T_ARRAY) reduce 139) ((T_ARRAY_CAST) reduce 139) ((T_ATTRIBUTE) reduce 139) ((T_BOOL_CAST) reduce 139) ((T_BREAK) reduce 139) ((T_CLASS) reduce 139) ((T_CLASS_C) reduce 139) ((T_CLONE) reduce 139) ((T_CONSTANT_ENCAPSED_STRING) reduce 139) ((T_CONTINUE) reduce 139) ((T_DEC) reduce 139) ((T_DECLARE) reduce 139) ((T_DIR) reduce 139) ((T_DNUMBER) reduce 139) ((T_DO) reduce 139) ((T_DOUBLE_CAST) reduce 139) ((T_ECHO) reduce 139) ((T_ELSE) reduce 139) ((T_ELSEIF) reduce 139) ((T_EMPTY) reduce 139) ((T_ENDIF) reduce 139) ((T_ENUM) reduce 139) ((T_EVAL) reduce 139) ((T_EXIT) reduce 139) ((T_FILE) reduce 139) ((T_FINAL) reduce 139) ((T_FN) reduce 139) ((T_FOR) reduce 139) ((T_FOREACH) reduce 139) ((T_FUNCTION) reduce 139) ((T_FUNC_C) reduce 139) ((T_GLOBAL) reduce 139) ((T_GOTO) reduce 139) ((T_HALT_COMPILER) reduce 139) ((T_IF) reduce 139) ((T_INC) reduce 139) ((T_INCLUDE) reduce 139) ((T_INCLUDE_ONCE) reduce 139) ((T_INLINE_HTML) reduce 139) ((T_INTERFACE) reduce 139) ((T_INT_CAST) reduce 139) ((T_ISSET) reduce 139) ((T_LINE) reduce 139) ((T_LIST) reduce 139) ((T_LNUMBER) reduce 139) ((T_MATCH) reduce 139) ((T_METHOD_C) reduce 139) ((T_NAME_FULLY_QUALIFIED) reduce 139) ((T_NAME_QUALIFIED) reduce 139) ((T_NAME_RELATIVE) reduce 139) ((T_NEW) reduce 139) ((T_NS_C) reduce 139) ((T_OBJECT_CAST) reduce 139) ((T_PRINT) reduce 139) ((T_REQUIRE) reduce 139) ((T_REQUIRE_ONCE) reduce 139) ((T_RETURN) reduce 139) ((T_START_HEREDOC) reduce 139) ((T_STATIC) reduce 139) ((T_STRING) reduce 139) ((T_STRING_CAST) reduce 139) ((T_SWITCH) reduce 139) ((T_THROW) reduce 139) ((T_TRAIT) reduce 139) ((T_TRAIT_C) reduce 139) ((T_TRY) reduce 139) ((T_UNSET) reduce 139) ((T_UNSET_CAST) reduce 139) ((T_VARIABLE) reduce 139) ((T_WHILE) reduce 139) ((T_YIELD) reduce 139) ((T_YIELD_FROM) reduce 139) (("[") reduce 139) (("`") reduce 139) (("{") reduce 139) (("~") reduce 139)) 15812 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_ELSE) reduce 105) ((T_ELSEIF) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDIF) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 15813 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_ELSE) reduce 141) ((T_ELSEIF) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDIF) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 15814 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_ELSE) reduce 104) ((T_ELSEIF) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDIF) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 15817 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDIF) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 15823 ((("!") reduce 235) (("\"") reduce 235) (("$") reduce 235) (("(") reduce 235) (("+") reduce 235) (("-") reduce 235) ((";") reduce 235) (("@") reduce 235) ((T_ABSTRACT) reduce 235) ((T_ARRAY) reduce 235) ((T_ARRAY_CAST) reduce 235) ((T_ATTRIBUTE) reduce 235) ((T_BOOL_CAST) reduce 235) ((T_BREAK) reduce 235) ((T_CLASS) reduce 235) ((T_CLASS_C) reduce 235) ((T_CLONE) reduce 235) ((T_CONSTANT_ENCAPSED_STRING) reduce 235) ((T_CONTINUE) reduce 235) ((T_DEC) reduce 235) ((T_DECLARE) reduce 235) ((T_DIR) reduce 235) ((T_DNUMBER) reduce 235) ((T_DO) reduce 235) ((T_DOUBLE_CAST) reduce 235) ((T_ECHO) reduce 235) ((T_ELSE) reduce 235) ((T_ELSEIF) reduce 235) ((T_EMPTY) reduce 235) ((T_ENDIF) reduce 235) ((T_ENUM) reduce 235) ((T_EVAL) reduce 235) ((T_EXIT) reduce 235) ((T_FILE) reduce 235) ((T_FINAL) reduce 235) ((T_FN) reduce 235) ((T_FOR) reduce 235) ((T_FOREACH) reduce 235) ((T_FUNCTION) reduce 235) ((T_FUNC_C) reduce 235) ((T_GLOBAL) reduce 235) ((T_GOTO) reduce 235) ((T_HALT_COMPILER) reduce 235) ((T_IF) reduce 235) ((T_INC) reduce 235) ((T_INCLUDE) reduce 235) ((T_INCLUDE_ONCE) reduce 235) ((T_INLINE_HTML) reduce 235) ((T_INTERFACE) reduce 235) ((T_INT_CAST) reduce 235) ((T_ISSET) reduce 235) ((T_LINE) reduce 235) ((T_LIST) reduce 235) ((T_LNUMBER) reduce 235) ((T_MATCH) reduce 235) ((T_METHOD_C) reduce 235) ((T_NAME_FULLY_QUALIFIED) reduce 235) ((T_NAME_QUALIFIED) reduce 235) ((T_NAME_RELATIVE) reduce 235) ((T_NEW) reduce 235) ((T_NS_C) reduce 235) ((T_OBJECT_CAST) reduce 235) ((T_PRINT) reduce 235) ((T_REQUIRE) reduce 235) ((T_REQUIRE_ONCE) reduce 235) ((T_RETURN) reduce 235) ((T_START_HEREDOC) reduce 235) ((T_STATIC) reduce 235) ((T_STRING) reduce 235) ((T_STRING_CAST) reduce 235) ((T_SWITCH) reduce 235) ((T_THROW) reduce 235) ((T_TRAIT) reduce 235) ((T_TRAIT_C) reduce 235) ((T_TRY) reduce 235) ((T_UNSET) reduce 235) ((T_UNSET_CAST) reduce 235) ((T_VARIABLE) reduce 235) ((T_WHILE) reduce 235) ((T_YIELD) reduce 235) ((T_YIELD_FROM) reduce 235) (("[") reduce 235) (("`") reduce 235) (("{") reduce 235) (("~") reduce 235)) 15824 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_ELSE) reduce 237) ((T_ELSEIF) reduce 237) ((T_EMPTY) reduce 237) ((T_ENDIF) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 15825 ((("(") reduce 79) ((T_STRING) reduce 79)) 15826 ((("(") reduce 80) ((T_STRING) reduce 80)) 15827 ((("(") reduce 447) ((T_STRING) reduce 447)) 15828 ((("(") reduce 443) ((T_STRING) shift)) 15838 ((("!") reduce 444) (("\"") reduce 444) (("$") reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) ((";") reduce 444) (("@") reduce 444) ((T_ABSTRACT) reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_BREAK) reduce 444) ((T_CLASS) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_CONTINUE) reduce 444) ((T_DEC) reduce 444) ((T_DECLARE) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DO) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_ECHO) reduce 444) ((T_ELSE) reduce 444) ((T_ELSEIF) reduce 444) ((T_EMPTY) reduce 444) ((T_ENDIF) reduce 444) ((T_ENUM) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FINAL) reduce 444) ((T_FN) reduce 444) ((T_FOR) reduce 444) ((T_FOREACH) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_GLOBAL) reduce 444) ((T_GOTO) reduce 444) ((T_HALT_COMPILER) reduce 444) ((T_IF) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INLINE_HTML) reduce 444) ((T_INTERFACE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_RETURN) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_SWITCH) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT) reduce 444) ((T_TRAIT_C) reduce 444) ((T_TRY) reduce 444) ((T_UNSET) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_VARIABLE) reduce 444) ((T_WHILE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("{") reduce 444) (("~") reduce 444)) 15839 ((("!") reduce 179) (("\"") reduce 179) (("$") reduce 179) (("(") reduce 179) (("+") reduce 179) (("-") reduce 179) ((";") reduce 179) (("@") reduce 179) ((T_ABSTRACT) reduce 179) ((T_ARRAY) reduce 179) ((T_ARRAY_CAST) reduce 179) ((T_ATTRIBUTE) reduce 179) ((T_BOOL_CAST) reduce 179) ((T_BREAK) reduce 179) ((T_CLASS) reduce 179) ((T_CLASS_C) reduce 179) ((T_CLONE) reduce 179) ((T_CONSTANT_ENCAPSED_STRING) reduce 179) ((T_CONTINUE) reduce 179) ((T_DEC) reduce 179) ((T_DECLARE) reduce 179) ((T_DIR) reduce 179) ((T_DNUMBER) reduce 179) ((T_DO) reduce 179) ((T_DOUBLE_CAST) reduce 179) ((T_ECHO) reduce 179) ((T_ELSE) reduce 179) ((T_ELSEIF) reduce 179) ((T_EMPTY) reduce 179) ((T_ENDIF) reduce 179) ((T_ENUM) reduce 179) ((T_EVAL) reduce 179) ((T_EXIT) reduce 179) ((T_FILE) reduce 179) ((T_FINAL) reduce 179) ((T_FN) reduce 179) ((T_FOR) reduce 179) ((T_FOREACH) reduce 179) ((T_FUNCTION) reduce 179) ((T_FUNC_C) reduce 179) ((T_GLOBAL) reduce 179) ((T_GOTO) reduce 179) ((T_HALT_COMPILER) reduce 179) ((T_IF) reduce 179) ((T_INC) reduce 179) ((T_INCLUDE) reduce 179) ((T_INCLUDE_ONCE) reduce 179) ((T_INLINE_HTML) reduce 179) ((T_INTERFACE) reduce 179) ((T_INT_CAST) reduce 179) ((T_ISSET) reduce 179) ((T_LINE) reduce 179) ((T_LIST) reduce 179) ((T_LNUMBER) reduce 179) ((T_MATCH) reduce 179) ((T_METHOD_C) reduce 179) ((T_NAME_FULLY_QUALIFIED) reduce 179) ((T_NAME_QUALIFIED) reduce 179) ((T_NAME_RELATIVE) reduce 179) ((T_NEW) reduce 179) ((T_NS_C) reduce 179) ((T_OBJECT_CAST) reduce 179) ((T_PRINT) reduce 179) ((T_REQUIRE) reduce 179) ((T_REQUIRE_ONCE) reduce 179) ((T_RETURN) reduce 179) ((T_START_HEREDOC) reduce 179) ((T_STATIC) reduce 179) ((T_STRING) reduce 179) ((T_STRING_CAST) reduce 179) ((T_SWITCH) reduce 179) ((T_THROW) reduce 179) ((T_TRAIT) reduce 179) ((T_TRAIT_C) reduce 179) ((T_TRY) reduce 179) ((T_UNSET) reduce 179) ((T_UNSET_CAST) reduce 179) ((T_VARIABLE) reduce 179) ((T_WHILE) reduce 179) ((T_YIELD) reduce 179) ((T_YIELD_FROM) reduce 179) (("[") reduce 179) (("`") reduce 179) (("{") reduce 179) (("~") reduce 179)) 15840 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDIF) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 15842 (((T_ABSTRACT) reduce 187) ((T_CLASS) reduce 187) ((T_FINAL) reduce 187)) 15849 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_ELSE) reduce 184) ((T_ELSEIF) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDIF) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 15851 (((T_ABSTRACT) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_CLASS) reduce 101) ((T_ENUM) reduce 101) ((T_FINAL) reduce 101) ((T_FN) reduce 101) ((T_FUNCTION) reduce 101) ((T_INTERFACE) reduce 101) ((T_STATIC) reduce 101) ((T_TRAIT) reduce 101)) 15852 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_ELSE) reduce 143) ((T_ELSEIF) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDIF) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 15853 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) ((";") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 15854 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 15857 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDIF) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 15858 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDIF) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 15859 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDIF) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 15860 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDIF) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 15874 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDIF) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 15884 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDIF) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 15886 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDIF) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 15888 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDIF) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 15890 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDIF) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 15893 ((("!") reduce 102) (("\"") reduce 102) (("$") reduce 102) (("(") reduce 102) (("+") reduce 102) (("-") reduce 102) ((";") reduce 102) (("@") reduce 102) ((T_ABSTRACT) reduce 102) ((T_ARRAY) reduce 102) ((T_ARRAY_CAST) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_BOOL_CAST) reduce 102) ((T_BREAK) reduce 102) ((T_CLASS) reduce 102) ((T_CLASS_C) reduce 102) ((T_CLONE) reduce 102) ((T_CONSTANT_ENCAPSED_STRING) reduce 102) ((T_CONTINUE) reduce 102) ((T_DEC) reduce 102) ((T_DECLARE) reduce 102) ((T_DIR) reduce 102) ((T_DNUMBER) reduce 102) ((T_DO) reduce 102) ((T_DOUBLE_CAST) reduce 102) ((T_ECHO) reduce 102) ((T_EMPTY) reduce 102) ((T_ENDIF) reduce 102) ((T_ENUM) reduce 102) ((T_EVAL) reduce 102) ((T_EXIT) reduce 102) ((T_FILE) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FOR) reduce 102) ((T_FOREACH) reduce 102) ((T_FUNCTION) reduce 102) ((T_FUNC_C) reduce 102) ((T_GLOBAL) reduce 102) ((T_GOTO) reduce 102) ((T_HALT_COMPILER) reduce 102) ((T_IF) reduce 102) ((T_INC) reduce 102) ((T_INCLUDE) reduce 102) ((T_INCLUDE_ONCE) reduce 102) ((T_INLINE_HTML) reduce 102) ((T_INTERFACE) reduce 102) ((T_INT_CAST) reduce 102) ((T_ISSET) reduce 102) ((T_LINE) reduce 102) ((T_LIST) reduce 102) ((T_LNUMBER) reduce 102) ((T_MATCH) reduce 102) ((T_METHOD_C) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_NEW) reduce 102) ((T_NS_C) reduce 102) ((T_OBJECT_CAST) reduce 102) ((T_PRINT) reduce 102) ((T_REQUIRE) reduce 102) ((T_REQUIRE_ONCE) reduce 102) ((T_RETURN) reduce 102) ((T_START_HEREDOC) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_STRING_CAST) reduce 102) ((T_SWITCH) reduce 102) ((T_THROW) reduce 102) ((T_TRAIT) reduce 102) ((T_TRAIT_C) reduce 102) ((T_TRY) reduce 102) ((T_UNSET) reduce 102) ((T_UNSET_CAST) reduce 102) ((T_VARIABLE) reduce 102) ((T_WHILE) reduce 102) ((T_YIELD) reduce 102) ((T_YIELD_FROM) reduce 102) (("[") reduce 102) (("`") reduce 102) (("{") reduce 102) (("~") reduce 102)) 15894 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDIF) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 15896 ((("!") reduce 139) (("\"") reduce 139) (("$") reduce 139) (("(") reduce 139) (("+") reduce 139) (("-") reduce 139) ((";") reduce 139) (("@") reduce 139) ((T_ABSTRACT) reduce 139) ((T_ARRAY) reduce 139) ((T_ARRAY_CAST) reduce 139) ((T_ATTRIBUTE) reduce 139) ((T_BOOL_CAST) reduce 139) ((T_BREAK) reduce 139) ((T_CLASS) reduce 139) ((T_CLASS_C) reduce 139) ((T_CLONE) reduce 139) ((T_CONSTANT_ENCAPSED_STRING) reduce 139) ((T_CONTINUE) reduce 139) ((T_DEC) reduce 139) ((T_DECLARE) reduce 139) ((T_DIR) reduce 139) ((T_DNUMBER) reduce 139) ((T_DO) reduce 139) ((T_DOUBLE_CAST) reduce 139) ((T_ECHO) reduce 139) ((T_EMPTY) reduce 139) ((T_ENDIF) reduce 139) ((T_ENUM) reduce 139) ((T_EVAL) reduce 139) ((T_EXIT) reduce 139) ((T_FILE) reduce 139) ((T_FINAL) reduce 139) ((T_FN) reduce 139) ((T_FOR) reduce 139) ((T_FOREACH) reduce 139) ((T_FUNCTION) reduce 139) ((T_FUNC_C) reduce 139) ((T_GLOBAL) reduce 139) ((T_GOTO) reduce 139) ((T_HALT_COMPILER) reduce 139) ((T_IF) reduce 139) ((T_INC) reduce 139) ((T_INCLUDE) reduce 139) ((T_INCLUDE_ONCE) reduce 139) ((T_INLINE_HTML) reduce 139) ((T_INTERFACE) reduce 139) ((T_INT_CAST) reduce 139) ((T_ISSET) reduce 139) ((T_LINE) reduce 139) ((T_LIST) reduce 139) ((T_LNUMBER) reduce 139) ((T_MATCH) reduce 139) ((T_METHOD_C) reduce 139) ((T_NAME_FULLY_QUALIFIED) reduce 139) ((T_NAME_QUALIFIED) reduce 139) ((T_NAME_RELATIVE) reduce 139) ((T_NEW) reduce 139) ((T_NS_C) reduce 139) ((T_OBJECT_CAST) reduce 139) ((T_PRINT) reduce 139) ((T_REQUIRE) reduce 139) ((T_REQUIRE_ONCE) reduce 139) ((T_RETURN) reduce 139) ((T_START_HEREDOC) reduce 139) ((T_STATIC) reduce 139) ((T_STRING) reduce 139) ((T_STRING_CAST) reduce 139) ((T_SWITCH) reduce 139) ((T_THROW) reduce 139) ((T_TRAIT) reduce 139) ((T_TRAIT_C) reduce 139) ((T_TRY) reduce 139) ((T_UNSET) reduce 139) ((T_UNSET_CAST) reduce 139) ((T_VARIABLE) reduce 139) ((T_WHILE) reduce 139) ((T_YIELD) reduce 139) ((T_YIELD_FROM) reduce 139) (("[") reduce 139) (("`") reduce 139) (("{") reduce 139) (("~") reduce 139)) 15897 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDIF) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 15898 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDIF) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 15899 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDIF) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 15902 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDIF) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 15904 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_EMPTY) reduce 237) ((T_ENDIF) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 15915 ((("!") reduce 444) (("\"") reduce 444) (("$") reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) ((";") reduce 444) (("@") reduce 444) ((T_ABSTRACT) reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_BREAK) reduce 444) ((T_CLASS) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_CONTINUE) reduce 444) ((T_DEC) reduce 444) ((T_DECLARE) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DO) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_ECHO) reduce 444) ((T_EMPTY) reduce 444) ((T_ENDIF) reduce 444) ((T_ENUM) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FINAL) reduce 444) ((T_FN) reduce 444) ((T_FOR) reduce 444) ((T_FOREACH) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_GLOBAL) reduce 444) ((T_GOTO) reduce 444) ((T_HALT_COMPILER) reduce 444) ((T_IF) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INLINE_HTML) reduce 444) ((T_INTERFACE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_RETURN) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_SWITCH) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT) reduce 444) ((T_TRAIT_C) reduce 444) ((T_TRY) reduce 444) ((T_UNSET) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_VARIABLE) reduce 444) ((T_WHILE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("{") reduce 444) (("~") reduce 444)) 15916 ((("!") reduce 179) (("\"") reduce 179) (("$") reduce 179) (("(") reduce 179) (("+") reduce 179) (("-") reduce 179) ((";") reduce 179) (("@") reduce 179) ((T_ABSTRACT) reduce 179) ((T_ARRAY) reduce 179) ((T_ARRAY_CAST) reduce 179) ((T_ATTRIBUTE) reduce 179) ((T_BOOL_CAST) reduce 179) ((T_BREAK) reduce 179) ((T_CLASS) reduce 179) ((T_CLASS_C) reduce 179) ((T_CLONE) reduce 179) ((T_CONSTANT_ENCAPSED_STRING) reduce 179) ((T_CONTINUE) reduce 179) ((T_DEC) reduce 179) ((T_DECLARE) reduce 179) ((T_DIR) reduce 179) ((T_DNUMBER) reduce 179) ((T_DO) reduce 179) ((T_DOUBLE_CAST) reduce 179) ((T_ECHO) reduce 179) ((T_EMPTY) reduce 179) ((T_ENDIF) reduce 179) ((T_ENUM) reduce 179) ((T_EVAL) reduce 179) ((T_EXIT) reduce 179) ((T_FILE) reduce 179) ((T_FINAL) reduce 179) ((T_FN) reduce 179) ((T_FOR) reduce 179) ((T_FOREACH) reduce 179) ((T_FUNCTION) reduce 179) ((T_FUNC_C) reduce 179) ((T_GLOBAL) reduce 179) ((T_GOTO) reduce 179) ((T_HALT_COMPILER) reduce 179) ((T_IF) reduce 179) ((T_INC) reduce 179) ((T_INCLUDE) reduce 179) ((T_INCLUDE_ONCE) reduce 179) ((T_INLINE_HTML) reduce 179) ((T_INTERFACE) reduce 179) ((T_INT_CAST) reduce 179) ((T_ISSET) reduce 179) ((T_LINE) reduce 179) ((T_LIST) reduce 179) ((T_LNUMBER) reduce 179) ((T_MATCH) reduce 179) ((T_METHOD_C) reduce 179) ((T_NAME_FULLY_QUALIFIED) reduce 179) ((T_NAME_QUALIFIED) reduce 179) ((T_NAME_RELATIVE) reduce 179) ((T_NEW) reduce 179) ((T_NS_C) reduce 179) ((T_OBJECT_CAST) reduce 179) ((T_PRINT) reduce 179) ((T_REQUIRE) reduce 179) ((T_REQUIRE_ONCE) reduce 179) ((T_RETURN) reduce 179) ((T_START_HEREDOC) reduce 179) ((T_STATIC) reduce 179) ((T_STRING) reduce 179) ((T_STRING_CAST) reduce 179) ((T_SWITCH) reduce 179) ((T_THROW) reduce 179) ((T_TRAIT) reduce 179) ((T_TRAIT_C) reduce 179) ((T_TRY) reduce 179) ((T_UNSET) reduce 179) ((T_UNSET_CAST) reduce 179) ((T_VARIABLE) reduce 179) ((T_WHILE) reduce 179) ((T_YIELD) reduce 179) ((T_YIELD_FROM) reduce 179) (("[") reduce 179) (("`") reduce 179) (("{") reduce 179) (("~") reduce 179)) 15917 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDIF) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 15925 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDIF) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 15926 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDIF) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 15929 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDIF) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 15933 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDIF) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 15936 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") shift) ((";") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 15937 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDWHILE) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 15938 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_EMPTY) reduce 232) ((T_ENDIF) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 15939 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDIF) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 15940 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 15941 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDWHILE) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 15956 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDWHILE) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 15966 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDWHILE) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 15968 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDWHILE) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 15970 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDWHILE) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 15972 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDWHILE) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 15975 ((("!") reduce 102) (("\"") reduce 102) (("$") reduce 102) (("(") reduce 102) (("+") reduce 102) (("-") reduce 102) ((";") reduce 102) (("@") reduce 102) ((T_ABSTRACT) reduce 102) ((T_ARRAY) reduce 102) ((T_ARRAY_CAST) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_BOOL_CAST) reduce 102) ((T_BREAK) reduce 102) ((T_CLASS) reduce 102) ((T_CLASS_C) reduce 102) ((T_CLONE) reduce 102) ((T_CONSTANT_ENCAPSED_STRING) reduce 102) ((T_CONTINUE) reduce 102) ((T_DEC) reduce 102) ((T_DECLARE) reduce 102) ((T_DIR) reduce 102) ((T_DNUMBER) reduce 102) ((T_DO) reduce 102) ((T_DOUBLE_CAST) reduce 102) ((T_ECHO) reduce 102) ((T_EMPTY) reduce 102) ((T_ENDWHILE) reduce 102) ((T_ENUM) reduce 102) ((T_EVAL) reduce 102) ((T_EXIT) reduce 102) ((T_FILE) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FOR) reduce 102) ((T_FOREACH) reduce 102) ((T_FUNCTION) reduce 102) ((T_FUNC_C) reduce 102) ((T_GLOBAL) reduce 102) ((T_GOTO) reduce 102) ((T_HALT_COMPILER) reduce 102) ((T_IF) reduce 102) ((T_INC) reduce 102) ((T_INCLUDE) reduce 102) ((T_INCLUDE_ONCE) reduce 102) ((T_INLINE_HTML) reduce 102) ((T_INTERFACE) reduce 102) ((T_INT_CAST) reduce 102) ((T_ISSET) reduce 102) ((T_LINE) reduce 102) ((T_LIST) reduce 102) ((T_LNUMBER) reduce 102) ((T_MATCH) reduce 102) ((T_METHOD_C) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_NEW) reduce 102) ((T_NS_C) reduce 102) ((T_OBJECT_CAST) reduce 102) ((T_PRINT) reduce 102) ((T_REQUIRE) reduce 102) ((T_REQUIRE_ONCE) reduce 102) ((T_RETURN) reduce 102) ((T_START_HEREDOC) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_STRING_CAST) reduce 102) ((T_SWITCH) reduce 102) ((T_THROW) reduce 102) ((T_TRAIT) reduce 102) ((T_TRAIT_C) reduce 102) ((T_TRY) reduce 102) ((T_UNSET) reduce 102) ((T_UNSET_CAST) reduce 102) ((T_VARIABLE) reduce 102) ((T_WHILE) reduce 102) ((T_YIELD) reduce 102) ((T_YIELD_FROM) reduce 102) (("[") reduce 102) (("`") reduce 102) (("{") reduce 102) (("~") reduce 102)) 15976 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDWHILE) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 15977 ((("!") reduce 236) (("\"") reduce 236) (("$") reduce 236) (("(") reduce 236) (("+") reduce 236) (("-") reduce 236) ((";") reduce 236) (("@") reduce 236) ((T_ABSTRACT) reduce 236) ((T_ARRAY) reduce 236) ((T_ARRAY_CAST) reduce 236) ((T_ATTRIBUTE) reduce 236) ((T_BOOL_CAST) reduce 236) ((T_BREAK) reduce 236) ((T_CLASS) reduce 236) ((T_CLASS_C) reduce 236) ((T_CLONE) reduce 236) ((T_CONSTANT_ENCAPSED_STRING) reduce 236) ((T_CONTINUE) reduce 236) ((T_DEC) reduce 236) ((T_DECLARE) reduce 236) ((T_DIR) reduce 236) ((T_DNUMBER) reduce 236) ((T_DO) reduce 236) ((T_DOUBLE_CAST) reduce 236) ((T_ECHO) reduce 236) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 236) ((T_ENDWHILE) reduce 236) ((T_ENUM) reduce 236) ((T_EVAL) reduce 236) ((T_EXIT) reduce 236) ((T_FILE) reduce 236) ((T_FINAL) reduce 236) ((T_FN) reduce 236) ((T_FOR) reduce 236) ((T_FOREACH) reduce 236) ((T_FUNCTION) reduce 236) ((T_FUNC_C) reduce 236) ((T_GLOBAL) reduce 236) ((T_GOTO) reduce 236) ((T_HALT_COMPILER) reduce 236) ((T_IF) reduce 236) ((T_INC) reduce 236) ((T_INCLUDE) reduce 236) ((T_INCLUDE_ONCE) reduce 236) ((T_INLINE_HTML) reduce 236) ((T_INTERFACE) reduce 236) ((T_INT_CAST) reduce 236) ((T_ISSET) reduce 236) ((T_LINE) reduce 236) ((T_LIST) reduce 236) ((T_LNUMBER) reduce 236) ((T_MATCH) reduce 236) ((T_METHOD_C) reduce 236) ((T_NAME_FULLY_QUALIFIED) reduce 236) ((T_NAME_QUALIFIED) reduce 236) ((T_NAME_RELATIVE) reduce 236) ((T_NEW) reduce 236) ((T_NS_C) reduce 236) ((T_OBJECT_CAST) reduce 236) ((T_PRINT) reduce 236) ((T_REQUIRE) reduce 236) ((T_REQUIRE_ONCE) reduce 236) ((T_RETURN) reduce 236) ((T_START_HEREDOC) reduce 236) ((T_STATIC) reduce 236) ((T_STRING) reduce 236) ((T_STRING_CAST) reduce 236) ((T_SWITCH) reduce 236) ((T_THROW) reduce 236) ((T_TRAIT) reduce 236) ((T_TRAIT_C) reduce 236) ((T_TRY) reduce 236) ((T_UNSET) reduce 236) ((T_UNSET_CAST) reduce 236) ((T_VARIABLE) reduce 236) ((T_WHILE) reduce 236) ((T_YIELD) reduce 236) ((T_YIELD_FROM) reduce 236) (("[") reduce 236) (("`") reduce 236) (("{") reduce 236) (("~") reduce 236)) 15978 ((("!") reduce 139) (("\"") reduce 139) (("$") reduce 139) (("(") reduce 139) (("+") reduce 139) (("-") reduce 139) ((";") reduce 139) (("@") reduce 139) ((T_ABSTRACT) reduce 139) ((T_ARRAY) reduce 139) ((T_ARRAY_CAST) reduce 139) ((T_ATTRIBUTE) reduce 139) ((T_BOOL_CAST) reduce 139) ((T_BREAK) reduce 139) ((T_CLASS) reduce 139) ((T_CLASS_C) reduce 139) ((T_CLONE) reduce 139) ((T_CONSTANT_ENCAPSED_STRING) reduce 139) ((T_CONTINUE) reduce 139) ((T_DEC) reduce 139) ((T_DECLARE) reduce 139) ((T_DIR) reduce 139) ((T_DNUMBER) reduce 139) ((T_DO) reduce 139) ((T_DOUBLE_CAST) reduce 139) ((T_ECHO) reduce 139) ((T_EMPTY) reduce 139) ((T_ENDWHILE) reduce 139) ((T_ENUM) reduce 139) ((T_EVAL) reduce 139) ((T_EXIT) reduce 139) ((T_FILE) reduce 139) ((T_FINAL) reduce 139) ((T_FN) reduce 139) ((T_FOR) reduce 139) ((T_FOREACH) reduce 139) ((T_FUNCTION) reduce 139) ((T_FUNC_C) reduce 139) ((T_GLOBAL) reduce 139) ((T_GOTO) reduce 139) ((T_HALT_COMPILER) reduce 139) ((T_IF) reduce 139) ((T_INC) reduce 139) ((T_INCLUDE) reduce 139) ((T_INCLUDE_ONCE) reduce 139) ((T_INLINE_HTML) reduce 139) ((T_INTERFACE) reduce 139) ((T_INT_CAST) reduce 139) ((T_ISSET) reduce 139) ((T_LINE) reduce 139) ((T_LIST) reduce 139) ((T_LNUMBER) reduce 139) ((T_MATCH) reduce 139) ((T_METHOD_C) reduce 139) ((T_NAME_FULLY_QUALIFIED) reduce 139) ((T_NAME_QUALIFIED) reduce 139) ((T_NAME_RELATIVE) reduce 139) ((T_NEW) reduce 139) ((T_NS_C) reduce 139) ((T_OBJECT_CAST) reduce 139) ((T_PRINT) reduce 139) ((T_REQUIRE) reduce 139) ((T_REQUIRE_ONCE) reduce 139) ((T_RETURN) reduce 139) ((T_START_HEREDOC) reduce 139) ((T_STATIC) reduce 139) ((T_STRING) reduce 139) ((T_STRING_CAST) reduce 139) ((T_SWITCH) reduce 139) ((T_THROW) reduce 139) ((T_TRAIT) reduce 139) ((T_TRAIT_C) reduce 139) ((T_TRY) reduce 139) ((T_UNSET) reduce 139) ((T_UNSET_CAST) reduce 139) ((T_VARIABLE) reduce 139) ((T_WHILE) reduce 139) ((T_YIELD) reduce 139) ((T_YIELD_FROM) reduce 139) (("[") reduce 139) (("`") reduce 139) (("{") reduce 139) (("~") reduce 139)) 15979 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDWHILE) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 15980 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDWHILE) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 15981 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDWHILE) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 15984 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDWHILE) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 15990 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDWHILE) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 16000 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDWHILE) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 16008 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDWHILE) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 16011 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDWHILE) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 16013 ((("!") reduce 235) (("\"") reduce 235) (("$") reduce 235) (("(") reduce 235) (("+") reduce 235) (("-") reduce 235) ((";") reduce 235) (("@") reduce 235) ((T_ABSTRACT) reduce 235) ((T_ARRAY) reduce 235) ((T_ARRAY_CAST) reduce 235) ((T_ATTRIBUTE) reduce 235) ((T_BOOL_CAST) reduce 235) ((T_BREAK) reduce 235) ((T_CLASS) reduce 235) ((T_CLASS_C) reduce 235) ((T_CLONE) reduce 235) ((T_CONSTANT_ENCAPSED_STRING) reduce 235) ((T_CONTINUE) reduce 235) ((T_DEC) reduce 235) ((T_DECLARE) reduce 235) ((T_DIR) reduce 235) ((T_DNUMBER) reduce 235) ((T_DO) reduce 235) ((T_DOUBLE_CAST) reduce 235) ((T_ECHO) reduce 235) ((T_ELSE) reduce 235) ((T_ELSEIF) reduce 235) ((T_EMPTY) reduce 235) ((T_ENDWHILE) reduce 235) ((T_ENUM) reduce 235) ((T_EVAL) reduce 235) ((T_EXIT) reduce 235) ((T_FILE) reduce 235) ((T_FINAL) reduce 235) ((T_FN) reduce 235) ((T_FOR) reduce 235) ((T_FOREACH) reduce 235) ((T_FUNCTION) reduce 235) ((T_FUNC_C) reduce 235) ((T_GLOBAL) reduce 235) ((T_GOTO) reduce 235) ((T_HALT_COMPILER) reduce 235) ((T_IF) reduce 235) ((T_INC) reduce 235) ((T_INCLUDE) reduce 235) ((T_INCLUDE_ONCE) reduce 235) ((T_INLINE_HTML) reduce 235) ((T_INTERFACE) reduce 235) ((T_INT_CAST) reduce 235) ((T_ISSET) reduce 235) ((T_LINE) reduce 235) ((T_LIST) reduce 235) ((T_LNUMBER) reduce 235) ((T_MATCH) reduce 235) ((T_METHOD_C) reduce 235) ((T_NAME_FULLY_QUALIFIED) reduce 235) ((T_NAME_QUALIFIED) reduce 235) ((T_NAME_RELATIVE) reduce 235) ((T_NEW) reduce 235) ((T_NS_C) reduce 235) ((T_OBJECT_CAST) reduce 235) ((T_PRINT) reduce 235) ((T_REQUIRE) reduce 235) ((T_REQUIRE_ONCE) reduce 235) ((T_RETURN) reduce 235) ((T_START_HEREDOC) reduce 235) ((T_STATIC) reduce 235) ((T_STRING) reduce 235) ((T_STRING_CAST) reduce 235) ((T_SWITCH) reduce 235) ((T_THROW) reduce 235) ((T_TRAIT) reduce 235) ((T_TRAIT_C) reduce 235) ((T_TRY) reduce 235) ((T_UNSET) reduce 235) ((T_UNSET_CAST) reduce 235) ((T_VARIABLE) reduce 235) ((T_WHILE) reduce 235) ((T_YIELD) reduce 235) ((T_YIELD_FROM) reduce 235) (("[") reduce 235) (("`") reduce 235) (("{") reduce 235) (("~") reduce 235)) 16016 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDWHILE) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 16018 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_ELSE) reduce 237) ((T_ELSEIF) reduce 237) ((T_EMPTY) reduce 237) ((T_ENDWHILE) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 16019 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDWHILE) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 16022 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDWHILE) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 16026 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDWHILE) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 16031 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_ELSE) reduce 232) ((T_ELSEIF) reduce 232) ((T_EMPTY) reduce 232) ((T_ENDWHILE) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 16032 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDWHILE) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 16035 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_ELSE) reduce 233) ((T_ELSEIF) reduce 233) ((T_EMPTY) reduce 233) ((T_ENDWHILE) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 16047 (((")") reduce 176) ((",") reduce 176)) 16049 (((")") reduce 178) ((",") reduce 178) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) (("[") reduce 496) (("{") reduce 496)) 16062 ((("\"") shift) (("$") shift) (("(") shift) ((")") reduce 124) ((T_ARRAY) shift) ((T_CLASS_C) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DIR) shift) ((T_FILE) shift) ((T_FUNC_C) shift) ((T_LINE) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NS_C) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_TRAIT_C) shift) ((T_VARIABLE) shift) (("[") shift)) 16065 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDWHILE) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 16066 (((")") reduce 177) ((",") reduce 177)) 16096 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDWHILE) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 16097 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_ELSE) reduce 174) ((T_ELSEIF) reduce 174) ((T_EMPTY) reduce 174) ((T_ENDWHILE) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 16100 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDWHILE) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 16103 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDWHILE) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 16105 (((")") reduce 93) ((T_VARIABLE) reduce 93) (("|") reduce 93)) 16106 (((")") reduce 92) ((T_VARIABLE) reduce 92) (("|") reduce 92)) 16107 (((")") reduce 94) ((T_VARIABLE) reduce 94) (("|") reduce 94)) 16108 (((")") reduce 458) ((T_VARIABLE) reduce 458) (("|") reduce 458)) 16109 (((")") reduce 91) ((T_VARIABLE) reduce 91) (("|") reduce 91)) 16110 (((")") reduce 172) ((T_VARIABLE) shift) (("|") shift)) 16111 (((")") reduce 170) ((T_VARIABLE) reduce 170) (("|") reduce 170)) 16112 (((")") reduce 459) ((T_VARIABLE) reduce 459) (("|") reduce 459)) 16113 (((")") reduce 173)) 16116 (((")") reduce 171) ((T_VARIABLE) reduce 171) (("|") reduce 171)) 16120 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDWHILE) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 16123 (((":") shift) (("{") shift)) 16124 (((";") shift) ((T_CASE) reduce 218) ((T_DEFAULT) reduce 218) ((T_ENDSWITCH) reduce 218)) 16125 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDWHILE) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 16126 (((";") shift) ((T_CASE) reduce 218) ((T_DEFAULT) reduce 218) (("}") reduce 218)) 16127 (((T_CASE) reduce 218) ((T_DEFAULT) reduce 218) (("}") reduce 218)) 16128 (((T_CASE) shift) ((T_DEFAULT) shift) (("}") shift)) 16130 (((":") shift) ((";") shift)) 16131 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_ELSE) reduce 214) ((T_ELSEIF) reduce 214) ((T_EMPTY) reduce 214) ((T_ENDWHILE) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 16132 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CASE) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DEFAULT) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 16133 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CASE) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DEFAULT) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 16134 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CASE) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DEFAULT) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("}") reduce 140) (("~") reduce 140)) 16135 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 220) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 220) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") reduce 220) (("~") shift)) 16136 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CASE) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DEFAULT) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 16150 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CASE) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DEFAULT) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 16160 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CASE) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DEFAULT) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 16162 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CASE) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DEFAULT) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("}") reduce 142) (("~") reduce 142)) 16164 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CASE) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DEFAULT) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("}") reduce 103) (("~") reduce 103)) 16166 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CASE) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DEFAULT) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("}") reduce 106) (("~") reduce 106)) 16169 ((("!") reduce 102) (("\"") reduce 102) (("$") reduce 102) (("(") reduce 102) (("+") reduce 102) (("-") reduce 102) ((";") reduce 102) (("@") reduce 102) ((T_ABSTRACT) reduce 102) ((T_ARRAY) reduce 102) ((T_ARRAY_CAST) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_BOOL_CAST) reduce 102) ((T_BREAK) reduce 102) ((T_CASE) reduce 102) ((T_CLASS) reduce 102) ((T_CLASS_C) reduce 102) ((T_CLONE) reduce 102) ((T_CONSTANT_ENCAPSED_STRING) reduce 102) ((T_CONTINUE) reduce 102) ((T_DEC) reduce 102) ((T_DECLARE) reduce 102) ((T_DEFAULT) reduce 102) ((T_DIR) reduce 102) ((T_DNUMBER) reduce 102) ((T_DO) reduce 102) ((T_DOUBLE_CAST) reduce 102) ((T_ECHO) reduce 102) ((T_EMPTY) reduce 102) ((T_ENUM) reduce 102) ((T_EVAL) reduce 102) ((T_EXIT) reduce 102) ((T_FILE) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FOR) reduce 102) ((T_FOREACH) reduce 102) ((T_FUNCTION) reduce 102) ((T_FUNC_C) reduce 102) ((T_GLOBAL) reduce 102) ((T_GOTO) reduce 102) ((T_HALT_COMPILER) reduce 102) ((T_IF) reduce 102) ((T_INC) reduce 102) ((T_INCLUDE) reduce 102) ((T_INCLUDE_ONCE) reduce 102) ((T_INLINE_HTML) reduce 102) ((T_INTERFACE) reduce 102) ((T_INT_CAST) reduce 102) ((T_ISSET) reduce 102) ((T_LINE) reduce 102) ((T_LIST) reduce 102) ((T_LNUMBER) reduce 102) ((T_MATCH) reduce 102) ((T_METHOD_C) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_NEW) reduce 102) ((T_NS_C) reduce 102) ((T_OBJECT_CAST) reduce 102) ((T_PRINT) reduce 102) ((T_REQUIRE) reduce 102) ((T_REQUIRE_ONCE) reduce 102) ((T_RETURN) reduce 102) ((T_START_HEREDOC) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_STRING_CAST) reduce 102) ((T_SWITCH) reduce 102) ((T_THROW) reduce 102) ((T_TRAIT) reduce 102) ((T_TRAIT_C) reduce 102) ((T_TRY) reduce 102) ((T_UNSET) reduce 102) ((T_UNSET_CAST) reduce 102) ((T_VARIABLE) reduce 102) ((T_WHILE) reduce 102) ((T_YIELD) reduce 102) ((T_YIELD_FROM) reduce 102) (("[") reduce 102) (("`") reduce 102) (("{") reduce 102) (("}") reduce 102) (("~") reduce 102)) 16170 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CASE) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DEFAULT) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 16171 ((("!") reduce 236) (("\"") reduce 236) (("$") reduce 236) (("(") reduce 236) (("+") reduce 236) (("-") reduce 236) ((";") reduce 236) (("@") reduce 236) ((T_ABSTRACT) reduce 236) ((T_ARRAY) reduce 236) ((T_ARRAY_CAST) reduce 236) ((T_ATTRIBUTE) reduce 236) ((T_BOOL_CAST) reduce 236) ((T_BREAK) reduce 236) ((T_CASE) reduce 236) ((T_CLASS) reduce 236) ((T_CLASS_C) reduce 236) ((T_CLONE) reduce 236) ((T_CONSTANT_ENCAPSED_STRING) reduce 236) ((T_CONTINUE) reduce 236) ((T_DEC) reduce 236) ((T_DECLARE) reduce 236) ((T_DEFAULT) reduce 236) ((T_DIR) reduce 236) ((T_DNUMBER) reduce 236) ((T_DO) reduce 236) ((T_DOUBLE_CAST) reduce 236) ((T_ECHO) reduce 236) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 236) ((T_ENUM) reduce 236) ((T_EVAL) reduce 236) ((T_EXIT) reduce 236) ((T_FILE) reduce 236) ((T_FINAL) reduce 236) ((T_FN) reduce 236) ((T_FOR) reduce 236) ((T_FOREACH) reduce 236) ((T_FUNCTION) reduce 236) ((T_FUNC_C) reduce 236) ((T_GLOBAL) reduce 236) ((T_GOTO) reduce 236) ((T_HALT_COMPILER) reduce 236) ((T_IF) reduce 236) ((T_INC) reduce 236) ((T_INCLUDE) reduce 236) ((T_INCLUDE_ONCE) reduce 236) ((T_INLINE_HTML) reduce 236) ((T_INTERFACE) reduce 236) ((T_INT_CAST) reduce 236) ((T_ISSET) reduce 236) ((T_LINE) reduce 236) ((T_LIST) reduce 236) ((T_LNUMBER) reduce 236) ((T_MATCH) reduce 236) ((T_METHOD_C) reduce 236) ((T_NAME_FULLY_QUALIFIED) reduce 236) ((T_NAME_QUALIFIED) reduce 236) ((T_NAME_RELATIVE) reduce 236) ((T_NEW) reduce 236) ((T_NS_C) reduce 236) ((T_OBJECT_CAST) reduce 236) ((T_PRINT) reduce 236) ((T_REQUIRE) reduce 236) ((T_REQUIRE_ONCE) reduce 236) ((T_RETURN) reduce 236) ((T_START_HEREDOC) reduce 236) ((T_STATIC) reduce 236) ((T_STRING) reduce 236) ((T_STRING_CAST) reduce 236) ((T_SWITCH) reduce 236) ((T_THROW) reduce 236) ((T_TRAIT) reduce 236) ((T_TRAIT_C) reduce 236) ((T_TRY) reduce 236) ((T_UNSET) reduce 236) ((T_UNSET_CAST) reduce 236) ((T_VARIABLE) reduce 236) ((T_WHILE) reduce 236) ((T_YIELD) reduce 236) ((T_YIELD_FROM) reduce 236) (("[") reduce 236) (("`") reduce 236) (("{") reduce 236) (("}") reduce 236) (("~") reduce 236)) 16172 ((("!") reduce 139) (("\"") reduce 139) (("$") reduce 139) (("(") reduce 139) (("+") reduce 139) (("-") reduce 139) ((";") reduce 139) (("@") reduce 139) ((T_ABSTRACT) reduce 139) ((T_ARRAY) reduce 139) ((T_ARRAY_CAST) reduce 139) ((T_ATTRIBUTE) reduce 139) ((T_BOOL_CAST) reduce 139) ((T_BREAK) reduce 139) ((T_CASE) reduce 139) ((T_CLASS) reduce 139) ((T_CLASS_C) reduce 139) ((T_CLONE) reduce 139) ((T_CONSTANT_ENCAPSED_STRING) reduce 139) ((T_CONTINUE) reduce 139) ((T_DEC) reduce 139) ((T_DECLARE) reduce 139) ((T_DEFAULT) reduce 139) ((T_DIR) reduce 139) ((T_DNUMBER) reduce 139) ((T_DO) reduce 139) ((T_DOUBLE_CAST) reduce 139) ((T_ECHO) reduce 139) ((T_EMPTY) reduce 139) ((T_ENUM) reduce 139) ((T_EVAL) reduce 139) ((T_EXIT) reduce 139) ((T_FILE) reduce 139) ((T_FINAL) reduce 139) ((T_FN) reduce 139) ((T_FOR) reduce 139) ((T_FOREACH) reduce 139) ((T_FUNCTION) reduce 139) ((T_FUNC_C) reduce 139) ((T_GLOBAL) reduce 139) ((T_GOTO) reduce 139) ((T_HALT_COMPILER) reduce 139) ((T_IF) reduce 139) ((T_INC) reduce 139) ((T_INCLUDE) reduce 139) ((T_INCLUDE_ONCE) reduce 139) ((T_INLINE_HTML) reduce 139) ((T_INTERFACE) reduce 139) ((T_INT_CAST) reduce 139) ((T_ISSET) reduce 139) ((T_LINE) reduce 139) ((T_LIST) reduce 139) ((T_LNUMBER) reduce 139) ((T_MATCH) reduce 139) ((T_METHOD_C) reduce 139) ((T_NAME_FULLY_QUALIFIED) reduce 139) ((T_NAME_QUALIFIED) reduce 139) ((T_NAME_RELATIVE) reduce 139) ((T_NEW) reduce 139) ((T_NS_C) reduce 139) ((T_OBJECT_CAST) reduce 139) ((T_PRINT) reduce 139) ((T_REQUIRE) reduce 139) ((T_REQUIRE_ONCE) reduce 139) ((T_RETURN) reduce 139) ((T_START_HEREDOC) reduce 139) ((T_STATIC) reduce 139) ((T_STRING) reduce 139) ((T_STRING_CAST) reduce 139) ((T_SWITCH) reduce 139) ((T_THROW) reduce 139) ((T_TRAIT) reduce 139) ((T_TRAIT_C) reduce 139) ((T_TRY) reduce 139) ((T_UNSET) reduce 139) ((T_UNSET_CAST) reduce 139) ((T_VARIABLE) reduce 139) ((T_WHILE) reduce 139) ((T_YIELD) reduce 139) ((T_YIELD_FROM) reduce 139) (("[") reduce 139) (("`") reduce 139) (("{") reduce 139) (("}") reduce 139) (("~") reduce 139)) 16173 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CASE) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DEFAULT) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("}") reduce 105) (("~") reduce 105)) 16174 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CASE) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DEFAULT) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("}") reduce 141) (("~") reduce 141)) 16175 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CASE) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DEFAULT) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("}") reduce 104) (("~") reduce 104)) 16178 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CASE) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DEFAULT) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("}") reduce 145) (("~") reduce 145)) 16184 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CASE) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DEFAULT) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 16194 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CASE) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DEFAULT) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 16202 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CASE) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DEFAULT) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 16205 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CASE) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DEFAULT) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 16207 ((("!") reduce 235) (("\"") reduce 235) (("$") reduce 235) (("(") reduce 235) (("+") reduce 235) (("-") reduce 235) ((";") reduce 235) (("@") reduce 235) ((T_ABSTRACT) reduce 235) ((T_ARRAY) reduce 235) ((T_ARRAY_CAST) reduce 235) ((T_ATTRIBUTE) reduce 235) ((T_BOOL_CAST) reduce 235) ((T_BREAK) reduce 235) ((T_CASE) reduce 235) ((T_CLASS) reduce 235) ((T_CLASS_C) reduce 235) ((T_CLONE) reduce 235) ((T_CONSTANT_ENCAPSED_STRING) reduce 235) ((T_CONTINUE) reduce 235) ((T_DEC) reduce 235) ((T_DECLARE) reduce 235) ((T_DEFAULT) reduce 235) ((T_DIR) reduce 235) ((T_DNUMBER) reduce 235) ((T_DO) reduce 235) ((T_DOUBLE_CAST) reduce 235) ((T_ECHO) reduce 235) ((T_ELSE) reduce 235) ((T_ELSEIF) reduce 235) ((T_EMPTY) reduce 235) ((T_ENUM) reduce 235) ((T_EVAL) reduce 235) ((T_EXIT) reduce 235) ((T_FILE) reduce 235) ((T_FINAL) reduce 235) ((T_FN) reduce 235) ((T_FOR) reduce 235) ((T_FOREACH) reduce 235) ((T_FUNCTION) reduce 235) ((T_FUNC_C) reduce 235) ((T_GLOBAL) reduce 235) ((T_GOTO) reduce 235) ((T_HALT_COMPILER) reduce 235) ((T_IF) reduce 235) ((T_INC) reduce 235) ((T_INCLUDE) reduce 235) ((T_INCLUDE_ONCE) reduce 235) ((T_INLINE_HTML) reduce 235) ((T_INTERFACE) reduce 235) ((T_INT_CAST) reduce 235) ((T_ISSET) reduce 235) ((T_LINE) reduce 235) ((T_LIST) reduce 235) ((T_LNUMBER) reduce 235) ((T_MATCH) reduce 235) ((T_METHOD_C) reduce 235) ((T_NAME_FULLY_QUALIFIED) reduce 235) ((T_NAME_QUALIFIED) reduce 235) ((T_NAME_RELATIVE) reduce 235) ((T_NEW) reduce 235) ((T_NS_C) reduce 235) ((T_OBJECT_CAST) reduce 235) ((T_PRINT) reduce 235) ((T_REQUIRE) reduce 235) ((T_REQUIRE_ONCE) reduce 235) ((T_RETURN) reduce 235) ((T_START_HEREDOC) reduce 235) ((T_STATIC) reduce 235) ((T_STRING) reduce 235) ((T_STRING_CAST) reduce 235) ((T_SWITCH) reduce 235) ((T_THROW) reduce 235) ((T_TRAIT) reduce 235) ((T_TRAIT_C) reduce 235) ((T_TRY) reduce 235) ((T_UNSET) reduce 235) ((T_UNSET_CAST) reduce 235) ((T_VARIABLE) reduce 235) ((T_WHILE) reduce 235) ((T_YIELD) reduce 235) ((T_YIELD_FROM) reduce 235) (("[") reduce 235) (("`") reduce 235) (("{") reduce 235) (("}") reduce 235) (("~") reduce 235)) 16210 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CASE) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DEFAULT) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("}") reduce 145) (("~") reduce 145)) 16212 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CASE) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DEFAULT) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_ELSE) reduce 237) ((T_ELSEIF) reduce 237) ((T_EMPTY) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("}") reduce 237) (("~") reduce 237)) 16213 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CASE) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DEFAULT) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 16216 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CASE) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DEFAULT) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("}") reduce 240) (("~") reduce 240)) 16220 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CASE) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DEFAULT) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("}") reduce 241) (("~") reduce 241)) 16225 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CASE) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DEFAULT) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_ELSE) reduce 232) ((T_ELSEIF) reduce 232) ((T_EMPTY) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("}") reduce 232) (("~") reduce 232)) 16226 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CASE) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DEFAULT) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 16229 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CASE) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DEFAULT) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_ELSE) reduce 233) ((T_ELSEIF) reduce 233) ((T_EMPTY) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("}") reduce 233) (("~") reduce 233)) 16234 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CASE) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DEFAULT) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 16237 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CASE) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DEFAULT) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 16238 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CASE) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DEFAULT) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_ELSE) reduce 174) ((T_ELSEIF) reduce 174) ((T_EMPTY) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("}") reduce 174) (("~") reduce 174)) 16241 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CASE) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DEFAULT) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 16244 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CASE) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DEFAULT) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("}") reduce 175) (("~") reduce 175)) 16251 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CASE) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DEFAULT) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 16256 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CASE) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DEFAULT) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 16260 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CASE) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DEFAULT) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_ELSE) reduce 214) ((T_ELSEIF) reduce 214) ((T_EMPTY) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("}") reduce 214) (("~") reduce 214)) 16262 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CASE) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DEFAULT) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_ELSE) reduce 215) ((T_ELSEIF) reduce 215) ((T_EMPTY) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("}") reduce 215) (("~") reduce 215)) 16263 (((T_CASE) reduce 218) ((T_DEFAULT) reduce 218) ((T_ENDSWITCH) reduce 218)) 16264 (((T_CASE) shift) ((T_DEFAULT) shift) ((T_ENDSWITCH) shift)) 16268 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CASE) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DEFAULT) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("}") reduce 216) (("~") reduce 216)) 16269 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CASE) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DEFAULT) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDSWITCH) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 16270 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CASE) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DEFAULT) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDSWITCH) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 16271 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CASE) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DEFAULT) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDSWITCH) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 16272 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 220) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 220) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDSWITCH) reduce 220) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 16273 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CASE) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DEFAULT) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDSWITCH) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 16287 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CASE) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DEFAULT) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDSWITCH) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 16297 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CASE) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DEFAULT) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDSWITCH) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 16299 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CASE) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DEFAULT) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDSWITCH) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 16301 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CASE) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DEFAULT) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDSWITCH) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 16303 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CASE) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DEFAULT) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDSWITCH) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 16306 ((("!") reduce 102) (("\"") reduce 102) (("$") reduce 102) (("(") reduce 102) (("+") reduce 102) (("-") reduce 102) ((";") reduce 102) (("@") reduce 102) ((T_ABSTRACT) reduce 102) ((T_ARRAY) reduce 102) ((T_ARRAY_CAST) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_BOOL_CAST) reduce 102) ((T_BREAK) reduce 102) ((T_CASE) reduce 102) ((T_CLASS) reduce 102) ((T_CLASS_C) reduce 102) ((T_CLONE) reduce 102) ((T_CONSTANT_ENCAPSED_STRING) reduce 102) ((T_CONTINUE) reduce 102) ((T_DEC) reduce 102) ((T_DECLARE) reduce 102) ((T_DEFAULT) reduce 102) ((T_DIR) reduce 102) ((T_DNUMBER) reduce 102) ((T_DO) reduce 102) ((T_DOUBLE_CAST) reduce 102) ((T_ECHO) reduce 102) ((T_EMPTY) reduce 102) ((T_ENDSWITCH) reduce 102) ((T_ENUM) reduce 102) ((T_EVAL) reduce 102) ((T_EXIT) reduce 102) ((T_FILE) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FOR) reduce 102) ((T_FOREACH) reduce 102) ((T_FUNCTION) reduce 102) ((T_FUNC_C) reduce 102) ((T_GLOBAL) reduce 102) ((T_GOTO) reduce 102) ((T_HALT_COMPILER) reduce 102) ((T_IF) reduce 102) ((T_INC) reduce 102) ((T_INCLUDE) reduce 102) ((T_INCLUDE_ONCE) reduce 102) ((T_INLINE_HTML) reduce 102) ((T_INTERFACE) reduce 102) ((T_INT_CAST) reduce 102) ((T_ISSET) reduce 102) ((T_LINE) reduce 102) ((T_LIST) reduce 102) ((T_LNUMBER) reduce 102) ((T_MATCH) reduce 102) ((T_METHOD_C) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_NEW) reduce 102) ((T_NS_C) reduce 102) ((T_OBJECT_CAST) reduce 102) ((T_PRINT) reduce 102) ((T_REQUIRE) reduce 102) ((T_REQUIRE_ONCE) reduce 102) ((T_RETURN) reduce 102) ((T_START_HEREDOC) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_STRING_CAST) reduce 102) ((T_SWITCH) reduce 102) ((T_THROW) reduce 102) ((T_TRAIT) reduce 102) ((T_TRAIT_C) reduce 102) ((T_TRY) reduce 102) ((T_UNSET) reduce 102) ((T_UNSET_CAST) reduce 102) ((T_VARIABLE) reduce 102) ((T_WHILE) reduce 102) ((T_YIELD) reduce 102) ((T_YIELD_FROM) reduce 102) (("[") reduce 102) (("`") reduce 102) (("{") reduce 102) (("~") reduce 102)) 16307 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CASE) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DEFAULT) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDSWITCH) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 16308 ((("!") reduce 236) (("\"") reduce 236) (("$") reduce 236) (("(") reduce 236) (("+") reduce 236) (("-") reduce 236) ((";") reduce 236) (("@") reduce 236) ((T_ABSTRACT) reduce 236) ((T_ARRAY) reduce 236) ((T_ARRAY_CAST) reduce 236) ((T_ATTRIBUTE) reduce 236) ((T_BOOL_CAST) reduce 236) ((T_BREAK) reduce 236) ((T_CASE) reduce 236) ((T_CLASS) reduce 236) ((T_CLASS_C) reduce 236) ((T_CLONE) reduce 236) ((T_CONSTANT_ENCAPSED_STRING) reduce 236) ((T_CONTINUE) reduce 236) ((T_DEC) reduce 236) ((T_DECLARE) reduce 236) ((T_DEFAULT) reduce 236) ((T_DIR) reduce 236) ((T_DNUMBER) reduce 236) ((T_DO) reduce 236) ((T_DOUBLE_CAST) reduce 236) ((T_ECHO) reduce 236) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 236) ((T_ENDSWITCH) reduce 236) ((T_ENUM) reduce 236) ((T_EVAL) reduce 236) ((T_EXIT) reduce 236) ((T_FILE) reduce 236) ((T_FINAL) reduce 236) ((T_FN) reduce 236) ((T_FOR) reduce 236) ((T_FOREACH) reduce 236) ((T_FUNCTION) reduce 236) ((T_FUNC_C) reduce 236) ((T_GLOBAL) reduce 236) ((T_GOTO) reduce 236) ((T_HALT_COMPILER) reduce 236) ((T_IF) reduce 236) ((T_INC) reduce 236) ((T_INCLUDE) reduce 236) ((T_INCLUDE_ONCE) reduce 236) ((T_INLINE_HTML) reduce 236) ((T_INTERFACE) reduce 236) ((T_INT_CAST) reduce 236) ((T_ISSET) reduce 236) ((T_LINE) reduce 236) ((T_LIST) reduce 236) ((T_LNUMBER) reduce 236) ((T_MATCH) reduce 236) ((T_METHOD_C) reduce 236) ((T_NAME_FULLY_QUALIFIED) reduce 236) ((T_NAME_QUALIFIED) reduce 236) ((T_NAME_RELATIVE) reduce 236) ((T_NEW) reduce 236) ((T_NS_C) reduce 236) ((T_OBJECT_CAST) reduce 236) ((T_PRINT) reduce 236) ((T_REQUIRE) reduce 236) ((T_REQUIRE_ONCE) reduce 236) ((T_RETURN) reduce 236) ((T_START_HEREDOC) reduce 236) ((T_STATIC) reduce 236) ((T_STRING) reduce 236) ((T_STRING_CAST) reduce 236) ((T_SWITCH) reduce 236) ((T_THROW) reduce 236) ((T_TRAIT) reduce 236) ((T_TRAIT_C) reduce 236) ((T_TRY) reduce 236) ((T_UNSET) reduce 236) ((T_UNSET_CAST) reduce 236) ((T_VARIABLE) reduce 236) ((T_WHILE) reduce 236) ((T_YIELD) reduce 236) ((T_YIELD_FROM) reduce 236) (("[") reduce 236) (("`") reduce 236) (("{") reduce 236) (("~") reduce 236)) 16309 ((("!") reduce 139) (("\"") reduce 139) (("$") reduce 139) (("(") reduce 139) (("+") reduce 139) (("-") reduce 139) ((";") reduce 139) (("@") reduce 139) ((T_ABSTRACT) reduce 139) ((T_ARRAY) reduce 139) ((T_ARRAY_CAST) reduce 139) ((T_ATTRIBUTE) reduce 139) ((T_BOOL_CAST) reduce 139) ((T_BREAK) reduce 139) ((T_CASE) reduce 139) ((T_CLASS) reduce 139) ((T_CLASS_C) reduce 139) ((T_CLONE) reduce 139) ((T_CONSTANT_ENCAPSED_STRING) reduce 139) ((T_CONTINUE) reduce 139) ((T_DEC) reduce 139) ((T_DECLARE) reduce 139) ((T_DEFAULT) reduce 139) ((T_DIR) reduce 139) ((T_DNUMBER) reduce 139) ((T_DO) reduce 139) ((T_DOUBLE_CAST) reduce 139) ((T_ECHO) reduce 139) ((T_EMPTY) reduce 139) ((T_ENDSWITCH) reduce 139) ((T_ENUM) reduce 139) ((T_EVAL) reduce 139) ((T_EXIT) reduce 139) ((T_FILE) reduce 139) ((T_FINAL) reduce 139) ((T_FN) reduce 139) ((T_FOR) reduce 139) ((T_FOREACH) reduce 139) ((T_FUNCTION) reduce 139) ((T_FUNC_C) reduce 139) ((T_GLOBAL) reduce 139) ((T_GOTO) reduce 139) ((T_HALT_COMPILER) reduce 139) ((T_IF) reduce 139) ((T_INC) reduce 139) ((T_INCLUDE) reduce 139) ((T_INCLUDE_ONCE) reduce 139) ((T_INLINE_HTML) reduce 139) ((T_INTERFACE) reduce 139) ((T_INT_CAST) reduce 139) ((T_ISSET) reduce 139) ((T_LINE) reduce 139) ((T_LIST) reduce 139) ((T_LNUMBER) reduce 139) ((T_MATCH) reduce 139) ((T_METHOD_C) reduce 139) ((T_NAME_FULLY_QUALIFIED) reduce 139) ((T_NAME_QUALIFIED) reduce 139) ((T_NAME_RELATIVE) reduce 139) ((T_NEW) reduce 139) ((T_NS_C) reduce 139) ((T_OBJECT_CAST) reduce 139) ((T_PRINT) reduce 139) ((T_REQUIRE) reduce 139) ((T_REQUIRE_ONCE) reduce 139) ((T_RETURN) reduce 139) ((T_START_HEREDOC) reduce 139) ((T_STATIC) reduce 139) ((T_STRING) reduce 139) ((T_STRING_CAST) reduce 139) ((T_SWITCH) reduce 139) ((T_THROW) reduce 139) ((T_TRAIT) reduce 139) ((T_TRAIT_C) reduce 139) ((T_TRY) reduce 139) ((T_UNSET) reduce 139) ((T_UNSET_CAST) reduce 139) ((T_VARIABLE) reduce 139) ((T_WHILE) reduce 139) ((T_YIELD) reduce 139) ((T_YIELD_FROM) reduce 139) (("[") reduce 139) (("`") reduce 139) (("{") reduce 139) (("~") reduce 139)) 16310 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CASE) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DEFAULT) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDSWITCH) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 16311 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CASE) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DEFAULT) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDSWITCH) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 16312 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CASE) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DEFAULT) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDSWITCH) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 16315 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CASE) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DEFAULT) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDSWITCH) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 16321 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CASE) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DEFAULT) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDSWITCH) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 16331 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CASE) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DEFAULT) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDSWITCH) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 16339 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CASE) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DEFAULT) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDSWITCH) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 16342 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CASE) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DEFAULT) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDSWITCH) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 16344 ((("!") reduce 235) (("\"") reduce 235) (("$") reduce 235) (("(") reduce 235) (("+") reduce 235) (("-") reduce 235) ((";") reduce 235) (("@") reduce 235) ((T_ABSTRACT) reduce 235) ((T_ARRAY) reduce 235) ((T_ARRAY_CAST) reduce 235) ((T_ATTRIBUTE) reduce 235) ((T_BOOL_CAST) reduce 235) ((T_BREAK) reduce 235) ((T_CASE) reduce 235) ((T_CLASS) reduce 235) ((T_CLASS_C) reduce 235) ((T_CLONE) reduce 235) ((T_CONSTANT_ENCAPSED_STRING) reduce 235) ((T_CONTINUE) reduce 235) ((T_DEC) reduce 235) ((T_DECLARE) reduce 235) ((T_DEFAULT) reduce 235) ((T_DIR) reduce 235) ((T_DNUMBER) reduce 235) ((T_DO) reduce 235) ((T_DOUBLE_CAST) reduce 235) ((T_ECHO) reduce 235) ((T_ELSE) reduce 235) ((T_ELSEIF) reduce 235) ((T_EMPTY) reduce 235) ((T_ENDSWITCH) reduce 235) ((T_ENUM) reduce 235) ((T_EVAL) reduce 235) ((T_EXIT) reduce 235) ((T_FILE) reduce 235) ((T_FINAL) reduce 235) ((T_FN) reduce 235) ((T_FOR) reduce 235) ((T_FOREACH) reduce 235) ((T_FUNCTION) reduce 235) ((T_FUNC_C) reduce 235) ((T_GLOBAL) reduce 235) ((T_GOTO) reduce 235) ((T_HALT_COMPILER) reduce 235) ((T_IF) reduce 235) ((T_INC) reduce 235) ((T_INCLUDE) reduce 235) ((T_INCLUDE_ONCE) reduce 235) ((T_INLINE_HTML) reduce 235) ((T_INTERFACE) reduce 235) ((T_INT_CAST) reduce 235) ((T_ISSET) reduce 235) ((T_LINE) reduce 235) ((T_LIST) reduce 235) ((T_LNUMBER) reduce 235) ((T_MATCH) reduce 235) ((T_METHOD_C) reduce 235) ((T_NAME_FULLY_QUALIFIED) reduce 235) ((T_NAME_QUALIFIED) reduce 235) ((T_NAME_RELATIVE) reduce 235) ((T_NEW) reduce 235) ((T_NS_C) reduce 235) ((T_OBJECT_CAST) reduce 235) ((T_PRINT) reduce 235) ((T_REQUIRE) reduce 235) ((T_REQUIRE_ONCE) reduce 235) ((T_RETURN) reduce 235) ((T_START_HEREDOC) reduce 235) ((T_STATIC) reduce 235) ((T_STRING) reduce 235) ((T_STRING_CAST) reduce 235) ((T_SWITCH) reduce 235) ((T_THROW) reduce 235) ((T_TRAIT) reduce 235) ((T_TRAIT_C) reduce 235) ((T_TRY) reduce 235) ((T_UNSET) reduce 235) ((T_UNSET_CAST) reduce 235) ((T_VARIABLE) reduce 235) ((T_WHILE) reduce 235) ((T_YIELD) reduce 235) ((T_YIELD_FROM) reduce 235) (("[") reduce 235) (("`") reduce 235) (("{") reduce 235) (("~") reduce 235)) 16347 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CASE) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DEFAULT) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDSWITCH) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 16349 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CASE) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DEFAULT) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_ELSE) reduce 237) ((T_ELSEIF) reduce 237) ((T_EMPTY) reduce 237) ((T_ENDSWITCH) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 16350 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CASE) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DEFAULT) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDSWITCH) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 16353 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CASE) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DEFAULT) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDSWITCH) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 16357 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CASE) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DEFAULT) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDSWITCH) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 16362 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CASE) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DEFAULT) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_ELSE) reduce 232) ((T_ELSEIF) reduce 232) ((T_EMPTY) reduce 232) ((T_ENDSWITCH) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 16363 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CASE) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DEFAULT) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDSWITCH) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 16366 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CASE) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DEFAULT) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_ELSE) reduce 233) ((T_ELSEIF) reduce 233) ((T_EMPTY) reduce 233) ((T_ENDSWITCH) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 16371 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CASE) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DEFAULT) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDSWITCH) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 16374 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CASE) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DEFAULT) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDSWITCH) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 16375 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CASE) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DEFAULT) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_ELSE) reduce 174) ((T_ELSEIF) reduce 174) ((T_EMPTY) reduce 174) ((T_ENDSWITCH) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 16378 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CASE) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DEFAULT) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDSWITCH) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 16381 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CASE) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DEFAULT) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDSWITCH) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 16388 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CASE) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DEFAULT) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDSWITCH) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 16393 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CASE) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DEFAULT) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDSWITCH) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 16397 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CASE) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DEFAULT) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_ELSE) reduce 214) ((T_ELSEIF) reduce 214) ((T_EMPTY) reduce 214) ((T_ENDSWITCH) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 16399 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CASE) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DEFAULT) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_ELSE) reduce 215) ((T_ELSEIF) reduce 215) ((T_EMPTY) reduce 215) ((T_ENDSWITCH) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 16403 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CASE) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DEFAULT) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDSWITCH) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 16406 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CASE) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DEFAULT) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDSWITCH) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 16407 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CASE) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DEFAULT) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDSWITCH) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 16408 (((",") reduce 294) ((";") reduce 294) (("=") shift)) 16409 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 16410 (((",") reduce 293) ((";") reduce 293)) 16413 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CASE) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DEFAULT) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDSWITCH) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 16414 (((",") reduce 292) ((";") reduce 292)) 16477 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 295) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 295) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17403 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 494) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17405 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CASE) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DEFAULT) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDSWITCH) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 17407 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CASE) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DEFAULT) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDSWITCH) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 17409 (((",") reduce 515) ((";") reduce 515)) 17410 (((",") reduce 290) ((";") reduce 290)) 17412 (((",") reduce 291) ((";") reduce 291)) 17414 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CASE) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DEFAULT) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDSWITCH) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 17415 (((",") reduce 289) ((";") reduce 289)) 17416 (((",") reduce 517) ((";") reduce 517)) 17419 (((",") reduce 516) ((";") reduce 516)) 17431 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AS) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 17433 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AS) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 17435 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AS) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 17436 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AS) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 17440 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AS) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 17441 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AS) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 17442 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AS) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 17448 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AS) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 17450 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AS) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 17452 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AS) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 17453 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 17454 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 17455 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 17457 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AS) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 17464 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AS) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 17467 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AS) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 17469 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_AS) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 17470 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_AS) reduce 429) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 17477 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_AS) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 17478 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AS) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 17480 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AS) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 17481 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AS) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 17482 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17485 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_AS) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 17486 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AS) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 17487 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AS) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 17488 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AS) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 17489 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AS) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 17490 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AS) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 17491 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AS) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 17492 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_AS) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 17493 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_AS) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 17494 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_AS) reduce 356) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 17497 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AS) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 17528 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 17529 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 17530 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AS) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 17531 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AS) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 17532 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AS) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 17533 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AS) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 17534 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 383) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17535 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 381) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17536 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 382) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17537 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AS) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 17538 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AS) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 17539 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AS) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 17540 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AS) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 17541 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AS) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 17542 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AS) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 17545 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 17546 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 17547 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 17548 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AS) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 17549 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AS) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 17550 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AS) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 17551 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AS) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 17552 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AS) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 17553 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AS) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 17554 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AS) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 17555 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AS) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 17562 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AS) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 17564 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AS) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 17565 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AS) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 17566 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AS) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 17567 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AS) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 17568 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AS) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 17571 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AS) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 17572 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AS) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 17574 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AS) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 17576 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AS) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 17577 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AS) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 17580 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AS) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 17581 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 415) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17582 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 379) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17583 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 380) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17584 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AS) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 17585 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AS) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 17589 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 413) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17590 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 414) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17591 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AS) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 17592 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AS) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 17593 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AS) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 17594 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AS) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 17595 ((("%") shift) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AS) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 17596 ((("%") shift) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AS) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 17597 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AS) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 17598 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AS) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 17600 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AS) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 17601 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AS) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 17602 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AS) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 17603 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AS) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 17604 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AS) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 17605 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AS) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 17606 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AS) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 17607 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AS) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 17608 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AS) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 17609 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AS) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 17610 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AS) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 17611 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AS) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 17612 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AS) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 17613 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AS) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 17614 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AS) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 17615 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AS) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 17616 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AS) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 17617 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AS) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 17618 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AS) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 17619 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AS) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 17620 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AS) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 17621 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AS) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 17622 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AS) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 17623 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AS) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 17624 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AS) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 17625 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AS) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 17626 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AS) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 17627 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AS) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 17628 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AS) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 17629 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AS) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 17630 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AS) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 17631 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AS) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 17632 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AS) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 17633 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AS) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 17634 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AS) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 17635 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AS) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 17636 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AS) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 17637 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AS) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 17638 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AS) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 17639 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AS) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 17640 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AS) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 17641 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AS) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 17642 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AS) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 17643 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AS) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 17644 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AS) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 17645 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AS) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 17646 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AS) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 17647 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AS) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 17648 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AS) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 17649 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AS) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 17650 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AS) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 17651 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AS) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 17652 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AS) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 17653 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AS) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 17654 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AS) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 17655 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AS) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 17656 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AS) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 17657 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AS) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 17658 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AS) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 17659 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AS) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 17660 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AS) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 17661 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AS) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 17662 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AS) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 17663 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AS) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 17664 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AS) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 17665 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AS) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 17666 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AS) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 17667 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AS) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 17668 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AS) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 17669 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AS) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 17670 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AS) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 17671 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AS) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 17672 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AS) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 17673 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AS) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 17674 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AS) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 17675 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AS) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 17676 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AS) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 17677 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AS) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 17678 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AS) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 17680 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AS) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 17681 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AS) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 17682 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_AS) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 17684 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_AS) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 17685 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_AS) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 17689 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_AS) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 17690 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_AS) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 17695 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AS) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 17697 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AS) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 17707 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 371) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17708 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 373) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17709 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 372) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17710 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 365) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17711 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 362) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17712 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 370) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17713 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 364) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17714 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 368) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17715 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 363) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17716 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 366) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17717 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 367) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17718 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 374) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17719 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 369) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17721 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 359) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17723 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AS) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 17726 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AS) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 17728 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AS) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 17730 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AS) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 17731 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AS) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 17732 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AS) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 17736 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AS) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 17738 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AS) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 17739 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AS) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 17743 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AS) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 17744 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AS) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 17745 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AS) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 17748 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AS) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 17749 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AS) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 17750 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AS) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 17756 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AS) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 17758 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AS) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 17759 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AS) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 17760 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AS) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 17761 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AS) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 17764 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AS) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 17765 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AS) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 17766 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AS) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 17767 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AS) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 17768 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AS) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 17771 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AS) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 17772 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_AS) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 17783 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AS) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 17784 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AS) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 17855 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 444) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18168 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AS) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 18171 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AS) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 18173 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_AS) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 18174 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_AS) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 18175 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_AS) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 18177 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AS) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 18178 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AS) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 18184 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_AS) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 18186 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_AS) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 18187 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_AS) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 18188 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_AS) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 18189 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_AS) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 18192 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_AS) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 18193 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_AS) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 18194 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_AS) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 18195 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_AS) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 18197 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AS) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 18199 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AS) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 18201 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 358) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18202 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 432) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18213 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AS) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 18215 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AS) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 18217 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AS) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 18218 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AS) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 18222 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AS) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 18223 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AS) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 18224 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AS) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 18230 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AS) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 18232 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AS) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 18234 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AS) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 18235 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 18236 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 18237 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 18239 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AS) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 18246 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AS) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 18249 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AS) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 18251 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_AS) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 18252 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_AS) reduce 429) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 18259 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_AS) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 18260 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AS) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 18262 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AS) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_DOUBLE_ARROW) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 18263 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AS) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_DOUBLE_ARROW) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 18264 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 430) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18267 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_AS) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 18268 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AS) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 18269 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AS) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 18270 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AS) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 18271 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AS) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 18272 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AS) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 18273 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AS) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 18274 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_AS) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 18275 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_AS) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 18276 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_AS) reduce 356) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 356) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 18279 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AS) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 18310 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 384) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 18311 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 387) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 18312 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AS) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 18313 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AS) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 18314 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AS) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 18315 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AS) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 18316 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 383) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18317 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 381) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 381) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18318 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 382) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 382) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18319 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AS) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 18320 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AS) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 18321 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AS) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 18322 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AS) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 18323 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AS) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 18324 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AS) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 18327 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 18328 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 18329 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 18330 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AS) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 18331 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AS) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 18332 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AS) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 18333 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AS) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 18334 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AS) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 18335 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AS) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 18336 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AS) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 18337 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AS) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 18344 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AS) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 18346 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AS) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 18347 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AS) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 18348 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AS) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 18349 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AS) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 18350 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AS) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 18353 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AS) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 18354 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AS) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 18356 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AS) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 18358 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AS) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 18359 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AS) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 18362 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AS) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 18363 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 415) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18364 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 379) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_DOUBLE_ARROW) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18365 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 380) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_DOUBLE_ARROW) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18366 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AS) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 18367 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AS) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 18371 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 413) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18372 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 414) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18373 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AS) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 18374 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AS) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 18375 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AS) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 18376 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AS) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_DOUBLE_ARROW) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 18377 ((("%") shift) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AS) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 18378 ((("%") shift) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AS) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 18379 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AS) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 18380 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AS) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 18382 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AS) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 18383 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AS) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 18384 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AS) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 18385 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AS) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 18386 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AS) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 18387 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AS) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 18388 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AS) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 18389 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AS) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 18390 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AS) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 18391 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AS) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 18392 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AS) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 18393 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AS) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 18394 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AS) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 18395 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AS) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 18396 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AS) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 18397 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AS) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 18398 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AS) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 18399 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AS) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 18400 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AS) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 18401 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AS) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 18402 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AS) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 18403 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AS) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 18404 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AS) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 18405 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AS) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 18406 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AS) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 18407 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AS) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 18408 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AS) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 18409 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AS) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 18410 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AS) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 18411 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AS) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 18412 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AS) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 18413 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AS) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 18414 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AS) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 18415 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AS) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 18416 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AS) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 18417 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AS) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 18418 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AS) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 18419 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AS) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 18420 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AS) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 18421 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AS) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 18422 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AS) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 18423 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AS) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 18424 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AS) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 18425 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AS) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 18426 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AS) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 18427 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AS) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 18428 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AS) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 18429 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AS) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 18430 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AS) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 18431 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AS) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 18432 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AS) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 18433 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AS) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 18434 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AS) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 18435 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AS) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 18436 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AS) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 18437 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AS) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 18438 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AS) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 18439 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AS) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 18440 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AS) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 18441 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AS) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 18442 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AS) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 18443 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AS) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 18444 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AS) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 18445 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AS) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 18446 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AS) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 18447 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AS) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 18448 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AS) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 18449 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AS) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_DOUBLE_ARROW) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 18450 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AS) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 18451 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AS) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 18452 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AS) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 18453 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AS) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 18454 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AS) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 18455 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AS) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 18456 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AS) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 18457 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AS) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 18458 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AS) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 18459 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AS) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 18460 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AS) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 18462 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AS) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 18463 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AS) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 18464 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_AS) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 18466 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_AS) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 18467 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_AS) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 18471 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_AS) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 18472 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_AS) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 18477 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AS) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_DOUBLE_ARROW) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 18479 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AS) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_DOUBLE_ARROW) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 18489 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 371) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 371) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18490 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 373) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 373) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18491 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 372) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 372) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18492 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 365) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 365) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18493 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 362) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 362) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18494 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 370) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 370) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18495 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 364) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 364) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18496 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 368) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 368) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18497 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 363) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 363) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18498 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 366) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 366) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18499 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 367) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 367) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18500 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 374) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 374) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18501 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 369) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 369) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18503 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 359) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 359) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18505 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AS) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 18508 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AS) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 18510 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AS) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 18512 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AS) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 18513 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AS) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 18514 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AS) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_DOUBLE_ARROW) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 18518 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AS) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 18520 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AS) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 18521 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AS) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 18525 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AS) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 18526 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AS) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 18527 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AS) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 18530 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AS) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 18531 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AS) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 18532 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AS) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 18538 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AS) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 18540 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AS) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 18541 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AS) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 18542 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AS) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 18543 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AS) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 18546 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AS) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 18547 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AS) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 18548 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AS) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 18549 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AS) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 18550 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AS) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 18553 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AS) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 18554 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_AS) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 18565 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AS) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 18566 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AS) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 18576 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 444) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18577 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AS) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 18579 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 431) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18581 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AS) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 18583 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_AS) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 18584 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_AS) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 18585 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_AS) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 18587 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AS) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 18588 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AS) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 18594 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_AS) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 18596 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_AS) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 18597 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_AS) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 18598 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_AS) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 18599 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_AS) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 18602 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_AS) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 18603 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_AS) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 18604 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_AS) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 18605 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_AS) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 18607 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AS) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 18609 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AS) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 18611 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 358) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 358) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18612 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 432) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18615 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 431) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18616 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AS) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 18617 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 433) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18618 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AS) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 18619 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AS) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 18621 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AS) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 18623 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AS) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 18624 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AS) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 18625 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 567) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18626 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 566) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18627 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 428) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18628 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AS) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 18631 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 18632 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 18633 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 18634 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AS) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 18635 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AS) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 18636 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AS) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_DOUBLE_ARROW) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 18638 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AS) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 18639 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AS) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 18640 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AS) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 18641 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AS) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 18642 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AS) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 18649 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AS) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 18651 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AS) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 18652 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AS) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 18653 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AS) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 18654 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AS) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 18656 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AS) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 18657 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AS) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_DOUBLE_ARROW) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 18658 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AS) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 18662 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AS) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 18663 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AS) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 18665 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AS) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 18666 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AS) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_DOUBLE_ARROW) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 18673 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AS) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_DOUBLE_ARROW) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 18675 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AS) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 18681 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AS) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_DOUBLE_ARROW) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 18686 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 357) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 357) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18690 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AS) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 18691 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AS) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 18692 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 564) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18693 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 563) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18694 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AS) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_DOUBLE_ARROW) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 18696 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AS) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 18698 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AS) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 18701 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AS) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 18704 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AS) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 18705 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AS) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 18706 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AS) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_DOUBLE_ARROW) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 18707 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AS) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_DOUBLE_ARROW) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 18708 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AS) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 18709 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AS) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 18712 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AS) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 18713 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AS) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 18714 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AS) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 18715 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AS) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 18717 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AS) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 18718 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_AS) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 18721 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_AS) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 18723 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AS) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 18724 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AS) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 18725 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AS) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 18726 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 433) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18727 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AS) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 18728 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AS) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 18730 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AS) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 18732 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AS) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 18733 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AS) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 18734 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 567) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18735 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 566) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18736 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 428) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18737 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AS) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 18740 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 18741 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 18742 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 18743 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AS) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 18744 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AS) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 18745 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AS) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 18747 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AS) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 18748 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AS) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 18749 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AS) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 18750 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AS) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 18751 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AS) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 18758 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AS) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 18760 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AS) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 18761 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AS) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 18762 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AS) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 18763 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AS) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 18765 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AS) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 18766 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AS) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 18767 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AS) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 18771 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AS) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 18772 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AS) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 18774 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AS) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 18775 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AS) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 18782 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AS) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 18784 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AS) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 18790 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AS) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 18795 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 357) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18799 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AS) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 18800 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AS) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 18801 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 564) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18802 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 563) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 18803 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AS) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 18805 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AS) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 18807 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AS) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 18810 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AS) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 18813 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AS) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 18814 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AS) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 18815 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AS) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 18816 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AS) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 18817 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AS) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 18818 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AS) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 18821 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AS) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 18822 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AS) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 18823 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AS) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 18824 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AS) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 18826 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AS) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 18827 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_AS) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 18830 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_AS) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 18832 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AS) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 18833 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AS) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 18834 ((("\"") shift) (("$") shift) (("(") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) shift) ((T_CLASS_C) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DIR) shift) ((T_FILE) shift) ((T_FUNC_C) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NS_C) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_TRAIT_C) shift) ((T_VARIABLE) shift) (("[") shift)) 18837 ((("(") reduce 515) ((")") reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) (("[") reduce 515) (("{") reduce 515)) 18842 ((("(") reduce 502) ((")") reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) (("[") reduce 511) (("{") reduce 511)) 18844 (((")") shift) ((T_DOUBLE_ARROW) shift)) 18845 ((("(") reduce 510) ((")") reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) (("[") reduce 510) (("{") reduce 510)) 18847 ((("(") reduce 505) ((")") reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) (("[") reduce 505) (("{") reduce 505)) 18848 (((")") reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) (("[") reduce 512) (("{") reduce 512)) 18849 (((")") reduce 204) ((T_DOUBLE_ARROW) reduce 204) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) (("[") reduce 496) (("{") reduce 496)) 18853 ((("(") reduce 529) ((")") reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) (("[") reduce 519) (("{") reduce 519)) 18855 ((("(") reduce 456) ((")") reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) (("[") reduce 456) (("{") reduce 456)) 18856 ((("(") reduce 281) ((")") reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) (("[") reduce 281) (("{") reduce 281)) 18860 ((("(") reduce 282) ((")") reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) (("[") reduce 282) (("{") reduce 282)) 18861 ((("(") reduce 283) ((")") reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) (("[") reduce 283) (("{") reduce 283)) 18862 ((("(") reduce 454) ((")") reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) (("[") reduce 454) (("{") reduce 454)) 18867 ((("(") reduce 515) ((")") reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) (("[") reduce 515) (("{") reduce 515)) 18872 ((("(") reduce 502) ((")") reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) (("[") reduce 511) (("{") reduce 511)) 18875 ((("(") reduce 510) ((")") reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) (("[") reduce 510) (("{") reduce 510)) 18877 ((("(") reduce 505) ((")") reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) (("[") reduce 505) (("{") reduce 505)) 18878 (((")") reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) (("[") reduce 512) (("{") reduce 512)) 18879 (((")") reduce 204) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) (("[") reduce 496) (("{") reduce 496)) 18883 ((("(") reduce 529) ((")") reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) (("[") reduce 519) (("{") reduce 519)) 18885 ((("(") reduce 456) ((")") reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) (("[") reduce 456) (("{") reduce 456)) 18886 ((("(") reduce 281) ((")") reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) (("[") reduce 281) (("{") reduce 281)) 18890 ((("(") reduce 282) ((")") reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) (("[") reduce 282) (("{") reduce 282)) 18891 ((("(") reduce 283) ((")") reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) (("[") reduce 283) (("{") reduce 283)) 18892 ((("(") reduce 454) ((")") reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) (("[") reduce 454) (("{") reduce 454)) 18894 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDFOREACH) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 18895 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CASE) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DEFAULT) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDSWITCH) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 18896 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CASE) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DEFAULT) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_ELSE) reduce 210) ((T_ELSEIF) reduce 210) ((T_EMPTY) reduce 210) ((T_ENDSWITCH) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 18897 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDFOREACH) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 18898 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDFOREACH) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 18913 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDFOREACH) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 18923 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDFOREACH) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 18925 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDFOREACH) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 18927 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDFOREACH) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 18929 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDFOREACH) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 18932 ((("!") reduce 102) (("\"") reduce 102) (("$") reduce 102) (("(") reduce 102) (("+") reduce 102) (("-") reduce 102) ((";") reduce 102) (("@") reduce 102) ((T_ABSTRACT) reduce 102) ((T_ARRAY) reduce 102) ((T_ARRAY_CAST) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_BOOL_CAST) reduce 102) ((T_BREAK) reduce 102) ((T_CLASS) reduce 102) ((T_CLASS_C) reduce 102) ((T_CLONE) reduce 102) ((T_CONSTANT_ENCAPSED_STRING) reduce 102) ((T_CONTINUE) reduce 102) ((T_DEC) reduce 102) ((T_DECLARE) reduce 102) ((T_DIR) reduce 102) ((T_DNUMBER) reduce 102) ((T_DO) reduce 102) ((T_DOUBLE_CAST) reduce 102) ((T_ECHO) reduce 102) ((T_EMPTY) reduce 102) ((T_ENDFOREACH) reduce 102) ((T_ENUM) reduce 102) ((T_EVAL) reduce 102) ((T_EXIT) reduce 102) ((T_FILE) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FOR) reduce 102) ((T_FOREACH) reduce 102) ((T_FUNCTION) reduce 102) ((T_FUNC_C) reduce 102) ((T_GLOBAL) reduce 102) ((T_GOTO) reduce 102) ((T_HALT_COMPILER) reduce 102) ((T_IF) reduce 102) ((T_INC) reduce 102) ((T_INCLUDE) reduce 102) ((T_INCLUDE_ONCE) reduce 102) ((T_INLINE_HTML) reduce 102) ((T_INTERFACE) reduce 102) ((T_INT_CAST) reduce 102) ((T_ISSET) reduce 102) ((T_LINE) reduce 102) ((T_LIST) reduce 102) ((T_LNUMBER) reduce 102) ((T_MATCH) reduce 102) ((T_METHOD_C) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_NEW) reduce 102) ((T_NS_C) reduce 102) ((T_OBJECT_CAST) reduce 102) ((T_PRINT) reduce 102) ((T_REQUIRE) reduce 102) ((T_REQUIRE_ONCE) reduce 102) ((T_RETURN) reduce 102) ((T_START_HEREDOC) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_STRING_CAST) reduce 102) ((T_SWITCH) reduce 102) ((T_THROW) reduce 102) ((T_TRAIT) reduce 102) ((T_TRAIT_C) reduce 102) ((T_TRY) reduce 102) ((T_UNSET) reduce 102) ((T_UNSET_CAST) reduce 102) ((T_VARIABLE) reduce 102) ((T_WHILE) reduce 102) ((T_YIELD) reduce 102) ((T_YIELD_FROM) reduce 102) (("[") reduce 102) (("`") reduce 102) (("{") reduce 102) (("~") reduce 102)) 18933 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDFOREACH) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 18934 ((("!") reduce 236) (("\"") reduce 236) (("$") reduce 236) (("(") reduce 236) (("+") reduce 236) (("-") reduce 236) ((";") reduce 236) (("@") reduce 236) ((T_ABSTRACT) reduce 236) ((T_ARRAY) reduce 236) ((T_ARRAY_CAST) reduce 236) ((T_ATTRIBUTE) reduce 236) ((T_BOOL_CAST) reduce 236) ((T_BREAK) reduce 236) ((T_CLASS) reduce 236) ((T_CLASS_C) reduce 236) ((T_CLONE) reduce 236) ((T_CONSTANT_ENCAPSED_STRING) reduce 236) ((T_CONTINUE) reduce 236) ((T_DEC) reduce 236) ((T_DECLARE) reduce 236) ((T_DIR) reduce 236) ((T_DNUMBER) reduce 236) ((T_DO) reduce 236) ((T_DOUBLE_CAST) reduce 236) ((T_ECHO) reduce 236) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 236) ((T_ENDFOREACH) reduce 236) ((T_ENUM) reduce 236) ((T_EVAL) reduce 236) ((T_EXIT) reduce 236) ((T_FILE) reduce 236) ((T_FINAL) reduce 236) ((T_FN) reduce 236) ((T_FOR) reduce 236) ((T_FOREACH) reduce 236) ((T_FUNCTION) reduce 236) ((T_FUNC_C) reduce 236) ((T_GLOBAL) reduce 236) ((T_GOTO) reduce 236) ((T_HALT_COMPILER) reduce 236) ((T_IF) reduce 236) ((T_INC) reduce 236) ((T_INCLUDE) reduce 236) ((T_INCLUDE_ONCE) reduce 236) ((T_INLINE_HTML) reduce 236) ((T_INTERFACE) reduce 236) ((T_INT_CAST) reduce 236) ((T_ISSET) reduce 236) ((T_LINE) reduce 236) ((T_LIST) reduce 236) ((T_LNUMBER) reduce 236) ((T_MATCH) reduce 236) ((T_METHOD_C) reduce 236) ((T_NAME_FULLY_QUALIFIED) reduce 236) ((T_NAME_QUALIFIED) reduce 236) ((T_NAME_RELATIVE) reduce 236) ((T_NEW) reduce 236) ((T_NS_C) reduce 236) ((T_OBJECT_CAST) reduce 236) ((T_PRINT) reduce 236) ((T_REQUIRE) reduce 236) ((T_REQUIRE_ONCE) reduce 236) ((T_RETURN) reduce 236) ((T_START_HEREDOC) reduce 236) ((T_STATIC) reduce 236) ((T_STRING) reduce 236) ((T_STRING_CAST) reduce 236) ((T_SWITCH) reduce 236) ((T_THROW) reduce 236) ((T_TRAIT) reduce 236) ((T_TRAIT_C) reduce 236) ((T_TRY) reduce 236) ((T_UNSET) reduce 236) ((T_UNSET_CAST) reduce 236) ((T_VARIABLE) reduce 236) ((T_WHILE) reduce 236) ((T_YIELD) reduce 236) ((T_YIELD_FROM) reduce 236) (("[") reduce 236) (("`") reduce 236) (("{") reduce 236) (("~") reduce 236)) 18935 ((("!") reduce 139) (("\"") reduce 139) (("$") reduce 139) (("(") reduce 139) (("+") reduce 139) (("-") reduce 139) ((";") reduce 139) (("@") reduce 139) ((T_ABSTRACT) reduce 139) ((T_ARRAY) reduce 139) ((T_ARRAY_CAST) reduce 139) ((T_ATTRIBUTE) reduce 139) ((T_BOOL_CAST) reduce 139) ((T_BREAK) reduce 139) ((T_CLASS) reduce 139) ((T_CLASS_C) reduce 139) ((T_CLONE) reduce 139) ((T_CONSTANT_ENCAPSED_STRING) reduce 139) ((T_CONTINUE) reduce 139) ((T_DEC) reduce 139) ((T_DECLARE) reduce 139) ((T_DIR) reduce 139) ((T_DNUMBER) reduce 139) ((T_DO) reduce 139) ((T_DOUBLE_CAST) reduce 139) ((T_ECHO) reduce 139) ((T_EMPTY) reduce 139) ((T_ENDFOREACH) reduce 139) ((T_ENUM) reduce 139) ((T_EVAL) reduce 139) ((T_EXIT) reduce 139) ((T_FILE) reduce 139) ((T_FINAL) reduce 139) ((T_FN) reduce 139) ((T_FOR) reduce 139) ((T_FOREACH) reduce 139) ((T_FUNCTION) reduce 139) ((T_FUNC_C) reduce 139) ((T_GLOBAL) reduce 139) ((T_GOTO) reduce 139) ((T_HALT_COMPILER) reduce 139) ((T_IF) reduce 139) ((T_INC) reduce 139) ((T_INCLUDE) reduce 139) ((T_INCLUDE_ONCE) reduce 139) ((T_INLINE_HTML) reduce 139) ((T_INTERFACE) reduce 139) ((T_INT_CAST) reduce 139) ((T_ISSET) reduce 139) ((T_LINE) reduce 139) ((T_LIST) reduce 139) ((T_LNUMBER) reduce 139) ((T_MATCH) reduce 139) ((T_METHOD_C) reduce 139) ((T_NAME_FULLY_QUALIFIED) reduce 139) ((T_NAME_QUALIFIED) reduce 139) ((T_NAME_RELATIVE) reduce 139) ((T_NEW) reduce 139) ((T_NS_C) reduce 139) ((T_OBJECT_CAST) reduce 139) ((T_PRINT) reduce 139) ((T_REQUIRE) reduce 139) ((T_REQUIRE_ONCE) reduce 139) ((T_RETURN) reduce 139) ((T_START_HEREDOC) reduce 139) ((T_STATIC) reduce 139) ((T_STRING) reduce 139) ((T_STRING_CAST) reduce 139) ((T_SWITCH) reduce 139) ((T_THROW) reduce 139) ((T_TRAIT) reduce 139) ((T_TRAIT_C) reduce 139) ((T_TRY) reduce 139) ((T_UNSET) reduce 139) ((T_UNSET_CAST) reduce 139) ((T_VARIABLE) reduce 139) ((T_WHILE) reduce 139) ((T_YIELD) reduce 139) ((T_YIELD_FROM) reduce 139) (("[") reduce 139) (("`") reduce 139) (("{") reduce 139) (("~") reduce 139)) 18936 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDFOREACH) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 18937 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDFOREACH) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 18938 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDFOREACH) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 18941 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDFOREACH) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 18947 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDFOREACH) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 18957 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDFOREACH) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 18965 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDFOREACH) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 18968 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDFOREACH) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 18970 ((("!") reduce 235) (("\"") reduce 235) (("$") reduce 235) (("(") reduce 235) (("+") reduce 235) (("-") reduce 235) ((";") reduce 235) (("@") reduce 235) ((T_ABSTRACT) reduce 235) ((T_ARRAY) reduce 235) ((T_ARRAY_CAST) reduce 235) ((T_ATTRIBUTE) reduce 235) ((T_BOOL_CAST) reduce 235) ((T_BREAK) reduce 235) ((T_CLASS) reduce 235) ((T_CLASS_C) reduce 235) ((T_CLONE) reduce 235) ((T_CONSTANT_ENCAPSED_STRING) reduce 235) ((T_CONTINUE) reduce 235) ((T_DEC) reduce 235) ((T_DECLARE) reduce 235) ((T_DIR) reduce 235) ((T_DNUMBER) reduce 235) ((T_DO) reduce 235) ((T_DOUBLE_CAST) reduce 235) ((T_ECHO) reduce 235) ((T_ELSE) reduce 235) ((T_ELSEIF) reduce 235) ((T_EMPTY) reduce 235) ((T_ENDFOREACH) reduce 235) ((T_ENUM) reduce 235) ((T_EVAL) reduce 235) ((T_EXIT) reduce 235) ((T_FILE) reduce 235) ((T_FINAL) reduce 235) ((T_FN) reduce 235) ((T_FOR) reduce 235) ((T_FOREACH) reduce 235) ((T_FUNCTION) reduce 235) ((T_FUNC_C) reduce 235) ((T_GLOBAL) reduce 235) ((T_GOTO) reduce 235) ((T_HALT_COMPILER) reduce 235) ((T_IF) reduce 235) ((T_INC) reduce 235) ((T_INCLUDE) reduce 235) ((T_INCLUDE_ONCE) reduce 235) ((T_INLINE_HTML) reduce 235) ((T_INTERFACE) reduce 235) ((T_INT_CAST) reduce 235) ((T_ISSET) reduce 235) ((T_LINE) reduce 235) ((T_LIST) reduce 235) ((T_LNUMBER) reduce 235) ((T_MATCH) reduce 235) ((T_METHOD_C) reduce 235) ((T_NAME_FULLY_QUALIFIED) reduce 235) ((T_NAME_QUALIFIED) reduce 235) ((T_NAME_RELATIVE) reduce 235) ((T_NEW) reduce 235) ((T_NS_C) reduce 235) ((T_OBJECT_CAST) reduce 235) ((T_PRINT) reduce 235) ((T_REQUIRE) reduce 235) ((T_REQUIRE_ONCE) reduce 235) ((T_RETURN) reduce 235) ((T_START_HEREDOC) reduce 235) ((T_STATIC) reduce 235) ((T_STRING) reduce 235) ((T_STRING_CAST) reduce 235) ((T_SWITCH) reduce 235) ((T_THROW) reduce 235) ((T_TRAIT) reduce 235) ((T_TRAIT_C) reduce 235) ((T_TRY) reduce 235) ((T_UNSET) reduce 235) ((T_UNSET_CAST) reduce 235) ((T_VARIABLE) reduce 235) ((T_WHILE) reduce 235) ((T_YIELD) reduce 235) ((T_YIELD_FROM) reduce 235) (("[") reduce 235) (("`") reduce 235) (("{") reduce 235) (("~") reduce 235)) 18973 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDFOREACH) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 18975 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_ELSE) reduce 237) ((T_ELSEIF) reduce 237) ((T_EMPTY) reduce 237) ((T_ENDFOREACH) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 18976 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDFOREACH) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 18979 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDFOREACH) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 18983 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDFOREACH) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 18988 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_ELSE) reduce 232) ((T_ELSEIF) reduce 232) ((T_EMPTY) reduce 232) ((T_ENDFOREACH) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 18989 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDFOREACH) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 18992 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_ELSE) reduce 233) ((T_ELSEIF) reduce 233) ((T_EMPTY) reduce 233) ((T_ENDFOREACH) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 18997 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDFOREACH) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 19000 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDFOREACH) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 19001 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_ELSE) reduce 174) ((T_ELSEIF) reduce 174) ((T_EMPTY) reduce 174) ((T_ENDFOREACH) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 19004 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDFOREACH) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 19007 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDFOREACH) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 19014 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDFOREACH) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 19019 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDFOREACH) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 19023 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_ELSE) reduce 214) ((T_ELSEIF) reduce 214) ((T_EMPTY) reduce 214) ((T_ENDFOREACH) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 19025 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_ELSE) reduce 215) ((T_ELSEIF) reduce 215) ((T_EMPTY) reduce 215) ((T_ENDFOREACH) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 19029 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDFOREACH) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 19032 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDFOREACH) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 19033 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDFOREACH) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 19035 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDFOREACH) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 19037 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDFOREACH) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 19039 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDFOREACH) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 19041 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDFOREACH) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 19051 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDFOREACH) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 19052 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_ELSE) reduce 210) ((T_ELSEIF) reduce 210) ((T_EMPTY) reduce 210) ((T_ENDFOREACH) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 19055 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_ELSE) reduce 211) ((T_ELSEIF) reduce 211) ((T_EMPTY) reduce 211) ((T_ENDFOREACH) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 19056 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDFOREACH) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 19057 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") reduce 348) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 19058 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 351) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 351) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19060 (((",") shift) ((";") reduce 349)) 19062 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 350) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 350) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19065 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 348) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 19066 ((("%") shift) ((")") reduce 351) (("*") shift) (("+") shift) ((",") reduce 351) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19068 (((")") reduce 349) ((",") shift)) 19070 ((("%") shift) ((")") reduce 350) (("*") shift) (("+") shift) ((",") reduce 350) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19072 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDFOR) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 19073 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDFOREACH) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 19074 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_ELSE) reduce 208) ((T_ELSEIF) reduce 208) ((T_EMPTY) reduce 208) ((T_ENDFOREACH) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 19075 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDFOR) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 19076 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDFOR) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 19091 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDFOR) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 19101 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDFOR) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 19103 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDFOR) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 19105 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDFOR) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 19107 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDFOR) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 19110 ((("!") reduce 102) (("\"") reduce 102) (("$") reduce 102) (("(") reduce 102) (("+") reduce 102) (("-") reduce 102) ((";") reduce 102) (("@") reduce 102) ((T_ABSTRACT) reduce 102) ((T_ARRAY) reduce 102) ((T_ARRAY_CAST) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_BOOL_CAST) reduce 102) ((T_BREAK) reduce 102) ((T_CLASS) reduce 102) ((T_CLASS_C) reduce 102) ((T_CLONE) reduce 102) ((T_CONSTANT_ENCAPSED_STRING) reduce 102) ((T_CONTINUE) reduce 102) ((T_DEC) reduce 102) ((T_DECLARE) reduce 102) ((T_DIR) reduce 102) ((T_DNUMBER) reduce 102) ((T_DO) reduce 102) ((T_DOUBLE_CAST) reduce 102) ((T_ECHO) reduce 102) ((T_EMPTY) reduce 102) ((T_ENDFOR) reduce 102) ((T_ENUM) reduce 102) ((T_EVAL) reduce 102) ((T_EXIT) reduce 102) ((T_FILE) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FOR) reduce 102) ((T_FOREACH) reduce 102) ((T_FUNCTION) reduce 102) ((T_FUNC_C) reduce 102) ((T_GLOBAL) reduce 102) ((T_GOTO) reduce 102) ((T_HALT_COMPILER) reduce 102) ((T_IF) reduce 102) ((T_INC) reduce 102) ((T_INCLUDE) reduce 102) ((T_INCLUDE_ONCE) reduce 102) ((T_INLINE_HTML) reduce 102) ((T_INTERFACE) reduce 102) ((T_INT_CAST) reduce 102) ((T_ISSET) reduce 102) ((T_LINE) reduce 102) ((T_LIST) reduce 102) ((T_LNUMBER) reduce 102) ((T_MATCH) reduce 102) ((T_METHOD_C) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_NEW) reduce 102) ((T_NS_C) reduce 102) ((T_OBJECT_CAST) reduce 102) ((T_PRINT) reduce 102) ((T_REQUIRE) reduce 102) ((T_REQUIRE_ONCE) reduce 102) ((T_RETURN) reduce 102) ((T_START_HEREDOC) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_STRING_CAST) reduce 102) ((T_SWITCH) reduce 102) ((T_THROW) reduce 102) ((T_TRAIT) reduce 102) ((T_TRAIT_C) reduce 102) ((T_TRY) reduce 102) ((T_UNSET) reduce 102) ((T_UNSET_CAST) reduce 102) ((T_VARIABLE) reduce 102) ((T_WHILE) reduce 102) ((T_YIELD) reduce 102) ((T_YIELD_FROM) reduce 102) (("[") reduce 102) (("`") reduce 102) (("{") reduce 102) (("~") reduce 102)) 19111 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDFOR) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 19112 ((("!") reduce 236) (("\"") reduce 236) (("$") reduce 236) (("(") reduce 236) (("+") reduce 236) (("-") reduce 236) ((";") reduce 236) (("@") reduce 236) ((T_ABSTRACT) reduce 236) ((T_ARRAY) reduce 236) ((T_ARRAY_CAST) reduce 236) ((T_ATTRIBUTE) reduce 236) ((T_BOOL_CAST) reduce 236) ((T_BREAK) reduce 236) ((T_CLASS) reduce 236) ((T_CLASS_C) reduce 236) ((T_CLONE) reduce 236) ((T_CONSTANT_ENCAPSED_STRING) reduce 236) ((T_CONTINUE) reduce 236) ((T_DEC) reduce 236) ((T_DECLARE) reduce 236) ((T_DIR) reduce 236) ((T_DNUMBER) reduce 236) ((T_DO) reduce 236) ((T_DOUBLE_CAST) reduce 236) ((T_ECHO) reduce 236) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 236) ((T_ENDFOR) reduce 236) ((T_ENUM) reduce 236) ((T_EVAL) reduce 236) ((T_EXIT) reduce 236) ((T_FILE) reduce 236) ((T_FINAL) reduce 236) ((T_FN) reduce 236) ((T_FOR) reduce 236) ((T_FOREACH) reduce 236) ((T_FUNCTION) reduce 236) ((T_FUNC_C) reduce 236) ((T_GLOBAL) reduce 236) ((T_GOTO) reduce 236) ((T_HALT_COMPILER) reduce 236) ((T_IF) reduce 236) ((T_INC) reduce 236) ((T_INCLUDE) reduce 236) ((T_INCLUDE_ONCE) reduce 236) ((T_INLINE_HTML) reduce 236) ((T_INTERFACE) reduce 236) ((T_INT_CAST) reduce 236) ((T_ISSET) reduce 236) ((T_LINE) reduce 236) ((T_LIST) reduce 236) ((T_LNUMBER) reduce 236) ((T_MATCH) reduce 236) ((T_METHOD_C) reduce 236) ((T_NAME_FULLY_QUALIFIED) reduce 236) ((T_NAME_QUALIFIED) reduce 236) ((T_NAME_RELATIVE) reduce 236) ((T_NEW) reduce 236) ((T_NS_C) reduce 236) ((T_OBJECT_CAST) reduce 236) ((T_PRINT) reduce 236) ((T_REQUIRE) reduce 236) ((T_REQUIRE_ONCE) reduce 236) ((T_RETURN) reduce 236) ((T_START_HEREDOC) reduce 236) ((T_STATIC) reduce 236) ((T_STRING) reduce 236) ((T_STRING_CAST) reduce 236) ((T_SWITCH) reduce 236) ((T_THROW) reduce 236) ((T_TRAIT) reduce 236) ((T_TRAIT_C) reduce 236) ((T_TRY) reduce 236) ((T_UNSET) reduce 236) ((T_UNSET_CAST) reduce 236) ((T_VARIABLE) reduce 236) ((T_WHILE) reduce 236) ((T_YIELD) reduce 236) ((T_YIELD_FROM) reduce 236) (("[") reduce 236) (("`") reduce 236) (("{") reduce 236) (("~") reduce 236)) 19113 ((("!") reduce 139) (("\"") reduce 139) (("$") reduce 139) (("(") reduce 139) (("+") reduce 139) (("-") reduce 139) ((";") reduce 139) (("@") reduce 139) ((T_ABSTRACT) reduce 139) ((T_ARRAY) reduce 139) ((T_ARRAY_CAST) reduce 139) ((T_ATTRIBUTE) reduce 139) ((T_BOOL_CAST) reduce 139) ((T_BREAK) reduce 139) ((T_CLASS) reduce 139) ((T_CLASS_C) reduce 139) ((T_CLONE) reduce 139) ((T_CONSTANT_ENCAPSED_STRING) reduce 139) ((T_CONTINUE) reduce 139) ((T_DEC) reduce 139) ((T_DECLARE) reduce 139) ((T_DIR) reduce 139) ((T_DNUMBER) reduce 139) ((T_DO) reduce 139) ((T_DOUBLE_CAST) reduce 139) ((T_ECHO) reduce 139) ((T_EMPTY) reduce 139) ((T_ENDFOR) reduce 139) ((T_ENUM) reduce 139) ((T_EVAL) reduce 139) ((T_EXIT) reduce 139) ((T_FILE) reduce 139) ((T_FINAL) reduce 139) ((T_FN) reduce 139) ((T_FOR) reduce 139) ((T_FOREACH) reduce 139) ((T_FUNCTION) reduce 139) ((T_FUNC_C) reduce 139) ((T_GLOBAL) reduce 139) ((T_GOTO) reduce 139) ((T_HALT_COMPILER) reduce 139) ((T_IF) reduce 139) ((T_INC) reduce 139) ((T_INCLUDE) reduce 139) ((T_INCLUDE_ONCE) reduce 139) ((T_INLINE_HTML) reduce 139) ((T_INTERFACE) reduce 139) ((T_INT_CAST) reduce 139) ((T_ISSET) reduce 139) ((T_LINE) reduce 139) ((T_LIST) reduce 139) ((T_LNUMBER) reduce 139) ((T_MATCH) reduce 139) ((T_METHOD_C) reduce 139) ((T_NAME_FULLY_QUALIFIED) reduce 139) ((T_NAME_QUALIFIED) reduce 139) ((T_NAME_RELATIVE) reduce 139) ((T_NEW) reduce 139) ((T_NS_C) reduce 139) ((T_OBJECT_CAST) reduce 139) ((T_PRINT) reduce 139) ((T_REQUIRE) reduce 139) ((T_REQUIRE_ONCE) reduce 139) ((T_RETURN) reduce 139) ((T_START_HEREDOC) reduce 139) ((T_STATIC) reduce 139) ((T_STRING) reduce 139) ((T_STRING_CAST) reduce 139) ((T_SWITCH) reduce 139) ((T_THROW) reduce 139) ((T_TRAIT) reduce 139) ((T_TRAIT_C) reduce 139) ((T_TRY) reduce 139) ((T_UNSET) reduce 139) ((T_UNSET_CAST) reduce 139) ((T_VARIABLE) reduce 139) ((T_WHILE) reduce 139) ((T_YIELD) reduce 139) ((T_YIELD_FROM) reduce 139) (("[") reduce 139) (("`") reduce 139) (("{") reduce 139) (("~") reduce 139)) 19114 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDFOR) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 19115 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDFOR) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 19116 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDFOR) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 19119 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDFOR) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 19125 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDFOR) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 19135 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDFOR) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 19143 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDFOR) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 19146 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDFOR) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 19148 ((("!") reduce 235) (("\"") reduce 235) (("$") reduce 235) (("(") reduce 235) (("+") reduce 235) (("-") reduce 235) ((";") reduce 235) (("@") reduce 235) ((T_ABSTRACT) reduce 235) ((T_ARRAY) reduce 235) ((T_ARRAY_CAST) reduce 235) ((T_ATTRIBUTE) reduce 235) ((T_BOOL_CAST) reduce 235) ((T_BREAK) reduce 235) ((T_CLASS) reduce 235) ((T_CLASS_C) reduce 235) ((T_CLONE) reduce 235) ((T_CONSTANT_ENCAPSED_STRING) reduce 235) ((T_CONTINUE) reduce 235) ((T_DEC) reduce 235) ((T_DECLARE) reduce 235) ((T_DIR) reduce 235) ((T_DNUMBER) reduce 235) ((T_DO) reduce 235) ((T_DOUBLE_CAST) reduce 235) ((T_ECHO) reduce 235) ((T_ELSE) reduce 235) ((T_ELSEIF) reduce 235) ((T_EMPTY) reduce 235) ((T_ENDFOR) reduce 235) ((T_ENUM) reduce 235) ((T_EVAL) reduce 235) ((T_EXIT) reduce 235) ((T_FILE) reduce 235) ((T_FINAL) reduce 235) ((T_FN) reduce 235) ((T_FOR) reduce 235) ((T_FOREACH) reduce 235) ((T_FUNCTION) reduce 235) ((T_FUNC_C) reduce 235) ((T_GLOBAL) reduce 235) ((T_GOTO) reduce 235) ((T_HALT_COMPILER) reduce 235) ((T_IF) reduce 235) ((T_INC) reduce 235) ((T_INCLUDE) reduce 235) ((T_INCLUDE_ONCE) reduce 235) ((T_INLINE_HTML) reduce 235) ((T_INTERFACE) reduce 235) ((T_INT_CAST) reduce 235) ((T_ISSET) reduce 235) ((T_LINE) reduce 235) ((T_LIST) reduce 235) ((T_LNUMBER) reduce 235) ((T_MATCH) reduce 235) ((T_METHOD_C) reduce 235) ((T_NAME_FULLY_QUALIFIED) reduce 235) ((T_NAME_QUALIFIED) reduce 235) ((T_NAME_RELATIVE) reduce 235) ((T_NEW) reduce 235) ((T_NS_C) reduce 235) ((T_OBJECT_CAST) reduce 235) ((T_PRINT) reduce 235) ((T_REQUIRE) reduce 235) ((T_REQUIRE_ONCE) reduce 235) ((T_RETURN) reduce 235) ((T_START_HEREDOC) reduce 235) ((T_STATIC) reduce 235) ((T_STRING) reduce 235) ((T_STRING_CAST) reduce 235) ((T_SWITCH) reduce 235) ((T_THROW) reduce 235) ((T_TRAIT) reduce 235) ((T_TRAIT_C) reduce 235) ((T_TRY) reduce 235) ((T_UNSET) reduce 235) ((T_UNSET_CAST) reduce 235) ((T_VARIABLE) reduce 235) ((T_WHILE) reduce 235) ((T_YIELD) reduce 235) ((T_YIELD_FROM) reduce 235) (("[") reduce 235) (("`") reduce 235) (("{") reduce 235) (("~") reduce 235)) 19151 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDFOR) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 19153 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_ELSE) reduce 237) ((T_ELSEIF) reduce 237) ((T_EMPTY) reduce 237) ((T_ENDFOR) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 19154 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDFOR) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 19157 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDFOR) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 19161 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDFOR) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 19166 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_ELSE) reduce 232) ((T_ELSEIF) reduce 232) ((T_EMPTY) reduce 232) ((T_ENDFOR) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 19167 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDFOR) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 19170 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_ELSE) reduce 233) ((T_ELSEIF) reduce 233) ((T_EMPTY) reduce 233) ((T_ENDFOR) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 19175 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDFOR) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 19178 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDFOR) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 19179 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_ELSE) reduce 174) ((T_ELSEIF) reduce 174) ((T_EMPTY) reduce 174) ((T_ENDFOR) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 19182 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDFOR) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 19185 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDFOR) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 19192 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDFOR) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 19197 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDFOR) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 19201 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_ELSE) reduce 214) ((T_ELSEIF) reduce 214) ((T_EMPTY) reduce 214) ((T_ENDFOR) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 19203 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_ELSE) reduce 215) ((T_ELSEIF) reduce 215) ((T_EMPTY) reduce 215) ((T_ENDFOR) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 19207 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDFOR) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 19210 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDFOR) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 19211 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDFOR) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 19213 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDFOR) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 19215 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDFOR) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 19217 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDFOR) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 19219 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDFOR) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 19229 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDFOR) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 19230 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_ELSE) reduce 210) ((T_ELSEIF) reduce 210) ((T_EMPTY) reduce 210) ((T_ENDFOR) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 19233 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_ELSE) reduce 211) ((T_ELSEIF) reduce 211) ((T_EMPTY) reduce 211) ((T_ENDFOR) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 19234 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDFOR) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 19243 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDFOR) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 19244 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_ELSE) reduce 208) ((T_ELSEIF) reduce 208) ((T_EMPTY) reduce 208) ((T_ENDFOR) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 19247 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_ELSE) reduce 209) ((T_ELSEIF) reduce 209) ((T_EMPTY) reduce 209) ((T_ENDFOR) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 19248 (((",") reduce 346) ((";") reduce 346)) 19250 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 347) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 347) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19252 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDFOR) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 19253 (((",") reduce 345) ((";") reduce 345)) 19254 (((T_WHILE) reduce 164)) 19265 (((T_WHILE) reduce 158)) 19273 (((T_WHILE) reduce 147)) 19276 (((T_WHILE) reduce 146)) 19277 (((T_ELSE) shift) ((T_ELSEIF) shift) ((T_WHILE) reduce 236)) 19278 (((T_WHILE) shift)) 19281 (((T_WHILE) reduce 145)) 19286 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDFOR) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 19292 (((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_WHILE) reduce 164)) 19302 (((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_WHILE) reduce 158)) 19310 (((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_WHILE) reduce 147)) 19313 (((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_WHILE) reduce 146)) 19315 (((T_ELSE) reduce 235) ((T_ELSEIF) reduce 235) ((T_WHILE) reduce 235)) 19318 (((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_WHILE) reduce 145)) 19320 (((T_ELSE) reduce 237) ((T_ELSEIF) reduce 237) ((T_WHILE) reduce 237)) 19321 (((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_WHILE) reduce 159)) 19324 (((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_WHILE) reduce 240)) 19328 (((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_WHILE) reduce 241)) 19333 (((T_ELSE) reduce 232) ((T_ELSEIF) reduce 232) ((T_WHILE) reduce 232)) 19334 (((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_WHILE) reduce 148)) 19337 (((T_ELSE) reduce 233) ((T_ELSEIF) reduce 233) ((T_WHILE) reduce 233)) 19342 (((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_WHILE) reduce 160)) 19345 (((T_CATCH) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_FINALLY) reduce 168) ((T_WHILE) reduce 168)) 19346 (((T_CATCH) shift) ((T_ELSE) reduce 174) ((T_ELSEIF) reduce 174) ((T_FINALLY) shift) ((T_WHILE) reduce 174)) 19349 (((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_WHILE) reduce 165)) 19352 (((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_WHILE) reduce 175)) 19359 (((T_CATCH) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_FINALLY) reduce 169) ((T_WHILE) reduce 169)) 19364 (((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_WHILE) reduce 151)) 19368 (((T_ELSE) reduce 214) ((T_ELSEIF) reduce 214) ((T_WHILE) reduce 214)) 19370 (((T_ELSE) reduce 215) ((T_ELSEIF) reduce 215) ((T_WHILE) reduce 215)) 19374 (((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_WHILE) reduce 216)) 19377 (((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_WHILE) reduce 217)) 19378 (((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_WHILE) reduce 167)) 19380 (((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_WHILE) reduce 156)) 19382 (((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_WHILE) reduce 154)) 19384 (((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_WHILE) reduce 166)) 19386 (((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_WHILE) reduce 155)) 19396 (((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_WHILE) reduce 162)) 19397 (((T_ELSE) reduce 210) ((T_ELSEIF) reduce 210) ((T_WHILE) reduce 210)) 19400 (((T_ELSE) reduce 211) ((T_ELSEIF) reduce 211) ((T_WHILE) reduce 211)) 19401 (((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_WHILE) reduce 161)) 19410 (((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_WHILE) reduce 150)) 19411 (((T_ELSE) reduce 208) ((T_ELSEIF) reduce 208) ((T_WHILE) reduce 208)) 19414 (((T_ELSE) reduce 209) ((T_ELSEIF) reduce 209) ((T_WHILE) reduce 209)) 19416 (((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_WHILE) reduce 157)) 19422 (((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_WHILE) reduce 149)) 19425 (((")") reduce 138) ((",") reduce 138)) 19426 (((")") shift) ((",") shift)) 19429 (((")") reduce 137) ((",") reduce 137)) 19430 ((("!") reduce 140) (("\"") reduce 140) (("$") reduce 140) (("(") reduce 140) (("+") reduce 140) (("-") reduce 140) ((";") reduce 140) (("@") reduce 140) ((T_ABSTRACT) reduce 140) ((T_ARRAY) reduce 140) ((T_ARRAY_CAST) reduce 140) ((T_ATTRIBUTE) reduce 140) ((T_BOOL_CAST) reduce 140) ((T_BREAK) reduce 140) ((T_CLASS) reduce 140) ((T_CLASS_C) reduce 140) ((T_CLONE) reduce 140) ((T_CONSTANT_ENCAPSED_STRING) reduce 140) ((T_CONTINUE) reduce 140) ((T_DEC) reduce 140) ((T_DECLARE) reduce 140) ((T_DIR) reduce 140) ((T_DNUMBER) reduce 140) ((T_DO) reduce 140) ((T_DOUBLE_CAST) reduce 140) ((T_ECHO) reduce 140) ((T_EMPTY) reduce 140) ((T_ENDDECLARE) reduce 140) ((T_ENUM) reduce 140) ((T_EVAL) reduce 140) ((T_EXIT) reduce 140) ((T_FILE) reduce 140) ((T_FINAL) reduce 140) ((T_FN) reduce 140) ((T_FOR) reduce 140) ((T_FOREACH) reduce 140) ((T_FUNCTION) reduce 140) ((T_FUNC_C) reduce 140) ((T_GLOBAL) reduce 140) ((T_GOTO) reduce 140) ((T_HALT_COMPILER) reduce 140) ((T_IF) reduce 140) ((T_INC) reduce 140) ((T_INCLUDE) reduce 140) ((T_INCLUDE_ONCE) reduce 140) ((T_INLINE_HTML) reduce 140) ((T_INTERFACE) reduce 140) ((T_INT_CAST) reduce 140) ((T_ISSET) reduce 140) ((T_LINE) reduce 140) ((T_LIST) reduce 140) ((T_LNUMBER) reduce 140) ((T_MATCH) reduce 140) ((T_METHOD_C) reduce 140) ((T_NAME_FULLY_QUALIFIED) reduce 140) ((T_NAME_QUALIFIED) reduce 140) ((T_NAME_RELATIVE) reduce 140) ((T_NEW) reduce 140) ((T_NS_C) reduce 140) ((T_OBJECT_CAST) reduce 140) ((T_PRINT) reduce 140) ((T_REQUIRE) reduce 140) ((T_REQUIRE_ONCE) reduce 140) ((T_RETURN) reduce 140) ((T_START_HEREDOC) reduce 140) ((T_STATIC) reduce 140) ((T_STRING) reduce 140) ((T_STRING_CAST) reduce 140) ((T_SWITCH) reduce 140) ((T_THROW) reduce 140) ((T_TRAIT) reduce 140) ((T_TRAIT_C) reduce 140) ((T_TRY) reduce 140) ((T_UNSET) reduce 140) ((T_UNSET_CAST) reduce 140) ((T_VARIABLE) reduce 140) ((T_WHILE) reduce 140) ((T_YIELD) reduce 140) ((T_YIELD_FROM) reduce 140) (("[") reduce 140) (("`") reduce 140) (("{") reduce 140) (("~") reduce 140)) 19431 (((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_WHILE) reduce 163)) 19432 (((T_ELSE) reduce 212) ((T_ELSEIF) reduce 212) ((T_WHILE) reduce 212)) 19433 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 19434 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDDECLARE) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 19449 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDDECLARE) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 19459 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDDECLARE) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 19461 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDDECLARE) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 19463 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDDECLARE) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 19465 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDDECLARE) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 19468 ((("!") reduce 102) (("\"") reduce 102) (("$") reduce 102) (("(") reduce 102) (("+") reduce 102) (("-") reduce 102) ((";") reduce 102) (("@") reduce 102) ((T_ABSTRACT) reduce 102) ((T_ARRAY) reduce 102) ((T_ARRAY_CAST) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_BOOL_CAST) reduce 102) ((T_BREAK) reduce 102) ((T_CLASS) reduce 102) ((T_CLASS_C) reduce 102) ((T_CLONE) reduce 102) ((T_CONSTANT_ENCAPSED_STRING) reduce 102) ((T_CONTINUE) reduce 102) ((T_DEC) reduce 102) ((T_DECLARE) reduce 102) ((T_DIR) reduce 102) ((T_DNUMBER) reduce 102) ((T_DO) reduce 102) ((T_DOUBLE_CAST) reduce 102) ((T_ECHO) reduce 102) ((T_EMPTY) reduce 102) ((T_ENDDECLARE) reduce 102) ((T_ENUM) reduce 102) ((T_EVAL) reduce 102) ((T_EXIT) reduce 102) ((T_FILE) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FOR) reduce 102) ((T_FOREACH) reduce 102) ((T_FUNCTION) reduce 102) ((T_FUNC_C) reduce 102) ((T_GLOBAL) reduce 102) ((T_GOTO) reduce 102) ((T_HALT_COMPILER) reduce 102) ((T_IF) reduce 102) ((T_INC) reduce 102) ((T_INCLUDE) reduce 102) ((T_INCLUDE_ONCE) reduce 102) ((T_INLINE_HTML) reduce 102) ((T_INTERFACE) reduce 102) ((T_INT_CAST) reduce 102) ((T_ISSET) reduce 102) ((T_LINE) reduce 102) ((T_LIST) reduce 102) ((T_LNUMBER) reduce 102) ((T_MATCH) reduce 102) ((T_METHOD_C) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_NEW) reduce 102) ((T_NS_C) reduce 102) ((T_OBJECT_CAST) reduce 102) ((T_PRINT) reduce 102) ((T_REQUIRE) reduce 102) ((T_REQUIRE_ONCE) reduce 102) ((T_RETURN) reduce 102) ((T_START_HEREDOC) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_STRING_CAST) reduce 102) ((T_SWITCH) reduce 102) ((T_THROW) reduce 102) ((T_TRAIT) reduce 102) ((T_TRAIT_C) reduce 102) ((T_TRY) reduce 102) ((T_UNSET) reduce 102) ((T_UNSET_CAST) reduce 102) ((T_VARIABLE) reduce 102) ((T_WHILE) reduce 102) ((T_YIELD) reduce 102) ((T_YIELD_FROM) reduce 102) (("[") reduce 102) (("`") reduce 102) (("{") reduce 102) (("~") reduce 102)) 19469 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDDECLARE) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 19470 ((("!") reduce 236) (("\"") reduce 236) (("$") reduce 236) (("(") reduce 236) (("+") reduce 236) (("-") reduce 236) ((";") reduce 236) (("@") reduce 236) ((T_ABSTRACT) reduce 236) ((T_ARRAY) reduce 236) ((T_ARRAY_CAST) reduce 236) ((T_ATTRIBUTE) reduce 236) ((T_BOOL_CAST) reduce 236) ((T_BREAK) reduce 236) ((T_CLASS) reduce 236) ((T_CLASS_C) reduce 236) ((T_CLONE) reduce 236) ((T_CONSTANT_ENCAPSED_STRING) reduce 236) ((T_CONTINUE) reduce 236) ((T_DEC) reduce 236) ((T_DECLARE) reduce 236) ((T_DIR) reduce 236) ((T_DNUMBER) reduce 236) ((T_DO) reduce 236) ((T_DOUBLE_CAST) reduce 236) ((T_ECHO) reduce 236) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 236) ((T_ENDDECLARE) reduce 236) ((T_ENUM) reduce 236) ((T_EVAL) reduce 236) ((T_EXIT) reduce 236) ((T_FILE) reduce 236) ((T_FINAL) reduce 236) ((T_FN) reduce 236) ((T_FOR) reduce 236) ((T_FOREACH) reduce 236) ((T_FUNCTION) reduce 236) ((T_FUNC_C) reduce 236) ((T_GLOBAL) reduce 236) ((T_GOTO) reduce 236) ((T_HALT_COMPILER) reduce 236) ((T_IF) reduce 236) ((T_INC) reduce 236) ((T_INCLUDE) reduce 236) ((T_INCLUDE_ONCE) reduce 236) ((T_INLINE_HTML) reduce 236) ((T_INTERFACE) reduce 236) ((T_INT_CAST) reduce 236) ((T_ISSET) reduce 236) ((T_LINE) reduce 236) ((T_LIST) reduce 236) ((T_LNUMBER) reduce 236) ((T_MATCH) reduce 236) ((T_METHOD_C) reduce 236) ((T_NAME_FULLY_QUALIFIED) reduce 236) ((T_NAME_QUALIFIED) reduce 236) ((T_NAME_RELATIVE) reduce 236) ((T_NEW) reduce 236) ((T_NS_C) reduce 236) ((T_OBJECT_CAST) reduce 236) ((T_PRINT) reduce 236) ((T_REQUIRE) reduce 236) ((T_REQUIRE_ONCE) reduce 236) ((T_RETURN) reduce 236) ((T_START_HEREDOC) reduce 236) ((T_STATIC) reduce 236) ((T_STRING) reduce 236) ((T_STRING_CAST) reduce 236) ((T_SWITCH) reduce 236) ((T_THROW) reduce 236) ((T_TRAIT) reduce 236) ((T_TRAIT_C) reduce 236) ((T_TRY) reduce 236) ((T_UNSET) reduce 236) ((T_UNSET_CAST) reduce 236) ((T_VARIABLE) reduce 236) ((T_WHILE) reduce 236) ((T_YIELD) reduce 236) ((T_YIELD_FROM) reduce 236) (("[") reduce 236) (("`") reduce 236) (("{") reduce 236) (("~") reduce 236)) 19471 ((("!") reduce 139) (("\"") reduce 139) (("$") reduce 139) (("(") reduce 139) (("+") reduce 139) (("-") reduce 139) ((";") reduce 139) (("@") reduce 139) ((T_ABSTRACT) reduce 139) ((T_ARRAY) reduce 139) ((T_ARRAY_CAST) reduce 139) ((T_ATTRIBUTE) reduce 139) ((T_BOOL_CAST) reduce 139) ((T_BREAK) reduce 139) ((T_CLASS) reduce 139) ((T_CLASS_C) reduce 139) ((T_CLONE) reduce 139) ((T_CONSTANT_ENCAPSED_STRING) reduce 139) ((T_CONTINUE) reduce 139) ((T_DEC) reduce 139) ((T_DECLARE) reduce 139) ((T_DIR) reduce 139) ((T_DNUMBER) reduce 139) ((T_DO) reduce 139) ((T_DOUBLE_CAST) reduce 139) ((T_ECHO) reduce 139) ((T_EMPTY) reduce 139) ((T_ENDDECLARE) reduce 139) ((T_ENUM) reduce 139) ((T_EVAL) reduce 139) ((T_EXIT) reduce 139) ((T_FILE) reduce 139) ((T_FINAL) reduce 139) ((T_FN) reduce 139) ((T_FOR) reduce 139) ((T_FOREACH) reduce 139) ((T_FUNCTION) reduce 139) ((T_FUNC_C) reduce 139) ((T_GLOBAL) reduce 139) ((T_GOTO) reduce 139) ((T_HALT_COMPILER) reduce 139) ((T_IF) reduce 139) ((T_INC) reduce 139) ((T_INCLUDE) reduce 139) ((T_INCLUDE_ONCE) reduce 139) ((T_INLINE_HTML) reduce 139) ((T_INTERFACE) reduce 139) ((T_INT_CAST) reduce 139) ((T_ISSET) reduce 139) ((T_LINE) reduce 139) ((T_LIST) reduce 139) ((T_LNUMBER) reduce 139) ((T_MATCH) reduce 139) ((T_METHOD_C) reduce 139) ((T_NAME_FULLY_QUALIFIED) reduce 139) ((T_NAME_QUALIFIED) reduce 139) ((T_NAME_RELATIVE) reduce 139) ((T_NEW) reduce 139) ((T_NS_C) reduce 139) ((T_OBJECT_CAST) reduce 139) ((T_PRINT) reduce 139) ((T_REQUIRE) reduce 139) ((T_REQUIRE_ONCE) reduce 139) ((T_RETURN) reduce 139) ((T_START_HEREDOC) reduce 139) ((T_STATIC) reduce 139) ((T_STRING) reduce 139) ((T_STRING_CAST) reduce 139) ((T_SWITCH) reduce 139) ((T_THROW) reduce 139) ((T_TRAIT) reduce 139) ((T_TRAIT_C) reduce 139) ((T_TRY) reduce 139) ((T_UNSET) reduce 139) ((T_UNSET_CAST) reduce 139) ((T_VARIABLE) reduce 139) ((T_WHILE) reduce 139) ((T_YIELD) reduce 139) ((T_YIELD_FROM) reduce 139) (("[") reduce 139) (("`") reduce 139) (("{") reduce 139) (("~") reduce 139)) 19472 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDDECLARE) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 19473 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDDECLARE) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 19474 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDDECLARE) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 19477 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDDECLARE) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 19483 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDDECLARE) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 19493 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDDECLARE) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 19501 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDDECLARE) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 19504 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDDECLARE) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 19506 ((("!") reduce 235) (("\"") reduce 235) (("$") reduce 235) (("(") reduce 235) (("+") reduce 235) (("-") reduce 235) ((";") reduce 235) (("@") reduce 235) ((T_ABSTRACT) reduce 235) ((T_ARRAY) reduce 235) ((T_ARRAY_CAST) reduce 235) ((T_ATTRIBUTE) reduce 235) ((T_BOOL_CAST) reduce 235) ((T_BREAK) reduce 235) ((T_CLASS) reduce 235) ((T_CLASS_C) reduce 235) ((T_CLONE) reduce 235) ((T_CONSTANT_ENCAPSED_STRING) reduce 235) ((T_CONTINUE) reduce 235) ((T_DEC) reduce 235) ((T_DECLARE) reduce 235) ((T_DIR) reduce 235) ((T_DNUMBER) reduce 235) ((T_DO) reduce 235) ((T_DOUBLE_CAST) reduce 235) ((T_ECHO) reduce 235) ((T_ELSE) reduce 235) ((T_ELSEIF) reduce 235) ((T_EMPTY) reduce 235) ((T_ENDDECLARE) reduce 235) ((T_ENUM) reduce 235) ((T_EVAL) reduce 235) ((T_EXIT) reduce 235) ((T_FILE) reduce 235) ((T_FINAL) reduce 235) ((T_FN) reduce 235) ((T_FOR) reduce 235) ((T_FOREACH) reduce 235) ((T_FUNCTION) reduce 235) ((T_FUNC_C) reduce 235) ((T_GLOBAL) reduce 235) ((T_GOTO) reduce 235) ((T_HALT_COMPILER) reduce 235) ((T_IF) reduce 235) ((T_INC) reduce 235) ((T_INCLUDE) reduce 235) ((T_INCLUDE_ONCE) reduce 235) ((T_INLINE_HTML) reduce 235) ((T_INTERFACE) reduce 235) ((T_INT_CAST) reduce 235) ((T_ISSET) reduce 235) ((T_LINE) reduce 235) ((T_LIST) reduce 235) ((T_LNUMBER) reduce 235) ((T_MATCH) reduce 235) ((T_METHOD_C) reduce 235) ((T_NAME_FULLY_QUALIFIED) reduce 235) ((T_NAME_QUALIFIED) reduce 235) ((T_NAME_RELATIVE) reduce 235) ((T_NEW) reduce 235) ((T_NS_C) reduce 235) ((T_OBJECT_CAST) reduce 235) ((T_PRINT) reduce 235) ((T_REQUIRE) reduce 235) ((T_REQUIRE_ONCE) reduce 235) ((T_RETURN) reduce 235) ((T_START_HEREDOC) reduce 235) ((T_STATIC) reduce 235) ((T_STRING) reduce 235) ((T_STRING_CAST) reduce 235) ((T_SWITCH) reduce 235) ((T_THROW) reduce 235) ((T_TRAIT) reduce 235) ((T_TRAIT_C) reduce 235) ((T_TRY) reduce 235) ((T_UNSET) reduce 235) ((T_UNSET_CAST) reduce 235) ((T_VARIABLE) reduce 235) ((T_WHILE) reduce 235) ((T_YIELD) reduce 235) ((T_YIELD_FROM) reduce 235) (("[") reduce 235) (("`") reduce 235) (("{") reduce 235) (("~") reduce 235)) 19509 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDDECLARE) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 19511 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_ELSE) reduce 237) ((T_ELSEIF) reduce 237) ((T_EMPTY) reduce 237) ((T_ENDDECLARE) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 19512 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDDECLARE) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 19515 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDDECLARE) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 19519 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDDECLARE) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 19524 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_ELSE) reduce 232) ((T_ELSEIF) reduce 232) ((T_EMPTY) reduce 232) ((T_ENDDECLARE) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 19525 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDDECLARE) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 19528 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_ELSE) reduce 233) ((T_ELSEIF) reduce 233) ((T_EMPTY) reduce 233) ((T_ENDDECLARE) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 19533 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDDECLARE) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 19536 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDDECLARE) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 19537 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_ELSE) reduce 174) ((T_ELSEIF) reduce 174) ((T_EMPTY) reduce 174) ((T_ENDDECLARE) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 19540 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDDECLARE) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 19543 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDDECLARE) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 19550 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDDECLARE) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 19555 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDDECLARE) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 19559 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_ELSE) reduce 214) ((T_ELSEIF) reduce 214) ((T_EMPTY) reduce 214) ((T_ENDDECLARE) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 19561 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_ELSE) reduce 215) ((T_ELSEIF) reduce 215) ((T_EMPTY) reduce 215) ((T_ENDDECLARE) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 19565 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDDECLARE) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 19568 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDDECLARE) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 19569 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDDECLARE) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 19571 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDDECLARE) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 19573 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDDECLARE) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 19575 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDDECLARE) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 19577 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDDECLARE) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 19587 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDDECLARE) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 19588 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_ELSE) reduce 210) ((T_ELSEIF) reduce 210) ((T_EMPTY) reduce 210) ((T_ENDDECLARE) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 19591 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_ELSE) reduce 211) ((T_ELSEIF) reduce 211) ((T_EMPTY) reduce 211) ((T_ENDDECLARE) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 19592 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDDECLARE) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 19601 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDDECLARE) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 19602 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_ELSE) reduce 208) ((T_ELSEIF) reduce 208) ((T_EMPTY) reduce 208) ((T_ENDDECLARE) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 19605 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_ELSE) reduce 209) ((T_ELSEIF) reduce 209) ((T_EMPTY) reduce 209) ((T_ENDDECLARE) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 19607 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDDECLARE) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 19613 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDDECLARE) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 19618 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDDECLARE) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 19619 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_ELSE) reduce 212) ((T_ELSEIF) reduce 212) ((T_EMPTY) reduce 212) ((T_ENDDECLARE) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 19622 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_ELSE) reduce 213) ((T_ELSEIF) reduce 213) ((T_EMPTY) reduce 213) ((T_ENDDECLARE) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 19624 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDDECLARE) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 19626 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDDECLARE) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 19627 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_EMPTY) reduce 237) ((T_ENDDECLARE) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 19638 ((("!") reduce 444) (("\"") reduce 444) (("$") reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) ((";") reduce 444) (("@") reduce 444) ((T_ABSTRACT) reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_BREAK) reduce 444) ((T_CLASS) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_CONTINUE) reduce 444) ((T_DEC) reduce 444) ((T_DECLARE) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DO) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_ECHO) reduce 444) ((T_EMPTY) reduce 444) ((T_ENDDECLARE) reduce 444) ((T_ENUM) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FINAL) reduce 444) ((T_FN) reduce 444) ((T_FOR) reduce 444) ((T_FOREACH) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_GLOBAL) reduce 444) ((T_GOTO) reduce 444) ((T_HALT_COMPILER) reduce 444) ((T_IF) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INLINE_HTML) reduce 444) ((T_INTERFACE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_RETURN) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_SWITCH) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT) reduce 444) ((T_TRAIT_C) reduce 444) ((T_TRY) reduce 444) ((T_UNSET) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_VARIABLE) reduce 444) ((T_WHILE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("{") reduce 444) (("~") reduce 444)) 19639 ((("!") reduce 179) (("\"") reduce 179) (("$") reduce 179) (("(") reduce 179) (("+") reduce 179) (("-") reduce 179) ((";") reduce 179) (("@") reduce 179) ((T_ABSTRACT) reduce 179) ((T_ARRAY) reduce 179) ((T_ARRAY_CAST) reduce 179) ((T_ATTRIBUTE) reduce 179) ((T_BOOL_CAST) reduce 179) ((T_BREAK) reduce 179) ((T_CLASS) reduce 179) ((T_CLASS_C) reduce 179) ((T_CLONE) reduce 179) ((T_CONSTANT_ENCAPSED_STRING) reduce 179) ((T_CONTINUE) reduce 179) ((T_DEC) reduce 179) ((T_DECLARE) reduce 179) ((T_DIR) reduce 179) ((T_DNUMBER) reduce 179) ((T_DO) reduce 179) ((T_DOUBLE_CAST) reduce 179) ((T_ECHO) reduce 179) ((T_EMPTY) reduce 179) ((T_ENDDECLARE) reduce 179) ((T_ENUM) reduce 179) ((T_EVAL) reduce 179) ((T_EXIT) reduce 179) ((T_FILE) reduce 179) ((T_FINAL) reduce 179) ((T_FN) reduce 179) ((T_FOR) reduce 179) ((T_FOREACH) reduce 179) ((T_FUNCTION) reduce 179) ((T_FUNC_C) reduce 179) ((T_GLOBAL) reduce 179) ((T_GOTO) reduce 179) ((T_HALT_COMPILER) reduce 179) ((T_IF) reduce 179) ((T_INC) reduce 179) ((T_INCLUDE) reduce 179) ((T_INCLUDE_ONCE) reduce 179) ((T_INLINE_HTML) reduce 179) ((T_INTERFACE) reduce 179) ((T_INT_CAST) reduce 179) ((T_ISSET) reduce 179) ((T_LINE) reduce 179) ((T_LIST) reduce 179) ((T_LNUMBER) reduce 179) ((T_MATCH) reduce 179) ((T_METHOD_C) reduce 179) ((T_NAME_FULLY_QUALIFIED) reduce 179) ((T_NAME_QUALIFIED) reduce 179) ((T_NAME_RELATIVE) reduce 179) ((T_NEW) reduce 179) ((T_NS_C) reduce 179) ((T_OBJECT_CAST) reduce 179) ((T_PRINT) reduce 179) ((T_REQUIRE) reduce 179) ((T_REQUIRE_ONCE) reduce 179) ((T_RETURN) reduce 179) ((T_START_HEREDOC) reduce 179) ((T_STATIC) reduce 179) ((T_STRING) reduce 179) ((T_STRING_CAST) reduce 179) ((T_SWITCH) reduce 179) ((T_THROW) reduce 179) ((T_TRAIT) reduce 179) ((T_TRAIT_C) reduce 179) ((T_TRY) reduce 179) ((T_UNSET) reduce 179) ((T_UNSET_CAST) reduce 179) ((T_VARIABLE) reduce 179) ((T_WHILE) reduce 179) ((T_YIELD) reduce 179) ((T_YIELD_FROM) reduce 179) (("[") reduce 179) (("`") reduce 179) (("{") reduce 179) (("~") reduce 179)) 19640 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDDECLARE) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 19648 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDDECLARE) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 19649 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDDECLARE) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 19652 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDDECLARE) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 19656 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDDECLARE) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 19661 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_EMPTY) reduce 232) ((T_ENDDECLARE) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 19662 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDDECLARE) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 19665 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_EMPTY) reduce 233) ((T_ENDDECLARE) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 19670 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDDECLARE) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 19673 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDDECLARE) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 19674 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_EMPTY) reduce 174) ((T_ENDDECLARE) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 19677 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDDECLARE) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 19680 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDDECLARE) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 19687 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDDECLARE) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 19692 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDDECLARE) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 19697 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDDECLARE) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 19701 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_EMPTY) reduce 214) ((T_ENDDECLARE) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 19703 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_EMPTY) reduce 215) ((T_ENDDECLARE) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 19707 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDDECLARE) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 19710 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDDECLARE) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 19711 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDDECLARE) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 19713 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDDECLARE) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 19715 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDDECLARE) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 19716 (((T_EXTENDS) shift) (("{") reduce 200)) 19722 ((("!") reduce 191) (("\"") reduce 191) (("$") reduce 191) (("(") reduce 191) (("+") reduce 191) (("-") reduce 191) ((";") reduce 191) (("@") reduce 191) ((T_ABSTRACT) reduce 191) ((T_ARRAY) reduce 191) ((T_ARRAY_CAST) reduce 191) ((T_ATTRIBUTE) reduce 191) ((T_BOOL_CAST) reduce 191) ((T_BREAK) reduce 191) ((T_CLASS) reduce 191) ((T_CLASS_C) reduce 191) ((T_CLONE) reduce 191) ((T_CONSTANT_ENCAPSED_STRING) reduce 191) ((T_CONTINUE) reduce 191) ((T_DEC) reduce 191) ((T_DECLARE) reduce 191) ((T_DIR) reduce 191) ((T_DNUMBER) reduce 191) ((T_DO) reduce 191) ((T_DOUBLE_CAST) reduce 191) ((T_ECHO) reduce 191) ((T_EMPTY) reduce 191) ((T_ENDDECLARE) reduce 191) ((T_ENUM) reduce 191) ((T_EVAL) reduce 191) ((T_EXIT) reduce 191) ((T_FILE) reduce 191) ((T_FINAL) reduce 191) ((T_FN) reduce 191) ((T_FOR) reduce 191) ((T_FOREACH) reduce 191) ((T_FUNCTION) reduce 191) ((T_FUNC_C) reduce 191) ((T_GLOBAL) reduce 191) ((T_GOTO) reduce 191) ((T_HALT_COMPILER) reduce 191) ((T_IF) reduce 191) ((T_INC) reduce 191) ((T_INCLUDE) reduce 191) ((T_INCLUDE_ONCE) reduce 191) ((T_INLINE_HTML) reduce 191) ((T_INTERFACE) reduce 191) ((T_INT_CAST) reduce 191) ((T_ISSET) reduce 191) ((T_LINE) reduce 191) ((T_LIST) reduce 191) ((T_LNUMBER) reduce 191) ((T_MATCH) reduce 191) ((T_METHOD_C) reduce 191) ((T_NAME_FULLY_QUALIFIED) reduce 191) ((T_NAME_QUALIFIED) reduce 191) ((T_NAME_RELATIVE) reduce 191) ((T_NEW) reduce 191) ((T_NS_C) reduce 191) ((T_OBJECT_CAST) reduce 191) ((T_PRINT) reduce 191) ((T_REQUIRE) reduce 191) ((T_REQUIRE_ONCE) reduce 191) ((T_RETURN) reduce 191) ((T_START_HEREDOC) reduce 191) ((T_STATIC) reduce 191) ((T_STRING) reduce 191) ((T_STRING_CAST) reduce 191) ((T_SWITCH) reduce 191) ((T_THROW) reduce 191) ((T_TRAIT) reduce 191) ((T_TRAIT_C) reduce 191) ((T_TRY) reduce 191) ((T_UNSET) reduce 191) ((T_UNSET_CAST) reduce 191) ((T_VARIABLE) reduce 191) ((T_WHILE) reduce 191) ((T_YIELD) reduce 191) ((T_YIELD_FROM) reduce 191) (("[") reduce 191) (("`") reduce 191) (("{") reduce 191) (("~") reduce 191)) 19723 (((",") shift) (("{") reduce 201)) 19728 ((("!") reduce 234) (("\"") reduce 234) (("$") reduce 234) (("(") reduce 234) (("+") reduce 234) (("-") reduce 234) ((";") reduce 234) (("@") reduce 234) ((T_ABSTRACT) reduce 234) ((T_ARRAY) reduce 234) ((T_ARRAY_CAST) reduce 234) ((T_ATTRIBUTE) reduce 234) ((T_BOOL_CAST) reduce 234) ((T_BREAK) reduce 234) ((T_CLASS) reduce 234) ((T_CLASS_C) reduce 234) ((T_CLONE) reduce 234) ((T_CONSTANT_ENCAPSED_STRING) reduce 234) ((T_CONTINUE) reduce 234) ((T_DEC) reduce 234) ((T_DECLARE) reduce 234) ((T_DIR) reduce 234) ((T_DNUMBER) reduce 234) ((T_DO) reduce 234) ((T_DOUBLE_CAST) reduce 234) ((T_ECHO) reduce 234) ((T_ELSE) reduce 234) ((T_ELSEIF) reduce 234) ((T_EMPTY) reduce 234) ((T_ENDDECLARE) reduce 234) ((T_ENUM) reduce 234) ((T_EVAL) reduce 234) ((T_EXIT) reduce 234) ((T_FILE) reduce 234) ((T_FINAL) reduce 234) ((T_FN) reduce 234) ((T_FOR) reduce 234) ((T_FOREACH) reduce 234) ((T_FUNCTION) reduce 234) ((T_FUNC_C) reduce 234) ((T_GLOBAL) reduce 234) ((T_GOTO) reduce 234) ((T_HALT_COMPILER) reduce 234) ((T_IF) reduce 234) ((T_INC) reduce 234) ((T_INCLUDE) reduce 234) ((T_INCLUDE_ONCE) reduce 234) ((T_INLINE_HTML) reduce 234) ((T_INTERFACE) reduce 234) ((T_INT_CAST) reduce 234) ((T_ISSET) reduce 234) ((T_LINE) reduce 234) ((T_LIST) reduce 234) ((T_LNUMBER) reduce 234) ((T_MATCH) reduce 234) ((T_METHOD_C) reduce 234) ((T_NAME_FULLY_QUALIFIED) reduce 234) ((T_NAME_QUALIFIED) reduce 234) ((T_NAME_RELATIVE) reduce 234) ((T_NEW) reduce 234) ((T_NS_C) reduce 234) ((T_OBJECT_CAST) reduce 234) ((T_PRINT) reduce 234) ((T_REQUIRE) reduce 234) ((T_REQUIRE_ONCE) reduce 234) ((T_RETURN) reduce 234) ((T_START_HEREDOC) reduce 234) ((T_STATIC) reduce 234) ((T_STRING) reduce 234) ((T_STRING_CAST) reduce 234) ((T_SWITCH) reduce 234) ((T_THROW) reduce 234) ((T_TRAIT) reduce 234) ((T_TRAIT_C) reduce 234) ((T_TRY) reduce 234) ((T_UNSET) reduce 234) ((T_UNSET_CAST) reduce 234) ((T_VARIABLE) reduce 234) ((T_WHILE) reduce 234) ((T_YIELD) reduce 234) ((T_YIELD_FROM) reduce 234) (("[") reduce 234) (("`") reduce 234) (("{") reduce 234) (("~") reduce 234)) 19729 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_ELSE) reduce 238) ((T_ELSEIF) reduce 238) ((T_EMPTY) shift) ((T_ENDIF) reduce 238) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 19732 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDDECLARE) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 19734 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDDECLARE) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 19736 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDDECLARE) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 19746 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDDECLARE) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 19747 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_EMPTY) reduce 210) ((T_ENDDECLARE) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 19750 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_EMPTY) reduce 211) ((T_ENDDECLARE) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 19751 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDDECLARE) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 19760 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDDECLARE) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 19761 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_EMPTY) reduce 208) ((T_ENDDECLARE) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 19764 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_EMPTY) reduce 209) ((T_ENDDECLARE) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 19765 (((":") shift) ((T_IMPLEMENTS) reduce 193) (("{") reduce 193)) 19772 ((("!") reduce 192) (("\"") reduce 192) (("$") reduce 192) (("(") reduce 192) (("+") reduce 192) (("-") reduce 192) ((";") reduce 192) (("@") reduce 192) ((T_ABSTRACT) reduce 192) ((T_ARRAY) reduce 192) ((T_ARRAY_CAST) reduce 192) ((T_ATTRIBUTE) reduce 192) ((T_BOOL_CAST) reduce 192) ((T_BREAK) reduce 192) ((T_CLASS) reduce 192) ((T_CLASS_C) reduce 192) ((T_CLONE) reduce 192) ((T_CONSTANT_ENCAPSED_STRING) reduce 192) ((T_CONTINUE) reduce 192) ((T_DEC) reduce 192) ((T_DECLARE) reduce 192) ((T_DIR) reduce 192) ((T_DNUMBER) reduce 192) ((T_DO) reduce 192) ((T_DOUBLE_CAST) reduce 192) ((T_ECHO) reduce 192) ((T_EMPTY) reduce 192) ((T_ENDDECLARE) reduce 192) ((T_ENUM) reduce 192) ((T_EVAL) reduce 192) ((T_EXIT) reduce 192) ((T_FILE) reduce 192) ((T_FINAL) reduce 192) ((T_FN) reduce 192) ((T_FOR) reduce 192) ((T_FOREACH) reduce 192) ((T_FUNCTION) reduce 192) ((T_FUNC_C) reduce 192) ((T_GLOBAL) reduce 192) ((T_GOTO) reduce 192) ((T_HALT_COMPILER) reduce 192) ((T_IF) reduce 192) ((T_INC) reduce 192) ((T_INCLUDE) reduce 192) ((T_INCLUDE_ONCE) reduce 192) ((T_INLINE_HTML) reduce 192) ((T_INTERFACE) reduce 192) ((T_INT_CAST) reduce 192) ((T_ISSET) reduce 192) ((T_LINE) reduce 192) ((T_LIST) reduce 192) ((T_LNUMBER) reduce 192) ((T_MATCH) reduce 192) ((T_METHOD_C) reduce 192) ((T_NAME_FULLY_QUALIFIED) reduce 192) ((T_NAME_QUALIFIED) reduce 192) ((T_NAME_RELATIVE) reduce 192) ((T_NEW) reduce 192) ((T_NS_C) reduce 192) ((T_OBJECT_CAST) reduce 192) ((T_PRINT) reduce 192) ((T_REQUIRE) reduce 192) ((T_REQUIRE_ONCE) reduce 192) ((T_RETURN) reduce 192) ((T_START_HEREDOC) reduce 192) ((T_STATIC) reduce 192) ((T_STRING) reduce 192) ((T_STRING_CAST) reduce 192) ((T_SWITCH) reduce 192) ((T_THROW) reduce 192) ((T_TRAIT) reduce 192) ((T_TRAIT_C) reduce 192) ((T_TRY) reduce 192) ((T_UNSET) reduce 192) ((T_UNSET_CAST) reduce 192) ((T_VARIABLE) reduce 192) ((T_WHILE) reduce 192) ((T_YIELD) reduce 192) ((T_YIELD_FROM) reduce 192) (("[") reduce 192) (("`") reduce 192) (("{") reduce 192) (("~") reduce 192)) 19774 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_IMPLEMENTS) reduce 272) (("{") reduce 272) (("|") reduce 272)) 19775 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_IMPLEMENTS) reduce 273) (("{") reduce 273) (("|") reduce 273)) 19776 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_IMPLEMENTS) reduce 93) (("{") reduce 93) (("|") reduce 93)) 19777 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_IMPLEMENTS) reduce 92) (("{") reduce 92) (("|") reduce 92)) 19778 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_IMPLEMENTS) reduce 94) (("{") reduce 94) (("|") reduce 94)) 19779 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 263) ((T_IMPLEMENTS) reduce 263) (("{") reduce 263) (("|") reduce 263)) 19780 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_IMPLEMENTS) reduce 91) (("{") reduce 91) (("|") reduce 91)) 19781 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_IMPLEMENTS) reduce 261) (("{") reduce 261)) 19782 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_IMPLEMENTS) reduce 274) (("{") reduce 274) (("|") reduce 274)) 19783 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_IMPLEMENTS) reduce 258) (("{") reduce 258) (("|") shift)) 19784 (((T_IMPLEMENTS) reduce 194) (("{") reduce 194)) 19785 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 262) ((T_IMPLEMENTS) reduce 262) (("{") reduce 262) (("|") reduce 262)) 19786 (((T_IMPLEMENTS) reduce 260) (("{") reduce 260) (("|") shift)) 19788 (((T_IMPLEMENTS) reduce 272) (("{") reduce 272) (("|") reduce 272)) 19789 (((T_IMPLEMENTS) reduce 273) (("{") reduce 273) (("|") reduce 273)) 19790 (((T_IMPLEMENTS) reduce 93) (("{") reduce 93) (("|") reduce 93)) 19791 (((T_IMPLEMENTS) reduce 92) (("{") reduce 92) (("|") reduce 92)) 19792 (((T_IMPLEMENTS) reduce 94) (("{") reduce 94) (("|") reduce 94)) 19793 (((T_IMPLEMENTS) reduce 263) (("{") reduce 263) (("|") reduce 263)) 19794 (((T_IMPLEMENTS) reduce 91) (("{") reduce 91) (("|") reduce 91)) 19795 (((T_IMPLEMENTS) reduce 274) (("{") reduce 274) (("|") reduce 274)) 19796 (((T_IMPLEMENTS) reduce 265) (("{") reduce 265) (("|") reduce 265)) 19797 (((T_IMPLEMENTS) reduce 262) (("{") reduce 262) (("|") reduce 262)) 19800 (((T_IMPLEMENTS) reduce 264) (("{") reduce 264) (("|") reduce 264)) 19801 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_IMPLEMENTS) reduce 272) (("{") reduce 272)) 19802 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_IMPLEMENTS) reduce 273) (("{") reduce 273)) 19803 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_IMPLEMENTS) reduce 93) (("{") reduce 93)) 19804 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_IMPLEMENTS) reduce 92) (("{") reduce 92)) 19805 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_IMPLEMENTS) reduce 94) (("{") reduce 94)) 19806 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 263) ((T_IMPLEMENTS) reduce 263) (("{") reduce 263)) 19807 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_IMPLEMENTS) reduce 91) (("{") reduce 91)) 19808 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_IMPLEMENTS) reduce 274) (("{") reduce 274)) 19809 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) ((T_IMPLEMENTS) reduce 266) (("{") reduce 266)) 19810 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 262) ((T_IMPLEMENTS) reduce 262) (("{") reduce 262)) 19812 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) ((T_IMPLEMENTS) reduce 267) (("{") reduce 267)) 19813 (((T_IMPLEMENTS) reduce 272) (("{") reduce 272)) 19814 (((T_IMPLEMENTS) reduce 273) (("{") reduce 273)) 19815 (((T_IMPLEMENTS) reduce 263) (("{") reduce 263)) 19816 (((T_IMPLEMENTS) reduce 274) (("{") reduce 274)) 19817 (((T_IMPLEMENTS) reduce 259) (("{") reduce 259)) 19818 (((T_IMPLEMENTS) reduce 262) (("{") reduce 262)) 19819 (((T_ELSE) reduce 213) ((T_ELSEIF) reduce 213) ((T_WHILE) reduce 213)) 19821 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDDECLARE) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 19827 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDDECLARE) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 19832 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDDECLARE) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 19833 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_EMPTY) reduce 212) ((T_ENDDECLARE) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 19836 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_EMPTY) reduce 213) ((T_ENDDECLARE) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 19838 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDDECLARE) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 19845 ((("!") reduce 185) (("\"") reduce 185) (("$") reduce 185) (("(") reduce 185) (("+") reduce 185) (("-") reduce 185) ((";") reduce 185) (("@") reduce 185) ((T_ABSTRACT) reduce 185) ((T_ARRAY) reduce 185) ((T_ARRAY_CAST) reduce 185) ((T_ATTRIBUTE) reduce 185) ((T_BOOL_CAST) reduce 185) ((T_BREAK) reduce 185) ((T_CLASS) reduce 185) ((T_CLASS_C) reduce 185) ((T_CLONE) reduce 185) ((T_CONSTANT_ENCAPSED_STRING) reduce 185) ((T_CONTINUE) reduce 185) ((T_DEC) reduce 185) ((T_DECLARE) reduce 185) ((T_DIR) reduce 185) ((T_DNUMBER) reduce 185) ((T_DO) reduce 185) ((T_DOUBLE_CAST) reduce 185) ((T_ECHO) reduce 185) ((T_EMPTY) reduce 185) ((T_ENDDECLARE) reduce 185) ((T_ENUM) reduce 185) ((T_EVAL) reduce 185) ((T_EXIT) reduce 185) ((T_FILE) reduce 185) ((T_FINAL) reduce 185) ((T_FN) reduce 185) ((T_FOR) reduce 185) ((T_FOREACH) reduce 185) ((T_FUNCTION) reduce 185) ((T_FUNC_C) reduce 185) ((T_GLOBAL) reduce 185) ((T_GOTO) reduce 185) ((T_HALT_COMPILER) reduce 185) ((T_IF) reduce 185) ((T_INC) reduce 185) ((T_INCLUDE) reduce 185) ((T_INCLUDE_ONCE) reduce 185) ((T_INLINE_HTML) reduce 185) ((T_INTERFACE) reduce 185) ((T_INT_CAST) reduce 185) ((T_ISSET) reduce 185) ((T_LINE) reduce 185) ((T_LIST) reduce 185) ((T_LNUMBER) reduce 185) ((T_MATCH) reduce 185) ((T_METHOD_C) reduce 185) ((T_NAME_FULLY_QUALIFIED) reduce 185) ((T_NAME_QUALIFIED) reduce 185) ((T_NAME_RELATIVE) reduce 185) ((T_NEW) reduce 185) ((T_NS_C) reduce 185) ((T_OBJECT_CAST) reduce 185) ((T_PRINT) reduce 185) ((T_REQUIRE) reduce 185) ((T_REQUIRE_ONCE) reduce 185) ((T_RETURN) reduce 185) ((T_START_HEREDOC) reduce 185) ((T_STATIC) reduce 185) ((T_STRING) reduce 185) ((T_STRING_CAST) reduce 185) ((T_SWITCH) reduce 185) ((T_THROW) reduce 185) ((T_TRAIT) reduce 185) ((T_TRAIT_C) reduce 185) ((T_TRY) reduce 185) ((T_UNSET) reduce 185) ((T_UNSET_CAST) reduce 185) ((T_VARIABLE) reduce 185) ((T_WHILE) reduce 185) ((T_YIELD) reduce 185) ((T_YIELD_FROM) reduce 185) (("[") reduce 185) (("`") reduce 185) (("{") reduce 185) (("~") reduce 185)) 19847 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDDECLARE) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 19849 ((("%") shift) ((")") reduce 443) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19850 (((")") reduce 344) ((",") reduce 344)) 19852 (((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_WHILE) reduce 153)) 19854 (((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_WHILE) reduce 152)) 19855 (((T_WHILE) reduce 237)) 19856 (((T_WHILE) reduce 159)) 19859 (((T_WHILE) reduce 240)) 19863 (((T_WHILE) reduce 241)) 19868 (((T_WHILE) reduce 232)) 19869 (((T_WHILE) reduce 148)) 19872 (((T_WHILE) reduce 233)) 19877 (((T_WHILE) reduce 160)) 19880 (((T_CATCH) reduce 168) ((T_FINALLY) reduce 168) ((T_WHILE) reduce 168)) 19881 (((T_CATCH) shift) ((T_FINALLY) shift) ((T_WHILE) reduce 174)) 19884 (((T_WHILE) reduce 165)) 19887 (((T_WHILE) reduce 175)) 19894 (((T_CATCH) reduce 169) ((T_FINALLY) reduce 169) ((T_WHILE) reduce 169)) 19899 (((T_WHILE) reduce 151)) 19903 (((T_WHILE) reduce 214)) 19905 (((T_WHILE) reduce 215)) 19909 (((T_WHILE) reduce 216)) 19912 (((T_WHILE) reduce 217)) 19913 (((T_WHILE) reduce 167)) 19915 (((T_WHILE) reduce 156)) 19917 (((T_WHILE) reduce 154)) 19921 (((T_ELSE) reduce 234) ((T_ELSEIF) reduce 234) ((T_WHILE) reduce 234)) 19923 (((T_WHILE) reduce 166)) 19925 (((T_WHILE) reduce 155)) 19935 (((T_WHILE) reduce 162)) 19936 (((T_WHILE) reduce 210)) 19939 (((T_WHILE) reduce 211)) 19940 (((T_WHILE) reduce 161)) 19949 (((T_WHILE) reduce 150)) 19950 (((T_WHILE) reduce 208)) 19953 (((T_WHILE) reduce 209)) 19955 (((T_WHILE) reduce 157)) 19961 (((T_WHILE) reduce 149)) 19966 (((T_WHILE) reduce 163)) 19967 (((T_WHILE) reduce 212)) 19970 (((T_WHILE) reduce 213)) 19972 (((T_WHILE) reduce 153)) 19974 (((T_WHILE) reduce 152)) 19979 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDFOR) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 19980 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_ELSE) reduce 212) ((T_ELSEIF) reduce 212) ((T_EMPTY) reduce 212) ((T_ENDFOR) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 19983 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_ELSE) reduce 213) ((T_ELSEIF) reduce 213) ((T_EMPTY) reduce 213) ((T_ENDFOR) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 19985 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDFOR) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 19987 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDFOR) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 19988 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_EMPTY) reduce 237) ((T_ENDFOR) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 19999 ((("!") reduce 444) (("\"") reduce 444) (("$") reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) ((";") reduce 444) (("@") reduce 444) ((T_ABSTRACT) reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_BREAK) reduce 444) ((T_CLASS) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_CONTINUE) reduce 444) ((T_DEC) reduce 444) ((T_DECLARE) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DO) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_ECHO) reduce 444) ((T_EMPTY) reduce 444) ((T_ENDFOR) reduce 444) ((T_ENUM) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FINAL) reduce 444) ((T_FN) reduce 444) ((T_FOR) reduce 444) ((T_FOREACH) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_GLOBAL) reduce 444) ((T_GOTO) reduce 444) ((T_HALT_COMPILER) reduce 444) ((T_IF) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INLINE_HTML) reduce 444) ((T_INTERFACE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_RETURN) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_SWITCH) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT) reduce 444) ((T_TRAIT_C) reduce 444) ((T_TRY) reduce 444) ((T_UNSET) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_VARIABLE) reduce 444) ((T_WHILE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("{") reduce 444) (("~") reduce 444)) 20000 ((("!") reduce 179) (("\"") reduce 179) (("$") reduce 179) (("(") reduce 179) (("+") reduce 179) (("-") reduce 179) ((";") reduce 179) (("@") reduce 179) ((T_ABSTRACT) reduce 179) ((T_ARRAY) reduce 179) ((T_ARRAY_CAST) reduce 179) ((T_ATTRIBUTE) reduce 179) ((T_BOOL_CAST) reduce 179) ((T_BREAK) reduce 179) ((T_CLASS) reduce 179) ((T_CLASS_C) reduce 179) ((T_CLONE) reduce 179) ((T_CONSTANT_ENCAPSED_STRING) reduce 179) ((T_CONTINUE) reduce 179) ((T_DEC) reduce 179) ((T_DECLARE) reduce 179) ((T_DIR) reduce 179) ((T_DNUMBER) reduce 179) ((T_DO) reduce 179) ((T_DOUBLE_CAST) reduce 179) ((T_ECHO) reduce 179) ((T_EMPTY) reduce 179) ((T_ENDFOR) reduce 179) ((T_ENUM) reduce 179) ((T_EVAL) reduce 179) ((T_EXIT) reduce 179) ((T_FILE) reduce 179) ((T_FINAL) reduce 179) ((T_FN) reduce 179) ((T_FOR) reduce 179) ((T_FOREACH) reduce 179) ((T_FUNCTION) reduce 179) ((T_FUNC_C) reduce 179) ((T_GLOBAL) reduce 179) ((T_GOTO) reduce 179) ((T_HALT_COMPILER) reduce 179) ((T_IF) reduce 179) ((T_INC) reduce 179) ((T_INCLUDE) reduce 179) ((T_INCLUDE_ONCE) reduce 179) ((T_INLINE_HTML) reduce 179) ((T_INTERFACE) reduce 179) ((T_INT_CAST) reduce 179) ((T_ISSET) reduce 179) ((T_LINE) reduce 179) ((T_LIST) reduce 179) ((T_LNUMBER) reduce 179) ((T_MATCH) reduce 179) ((T_METHOD_C) reduce 179) ((T_NAME_FULLY_QUALIFIED) reduce 179) ((T_NAME_QUALIFIED) reduce 179) ((T_NAME_RELATIVE) reduce 179) ((T_NEW) reduce 179) ((T_NS_C) reduce 179) ((T_OBJECT_CAST) reduce 179) ((T_PRINT) reduce 179) ((T_REQUIRE) reduce 179) ((T_REQUIRE_ONCE) reduce 179) ((T_RETURN) reduce 179) ((T_START_HEREDOC) reduce 179) ((T_STATIC) reduce 179) ((T_STRING) reduce 179) ((T_STRING_CAST) reduce 179) ((T_SWITCH) reduce 179) ((T_THROW) reduce 179) ((T_TRAIT) reduce 179) ((T_TRAIT_C) reduce 179) ((T_TRY) reduce 179) ((T_UNSET) reduce 179) ((T_UNSET_CAST) reduce 179) ((T_VARIABLE) reduce 179) ((T_WHILE) reduce 179) ((T_YIELD) reduce 179) ((T_YIELD_FROM) reduce 179) (("[") reduce 179) (("`") reduce 179) (("{") reduce 179) (("~") reduce 179)) 20001 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDFOR) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 20009 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDFOR) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 20010 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDFOR) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 20013 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDFOR) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 20017 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDFOR) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 20022 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_EMPTY) reduce 232) ((T_ENDFOR) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 20023 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDFOR) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 20026 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_EMPTY) reduce 233) ((T_ENDFOR) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 20031 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDFOR) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 20034 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDFOR) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 20035 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_EMPTY) reduce 174) ((T_ENDFOR) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 20038 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDFOR) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 20041 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDFOR) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 20048 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDFOR) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 20053 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDFOR) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 20058 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDFOR) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 20062 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_EMPTY) reduce 214) ((T_ENDFOR) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 20064 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_EMPTY) reduce 215) ((T_ENDFOR) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 20068 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDFOR) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 20071 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDFOR) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 20072 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDFOR) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 20074 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDFOR) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 20076 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDFOR) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 20082 ((("!") reduce 191) (("\"") reduce 191) (("$") reduce 191) (("(") reduce 191) (("+") reduce 191) (("-") reduce 191) ((";") reduce 191) (("@") reduce 191) ((T_ABSTRACT) reduce 191) ((T_ARRAY) reduce 191) ((T_ARRAY_CAST) reduce 191) ((T_ATTRIBUTE) reduce 191) ((T_BOOL_CAST) reduce 191) ((T_BREAK) reduce 191) ((T_CLASS) reduce 191) ((T_CLASS_C) reduce 191) ((T_CLONE) reduce 191) ((T_CONSTANT_ENCAPSED_STRING) reduce 191) ((T_CONTINUE) reduce 191) ((T_DEC) reduce 191) ((T_DECLARE) reduce 191) ((T_DIR) reduce 191) ((T_DNUMBER) reduce 191) ((T_DO) reduce 191) ((T_DOUBLE_CAST) reduce 191) ((T_ECHO) reduce 191) ((T_EMPTY) reduce 191) ((T_ENDFOR) reduce 191) ((T_ENUM) reduce 191) ((T_EVAL) reduce 191) ((T_EXIT) reduce 191) ((T_FILE) reduce 191) ((T_FINAL) reduce 191) ((T_FN) reduce 191) ((T_FOR) reduce 191) ((T_FOREACH) reduce 191) ((T_FUNCTION) reduce 191) ((T_FUNC_C) reduce 191) ((T_GLOBAL) reduce 191) ((T_GOTO) reduce 191) ((T_HALT_COMPILER) reduce 191) ((T_IF) reduce 191) ((T_INC) reduce 191) ((T_INCLUDE) reduce 191) ((T_INCLUDE_ONCE) reduce 191) ((T_INLINE_HTML) reduce 191) ((T_INTERFACE) reduce 191) ((T_INT_CAST) reduce 191) ((T_ISSET) reduce 191) ((T_LINE) reduce 191) ((T_LIST) reduce 191) ((T_LNUMBER) reduce 191) ((T_MATCH) reduce 191) ((T_METHOD_C) reduce 191) ((T_NAME_FULLY_QUALIFIED) reduce 191) ((T_NAME_QUALIFIED) reduce 191) ((T_NAME_RELATIVE) reduce 191) ((T_NEW) reduce 191) ((T_NS_C) reduce 191) ((T_OBJECT_CAST) reduce 191) ((T_PRINT) reduce 191) ((T_REQUIRE) reduce 191) ((T_REQUIRE_ONCE) reduce 191) ((T_RETURN) reduce 191) ((T_START_HEREDOC) reduce 191) ((T_STATIC) reduce 191) ((T_STRING) reduce 191) ((T_STRING_CAST) reduce 191) ((T_SWITCH) reduce 191) ((T_THROW) reduce 191) ((T_TRAIT) reduce 191) ((T_TRAIT_C) reduce 191) ((T_TRY) reduce 191) ((T_UNSET) reduce 191) ((T_UNSET_CAST) reduce 191) ((T_VARIABLE) reduce 191) ((T_WHILE) reduce 191) ((T_YIELD) reduce 191) ((T_YIELD_FROM) reduce 191) (("[") reduce 191) (("`") reduce 191) (("{") reduce 191) (("~") reduce 191)) 20086 ((("!") reduce 234) (("\"") reduce 234) (("$") reduce 234) (("(") reduce 234) (("+") reduce 234) (("-") reduce 234) ((";") reduce 234) (("@") reduce 234) ((T_ABSTRACT) reduce 234) ((T_ARRAY) reduce 234) ((T_ARRAY_CAST) reduce 234) ((T_ATTRIBUTE) reduce 234) ((T_BOOL_CAST) reduce 234) ((T_BREAK) reduce 234) ((T_CLASS) reduce 234) ((T_CLASS_C) reduce 234) ((T_CLONE) reduce 234) ((T_CONSTANT_ENCAPSED_STRING) reduce 234) ((T_CONTINUE) reduce 234) ((T_DEC) reduce 234) ((T_DECLARE) reduce 234) ((T_DIR) reduce 234) ((T_DNUMBER) reduce 234) ((T_DO) reduce 234) ((T_DOUBLE_CAST) reduce 234) ((T_ECHO) reduce 234) ((T_ELSE) reduce 234) ((T_ELSEIF) reduce 234) ((T_EMPTY) reduce 234) ((T_ENDFOR) reduce 234) ((T_ENUM) reduce 234) ((T_EVAL) reduce 234) ((T_EXIT) reduce 234) ((T_FILE) reduce 234) ((T_FINAL) reduce 234) ((T_FN) reduce 234) ((T_FOR) reduce 234) ((T_FOREACH) reduce 234) ((T_FUNCTION) reduce 234) ((T_FUNC_C) reduce 234) ((T_GLOBAL) reduce 234) ((T_GOTO) reduce 234) ((T_HALT_COMPILER) reduce 234) ((T_IF) reduce 234) ((T_INC) reduce 234) ((T_INCLUDE) reduce 234) ((T_INCLUDE_ONCE) reduce 234) ((T_INLINE_HTML) reduce 234) ((T_INTERFACE) reduce 234) ((T_INT_CAST) reduce 234) ((T_ISSET) reduce 234) ((T_LINE) reduce 234) ((T_LIST) reduce 234) ((T_LNUMBER) reduce 234) ((T_MATCH) reduce 234) ((T_METHOD_C) reduce 234) ((T_NAME_FULLY_QUALIFIED) reduce 234) ((T_NAME_QUALIFIED) reduce 234) ((T_NAME_RELATIVE) reduce 234) ((T_NEW) reduce 234) ((T_NS_C) reduce 234) ((T_OBJECT_CAST) reduce 234) ((T_PRINT) reduce 234) ((T_REQUIRE) reduce 234) ((T_REQUIRE_ONCE) reduce 234) ((T_RETURN) reduce 234) ((T_START_HEREDOC) reduce 234) ((T_STATIC) reduce 234) ((T_STRING) reduce 234) ((T_STRING_CAST) reduce 234) ((T_SWITCH) reduce 234) ((T_THROW) reduce 234) ((T_TRAIT) reduce 234) ((T_TRAIT_C) reduce 234) ((T_TRY) reduce 234) ((T_UNSET) reduce 234) ((T_UNSET_CAST) reduce 234) ((T_VARIABLE) reduce 234) ((T_WHILE) reduce 234) ((T_YIELD) reduce 234) ((T_YIELD_FROM) reduce 234) (("[") reduce 234) (("`") reduce 234) (("{") reduce 234) (("~") reduce 234)) 20089 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDFOR) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 20091 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDFOR) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 20093 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDFOR) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 20103 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDFOR) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 20104 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_EMPTY) reduce 210) ((T_ENDFOR) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 20107 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_EMPTY) reduce 211) ((T_ENDFOR) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 20108 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDFOR) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 20117 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDFOR) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 20118 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_EMPTY) reduce 208) ((T_ENDFOR) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 20121 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_EMPTY) reduce 209) ((T_ENDFOR) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 20128 ((("!") reduce 192) (("\"") reduce 192) (("$") reduce 192) (("(") reduce 192) (("+") reduce 192) (("-") reduce 192) ((";") reduce 192) (("@") reduce 192) ((T_ABSTRACT) reduce 192) ((T_ARRAY) reduce 192) ((T_ARRAY_CAST) reduce 192) ((T_ATTRIBUTE) reduce 192) ((T_BOOL_CAST) reduce 192) ((T_BREAK) reduce 192) ((T_CLASS) reduce 192) ((T_CLASS_C) reduce 192) ((T_CLONE) reduce 192) ((T_CONSTANT_ENCAPSED_STRING) reduce 192) ((T_CONTINUE) reduce 192) ((T_DEC) reduce 192) ((T_DECLARE) reduce 192) ((T_DIR) reduce 192) ((T_DNUMBER) reduce 192) ((T_DO) reduce 192) ((T_DOUBLE_CAST) reduce 192) ((T_ECHO) reduce 192) ((T_EMPTY) reduce 192) ((T_ENDFOR) reduce 192) ((T_ENUM) reduce 192) ((T_EVAL) reduce 192) ((T_EXIT) reduce 192) ((T_FILE) reduce 192) ((T_FINAL) reduce 192) ((T_FN) reduce 192) ((T_FOR) reduce 192) ((T_FOREACH) reduce 192) ((T_FUNCTION) reduce 192) ((T_FUNC_C) reduce 192) ((T_GLOBAL) reduce 192) ((T_GOTO) reduce 192) ((T_HALT_COMPILER) reduce 192) ((T_IF) reduce 192) ((T_INC) reduce 192) ((T_INCLUDE) reduce 192) ((T_INCLUDE_ONCE) reduce 192) ((T_INLINE_HTML) reduce 192) ((T_INTERFACE) reduce 192) ((T_INT_CAST) reduce 192) ((T_ISSET) reduce 192) ((T_LINE) reduce 192) ((T_LIST) reduce 192) ((T_LNUMBER) reduce 192) ((T_MATCH) reduce 192) ((T_METHOD_C) reduce 192) ((T_NAME_FULLY_QUALIFIED) reduce 192) ((T_NAME_QUALIFIED) reduce 192) ((T_NAME_RELATIVE) reduce 192) ((T_NEW) reduce 192) ((T_NS_C) reduce 192) ((T_OBJECT_CAST) reduce 192) ((T_PRINT) reduce 192) ((T_REQUIRE) reduce 192) ((T_REQUIRE_ONCE) reduce 192) ((T_RETURN) reduce 192) ((T_START_HEREDOC) reduce 192) ((T_STATIC) reduce 192) ((T_STRING) reduce 192) ((T_STRING_CAST) reduce 192) ((T_SWITCH) reduce 192) ((T_THROW) reduce 192) ((T_TRAIT) reduce 192) ((T_TRAIT_C) reduce 192) ((T_TRY) reduce 192) ((T_UNSET) reduce 192) ((T_UNSET_CAST) reduce 192) ((T_VARIABLE) reduce 192) ((T_WHILE) reduce 192) ((T_YIELD) reduce 192) ((T_YIELD_FROM) reduce 192) (("[") reduce 192) (("`") reduce 192) (("{") reduce 192) (("~") reduce 192)) 20129 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_ELSE) reduce 209) ((T_ELSEIF) reduce 209) ((T_EMPTY) reduce 209) ((T_ENDFOREACH) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 20131 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDFOR) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 20137 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDFOR) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 20142 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDFOR) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 20143 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_EMPTY) reduce 212) ((T_ENDFOR) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 20146 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_EMPTY) reduce 213) ((T_ENDFOR) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 20148 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDFOR) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 20155 ((("!") reduce 185) (("\"") reduce 185) (("$") reduce 185) (("(") reduce 185) (("+") reduce 185) (("-") reduce 185) ((";") reduce 185) (("@") reduce 185) ((T_ABSTRACT) reduce 185) ((T_ARRAY) reduce 185) ((T_ARRAY_CAST) reduce 185) ((T_ATTRIBUTE) reduce 185) ((T_BOOL_CAST) reduce 185) ((T_BREAK) reduce 185) ((T_CLASS) reduce 185) ((T_CLASS_C) reduce 185) ((T_CLONE) reduce 185) ((T_CONSTANT_ENCAPSED_STRING) reduce 185) ((T_CONTINUE) reduce 185) ((T_DEC) reduce 185) ((T_DECLARE) reduce 185) ((T_DIR) reduce 185) ((T_DNUMBER) reduce 185) ((T_DO) reduce 185) ((T_DOUBLE_CAST) reduce 185) ((T_ECHO) reduce 185) ((T_EMPTY) reduce 185) ((T_ENDFOR) reduce 185) ((T_ENUM) reduce 185) ((T_EVAL) reduce 185) ((T_EXIT) reduce 185) ((T_FILE) reduce 185) ((T_FINAL) reduce 185) ((T_FN) reduce 185) ((T_FOR) reduce 185) ((T_FOREACH) reduce 185) ((T_FUNCTION) reduce 185) ((T_FUNC_C) reduce 185) ((T_GLOBAL) reduce 185) ((T_GOTO) reduce 185) ((T_HALT_COMPILER) reduce 185) ((T_IF) reduce 185) ((T_INC) reduce 185) ((T_INCLUDE) reduce 185) ((T_INCLUDE_ONCE) reduce 185) ((T_INLINE_HTML) reduce 185) ((T_INTERFACE) reduce 185) ((T_INT_CAST) reduce 185) ((T_ISSET) reduce 185) ((T_LINE) reduce 185) ((T_LIST) reduce 185) ((T_LNUMBER) reduce 185) ((T_MATCH) reduce 185) ((T_METHOD_C) reduce 185) ((T_NAME_FULLY_QUALIFIED) reduce 185) ((T_NAME_QUALIFIED) reduce 185) ((T_NAME_RELATIVE) reduce 185) ((T_NEW) reduce 185) ((T_NS_C) reduce 185) ((T_OBJECT_CAST) reduce 185) ((T_PRINT) reduce 185) ((T_REQUIRE) reduce 185) ((T_REQUIRE_ONCE) reduce 185) ((T_RETURN) reduce 185) ((T_START_HEREDOC) reduce 185) ((T_STATIC) reduce 185) ((T_STRING) reduce 185) ((T_STRING_CAST) reduce 185) ((T_SWITCH) reduce 185) ((T_THROW) reduce 185) ((T_TRAIT) reduce 185) ((T_TRAIT_C) reduce 185) ((T_TRY) reduce 185) ((T_UNSET) reduce 185) ((T_UNSET_CAST) reduce 185) ((T_VARIABLE) reduce 185) ((T_WHILE) reduce 185) ((T_YIELD) reduce 185) ((T_YIELD_FROM) reduce 185) (("[") reduce 185) (("`") reduce 185) (("{") reduce 185) (("~") reduce 185)) 20157 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDFOR) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 20159 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDFOREACH) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 20165 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDFOREACH) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 20170 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDFOREACH) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 20171 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_ELSE) reduce 212) ((T_ELSEIF) reduce 212) ((T_EMPTY) reduce 212) ((T_ENDFOREACH) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 20174 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_ELSE) reduce 213) ((T_ELSEIF) reduce 213) ((T_EMPTY) reduce 213) ((T_ENDFOREACH) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 20176 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDFOREACH) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 20178 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDFOREACH) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 20179 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_EMPTY) reduce 237) ((T_ENDFOREACH) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 20190 ((("!") reduce 444) (("\"") reduce 444) (("$") reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) ((";") reduce 444) (("@") reduce 444) ((T_ABSTRACT) reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_BREAK) reduce 444) ((T_CLASS) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_CONTINUE) reduce 444) ((T_DEC) reduce 444) ((T_DECLARE) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DO) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_ECHO) reduce 444) ((T_EMPTY) reduce 444) ((T_ENDFOREACH) reduce 444) ((T_ENUM) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FINAL) reduce 444) ((T_FN) reduce 444) ((T_FOR) reduce 444) ((T_FOREACH) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_GLOBAL) reduce 444) ((T_GOTO) reduce 444) ((T_HALT_COMPILER) reduce 444) ((T_IF) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INLINE_HTML) reduce 444) ((T_INTERFACE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_RETURN) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_SWITCH) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT) reduce 444) ((T_TRAIT_C) reduce 444) ((T_TRY) reduce 444) ((T_UNSET) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_VARIABLE) reduce 444) ((T_WHILE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("{") reduce 444) (("~") reduce 444)) 20191 ((("!") reduce 179) (("\"") reduce 179) (("$") reduce 179) (("(") reduce 179) (("+") reduce 179) (("-") reduce 179) ((";") reduce 179) (("@") reduce 179) ((T_ABSTRACT) reduce 179) ((T_ARRAY) reduce 179) ((T_ARRAY_CAST) reduce 179) ((T_ATTRIBUTE) reduce 179) ((T_BOOL_CAST) reduce 179) ((T_BREAK) reduce 179) ((T_CLASS) reduce 179) ((T_CLASS_C) reduce 179) ((T_CLONE) reduce 179) ((T_CONSTANT_ENCAPSED_STRING) reduce 179) ((T_CONTINUE) reduce 179) ((T_DEC) reduce 179) ((T_DECLARE) reduce 179) ((T_DIR) reduce 179) ((T_DNUMBER) reduce 179) ((T_DO) reduce 179) ((T_DOUBLE_CAST) reduce 179) ((T_ECHO) reduce 179) ((T_EMPTY) reduce 179) ((T_ENDFOREACH) reduce 179) ((T_ENUM) reduce 179) ((T_EVAL) reduce 179) ((T_EXIT) reduce 179) ((T_FILE) reduce 179) ((T_FINAL) reduce 179) ((T_FN) reduce 179) ((T_FOR) reduce 179) ((T_FOREACH) reduce 179) ((T_FUNCTION) reduce 179) ((T_FUNC_C) reduce 179) ((T_GLOBAL) reduce 179) ((T_GOTO) reduce 179) ((T_HALT_COMPILER) reduce 179) ((T_IF) reduce 179) ((T_INC) reduce 179) ((T_INCLUDE) reduce 179) ((T_INCLUDE_ONCE) reduce 179) ((T_INLINE_HTML) reduce 179) ((T_INTERFACE) reduce 179) ((T_INT_CAST) reduce 179) ((T_ISSET) reduce 179) ((T_LINE) reduce 179) ((T_LIST) reduce 179) ((T_LNUMBER) reduce 179) ((T_MATCH) reduce 179) ((T_METHOD_C) reduce 179) ((T_NAME_FULLY_QUALIFIED) reduce 179) ((T_NAME_QUALIFIED) reduce 179) ((T_NAME_RELATIVE) reduce 179) ((T_NEW) reduce 179) ((T_NS_C) reduce 179) ((T_OBJECT_CAST) reduce 179) ((T_PRINT) reduce 179) ((T_REQUIRE) reduce 179) ((T_REQUIRE_ONCE) reduce 179) ((T_RETURN) reduce 179) ((T_START_HEREDOC) reduce 179) ((T_STATIC) reduce 179) ((T_STRING) reduce 179) ((T_STRING_CAST) reduce 179) ((T_SWITCH) reduce 179) ((T_THROW) reduce 179) ((T_TRAIT) reduce 179) ((T_TRAIT_C) reduce 179) ((T_TRY) reduce 179) ((T_UNSET) reduce 179) ((T_UNSET_CAST) reduce 179) ((T_VARIABLE) reduce 179) ((T_WHILE) reduce 179) ((T_YIELD) reduce 179) ((T_YIELD_FROM) reduce 179) (("[") reduce 179) (("`") reduce 179) (("{") reduce 179) (("~") reduce 179)) 20192 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDFOREACH) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 20200 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDFOREACH) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 20201 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDFOREACH) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 20204 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDFOREACH) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 20208 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDFOREACH) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 20213 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_EMPTY) reduce 232) ((T_ENDFOREACH) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 20214 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDFOREACH) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 20217 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_EMPTY) reduce 233) ((T_ENDFOREACH) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 20222 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDFOREACH) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 20225 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDFOREACH) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 20226 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_EMPTY) reduce 174) ((T_ENDFOREACH) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 20229 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDFOREACH) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 20232 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDFOREACH) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 20239 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDFOREACH) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 20244 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDFOREACH) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 20249 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDFOREACH) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 20253 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_EMPTY) reduce 214) ((T_ENDFOREACH) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 20255 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_EMPTY) reduce 215) ((T_ENDFOREACH) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 20259 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDFOREACH) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 20262 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDFOREACH) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 20263 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDFOREACH) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 20265 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDFOREACH) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 20267 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDFOREACH) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 20273 ((("!") reduce 191) (("\"") reduce 191) (("$") reduce 191) (("(") reduce 191) (("+") reduce 191) (("-") reduce 191) ((";") reduce 191) (("@") reduce 191) ((T_ABSTRACT) reduce 191) ((T_ARRAY) reduce 191) ((T_ARRAY_CAST) reduce 191) ((T_ATTRIBUTE) reduce 191) ((T_BOOL_CAST) reduce 191) ((T_BREAK) reduce 191) ((T_CLASS) reduce 191) ((T_CLASS_C) reduce 191) ((T_CLONE) reduce 191) ((T_CONSTANT_ENCAPSED_STRING) reduce 191) ((T_CONTINUE) reduce 191) ((T_DEC) reduce 191) ((T_DECLARE) reduce 191) ((T_DIR) reduce 191) ((T_DNUMBER) reduce 191) ((T_DO) reduce 191) ((T_DOUBLE_CAST) reduce 191) ((T_ECHO) reduce 191) ((T_EMPTY) reduce 191) ((T_ENDFOREACH) reduce 191) ((T_ENUM) reduce 191) ((T_EVAL) reduce 191) ((T_EXIT) reduce 191) ((T_FILE) reduce 191) ((T_FINAL) reduce 191) ((T_FN) reduce 191) ((T_FOR) reduce 191) ((T_FOREACH) reduce 191) ((T_FUNCTION) reduce 191) ((T_FUNC_C) reduce 191) ((T_GLOBAL) reduce 191) ((T_GOTO) reduce 191) ((T_HALT_COMPILER) reduce 191) ((T_IF) reduce 191) ((T_INC) reduce 191) ((T_INCLUDE) reduce 191) ((T_INCLUDE_ONCE) reduce 191) ((T_INLINE_HTML) reduce 191) ((T_INTERFACE) reduce 191) ((T_INT_CAST) reduce 191) ((T_ISSET) reduce 191) ((T_LINE) reduce 191) ((T_LIST) reduce 191) ((T_LNUMBER) reduce 191) ((T_MATCH) reduce 191) ((T_METHOD_C) reduce 191) ((T_NAME_FULLY_QUALIFIED) reduce 191) ((T_NAME_QUALIFIED) reduce 191) ((T_NAME_RELATIVE) reduce 191) ((T_NEW) reduce 191) ((T_NS_C) reduce 191) ((T_OBJECT_CAST) reduce 191) ((T_PRINT) reduce 191) ((T_REQUIRE) reduce 191) ((T_REQUIRE_ONCE) reduce 191) ((T_RETURN) reduce 191) ((T_START_HEREDOC) reduce 191) ((T_STATIC) reduce 191) ((T_STRING) reduce 191) ((T_STRING_CAST) reduce 191) ((T_SWITCH) reduce 191) ((T_THROW) reduce 191) ((T_TRAIT) reduce 191) ((T_TRAIT_C) reduce 191) ((T_TRY) reduce 191) ((T_UNSET) reduce 191) ((T_UNSET_CAST) reduce 191) ((T_VARIABLE) reduce 191) ((T_WHILE) reduce 191) ((T_YIELD) reduce 191) ((T_YIELD_FROM) reduce 191) (("[") reduce 191) (("`") reduce 191) (("{") reduce 191) (("~") reduce 191)) 20277 ((("!") reduce 234) (("\"") reduce 234) (("$") reduce 234) (("(") reduce 234) (("+") reduce 234) (("-") reduce 234) ((";") reduce 234) (("@") reduce 234) ((T_ABSTRACT) reduce 234) ((T_ARRAY) reduce 234) ((T_ARRAY_CAST) reduce 234) ((T_ATTRIBUTE) reduce 234) ((T_BOOL_CAST) reduce 234) ((T_BREAK) reduce 234) ((T_CLASS) reduce 234) ((T_CLASS_C) reduce 234) ((T_CLONE) reduce 234) ((T_CONSTANT_ENCAPSED_STRING) reduce 234) ((T_CONTINUE) reduce 234) ((T_DEC) reduce 234) ((T_DECLARE) reduce 234) ((T_DIR) reduce 234) ((T_DNUMBER) reduce 234) ((T_DO) reduce 234) ((T_DOUBLE_CAST) reduce 234) ((T_ECHO) reduce 234) ((T_ELSE) reduce 234) ((T_ELSEIF) reduce 234) ((T_EMPTY) reduce 234) ((T_ENDFOREACH) reduce 234) ((T_ENUM) reduce 234) ((T_EVAL) reduce 234) ((T_EXIT) reduce 234) ((T_FILE) reduce 234) ((T_FINAL) reduce 234) ((T_FN) reduce 234) ((T_FOR) reduce 234) ((T_FOREACH) reduce 234) ((T_FUNCTION) reduce 234) ((T_FUNC_C) reduce 234) ((T_GLOBAL) reduce 234) ((T_GOTO) reduce 234) ((T_HALT_COMPILER) reduce 234) ((T_IF) reduce 234) ((T_INC) reduce 234) ((T_INCLUDE) reduce 234) ((T_INCLUDE_ONCE) reduce 234) ((T_INLINE_HTML) reduce 234) ((T_INTERFACE) reduce 234) ((T_INT_CAST) reduce 234) ((T_ISSET) reduce 234) ((T_LINE) reduce 234) ((T_LIST) reduce 234) ((T_LNUMBER) reduce 234) ((T_MATCH) reduce 234) ((T_METHOD_C) reduce 234) ((T_NAME_FULLY_QUALIFIED) reduce 234) ((T_NAME_QUALIFIED) reduce 234) ((T_NAME_RELATIVE) reduce 234) ((T_NEW) reduce 234) ((T_NS_C) reduce 234) ((T_OBJECT_CAST) reduce 234) ((T_PRINT) reduce 234) ((T_REQUIRE) reduce 234) ((T_REQUIRE_ONCE) reduce 234) ((T_RETURN) reduce 234) ((T_START_HEREDOC) reduce 234) ((T_STATIC) reduce 234) ((T_STRING) reduce 234) ((T_STRING_CAST) reduce 234) ((T_SWITCH) reduce 234) ((T_THROW) reduce 234) ((T_TRAIT) reduce 234) ((T_TRAIT_C) reduce 234) ((T_TRY) reduce 234) ((T_UNSET) reduce 234) ((T_UNSET_CAST) reduce 234) ((T_VARIABLE) reduce 234) ((T_WHILE) reduce 234) ((T_YIELD) reduce 234) ((T_YIELD_FROM) reduce 234) (("[") reduce 234) (("`") reduce 234) (("{") reduce 234) (("~") reduce 234)) 20280 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDFOREACH) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 20282 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDFOREACH) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 20284 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDFOREACH) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 20294 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDFOREACH) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 20295 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_EMPTY) reduce 210) ((T_ENDFOREACH) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 20298 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_EMPTY) reduce 211) ((T_ENDFOREACH) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 20299 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDFOREACH) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 20308 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDFOREACH) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 20309 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_EMPTY) reduce 208) ((T_ENDFOREACH) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 20312 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_EMPTY) reduce 209) ((T_ENDFOREACH) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 20319 ((("!") reduce 192) (("\"") reduce 192) (("$") reduce 192) (("(") reduce 192) (("+") reduce 192) (("-") reduce 192) ((";") reduce 192) (("@") reduce 192) ((T_ABSTRACT) reduce 192) ((T_ARRAY) reduce 192) ((T_ARRAY_CAST) reduce 192) ((T_ATTRIBUTE) reduce 192) ((T_BOOL_CAST) reduce 192) ((T_BREAK) reduce 192) ((T_CLASS) reduce 192) ((T_CLASS_C) reduce 192) ((T_CLONE) reduce 192) ((T_CONSTANT_ENCAPSED_STRING) reduce 192) ((T_CONTINUE) reduce 192) ((T_DEC) reduce 192) ((T_DECLARE) reduce 192) ((T_DIR) reduce 192) ((T_DNUMBER) reduce 192) ((T_DO) reduce 192) ((T_DOUBLE_CAST) reduce 192) ((T_ECHO) reduce 192) ((T_EMPTY) reduce 192) ((T_ENDFOREACH) reduce 192) ((T_ENUM) reduce 192) ((T_EVAL) reduce 192) ((T_EXIT) reduce 192) ((T_FILE) reduce 192) ((T_FINAL) reduce 192) ((T_FN) reduce 192) ((T_FOR) reduce 192) ((T_FOREACH) reduce 192) ((T_FUNCTION) reduce 192) ((T_FUNC_C) reduce 192) ((T_GLOBAL) reduce 192) ((T_GOTO) reduce 192) ((T_HALT_COMPILER) reduce 192) ((T_IF) reduce 192) ((T_INC) reduce 192) ((T_INCLUDE) reduce 192) ((T_INCLUDE_ONCE) reduce 192) ((T_INLINE_HTML) reduce 192) ((T_INTERFACE) reduce 192) ((T_INT_CAST) reduce 192) ((T_ISSET) reduce 192) ((T_LINE) reduce 192) ((T_LIST) reduce 192) ((T_LNUMBER) reduce 192) ((T_MATCH) reduce 192) ((T_METHOD_C) reduce 192) ((T_NAME_FULLY_QUALIFIED) reduce 192) ((T_NAME_QUALIFIED) reduce 192) ((T_NAME_RELATIVE) reduce 192) ((T_NEW) reduce 192) ((T_NS_C) reduce 192) ((T_OBJECT_CAST) reduce 192) ((T_PRINT) reduce 192) ((T_REQUIRE) reduce 192) ((T_REQUIRE_ONCE) reduce 192) ((T_RETURN) reduce 192) ((T_START_HEREDOC) reduce 192) ((T_STATIC) reduce 192) ((T_STRING) reduce 192) ((T_STRING_CAST) reduce 192) ((T_SWITCH) reduce 192) ((T_THROW) reduce 192) ((T_TRAIT) reduce 192) ((T_TRAIT_C) reduce 192) ((T_TRY) reduce 192) ((T_UNSET) reduce 192) ((T_UNSET_CAST) reduce 192) ((T_VARIABLE) reduce 192) ((T_WHILE) reduce 192) ((T_YIELD) reduce 192) ((T_YIELD_FROM) reduce 192) (("[") reduce 192) (("`") reduce 192) (("{") reduce 192) (("~") reduce 192)) 20320 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CASE) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DEFAULT) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_ELSE) reduce 211) ((T_ELSEIF) reduce 211) ((T_EMPTY) reduce 211) ((T_ENDSWITCH) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 20322 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDFOREACH) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 20328 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDFOREACH) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 20333 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDFOREACH) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 20334 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_EMPTY) reduce 212) ((T_ENDFOREACH) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 20337 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_EMPTY) reduce 213) ((T_ENDFOREACH) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 20339 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDFOREACH) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 20346 ((("!") reduce 185) (("\"") reduce 185) (("$") reduce 185) (("(") reduce 185) (("+") reduce 185) (("-") reduce 185) ((";") reduce 185) (("@") reduce 185) ((T_ABSTRACT) reduce 185) ((T_ARRAY) reduce 185) ((T_ARRAY_CAST) reduce 185) ((T_ATTRIBUTE) reduce 185) ((T_BOOL_CAST) reduce 185) ((T_BREAK) reduce 185) ((T_CLASS) reduce 185) ((T_CLASS_C) reduce 185) ((T_CLONE) reduce 185) ((T_CONSTANT_ENCAPSED_STRING) reduce 185) ((T_CONTINUE) reduce 185) ((T_DEC) reduce 185) ((T_DECLARE) reduce 185) ((T_DIR) reduce 185) ((T_DNUMBER) reduce 185) ((T_DO) reduce 185) ((T_DOUBLE_CAST) reduce 185) ((T_ECHO) reduce 185) ((T_EMPTY) reduce 185) ((T_ENDFOREACH) reduce 185) ((T_ENUM) reduce 185) ((T_EVAL) reduce 185) ((T_EXIT) reduce 185) ((T_FILE) reduce 185) ((T_FINAL) reduce 185) ((T_FN) reduce 185) ((T_FOR) reduce 185) ((T_FOREACH) reduce 185) ((T_FUNCTION) reduce 185) ((T_FUNC_C) reduce 185) ((T_GLOBAL) reduce 185) ((T_GOTO) reduce 185) ((T_HALT_COMPILER) reduce 185) ((T_IF) reduce 185) ((T_INC) reduce 185) ((T_INCLUDE) reduce 185) ((T_INCLUDE_ONCE) reduce 185) ((T_INLINE_HTML) reduce 185) ((T_INTERFACE) reduce 185) ((T_INT_CAST) reduce 185) ((T_ISSET) reduce 185) ((T_LINE) reduce 185) ((T_LIST) reduce 185) ((T_LNUMBER) reduce 185) ((T_MATCH) reduce 185) ((T_METHOD_C) reduce 185) ((T_NAME_FULLY_QUALIFIED) reduce 185) ((T_NAME_QUALIFIED) reduce 185) ((T_NAME_RELATIVE) reduce 185) ((T_NEW) reduce 185) ((T_NS_C) reduce 185) ((T_OBJECT_CAST) reduce 185) ((T_PRINT) reduce 185) ((T_REQUIRE) reduce 185) ((T_REQUIRE_ONCE) reduce 185) ((T_RETURN) reduce 185) ((T_START_HEREDOC) reduce 185) ((T_STATIC) reduce 185) ((T_STRING) reduce 185) ((T_STRING_CAST) reduce 185) ((T_SWITCH) reduce 185) ((T_THROW) reduce 185) ((T_TRAIT) reduce 185) ((T_TRAIT_C) reduce 185) ((T_TRY) reduce 185) ((T_UNSET) reduce 185) ((T_UNSET_CAST) reduce 185) ((T_VARIABLE) reduce 185) ((T_WHILE) reduce 185) ((T_YIELD) reduce 185) ((T_YIELD_FROM) reduce 185) (("[") reduce 185) (("`") reduce 185) (("{") reduce 185) (("~") reduce 185)) 20348 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDFOREACH) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 20351 ((("(") reduce 529) ((")") reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) (("[") reduce 518) (("{") reduce 518)) 20352 ((("(") reduce 455) ((")") reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) (("[") reduce 455) (("{") reduce 455)) 20353 ((("(") reduce 457) ((")") reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) (("[") reduce 457) (("{") reduce 457)) 20359 ((("(") reduce 507) ((")") reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) (("[") reduce 507) (("{") reduce 507)) 20361 ((("(") reduce 506) ((")") reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) (("[") reduce 506) (("{") reduce 506)) 20362 ((("(") reduce 530) ((")") reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) (("[") reduce 530) (("{") reduce 530)) 20363 ((("(") shift) ((")") reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) (("[") reduce 513) (("{") reduce 513)) 20364 ((("(") reduce 532) ((")") reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) (("[") reduce 532) (("{") reduce 532)) 20367 ((("(") reduce 531) ((")") reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531) (("{") reduce 531)) 20368 ((("(") reduce 508) ((")") reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) (("[") reduce 508) (("{") reduce 508)) 20369 ((("(") shift) ((")") reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) (("[") reduce 514) (("{") reduce 514)) 20370 ((("(") reduce 509) ((")") reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) (("[") reduce 509) (("{") reduce 509)) 20371 (((")") reduce 205) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) (("[") reduce 496) (("{") reduce 496)) 20373 ((("(") reduce 471) ((")") reduce 207) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) (("[") reduce 471) (("{") reduce 471)) 20376 (((")") reduce 206)) 20377 ((("(") reduce 517) ((")") reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) (("[") reduce 517) (("{") reduce 517)) 20380 ((("(") reduce 516) ((")") reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) (("[") reduce 516) (("{") reduce 516)) 20381 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CASE) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DEFAULT) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDSWITCH) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 20384 ((("(") reduce 529) ((")") reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) (("[") reduce 518) (("{") reduce 518)) 20385 ((("(") reduce 455) ((")") reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) (("[") reduce 455) (("{") reduce 455)) 20386 ((("(") reduce 457) ((")") reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) (("[") reduce 457) (("{") reduce 457)) 20392 ((("(") reduce 507) ((")") reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) (("[") reduce 507) (("{") reduce 507)) 20394 ((("(") reduce 506) ((")") reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) (("[") reduce 506) (("{") reduce 506)) 20395 ((("(") reduce 530) ((")") reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) (("[") reduce 530) (("{") reduce 530)) 20396 ((("(") shift) ((")") reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) (("[") reduce 513) (("{") reduce 513)) 20397 ((("(") reduce 532) ((")") reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) (("[") reduce 532) (("{") reduce 532)) 20400 ((("(") reduce 531) ((")") reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) (("[") reduce 531) (("{") reduce 531)) 20401 ((("(") reduce 508) ((")") reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) (("[") reduce 508) (("{") reduce 508)) 20402 ((("(") shift) ((")") reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) (("[") reduce 514) (("{") reduce 514)) 20403 ((("(") reduce 509) ((")") reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) (("[") reduce 509) (("{") reduce 509)) 20404 (((")") reduce 205) ((T_DOUBLE_ARROW) reduce 205) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) (("[") reduce 496) (("{") reduce 496)) 20406 ((("(") reduce 471) ((")") reduce 207) ((T_DOUBLE_ARROW) reduce 207) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) (("[") reduce 471) (("{") reduce 471)) 20409 (((")") reduce 206) ((T_DOUBLE_ARROW) reduce 206)) 20410 ((("(") reduce 517) ((")") reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) (("[") reduce 517) (("{") reduce 517)) 20413 ((("(") reduce 516) ((")") reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) (("[") reduce 516) (("{") reduce 516)) 21034 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CASE) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DEFAULT) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDSWITCH) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 21035 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CASE) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DEFAULT) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_ELSE) reduce 208) ((T_ELSEIF) reduce 208) ((T_EMPTY) reduce 208) ((T_ENDSWITCH) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 21038 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CASE) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DEFAULT) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_ELSE) reduce 209) ((T_ELSEIF) reduce 209) ((T_EMPTY) reduce 209) ((T_ENDSWITCH) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 21040 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CASE) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DEFAULT) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDSWITCH) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 21046 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CASE) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DEFAULT) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDSWITCH) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 21051 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CASE) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DEFAULT) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDSWITCH) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 21052 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CASE) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DEFAULT) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_ELSE) reduce 212) ((T_ELSEIF) reduce 212) ((T_EMPTY) reduce 212) ((T_ENDSWITCH) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 21055 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CASE) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DEFAULT) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_ELSE) reduce 213) ((T_ELSEIF) reduce 213) ((T_EMPTY) reduce 213) ((T_ENDSWITCH) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 21057 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CASE) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DEFAULT) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDSWITCH) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 21059 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CASE) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DEFAULT) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDSWITCH) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 21060 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CASE) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DEFAULT) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_EMPTY) reduce 237) ((T_ENDSWITCH) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 21071 ((("!") reduce 444) (("\"") reduce 444) (("$") reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) ((";") reduce 444) (("@") reduce 444) ((T_ABSTRACT) reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_BREAK) reduce 444) ((T_CASE) reduce 444) ((T_CLASS) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_CONTINUE) reduce 444) ((T_DEC) reduce 444) ((T_DECLARE) reduce 444) ((T_DEFAULT) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DO) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_ECHO) reduce 444) ((T_EMPTY) reduce 444) ((T_ENDSWITCH) reduce 444) ((T_ENUM) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FINAL) reduce 444) ((T_FN) reduce 444) ((T_FOR) reduce 444) ((T_FOREACH) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_GLOBAL) reduce 444) ((T_GOTO) reduce 444) ((T_HALT_COMPILER) reduce 444) ((T_IF) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INLINE_HTML) reduce 444) ((T_INTERFACE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_RETURN) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_SWITCH) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT) reduce 444) ((T_TRAIT_C) reduce 444) ((T_TRY) reduce 444) ((T_UNSET) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_VARIABLE) reduce 444) ((T_WHILE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("{") reduce 444) (("~") reduce 444)) 21072 ((("!") reduce 179) (("\"") reduce 179) (("$") reduce 179) (("(") reduce 179) (("+") reduce 179) (("-") reduce 179) ((";") reduce 179) (("@") reduce 179) ((T_ABSTRACT) reduce 179) ((T_ARRAY) reduce 179) ((T_ARRAY_CAST) reduce 179) ((T_ATTRIBUTE) reduce 179) ((T_BOOL_CAST) reduce 179) ((T_BREAK) reduce 179) ((T_CASE) reduce 179) ((T_CLASS) reduce 179) ((T_CLASS_C) reduce 179) ((T_CLONE) reduce 179) ((T_CONSTANT_ENCAPSED_STRING) reduce 179) ((T_CONTINUE) reduce 179) ((T_DEC) reduce 179) ((T_DECLARE) reduce 179) ((T_DEFAULT) reduce 179) ((T_DIR) reduce 179) ((T_DNUMBER) reduce 179) ((T_DO) reduce 179) ((T_DOUBLE_CAST) reduce 179) ((T_ECHO) reduce 179) ((T_EMPTY) reduce 179) ((T_ENDSWITCH) reduce 179) ((T_ENUM) reduce 179) ((T_EVAL) reduce 179) ((T_EXIT) reduce 179) ((T_FILE) reduce 179) ((T_FINAL) reduce 179) ((T_FN) reduce 179) ((T_FOR) reduce 179) ((T_FOREACH) reduce 179) ((T_FUNCTION) reduce 179) ((T_FUNC_C) reduce 179) ((T_GLOBAL) reduce 179) ((T_GOTO) reduce 179) ((T_HALT_COMPILER) reduce 179) ((T_IF) reduce 179) ((T_INC) reduce 179) ((T_INCLUDE) reduce 179) ((T_INCLUDE_ONCE) reduce 179) ((T_INLINE_HTML) reduce 179) ((T_INTERFACE) reduce 179) ((T_INT_CAST) reduce 179) ((T_ISSET) reduce 179) ((T_LINE) reduce 179) ((T_LIST) reduce 179) ((T_LNUMBER) reduce 179) ((T_MATCH) reduce 179) ((T_METHOD_C) reduce 179) ((T_NAME_FULLY_QUALIFIED) reduce 179) ((T_NAME_QUALIFIED) reduce 179) ((T_NAME_RELATIVE) reduce 179) ((T_NEW) reduce 179) ((T_NS_C) reduce 179) ((T_OBJECT_CAST) reduce 179) ((T_PRINT) reduce 179) ((T_REQUIRE) reduce 179) ((T_REQUIRE_ONCE) reduce 179) ((T_RETURN) reduce 179) ((T_START_HEREDOC) reduce 179) ((T_STATIC) reduce 179) ((T_STRING) reduce 179) ((T_STRING_CAST) reduce 179) ((T_SWITCH) reduce 179) ((T_THROW) reduce 179) ((T_TRAIT) reduce 179) ((T_TRAIT_C) reduce 179) ((T_TRY) reduce 179) ((T_UNSET) reduce 179) ((T_UNSET_CAST) reduce 179) ((T_VARIABLE) reduce 179) ((T_WHILE) reduce 179) ((T_YIELD) reduce 179) ((T_YIELD_FROM) reduce 179) (("[") reduce 179) (("`") reduce 179) (("{") reduce 179) (("~") reduce 179)) 21073 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CASE) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DEFAULT) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDSWITCH) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 21081 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CASE) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DEFAULT) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDSWITCH) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 21082 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CASE) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DEFAULT) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDSWITCH) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 21085 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CASE) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DEFAULT) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDSWITCH) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 21089 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CASE) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DEFAULT) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDSWITCH) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 21094 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CASE) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DEFAULT) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_EMPTY) reduce 232) ((T_ENDSWITCH) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 21095 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CASE) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DEFAULT) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDSWITCH) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 21098 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CASE) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DEFAULT) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_EMPTY) reduce 233) ((T_ENDSWITCH) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 21103 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CASE) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DEFAULT) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDSWITCH) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 21106 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CASE) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DEFAULT) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDSWITCH) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 21107 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CASE) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DEFAULT) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_EMPTY) reduce 174) ((T_ENDSWITCH) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 21110 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CASE) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DEFAULT) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDSWITCH) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 21113 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CASE) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DEFAULT) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDSWITCH) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 21120 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CASE) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DEFAULT) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDSWITCH) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 21125 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CASE) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DEFAULT) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDSWITCH) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 21130 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CASE) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DEFAULT) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDSWITCH) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 21134 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CASE) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DEFAULT) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_EMPTY) reduce 214) ((T_ENDSWITCH) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 21136 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CASE) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DEFAULT) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_EMPTY) reduce 215) ((T_ENDSWITCH) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 21140 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CASE) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DEFAULT) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDSWITCH) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 21143 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CASE) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DEFAULT) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDSWITCH) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 21144 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CASE) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DEFAULT) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDSWITCH) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 21146 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CASE) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DEFAULT) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDSWITCH) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 21148 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CASE) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DEFAULT) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDSWITCH) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 21154 ((("!") reduce 191) (("\"") reduce 191) (("$") reduce 191) (("(") reduce 191) (("+") reduce 191) (("-") reduce 191) ((";") reduce 191) (("@") reduce 191) ((T_ABSTRACT) reduce 191) ((T_ARRAY) reduce 191) ((T_ARRAY_CAST) reduce 191) ((T_ATTRIBUTE) reduce 191) ((T_BOOL_CAST) reduce 191) ((T_BREAK) reduce 191) ((T_CASE) reduce 191) ((T_CLASS) reduce 191) ((T_CLASS_C) reduce 191) ((T_CLONE) reduce 191) ((T_CONSTANT_ENCAPSED_STRING) reduce 191) ((T_CONTINUE) reduce 191) ((T_DEC) reduce 191) ((T_DECLARE) reduce 191) ((T_DEFAULT) reduce 191) ((T_DIR) reduce 191) ((T_DNUMBER) reduce 191) ((T_DO) reduce 191) ((T_DOUBLE_CAST) reduce 191) ((T_ECHO) reduce 191) ((T_EMPTY) reduce 191) ((T_ENDSWITCH) reduce 191) ((T_ENUM) reduce 191) ((T_EVAL) reduce 191) ((T_EXIT) reduce 191) ((T_FILE) reduce 191) ((T_FINAL) reduce 191) ((T_FN) reduce 191) ((T_FOR) reduce 191) ((T_FOREACH) reduce 191) ((T_FUNCTION) reduce 191) ((T_FUNC_C) reduce 191) ((T_GLOBAL) reduce 191) ((T_GOTO) reduce 191) ((T_HALT_COMPILER) reduce 191) ((T_IF) reduce 191) ((T_INC) reduce 191) ((T_INCLUDE) reduce 191) ((T_INCLUDE_ONCE) reduce 191) ((T_INLINE_HTML) reduce 191) ((T_INTERFACE) reduce 191) ((T_INT_CAST) reduce 191) ((T_ISSET) reduce 191) ((T_LINE) reduce 191) ((T_LIST) reduce 191) ((T_LNUMBER) reduce 191) ((T_MATCH) reduce 191) ((T_METHOD_C) reduce 191) ((T_NAME_FULLY_QUALIFIED) reduce 191) ((T_NAME_QUALIFIED) reduce 191) ((T_NAME_RELATIVE) reduce 191) ((T_NEW) reduce 191) ((T_NS_C) reduce 191) ((T_OBJECT_CAST) reduce 191) ((T_PRINT) reduce 191) ((T_REQUIRE) reduce 191) ((T_REQUIRE_ONCE) reduce 191) ((T_RETURN) reduce 191) ((T_START_HEREDOC) reduce 191) ((T_STATIC) reduce 191) ((T_STRING) reduce 191) ((T_STRING_CAST) reduce 191) ((T_SWITCH) reduce 191) ((T_THROW) reduce 191) ((T_TRAIT) reduce 191) ((T_TRAIT_C) reduce 191) ((T_TRY) reduce 191) ((T_UNSET) reduce 191) ((T_UNSET_CAST) reduce 191) ((T_VARIABLE) reduce 191) ((T_WHILE) reduce 191) ((T_YIELD) reduce 191) ((T_YIELD_FROM) reduce 191) (("[") reduce 191) (("`") reduce 191) (("{") reduce 191) (("~") reduce 191)) 21158 ((("!") reduce 234) (("\"") reduce 234) (("$") reduce 234) (("(") reduce 234) (("+") reduce 234) (("-") reduce 234) ((";") reduce 234) (("@") reduce 234) ((T_ABSTRACT) reduce 234) ((T_ARRAY) reduce 234) ((T_ARRAY_CAST) reduce 234) ((T_ATTRIBUTE) reduce 234) ((T_BOOL_CAST) reduce 234) ((T_BREAK) reduce 234) ((T_CASE) reduce 234) ((T_CLASS) reduce 234) ((T_CLASS_C) reduce 234) ((T_CLONE) reduce 234) ((T_CONSTANT_ENCAPSED_STRING) reduce 234) ((T_CONTINUE) reduce 234) ((T_DEC) reduce 234) ((T_DECLARE) reduce 234) ((T_DEFAULT) reduce 234) ((T_DIR) reduce 234) ((T_DNUMBER) reduce 234) ((T_DO) reduce 234) ((T_DOUBLE_CAST) reduce 234) ((T_ECHO) reduce 234) ((T_ELSE) reduce 234) ((T_ELSEIF) reduce 234) ((T_EMPTY) reduce 234) ((T_ENDSWITCH) reduce 234) ((T_ENUM) reduce 234) ((T_EVAL) reduce 234) ((T_EXIT) reduce 234) ((T_FILE) reduce 234) ((T_FINAL) reduce 234) ((T_FN) reduce 234) ((T_FOR) reduce 234) ((T_FOREACH) reduce 234) ((T_FUNCTION) reduce 234) ((T_FUNC_C) reduce 234) ((T_GLOBAL) reduce 234) ((T_GOTO) reduce 234) ((T_HALT_COMPILER) reduce 234) ((T_IF) reduce 234) ((T_INC) reduce 234) ((T_INCLUDE) reduce 234) ((T_INCLUDE_ONCE) reduce 234) ((T_INLINE_HTML) reduce 234) ((T_INTERFACE) reduce 234) ((T_INT_CAST) reduce 234) ((T_ISSET) reduce 234) ((T_LINE) reduce 234) ((T_LIST) reduce 234) ((T_LNUMBER) reduce 234) ((T_MATCH) reduce 234) ((T_METHOD_C) reduce 234) ((T_NAME_FULLY_QUALIFIED) reduce 234) ((T_NAME_QUALIFIED) reduce 234) ((T_NAME_RELATIVE) reduce 234) ((T_NEW) reduce 234) ((T_NS_C) reduce 234) ((T_OBJECT_CAST) reduce 234) ((T_PRINT) reduce 234) ((T_REQUIRE) reduce 234) ((T_REQUIRE_ONCE) reduce 234) ((T_RETURN) reduce 234) ((T_START_HEREDOC) reduce 234) ((T_STATIC) reduce 234) ((T_STRING) reduce 234) ((T_STRING_CAST) reduce 234) ((T_SWITCH) reduce 234) ((T_THROW) reduce 234) ((T_TRAIT) reduce 234) ((T_TRAIT_C) reduce 234) ((T_TRY) reduce 234) ((T_UNSET) reduce 234) ((T_UNSET_CAST) reduce 234) ((T_VARIABLE) reduce 234) ((T_WHILE) reduce 234) ((T_YIELD) reduce 234) ((T_YIELD_FROM) reduce 234) (("[") reduce 234) (("`") reduce 234) (("{") reduce 234) (("~") reduce 234)) 21161 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CASE) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DEFAULT) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDSWITCH) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 21163 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CASE) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DEFAULT) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDSWITCH) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 21165 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CASE) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DEFAULT) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDSWITCH) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 21175 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CASE) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DEFAULT) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDSWITCH) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 21176 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CASE) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DEFAULT) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_EMPTY) reduce 210) ((T_ENDSWITCH) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 21179 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CASE) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DEFAULT) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_EMPTY) reduce 211) ((T_ENDSWITCH) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 21180 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CASE) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DEFAULT) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDSWITCH) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 21189 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CASE) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DEFAULT) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDSWITCH) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 21190 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CASE) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DEFAULT) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_EMPTY) reduce 208) ((T_ENDSWITCH) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 21193 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CASE) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DEFAULT) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_EMPTY) reduce 209) ((T_ENDSWITCH) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 21200 ((("!") reduce 192) (("\"") reduce 192) (("$") reduce 192) (("(") reduce 192) (("+") reduce 192) (("-") reduce 192) ((";") reduce 192) (("@") reduce 192) ((T_ABSTRACT) reduce 192) ((T_ARRAY) reduce 192) ((T_ARRAY_CAST) reduce 192) ((T_ATTRIBUTE) reduce 192) ((T_BOOL_CAST) reduce 192) ((T_BREAK) reduce 192) ((T_CASE) reduce 192) ((T_CLASS) reduce 192) ((T_CLASS_C) reduce 192) ((T_CLONE) reduce 192) ((T_CONSTANT_ENCAPSED_STRING) reduce 192) ((T_CONTINUE) reduce 192) ((T_DEC) reduce 192) ((T_DECLARE) reduce 192) ((T_DEFAULT) reduce 192) ((T_DIR) reduce 192) ((T_DNUMBER) reduce 192) ((T_DO) reduce 192) ((T_DOUBLE_CAST) reduce 192) ((T_ECHO) reduce 192) ((T_EMPTY) reduce 192) ((T_ENDSWITCH) reduce 192) ((T_ENUM) reduce 192) ((T_EVAL) reduce 192) ((T_EXIT) reduce 192) ((T_FILE) reduce 192) ((T_FINAL) reduce 192) ((T_FN) reduce 192) ((T_FOR) reduce 192) ((T_FOREACH) reduce 192) ((T_FUNCTION) reduce 192) ((T_FUNC_C) reduce 192) ((T_GLOBAL) reduce 192) ((T_GOTO) reduce 192) ((T_HALT_COMPILER) reduce 192) ((T_IF) reduce 192) ((T_INC) reduce 192) ((T_INCLUDE) reduce 192) ((T_INCLUDE_ONCE) reduce 192) ((T_INLINE_HTML) reduce 192) ((T_INTERFACE) reduce 192) ((T_INT_CAST) reduce 192) ((T_ISSET) reduce 192) ((T_LINE) reduce 192) ((T_LIST) reduce 192) ((T_LNUMBER) reduce 192) ((T_MATCH) reduce 192) ((T_METHOD_C) reduce 192) ((T_NAME_FULLY_QUALIFIED) reduce 192) ((T_NAME_QUALIFIED) reduce 192) ((T_NAME_RELATIVE) reduce 192) ((T_NEW) reduce 192) ((T_NS_C) reduce 192) ((T_OBJECT_CAST) reduce 192) ((T_PRINT) reduce 192) ((T_REQUIRE) reduce 192) ((T_REQUIRE_ONCE) reduce 192) ((T_RETURN) reduce 192) ((T_START_HEREDOC) reduce 192) ((T_STATIC) reduce 192) ((T_STRING) reduce 192) ((T_STRING_CAST) reduce 192) ((T_SWITCH) reduce 192) ((T_THROW) reduce 192) ((T_TRAIT) reduce 192) ((T_TRAIT_C) reduce 192) ((T_TRY) reduce 192) ((T_UNSET) reduce 192) ((T_UNSET_CAST) reduce 192) ((T_VARIABLE) reduce 192) ((T_WHILE) reduce 192) ((T_YIELD) reduce 192) ((T_YIELD_FROM) reduce 192) (("[") reduce 192) (("`") reduce 192) (("{") reduce 192) (("~") reduce 192)) 21202 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CASE) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DEFAULT) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDSWITCH) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 21208 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CASE) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DEFAULT) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDSWITCH) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 21213 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CASE) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DEFAULT) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDSWITCH) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 21214 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CASE) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DEFAULT) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_EMPTY) reduce 212) ((T_ENDSWITCH) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 21217 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CASE) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DEFAULT) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_EMPTY) reduce 213) ((T_ENDSWITCH) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 21219 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CASE) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DEFAULT) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDSWITCH) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 21226 ((("!") reduce 185) (("\"") reduce 185) (("$") reduce 185) (("(") reduce 185) (("+") reduce 185) (("-") reduce 185) ((";") reduce 185) (("@") reduce 185) ((T_ABSTRACT) reduce 185) ((T_ARRAY) reduce 185) ((T_ARRAY_CAST) reduce 185) ((T_ATTRIBUTE) reduce 185) ((T_BOOL_CAST) reduce 185) ((T_BREAK) reduce 185) ((T_CASE) reduce 185) ((T_CLASS) reduce 185) ((T_CLASS_C) reduce 185) ((T_CLONE) reduce 185) ((T_CONSTANT_ENCAPSED_STRING) reduce 185) ((T_CONTINUE) reduce 185) ((T_DEC) reduce 185) ((T_DECLARE) reduce 185) ((T_DEFAULT) reduce 185) ((T_DIR) reduce 185) ((T_DNUMBER) reduce 185) ((T_DO) reduce 185) ((T_DOUBLE_CAST) reduce 185) ((T_ECHO) reduce 185) ((T_EMPTY) reduce 185) ((T_ENDSWITCH) reduce 185) ((T_ENUM) reduce 185) ((T_EVAL) reduce 185) ((T_EXIT) reduce 185) ((T_FILE) reduce 185) ((T_FINAL) reduce 185) ((T_FN) reduce 185) ((T_FOR) reduce 185) ((T_FOREACH) reduce 185) ((T_FUNCTION) reduce 185) ((T_FUNC_C) reduce 185) ((T_GLOBAL) reduce 185) ((T_GOTO) reduce 185) ((T_HALT_COMPILER) reduce 185) ((T_IF) reduce 185) ((T_INC) reduce 185) ((T_INCLUDE) reduce 185) ((T_INCLUDE_ONCE) reduce 185) ((T_INLINE_HTML) reduce 185) ((T_INTERFACE) reduce 185) ((T_INT_CAST) reduce 185) ((T_ISSET) reduce 185) ((T_LINE) reduce 185) ((T_LIST) reduce 185) ((T_LNUMBER) reduce 185) ((T_MATCH) reduce 185) ((T_METHOD_C) reduce 185) ((T_NAME_FULLY_QUALIFIED) reduce 185) ((T_NAME_QUALIFIED) reduce 185) ((T_NAME_RELATIVE) reduce 185) ((T_NEW) reduce 185) ((T_NS_C) reduce 185) ((T_OBJECT_CAST) reduce 185) ((T_PRINT) reduce 185) ((T_REQUIRE) reduce 185) ((T_REQUIRE_ONCE) reduce 185) ((T_RETURN) reduce 185) ((T_START_HEREDOC) reduce 185) ((T_STATIC) reduce 185) ((T_STRING) reduce 185) ((T_STRING_CAST) reduce 185) ((T_SWITCH) reduce 185) ((T_THROW) reduce 185) ((T_TRAIT) reduce 185) ((T_TRAIT_C) reduce 185) ((T_TRY) reduce 185) ((T_UNSET) reduce 185) ((T_UNSET_CAST) reduce 185) ((T_VARIABLE) reduce 185) ((T_WHILE) reduce 185) ((T_YIELD) reduce 185) ((T_YIELD_FROM) reduce 185) (("[") reduce 185) (("`") reduce 185) (("{") reduce 185) (("~") reduce 185)) 21228 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CASE) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DEFAULT) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDSWITCH) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 21239 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 21241 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((":") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 21243 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((":") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 21244 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((":") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 21248 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) ((":") reduce 463) ((";") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 21249 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((":") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 21250 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 21256 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) ((":") reduce 483) ((";") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 21258 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((":") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 21260 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 21261 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 21262 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 21263 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 21265 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 21272 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 21275 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((":") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 21277 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) ((";") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 21278 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) ((":") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 21285 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) ((";") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 21286 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) ((":") reduce 481) ((";") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 21288 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) ((":") reduce 480) ((";") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 21289 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) ((":") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 21290 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") shift) ((";") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21293 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) ((";") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 21294 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) ((":") reduce 434) ((";") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 21295 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) ((":") reduce 416) ((";") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 21296 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) ((":") reduce 438) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 21297 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) ((":") reduce 482) ((";") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 21298 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) ((":") reduce 412) ((";") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 21299 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((":") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 21300 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 21301 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) ((";") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 21302 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) ((":") reduce 356) ((";") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 21305 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((":") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 21336 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 384) ((";") reduce 384) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 21337 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 387) ((";") reduce 387) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 21338 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) ((":") reduce 396) ((";") reduce 396) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 21339 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 409) ((";") reduce 409) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 21340 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) ((":") reduce 395) ((";") reduce 395) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 21341 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((":") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 21342 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 383) ((";") reduce 383) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21343 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 381) ((";") reduce 381) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21344 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 382) ((";") reduce 382) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21345 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 406) ((";") reduce 406) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 21346 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 402) ((";") reduce 402) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 21347 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 404) ((";") reduce 404) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 21348 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 401) ((";") reduce 401) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 21349 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 408) ((";") reduce 408) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 21350 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 403) ((";") reduce 403) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 21353 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 21354 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 21355 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 21356 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((":") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 21357 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 21358 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 21359 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 21360 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((":") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 21361 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 21362 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((":") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 21363 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 21370 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 21372 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 21373 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 21374 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 21375 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 21376 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 21379 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 21380 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 21382 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 21384 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 21385 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 21388 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 21389 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 415) ((";") reduce 415) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21390 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 379) ((";") reduce 379) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21391 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 380) ((";") reduce 380) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21392 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 385) ((";") reduce 385) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 21393 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 386) ((";") reduce 386) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 21397 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 413) ((";") reduce 413) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21398 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 414) ((";") reduce 414) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21399 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 407) ((";") reduce 407) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 21400 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 405) ((";") reduce 405) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 21401 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((":") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 21402 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) ((":") reduce 388) ((";") reduce 388) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 21403 ((("%") shift) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) ((":") reduce 390) ((";") reduce 390) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 21404 ((("%") shift) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) ((":") reduce 389) ((";") reduce 389) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 21405 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) ((":") reduce 391) ((";") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 21406 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) ((":") reduce 394) ((";") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 21408 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((":") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 21409 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((":") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 21410 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((":") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 21411 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((":") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 21412 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((":") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 21413 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((":") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 21414 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((":") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 21415 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((":") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 21416 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((":") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 21417 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((":") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 21418 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((":") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 21419 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((":") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 21420 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((":") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 21421 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((":") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 21422 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((":") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 21423 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((":") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 21424 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((":") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 21425 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((":") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 21426 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((":") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 21427 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((":") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 21428 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((":") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 21429 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((":") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 21430 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((":") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 21431 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((":") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 21432 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((":") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 21433 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((":") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 21434 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((":") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 21435 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((":") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 21436 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((":") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 21437 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((":") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 21438 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((":") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 21439 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((":") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 21440 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((":") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 21441 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((":") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 21442 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((":") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 21443 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((":") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 21444 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((":") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 21445 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((":") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 21446 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((":") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 21447 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((":") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 21448 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((":") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 21449 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((":") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 21450 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((":") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 21451 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((":") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 21452 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((":") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 21453 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((":") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 21454 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((":") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 21455 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((":") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 21456 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((":") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 21457 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((":") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 21458 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((":") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 21459 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((":") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 21460 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((":") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 21461 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((":") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 21462 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((":") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 21463 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((":") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 21464 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((":") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 21465 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((":") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 21466 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((":") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 21467 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((":") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 21468 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((":") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 21469 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((":") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 21470 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((":") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 21471 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((":") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 21472 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((":") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 21473 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((":") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 21474 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((":") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 21475 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) ((":") reduce 81) ((";") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 21476 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((":") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 21477 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((":") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 21478 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((":") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 21479 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((":") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 21480 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((":") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 21481 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((":") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 21482 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((":") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 21483 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((":") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 21484 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((":") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 21485 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((":") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 21486 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((":") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 21488 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((":") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 21489 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((":") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 21490 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) ((";") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 21492 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((":") reduce 456) ((";") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 21493 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((":") reduce 281) ((";") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 21497 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((":") reduce 282) ((";") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 21498 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((":") reduce 283) ((";") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 21503 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) ((":") reduce 377) ((";") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 21505 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) ((":") reduce 375) ((";") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 21515 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 371) ((";") reduce 371) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21516 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 373) ((";") reduce 373) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21517 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 372) ((";") reduce 372) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21518 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 365) ((";") reduce 365) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21519 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 362) ((";") reduce 362) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21520 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 370) ((";") reduce 370) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21521 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 364) ((";") reduce 364) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21522 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 368) ((";") reduce 368) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21523 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 363) ((";") reduce 363) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21524 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 366) ((";") reduce 366) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21525 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 367) ((";") reduce 367) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21526 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 374) ((";") reduce 374) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21527 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 369) ((";") reduce 369) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21529 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 359) ((";") reduce 359) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21531 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 21534 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 21536 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 21538 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 21539 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 21540 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) ((":") reduce 360) ((";") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 21544 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 21546 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((":") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 21547 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((":") reduce 281) ((";") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 21551 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((":") reduce 282) ((";") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 21552 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((":") reduce 283) ((";") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 21553 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((":") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 21556 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 21557 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((":") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 21558 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((":") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 21564 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 21566 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 21567 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 21568 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 21569 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 21572 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 21573 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 21574 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((":") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 21575 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 21576 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 21579 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 21580 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((":") reduce 454) ((";") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 21591 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((":") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 21592 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((":") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 21663 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 444) ((";") reduce 444) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21976 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((":") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 21979 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((":") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 21981 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) ((";") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 21982 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((":") reduce 455) ((";") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 21983 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((":") reduce 457) ((";") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 21985 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) ((":") reduce 435) ((";") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 21986 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) ((":") reduce 437) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 21992 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 21994 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 21995 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) ((";") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 21996 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) ((";") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 21997 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) ((";") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 22000 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) ((";") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 22001 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 22002 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((":") reduce 514) ((";") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 22003 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) ((";") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 22005 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) ((":") reduce 427) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 22007 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((":") reduce 471) ((";") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 22009 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 358) ((";") reduce 358) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22010 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 432) ((";") reduce 432) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22021 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 22023 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((":") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 22025 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((":") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 22026 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((":") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 22030 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) ((":") reduce 463) ((";") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 22031 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((":") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 22032 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 22038 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) ((":") reduce 483) ((";") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 22040 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((":") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 22042 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 22043 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 22044 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 22045 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 22047 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 22054 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 22057 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((":") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 22059 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) ((";") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 22060 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) ((":") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 22067 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) ((";") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 22068 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) ((":") reduce 481) ((";") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 22070 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) ((":") reduce 480) ((";") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_DOUBLE_ARROW) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 22071 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) ((":") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_DOUBLE_ARROW) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 22072 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 430) ((";") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22075 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) ((";") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 22076 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) ((":") reduce 434) ((";") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 22077 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) ((":") reduce 416) ((";") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 22078 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) ((":") reduce 438) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 22079 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) ((":") reduce 482) ((";") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 22080 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) ((":") reduce 412) ((";") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 22081 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((":") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 22082 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 22083 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) ((";") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 22084 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) ((":") reduce 356) ((";") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 356) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 22087 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((":") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 22118 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 384) ((";") reduce 384) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 22119 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 387) ((";") reduce 387) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 22120 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) ((":") reduce 396) ((";") reduce 396) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 22121 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 409) ((";") reduce 409) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 22122 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) ((":") reduce 395) ((";") reduce 395) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 22123 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((":") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 22124 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 383) ((";") reduce 383) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22125 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 381) ((";") reduce 381) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 381) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22126 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 382) ((";") reduce 382) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 382) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22127 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 406) ((";") reduce 406) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 22128 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 402) ((";") reduce 402) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 22129 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 404) ((";") reduce 404) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 22130 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 401) ((";") reduce 401) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 22131 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 408) ((";") reduce 408) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 22132 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 403) ((";") reduce 403) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 22135 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 22136 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 22137 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 22138 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((":") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 22139 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 22140 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 22141 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 22142 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((":") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 22143 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 22144 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((":") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 22145 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 22152 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 22154 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 22155 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 22156 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 22157 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 22158 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 22161 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 22162 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 22164 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 22166 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 22167 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 22170 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 22171 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 415) ((";") reduce 415) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22172 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 379) ((";") reduce 379) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_DOUBLE_ARROW) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22173 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 380) ((";") reduce 380) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_DOUBLE_ARROW) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22174 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 385) ((";") reduce 385) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 22175 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 386) ((";") reduce 386) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 22179 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 413) ((";") reduce 413) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22180 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 414) ((";") reduce 414) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22181 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 407) ((";") reduce 407) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 22182 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 405) ((";") reduce 405) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 22183 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((":") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 22184 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) ((":") reduce 388) ((";") reduce 388) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_DOUBLE_ARROW) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 22185 ((("%") shift) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) ((":") reduce 390) ((";") reduce 390) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 22186 ((("%") shift) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) ((":") reduce 389) ((";") reduce 389) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 22187 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) ((":") reduce 391) ((";") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 22188 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) ((":") reduce 394) ((";") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 22190 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((":") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 22191 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((":") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 22192 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((":") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 22193 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((":") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 22194 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((":") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 22195 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((":") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 22196 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((":") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 22197 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((":") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 22198 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((":") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 22199 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((":") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 22200 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((":") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 22201 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((":") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 22202 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((":") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 22203 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((":") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 22204 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((":") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 22205 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((":") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 22206 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((":") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 22207 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((":") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 22208 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((":") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 22209 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((":") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 22210 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((":") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 22211 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((":") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 22212 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((":") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 22213 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((":") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 22214 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((":") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 22215 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((":") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 22216 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((":") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 22217 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((":") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 22218 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((":") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 22219 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((":") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 22220 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((":") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 22221 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((":") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 22222 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((":") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 22223 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((":") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 22224 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((":") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 22225 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((":") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 22226 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((":") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 22227 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((":") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 22228 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((":") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 22229 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((":") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 22230 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((":") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 22231 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((":") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 22232 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((":") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 22233 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((":") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 22234 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((":") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 22235 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((":") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 22236 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((":") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 22237 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((":") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 22238 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((":") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 22239 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((":") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 22240 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((":") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 22241 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((":") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 22242 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((":") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 22243 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((":") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 22244 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((":") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 22245 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((":") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 22246 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((":") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 22247 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((":") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 22248 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((":") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 22249 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((":") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 22250 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((":") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 22251 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((":") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 22252 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((":") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 22253 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((":") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 22254 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((":") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 22255 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((":") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 22256 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((":") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 22257 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) ((":") reduce 81) ((";") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_DOUBLE_ARROW) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 22258 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((":") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 22259 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((":") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 22260 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((":") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 22261 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((":") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 22262 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((":") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 22263 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((":") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 22264 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((":") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 22265 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((":") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 22266 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((":") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 22267 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((":") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 22268 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((":") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 22270 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((":") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 22271 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((":") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 22272 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) ((";") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 22274 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((":") reduce 456) ((";") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 22275 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((":") reduce 281) ((";") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 22279 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((":") reduce 282) ((";") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 22280 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((":") reduce 283) ((";") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 22285 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) ((":") reduce 377) ((";") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_DOUBLE_ARROW) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 22287 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) ((":") reduce 375) ((";") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_DOUBLE_ARROW) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 22297 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 371) ((";") reduce 371) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 371) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22298 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 373) ((";") reduce 373) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 373) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22299 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 372) ((";") reduce 372) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 372) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22300 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 365) ((";") reduce 365) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 365) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22301 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 362) ((";") reduce 362) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 362) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22302 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 370) ((";") reduce 370) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 370) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22303 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 364) ((";") reduce 364) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 364) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22304 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 368) ((";") reduce 368) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 368) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22305 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 363) ((";") reduce 363) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 363) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22306 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 366) ((";") reduce 366) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 366) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22307 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 367) ((";") reduce 367) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 367) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22308 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 374) ((";") reduce 374) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 374) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22309 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 369) ((";") reduce 369) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 369) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22311 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 359) ((";") reduce 359) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 359) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22313 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((":") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 22316 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 22318 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 22320 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 22321 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 22322 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) ((":") reduce 360) ((";") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_DOUBLE_ARROW) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 22326 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 22328 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((":") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 22329 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((":") reduce 281) ((";") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 22333 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((":") reduce 282) ((";") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 22334 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((":") reduce 283) ((";") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 22335 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((":") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 22338 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 22339 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((":") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 22340 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((":") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 22346 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 22348 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 22349 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 22350 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 22351 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 22354 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 22355 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 22356 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((":") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 22357 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 22358 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 22361 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 22362 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((":") reduce 454) ((";") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 22373 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((":") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 22374 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((":") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 22384 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 444) ((";") reduce 444) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22385 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((":") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 22387 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 431) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22389 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((":") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 22391 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) ((";") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 22392 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((":") reduce 455) ((";") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 22393 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((":") reduce 457) ((";") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 22395 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) ((":") reduce 435) ((";") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 22396 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) ((":") reduce 437) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 22402 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 22404 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 22405 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) ((";") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 22406 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) ((";") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 22407 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) ((";") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 22410 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) ((";") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 22411 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 22412 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((":") reduce 514) ((";") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 22413 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) ((";") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 22415 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) ((":") reduce 427) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 22417 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((":") reduce 471) ((";") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 22419 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 358) ((";") reduce 358) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 358) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22420 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 432) ((";") reduce 432) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22423 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 431) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22424 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((":") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 22425 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 433) ((";") reduce 433) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22426 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) ((":") reduce 419) ((";") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 22427 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) ((":") reduce 436) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 22429 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((":") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 22431 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) ((":") reduce 478) ((";") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 22432 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((":") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 22433 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 567) ((";") reduce 567) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22434 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 566) ((";") reduce 566) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22435 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 428) ((";") reduce 428) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22436 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) ((":") reduce 421) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 22439 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 22440 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 22441 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 22442 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((":") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 22443 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 22444 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) ((":") reduce 354) ((";") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_DOUBLE_ARROW) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 22446 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 22447 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((":") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 22448 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 22449 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((":") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 22450 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 22457 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 22459 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 22460 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 22461 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 22462 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 22464 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((":") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 22465 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) ((":") reduce 353) ((";") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_DOUBLE_ARROW) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 22466 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((":") reduce 281) ((";") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 22470 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((":") reduce 282) ((";") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 22471 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((":") reduce 283) ((";") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 22473 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 22474 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) ((":") reduce 355) ((";") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_DOUBLE_ARROW) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 22481 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) ((":") reduce 352) ((";") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_DOUBLE_ARROW) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 22483 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 22489 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) ((":") reduce 223) ((";") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_DOUBLE_ARROW) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 22494 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 357) ((";") reduce 357) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 357) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22498 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 22499 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) ((":") reduce 417) ((";") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 22500 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 564) ((";") reduce 564) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22501 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 563) ((";") reduce 563) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22502 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) ((":") reduce 376) ((";") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_DOUBLE_ARROW) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 22504 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((":") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 22506 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) ((":") reduce 464) ((";") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 22509 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 22512 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((":") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 22513 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) ((":") reduce 418) ((";") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 22514 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) ((":") reduce 378) ((";") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_DOUBLE_ARROW) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 22515 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) ((":") reduce 361) ((";") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_DOUBLE_ARROW) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 22516 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) ((":") reduce 422) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 22517 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) ((":") reduce 420) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 22520 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((":") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 22521 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((":") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 22522 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((":") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 22523 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) ((":") reduce 397) ((";") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 22525 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) ((":") reduce 411) ((";") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 22526 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) ((";") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 22529 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) ((";") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 22531 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((":") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 22532 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((":") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 22533 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((":") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 22534 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 433) ((";") reduce 433) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22535 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) ((":") reduce 419) ((";") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 22536 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) ((":") reduce 436) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 22538 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((":") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 22540 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) ((":") reduce 478) ((";") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 22541 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((":") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 22542 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 567) ((";") reduce 567) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22543 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 566) ((";") reduce 566) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22544 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 428) ((";") reduce 428) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22545 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) ((":") reduce 421) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 22548 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 22549 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 22550 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 22551 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((":") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 22552 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((":") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 22553 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) ((":") reduce 354) ((";") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 22555 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 22556 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((":") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 22557 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 22558 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((":") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 22559 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 22566 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 22568 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 22569 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 22570 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 22571 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 22573 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((":") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 22574 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) ((":") reduce 353) ((";") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 22575 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((":") reduce 281) ((";") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 22579 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((":") reduce 282) ((";") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 22580 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((":") reduce 283) ((";") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 22582 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 22583 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) ((":") reduce 355) ((";") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 22590 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) ((":") reduce 352) ((";") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 22592 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 22598 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) ((":") reduce 223) ((";") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 22603 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 357) ((";") reduce 357) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22607 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 22608 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) ((":") reduce 417) ((";") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 22609 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 564) ((";") reduce 564) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22610 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 563) ((";") reduce 563) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22611 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) ((":") reduce 376) ((";") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 22613 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((":") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 22615 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) ((":") reduce 464) ((";") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 22618 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 22621 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((":") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 22622 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) ((":") reduce 418) ((";") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 22623 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) ((":") reduce 378) ((";") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 22624 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) ((":") reduce 361) ((";") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 22625 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) ((":") reduce 422) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 22626 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) ((":") reduce 420) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 22629 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((":") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 22630 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((":") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 22631 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((":") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 22632 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) ((":") reduce 397) ((";") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 22634 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) ((":") reduce 411) ((";") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 22635 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) ((";") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 22638 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((":") reduce 516) ((";") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 22640 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((":") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 22641 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((":") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 22643 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 219) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 219) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDSWITCH) reduce 219) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 23258 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CASE) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DEFAULT) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("}") reduce 217) (("~") reduce 217)) 23259 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CASE) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DEFAULT) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 23261 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CASE) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DEFAULT) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 23263 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CASE) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DEFAULT) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 23265 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CASE) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DEFAULT) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 23267 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CASE) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DEFAULT) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 23277 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CASE) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DEFAULT) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 23278 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CASE) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DEFAULT) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_ELSE) reduce 210) ((T_ELSEIF) reduce 210) ((T_EMPTY) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("}") reduce 210) (("~") reduce 210)) 23281 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CASE) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DEFAULT) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_ELSE) reduce 211) ((T_ELSEIF) reduce 211) ((T_EMPTY) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("}") reduce 211) (("~") reduce 211)) 23282 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CASE) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DEFAULT) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 23291 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CASE) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DEFAULT) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 23292 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CASE) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DEFAULT) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_ELSE) reduce 208) ((T_ELSEIF) reduce 208) ((T_EMPTY) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("}") reduce 208) (("~") reduce 208)) 23295 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CASE) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DEFAULT) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_ELSE) reduce 209) ((T_ELSEIF) reduce 209) ((T_EMPTY) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("}") reduce 209) (("~") reduce 209)) 23297 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CASE) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DEFAULT) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 23303 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CASE) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DEFAULT) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 23308 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CASE) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DEFAULT) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 23309 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CASE) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DEFAULT) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_ELSE) reduce 212) ((T_ELSEIF) reduce 212) ((T_EMPTY) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("}") reduce 212) (("~") reduce 212)) 23312 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CASE) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DEFAULT) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_ELSE) reduce 213) ((T_ELSEIF) reduce 213) ((T_EMPTY) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("}") reduce 213) (("~") reduce 213)) 23314 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CASE) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DEFAULT) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 23316 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CASE) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DEFAULT) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 23317 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CASE) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DEFAULT) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_EMPTY) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("}") reduce 237) (("~") reduce 237)) 23328 ((("!") reduce 444) (("\"") reduce 444) (("$") reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) ((";") reduce 444) (("@") reduce 444) ((T_ABSTRACT) reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_BREAK) reduce 444) ((T_CASE) reduce 444) ((T_CLASS) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_CONTINUE) reduce 444) ((T_DEC) reduce 444) ((T_DECLARE) reduce 444) ((T_DEFAULT) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DO) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_ECHO) reduce 444) ((T_EMPTY) reduce 444) ((T_ENUM) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FINAL) reduce 444) ((T_FN) reduce 444) ((T_FOR) reduce 444) ((T_FOREACH) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_GLOBAL) reduce 444) ((T_GOTO) reduce 444) ((T_HALT_COMPILER) reduce 444) ((T_IF) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INLINE_HTML) reduce 444) ((T_INTERFACE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_RETURN) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_SWITCH) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT) reduce 444) ((T_TRAIT_C) reduce 444) ((T_TRY) reduce 444) ((T_UNSET) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_VARIABLE) reduce 444) ((T_WHILE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("{") reduce 444) (("}") reduce 444) (("~") reduce 444)) 23329 ((("!") reduce 179) (("\"") reduce 179) (("$") reduce 179) (("(") reduce 179) (("+") reduce 179) (("-") reduce 179) ((";") reduce 179) (("@") reduce 179) ((T_ABSTRACT) reduce 179) ((T_ARRAY) reduce 179) ((T_ARRAY_CAST) reduce 179) ((T_ATTRIBUTE) reduce 179) ((T_BOOL_CAST) reduce 179) ((T_BREAK) reduce 179) ((T_CASE) reduce 179) ((T_CLASS) reduce 179) ((T_CLASS_C) reduce 179) ((T_CLONE) reduce 179) ((T_CONSTANT_ENCAPSED_STRING) reduce 179) ((T_CONTINUE) reduce 179) ((T_DEC) reduce 179) ((T_DECLARE) reduce 179) ((T_DEFAULT) reduce 179) ((T_DIR) reduce 179) ((T_DNUMBER) reduce 179) ((T_DO) reduce 179) ((T_DOUBLE_CAST) reduce 179) ((T_ECHO) reduce 179) ((T_EMPTY) reduce 179) ((T_ENUM) reduce 179) ((T_EVAL) reduce 179) ((T_EXIT) reduce 179) ((T_FILE) reduce 179) ((T_FINAL) reduce 179) ((T_FN) reduce 179) ((T_FOR) reduce 179) ((T_FOREACH) reduce 179) ((T_FUNCTION) reduce 179) ((T_FUNC_C) reduce 179) ((T_GLOBAL) reduce 179) ((T_GOTO) reduce 179) ((T_HALT_COMPILER) reduce 179) ((T_IF) reduce 179) ((T_INC) reduce 179) ((T_INCLUDE) reduce 179) ((T_INCLUDE_ONCE) reduce 179) ((T_INLINE_HTML) reduce 179) ((T_INTERFACE) reduce 179) ((T_INT_CAST) reduce 179) ((T_ISSET) reduce 179) ((T_LINE) reduce 179) ((T_LIST) reduce 179) ((T_LNUMBER) reduce 179) ((T_MATCH) reduce 179) ((T_METHOD_C) reduce 179) ((T_NAME_FULLY_QUALIFIED) reduce 179) ((T_NAME_QUALIFIED) reduce 179) ((T_NAME_RELATIVE) reduce 179) ((T_NEW) reduce 179) ((T_NS_C) reduce 179) ((T_OBJECT_CAST) reduce 179) ((T_PRINT) reduce 179) ((T_REQUIRE) reduce 179) ((T_REQUIRE_ONCE) reduce 179) ((T_RETURN) reduce 179) ((T_START_HEREDOC) reduce 179) ((T_STATIC) reduce 179) ((T_STRING) reduce 179) ((T_STRING_CAST) reduce 179) ((T_SWITCH) reduce 179) ((T_THROW) reduce 179) ((T_TRAIT) reduce 179) ((T_TRAIT_C) reduce 179) ((T_TRY) reduce 179) ((T_UNSET) reduce 179) ((T_UNSET_CAST) reduce 179) ((T_VARIABLE) reduce 179) ((T_WHILE) reduce 179) ((T_YIELD) reduce 179) ((T_YIELD_FROM) reduce 179) (("[") reduce 179) (("`") reduce 179) (("{") reduce 179) (("}") reduce 179) (("~") reduce 179)) 23330 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CASE) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DEFAULT) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 23338 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CASE) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DEFAULT) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("}") reduce 184) (("~") reduce 184)) 23339 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CASE) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DEFAULT) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("}") reduce 143) (("~") reduce 143)) 23342 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CASE) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DEFAULT) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("}") reduce 240) (("~") reduce 240)) 23346 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CASE) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DEFAULT) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("}") reduce 241) (("~") reduce 241)) 23351 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CASE) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DEFAULT) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_EMPTY) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("}") reduce 232) (("~") reduce 232)) 23352 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CASE) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DEFAULT) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 23355 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CASE) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DEFAULT) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_EMPTY) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("}") reduce 233) (("~") reduce 233)) 23360 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CASE) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DEFAULT) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 23363 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CASE) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DEFAULT) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 23364 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CASE) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DEFAULT) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_EMPTY) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("}") reduce 174) (("~") reduce 174)) 23367 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CASE) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DEFAULT) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 23370 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CASE) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DEFAULT) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("}") reduce 175) (("~") reduce 175)) 23377 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CASE) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DEFAULT) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 23382 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CASE) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DEFAULT) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("}") reduce 190) (("~") reduce 190)) 23387 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CASE) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DEFAULT) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 23391 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CASE) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DEFAULT) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_EMPTY) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("}") reduce 214) (("~") reduce 214)) 23393 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CASE) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DEFAULT) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_EMPTY) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("}") reduce 215) (("~") reduce 215)) 23397 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CASE) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DEFAULT) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("}") reduce 216) (("~") reduce 216)) 23400 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CASE) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DEFAULT) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("}") reduce 217) (("~") reduce 217)) 23401 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CASE) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DEFAULT) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 23403 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CASE) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DEFAULT) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 23405 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CASE) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DEFAULT) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 23411 ((("!") reduce 191) (("\"") reduce 191) (("$") reduce 191) (("(") reduce 191) (("+") reduce 191) (("-") reduce 191) ((";") reduce 191) (("@") reduce 191) ((T_ABSTRACT) reduce 191) ((T_ARRAY) reduce 191) ((T_ARRAY_CAST) reduce 191) ((T_ATTRIBUTE) reduce 191) ((T_BOOL_CAST) reduce 191) ((T_BREAK) reduce 191) ((T_CASE) reduce 191) ((T_CLASS) reduce 191) ((T_CLASS_C) reduce 191) ((T_CLONE) reduce 191) ((T_CONSTANT_ENCAPSED_STRING) reduce 191) ((T_CONTINUE) reduce 191) ((T_DEC) reduce 191) ((T_DECLARE) reduce 191) ((T_DEFAULT) reduce 191) ((T_DIR) reduce 191) ((T_DNUMBER) reduce 191) ((T_DO) reduce 191) ((T_DOUBLE_CAST) reduce 191) ((T_ECHO) reduce 191) ((T_EMPTY) reduce 191) ((T_ENUM) reduce 191) ((T_EVAL) reduce 191) ((T_EXIT) reduce 191) ((T_FILE) reduce 191) ((T_FINAL) reduce 191) ((T_FN) reduce 191) ((T_FOR) reduce 191) ((T_FOREACH) reduce 191) ((T_FUNCTION) reduce 191) ((T_FUNC_C) reduce 191) ((T_GLOBAL) reduce 191) ((T_GOTO) reduce 191) ((T_HALT_COMPILER) reduce 191) ((T_IF) reduce 191) ((T_INC) reduce 191) ((T_INCLUDE) reduce 191) ((T_INCLUDE_ONCE) reduce 191) ((T_INLINE_HTML) reduce 191) ((T_INTERFACE) reduce 191) ((T_INT_CAST) reduce 191) ((T_ISSET) reduce 191) ((T_LINE) reduce 191) ((T_LIST) reduce 191) ((T_LNUMBER) reduce 191) ((T_MATCH) reduce 191) ((T_METHOD_C) reduce 191) ((T_NAME_FULLY_QUALIFIED) reduce 191) ((T_NAME_QUALIFIED) reduce 191) ((T_NAME_RELATIVE) reduce 191) ((T_NEW) reduce 191) ((T_NS_C) reduce 191) ((T_OBJECT_CAST) reduce 191) ((T_PRINT) reduce 191) ((T_REQUIRE) reduce 191) ((T_REQUIRE_ONCE) reduce 191) ((T_RETURN) reduce 191) ((T_START_HEREDOC) reduce 191) ((T_STATIC) reduce 191) ((T_STRING) reduce 191) ((T_STRING_CAST) reduce 191) ((T_SWITCH) reduce 191) ((T_THROW) reduce 191) ((T_TRAIT) reduce 191) ((T_TRAIT_C) reduce 191) ((T_TRY) reduce 191) ((T_UNSET) reduce 191) ((T_UNSET_CAST) reduce 191) ((T_VARIABLE) reduce 191) ((T_WHILE) reduce 191) ((T_YIELD) reduce 191) ((T_YIELD_FROM) reduce 191) (("[") reduce 191) (("`") reduce 191) (("{") reduce 191) (("}") reduce 191) (("~") reduce 191)) 23415 ((("!") reduce 234) (("\"") reduce 234) (("$") reduce 234) (("(") reduce 234) (("+") reduce 234) (("-") reduce 234) ((";") reduce 234) (("@") reduce 234) ((T_ABSTRACT) reduce 234) ((T_ARRAY) reduce 234) ((T_ARRAY_CAST) reduce 234) ((T_ATTRIBUTE) reduce 234) ((T_BOOL_CAST) reduce 234) ((T_BREAK) reduce 234) ((T_CASE) reduce 234) ((T_CLASS) reduce 234) ((T_CLASS_C) reduce 234) ((T_CLONE) reduce 234) ((T_CONSTANT_ENCAPSED_STRING) reduce 234) ((T_CONTINUE) reduce 234) ((T_DEC) reduce 234) ((T_DECLARE) reduce 234) ((T_DEFAULT) reduce 234) ((T_DIR) reduce 234) ((T_DNUMBER) reduce 234) ((T_DO) reduce 234) ((T_DOUBLE_CAST) reduce 234) ((T_ECHO) reduce 234) ((T_ELSE) reduce 234) ((T_ELSEIF) reduce 234) ((T_EMPTY) reduce 234) ((T_ENUM) reduce 234) ((T_EVAL) reduce 234) ((T_EXIT) reduce 234) ((T_FILE) reduce 234) ((T_FINAL) reduce 234) ((T_FN) reduce 234) ((T_FOR) reduce 234) ((T_FOREACH) reduce 234) ((T_FUNCTION) reduce 234) ((T_FUNC_C) reduce 234) ((T_GLOBAL) reduce 234) ((T_GOTO) reduce 234) ((T_HALT_COMPILER) reduce 234) ((T_IF) reduce 234) ((T_INC) reduce 234) ((T_INCLUDE) reduce 234) ((T_INCLUDE_ONCE) reduce 234) ((T_INLINE_HTML) reduce 234) ((T_INTERFACE) reduce 234) ((T_INT_CAST) reduce 234) ((T_ISSET) reduce 234) ((T_LINE) reduce 234) ((T_LIST) reduce 234) ((T_LNUMBER) reduce 234) ((T_MATCH) reduce 234) ((T_METHOD_C) reduce 234) ((T_NAME_FULLY_QUALIFIED) reduce 234) ((T_NAME_QUALIFIED) reduce 234) ((T_NAME_RELATIVE) reduce 234) ((T_NEW) reduce 234) ((T_NS_C) reduce 234) ((T_OBJECT_CAST) reduce 234) ((T_PRINT) reduce 234) ((T_REQUIRE) reduce 234) ((T_REQUIRE_ONCE) reduce 234) ((T_RETURN) reduce 234) ((T_START_HEREDOC) reduce 234) ((T_STATIC) reduce 234) ((T_STRING) reduce 234) ((T_STRING_CAST) reduce 234) ((T_SWITCH) reduce 234) ((T_THROW) reduce 234) ((T_TRAIT) reduce 234) ((T_TRAIT_C) reduce 234) ((T_TRY) reduce 234) ((T_UNSET) reduce 234) ((T_UNSET_CAST) reduce 234) ((T_VARIABLE) reduce 234) ((T_WHILE) reduce 234) ((T_YIELD) reduce 234) ((T_YIELD_FROM) reduce 234) (("[") reduce 234) (("`") reduce 234) (("{") reduce 234) (("}") reduce 234) (("~") reduce 234)) 23418 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CASE) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DEFAULT) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("}") reduce 144) (("~") reduce 144)) 23420 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CASE) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DEFAULT) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 23422 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CASE) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DEFAULT) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 23432 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CASE) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DEFAULT) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 23433 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CASE) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DEFAULT) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_EMPTY) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("}") reduce 210) (("~") reduce 210)) 23436 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CASE) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DEFAULT) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_EMPTY) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("}") reduce 211) (("~") reduce 211)) 23437 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CASE) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DEFAULT) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 23446 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CASE) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DEFAULT) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 23447 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CASE) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DEFAULT) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_EMPTY) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("}") reduce 208) (("~") reduce 208)) 23450 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CASE) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DEFAULT) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_EMPTY) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("}") reduce 209) (("~") reduce 209)) 23457 ((("!") reduce 192) (("\"") reduce 192) (("$") reduce 192) (("(") reduce 192) (("+") reduce 192) (("-") reduce 192) ((";") reduce 192) (("@") reduce 192) ((T_ABSTRACT) reduce 192) ((T_ARRAY) reduce 192) ((T_ARRAY_CAST) reduce 192) ((T_ATTRIBUTE) reduce 192) ((T_BOOL_CAST) reduce 192) ((T_BREAK) reduce 192) ((T_CASE) reduce 192) ((T_CLASS) reduce 192) ((T_CLASS_C) reduce 192) ((T_CLONE) reduce 192) ((T_CONSTANT_ENCAPSED_STRING) reduce 192) ((T_CONTINUE) reduce 192) ((T_DEC) reduce 192) ((T_DECLARE) reduce 192) ((T_DEFAULT) reduce 192) ((T_DIR) reduce 192) ((T_DNUMBER) reduce 192) ((T_DO) reduce 192) ((T_DOUBLE_CAST) reduce 192) ((T_ECHO) reduce 192) ((T_EMPTY) reduce 192) ((T_ENUM) reduce 192) ((T_EVAL) reduce 192) ((T_EXIT) reduce 192) ((T_FILE) reduce 192) ((T_FINAL) reduce 192) ((T_FN) reduce 192) ((T_FOR) reduce 192) ((T_FOREACH) reduce 192) ((T_FUNCTION) reduce 192) ((T_FUNC_C) reduce 192) ((T_GLOBAL) reduce 192) ((T_GOTO) reduce 192) ((T_HALT_COMPILER) reduce 192) ((T_IF) reduce 192) ((T_INC) reduce 192) ((T_INCLUDE) reduce 192) ((T_INCLUDE_ONCE) reduce 192) ((T_INLINE_HTML) reduce 192) ((T_INTERFACE) reduce 192) ((T_INT_CAST) reduce 192) ((T_ISSET) reduce 192) ((T_LINE) reduce 192) ((T_LIST) reduce 192) ((T_LNUMBER) reduce 192) ((T_MATCH) reduce 192) ((T_METHOD_C) reduce 192) ((T_NAME_FULLY_QUALIFIED) reduce 192) ((T_NAME_QUALIFIED) reduce 192) ((T_NAME_RELATIVE) reduce 192) ((T_NEW) reduce 192) ((T_NS_C) reduce 192) ((T_OBJECT_CAST) reduce 192) ((T_PRINT) reduce 192) ((T_REQUIRE) reduce 192) ((T_REQUIRE_ONCE) reduce 192) ((T_RETURN) reduce 192) ((T_START_HEREDOC) reduce 192) ((T_STATIC) reduce 192) ((T_STRING) reduce 192) ((T_STRING_CAST) reduce 192) ((T_SWITCH) reduce 192) ((T_THROW) reduce 192) ((T_TRAIT) reduce 192) ((T_TRAIT_C) reduce 192) ((T_TRY) reduce 192) ((T_UNSET) reduce 192) ((T_UNSET_CAST) reduce 192) ((T_VARIABLE) reduce 192) ((T_WHILE) reduce 192) ((T_YIELD) reduce 192) ((T_YIELD_FROM) reduce 192) (("[") reduce 192) (("`") reduce 192) (("{") reduce 192) (("}") reduce 192) (("~") reduce 192)) 23459 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CASE) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DEFAULT) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 23465 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CASE) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DEFAULT) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 23470 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CASE) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DEFAULT) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 23471 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CASE) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DEFAULT) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_EMPTY) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("}") reduce 212) (("~") reduce 212)) 23474 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CASE) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DEFAULT) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_EMPTY) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("}") reduce 213) (("~") reduce 213)) 23476 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CASE) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DEFAULT) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 23483 ((("!") reduce 185) (("\"") reduce 185) (("$") reduce 185) (("(") reduce 185) (("+") reduce 185) (("-") reduce 185) ((";") reduce 185) (("@") reduce 185) ((T_ABSTRACT) reduce 185) ((T_ARRAY) reduce 185) ((T_ARRAY_CAST) reduce 185) ((T_ATTRIBUTE) reduce 185) ((T_BOOL_CAST) reduce 185) ((T_BREAK) reduce 185) ((T_CASE) reduce 185) ((T_CLASS) reduce 185) ((T_CLASS_C) reduce 185) ((T_CLONE) reduce 185) ((T_CONSTANT_ENCAPSED_STRING) reduce 185) ((T_CONTINUE) reduce 185) ((T_DEC) reduce 185) ((T_DECLARE) reduce 185) ((T_DEFAULT) reduce 185) ((T_DIR) reduce 185) ((T_DNUMBER) reduce 185) ((T_DO) reduce 185) ((T_DOUBLE_CAST) reduce 185) ((T_ECHO) reduce 185) ((T_EMPTY) reduce 185) ((T_ENUM) reduce 185) ((T_EVAL) reduce 185) ((T_EXIT) reduce 185) ((T_FILE) reduce 185) ((T_FINAL) reduce 185) ((T_FN) reduce 185) ((T_FOR) reduce 185) ((T_FOREACH) reduce 185) ((T_FUNCTION) reduce 185) ((T_FUNC_C) reduce 185) ((T_GLOBAL) reduce 185) ((T_GOTO) reduce 185) ((T_HALT_COMPILER) reduce 185) ((T_IF) reduce 185) ((T_INC) reduce 185) ((T_INCLUDE) reduce 185) ((T_INCLUDE_ONCE) reduce 185) ((T_INLINE_HTML) reduce 185) ((T_INTERFACE) reduce 185) ((T_INT_CAST) reduce 185) ((T_ISSET) reduce 185) ((T_LINE) reduce 185) ((T_LIST) reduce 185) ((T_LNUMBER) reduce 185) ((T_MATCH) reduce 185) ((T_METHOD_C) reduce 185) ((T_NAME_FULLY_QUALIFIED) reduce 185) ((T_NAME_QUALIFIED) reduce 185) ((T_NAME_RELATIVE) reduce 185) ((T_NEW) reduce 185) ((T_NS_C) reduce 185) ((T_OBJECT_CAST) reduce 185) ((T_PRINT) reduce 185) ((T_REQUIRE) reduce 185) ((T_REQUIRE_ONCE) reduce 185) ((T_RETURN) reduce 185) ((T_START_HEREDOC) reduce 185) ((T_STATIC) reduce 185) ((T_STRING) reduce 185) ((T_STRING_CAST) reduce 185) ((T_SWITCH) reduce 185) ((T_THROW) reduce 185) ((T_TRAIT) reduce 185) ((T_TRAIT_C) reduce 185) ((T_TRY) reduce 185) ((T_UNSET) reduce 185) ((T_UNSET_CAST) reduce 185) ((T_VARIABLE) reduce 185) ((T_WHILE) reduce 185) ((T_YIELD) reduce 185) ((T_YIELD_FROM) reduce 185) (("[") reduce 185) (("`") reduce 185) (("{") reduce 185) (("}") reduce 185) (("~") reduce 185)) 23485 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CASE) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DEFAULT) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 23488 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 219) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 219) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") reduce 219) (("~") shift)) 23490 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_ELSE) reduce 215) ((T_ELSEIF) reduce 215) ((T_EMPTY) reduce 215) ((T_ENDWHILE) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 23494 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDWHILE) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 23497 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDWHILE) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 23498 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDWHILE) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 23500 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDWHILE) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 23502 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDWHILE) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 23504 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDWHILE) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 23506 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDWHILE) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 23516 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDWHILE) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 23517 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_ELSE) reduce 210) ((T_ELSEIF) reduce 210) ((T_EMPTY) reduce 210) ((T_ENDWHILE) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 23520 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_ELSE) reduce 211) ((T_ELSEIF) reduce 211) ((T_EMPTY) reduce 211) ((T_ENDWHILE) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 23521 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDWHILE) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 23530 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDWHILE) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 23531 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_ELSE) reduce 208) ((T_ELSEIF) reduce 208) ((T_EMPTY) reduce 208) ((T_ENDWHILE) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 23534 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_ELSE) reduce 209) ((T_ELSEIF) reduce 209) ((T_EMPTY) reduce 209) ((T_ENDWHILE) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 23536 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDWHILE) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 23542 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDWHILE) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 23547 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDWHILE) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 23548 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_ELSE) reduce 212) ((T_ELSEIF) reduce 212) ((T_EMPTY) reduce 212) ((T_ENDWHILE) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 23551 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_ELSE) reduce 213) ((T_ELSEIF) reduce 213) ((T_EMPTY) reduce 213) ((T_ENDWHILE) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 23553 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDWHILE) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 23555 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDWHILE) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 23556 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_EMPTY) reduce 237) ((T_ENDWHILE) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 23567 ((("!") reduce 444) (("\"") reduce 444) (("$") reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) ((";") reduce 444) (("@") reduce 444) ((T_ABSTRACT) reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_BREAK) reduce 444) ((T_CLASS) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_CONTINUE) reduce 444) ((T_DEC) reduce 444) ((T_DECLARE) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DO) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_ECHO) reduce 444) ((T_EMPTY) reduce 444) ((T_ENDWHILE) reduce 444) ((T_ENUM) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FINAL) reduce 444) ((T_FN) reduce 444) ((T_FOR) reduce 444) ((T_FOREACH) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_GLOBAL) reduce 444) ((T_GOTO) reduce 444) ((T_HALT_COMPILER) reduce 444) ((T_IF) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INLINE_HTML) reduce 444) ((T_INTERFACE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_RETURN) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_SWITCH) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT) reduce 444) ((T_TRAIT_C) reduce 444) ((T_TRY) reduce 444) ((T_UNSET) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_VARIABLE) reduce 444) ((T_WHILE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("{") reduce 444) (("~") reduce 444)) 23568 ((("!") reduce 179) (("\"") reduce 179) (("$") reduce 179) (("(") reduce 179) (("+") reduce 179) (("-") reduce 179) ((";") reduce 179) (("@") reduce 179) ((T_ABSTRACT) reduce 179) ((T_ARRAY) reduce 179) ((T_ARRAY_CAST) reduce 179) ((T_ATTRIBUTE) reduce 179) ((T_BOOL_CAST) reduce 179) ((T_BREAK) reduce 179) ((T_CLASS) reduce 179) ((T_CLASS_C) reduce 179) ((T_CLONE) reduce 179) ((T_CONSTANT_ENCAPSED_STRING) reduce 179) ((T_CONTINUE) reduce 179) ((T_DEC) reduce 179) ((T_DECLARE) reduce 179) ((T_DIR) reduce 179) ((T_DNUMBER) reduce 179) ((T_DO) reduce 179) ((T_DOUBLE_CAST) reduce 179) ((T_ECHO) reduce 179) ((T_EMPTY) reduce 179) ((T_ENDWHILE) reduce 179) ((T_ENUM) reduce 179) ((T_EVAL) reduce 179) ((T_EXIT) reduce 179) ((T_FILE) reduce 179) ((T_FINAL) reduce 179) ((T_FN) reduce 179) ((T_FOR) reduce 179) ((T_FOREACH) reduce 179) ((T_FUNCTION) reduce 179) ((T_FUNC_C) reduce 179) ((T_GLOBAL) reduce 179) ((T_GOTO) reduce 179) ((T_HALT_COMPILER) reduce 179) ((T_IF) reduce 179) ((T_INC) reduce 179) ((T_INCLUDE) reduce 179) ((T_INCLUDE_ONCE) reduce 179) ((T_INLINE_HTML) reduce 179) ((T_INTERFACE) reduce 179) ((T_INT_CAST) reduce 179) ((T_ISSET) reduce 179) ((T_LINE) reduce 179) ((T_LIST) reduce 179) ((T_LNUMBER) reduce 179) ((T_MATCH) reduce 179) ((T_METHOD_C) reduce 179) ((T_NAME_FULLY_QUALIFIED) reduce 179) ((T_NAME_QUALIFIED) reduce 179) ((T_NAME_RELATIVE) reduce 179) ((T_NEW) reduce 179) ((T_NS_C) reduce 179) ((T_OBJECT_CAST) reduce 179) ((T_PRINT) reduce 179) ((T_REQUIRE) reduce 179) ((T_REQUIRE_ONCE) reduce 179) ((T_RETURN) reduce 179) ((T_START_HEREDOC) reduce 179) ((T_STATIC) reduce 179) ((T_STRING) reduce 179) ((T_STRING_CAST) reduce 179) ((T_SWITCH) reduce 179) ((T_THROW) reduce 179) ((T_TRAIT) reduce 179) ((T_TRAIT_C) reduce 179) ((T_TRY) reduce 179) ((T_UNSET) reduce 179) ((T_UNSET_CAST) reduce 179) ((T_VARIABLE) reduce 179) ((T_WHILE) reduce 179) ((T_YIELD) reduce 179) ((T_YIELD_FROM) reduce 179) (("[") reduce 179) (("`") reduce 179) (("{") reduce 179) (("~") reduce 179)) 23569 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDWHILE) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 23577 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDWHILE) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 23578 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDWHILE) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 23581 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENDWHILE) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 23585 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDWHILE) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 23590 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_EMPTY) reduce 232) ((T_ENDWHILE) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 23591 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDWHILE) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 23594 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_EMPTY) reduce 233) ((T_ENDWHILE) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 23599 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDWHILE) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 23602 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDWHILE) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 23603 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_EMPTY) reduce 174) ((T_ENDWHILE) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 23606 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDWHILE) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 23609 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDWHILE) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 23616 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDWHILE) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 23621 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDWHILE) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 23626 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDWHILE) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 23630 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_EMPTY) reduce 214) ((T_ENDWHILE) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 23632 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_EMPTY) reduce 215) ((T_ENDWHILE) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 23636 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDWHILE) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 23639 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDWHILE) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 23640 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDWHILE) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 23642 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDWHILE) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 23644 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDWHILE) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 23650 ((("!") reduce 191) (("\"") reduce 191) (("$") reduce 191) (("(") reduce 191) (("+") reduce 191) (("-") reduce 191) ((";") reduce 191) (("@") reduce 191) ((T_ABSTRACT) reduce 191) ((T_ARRAY) reduce 191) ((T_ARRAY_CAST) reduce 191) ((T_ATTRIBUTE) reduce 191) ((T_BOOL_CAST) reduce 191) ((T_BREAK) reduce 191) ((T_CLASS) reduce 191) ((T_CLASS_C) reduce 191) ((T_CLONE) reduce 191) ((T_CONSTANT_ENCAPSED_STRING) reduce 191) ((T_CONTINUE) reduce 191) ((T_DEC) reduce 191) ((T_DECLARE) reduce 191) ((T_DIR) reduce 191) ((T_DNUMBER) reduce 191) ((T_DO) reduce 191) ((T_DOUBLE_CAST) reduce 191) ((T_ECHO) reduce 191) ((T_EMPTY) reduce 191) ((T_ENDWHILE) reduce 191) ((T_ENUM) reduce 191) ((T_EVAL) reduce 191) ((T_EXIT) reduce 191) ((T_FILE) reduce 191) ((T_FINAL) reduce 191) ((T_FN) reduce 191) ((T_FOR) reduce 191) ((T_FOREACH) reduce 191) ((T_FUNCTION) reduce 191) ((T_FUNC_C) reduce 191) ((T_GLOBAL) reduce 191) ((T_GOTO) reduce 191) ((T_HALT_COMPILER) reduce 191) ((T_IF) reduce 191) ((T_INC) reduce 191) ((T_INCLUDE) reduce 191) ((T_INCLUDE_ONCE) reduce 191) ((T_INLINE_HTML) reduce 191) ((T_INTERFACE) reduce 191) ((T_INT_CAST) reduce 191) ((T_ISSET) reduce 191) ((T_LINE) reduce 191) ((T_LIST) reduce 191) ((T_LNUMBER) reduce 191) ((T_MATCH) reduce 191) ((T_METHOD_C) reduce 191) ((T_NAME_FULLY_QUALIFIED) reduce 191) ((T_NAME_QUALIFIED) reduce 191) ((T_NAME_RELATIVE) reduce 191) ((T_NEW) reduce 191) ((T_NS_C) reduce 191) ((T_OBJECT_CAST) reduce 191) ((T_PRINT) reduce 191) ((T_REQUIRE) reduce 191) ((T_REQUIRE_ONCE) reduce 191) ((T_RETURN) reduce 191) ((T_START_HEREDOC) reduce 191) ((T_STATIC) reduce 191) ((T_STRING) reduce 191) ((T_STRING_CAST) reduce 191) ((T_SWITCH) reduce 191) ((T_THROW) reduce 191) ((T_TRAIT) reduce 191) ((T_TRAIT_C) reduce 191) ((T_TRY) reduce 191) ((T_UNSET) reduce 191) ((T_UNSET_CAST) reduce 191) ((T_VARIABLE) reduce 191) ((T_WHILE) reduce 191) ((T_YIELD) reduce 191) ((T_YIELD_FROM) reduce 191) (("[") reduce 191) (("`") reduce 191) (("{") reduce 191) (("~") reduce 191)) 23654 ((("!") reduce 234) (("\"") reduce 234) (("$") reduce 234) (("(") reduce 234) (("+") reduce 234) (("-") reduce 234) ((";") reduce 234) (("@") reduce 234) ((T_ABSTRACT) reduce 234) ((T_ARRAY) reduce 234) ((T_ARRAY_CAST) reduce 234) ((T_ATTRIBUTE) reduce 234) ((T_BOOL_CAST) reduce 234) ((T_BREAK) reduce 234) ((T_CLASS) reduce 234) ((T_CLASS_C) reduce 234) ((T_CLONE) reduce 234) ((T_CONSTANT_ENCAPSED_STRING) reduce 234) ((T_CONTINUE) reduce 234) ((T_DEC) reduce 234) ((T_DECLARE) reduce 234) ((T_DIR) reduce 234) ((T_DNUMBER) reduce 234) ((T_DO) reduce 234) ((T_DOUBLE_CAST) reduce 234) ((T_ECHO) reduce 234) ((T_ELSE) reduce 234) ((T_ELSEIF) reduce 234) ((T_EMPTY) reduce 234) ((T_ENDWHILE) reduce 234) ((T_ENUM) reduce 234) ((T_EVAL) reduce 234) ((T_EXIT) reduce 234) ((T_FILE) reduce 234) ((T_FINAL) reduce 234) ((T_FN) reduce 234) ((T_FOR) reduce 234) ((T_FOREACH) reduce 234) ((T_FUNCTION) reduce 234) ((T_FUNC_C) reduce 234) ((T_GLOBAL) reduce 234) ((T_GOTO) reduce 234) ((T_HALT_COMPILER) reduce 234) ((T_IF) reduce 234) ((T_INC) reduce 234) ((T_INCLUDE) reduce 234) ((T_INCLUDE_ONCE) reduce 234) ((T_INLINE_HTML) reduce 234) ((T_INTERFACE) reduce 234) ((T_INT_CAST) reduce 234) ((T_ISSET) reduce 234) ((T_LINE) reduce 234) ((T_LIST) reduce 234) ((T_LNUMBER) reduce 234) ((T_MATCH) reduce 234) ((T_METHOD_C) reduce 234) ((T_NAME_FULLY_QUALIFIED) reduce 234) ((T_NAME_QUALIFIED) reduce 234) ((T_NAME_RELATIVE) reduce 234) ((T_NEW) reduce 234) ((T_NS_C) reduce 234) ((T_OBJECT_CAST) reduce 234) ((T_PRINT) reduce 234) ((T_REQUIRE) reduce 234) ((T_REQUIRE_ONCE) reduce 234) ((T_RETURN) reduce 234) ((T_START_HEREDOC) reduce 234) ((T_STATIC) reduce 234) ((T_STRING) reduce 234) ((T_STRING_CAST) reduce 234) ((T_SWITCH) reduce 234) ((T_THROW) reduce 234) ((T_TRAIT) reduce 234) ((T_TRAIT_C) reduce 234) ((T_TRY) reduce 234) ((T_UNSET) reduce 234) ((T_UNSET_CAST) reduce 234) ((T_VARIABLE) reduce 234) ((T_WHILE) reduce 234) ((T_YIELD) reduce 234) ((T_YIELD_FROM) reduce 234) (("[") reduce 234) (("`") reduce 234) (("{") reduce 234) (("~") reduce 234)) 23657 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDWHILE) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 23659 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDWHILE) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 23661 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDWHILE) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 23671 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDWHILE) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 23672 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_EMPTY) reduce 210) ((T_ENDWHILE) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 23675 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_EMPTY) reduce 211) ((T_ENDWHILE) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 23676 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDWHILE) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 23685 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDWHILE) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 23686 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_EMPTY) reduce 208) ((T_ENDWHILE) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 23689 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_EMPTY) reduce 209) ((T_ENDWHILE) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 23696 ((("!") reduce 192) (("\"") reduce 192) (("$") reduce 192) (("(") reduce 192) (("+") reduce 192) (("-") reduce 192) ((";") reduce 192) (("@") reduce 192) ((T_ABSTRACT) reduce 192) ((T_ARRAY) reduce 192) ((T_ARRAY_CAST) reduce 192) ((T_ATTRIBUTE) reduce 192) ((T_BOOL_CAST) reduce 192) ((T_BREAK) reduce 192) ((T_CLASS) reduce 192) ((T_CLASS_C) reduce 192) ((T_CLONE) reduce 192) ((T_CONSTANT_ENCAPSED_STRING) reduce 192) ((T_CONTINUE) reduce 192) ((T_DEC) reduce 192) ((T_DECLARE) reduce 192) ((T_DIR) reduce 192) ((T_DNUMBER) reduce 192) ((T_DO) reduce 192) ((T_DOUBLE_CAST) reduce 192) ((T_ECHO) reduce 192) ((T_EMPTY) reduce 192) ((T_ENDWHILE) reduce 192) ((T_ENUM) reduce 192) ((T_EVAL) reduce 192) ((T_EXIT) reduce 192) ((T_FILE) reduce 192) ((T_FINAL) reduce 192) ((T_FN) reduce 192) ((T_FOR) reduce 192) ((T_FOREACH) reduce 192) ((T_FUNCTION) reduce 192) ((T_FUNC_C) reduce 192) ((T_GLOBAL) reduce 192) ((T_GOTO) reduce 192) ((T_HALT_COMPILER) reduce 192) ((T_IF) reduce 192) ((T_INC) reduce 192) ((T_INCLUDE) reduce 192) ((T_INCLUDE_ONCE) reduce 192) ((T_INLINE_HTML) reduce 192) ((T_INTERFACE) reduce 192) ((T_INT_CAST) reduce 192) ((T_ISSET) reduce 192) ((T_LINE) reduce 192) ((T_LIST) reduce 192) ((T_LNUMBER) reduce 192) ((T_MATCH) reduce 192) ((T_METHOD_C) reduce 192) ((T_NAME_FULLY_QUALIFIED) reduce 192) ((T_NAME_QUALIFIED) reduce 192) ((T_NAME_RELATIVE) reduce 192) ((T_NEW) reduce 192) ((T_NS_C) reduce 192) ((T_OBJECT_CAST) reduce 192) ((T_PRINT) reduce 192) ((T_REQUIRE) reduce 192) ((T_REQUIRE_ONCE) reduce 192) ((T_RETURN) reduce 192) ((T_START_HEREDOC) reduce 192) ((T_STATIC) reduce 192) ((T_STRING) reduce 192) ((T_STRING_CAST) reduce 192) ((T_SWITCH) reduce 192) ((T_THROW) reduce 192) ((T_TRAIT) reduce 192) ((T_TRAIT_C) reduce 192) ((T_TRY) reduce 192) ((T_UNSET) reduce 192) ((T_UNSET_CAST) reduce 192) ((T_VARIABLE) reduce 192) ((T_WHILE) reduce 192) ((T_YIELD) reduce 192) ((T_YIELD_FROM) reduce 192) (("[") reduce 192) (("`") reduce 192) (("{") reduce 192) (("~") reduce 192)) 23697 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_EMPTY) reduce 233) ((T_ENDIF) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 23699 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDWHILE) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 23705 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDWHILE) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 23710 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDWHILE) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 23711 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_EMPTY) reduce 212) ((T_ENDWHILE) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 23714 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_EMPTY) reduce 213) ((T_ENDWHILE) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 23716 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDWHILE) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 23723 ((("!") reduce 185) (("\"") reduce 185) (("$") reduce 185) (("(") reduce 185) (("+") reduce 185) (("-") reduce 185) ((";") reduce 185) (("@") reduce 185) ((T_ABSTRACT) reduce 185) ((T_ARRAY) reduce 185) ((T_ARRAY_CAST) reduce 185) ((T_ATTRIBUTE) reduce 185) ((T_BOOL_CAST) reduce 185) ((T_BREAK) reduce 185) ((T_CLASS) reduce 185) ((T_CLASS_C) reduce 185) ((T_CLONE) reduce 185) ((T_CONSTANT_ENCAPSED_STRING) reduce 185) ((T_CONTINUE) reduce 185) ((T_DEC) reduce 185) ((T_DECLARE) reduce 185) ((T_DIR) reduce 185) ((T_DNUMBER) reduce 185) ((T_DO) reduce 185) ((T_DOUBLE_CAST) reduce 185) ((T_ECHO) reduce 185) ((T_EMPTY) reduce 185) ((T_ENDWHILE) reduce 185) ((T_ENUM) reduce 185) ((T_EVAL) reduce 185) ((T_EXIT) reduce 185) ((T_FILE) reduce 185) ((T_FINAL) reduce 185) ((T_FN) reduce 185) ((T_FOR) reduce 185) ((T_FOREACH) reduce 185) ((T_FUNCTION) reduce 185) ((T_FUNC_C) reduce 185) ((T_GLOBAL) reduce 185) ((T_GOTO) reduce 185) ((T_HALT_COMPILER) reduce 185) ((T_IF) reduce 185) ((T_INC) reduce 185) ((T_INCLUDE) reduce 185) ((T_INCLUDE_ONCE) reduce 185) ((T_INLINE_HTML) reduce 185) ((T_INTERFACE) reduce 185) ((T_INT_CAST) reduce 185) ((T_ISSET) reduce 185) ((T_LINE) reduce 185) ((T_LIST) reduce 185) ((T_LNUMBER) reduce 185) ((T_MATCH) reduce 185) ((T_METHOD_C) reduce 185) ((T_NAME_FULLY_QUALIFIED) reduce 185) ((T_NAME_QUALIFIED) reduce 185) ((T_NAME_RELATIVE) reduce 185) ((T_NEW) reduce 185) ((T_NS_C) reduce 185) ((T_OBJECT_CAST) reduce 185) ((T_PRINT) reduce 185) ((T_REQUIRE) reduce 185) ((T_REQUIRE_ONCE) reduce 185) ((T_RETURN) reduce 185) ((T_START_HEREDOC) reduce 185) ((T_STATIC) reduce 185) ((T_STRING) reduce 185) ((T_STRING_CAST) reduce 185) ((T_SWITCH) reduce 185) ((T_THROW) reduce 185) ((T_TRAIT) reduce 185) ((T_TRAIT_C) reduce 185) ((T_TRY) reduce 185) ((T_UNSET) reduce 185) ((T_UNSET_CAST) reduce 185) ((T_VARIABLE) reduce 185) ((T_WHILE) reduce 185) ((T_YIELD) reduce 185) ((T_YIELD_FROM) reduce 185) (("[") reduce 185) (("`") reduce 185) (("{") reduce 185) (("~") reduce 185)) 23725 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDWHILE) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 23730 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDIF) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 23733 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDIF) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 23734 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_EMPTY) reduce 174) ((T_ENDIF) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 23737 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDIF) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 23740 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDIF) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 23747 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDIF) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 23752 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDIF) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 23757 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDIF) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 23761 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_EMPTY) reduce 214) ((T_ENDIF) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 23763 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_EMPTY) reduce 215) ((T_ENDIF) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 23767 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDIF) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 23770 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDIF) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 23771 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDIF) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 23773 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDIF) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 23775 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDIF) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 23781 ((("!") reduce 191) (("\"") reduce 191) (("$") reduce 191) (("(") reduce 191) (("+") reduce 191) (("-") reduce 191) ((";") reduce 191) (("@") reduce 191) ((T_ABSTRACT) reduce 191) ((T_ARRAY) reduce 191) ((T_ARRAY_CAST) reduce 191) ((T_ATTRIBUTE) reduce 191) ((T_BOOL_CAST) reduce 191) ((T_BREAK) reduce 191) ((T_CLASS) reduce 191) ((T_CLASS_C) reduce 191) ((T_CLONE) reduce 191) ((T_CONSTANT_ENCAPSED_STRING) reduce 191) ((T_CONTINUE) reduce 191) ((T_DEC) reduce 191) ((T_DECLARE) reduce 191) ((T_DIR) reduce 191) ((T_DNUMBER) reduce 191) ((T_DO) reduce 191) ((T_DOUBLE_CAST) reduce 191) ((T_ECHO) reduce 191) ((T_EMPTY) reduce 191) ((T_ENDIF) reduce 191) ((T_ENUM) reduce 191) ((T_EVAL) reduce 191) ((T_EXIT) reduce 191) ((T_FILE) reduce 191) ((T_FINAL) reduce 191) ((T_FN) reduce 191) ((T_FOR) reduce 191) ((T_FOREACH) reduce 191) ((T_FUNCTION) reduce 191) ((T_FUNC_C) reduce 191) ((T_GLOBAL) reduce 191) ((T_GOTO) reduce 191) ((T_HALT_COMPILER) reduce 191) ((T_IF) reduce 191) ((T_INC) reduce 191) ((T_INCLUDE) reduce 191) ((T_INCLUDE_ONCE) reduce 191) ((T_INLINE_HTML) reduce 191) ((T_INTERFACE) reduce 191) ((T_INT_CAST) reduce 191) ((T_ISSET) reduce 191) ((T_LINE) reduce 191) ((T_LIST) reduce 191) ((T_LNUMBER) reduce 191) ((T_MATCH) reduce 191) ((T_METHOD_C) reduce 191) ((T_NAME_FULLY_QUALIFIED) reduce 191) ((T_NAME_QUALIFIED) reduce 191) ((T_NAME_RELATIVE) reduce 191) ((T_NEW) reduce 191) ((T_NS_C) reduce 191) ((T_OBJECT_CAST) reduce 191) ((T_PRINT) reduce 191) ((T_REQUIRE) reduce 191) ((T_REQUIRE_ONCE) reduce 191) ((T_RETURN) reduce 191) ((T_START_HEREDOC) reduce 191) ((T_STATIC) reduce 191) ((T_STRING) reduce 191) ((T_STRING_CAST) reduce 191) ((T_SWITCH) reduce 191) ((T_THROW) reduce 191) ((T_TRAIT) reduce 191) ((T_TRAIT_C) reduce 191) ((T_TRY) reduce 191) ((T_UNSET) reduce 191) ((T_UNSET_CAST) reduce 191) ((T_VARIABLE) reduce 191) ((T_WHILE) reduce 191) ((T_YIELD) reduce 191) ((T_YIELD_FROM) reduce 191) (("[") reduce 191) (("`") reduce 191) (("{") reduce 191) (("~") reduce 191)) 23784 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDIF) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 23786 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDIF) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 23788 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDIF) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 23798 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDIF) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 23799 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_EMPTY) reduce 210) ((T_ENDIF) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 23802 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_EMPTY) reduce 211) ((T_ENDIF) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 23803 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDIF) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 23812 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDIF) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 23813 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_EMPTY) reduce 208) ((T_ENDIF) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 23816 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_EMPTY) reduce 209) ((T_ENDIF) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 23823 ((("!") reduce 192) (("\"") reduce 192) (("$") reduce 192) (("(") reduce 192) (("+") reduce 192) (("-") reduce 192) ((";") reduce 192) (("@") reduce 192) ((T_ABSTRACT) reduce 192) ((T_ARRAY) reduce 192) ((T_ARRAY_CAST) reduce 192) ((T_ATTRIBUTE) reduce 192) ((T_BOOL_CAST) reduce 192) ((T_BREAK) reduce 192) ((T_CLASS) reduce 192) ((T_CLASS_C) reduce 192) ((T_CLONE) reduce 192) ((T_CONSTANT_ENCAPSED_STRING) reduce 192) ((T_CONTINUE) reduce 192) ((T_DEC) reduce 192) ((T_DECLARE) reduce 192) ((T_DIR) reduce 192) ((T_DNUMBER) reduce 192) ((T_DO) reduce 192) ((T_DOUBLE_CAST) reduce 192) ((T_ECHO) reduce 192) ((T_EMPTY) reduce 192) ((T_ENDIF) reduce 192) ((T_ENUM) reduce 192) ((T_EVAL) reduce 192) ((T_EXIT) reduce 192) ((T_FILE) reduce 192) ((T_FINAL) reduce 192) ((T_FN) reduce 192) ((T_FOR) reduce 192) ((T_FOREACH) reduce 192) ((T_FUNCTION) reduce 192) ((T_FUNC_C) reduce 192) ((T_GLOBAL) reduce 192) ((T_GOTO) reduce 192) ((T_HALT_COMPILER) reduce 192) ((T_IF) reduce 192) ((T_INC) reduce 192) ((T_INCLUDE) reduce 192) ((T_INCLUDE_ONCE) reduce 192) ((T_INLINE_HTML) reduce 192) ((T_INTERFACE) reduce 192) ((T_INT_CAST) reduce 192) ((T_ISSET) reduce 192) ((T_LINE) reduce 192) ((T_LIST) reduce 192) ((T_LNUMBER) reduce 192) ((T_MATCH) reduce 192) ((T_METHOD_C) reduce 192) ((T_NAME_FULLY_QUALIFIED) reduce 192) ((T_NAME_QUALIFIED) reduce 192) ((T_NAME_RELATIVE) reduce 192) ((T_NEW) reduce 192) ((T_NS_C) reduce 192) ((T_OBJECT_CAST) reduce 192) ((T_PRINT) reduce 192) ((T_REQUIRE) reduce 192) ((T_REQUIRE_ONCE) reduce 192) ((T_RETURN) reduce 192) ((T_START_HEREDOC) reduce 192) ((T_STATIC) reduce 192) ((T_STRING) reduce 192) ((T_STRING_CAST) reduce 192) ((T_SWITCH) reduce 192) ((T_THROW) reduce 192) ((T_TRAIT) reduce 192) ((T_TRAIT_C) reduce 192) ((T_TRY) reduce 192) ((T_UNSET) reduce 192) ((T_UNSET_CAST) reduce 192) ((T_VARIABLE) reduce 192) ((T_WHILE) reduce 192) ((T_YIELD) reduce 192) ((T_YIELD_FROM) reduce 192) (("[") reduce 192) (("`") reduce 192) (("{") reduce 192) (("~") reduce 192)) 23824 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENDIF) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 23826 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDIF) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 23832 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDIF) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 23837 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDIF) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 23838 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_EMPTY) reduce 212) ((T_ENDIF) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 23841 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_EMPTY) reduce 213) ((T_ENDIF) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 23843 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDIF) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 23850 ((("!") reduce 185) (("\"") reduce 185) (("$") reduce 185) (("(") reduce 185) (("+") reduce 185) (("-") reduce 185) ((";") reduce 185) (("@") reduce 185) ((T_ABSTRACT) reduce 185) ((T_ARRAY) reduce 185) ((T_ARRAY_CAST) reduce 185) ((T_ATTRIBUTE) reduce 185) ((T_BOOL_CAST) reduce 185) ((T_BREAK) reduce 185) ((T_CLASS) reduce 185) ((T_CLASS_C) reduce 185) ((T_CLONE) reduce 185) ((T_CONSTANT_ENCAPSED_STRING) reduce 185) ((T_CONTINUE) reduce 185) ((T_DEC) reduce 185) ((T_DECLARE) reduce 185) ((T_DIR) reduce 185) ((T_DNUMBER) reduce 185) ((T_DO) reduce 185) ((T_DOUBLE_CAST) reduce 185) ((T_ECHO) reduce 185) ((T_EMPTY) reduce 185) ((T_ENDIF) reduce 185) ((T_ENUM) reduce 185) ((T_EVAL) reduce 185) ((T_EXIT) reduce 185) ((T_FILE) reduce 185) ((T_FINAL) reduce 185) ((T_FN) reduce 185) ((T_FOR) reduce 185) ((T_FOREACH) reduce 185) ((T_FUNCTION) reduce 185) ((T_FUNC_C) reduce 185) ((T_GLOBAL) reduce 185) ((T_GOTO) reduce 185) ((T_HALT_COMPILER) reduce 185) ((T_IF) reduce 185) ((T_INC) reduce 185) ((T_INCLUDE) reduce 185) ((T_INCLUDE_ONCE) reduce 185) ((T_INLINE_HTML) reduce 185) ((T_INTERFACE) reduce 185) ((T_INT_CAST) reduce 185) ((T_ISSET) reduce 185) ((T_LINE) reduce 185) ((T_LIST) reduce 185) ((T_LNUMBER) reduce 185) ((T_MATCH) reduce 185) ((T_METHOD_C) reduce 185) ((T_NAME_FULLY_QUALIFIED) reduce 185) ((T_NAME_QUALIFIED) reduce 185) ((T_NAME_RELATIVE) reduce 185) ((T_NEW) reduce 185) ((T_NS_C) reduce 185) ((T_OBJECT_CAST) reduce 185) ((T_PRINT) reduce 185) ((T_REQUIRE) reduce 185) ((T_REQUIRE_ONCE) reduce 185) ((T_RETURN) reduce 185) ((T_START_HEREDOC) reduce 185) ((T_STATIC) reduce 185) ((T_STRING) reduce 185) ((T_STRING_CAST) reduce 185) ((T_SWITCH) reduce 185) ((T_THROW) reduce 185) ((T_TRAIT) reduce 185) ((T_TRAIT_C) reduce 185) ((T_TRY) reduce 185) ((T_UNSET) reduce 185) ((T_UNSET_CAST) reduce 185) ((T_VARIABLE) reduce 185) ((T_WHILE) reduce 185) ((T_YIELD) reduce 185) ((T_YIELD_FROM) reduce 185) (("[") reduce 185) (("`") reduce 185) (("{") reduce 185) (("~") reduce 185)) 23852 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDIF) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 23857 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_ELSE) reduce 232) ((T_ELSEIF) reduce 232) ((T_EMPTY) reduce 232) ((T_ENDIF) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 23858 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDIF) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 23861 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_ELSE) reduce 233) ((T_ELSEIF) reduce 233) ((T_EMPTY) reduce 233) ((T_ENDIF) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 23866 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDIF) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 23869 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDIF) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 23870 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_ELSE) reduce 174) ((T_ELSEIF) reduce 174) ((T_EMPTY) reduce 174) ((T_ENDIF) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 23873 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDIF) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 23876 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENDIF) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 23883 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDIF) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 23888 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_ELSE) reduce 190) ((T_ELSEIF) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDIF) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 23893 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDIF) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 23897 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_ELSE) reduce 214) ((T_ELSEIF) reduce 214) ((T_EMPTY) reduce 214) ((T_ENDIF) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 23899 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_ELSE) reduce 215) ((T_ELSEIF) reduce 215) ((T_EMPTY) reduce 215) ((T_ENDIF) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 23903 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENDIF) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 23906 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENDIF) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 23907 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDIF) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 23909 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDIF) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 23911 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDIF) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 23917 ((("!") reduce 191) (("\"") reduce 191) (("$") reduce 191) (("(") reduce 191) (("+") reduce 191) (("-") reduce 191) ((";") reduce 191) (("@") reduce 191) ((T_ABSTRACT) reduce 191) ((T_ARRAY) reduce 191) ((T_ARRAY_CAST) reduce 191) ((T_ATTRIBUTE) reduce 191) ((T_BOOL_CAST) reduce 191) ((T_BREAK) reduce 191) ((T_CLASS) reduce 191) ((T_CLASS_C) reduce 191) ((T_CLONE) reduce 191) ((T_CONSTANT_ENCAPSED_STRING) reduce 191) ((T_CONTINUE) reduce 191) ((T_DEC) reduce 191) ((T_DECLARE) reduce 191) ((T_DIR) reduce 191) ((T_DNUMBER) reduce 191) ((T_DO) reduce 191) ((T_DOUBLE_CAST) reduce 191) ((T_ECHO) reduce 191) ((T_ELSE) reduce 191) ((T_ELSEIF) reduce 191) ((T_EMPTY) reduce 191) ((T_ENDIF) reduce 191) ((T_ENUM) reduce 191) ((T_EVAL) reduce 191) ((T_EXIT) reduce 191) ((T_FILE) reduce 191) ((T_FINAL) reduce 191) ((T_FN) reduce 191) ((T_FOR) reduce 191) ((T_FOREACH) reduce 191) ((T_FUNCTION) reduce 191) ((T_FUNC_C) reduce 191) ((T_GLOBAL) reduce 191) ((T_GOTO) reduce 191) ((T_HALT_COMPILER) reduce 191) ((T_IF) reduce 191) ((T_INC) reduce 191) ((T_INCLUDE) reduce 191) ((T_INCLUDE_ONCE) reduce 191) ((T_INLINE_HTML) reduce 191) ((T_INTERFACE) reduce 191) ((T_INT_CAST) reduce 191) ((T_ISSET) reduce 191) ((T_LINE) reduce 191) ((T_LIST) reduce 191) ((T_LNUMBER) reduce 191) ((T_MATCH) reduce 191) ((T_METHOD_C) reduce 191) ((T_NAME_FULLY_QUALIFIED) reduce 191) ((T_NAME_QUALIFIED) reduce 191) ((T_NAME_RELATIVE) reduce 191) ((T_NEW) reduce 191) ((T_NS_C) reduce 191) ((T_OBJECT_CAST) reduce 191) ((T_PRINT) reduce 191) ((T_REQUIRE) reduce 191) ((T_REQUIRE_ONCE) reduce 191) ((T_RETURN) reduce 191) ((T_START_HEREDOC) reduce 191) ((T_STATIC) reduce 191) ((T_STRING) reduce 191) ((T_STRING_CAST) reduce 191) ((T_SWITCH) reduce 191) ((T_THROW) reduce 191) ((T_TRAIT) reduce 191) ((T_TRAIT_C) reduce 191) ((T_TRY) reduce 191) ((T_UNSET) reduce 191) ((T_UNSET_CAST) reduce 191) ((T_VARIABLE) reduce 191) ((T_WHILE) reduce 191) ((T_YIELD) reduce 191) ((T_YIELD_FROM) reduce 191) (("[") reduce 191) (("`") reduce 191) (("{") reduce 191) (("~") reduce 191)) 23921 ((("!") reduce 234) (("\"") reduce 234) (("$") reduce 234) (("(") reduce 234) (("+") reduce 234) (("-") reduce 234) ((";") reduce 234) (("@") reduce 234) ((T_ABSTRACT) reduce 234) ((T_ARRAY) reduce 234) ((T_ARRAY_CAST) reduce 234) ((T_ATTRIBUTE) reduce 234) ((T_BOOL_CAST) reduce 234) ((T_BREAK) reduce 234) ((T_CLASS) reduce 234) ((T_CLASS_C) reduce 234) ((T_CLONE) reduce 234) ((T_CONSTANT_ENCAPSED_STRING) reduce 234) ((T_CONTINUE) reduce 234) ((T_DEC) reduce 234) ((T_DECLARE) reduce 234) ((T_DIR) reduce 234) ((T_DNUMBER) reduce 234) ((T_DO) reduce 234) ((T_DOUBLE_CAST) reduce 234) ((T_ECHO) reduce 234) ((T_ELSE) reduce 234) ((T_ELSEIF) reduce 234) ((T_EMPTY) reduce 234) ((T_ENDIF) reduce 234) ((T_ENUM) reduce 234) ((T_EVAL) reduce 234) ((T_EXIT) reduce 234) ((T_FILE) reduce 234) ((T_FINAL) reduce 234) ((T_FN) reduce 234) ((T_FOR) reduce 234) ((T_FOREACH) reduce 234) ((T_FUNCTION) reduce 234) ((T_FUNC_C) reduce 234) ((T_GLOBAL) reduce 234) ((T_GOTO) reduce 234) ((T_HALT_COMPILER) reduce 234) ((T_IF) reduce 234) ((T_INC) reduce 234) ((T_INCLUDE) reduce 234) ((T_INCLUDE_ONCE) reduce 234) ((T_INLINE_HTML) reduce 234) ((T_INTERFACE) reduce 234) ((T_INT_CAST) reduce 234) ((T_ISSET) reduce 234) ((T_LINE) reduce 234) ((T_LIST) reduce 234) ((T_LNUMBER) reduce 234) ((T_MATCH) reduce 234) ((T_METHOD_C) reduce 234) ((T_NAME_FULLY_QUALIFIED) reduce 234) ((T_NAME_QUALIFIED) reduce 234) ((T_NAME_RELATIVE) reduce 234) ((T_NEW) reduce 234) ((T_NS_C) reduce 234) ((T_OBJECT_CAST) reduce 234) ((T_PRINT) reduce 234) ((T_REQUIRE) reduce 234) ((T_REQUIRE_ONCE) reduce 234) ((T_RETURN) reduce 234) ((T_START_HEREDOC) reduce 234) ((T_STATIC) reduce 234) ((T_STRING) reduce 234) ((T_STRING_CAST) reduce 234) ((T_SWITCH) reduce 234) ((T_THROW) reduce 234) ((T_TRAIT) reduce 234) ((T_TRAIT_C) reduce 234) ((T_TRY) reduce 234) ((T_UNSET) reduce 234) ((T_UNSET_CAST) reduce 234) ((T_VARIABLE) reduce 234) ((T_WHILE) reduce 234) ((T_YIELD) reduce 234) ((T_YIELD_FROM) reduce 234) (("[") reduce 234) (("`") reduce 234) (("{") reduce 234) (("~") reduce 234)) 23924 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_ELSE) reduce 144) ((T_ELSEIF) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDIF) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 23926 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDIF) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 23928 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDIF) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 23938 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDIF) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 23939 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_ELSE) reduce 210) ((T_ELSEIF) reduce 210) ((T_EMPTY) reduce 210) ((T_ENDIF) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 23942 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_ELSE) reduce 211) ((T_ELSEIF) reduce 211) ((T_EMPTY) reduce 211) ((T_ENDIF) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 23943 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDIF) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 23952 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDIF) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 23953 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_ELSE) reduce 208) ((T_ELSEIF) reduce 208) ((T_EMPTY) reduce 208) ((T_ENDIF) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 23956 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_ELSE) reduce 209) ((T_ELSEIF) reduce 209) ((T_EMPTY) reduce 209) ((T_ENDIF) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 23963 ((("!") reduce 192) (("\"") reduce 192) (("$") reduce 192) (("(") reduce 192) (("+") reduce 192) (("-") reduce 192) ((";") reduce 192) (("@") reduce 192) ((T_ABSTRACT) reduce 192) ((T_ARRAY) reduce 192) ((T_ARRAY_CAST) reduce 192) ((T_ATTRIBUTE) reduce 192) ((T_BOOL_CAST) reduce 192) ((T_BREAK) reduce 192) ((T_CLASS) reduce 192) ((T_CLASS_C) reduce 192) ((T_CLONE) reduce 192) ((T_CONSTANT_ENCAPSED_STRING) reduce 192) ((T_CONTINUE) reduce 192) ((T_DEC) reduce 192) ((T_DECLARE) reduce 192) ((T_DIR) reduce 192) ((T_DNUMBER) reduce 192) ((T_DO) reduce 192) ((T_DOUBLE_CAST) reduce 192) ((T_ECHO) reduce 192) ((T_ELSE) reduce 192) ((T_ELSEIF) reduce 192) ((T_EMPTY) reduce 192) ((T_ENDIF) reduce 192) ((T_ENUM) reduce 192) ((T_EVAL) reduce 192) ((T_EXIT) reduce 192) ((T_FILE) reduce 192) ((T_FINAL) reduce 192) ((T_FN) reduce 192) ((T_FOR) reduce 192) ((T_FOREACH) reduce 192) ((T_FUNCTION) reduce 192) ((T_FUNC_C) reduce 192) ((T_GLOBAL) reduce 192) ((T_GOTO) reduce 192) ((T_HALT_COMPILER) reduce 192) ((T_IF) reduce 192) ((T_INC) reduce 192) ((T_INCLUDE) reduce 192) ((T_INCLUDE_ONCE) reduce 192) ((T_INLINE_HTML) reduce 192) ((T_INTERFACE) reduce 192) ((T_INT_CAST) reduce 192) ((T_ISSET) reduce 192) ((T_LINE) reduce 192) ((T_LIST) reduce 192) ((T_LNUMBER) reduce 192) ((T_MATCH) reduce 192) ((T_METHOD_C) reduce 192) ((T_NAME_FULLY_QUALIFIED) reduce 192) ((T_NAME_QUALIFIED) reduce 192) ((T_NAME_RELATIVE) reduce 192) ((T_NEW) reduce 192) ((T_NS_C) reduce 192) ((T_OBJECT_CAST) reduce 192) ((T_PRINT) reduce 192) ((T_REQUIRE) reduce 192) ((T_REQUIRE_ONCE) reduce 192) ((T_RETURN) reduce 192) ((T_START_HEREDOC) reduce 192) ((T_STATIC) reduce 192) ((T_STRING) reduce 192) ((T_STRING_CAST) reduce 192) ((T_SWITCH) reduce 192) ((T_THROW) reduce 192) ((T_TRAIT) reduce 192) ((T_TRAIT_C) reduce 192) ((T_TRY) reduce 192) ((T_UNSET) reduce 192) ((T_UNSET_CAST) reduce 192) ((T_VARIABLE) reduce 192) ((T_WHILE) reduce 192) ((T_YIELD) reduce 192) ((T_YIELD_FROM) reduce 192) (("[") reduce 192) (("`") reduce 192) (("{") reduce 192) (("~") reduce 192)) 23965 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDIF) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 23971 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDIF) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 23976 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDIF) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 23977 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_ELSE) reduce 212) ((T_ELSEIF) reduce 212) ((T_EMPTY) reduce 212) ((T_ENDIF) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 23980 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_ELSE) reduce 213) ((T_ELSEIF) reduce 213) ((T_EMPTY) reduce 213) ((T_ENDIF) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 23982 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDIF) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 23989 ((("!") reduce 185) (("\"") reduce 185) (("$") reduce 185) (("(") reduce 185) (("+") reduce 185) (("-") reduce 185) ((";") reduce 185) (("@") reduce 185) ((T_ABSTRACT) reduce 185) ((T_ARRAY) reduce 185) ((T_ARRAY_CAST) reduce 185) ((T_ATTRIBUTE) reduce 185) ((T_BOOL_CAST) reduce 185) ((T_BREAK) reduce 185) ((T_CLASS) reduce 185) ((T_CLASS_C) reduce 185) ((T_CLONE) reduce 185) ((T_CONSTANT_ENCAPSED_STRING) reduce 185) ((T_CONTINUE) reduce 185) ((T_DEC) reduce 185) ((T_DECLARE) reduce 185) ((T_DIR) reduce 185) ((T_DNUMBER) reduce 185) ((T_DO) reduce 185) ((T_DOUBLE_CAST) reduce 185) ((T_ECHO) reduce 185) ((T_ELSE) reduce 185) ((T_ELSEIF) reduce 185) ((T_EMPTY) reduce 185) ((T_ENDIF) reduce 185) ((T_ENUM) reduce 185) ((T_EVAL) reduce 185) ((T_EXIT) reduce 185) ((T_FILE) reduce 185) ((T_FINAL) reduce 185) ((T_FN) reduce 185) ((T_FOR) reduce 185) ((T_FOREACH) reduce 185) ((T_FUNCTION) reduce 185) ((T_FUNC_C) reduce 185) ((T_GLOBAL) reduce 185) ((T_GOTO) reduce 185) ((T_HALT_COMPILER) reduce 185) ((T_IF) reduce 185) ((T_INC) reduce 185) ((T_INCLUDE) reduce 185) ((T_INCLUDE_ONCE) reduce 185) ((T_INLINE_HTML) reduce 185) ((T_INTERFACE) reduce 185) ((T_INT_CAST) reduce 185) ((T_ISSET) reduce 185) ((T_LINE) reduce 185) ((T_LIST) reduce 185) ((T_LNUMBER) reduce 185) ((T_MATCH) reduce 185) ((T_METHOD_C) reduce 185) ((T_NAME_FULLY_QUALIFIED) reduce 185) ((T_NAME_QUALIFIED) reduce 185) ((T_NAME_RELATIVE) reduce 185) ((T_NEW) reduce 185) ((T_NS_C) reduce 185) ((T_OBJECT_CAST) reduce 185) ((T_PRINT) reduce 185) ((T_REQUIRE) reduce 185) ((T_REQUIRE_ONCE) reduce 185) ((T_RETURN) reduce 185) ((T_START_HEREDOC) reduce 185) ((T_STATIC) reduce 185) ((T_STRING) reduce 185) ((T_STRING_CAST) reduce 185) ((T_SWITCH) reduce 185) ((T_THROW) reduce 185) ((T_TRAIT) reduce 185) ((T_TRAIT_C) reduce 185) ((T_TRY) reduce 185) ((T_UNSET) reduce 185) ((T_UNSET_CAST) reduce 185) ((T_VARIABLE) reduce 185) ((T_WHILE) reduce 185) ((T_YIELD) reduce 185) ((T_YIELD_FROM) reduce 185) (("[") reduce 185) (("`") reduce 185) (("{") reduce 185) (("~") reduce 185)) 23991 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDIF) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 23995 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("}") reduce 241) (("~") reduce 241)) 24000 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_ELSE) reduce 232) ((T_ELSEIF) reduce 232) ((T_EMPTY) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("}") reduce 232) (("~") reduce 232)) 24001 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 24004 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_ELSE) reduce 233) ((T_ELSEIF) reduce 233) ((T_EMPTY) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("}") reduce 233) (("~") reduce 233)) 24009 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 24012 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 24013 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_ELSE) reduce 174) ((T_ELSEIF) reduce 174) ((T_EMPTY) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("}") reduce 174) (("~") reduce 174)) 24016 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 24019 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("}") reduce 175) (("~") reduce 175)) 24026 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 24031 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 24035 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_ELSE) reduce 214) ((T_ELSEIF) reduce 214) ((T_EMPTY) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("}") reduce 214) (("~") reduce 214)) 24037 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_ELSE) reduce 215) ((T_ELSEIF) reduce 215) ((T_EMPTY) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("}") reduce 215) (("~") reduce 215)) 24041 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("}") reduce 216) (("~") reduce 216)) 24044 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("}") reduce 217) (("~") reduce 217)) 24045 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 24047 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 24049 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 24051 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 24053 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 24063 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 24064 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_ELSE) reduce 210) ((T_ELSEIF) reduce 210) ((T_EMPTY) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("}") reduce 210) (("~") reduce 210)) 24067 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_ELSE) reduce 211) ((T_ELSEIF) reduce 211) ((T_EMPTY) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("}") reduce 211) (("~") reduce 211)) 24068 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 24077 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 24078 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_ELSE) reduce 208) ((T_ELSEIF) reduce 208) ((T_EMPTY) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("}") reduce 208) (("~") reduce 208)) 24081 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_ELSE) reduce 209) ((T_ELSEIF) reduce 209) ((T_EMPTY) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("}") reduce 209) (("~") reduce 209)) 24083 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 24089 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 24094 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 24095 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_ELSE) reduce 212) ((T_ELSEIF) reduce 212) ((T_EMPTY) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("}") reduce 212) (("~") reduce 212)) 24098 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_ELSE) reduce 213) ((T_ELSEIF) reduce 213) ((T_EMPTY) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("}") reduce 213) (("~") reduce 213)) 24100 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 24102 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 24715 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_EMPTY) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("}") reduce 237) (("~") reduce 237)) 24726 ((("!") reduce 444) (("\"") reduce 444) (("$") reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) ((";") reduce 444) (("@") reduce 444) ((T_ABSTRACT) reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_BREAK) reduce 444) ((T_CLASS) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_CONTINUE) reduce 444) ((T_DEC) reduce 444) ((T_DECLARE) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DO) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_ECHO) reduce 444) ((T_EMPTY) reduce 444) ((T_ENUM) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FINAL) reduce 444) ((T_FN) reduce 444) ((T_FOR) reduce 444) ((T_FOREACH) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_GLOBAL) reduce 444) ((T_GOTO) reduce 444) ((T_HALT_COMPILER) reduce 444) ((T_IF) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INLINE_HTML) reduce 444) ((T_INTERFACE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_RETURN) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_SWITCH) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT) reduce 444) ((T_TRAIT_C) reduce 444) ((T_TRY) reduce 444) ((T_UNSET) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_VARIABLE) reduce 444) ((T_WHILE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("{") reduce 444) (("}") reduce 444) (("~") reduce 444)) 24727 ((("!") reduce 179) (("\"") reduce 179) (("$") reduce 179) (("(") reduce 179) (("+") reduce 179) (("-") reduce 179) ((";") reduce 179) (("@") reduce 179) ((T_ABSTRACT) reduce 179) ((T_ARRAY) reduce 179) ((T_ARRAY_CAST) reduce 179) ((T_ATTRIBUTE) reduce 179) ((T_BOOL_CAST) reduce 179) ((T_BREAK) reduce 179) ((T_CLASS) reduce 179) ((T_CLASS_C) reduce 179) ((T_CLONE) reduce 179) ((T_CONSTANT_ENCAPSED_STRING) reduce 179) ((T_CONTINUE) reduce 179) ((T_DEC) reduce 179) ((T_DECLARE) reduce 179) ((T_DIR) reduce 179) ((T_DNUMBER) reduce 179) ((T_DO) reduce 179) ((T_DOUBLE_CAST) reduce 179) ((T_ECHO) reduce 179) ((T_EMPTY) reduce 179) ((T_ENUM) reduce 179) ((T_EVAL) reduce 179) ((T_EXIT) reduce 179) ((T_FILE) reduce 179) ((T_FINAL) reduce 179) ((T_FN) reduce 179) ((T_FOR) reduce 179) ((T_FOREACH) reduce 179) ((T_FUNCTION) reduce 179) ((T_FUNC_C) reduce 179) ((T_GLOBAL) reduce 179) ((T_GOTO) reduce 179) ((T_HALT_COMPILER) reduce 179) ((T_IF) reduce 179) ((T_INC) reduce 179) ((T_INCLUDE) reduce 179) ((T_INCLUDE_ONCE) reduce 179) ((T_INLINE_HTML) reduce 179) ((T_INTERFACE) reduce 179) ((T_INT_CAST) reduce 179) ((T_ISSET) reduce 179) ((T_LINE) reduce 179) ((T_LIST) reduce 179) ((T_LNUMBER) reduce 179) ((T_MATCH) reduce 179) ((T_METHOD_C) reduce 179) ((T_NAME_FULLY_QUALIFIED) reduce 179) ((T_NAME_QUALIFIED) reduce 179) ((T_NAME_RELATIVE) reduce 179) ((T_NEW) reduce 179) ((T_NS_C) reduce 179) ((T_OBJECT_CAST) reduce 179) ((T_PRINT) reduce 179) ((T_REQUIRE) reduce 179) ((T_REQUIRE_ONCE) reduce 179) ((T_RETURN) reduce 179) ((T_START_HEREDOC) reduce 179) ((T_STATIC) reduce 179) ((T_STRING) reduce 179) ((T_STRING_CAST) reduce 179) ((T_SWITCH) reduce 179) ((T_THROW) reduce 179) ((T_TRAIT) reduce 179) ((T_TRAIT_C) reduce 179) ((T_TRY) reduce 179) ((T_UNSET) reduce 179) ((T_UNSET_CAST) reduce 179) ((T_VARIABLE) reduce 179) ((T_WHILE) reduce 179) ((T_YIELD) reduce 179) ((T_YIELD_FROM) reduce 179) (("[") reduce 179) (("`") reduce 179) (("{") reduce 179) (("}") reduce 179) (("~") reduce 179)) 24728 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 24736 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("}") reduce 184) (("~") reduce 184)) 24737 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("}") reduce 143) (("~") reduce 143)) 24740 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("}") reduce 240) (("~") reduce 240)) 24744 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("}") reduce 241) (("~") reduce 241)) 24749 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_EMPTY) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("}") reduce 232) (("~") reduce 232)) 24750 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 24753 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_EMPTY) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("}") reduce 233) (("~") reduce 233)) 24758 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 24761 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 24762 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_EMPTY) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("}") reduce 174) (("~") reduce 174)) 24765 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 24768 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("}") reduce 175) (("~") reduce 175)) 24775 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 24780 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("}") reduce 190) (("~") reduce 190)) 24785 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 24789 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_EMPTY) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("}") reduce 214) (("~") reduce 214)) 24791 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_EMPTY) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("}") reduce 215) (("~") reduce 215)) 24795 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("}") reduce 216) (("~") reduce 216)) 24798 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("}") reduce 217) (("~") reduce 217)) 24799 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 24801 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 24803 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 24809 ((("!") reduce 191) (("\"") reduce 191) (("$") reduce 191) (("(") reduce 191) (("+") reduce 191) (("-") reduce 191) ((";") reduce 191) (("@") reduce 191) ((T_ABSTRACT) reduce 191) ((T_ARRAY) reduce 191) ((T_ARRAY_CAST) reduce 191) ((T_ATTRIBUTE) reduce 191) ((T_BOOL_CAST) reduce 191) ((T_BREAK) reduce 191) ((T_CLASS) reduce 191) ((T_CLASS_C) reduce 191) ((T_CLONE) reduce 191) ((T_CONSTANT_ENCAPSED_STRING) reduce 191) ((T_CONTINUE) reduce 191) ((T_DEC) reduce 191) ((T_DECLARE) reduce 191) ((T_DIR) reduce 191) ((T_DNUMBER) reduce 191) ((T_DO) reduce 191) ((T_DOUBLE_CAST) reduce 191) ((T_ECHO) reduce 191) ((T_EMPTY) reduce 191) ((T_ENUM) reduce 191) ((T_EVAL) reduce 191) ((T_EXIT) reduce 191) ((T_FILE) reduce 191) ((T_FINAL) reduce 191) ((T_FN) reduce 191) ((T_FOR) reduce 191) ((T_FOREACH) reduce 191) ((T_FUNCTION) reduce 191) ((T_FUNC_C) reduce 191) ((T_GLOBAL) reduce 191) ((T_GOTO) reduce 191) ((T_HALT_COMPILER) reduce 191) ((T_IF) reduce 191) ((T_INC) reduce 191) ((T_INCLUDE) reduce 191) ((T_INCLUDE_ONCE) reduce 191) ((T_INLINE_HTML) reduce 191) ((T_INTERFACE) reduce 191) ((T_INT_CAST) reduce 191) ((T_ISSET) reduce 191) ((T_LINE) reduce 191) ((T_LIST) reduce 191) ((T_LNUMBER) reduce 191) ((T_MATCH) reduce 191) ((T_METHOD_C) reduce 191) ((T_NAME_FULLY_QUALIFIED) reduce 191) ((T_NAME_QUALIFIED) reduce 191) ((T_NAME_RELATIVE) reduce 191) ((T_NEW) reduce 191) ((T_NS_C) reduce 191) ((T_OBJECT_CAST) reduce 191) ((T_PRINT) reduce 191) ((T_REQUIRE) reduce 191) ((T_REQUIRE_ONCE) reduce 191) ((T_RETURN) reduce 191) ((T_START_HEREDOC) reduce 191) ((T_STATIC) reduce 191) ((T_STRING) reduce 191) ((T_STRING_CAST) reduce 191) ((T_SWITCH) reduce 191) ((T_THROW) reduce 191) ((T_TRAIT) reduce 191) ((T_TRAIT_C) reduce 191) ((T_TRY) reduce 191) ((T_UNSET) reduce 191) ((T_UNSET_CAST) reduce 191) ((T_VARIABLE) reduce 191) ((T_WHILE) reduce 191) ((T_YIELD) reduce 191) ((T_YIELD_FROM) reduce 191) (("[") reduce 191) (("`") reduce 191) (("{") reduce 191) (("}") reduce 191) (("~") reduce 191)) 24813 ((("!") reduce 234) (("\"") reduce 234) (("$") reduce 234) (("(") reduce 234) (("+") reduce 234) (("-") reduce 234) ((";") reduce 234) (("@") reduce 234) ((T_ABSTRACT) reduce 234) ((T_ARRAY) reduce 234) ((T_ARRAY_CAST) reduce 234) ((T_ATTRIBUTE) reduce 234) ((T_BOOL_CAST) reduce 234) ((T_BREAK) reduce 234) ((T_CLASS) reduce 234) ((T_CLASS_C) reduce 234) ((T_CLONE) reduce 234) ((T_CONSTANT_ENCAPSED_STRING) reduce 234) ((T_CONTINUE) reduce 234) ((T_DEC) reduce 234) ((T_DECLARE) reduce 234) ((T_DIR) reduce 234) ((T_DNUMBER) reduce 234) ((T_DO) reduce 234) ((T_DOUBLE_CAST) reduce 234) ((T_ECHO) reduce 234) ((T_ELSE) reduce 234) ((T_ELSEIF) reduce 234) ((T_EMPTY) reduce 234) ((T_ENUM) reduce 234) ((T_EVAL) reduce 234) ((T_EXIT) reduce 234) ((T_FILE) reduce 234) ((T_FINAL) reduce 234) ((T_FN) reduce 234) ((T_FOR) reduce 234) ((T_FOREACH) reduce 234) ((T_FUNCTION) reduce 234) ((T_FUNC_C) reduce 234) ((T_GLOBAL) reduce 234) ((T_GOTO) reduce 234) ((T_HALT_COMPILER) reduce 234) ((T_IF) reduce 234) ((T_INC) reduce 234) ((T_INCLUDE) reduce 234) ((T_INCLUDE_ONCE) reduce 234) ((T_INLINE_HTML) reduce 234) ((T_INTERFACE) reduce 234) ((T_INT_CAST) reduce 234) ((T_ISSET) reduce 234) ((T_LINE) reduce 234) ((T_LIST) reduce 234) ((T_LNUMBER) reduce 234) ((T_MATCH) reduce 234) ((T_METHOD_C) reduce 234) ((T_NAME_FULLY_QUALIFIED) reduce 234) ((T_NAME_QUALIFIED) reduce 234) ((T_NAME_RELATIVE) reduce 234) ((T_NEW) reduce 234) ((T_NS_C) reduce 234) ((T_OBJECT_CAST) reduce 234) ((T_PRINT) reduce 234) ((T_REQUIRE) reduce 234) ((T_REQUIRE_ONCE) reduce 234) ((T_RETURN) reduce 234) ((T_START_HEREDOC) reduce 234) ((T_STATIC) reduce 234) ((T_STRING) reduce 234) ((T_STRING_CAST) reduce 234) ((T_SWITCH) reduce 234) ((T_THROW) reduce 234) ((T_TRAIT) reduce 234) ((T_TRAIT_C) reduce 234) ((T_TRY) reduce 234) ((T_UNSET) reduce 234) ((T_UNSET_CAST) reduce 234) ((T_VARIABLE) reduce 234) ((T_WHILE) reduce 234) ((T_YIELD) reduce 234) ((T_YIELD_FROM) reduce 234) (("[") reduce 234) (("`") reduce 234) (("{") reduce 234) (("}") reduce 234) (("~") reduce 234)) 24816 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("}") reduce 144) (("~") reduce 144)) 24818 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 24820 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 24830 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 24831 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_EMPTY) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("}") reduce 210) (("~") reduce 210)) 24834 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_EMPTY) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("}") reduce 211) (("~") reduce 211)) 24835 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 24844 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 24845 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_EMPTY) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("}") reduce 208) (("~") reduce 208)) 24848 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_EMPTY) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("}") reduce 209) (("~") reduce 209)) 24855 ((("!") reduce 192) (("\"") reduce 192) (("$") reduce 192) (("(") reduce 192) (("+") reduce 192) (("-") reduce 192) ((";") reduce 192) (("@") reduce 192) ((T_ABSTRACT) reduce 192) ((T_ARRAY) reduce 192) ((T_ARRAY_CAST) reduce 192) ((T_ATTRIBUTE) reduce 192) ((T_BOOL_CAST) reduce 192) ((T_BREAK) reduce 192) ((T_CLASS) reduce 192) ((T_CLASS_C) reduce 192) ((T_CLONE) reduce 192) ((T_CONSTANT_ENCAPSED_STRING) reduce 192) ((T_CONTINUE) reduce 192) ((T_DEC) reduce 192) ((T_DECLARE) reduce 192) ((T_DIR) reduce 192) ((T_DNUMBER) reduce 192) ((T_DO) reduce 192) ((T_DOUBLE_CAST) reduce 192) ((T_ECHO) reduce 192) ((T_EMPTY) reduce 192) ((T_ENUM) reduce 192) ((T_EVAL) reduce 192) ((T_EXIT) reduce 192) ((T_FILE) reduce 192) ((T_FINAL) reduce 192) ((T_FN) reduce 192) ((T_FOR) reduce 192) ((T_FOREACH) reduce 192) ((T_FUNCTION) reduce 192) ((T_FUNC_C) reduce 192) ((T_GLOBAL) reduce 192) ((T_GOTO) reduce 192) ((T_HALT_COMPILER) reduce 192) ((T_IF) reduce 192) ((T_INC) reduce 192) ((T_INCLUDE) reduce 192) ((T_INCLUDE_ONCE) reduce 192) ((T_INLINE_HTML) reduce 192) ((T_INTERFACE) reduce 192) ((T_INT_CAST) reduce 192) ((T_ISSET) reduce 192) ((T_LINE) reduce 192) ((T_LIST) reduce 192) ((T_LNUMBER) reduce 192) ((T_MATCH) reduce 192) ((T_METHOD_C) reduce 192) ((T_NAME_FULLY_QUALIFIED) reduce 192) ((T_NAME_QUALIFIED) reduce 192) ((T_NAME_RELATIVE) reduce 192) ((T_NEW) reduce 192) ((T_NS_C) reduce 192) ((T_OBJECT_CAST) reduce 192) ((T_PRINT) reduce 192) ((T_REQUIRE) reduce 192) ((T_REQUIRE_ONCE) reduce 192) ((T_RETURN) reduce 192) ((T_START_HEREDOC) reduce 192) ((T_STATIC) reduce 192) ((T_STRING) reduce 192) ((T_STRING_CAST) reduce 192) ((T_SWITCH) reduce 192) ((T_THROW) reduce 192) ((T_TRAIT) reduce 192) ((T_TRAIT_C) reduce 192) ((T_TRY) reduce 192) ((T_UNSET) reduce 192) ((T_UNSET_CAST) reduce 192) ((T_VARIABLE) reduce 192) ((T_WHILE) reduce 192) ((T_YIELD) reduce 192) ((T_YIELD_FROM) reduce 192) (("[") reduce 192) (("`") reduce 192) (("{") reduce 192) (("}") reduce 192) (("~") reduce 192)) 24857 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 24863 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 24868 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 24869 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_EMPTY) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("}") reduce 212) (("~") reduce 212)) 24872 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_EMPTY) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("}") reduce 213) (("~") reduce 213)) 24874 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 24881 ((("!") reduce 185) (("\"") reduce 185) (("$") reduce 185) (("(") reduce 185) (("+") reduce 185) (("-") reduce 185) ((";") reduce 185) (("@") reduce 185) ((T_ABSTRACT) reduce 185) ((T_ARRAY) reduce 185) ((T_ARRAY_CAST) reduce 185) ((T_ATTRIBUTE) reduce 185) ((T_BOOL_CAST) reduce 185) ((T_BREAK) reduce 185) ((T_CLASS) reduce 185) ((T_CLASS_C) reduce 185) ((T_CLONE) reduce 185) ((T_CONSTANT_ENCAPSED_STRING) reduce 185) ((T_CONTINUE) reduce 185) ((T_DEC) reduce 185) ((T_DECLARE) reduce 185) ((T_DIR) reduce 185) ((T_DNUMBER) reduce 185) ((T_DO) reduce 185) ((T_DOUBLE_CAST) reduce 185) ((T_ECHO) reduce 185) ((T_EMPTY) reduce 185) ((T_ENUM) reduce 185) ((T_EVAL) reduce 185) ((T_EXIT) reduce 185) ((T_FILE) reduce 185) ((T_FINAL) reduce 185) ((T_FN) reduce 185) ((T_FOR) reduce 185) ((T_FOREACH) reduce 185) ((T_FUNCTION) reduce 185) ((T_FUNC_C) reduce 185) ((T_GLOBAL) reduce 185) ((T_GOTO) reduce 185) ((T_HALT_COMPILER) reduce 185) ((T_IF) reduce 185) ((T_INC) reduce 185) ((T_INCLUDE) reduce 185) ((T_INCLUDE_ONCE) reduce 185) ((T_INLINE_HTML) reduce 185) ((T_INTERFACE) reduce 185) ((T_INT_CAST) reduce 185) ((T_ISSET) reduce 185) ((T_LINE) reduce 185) ((T_LIST) reduce 185) ((T_LNUMBER) reduce 185) ((T_MATCH) reduce 185) ((T_METHOD_C) reduce 185) ((T_NAME_FULLY_QUALIFIED) reduce 185) ((T_NAME_QUALIFIED) reduce 185) ((T_NAME_RELATIVE) reduce 185) ((T_NEW) reduce 185) ((T_NS_C) reduce 185) ((T_OBJECT_CAST) reduce 185) ((T_PRINT) reduce 185) ((T_REQUIRE) reduce 185) ((T_REQUIRE_ONCE) reduce 185) ((T_RETURN) reduce 185) ((T_START_HEREDOC) reduce 185) ((T_STATIC) reduce 185) ((T_STRING) reduce 185) ((T_STRING_CAST) reduce 185) ((T_SWITCH) reduce 185) ((T_THROW) reduce 185) ((T_TRAIT) reduce 185) ((T_TRAIT_C) reduce 185) ((T_TRY) reduce 185) ((T_UNSET) reduce 185) ((T_UNSET_CAST) reduce 185) ((T_VARIABLE) reduce 185) ((T_WHILE) reduce 185) ((T_YIELD) reduce 185) ((T_YIELD_FROM) reduce 185) (("[") reduce 185) (("`") reduce 185) (("{") reduce 185) (("}") reduce 185) (("~") reduce 185)) 24883 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 24885 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) (("{") reduce 272) (("|") reduce 272)) 24886 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) (("{") reduce 273) (("|") reduce 273)) 24887 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) (("{") reduce 93) (("|") reduce 93)) 24888 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) (("{") reduce 92) (("|") reduce 92)) 24889 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) (("{") reduce 94) (("|") reduce 94)) 24890 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 263) (("{") reduce 263) (("|") reduce 263)) 24891 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) (("{") reduce 91) (("|") reduce 91)) 24892 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) (("{") reduce 261)) 24893 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) (("{") reduce 274) (("|") reduce 274)) 24894 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) (("{") reduce 258) (("|") shift)) 24895 ((("{") reduce 280)) 24896 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 262) (("{") reduce 262) (("|") reduce 262)) 24897 ((("{") reduce 260) (("|") shift)) 24899 ((("{") reduce 272) (("|") reduce 272)) 24900 ((("{") reduce 273) (("|") reduce 273)) 24901 ((("{") reduce 93) (("|") reduce 93)) 24902 ((("{") reduce 92) (("|") reduce 92)) 24903 ((("{") reduce 94) (("|") reduce 94)) 24904 ((("{") reduce 263) (("|") reduce 263)) 24905 ((("{") reduce 91) (("|") reduce 91)) 24906 ((("{") reduce 274) (("|") reduce 274)) 24907 ((("{") reduce 265) (("|") reduce 265)) 24908 ((("{") reduce 262) (("|") reduce 262)) 24911 ((("{") reduce 264) (("|") reduce 264)) 24912 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) (("{") reduce 272)) 24913 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) (("{") reduce 273)) 24914 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) (("{") reduce 93)) 24915 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) (("{") reduce 92)) 24916 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) (("{") reduce 94)) 24917 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 263) (("{") reduce 263)) 24918 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) (("{") reduce 91)) 24919 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) (("{") reduce 274)) 24920 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 266) (("{") reduce 266)) 24921 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 262) (("{") reduce 262)) 24923 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 267) (("{") reduce 267)) 24924 ((("{") reduce 272)) 24925 ((("{") reduce 273)) 24926 ((("{") reduce 93)) 24927 ((("{") reduce 92)) 24928 ((("{") reduce 94)) 24929 ((("{") reduce 263)) 24930 ((("{") reduce 91)) 24931 ((("{") reduce 274)) 24932 ((("{") reduce 259)) 24933 ((("{") reduce 262)) 24934 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_VARIABLE) shift)) 24935 (((T_VARIABLE) reduce 79)) 24936 (((T_VARIABLE) reduce 80)) 24937 (((")") reduce 452) ((",") reduce 452)) 24939 (((")") reduce 451) ((",") reduce 451)) 24941 (((")") reduce 124) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_VARIABLE) shift)) 24943 (((":") reduce 449) (("{") reduce 449)) 24944 (((")") reduce 450) ((",") reduce 450)) 24945 (((")") reduce 453) ((",") reduce 453)) 24947 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_ELLIPSIS) reduce 272) ((T_VARIABLE) reduce 272) (("|") reduce 272)) 24948 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_ELLIPSIS) reduce 273) ((T_VARIABLE) reduce 273) (("|") reduce 273)) 24949 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ELLIPSIS) reduce 93) ((T_VARIABLE) reduce 93) (("|") reduce 93)) 24950 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ELLIPSIS) reduce 92) ((T_VARIABLE) reduce 92) (("|") reduce 92)) 24951 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ELLIPSIS) reduce 94) ((T_VARIABLE) reduce 94) (("|") reduce 94)) 24952 ((("?") reduce 252) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 252) ((T_ARRAY) reduce 252) ((T_CALLABLE) reduce 252) ((T_ELLIPSIS) reduce 252) ((T_NAME_FULLY_QUALIFIED) reduce 252) ((T_NAME_QUALIFIED) reduce 252) ((T_NAME_RELATIVE) reduce 252) ((T_PRIVATE) reduce 252) ((T_PROTECTED) reduce 252) ((T_PUBLIC) reduce 252) ((T_READONLY) reduce 252) ((T_STRING) reduce 252) ((T_VARIABLE) reduce 252)) 24953 ((("?") reduce 251) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 251) ((T_ARRAY) reduce 251) ((T_CALLABLE) reduce 251) ((T_ELLIPSIS) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_PRIVATE) reduce 251) ((T_PROTECTED) reduce 251) ((T_PUBLIC) reduce 251) ((T_READONLY) reduce 251) ((T_STRING) reduce 251) ((T_VARIABLE) reduce 251)) 24954 ((("?") reduce 250) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 250) ((T_ARRAY) reduce 250) ((T_CALLABLE) reduce 250) ((T_ELLIPSIS) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_PRIVATE) reduce 250) ((T_PROTECTED) reduce 250) ((T_PUBLIC) reduce 250) ((T_READONLY) reduce 250) ((T_STRING) reduce 250) ((T_VARIABLE) reduce 250)) 24955 ((("?") reduce 253) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 253) ((T_ARRAY) reduce 253) ((T_CALLABLE) reduce 253) ((T_ELLIPSIS) reduce 253) ((T_NAME_FULLY_QUALIFIED) reduce 253) ((T_NAME_QUALIFIED) reduce 253) ((T_NAME_RELATIVE) reduce 253) ((T_PRIVATE) reduce 253) ((T_PROTECTED) reduce 253) ((T_PUBLIC) reduce 253) ((T_READONLY) reduce 253) ((T_STRING) reduce 253) ((T_VARIABLE) reduce 253)) 24956 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_ELLIPSIS) reduce 91) ((T_VARIABLE) reduce 91) (("|") reduce 91)) 24957 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 271) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ELLIPSIS) reduce 271) ((T_VARIABLE) reduce 271)) 24958 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_ELLIPSIS) reduce 274) ((T_VARIABLE) reduce 274) (("|") reduce 274)) 24959 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ELLIPSIS) reduce 180) ((T_VARIABLE) reduce 180)) 24960 ((("?") reduce 249) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 249) ((T_ARRAY) reduce 249) ((T_CALLABLE) reduce 249) ((T_ELLIPSIS) reduce 249) ((T_NAME_FULLY_QUALIFIED) reduce 249) ((T_NAME_QUALIFIED) reduce 249) ((T_NAME_RELATIVE) reduce 249) ((T_PRIVATE) reduce 249) ((T_PROTECTED) reduce 249) ((T_PUBLIC) reduce 249) ((T_READONLY) reduce 249) ((T_STRING) reduce 249) ((T_VARIABLE) reduce 249)) 24961 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 257) ((T_ELLIPSIS) reduce 257) ((T_VARIABLE) reduce 257)) 24962 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 268) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ELLIPSIS) reduce 268) ((T_VARIABLE) reduce 268) (("|") shift)) 24963 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 270) ((T_ELLIPSIS) reduce 270) ((T_VARIABLE) reduce 270) (("|") shift)) 24965 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_ELLIPSIS) reduce 272) ((T_VARIABLE) reduce 272) (("|") reduce 272)) 24966 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_ELLIPSIS) reduce 273) ((T_VARIABLE) reduce 273) (("|") reduce 273)) 24967 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ELLIPSIS) reduce 93) ((T_VARIABLE) reduce 93) (("|") reduce 93)) 24968 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ELLIPSIS) reduce 92) ((T_VARIABLE) reduce 92) (("|") reduce 92)) 24969 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ELLIPSIS) reduce 94) ((T_VARIABLE) reduce 94) (("|") reduce 94)) 24970 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_ELLIPSIS) reduce 91) ((T_VARIABLE) reduce 91) (("|") reduce 91)) 24971 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_ELLIPSIS) reduce 274) ((T_VARIABLE) reduce 274) (("|") reduce 274)) 24972 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 276) ((T_ELLIPSIS) reduce 276) ((T_VARIABLE) reduce 276) (("|") reduce 276)) 24975 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 275) ((T_ELLIPSIS) reduce 275) ((T_VARIABLE) reduce 275) (("|") reduce 275)) 24976 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_ELLIPSIS) reduce 272) ((T_VARIABLE) reduce 272)) 24977 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_ELLIPSIS) reduce 273) ((T_VARIABLE) reduce 273)) 24978 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ELLIPSIS) reduce 93) ((T_VARIABLE) reduce 93)) 24979 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ELLIPSIS) reduce 92) ((T_VARIABLE) reduce 92)) 24980 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ELLIPSIS) reduce 94) ((T_VARIABLE) reduce 94)) 24981 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_ELLIPSIS) reduce 91) ((T_VARIABLE) reduce 91)) 24982 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_ELLIPSIS) reduce 274) ((T_VARIABLE) reduce 274)) 24983 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 277) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 277) ((T_ELLIPSIS) reduce 277) ((T_VARIABLE) reduce 277)) 24984 (((T_ELLIPSIS) reduce 181) ((T_VARIABLE) reduce 181)) 24985 (((T_ELLIPSIS) shift) ((T_VARIABLE) reduce 182)) 24986 (((T_VARIABLE) reduce 183)) 24988 (((")") reduce 443) ((",") reduce 443) (("=") reduce 443)) 24989 (((")") reduce 254) ((",") reduce 254) (("=") shift)) 24991 ((("%") shift) ((")") reduce 255) (("*") shift) (("+") shift) ((",") reduce 255) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 24993 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_ELLIPSIS) reduce 278) ((T_VARIABLE) reduce 278)) 24994 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 272) ((T_ELLIPSIS) reduce 272) ((T_VARIABLE) reduce 272)) 24995 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 273) ((T_ELLIPSIS) reduce 273) ((T_VARIABLE) reduce 273)) 24996 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ELLIPSIS) reduce 93) ((T_VARIABLE) reduce 93)) 24997 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ELLIPSIS) reduce 92) ((T_VARIABLE) reduce 92)) 24998 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ELLIPSIS) reduce 94) ((T_VARIABLE) reduce 94)) 24999 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_ELLIPSIS) reduce 91) ((T_VARIABLE) reduce 91)) 25000 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_ELLIPSIS) reduce 274) ((T_VARIABLE) reduce 274)) 25001 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 269) ((T_ELLIPSIS) reduce 269) ((T_VARIABLE) reduce 269)) 25002 (((")") reduce 124) (("?") reduce 248) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 248) ((T_ARRAY) reduce 248) ((T_ATTRIBUTE) shift) ((T_CALLABLE) reduce 248) ((T_ELLIPSIS) reduce 248) ((T_NAME_FULLY_QUALIFIED) reduce 248) ((T_NAME_QUALIFIED) reduce 248) ((T_NAME_RELATIVE) reduce 248) ((T_PRIVATE) reduce 248) ((T_PROTECTED) reduce 248) ((T_PUBLIC) reduce 248) ((T_READONLY) reduce 248) ((T_STRING) reduce 248) ((T_VARIABLE) reduce 248)) 25003 (((")") reduce 242)) 25004 (((")") reduce 245) ((",") reduce 245)) 25005 ((("?") reduce 101) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 101) ((T_ARRAY) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_CALLABLE) reduce 101) ((T_ELLIPSIS) reduce 101) ((T_NAME_FULLY_QUALIFIED) reduce 101) ((T_NAME_QUALIFIED) reduce 101) ((T_NAME_RELATIVE) reduce 101) ((T_PRIVATE) reduce 101) ((T_PROTECTED) reduce 101) ((T_PUBLIC) reduce 101) ((T_READONLY) reduce 101) ((T_STRING) reduce 101) ((T_VARIABLE) reduce 101)) 25006 (((")") reduce 246) ((",") reduce 246)) 25009 ((("?") reduce 99) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 99) ((T_ARRAY) reduce 99) ((T_ATTRIBUTE) reduce 99) ((T_CALLABLE) reduce 99) ((T_ELLIPSIS) reduce 99) ((T_NAME_FULLY_QUALIFIED) reduce 99) ((T_NAME_QUALIFIED) reduce 99) ((T_NAME_RELATIVE) reduce 99) ((T_PRIVATE) reduce 99) ((T_PROTECTED) reduce 99) ((T_PUBLIC) reduce 99) ((T_READONLY) reduce 99) ((T_STRING) reduce 99) ((T_VARIABLE) reduce 99)) 25012 (((T_ATTRIBUTE) reduce 99) ((T_FN) reduce 99) ((T_FUNCTION) reduce 99) ((T_STATIC) reduce 99)) 25014 ((("!") reduce 145) (("\"") reduce 145) (($) reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONST) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAMESPACE) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_USE) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 25016 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 25017 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 25018 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 25019 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 25020 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 25021 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 25022 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 25023 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 25024 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 25025 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 25026 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 25027 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 25028 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 25029 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 25030 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 25031 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 25032 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 25033 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 25034 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 25035 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 25036 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 25037 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 25038 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 25039 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 25040 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 25041 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 25042 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 25043 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 25044 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 25045 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 25046 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 25047 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 25048 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 25049 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 25050 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 25051 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 25052 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 25053 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 25054 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 25055 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 25056 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 25057 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 25058 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 25059 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 25060 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 25061 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 25062 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 25063 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 25064 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 25065 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 25066 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 25067 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 25068 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 25069 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 25070 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 25071 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 25072 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 25073 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 25074 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 25075 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 25076 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 25077 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 25078 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 25079 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 25080 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 25081 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 25082 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 25083 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) ((";") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 25084 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 25085 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 25086 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 25087 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 25088 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 25089 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 25090 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 25091 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 25092 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 25093 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 25094 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 25096 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 25097 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 25098 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 25100 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 25101 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((";") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 25105 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((";") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 25106 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((";") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 25111 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) ((";") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 25113 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) ((";") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 25123 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 371) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25124 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 373) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25125 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 372) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25126 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 365) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25127 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 362) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25128 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 370) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25129 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 364) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25130 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 368) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25131 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 363) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25132 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 366) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25133 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 367) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25134 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 374) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25135 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 369) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25137 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 359) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25139 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 25142 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 25144 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 25146 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 25147 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 25148 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) ((";") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 25152 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 25154 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 25155 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((";") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 25159 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((";") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 25160 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((";") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 25161 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 25164 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 25165 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 25166 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 25172 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 25174 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 25175 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 25176 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 25177 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 25180 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 25181 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 25182 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 25183 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 25184 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 25187 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 25188 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 25194 ((("!") reduce 164) (("\"") reduce 164) (($) reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 25204 ((("!") reduce 158) (("\"") reduce 158) (($) reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 25212 ((("!") reduce 147) (("\"") reduce 147) (($) reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 25215 ((("!") reduce 146) (("\"") reduce 146) (($) reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONST) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAMESPACE) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_USE) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 25217 ((("!") reduce 235) (("\"") reduce 235) (("$") reduce 235) (($) reduce 235) (("(") reduce 235) (("+") reduce 235) (("-") reduce 235) ((";") reduce 235) (("@") reduce 235) ((T_ABSTRACT) reduce 235) ((T_ARRAY) reduce 235) ((T_ARRAY_CAST) reduce 235) ((T_ATTRIBUTE) reduce 235) ((T_BOOL_CAST) reduce 235) ((T_BREAK) reduce 235) ((T_CLASS) reduce 235) ((T_CLASS_C) reduce 235) ((T_CLONE) reduce 235) ((T_CONST) reduce 235) ((T_CONSTANT_ENCAPSED_STRING) reduce 235) ((T_CONTINUE) reduce 235) ((T_DEC) reduce 235) ((T_DECLARE) reduce 235) ((T_DIR) reduce 235) ((T_DNUMBER) reduce 235) ((T_DO) reduce 235) ((T_DOUBLE_CAST) reduce 235) ((T_ECHO) reduce 235) ((T_ELSE) reduce 235) ((T_ELSEIF) reduce 235) ((T_EMPTY) reduce 235) ((T_ENUM) reduce 235) ((T_EVAL) reduce 235) ((T_EXIT) reduce 235) ((T_FILE) reduce 235) ((T_FINAL) reduce 235) ((T_FN) reduce 235) ((T_FOR) reduce 235) ((T_FOREACH) reduce 235) ((T_FUNCTION) reduce 235) ((T_FUNC_C) reduce 235) ((T_GLOBAL) reduce 235) ((T_GOTO) reduce 235) ((T_HALT_COMPILER) reduce 235) ((T_IF) reduce 235) ((T_INC) reduce 235) ((T_INCLUDE) reduce 235) ((T_INCLUDE_ONCE) reduce 235) ((T_INLINE_HTML) reduce 235) ((T_INTERFACE) reduce 235) ((T_INT_CAST) reduce 235) ((T_ISSET) reduce 235) ((T_LINE) reduce 235) ((T_LIST) reduce 235) ((T_LNUMBER) reduce 235) ((T_MATCH) reduce 235) ((T_METHOD_C) reduce 235) ((T_NAMESPACE) reduce 235) ((T_NAME_FULLY_QUALIFIED) reduce 235) ((T_NAME_QUALIFIED) reduce 235) ((T_NAME_RELATIVE) reduce 235) ((T_NEW) reduce 235) ((T_NS_C) reduce 235) ((T_OBJECT_CAST) reduce 235) ((T_PRINT) reduce 235) ((T_REQUIRE) reduce 235) ((T_REQUIRE_ONCE) reduce 235) ((T_RETURN) reduce 235) ((T_START_HEREDOC) reduce 235) ((T_STATIC) reduce 235) ((T_STRING) reduce 235) ((T_STRING_CAST) reduce 235) ((T_SWITCH) reduce 235) ((T_THROW) reduce 235) ((T_TRAIT) reduce 235) ((T_TRAIT_C) reduce 235) ((T_TRY) reduce 235) ((T_UNSET) reduce 235) ((T_UNSET_CAST) reduce 235) ((T_USE) reduce 235) ((T_VARIABLE) reduce 235) ((T_WHILE) reduce 235) ((T_YIELD) reduce 235) ((T_YIELD_FROM) reduce 235) (("[") reduce 235) (("`") reduce 235) (("{") reduce 235) (("~") reduce 235)) 25220 ((("!") reduce 145) (("\"") reduce 145) (($) reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONST) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAMESPACE) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_USE) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 25222 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (($) reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONST) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_ELSE) reduce 237) ((T_ELSEIF) reduce 237) ((T_EMPTY) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAMESPACE) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_USE) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 25223 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (($) reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 25226 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (($) reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONST) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAMESPACE) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_USE) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 25230 ((("!") reduce 241) (("\"") reduce 241) (($) reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONST) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAMESPACE) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_USE) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 25235 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (($) reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONST) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_ELSE) reduce 232) ((T_ELSEIF) reduce 232) ((T_EMPTY) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAMESPACE) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_USE) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 25236 ((("!") reduce 148) (("\"") reduce 148) (($) reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 25239 ((("!") reduce 233) (("\"") reduce 233) (($) reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONST) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_ELSE) reduce 233) ((T_ELSEIF) reduce 233) ((T_EMPTY) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAMESPACE) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_USE) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 25244 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (($) reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 25247 ((("!") reduce 168) (("\"") reduce 168) (($) reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 25248 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (($) reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONST) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_ELSE) reduce 174) ((T_ELSEIF) reduce 174) ((T_EMPTY) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAMESPACE) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_USE) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 25251 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (($) reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 25254 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (($) reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONST) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAMESPACE) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_USE) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 25261 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (($) reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 25266 ((("!") reduce 151) (("\"") reduce 151) (($) reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 25270 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (($) reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONST) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_ELSE) reduce 214) ((T_ELSEIF) reduce 214) ((T_EMPTY) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAMESPACE) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_USE) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 25272 ((("!") reduce 215) (("\"") reduce 215) (($) reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONST) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_ELSE) reduce 215) ((T_ELSEIF) reduce 215) ((T_EMPTY) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAMESPACE) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_USE) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 25276 ((("!") reduce 216) (("\"") reduce 216) (($) reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONST) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAMESPACE) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_USE) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 25279 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (($) reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONST) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAMESPACE) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_USE) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 25280 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (($) reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 25282 ((("!") reduce 156) (("\"") reduce 156) (($) reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 25284 ((("!") reduce 154) (("\"") reduce 154) (($) reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 25286 ((("!") reduce 166) (("\"") reduce 166) (($) reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 25288 ((("!") reduce 155) (("\"") reduce 155) (($) reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 25298 ((("!") reduce 162) (("\"") reduce 162) (($) reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 25299 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (($) reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONST) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_ELSE) reduce 210) ((T_ELSEIF) reduce 210) ((T_EMPTY) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAMESPACE) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_USE) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 25302 ((("!") reduce 211) (("\"") reduce 211) (($) reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONST) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_ELSE) reduce 211) ((T_ELSEIF) reduce 211) ((T_EMPTY) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAMESPACE) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_USE) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 25303 ((("!") reduce 161) (("\"") reduce 161) (($) reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 25312 ((("!") reduce 150) (("\"") reduce 150) (($) reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 25313 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (($) reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONST) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_ELSE) reduce 208) ((T_ELSEIF) reduce 208) ((T_EMPTY) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAMESPACE) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_USE) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 25316 ((("!") reduce 209) (("\"") reduce 209) (($) reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONST) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_ELSE) reduce 209) ((T_ELSEIF) reduce 209) ((T_EMPTY) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAMESPACE) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_USE) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 25318 ((("!") reduce 157) (("\"") reduce 157) (($) reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 25324 ((("!") reduce 149) (("\"") reduce 149) (($) reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 25329 ((("!") reduce 163) (("\"") reduce 163) (($) reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 25330 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (($) reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONST) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_ELSE) reduce 212) ((T_ELSEIF) reduce 212) ((T_EMPTY) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAMESPACE) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_USE) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 25333 ((("!") reduce 213) (("\"") reduce 213) (($) reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONST) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_ELSE) reduce 213) ((T_ELSEIF) reduce 213) ((T_EMPTY) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAMESPACE) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_USE) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 25335 ((("!") reduce 153) (("\"") reduce 153) (($) reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 25337 ((("!") reduce 152) (("\"") reduce 152) (($) reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 25338 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (($) reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONST) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_EMPTY) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAMESPACE) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_USE) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 25349 ((("!") reduce 444) (("\"") reduce 444) (("$") reduce 444) (($) reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) ((";") reduce 444) (("@") reduce 444) ((T_ABSTRACT) reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_BREAK) reduce 444) ((T_CLASS) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONST) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_CONTINUE) reduce 444) ((T_DEC) reduce 444) ((T_DECLARE) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DO) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_ECHO) reduce 444) ((T_EMPTY) reduce 444) ((T_ENUM) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FINAL) reduce 444) ((T_FN) reduce 444) ((T_FOR) reduce 444) ((T_FOREACH) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_GLOBAL) reduce 444) ((T_GOTO) reduce 444) ((T_HALT_COMPILER) reduce 444) ((T_IF) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INLINE_HTML) reduce 444) ((T_INTERFACE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAMESPACE) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_RETURN) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_SWITCH) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT) reduce 444) ((T_TRAIT_C) reduce 444) ((T_TRY) reduce 444) ((T_UNSET) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_USE) reduce 444) ((T_VARIABLE) reduce 444) ((T_WHILE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("{") reduce 444) (("~") reduce 444)) 25350 ((("!") reduce 179) (("\"") reduce 179) (("$") reduce 179) (($) reduce 179) (("(") reduce 179) (("+") reduce 179) (("-") reduce 179) ((";") reduce 179) (("@") reduce 179) ((T_ABSTRACT) reduce 179) ((T_ARRAY) reduce 179) ((T_ARRAY_CAST) reduce 179) ((T_ATTRIBUTE) reduce 179) ((T_BOOL_CAST) reduce 179) ((T_BREAK) reduce 179) ((T_CLASS) reduce 179) ((T_CLASS_C) reduce 179) ((T_CLONE) reduce 179) ((T_CONST) reduce 179) ((T_CONSTANT_ENCAPSED_STRING) reduce 179) ((T_CONTINUE) reduce 179) ((T_DEC) reduce 179) ((T_DECLARE) reduce 179) ((T_DIR) reduce 179) ((T_DNUMBER) reduce 179) ((T_DO) reduce 179) ((T_DOUBLE_CAST) reduce 179) ((T_ECHO) reduce 179) ((T_EMPTY) reduce 179) ((T_ENUM) reduce 179) ((T_EVAL) reduce 179) ((T_EXIT) reduce 179) ((T_FILE) reduce 179) ((T_FINAL) reduce 179) ((T_FN) reduce 179) ((T_FOR) reduce 179) ((T_FOREACH) reduce 179) ((T_FUNCTION) reduce 179) ((T_FUNC_C) reduce 179) ((T_GLOBAL) reduce 179) ((T_GOTO) reduce 179) ((T_HALT_COMPILER) reduce 179) ((T_IF) reduce 179) ((T_INC) reduce 179) ((T_INCLUDE) reduce 179) ((T_INCLUDE_ONCE) reduce 179) ((T_INLINE_HTML) reduce 179) ((T_INTERFACE) reduce 179) ((T_INT_CAST) reduce 179) ((T_ISSET) reduce 179) ((T_LINE) reduce 179) ((T_LIST) reduce 179) ((T_LNUMBER) reduce 179) ((T_MATCH) reduce 179) ((T_METHOD_C) reduce 179) ((T_NAMESPACE) reduce 179) ((T_NAME_FULLY_QUALIFIED) reduce 179) ((T_NAME_QUALIFIED) reduce 179) ((T_NAME_RELATIVE) reduce 179) ((T_NEW) reduce 179) ((T_NS_C) reduce 179) ((T_OBJECT_CAST) reduce 179) ((T_PRINT) reduce 179) ((T_REQUIRE) reduce 179) ((T_REQUIRE_ONCE) reduce 179) ((T_RETURN) reduce 179) ((T_START_HEREDOC) reduce 179) ((T_STATIC) reduce 179) ((T_STRING) reduce 179) ((T_STRING_CAST) reduce 179) ((T_SWITCH) reduce 179) ((T_THROW) reduce 179) ((T_TRAIT) reduce 179) ((T_TRAIT_C) reduce 179) ((T_TRY) reduce 179) ((T_UNSET) reduce 179) ((T_UNSET_CAST) reduce 179) ((T_USE) reduce 179) ((T_VARIABLE) reduce 179) ((T_WHILE) reduce 179) ((T_YIELD) reduce 179) ((T_YIELD_FROM) reduce 179) (("[") reduce 179) (("`") reduce 179) (("{") reduce 179) (("~") reduce 179)) 25421 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 444) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25734 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 25737 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 25739 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 25740 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 25741 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 25750 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 25752 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 25753 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 25754 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 25755 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 25758 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 25759 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 25760 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 25761 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 25763 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 25765 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 25767 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 358) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25768 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 432) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25779 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 490) ((T_OBJECT_OPERATOR) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("[") reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 25781 ((("%") reduce 472) (("(") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 472) ((T_OBJECT_OPERATOR) reduce 472) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 472) ((T_POW) reduce 472) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("[") reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 25783 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 485) ((T_OBJECT_OPERATOR) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("[") reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 25784 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_POW) reduce 475) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 25788 ((("%") reduce 463) (("(") shift) (("*") reduce 463) (("+") reduce 463) (("-") reduce 463) ((".") reduce 463) (("/") reduce 463) ((";") reduce 463) (("<") reduce 463) ((">") reduce 463) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 463) ((T_BOOLEAN_AND) reduce 463) ((T_BOOLEAN_OR) reduce 463) ((T_COALESCE) reduce 463) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) reduce 463) ((T_IS_EQUAL) reduce 463) ((T_IS_GREATER_OR_EQUAL) reduce 463) ((T_IS_IDENTICAL) reduce 463) ((T_IS_NOT_EQUAL) reduce 463) ((T_IS_NOT_IDENTICAL) reduce 463) ((T_IS_SMALLER_OR_EQUAL) reduce 463) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_POW) reduce 463) ((T_SL) reduce 463) ((T_SPACESHIP) reduce 463) ((T_SR) reduce 463) (("^") reduce 463) (("|") reduce 463)) 25789 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 484) ((T_OBJECT_OPERATOR) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("[") reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 25790 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 488) ((T_OBJECT_OPERATOR) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("[") reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 25796 ((("%") reduce 483) (("*") reduce 483) (("+") reduce 483) (("-") reduce 483) ((".") reduce 483) (("/") reduce 483) ((";") reduce 483) (("<") reduce 483) ((">") reduce 483) (("?") reduce 483) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 483) ((T_BOOLEAN_AND) reduce 483) ((T_BOOLEAN_OR) reduce 483) ((T_COALESCE) reduce 483) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) reduce 483) ((T_IS_EQUAL) reduce 483) ((T_IS_GREATER_OR_EQUAL) reduce 483) ((T_IS_IDENTICAL) reduce 483) ((T_IS_NOT_EQUAL) reduce 483) ((T_IS_NOT_IDENTICAL) reduce 483) ((T_IS_SMALLER_OR_EQUAL) reduce 483) ((T_LOGICAL_AND) reduce 483) ((T_LOGICAL_OR) reduce 483) ((T_LOGICAL_XOR) reduce 483) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 483) ((T_OBJECT_OPERATOR) reduce 483) ((T_POW) reduce 483) ((T_SL) reduce 483) ((T_SPACESHIP) reduce 483) ((T_SR) reduce 483) (("[") reduce 483) (("^") reduce 483) (("{") reduce 483) (("|") reduce 483)) 25798 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 25800 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 487) ((T_OBJECT_OPERATOR) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("[") reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 25801 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 25802 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 25803 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 25805 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 489) ((T_OBJECT_OPERATOR) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("[") reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 25812 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 91) ((T_OBJECT_OPERATOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("[") reduce 91) (("^") reduce 91) (("{") reduce 91) (("|") reduce 91)) 25815 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 486) ((T_OBJECT_OPERATOR) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("[") reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 25817 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) (("=") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AND_EQUAL) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_COALESCE_EQUAL) reduce 515) ((T_CONCAT_EQUAL) reduce 515) ((T_DEC) reduce 515) ((T_DIV_EQUAL) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INC) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_MINUS_EQUAL) reduce 515) ((T_MOD_EQUAL) reduce 515) ((T_MUL_EQUAL) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_OR_EQUAL) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_PLUS_EQUAL) reduce 515) ((T_POW) reduce 515) ((T_POW_EQUAL) reduce 515) ((T_SL) reduce 515) ((T_SL_EQUAL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) ((T_SR_EQUAL) reduce 515) ((T_XOR_EQUAL) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 25818 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 429) (("(") shift) (("*") reduce 429) (("+") shift) (("-") shift) ((".") reduce 429) (("/") reduce 429) ((";") reduce 429) (("<") reduce 429) ((">") reduce 429) (("?") reduce 429) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 429) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 429) ((T_BOOLEAN_OR) reduce 429) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 429) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 429) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 429) ((T_IS_GREATER_OR_EQUAL) reduce 429) ((T_IS_IDENTICAL) reduce 429) ((T_IS_NOT_EQUAL) reduce 429) ((T_IS_NOT_IDENTICAL) reduce 429) ((T_IS_SMALLER_OR_EQUAL) reduce 429) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 429) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) reduce 429) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_POW) reduce 429) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 429) ((T_SPACESHIP) reduce 429) ((T_SR) reduce 429) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 429) (("`") shift) (("|") reduce 429) (("~") shift)) 25825 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) (("=") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AND_EQUAL) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_COALESCE_EQUAL) reduce 511) ((T_CONCAT_EQUAL) reduce 511) ((T_DEC) reduce 511) ((T_DIV_EQUAL) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INC) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_MINUS_EQUAL) reduce 511) ((T_MOD_EQUAL) reduce 511) ((T_MUL_EQUAL) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_OR_EQUAL) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_PLUS_EQUAL) reduce 511) ((T_POW) reduce 511) ((T_POW_EQUAL) reduce 511) ((T_SL) reduce 511) ((T_SL_EQUAL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) ((T_SR_EQUAL) reduce 511) ((T_XOR_EQUAL) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 25826 ((("%") reduce 481) (("*") reduce 481) (("+") reduce 481) (("-") reduce 481) ((".") reduce 481) (("/") reduce 481) ((";") reduce 481) (("<") reduce 481) ((">") reduce 481) (("?") reduce 481) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 481) ((T_BOOLEAN_AND) reduce 481) ((T_BOOLEAN_OR) reduce 481) ((T_COALESCE) reduce 481) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) reduce 481) ((T_IS_EQUAL) reduce 481) ((T_IS_GREATER_OR_EQUAL) reduce 481) ((T_IS_IDENTICAL) reduce 481) ((T_IS_NOT_EQUAL) reduce 481) ((T_IS_NOT_IDENTICAL) reduce 481) ((T_IS_SMALLER_OR_EQUAL) reduce 481) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 499) ((T_OBJECT_OPERATOR) reduce 499) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 499) ((T_POW) reduce 481) ((T_SL) reduce 481) ((T_SPACESHIP) reduce 481) ((T_SR) reduce 481) (("[") reduce 499) (("^") reduce 481) (("{") reduce 499) (("|") reduce 481)) 25828 ((("%") reduce 480) (("*") reduce 480) (("+") reduce 480) (("-") reduce 480) ((".") reduce 480) (("/") reduce 480) ((";") reduce 480) (("<") reduce 480) ((">") reduce 480) (("?") reduce 480) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 480) ((T_BOOLEAN_AND) reduce 480) ((T_BOOLEAN_OR) reduce 480) ((T_COALESCE) reduce 480) ((T_DOUBLE_ARROW) reduce 480) ((T_INSTANCEOF) reduce 480) ((T_IS_EQUAL) reduce 480) ((T_IS_GREATER_OR_EQUAL) reduce 480) ((T_IS_IDENTICAL) reduce 480) ((T_IS_NOT_EQUAL) reduce 480) ((T_IS_NOT_IDENTICAL) reduce 480) ((T_IS_SMALLER_OR_EQUAL) reduce 480) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 501) ((T_OBJECT_OPERATOR) reduce 501) ((T_POW) reduce 480) ((T_SL) reduce 480) ((T_SPACESHIP) reduce 480) ((T_SR) reduce 480) (("[") reduce 501) (("^") reduce 480) (("{") reduce 501) (("|") reduce 480)) 25829 ((("%") reduce 479) (("(") reduce 504) (("*") reduce 479) (("+") reduce 479) (("-") reduce 479) ((".") reduce 479) (("/") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_COALESCE) reduce 479) ((T_DOUBLE_ARROW) reduce 479) ((T_INSTANCEOF) reduce 479) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 498) ((T_OBJECT_OPERATOR) reduce 498) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 498) ((T_POW) reduce 479) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) (("[") reduce 498) (("^") reduce 479) (("{") reduce 498) (("|") reduce 479)) 25830 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25833 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) (("=") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AND_EQUAL) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_COALESCE_EQUAL) reduce 510) ((T_CONCAT_EQUAL) reduce 510) ((T_DEC) reduce 510) ((T_DIV_EQUAL) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INC) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_MINUS_EQUAL) reduce 510) ((T_MOD_EQUAL) reduce 510) ((T_MUL_EQUAL) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_OR_EQUAL) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PLUS_EQUAL) reduce 510) ((T_POW) reduce 510) ((T_POW_EQUAL) reduce 510) ((T_SL) reduce 510) ((T_SL_EQUAL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) ((T_SR_EQUAL) reduce 510) ((T_XOR_EQUAL) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 25834 ((("%") reduce 434) (("*") reduce 434) (("+") reduce 434) (("-") reduce 434) ((".") reduce 434) (("/") reduce 434) ((";") reduce 434) (("<") reduce 434) ((">") reduce 434) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) reduce 434) ((T_IS_EQUAL) reduce 434) ((T_IS_GREATER_OR_EQUAL) reduce 434) ((T_IS_IDENTICAL) reduce 434) ((T_IS_NOT_EQUAL) reduce 434) ((T_IS_NOT_IDENTICAL) reduce 434) ((T_IS_SMALLER_OR_EQUAL) reduce 434) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_POW) reduce 434) ((T_SL) reduce 434) ((T_SPACESHIP) reduce 434) ((T_SR) reduce 434) (("^") reduce 434) (("|") reduce 434)) 25835 ((("%") reduce 416) (("*") reduce 416) (("+") reduce 416) (("-") reduce 416) ((".") reduce 416) (("/") reduce 416) ((";") reduce 416) (("<") reduce 416) ((">") reduce 416) (("?") reduce 416) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 416) ((T_BOOLEAN_AND) reduce 416) ((T_BOOLEAN_OR) reduce 416) ((T_COALESCE) reduce 416) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) reduce 416) ((T_IS_EQUAL) reduce 416) ((T_IS_GREATER_OR_EQUAL) reduce 416) ((T_IS_IDENTICAL) reduce 416) ((T_IS_NOT_EQUAL) reduce 416) ((T_IS_NOT_IDENTICAL) reduce 416) ((T_IS_SMALLER_OR_EQUAL) reduce 416) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_POW) reduce 416) ((T_SL) reduce 416) ((T_SPACESHIP) reduce 416) ((T_SR) reduce 416) (("^") reduce 416) (("|") reduce 416)) 25836 ((("%") reduce 438) (("*") reduce 438) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") reduce 438) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) reduce 438) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_POW) reduce 438) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 25837 ((("%") reduce 482) (("(") shift) (("*") reduce 482) (("+") reduce 482) (("-") reduce 482) ((".") reduce 482) (("/") reduce 482) ((";") reduce 482) (("<") reduce 482) ((">") reduce 482) (("?") reduce 482) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 482) ((T_BOOLEAN_AND) reduce 482) ((T_BOOLEAN_OR) reduce 482) ((T_COALESCE) reduce 482) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) reduce 482) ((T_IS_EQUAL) reduce 482) ((T_IS_GREATER_OR_EQUAL) reduce 482) ((T_IS_IDENTICAL) reduce 482) ((T_IS_NOT_EQUAL) reduce 482) ((T_IS_NOT_IDENTICAL) reduce 482) ((T_IS_SMALLER_OR_EQUAL) reduce 482) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 482) ((T_OBJECT_OPERATOR) reduce 482) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 482) ((T_SL) reduce 482) ((T_SPACESHIP) reduce 482) ((T_SR) reduce 482) (("[") reduce 482) (("^") reduce 482) (("{") reduce 482) (("|") reduce 482)) 25838 ((("%") reduce 412) (("*") reduce 412) (("+") reduce 412) (("-") reduce 412) ((".") reduce 412) (("/") reduce 412) ((";") reduce 412) (("<") reduce 412) ((">") reduce 412) (("?") reduce 412) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 412) ((T_BOOLEAN_AND) reduce 412) ((T_BOOLEAN_OR) reduce 412) ((T_COALESCE) reduce 412) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) reduce 412) ((T_IS_EQUAL) reduce 412) ((T_IS_GREATER_OR_EQUAL) reduce 412) ((T_IS_IDENTICAL) reduce 412) ((T_IS_NOT_EQUAL) reduce 412) ((T_IS_NOT_IDENTICAL) reduce 412) ((T_IS_SMALLER_OR_EQUAL) reduce 412) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_POW) reduce 412) ((T_SL) reduce 412) ((T_SPACESHIP) reduce 412) ((T_SR) reduce 412) (("^") reduce 412) (("|") reduce 412)) 25839 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("^") reduce 426) (("|") reduce 426)) 25840 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 25841 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) (("=") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AND_EQUAL) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_COALESCE_EQUAL) reduce 512) ((T_CONCAT_EQUAL) reduce 512) ((T_DEC) reduce 512) ((T_DIV_EQUAL) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INC) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_MINUS_EQUAL) reduce 512) ((T_MOD_EQUAL) reduce 512) ((T_MUL_EQUAL) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_OR_EQUAL) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_PLUS_EQUAL) reduce 512) ((T_POW) reduce 512) ((T_POW_EQUAL) reduce 512) ((T_SL) reduce 512) ((T_SL_EQUAL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) ((T_SR_EQUAL) reduce 512) ((T_XOR_EQUAL) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 25842 ((("%") reduce 356) (("*") reduce 356) (("+") reduce 356) (("-") reduce 356) ((".") reduce 356) (("/") reduce 356) ((";") reduce 356) (("<") reduce 356) (("=") shift) ((">") reduce 356) (("?") reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 356) ((T_BOOLEAN_OR) reduce 356) ((T_COALESCE) reduce 356) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 356) ((T_INC) shift) ((T_INSTANCEOF) reduce 356) ((T_IS_EQUAL) reduce 356) ((T_IS_GREATER_OR_EQUAL) reduce 356) ((T_IS_IDENTICAL) reduce 356) ((T_IS_NOT_EQUAL) reduce 356) ((T_IS_NOT_IDENTICAL) reduce 356) ((T_IS_SMALLER_OR_EQUAL) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 356) ((T_POW_EQUAL) shift) ((T_SL) reduce 356) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 356) ((T_SR) reduce 356) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 496) (("^") reduce 356) (("{") reduce 496) (("|") reduce 356)) 25845 ((("%") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_POW) shift) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("^") reduce 400) (("|") reduce 400)) 25876 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 384) (("<") shift) ((">") shift) (("?") reduce 384) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 384) ((T_BOOLEAN_OR) reduce 384) ((T_COALESCE) reduce 384) ((T_DOUBLE_ARROW) reduce 384) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 384) ((T_LOGICAL_OR) reduce 384) ((T_LOGICAL_XOR) reduce 384) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 384)) 25877 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 387) (("<") shift) ((">") shift) (("?") reduce 387) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 387) ((T_BOOLEAN_OR) reduce 387) ((T_COALESCE) reduce 387) ((T_DOUBLE_ARROW) reduce 387) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 387) ((T_LOGICAL_OR) reduce 387) ((T_LOGICAL_XOR) reduce 387) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 387) (("|") reduce 387)) 25878 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 396) (("/") shift) ((";") reduce 396) (("<") reduce 396) ((">") reduce 396) (("?") reduce 396) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 396) ((T_BOOLEAN_AND) reduce 396) ((T_BOOLEAN_OR) reduce 396) ((T_COALESCE) reduce 396) ((T_DOUBLE_ARROW) reduce 396) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 396) ((T_IS_GREATER_OR_EQUAL) reduce 396) ((T_IS_IDENTICAL) reduce 396) ((T_IS_NOT_EQUAL) reduce 396) ((T_IS_NOT_IDENTICAL) reduce 396) ((T_IS_SMALLER_OR_EQUAL) reduce 396) ((T_LOGICAL_AND) reduce 396) ((T_LOGICAL_OR) reduce 396) ((T_LOGICAL_XOR) reduce 396) ((T_POW) shift) ((T_SL) reduce 396) ((T_SPACESHIP) reduce 396) ((T_SR) reduce 396) (("^") reduce 396) (("|") reduce 396)) 25879 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 409) (("<") shift) ((">") shift) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 409) (("|") reduce 409)) 25880 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 395) (("/") shift) ((";") reduce 395) (("<") reduce 395) ((">") reduce 395) (("?") reduce 395) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 395) ((T_BOOLEAN_AND) reduce 395) ((T_BOOLEAN_OR) reduce 395) ((T_COALESCE) reduce 395) ((T_DOUBLE_ARROW) reduce 395) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 395) ((T_IS_GREATER_OR_EQUAL) reduce 395) ((T_IS_IDENTICAL) reduce 395) ((T_IS_NOT_EQUAL) reduce 395) ((T_IS_NOT_IDENTICAL) reduce 395) ((T_IS_SMALLER_OR_EQUAL) reduce 395) ((T_LOGICAL_AND) reduce 395) ((T_LOGICAL_OR) reduce 395) ((T_LOGICAL_XOR) reduce 395) ((T_POW) shift) ((T_SL) reduce 395) ((T_SPACESHIP) reduce 395) ((T_SR) reduce 395) (("^") reduce 395) (("|") reduce 395)) 25881 ((("%") reduce 392) (("*") reduce 392) (("+") reduce 392) (("-") reduce 392) ((".") reduce 392) (("/") reduce 392) ((";") reduce 392) (("<") reduce 392) ((">") reduce 392) (("?") reduce 392) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 392) ((T_BOOLEAN_AND) reduce 392) ((T_BOOLEAN_OR) reduce 392) ((T_COALESCE) reduce 392) ((T_DOUBLE_ARROW) reduce 392) ((T_INSTANCEOF) reduce 392) ((T_IS_EQUAL) reduce 392) ((T_IS_GREATER_OR_EQUAL) reduce 392) ((T_IS_IDENTICAL) reduce 392) ((T_IS_NOT_EQUAL) reduce 392) ((T_IS_NOT_IDENTICAL) reduce 392) ((T_IS_SMALLER_OR_EQUAL) reduce 392) ((T_LOGICAL_AND) reduce 392) ((T_LOGICAL_OR) reduce 392) ((T_LOGICAL_XOR) reduce 392) ((T_POW) shift) ((T_SL) reduce 392) ((T_SPACESHIP) reduce 392) ((T_SR) reduce 392) (("^") reduce 392) (("|") reduce 392)) 25882 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 383) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 383) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 383) ((T_LOGICAL_XOR) reduce 383) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25883 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 381) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 381) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 381) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25884 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 382) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 382) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 382) ((T_LOGICAL_OR) reduce 382) ((T_LOGICAL_XOR) reduce 382) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25885 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") reduce 406) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 406) ((T_BOOLEAN_AND) reduce 406) ((T_BOOLEAN_OR) reduce 406) ((T_COALESCE) reduce 406) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 406) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 406) ((T_IS_NOT_EQUAL) reduce 406) ((T_IS_NOT_IDENTICAL) reduce 406) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 406) ((T_SR) shift) (("^") reduce 406) (("|") reduce 406)) 25886 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 402) (("<") shift) ((">") shift) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 402) (("|") reduce 402)) 25887 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") reduce 404) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 404) ((T_BOOLEAN_AND) reduce 404) ((T_BOOLEAN_OR) reduce 404) ((T_COALESCE) reduce 404) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 404) (("|") reduce 404)) 25888 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 401) (("<") shift) ((">") shift) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 401) (("|") reduce 401)) 25889 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 408) (("<") shift) ((">") shift) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 408) ((T_SR) shift) (("^") reduce 408) (("|") reduce 408)) 25890 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 403) (("<") shift) ((">") shift) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_DOUBLE_ARROW) reduce 403) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 403) (("|") reduce 403)) 25893 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 25894 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 25895 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 25896 ((("%") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 25897 ((("%") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 25898 ((("%") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 25899 ((("%") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 25900 ((("%") reduce 410) (("*") reduce 410) (("+") reduce 410) (("-") reduce 410) ((".") reduce 410) (("/") reduce 410) ((";") reduce 410) (("<") reduce 410) ((">") reduce 410) (("?") reduce 410) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 410) ((T_BOOLEAN_AND) reduce 410) ((T_BOOLEAN_OR) reduce 410) ((T_COALESCE) reduce 410) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) reduce 410) ((T_IS_EQUAL) reduce 410) ((T_IS_GREATER_OR_EQUAL) reduce 410) ((T_IS_IDENTICAL) reduce 410) ((T_IS_NOT_EQUAL) reduce 410) ((T_IS_NOT_IDENTICAL) reduce 410) ((T_IS_SMALLER_OR_EQUAL) reduce 410) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_POW) reduce 410) ((T_SL) reduce 410) ((T_SPACESHIP) reduce 410) ((T_SR) reduce 410) (("^") reduce 410) (("|") reduce 410)) 25901 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 25902 ((("%") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 25903 ((("%") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 25910 ((("%") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 25912 ((("%") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 25913 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 25914 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 25915 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 25916 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 25919 ((("%") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 25920 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 25922 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 25924 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 25925 ((("%") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 25928 ((("%") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 25929 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") reduce 415) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25930 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 379) (("<") shift) ((">") shift) (("?") reduce 379) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 379) ((T_COALESCE) reduce 379) ((T_DOUBLE_ARROW) reduce 379) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 379) ((T_LOGICAL_OR) reduce 379) ((T_LOGICAL_XOR) reduce 379) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25931 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 380) (("<") shift) ((">") shift) (("?") reduce 380) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 380) ((T_BOOLEAN_OR) reduce 380) ((T_COALESCE) reduce 380) ((T_DOUBLE_ARROW) reduce 380) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 380) ((T_LOGICAL_OR) reduce 380) ((T_LOGICAL_XOR) reduce 380) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25932 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 385) (("<") shift) ((">") shift) (("?") reduce 385) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 385) ((T_BOOLEAN_AND) reduce 385) ((T_BOOLEAN_OR) reduce 385) ((T_COALESCE) reduce 385) ((T_DOUBLE_ARROW) reduce 385) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 385) ((T_LOGICAL_OR) reduce 385) ((T_LOGICAL_XOR) reduce 385) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 385) (("|") reduce 385)) 25933 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 386) (("<") shift) ((">") shift) (("?") reduce 386) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 386) ((T_BOOLEAN_AND) reduce 386) ((T_BOOLEAN_OR) reduce 386) ((T_COALESCE) reduce 386) ((T_DOUBLE_ARROW) reduce 386) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 386) ((T_LOGICAL_OR) reduce 386) ((T_LOGICAL_XOR) reduce 386) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 386) (("|") reduce 386)) 25937 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") reduce 413) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25938 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") reduce 414) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25939 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 407) (("<") shift) ((">") shift) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 407) ((T_SR) shift) (("^") reduce 407) (("|") reduce 407)) 25940 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") reduce 405) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 405) ((T_BOOLEAN_AND) reduce 405) ((T_BOOLEAN_OR) reduce 405) ((T_COALESCE) reduce 405) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 405) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 405) ((T_IS_NOT_EQUAL) reduce 405) ((T_IS_NOT_IDENTICAL) reduce 405) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 405) ((T_SR) shift) (("^") reduce 405) (("|") reduce 405)) 25941 ((("%") reduce 393) (("*") reduce 393) (("+") reduce 393) (("-") reduce 393) ((".") reduce 393) (("/") reduce 393) ((";") reduce 393) (("<") reduce 393) ((">") reduce 393) (("?") reduce 393) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 393) ((T_BOOLEAN_AND) reduce 393) ((T_BOOLEAN_OR) reduce 393) ((T_COALESCE) reduce 393) ((T_DOUBLE_ARROW) reduce 393) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 393) ((T_IS_GREATER_OR_EQUAL) reduce 393) ((T_IS_IDENTICAL) reduce 393) ((T_IS_NOT_EQUAL) reduce 393) ((T_IS_NOT_IDENTICAL) reduce 393) ((T_IS_SMALLER_OR_EQUAL) reduce 393) ((T_LOGICAL_AND) reduce 393) ((T_LOGICAL_OR) reduce 393) ((T_LOGICAL_XOR) reduce 393) ((T_POW) shift) ((T_SL) reduce 393) ((T_SPACESHIP) reduce 393) ((T_SR) reduce 393) (("^") reduce 393) (("|") reduce 393)) 25942 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 388) (("/") shift) ((";") reduce 388) (("<") reduce 388) ((">") reduce 388) (("?") reduce 388) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 388) ((T_BOOLEAN_AND) reduce 388) ((T_BOOLEAN_OR) reduce 388) ((T_COALESCE) reduce 388) ((T_DOUBLE_ARROW) reduce 388) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 388) ((T_IS_GREATER_OR_EQUAL) reduce 388) ((T_IS_IDENTICAL) reduce 388) ((T_IS_NOT_EQUAL) reduce 388) ((T_IS_NOT_IDENTICAL) reduce 388) ((T_IS_SMALLER_OR_EQUAL) reduce 388) ((T_LOGICAL_AND) reduce 388) ((T_LOGICAL_OR) reduce 388) ((T_LOGICAL_XOR) reduce 388) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 388) ((T_SR) shift) (("^") reduce 388) (("|") reduce 388)) 25943 ((("%") shift) (("*") shift) (("+") reduce 390) (("-") reduce 390) ((".") reduce 390) (("/") shift) ((";") reduce 390) (("<") reduce 390) ((">") reduce 390) (("?") reduce 390) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 390) ((T_BOOLEAN_AND) reduce 390) ((T_BOOLEAN_OR) reduce 390) ((T_COALESCE) reduce 390) ((T_DOUBLE_ARROW) reduce 390) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 390) ((T_IS_GREATER_OR_EQUAL) reduce 390) ((T_IS_IDENTICAL) reduce 390) ((T_IS_NOT_EQUAL) reduce 390) ((T_IS_NOT_IDENTICAL) reduce 390) ((T_IS_SMALLER_OR_EQUAL) reduce 390) ((T_LOGICAL_AND) reduce 390) ((T_LOGICAL_OR) reduce 390) ((T_LOGICAL_XOR) reduce 390) ((T_POW) shift) ((T_SL) reduce 390) ((T_SPACESHIP) reduce 390) ((T_SR) reduce 390) (("^") reduce 390) (("|") reduce 390)) 25944 ((("%") shift) (("*") shift) (("+") reduce 389) (("-") reduce 389) ((".") reduce 389) (("/") shift) ((";") reduce 389) (("<") reduce 389) ((">") reduce 389) (("?") reduce 389) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 389) ((T_BOOLEAN_AND) reduce 389) ((T_BOOLEAN_OR) reduce 389) ((T_COALESCE) reduce 389) ((T_DOUBLE_ARROW) reduce 389) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 389) ((T_IS_GREATER_OR_EQUAL) reduce 389) ((T_IS_IDENTICAL) reduce 389) ((T_IS_NOT_EQUAL) reduce 389) ((T_IS_NOT_IDENTICAL) reduce 389) ((T_IS_SMALLER_OR_EQUAL) reduce 389) ((T_LOGICAL_AND) reduce 389) ((T_LOGICAL_OR) reduce 389) ((T_LOGICAL_XOR) reduce 389) ((T_POW) shift) ((T_SL) reduce 389) ((T_SPACESHIP) reduce 389) ((T_SR) reduce 389) (("^") reduce 389) (("|") reduce 389)) 25945 ((("%") reduce 391) (("*") reduce 391) (("+") reduce 391) (("-") reduce 391) ((".") reduce 391) (("/") reduce 391) ((";") reduce 391) (("<") reduce 391) ((">") reduce 391) (("?") reduce 391) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 391) ((T_BOOLEAN_AND) reduce 391) ((T_BOOLEAN_OR) reduce 391) ((T_COALESCE) reduce 391) ((T_DOUBLE_ARROW) reduce 391) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 391) ((T_IS_GREATER_OR_EQUAL) reduce 391) ((T_IS_IDENTICAL) reduce 391) ((T_IS_NOT_EQUAL) reduce 391) ((T_IS_NOT_IDENTICAL) reduce 391) ((T_IS_SMALLER_OR_EQUAL) reduce 391) ((T_LOGICAL_AND) reduce 391) ((T_LOGICAL_OR) reduce 391) ((T_LOGICAL_XOR) reduce 391) ((T_POW) shift) ((T_SL) reduce 391) ((T_SPACESHIP) reduce 391) ((T_SR) reduce 391) (("^") reduce 391) (("|") reduce 391)) 25946 ((("%") reduce 394) (("*") reduce 394) (("+") reduce 394) (("-") reduce 394) ((".") reduce 394) (("/") reduce 394) ((";") reduce 394) (("<") reduce 394) ((">") reduce 394) (("?") reduce 394) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 394) ((T_BOOLEAN_AND) reduce 394) ((T_BOOLEAN_OR) reduce 394) ((T_COALESCE) reduce 394) ((T_DOUBLE_ARROW) reduce 394) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 394) ((T_IS_GREATER_OR_EQUAL) reduce 394) ((T_IS_IDENTICAL) reduce 394) ((T_IS_NOT_EQUAL) reduce 394) ((T_IS_NOT_IDENTICAL) reduce 394) ((T_IS_SMALLER_OR_EQUAL) reduce 394) ((T_LOGICAL_AND) reduce 394) ((T_LOGICAL_OR) reduce 394) ((T_LOGICAL_XOR) reduce 394) ((T_POW) shift) ((T_SL) reduce 394) ((T_SPACESHIP) reduce 394) ((T_SR) reduce 394) (("^") reduce 394) (("|") reduce 394)) 25948 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 25949 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 25950 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 25951 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 25952 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 25953 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 25954 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 25955 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 25956 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 25957 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 25958 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 25959 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 25960 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 25961 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 25962 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 25963 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 25964 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 25965 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 25966 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 25967 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 25968 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 25969 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 25970 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 25971 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 25972 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 25973 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 25974 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 25975 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 25976 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 25977 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 25978 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 25979 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 25980 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 25981 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 25982 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 25983 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 25984 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 25985 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 25986 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 25987 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 25988 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 25989 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 25990 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 25991 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 25992 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 25993 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 25994 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 25995 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 25996 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 25997 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 25998 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 25999 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 26000 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 26001 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 26002 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 26003 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 26004 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 26005 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 26006 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 26007 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 26008 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 26009 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 26010 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 26011 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 26012 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 26013 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 26014 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 26015 ((("%") reduce 81) (("(") reduce 81) (("*") reduce 81) (("+") reduce 81) (("-") reduce 81) ((".") reduce 81) (("/") reduce 81) ((";") reduce 81) (("<") reduce 81) ((">") reduce 81) (("?") reduce 81) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 81) ((T_BOOLEAN_AND) reduce 81) ((T_BOOLEAN_OR) reduce 81) ((T_COALESCE) reduce 81) ((T_DOUBLE_ARROW) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_IS_EQUAL) reduce 81) ((T_IS_GREATER_OR_EQUAL) reduce 81) ((T_IS_IDENTICAL) reduce 81) ((T_IS_NOT_EQUAL) reduce 81) ((T_IS_NOT_IDENTICAL) reduce 81) ((T_IS_SMALLER_OR_EQUAL) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 81) ((T_OBJECT_OPERATOR) reduce 81) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 81) ((T_POW) reduce 81) ((T_SL) reduce 81) ((T_SPACESHIP) reduce 81) ((T_SR) reduce 81) (("[") reduce 81) (("^") reduce 81) (("{") reduce 81) (("|") reduce 81)) 26016 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 26017 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 26018 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 26019 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 26020 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 26021 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 26022 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 26023 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 26024 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 26025 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 26026 ((("%") reduce 492) (("(") reduce 527) (("*") reduce 492) (("+") reduce 492) (("-") reduce 492) ((".") reduce 492) (("/") reduce 492) ((";") reduce 492) (("<") reduce 492) ((">") reduce 492) (("?") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_BOOLEAN_AND) reduce 492) ((T_BOOLEAN_OR) reduce 492) ((T_COALESCE) reduce 492) ((T_DOUBLE_ARROW) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_IS_EQUAL) reduce 492) ((T_IS_GREATER_OR_EQUAL) reduce 492) ((T_IS_IDENTICAL) reduce 492) ((T_IS_NOT_EQUAL) reduce 492) ((T_IS_NOT_IDENTICAL) reduce 492) ((T_IS_SMALLER_OR_EQUAL) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 492) ((T_OBJECT_OPERATOR) reduce 492) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 492) ((T_POW) reduce 492) ((T_SL) reduce 492) ((T_SPACESHIP) reduce 492) ((T_SR) reduce 492) (("[") reduce 492) (("^") reduce 492) (("{") reduce 492) (("|") reduce 492)) 26028 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 26029 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 26030 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) (("=") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AND_EQUAL) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_COALESCE_EQUAL) reduce 519) ((T_CONCAT_EQUAL) reduce 519) ((T_DEC) reduce 519) ((T_DIV_EQUAL) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INC) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_MINUS_EQUAL) reduce 519) ((T_MOD_EQUAL) reduce 519) ((T_MUL_EQUAL) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_OR_EQUAL) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PLUS_EQUAL) reduce 519) ((T_POW) reduce 519) ((T_POW_EQUAL) reduce 519) ((T_SL) reduce 519) ((T_SL_EQUAL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) ((T_SR_EQUAL) reduce 519) ((T_XOR_EQUAL) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 26032 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) (("=") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AND_EQUAL) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_COALESCE_EQUAL) reduce 456) ((T_CONCAT_EQUAL) reduce 456) ((T_DEC) reduce 456) ((T_DIV_EQUAL) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INC) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_MINUS_EQUAL) reduce 456) ((T_MOD_EQUAL) reduce 456) ((T_MUL_EQUAL) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_OR_EQUAL) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_PLUS_EQUAL) reduce 456) ((T_POW) reduce 456) ((T_POW_EQUAL) reduce 456) ((T_SL) reduce 456) ((T_SL_EQUAL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) ((T_SR_EQUAL) reduce 456) ((T_XOR_EQUAL) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 26033 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((";") reduce 281) (("<") reduce 281) (("=") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AND_EQUAL) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_COALESCE_EQUAL) reduce 281) ((T_CONCAT_EQUAL) reduce 281) ((T_DEC) reduce 281) ((T_DIV_EQUAL) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INC) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MINUS_EQUAL) reduce 281) ((T_MOD_EQUAL) reduce 281) ((T_MUL_EQUAL) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_OR_EQUAL) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_PLUS_EQUAL) reduce 281) ((T_POW) reduce 281) ((T_POW_EQUAL) reduce 281) ((T_SL) reduce 281) ((T_SL_EQUAL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) ((T_SR_EQUAL) reduce 281) ((T_XOR_EQUAL) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 26037 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((";") reduce 282) (("<") reduce 282) (("=") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AND_EQUAL) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_COALESCE_EQUAL) reduce 282) ((T_CONCAT_EQUAL) reduce 282) ((T_DEC) reduce 282) ((T_DIV_EQUAL) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INC) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MINUS_EQUAL) reduce 282) ((T_MOD_EQUAL) reduce 282) ((T_MUL_EQUAL) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_OR_EQUAL) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_PLUS_EQUAL) reduce 282) ((T_POW) reduce 282) ((T_POW_EQUAL) reduce 282) ((T_SL) reduce 282) ((T_SL_EQUAL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) ((T_SR_EQUAL) reduce 282) ((T_XOR_EQUAL) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 26038 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((";") reduce 283) (("<") reduce 283) (("=") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AND_EQUAL) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_COALESCE_EQUAL) reduce 283) ((T_CONCAT_EQUAL) reduce 283) ((T_DEC) reduce 283) ((T_DIV_EQUAL) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INC) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_MINUS_EQUAL) reduce 283) ((T_MOD_EQUAL) reduce 283) ((T_MUL_EQUAL) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_OR_EQUAL) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_PLUS_EQUAL) reduce 283) ((T_POW) reduce 283) ((T_POW_EQUAL) reduce 283) ((T_SL) reduce 283) ((T_SL_EQUAL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) ((T_SR_EQUAL) reduce 283) ((T_XOR_EQUAL) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 26043 ((("%") reduce 377) (("*") reduce 377) (("+") reduce 377) (("-") reduce 377) ((".") reduce 377) (("/") reduce 377) ((";") reduce 377) (("<") reduce 377) ((">") reduce 377) (("?") reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_BOOLEAN_AND) reduce 377) ((T_BOOLEAN_OR) reduce 377) ((T_COALESCE) reduce 377) ((T_DOUBLE_ARROW) reduce 377) ((T_INSTANCEOF) reduce 377) ((T_IS_EQUAL) reduce 377) ((T_IS_GREATER_OR_EQUAL) reduce 377) ((T_IS_IDENTICAL) reduce 377) ((T_IS_NOT_EQUAL) reduce 377) ((T_IS_NOT_IDENTICAL) reduce 377) ((T_IS_SMALLER_OR_EQUAL) reduce 377) ((T_LOGICAL_AND) reduce 377) ((T_LOGICAL_OR) reduce 377) ((T_LOGICAL_XOR) reduce 377) ((T_POW) reduce 377) ((T_SL) reduce 377) ((T_SPACESHIP) reduce 377) ((T_SR) reduce 377) (("^") reduce 377) (("|") reduce 377)) 26045 ((("%") reduce 375) (("*") reduce 375) (("+") reduce 375) (("-") reduce 375) ((".") reduce 375) (("/") reduce 375) ((";") reduce 375) (("<") reduce 375) ((">") reduce 375) (("?") reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_BOOLEAN_AND) reduce 375) ((T_BOOLEAN_OR) reduce 375) ((T_COALESCE) reduce 375) ((T_DOUBLE_ARROW) reduce 375) ((T_INSTANCEOF) reduce 375) ((T_IS_EQUAL) reduce 375) ((T_IS_GREATER_OR_EQUAL) reduce 375) ((T_IS_IDENTICAL) reduce 375) ((T_IS_NOT_EQUAL) reduce 375) ((T_IS_NOT_IDENTICAL) reduce 375) ((T_IS_SMALLER_OR_EQUAL) reduce 375) ((T_LOGICAL_AND) reduce 375) ((T_LOGICAL_OR) reduce 375) ((T_LOGICAL_XOR) reduce 375) ((T_POW) reduce 375) ((T_SL) reduce 375) ((T_SPACESHIP) reduce 375) ((T_SR) reduce 375) (("^") reduce 375) (("|") reduce 375)) 26055 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 371) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 371) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 371) ((T_LOGICAL_OR) reduce 371) ((T_LOGICAL_XOR) reduce 371) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26056 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 373) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 373) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 373) ((T_LOGICAL_OR) reduce 373) ((T_LOGICAL_XOR) reduce 373) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26057 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 372) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 372) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 372) ((T_LOGICAL_OR) reduce 372) ((T_LOGICAL_XOR) reduce 372) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26058 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 365) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 365) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 365) ((T_LOGICAL_OR) reduce 365) ((T_LOGICAL_XOR) reduce 365) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26059 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 362) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 362) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 362) ((T_LOGICAL_OR) reduce 362) ((T_LOGICAL_XOR) reduce 362) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26060 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 370) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 370) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 370) ((T_LOGICAL_OR) reduce 370) ((T_LOGICAL_XOR) reduce 370) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26061 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 364) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 364) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 364) ((T_LOGICAL_OR) reduce 364) ((T_LOGICAL_XOR) reduce 364) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26062 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 368) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 368) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 368) ((T_LOGICAL_OR) reduce 368) ((T_LOGICAL_XOR) reduce 368) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26063 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 363) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 363) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 363) ((T_LOGICAL_OR) reduce 363) ((T_LOGICAL_XOR) reduce 363) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26064 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 366) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 366) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 366) ((T_LOGICAL_OR) reduce 366) ((T_LOGICAL_XOR) reduce 366) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26065 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 367) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 367) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 367) ((T_LOGICAL_OR) reduce 367) ((T_LOGICAL_XOR) reduce 367) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26066 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 374) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 374) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 374) ((T_LOGICAL_OR) reduce 374) ((T_LOGICAL_XOR) reduce 374) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26067 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 369) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 369) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 369) ((T_LOGICAL_OR) reduce 369) ((T_LOGICAL_XOR) reduce 369) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26069 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 359) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 359) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26071 ((("%") reduce 515) (("(") reduce 515) (("*") reduce 515) (("+") reduce 515) (("-") reduce 515) ((".") reduce 515) (("/") reduce 515) ((";") reduce 515) (("<") reduce 515) ((">") reduce 515) (("?") reduce 515) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 515) ((T_BOOLEAN_AND) reduce 515) ((T_BOOLEAN_OR) reduce 515) ((T_COALESCE) reduce 515) ((T_DOUBLE_ARROW) reduce 515) ((T_INSTANCEOF) reduce 515) ((T_IS_EQUAL) reduce 515) ((T_IS_GREATER_OR_EQUAL) reduce 515) ((T_IS_IDENTICAL) reduce 515) ((T_IS_NOT_EQUAL) reduce 515) ((T_IS_NOT_IDENTICAL) reduce 515) ((T_IS_SMALLER_OR_EQUAL) reduce 515) ((T_LOGICAL_AND) reduce 515) ((T_LOGICAL_OR) reduce 515) ((T_LOGICAL_XOR) reduce 515) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 515) ((T_OBJECT_OPERATOR) reduce 515) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 515) ((T_POW) reduce 515) ((T_SL) reduce 515) ((T_SPACESHIP) reduce 515) ((T_SR) reduce 515) (("[") reduce 515) (("^") reduce 515) (("{") reduce 515) (("|") reduce 515)) 26074 ((("%") reduce 511) (("(") reduce 502) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 511) ((T_OBJECT_OPERATOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("[") reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 26076 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 510) ((T_OBJECT_OPERATOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("[") reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 26078 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 26079 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 512) ((T_OBJECT_OPERATOR) reduce 512) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") reduce 512) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 26080 ((("%") reduce 360) (("*") reduce 360) (("+") reduce 360) (("-") reduce 360) ((".") reduce 360) (("/") reduce 360) ((";") reduce 360) (("<") reduce 360) ((">") reduce 360) (("?") reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_BOOLEAN_AND) reduce 360) ((T_BOOLEAN_OR) reduce 360) ((T_COALESCE) reduce 360) ((T_DOUBLE_ARROW) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_IS_EQUAL) reduce 360) ((T_IS_GREATER_OR_EQUAL) reduce 360) ((T_IS_IDENTICAL) reduce 360) ((T_IS_NOT_EQUAL) reduce 360) ((T_IS_NOT_IDENTICAL) reduce 360) ((T_IS_SMALLER_OR_EQUAL) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 360) ((T_SL) reduce 360) ((T_SPACESHIP) reduce 360) ((T_SR) reduce 360) (("[") reduce 496) (("^") reduce 360) (("{") reduce 496) (("|") reduce 360)) 26084 ((("%") reduce 519) (("(") reduce 529) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 26086 ((("%") reduce 456) (("(") reduce 456) (("*") reduce 456) (("+") reduce 456) (("-") reduce 456) ((".") reduce 456) (("/") reduce 456) ((";") reduce 456) (("<") reduce 456) ((">") reduce 456) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) reduce 456) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) reduce 456) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) reduce 456) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 456) ((T_OBJECT_OPERATOR) reduce 456) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 456) ((T_POW) reduce 456) ((T_SL) reduce 456) ((T_SPACESHIP) reduce 456) ((T_SR) reduce 456) (("[") reduce 456) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 26087 ((("%") reduce 281) (("(") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((";") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 281) ((T_OBJECT_OPERATOR) reduce 281) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("[") reduce 281) (("^") reduce 281) (("{") reduce 281) (("|") reduce 281)) 26091 ((("%") reduce 282) (("(") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((";") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 282) ((T_OBJECT_OPERATOR) reduce 282) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("[") reduce 282) (("^") reduce 282) (("{") reduce 282) (("|") reduce 282)) 26092 ((("%") reduce 283) (("(") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((";") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 283) ((T_OBJECT_OPERATOR) reduce 283) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("[") reduce 283) (("^") reduce 283) (("{") reduce 283) (("|") reduce 283)) 26093 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_POW) reduce 454) ((T_SL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 26096 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 26097 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_POW) reduce 455) ((T_SL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 26098 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_POW) reduce 457) ((T_SL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 26104 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 26106 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 26107 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 26108 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 26109 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 26112 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 26113 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 26114 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_POW) reduce 514) ((T_SL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 26115 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 26116 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 26119 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_POW) reduce 516) ((T_SL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 26120 ((("%") reduce 454) (("(") reduce 454) (("*") reduce 454) (("+") reduce 454) (("-") reduce 454) ((".") reduce 454) (("/") reduce 454) ((";") reduce 454) (("<") reduce 454) (("=") reduce 454) ((">") reduce 454) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AND_EQUAL) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_COALESCE_EQUAL) reduce 454) ((T_CONCAT_EQUAL) reduce 454) ((T_DEC) reduce 454) ((T_DIV_EQUAL) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INC) reduce 454) ((T_INSTANCEOF) reduce 454) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) reduce 454) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) reduce 454) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_MINUS_EQUAL) reduce 454) ((T_MOD_EQUAL) reduce 454) ((T_MUL_EQUAL) reduce 454) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 454) ((T_OBJECT_OPERATOR) reduce 454) ((T_OR_EQUAL) reduce 454) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 454) ((T_PLUS_EQUAL) reduce 454) ((T_POW) reduce 454) ((T_POW_EQUAL) reduce 454) ((T_SL) reduce 454) ((T_SL_EQUAL) reduce 454) ((T_SPACESHIP) reduce 454) ((T_SR) reduce 454) ((T_SR_EQUAL) reduce 454) ((T_XOR_EQUAL) reduce 454) (("[") reduce 454) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 26131 ((("%") reduce 444) (("*") reduce 444) (("+") reduce 444) (("-") reduce 444) ((".") reduce 444) (("/") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) reduce 444) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_POW) reduce 444) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 26132 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) reduce 439) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_POW) reduce 439) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 26142 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 444) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26143 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_POW) reduce 440) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 26145 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26147 ((("%") reduce 491) (("(") reduce 527) (("*") reduce 491) (("+") reduce 491) (("-") reduce 491) ((".") reduce 491) (("/") reduce 491) ((";") reduce 491) (("<") reduce 491) ((">") reduce 491) (("?") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_BOOLEAN_AND) reduce 491) ((T_BOOLEAN_OR) reduce 491) ((T_COALESCE) reduce 491) ((T_DOUBLE_ARROW) reduce 491) ((T_INSTANCEOF) reduce 491) ((T_IS_EQUAL) reduce 491) ((T_IS_GREATER_OR_EQUAL) reduce 491) ((T_IS_IDENTICAL) reduce 491) ((T_IS_NOT_EQUAL) reduce 491) ((T_IS_NOT_IDENTICAL) reduce 491) ((T_IS_SMALLER_OR_EQUAL) reduce 491) ((T_LOGICAL_AND) reduce 491) ((T_LOGICAL_OR) reduce 491) ((T_LOGICAL_XOR) reduce 491) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 491) ((T_OBJECT_OPERATOR) reduce 491) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 491) ((T_POW) reduce 491) ((T_SL) reduce 491) ((T_SPACESHIP) reduce 491) ((T_SR) reduce 491) (("[") reduce 491) (("^") reduce 491) (("{") reduce 491) (("|") reduce 491)) 26149 ((("%") reduce 518) (("(") reduce 529) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) (("=") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AND_EQUAL) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_COALESCE_EQUAL) reduce 518) ((T_CONCAT_EQUAL) reduce 518) ((T_DEC) reduce 518) ((T_DIV_EQUAL) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INC) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_MINUS_EQUAL) reduce 518) ((T_MOD_EQUAL) reduce 518) ((T_MUL_EQUAL) reduce 518) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 518) ((T_OBJECT_OPERATOR) reduce 518) ((T_OR_EQUAL) reduce 518) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 518) ((T_PLUS_EQUAL) reduce 518) ((T_POW) reduce 518) ((T_POW_EQUAL) reduce 518) ((T_SL) reduce 518) ((T_SL_EQUAL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) ((T_SR_EQUAL) reduce 518) ((T_XOR_EQUAL) reduce 518) (("[") reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 26150 ((("%") reduce 455) (("(") reduce 455) (("*") reduce 455) (("+") reduce 455) (("-") reduce 455) ((".") reduce 455) (("/") reduce 455) ((";") reduce 455) (("<") reduce 455) (("=") reduce 455) ((">") reduce 455) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AND_EQUAL) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_COALESCE_EQUAL) reduce 455) ((T_CONCAT_EQUAL) reduce 455) ((T_DEC) reduce 455) ((T_DIV_EQUAL) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INC) reduce 455) ((T_INSTANCEOF) reduce 455) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) reduce 455) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) reduce 455) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_MINUS_EQUAL) reduce 455) ((T_MOD_EQUAL) reduce 455) ((T_MUL_EQUAL) reduce 455) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 455) ((T_OBJECT_OPERATOR) reduce 455) ((T_OR_EQUAL) reduce 455) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 455) ((T_PLUS_EQUAL) reduce 455) ((T_POW) reduce 455) ((T_POW_EQUAL) reduce 455) ((T_SL) reduce 455) ((T_SL_EQUAL) reduce 455) ((T_SPACESHIP) reduce 455) ((T_SR) reduce 455) ((T_SR_EQUAL) reduce 455) ((T_XOR_EQUAL) reduce 455) (("[") reduce 455) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 26151 ((("%") reduce 457) (("(") reduce 457) (("*") reduce 457) (("+") reduce 457) (("-") reduce 457) ((".") reduce 457) (("/") reduce 457) ((";") reduce 457) (("<") reduce 457) (("=") reduce 457) ((">") reduce 457) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AND_EQUAL) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_COALESCE_EQUAL) reduce 457) ((T_CONCAT_EQUAL) reduce 457) ((T_DEC) reduce 457) ((T_DIV_EQUAL) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INC) reduce 457) ((T_INSTANCEOF) reduce 457) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) reduce 457) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) reduce 457) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_MINUS_EQUAL) reduce 457) ((T_MOD_EQUAL) reduce 457) ((T_MUL_EQUAL) reduce 457) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 457) ((T_OBJECT_OPERATOR) reduce 457) ((T_OR_EQUAL) reduce 457) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 457) ((T_PLUS_EQUAL) reduce 457) ((T_POW) reduce 457) ((T_POW_EQUAL) reduce 457) ((T_SL) reduce 457) ((T_SL_EQUAL) reduce 457) ((T_SPACESHIP) reduce 457) ((T_SR) reduce 457) ((T_SR_EQUAL) reduce 457) ((T_XOR_EQUAL) reduce 457) (("[") reduce 457) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 26153 ((("%") reduce 435) (("*") reduce 435) (("+") reduce 435) (("-") reduce 435) ((".") reduce 435) (("/") reduce 435) ((";") reduce 435) (("<") reduce 435) ((">") reduce 435) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) reduce 435) ((T_IS_EQUAL) reduce 435) ((T_IS_GREATER_OR_EQUAL) reduce 435) ((T_IS_IDENTICAL) reduce 435) ((T_IS_NOT_EQUAL) reduce 435) ((T_IS_NOT_IDENTICAL) reduce 435) ((T_IS_SMALLER_OR_EQUAL) reduce 435) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_POW) reduce 435) ((T_SL) reduce 435) ((T_SPACESHIP) reduce 435) ((T_SR) reduce 435) (("^") reduce 435) (("|") reduce 435)) 26154 ((("%") reduce 437) (("*") reduce 437) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") reduce 437) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) reduce 437) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_POW) reduce 437) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 26160 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 26162 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 26163 ((("%") reduce 530) (("(") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) (("=") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AND_EQUAL) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_COALESCE_EQUAL) reduce 530) ((T_CONCAT_EQUAL) reduce 530) ((T_DEC) reduce 530) ((T_DIV_EQUAL) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INC) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_MINUS_EQUAL) reduce 530) ((T_MOD_EQUAL) reduce 530) ((T_MUL_EQUAL) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 530) ((T_OBJECT_OPERATOR) reduce 530) ((T_OR_EQUAL) reduce 530) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 530) ((T_PLUS_EQUAL) reduce 530) ((T_POW) reduce 530) ((T_POW_EQUAL) reduce 530) ((T_SL) reduce 530) ((T_SL_EQUAL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) ((T_SR_EQUAL) reduce 530) ((T_XOR_EQUAL) reduce 530) (("[") reduce 530) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 26164 ((("%") reduce 513) (("(") shift) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) (("=") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AND_EQUAL) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_COALESCE_EQUAL) reduce 513) ((T_CONCAT_EQUAL) reduce 513) ((T_DEC) reduce 513) ((T_DIV_EQUAL) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INC) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_MINUS_EQUAL) reduce 513) ((T_MOD_EQUAL) reduce 513) ((T_MUL_EQUAL) reduce 513) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 513) ((T_OBJECT_OPERATOR) reduce 513) ((T_OR_EQUAL) reduce 513) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 513) ((T_PLUS_EQUAL) reduce 513) ((T_POW) reduce 513) ((T_POW_EQUAL) reduce 513) ((T_SL) reduce 513) ((T_SL_EQUAL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) ((T_SR_EQUAL) reduce 513) ((T_XOR_EQUAL) reduce 513) (("[") reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 26165 ((("%") reduce 532) (("(") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) (("=") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AND_EQUAL) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_COALESCE_EQUAL) reduce 532) ((T_CONCAT_EQUAL) reduce 532) ((T_DEC) reduce 532) ((T_DIV_EQUAL) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INC) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_MINUS_EQUAL) reduce 532) ((T_MOD_EQUAL) reduce 532) ((T_MUL_EQUAL) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_OR_EQUAL) reduce 532) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 532) ((T_PLUS_EQUAL) reduce 532) ((T_POW) reduce 532) ((T_POW_EQUAL) reduce 532) ((T_SL) reduce 532) ((T_SL_EQUAL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) ((T_SR_EQUAL) reduce 532) ((T_XOR_EQUAL) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 26168 ((("%") reduce 531) (("(") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) (("=") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AND_EQUAL) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_COALESCE_EQUAL) reduce 531) ((T_CONCAT_EQUAL) reduce 531) ((T_DEC) reduce 531) ((T_DIV_EQUAL) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INC) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_MINUS_EQUAL) reduce 531) ((T_MOD_EQUAL) reduce 531) ((T_MUL_EQUAL) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_OR_EQUAL) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 531) ((T_PLUS_EQUAL) reduce 531) ((T_POW) reduce 531) ((T_POW_EQUAL) reduce 531) ((T_SL) reduce 531) ((T_SL_EQUAL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) ((T_SR_EQUAL) reduce 531) ((T_XOR_EQUAL) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 26169 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 26170 ((("%") reduce 514) (("(") shift) (("*") reduce 514) (("+") reduce 514) (("-") reduce 514) ((".") reduce 514) (("/") reduce 514) ((";") reduce 514) (("<") reduce 514) (("=") reduce 514) ((">") reduce 514) (("?") reduce 514) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 514) ((T_AND_EQUAL) reduce 514) ((T_BOOLEAN_AND) reduce 514) ((T_BOOLEAN_OR) reduce 514) ((T_COALESCE) reduce 514) ((T_COALESCE_EQUAL) reduce 514) ((T_CONCAT_EQUAL) reduce 514) ((T_DEC) reduce 514) ((T_DIV_EQUAL) reduce 514) ((T_DOUBLE_ARROW) reduce 514) ((T_INC) reduce 514) ((T_INSTANCEOF) reduce 514) ((T_IS_EQUAL) reduce 514) ((T_IS_GREATER_OR_EQUAL) reduce 514) ((T_IS_IDENTICAL) reduce 514) ((T_IS_NOT_EQUAL) reduce 514) ((T_IS_NOT_IDENTICAL) reduce 514) ((T_IS_SMALLER_OR_EQUAL) reduce 514) ((T_LOGICAL_AND) reduce 514) ((T_LOGICAL_OR) reduce 514) ((T_LOGICAL_XOR) reduce 514) ((T_MINUS_EQUAL) reduce 514) ((T_MOD_EQUAL) reduce 514) ((T_MUL_EQUAL) reduce 514) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 514) ((T_OBJECT_OPERATOR) reduce 514) ((T_OR_EQUAL) reduce 514) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 514) ((T_PLUS_EQUAL) reduce 514) ((T_POW) reduce 514) ((T_POW_EQUAL) reduce 514) ((T_SL) reduce 514) ((T_SL_EQUAL) reduce 514) ((T_SPACESHIP) reduce 514) ((T_SR) reduce 514) ((T_SR_EQUAL) reduce 514) ((T_XOR_EQUAL) reduce 514) (("[") reduce 514) (("^") reduce 514) (("{") reduce 514) (("|") reduce 514)) 26171 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) (("=") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AND_EQUAL) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_COALESCE_EQUAL) reduce 509) ((T_CONCAT_EQUAL) reduce 509) ((T_DEC) reduce 509) ((T_DIV_EQUAL) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INC) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_MINUS_EQUAL) reduce 509) ((T_MOD_EQUAL) reduce 509) ((T_MUL_EQUAL) reduce 509) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 509) ((T_OBJECT_OPERATOR) reduce 509) ((T_OR_EQUAL) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PLUS_EQUAL) reduce 509) ((T_POW) reduce 509) ((T_POW_EQUAL) reduce 509) ((T_SL) reduce 509) ((T_SL_EQUAL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) ((T_SR_EQUAL) reduce 509) ((T_XOR_EQUAL) reduce 509) (("[") reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 26173 ((("%") reduce 427) (("*") reduce 427) (("+") reduce 427) (("-") reduce 427) ((".") reduce 427) (("/") reduce 427) ((";") reduce 427) (("<") reduce 427) ((">") reduce 427) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 427) ((T_BOOLEAN_AND) reduce 427) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) reduce 427) ((T_IS_EQUAL) reduce 427) ((T_IS_GREATER_OR_EQUAL) reduce 427) ((T_IS_IDENTICAL) reduce 427) ((T_IS_NOT_EQUAL) reduce 427) ((T_IS_NOT_IDENTICAL) reduce 427) ((T_IS_SMALLER_OR_EQUAL) reduce 427) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_POW) reduce 427) ((T_SL) reduce 427) ((T_SPACESHIP) reduce 427) ((T_SR) reduce 427) (("^") reduce 427) (("|") reduce 427)) 26175 ((("%") reduce 471) (("(") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) (("=") shift) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 471) ((T_OBJECT_OPERATOR) reduce 471) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 471) ((T_POW) reduce 471) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("[") reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 26177 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 358) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 358) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26178 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 432) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26181 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 431) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26182 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 26183 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26184 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) ((";") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 26185 ((("%") reduce 436) (("*") reduce 436) (("+") reduce 436) (("-") reduce 436) ((".") reduce 436) (("/") reduce 436) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) reduce 436) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_POW) reduce 436) ((T_SL) reduce 436) ((T_SPACESHIP) reduce 436) ((T_SR) reduce 436) (("^") reduce 436) (("|") reduce 436)) 26187 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 26189 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) ((";") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 26190 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 26191 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 567) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26192 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 566) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26193 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 428) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26194 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 26197 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 26198 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 26199 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 26200 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 26201 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_DOUBLE_ARROW) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 26202 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) ((";") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_DOUBLE_ARROW) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 26204 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 26205 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 26206 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 26207 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 26208 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 26215 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 26217 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 26218 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 26219 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 26220 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 26222 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 26223 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) ((";") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_DOUBLE_ARROW) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 26224 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((";") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_DOUBLE_ARROW) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 26228 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((";") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_DOUBLE_ARROW) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 26229 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((";") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_DOUBLE_ARROW) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 26231 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 26232 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) ((";") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_DOUBLE_ARROW) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 26239 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) ((";") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_DOUBLE_ARROW) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 26241 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 26247 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) ((";") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_DOUBLE_ARROW) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 26252 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 357) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 357) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26256 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 26257 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) ((";") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 26258 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 564) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26259 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 563) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26260 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) ((";") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_DOUBLE_ARROW) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 26262 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 26264 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) ((";") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 26267 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 26270 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 26271 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) ((";") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 26272 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) ((";") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_DOUBLE_ARROW) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 26273 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) ((";") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_DOUBLE_ARROW) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 26274 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 26275 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 26278 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 26279 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 26280 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 26281 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) ((";") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_DOUBLE_ARROW) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 26283 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) ((";") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 26284 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 26287 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_DOUBLE_ARROW) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 26289 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 26290 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 26291 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_POW) shift) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 26292 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26293 ((("%") reduce 419) (("*") reduce 419) (("+") reduce 419) (("-") reduce 419) ((".") reduce 419) (("/") reduce 419) ((";") reduce 419) (("<") reduce 419) ((">") reduce 419) (("?") reduce 419) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 419) ((T_BOOLEAN_AND) reduce 419) ((T_BOOLEAN_OR) reduce 419) ((T_COALESCE) reduce 419) ((T_INSTANCEOF) reduce 419) ((T_IS_EQUAL) reduce 419) ((T_IS_GREATER_OR_EQUAL) reduce 419) ((T_IS_IDENTICAL) reduce 419) ((T_IS_NOT_EQUAL) reduce 419) ((T_IS_NOT_IDENTICAL) reduce 419) ((T_IS_SMALLER_OR_EQUAL) reduce 419) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_POW) shift) ((T_SL) reduce 419) ((T_SPACESHIP) reduce 419) ((T_SR) reduce 419) (("^") reduce 419) (("|") reduce 419)) 26296 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 26298 ((("%") reduce 478) (("*") reduce 478) (("+") reduce 478) (("-") reduce 478) ((".") reduce 478) (("/") reduce 478) ((";") reduce 478) (("<") reduce 478) ((">") reduce 478) (("?") reduce 478) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 478) ((T_BOOLEAN_AND) reduce 478) ((T_BOOLEAN_OR) reduce 478) ((T_COALESCE) reduce 478) ((T_INSTANCEOF) reduce 478) ((T_IS_EQUAL) reduce 478) ((T_IS_GREATER_OR_EQUAL) reduce 478) ((T_IS_IDENTICAL) reduce 478) ((T_IS_NOT_EQUAL) reduce 478) ((T_IS_NOT_IDENTICAL) reduce 478) ((T_IS_SMALLER_OR_EQUAL) reduce 478) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_POW) reduce 478) ((T_SL) reduce 478) ((T_SPACESHIP) reduce 478) ((T_SR) reduce 478) (("^") reduce 478) (("|") reduce 478)) 26299 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 26300 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 567) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26301 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 566) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26302 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 428) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26303 ((("%") reduce 421) (("*") reduce 421) (("+") reduce 421) (("-") reduce 421) ((".") reduce 421) (("/") reduce 421) ((";") reduce 421) (("<") reduce 421) ((">") reduce 421) (("?") reduce 421) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 421) ((T_BOOLEAN_AND) reduce 421) ((T_BOOLEAN_OR) reduce 421) ((T_COALESCE) reduce 421) ((T_INSTANCEOF) reduce 421) ((T_IS_EQUAL) reduce 421) ((T_IS_GREATER_OR_EQUAL) reduce 421) ((T_IS_IDENTICAL) reduce 421) ((T_IS_NOT_EQUAL) reduce 421) ((T_IS_NOT_IDENTICAL) reduce 421) ((T_IS_SMALLER_OR_EQUAL) reduce 421) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_POW) shift) ((T_SL) reduce 421) ((T_SPACESHIP) reduce 421) ((T_SR) reduce 421) (("^") reduce 421) (("|") reduce 421)) 26306 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 26307 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 26308 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 26309 ((("%") reduce 458) (("(") reduce 458) (("*") reduce 458) (("+") reduce 458) (("-") reduce 458) ((".") reduce 458) (("/") reduce 458) ((";") reduce 458) (("<") reduce 458) ((">") reduce 458) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) reduce 458) ((T_IS_EQUAL) reduce 458) ((T_IS_GREATER_OR_EQUAL) reduce 458) ((T_IS_IDENTICAL) reduce 458) ((T_IS_NOT_EQUAL) reduce 458) ((T_IS_NOT_IDENTICAL) reduce 458) ((T_IS_SMALLER_OR_EQUAL) reduce 458) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 458) ((T_POW) reduce 458) ((T_SL) reduce 458) ((T_SPACESHIP) reduce 458) ((T_SR) reduce 458) (("^") reduce 458) (("|") reduce 458)) 26310 ((("%") reduce 91) (("(") reduce 91) (("*") reduce 91) (("+") reduce 91) (("-") reduce 91) ((".") reduce 91) (("/") reduce 91) ((";") reduce 91) (("<") reduce 91) ((">") reduce 91) (("?") reduce 91) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 91) ((T_BOOLEAN_AND) reduce 91) ((T_BOOLEAN_OR) reduce 91) ((T_COALESCE) reduce 91) ((T_INSTANCEOF) reduce 91) ((T_IS_EQUAL) reduce 91) ((T_IS_GREATER_OR_EQUAL) reduce 91) ((T_IS_IDENTICAL) reduce 91) ((T_IS_NOT_EQUAL) reduce 91) ((T_IS_NOT_IDENTICAL) reduce 91) ((T_IS_SMALLER_OR_EQUAL) reduce 91) ((T_LOGICAL_AND) reduce 91) ((T_LOGICAL_OR) reduce 91) ((T_LOGICAL_XOR) reduce 91) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 91) ((T_POW) reduce 91) ((T_SL) reduce 91) ((T_SPACESHIP) reduce 91) ((T_SR) reduce 91) (("^") reduce 91) (("|") reduce 91)) 26311 ((("%") reduce 354) (("*") reduce 354) (("+") reduce 354) (("-") reduce 354) ((".") reduce 354) (("/") reduce 354) ((";") reduce 354) (("<") reduce 354) ((">") reduce 354) (("?") reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_BOOLEAN_AND) reduce 354) ((T_BOOLEAN_OR) reduce 354) ((T_COALESCE) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_IS_EQUAL) reduce 354) ((T_IS_GREATER_OR_EQUAL) reduce 354) ((T_IS_IDENTICAL) reduce 354) ((T_IS_NOT_EQUAL) reduce 354) ((T_IS_NOT_IDENTICAL) reduce 354) ((T_IS_SMALLER_OR_EQUAL) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_POW) reduce 354) ((T_SL) reduce 354) ((T_SPACESHIP) reduce 354) ((T_SR) reduce 354) (("^") reduce 354) (("|") reduce 354)) 26313 ((("%") reduce 460) (("(") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("^") reduce 460) (("|") reduce 460)) 26314 ((("%") reduce 468) (("(") shift) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_POW) reduce 468) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 26315 ((("%") reduce 459) (("(") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 26316 ((("%") reduce 461) (("(") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") shift) (("^") reduce 461) (("{") shift) (("|") reduce 461)) 26317 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 26324 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 26326 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 26327 ((("%") reduce 526) (("(") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 526) ((T_OBJECT_OPERATOR) reduce 526) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("[") reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 26328 ((("%") reduce 523) (("(") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 523) ((T_OBJECT_OPERATOR) reduce 523) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("[") reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 26329 ((("%") reduce 524) (("(") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 524) ((T_OBJECT_OPERATOR) reduce 524) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("[") reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 26331 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_POW) reduce 469) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 26332 ((("%") reduce 353) (("*") reduce 353) (("+") reduce 353) (("-") reduce 353) ((".") reduce 353) (("/") reduce 353) ((";") reduce 353) (("<") reduce 353) ((">") reduce 353) (("?") reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_BOOLEAN_AND) reduce 353) ((T_BOOLEAN_OR) reduce 353) ((T_COALESCE) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_IS_EQUAL) reduce 353) ((T_IS_GREATER_OR_EQUAL) reduce 353) ((T_IS_IDENTICAL) reduce 353) ((T_IS_NOT_EQUAL) reduce 353) ((T_IS_NOT_IDENTICAL) reduce 353) ((T_IS_SMALLER_OR_EQUAL) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_POW) reduce 353) ((T_SL) reduce 353) ((T_SPACESHIP) reduce 353) ((T_SR) reduce 353) (("^") reduce 353) (("|") reduce 353)) 26333 ((("%") reduce 281) (("*") reduce 281) (("+") reduce 281) (("-") reduce 281) ((".") reduce 281) (("/") reduce 281) ((";") reduce 281) (("<") reduce 281) ((">") reduce 281) (("?") reduce 281) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_BOOLEAN_AND) reduce 281) ((T_BOOLEAN_OR) reduce 281) ((T_COALESCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_IS_EQUAL) reduce 281) ((T_IS_GREATER_OR_EQUAL) reduce 281) ((T_IS_IDENTICAL) reduce 281) ((T_IS_NOT_EQUAL) reduce 281) ((T_IS_NOT_IDENTICAL) reduce 281) ((T_IS_SMALLER_OR_EQUAL) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_POW) reduce 281) ((T_SL) reduce 281) ((T_SPACESHIP) reduce 281) ((T_SR) reduce 281) (("^") reduce 281) (("|") reduce 281)) 26337 ((("%") reduce 282) (("*") reduce 282) (("+") reduce 282) (("-") reduce 282) ((".") reduce 282) (("/") reduce 282) ((";") reduce 282) (("<") reduce 282) ((">") reduce 282) (("?") reduce 282) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_BOOLEAN_AND) reduce 282) ((T_BOOLEAN_OR) reduce 282) ((T_COALESCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_IS_EQUAL) reduce 282) ((T_IS_GREATER_OR_EQUAL) reduce 282) ((T_IS_IDENTICAL) reduce 282) ((T_IS_NOT_EQUAL) reduce 282) ((T_IS_NOT_IDENTICAL) reduce 282) ((T_IS_SMALLER_OR_EQUAL) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_POW) reduce 282) ((T_SL) reduce 282) ((T_SPACESHIP) reduce 282) ((T_SR) reduce 282) (("^") reduce 282) (("|") reduce 282)) 26338 ((("%") reduce 283) (("*") reduce 283) (("+") reduce 283) (("-") reduce 283) ((".") reduce 283) (("/") reduce 283) ((";") reduce 283) (("<") reduce 283) ((">") reduce 283) (("?") reduce 283) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_BOOLEAN_AND) reduce 283) ((T_BOOLEAN_OR) reduce 283) ((T_COALESCE) reduce 283) ((T_INSTANCEOF) reduce 283) ((T_IS_EQUAL) reduce 283) ((T_IS_GREATER_OR_EQUAL) reduce 283) ((T_IS_IDENTICAL) reduce 283) ((T_IS_NOT_EQUAL) reduce 283) ((T_IS_NOT_IDENTICAL) reduce 283) ((T_IS_SMALLER_OR_EQUAL) reduce 283) ((T_LOGICAL_AND) reduce 283) ((T_LOGICAL_OR) reduce 283) ((T_LOGICAL_XOR) reduce 283) ((T_POW) reduce 283) ((T_SL) reduce 283) ((T_SPACESHIP) reduce 283) ((T_SR) reduce 283) (("^") reduce 283) (("|") reduce 283)) 26340 ((("%") reduce 525) (("(") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 525) ((T_OBJECT_OPERATOR) reduce 525) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("[") reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 26341 ((("%") reduce 355) (("*") reduce 355) (("+") reduce 355) (("-") reduce 355) ((".") reduce 355) (("/") reduce 355) ((";") reduce 355) (("<") reduce 355) ((">") reduce 355) (("?") reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_BOOLEAN_AND) reduce 355) ((T_BOOLEAN_OR) reduce 355) ((T_COALESCE) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_IS_EQUAL) reduce 355) ((T_IS_GREATER_OR_EQUAL) reduce 355) ((T_IS_IDENTICAL) reduce 355) ((T_IS_NOT_EQUAL) reduce 355) ((T_IS_NOT_IDENTICAL) reduce 355) ((T_IS_SMALLER_OR_EQUAL) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_POW) reduce 355) ((T_SL) reduce 355) ((T_SPACESHIP) reduce 355) ((T_SR) reduce 355) (("^") reduce 355) (("|") reduce 355)) 26348 ((("%") reduce 352) (("*") reduce 352) (("+") reduce 352) (("-") reduce 352) ((".") reduce 352) (("/") reduce 352) ((";") reduce 352) (("<") reduce 352) ((">") reduce 352) (("?") reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_BOOLEAN_AND) reduce 352) ((T_BOOLEAN_OR) reduce 352) ((T_COALESCE) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_IS_EQUAL) reduce 352) ((T_IS_GREATER_OR_EQUAL) reduce 352) ((T_IS_IDENTICAL) reduce 352) ((T_IS_NOT_EQUAL) reduce 352) ((T_IS_NOT_IDENTICAL) reduce 352) ((T_IS_SMALLER_OR_EQUAL) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_POW) reduce 352) ((T_SL) reduce 352) ((T_SPACESHIP) reduce 352) ((T_SR) reduce 352) (("^") reduce 352) (("|") reduce 352)) 26350 ((("%") reduce 462) (("(") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 26356 ((("%") reduce 223) (("*") reduce 223) (("+") reduce 223) (("-") reduce 223) ((".") reduce 223) (("/") reduce 223) ((";") reduce 223) (("<") reduce 223) ((">") reduce 223) (("?") reduce 223) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 223) ((T_BOOLEAN_AND) reduce 223) ((T_BOOLEAN_OR) reduce 223) ((T_COALESCE) reduce 223) ((T_INSTANCEOF) reduce 223) ((T_IS_EQUAL) reduce 223) ((T_IS_GREATER_OR_EQUAL) reduce 223) ((T_IS_IDENTICAL) reduce 223) ((T_IS_NOT_EQUAL) reduce 223) ((T_IS_NOT_IDENTICAL) reduce 223) ((T_IS_SMALLER_OR_EQUAL) reduce 223) ((T_LOGICAL_AND) reduce 223) ((T_LOGICAL_OR) reduce 223) ((T_LOGICAL_XOR) reduce 223) ((T_POW) reduce 223) ((T_SL) reduce 223) ((T_SPACESHIP) reduce 223) ((T_SR) reduce 223) (("^") reduce 223) (("|") reduce 223)) 26361 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 357) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26365 ((("%") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("^") reduce 561) (("|") reduce 561)) 26366 ((("%") reduce 417) (("*") reduce 417) (("+") reduce 417) (("-") reduce 417) ((".") reduce 417) (("/") reduce 417) ((";") reduce 417) (("<") reduce 417) ((">") reduce 417) (("?") reduce 417) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 417) ((T_BOOLEAN_AND) reduce 417) ((T_BOOLEAN_OR) reduce 417) ((T_COALESCE) reduce 417) ((T_INSTANCEOF) reduce 417) ((T_IS_EQUAL) reduce 417) ((T_IS_GREATER_OR_EQUAL) reduce 417) ((T_IS_IDENTICAL) reduce 417) ((T_IS_NOT_EQUAL) reduce 417) ((T_IS_NOT_IDENTICAL) reduce 417) ((T_IS_SMALLER_OR_EQUAL) reduce 417) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_POW) shift) ((T_SL) reduce 417) ((T_SPACESHIP) reduce 417) ((T_SR) reduce 417) (("^") reduce 417) (("|") reduce 417)) 26367 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 564) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26368 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 563) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26369 ((("%") reduce 376) (("*") reduce 376) (("+") reduce 376) (("-") reduce 376) ((".") reduce 376) (("/") reduce 376) ((";") reduce 376) (("<") reduce 376) ((">") reduce 376) (("?") reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_BOOLEAN_AND) reduce 376) ((T_BOOLEAN_OR) reduce 376) ((T_COALESCE) reduce 376) ((T_INSTANCEOF) reduce 376) ((T_IS_EQUAL) reduce 376) ((T_IS_GREATER_OR_EQUAL) reduce 376) ((T_IS_IDENTICAL) reduce 376) ((T_IS_NOT_EQUAL) reduce 376) ((T_IS_NOT_IDENTICAL) reduce 376) ((T_IS_SMALLER_OR_EQUAL) reduce 376) ((T_LOGICAL_AND) reduce 376) ((T_LOGICAL_OR) reduce 376) ((T_LOGICAL_XOR) reduce 376) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 376) ((T_SL) reduce 376) ((T_SPACESHIP) reduce 376) ((T_SR) reduce 376) (("[") reduce 496) (("^") reduce 376) (("{") reduce 496) (("|") reduce 376)) 26371 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("^") reduce 424) (("|") reduce 424)) 26373 ((("%") reduce 464) (("*") reduce 464) (("+") reduce 464) (("-") reduce 464) ((".") reduce 464) (("/") reduce 464) ((";") reduce 464) (("<") reduce 464) ((">") reduce 464) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 464) ((T_BOOLEAN_AND) reduce 464) ((T_BOOLEAN_OR) reduce 464) ((T_COALESCE) reduce 464) ((T_INSTANCEOF) reduce 464) ((T_IS_EQUAL) reduce 464) ((T_IS_GREATER_OR_EQUAL) reduce 464) ((T_IS_IDENTICAL) reduce 464) ((T_IS_NOT_EQUAL) reduce 464) ((T_IS_NOT_IDENTICAL) reduce 464) ((T_IS_SMALLER_OR_EQUAL) reduce 464) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_POW) reduce 464) ((T_SL) reduce 464) ((T_SPACESHIP) reduce 464) ((T_SR) reduce 464) (("^") reduce 464) (("|") reduce 464)) 26376 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("^") reduce 565) (("|") reduce 565)) 26379 ((("%") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("^") reduce 562) (("|") reduce 562)) 26380 ((("%") reduce 418) (("*") reduce 418) (("+") reduce 418) (("-") reduce 418) ((".") reduce 418) (("/") reduce 418) ((";") reduce 418) (("<") reduce 418) ((">") reduce 418) (("?") reduce 418) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 418) ((T_BOOLEAN_AND) reduce 418) ((T_BOOLEAN_OR) reduce 418) ((T_COALESCE) reduce 418) ((T_INSTANCEOF) reduce 418) ((T_IS_EQUAL) reduce 418) ((T_IS_GREATER_OR_EQUAL) reduce 418) ((T_IS_IDENTICAL) reduce 418) ((T_IS_NOT_EQUAL) reduce 418) ((T_IS_NOT_IDENTICAL) reduce 418) ((T_IS_SMALLER_OR_EQUAL) reduce 418) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_POW) shift) ((T_SL) reduce 418) ((T_SPACESHIP) reduce 418) ((T_SR) reduce 418) (("^") reduce 418) (("|") reduce 418)) 26381 ((("%") reduce 378) (("*") reduce 378) (("+") reduce 378) (("-") reduce 378) ((".") reduce 378) (("/") reduce 378) ((";") reduce 378) (("<") reduce 378) ((">") reduce 378) (("?") reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_BOOLEAN_AND) reduce 378) ((T_BOOLEAN_OR) reduce 378) ((T_COALESCE) reduce 378) ((T_INSTANCEOF) reduce 378) ((T_IS_EQUAL) reduce 378) ((T_IS_GREATER_OR_EQUAL) reduce 378) ((T_IS_IDENTICAL) reduce 378) ((T_IS_NOT_EQUAL) reduce 378) ((T_IS_NOT_IDENTICAL) reduce 378) ((T_IS_SMALLER_OR_EQUAL) reduce 378) ((T_LOGICAL_AND) reduce 378) ((T_LOGICAL_OR) reduce 378) ((T_LOGICAL_XOR) reduce 378) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 496) ((T_OBJECT_OPERATOR) reduce 496) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 496) ((T_POW) reduce 378) ((T_SL) reduce 378) ((T_SPACESHIP) reduce 378) ((T_SR) reduce 378) (("[") reduce 496) (("^") reduce 378) (("{") reduce 496) (("|") reduce 378)) 26382 ((("%") reduce 361) (("*") reduce 361) (("+") reduce 361) (("-") reduce 361) ((".") reduce 361) (("/") reduce 361) ((";") reduce 361) (("<") reduce 361) ((">") reduce 361) (("?") reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_BOOLEAN_AND) reduce 361) ((T_BOOLEAN_OR) reduce 361) ((T_COALESCE) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_IS_EQUAL) reduce 361) ((T_IS_GREATER_OR_EQUAL) reduce 361) ((T_IS_IDENTICAL) reduce 361) ((T_IS_NOT_EQUAL) reduce 361) ((T_IS_NOT_IDENTICAL) reduce 361) ((T_IS_SMALLER_OR_EQUAL) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_POW) reduce 361) ((T_SL) reduce 361) ((T_SPACESHIP) reduce 361) ((T_SR) reduce 361) (("^") reduce 361) (("|") reduce 361)) 26383 ((("%") reduce 422) (("*") reduce 422) (("+") reduce 422) (("-") reduce 422) ((".") reduce 422) (("/") reduce 422) ((";") reduce 422) (("<") reduce 422) ((">") reduce 422) (("?") reduce 422) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 422) ((T_BOOLEAN_AND) reduce 422) ((T_BOOLEAN_OR) reduce 422) ((T_COALESCE) reduce 422) ((T_INSTANCEOF) reduce 422) ((T_IS_EQUAL) reduce 422) ((T_IS_GREATER_OR_EQUAL) reduce 422) ((T_IS_IDENTICAL) reduce 422) ((T_IS_NOT_EQUAL) reduce 422) ((T_IS_NOT_IDENTICAL) reduce 422) ((T_IS_SMALLER_OR_EQUAL) reduce 422) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_POW) shift) ((T_SL) reduce 422) ((T_SPACESHIP) reduce 422) ((T_SR) reduce 422) (("^") reduce 422) (("|") reduce 422)) 26384 ((("%") reduce 420) (("*") reduce 420) (("+") reduce 420) (("-") reduce 420) ((".") reduce 420) (("/") reduce 420) ((";") reduce 420) (("<") reduce 420) ((">") reduce 420) (("?") reduce 420) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 420) ((T_BOOLEAN_AND) reduce 420) ((T_BOOLEAN_OR) reduce 420) ((T_COALESCE) reduce 420) ((T_INSTANCEOF) reduce 420) ((T_IS_EQUAL) reduce 420) ((T_IS_GREATER_OR_EQUAL) reduce 420) ((T_IS_IDENTICAL) reduce 420) ((T_IS_NOT_EQUAL) reduce 420) ((T_IS_NOT_IDENTICAL) reduce 420) ((T_IS_SMALLER_OR_EQUAL) reduce 420) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_POW) shift) ((T_SL) reduce 420) ((T_SPACESHIP) reduce 420) ((T_SR) reduce 420) (("^") reduce 420) (("|") reduce 420)) 26387 ((("%") reduce 470) (("(") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 470) ((T_OBJECT_OPERATOR) reduce 470) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 470) ((T_POW) reduce 470) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("[") reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 26388 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_POW) shift) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 26389 ((("%") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_POW) shift) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("^") reduce 398) (("|") reduce 398)) 26390 ((("%") reduce 397) (("*") reduce 397) (("+") reduce 397) (("-") reduce 397) ((".") reduce 397) (("/") reduce 397) ((";") reduce 397) (("<") reduce 397) ((">") reduce 397) (("?") reduce 397) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 397) ((T_BOOLEAN_AND) reduce 397) ((T_BOOLEAN_OR) reduce 397) ((T_COALESCE) reduce 397) ((T_INSTANCEOF) reduce 397) ((T_IS_EQUAL) reduce 397) ((T_IS_GREATER_OR_EQUAL) reduce 397) ((T_IS_IDENTICAL) reduce 397) ((T_IS_NOT_EQUAL) reduce 397) ((T_IS_NOT_IDENTICAL) reduce 397) ((T_IS_SMALLER_OR_EQUAL) reduce 397) ((T_LOGICAL_AND) reduce 397) ((T_LOGICAL_OR) reduce 397) ((T_LOGICAL_XOR) reduce 397) ((T_POW) shift) ((T_SL) reduce 397) ((T_SPACESHIP) reduce 397) ((T_SR) reduce 397) (("^") reduce 397) (("|") reduce 397)) 26392 ((("%") reduce 411) (("(") reduce 503) (("*") reduce 411) (("+") reduce 411) (("-") reduce 411) ((".") reduce 411) (("/") reduce 411) ((";") reduce 411) (("<") reduce 411) ((">") reduce 411) (("?") reduce 411) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 411) ((T_BOOLEAN_AND) reduce 411) ((T_BOOLEAN_OR) reduce 411) ((T_COALESCE) reduce 411) ((T_INSTANCEOF) reduce 411) ((T_IS_EQUAL) reduce 411) ((T_IS_GREATER_OR_EQUAL) reduce 411) ((T_IS_IDENTICAL) reduce 411) ((T_IS_NOT_EQUAL) reduce 411) ((T_IS_NOT_IDENTICAL) reduce 411) ((T_IS_SMALLER_OR_EQUAL) reduce 411) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 497) ((T_OBJECT_OPERATOR) reduce 497) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 497) ((T_POW) reduce 411) ((T_SL) reduce 411) ((T_SPACESHIP) reduce 411) ((T_SR) reduce 411) (("[") reduce 497) (("^") reduce 411) (("{") reduce 497) (("|") reduce 411)) 26393 ((("%") reduce 517) (("(") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) (("=") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AND_EQUAL) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_COALESCE_EQUAL) reduce 517) ((T_CONCAT_EQUAL) reduce 517) ((T_DEC) reduce 517) ((T_DIV_EQUAL) reduce 517) ((T_INC) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_MINUS_EQUAL) reduce 517) ((T_MOD_EQUAL) reduce 517) ((T_MUL_EQUAL) reduce 517) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 517) ((T_OBJECT_OPERATOR) reduce 517) ((T_OR_EQUAL) reduce 517) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 517) ((T_PLUS_EQUAL) reduce 517) ((T_POW) reduce 517) ((T_POW_EQUAL) reduce 517) ((T_SL) reduce 517) ((T_SL_EQUAL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) ((T_SR_EQUAL) reduce 517) ((T_XOR_EQUAL) reduce 517) (("[") reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 26396 ((("%") reduce 516) (("(") reduce 516) (("*") reduce 516) (("+") reduce 516) (("-") reduce 516) ((".") reduce 516) (("/") reduce 516) ((";") reduce 516) (("<") reduce 516) (("=") reduce 516) ((">") reduce 516) (("?") reduce 516) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 516) ((T_AND_EQUAL) reduce 516) ((T_BOOLEAN_AND) reduce 516) ((T_BOOLEAN_OR) reduce 516) ((T_COALESCE) reduce 516) ((T_COALESCE_EQUAL) reduce 516) ((T_CONCAT_EQUAL) reduce 516) ((T_DEC) reduce 516) ((T_DIV_EQUAL) reduce 516) ((T_INC) reduce 516) ((T_INSTANCEOF) reduce 516) ((T_IS_EQUAL) reduce 516) ((T_IS_GREATER_OR_EQUAL) reduce 516) ((T_IS_IDENTICAL) reduce 516) ((T_IS_NOT_EQUAL) reduce 516) ((T_IS_NOT_IDENTICAL) reduce 516) ((T_IS_SMALLER_OR_EQUAL) reduce 516) ((T_LOGICAL_AND) reduce 516) ((T_LOGICAL_OR) reduce 516) ((T_LOGICAL_XOR) reduce 516) ((T_MINUS_EQUAL) reduce 516) ((T_MOD_EQUAL) reduce 516) ((T_MUL_EQUAL) reduce 516) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 516) ((T_OBJECT_OPERATOR) reduce 516) ((T_OR_EQUAL) reduce 516) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 516) ((T_PLUS_EQUAL) reduce 516) ((T_POW) reduce 516) ((T_POW_EQUAL) reduce 516) ((T_SL) reduce 516) ((T_SL_EQUAL) reduce 516) ((T_SPACESHIP) reduce 516) ((T_SR) reduce 516) ((T_SR_EQUAL) reduce 516) ((T_XOR_EQUAL) reduce 516) (("[") reduce 516) (("^") reduce 516) (("{") reduce 516) (("|") reduce 516)) 26398 ((("%") reduce 473) (("(") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 473) ((T_OBJECT_OPERATOR) reduce 473) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 473) ((T_POW) reduce 473) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("[") reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 26399 ((("%") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_POW) shift) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("^") reduce 399) (("|") reduce 399)) 26400 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (($) reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 26413 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (($) reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONST) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAMESPACE) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_USE) reduce 184) ((T_VARIABLE) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 26419 ((("!") reduce 109) (("\"") reduce 109) (($) reduce 109) (("$") reduce 109) (("(") reduce 109) (("+") reduce 109) (("-") reduce 109) ((";") reduce 109) (("@") reduce 109) ((T_ABSTRACT) reduce 109) ((T_ARRAY) reduce 109) ((T_ARRAY_CAST) reduce 109) ((T_ATTRIBUTE) reduce 109) ((T_BOOL_CAST) reduce 109) ((T_BREAK) reduce 109) ((T_CLASS) reduce 109) ((T_CLASS_C) reduce 109) ((T_CLONE) reduce 109) ((T_CONST) reduce 109) ((T_CONSTANT_ENCAPSED_STRING) reduce 109) ((T_CONTINUE) reduce 109) ((T_DEC) reduce 109) ((T_DECLARE) reduce 109) ((T_DIR) reduce 109) ((T_DNUMBER) reduce 109) ((T_DO) reduce 109) ((T_DOUBLE_CAST) reduce 109) ((T_ECHO) reduce 109) ((T_EMPTY) reduce 109) ((T_ENUM) reduce 109) ((T_EVAL) reduce 109) ((T_EXIT) reduce 109) ((T_FILE) reduce 109) ((T_FINAL) reduce 109) ((T_FN) reduce 109) ((T_FOR) reduce 109) ((T_FOREACH) reduce 109) ((T_FUNCTION) reduce 109) ((T_FUNC_C) reduce 109) ((T_GLOBAL) reduce 109) ((T_GOTO) reduce 109) ((T_HALT_COMPILER) reduce 109) ((T_IF) reduce 109) ((T_INC) reduce 109) ((T_INCLUDE) reduce 109) ((T_INCLUDE_ONCE) reduce 109) ((T_INLINE_HTML) reduce 109) ((T_INTERFACE) reduce 109) ((T_INT_CAST) reduce 109) ((T_ISSET) reduce 109) ((T_LINE) reduce 109) ((T_LIST) reduce 109) ((T_LNUMBER) reduce 109) ((T_MATCH) reduce 109) ((T_METHOD_C) reduce 109) ((T_NAMESPACE) reduce 109) ((T_NAME_FULLY_QUALIFIED) reduce 109) ((T_NAME_QUALIFIED) reduce 109) ((T_NAME_RELATIVE) reduce 109) ((T_NEW) reduce 109) ((T_NS_C) reduce 109) ((T_OBJECT_CAST) reduce 109) ((T_PRINT) reduce 109) ((T_REQUIRE) reduce 109) ((T_REQUIRE_ONCE) reduce 109) ((T_RETURN) reduce 109) ((T_START_HEREDOC) reduce 109) ((T_STATIC) reduce 109) ((T_STRING) reduce 109) ((T_STRING_CAST) reduce 109) ((T_SWITCH) reduce 109) ((T_THROW) reduce 109) ((T_TRAIT) reduce 109) ((T_TRAIT_C) reduce 109) ((T_TRY) reduce 109) ((T_UNSET) reduce 109) ((T_UNSET_CAST) reduce 109) ((T_USE) reduce 109) ((T_VARIABLE) reduce 109) ((T_WHILE) reduce 109) ((T_YIELD) reduce 109) ((T_YIELD_FROM) reduce 109) (("[") reduce 109) (("`") reduce 109) (("{") reduce 109) (("~") reduce 109)) 26420 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (($) reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONST) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAMESPACE) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_USE) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 26422 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (($) reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONST) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAMESPACE) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_USE) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 26424 ((("!") reduce 102) (("\"") reduce 102) (("$") reduce 102) (($) reduce 102) (("(") reduce 102) (("+") reduce 102) (("-") reduce 102) ((";") reduce 102) (("@") reduce 102) ((T_ABSTRACT) reduce 102) ((T_ARRAY) reduce 102) ((T_ARRAY_CAST) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_BOOL_CAST) reduce 102) ((T_BREAK) reduce 102) ((T_CLASS) reduce 102) ((T_CLASS_C) reduce 102) ((T_CLONE) reduce 102) ((T_CONST) reduce 102) ((T_CONSTANT_ENCAPSED_STRING) reduce 102) ((T_CONTINUE) reduce 102) ((T_DEC) reduce 102) ((T_DECLARE) reduce 102) ((T_DIR) reduce 102) ((T_DNUMBER) reduce 102) ((T_DO) reduce 102) ((T_DOUBLE_CAST) reduce 102) ((T_ECHO) reduce 102) ((T_EMPTY) reduce 102) ((T_ENUM) reduce 102) ((T_EVAL) reduce 102) ((T_EXIT) reduce 102) ((T_FILE) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FOR) reduce 102) ((T_FOREACH) reduce 102) ((T_FUNCTION) reduce 102) ((T_FUNC_C) reduce 102) ((T_GLOBAL) reduce 102) ((T_GOTO) reduce 102) ((T_HALT_COMPILER) reduce 102) ((T_IF) reduce 102) ((T_INC) reduce 102) ((T_INCLUDE) reduce 102) ((T_INCLUDE_ONCE) reduce 102) ((T_INLINE_HTML) reduce 102) ((T_INTERFACE) reduce 102) ((T_INT_CAST) reduce 102) ((T_ISSET) reduce 102) ((T_LINE) reduce 102) ((T_LIST) reduce 102) ((T_LNUMBER) reduce 102) ((T_MATCH) reduce 102) ((T_METHOD_C) reduce 102) ((T_NAMESPACE) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_NEW) reduce 102) ((T_NS_C) reduce 102) ((T_OBJECT_CAST) reduce 102) ((T_PRINT) reduce 102) ((T_REQUIRE) reduce 102) ((T_REQUIRE_ONCE) reduce 102) ((T_RETURN) reduce 102) ((T_START_HEREDOC) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_STRING_CAST) reduce 102) ((T_SWITCH) reduce 102) ((T_THROW) reduce 102) ((T_TRAIT) reduce 102) ((T_TRAIT_C) reduce 102) ((T_TRY) reduce 102) ((T_UNSET) reduce 102) ((T_UNSET_CAST) reduce 102) ((T_USE) reduce 102) ((T_VARIABLE) reduce 102) ((T_WHILE) reduce 102) ((T_YIELD) reduce 102) ((T_YIELD_FROM) reduce 102) (("[") reduce 102) (("`") reduce 102) (("{") reduce 102) (("~") reduce 102)) 26425 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (($) reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONST) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAMESPACE) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_USE) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 26426 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (($) reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONST) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAMESPACE) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_USE) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 26437 ((("!") reduce 444) (("\"") reduce 444) (($) reduce 444) (("$") reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) ((";") reduce 444) (("@") reduce 444) ((T_ABSTRACT) reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_BREAK) reduce 444) ((T_CLASS) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONST) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_CONTINUE) reduce 444) ((T_DEC) reduce 444) ((T_DECLARE) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DO) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_ECHO) reduce 444) ((T_EMPTY) reduce 444) ((T_ENUM) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FINAL) reduce 444) ((T_FN) reduce 444) ((T_FOR) reduce 444) ((T_FOREACH) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_GLOBAL) reduce 444) ((T_GOTO) reduce 444) ((T_HALT_COMPILER) reduce 444) ((T_IF) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INLINE_HTML) reduce 444) ((T_INTERFACE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAMESPACE) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_RETURN) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_SWITCH) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT) reduce 444) ((T_TRAIT_C) reduce 444) ((T_TRY) reduce 444) ((T_UNSET) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_USE) reduce 444) ((T_VARIABLE) reduce 444) ((T_WHILE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("{") reduce 444) (("~") reduce 444)) 26438 ((("!") reduce 179) (("\"") reduce 179) (($) reduce 179) (("$") reduce 179) (("(") reduce 179) (("+") reduce 179) (("-") reduce 179) ((";") reduce 179) (("@") reduce 179) ((T_ABSTRACT) reduce 179) ((T_ARRAY) reduce 179) ((T_ARRAY_CAST) reduce 179) ((T_ATTRIBUTE) reduce 179) ((T_BOOL_CAST) reduce 179) ((T_BREAK) reduce 179) ((T_CLASS) reduce 179) ((T_CLASS_C) reduce 179) ((T_CLONE) reduce 179) ((T_CONST) reduce 179) ((T_CONSTANT_ENCAPSED_STRING) reduce 179) ((T_CONTINUE) reduce 179) ((T_DEC) reduce 179) ((T_DECLARE) reduce 179) ((T_DIR) reduce 179) ((T_DNUMBER) reduce 179) ((T_DO) reduce 179) ((T_DOUBLE_CAST) reduce 179) ((T_ECHO) reduce 179) ((T_EMPTY) reduce 179) ((T_ENUM) reduce 179) ((T_EVAL) reduce 179) ((T_EXIT) reduce 179) ((T_FILE) reduce 179) ((T_FINAL) reduce 179) ((T_FN) reduce 179) ((T_FOR) reduce 179) ((T_FOREACH) reduce 179) ((T_FUNCTION) reduce 179) ((T_FUNC_C) reduce 179) ((T_GLOBAL) reduce 179) ((T_GOTO) reduce 179) ((T_HALT_COMPILER) reduce 179) ((T_IF) reduce 179) ((T_INC) reduce 179) ((T_INCLUDE) reduce 179) ((T_INCLUDE_ONCE) reduce 179) ((T_INLINE_HTML) reduce 179) ((T_INTERFACE) reduce 179) ((T_INT_CAST) reduce 179) ((T_ISSET) reduce 179) ((T_LINE) reduce 179) ((T_LIST) reduce 179) ((T_LNUMBER) reduce 179) ((T_MATCH) reduce 179) ((T_METHOD_C) reduce 179) ((T_NAMESPACE) reduce 179) ((T_NAME_FULLY_QUALIFIED) reduce 179) ((T_NAME_QUALIFIED) reduce 179) ((T_NAME_RELATIVE) reduce 179) ((T_NEW) reduce 179) ((T_NS_C) reduce 179) ((T_OBJECT_CAST) reduce 179) ((T_PRINT) reduce 179) ((T_REQUIRE) reduce 179) ((T_REQUIRE_ONCE) reduce 179) ((T_RETURN) reduce 179) ((T_START_HEREDOC) reduce 179) ((T_STATIC) reduce 179) ((T_STRING) reduce 179) ((T_STRING_CAST) reduce 179) ((T_SWITCH) reduce 179) ((T_THROW) reduce 179) ((T_TRAIT) reduce 179) ((T_TRAIT_C) reduce 179) ((T_TRY) reduce 179) ((T_UNSET) reduce 179) ((T_UNSET_CAST) reduce 179) ((T_USE) reduce 179) ((T_VARIABLE) reduce 179) ((T_WHILE) reduce 179) ((T_YIELD) reduce 179) ((T_YIELD_FROM) reduce 179) (("[") reduce 179) (("`") reduce 179) (("{") reduce 179) (("~") reduce 179)) 26446 ((("!") reduce 184) (("\"") reduce 184) (($) reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONST) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAMESPACE) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_USE) reduce 184) ((T_VARIABLE) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 26451 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (($) reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONST) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAMESPACE) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_USE) reduce 190) ((T_VARIABLE) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 26457 ((("!") reduce 191) (("\"") reduce 191) (($) reduce 191) (("$") reduce 191) (("(") reduce 191) (("+") reduce 191) (("-") reduce 191) ((";") reduce 191) (("@") reduce 191) ((T_ABSTRACT) reduce 191) ((T_ARRAY) reduce 191) ((T_ARRAY_CAST) reduce 191) ((T_ATTRIBUTE) reduce 191) ((T_BOOL_CAST) reduce 191) ((T_BREAK) reduce 191) ((T_CLASS) reduce 191) ((T_CLASS_C) reduce 191) ((T_CLONE) reduce 191) ((T_CONST) reduce 191) ((T_CONSTANT_ENCAPSED_STRING) reduce 191) ((T_CONTINUE) reduce 191) ((T_DEC) reduce 191) ((T_DECLARE) reduce 191) ((T_DIR) reduce 191) ((T_DNUMBER) reduce 191) ((T_DO) reduce 191) ((T_DOUBLE_CAST) reduce 191) ((T_ECHO) reduce 191) ((T_EMPTY) reduce 191) ((T_ENUM) reduce 191) ((T_EVAL) reduce 191) ((T_EXIT) reduce 191) ((T_FILE) reduce 191) ((T_FINAL) reduce 191) ((T_FN) reduce 191) ((T_FOR) reduce 191) ((T_FOREACH) reduce 191) ((T_FUNCTION) reduce 191) ((T_FUNC_C) reduce 191) ((T_GLOBAL) reduce 191) ((T_GOTO) reduce 191) ((T_HALT_COMPILER) reduce 191) ((T_IF) reduce 191) ((T_INC) reduce 191) ((T_INCLUDE) reduce 191) ((T_INCLUDE_ONCE) reduce 191) ((T_INLINE_HTML) reduce 191) ((T_INTERFACE) reduce 191) ((T_INT_CAST) reduce 191) ((T_ISSET) reduce 191) ((T_LINE) reduce 191) ((T_LIST) reduce 191) ((T_LNUMBER) reduce 191) ((T_MATCH) reduce 191) ((T_METHOD_C) reduce 191) ((T_NAMESPACE) reduce 191) ((T_NAME_FULLY_QUALIFIED) reduce 191) ((T_NAME_QUALIFIED) reduce 191) ((T_NAME_RELATIVE) reduce 191) ((T_NEW) reduce 191) ((T_NS_C) reduce 191) ((T_OBJECT_CAST) reduce 191) ((T_PRINT) reduce 191) ((T_REQUIRE) reduce 191) ((T_REQUIRE_ONCE) reduce 191) ((T_RETURN) reduce 191) ((T_START_HEREDOC) reduce 191) ((T_STATIC) reduce 191) ((T_STRING) reduce 191) ((T_STRING_CAST) reduce 191) ((T_SWITCH) reduce 191) ((T_THROW) reduce 191) ((T_TRAIT) reduce 191) ((T_TRAIT_C) reduce 191) ((T_TRY) reduce 191) ((T_UNSET) reduce 191) ((T_UNSET_CAST) reduce 191) ((T_USE) reduce 191) ((T_VARIABLE) reduce 191) ((T_WHILE) reduce 191) ((T_YIELD) reduce 191) ((T_YIELD_FROM) reduce 191) (("[") reduce 191) (("`") reduce 191) (("{") reduce 191) (("~") reduce 191)) 26464 ((("!") reduce 192) (("\"") reduce 192) (("$") reduce 192) (($) reduce 192) (("(") reduce 192) (("+") reduce 192) (("-") reduce 192) ((";") reduce 192) (("@") reduce 192) ((T_ABSTRACT) reduce 192) ((T_ARRAY) reduce 192) ((T_ARRAY_CAST) reduce 192) ((T_ATTRIBUTE) reduce 192) ((T_BOOL_CAST) reduce 192) ((T_BREAK) reduce 192) ((T_CLASS) reduce 192) ((T_CLASS_C) reduce 192) ((T_CLONE) reduce 192) ((T_CONST) reduce 192) ((T_CONSTANT_ENCAPSED_STRING) reduce 192) ((T_CONTINUE) reduce 192) ((T_DEC) reduce 192) ((T_DECLARE) reduce 192) ((T_DIR) reduce 192) ((T_DNUMBER) reduce 192) ((T_DO) reduce 192) ((T_DOUBLE_CAST) reduce 192) ((T_ECHO) reduce 192) ((T_EMPTY) reduce 192) ((T_ENUM) reduce 192) ((T_EVAL) reduce 192) ((T_EXIT) reduce 192) ((T_FILE) reduce 192) ((T_FINAL) reduce 192) ((T_FN) reduce 192) ((T_FOR) reduce 192) ((T_FOREACH) reduce 192) ((T_FUNCTION) reduce 192) ((T_FUNC_C) reduce 192) ((T_GLOBAL) reduce 192) ((T_GOTO) reduce 192) ((T_HALT_COMPILER) reduce 192) ((T_IF) reduce 192) ((T_INC) reduce 192) ((T_INCLUDE) reduce 192) ((T_INCLUDE_ONCE) reduce 192) ((T_INLINE_HTML) reduce 192) ((T_INTERFACE) reduce 192) ((T_INT_CAST) reduce 192) ((T_ISSET) reduce 192) ((T_LINE) reduce 192) ((T_LIST) reduce 192) ((T_LNUMBER) reduce 192) ((T_MATCH) reduce 192) ((T_METHOD_C) reduce 192) ((T_NAMESPACE) reduce 192) ((T_NAME_FULLY_QUALIFIED) reduce 192) ((T_NAME_QUALIFIED) reduce 192) ((T_NAME_RELATIVE) reduce 192) ((T_NEW) reduce 192) ((T_NS_C) reduce 192) ((T_OBJECT_CAST) reduce 192) ((T_PRINT) reduce 192) ((T_REQUIRE) reduce 192) ((T_REQUIRE_ONCE) reduce 192) ((T_RETURN) reduce 192) ((T_START_HEREDOC) reduce 192) ((T_STATIC) reduce 192) ((T_STRING) reduce 192) ((T_STRING_CAST) reduce 192) ((T_SWITCH) reduce 192) ((T_THROW) reduce 192) ((T_TRAIT) reduce 192) ((T_TRAIT_C) reduce 192) ((T_TRY) reduce 192) ((T_UNSET) reduce 192) ((T_UNSET_CAST) reduce 192) ((T_USE) reduce 192) ((T_VARIABLE) reduce 192) ((T_WHILE) reduce 192) ((T_YIELD) reduce 192) ((T_YIELD_FROM) reduce 192) (("[") reduce 192) (("`") reduce 192) (("{") reduce 192) (("~") reduce 192)) 26471 ((("!") reduce 185) (("\"") reduce 185) (("$") reduce 185) (($) reduce 185) (("(") reduce 185) (("+") reduce 185) (("-") reduce 185) ((";") reduce 185) (("@") reduce 185) ((T_ABSTRACT) reduce 185) ((T_ARRAY) reduce 185) ((T_ARRAY_CAST) reduce 185) ((T_ATTRIBUTE) reduce 185) ((T_BOOL_CAST) reduce 185) ((T_BREAK) reduce 185) ((T_CLASS) reduce 185) ((T_CLASS_C) reduce 185) ((T_CLONE) reduce 185) ((T_CONST) reduce 185) ((T_CONSTANT_ENCAPSED_STRING) reduce 185) ((T_CONTINUE) reduce 185) ((T_DEC) reduce 185) ((T_DECLARE) reduce 185) ((T_DIR) reduce 185) ((T_DNUMBER) reduce 185) ((T_DO) reduce 185) ((T_DOUBLE_CAST) reduce 185) ((T_ECHO) reduce 185) ((T_EMPTY) reduce 185) ((T_ENUM) reduce 185) ((T_EVAL) reduce 185) ((T_EXIT) reduce 185) ((T_FILE) reduce 185) ((T_FINAL) reduce 185) ((T_FN) reduce 185) ((T_FOR) reduce 185) ((T_FOREACH) reduce 185) ((T_FUNCTION) reduce 185) ((T_FUNC_C) reduce 185) ((T_GLOBAL) reduce 185) ((T_GOTO) reduce 185) ((T_HALT_COMPILER) reduce 185) ((T_IF) reduce 185) ((T_INC) reduce 185) ((T_INCLUDE) reduce 185) ((T_INCLUDE_ONCE) reduce 185) ((T_INLINE_HTML) reduce 185) ((T_INTERFACE) reduce 185) ((T_INT_CAST) reduce 185) ((T_ISSET) reduce 185) ((T_LINE) reduce 185) ((T_LIST) reduce 185) ((T_LNUMBER) reduce 185) ((T_MATCH) reduce 185) ((T_METHOD_C) reduce 185) ((T_NAMESPACE) reduce 185) ((T_NAME_FULLY_QUALIFIED) reduce 185) ((T_NAME_QUALIFIED) reduce 185) ((T_NAME_RELATIVE) reduce 185) ((T_NEW) reduce 185) ((T_NS_C) reduce 185) ((T_OBJECT_CAST) reduce 185) ((T_PRINT) reduce 185) ((T_REQUIRE) reduce 185) ((T_REQUIRE_ONCE) reduce 185) ((T_RETURN) reduce 185) ((T_START_HEREDOC) reduce 185) ((T_STATIC) reduce 185) ((T_STRING) reduce 185) ((T_STRING_CAST) reduce 185) ((T_SWITCH) reduce 185) ((T_THROW) reduce 185) ((T_TRAIT) reduce 185) ((T_TRAIT_C) reduce 185) ((T_TRY) reduce 185) ((T_UNSET) reduce 185) ((T_UNSET_CAST) reduce 185) ((T_USE) reduce 185) ((T_VARIABLE) reduce 185) ((T_WHILE) reduce 185) ((T_YIELD) reduce 185) ((T_YIELD_FROM) reduce 185) (("[") reduce 185) (("`") reduce 185) (("{") reduce 185) (("~") reduce 185)) 26491 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (($) reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONST) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAMESPACE) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_USE) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 26495 ((("!") reduce 241) (("\"") reduce 241) (($) reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONST) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAMESPACE) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_USE) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 27003 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (($) reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONST) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_EMPTY) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAMESPACE) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_USE) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 27004 ((("!") reduce 148) (("\"") reduce 148) (($) reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 27007 ((("!") reduce 233) (("\"") reduce 233) (($) reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONST) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_EMPTY) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAMESPACE) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_USE) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 27008 (((T_NAME_FULLY_QUALIFIED) reduce 120) ((T_NAME_QUALIFIED) reduce 120) ((T_STRING) reduce 120)) 27009 (((T_NAME_FULLY_QUALIFIED) reduce 119) ((T_NAME_QUALIFIED) reduce 119) ((T_STRING) reduce 119)) 27010 (((",") reduce 90) ((";") reduce 90) ((T_AS) reduce 90) ((T_NS_SEPARATOR) reduce 90)) 27011 (((",") reduce 88) ((";") reduce 88) ((T_AS) reduce 88) ((T_NS_SEPARATOR) reduce 88)) 27012 (((",") reduce 87) ((";") reduce 87) ((T_AS) reduce 87) ((T_NS_SEPARATOR) reduce 87)) 27013 (((",") reduce 135) ((";") reduce 135) ((T_AS) shift) ((T_NS_SEPARATOR) shift)) 27015 (((",") reduce 89) ((";") reduce 89) ((T_AS) reduce 89) ((T_NS_SEPARATOR) reduce 89)) 27016 (((",") reduce 130) ((";") reduce 130)) 27018 (((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_STRING) shift)) 27023 ((("!") reduce 117) (("\"") reduce 117) (($) reduce 117) (("$") reduce 117) (("(") reduce 117) (("+") reduce 117) (("-") reduce 117) ((";") reduce 117) (("@") reduce 117) ((T_ABSTRACT) reduce 117) ((T_ARRAY) reduce 117) ((T_ARRAY_CAST) reduce 117) ((T_ATTRIBUTE) reduce 117) ((T_BOOL_CAST) reduce 117) ((T_BREAK) reduce 117) ((T_CLASS) reduce 117) ((T_CLASS_C) reduce 117) ((T_CLONE) reduce 117) ((T_CONST) reduce 117) ((T_CONSTANT_ENCAPSED_STRING) reduce 117) ((T_CONTINUE) reduce 117) ((T_DEC) reduce 117) ((T_DECLARE) reduce 117) ((T_DIR) reduce 117) ((T_DNUMBER) reduce 117) ((T_DO) reduce 117) ((T_DOUBLE_CAST) reduce 117) ((T_ECHO) reduce 117) ((T_EMPTY) reduce 117) ((T_ENUM) reduce 117) ((T_EVAL) reduce 117) ((T_EXIT) reduce 117) ((T_FILE) reduce 117) ((T_FINAL) reduce 117) ((T_FN) reduce 117) ((T_FOR) reduce 117) ((T_FOREACH) reduce 117) ((T_FUNCTION) reduce 117) ((T_FUNC_C) reduce 117) ((T_GLOBAL) reduce 117) ((T_GOTO) reduce 117) ((T_HALT_COMPILER) reduce 117) ((T_IF) reduce 117) ((T_INC) reduce 117) ((T_INCLUDE) reduce 117) ((T_INCLUDE_ONCE) reduce 117) ((T_INLINE_HTML) reduce 117) ((T_INTERFACE) reduce 117) ((T_INT_CAST) reduce 117) ((T_ISSET) reduce 117) ((T_LINE) reduce 117) ((T_LIST) reduce 117) ((T_LNUMBER) reduce 117) ((T_MATCH) reduce 117) ((T_METHOD_C) reduce 117) ((T_NAMESPACE) reduce 117) ((T_NAME_FULLY_QUALIFIED) reduce 117) ((T_NAME_QUALIFIED) reduce 117) ((T_NAME_RELATIVE) reduce 117) ((T_NEW) reduce 117) ((T_NS_C) reduce 117) ((T_OBJECT_CAST) reduce 117) ((T_PRINT) reduce 117) ((T_REQUIRE) reduce 117) ((T_REQUIRE_ONCE) reduce 117) ((T_RETURN) reduce 117) ((T_START_HEREDOC) reduce 117) ((T_STATIC) reduce 117) ((T_STRING) reduce 117) ((T_STRING_CAST) reduce 117) ((T_SWITCH) reduce 117) ((T_THROW) reduce 117) ((T_TRAIT) reduce 117) ((T_TRAIT_C) reduce 117) ((T_TRY) reduce 117) ((T_UNSET) reduce 117) ((T_UNSET_CAST) reduce 117) ((T_USE) reduce 117) ((T_VARIABLE) reduce 117) ((T_WHILE) reduce 117) ((T_YIELD) reduce 117) ((T_YIELD_FROM) reduce 117) (("[") reduce 117) (("`") reduce 117) (("{") reduce 117) (("~") reduce 117)) 27024 (((",") reduce 90) ((";") reduce 90) ((T_AS) reduce 90)) 27025 (((",") reduce 88) ((";") reduce 88) ((T_AS) reduce 88)) 27026 (((",") reduce 87) ((";") reduce 87) ((T_AS) reduce 87)) 27027 (((",") reduce 135) ((";") reduce 135) ((T_AS) shift)) 27028 (((",") reduce 89) ((";") reduce 89) ((T_AS) reduce 89)) 27029 (((",") reduce 129) ((";") reduce 129)) 27031 (((",") reduce 136) ((";") reduce 136)) 27033 (((T_NAME_QUALIFIED) shift) ((T_STRING) shift)) 27034 (((",") reduce 88) ((T_AS) reduce 88) (("}") reduce 88)) 27035 (((",") reduce 87) ((T_AS) reduce 87) (("}") reduce 87)) 27036 (((",") reduce 133) ((T_AS) shift) (("}") reduce 133)) 27037 (((",") reduce 128) (("}") reduce 128)) 27039 (((T_NAME_QUALIFIED) shift) ((T_STRING) shift) (("}") reduce 124)) 27041 (((";") reduce 121)) 27042 (((",") reduce 127) (("}") reduce 127)) 27044 (((",") reduce 134) (("}") reduce 134)) 27045 ((("!") reduce 115) (("\"") reduce 115) (($) reduce 115) (("$") reduce 115) (("(") reduce 115) (("+") reduce 115) (("-") reduce 115) ((";") reduce 115) (("@") reduce 115) ((T_ABSTRACT) reduce 115) ((T_ARRAY) reduce 115) ((T_ARRAY_CAST) reduce 115) ((T_ATTRIBUTE) reduce 115) ((T_BOOL_CAST) reduce 115) ((T_BREAK) reduce 115) ((T_CLASS) reduce 115) ((T_CLASS_C) reduce 115) ((T_CLONE) reduce 115) ((T_CONST) reduce 115) ((T_CONSTANT_ENCAPSED_STRING) reduce 115) ((T_CONTINUE) reduce 115) ((T_DEC) reduce 115) ((T_DECLARE) reduce 115) ((T_DIR) reduce 115) ((T_DNUMBER) reduce 115) ((T_DO) reduce 115) ((T_DOUBLE_CAST) reduce 115) ((T_ECHO) reduce 115) ((T_EMPTY) reduce 115) ((T_ENUM) reduce 115) ((T_EVAL) reduce 115) ((T_EXIT) reduce 115) ((T_FILE) reduce 115) ((T_FINAL) reduce 115) ((T_FN) reduce 115) ((T_FOR) reduce 115) ((T_FOREACH) reduce 115) ((T_FUNCTION) reduce 115) ((T_FUNC_C) reduce 115) ((T_GLOBAL) reduce 115) ((T_GOTO) reduce 115) ((T_HALT_COMPILER) reduce 115) ((T_IF) reduce 115) ((T_INC) reduce 115) ((T_INCLUDE) reduce 115) ((T_INCLUDE_ONCE) reduce 115) ((T_INLINE_HTML) reduce 115) ((T_INTERFACE) reduce 115) ((T_INT_CAST) reduce 115) ((T_ISSET) reduce 115) ((T_LINE) reduce 115) ((T_LIST) reduce 115) ((T_LNUMBER) reduce 115) ((T_MATCH) reduce 115) ((T_METHOD_C) reduce 115) ((T_NAMESPACE) reduce 115) ((T_NAME_FULLY_QUALIFIED) reduce 115) ((T_NAME_QUALIFIED) reduce 115) ((T_NAME_RELATIVE) reduce 115) ((T_NEW) reduce 115) ((T_NS_C) reduce 115) ((T_OBJECT_CAST) reduce 115) ((T_PRINT) reduce 115) ((T_REQUIRE) reduce 115) ((T_REQUIRE_ONCE) reduce 115) ((T_RETURN) reduce 115) ((T_START_HEREDOC) reduce 115) ((T_STATIC) reduce 115) ((T_STRING) reduce 115) ((T_STRING_CAST) reduce 115) ((T_SWITCH) reduce 115) ((T_THROW) reduce 115) ((T_TRAIT) reduce 115) ((T_TRAIT_C) reduce 115) ((T_TRY) reduce 115) ((T_UNSET) reduce 115) ((T_UNSET_CAST) reduce 115) ((T_USE) reduce 115) ((T_VARIABLE) reduce 115) ((T_WHILE) reduce 115) ((T_YIELD) reduce 115) ((T_YIELD_FROM) reduce 115) (("[") reduce 115) (("`") reduce 115) (("{") reduce 115) (("~") reduce 115)) 27046 ((("!") reduce 116) (("\"") reduce 116) (("$") reduce 116) (($) reduce 116) (("(") reduce 116) (("+") reduce 116) (("-") reduce 116) ((";") reduce 116) (("@") reduce 116) ((T_ABSTRACT) reduce 116) ((T_ARRAY) reduce 116) ((T_ARRAY_CAST) reduce 116) ((T_ATTRIBUTE) reduce 116) ((T_BOOL_CAST) reduce 116) ((T_BREAK) reduce 116) ((T_CLASS) reduce 116) ((T_CLASS_C) reduce 116) ((T_CLONE) reduce 116) ((T_CONST) reduce 116) ((T_CONSTANT_ENCAPSED_STRING) reduce 116) ((T_CONTINUE) reduce 116) ((T_DEC) reduce 116) ((T_DECLARE) reduce 116) ((T_DIR) reduce 116) ((T_DNUMBER) reduce 116) ((T_DO) reduce 116) ((T_DOUBLE_CAST) reduce 116) ((T_ECHO) reduce 116) ((T_EMPTY) reduce 116) ((T_ENUM) reduce 116) ((T_EVAL) reduce 116) ((T_EXIT) reduce 116) ((T_FILE) reduce 116) ((T_FINAL) reduce 116) ((T_FN) reduce 116) ((T_FOR) reduce 116) ((T_FOREACH) reduce 116) ((T_FUNCTION) reduce 116) ((T_FUNC_C) reduce 116) ((T_GLOBAL) reduce 116) ((T_GOTO) reduce 116) ((T_HALT_COMPILER) reduce 116) ((T_IF) reduce 116) ((T_INC) reduce 116) ((T_INCLUDE) reduce 116) ((T_INCLUDE_ONCE) reduce 116) ((T_INLINE_HTML) reduce 116) ((T_INTERFACE) reduce 116) ((T_INT_CAST) reduce 116) ((T_ISSET) reduce 116) ((T_LINE) reduce 116) ((T_LIST) reduce 116) ((T_LNUMBER) reduce 116) ((T_MATCH) reduce 116) ((T_METHOD_C) reduce 116) ((T_NAMESPACE) reduce 116) ((T_NAME_FULLY_QUALIFIED) reduce 116) ((T_NAME_QUALIFIED) reduce 116) ((T_NAME_RELATIVE) reduce 116) ((T_NEW) reduce 116) ((T_NS_C) reduce 116) ((T_OBJECT_CAST) reduce 116) ((T_PRINT) reduce 116) ((T_REQUIRE) reduce 116) ((T_REQUIRE_ONCE) reduce 116) ((T_RETURN) reduce 116) ((T_START_HEREDOC) reduce 116) ((T_STATIC) reduce 116) ((T_STRING) reduce 116) ((T_STRING_CAST) reduce 116) ((T_SWITCH) reduce 116) ((T_THROW) reduce 116) ((T_TRAIT) reduce 116) ((T_TRAIT_C) reduce 116) ((T_TRY) reduce 116) ((T_UNSET) reduce 116) ((T_UNSET_CAST) reduce 116) ((T_USE) reduce 116) ((T_VARIABLE) reduce 116) ((T_WHILE) reduce 116) ((T_YIELD) reduce 116) ((T_YIELD_FROM) reduce 116) (("[") reduce 116) (("`") reduce 116) (("{") reduce 116) (("~") reduce 116)) 27047 ((("!") reduce 114) (("\"") reduce 114) (("$") reduce 114) (($) reduce 114) (("(") reduce 114) (("+") reduce 114) (("-") reduce 114) ((";") reduce 114) (("@") reduce 114) ((T_ABSTRACT) reduce 114) ((T_ARRAY) reduce 114) ((T_ARRAY_CAST) reduce 114) ((T_ATTRIBUTE) reduce 114) ((T_BOOL_CAST) reduce 114) ((T_BREAK) reduce 114) ((T_CLASS) reduce 114) ((T_CLASS_C) reduce 114) ((T_CLONE) reduce 114) ((T_CONST) reduce 114) ((T_CONSTANT_ENCAPSED_STRING) reduce 114) ((T_CONTINUE) reduce 114) ((T_DEC) reduce 114) ((T_DECLARE) reduce 114) ((T_DIR) reduce 114) ((T_DNUMBER) reduce 114) ((T_DO) reduce 114) ((T_DOUBLE_CAST) reduce 114) ((T_ECHO) reduce 114) ((T_EMPTY) reduce 114) ((T_ENUM) reduce 114) ((T_EVAL) reduce 114) ((T_EXIT) reduce 114) ((T_FILE) reduce 114) ((T_FINAL) reduce 114) ((T_FN) reduce 114) ((T_FOR) reduce 114) ((T_FOREACH) reduce 114) ((T_FUNCTION) reduce 114) ((T_FUNC_C) reduce 114) ((T_GLOBAL) reduce 114) ((T_GOTO) reduce 114) ((T_HALT_COMPILER) reduce 114) ((T_IF) reduce 114) ((T_INC) reduce 114) ((T_INCLUDE) reduce 114) ((T_INCLUDE_ONCE) reduce 114) ((T_INLINE_HTML) reduce 114) ((T_INTERFACE) reduce 114) ((T_INT_CAST) reduce 114) ((T_ISSET) reduce 114) ((T_LINE) reduce 114) ((T_LIST) reduce 114) ((T_LNUMBER) reduce 114) ((T_MATCH) reduce 114) ((T_METHOD_C) reduce 114) ((T_NAMESPACE) reduce 114) ((T_NAME_FULLY_QUALIFIED) reduce 114) ((T_NAME_QUALIFIED) reduce 114) ((T_NAME_RELATIVE) reduce 114) ((T_NEW) reduce 114) ((T_NS_C) reduce 114) ((T_OBJECT_CAST) reduce 114) ((T_PRINT) reduce 114) ((T_REQUIRE) reduce 114) ((T_REQUIRE_ONCE) reduce 114) ((T_RETURN) reduce 114) ((T_START_HEREDOC) reduce 114) ((T_STATIC) reduce 114) ((T_STRING) reduce 114) ((T_STRING_CAST) reduce 114) ((T_SWITCH) reduce 114) ((T_THROW) reduce 114) ((T_TRAIT) reduce 114) ((T_TRAIT_C) reduce 114) ((T_TRY) reduce 114) ((T_UNSET) reduce 114) ((T_UNSET_CAST) reduce 114) ((T_USE) reduce 114) ((T_VARIABLE) reduce 114) ((T_WHILE) reduce 114) ((T_YIELD) reduce 114) ((T_YIELD_FROM) reduce 114) (("[") reduce 114) (("`") reduce 114) (("{") reduce 114) (("~") reduce 114)) 27049 (((T_CONST) shift) ((T_FUNCTION) shift) ((T_NAME_QUALIFIED) shift) ((T_STRING) shift)) 27050 (((T_NAME_QUALIFIED) reduce 120) ((T_STRING) reduce 120)) 27051 (((T_NAME_QUALIFIED) reduce 119) ((T_STRING) reduce 119)) 27052 (((",") reduce 126) (("}") reduce 126)) 27054 (((",") reduce 131) (("}") reduce 131)) 27056 (((",") reduce 132) (("}") reduce 132)) 27057 (((T_CONST) shift) ((T_FUNCTION) shift) ((T_NAME_QUALIFIED) shift) ((T_STRING) shift) (("}") reduce 124)) 27059 (((";") reduce 122)) 27060 (((",") reduce 125) (("}") reduce 125)) 27066 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (($) reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 27069 ((("!") reduce 168) (("\"") reduce 168) (($) reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 27070 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (($) reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONST) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_EMPTY) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAMESPACE) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_USE) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 27073 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (($) reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 27076 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (($) reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONST) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAMESPACE) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_USE) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 27083 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (($) reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 27088 ((("!") reduce 190) (("\"") reduce 190) (($) reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONST) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAMESPACE) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_USE) reduce 190) ((T_VARIABLE) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 27094 ((("!") reduce 151) (("\"") reduce 151) (($) reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 27098 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (($) reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONST) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_EMPTY) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAMESPACE) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_USE) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 27100 ((("!") reduce 215) (("\"") reduce 215) (($) reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONST) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_EMPTY) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAMESPACE) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_USE) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 27104 ((("!") reduce 216) (("\"") reduce 216) (($) reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONST) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAMESPACE) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_USE) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 27107 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (($) reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONST) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAMESPACE) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_USE) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 27109 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (($) reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 27111 ((("!") reduce 156) (("\"") reduce 156) (($) reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 27118 ((("!") reduce 154) (("\"") reduce 154) (($) reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 27144 (((";") reduce 73) (("{") reduce 73)) 27145 (((";") reduce 52) (("{") reduce 52)) 27146 (((";") reduce 27) (("{") reduce 27)) 27147 (((";") reduce 51) (("{") reduce 51)) 27148 (((";") reduce 53) (("{") reduce 53)) 27149 (((";") reduce 49) (("{") reduce 49)) 27150 (((";") reduce 29) (("{") reduce 29)) 27151 (((";") reduce 59) (("{") reduce 59)) 27152 (((";") reduce 60) (("{") reduce 60)) 27153 (((";") reduce 11) (("{") reduce 11)) 27154 (((";") reduce 42) (("{") reduce 42)) 27155 (((";") reduce 39) (("{") reduce 39)) 27156 (((";") reduce 25) (("{") reduce 25)) 27157 (((";") reduce 50) (("{") reduce 50)) 27158 (((";") reduce 66) (("{") reduce 66)) 27159 (((";") reduce 18) (("{") reduce 18)) 27160 (((";") reduce 17) (("{") reduce 17)) 27161 (((";") reduce 15) (("{") reduce 15)) 27162 (((";") reduce 14) (("{") reduce 14)) 27163 (((";") reduce 38) (("{") reduce 38)) 27164 (((";") reduce 26) (("{") reduce 26)) 27165 (((";") reduce 22) (("{") reduce 22)) 27166 (((";") reduce 24) (("{") reduce 24)) 27167 (((";") reduce 16) (("{") reduce 16)) 27168 (((";") reduce 48) (("{") reduce 48)) 27169 (((";") reduce 20) (("{") reduce 20)) 27170 (((";") reduce 70) (("{") reduce 70)) 27171 (((";") reduce 3) (("{") reduce 3)) 27172 (((";") reduce 12) (("{") reduce 12)) 27173 (((";") reduce 54) (("{") reduce 54)) 27174 (((";") reduce 65) (("{") reduce 65)) 27175 (((";") reduce 74) (("{") reduce 74)) 27176 (((";") reduce 30) (("{") reduce 30)) 27177 (((";") reduce 68) (("{") reduce 68)) 27178 (((";") reduce 21) (("{") reduce 21)) 27179 (((";") reduce 23) (("{") reduce 23)) 27180 (((";") reduce 41) (("{") reduce 41)) 27181 (((";") reduce 62) (("{") reduce 62)) 27182 (((";") reduce 34) (("{") reduce 34)) 27183 (((";") reduce 40) (("{") reduce 40)) 27184 (((";") reduce 13) (("{") reduce 13)) 27185 (((";") reduce 55) (("{") reduce 55)) 27186 (((";") reduce 1) (("{") reduce 1)) 27187 (((";") reduce 2) (("{") reduce 2)) 27188 (((";") reduce 9) (("{") reduce 9)) 27189 (((";") reduce 33) (("{") reduce 33)) 27190 (((";") reduce 58) (("{") reduce 58)) 27191 (((";") reduce 37) (("{") reduce 37)) 27192 (((";") reduce 64) (("{") reduce 64)) 27193 (((";") reduce 46) (("{") reduce 46)) 27194 (((";") reduce 8) (("{") reduce 8)) 27195 (((";") reduce 6) (("{") reduce 6)) 27196 (((";") reduce 7) (("{") reduce 7)) 27197 (((";") reduce 69) (("{") reduce 69)) 27198 (((";") reduce 63) (("{") reduce 63)) 27199 (((";") reduce 56) (("{") reduce 56)) 27200 (((";") reduce 86) (("{") reduce 86)) 27201 (((";") reduce 10) (("{") reduce 10)) 27202 (((";") reduce 67) (("{") reduce 67)) 27203 (((";") reduce 44) (("{") reduce 44)) 27204 (((";") reduce 75) (("{") reduce 75)) 27205 (((";") reduce 76) (("{") reduce 76)) 27206 (((";") reduce 77) (("{") reduce 77)) 27207 (((";") reduce 78) (("{") reduce 78)) 27208 (((";") reduce 4) (("{") reduce 4)) 27209 (((";") reduce 5) (("{") reduce 5)) 27210 (((";") reduce 43) (("{") reduce 43)) 27211 (((";") reduce 72) (("{") reduce 72)) 27212 (((";") reduce 81) (("{") reduce 81)) 27213 (((";") reduce 47) (("{") reduce 47)) 27214 (((";") reduce 31) (("{") reduce 31)) 27215 (((";") reduce 57) (("{") reduce 57)) 27216 (((";") reduce 61) (("{") reduce 61)) 27217 (((";") reduce 28) (("{") reduce 28)) 27218 (((";") reduce 36) (("{") reduce 36)) 27219 (((";") reduce 32) (("{") reduce 32)) 27220 (((";") reduce 35) (("{") reduce 35)) 27221 (((";") reduce 19) (("{") reduce 19)) 27222 (((";") reduce 45) (("{") reduce 45)) 27223 (((";") reduce 85) (("{") reduce 85)) 27225 (((";") reduce 71) (("{") reduce 71)) 27226 (((";") reduce 82) (("{") reduce 82)) 27227 ((("!") reduce 84) (("\"") reduce 84) (("$") reduce 84) (("(") reduce 84) (("+") reduce 84) (("-") reduce 84) ((";") reduce 84) (("@") reduce 84) ((T_ABSTRACT) reduce 84) ((T_ARRAY) reduce 84) ((T_ARRAY_CAST) reduce 84) ((T_ATTRIBUTE) reduce 84) ((T_BOOL_CAST) reduce 84) ((T_BREAK) reduce 84) ((T_CLASS) reduce 84) ((T_CLASS_C) reduce 84) ((T_CLONE) reduce 84) ((T_CONST) reduce 84) ((T_CONSTANT_ENCAPSED_STRING) reduce 84) ((T_CONTINUE) reduce 84) ((T_DEC) reduce 84) ((T_DECLARE) reduce 84) ((T_DIR) reduce 84) ((T_DNUMBER) reduce 84) ((T_DO) reduce 84) ((T_DOUBLE_CAST) reduce 84) ((T_ECHO) reduce 84) ((T_EMPTY) reduce 84) ((T_ENUM) reduce 84) ((T_EVAL) reduce 84) ((T_EXIT) reduce 84) ((T_FILE) reduce 84) ((T_FINAL) reduce 84) ((T_FN) reduce 84) ((T_FOR) reduce 84) ((T_FOREACH) reduce 84) ((T_FUNCTION) reduce 84) ((T_FUNC_C) reduce 84) ((T_GLOBAL) reduce 84) ((T_GOTO) reduce 84) ((T_HALT_COMPILER) reduce 84) ((T_IF) reduce 84) ((T_INC) reduce 84) ((T_INCLUDE) reduce 84) ((T_INCLUDE_ONCE) reduce 84) ((T_INLINE_HTML) reduce 84) ((T_INTERFACE) reduce 84) ((T_INT_CAST) reduce 84) ((T_ISSET) reduce 84) ((T_LINE) reduce 84) ((T_LIST) reduce 84) ((T_LNUMBER) reduce 84) ((T_MATCH) reduce 84) ((T_METHOD_C) reduce 84) ((T_NAMESPACE) reduce 84) ((T_NAME_FULLY_QUALIFIED) reduce 84) ((T_NAME_QUALIFIED) reduce 84) ((T_NAME_RELATIVE) reduce 84) ((T_NEW) reduce 84) ((T_NS_C) reduce 84) ((T_OBJECT_CAST) reduce 84) ((T_PRINT) reduce 84) ((T_REQUIRE) reduce 84) ((T_REQUIRE_ONCE) reduce 84) ((T_RETURN) reduce 84) ((T_START_HEREDOC) reduce 84) ((T_STATIC) reduce 84) ((T_STRING) reduce 84) ((T_STRING_CAST) reduce 84) ((T_SWITCH) reduce 84) ((T_THROW) reduce 84) ((T_TRAIT) reduce 84) ((T_TRAIT_C) reduce 84) ((T_TRY) reduce 84) ((T_UNSET) reduce 84) ((T_UNSET_CAST) reduce 84) ((T_USE) reduce 84) ((T_VARIABLE) reduce 84) ((T_WHILE) reduce 84) ((T_YIELD) reduce 84) ((T_YIELD_FROM) reduce 84) (("[") reduce 84) (("`") reduce 84) (("{") reduce 84) (("}") reduce 84) (("~") reduce 84)) 27228 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_USE) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") shift) (("~") shift)) 27229 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 27244 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 27256 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 27258 ((("!") reduce 108) (("\"") reduce 108) (("$") reduce 108) (("(") reduce 108) (("+") reduce 108) (("-") reduce 108) ((";") reduce 108) (("@") reduce 108) ((T_ABSTRACT) reduce 108) ((T_ARRAY) reduce 108) ((T_ARRAY_CAST) reduce 108) ((T_ATTRIBUTE) reduce 108) ((T_BOOL_CAST) reduce 108) ((T_BREAK) reduce 108) ((T_CLASS) reduce 108) ((T_CLASS_C) reduce 108) ((T_CLONE) reduce 108) ((T_CONST) reduce 108) ((T_CONSTANT_ENCAPSED_STRING) reduce 108) ((T_CONTINUE) reduce 108) ((T_DEC) reduce 108) ((T_DECLARE) reduce 108) ((T_DIR) reduce 108) ((T_DNUMBER) reduce 108) ((T_DO) reduce 108) ((T_DOUBLE_CAST) reduce 108) ((T_ECHO) reduce 108) ((T_EMPTY) reduce 108) ((T_ENUM) reduce 108) ((T_EVAL) reduce 108) ((T_EXIT) reduce 108) ((T_FILE) reduce 108) ((T_FINAL) reduce 108) ((T_FN) reduce 108) ((T_FOR) reduce 108) ((T_FOREACH) reduce 108) ((T_FUNCTION) reduce 108) ((T_FUNC_C) reduce 108) ((T_GLOBAL) reduce 108) ((T_GOTO) reduce 108) ((T_HALT_COMPILER) reduce 108) ((T_IF) reduce 108) ((T_INC) reduce 108) ((T_INCLUDE) reduce 108) ((T_INCLUDE_ONCE) reduce 108) ((T_INLINE_HTML) reduce 108) ((T_INTERFACE) reduce 108) ((T_INT_CAST) reduce 108) ((T_ISSET) reduce 108) ((T_LINE) reduce 108) ((T_LIST) reduce 108) ((T_LNUMBER) reduce 108) ((T_MATCH) reduce 108) ((T_METHOD_C) reduce 108) ((T_NAMESPACE) reduce 108) ((T_NAME_FULLY_QUALIFIED) reduce 108) ((T_NAME_QUALIFIED) reduce 108) ((T_NAME_RELATIVE) reduce 108) ((T_NEW) reduce 108) ((T_NS_C) reduce 108) ((T_OBJECT_CAST) reduce 108) ((T_PRINT) reduce 108) ((T_REQUIRE) reduce 108) ((T_REQUIRE_ONCE) reduce 108) ((T_RETURN) reduce 108) ((T_START_HEREDOC) reduce 108) ((T_STATIC) reduce 108) ((T_STRING) reduce 108) ((T_STRING_CAST) reduce 108) ((T_SWITCH) reduce 108) ((T_THROW) reduce 108) ((T_TRAIT) reduce 108) ((T_TRAIT_C) reduce 108) ((T_TRY) reduce 108) ((T_UNSET) reduce 108) ((T_UNSET_CAST) reduce 108) ((T_USE) reduce 108) ((T_VARIABLE) reduce 108) ((T_WHILE) reduce 108) ((T_YIELD) reduce 108) ((T_YIELD_FROM) reduce 108) (("[") reduce 108) (("`") reduce 108) (("{") reduce 108) (("}") reduce 108) (("~") reduce 108)) 27260 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONST) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAMESPACE) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_USE) reduce 103) ((T_VARIABLE) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("}") reduce 103) (("~") reduce 103)) 27262 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONST) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAMESPACE) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_USE) reduce 106) ((T_VARIABLE) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("}") reduce 106) (("~") reduce 106)) 27265 ((("!") reduce 102) (("\"") reduce 102) (("$") reduce 102) (("(") reduce 102) (("+") reduce 102) (("-") reduce 102) ((";") reduce 102) (("@") reduce 102) ((T_ABSTRACT) reduce 102) ((T_ARRAY) reduce 102) ((T_ARRAY_CAST) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_BOOL_CAST) reduce 102) ((T_BREAK) reduce 102) ((T_CLASS) reduce 102) ((T_CLASS_C) reduce 102) ((T_CLONE) reduce 102) ((T_CONST) reduce 102) ((T_CONSTANT_ENCAPSED_STRING) reduce 102) ((T_CONTINUE) reduce 102) ((T_DEC) reduce 102) ((T_DECLARE) reduce 102) ((T_DIR) reduce 102) ((T_DNUMBER) reduce 102) ((T_DO) reduce 102) ((T_DOUBLE_CAST) reduce 102) ((T_ECHO) reduce 102) ((T_EMPTY) reduce 102) ((T_ENUM) reduce 102) ((T_EVAL) reduce 102) ((T_EXIT) reduce 102) ((T_FILE) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FOR) reduce 102) ((T_FOREACH) reduce 102) ((T_FUNCTION) reduce 102) ((T_FUNC_C) reduce 102) ((T_GLOBAL) reduce 102) ((T_GOTO) reduce 102) ((T_HALT_COMPILER) reduce 102) ((T_IF) reduce 102) ((T_INC) reduce 102) ((T_INCLUDE) reduce 102) ((T_INCLUDE_ONCE) reduce 102) ((T_INLINE_HTML) reduce 102) ((T_INTERFACE) reduce 102) ((T_INT_CAST) reduce 102) ((T_ISSET) reduce 102) ((T_LINE) reduce 102) ((T_LIST) reduce 102) ((T_LNUMBER) reduce 102) ((T_MATCH) reduce 102) ((T_METHOD_C) reduce 102) ((T_NAMESPACE) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_NEW) reduce 102) ((T_NS_C) reduce 102) ((T_OBJECT_CAST) reduce 102) ((T_PRINT) reduce 102) ((T_REQUIRE) reduce 102) ((T_REQUIRE_ONCE) reduce 102) ((T_RETURN) reduce 102) ((T_START_HEREDOC) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_STRING_CAST) reduce 102) ((T_SWITCH) reduce 102) ((T_THROW) reduce 102) ((T_TRAIT) reduce 102) ((T_TRAIT_C) reduce 102) ((T_TRY) reduce 102) ((T_UNSET) reduce 102) ((T_UNSET_CAST) reduce 102) ((T_USE) reduce 102) ((T_VARIABLE) reduce 102) ((T_WHILE) reduce 102) ((T_YIELD) reduce 102) ((T_YIELD_FROM) reduce 102) (("[") reduce 102) (("`") reduce 102) (("{") reduce 102) (("}") reduce 102) (("~") reduce 102)) 27266 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONST) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAMESPACE) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_USE) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 27267 ((("!") reduce 236) (("\"") reduce 236) (("$") reduce 236) (("(") reduce 236) (("+") reduce 236) (("-") reduce 236) ((";") reduce 236) (("@") reduce 236) ((T_ABSTRACT) reduce 236) ((T_ARRAY) reduce 236) ((T_ARRAY_CAST) reduce 236) ((T_ATTRIBUTE) reduce 236) ((T_BOOL_CAST) reduce 236) ((T_BREAK) reduce 236) ((T_CLASS) reduce 236) ((T_CLASS_C) reduce 236) ((T_CLONE) reduce 236) ((T_CONST) reduce 236) ((T_CONSTANT_ENCAPSED_STRING) reduce 236) ((T_CONTINUE) reduce 236) ((T_DEC) reduce 236) ((T_DECLARE) reduce 236) ((T_DIR) reduce 236) ((T_DNUMBER) reduce 236) ((T_DO) reduce 236) ((T_DOUBLE_CAST) reduce 236) ((T_ECHO) reduce 236) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 236) ((T_ENUM) reduce 236) ((T_EVAL) reduce 236) ((T_EXIT) reduce 236) ((T_FILE) reduce 236) ((T_FINAL) reduce 236) ((T_FN) reduce 236) ((T_FOR) reduce 236) ((T_FOREACH) reduce 236) ((T_FUNCTION) reduce 236) ((T_FUNC_C) reduce 236) ((T_GLOBAL) reduce 236) ((T_GOTO) reduce 236) ((T_HALT_COMPILER) reduce 236) ((T_IF) reduce 236) ((T_INC) reduce 236) ((T_INCLUDE) reduce 236) ((T_INCLUDE_ONCE) reduce 236) ((T_INLINE_HTML) reduce 236) ((T_INTERFACE) reduce 236) ((T_INT_CAST) reduce 236) ((T_ISSET) reduce 236) ((T_LINE) reduce 236) ((T_LIST) reduce 236) ((T_LNUMBER) reduce 236) ((T_MATCH) reduce 236) ((T_METHOD_C) reduce 236) ((T_NAMESPACE) reduce 236) ((T_NAME_FULLY_QUALIFIED) reduce 236) ((T_NAME_QUALIFIED) reduce 236) ((T_NAME_RELATIVE) reduce 236) ((T_NEW) reduce 236) ((T_NS_C) reduce 236) ((T_OBJECT_CAST) reduce 236) ((T_PRINT) reduce 236) ((T_REQUIRE) reduce 236) ((T_REQUIRE_ONCE) reduce 236) ((T_RETURN) reduce 236) ((T_START_HEREDOC) reduce 236) ((T_STATIC) reduce 236) ((T_STRING) reduce 236) ((T_STRING_CAST) reduce 236) ((T_SWITCH) reduce 236) ((T_THROW) reduce 236) ((T_TRAIT) reduce 236) ((T_TRAIT_C) reduce 236) ((T_TRY) reduce 236) ((T_UNSET) reduce 236) ((T_UNSET_CAST) reduce 236) ((T_USE) reduce 236) ((T_VARIABLE) reduce 236) ((T_WHILE) reduce 236) ((T_YIELD) reduce 236) ((T_YIELD_FROM) reduce 236) (("[") reduce 236) (("`") reduce 236) (("{") reduce 236) (("}") reduce 236) (("~") reduce 236)) 27268 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONST) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAMESPACE) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_USE) reduce 105) ((T_VARIABLE) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("}") reduce 105) (("~") reduce 105)) 27269 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONST) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAMESPACE) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_USE) reduce 107) ((T_VARIABLE) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("}") reduce 107) (("~") reduce 107)) 27270 ((("!") reduce 83) (("\"") reduce 83) (("$") reduce 83) (("(") reduce 83) (("+") reduce 83) (("-") reduce 83) ((";") reduce 83) (("@") reduce 83) ((T_ABSTRACT) reduce 83) ((T_ARRAY) reduce 83) ((T_ARRAY_CAST) reduce 83) ((T_ATTRIBUTE) reduce 83) ((T_BOOL_CAST) reduce 83) ((T_BREAK) reduce 83) ((T_CLASS) reduce 83) ((T_CLASS_C) reduce 83) ((T_CLONE) reduce 83) ((T_CONST) reduce 83) ((T_CONSTANT_ENCAPSED_STRING) reduce 83) ((T_CONTINUE) reduce 83) ((T_DEC) reduce 83) ((T_DECLARE) reduce 83) ((T_DIR) reduce 83) ((T_DNUMBER) reduce 83) ((T_DO) reduce 83) ((T_DOUBLE_CAST) reduce 83) ((T_ECHO) reduce 83) ((T_EMPTY) reduce 83) ((T_ENUM) reduce 83) ((T_EVAL) reduce 83) ((T_EXIT) reduce 83) ((T_FILE) reduce 83) ((T_FINAL) reduce 83) ((T_FN) reduce 83) ((T_FOR) reduce 83) ((T_FOREACH) reduce 83) ((T_FUNCTION) reduce 83) ((T_FUNC_C) reduce 83) ((T_GLOBAL) reduce 83) ((T_GOTO) reduce 83) ((T_HALT_COMPILER) reduce 83) ((T_IF) reduce 83) ((T_INC) reduce 83) ((T_INCLUDE) reduce 83) ((T_INCLUDE_ONCE) reduce 83) ((T_INLINE_HTML) reduce 83) ((T_INTERFACE) reduce 83) ((T_INT_CAST) reduce 83) ((T_ISSET) reduce 83) ((T_LINE) reduce 83) ((T_LIST) reduce 83) ((T_LNUMBER) reduce 83) ((T_MATCH) reduce 83) ((T_METHOD_C) reduce 83) ((T_NAMESPACE) reduce 83) ((T_NAME_FULLY_QUALIFIED) reduce 83) ((T_NAME_QUALIFIED) reduce 83) ((T_NAME_RELATIVE) reduce 83) ((T_NEW) reduce 83) ((T_NS_C) reduce 83) ((T_OBJECT_CAST) reduce 83) ((T_PRINT) reduce 83) ((T_REQUIRE) reduce 83) ((T_REQUIRE_ONCE) reduce 83) ((T_RETURN) reduce 83) ((T_START_HEREDOC) reduce 83) ((T_STATIC) reduce 83) ((T_STRING) reduce 83) ((T_STRING_CAST) reduce 83) ((T_SWITCH) reduce 83) ((T_THROW) reduce 83) ((T_TRAIT) reduce 83) ((T_TRAIT_C) reduce 83) ((T_TRY) reduce 83) ((T_UNSET) reduce 83) ((T_UNSET_CAST) reduce 83) ((T_USE) reduce 83) ((T_VARIABLE) reduce 83) ((T_WHILE) reduce 83) ((T_YIELD) reduce 83) ((T_YIELD_FROM) reduce 83) (("[") reduce 83) (("`") reduce 83) (("{") reduce 83) (("}") reduce 83) (("~") reduce 83)) 27271 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONST) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAMESPACE) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_USE) reduce 104) ((T_VARIABLE) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("}") reduce 104) (("~") reduce 104)) 27273 ((("!") reduce 113) (("\"") reduce 113) (($) reduce 113) (("$") reduce 113) (("(") reduce 113) (("+") reduce 113) (("-") reduce 113) ((";") reduce 113) (("@") reduce 113) ((T_ABSTRACT) reduce 113) ((T_ARRAY) reduce 113) ((T_ARRAY_CAST) reduce 113) ((T_ATTRIBUTE) reduce 113) ((T_BOOL_CAST) reduce 113) ((T_BREAK) reduce 113) ((T_CLASS) reduce 113) ((T_CLASS_C) reduce 113) ((T_CLONE) reduce 113) ((T_CONST) reduce 113) ((T_CONSTANT_ENCAPSED_STRING) reduce 113) ((T_CONTINUE) reduce 113) ((T_DEC) reduce 113) ((T_DECLARE) reduce 113) ((T_DIR) reduce 113) ((T_DNUMBER) reduce 113) ((T_DO) reduce 113) ((T_DOUBLE_CAST) reduce 113) ((T_ECHO) reduce 113) ((T_EMPTY) reduce 113) ((T_ENUM) reduce 113) ((T_EVAL) reduce 113) ((T_EXIT) reduce 113) ((T_FILE) reduce 113) ((T_FINAL) reduce 113) ((T_FN) reduce 113) ((T_FOR) reduce 113) ((T_FOREACH) reduce 113) ((T_FUNCTION) reduce 113) ((T_FUNC_C) reduce 113) ((T_GLOBAL) reduce 113) ((T_GOTO) reduce 113) ((T_HALT_COMPILER) reduce 113) ((T_IF) reduce 113) ((T_INC) reduce 113) ((T_INCLUDE) reduce 113) ((T_INCLUDE_ONCE) reduce 113) ((T_INLINE_HTML) reduce 113) ((T_INTERFACE) reduce 113) ((T_INT_CAST) reduce 113) ((T_ISSET) reduce 113) ((T_LINE) reduce 113) ((T_LIST) reduce 113) ((T_LNUMBER) reduce 113) ((T_MATCH) reduce 113) ((T_METHOD_C) reduce 113) ((T_NAMESPACE) reduce 113) ((T_NAME_FULLY_QUALIFIED) reduce 113) ((T_NAME_QUALIFIED) reduce 113) ((T_NAME_RELATIVE) reduce 113) ((T_NEW) reduce 113) ((T_NS_C) reduce 113) ((T_OBJECT_CAST) reduce 113) ((T_PRINT) reduce 113) ((T_REQUIRE) reduce 113) ((T_REQUIRE_ONCE) reduce 113) ((T_RETURN) reduce 113) ((T_START_HEREDOC) reduce 113) ((T_STATIC) reduce 113) ((T_STRING) reduce 113) ((T_STRING_CAST) reduce 113) ((T_SWITCH) reduce 113) ((T_THROW) reduce 113) ((T_TRAIT) reduce 113) ((T_TRAIT_C) reduce 113) ((T_TRY) reduce 113) ((T_UNSET) reduce 113) ((T_UNSET_CAST) reduce 113) ((T_USE) reduce 113) ((T_VARIABLE) reduce 113) ((T_WHILE) reduce 113) ((T_YIELD) reduce 113) ((T_YIELD_FROM) reduce 113) (("[") reduce 113) (("`") reduce 113) (("{") reduce 113) (("~") reduce 113)) 27275 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONST) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAMESPACE) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_USE) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("}") reduce 145) (("~") reduce 145)) 27281 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 27291 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 27299 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 27302 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONST) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAMESPACE) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_USE) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 27304 ((("!") reduce 235) (("\"") reduce 235) (("$") reduce 235) (("(") reduce 235) (("+") reduce 235) (("-") reduce 235) ((";") reduce 235) (("@") reduce 235) ((T_ABSTRACT) reduce 235) ((T_ARRAY) reduce 235) ((T_ARRAY_CAST) reduce 235) ((T_ATTRIBUTE) reduce 235) ((T_BOOL_CAST) reduce 235) ((T_BREAK) reduce 235) ((T_CLASS) reduce 235) ((T_CLASS_C) reduce 235) ((T_CLONE) reduce 235) ((T_CONST) reduce 235) ((T_CONSTANT_ENCAPSED_STRING) reduce 235) ((T_CONTINUE) reduce 235) ((T_DEC) reduce 235) ((T_DECLARE) reduce 235) ((T_DIR) reduce 235) ((T_DNUMBER) reduce 235) ((T_DO) reduce 235) ((T_DOUBLE_CAST) reduce 235) ((T_ECHO) reduce 235) ((T_ELSE) reduce 235) ((T_ELSEIF) reduce 235) ((T_EMPTY) reduce 235) ((T_ENUM) reduce 235) ((T_EVAL) reduce 235) ((T_EXIT) reduce 235) ((T_FILE) reduce 235) ((T_FINAL) reduce 235) ((T_FN) reduce 235) ((T_FOR) reduce 235) ((T_FOREACH) reduce 235) ((T_FUNCTION) reduce 235) ((T_FUNC_C) reduce 235) ((T_GLOBAL) reduce 235) ((T_GOTO) reduce 235) ((T_HALT_COMPILER) reduce 235) ((T_IF) reduce 235) ((T_INC) reduce 235) ((T_INCLUDE) reduce 235) ((T_INCLUDE_ONCE) reduce 235) ((T_INLINE_HTML) reduce 235) ((T_INTERFACE) reduce 235) ((T_INT_CAST) reduce 235) ((T_ISSET) reduce 235) ((T_LINE) reduce 235) ((T_LIST) reduce 235) ((T_LNUMBER) reduce 235) ((T_MATCH) reduce 235) ((T_METHOD_C) reduce 235) ((T_NAMESPACE) reduce 235) ((T_NAME_FULLY_QUALIFIED) reduce 235) ((T_NAME_QUALIFIED) reduce 235) ((T_NAME_RELATIVE) reduce 235) ((T_NEW) reduce 235) ((T_NS_C) reduce 235) ((T_OBJECT_CAST) reduce 235) ((T_PRINT) reduce 235) ((T_REQUIRE) reduce 235) ((T_REQUIRE_ONCE) reduce 235) ((T_RETURN) reduce 235) ((T_START_HEREDOC) reduce 235) ((T_STATIC) reduce 235) ((T_STRING) reduce 235) ((T_STRING_CAST) reduce 235) ((T_SWITCH) reduce 235) ((T_THROW) reduce 235) ((T_TRAIT) reduce 235) ((T_TRAIT_C) reduce 235) ((T_TRY) reduce 235) ((T_UNSET) reduce 235) ((T_UNSET_CAST) reduce 235) ((T_USE) reduce 235) ((T_VARIABLE) reduce 235) ((T_WHILE) reduce 235) ((T_YIELD) reduce 235) ((T_YIELD_FROM) reduce 235) (("[") reduce 235) (("`") reduce 235) (("{") reduce 235) (("}") reduce 235) (("~") reduce 235)) 27307 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONST) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAMESPACE) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_USE) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("}") reduce 145) (("~") reduce 145)) 27309 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONST) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_ELSE) reduce 237) ((T_ELSEIF) reduce 237) ((T_EMPTY) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAMESPACE) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_USE) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("}") reduce 237) (("~") reduce 237)) 27310 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 27313 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONST) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAMESPACE) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_USE) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("}") reduce 240) (("~") reduce 240)) 27317 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONST) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAMESPACE) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_USE) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("}") reduce 241) (("~") reduce 241)) 27322 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONST) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_ELSE) reduce 232) ((T_ELSEIF) reduce 232) ((T_EMPTY) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAMESPACE) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_USE) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("}") reduce 232) (("~") reduce 232)) 27323 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 27326 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONST) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_ELSE) reduce 233) ((T_ELSEIF) reduce 233) ((T_EMPTY) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAMESPACE) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_USE) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("}") reduce 233) (("~") reduce 233)) 27331 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 27334 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 27335 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONST) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_ELSE) reduce 174) ((T_ELSEIF) reduce 174) ((T_EMPTY) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAMESPACE) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_USE) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("}") reduce 174) (("~") reduce 174)) 27338 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 27341 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONST) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAMESPACE) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_USE) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("}") reduce 175) (("~") reduce 175)) 27348 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 27353 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 27357 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONST) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_ELSE) reduce 214) ((T_ELSEIF) reduce 214) ((T_EMPTY) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAMESPACE) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_USE) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("}") reduce 214) (("~") reduce 214)) 27359 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONST) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_ELSE) reduce 215) ((T_ELSEIF) reduce 215) ((T_EMPTY) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAMESPACE) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_USE) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("}") reduce 215) (("~") reduce 215)) 27363 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONST) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAMESPACE) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_USE) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("}") reduce 216) (("~") reduce 216)) 27366 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONST) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAMESPACE) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_USE) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("}") reduce 217) (("~") reduce 217)) 27367 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 27369 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 27371 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 27373 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 27375 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 27385 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 27386 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONST) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_ELSE) reduce 210) ((T_ELSEIF) reduce 210) ((T_EMPTY) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAMESPACE) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_USE) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("}") reduce 210) (("~") reduce 210)) 27389 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONST) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_ELSE) reduce 211) ((T_ELSEIF) reduce 211) ((T_EMPTY) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAMESPACE) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_USE) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("}") reduce 211) (("~") reduce 211)) 27390 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 27399 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 27400 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONST) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_ELSE) reduce 208) ((T_ELSEIF) reduce 208) ((T_EMPTY) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAMESPACE) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_USE) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("}") reduce 208) (("~") reduce 208)) 27403 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONST) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_ELSE) reduce 209) ((T_ELSEIF) reduce 209) ((T_EMPTY) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAMESPACE) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_USE) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("}") reduce 209) (("~") reduce 209)) 27405 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 27411 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 27416 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 27417 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONST) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_ELSE) reduce 212) ((T_ELSEIF) reduce 212) ((T_EMPTY) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAMESPACE) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_USE) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("}") reduce 212) (("~") reduce 212)) 27420 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONST) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_ELSE) reduce 213) ((T_ELSEIF) reduce 213) ((T_EMPTY) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAMESPACE) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_USE) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("}") reduce 213) (("~") reduce 213)) 27422 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 27424 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 27425 ((("!") reduce 237) (("\"") reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONST) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_EMPTY) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAMESPACE) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_USE) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("}") reduce 237) (("~") reduce 237)) 27436 ((("!") reduce 444) (("\"") reduce 444) (("$") reduce 444) (("(") reduce 444) (("+") reduce 444) (("-") reduce 444) ((";") reduce 444) (("@") reduce 444) ((T_ABSTRACT) reduce 444) ((T_ARRAY) reduce 444) ((T_ARRAY_CAST) reduce 444) ((T_ATTRIBUTE) reduce 444) ((T_BOOL_CAST) reduce 444) ((T_BREAK) reduce 444) ((T_CLASS) reduce 444) ((T_CLASS_C) reduce 444) ((T_CLONE) reduce 444) ((T_CONST) reduce 444) ((T_CONSTANT_ENCAPSED_STRING) reduce 444) ((T_CONTINUE) reduce 444) ((T_DEC) reduce 444) ((T_DECLARE) reduce 444) ((T_DIR) reduce 444) ((T_DNUMBER) reduce 444) ((T_DO) reduce 444) ((T_DOUBLE_CAST) reduce 444) ((T_ECHO) reduce 444) ((T_EMPTY) reduce 444) ((T_ENUM) reduce 444) ((T_EVAL) reduce 444) ((T_EXIT) reduce 444) ((T_FILE) reduce 444) ((T_FINAL) reduce 444) ((T_FN) reduce 444) ((T_FOR) reduce 444) ((T_FOREACH) reduce 444) ((T_FUNCTION) reduce 444) ((T_FUNC_C) reduce 444) ((T_GLOBAL) reduce 444) ((T_GOTO) reduce 444) ((T_HALT_COMPILER) reduce 444) ((T_IF) reduce 444) ((T_INC) reduce 444) ((T_INCLUDE) reduce 444) ((T_INCLUDE_ONCE) reduce 444) ((T_INLINE_HTML) reduce 444) ((T_INTERFACE) reduce 444) ((T_INT_CAST) reduce 444) ((T_ISSET) reduce 444) ((T_LINE) reduce 444) ((T_LIST) reduce 444) ((T_LNUMBER) reduce 444) ((T_MATCH) reduce 444) ((T_METHOD_C) reduce 444) ((T_NAMESPACE) reduce 444) ((T_NAME_FULLY_QUALIFIED) reduce 444) ((T_NAME_QUALIFIED) reduce 444) ((T_NAME_RELATIVE) reduce 444) ((T_NEW) reduce 444) ((T_NS_C) reduce 444) ((T_OBJECT_CAST) reduce 444) ((T_PRINT) reduce 444) ((T_REQUIRE) reduce 444) ((T_REQUIRE_ONCE) reduce 444) ((T_RETURN) reduce 444) ((T_START_HEREDOC) reduce 444) ((T_STATIC) reduce 444) ((T_STRING) reduce 444) ((T_STRING_CAST) reduce 444) ((T_SWITCH) reduce 444) ((T_THROW) reduce 444) ((T_TRAIT) reduce 444) ((T_TRAIT_C) reduce 444) ((T_TRY) reduce 444) ((T_UNSET) reduce 444) ((T_UNSET_CAST) reduce 444) ((T_USE) reduce 444) ((T_VARIABLE) reduce 444) ((T_WHILE) reduce 444) ((T_YIELD) reduce 444) ((T_YIELD_FROM) reduce 444) (("[") reduce 444) (("`") reduce 444) (("{") reduce 444) (("}") reduce 444) (("~") reduce 444)) 27437 ((("!") reduce 179) (("\"") reduce 179) (("$") reduce 179) (("(") reduce 179) (("+") reduce 179) (("-") reduce 179) ((";") reduce 179) (("@") reduce 179) ((T_ABSTRACT) reduce 179) ((T_ARRAY) reduce 179) ((T_ARRAY_CAST) reduce 179) ((T_ATTRIBUTE) reduce 179) ((T_BOOL_CAST) reduce 179) ((T_BREAK) reduce 179) ((T_CLASS) reduce 179) ((T_CLASS_C) reduce 179) ((T_CLONE) reduce 179) ((T_CONST) reduce 179) ((T_CONSTANT_ENCAPSED_STRING) reduce 179) ((T_CONTINUE) reduce 179) ((T_DEC) reduce 179) ((T_DECLARE) reduce 179) ((T_DIR) reduce 179) ((T_DNUMBER) reduce 179) ((T_DO) reduce 179) ((T_DOUBLE_CAST) reduce 179) ((T_ECHO) reduce 179) ((T_EMPTY) reduce 179) ((T_ENUM) reduce 179) ((T_EVAL) reduce 179) ((T_EXIT) reduce 179) ((T_FILE) reduce 179) ((T_FINAL) reduce 179) ((T_FN) reduce 179) ((T_FOR) reduce 179) ((T_FOREACH) reduce 179) ((T_FUNCTION) reduce 179) ((T_FUNC_C) reduce 179) ((T_GLOBAL) reduce 179) ((T_GOTO) reduce 179) ((T_HALT_COMPILER) reduce 179) ((T_IF) reduce 179) ((T_INC) reduce 179) ((T_INCLUDE) reduce 179) ((T_INCLUDE_ONCE) reduce 179) ((T_INLINE_HTML) reduce 179) ((T_INTERFACE) reduce 179) ((T_INT_CAST) reduce 179) ((T_ISSET) reduce 179) ((T_LINE) reduce 179) ((T_LIST) reduce 179) ((T_LNUMBER) reduce 179) ((T_MATCH) reduce 179) ((T_METHOD_C) reduce 179) ((T_NAMESPACE) reduce 179) ((T_NAME_FULLY_QUALIFIED) reduce 179) ((T_NAME_QUALIFIED) reduce 179) ((T_NAME_RELATIVE) reduce 179) ((T_NEW) reduce 179) ((T_NS_C) reduce 179) ((T_OBJECT_CAST) reduce 179) ((T_PRINT) reduce 179) ((T_REQUIRE) reduce 179) ((T_REQUIRE_ONCE) reduce 179) ((T_RETURN) reduce 179) ((T_START_HEREDOC) reduce 179) ((T_STATIC) reduce 179) ((T_STRING) reduce 179) ((T_STRING_CAST) reduce 179) ((T_SWITCH) reduce 179) ((T_THROW) reduce 179) ((T_TRAIT) reduce 179) ((T_TRAIT_C) reduce 179) ((T_TRY) reduce 179) ((T_UNSET) reduce 179) ((T_UNSET_CAST) reduce 179) ((T_USE) reduce 179) ((T_VARIABLE) reduce 179) ((T_WHILE) reduce 179) ((T_YIELD) reduce 179) ((T_YIELD_FROM) reduce 179) (("[") reduce 179) (("`") reduce 179) (("{") reduce 179) (("}") reduce 179) (("~") reduce 179)) 27438 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 27446 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONST) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAMESPACE) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_USE) reduce 184) ((T_VARIABLE) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("}") reduce 184) (("~") reduce 184)) 27447 ((("!") reduce 109) (("\"") reduce 109) (("$") reduce 109) (("(") reduce 109) (("+") reduce 109) (("-") reduce 109) ((";") reduce 109) (("@") reduce 109) ((T_ABSTRACT) reduce 109) ((T_ARRAY) reduce 109) ((T_ARRAY_CAST) reduce 109) ((T_ATTRIBUTE) reduce 109) ((T_BOOL_CAST) reduce 109) ((T_BREAK) reduce 109) ((T_CLASS) reduce 109) ((T_CLASS_C) reduce 109) ((T_CLONE) reduce 109) ((T_CONST) reduce 109) ((T_CONSTANT_ENCAPSED_STRING) reduce 109) ((T_CONTINUE) reduce 109) ((T_DEC) reduce 109) ((T_DECLARE) reduce 109) ((T_DIR) reduce 109) ((T_DNUMBER) reduce 109) ((T_DO) reduce 109) ((T_DOUBLE_CAST) reduce 109) ((T_ECHO) reduce 109) ((T_EMPTY) reduce 109) ((T_ENUM) reduce 109) ((T_EVAL) reduce 109) ((T_EXIT) reduce 109) ((T_FILE) reduce 109) ((T_FINAL) reduce 109) ((T_FN) reduce 109) ((T_FOR) reduce 109) ((T_FOREACH) reduce 109) ((T_FUNCTION) reduce 109) ((T_FUNC_C) reduce 109) ((T_GLOBAL) reduce 109) ((T_GOTO) reduce 109) ((T_HALT_COMPILER) reduce 109) ((T_IF) reduce 109) ((T_INC) reduce 109) ((T_INCLUDE) reduce 109) ((T_INCLUDE_ONCE) reduce 109) ((T_INLINE_HTML) reduce 109) ((T_INTERFACE) reduce 109) ((T_INT_CAST) reduce 109) ((T_ISSET) reduce 109) ((T_LINE) reduce 109) ((T_LIST) reduce 109) ((T_LNUMBER) reduce 109) ((T_MATCH) reduce 109) ((T_METHOD_C) reduce 109) ((T_NAMESPACE) reduce 109) ((T_NAME_FULLY_QUALIFIED) reduce 109) ((T_NAME_QUALIFIED) reduce 109) ((T_NAME_RELATIVE) reduce 109) ((T_NEW) reduce 109) ((T_NS_C) reduce 109) ((T_OBJECT_CAST) reduce 109) ((T_PRINT) reduce 109) ((T_REQUIRE) reduce 109) ((T_REQUIRE_ONCE) reduce 109) ((T_RETURN) reduce 109) ((T_START_HEREDOC) reduce 109) ((T_STATIC) reduce 109) ((T_STRING) reduce 109) ((T_STRING_CAST) reduce 109) ((T_SWITCH) reduce 109) ((T_THROW) reduce 109) ((T_TRAIT) reduce 109) ((T_TRAIT_C) reduce 109) ((T_TRY) reduce 109) ((T_UNSET) reduce 109) ((T_UNSET_CAST) reduce 109) ((T_USE) reduce 109) ((T_VARIABLE) reduce 109) ((T_WHILE) reduce 109) ((T_YIELD) reduce 109) ((T_YIELD_FROM) reduce 109) (("[") reduce 109) (("`") reduce 109) (("{") reduce 109) (("}") reduce 109) (("~") reduce 109)) 27450 ((("!") reduce 240) (("\"") reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONST) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAMESPACE) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_USE) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("}") reduce 240) (("~") reduce 240)) 27454 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONST) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAMESPACE) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_USE) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("}") reduce 241) (("~") reduce 241)) 27459 ((("!") reduce 232) (("\"") reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONST) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_EMPTY) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAMESPACE) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_USE) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("}") reduce 232) (("~") reduce 232)) 27460 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 27463 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONST) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_EMPTY) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAMESPACE) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_USE) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("}") reduce 233) (("~") reduce 233)) 27469 ((("!") reduce 117) (("\"") reduce 117) (("$") reduce 117) (("(") reduce 117) (("+") reduce 117) (("-") reduce 117) ((";") reduce 117) (("@") reduce 117) ((T_ABSTRACT) reduce 117) ((T_ARRAY) reduce 117) ((T_ARRAY_CAST) reduce 117) ((T_ATTRIBUTE) reduce 117) ((T_BOOL_CAST) reduce 117) ((T_BREAK) reduce 117) ((T_CLASS) reduce 117) ((T_CLASS_C) reduce 117) ((T_CLONE) reduce 117) ((T_CONST) reduce 117) ((T_CONSTANT_ENCAPSED_STRING) reduce 117) ((T_CONTINUE) reduce 117) ((T_DEC) reduce 117) ((T_DECLARE) reduce 117) ((T_DIR) reduce 117) ((T_DNUMBER) reduce 117) ((T_DO) reduce 117) ((T_DOUBLE_CAST) reduce 117) ((T_ECHO) reduce 117) ((T_EMPTY) reduce 117) ((T_ENUM) reduce 117) ((T_EVAL) reduce 117) ((T_EXIT) reduce 117) ((T_FILE) reduce 117) ((T_FINAL) reduce 117) ((T_FN) reduce 117) ((T_FOR) reduce 117) ((T_FOREACH) reduce 117) ((T_FUNCTION) reduce 117) ((T_FUNC_C) reduce 117) ((T_GLOBAL) reduce 117) ((T_GOTO) reduce 117) ((T_HALT_COMPILER) reduce 117) ((T_IF) reduce 117) ((T_INC) reduce 117) ((T_INCLUDE) reduce 117) ((T_INCLUDE_ONCE) reduce 117) ((T_INLINE_HTML) reduce 117) ((T_INTERFACE) reduce 117) ((T_INT_CAST) reduce 117) ((T_ISSET) reduce 117) ((T_LINE) reduce 117) ((T_LIST) reduce 117) ((T_LNUMBER) reduce 117) ((T_MATCH) reduce 117) ((T_METHOD_C) reduce 117) ((T_NAMESPACE) reduce 117) ((T_NAME_FULLY_QUALIFIED) reduce 117) ((T_NAME_QUALIFIED) reduce 117) ((T_NAME_RELATIVE) reduce 117) ((T_NEW) reduce 117) ((T_NS_C) reduce 117) ((T_OBJECT_CAST) reduce 117) ((T_PRINT) reduce 117) ((T_REQUIRE) reduce 117) ((T_REQUIRE_ONCE) reduce 117) ((T_RETURN) reduce 117) ((T_START_HEREDOC) reduce 117) ((T_STATIC) reduce 117) ((T_STRING) reduce 117) ((T_STRING_CAST) reduce 117) ((T_SWITCH) reduce 117) ((T_THROW) reduce 117) ((T_TRAIT) reduce 117) ((T_TRAIT_C) reduce 117) ((T_TRY) reduce 117) ((T_UNSET) reduce 117) ((T_UNSET_CAST) reduce 117) ((T_USE) reduce 117) ((T_VARIABLE) reduce 117) ((T_WHILE) reduce 117) ((T_YIELD) reduce 117) ((T_YIELD_FROM) reduce 117) (("[") reduce 117) (("`") reduce 117) (("{") reduce 117) (("}") reduce 117) (("~") reduce 117)) 27470 ((("!") reduce 115) (("\"") reduce 115) (("$") reduce 115) (("(") reduce 115) (("+") reduce 115) (("-") reduce 115) ((";") reduce 115) (("@") reduce 115) ((T_ABSTRACT) reduce 115) ((T_ARRAY) reduce 115) ((T_ARRAY_CAST) reduce 115) ((T_ATTRIBUTE) reduce 115) ((T_BOOL_CAST) reduce 115) ((T_BREAK) reduce 115) ((T_CLASS) reduce 115) ((T_CLASS_C) reduce 115) ((T_CLONE) reduce 115) ((T_CONST) reduce 115) ((T_CONSTANT_ENCAPSED_STRING) reduce 115) ((T_CONTINUE) reduce 115) ((T_DEC) reduce 115) ((T_DECLARE) reduce 115) ((T_DIR) reduce 115) ((T_DNUMBER) reduce 115) ((T_DO) reduce 115) ((T_DOUBLE_CAST) reduce 115) ((T_ECHO) reduce 115) ((T_EMPTY) reduce 115) ((T_ENUM) reduce 115) ((T_EVAL) reduce 115) ((T_EXIT) reduce 115) ((T_FILE) reduce 115) ((T_FINAL) reduce 115) ((T_FN) reduce 115) ((T_FOR) reduce 115) ((T_FOREACH) reduce 115) ((T_FUNCTION) reduce 115) ((T_FUNC_C) reduce 115) ((T_GLOBAL) reduce 115) ((T_GOTO) reduce 115) ((T_HALT_COMPILER) reduce 115) ((T_IF) reduce 115) ((T_INC) reduce 115) ((T_INCLUDE) reduce 115) ((T_INCLUDE_ONCE) reduce 115) ((T_INLINE_HTML) reduce 115) ((T_INTERFACE) reduce 115) ((T_INT_CAST) reduce 115) ((T_ISSET) reduce 115) ((T_LINE) reduce 115) ((T_LIST) reduce 115) ((T_LNUMBER) reduce 115) ((T_MATCH) reduce 115) ((T_METHOD_C) reduce 115) ((T_NAMESPACE) reduce 115) ((T_NAME_FULLY_QUALIFIED) reduce 115) ((T_NAME_QUALIFIED) reduce 115) ((T_NAME_RELATIVE) reduce 115) ((T_NEW) reduce 115) ((T_NS_C) reduce 115) ((T_OBJECT_CAST) reduce 115) ((T_PRINT) reduce 115) ((T_REQUIRE) reduce 115) ((T_REQUIRE_ONCE) reduce 115) ((T_RETURN) reduce 115) ((T_START_HEREDOC) reduce 115) ((T_STATIC) reduce 115) ((T_STRING) reduce 115) ((T_STRING_CAST) reduce 115) ((T_SWITCH) reduce 115) ((T_THROW) reduce 115) ((T_TRAIT) reduce 115) ((T_TRAIT_C) reduce 115) ((T_TRY) reduce 115) ((T_UNSET) reduce 115) ((T_UNSET_CAST) reduce 115) ((T_USE) reduce 115) ((T_VARIABLE) reduce 115) ((T_WHILE) reduce 115) ((T_YIELD) reduce 115) ((T_YIELD_FROM) reduce 115) (("[") reduce 115) (("`") reduce 115) (("{") reduce 115) (("}") reduce 115) (("~") reduce 115)) 27471 ((("!") reduce 116) (("\"") reduce 116) (("$") reduce 116) (("(") reduce 116) (("+") reduce 116) (("-") reduce 116) ((";") reduce 116) (("@") reduce 116) ((T_ABSTRACT) reduce 116) ((T_ARRAY) reduce 116) ((T_ARRAY_CAST) reduce 116) ((T_ATTRIBUTE) reduce 116) ((T_BOOL_CAST) reduce 116) ((T_BREAK) reduce 116) ((T_CLASS) reduce 116) ((T_CLASS_C) reduce 116) ((T_CLONE) reduce 116) ((T_CONST) reduce 116) ((T_CONSTANT_ENCAPSED_STRING) reduce 116) ((T_CONTINUE) reduce 116) ((T_DEC) reduce 116) ((T_DECLARE) reduce 116) ((T_DIR) reduce 116) ((T_DNUMBER) reduce 116) ((T_DO) reduce 116) ((T_DOUBLE_CAST) reduce 116) ((T_ECHO) reduce 116) ((T_EMPTY) reduce 116) ((T_ENUM) reduce 116) ((T_EVAL) reduce 116) ((T_EXIT) reduce 116) ((T_FILE) reduce 116) ((T_FINAL) reduce 116) ((T_FN) reduce 116) ((T_FOR) reduce 116) ((T_FOREACH) reduce 116) ((T_FUNCTION) reduce 116) ((T_FUNC_C) reduce 116) ((T_GLOBAL) reduce 116) ((T_GOTO) reduce 116) ((T_HALT_COMPILER) reduce 116) ((T_IF) reduce 116) ((T_INC) reduce 116) ((T_INCLUDE) reduce 116) ((T_INCLUDE_ONCE) reduce 116) ((T_INLINE_HTML) reduce 116) ((T_INTERFACE) reduce 116) ((T_INT_CAST) reduce 116) ((T_ISSET) reduce 116) ((T_LINE) reduce 116) ((T_LIST) reduce 116) ((T_LNUMBER) reduce 116) ((T_MATCH) reduce 116) ((T_METHOD_C) reduce 116) ((T_NAMESPACE) reduce 116) ((T_NAME_FULLY_QUALIFIED) reduce 116) ((T_NAME_QUALIFIED) reduce 116) ((T_NAME_RELATIVE) reduce 116) ((T_NEW) reduce 116) ((T_NS_C) reduce 116) ((T_OBJECT_CAST) reduce 116) ((T_PRINT) reduce 116) ((T_REQUIRE) reduce 116) ((T_REQUIRE_ONCE) reduce 116) ((T_RETURN) reduce 116) ((T_START_HEREDOC) reduce 116) ((T_STATIC) reduce 116) ((T_STRING) reduce 116) ((T_STRING_CAST) reduce 116) ((T_SWITCH) reduce 116) ((T_THROW) reduce 116) ((T_TRAIT) reduce 116) ((T_TRAIT_C) reduce 116) ((T_TRY) reduce 116) ((T_UNSET) reduce 116) ((T_UNSET_CAST) reduce 116) ((T_USE) reduce 116) ((T_VARIABLE) reduce 116) ((T_WHILE) reduce 116) ((T_YIELD) reduce 116) ((T_YIELD_FROM) reduce 116) (("[") reduce 116) (("`") reduce 116) (("{") reduce 116) (("}") reduce 116) (("~") reduce 116)) 27472 ((("!") reduce 114) (("\"") reduce 114) (("$") reduce 114) (("(") reduce 114) (("+") reduce 114) (("-") reduce 114) ((";") reduce 114) (("@") reduce 114) ((T_ABSTRACT) reduce 114) ((T_ARRAY) reduce 114) ((T_ARRAY_CAST) reduce 114) ((T_ATTRIBUTE) reduce 114) ((T_BOOL_CAST) reduce 114) ((T_BREAK) reduce 114) ((T_CLASS) reduce 114) ((T_CLASS_C) reduce 114) ((T_CLONE) reduce 114) ((T_CONST) reduce 114) ((T_CONSTANT_ENCAPSED_STRING) reduce 114) ((T_CONTINUE) reduce 114) ((T_DEC) reduce 114) ((T_DECLARE) reduce 114) ((T_DIR) reduce 114) ((T_DNUMBER) reduce 114) ((T_DO) reduce 114) ((T_DOUBLE_CAST) reduce 114) ((T_ECHO) reduce 114) ((T_EMPTY) reduce 114) ((T_ENUM) reduce 114) ((T_EVAL) reduce 114) ((T_EXIT) reduce 114) ((T_FILE) reduce 114) ((T_FINAL) reduce 114) ((T_FN) reduce 114) ((T_FOR) reduce 114) ((T_FOREACH) reduce 114) ((T_FUNCTION) reduce 114) ((T_FUNC_C) reduce 114) ((T_GLOBAL) reduce 114) ((T_GOTO) reduce 114) ((T_HALT_COMPILER) reduce 114) ((T_IF) reduce 114) ((T_INC) reduce 114) ((T_INCLUDE) reduce 114) ((T_INCLUDE_ONCE) reduce 114) ((T_INLINE_HTML) reduce 114) ((T_INTERFACE) reduce 114) ((T_INT_CAST) reduce 114) ((T_ISSET) reduce 114) ((T_LINE) reduce 114) ((T_LIST) reduce 114) ((T_LNUMBER) reduce 114) ((T_MATCH) reduce 114) ((T_METHOD_C) reduce 114) ((T_NAMESPACE) reduce 114) ((T_NAME_FULLY_QUALIFIED) reduce 114) ((T_NAME_QUALIFIED) reduce 114) ((T_NAME_RELATIVE) reduce 114) ((T_NEW) reduce 114) ((T_NS_C) reduce 114) ((T_OBJECT_CAST) reduce 114) ((T_PRINT) reduce 114) ((T_REQUIRE) reduce 114) ((T_REQUIRE_ONCE) reduce 114) ((T_RETURN) reduce 114) ((T_START_HEREDOC) reduce 114) ((T_STATIC) reduce 114) ((T_STRING) reduce 114) ((T_STRING_CAST) reduce 114) ((T_SWITCH) reduce 114) ((T_THROW) reduce 114) ((T_TRAIT) reduce 114) ((T_TRAIT_C) reduce 114) ((T_TRY) reduce 114) ((T_UNSET) reduce 114) ((T_UNSET_CAST) reduce 114) ((T_USE) reduce 114) ((T_VARIABLE) reduce 114) ((T_WHILE) reduce 114) ((T_YIELD) reduce 114) ((T_YIELD_FROM) reduce 114) (("[") reduce 114) (("`") reduce 114) (("{") reduce 114) (("}") reduce 114) (("~") reduce 114)) 27477 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 27480 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 27481 ((("!") reduce 174) (("\"") reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONST) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_EMPTY) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAMESPACE) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_USE) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("}") reduce 174) (("~") reduce 174)) 27484 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 27487 ((("!") reduce 175) (("\"") reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONST) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAMESPACE) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_USE) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("}") reduce 175) (("~") reduce 175)) 27494 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 27499 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONST) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAMESPACE) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_USE) reduce 190) ((T_VARIABLE) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("}") reduce 190) (("~") reduce 190)) 27504 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 27508 ((("!") reduce 214) (("\"") reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONST) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_EMPTY) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAMESPACE) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_USE) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("}") reduce 214) (("~") reduce 214)) 27510 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONST) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_EMPTY) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAMESPACE) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_USE) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("}") reduce 215) (("~") reduce 215)) 27514 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONST) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAMESPACE) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_USE) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("}") reduce 216) (("~") reduce 216)) 27517 ((("!") reduce 217) (("\"") reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONST) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAMESPACE) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_USE) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("}") reduce 217) (("~") reduce 217)) 27518 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 27520 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 27522 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 27526 ((("!") reduce 113) (("\"") reduce 113) (("$") reduce 113) (("(") reduce 113) (("+") reduce 113) (("-") reduce 113) ((";") reduce 113) (("@") reduce 113) ((T_ABSTRACT) reduce 113) ((T_ARRAY) reduce 113) ((T_ARRAY_CAST) reduce 113) ((T_ATTRIBUTE) reduce 113) ((T_BOOL_CAST) reduce 113) ((T_BREAK) reduce 113) ((T_CLASS) reduce 113) ((T_CLASS_C) reduce 113) ((T_CLONE) reduce 113) ((T_CONST) reduce 113) ((T_CONSTANT_ENCAPSED_STRING) reduce 113) ((T_CONTINUE) reduce 113) ((T_DEC) reduce 113) ((T_DECLARE) reduce 113) ((T_DIR) reduce 113) ((T_DNUMBER) reduce 113) ((T_DO) reduce 113) ((T_DOUBLE_CAST) reduce 113) ((T_ECHO) reduce 113) ((T_EMPTY) reduce 113) ((T_ENUM) reduce 113) ((T_EVAL) reduce 113) ((T_EXIT) reduce 113) ((T_FILE) reduce 113) ((T_FINAL) reduce 113) ((T_FN) reduce 113) ((T_FOR) reduce 113) ((T_FOREACH) reduce 113) ((T_FUNCTION) reduce 113) ((T_FUNC_C) reduce 113) ((T_GLOBAL) reduce 113) ((T_GOTO) reduce 113) ((T_HALT_COMPILER) reduce 113) ((T_IF) reduce 113) ((T_INC) reduce 113) ((T_INCLUDE) reduce 113) ((T_INCLUDE_ONCE) reduce 113) ((T_INLINE_HTML) reduce 113) ((T_INTERFACE) reduce 113) ((T_INT_CAST) reduce 113) ((T_ISSET) reduce 113) ((T_LINE) reduce 113) ((T_LIST) reduce 113) ((T_LNUMBER) reduce 113) ((T_MATCH) reduce 113) ((T_METHOD_C) reduce 113) ((T_NAMESPACE) reduce 113) ((T_NAME_FULLY_QUALIFIED) reduce 113) ((T_NAME_QUALIFIED) reduce 113) ((T_NAME_RELATIVE) reduce 113) ((T_NEW) reduce 113) ((T_NS_C) reduce 113) ((T_OBJECT_CAST) reduce 113) ((T_PRINT) reduce 113) ((T_REQUIRE) reduce 113) ((T_REQUIRE_ONCE) reduce 113) ((T_RETURN) reduce 113) ((T_START_HEREDOC) reduce 113) ((T_STATIC) reduce 113) ((T_STRING) reduce 113) ((T_STRING_CAST) reduce 113) ((T_SWITCH) reduce 113) ((T_THROW) reduce 113) ((T_TRAIT) reduce 113) ((T_TRAIT_C) reduce 113) ((T_TRY) reduce 113) ((T_UNSET) reduce 113) ((T_UNSET_CAST) reduce 113) ((T_USE) reduce 113) ((T_VARIABLE) reduce 113) ((T_WHILE) reduce 113) ((T_YIELD) reduce 113) ((T_YIELD_FROM) reduce 113) (("[") reduce 113) (("`") reduce 113) (("{") reduce 113) (("}") reduce 113) (("~") reduce 113)) 27527 ((("!") reduce 111) (("\"") reduce 111) (("$") reduce 111) (("(") reduce 111) (("+") reduce 111) (("-") reduce 111) ((";") reduce 111) (("@") reduce 111) ((T_ABSTRACT) reduce 111) ((T_ARRAY) reduce 111) ((T_ARRAY_CAST) reduce 111) ((T_ATTRIBUTE) reduce 111) ((T_BOOL_CAST) reduce 111) ((T_BREAK) reduce 111) ((T_CLASS) reduce 111) ((T_CLASS_C) reduce 111) ((T_CLONE) reduce 111) ((T_CONST) reduce 111) ((T_CONSTANT_ENCAPSED_STRING) reduce 111) ((T_CONTINUE) reduce 111) ((T_DEC) reduce 111) ((T_DECLARE) reduce 111) ((T_DIR) reduce 111) ((T_DNUMBER) reduce 111) ((T_DO) reduce 111) ((T_DOUBLE_CAST) reduce 111) ((T_ECHO) reduce 111) ((T_EMPTY) reduce 111) ((T_ENUM) reduce 111) ((T_EVAL) reduce 111) ((T_EXIT) reduce 111) ((T_FILE) reduce 111) ((T_FINAL) reduce 111) ((T_FN) reduce 111) ((T_FOR) reduce 111) ((T_FOREACH) reduce 111) ((T_FUNCTION) reduce 111) ((T_FUNC_C) reduce 111) ((T_GLOBAL) reduce 111) ((T_GOTO) reduce 111) ((T_HALT_COMPILER) reduce 111) ((T_IF) reduce 111) ((T_INC) reduce 111) ((T_INCLUDE) reduce 111) ((T_INCLUDE_ONCE) reduce 111) ((T_INLINE_HTML) reduce 111) ((T_INTERFACE) reduce 111) ((T_INT_CAST) reduce 111) ((T_ISSET) reduce 111) ((T_LINE) reduce 111) ((T_LIST) reduce 111) ((T_LNUMBER) reduce 111) ((T_MATCH) reduce 111) ((T_METHOD_C) reduce 111) ((T_NAMESPACE) reduce 111) ((T_NAME_FULLY_QUALIFIED) reduce 111) ((T_NAME_QUALIFIED) reduce 111) ((T_NAME_RELATIVE) reduce 111) ((T_NEW) reduce 111) ((T_NS_C) reduce 111) ((T_OBJECT_CAST) reduce 111) ((T_PRINT) reduce 111) ((T_REQUIRE) reduce 111) ((T_REQUIRE_ONCE) reduce 111) ((T_RETURN) reduce 111) ((T_START_HEREDOC) reduce 111) ((T_STATIC) reduce 111) ((T_STRING) reduce 111) ((T_STRING_CAST) reduce 111) ((T_SWITCH) reduce 111) ((T_THROW) reduce 111) ((T_TRAIT) reduce 111) ((T_TRAIT_C) reduce 111) ((T_TRY) reduce 111) ((T_UNSET) reduce 111) ((T_UNSET_CAST) reduce 111) ((T_USE) reduce 111) ((T_VARIABLE) reduce 111) ((T_WHILE) reduce 111) ((T_YIELD) reduce 111) ((T_YIELD_FROM) reduce 111) (("[") reduce 111) (("`") reduce 111) (("{") reduce 111) (("}") reduce 111) (("~") reduce 111)) 27530 ((("!") reduce 112) (("\"") reduce 112) (("$") reduce 112) (("(") reduce 112) (("+") reduce 112) (("-") reduce 112) ((";") reduce 112) (("@") reduce 112) ((T_ABSTRACT) reduce 112) ((T_ARRAY) reduce 112) ((T_ARRAY_CAST) reduce 112) ((T_ATTRIBUTE) reduce 112) ((T_BOOL_CAST) reduce 112) ((T_BREAK) reduce 112) ((T_CLASS) reduce 112) ((T_CLASS_C) reduce 112) ((T_CLONE) reduce 112) ((T_CONST) reduce 112) ((T_CONSTANT_ENCAPSED_STRING) reduce 112) ((T_CONTINUE) reduce 112) ((T_DEC) reduce 112) ((T_DECLARE) reduce 112) ((T_DIR) reduce 112) ((T_DNUMBER) reduce 112) ((T_DO) reduce 112) ((T_DOUBLE_CAST) reduce 112) ((T_ECHO) reduce 112) ((T_EMPTY) reduce 112) ((T_ENUM) reduce 112) ((T_EVAL) reduce 112) ((T_EXIT) reduce 112) ((T_FILE) reduce 112) ((T_FINAL) reduce 112) ((T_FN) reduce 112) ((T_FOR) reduce 112) ((T_FOREACH) reduce 112) ((T_FUNCTION) reduce 112) ((T_FUNC_C) reduce 112) ((T_GLOBAL) reduce 112) ((T_GOTO) reduce 112) ((T_HALT_COMPILER) reduce 112) ((T_IF) reduce 112) ((T_INC) reduce 112) ((T_INCLUDE) reduce 112) ((T_INCLUDE_ONCE) reduce 112) ((T_INLINE_HTML) reduce 112) ((T_INTERFACE) reduce 112) ((T_INT_CAST) reduce 112) ((T_ISSET) reduce 112) ((T_LINE) reduce 112) ((T_LIST) reduce 112) ((T_LNUMBER) reduce 112) ((T_MATCH) reduce 112) ((T_METHOD_C) reduce 112) ((T_NAMESPACE) reduce 112) ((T_NAME_FULLY_QUALIFIED) reduce 112) ((T_NAME_QUALIFIED) reduce 112) ((T_NAME_RELATIVE) reduce 112) ((T_NEW) reduce 112) ((T_NS_C) reduce 112) ((T_OBJECT_CAST) reduce 112) ((T_PRINT) reduce 112) ((T_REQUIRE) reduce 112) ((T_REQUIRE_ONCE) reduce 112) ((T_RETURN) reduce 112) ((T_START_HEREDOC) reduce 112) ((T_STATIC) reduce 112) ((T_STRING) reduce 112) ((T_STRING_CAST) reduce 112) ((T_SWITCH) reduce 112) ((T_THROW) reduce 112) ((T_TRAIT) reduce 112) ((T_TRAIT_C) reduce 112) ((T_TRY) reduce 112) ((T_UNSET) reduce 112) ((T_UNSET_CAST) reduce 112) ((T_USE) reduce 112) ((T_VARIABLE) reduce 112) ((T_WHILE) reduce 112) ((T_YIELD) reduce 112) ((T_YIELD_FROM) reduce 112) (("[") reduce 112) (("`") reduce 112) (("{") reduce 112) (("}") reduce 112) (("~") reduce 112)) 27536 ((("!") reduce 191) (("\"") reduce 191) (("$") reduce 191) (("(") reduce 191) (("+") reduce 191) (("-") reduce 191) ((";") reduce 191) (("@") reduce 191) ((T_ABSTRACT) reduce 191) ((T_ARRAY) reduce 191) ((T_ARRAY_CAST) reduce 191) ((T_ATTRIBUTE) reduce 191) ((T_BOOL_CAST) reduce 191) ((T_BREAK) reduce 191) ((T_CLASS) reduce 191) ((T_CLASS_C) reduce 191) ((T_CLONE) reduce 191) ((T_CONST) reduce 191) ((T_CONSTANT_ENCAPSED_STRING) reduce 191) ((T_CONTINUE) reduce 191) ((T_DEC) reduce 191) ((T_DECLARE) reduce 191) ((T_DIR) reduce 191) ((T_DNUMBER) reduce 191) ((T_DO) reduce 191) ((T_DOUBLE_CAST) reduce 191) ((T_ECHO) reduce 191) ((T_EMPTY) reduce 191) ((T_ENUM) reduce 191) ((T_EVAL) reduce 191) ((T_EXIT) reduce 191) ((T_FILE) reduce 191) ((T_FINAL) reduce 191) ((T_FN) reduce 191) ((T_FOR) reduce 191) ((T_FOREACH) reduce 191) ((T_FUNCTION) reduce 191) ((T_FUNC_C) reduce 191) ((T_GLOBAL) reduce 191) ((T_GOTO) reduce 191) ((T_HALT_COMPILER) reduce 191) ((T_IF) reduce 191) ((T_INC) reduce 191) ((T_INCLUDE) reduce 191) ((T_INCLUDE_ONCE) reduce 191) ((T_INLINE_HTML) reduce 191) ((T_INTERFACE) reduce 191) ((T_INT_CAST) reduce 191) ((T_ISSET) reduce 191) ((T_LINE) reduce 191) ((T_LIST) reduce 191) ((T_LNUMBER) reduce 191) ((T_MATCH) reduce 191) ((T_METHOD_C) reduce 191) ((T_NAMESPACE) reduce 191) ((T_NAME_FULLY_QUALIFIED) reduce 191) ((T_NAME_QUALIFIED) reduce 191) ((T_NAME_RELATIVE) reduce 191) ((T_NEW) reduce 191) ((T_NS_C) reduce 191) ((T_OBJECT_CAST) reduce 191) ((T_PRINT) reduce 191) ((T_REQUIRE) reduce 191) ((T_REQUIRE_ONCE) reduce 191) ((T_RETURN) reduce 191) ((T_START_HEREDOC) reduce 191) ((T_STATIC) reduce 191) ((T_STRING) reduce 191) ((T_STRING_CAST) reduce 191) ((T_SWITCH) reduce 191) ((T_THROW) reduce 191) ((T_TRAIT) reduce 191) ((T_TRAIT_C) reduce 191) ((T_TRY) reduce 191) ((T_UNSET) reduce 191) ((T_UNSET_CAST) reduce 191) ((T_USE) reduce 191) ((T_VARIABLE) reduce 191) ((T_WHILE) reduce 191) ((T_YIELD) reduce 191) ((T_YIELD_FROM) reduce 191) (("[") reduce 191) (("`") reduce 191) (("{") reduce 191) (("}") reduce 191) (("~") reduce 191)) 27540 ((("!") reduce 234) (("\"") reduce 234) (("$") reduce 234) (("(") reduce 234) (("+") reduce 234) (("-") reduce 234) ((";") reduce 234) (("@") reduce 234) ((T_ABSTRACT) reduce 234) ((T_ARRAY) reduce 234) ((T_ARRAY_CAST) reduce 234) ((T_ATTRIBUTE) reduce 234) ((T_BOOL_CAST) reduce 234) ((T_BREAK) reduce 234) ((T_CLASS) reduce 234) ((T_CLASS_C) reduce 234) ((T_CLONE) reduce 234) ((T_CONST) reduce 234) ((T_CONSTANT_ENCAPSED_STRING) reduce 234) ((T_CONTINUE) reduce 234) ((T_DEC) reduce 234) ((T_DECLARE) reduce 234) ((T_DIR) reduce 234) ((T_DNUMBER) reduce 234) ((T_DO) reduce 234) ((T_DOUBLE_CAST) reduce 234) ((T_ECHO) reduce 234) ((T_ELSE) reduce 234) ((T_ELSEIF) reduce 234) ((T_EMPTY) reduce 234) ((T_ENUM) reduce 234) ((T_EVAL) reduce 234) ((T_EXIT) reduce 234) ((T_FILE) reduce 234) ((T_FINAL) reduce 234) ((T_FN) reduce 234) ((T_FOR) reduce 234) ((T_FOREACH) reduce 234) ((T_FUNCTION) reduce 234) ((T_FUNC_C) reduce 234) ((T_GLOBAL) reduce 234) ((T_GOTO) reduce 234) ((T_HALT_COMPILER) reduce 234) ((T_IF) reduce 234) ((T_INC) reduce 234) ((T_INCLUDE) reduce 234) ((T_INCLUDE_ONCE) reduce 234) ((T_INLINE_HTML) reduce 234) ((T_INTERFACE) reduce 234) ((T_INT_CAST) reduce 234) ((T_ISSET) reduce 234) ((T_LINE) reduce 234) ((T_LIST) reduce 234) ((T_LNUMBER) reduce 234) ((T_MATCH) reduce 234) ((T_METHOD_C) reduce 234) ((T_NAMESPACE) reduce 234) ((T_NAME_FULLY_QUALIFIED) reduce 234) ((T_NAME_QUALIFIED) reduce 234) ((T_NAME_RELATIVE) reduce 234) ((T_NEW) reduce 234) ((T_NS_C) reduce 234) ((T_OBJECT_CAST) reduce 234) ((T_PRINT) reduce 234) ((T_REQUIRE) reduce 234) ((T_REQUIRE_ONCE) reduce 234) ((T_RETURN) reduce 234) ((T_START_HEREDOC) reduce 234) ((T_STATIC) reduce 234) ((T_STRING) reduce 234) ((T_STRING_CAST) reduce 234) ((T_SWITCH) reduce 234) ((T_THROW) reduce 234) ((T_TRAIT) reduce 234) ((T_TRAIT_C) reduce 234) ((T_TRY) reduce 234) ((T_UNSET) reduce 234) ((T_UNSET_CAST) reduce 234) ((T_USE) reduce 234) ((T_VARIABLE) reduce 234) ((T_WHILE) reduce 234) ((T_YIELD) reduce 234) ((T_YIELD_FROM) reduce 234) (("[") reduce 234) (("`") reduce 234) (("{") reduce 234) (("}") reduce 234) (("~") reduce 234)) 27543 ((("!") reduce 110) (("\"") reduce 110) (("$") reduce 110) (("(") reduce 110) (("+") reduce 110) (("-") reduce 110) ((";") reduce 110) (("@") reduce 110) ((T_ABSTRACT) reduce 110) ((T_ARRAY) reduce 110) ((T_ARRAY_CAST) reduce 110) ((T_ATTRIBUTE) reduce 110) ((T_BOOL_CAST) reduce 110) ((T_BREAK) reduce 110) ((T_CLASS) reduce 110) ((T_CLASS_C) reduce 110) ((T_CLONE) reduce 110) ((T_CONST) reduce 110) ((T_CONSTANT_ENCAPSED_STRING) reduce 110) ((T_CONTINUE) reduce 110) ((T_DEC) reduce 110) ((T_DECLARE) reduce 110) ((T_DIR) reduce 110) ((T_DNUMBER) reduce 110) ((T_DO) reduce 110) ((T_DOUBLE_CAST) reduce 110) ((T_ECHO) reduce 110) ((T_EMPTY) reduce 110) ((T_ENUM) reduce 110) ((T_EVAL) reduce 110) ((T_EXIT) reduce 110) ((T_FILE) reduce 110) ((T_FINAL) reduce 110) ((T_FN) reduce 110) ((T_FOR) reduce 110) ((T_FOREACH) reduce 110) ((T_FUNCTION) reduce 110) ((T_FUNC_C) reduce 110) ((T_GLOBAL) reduce 110) ((T_GOTO) reduce 110) ((T_HALT_COMPILER) reduce 110) ((T_IF) reduce 110) ((T_INC) reduce 110) ((T_INCLUDE) reduce 110) ((T_INCLUDE_ONCE) reduce 110) ((T_INLINE_HTML) reduce 110) ((T_INTERFACE) reduce 110) ((T_INT_CAST) reduce 110) ((T_ISSET) reduce 110) ((T_LINE) reduce 110) ((T_LIST) reduce 110) ((T_LNUMBER) reduce 110) ((T_MATCH) reduce 110) ((T_METHOD_C) reduce 110) ((T_NAMESPACE) reduce 110) ((T_NAME_FULLY_QUALIFIED) reduce 110) ((T_NAME_QUALIFIED) reduce 110) ((T_NAME_RELATIVE) reduce 110) ((T_NEW) reduce 110) ((T_NS_C) reduce 110) ((T_OBJECT_CAST) reduce 110) ((T_PRINT) reduce 110) ((T_REQUIRE) reduce 110) ((T_REQUIRE_ONCE) reduce 110) ((T_RETURN) reduce 110) ((T_START_HEREDOC) reduce 110) ((T_STATIC) reduce 110) ((T_STRING) reduce 110) ((T_STRING_CAST) reduce 110) ((T_SWITCH) reduce 110) ((T_THROW) reduce 110) ((T_TRAIT) reduce 110) ((T_TRAIT_C) reduce 110) ((T_TRY) reduce 110) ((T_UNSET) reduce 110) ((T_UNSET_CAST) reduce 110) ((T_USE) reduce 110) ((T_VARIABLE) reduce 110) ((T_WHILE) reduce 110) ((T_YIELD) reduce 110) ((T_YIELD_FROM) reduce 110) (("[") reduce 110) (("`") reduce 110) (("{") reduce 110) (("}") reduce 110) (("~") reduce 110)) 27545 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 27547 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 27557 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 27558 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONST) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_EMPTY) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAMESPACE) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_USE) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("}") reduce 210) (("~") reduce 210)) 27561 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONST) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_EMPTY) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAMESPACE) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_USE) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("}") reduce 211) (("~") reduce 211)) 27562 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 27571 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 27572 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONST) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_EMPTY) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAMESPACE) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_USE) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("}") reduce 208) (("~") reduce 208)) 27575 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONST) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_EMPTY) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAMESPACE) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_USE) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("}") reduce 209) (("~") reduce 209)) 27582 ((("!") reduce 192) (("\"") reduce 192) (("$") reduce 192) (("(") reduce 192) (("+") reduce 192) (("-") reduce 192) ((";") reduce 192) (("@") reduce 192) ((T_ABSTRACT) reduce 192) ((T_ARRAY) reduce 192) ((T_ARRAY_CAST) reduce 192) ((T_ATTRIBUTE) reduce 192) ((T_BOOL_CAST) reduce 192) ((T_BREAK) reduce 192) ((T_CLASS) reduce 192) ((T_CLASS_C) reduce 192) ((T_CLONE) reduce 192) ((T_CONST) reduce 192) ((T_CONSTANT_ENCAPSED_STRING) reduce 192) ((T_CONTINUE) reduce 192) ((T_DEC) reduce 192) ((T_DECLARE) reduce 192) ((T_DIR) reduce 192) ((T_DNUMBER) reduce 192) ((T_DO) reduce 192) ((T_DOUBLE_CAST) reduce 192) ((T_ECHO) reduce 192) ((T_EMPTY) reduce 192) ((T_ENUM) reduce 192) ((T_EVAL) reduce 192) ((T_EXIT) reduce 192) ((T_FILE) reduce 192) ((T_FINAL) reduce 192) ((T_FN) reduce 192) ((T_FOR) reduce 192) ((T_FOREACH) reduce 192) ((T_FUNCTION) reduce 192) ((T_FUNC_C) reduce 192) ((T_GLOBAL) reduce 192) ((T_GOTO) reduce 192) ((T_HALT_COMPILER) reduce 192) ((T_IF) reduce 192) ((T_INC) reduce 192) ((T_INCLUDE) reduce 192) ((T_INCLUDE_ONCE) reduce 192) ((T_INLINE_HTML) reduce 192) ((T_INTERFACE) reduce 192) ((T_INT_CAST) reduce 192) ((T_ISSET) reduce 192) ((T_LINE) reduce 192) ((T_LIST) reduce 192) ((T_LNUMBER) reduce 192) ((T_MATCH) reduce 192) ((T_METHOD_C) reduce 192) ((T_NAMESPACE) reduce 192) ((T_NAME_FULLY_QUALIFIED) reduce 192) ((T_NAME_QUALIFIED) reduce 192) ((T_NAME_RELATIVE) reduce 192) ((T_NEW) reduce 192) ((T_NS_C) reduce 192) ((T_OBJECT_CAST) reduce 192) ((T_PRINT) reduce 192) ((T_REQUIRE) reduce 192) ((T_REQUIRE_ONCE) reduce 192) ((T_RETURN) reduce 192) ((T_START_HEREDOC) reduce 192) ((T_STATIC) reduce 192) ((T_STRING) reduce 192) ((T_STRING_CAST) reduce 192) ((T_SWITCH) reduce 192) ((T_THROW) reduce 192) ((T_TRAIT) reduce 192) ((T_TRAIT_C) reduce 192) ((T_TRY) reduce 192) ((T_UNSET) reduce 192) ((T_UNSET_CAST) reduce 192) ((T_USE) reduce 192) ((T_VARIABLE) reduce 192) ((T_WHILE) reduce 192) ((T_YIELD) reduce 192) ((T_YIELD_FROM) reduce 192) (("[") reduce 192) (("`") reduce 192) (("{") reduce 192) (("}") reduce 192) (("~") reduce 192)) 27584 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 27590 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 27595 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 27596 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONST) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_EMPTY) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAMESPACE) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_USE) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("}") reduce 212) (("~") reduce 212)) 27599 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONST) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_EMPTY) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAMESPACE) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_USE) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("}") reduce 213) (("~") reduce 213)) 27601 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 27603 (((",") reduce 138) ((";") reduce 138)) 27606 ((("!") reduce 118) (("\"") reduce 118) (("$") reduce 118) (("(") reduce 118) (("+") reduce 118) (("-") reduce 118) ((";") reduce 118) (("@") reduce 118) ((T_ABSTRACT) reduce 118) ((T_ARRAY) reduce 118) ((T_ARRAY_CAST) reduce 118) ((T_ATTRIBUTE) reduce 118) ((T_BOOL_CAST) reduce 118) ((T_BREAK) reduce 118) ((T_CLASS) reduce 118) ((T_CLASS_C) reduce 118) ((T_CLONE) reduce 118) ((T_CONST) reduce 118) ((T_CONSTANT_ENCAPSED_STRING) reduce 118) ((T_CONTINUE) reduce 118) ((T_DEC) reduce 118) ((T_DECLARE) reduce 118) ((T_DIR) reduce 118) ((T_DNUMBER) reduce 118) ((T_DO) reduce 118) ((T_DOUBLE_CAST) reduce 118) ((T_ECHO) reduce 118) ((T_EMPTY) reduce 118) ((T_ENUM) reduce 118) ((T_EVAL) reduce 118) ((T_EXIT) reduce 118) ((T_FILE) reduce 118) ((T_FINAL) reduce 118) ((T_FN) reduce 118) ((T_FOR) reduce 118) ((T_FOREACH) reduce 118) ((T_FUNCTION) reduce 118) ((T_FUNC_C) reduce 118) ((T_GLOBAL) reduce 118) ((T_GOTO) reduce 118) ((T_HALT_COMPILER) reduce 118) ((T_IF) reduce 118) ((T_INC) reduce 118) ((T_INCLUDE) reduce 118) ((T_INCLUDE_ONCE) reduce 118) ((T_INLINE_HTML) reduce 118) ((T_INTERFACE) reduce 118) ((T_INT_CAST) reduce 118) ((T_ISSET) reduce 118) ((T_LINE) reduce 118) ((T_LIST) reduce 118) ((T_LNUMBER) reduce 118) ((T_MATCH) reduce 118) ((T_METHOD_C) reduce 118) ((T_NAMESPACE) reduce 118) ((T_NAME_FULLY_QUALIFIED) reduce 118) ((T_NAME_QUALIFIED) reduce 118) ((T_NAME_RELATIVE) reduce 118) ((T_NEW) reduce 118) ((T_NS_C) reduce 118) ((T_OBJECT_CAST) reduce 118) ((T_PRINT) reduce 118) ((T_REQUIRE) reduce 118) ((T_REQUIRE_ONCE) reduce 118) ((T_RETURN) reduce 118) ((T_START_HEREDOC) reduce 118) ((T_STATIC) reduce 118) ((T_STRING) reduce 118) ((T_STRING_CAST) reduce 118) ((T_SWITCH) reduce 118) ((T_THROW) reduce 118) ((T_TRAIT) reduce 118) ((T_TRAIT_C) reduce 118) ((T_TRY) reduce 118) ((T_UNSET) reduce 118) ((T_UNSET_CAST) reduce 118) ((T_USE) reduce 118) ((T_VARIABLE) reduce 118) ((T_WHILE) reduce 118) ((T_YIELD) reduce 118) ((T_YIELD_FROM) reduce 118) (("[") reduce 118) (("`") reduce 118) (("{") reduce 118) (("}") reduce 118) (("~") reduce 118)) 27607 (((",") reduce 137) ((";") reduce 137)) 27610 (((",") reduce 344) ((";") reduce 344)) 27617 ((("!") reduce 185) (("\"") reduce 185) (("$") reduce 185) (("(") reduce 185) (("+") reduce 185) (("-") reduce 185) ((";") reduce 185) (("@") reduce 185) ((T_ABSTRACT) reduce 185) ((T_ARRAY) reduce 185) ((T_ARRAY_CAST) reduce 185) ((T_ATTRIBUTE) reduce 185) ((T_BOOL_CAST) reduce 185) ((T_BREAK) reduce 185) ((T_CLASS) reduce 185) ((T_CLASS_C) reduce 185) ((T_CLONE) reduce 185) ((T_CONST) reduce 185) ((T_CONSTANT_ENCAPSED_STRING) reduce 185) ((T_CONTINUE) reduce 185) ((T_DEC) reduce 185) ((T_DECLARE) reduce 185) ((T_DIR) reduce 185) ((T_DNUMBER) reduce 185) ((T_DO) reduce 185) ((T_DOUBLE_CAST) reduce 185) ((T_ECHO) reduce 185) ((T_EMPTY) reduce 185) ((T_ENUM) reduce 185) ((T_EVAL) reduce 185) ((T_EXIT) reduce 185) ((T_FILE) reduce 185) ((T_FINAL) reduce 185) ((T_FN) reduce 185) ((T_FOR) reduce 185) ((T_FOREACH) reduce 185) ((T_FUNCTION) reduce 185) ((T_FUNC_C) reduce 185) ((T_GLOBAL) reduce 185) ((T_GOTO) reduce 185) ((T_HALT_COMPILER) reduce 185) ((T_IF) reduce 185) ((T_INC) reduce 185) ((T_INCLUDE) reduce 185) ((T_INCLUDE_ONCE) reduce 185) ((T_INLINE_HTML) reduce 185) ((T_INTERFACE) reduce 185) ((T_INT_CAST) reduce 185) ((T_ISSET) reduce 185) ((T_LINE) reduce 185) ((T_LIST) reduce 185) ((T_LNUMBER) reduce 185) ((T_MATCH) reduce 185) ((T_METHOD_C) reduce 185) ((T_NAMESPACE) reduce 185) ((T_NAME_FULLY_QUALIFIED) reduce 185) ((T_NAME_QUALIFIED) reduce 185) ((T_NAME_RELATIVE) reduce 185) ((T_NEW) reduce 185) ((T_NS_C) reduce 185) ((T_OBJECT_CAST) reduce 185) ((T_PRINT) reduce 185) ((T_REQUIRE) reduce 185) ((T_REQUIRE_ONCE) reduce 185) ((T_RETURN) reduce 185) ((T_START_HEREDOC) reduce 185) ((T_STATIC) reduce 185) ((T_STRING) reduce 185) ((T_STRING_CAST) reduce 185) ((T_SWITCH) reduce 185) ((T_THROW) reduce 185) ((T_TRAIT) reduce 185) ((T_TRAIT_C) reduce 185) ((T_TRY) reduce 185) ((T_UNSET) reduce 185) ((T_UNSET_CAST) reduce 185) ((T_USE) reduce 185) ((T_VARIABLE) reduce 185) ((T_WHILE) reduce 185) ((T_YIELD) reduce 185) ((T_YIELD_FROM) reduce 185) (("[") reduce 185) (("`") reduce 185) (("{") reduce 185) (("}") reduce 185) (("~") reduce 185)) 27619 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 27620 ((("!") reduce 111) (("\"") reduce 111) (("$") reduce 111) (($) reduce 111) (("(") reduce 111) (("+") reduce 111) (("-") reduce 111) ((";") reduce 111) (("@") reduce 111) ((T_ABSTRACT) reduce 111) ((T_ARRAY) reduce 111) ((T_ARRAY_CAST) reduce 111) ((T_ATTRIBUTE) reduce 111) ((T_BOOL_CAST) reduce 111) ((T_BREAK) reduce 111) ((T_CLASS) reduce 111) ((T_CLASS_C) reduce 111) ((T_CLONE) reduce 111) ((T_CONST) reduce 111) ((T_CONSTANT_ENCAPSED_STRING) reduce 111) ((T_CONTINUE) reduce 111) ((T_DEC) reduce 111) ((T_DECLARE) reduce 111) ((T_DIR) reduce 111) ((T_DNUMBER) reduce 111) ((T_DO) reduce 111) ((T_DOUBLE_CAST) reduce 111) ((T_ECHO) reduce 111) ((T_EMPTY) reduce 111) ((T_ENUM) reduce 111) ((T_EVAL) reduce 111) ((T_EXIT) reduce 111) ((T_FILE) reduce 111) ((T_FINAL) reduce 111) ((T_FN) reduce 111) ((T_FOR) reduce 111) ((T_FOREACH) reduce 111) ((T_FUNCTION) reduce 111) ((T_FUNC_C) reduce 111) ((T_GLOBAL) reduce 111) ((T_GOTO) reduce 111) ((T_HALT_COMPILER) reduce 111) ((T_IF) reduce 111) ((T_INC) reduce 111) ((T_INCLUDE) reduce 111) ((T_INCLUDE_ONCE) reduce 111) ((T_INLINE_HTML) reduce 111) ((T_INTERFACE) reduce 111) ((T_INT_CAST) reduce 111) ((T_ISSET) reduce 111) ((T_LINE) reduce 111) ((T_LIST) reduce 111) ((T_LNUMBER) reduce 111) ((T_MATCH) reduce 111) ((T_METHOD_C) reduce 111) ((T_NAMESPACE) reduce 111) ((T_NAME_FULLY_QUALIFIED) reduce 111) ((T_NAME_QUALIFIED) reduce 111) ((T_NAME_RELATIVE) reduce 111) ((T_NEW) reduce 111) ((T_NS_C) reduce 111) ((T_OBJECT_CAST) reduce 111) ((T_PRINT) reduce 111) ((T_REQUIRE) reduce 111) ((T_REQUIRE_ONCE) reduce 111) ((T_RETURN) reduce 111) ((T_START_HEREDOC) reduce 111) ((T_STATIC) reduce 111) ((T_STRING) reduce 111) ((T_STRING_CAST) reduce 111) ((T_SWITCH) reduce 111) ((T_THROW) reduce 111) ((T_TRAIT) reduce 111) ((T_TRAIT_C) reduce 111) ((T_TRY) reduce 111) ((T_UNSET) reduce 111) ((T_UNSET_CAST) reduce 111) ((T_USE) reduce 111) ((T_VARIABLE) reduce 111) ((T_WHILE) reduce 111) ((T_YIELD) reduce 111) ((T_YIELD_FROM) reduce 111) (("[") reduce 111) (("`") reduce 111) (("{") reduce 111) (("~") reduce 111)) 27623 ((("!") reduce 112) (("\"") reduce 112) (("$") reduce 112) (($) reduce 112) (("(") reduce 112) (("+") reduce 112) (("-") reduce 112) ((";") reduce 112) (("@") reduce 112) ((T_ABSTRACT) reduce 112) ((T_ARRAY) reduce 112) ((T_ARRAY_CAST) reduce 112) ((T_ATTRIBUTE) reduce 112) ((T_BOOL_CAST) reduce 112) ((T_BREAK) reduce 112) ((T_CLASS) reduce 112) ((T_CLASS_C) reduce 112) ((T_CLONE) reduce 112) ((T_CONST) reduce 112) ((T_CONSTANT_ENCAPSED_STRING) reduce 112) ((T_CONTINUE) reduce 112) ((T_DEC) reduce 112) ((T_DECLARE) reduce 112) ((T_DIR) reduce 112) ((T_DNUMBER) reduce 112) ((T_DO) reduce 112) ((T_DOUBLE_CAST) reduce 112) ((T_ECHO) reduce 112) ((T_EMPTY) reduce 112) ((T_ENUM) reduce 112) ((T_EVAL) reduce 112) ((T_EXIT) reduce 112) ((T_FILE) reduce 112) ((T_FINAL) reduce 112) ((T_FN) reduce 112) ((T_FOR) reduce 112) ((T_FOREACH) reduce 112) ((T_FUNCTION) reduce 112) ((T_FUNC_C) reduce 112) ((T_GLOBAL) reduce 112) ((T_GOTO) reduce 112) ((T_HALT_COMPILER) reduce 112) ((T_IF) reduce 112) ((T_INC) reduce 112) ((T_INCLUDE) reduce 112) ((T_INCLUDE_ONCE) reduce 112) ((T_INLINE_HTML) reduce 112) ((T_INTERFACE) reduce 112) ((T_INT_CAST) reduce 112) ((T_ISSET) reduce 112) ((T_LINE) reduce 112) ((T_LIST) reduce 112) ((T_LNUMBER) reduce 112) ((T_MATCH) reduce 112) ((T_METHOD_C) reduce 112) ((T_NAMESPACE) reduce 112) ((T_NAME_FULLY_QUALIFIED) reduce 112) ((T_NAME_QUALIFIED) reduce 112) ((T_NAME_RELATIVE) reduce 112) ((T_NEW) reduce 112) ((T_NS_C) reduce 112) ((T_OBJECT_CAST) reduce 112) ((T_PRINT) reduce 112) ((T_REQUIRE) reduce 112) ((T_REQUIRE_ONCE) reduce 112) ((T_RETURN) reduce 112) ((T_START_HEREDOC) reduce 112) ((T_STATIC) reduce 112) ((T_STRING) reduce 112) ((T_STRING_CAST) reduce 112) ((T_SWITCH) reduce 112) ((T_THROW) reduce 112) ((T_TRAIT) reduce 112) ((T_TRAIT_C) reduce 112) ((T_TRY) reduce 112) ((T_UNSET) reduce 112) ((T_UNSET_CAST) reduce 112) ((T_USE) reduce 112) ((T_VARIABLE) reduce 112) ((T_WHILE) reduce 112) ((T_YIELD) reduce 112) ((T_YIELD_FROM) reduce 112) (("[") reduce 112) (("`") reduce 112) (("{") reduce 112) (("~") reduce 112)) 27645 ((("!") reduce 191) (("\"") reduce 191) (("$") reduce 191) (($) reduce 191) (("(") reduce 191) (("+") reduce 191) (("-") reduce 191) ((";") reduce 191) (("@") reduce 191) ((T_ABSTRACT) reduce 191) ((T_ARRAY) reduce 191) ((T_ARRAY_CAST) reduce 191) ((T_ATTRIBUTE) reduce 191) ((T_BOOL_CAST) reduce 191) ((T_BREAK) reduce 191) ((T_CLASS) reduce 191) ((T_CLASS_C) reduce 191) ((T_CLONE) reduce 191) ((T_CONST) reduce 191) ((T_CONSTANT_ENCAPSED_STRING) reduce 191) ((T_CONTINUE) reduce 191) ((T_DEC) reduce 191) ((T_DECLARE) reduce 191) ((T_DIR) reduce 191) ((T_DNUMBER) reduce 191) ((T_DO) reduce 191) ((T_DOUBLE_CAST) reduce 191) ((T_ECHO) reduce 191) ((T_EMPTY) reduce 191) ((T_ENUM) reduce 191) ((T_EVAL) reduce 191) ((T_EXIT) reduce 191) ((T_FILE) reduce 191) ((T_FINAL) reduce 191) ((T_FN) reduce 191) ((T_FOR) reduce 191) ((T_FOREACH) reduce 191) ((T_FUNCTION) reduce 191) ((T_FUNC_C) reduce 191) ((T_GLOBAL) reduce 191) ((T_GOTO) reduce 191) ((T_HALT_COMPILER) reduce 191) ((T_IF) reduce 191) ((T_INC) reduce 191) ((T_INCLUDE) reduce 191) ((T_INCLUDE_ONCE) reduce 191) ((T_INLINE_HTML) reduce 191) ((T_INTERFACE) reduce 191) ((T_INT_CAST) reduce 191) ((T_ISSET) reduce 191) ((T_LINE) reduce 191) ((T_LIST) reduce 191) ((T_LNUMBER) reduce 191) ((T_MATCH) reduce 191) ((T_METHOD_C) reduce 191) ((T_NAMESPACE) reduce 191) ((T_NAME_FULLY_QUALIFIED) reduce 191) ((T_NAME_QUALIFIED) reduce 191) ((T_NAME_RELATIVE) reduce 191) ((T_NEW) reduce 191) ((T_NS_C) reduce 191) ((T_OBJECT_CAST) reduce 191) ((T_PRINT) reduce 191) ((T_REQUIRE) reduce 191) ((T_REQUIRE_ONCE) reduce 191) ((T_RETURN) reduce 191) ((T_START_HEREDOC) reduce 191) ((T_STATIC) reduce 191) ((T_STRING) reduce 191) ((T_STRING_CAST) reduce 191) ((T_SWITCH) reduce 191) ((T_THROW) reduce 191) ((T_TRAIT) reduce 191) ((T_TRAIT_C) reduce 191) ((T_TRY) reduce 191) ((T_UNSET) reduce 191) ((T_UNSET_CAST) reduce 191) ((T_USE) reduce 191) ((T_VARIABLE) reduce 191) ((T_WHILE) reduce 191) ((T_YIELD) reduce 191) ((T_YIELD_FROM) reduce 191) (("[") reduce 191) (("`") reduce 191) (("{") reduce 191) (("~") reduce 191)) 27652 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (($) reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 27663 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (($) reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 27671 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (($) reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 27674 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (($) reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONST) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAMESPACE) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_USE) reduce 146) ((T_VARIABLE) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 27675 ((("!") reduce 236) (("\"") reduce 236) (($) reduce 236) (("$") reduce 236) (("(") reduce 236) (("+") reduce 236) (("-") reduce 236) ((";") reduce 236) (("@") reduce 236) ((T_ABSTRACT) reduce 236) ((T_ARRAY) reduce 236) ((T_ARRAY_CAST) reduce 236) ((T_ATTRIBUTE) reduce 236) ((T_BOOL_CAST) reduce 236) ((T_BREAK) reduce 236) ((T_CLASS) reduce 236) ((T_CLASS_C) reduce 236) ((T_CLONE) reduce 236) ((T_CONST) reduce 236) ((T_CONSTANT_ENCAPSED_STRING) reduce 236) ((T_CONTINUE) reduce 236) ((T_DEC) reduce 236) ((T_DECLARE) reduce 236) ((T_DIR) reduce 236) ((T_DNUMBER) reduce 236) ((T_DO) reduce 236) ((T_DOUBLE_CAST) reduce 236) ((T_ECHO) reduce 236) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 236) ((T_ENUM) reduce 236) ((T_EVAL) reduce 236) ((T_EXIT) reduce 236) ((T_FILE) reduce 236) ((T_FINAL) reduce 236) ((T_FN) reduce 236) ((T_FOR) reduce 236) ((T_FOREACH) reduce 236) ((T_FUNCTION) reduce 236) ((T_FUNC_C) reduce 236) ((T_GLOBAL) reduce 236) ((T_GOTO) reduce 236) ((T_HALT_COMPILER) reduce 236) ((T_IF) reduce 236) ((T_INC) reduce 236) ((T_INCLUDE) reduce 236) ((T_INCLUDE_ONCE) reduce 236) ((T_INLINE_HTML) reduce 236) ((T_INTERFACE) reduce 236) ((T_INT_CAST) reduce 236) ((T_ISSET) reduce 236) ((T_LINE) reduce 236) ((T_LIST) reduce 236) ((T_LNUMBER) reduce 236) ((T_MATCH) reduce 236) ((T_METHOD_C) reduce 236) ((T_NAMESPACE) reduce 236) ((T_NAME_FULLY_QUALIFIED) reduce 236) ((T_NAME_QUALIFIED) reduce 236) ((T_NAME_RELATIVE) reduce 236) ((T_NEW) reduce 236) ((T_NS_C) reduce 236) ((T_OBJECT_CAST) reduce 236) ((T_PRINT) reduce 236) ((T_REQUIRE) reduce 236) ((T_REQUIRE_ONCE) reduce 236) ((T_RETURN) reduce 236) ((T_START_HEREDOC) reduce 236) ((T_STATIC) reduce 236) ((T_STRING) reduce 236) ((T_STRING_CAST) reduce 236) ((T_SWITCH) reduce 236) ((T_THROW) reduce 236) ((T_TRAIT) reduce 236) ((T_TRAIT_C) reduce 236) ((T_TRY) reduce 236) ((T_UNSET) reduce 236) ((T_UNSET_CAST) reduce 236) ((T_USE) reduce 236) ((T_VARIABLE) reduce 236) ((T_WHILE) reduce 236) ((T_YIELD) reduce 236) ((T_YIELD_FROM) reduce 236) (("[") reduce 236) (("`") reduce 236) (("{") reduce 236) (("~") reduce 236)) 27676 ((("!") reduce 234) (("\"") reduce 234) (($) reduce 234) (("$") reduce 234) (("(") reduce 234) (("+") reduce 234) (("-") reduce 234) ((";") reduce 234) (("@") reduce 234) ((T_ABSTRACT) reduce 234) ((T_ARRAY) reduce 234) ((T_ARRAY_CAST) reduce 234) ((T_ATTRIBUTE) reduce 234) ((T_BOOL_CAST) reduce 234) ((T_BREAK) reduce 234) ((T_CLASS) reduce 234) ((T_CLASS_C) reduce 234) ((T_CLONE) reduce 234) ((T_CONST) reduce 234) ((T_CONSTANT_ENCAPSED_STRING) reduce 234) ((T_CONTINUE) reduce 234) ((T_DEC) reduce 234) ((T_DECLARE) reduce 234) ((T_DIR) reduce 234) ((T_DNUMBER) reduce 234) ((T_DO) reduce 234) ((T_DOUBLE_CAST) reduce 234) ((T_ECHO) reduce 234) ((T_ELSE) reduce 234) ((T_ELSEIF) reduce 234) ((T_EMPTY) reduce 234) ((T_ENUM) reduce 234) ((T_EVAL) reduce 234) ((T_EXIT) reduce 234) ((T_FILE) reduce 234) ((T_FINAL) reduce 234) ((T_FN) reduce 234) ((T_FOR) reduce 234) ((T_FOREACH) reduce 234) ((T_FUNCTION) reduce 234) ((T_FUNC_C) reduce 234) ((T_GLOBAL) reduce 234) ((T_GOTO) reduce 234) ((T_HALT_COMPILER) reduce 234) ((T_IF) reduce 234) ((T_INC) reduce 234) ((T_INCLUDE) reduce 234) ((T_INCLUDE_ONCE) reduce 234) ((T_INLINE_HTML) reduce 234) ((T_INTERFACE) reduce 234) ((T_INT_CAST) reduce 234) ((T_ISSET) reduce 234) ((T_LINE) reduce 234) ((T_LIST) reduce 234) ((T_LNUMBER) reduce 234) ((T_MATCH) reduce 234) ((T_METHOD_C) reduce 234) ((T_NAMESPACE) reduce 234) ((T_NAME_FULLY_QUALIFIED) reduce 234) ((T_NAME_QUALIFIED) reduce 234) ((T_NAME_RELATIVE) reduce 234) ((T_NEW) reduce 234) ((T_NS_C) reduce 234) ((T_OBJECT_CAST) reduce 234) ((T_PRINT) reduce 234) ((T_REQUIRE) reduce 234) ((T_REQUIRE_ONCE) reduce 234) ((T_RETURN) reduce 234) ((T_START_HEREDOC) reduce 234) ((T_STATIC) reduce 234) ((T_STRING) reduce 234) ((T_STRING_CAST) reduce 234) ((T_SWITCH) reduce 234) ((T_THROW) reduce 234) ((T_TRAIT) reduce 234) ((T_TRAIT_C) reduce 234) ((T_TRY) reduce 234) ((T_UNSET) reduce 234) ((T_UNSET_CAST) reduce 234) ((T_USE) reduce 234) ((T_VARIABLE) reduce 234) ((T_WHILE) reduce 234) ((T_YIELD) reduce 234) ((T_YIELD_FROM) reduce 234) (("[") reduce 234) (("`") reduce 234) (("{") reduce 234) (("~") reduce 234)) 27679 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (($) reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONST) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAMESPACE) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_USE) reduce 145) ((T_VARIABLE) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 27685 ((("!") reduce 235) (("\"") reduce 235) (($) reduce 235) (("$") reduce 235) (("(") reduce 235) (("+") reduce 235) (("-") reduce 235) ((";") reduce 235) (("@") reduce 235) ((T_ABSTRACT) reduce 235) ((T_ARRAY) reduce 235) ((T_ARRAY_CAST) reduce 235) ((T_ATTRIBUTE) reduce 235) ((T_BOOL_CAST) reduce 235) ((T_BREAK) reduce 235) ((T_CLASS) reduce 235) ((T_CLASS_C) reduce 235) ((T_CLONE) reduce 235) ((T_CONST) reduce 235) ((T_CONSTANT_ENCAPSED_STRING) reduce 235) ((T_CONTINUE) reduce 235) ((T_DEC) reduce 235) ((T_DECLARE) reduce 235) ((T_DIR) reduce 235) ((T_DNUMBER) reduce 235) ((T_DO) reduce 235) ((T_DOUBLE_CAST) reduce 235) ((T_ECHO) reduce 235) ((T_ELSE) reduce 235) ((T_ELSEIF) reduce 235) ((T_EMPTY) reduce 235) ((T_ENUM) reduce 235) ((T_EVAL) reduce 235) ((T_EXIT) reduce 235) ((T_FILE) reduce 235) ((T_FINAL) reduce 235) ((T_FN) reduce 235) ((T_FOR) reduce 235) ((T_FOREACH) reduce 235) ((T_FUNCTION) reduce 235) ((T_FUNC_C) reduce 235) ((T_GLOBAL) reduce 235) ((T_GOTO) reduce 235) ((T_HALT_COMPILER) reduce 235) ((T_IF) reduce 235) ((T_INC) reduce 235) ((T_INCLUDE) reduce 235) ((T_INCLUDE_ONCE) reduce 235) ((T_INLINE_HTML) reduce 235) ((T_INTERFACE) reduce 235) ((T_INT_CAST) reduce 235) ((T_ISSET) reduce 235) ((T_LINE) reduce 235) ((T_LIST) reduce 235) ((T_LNUMBER) reduce 235) ((T_MATCH) reduce 235) ((T_METHOD_C) reduce 235) ((T_NAMESPACE) reduce 235) ((T_NAME_FULLY_QUALIFIED) reduce 235) ((T_NAME_QUALIFIED) reduce 235) ((T_NAME_RELATIVE) reduce 235) ((T_NEW) reduce 235) ((T_NS_C) reduce 235) ((T_OBJECT_CAST) reduce 235) ((T_PRINT) reduce 235) ((T_REQUIRE) reduce 235) ((T_REQUIRE_ONCE) reduce 235) ((T_RETURN) reduce 235) ((T_START_HEREDOC) reduce 235) ((T_STATIC) reduce 235) ((T_STRING) reduce 235) ((T_STRING_CAST) reduce 235) ((T_SWITCH) reduce 235) ((T_THROW) reduce 235) ((T_TRAIT) reduce 235) ((T_TRAIT_C) reduce 235) ((T_TRY) reduce 235) ((T_UNSET) reduce 235) ((T_UNSET_CAST) reduce 235) ((T_USE) reduce 235) ((T_VARIABLE) reduce 235) ((T_WHILE) reduce 235) ((T_YIELD) reduce 235) ((T_YIELD_FROM) reduce 235) (("[") reduce 235) (("`") reduce 235) (("{") reduce 235) (("~") reduce 235)) 27686 ((("!") reduce 237) (("\"") reduce 237) (($) reduce 237) (("$") reduce 237) (("(") reduce 237) (("+") reduce 237) (("-") reduce 237) ((";") reduce 237) (("@") reduce 237) ((T_ABSTRACT) reduce 237) ((T_ARRAY) reduce 237) ((T_ARRAY_CAST) reduce 237) ((T_ATTRIBUTE) reduce 237) ((T_BOOL_CAST) reduce 237) ((T_BREAK) reduce 237) ((T_CLASS) reduce 237) ((T_CLASS_C) reduce 237) ((T_CLONE) reduce 237) ((T_CONST) reduce 237) ((T_CONSTANT_ENCAPSED_STRING) reduce 237) ((T_CONTINUE) reduce 237) ((T_DEC) reduce 237) ((T_DECLARE) reduce 237) ((T_DIR) reduce 237) ((T_DNUMBER) reduce 237) ((T_DO) reduce 237) ((T_DOUBLE_CAST) reduce 237) ((T_ECHO) reduce 237) ((T_ELSE) reduce 237) ((T_ELSEIF) reduce 237) ((T_EMPTY) reduce 237) ((T_ENUM) reduce 237) ((T_EVAL) reduce 237) ((T_EXIT) reduce 237) ((T_FILE) reduce 237) ((T_FINAL) reduce 237) ((T_FN) reduce 237) ((T_FOR) reduce 237) ((T_FOREACH) reduce 237) ((T_FUNCTION) reduce 237) ((T_FUNC_C) reduce 237) ((T_GLOBAL) reduce 237) ((T_GOTO) reduce 237) ((T_HALT_COMPILER) reduce 237) ((T_IF) reduce 237) ((T_INC) reduce 237) ((T_INCLUDE) reduce 237) ((T_INCLUDE_ONCE) reduce 237) ((T_INLINE_HTML) reduce 237) ((T_INTERFACE) reduce 237) ((T_INT_CAST) reduce 237) ((T_ISSET) reduce 237) ((T_LINE) reduce 237) ((T_LIST) reduce 237) ((T_LNUMBER) reduce 237) ((T_MATCH) reduce 237) ((T_METHOD_C) reduce 237) ((T_NAMESPACE) reduce 237) ((T_NAME_FULLY_QUALIFIED) reduce 237) ((T_NAME_QUALIFIED) reduce 237) ((T_NAME_RELATIVE) reduce 237) ((T_NEW) reduce 237) ((T_NS_C) reduce 237) ((T_OBJECT_CAST) reduce 237) ((T_PRINT) reduce 237) ((T_REQUIRE) reduce 237) ((T_REQUIRE_ONCE) reduce 237) ((T_RETURN) reduce 237) ((T_START_HEREDOC) reduce 237) ((T_STATIC) reduce 237) ((T_STRING) reduce 237) ((T_STRING_CAST) reduce 237) ((T_SWITCH) reduce 237) ((T_THROW) reduce 237) ((T_TRAIT) reduce 237) ((T_TRAIT_C) reduce 237) ((T_TRY) reduce 237) ((T_UNSET) reduce 237) ((T_UNSET_CAST) reduce 237) ((T_USE) reduce 237) ((T_VARIABLE) reduce 237) ((T_WHILE) reduce 237) ((T_YIELD) reduce 237) ((T_YIELD_FROM) reduce 237) (("[") reduce 237) (("`") reduce 237) (("{") reduce 237) (("~") reduce 237)) 27687 ((("!") reduce 159) (("\"") reduce 159) (($) reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 27690 ((("!") reduce 240) (("\"") reduce 240) (($) reduce 240) (("$") reduce 240) (("(") reduce 240) (("+") reduce 240) (("-") reduce 240) ((";") reduce 240) (("@") reduce 240) ((T_ABSTRACT) reduce 240) ((T_ARRAY) reduce 240) ((T_ARRAY_CAST) reduce 240) ((T_ATTRIBUTE) reduce 240) ((T_BOOL_CAST) reduce 240) ((T_BREAK) reduce 240) ((T_CLASS) reduce 240) ((T_CLASS_C) reduce 240) ((T_CLONE) reduce 240) ((T_CONST) reduce 240) ((T_CONSTANT_ENCAPSED_STRING) reduce 240) ((T_CONTINUE) reduce 240) ((T_DEC) reduce 240) ((T_DECLARE) reduce 240) ((T_DIR) reduce 240) ((T_DNUMBER) reduce 240) ((T_DO) reduce 240) ((T_DOUBLE_CAST) reduce 240) ((T_ECHO) reduce 240) ((T_ELSE) reduce 240) ((T_ELSEIF) reduce 240) ((T_EMPTY) reduce 240) ((T_ENUM) reduce 240) ((T_EVAL) reduce 240) ((T_EXIT) reduce 240) ((T_FILE) reduce 240) ((T_FINAL) reduce 240) ((T_FN) reduce 240) ((T_FOR) reduce 240) ((T_FOREACH) reduce 240) ((T_FUNCTION) reduce 240) ((T_FUNC_C) reduce 240) ((T_GLOBAL) reduce 240) ((T_GOTO) reduce 240) ((T_HALT_COMPILER) reduce 240) ((T_IF) reduce 240) ((T_INC) reduce 240) ((T_INCLUDE) reduce 240) ((T_INCLUDE_ONCE) reduce 240) ((T_INLINE_HTML) reduce 240) ((T_INTERFACE) reduce 240) ((T_INT_CAST) reduce 240) ((T_ISSET) reduce 240) ((T_LINE) reduce 240) ((T_LIST) reduce 240) ((T_LNUMBER) reduce 240) ((T_MATCH) reduce 240) ((T_METHOD_C) reduce 240) ((T_NAMESPACE) reduce 240) ((T_NAME_FULLY_QUALIFIED) reduce 240) ((T_NAME_QUALIFIED) reduce 240) ((T_NAME_RELATIVE) reduce 240) ((T_NEW) reduce 240) ((T_NS_C) reduce 240) ((T_OBJECT_CAST) reduce 240) ((T_PRINT) reduce 240) ((T_REQUIRE) reduce 240) ((T_REQUIRE_ONCE) reduce 240) ((T_RETURN) reduce 240) ((T_START_HEREDOC) reduce 240) ((T_STATIC) reduce 240) ((T_STRING) reduce 240) ((T_STRING_CAST) reduce 240) ((T_SWITCH) reduce 240) ((T_THROW) reduce 240) ((T_TRAIT) reduce 240) ((T_TRAIT_C) reduce 240) ((T_TRY) reduce 240) ((T_UNSET) reduce 240) ((T_UNSET_CAST) reduce 240) ((T_USE) reduce 240) ((T_VARIABLE) reduce 240) ((T_WHILE) reduce 240) ((T_YIELD) reduce 240) ((T_YIELD_FROM) reduce 240) (("[") reduce 240) (("`") reduce 240) (("{") reduce 240) (("~") reduce 240)) 27694 ((("!") reduce 241) (("\"") reduce 241) (("$") reduce 241) (($) reduce 241) (("(") reduce 241) (("+") reduce 241) (("-") reduce 241) ((";") reduce 241) (("@") reduce 241) ((T_ABSTRACT) reduce 241) ((T_ARRAY) reduce 241) ((T_ARRAY_CAST) reduce 241) ((T_ATTRIBUTE) reduce 241) ((T_BOOL_CAST) reduce 241) ((T_BREAK) reduce 241) ((T_CLASS) reduce 241) ((T_CLASS_C) reduce 241) ((T_CLONE) reduce 241) ((T_CONST) reduce 241) ((T_CONSTANT_ENCAPSED_STRING) reduce 241) ((T_CONTINUE) reduce 241) ((T_DEC) reduce 241) ((T_DECLARE) reduce 241) ((T_DIR) reduce 241) ((T_DNUMBER) reduce 241) ((T_DO) reduce 241) ((T_DOUBLE_CAST) reduce 241) ((T_ECHO) reduce 241) ((T_ELSE) reduce 241) ((T_ELSEIF) reduce 241) ((T_EMPTY) reduce 241) ((T_ENUM) reduce 241) ((T_EVAL) reduce 241) ((T_EXIT) reduce 241) ((T_FILE) reduce 241) ((T_FINAL) reduce 241) ((T_FN) reduce 241) ((T_FOR) reduce 241) ((T_FOREACH) reduce 241) ((T_FUNCTION) reduce 241) ((T_FUNC_C) reduce 241) ((T_GLOBAL) reduce 241) ((T_GOTO) reduce 241) ((T_HALT_COMPILER) reduce 241) ((T_IF) reduce 241) ((T_INC) reduce 241) ((T_INCLUDE) reduce 241) ((T_INCLUDE_ONCE) reduce 241) ((T_INLINE_HTML) reduce 241) ((T_INTERFACE) reduce 241) ((T_INT_CAST) reduce 241) ((T_ISSET) reduce 241) ((T_LINE) reduce 241) ((T_LIST) reduce 241) ((T_LNUMBER) reduce 241) ((T_MATCH) reduce 241) ((T_METHOD_C) reduce 241) ((T_NAMESPACE) reduce 241) ((T_NAME_FULLY_QUALIFIED) reduce 241) ((T_NAME_QUALIFIED) reduce 241) ((T_NAME_RELATIVE) reduce 241) ((T_NEW) reduce 241) ((T_NS_C) reduce 241) ((T_OBJECT_CAST) reduce 241) ((T_PRINT) reduce 241) ((T_REQUIRE) reduce 241) ((T_REQUIRE_ONCE) reduce 241) ((T_RETURN) reduce 241) ((T_START_HEREDOC) reduce 241) ((T_STATIC) reduce 241) ((T_STRING) reduce 241) ((T_STRING_CAST) reduce 241) ((T_SWITCH) reduce 241) ((T_THROW) reduce 241) ((T_TRAIT) reduce 241) ((T_TRAIT_C) reduce 241) ((T_TRY) reduce 241) ((T_UNSET) reduce 241) ((T_UNSET_CAST) reduce 241) ((T_USE) reduce 241) ((T_VARIABLE) reduce 241) ((T_WHILE) reduce 241) ((T_YIELD) reduce 241) ((T_YIELD_FROM) reduce 241) (("[") reduce 241) (("`") reduce 241) (("{") reduce 241) (("~") reduce 241)) 27699 ((("!") reduce 232) (("\"") reduce 232) (($) reduce 232) (("$") reduce 232) (("(") reduce 232) (("+") reduce 232) (("-") reduce 232) ((";") reduce 232) (("@") reduce 232) ((T_ABSTRACT) reduce 232) ((T_ARRAY) reduce 232) ((T_ARRAY_CAST) reduce 232) ((T_ATTRIBUTE) reduce 232) ((T_BOOL_CAST) reduce 232) ((T_BREAK) reduce 232) ((T_CLASS) reduce 232) ((T_CLASS_C) reduce 232) ((T_CLONE) reduce 232) ((T_CONST) reduce 232) ((T_CONSTANT_ENCAPSED_STRING) reduce 232) ((T_CONTINUE) reduce 232) ((T_DEC) reduce 232) ((T_DECLARE) reduce 232) ((T_DIR) reduce 232) ((T_DNUMBER) reduce 232) ((T_DO) reduce 232) ((T_DOUBLE_CAST) reduce 232) ((T_ECHO) reduce 232) ((T_ELSE) reduce 232) ((T_ELSEIF) reduce 232) ((T_EMPTY) reduce 232) ((T_ENUM) reduce 232) ((T_EVAL) reduce 232) ((T_EXIT) reduce 232) ((T_FILE) reduce 232) ((T_FINAL) reduce 232) ((T_FN) reduce 232) ((T_FOR) reduce 232) ((T_FOREACH) reduce 232) ((T_FUNCTION) reduce 232) ((T_FUNC_C) reduce 232) ((T_GLOBAL) reduce 232) ((T_GOTO) reduce 232) ((T_HALT_COMPILER) reduce 232) ((T_IF) reduce 232) ((T_INC) reduce 232) ((T_INCLUDE) reduce 232) ((T_INCLUDE_ONCE) reduce 232) ((T_INLINE_HTML) reduce 232) ((T_INTERFACE) reduce 232) ((T_INT_CAST) reduce 232) ((T_ISSET) reduce 232) ((T_LINE) reduce 232) ((T_LIST) reduce 232) ((T_LNUMBER) reduce 232) ((T_MATCH) reduce 232) ((T_METHOD_C) reduce 232) ((T_NAMESPACE) reduce 232) ((T_NAME_FULLY_QUALIFIED) reduce 232) ((T_NAME_QUALIFIED) reduce 232) ((T_NAME_RELATIVE) reduce 232) ((T_NEW) reduce 232) ((T_NS_C) reduce 232) ((T_OBJECT_CAST) reduce 232) ((T_PRINT) reduce 232) ((T_REQUIRE) reduce 232) ((T_REQUIRE_ONCE) reduce 232) ((T_RETURN) reduce 232) ((T_START_HEREDOC) reduce 232) ((T_STATIC) reduce 232) ((T_STRING) reduce 232) ((T_STRING_CAST) reduce 232) ((T_SWITCH) reduce 232) ((T_THROW) reduce 232) ((T_TRAIT) reduce 232) ((T_TRAIT_C) reduce 232) ((T_TRY) reduce 232) ((T_UNSET) reduce 232) ((T_UNSET_CAST) reduce 232) ((T_USE) reduce 232) ((T_VARIABLE) reduce 232) ((T_WHILE) reduce 232) ((T_YIELD) reduce 232) ((T_YIELD_FROM) reduce 232) (("[") reduce 232) (("`") reduce 232) (("{") reduce 232) (("~") reduce 232)) 27700 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (($) reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 27703 ((("!") reduce 233) (("\"") reduce 233) (("$") reduce 233) (($) reduce 233) (("(") reduce 233) (("+") reduce 233) (("-") reduce 233) ((";") reduce 233) (("@") reduce 233) ((T_ABSTRACT) reduce 233) ((T_ARRAY) reduce 233) ((T_ARRAY_CAST) reduce 233) ((T_ATTRIBUTE) reduce 233) ((T_BOOL_CAST) reduce 233) ((T_BREAK) reduce 233) ((T_CLASS) reduce 233) ((T_CLASS_C) reduce 233) ((T_CLONE) reduce 233) ((T_CONST) reduce 233) ((T_CONSTANT_ENCAPSED_STRING) reduce 233) ((T_CONTINUE) reduce 233) ((T_DEC) reduce 233) ((T_DECLARE) reduce 233) ((T_DIR) reduce 233) ((T_DNUMBER) reduce 233) ((T_DO) reduce 233) ((T_DOUBLE_CAST) reduce 233) ((T_ECHO) reduce 233) ((T_ELSE) reduce 233) ((T_ELSEIF) reduce 233) ((T_EMPTY) reduce 233) ((T_ENUM) reduce 233) ((T_EVAL) reduce 233) ((T_EXIT) reduce 233) ((T_FILE) reduce 233) ((T_FINAL) reduce 233) ((T_FN) reduce 233) ((T_FOR) reduce 233) ((T_FOREACH) reduce 233) ((T_FUNCTION) reduce 233) ((T_FUNC_C) reduce 233) ((T_GLOBAL) reduce 233) ((T_GOTO) reduce 233) ((T_HALT_COMPILER) reduce 233) ((T_IF) reduce 233) ((T_INC) reduce 233) ((T_INCLUDE) reduce 233) ((T_INCLUDE_ONCE) reduce 233) ((T_INLINE_HTML) reduce 233) ((T_INTERFACE) reduce 233) ((T_INT_CAST) reduce 233) ((T_ISSET) reduce 233) ((T_LINE) reduce 233) ((T_LIST) reduce 233) ((T_LNUMBER) reduce 233) ((T_MATCH) reduce 233) ((T_METHOD_C) reduce 233) ((T_NAMESPACE) reduce 233) ((T_NAME_FULLY_QUALIFIED) reduce 233) ((T_NAME_QUALIFIED) reduce 233) ((T_NAME_RELATIVE) reduce 233) ((T_NEW) reduce 233) ((T_NS_C) reduce 233) ((T_OBJECT_CAST) reduce 233) ((T_PRINT) reduce 233) ((T_REQUIRE) reduce 233) ((T_REQUIRE_ONCE) reduce 233) ((T_RETURN) reduce 233) ((T_START_HEREDOC) reduce 233) ((T_STATIC) reduce 233) ((T_STRING) reduce 233) ((T_STRING_CAST) reduce 233) ((T_SWITCH) reduce 233) ((T_THROW) reduce 233) ((T_TRAIT) reduce 233) ((T_TRAIT_C) reduce 233) ((T_TRY) reduce 233) ((T_UNSET) reduce 233) ((T_UNSET_CAST) reduce 233) ((T_USE) reduce 233) ((T_VARIABLE) reduce 233) ((T_WHILE) reduce 233) ((T_YIELD) reduce 233) ((T_YIELD_FROM) reduce 233) (("[") reduce 233) (("`") reduce 233) (("{") reduce 233) (("~") reduce 233)) 27708 ((("!") reduce 160) (("\"") reduce 160) (($) reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 27711 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (($) reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CATCH) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FINALLY) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 27712 ((("!") reduce 174) (("\"") reduce 174) (($) reduce 174) (("$") reduce 174) (("(") reduce 174) (("+") reduce 174) (("-") reduce 174) ((";") reduce 174) (("@") reduce 174) ((T_ABSTRACT) reduce 174) ((T_ARRAY) reduce 174) ((T_ARRAY_CAST) reduce 174) ((T_ATTRIBUTE) reduce 174) ((T_BOOL_CAST) reduce 174) ((T_BREAK) reduce 174) ((T_CATCH) shift) ((T_CLASS) reduce 174) ((T_CLASS_C) reduce 174) ((T_CLONE) reduce 174) ((T_CONST) reduce 174) ((T_CONSTANT_ENCAPSED_STRING) reduce 174) ((T_CONTINUE) reduce 174) ((T_DEC) reduce 174) ((T_DECLARE) reduce 174) ((T_DIR) reduce 174) ((T_DNUMBER) reduce 174) ((T_DO) reduce 174) ((T_DOUBLE_CAST) reduce 174) ((T_ECHO) reduce 174) ((T_ELSE) reduce 174) ((T_ELSEIF) reduce 174) ((T_EMPTY) reduce 174) ((T_ENUM) reduce 174) ((T_EVAL) reduce 174) ((T_EXIT) reduce 174) ((T_FILE) reduce 174) ((T_FINAL) reduce 174) ((T_FINALLY) shift) ((T_FN) reduce 174) ((T_FOR) reduce 174) ((T_FOREACH) reduce 174) ((T_FUNCTION) reduce 174) ((T_FUNC_C) reduce 174) ((T_GLOBAL) reduce 174) ((T_GOTO) reduce 174) ((T_HALT_COMPILER) reduce 174) ((T_IF) reduce 174) ((T_INC) reduce 174) ((T_INCLUDE) reduce 174) ((T_INCLUDE_ONCE) reduce 174) ((T_INLINE_HTML) reduce 174) ((T_INTERFACE) reduce 174) ((T_INT_CAST) reduce 174) ((T_ISSET) reduce 174) ((T_LINE) reduce 174) ((T_LIST) reduce 174) ((T_LNUMBER) reduce 174) ((T_MATCH) reduce 174) ((T_METHOD_C) reduce 174) ((T_NAMESPACE) reduce 174) ((T_NAME_FULLY_QUALIFIED) reduce 174) ((T_NAME_QUALIFIED) reduce 174) ((T_NAME_RELATIVE) reduce 174) ((T_NEW) reduce 174) ((T_NS_C) reduce 174) ((T_OBJECT_CAST) reduce 174) ((T_PRINT) reduce 174) ((T_REQUIRE) reduce 174) ((T_REQUIRE_ONCE) reduce 174) ((T_RETURN) reduce 174) ((T_START_HEREDOC) reduce 174) ((T_STATIC) reduce 174) ((T_STRING) reduce 174) ((T_STRING_CAST) reduce 174) ((T_SWITCH) reduce 174) ((T_THROW) reduce 174) ((T_TRAIT) reduce 174) ((T_TRAIT_C) reduce 174) ((T_TRY) reduce 174) ((T_UNSET) reduce 174) ((T_UNSET_CAST) reduce 174) ((T_USE) reduce 174) ((T_VARIABLE) reduce 174) ((T_WHILE) reduce 174) ((T_YIELD) reduce 174) ((T_YIELD_FROM) reduce 174) (("[") reduce 174) (("`") reduce 174) (("{") reduce 174) (("~") reduce 174)) 27715 ((("!") reduce 165) (("\"") reduce 165) (($) reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 27718 ((("!") reduce 175) (("\"") reduce 175) (($) reduce 175) (("$") reduce 175) (("(") reduce 175) (("+") reduce 175) (("-") reduce 175) ((";") reduce 175) (("@") reduce 175) ((T_ABSTRACT) reduce 175) ((T_ARRAY) reduce 175) ((T_ARRAY_CAST) reduce 175) ((T_ATTRIBUTE) reduce 175) ((T_BOOL_CAST) reduce 175) ((T_BREAK) reduce 175) ((T_CLASS) reduce 175) ((T_CLASS_C) reduce 175) ((T_CLONE) reduce 175) ((T_CONST) reduce 175) ((T_CONSTANT_ENCAPSED_STRING) reduce 175) ((T_CONTINUE) reduce 175) ((T_DEC) reduce 175) ((T_DECLARE) reduce 175) ((T_DIR) reduce 175) ((T_DNUMBER) reduce 175) ((T_DO) reduce 175) ((T_DOUBLE_CAST) reduce 175) ((T_ECHO) reduce 175) ((T_ELSE) reduce 175) ((T_ELSEIF) reduce 175) ((T_EMPTY) reduce 175) ((T_ENUM) reduce 175) ((T_EVAL) reduce 175) ((T_EXIT) reduce 175) ((T_FILE) reduce 175) ((T_FINAL) reduce 175) ((T_FN) reduce 175) ((T_FOR) reduce 175) ((T_FOREACH) reduce 175) ((T_FUNCTION) reduce 175) ((T_FUNC_C) reduce 175) ((T_GLOBAL) reduce 175) ((T_GOTO) reduce 175) ((T_HALT_COMPILER) reduce 175) ((T_IF) reduce 175) ((T_INC) reduce 175) ((T_INCLUDE) reduce 175) ((T_INCLUDE_ONCE) reduce 175) ((T_INLINE_HTML) reduce 175) ((T_INTERFACE) reduce 175) ((T_INT_CAST) reduce 175) ((T_ISSET) reduce 175) ((T_LINE) reduce 175) ((T_LIST) reduce 175) ((T_LNUMBER) reduce 175) ((T_MATCH) reduce 175) ((T_METHOD_C) reduce 175) ((T_NAMESPACE) reduce 175) ((T_NAME_FULLY_QUALIFIED) reduce 175) ((T_NAME_QUALIFIED) reduce 175) ((T_NAME_RELATIVE) reduce 175) ((T_NEW) reduce 175) ((T_NS_C) reduce 175) ((T_OBJECT_CAST) reduce 175) ((T_PRINT) reduce 175) ((T_REQUIRE) reduce 175) ((T_REQUIRE_ONCE) reduce 175) ((T_RETURN) reduce 175) ((T_START_HEREDOC) reduce 175) ((T_STATIC) reduce 175) ((T_STRING) reduce 175) ((T_STRING_CAST) reduce 175) ((T_SWITCH) reduce 175) ((T_THROW) reduce 175) ((T_TRAIT) reduce 175) ((T_TRAIT_C) reduce 175) ((T_TRY) reduce 175) ((T_UNSET) reduce 175) ((T_UNSET_CAST) reduce 175) ((T_USE) reduce 175) ((T_VARIABLE) reduce 175) ((T_WHILE) reduce 175) ((T_YIELD) reduce 175) ((T_YIELD_FROM) reduce 175) (("[") reduce 175) (("`") reduce 175) (("{") reduce 175) (("~") reduce 175)) 27725 ((("!") reduce 169) (("\"") reduce 169) (($) reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CATCH) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FINALLY) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 27730 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (($) reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 27734 ((("!") reduce 214) (("\"") reduce 214) (($) reduce 214) (("$") reduce 214) (("(") reduce 214) (("+") reduce 214) (("-") reduce 214) ((";") reduce 214) (("@") reduce 214) ((T_ABSTRACT) reduce 214) ((T_ARRAY) reduce 214) ((T_ARRAY_CAST) reduce 214) ((T_ATTRIBUTE) reduce 214) ((T_BOOL_CAST) reduce 214) ((T_BREAK) reduce 214) ((T_CLASS) reduce 214) ((T_CLASS_C) reduce 214) ((T_CLONE) reduce 214) ((T_CONST) reduce 214) ((T_CONSTANT_ENCAPSED_STRING) reduce 214) ((T_CONTINUE) reduce 214) ((T_DEC) reduce 214) ((T_DECLARE) reduce 214) ((T_DIR) reduce 214) ((T_DNUMBER) reduce 214) ((T_DO) reduce 214) ((T_DOUBLE_CAST) reduce 214) ((T_ECHO) reduce 214) ((T_ELSE) reduce 214) ((T_ELSEIF) reduce 214) ((T_EMPTY) reduce 214) ((T_ENUM) reduce 214) ((T_EVAL) reduce 214) ((T_EXIT) reduce 214) ((T_FILE) reduce 214) ((T_FINAL) reduce 214) ((T_FN) reduce 214) ((T_FOR) reduce 214) ((T_FOREACH) reduce 214) ((T_FUNCTION) reduce 214) ((T_FUNC_C) reduce 214) ((T_GLOBAL) reduce 214) ((T_GOTO) reduce 214) ((T_HALT_COMPILER) reduce 214) ((T_IF) reduce 214) ((T_INC) reduce 214) ((T_INCLUDE) reduce 214) ((T_INCLUDE_ONCE) reduce 214) ((T_INLINE_HTML) reduce 214) ((T_INTERFACE) reduce 214) ((T_INT_CAST) reduce 214) ((T_ISSET) reduce 214) ((T_LINE) reduce 214) ((T_LIST) reduce 214) ((T_LNUMBER) reduce 214) ((T_MATCH) reduce 214) ((T_METHOD_C) reduce 214) ((T_NAMESPACE) reduce 214) ((T_NAME_FULLY_QUALIFIED) reduce 214) ((T_NAME_QUALIFIED) reduce 214) ((T_NAME_RELATIVE) reduce 214) ((T_NEW) reduce 214) ((T_NS_C) reduce 214) ((T_OBJECT_CAST) reduce 214) ((T_PRINT) reduce 214) ((T_REQUIRE) reduce 214) ((T_REQUIRE_ONCE) reduce 214) ((T_RETURN) reduce 214) ((T_START_HEREDOC) reduce 214) ((T_STATIC) reduce 214) ((T_STRING) reduce 214) ((T_STRING_CAST) reduce 214) ((T_SWITCH) reduce 214) ((T_THROW) reduce 214) ((T_TRAIT) reduce 214) ((T_TRAIT_C) reduce 214) ((T_TRY) reduce 214) ((T_UNSET) reduce 214) ((T_UNSET_CAST) reduce 214) ((T_USE) reduce 214) ((T_VARIABLE) reduce 214) ((T_WHILE) reduce 214) ((T_YIELD) reduce 214) ((T_YIELD_FROM) reduce 214) (("[") reduce 214) (("`") reduce 214) (("{") reduce 214) (("~") reduce 214)) 27736 ((("!") reduce 215) (("\"") reduce 215) (("$") reduce 215) (($) reduce 215) (("(") reduce 215) (("+") reduce 215) (("-") reduce 215) ((";") reduce 215) (("@") reduce 215) ((T_ABSTRACT) reduce 215) ((T_ARRAY) reduce 215) ((T_ARRAY_CAST) reduce 215) ((T_ATTRIBUTE) reduce 215) ((T_BOOL_CAST) reduce 215) ((T_BREAK) reduce 215) ((T_CLASS) reduce 215) ((T_CLASS_C) reduce 215) ((T_CLONE) reduce 215) ((T_CONST) reduce 215) ((T_CONSTANT_ENCAPSED_STRING) reduce 215) ((T_CONTINUE) reduce 215) ((T_DEC) reduce 215) ((T_DECLARE) reduce 215) ((T_DIR) reduce 215) ((T_DNUMBER) reduce 215) ((T_DO) reduce 215) ((T_DOUBLE_CAST) reduce 215) ((T_ECHO) reduce 215) ((T_ELSE) reduce 215) ((T_ELSEIF) reduce 215) ((T_EMPTY) reduce 215) ((T_ENUM) reduce 215) ((T_EVAL) reduce 215) ((T_EXIT) reduce 215) ((T_FILE) reduce 215) ((T_FINAL) reduce 215) ((T_FN) reduce 215) ((T_FOR) reduce 215) ((T_FOREACH) reduce 215) ((T_FUNCTION) reduce 215) ((T_FUNC_C) reduce 215) ((T_GLOBAL) reduce 215) ((T_GOTO) reduce 215) ((T_HALT_COMPILER) reduce 215) ((T_IF) reduce 215) ((T_INC) reduce 215) ((T_INCLUDE) reduce 215) ((T_INCLUDE_ONCE) reduce 215) ((T_INLINE_HTML) reduce 215) ((T_INTERFACE) reduce 215) ((T_INT_CAST) reduce 215) ((T_ISSET) reduce 215) ((T_LINE) reduce 215) ((T_LIST) reduce 215) ((T_LNUMBER) reduce 215) ((T_MATCH) reduce 215) ((T_METHOD_C) reduce 215) ((T_NAMESPACE) reduce 215) ((T_NAME_FULLY_QUALIFIED) reduce 215) ((T_NAME_QUALIFIED) reduce 215) ((T_NAME_RELATIVE) reduce 215) ((T_NEW) reduce 215) ((T_NS_C) reduce 215) ((T_OBJECT_CAST) reduce 215) ((T_PRINT) reduce 215) ((T_REQUIRE) reduce 215) ((T_REQUIRE_ONCE) reduce 215) ((T_RETURN) reduce 215) ((T_START_HEREDOC) reduce 215) ((T_STATIC) reduce 215) ((T_STRING) reduce 215) ((T_STRING_CAST) reduce 215) ((T_SWITCH) reduce 215) ((T_THROW) reduce 215) ((T_TRAIT) reduce 215) ((T_TRAIT_C) reduce 215) ((T_TRY) reduce 215) ((T_UNSET) reduce 215) ((T_UNSET_CAST) reduce 215) ((T_USE) reduce 215) ((T_VARIABLE) reduce 215) ((T_WHILE) reduce 215) ((T_YIELD) reduce 215) ((T_YIELD_FROM) reduce 215) (("[") reduce 215) (("`") reduce 215) (("{") reduce 215) (("~") reduce 215)) 27740 ((("!") reduce 216) (("\"") reduce 216) (("$") reduce 216) (($) reduce 216) (("(") reduce 216) (("+") reduce 216) (("-") reduce 216) ((";") reduce 216) (("@") reduce 216) ((T_ABSTRACT) reduce 216) ((T_ARRAY) reduce 216) ((T_ARRAY_CAST) reduce 216) ((T_ATTRIBUTE) reduce 216) ((T_BOOL_CAST) reduce 216) ((T_BREAK) reduce 216) ((T_CLASS) reduce 216) ((T_CLASS_C) reduce 216) ((T_CLONE) reduce 216) ((T_CONST) reduce 216) ((T_CONSTANT_ENCAPSED_STRING) reduce 216) ((T_CONTINUE) reduce 216) ((T_DEC) reduce 216) ((T_DECLARE) reduce 216) ((T_DIR) reduce 216) ((T_DNUMBER) reduce 216) ((T_DO) reduce 216) ((T_DOUBLE_CAST) reduce 216) ((T_ECHO) reduce 216) ((T_ELSE) reduce 216) ((T_ELSEIF) reduce 216) ((T_EMPTY) reduce 216) ((T_ENUM) reduce 216) ((T_EVAL) reduce 216) ((T_EXIT) reduce 216) ((T_FILE) reduce 216) ((T_FINAL) reduce 216) ((T_FN) reduce 216) ((T_FOR) reduce 216) ((T_FOREACH) reduce 216) ((T_FUNCTION) reduce 216) ((T_FUNC_C) reduce 216) ((T_GLOBAL) reduce 216) ((T_GOTO) reduce 216) ((T_HALT_COMPILER) reduce 216) ((T_IF) reduce 216) ((T_INC) reduce 216) ((T_INCLUDE) reduce 216) ((T_INCLUDE_ONCE) reduce 216) ((T_INLINE_HTML) reduce 216) ((T_INTERFACE) reduce 216) ((T_INT_CAST) reduce 216) ((T_ISSET) reduce 216) ((T_LINE) reduce 216) ((T_LIST) reduce 216) ((T_LNUMBER) reduce 216) ((T_MATCH) reduce 216) ((T_METHOD_C) reduce 216) ((T_NAMESPACE) reduce 216) ((T_NAME_FULLY_QUALIFIED) reduce 216) ((T_NAME_QUALIFIED) reduce 216) ((T_NAME_RELATIVE) reduce 216) ((T_NEW) reduce 216) ((T_NS_C) reduce 216) ((T_OBJECT_CAST) reduce 216) ((T_PRINT) reduce 216) ((T_REQUIRE) reduce 216) ((T_REQUIRE_ONCE) reduce 216) ((T_RETURN) reduce 216) ((T_START_HEREDOC) reduce 216) ((T_STATIC) reduce 216) ((T_STRING) reduce 216) ((T_STRING_CAST) reduce 216) ((T_SWITCH) reduce 216) ((T_THROW) reduce 216) ((T_TRAIT) reduce 216) ((T_TRAIT_C) reduce 216) ((T_TRY) reduce 216) ((T_UNSET) reduce 216) ((T_UNSET_CAST) reduce 216) ((T_USE) reduce 216) ((T_VARIABLE) reduce 216) ((T_WHILE) reduce 216) ((T_YIELD) reduce 216) ((T_YIELD_FROM) reduce 216) (("[") reduce 216) (("`") reduce 216) (("{") reduce 216) (("~") reduce 216)) 27743 ((("!") reduce 217) (("\"") reduce 217) (($) reduce 217) (("$") reduce 217) (("(") reduce 217) (("+") reduce 217) (("-") reduce 217) ((";") reduce 217) (("@") reduce 217) ((T_ABSTRACT) reduce 217) ((T_ARRAY) reduce 217) ((T_ARRAY_CAST) reduce 217) ((T_ATTRIBUTE) reduce 217) ((T_BOOL_CAST) reduce 217) ((T_BREAK) reduce 217) ((T_CLASS) reduce 217) ((T_CLASS_C) reduce 217) ((T_CLONE) reduce 217) ((T_CONST) reduce 217) ((T_CONSTANT_ENCAPSED_STRING) reduce 217) ((T_CONTINUE) reduce 217) ((T_DEC) reduce 217) ((T_DECLARE) reduce 217) ((T_DIR) reduce 217) ((T_DNUMBER) reduce 217) ((T_DO) reduce 217) ((T_DOUBLE_CAST) reduce 217) ((T_ECHO) reduce 217) ((T_ELSE) reduce 217) ((T_ELSEIF) reduce 217) ((T_EMPTY) reduce 217) ((T_ENUM) reduce 217) ((T_EVAL) reduce 217) ((T_EXIT) reduce 217) ((T_FILE) reduce 217) ((T_FINAL) reduce 217) ((T_FN) reduce 217) ((T_FOR) reduce 217) ((T_FOREACH) reduce 217) ((T_FUNCTION) reduce 217) ((T_FUNC_C) reduce 217) ((T_GLOBAL) reduce 217) ((T_GOTO) reduce 217) ((T_HALT_COMPILER) reduce 217) ((T_IF) reduce 217) ((T_INC) reduce 217) ((T_INCLUDE) reduce 217) ((T_INCLUDE_ONCE) reduce 217) ((T_INLINE_HTML) reduce 217) ((T_INTERFACE) reduce 217) ((T_INT_CAST) reduce 217) ((T_ISSET) reduce 217) ((T_LINE) reduce 217) ((T_LIST) reduce 217) ((T_LNUMBER) reduce 217) ((T_MATCH) reduce 217) ((T_METHOD_C) reduce 217) ((T_NAMESPACE) reduce 217) ((T_NAME_FULLY_QUALIFIED) reduce 217) ((T_NAME_QUALIFIED) reduce 217) ((T_NAME_RELATIVE) reduce 217) ((T_NEW) reduce 217) ((T_NS_C) reduce 217) ((T_OBJECT_CAST) reduce 217) ((T_PRINT) reduce 217) ((T_REQUIRE) reduce 217) ((T_REQUIRE_ONCE) reduce 217) ((T_RETURN) reduce 217) ((T_START_HEREDOC) reduce 217) ((T_STATIC) reduce 217) ((T_STRING) reduce 217) ((T_STRING_CAST) reduce 217) ((T_SWITCH) reduce 217) ((T_THROW) reduce 217) ((T_TRAIT) reduce 217) ((T_TRAIT_C) reduce 217) ((T_TRY) reduce 217) ((T_UNSET) reduce 217) ((T_UNSET_CAST) reduce 217) ((T_USE) reduce 217) ((T_VARIABLE) reduce 217) ((T_WHILE) reduce 217) ((T_YIELD) reduce 217) ((T_YIELD_FROM) reduce 217) (("[") reduce 217) (("`") reduce 217) (("{") reduce 217) (("~") reduce 217)) 27744 ((("!") reduce 167) (("\"") reduce 167) (($) reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 27746 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (($) reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 27748 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (($) reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 27752 ((("!") reduce 234) (("\"") reduce 234) (("$") reduce 234) (($) reduce 234) (("(") reduce 234) (("+") reduce 234) (("-") reduce 234) ((";") reduce 234) (("@") reduce 234) ((T_ABSTRACT) reduce 234) ((T_ARRAY) reduce 234) ((T_ARRAY_CAST) reduce 234) ((T_ATTRIBUTE) reduce 234) ((T_BOOL_CAST) reduce 234) ((T_BREAK) reduce 234) ((T_CLASS) reduce 234) ((T_CLASS_C) reduce 234) ((T_CLONE) reduce 234) ((T_CONST) reduce 234) ((T_CONSTANT_ENCAPSED_STRING) reduce 234) ((T_CONTINUE) reduce 234) ((T_DEC) reduce 234) ((T_DECLARE) reduce 234) ((T_DIR) reduce 234) ((T_DNUMBER) reduce 234) ((T_DO) reduce 234) ((T_DOUBLE_CAST) reduce 234) ((T_ECHO) reduce 234) ((T_ELSE) reduce 234) ((T_ELSEIF) reduce 234) ((T_EMPTY) reduce 234) ((T_ENUM) reduce 234) ((T_EVAL) reduce 234) ((T_EXIT) reduce 234) ((T_FILE) reduce 234) ((T_FINAL) reduce 234) ((T_FN) reduce 234) ((T_FOR) reduce 234) ((T_FOREACH) reduce 234) ((T_FUNCTION) reduce 234) ((T_FUNC_C) reduce 234) ((T_GLOBAL) reduce 234) ((T_GOTO) reduce 234) ((T_HALT_COMPILER) reduce 234) ((T_IF) reduce 234) ((T_INC) reduce 234) ((T_INCLUDE) reduce 234) ((T_INCLUDE_ONCE) reduce 234) ((T_INLINE_HTML) reduce 234) ((T_INTERFACE) reduce 234) ((T_INT_CAST) reduce 234) ((T_ISSET) reduce 234) ((T_LINE) reduce 234) ((T_LIST) reduce 234) ((T_LNUMBER) reduce 234) ((T_MATCH) reduce 234) ((T_METHOD_C) reduce 234) ((T_NAMESPACE) reduce 234) ((T_NAME_FULLY_QUALIFIED) reduce 234) ((T_NAME_QUALIFIED) reduce 234) ((T_NAME_RELATIVE) reduce 234) ((T_NEW) reduce 234) ((T_NS_C) reduce 234) ((T_OBJECT_CAST) reduce 234) ((T_PRINT) reduce 234) ((T_REQUIRE) reduce 234) ((T_REQUIRE_ONCE) reduce 234) ((T_RETURN) reduce 234) ((T_START_HEREDOC) reduce 234) ((T_STATIC) reduce 234) ((T_STRING) reduce 234) ((T_STRING_CAST) reduce 234) ((T_SWITCH) reduce 234) ((T_THROW) reduce 234) ((T_TRAIT) reduce 234) ((T_TRAIT_C) reduce 234) ((T_TRY) reduce 234) ((T_UNSET) reduce 234) ((T_UNSET_CAST) reduce 234) ((T_USE) reduce 234) ((T_VARIABLE) reduce 234) ((T_WHILE) reduce 234) ((T_YIELD) reduce 234) ((T_YIELD_FROM) reduce 234) (("[") reduce 234) (("`") reduce 234) (("{") reduce 234) (("~") reduce 234)) 27754 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (($) reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 27756 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (($) reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 27766 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (($) reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 27767 ((("!") reduce 210) (("\"") reduce 210) (($) reduce 210) (("$") reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONST) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_ELSE) reduce 210) ((T_ELSEIF) reduce 210) ((T_EMPTY) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAMESPACE) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_USE) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 27770 ((("!") reduce 211) (("\"") reduce 211) (("$") reduce 211) (($) reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONST) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_ELSE) reduce 211) ((T_ELSEIF) reduce 211) ((T_EMPTY) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAMESPACE) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_USE) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 27771 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (($) reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 27780 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (($) reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 27781 ((("!") reduce 208) (("\"") reduce 208) (($) reduce 208) (("$") reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONST) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_ELSE) reduce 208) ((T_ELSEIF) reduce 208) ((T_EMPTY) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAMESPACE) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_USE) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 27784 ((("!") reduce 209) (("\"") reduce 209) (("$") reduce 209) (($) reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONST) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_ELSE) reduce 209) ((T_ELSEIF) reduce 209) ((T_EMPTY) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAMESPACE) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_USE) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 27786 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (($) reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 27792 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (($) reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 27797 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (($) reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 27798 ((("!") reduce 212) (("\"") reduce 212) (($) reduce 212) (("$") reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONST) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_ELSE) reduce 212) ((T_ELSEIF) reduce 212) ((T_EMPTY) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAMESPACE) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_USE) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 27801 ((("!") reduce 213) (("\"") reduce 213) (("$") reduce 213) (($) reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONST) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_ELSE) reduce 213) ((T_ELSEIF) reduce 213) ((T_EMPTY) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAMESPACE) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_USE) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 27803 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (($) reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 27805 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (($) reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 27808 ((("!") reduce 110) (("\"") reduce 110) (($) reduce 110) (("$") reduce 110) (("(") reduce 110) (("+") reduce 110) (("-") reduce 110) ((";") reduce 110) (("@") reduce 110) ((T_ABSTRACT) reduce 110) ((T_ARRAY) reduce 110) ((T_ARRAY_CAST) reduce 110) ((T_ATTRIBUTE) reduce 110) ((T_BOOL_CAST) reduce 110) ((T_BREAK) reduce 110) ((T_CLASS) reduce 110) ((T_CLASS_C) reduce 110) ((T_CLONE) reduce 110) ((T_CONST) reduce 110) ((T_CONSTANT_ENCAPSED_STRING) reduce 110) ((T_CONTINUE) reduce 110) ((T_DEC) reduce 110) ((T_DECLARE) reduce 110) ((T_DIR) reduce 110) ((T_DNUMBER) reduce 110) ((T_DO) reduce 110) ((T_DOUBLE_CAST) reduce 110) ((T_ECHO) reduce 110) ((T_EMPTY) reduce 110) ((T_ENUM) reduce 110) ((T_EVAL) reduce 110) ((T_EXIT) reduce 110) ((T_FILE) reduce 110) ((T_FINAL) reduce 110) ((T_FN) reduce 110) ((T_FOR) reduce 110) ((T_FOREACH) reduce 110) ((T_FUNCTION) reduce 110) ((T_FUNC_C) reduce 110) ((T_GLOBAL) reduce 110) ((T_GOTO) reduce 110) ((T_HALT_COMPILER) reduce 110) ((T_IF) reduce 110) ((T_INC) reduce 110) ((T_INCLUDE) reduce 110) ((T_INCLUDE_ONCE) reduce 110) ((T_INLINE_HTML) reduce 110) ((T_INTERFACE) reduce 110) ((T_INT_CAST) reduce 110) ((T_ISSET) reduce 110) ((T_LINE) reduce 110) ((T_LIST) reduce 110) ((T_LNUMBER) reduce 110) ((T_MATCH) reduce 110) ((T_METHOD_C) reduce 110) ((T_NAMESPACE) reduce 110) ((T_NAME_FULLY_QUALIFIED) reduce 110) ((T_NAME_QUALIFIED) reduce 110) ((T_NAME_RELATIVE) reduce 110) ((T_NEW) reduce 110) ((T_NS_C) reduce 110) ((T_OBJECT_CAST) reduce 110) ((T_PRINT) reduce 110) ((T_REQUIRE) reduce 110) ((T_REQUIRE_ONCE) reduce 110) ((T_RETURN) reduce 110) ((T_START_HEREDOC) reduce 110) ((T_STATIC) reduce 110) ((T_STRING) reduce 110) ((T_STRING_CAST) reduce 110) ((T_SWITCH) reduce 110) ((T_THROW) reduce 110) ((T_TRAIT) reduce 110) ((T_TRAIT_C) reduce 110) ((T_TRY) reduce 110) ((T_UNSET) reduce 110) ((T_UNSET_CAST) reduce 110) ((T_USE) reduce 110) ((T_VARIABLE) reduce 110) ((T_WHILE) reduce 110) ((T_YIELD) reduce 110) ((T_YIELD_FROM) reduce 110) (("[") reduce 110) (("`") reduce 110) (("{") reduce 110) (("~") reduce 110)) 27810 ((("!") reduce 166) (("\"") reduce 166) (($) reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 27812 ((("!") reduce 155) (("\"") reduce 155) (($) reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 27822 ((("!") reduce 162) (("\"") reduce 162) (($) reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 27823 ((("!") reduce 210) (("\"") reduce 210) (("$") reduce 210) (($) reduce 210) (("(") reduce 210) (("+") reduce 210) (("-") reduce 210) ((";") reduce 210) (("@") reduce 210) ((T_ABSTRACT) reduce 210) ((T_ARRAY) reduce 210) ((T_ARRAY_CAST) reduce 210) ((T_ATTRIBUTE) reduce 210) ((T_BOOL_CAST) reduce 210) ((T_BREAK) reduce 210) ((T_CLASS) reduce 210) ((T_CLASS_C) reduce 210) ((T_CLONE) reduce 210) ((T_CONST) reduce 210) ((T_CONSTANT_ENCAPSED_STRING) reduce 210) ((T_CONTINUE) reduce 210) ((T_DEC) reduce 210) ((T_DECLARE) reduce 210) ((T_DIR) reduce 210) ((T_DNUMBER) reduce 210) ((T_DO) reduce 210) ((T_DOUBLE_CAST) reduce 210) ((T_ECHO) reduce 210) ((T_EMPTY) reduce 210) ((T_ENUM) reduce 210) ((T_EVAL) reduce 210) ((T_EXIT) reduce 210) ((T_FILE) reduce 210) ((T_FINAL) reduce 210) ((T_FN) reduce 210) ((T_FOR) reduce 210) ((T_FOREACH) reduce 210) ((T_FUNCTION) reduce 210) ((T_FUNC_C) reduce 210) ((T_GLOBAL) reduce 210) ((T_GOTO) reduce 210) ((T_HALT_COMPILER) reduce 210) ((T_IF) reduce 210) ((T_INC) reduce 210) ((T_INCLUDE) reduce 210) ((T_INCLUDE_ONCE) reduce 210) ((T_INLINE_HTML) reduce 210) ((T_INTERFACE) reduce 210) ((T_INT_CAST) reduce 210) ((T_ISSET) reduce 210) ((T_LINE) reduce 210) ((T_LIST) reduce 210) ((T_LNUMBER) reduce 210) ((T_MATCH) reduce 210) ((T_METHOD_C) reduce 210) ((T_NAMESPACE) reduce 210) ((T_NAME_FULLY_QUALIFIED) reduce 210) ((T_NAME_QUALIFIED) reduce 210) ((T_NAME_RELATIVE) reduce 210) ((T_NEW) reduce 210) ((T_NS_C) reduce 210) ((T_OBJECT_CAST) reduce 210) ((T_PRINT) reduce 210) ((T_REQUIRE) reduce 210) ((T_REQUIRE_ONCE) reduce 210) ((T_RETURN) reduce 210) ((T_START_HEREDOC) reduce 210) ((T_STATIC) reduce 210) ((T_STRING) reduce 210) ((T_STRING_CAST) reduce 210) ((T_SWITCH) reduce 210) ((T_THROW) reduce 210) ((T_TRAIT) reduce 210) ((T_TRAIT_C) reduce 210) ((T_TRY) reduce 210) ((T_UNSET) reduce 210) ((T_UNSET_CAST) reduce 210) ((T_USE) reduce 210) ((T_VARIABLE) reduce 210) ((T_WHILE) reduce 210) ((T_YIELD) reduce 210) ((T_YIELD_FROM) reduce 210) (("[") reduce 210) (("`") reduce 210) (("{") reduce 210) (("~") reduce 210)) 27826 ((("!") reduce 211) (("\"") reduce 211) (($) reduce 211) (("$") reduce 211) (("(") reduce 211) (("+") reduce 211) (("-") reduce 211) ((";") reduce 211) (("@") reduce 211) ((T_ABSTRACT) reduce 211) ((T_ARRAY) reduce 211) ((T_ARRAY_CAST) reduce 211) ((T_ATTRIBUTE) reduce 211) ((T_BOOL_CAST) reduce 211) ((T_BREAK) reduce 211) ((T_CLASS) reduce 211) ((T_CLASS_C) reduce 211) ((T_CLONE) reduce 211) ((T_CONST) reduce 211) ((T_CONSTANT_ENCAPSED_STRING) reduce 211) ((T_CONTINUE) reduce 211) ((T_DEC) reduce 211) ((T_DECLARE) reduce 211) ((T_DIR) reduce 211) ((T_DNUMBER) reduce 211) ((T_DO) reduce 211) ((T_DOUBLE_CAST) reduce 211) ((T_ECHO) reduce 211) ((T_EMPTY) reduce 211) ((T_ENUM) reduce 211) ((T_EVAL) reduce 211) ((T_EXIT) reduce 211) ((T_FILE) reduce 211) ((T_FINAL) reduce 211) ((T_FN) reduce 211) ((T_FOR) reduce 211) ((T_FOREACH) reduce 211) ((T_FUNCTION) reduce 211) ((T_FUNC_C) reduce 211) ((T_GLOBAL) reduce 211) ((T_GOTO) reduce 211) ((T_HALT_COMPILER) reduce 211) ((T_IF) reduce 211) ((T_INC) reduce 211) ((T_INCLUDE) reduce 211) ((T_INCLUDE_ONCE) reduce 211) ((T_INLINE_HTML) reduce 211) ((T_INTERFACE) reduce 211) ((T_INT_CAST) reduce 211) ((T_ISSET) reduce 211) ((T_LINE) reduce 211) ((T_LIST) reduce 211) ((T_LNUMBER) reduce 211) ((T_MATCH) reduce 211) ((T_METHOD_C) reduce 211) ((T_NAMESPACE) reduce 211) ((T_NAME_FULLY_QUALIFIED) reduce 211) ((T_NAME_QUALIFIED) reduce 211) ((T_NAME_RELATIVE) reduce 211) ((T_NEW) reduce 211) ((T_NS_C) reduce 211) ((T_OBJECT_CAST) reduce 211) ((T_PRINT) reduce 211) ((T_REQUIRE) reduce 211) ((T_REQUIRE_ONCE) reduce 211) ((T_RETURN) reduce 211) ((T_START_HEREDOC) reduce 211) ((T_STATIC) reduce 211) ((T_STRING) reduce 211) ((T_STRING_CAST) reduce 211) ((T_SWITCH) reduce 211) ((T_THROW) reduce 211) ((T_TRAIT) reduce 211) ((T_TRAIT_C) reduce 211) ((T_TRY) reduce 211) ((T_UNSET) reduce 211) ((T_UNSET_CAST) reduce 211) ((T_USE) reduce 211) ((T_VARIABLE) reduce 211) ((T_WHILE) reduce 211) ((T_YIELD) reduce 211) ((T_YIELD_FROM) reduce 211) (("[") reduce 211) (("`") reduce 211) (("{") reduce 211) (("~") reduce 211)) 27827 ((("!") reduce 161) (("\"") reduce 161) (($) reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 27836 ((("!") reduce 150) (("\"") reduce 150) (($) reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 27837 ((("!") reduce 208) (("\"") reduce 208) (("$") reduce 208) (($) reduce 208) (("(") reduce 208) (("+") reduce 208) (("-") reduce 208) ((";") reduce 208) (("@") reduce 208) ((T_ABSTRACT) reduce 208) ((T_ARRAY) reduce 208) ((T_ARRAY_CAST) reduce 208) ((T_ATTRIBUTE) reduce 208) ((T_BOOL_CAST) reduce 208) ((T_BREAK) reduce 208) ((T_CLASS) reduce 208) ((T_CLASS_C) reduce 208) ((T_CLONE) reduce 208) ((T_CONST) reduce 208) ((T_CONSTANT_ENCAPSED_STRING) reduce 208) ((T_CONTINUE) reduce 208) ((T_DEC) reduce 208) ((T_DECLARE) reduce 208) ((T_DIR) reduce 208) ((T_DNUMBER) reduce 208) ((T_DO) reduce 208) ((T_DOUBLE_CAST) reduce 208) ((T_ECHO) reduce 208) ((T_EMPTY) reduce 208) ((T_ENUM) reduce 208) ((T_EVAL) reduce 208) ((T_EXIT) reduce 208) ((T_FILE) reduce 208) ((T_FINAL) reduce 208) ((T_FN) reduce 208) ((T_FOR) reduce 208) ((T_FOREACH) reduce 208) ((T_FUNCTION) reduce 208) ((T_FUNC_C) reduce 208) ((T_GLOBAL) reduce 208) ((T_GOTO) reduce 208) ((T_HALT_COMPILER) reduce 208) ((T_IF) reduce 208) ((T_INC) reduce 208) ((T_INCLUDE) reduce 208) ((T_INCLUDE_ONCE) reduce 208) ((T_INLINE_HTML) reduce 208) ((T_INTERFACE) reduce 208) ((T_INT_CAST) reduce 208) ((T_ISSET) reduce 208) ((T_LINE) reduce 208) ((T_LIST) reduce 208) ((T_LNUMBER) reduce 208) ((T_MATCH) reduce 208) ((T_METHOD_C) reduce 208) ((T_NAMESPACE) reduce 208) ((T_NAME_FULLY_QUALIFIED) reduce 208) ((T_NAME_QUALIFIED) reduce 208) ((T_NAME_RELATIVE) reduce 208) ((T_NEW) reduce 208) ((T_NS_C) reduce 208) ((T_OBJECT_CAST) reduce 208) ((T_PRINT) reduce 208) ((T_REQUIRE) reduce 208) ((T_REQUIRE_ONCE) reduce 208) ((T_RETURN) reduce 208) ((T_START_HEREDOC) reduce 208) ((T_STATIC) reduce 208) ((T_STRING) reduce 208) ((T_STRING_CAST) reduce 208) ((T_SWITCH) reduce 208) ((T_THROW) reduce 208) ((T_TRAIT) reduce 208) ((T_TRAIT_C) reduce 208) ((T_TRY) reduce 208) ((T_UNSET) reduce 208) ((T_UNSET_CAST) reduce 208) ((T_USE) reduce 208) ((T_VARIABLE) reduce 208) ((T_WHILE) reduce 208) ((T_YIELD) reduce 208) ((T_YIELD_FROM) reduce 208) (("[") reduce 208) (("`") reduce 208) (("{") reduce 208) (("~") reduce 208)) 27840 ((("!") reduce 209) (("\"") reduce 209) (($) reduce 209) (("$") reduce 209) (("(") reduce 209) (("+") reduce 209) (("-") reduce 209) ((";") reduce 209) (("@") reduce 209) ((T_ABSTRACT) reduce 209) ((T_ARRAY) reduce 209) ((T_ARRAY_CAST) reduce 209) ((T_ATTRIBUTE) reduce 209) ((T_BOOL_CAST) reduce 209) ((T_BREAK) reduce 209) ((T_CLASS) reduce 209) ((T_CLASS_C) reduce 209) ((T_CLONE) reduce 209) ((T_CONST) reduce 209) ((T_CONSTANT_ENCAPSED_STRING) reduce 209) ((T_CONTINUE) reduce 209) ((T_DEC) reduce 209) ((T_DECLARE) reduce 209) ((T_DIR) reduce 209) ((T_DNUMBER) reduce 209) ((T_DO) reduce 209) ((T_DOUBLE_CAST) reduce 209) ((T_ECHO) reduce 209) ((T_EMPTY) reduce 209) ((T_ENUM) reduce 209) ((T_EVAL) reduce 209) ((T_EXIT) reduce 209) ((T_FILE) reduce 209) ((T_FINAL) reduce 209) ((T_FN) reduce 209) ((T_FOR) reduce 209) ((T_FOREACH) reduce 209) ((T_FUNCTION) reduce 209) ((T_FUNC_C) reduce 209) ((T_GLOBAL) reduce 209) ((T_GOTO) reduce 209) ((T_HALT_COMPILER) reduce 209) ((T_IF) reduce 209) ((T_INC) reduce 209) ((T_INCLUDE) reduce 209) ((T_INCLUDE_ONCE) reduce 209) ((T_INLINE_HTML) reduce 209) ((T_INTERFACE) reduce 209) ((T_INT_CAST) reduce 209) ((T_ISSET) reduce 209) ((T_LINE) reduce 209) ((T_LIST) reduce 209) ((T_LNUMBER) reduce 209) ((T_MATCH) reduce 209) ((T_METHOD_C) reduce 209) ((T_NAMESPACE) reduce 209) ((T_NAME_FULLY_QUALIFIED) reduce 209) ((T_NAME_QUALIFIED) reduce 209) ((T_NAME_RELATIVE) reduce 209) ((T_NEW) reduce 209) ((T_NS_C) reduce 209) ((T_OBJECT_CAST) reduce 209) ((T_PRINT) reduce 209) ((T_REQUIRE) reduce 209) ((T_REQUIRE_ONCE) reduce 209) ((T_RETURN) reduce 209) ((T_START_HEREDOC) reduce 209) ((T_STATIC) reduce 209) ((T_STRING) reduce 209) ((T_STRING_CAST) reduce 209) ((T_SWITCH) reduce 209) ((T_THROW) reduce 209) ((T_TRAIT) reduce 209) ((T_TRAIT_C) reduce 209) ((T_TRY) reduce 209) ((T_UNSET) reduce 209) ((T_UNSET_CAST) reduce 209) ((T_USE) reduce 209) ((T_VARIABLE) reduce 209) ((T_WHILE) reduce 209) ((T_YIELD) reduce 209) ((T_YIELD_FROM) reduce 209) (("[") reduce 209) (("`") reduce 209) (("{") reduce 209) (("~") reduce 209)) 27854 ((("!") reduce 192) (("\"") reduce 192) (($) reduce 192) (("$") reduce 192) (("(") reduce 192) (("+") reduce 192) (("-") reduce 192) ((";") reduce 192) (("@") reduce 192) ((T_ABSTRACT) reduce 192) ((T_ARRAY) reduce 192) ((T_ARRAY_CAST) reduce 192) ((T_ATTRIBUTE) reduce 192) ((T_BOOL_CAST) reduce 192) ((T_BREAK) reduce 192) ((T_CLASS) reduce 192) ((T_CLASS_C) reduce 192) ((T_CLONE) reduce 192) ((T_CONST) reduce 192) ((T_CONSTANT_ENCAPSED_STRING) reduce 192) ((T_CONTINUE) reduce 192) ((T_DEC) reduce 192) ((T_DECLARE) reduce 192) ((T_DIR) reduce 192) ((T_DNUMBER) reduce 192) ((T_DO) reduce 192) ((T_DOUBLE_CAST) reduce 192) ((T_ECHO) reduce 192) ((T_EMPTY) reduce 192) ((T_ENUM) reduce 192) ((T_EVAL) reduce 192) ((T_EXIT) reduce 192) ((T_FILE) reduce 192) ((T_FINAL) reduce 192) ((T_FN) reduce 192) ((T_FOR) reduce 192) ((T_FOREACH) reduce 192) ((T_FUNCTION) reduce 192) ((T_FUNC_C) reduce 192) ((T_GLOBAL) reduce 192) ((T_GOTO) reduce 192) ((T_HALT_COMPILER) reduce 192) ((T_IF) reduce 192) ((T_INC) reduce 192) ((T_INCLUDE) reduce 192) ((T_INCLUDE_ONCE) reduce 192) ((T_INLINE_HTML) reduce 192) ((T_INTERFACE) reduce 192) ((T_INT_CAST) reduce 192) ((T_ISSET) reduce 192) ((T_LINE) reduce 192) ((T_LIST) reduce 192) ((T_LNUMBER) reduce 192) ((T_MATCH) reduce 192) ((T_METHOD_C) reduce 192) ((T_NAMESPACE) reduce 192) ((T_NAME_FULLY_QUALIFIED) reduce 192) ((T_NAME_QUALIFIED) reduce 192) ((T_NAME_RELATIVE) reduce 192) ((T_NEW) reduce 192) ((T_NS_C) reduce 192) ((T_OBJECT_CAST) reduce 192) ((T_PRINT) reduce 192) ((T_REQUIRE) reduce 192) ((T_REQUIRE_ONCE) reduce 192) ((T_RETURN) reduce 192) ((T_START_HEREDOC) reduce 192) ((T_STATIC) reduce 192) ((T_STRING) reduce 192) ((T_STRING_CAST) reduce 192) ((T_SWITCH) reduce 192) ((T_THROW) reduce 192) ((T_TRAIT) reduce 192) ((T_TRAIT_C) reduce 192) ((T_TRY) reduce 192) ((T_UNSET) reduce 192) ((T_UNSET_CAST) reduce 192) ((T_USE) reduce 192) ((T_VARIABLE) reduce 192) ((T_WHILE) reduce 192) ((T_YIELD) reduce 192) ((T_YIELD_FROM) reduce 192) (("[") reduce 192) (("`") reduce 192) (("{") reduce 192) (("~") reduce 192)) 27859 ((("!") reduce 157) (("\"") reduce 157) (($) reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 27866 ((("!") reduce 149) (("\"") reduce 149) (($) reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 27871 ((("!") reduce 163) (("\"") reduce 163) (($) reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 27872 ((("!") reduce 212) (("\"") reduce 212) (("$") reduce 212) (($) reduce 212) (("(") reduce 212) (("+") reduce 212) (("-") reduce 212) ((";") reduce 212) (("@") reduce 212) ((T_ABSTRACT) reduce 212) ((T_ARRAY) reduce 212) ((T_ARRAY_CAST) reduce 212) ((T_ATTRIBUTE) reduce 212) ((T_BOOL_CAST) reduce 212) ((T_BREAK) reduce 212) ((T_CLASS) reduce 212) ((T_CLASS_C) reduce 212) ((T_CLONE) reduce 212) ((T_CONST) reduce 212) ((T_CONSTANT_ENCAPSED_STRING) reduce 212) ((T_CONTINUE) reduce 212) ((T_DEC) reduce 212) ((T_DECLARE) reduce 212) ((T_DIR) reduce 212) ((T_DNUMBER) reduce 212) ((T_DO) reduce 212) ((T_DOUBLE_CAST) reduce 212) ((T_ECHO) reduce 212) ((T_EMPTY) reduce 212) ((T_ENUM) reduce 212) ((T_EVAL) reduce 212) ((T_EXIT) reduce 212) ((T_FILE) reduce 212) ((T_FINAL) reduce 212) ((T_FN) reduce 212) ((T_FOR) reduce 212) ((T_FOREACH) reduce 212) ((T_FUNCTION) reduce 212) ((T_FUNC_C) reduce 212) ((T_GLOBAL) reduce 212) ((T_GOTO) reduce 212) ((T_HALT_COMPILER) reduce 212) ((T_IF) reduce 212) ((T_INC) reduce 212) ((T_INCLUDE) reduce 212) ((T_INCLUDE_ONCE) reduce 212) ((T_INLINE_HTML) reduce 212) ((T_INTERFACE) reduce 212) ((T_INT_CAST) reduce 212) ((T_ISSET) reduce 212) ((T_LINE) reduce 212) ((T_LIST) reduce 212) ((T_LNUMBER) reduce 212) ((T_MATCH) reduce 212) ((T_METHOD_C) reduce 212) ((T_NAMESPACE) reduce 212) ((T_NAME_FULLY_QUALIFIED) reduce 212) ((T_NAME_QUALIFIED) reduce 212) ((T_NAME_RELATIVE) reduce 212) ((T_NEW) reduce 212) ((T_NS_C) reduce 212) ((T_OBJECT_CAST) reduce 212) ((T_PRINT) reduce 212) ((T_REQUIRE) reduce 212) ((T_REQUIRE_ONCE) reduce 212) ((T_RETURN) reduce 212) ((T_START_HEREDOC) reduce 212) ((T_STATIC) reduce 212) ((T_STRING) reduce 212) ((T_STRING_CAST) reduce 212) ((T_SWITCH) reduce 212) ((T_THROW) reduce 212) ((T_TRAIT) reduce 212) ((T_TRAIT_C) reduce 212) ((T_TRY) reduce 212) ((T_UNSET) reduce 212) ((T_UNSET_CAST) reduce 212) ((T_USE) reduce 212) ((T_VARIABLE) reduce 212) ((T_WHILE) reduce 212) ((T_YIELD) reduce 212) ((T_YIELD_FROM) reduce 212) (("[") reduce 212) (("`") reduce 212) (("{") reduce 212) (("~") reduce 212)) 27875 ((("!") reduce 213) (("\"") reduce 213) (($) reduce 213) (("$") reduce 213) (("(") reduce 213) (("+") reduce 213) (("-") reduce 213) ((";") reduce 213) (("@") reduce 213) ((T_ABSTRACT) reduce 213) ((T_ARRAY) reduce 213) ((T_ARRAY_CAST) reduce 213) ((T_ATTRIBUTE) reduce 213) ((T_BOOL_CAST) reduce 213) ((T_BREAK) reduce 213) ((T_CLASS) reduce 213) ((T_CLASS_C) reduce 213) ((T_CLONE) reduce 213) ((T_CONST) reduce 213) ((T_CONSTANT_ENCAPSED_STRING) reduce 213) ((T_CONTINUE) reduce 213) ((T_DEC) reduce 213) ((T_DECLARE) reduce 213) ((T_DIR) reduce 213) ((T_DNUMBER) reduce 213) ((T_DO) reduce 213) ((T_DOUBLE_CAST) reduce 213) ((T_ECHO) reduce 213) ((T_EMPTY) reduce 213) ((T_ENUM) reduce 213) ((T_EVAL) reduce 213) ((T_EXIT) reduce 213) ((T_FILE) reduce 213) ((T_FINAL) reduce 213) ((T_FN) reduce 213) ((T_FOR) reduce 213) ((T_FOREACH) reduce 213) ((T_FUNCTION) reduce 213) ((T_FUNC_C) reduce 213) ((T_GLOBAL) reduce 213) ((T_GOTO) reduce 213) ((T_HALT_COMPILER) reduce 213) ((T_IF) reduce 213) ((T_INC) reduce 213) ((T_INCLUDE) reduce 213) ((T_INCLUDE_ONCE) reduce 213) ((T_INLINE_HTML) reduce 213) ((T_INTERFACE) reduce 213) ((T_INT_CAST) reduce 213) ((T_ISSET) reduce 213) ((T_LINE) reduce 213) ((T_LIST) reduce 213) ((T_LNUMBER) reduce 213) ((T_MATCH) reduce 213) ((T_METHOD_C) reduce 213) ((T_NAMESPACE) reduce 213) ((T_NAME_FULLY_QUALIFIED) reduce 213) ((T_NAME_QUALIFIED) reduce 213) ((T_NAME_RELATIVE) reduce 213) ((T_NEW) reduce 213) ((T_NS_C) reduce 213) ((T_OBJECT_CAST) reduce 213) ((T_PRINT) reduce 213) ((T_REQUIRE) reduce 213) ((T_REQUIRE_ONCE) reduce 213) ((T_RETURN) reduce 213) ((T_START_HEREDOC) reduce 213) ((T_STATIC) reduce 213) ((T_STRING) reduce 213) ((T_STRING_CAST) reduce 213) ((T_SWITCH) reduce 213) ((T_THROW) reduce 213) ((T_TRAIT) reduce 213) ((T_TRAIT_C) reduce 213) ((T_TRY) reduce 213) ((T_UNSET) reduce 213) ((T_UNSET_CAST) reduce 213) ((T_USE) reduce 213) ((T_VARIABLE) reduce 213) ((T_WHILE) reduce 213) ((T_YIELD) reduce 213) ((T_YIELD_FROM) reduce 213) (("[") reduce 213) (("`") reduce 213) (("{") reduce 213) (("~") reduce 213)) 27878 ((("!") reduce 153) (("\"") reduce 153) (($) reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 27880 ((("!") reduce 118) (("\"") reduce 118) (("$") reduce 118) (($) reduce 118) (("(") reduce 118) (("+") reduce 118) (("-") reduce 118) ((";") reduce 118) (("@") reduce 118) ((T_ABSTRACT) reduce 118) ((T_ARRAY) reduce 118) ((T_ARRAY_CAST) reduce 118) ((T_ATTRIBUTE) reduce 118) ((T_BOOL_CAST) reduce 118) ((T_BREAK) reduce 118) ((T_CLASS) reduce 118) ((T_CLASS_C) reduce 118) ((T_CLONE) reduce 118) ((T_CONST) reduce 118) ((T_CONSTANT_ENCAPSED_STRING) reduce 118) ((T_CONTINUE) reduce 118) ((T_DEC) reduce 118) ((T_DECLARE) reduce 118) ((T_DIR) reduce 118) ((T_DNUMBER) reduce 118) ((T_DO) reduce 118) ((T_DOUBLE_CAST) reduce 118) ((T_ECHO) reduce 118) ((T_EMPTY) reduce 118) ((T_ENUM) reduce 118) ((T_EVAL) reduce 118) ((T_EXIT) reduce 118) ((T_FILE) reduce 118) ((T_FINAL) reduce 118) ((T_FN) reduce 118) ((T_FOR) reduce 118) ((T_FOREACH) reduce 118) ((T_FUNCTION) reduce 118) ((T_FUNC_C) reduce 118) ((T_GLOBAL) reduce 118) ((T_GOTO) reduce 118) ((T_HALT_COMPILER) reduce 118) ((T_IF) reduce 118) ((T_INC) reduce 118) ((T_INCLUDE) reduce 118) ((T_INCLUDE_ONCE) reduce 118) ((T_INLINE_HTML) reduce 118) ((T_INTERFACE) reduce 118) ((T_INT_CAST) reduce 118) ((T_ISSET) reduce 118) ((T_LINE) reduce 118) ((T_LIST) reduce 118) ((T_LNUMBER) reduce 118) ((T_MATCH) reduce 118) ((T_METHOD_C) reduce 118) ((T_NAMESPACE) reduce 118) ((T_NAME_FULLY_QUALIFIED) reduce 118) ((T_NAME_QUALIFIED) reduce 118) ((T_NAME_RELATIVE) reduce 118) ((T_NEW) reduce 118) ((T_NS_C) reduce 118) ((T_OBJECT_CAST) reduce 118) ((T_PRINT) reduce 118) ((T_REQUIRE) reduce 118) ((T_REQUIRE_ONCE) reduce 118) ((T_RETURN) reduce 118) ((T_START_HEREDOC) reduce 118) ((T_STATIC) reduce 118) ((T_STRING) reduce 118) ((T_STRING_CAST) reduce 118) ((T_SWITCH) reduce 118) ((T_THROW) reduce 118) ((T_TRAIT) reduce 118) ((T_TRAIT_C) reduce 118) ((T_TRY) reduce 118) ((T_UNSET) reduce 118) ((T_UNSET_CAST) reduce 118) ((T_USE) reduce 118) ((T_VARIABLE) reduce 118) ((T_WHILE) reduce 118) ((T_YIELD) reduce 118) ((T_YIELD_FROM) reduce 118) (("[") reduce 118) (("`") reduce 118) (("{") reduce 118) (("~") reduce 118)) 27888 ((("!") reduce 185) (("\"") reduce 185) (($) reduce 185) (("$") reduce 185) (("(") reduce 185) (("+") reduce 185) (("-") reduce 185) ((";") reduce 185) (("@") reduce 185) ((T_ABSTRACT) reduce 185) ((T_ARRAY) reduce 185) ((T_ARRAY_CAST) reduce 185) ((T_ATTRIBUTE) reduce 185) ((T_BOOL_CAST) reduce 185) ((T_BREAK) reduce 185) ((T_CLASS) reduce 185) ((T_CLASS_C) reduce 185) ((T_CLONE) reduce 185) ((T_CONST) reduce 185) ((T_CONSTANT_ENCAPSED_STRING) reduce 185) ((T_CONTINUE) reduce 185) ((T_DEC) reduce 185) ((T_DECLARE) reduce 185) ((T_DIR) reduce 185) ((T_DNUMBER) reduce 185) ((T_DO) reduce 185) ((T_DOUBLE_CAST) reduce 185) ((T_ECHO) reduce 185) ((T_EMPTY) reduce 185) ((T_ENUM) reduce 185) ((T_EVAL) reduce 185) ((T_EXIT) reduce 185) ((T_FILE) reduce 185) ((T_FINAL) reduce 185) ((T_FN) reduce 185) ((T_FOR) reduce 185) ((T_FOREACH) reduce 185) ((T_FUNCTION) reduce 185) ((T_FUNC_C) reduce 185) ((T_GLOBAL) reduce 185) ((T_GOTO) reduce 185) ((T_HALT_COMPILER) reduce 185) ((T_IF) reduce 185) ((T_INC) reduce 185) ((T_INCLUDE) reduce 185) ((T_INCLUDE_ONCE) reduce 185) ((T_INLINE_HTML) reduce 185) ((T_INTERFACE) reduce 185) ((T_INT_CAST) reduce 185) ((T_ISSET) reduce 185) ((T_LINE) reduce 185) ((T_LIST) reduce 185) ((T_LNUMBER) reduce 185) ((T_MATCH) reduce 185) ((T_METHOD_C) reduce 185) ((T_NAMESPACE) reduce 185) ((T_NAME_FULLY_QUALIFIED) reduce 185) ((T_NAME_QUALIFIED) reduce 185) ((T_NAME_RELATIVE) reduce 185) ((T_NEW) reduce 185) ((T_NS_C) reduce 185) ((T_OBJECT_CAST) reduce 185) ((T_PRINT) reduce 185) ((T_REQUIRE) reduce 185) ((T_REQUIRE_ONCE) reduce 185) ((T_RETURN) reduce 185) ((T_START_HEREDOC) reduce 185) ((T_STATIC) reduce 185) ((T_STRING) reduce 185) ((T_STRING_CAST) reduce 185) ((T_SWITCH) reduce 185) ((T_THROW) reduce 185) ((T_TRAIT) reduce 185) ((T_TRAIT_C) reduce 185) ((T_TRY) reduce 185) ((T_UNSET) reduce 185) ((T_UNSET_CAST) reduce 185) ((T_USE) reduce 185) ((T_VARIABLE) reduce 185) ((T_WHILE) reduce 185) ((T_YIELD) reduce 185) ((T_YIELD_FROM) reduce 185) (("[") reduce 185) (("`") reduce 185) (("{") reduce 185) (("~") reduce 185)) 27890 ((("!") reduce 152) (("\"") reduce 152) (($) reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 27894 (((T_ABSTRACT) reduce 99) ((T_ATTRIBUTE) reduce 99) ((T_CLASS) reduce 99) ((T_ENUM) reduce 99) ((T_FINAL) reduce 99) ((T_FN) reduce 99) ((T_FUNCTION) reduce 99) ((T_INTERFACE) reduce 99) ((T_STATIC) reduce 99) ((T_TRAIT) reduce 99))))
  "The generated distinct action-tables.")

(defvar
  phps-mode-parser--goto-tables
  #s(hash-table size 28048 test equal rehash-size 1.5 rehash-threshold 0.8125 data (0 0 1 1 2 2 3 3 4 4 5 5 6 6 7 7 8 8 9 9 10 8 11 10 12 11 13 12 14 13 15 14 16 15 17 8 18 16 19 17 20 8 21 18 22 19 23 20 24 8 25 8 26 21 27 22 28 23 29 24 30 25 31 26 32 27 33 8 34 8 35 8 36 28 37 29 38 8 39 8 40 30 41 31 42 32 43 33 44 34 45 35 46 36 47 8 48 37 49 38 50 39 51 8 52 40 53 8 54 41 55 8 56 42 57 8 58 8 59 8 60 43 61 8 62 44 63 45 64 46 65 47 66 48 67 49 68 50 69 51 70 52 71 53 72 54 73 55 74 8 75 56 76 57 77 58 78 59 79 8 80 60 81 61 82 62 83 63 84 64 85 8 86 65 87 66 88 8 89 8 90 67 91 68 92 8 93 8 94 8 95 8 96 69 97 70 98 8 99 8 100 8 101 71 102 72 103 8 104 73 105 8 106 8 107 8 108 74 109 8 110 8 111 8 112 8 113 75 114 8 115 8 116 8 117 8 118 8 119 8 120 8 121 76 122 77 123 78 124 79 125 80 126 8 127 81 128 8 129 8 130 82 131 83 132 84 133 8 134 8 135 8 136 85 137 86 138 87 139 88 140 8 141 8 142 89 143 90 144 91 145 8 146 92 147 93 148 94 149 95 150 96 151 97 152 98 153 99 154 100 155 8 156 101 157 102 158 103 159 104 160 105 161 106 162 107 163 108 164 109 165 110 166 111 167 112 168 113 169 8 170 114 171 115 172 116 173 117 174 118 175 119 176 120 177 121 178 122 179 8 180 123 181 8 182 124 183 8 184 125 185 8 186 126 187 127 188 8 189 8 190 128 191 8 192 8 193 8 194 8 195 129 196 130 197 8 198 131 199 8 200 132 201 133 202 134 203 135 204 136 205 137 206 138 207 139 208 140 209 141 210 142 211 143 212 144 213 8 214 145 215 8 216 146 217 8 218 8 219 147 220 148 221 149 222 150 223 8 224 8 225 151 226 152 227 153 228 154 229 155 230 8 231 156 232 8 233 157 234 8 235 8 236 8 237 8 238 158 239 8 240 159 241 160 242 161 243 162 244 163 245 164 246 8 247 165 248 166 249 8 250 167 251 8 252 168 253 169 254 170 255 171 256 172 257 173 258 174 259 8 260 8 261 175 262 8 263 8 264 176 265 177 266 178 267 8 268 8 269 8 270 8 271 179 272 8 273 8 274 8 275 8 276 180 277 181 278 182 279 183 280 184 281 185 282 186 283 187 284 188 285 189 286 190 287 191 288 192 289 193 290 194 291 195 292 196 293 197 294 198 295 199 296 200 297 201 298 202 299 203 300 204 301 205 302 206 303 207 304 208 305 209 306 210 307 211 308 212 309 213 310 183 311 183 312 183 313 183 314 183 315 183 316 183 317 183 318 183 319 183 320 183 321 183 322 183 323 183 324 183 325 214 326 215 327 8 328 8 329 8 330 8 331 8 332 8 333 216 334 8 335 8 336 217 337 8 338 218 339 219 340 220 341 221 342 222 343 223 344 224 345 225 346 226 347 227 348 228 349 229 350 230 351 231 352 232 353 8 354 233 355 8 356 234 357 8 358 8 359 235 360 236 361 237 362 238 363 8 364 8 365 239 366 240 367 241 368 242 369 243 370 8 371 244 372 8 373 245 374 8 375 8 376 8 377 8 378 246 379 8 380 247 381 248 382 249 383 250 384 251 385 252 386 8 387 253 388 254 389 8 390 255 391 8 392 256 393 257 394 258 395 259 396 260 397 261 398 262 399 8 400 8 401 263 402 8 403 8 404 264 405 265 406 266 407 8 408 8 409 8 410 8 411 267 412 8 413 8 414 8 415 8 416 268 417 269 418 270 419 271 420 272 421 273 422 274 423 275 424 276 425 277 426 278 427 279 428 280 429 281 430 282 431 283 432 284 433 285 434 286 435 287 436 288 437 289 438 290 439 291 440 292 441 293 442 294 443 295 444 296 445 297 446 298 447 299 448 300 449 301 450 271 451 271 452 271 453 271 454 271 455 271 456 271 457 271 458 271 459 271 460 271 461 271 462 271 463 271 464 271 465 302 466 303 467 8 468 8 469 8 470 8 471 8 472 8 473 304 474 8 475 8 476 305 477 8 478 306 479 307 480 308 481 309 482 310 483 311 484 8 485 312 486 313 487 314 488 315 489 316 490 317 491 318 492 319 493 320 494 321 495 8 496 322 497 8 498 323 499 8 500 8 501 324 502 325 503 326 504 327 505 8 506 8 507 328 508 329 509 330 510 331 511 332 512 8 513 333 514 8 515 334 516 8 517 8 518 8 519 8 520 335 521 8 522 336 523 337 524 338 525 339 526 340 527 341 528 8 529 342 530 343 531 8 532 344 533 8 534 345 535 346 536 347 537 348 538 349 539 350 540 351 541 8 542 8 543 352 544 8 545 8 546 353 547 354 548 355 549 8 550 8 551 8 552 8 553 356 554 8 555 357 556 8 557 8 558 8 559 358 560 359 561 360 562 353 563 361 564 362 565 363 566 364 567 365 568 366 569 367 570 368 571 369 572 370 573 371 574 372 575 373 576 374 577 375 578 376 579 377 580 378 581 379 582 380 583 381 584 382 585 383 586 384 587 385 588 386 589 387 590 388 591 389 592 390 593 353 594 353 595 353 596 353 597 353 598 353 599 353 600 353 601 353 602 353 603 353 604 353 605 353 606 353 607 353 608 391 609 392 610 8 611 8 612 8 613 8 614 8 615 8 616 393 617 8 618 8 619 394 620 8 621 395 622 396 623 397 624 398 625 399 626 400 627 8 628 401 629 8 630 8 631 8 632 8 633 8 634 402 635 403 636 8 637 8 638 404 639 8 640 405 641 8 642 8 643 406 644 407 645 8 646 353 647 353 648 353 649 353 650 353 651 408 652 409 653 410 654 411 655 412 656 413 657 414 658 415 659 416 660 417 661 418 662 8 663 419 664 8 665 420 666 8 667 8 668 421 669 422 670 423 671 424 672 8 673 8 674 425 675 426 676 427 677 428 678 429 679 8 680 430 681 8 682 431 683 8 684 8 685 8 686 8 687 432 688 8 689 433 690 434 691 435 692 436 693 437 694 438 695 8 696 439 697 440 698 8 699 441 700 8 701 442 702 443 703 444 704 445 705 446 706 447 707 448 708 8 709 8 710 449 711 8 712 8 713 450 714 451 715 452 716 8 717 8 718 8 719 8 720 453 721 8 722 8 723 8 724 8 725 454 726 455 727 456 728 457 729 458 730 459 731 460 732 461 733 462 734 463 735 464 736 465 737 466 738 467 739 468 740 469 741 470 742 471 743 472 744 473 745 474 746 475 747 476 748 477 749 478 750 479 751 480 752 481 753 482 754 483 755 484 756 485 757 486 758 487 759 457 760 457 761 457 762 457 763 457 764 457 765 457 766 457 767 457 768 457 769 457 770 457 771 457 772 457 773 457 774 488 775 489 776 8 777 8 778 8 779 8 780 8 781 8 782 490 783 8 784 8 785 491 786 8 787 492 788 493 789 494 790 495 791 496 792 497 793 8 794 498 795 8 796 8 797 8 798 8 799 8 800 499 801 500 802 8 803 8 804 501 805 8 806 502 807 8 808 8 809 503 810 504 811 8 812 457 813 457 814 457 815 457 816 457 817 505 818 506 819 507 820 457 821 457 822 457 823 457 824 457 825 457 826 457 827 457 828 457 829 457 830 508 831 8 832 8 833 8 834 8 835 8 836 8 837 8 838 8 839 8 840 8 841 8 842 8 843 8 844 8 845 8 846 8 847 8 848 8 849 8 850 8 851 8 852 8 853 8 854 8 855 8 856 8 857 8 858 8 859 8 860 8 861 8 862 8 863 8 864 8 865 8 866 8 867 8 868 8 869 8 870 8 871 8 872 8 873 8 874 8 875 8 876 8 877 8 878 8 879 8 880 8 881 8 882 8 883 8 884 8 885 8 886 8 887 8 888 8 889 8 890 8 891 8 892 8 893 8 894 8 895 8 896 8 897 8 898 8 899 8 900 8 901 8 902 8 903 8 904 8 905 8 906 8 907 8 908 8 909 8 910 509 911 8 912 8 913 8 914 510 915 511 916 8 917 512 918 8 919 513 920 514 921 515 922 516 923 8 924 517 925 518 926 519 927 8 928 520 929 521 930 8 931 522 932 8 933 8 934 8 935 8 936 8 937 8 938 523 939 8 940 8 941 8 942 524 943 8 944 8 945 8 946 8 947 8 948 525 949 8 950 526 951 8 952 8 953 527 954 8 955 8 956 8 957 8 958 8 959 8 960 8 961 528 962 529 963 8 964 8 965 8 966 8 967 8 968 8 969 8 970 8 971 8 972 8 973 8 974 8 975 8 976 530 977 531 978 532 979 8 980 8 981 8 982 533 983 534 984 8 985 535 986 8 987 8 988 8 989 8 990 536 991 8 992 8 993 8 994 8 995 8 996 537 997 8 998 538 999 539 1000 8 1001 8 1002 8 1003 8 1004 540 1005 541 1006 8 1007 542 1008 543 1009 8 1010 544 1011 8 1012 545 1013 8 1014 8 1015 8 1016 8 1017 546 1018 8 1019 8 1020 8 1021 8 1022 547 1023 548 1024 549 1025 550 1026 8 1027 551 1028 552 1029 553 1030 8 1031 8 1032 554 1033 8 1034 8 1035 555 1036 556 1037 557 1038 8 1039 558 1040 8 1041 8 1042 8 1043 559 1044 8 1045 560 1046 8 1047 8 1048 8 1049 8 1050 8 1051 561 1052 562 1053 563 1054 520 1055 8 1056 523 1057 8 1058 527 1059 528 1060 529 1061 8 1062 8 1063 531 1064 532 1065 534 1066 8 1067 535 1068 536 1069 8 1070 537 1071 8 1072 539 1073 540 1074 541 1075 543 1076 8 1077 545 1078 8 1079 547 1080 555 1081 564 1082 565 1083 566 1084 567 1085 568 1086 569 1087 570 1088 571 1089 572 1090 573 1091 574 1092 575 1093 576 1094 577 1095 578 1096 579 1097 580 1098 581 1099 582 1100 583 1101 584 1102 585 1103 586 1104 587 1105 588 1106 589 1107 590 1108 591 1109 592 1110 593 1111 555 1112 555 1113 555 1114 555 1115 555 1116 555 1117 555 1118 555 1119 555 1120 555 1121 555 1122 555 1123 555 1124 555 1125 555 1126 594 1127 595 1128 8 1129 8 1130 8 1131 8 1132 8 1133 8 1134 596 1135 8 1136 8 1137 597 1138 8 1139 598 1140 599 1141 600 1142 601 1143 602 1144 603 1145 8 1146 604 1147 8 1148 8 1149 8 1150 8 1151 8 1152 605 1153 606 1154 8 1155 8 1156 607 1157 8 1158 608 1159 8 1160 8 1161 609 1162 610 1163 8 1164 555 1165 555 1166 555 1167 555 1168 555 1169 611 1170 612 1171 613 1172 555 1173 555 1174 555 1175 555 1176 555 1177 555 1178 555 1179 555 1180 555 1181 555 1182 614 1183 615 1184 616 1185 617 1186 618 1187 619 1188 620 1189 621 1190 622 1191 623 1192 8 1193 624 1194 8 1195 625 1196 8 1197 8 1198 626 1199 627 1200 628 1201 629 1202 8 1203 8 1204 630 1205 631 1206 632 1207 633 1208 634 1209 8 1210 635 1211 8 1212 636 1213 8 1214 8 1215 8 1216 8 1217 637 1218 8 1219 638 1220 639 1221 640 1222 641 1223 642 1224 643 1225 8 1226 644 1227 645 1228 8 1229 646 1230 8 1231 647 1232 648 1233 649 1234 650 1235 651 1236 652 1237 653 1238 8 1239 8 1240 654 1241 8 1242 8 1243 655 1244 656 1245 657 1246 8 1247 8 1248 8 1249 8 1250 658 1251 8 1252 8 1253 8 1254 8 1255 659 1256 660 1257 661 1258 662 1259 663 1260 664 1261 665 1262 666 1263 667 1264 668 1265 669 1266 670 1267 671 1268 672 1269 673 1270 674 1271 675 1272 676 1273 677 1274 678 1275 679 1276 680 1277 681 1278 682 1279 683 1280 684 1281 685 1282 686 1283 687 1284 688 1285 689 1286 690 1287 691 1288 692 1289 662 1290 662 1291 662 1292 662 1293 662 1294 662 1295 662 1296 662 1297 662 1298 662 1299 662 1300 662 1301 662 1302 662 1303 662 1304 693 1305 694 1306 8 1307 8 1308 8 1309 8 1310 8 1311 8 1312 695 1313 8 1314 8 1315 696 1316 8 1317 697 1318 698 1319 699 1320 700 1321 701 1322 702 1323 8 1324 703 1325 8 1326 8 1327 8 1328 8 1329 8 1330 704 1331 705 1332 8 1333 8 1334 706 1335 8 1336 707 1337 8 1338 8 1339 708 1340 709 1341 8 1342 662 1343 662 1344 662 1345 662 1346 662 1347 710 1348 711 1349 712 1350 662 1351 662 1352 662 1353 662 1354 662 1355 662 1356 662 1357 662 1358 662 1359 662 1360 713 1361 8 1362 8 1363 8 1364 8 1365 8 1366 8 1367 8 1368 8 1369 8 1370 8 1371 8 1372 8 1373 8 1374 8 1375 8 1376 8 1377 8 1378 8 1379 8 1380 8 1381 8 1382 8 1383 8 1384 8 1385 8 1386 8 1387 8 1388 8 1389 8 1390 8 1391 8 1392 8 1393 8 1394 8 1395 8 1396 8 1397 8 1398 8 1399 8 1400 8 1401 8 1402 8 1403 8 1404 8 1405 8 1406 8 1407 8 1408 8 1409 8 1410 8 1411 8 1412 8 1413 8 1414 8 1415 8 1416 8 1417 8 1418 8 1419 8 1420 8 1421 8 1422 8 1423 8 1424 8 1425 8 1426 8 1427 8 1428 8 1429 8 1430 8 1431 8 1432 8 1433 8 1434 8 1435 8 1436 8 1437 8 1438 8 1439 8 1440 714 1441 8 1442 8 1443 8 1444 715 1445 8 1446 8 1447 716 1448 717 1449 718 1450 719 1451 8 1452 720 1453 8 1454 555 1455 8 1456 721 1457 722 1458 723 1459 724 1460 8 1461 725 1462 8 1463 726 1464 727 1465 728 1466 729 1467 730 1468 731 1469 732 1470 733 1471 734 1472 662 1473 662 1474 662 1475 662 1476 662 1477 662 1478 662 1479 662 1480 662 1481 662 1482 662 1483 662 1484 662 1485 8 1486 8 1487 735 1488 662 1489 736 1490 737 1491 738 1492 739 1493 8 1494 8 1495 8 1496 8 1497 8 1498 8 1499 8 1500 8 1501 8 1502 8 1503 8 1504 8 1505 8 1506 8 1507 8 1508 740 1509 741 1510 742 1511 8 1512 8 1513 743 1514 8 1515 8 1516 8 1517 744 1518 8 1519 8 1520 8 1521 745 1522 746 1523 8 1524 8 1525 8 1526 8 1527 8 1528 8 1529 8 1530 8 1531 8 1532 8 1533 8 1534 8 1535 8 1536 8 1537 8 1538 8 1539 8 1540 8 1541 8 1542 8 1543 8 1544 8 1545 8 1546 8 1547 8 1548 8 1549 8 1550 8 1551 8 1552 8 1553 8 1554 8 1555 8 1556 8 1557 8 1558 8 1559 8 1560 8 1561 8 1562 8 1563 8 1564 8 1565 8 1566 8 1567 8 1568 8 1569 8 1570 8 1571 8 1572 8 1573 8 1574 8 1575 8 1576 8 1577 8 1578 8 1579 8 1580 8 1581 8 1582 8 1583 8 1584 8 1585 8 1586 8 1587 8 1588 8 1589 8 1590 8 1591 8 1592 8 1593 8 1594 8 1595 8 1596 8 1597 8 1598 8 1599 8 1600 8 1601 8 1602 747 1603 8 1604 8 1605 8 1606 748 1607 8 1608 8 1609 749 1610 750 1611 751 1612 8 1613 8 1614 8 1615 752 1616 8 1617 753 1618 8 1619 8 1620 8 1621 754 1622 755 1623 756 1624 757 1625 758 1626 8 1627 759 1628 8 1629 8 1630 760 1631 8 1632 761 1633 762 1634 8 1635 8 1636 763 1637 8 1638 764 1639 765 1640 766 1641 767 1642 768 1643 769 1644 770 1645 771 1646 772 1647 773 1648 8 1649 774 1650 8 1651 775 1652 8 1653 8 1654 776 1655 777 1656 778 1657 779 1658 780 1659 8 1660 8 1661 781 1662 782 1663 783 1664 784 1665 785 1666 8 1667 786 1668 8 1669 787 1670 8 1671 8 1672 8 1673 8 1674 788 1675 8 1676 789 1677 790 1678 791 1679 792 1680 793 1681 794 1682 8 1683 795 1684 796 1685 8 1686 797 1687 8 1688 798 1689 799 1690 800 1691 801 1692 802 1693 803 1694 8 1695 804 1696 805 1697 806 1698 8 1699 8 1700 807 1701 8 1702 8 1703 808 1704 809 1705 810 1706 8 1707 8 1708 8 1709 8 1710 811 1711 8 1712 812 1713 8 1714 8 1715 8 1716 8 1717 813 1718 814 1719 815 1720 816 1721 817 1722 818 1723 819 1724 820 1725 821 1726 822 1727 823 1728 824 1729 825 1730 826 1731 827 1732 828 1733 829 1734 830 1735 831 1736 832 1737 833 1738 834 1739 835 1740 836 1741 837 1742 838 1743 839 1744 840 1745 841 1746 842 1747 843 1748 844 1749 845 1750 846 1751 847 1752 817 1753 817 1754 817 1755 817 1756 817 1757 817 1758 817 1759 817 1760 817 1761 817 1762 817 1763 817 1764 817 1765 817 1766 817 1767 848 1768 849 1769 8 1770 8 1771 8 1772 8 1773 8 1774 8 1775 850 1776 8 1777 8 1778 851 1779 8 1780 852 1781 853 1782 854 1783 855 1784 856 1785 857 1786 8 1787 858 1788 8 1789 8 1790 8 1791 8 1792 8 1793 859 1794 860 1795 8 1796 8 1797 861 1798 8 1799 862 1800 8 1801 8 1802 863 1803 864 1804 8 1805 817 1806 817 1807 817 1808 817 1809 817 1810 865 1811 866 1812 867 1813 817 1814 817 1815 817 1816 817 1817 817 1818 817 1819 817 1820 817 1821 817 1822 817 1823 868 1824 869 1825 870 1826 871 1827 872 1828 873 1829 874 1830 875 1831 876 1832 877 1833 878 1834 8 1835 879 1836 8 1837 880 1838 8 1839 8 1840 881 1841 882 1842 883 1843 884 1844 885 1845 8 1846 8 1847 886 1848 887 1849 888 1850 889 1851 890 1852 8 1853 891 1854 8 1855 892 1856 8 1857 8 1858 8 1859 8 1860 893 1861 8 1862 894 1863 895 1864 896 1865 897 1866 898 1867 899 1868 8 1869 900 1870 901 1871 8 1872 902 1873 8 1874 903 1875 904 1876 905 1877 906 1878 907 1879 908 1880 8 1881 909 1882 910 1883 911 1884 8 1885 8 1886 912 1887 8 1888 8 1889 913 1890 914 1891 915 1892 8 1893 8 1894 8 1895 8 1896 916 1897 8 1898 917 1899 8 1900 8 1901 8 1902 8 1903 918 1904 919 1905 920 1906 921 1907 922 1908 923 1909 924 1910 925 1911 926 1912 927 1913 928 1914 929 1915 930 1916 931 1917 932 1918 933 1919 934 1920 935 1921 936 1922 937 1923 938 1924 939 1925 940 1926 941 1927 942 1928 943 1929 944 1930 945 1931 946 1932 947 1933 948 1934 949 1935 950 1936 951 1937 952 1938 922 1939 922 1940 922 1941 922 1942 922 1943 922 1944 922 1945 922 1946 922 1947 922 1948 922 1949 922 1950 922 1951 922 1952 922 1953 953 1954 954 1955 8 1956 8 1957 8 1958 8 1959 8 1960 8 1961 955 1962 8 1963 8 1964 956 1965 8 1966 957 1967 958 1968 959 1969 960 1970 961 1971 962 1972 8 1973 963 1974 8 1975 8 1976 8 1977 8 1978 8 1979 964 1980 965 1981 8 1982 8 1983 966 1984 8 1985 967 1986 8 1987 8 1988 968 1989 969 1990 8 1991 922 1992 922 1993 922 1994 922 1995 922 1996 970 1997 971 1998 972 1999 922 2000 922 2001 922 2002 922 2003 922 2004 922 2005 922 2006 922 2007 922 2008 922 2009 973 2010 974 2011 975 2012 976 2013 922 2014 977 2015 8 2016 8 2017 8 2018 8 2019 8 2020 8 2021 8 2022 8 2023 8 2024 8 2025 8 2026 8 2027 8 2028 8 2029 8 2030 8 2031 8 2032 8 2033 8 2034 8 2035 8 2036 8 2037 8 2038 8 2039 8 2040 8 2041 8 2042 8 2043 8 2044 8 2045 8 2046 8 2047 8 2048 8 2049 8 2050 8 2051 8 2052 8 2053 8 2054 8 2055 8 2056 8 2057 8 2058 8 2059 8 2060 8 2061 8 2062 8 2063 8 2064 8 2065 8 2066 8 2067 8 2068 8 2069 8 2070 8 2071 8 2072 8 2073 8 2074 8 2075 8 2076 8 2077 8 2078 8 2079 8 2080 8 2081 8 2082 8 2083 8 2084 8 2085 8 2086 8 2087 8 2088 8 2089 8 2090 8 2091 8 2092 8 2093 8 2094 978 2095 8 2096 8 2097 8 2098 979 2099 8 2100 8 2101 980 2102 981 2103 982 2104 8 2105 8 2106 983 2107 984 2108 985 2109 986 2110 8 2111 987 2112 8 2113 988 2114 989 2115 990 2116 991 2117 992 2118 993 2119 994 2120 995 2121 996 2122 922 2123 922 2124 922 2125 922 2126 922 2127 922 2128 922 2129 922 2130 922 2131 922 2132 922 2133 922 2134 922 2135 997 2136 922 2137 998 2138 8 2139 999 2140 1000 2141 8 2142 1001 2143 8 2144 1002 2145 8 2146 8 2147 8 2148 1003 2149 1004 2150 1005 2151 8 2152 1006 2153 8 2154 8 2155 1007 2156 1008 2157 1009 2158 8 2159 8 2160 8 2161 1010 2162 1011 2163 8 2164 8 2165 8 2166 1012 2167 1013 2168 1014 2169 1015 2170 1016 2171 8 2172 1017 2173 8 2174 8 2175 1018 2176 8 2177 1019 2178 1020 2179 8 2180 8 2181 1021 2182 8 2183 8 2184 1022 2185 1023 2186 8 2187 1024 2188 8 2189 8 2190 1025 2191 1026 2192 1027 2193 1028 2194 1029 2195 1030 2196 1031 2197 1032 2198 1033 2199 1034 2200 1035 2201 8 2202 1036 2203 1037 2204 1038 2205 1039 2206 1040 2207 1041 2208 1042 2209 1043 2210 1044 2211 1045 2212 1046 2213 8 2214 1047 2215 8 2216 8 2217 8 2218 8 2219 8 2220 8 2221 8 2222 1048 2223 8 2224 1049 2225 8 2226 8 2227 1050 2228 1051 2229 8 2230 8 2231 8 2232 8 2233 8 2234 8 2235 8 2236 8 2237 8 2238 8 2239 1052 2240 1053 2241 8 2242 8 2243 8 2244 8 2245 8 2246 8 2247 8 2248 8 2249 8 2250 8 2251 8 2252 1054 2253 8 2254 8 2255 8 2256 8 2257 8 2258 8 2259 8 2260 8 2261 8 2262 8 2263 8 2264 1055 2265 1056 2266 1057 2267 1058 2268 1059 2269 1060 2270 1061 2271 1062 2272 1063 2273 1064 2274 1065 2275 8 2276 1066 2277 8 2278 1067 2279 8 2280 8 2281 1068 2282 1069 2283 1070 2284 1071 2285 8 2286 8 2287 1072 2288 1073 2289 1074 2290 1075 2291 1076 2292 8 2293 1077 2294 8 2295 1078 2296 8 2297 8 2298 8 2299 8 2300 1079 2301 8 2302 1080 2303 1081 2304 1082 2305 1083 2306 1084 2307 1085 2308 8 2309 1086 2310 1087 2311 8 2312 1088 2313 8 2314 1089 2315 1090 2316 1091 2317 1092 2318 1093 2319 1094 2320 1095 2321 1096 2322 8 2323 8 2324 1097 2325 8 2326 8 2327 1098 2328 1099 2329 1100 2330 8 2331 8 2332 8 2333 8 2334 1101 2335 8 2336 8 2337 8 2338 8 2339 1102 2340 1103 2341 1104 2342 1105 2343 1098 2344 1106 2345 1107 2346 1108 2347 1109 2348 1110 2349 1111 2350 1112 2351 1113 2352 1114 2353 1115 2354 1116 2355 1117 2356 1118 2357 1119 2358 1120 2359 1121 2360 1122 2361 1123 2362 1124 2363 1125 2364 1126 2365 1127 2366 1128 2367 1129 2368 1130 2369 1131 2370 1132 2371 1133 2372 1134 2373 1135 2374 1098 2375 1098 2376 1098 2377 1098 2378 1098 2379 1098 2380 1098 2381 1098 2382 1098 2383 1098 2384 1098 2385 1098 2386 1098 2387 1098 2388 1098 2389 1136 2390 1137 2391 8 2392 8 2393 8 2394 8 2395 8 2396 8 2397 1138 2398 8 2399 8 2400 1139 2401 8 2402 1140 2403 1141 2404 1142 2405 1143 2406 1144 2407 1145 2408 8 2409 1146 2410 8 2411 8 2412 8 2413 8 2414 8 2415 1147 2416 1148 2417 8 2418 8 2419 1149 2420 8 2421 1150 2422 8 2423 8 2424 1151 2425 1152 2426 8 2427 1098 2428 1098 2429 1098 2430 1098 2431 1098 2432 1153 2433 1154 2434 1155 2435 1098 2436 1098 2437 1098 2438 1098 2439 1098 2440 1098 2441 1098 2442 1098 2443 1098 2444 1098 2445 1156 2446 1157 2447 1158 2448 1159 2449 1098 2450 1160 2451 8 2452 8 2453 8 2454 8 2455 8 2456 8 2457 8 2458 8 2459 8 2460 8 2461 8 2462 8 2463 8 2464 8 2465 8 2466 8 2467 8 2468 8 2469 8 2470 8 2471 8 2472 8 2473 8 2474 8 2475 8 2476 8 2477 8 2478 8 2479 8 2480 8 2481 8 2482 8 2483 8 2484 8 2485 8 2486 8 2487 8 2488 8 2489 8 2490 8 2491 8 2492 8 2493 8 2494 8 2495 8 2496 8 2497 8 2498 8 2499 8 2500 8 2501 8 2502 8 2503 8 2504 8 2505 8 2506 8 2507 8 2508 8 2509 8 2510 8 2511 8 2512 8 2513 8 2514 8 2515 8 2516 8 2517 8 2518 8 2519 8 2520 8 2521 8 2522 8 2523 8 2524 8 2525 8 2526 8 2527 8 2528 8 2529 8 2530 1161 2531 8 2532 8 2533 8 2534 1162 2535 8 2536 8 2537 1163 2538 1164 2539 1165 2540 8 2541 8 2542 1166 2543 1167 2544 1168 2545 1169 2546 8 2547 1170 2548 8 2549 1171 2550 1172 2551 1173 2552 1174 2553 1175 2554 1176 2555 1177 2556 1178 2557 1179 2558 1098 2559 1098 2560 1098 2561 1098 2562 1098 2563 1098 2564 1098 2565 1098 2566 1098 2567 1098 2568 1098 2569 1098 2570 1098 2571 1180 2572 1098 2573 1181 2574 8 2575 1182 2576 1183 2577 8 2578 1184 2579 8 2580 1185 2581 8 2582 8 2583 8 2584 1186 2585 1187 2586 1188 2587 8 2588 1189 2589 8 2590 8 2591 1190 2592 1191 2593 1192 2594 8 2595 8 2596 8 2597 1193 2598 1194 2599 8 2600 8 2601 8 2602 1195 2603 1196 2604 1197 2605 1198 2606 1199 2607 8 2608 1200 2609 8 2610 8 2611 1201 2612 8 2613 1202 2614 1203 2615 8 2616 8 2617 1204 2618 8 2619 8 2620 1205 2621 1206 2622 8 2623 8 2624 1207 2625 1208 2626 1209 2627 1210 2628 1211 2629 1212 2630 1213 2631 1214 2632 1215 2633 1216 2634 1217 2635 8 2636 1218 2637 1219 2638 1220 2639 1221 2640 1222 2641 1223 2642 1224 2643 1225 2644 1226 2645 1227 2646 8 2647 1228 2648 8 2649 1229 2650 8 2651 8 2652 8 2653 1230 2654 8 2655 8 2656 8 2657 1231 2658 1232 2659 1233 2660 1234 2661 1235 2662 8 2663 1236 2664 8 2665 8 2666 1237 2667 8 2668 1238 2669 1239 2670 8 2671 8 2672 1240 2673 8 2674 1241 2675 8 2676 1242 2677 1243 2678 8 2679 1244 2680 8 2681 1245 2682 8 2683 8 2684 8 2685 1246 2686 1247 2687 1248 2688 8 2689 1249 2690 8 2691 8 2692 1250 2693 1251 2694 1252 2695 8 2696 8 2697 8 2698 1253 2699 1254 2700 8 2701 8 2702 8 2703 1255 2704 1256 2705 1257 2706 1258 2707 1259 2708 8 2709 1260 2710 8 2711 8 2712 1261 2713 8 2714 1262 2715 1263 2716 8 2717 8 2718 1264 2719 8 2720 8 2721 1265 2722 1266 2723 8 2724 1267 2725 1268 2726 1269 2727 1270 2728 1271 2729 1272 2730 8 2731 8 2732 8 2733 8 2734 1273 2735 1274 2736 1275 2737 1276 2738 8 2739 8 2740 8 2741 1277 2742 8 2743 8 2744 8 2745 8 2746 8 2747 1278 2748 1279 2749 8 2750 1280 2751 8 2752 1281 2753 1282 2754 8 2755 1283 2756 8 2757 1284 2758 1285 2759 8 2760 1286 2761 8 2762 1287 2763 8 2764 8 2765 1288 2766 1289 2767 1290 2768 1291 2769 8 2770 1292 2771 8 2772 8 2773 1293 2774 1294 2775 1295 2776 8 2777 8 2778 8 2779 8 2780 1296 2781 1297 2782 8 2783 8 2784 8 2785 1298 2786 1299 2787 1300 2788 1301 2789 1302 2790 8 2791 1303 2792 8 2793 8 2794 1304 2795 8 2796 1305 2797 1306 2798 8 2799 8 2800 1307 2801 8 2802 8 2803 1308 2804 1309 2805 8 2806 1310 2807 1311 2808 1312 2809 1098 2810 1098 2811 1313 2812 1314 2813 1098 2814 1098 2815 1098 2816 1098 2817 8 2818 1315 2819 1316 2820 1317 2821 8 2822 1318 2823 1319 2824 8 2825 8 2826 8 2827 8 2828 1320 2829 1321 2830 1322 2831 1323 2832 8 2833 8 2834 8 2835 8 2836 8 2837 1324 2838 1325 2839 8 2840 1326 2841 8 2842 1327 2843 1328 2844 8 2845 1329 2846 8 2847 1098 2848 1098 2849 1098 2850 1098 2851 1330 2852 1331 2853 1332 2854 1333 2855 8 2856 8 2857 8 2858 8 2859 8 2860 8 2861 8 2862 8 2863 1334 2864 1335 2865 1336 2866 8 2867 1337 2868 8 2869 1338 2870 1339 2871 1340 2872 1341 2873 1342 2874 1343 2875 8 2876 1344 2877 8 2878 8 2879 8 2880 8 2881 8 2882 1345 2883 1346 2884 8 2885 8 2886 1347 2887 8 2888 8 2889 8 2890 1348 2891 1349 2892 1350 2893 8 2894 8 2895 1351 2896 8 2897 8 2898 8 2899 1352 2900 8 2901 1353 2902 1354 2903 1355 2904 1356 2905 1357 2906 1358 2907 1359 2908 1360 2909 8 2910 1361 2911 1362 2912 8 2913 8 2914 8 2915 8 2916 8 2917 8 2918 1363 2919 8 2920 8 2921 8 2922 1364 2923 8 2924 8 2925 8 2926 1365 2927 1366 2928 1367 2929 8 2930 1368 2931 8 2932 8 2933 8 2934 8 2935 8 2936 8 2937 1369 2938 8 2939 1370 2940 8 2941 1371 2942 1372 2943 1373 2944 8 2945 8 2946 8 2947 8 2948 8 2949 8 2950 8 2951 8 2952 1374 2953 1375 2954 8 2955 8 2956 8 2957 8 2958 8 2959 8 2960 8 2961 8 2962 8 2963 1376 2964 8 2965 1377 2966 1378 2967 8 2968 8 2969 1379 2970 8 2971 1380 2972 1381 2973 1382 2974 1383 2975 1384 2976 1385 2977 1386 2978 1387 2979 1388 2980 1389 2981 8 2982 1390 2983 8 2984 1391 2985 8 2986 8 2987 1392 2988 1393 2989 1394 2990 1395 2991 8 2992 8 2993 1396 2994 1397 2995 1398 2996 1399 2997 1400 2998 8 2999 1401 3000 8 3001 1402 3002 8 3003 8 3004 8 3005 8 3006 1403 3007 8 3008 1404 3009 1405 3010 1406 3011 1407 3012 1408 3013 1409 3014 8 3015 1410 3016 1411 3017 8 3018 1412 3019 8 3020 1413 3021 1414 3022 1415 3023 1416 3024 1417 3025 1418 3026 1419 3027 8 3028 8 3029 1420 3030 8 3031 8 3032 1421 3033 1422 3034 1423 3035 8 3036 8 3037 8 3038 8 3039 1424 3040 8 3041 8 3042 8 3043 8 3044 1425 3045 1426 3046 1427 3047 1428 3048 1429 3049 1430 3050 1431 3051 1432 3052 1433 3053 1434 3054 1435 3055 1436 3056 1437 3057 1438 3058 1439 3059 1440 3060 1441 3061 1442 3062 1443 3063 1444 3064 1445 3065 1446 3066 1447 3067 1448 3068 1449 3069 1450 3070 1451 3071 1452 3072 1453 3073 1454 3074 1455 3075 1456 3076 1457 3077 1458 3078 1428 3079 1428 3080 1428 3081 1428 3082 1428 3083 1428 3084 1428 3085 1428 3086 1428 3087 1428 3088 1428 3089 1428 3090 1428 3091 1428 3092 1428 3093 1459 3094 1460 3095 8 3096 8 3097 8 3098 8 3099 8 3100 8 3101 1461 3102 8 3103 8 3104 1462 3105 8 3106 1463 3107 1464 3108 1465 3109 1466 3110 1467 3111 1468 3112 8 3113 1469 3114 8 3115 8 3116 8 3117 8 3118 8 3119 1470 3120 1471 3121 8 3122 8 3123 1472 3124 8 3125 1473 3126 8 3127 8 3128 1474 3129 1475 3130 8 3131 1428 3132 1428 3133 1428 3134 1428 3135 1428 3136 1476 3137 1477 3138 1478 3139 1428 3140 1428 3141 1428 3142 1428 3143 1428 3144 1428 3145 1428 3146 1428 3147 1428 3148 1428 3149 1479 3150 8 3151 8 3152 8 3153 8 3154 8 3155 8 3156 8 3157 8 3158 8 3159 8 3160 8 3161 8 3162 8 3163 8 3164 8 3165 8 3166 8 3167 8 3168 8 3169 8 3170 8 3171 8 3172 8 3173 8 3174 8 3175 8 3176 8 3177 8 3178 8 3179 8 3180 8 3181 8 3182 8 3183 8 3184 8 3185 8 3186 8 3187 8 3188 8 3189 8 3190 8 3191 8 3192 8 3193 8 3194 8 3195 8 3196 8 3197 8 3198 8 3199 8 3200 8 3201 8 3202 8 3203 8 3204 8 3205 8 3206 8 3207 8 3208 8 3209 8 3210 8 3211 8 3212 8 3213 8 3214 8 3215 8 3216 8 3217 8 3218 8 3219 8 3220 8 3221 8 3222 8 3223 8 3224 8 3225 8 3226 8 3227 8 3228 8 3229 1480 3230 8 3231 8 3232 8 3233 1481 3234 8 3235 8 3236 1482 3237 1483 3238 1484 3239 8 3240 8 3241 1485 3242 1486 3243 1487 3244 1488 3245 8 3246 1489 3247 8 3248 1490 3249 1491 3250 1492 3251 1493 3252 1494 3253 1495 3254 1496 3255 1497 3256 1498 3257 1428 3258 1428 3259 1428 3260 1428 3261 1428 3262 1428 3263 1428 3264 1428 3265 1428 3266 1428 3267 1428 3268 1428 3269 1428 3270 1499 3271 1428 3272 1500 3273 8 3274 1501 3275 1502 3276 8 3277 1503 3278 8 3279 1504 3280 8 3281 8 3282 8 3283 1505 3284 1506 3285 1507 3286 8 3287 1508 3288 8 3289 8 3290 1509 3291 1510 3292 1511 3293 8 3294 8 3295 8 3296 1512 3297 1513 3298 8 3299 8 3300 8 3301 1514 3302 1515 3303 1516 3304 1517 3305 1518 3306 8 3307 1519 3308 8 3309 8 3310 1520 3311 8 3312 1521 3313 1522 3314 8 3315 8 3316 1523 3317 8 3318 8 3319 1524 3320 1525 3321 8 3322 8 3323 1526 3324 1527 3325 1528 3326 1529 3327 1530 3328 1531 3329 1532 3330 1533 3331 1534 3332 1535 3333 1536 3334 8 3335 1537 3336 1538 3337 1539 3338 1540 3339 1541 3340 1542 3341 1543 3342 1544 3343 1545 3344 1546 3345 8 3346 8 3347 1547 3348 8 3349 1548 3350 8 3351 8 3352 8 3353 1549 3354 8 3355 8 3356 1550 3357 1551 3358 1552 3359 1553 3360 1554 3361 8 3362 1555 3363 8 3364 8 3365 1556 3366 8 3367 1557 3368 1558 3369 8 3370 8 3371 1559 3372 8 3373 1560 3374 8 3375 1561 3376 1562 3377 1563 3378 1428 3379 1428 3380 1564 3381 1565 3382 1566 3383 1567 3384 1568 3385 1569 3386 1570 3387 1571 3388 1572 3389 1573 3390 8 3391 1574 3392 8 3393 1575 3394 8 3395 8 3396 1576 3397 1577 3398 1578 3399 1579 3400 8 3401 8 3402 1580 3403 1581 3404 1582 3405 1583 3406 1584 3407 8 3408 1585 3409 8 3410 1586 3411 8 3412 8 3413 8 3414 8 3415 1587 3416 8 3417 1588 3418 1589 3419 1590 3420 1591 3421 1592 3422 1593 3423 8 3424 1594 3425 1595 3426 8 3427 1596 3428 8 3429 1597 3430 1598 3431 1599 3432 1600 3433 1601 3434 1602 3435 1603 3436 8 3437 8 3438 1604 3439 8 3440 8 3441 1605 3442 1606 3443 1607 3444 8 3445 8 3446 8 3447 8 3448 1608 3449 8 3450 8 3451 8 3452 8 3453 1609 3454 1610 3455 1611 3456 1612 3457 1613 3458 1614 3459 1615 3460 1616 3461 1617 3462 1618 3463 1619 3464 1620 3465 1621 3466 1622 3467 1623 3468 1624 3469 1625 3470 1626 3471 1627 3472 1628 3473 1629 3474 1630 3475 1631 3476 1632 3477 1633 3478 1634 3479 1635 3480 1636 3481 1637 3482 1638 3483 1639 3484 1640 3485 1641 3486 1642 3487 1612 3488 1612 3489 1612 3490 1612 3491 1612 3492 1612 3493 1612 3494 1612 3495 1612 3496 1612 3497 1612 3498 1612 3499 1612 3500 1612 3501 1612 3502 1643 3503 1644 3504 8 3505 8 3506 8 3507 8 3508 8 3509 8 3510 1645 3511 8 3512 8 3513 1646 3514 8 3515 1647 3516 1648 3517 1649 3518 1650 3519 1651 3520 1652 3521 8 3522 1653 3523 8 3524 8 3525 8 3526 8 3527 8 3528 1654 3529 1655 3530 8 3531 8 3532 1656 3533 8 3534 1657 3535 8 3536 8 3537 1658 3538 1659 3539 8 3540 1612 3541 1612 3542 1612 3543 1612 3544 1612 3545 1660 3546 1661 3547 1662 3548 1612 3549 1612 3550 1612 3551 1612 3552 1612 3553 1612 3554 1612 3555 1612 3556 1612 3557 1612 3558 1663 3559 8 3560 8 3561 8 3562 8 3563 8 3564 8 3565 8 3566 8 3567 8 3568 8 3569 8 3570 8 3571 8 3572 8 3573 8 3574 8 3575 8 3576 8 3577 8 3578 8 3579 8 3580 8 3581 8 3582 8 3583 8 3584 8 3585 8 3586 8 3587 8 3588 8 3589 8 3590 8 3591 8 3592 8 3593 8 3594 8 3595 8 3596 8 3597 8 3598 8 3599 8 3600 8 3601 8 3602 8 3603 8 3604 8 3605 8 3606 8 3607 8 3608 8 3609 8 3610 8 3611 8 3612 8 3613 8 3614 8 3615 8 3616 8 3617 8 3618 8 3619 8 3620 8 3621 8 3622 8 3623 8 3624 8 3625 8 3626 8 3627 8 3628 8 3629 8 3630 8 3631 8 3632 8 3633 8 3634 8 3635 8 3636 8 3637 8 3638 1664 3639 8 3640 8 3641 8 3642 1665 3643 8 3644 8 3645 1666 3646 1667 3647 1668 3648 8 3649 8 3650 1669 3651 1670 3652 1671 3653 1672 3654 8 3655 1673 3656 8 3657 1674 3658 1675 3659 1676 3660 1677 3661 1678 3662 1679 3663 1680 3664 1681 3665 1682 3666 1612 3667 1612 3668 1612 3669 1612 3670 1612 3671 1612 3672 1612 3673 1612 3674 1612 3675 1612 3676 1612 3677 1612 3678 1612 3679 1683 3680 1612 3681 1684 3682 8 3683 1685 3684 1686 3685 8 3686 1687 3687 8 3688 1688 3689 8 3690 8 3691 8 3692 1689 3693 1690 3694 1691 3695 8 3696 1692 3697 8 3698 8 3699 1693 3700 1694 3701 1695 3702 8 3703 8 3704 8 3705 1696 3706 1697 3707 8 3708 8 3709 8 3710 1698 3711 1699 3712 1700 3713 1701 3714 1702 3715 8 3716 1703 3717 8 3718 8 3719 1704 3720 8 3721 1705 3722 1706 3723 8 3724 8 3725 1707 3726 8 3727 8 3728 1708 3729 1709 3730 8 3731 8 3732 1710 3733 1711 3734 1712 3735 1713 3736 1714 3737 1715 3738 1716 3739 1717 3740 1718 3741 1719 3742 1720 3743 8 3744 1721 3745 1722 3746 1723 3747 1724 3748 1725 3749 1726 3750 1727 3751 1728 3752 1729 3753 1730 3754 8 3755 1731 3756 1428 3757 1732 3758 8 3759 1733 3760 8 3761 8 3762 8 3763 1734 3764 8 3765 8 3766 1735 3767 1736 3768 1737 3769 1738 3770 1739 3771 8 3772 1740 3773 8 3774 8 3775 1741 3776 8 3777 1742 3778 1743 3779 8 3780 8 3781 1744 3782 8 3783 1745 3784 8 3785 1746 3786 1747 3787 1748 3788 1612 3789 1612 3790 1749 3791 1750 3792 1612 3793 1612 3794 1612 3795 1612 3796 8 3797 1751 3798 8 3799 1752 3800 8 3801 8 3802 1612 3803 1612 3804 1612 3805 1612 3806 1753 3807 1754 3808 8 3809 8 3810 8 3811 8 3812 8 3813 8 3814 1755 3815 1756 3816 1757 3817 8 3818 1758 3819 8 3820 1759 3821 1760 3822 1761 3823 1762 3824 1763 3825 1764 3826 8 3827 1765 3828 8 3829 8 3830 8 3831 8 3832 1766 3833 8 3834 8 3835 8 3836 1767 3837 1768 3838 1769 3839 8 3840 8 3841 1770 3842 8 3843 8 3844 1771 3845 1772 3846 1773 3847 1774 3848 1775 3849 1776 3850 8 3851 1777 3852 8 3853 1778 3854 1779 3855 1780 3856 1781 3857 1782 3858 1783 3859 1784 3860 1785 3861 1786 3862 1787 3863 1788 3864 1789 3865 1790 3866 1791 3867 8 3868 1792 3869 8 3870 1793 3871 1794 3872 8 3873 8 3874 1795 3875 1796 3876 1797 3877 1798 3878 8 3879 8 3880 1799 3881 1800 3882 1801 3883 1802 3884 1803 3885 8 3886 1804 3887 8 3888 1805 3889 8 3890 8 3891 8 3892 8 3893 1806 3894 8 3895 1807 3896 1808 3897 1809 3898 1810 3899 1811 3900 1812 3901 8 3902 1813 3903 1814 3904 8 3905 1815 3906 8 3907 1816 3908 1817 3909 1818 3910 1819 3911 1820 3912 1821 3913 1822 3914 8 3915 8 3916 1823 3917 8 3918 8 3919 1824 3920 1825 3921 1826 3922 8 3923 8 3924 8 3925 8 3926 8 3927 1827 3928 1828 3929 1829 3930 8 3931 1830 3932 8 3933 8 3934 8 3935 1831 3936 1832 3937 1833 3938 1824 3939 1834 3940 1835 3941 1836 3942 1837 3943 1838 3944 1839 3945 1840 3946 1841 3947 1842 3948 1843 3949 1844 3950 1845 3951 1846 3952 1847 3953 1848 3954 1849 3955 1850 3956 1851 3957 1852 3958 1853 3959 1854 3960 1855 3961 1856 3962 1857 3963 1858 3964 1859 3965 1860 3966 1861 3967 1862 3968 1863 3969 1824 3970 1824 3971 1824 3972 1824 3973 1824 3974 1824 3975 1824 3976 1824 3977 1824 3978 1824 3979 1824 3980 1824 3981 1824 3982 1824 3983 1824 3984 1864 3985 1865 3986 8 3987 8 3988 8 3989 8 3990 8 3991 8 3992 1866 3993 8 3994 8 3995 1867 3996 8 3997 1868 3998 1869 3999 1870 4000 1871 4001 1872 4002 1873 4003 8 4004 1874 4005 8 4006 8 4007 8 4008 8 4009 8 4010 1875 4011 1876 4012 8 4013 8 4014 1877 4015 8 4016 1878 4017 8 4018 8 4019 1879 4020 1880 4021 8 4022 1824 4023 1824 4024 1824 4025 1824 4026 1824 4027 1881 4028 1882 4029 1883 4030 1824 4031 1824 4032 1824 4033 1824 4034 1824 4035 1824 4036 1824 4037 1824 4038 1824 4039 1824 4040 1884 4041 8 4042 8 4043 8 4044 8 4045 8 4046 8 4047 8 4048 8 4049 8 4050 8 4051 8 4052 8 4053 8 4054 8 4055 8 4056 8 4057 8 4058 8 4059 8 4060 8 4061 8 4062 8 4063 8 4064 8 4065 8 4066 8 4067 8 4068 8 4069 8 4070 8 4071 8 4072 8 4073 8 4074 8 4075 8 4076 8 4077 8 4078 8 4079 8 4080 8 4081 8 4082 8 4083 8 4084 8 4085 8 4086 8 4087 8 4088 8 4089 8 4090 8 4091 8 4092 8 4093 8 4094 8 4095 8 4096 8 4097 8 4098 8 4099 8 4100 8 4101 8 4102 8 4103 8 4104 8 4105 8 4106 8 4107 8 4108 8 4109 8 4110 8 4111 8 4112 8 4113 8 4114 8 4115 8 4116 8 4117 8 4118 8 4119 8 4120 1885 4121 8 4122 8 4123 8 4124 1886 4125 8 4126 8 4127 1887 4128 1888 4129 1889 4130 8 4131 8 4132 1890 4133 1891 4134 1892 4135 1893 4136 8 4137 1894 4138 8 4139 1895 4140 1896 4141 1897 4142 1898 4143 1899 4144 1900 4145 1901 4146 1902 4147 1903 4148 1824 4149 1824 4150 1824 4151 1824 4152 1824 4153 1824 4154 1824 4155 1824 4156 1824 4157 1824 4158 1824 4159 1824 4160 1824 4161 1904 4162 1824 4163 1905 4164 8 4165 1906 4166 1907 4167 8 4168 1908 4169 8 4170 1909 4171 8 4172 8 4173 8 4174 1910 4175 1911 4176 1912 4177 8 4178 1913 4179 8 4180 8 4181 1914 4182 1915 4183 1916 4184 8 4185 8 4186 8 4187 1917 4188 1918 4189 8 4190 8 4191 8 4192 1919 4193 1920 4194 1921 4195 1922 4196 1923 4197 8 4198 1924 4199 8 4200 8 4201 1925 4202 8 4203 1926 4204 1927 4205 8 4206 8 4207 1928 4208 8 4209 8 4210 1929 4211 1930 4212 8 4213 1931 4214 8 4215 8 4216 8 4217 8 4218 1932 4219 1933 4220 1934 4221 1935 4222 1936 4223 1937 4224 1938 4225 1939 4226 1940 4227 1941 4228 1942 4229 1943 4230 1944 4231 8 4232 1945 4233 8 4234 1946 4235 8 4236 8 4237 1947 4238 1948 4239 1949 4240 1950 4241 8 4242 8 4243 1951 4244 1952 4245 1953 4246 1954 4247 1955 4248 8 4249 1956 4250 8 4251 1957 4252 8 4253 8 4254 8 4255 8 4256 1958 4257 8 4258 1959 4259 1960 4260 1961 4261 1962 4262 1963 4263 1964 4264 8 4265 1965 4266 1966 4267 8 4268 1967 4269 8 4270 1968 4271 1969 4272 1970 4273 1971 4274 1972 4275 1973 4276 1974 4277 8 4278 8 4279 1975 4280 8 4281 8 4282 1976 4283 1977 4284 1978 4285 8 4286 8 4287 8 4288 8 4289 1979 4290 8 4291 8 4292 8 4293 8 4294 1980 4295 1981 4296 1982 4297 1976 4298 1983 4299 1984 4300 1985 4301 1986 4302 1987 4303 1988 4304 1989 4305 1990 4306 1991 4307 1992 4308 1993 4309 1994 4310 1995 4311 1996 4312 1997 4313 1998 4314 1999 4315 2000 4316 2001 4317 2002 4318 2003 4319 2004 4320 2005 4321 2006 4322 2007 4323 2008 4324 2009 4325 2010 4326 2011 4327 2012 4328 1976 4329 1976 4330 1976 4331 1976 4332 1976 4333 1976 4334 1976 4335 1976 4336 1976 4337 1976 4338 1976 4339 1976 4340 1976 4341 1976 4342 1976 4343 2013 4344 2014 4345 8 4346 8 4347 8 4348 8 4349 8 4350 8 4351 2015 4352 8 4353 8 4354 2016 4355 8 4356 2017 4357 2018 4358 2019 4359 2020 4360 2021 4361 2022 4362 8 4363 2023 4364 8 4365 8 4366 8 4367 8 4368 8 4369 2024 4370 2025 4371 8 4372 8 4373 2026 4374 8 4375 2027 4376 8 4377 8 4378 2028 4379 2029 4380 8 4381 1976 4382 1976 4383 1976 4384 1976 4385 1976 4386 2030 4387 2031 4388 2032 4389 1976 4390 1976 4391 1976 4392 1976 4393 1976 4394 1976 4395 1976 4396 1976 4397 1976 4398 1976 4399 2033 4400 8 4401 8 4402 8 4403 8 4404 8 4405 8 4406 8 4407 8 4408 8 4409 8 4410 8 4411 8 4412 8 4413 8 4414 8 4415 8 4416 8 4417 8 4418 8 4419 8 4420 8 4421 8 4422 8 4423 8 4424 8 4425 8 4426 8 4427 8 4428 8 4429 8 4430 8 4431 8 4432 8 4433 8 4434 8 4435 8 4436 8 4437 8 4438 8 4439 8 4440 8 4441 8 4442 8 4443 8 4444 8 4445 8 4446 8 4447 8 4448 8 4449 8 4450 8 4451 8 4452 8 4453 8 4454 8 4455 8 4456 8 4457 8 4458 8 4459 8 4460 8 4461 8 4462 8 4463 8 4464 8 4465 8 4466 8 4467 8 4468 8 4469 8 4470 8 4471 8 4472 8 4473 8 4474 8 4475 8 4476 8 4477 8 4478 8 4479 2034 4480 8 4481 8 4482 8 4483 2035 4484 8 4485 8 4486 2036 4487 2037 4488 2038 4489 8 4490 8 4491 2039 4492 2040 4493 2041 4494 2042 4495 8 4496 2043 4497 8 4498 2044 4499 2045 4500 2046 4501 2047 4502 2048 4503 2049 4504 2050 4505 2051 4506 2052 4507 1976 4508 1976 4509 1976 4510 1976 4511 1976 4512 1976 4513 1976 4514 1976 4515 1976 4516 1976 4517 1976 4518 1976 4519 1976 4520 2053 4521 1976 4522 2054 4523 8 4524 2055 4525 2056 4526 8 4527 2057 4528 8 4529 2058 4530 8 4531 8 4532 8 4533 2059 4534 2060 4535 2061 4536 8 4537 2062 4538 8 4539 8 4540 2063 4541 2064 4542 2065 4543 8 4544 8 4545 8 4546 2066 4547 2067 4548 8 4549 8 4550 8 4551 2068 4552 2069 4553 2070 4554 2071 4555 2072 4556 8 4557 2073 4558 8 4559 8 4560 2074 4561 8 4562 2075 4563 2076 4564 8 4565 8 4566 2077 4567 8 4568 8 4569 2078 4570 2079 4571 8 4572 8 4573 2080 4574 2081 4575 2082 4576 2083 4577 2084 4578 2085 4579 2086 4580 2087 4581 2088 4582 2089 4583 2090 4584 8 4585 2091 4586 2092 4587 2093 4588 2094 4589 2095 4590 2096 4591 2097 4592 2098 4593 2099 4594 2100 4595 8 4596 2101 4597 8 4598 2102 4599 8 4600 8 4601 8 4602 2103 4603 8 4604 8 4605 2104 4606 2105 4607 2106 4608 2107 4609 2108 4610 8 4611 2109 4612 8 4613 8 4614 2110 4615 8 4616 2111 4617 2112 4618 8 4619 8 4620 2113 4621 8 4622 2114 4623 8 4624 2115 4625 2116 4626 2117 4627 1976 4628 1976 4629 2118 4630 2119 4631 2120 4632 2121 4633 2122 4634 2123 4635 2124 4636 2125 4637 2126 4638 2127 4639 8 4640 2128 4641 8 4642 2129 4643 8 4644 8 4645 2130 4646 2131 4647 2132 4648 2133 4649 8 4650 8 4651 2134 4652 2135 4653 2136 4654 2137 4655 2138 4656 8 4657 2139 4658 8 4659 2140 4660 8 4661 8 4662 8 4663 8 4664 2141 4665 8 4666 2142 4667 2143 4668 2144 4669 2145 4670 2146 4671 2147 4672 8 4673 2148 4674 2149 4675 8 4676 2150 4677 8 4678 2151 4679 2152 4680 2153 4681 2154 4682 2155 4683 2156 4684 2157 4685 8 4686 8 4687 2158 4688 8 4689 8 4690 2159 4691 2160 4692 2161 4693 8 4694 8 4695 8 4696 8 4697 2162 4698 8 4699 8 4700 8 4701 8 4702 2163 4703 2164 4704 2165 4705 2166 4706 2167 4707 2168 4708 2169 4709 2170 4710 2171 4711 2172 4712 2173 4713 2174 4714 2175 4715 2176 4716 2177 4717 2178 4718 2179 4719 2180 4720 2181 4721 2182 4722 2183 4723 2184 4724 2185 4725 2186 4726 2187 4727 2188 4728 2189 4729 2190 4730 2191 4731 2192 4732 2193 4733 2194 4734 2195 4735 2196 4736 2166 4737 2166 4738 2166 4739 2166 4740 2166 4741 2166 4742 2166 4743 2166 4744 2166 4745 2166 4746 2166 4747 2166 4748 2166 4749 2166 4750 2166 4751 2197 4752 2198 4753 8 4754 8 4755 8 4756 8 4757 8 4758 8 4759 2199 4760 8 4761 8 4762 2200 4763 8 4764 2201 4765 2202 4766 2203 4767 2204 4768 2205 4769 2206 4770 8 4771 2207 4772 8 4773 8 4774 8 4775 8 4776 8 4777 2208 4778 2209 4779 8 4780 8 4781 2210 4782 8 4783 2211 4784 8 4785 8 4786 2212 4787 2213 4788 8 4789 2166 4790 2166 4791 2166 4792 2166 4793 2166 4794 2214 4795 2215 4796 2216 4797 2166 4798 2166 4799 2166 4800 2166 4801 2166 4802 2166 4803 2166 4804 2166 4805 2166 4806 2166 4807 2217 4808 8 4809 8 4810 8 4811 8 4812 8 4813 8 4814 8 4815 8 4816 8 4817 8 4818 8 4819 8 4820 8 4821 8 4822 8 4823 8 4824 8 4825 8 4826 8 4827 8 4828 8 4829 8 4830 8 4831 8 4832 8 4833 8 4834 8 4835 8 4836 8 4837 8 4838 8 4839 8 4840 8 4841 8 4842 8 4843 8 4844 8 4845 8 4846 8 4847 8 4848 8 4849 8 4850 8 4851 8 4852 8 4853 8 4854 8 4855 8 4856 8 4857 8 4858 8 4859 8 4860 8 4861 8 4862 8 4863 8 4864 8 4865 8 4866 8 4867 8 4868 8 4869 8 4870 8 4871 8 4872 8 4873 8 4874 8 4875 8 4876 8 4877 8 4878 8 4879 8 4880 8 4881 8 4882 8 4883 8 4884 8 4885 8 4886 8 4887 2218 4888 8 4889 8 4890 8 4891 2219 4892 8 4893 8 4894 2220 4895 2221 4896 2222 4897 8 4898 8 4899 2223 4900 2224 4901 2225 4902 2226 4903 8 4904 2227 4905 8 4906 2228 4907 2229 4908 2230 4909 2231 4910 2232 4911 2233 4912 2234 4913 2235 4914 2236 4915 2166 4916 2166 4917 2166 4918 2166 4919 2166 4920 2166 4921 2166 4922 2166 4923 2166 4924 2166 4925 2166 4926 2166 4927 2166 4928 2237 4929 2166 4930 2238 4931 8 4932 2239 4933 2240 4934 8 4935 2241 4936 8 4937 2242 4938 8 4939 8 4940 8 4941 2243 4942 2244 4943 2245 4944 8 4945 2246 4946 8 4947 8 4948 2247 4949 2248 4950 2249 4951 8 4952 8 4953 8 4954 2250 4955 2251 4956 8 4957 8 4958 8 4959 2252 4960 2253 4961 2254 4962 2255 4963 2256 4964 8 4965 2257 4966 8 4967 8 4968 2258 4969 8 4970 2259 4971 2260 4972 8 4973 8 4974 2261 4975 8 4976 8 4977 2262 4978 2263 4979 8 4980 8 4981 2264 4982 2265 4983 2266 4984 2267 4985 2268 4986 2269 4987 2270 4988 2271 4989 2272 4990 2273 4991 2274 4992 8 4993 2275 4994 2276 4995 2277 4996 2278 4997 2279 4998 2280 4999 2281 5000 2282 5001 2283 5002 2284 5003 8 5004 2285 5005 1976 5006 2286 5007 8 5008 2287 5009 8 5010 8 5011 8 5012 2288 5013 8 5014 8 5015 2289 5016 2290 5017 2291 5018 2292 5019 2293 5020 8 5021 2294 5022 8 5023 8 5024 2295 5025 8 5026 2296 5027 2297 5028 8 5029 8 5030 2298 5031 8 5032 2299 5033 8 5034 2300 5035 2301 5036 2302 5037 2166 5038 2166 5039 2303 5040 2304 5041 2166 5042 2166 5043 2166 5044 2166 5045 8 5046 2305 5047 8 5048 2306 5049 8 5050 8 5051 2166 5052 2166 5053 2166 5054 2166 5055 2307 5056 2308 5057 8 5058 8 5059 8 5060 8 5061 8 5062 8 5063 2309 5064 2310 5065 2311 5066 8 5067 2312 5068 8 5069 2313 5070 2314 5071 2315 5072 2316 5073 2317 5074 2318 5075 8 5076 2319 5077 8 5078 8 5079 8 5080 8 5081 2320 5082 8 5083 8 5084 8 5085 2321 5086 2322 5087 2323 5088 8 5089 8 5090 2324 5091 8 5092 8 5093 2325 5094 2326 5095 2327 5096 2328 5097 2329 5098 2330 5099 8 5100 2331 5101 8 5102 2332 5103 2333 5104 2334 5105 2335 5106 2336 5107 8 5108 2337 5109 2338 5110 2339 5111 2340 5112 2166 5113 2341 5114 555 5115 8 5116 2342 5117 2343 5118 2344 5119 8 5120 8 5121 2166 5122 2166 5123 2166 5124 8 5125 2345 5126 8 5127 183 5128 2346 5129 8 5130 2347 5131 2348 5132 8 5133 2349 5134 2350 5135 8 5136 2166 5137 8 5138 2166 5139 2166 5140 2166 5141 2351 5142 2352 5143 8 5144 2166 5145 2166 5146 2166 5147 2353 5148 8 5149 8 5150 2354 5151 2355 5152 8 5153 2356 5154 2357 5155 2358 5156 2359 5157 2360 5158 8 5159 8 5160 8 5161 8 5162 2361 5163 2362 5164 2363 5165 2364 5166 8 5167 8 5168 8 5169 8 5170 2365 5171 2366 5172 8 5173 2367 5174 8 5175 2368 5176 2369 5177 8 5178 2370 5179 8 5180 2166 5181 1976 5182 1976 5183 1976 5184 8 5185 2371 5186 8 5187 2372 5188 8 5189 8 5190 1976 5191 1976 5192 1976 5193 1976 5194 2373 5195 2374 5196 8 5197 8 5198 8 5199 8 5200 8 5201 8 5202 2375 5203 2376 5204 2377 5205 8 5206 2378 5207 8 5208 2379 5209 2380 5210 2381 5211 2382 5212 2383 5213 2384 5214 8 5215 2385 5216 8 5217 8 5218 8 5219 8 5220 2386 5221 8 5222 8 5223 8 5224 2387 5225 2388 5226 2389 5227 8 5228 8 5229 2390 5230 8 5231 8 5232 2391 5233 2392 5234 2393 5235 2394 5236 2395 5237 2396 5238 8 5239 2397 5240 8 5241 2398 5242 2399 5243 2400 5244 2401 5245 2402 5246 8 5247 2403 5248 2404 5249 2405 5250 2406 5251 1976 5252 2407 5253 2408 5254 2409 5255 8 5256 1976 5257 1976 5258 1976 5259 8 5260 2410 5261 8 5262 2411 5263 8 5264 2412 5265 2413 5266 8 5267 2414 5268 2415 5269 8 5270 1976 5271 8 5272 1976 5273 1976 5274 1976 5275 2416 5276 2417 5277 8 5278 1976 5279 1976 5280 1976 5281 2418 5282 8 5283 8 5284 2419 5285 2420 5286 8 5287 2421 5288 8 5289 1976 5290 1824 5291 2422 5292 2423 5293 2424 5294 2425 5295 2426 5296 2427 5297 2428 5298 2429 5299 2430 5300 2431 5301 2432 5302 8 5303 2433 5304 2434 5305 2435 5306 2436 5307 2437 5308 2438 5309 2439 5310 2440 5311 2441 5312 2442 5313 8 5314 2443 5315 8 5316 2444 5317 8 5318 8 5319 8 5320 2445 5321 8 5322 8 5323 2446 5324 2447 5325 2448 5326 2449 5327 2450 5328 8 5329 2451 5330 8 5331 8 5332 2452 5333 8 5334 2453 5335 2454 5336 8 5337 8 5338 2455 5339 8 5340 2456 5341 8 5342 2457 5343 2458 5344 2459 5345 1824 5346 1824 5347 2460 5348 2461 5349 1824 5350 1824 5351 1824 5352 1824 5353 8 5354 2462 5355 8 5356 2463 5357 8 5358 8 5359 1824 5360 1824 5361 1824 5362 1824 5363 2464 5364 2465 5365 8 5366 8 5367 8 5368 8 5369 8 5370 8 5371 2466 5372 2467 5373 2468 5374 8 5375 2469 5376 8 5377 2470 5378 2471 5379 2472 5380 2473 5381 2474 5382 2475 5383 8 5384 2476 5385 8 5386 8 5387 8 5388 8 5389 2477 5390 8 5391 8 5392 8 5393 2478 5394 2479 5395 2480 5396 8 5397 8 5398 2481 5399 8 5400 8 5401 2482 5402 2483 5403 2484 5404 2485 5405 2486 5406 2487 5407 8 5408 2488 5409 8 5410 2489 5411 2490 5412 2491 5413 2492 5414 2493 5415 8 5416 2494 5417 2495 5418 2496 5419 2497 5420 1824 5421 2498 5422 2499 5423 2500 5424 8 5425 1824 5426 1824 5427 1824 5428 8 5429 2501 5430 8 5431 2502 5432 8 5433 2503 5434 2504 5435 8 5436 2505 5437 2506 5438 8 5439 1824 5440 8 5441 2507 5442 2508 5443 1976 5444 8 5445 1824 5446 1824 5447 1824 5448 2509 5449 2510 5450 8 5451 1824 5452 1824 5453 1824 5454 2511 5455 8 5456 8 5457 2512 5458 2513 5459 8 5460 2514 5461 8 5462 1824 5463 2515 5464 2516 5465 2517 5466 2518 5467 1612 5468 2519 5469 2520 5470 2521 5471 8 5472 1612 5473 1612 5474 1612 5475 8 5476 2522 5477 8 5478 2523 5479 8 5480 2524 5481 2525 5482 8 5483 2526 5484 2527 5485 8 5486 1612 5487 8 5488 1612 5489 1612 5490 1612 5491 2528 5492 2529 5493 8 5494 1612 5495 1612 5496 1612 5497 2530 5498 8 5499 8 5500 2531 5501 2532 5502 8 5503 2533 5504 8 5505 1612 5506 1428 5507 1428 5508 1428 5509 8 5510 2534 5511 8 5512 2535 5513 8 5514 8 5515 1428 5516 1428 5517 1428 5518 1428 5519 2536 5520 2537 5521 8 5522 8 5523 8 5524 8 5525 8 5526 8 5527 2538 5528 2539 5529 2540 5530 8 5531 2541 5532 8 5533 2542 5534 2543 5535 2544 5536 2545 5537 2546 5538 2547 5539 8 5540 2548 5541 8 5542 8 5543 8 5544 8 5545 2549 5546 8 5547 8 5548 8 5549 2550 5550 2551 5551 2552 5552 8 5553 8 5554 2553 5555 8 5556 8 5557 2554 5558 2555 5559 2556 5560 2557 5561 2558 5562 2559 5563 8 5564 2560 5565 8 5566 2561 5567 2562 5568 2563 5569 2564 5570 2565 5571 8 5572 2566 5573 2567 5574 2568 5575 2569 5576 1428 5577 2570 5578 2571 5579 2572 5580 8 5581 1428 5582 1428 5583 1428 5584 8 5585 2573 5586 8 5587 2574 5588 8 5589 2575 5590 2576 5591 8 5592 2577 5593 2578 5594 8 5595 1428 5596 8 5597 1428 5598 1428 5599 1428 5600 2579 5601 2580 5602 8 5603 1428 5604 1428 5605 1428 5606 2581 5607 8 5608 8 5609 2582 5610 2583 5611 8 5612 2584 5613 8 5614 1428 5615 2585 5616 8 5617 8 5618 8 5619 8 5620 8 5621 8 5622 8 5623 8 5624 8 5625 8 5626 2586 5627 8 5628 2587 5629 8 5630 8 5631 8 5632 2588 5633 8 5634 8 5635 8 5636 8 5637 8 5638 8 5639 8 5640 8 5641 8 5642 8 5643 8 5644 8 5645 8 5646 8 5647 8 5648 8 5649 8 5650 8 5651 8 5652 8 5653 8 5654 8 5655 8 5656 8 5657 8 5658 8 5659 8 5660 8 5661 8 5662 8 5663 8 5664 8 5665 8 5666 8 5667 8 5668 8 5669 8 5670 8 5671 8 5672 8 5673 8 5674 8 5675 8 5676 8 5677 8 5678 8 5679 8 5680 8 5681 8 5682 8 5683 8 5684 8 5685 8 5686 8 5687 8 5688 8 5689 8 5690 8 5691 8 5692 8 5693 8 5694 8 5695 8 5696 8 5697 8 5698 8 5699 8 5700 8 5701 8 5702 8 5703 8 5704 8 5705 8 5706 8 5707 8 5708 8 5709 8 5710 8 5711 2589 5712 8 5713 8 5714 2590 5715 2591 5716 2592 5717 2593 5718 2594 5719 2595 5720 2596 5721 8 5722 2597 5723 2598 5724 2599 5725 8 5726 8 5727 2600 5728 8 5729 8 5730 8 5731 8 5732 8 5733 8 5734 8 5735 2601 5736 8 5737 2602 5738 8 5739 8 5740 2603 5741 2604 5742 8 5743 8 5744 8 5745 8 5746 8 5747 8 5748 8 5749 8 5750 8 5751 8 5752 2605 5753 2606 5754 8 5755 8 5756 8 5757 8 5758 8 5759 8 5760 8 5761 8 5762 8 5763 8 5764 8 5765 2607 5766 8 5767 8 5768 8 5769 8 5770 8 5771 8 5772 8 5773 8 5774 8 5775 8 5776 8 5777 8 5778 8 5779 8 5780 8 5781 8 5782 8 5783 8 5784 8 5785 8 5786 8 5787 8 5788 8 5789 8 5790 8 5791 8 5792 8 5793 8 5794 8 5795 8 5796 8 5797 8 5798 8 5799 8 5800 8 5801 8 5802 8 5803 8 5804 8 5805 8 5806 8 5807 8 5808 8 5809 8 5810 8 5811 8 5812 8 5813 8 5814 8 5815 8 5816 8 5817 8 5818 8 5819 8 5820 8 5821 8 5822 8 5823 8 5824 8 5825 8 5826 8 5827 8 5828 8 5829 8 5830 8 5831 8 5832 8 5833 8 5834 8 5835 8 5836 8 5837 8 5838 8 5839 8 5840 8 5841 8 5842 8 5843 8 5844 8 5845 8 5846 8 5847 8 5848 8 5849 8 5850 8 5851 8 5852 8 5853 8 5854 8 5855 8 5856 2608 5857 2609 5858 8 5859 8 5860 2610 5861 2611 5862 8 5863 2612 5864 8 5865 8 5866 8 5867 8 5868 8 5869 8 5870 8 5871 8 5872 8 5873 8 5874 2613 5875 8 5876 2614 5877 8 5878 8 5879 2615 5880 8 5881 8 5882 8 5883 8 5884 8 5885 8 5886 8 5887 8 5888 8 5889 8 5890 8 5891 8 5892 8 5893 8 5894 8 5895 8 5896 8 5897 8 5898 8 5899 8 5900 8 5901 8 5902 8 5903 8 5904 8 5905 8 5906 8 5907 8 5908 8 5909 8 5910 8 5911 8 5912 8 5913 8 5914 8 5915 8 5916 8 5917 8 5918 8 5919 8 5920 8 5921 8 5922 8 5923 8 5924 8 5925 8 5926 8 5927 8 5928 8 5929 8 5930 8 5931 8 5932 8 5933 8 5934 8 5935 8 5936 8 5937 8 5938 8 5939 8 5940 8 5941 8 5942 8 5943 8 5944 8 5945 8 5946 8 5947 8 5948 8 5949 8 5950 8 5951 8 5952 8 5953 8 5954 8 5955 8 5956 8 5957 8 5958 8 5959 8 5960 8 5961 2616 5962 2617 5963 8 5964 8 5965 8 5966 8 5967 8 5968 2618 5969 2619 5970 2620 5971 2621 5972 8 5973 8 5974 2622 5975 8 5976 8 5977 8 5978 8 5979 8 5980 8 5981 8 5982 8 5983 8 5984 8 5985 8 5986 8 5987 8 5988 8 5989 8 5990 8 5991 8 5992 8 5993 8 5994 8 5995 8 5996 8 5997 8 5998 8 5999 8 6000 8 6001 8 6002 8 6003 8 6004 8 6005 8 6006 8 6007 8 6008 8 6009 8 6010 8 6011 8 6012 8 6013 8 6014 8 6015 8 6016 8 6017 8 6018 8 6019 8 6020 8 6021 8 6022 8 6023 8 6024 8 6025 8 6026 8 6027 8 6028 8 6029 8 6030 8 6031 8 6032 8 6033 8 6034 8 6035 8 6036 8 6037 8 6038 8 6039 8 6040 8 6041 8 6042 8 6043 8 6044 8 6045 8 6046 8 6047 8 6048 8 6049 8 6050 8 6051 8 6052 8 6053 2623 6054 8 6055 8 6056 8 6057 8 6058 8 6059 8 6060 8 6061 8 6062 8 6063 8 6064 8 6065 8 6066 8 6067 8 6068 8 6069 2624 6070 8 6071 8 6072 8 6073 8 6074 8 6075 8 6076 8 6077 8 6078 8 6079 8 6080 8 6081 8 6082 8 6083 8 6084 8 6085 8 6086 8 6087 8 6088 8 6089 8 6090 8 6091 8 6092 8 6093 8 6094 8 6095 8 6096 8 6097 8 6098 8 6099 8 6100 8 6101 8 6102 8 6103 8 6104 8 6105 8 6106 8 6107 8 6108 8 6109 8 6110 8 6111 8 6112 8 6113 8 6114 8 6115 8 6116 8 6117 8 6118 8 6119 8 6120 8 6121 8 6122 8 6123 8 6124 8 6125 8 6126 8 6127 8 6128 8 6129 8 6130 8 6131 8 6132 8 6133 8 6134 8 6135 8 6136 8 6137 8 6138 8 6139 8 6140 8 6141 8 6142 8 6143 8 6144 8 6145 8 6146 8 6147 8 6148 8 6149 8 6150 8 6151 2625 6152 8 6153 8 6154 8 6155 8 6156 8 6157 8 6158 2626 6159 8 6160 2627 6161 8 6162 8 6163 2628 6164 8 6165 8 6166 8 6167 8 6168 8 6169 8 6170 8 6171 8 6172 8 6173 8 6174 8 6175 8 6176 8 6177 8 6178 8 6179 8 6180 8 6181 8 6182 8 6183 8 6184 8 6185 8 6186 8 6187 8 6188 8 6189 8 6190 8 6191 8 6192 8 6193 8 6194 8 6195 8 6196 8 6197 8 6198 8 6199 8 6200 8 6201 8 6202 8 6203 8 6204 8 6205 8 6206 8 6207 8 6208 8 6209 8 6210 8 6211 8 6212 8 6213 8 6214 8 6215 8 6216 8 6217 8 6218 8 6219 8 6220 8 6221 8 6222 8 6223 8 6224 8 6225 8 6226 8 6227 8 6228 8 6229 8 6230 8 6231 8 6232 8 6233 8 6234 8 6235 8 6236 8 6237 8 6238 8 6239 8 6240 8 6241 8 6242 2629 6243 8 6244 8 6245 2630 6246 2631 6247 8 6248 83 6249 2632 6250 2633 6251 2634 6252 2635 6253 2636 6254 2637 6255 2638 6256 2639 6257 2640 6258 2641 6259 2642 6260 2643 6261 2644 6262 2645 6263 2646 6264 2647 6265 2648 6266 2649 6267 2650 6268 2651 6269 2652 6270 2653 6271 2654 6272 2655 6273 2656 6274 2657 6275 2658 6276 2659 6277 2660 6278 2661 6279 83 6280 83 6281 83 6282 83 6283 83 6284 83 6285 83 6286 83 6287 83 6288 83 6289 83 6290 83 6291 83 6292 83 6293 83 6294 2662 6295 2663 6296 8 6297 8 6298 8 6299 8 6300 8 6301 8 6302 2664 6303 8 6304 8 6305 2665 6306 8 6307 2666 6308 2667 6309 2668 6310 2669 6311 2670 6312 2671 6313 8 6314 2672 6315 8 6316 8 6317 8 6318 8 6319 8 6320 2673 6321 2674 6322 8 6323 8 6324 2675 6325 8 6326 2676 6327 8 6328 8 6329 2677 6330 2678 6331 8 6332 83 6333 83 6334 83 6335 83 6336 83 6337 2679 6338 2680 6339 2681 6340 83 6341 83 6342 83 6343 83 6344 83 6345 83 6346 83 6347 83 6348 83 6349 83 6350 8 6351 8 6352 8 6353 8 6354 8 6355 8 6356 2682 6357 2683 6358 8 6359 2684 6360 8 6361 8 6362 2685 6363 2686 6364 2687 6365 8 6366 8 6367 2688 6368 2689 6369 8 6370 8 6371 8 6372 8 6373 8 6374 8 6375 8 6376 8 6377 2690 6378 8 6379 2691 6380 8 6381 8 6382 8 6383 8 6384 8 6385 8 6386 8 6387 8 6388 8 6389 2692 6390 2693 6391 2694 6392 8 6393 8 6394 2695 6395 2696 6396 8 6397 2697 6398 8 6399 8 6400 2698 6401 2699 6402 8 6403 2700 6404 2701 6405 2702 6406 2703 6407 2704 6408 8 6409 2705 6410 2706 6411 1158 6412 2707 6413 2708 6414 2709 6415 8 6416 1098 6417 1098 6418 1098 6419 8 6420 2710 6421 8 6422 2711 6423 8 6424 2712 6425 2713 6426 8 6427 2714 6428 2715 6429 8 6430 1098 6431 8 6432 1098 6433 1098 6434 1098 6435 2716 6436 2717 6437 8 6438 1098 6439 1098 6440 1098 6441 2718 6442 8 6443 8 6444 2719 6445 2720 6446 8 6447 2721 6448 8 6449 1098 6450 2722 6451 8 6452 2723 6453 8 6454 8 6455 8 6456 2724 6457 8 6458 8 6459 2725 6460 2726 6461 817 6462 2727 6463 2728 6464 2729 6465 2730 6466 2731 6467 8 6468 2732 6469 8 6470 8 6471 2733 6472 8 6473 2734 6474 2735 6475 8 6476 8 6477 2736 6478 8 6479 8 6480 2737 6481 8 6482 2738 6483 2739 6484 2740 6485 922 6486 922 6487 2741 6488 2742 6489 922 6490 922 6491 922 6492 922 6493 8 6494 2743 6495 8 6496 2744 6497 8 6498 8 6499 922 6500 922 6501 922 6502 922 6503 2745 6504 2746 6505 8 6506 8 6507 8 6508 8 6509 8 6510 8 6511 2747 6512 2748 6513 2749 6514 8 6515 2750 6516 8 6517 2751 6518 2752 6519 2753 6520 2754 6521 2755 6522 2756 6523 8 6524 2757 6525 8 6526 8 6527 8 6528 8 6529 2758 6530 8 6531 8 6532 8 6533 2759 6534 2760 6535 2761 6536 8 6537 8 6538 2762 6539 8 6540 8 6541 2763 6542 2764 6543 2765 6544 2766 6545 2767 6546 2768 6547 8 6548 2769 6549 8 6550 2770 6551 2771 6552 2772 6553 2773 6554 2774 6555 8 6556 2775 6557 2776 6558 975 6559 2777 6560 2778 6561 2779 6562 8 6563 922 6564 922 6565 922 6566 8 6567 2780 6568 8 6569 2781 6570 8 6571 2782 6572 2783 6573 8 6574 2784 6575 2785 6576 8 6577 1098 6578 922 6579 8 6580 922 6581 922 6582 922 6583 2786 6584 2787 6585 8 6586 922 6587 922 6588 922 6589 2788 6590 8 6591 8 6592 2789 6593 2790 6594 8 6595 2791 6596 8 6597 922 6598 2792 6599 8 6600 8 6601 8 6602 8 6603 8 6604 8 6605 8 6606 8 6607 8 6608 8 6609 8 6610 8 6611 8 6612 8 6613 8 6614 8 6615 8 6616 8 6617 8 6618 8 6619 8 6620 8 6621 8 6622 8 6623 8 6624 8 6625 8 6626 8 6627 8 6628 8 6629 8 6630 8 6631 8 6632 8 6633 8 6634 8 6635 8 6636 8 6637 8 6638 8 6639 8 6640 8 6641 8 6642 8 6643 8 6644 8 6645 8 6646 8 6647 8 6648 8 6649 8 6650 8 6651 8 6652 8 6653 8 6654 8 6655 8 6656 8 6657 8 6658 8 6659 8 6660 8 6661 8 6662 8 6663 8 6664 8 6665 8 6666 8 6667 8 6668 8 6669 8 6670 8 6671 8 6672 8 6673 8 6674 8 6675 8 6676 8 6677 8 6678 2793 6679 8 6680 8 6681 8 6682 2794 6683 8 6684 8 6685 2795 6686 2796 6687 2797 6688 8 6689 8 6690 2798 6691 2799 6692 2800 6693 2801 6694 8 6695 2802 6696 8 6697 2803 6698 2804 6699 2805 6700 2806 6701 2807 6702 2808 6703 2809 6704 2810 6705 2811 6706 817 6707 817 6708 817 6709 817 6710 817 6711 817 6712 817 6713 817 6714 817 6715 817 6716 817 6717 817 6718 817 6719 2812 6720 817 6721 2813 6722 8 6723 2814 6724 2815 6725 8 6726 2816 6727 8 6728 2817 6729 8 6730 8 6731 8 6732 2818 6733 2819 6734 2820 6735 8 6736 2821 6737 8 6738 8 6739 2822 6740 2823 6741 2824 6742 8 6743 8 6744 8 6745 2825 6746 2826 6747 8 6748 8 6749 8 6750 2827 6751 2828 6752 2829 6753 2830 6754 2831 6755 8 6756 2832 6757 8 6758 8 6759 2833 6760 8 6761 2834 6762 2835 6763 8 6764 8 6765 2836 6766 8 6767 8 6768 2837 6769 2838 6770 8 6771 2839 6772 8 6773 8 6774 2840 6775 2841 6776 2842 6777 2843 6778 2844 6779 2845 6780 2846 6781 2847 6782 2848 6783 2849 6784 2850 6785 8 6786 2851 6787 2852 6788 2853 6789 2854 6790 2855 6791 2856 6792 2857 6793 2858 6794 2859 6795 2860 6796 2861 6797 2862 6798 2863 6799 2864 6800 2865 6801 2866 6802 2867 6803 2868 6804 2869 6805 8 6806 2870 6807 8 6808 2871 6809 8 6810 8 6811 2872 6812 2873 6813 2874 6814 2875 6815 8 6816 8 6817 2876 6818 2877 6819 2878 6820 2879 6821 2880 6822 8 6823 2881 6824 8 6825 2882 6826 8 6827 8 6828 8 6829 8 6830 2883 6831 8 6832 2884 6833 2885 6834 2886 6835 2887 6836 2888 6837 2889 6838 8 6839 2890 6840 2891 6841 8 6842 2892 6843 8 6844 2893 6845 2894 6846 2895 6847 2896 6848 2897 6849 2898 6850 2899 6851 8 6852 8 6853 2900 6854 8 6855 8 6856 2901 6857 2902 6858 2903 6859 8 6860 8 6861 8 6862 8 6863 2904 6864 8 6865 8 6866 8 6867 8 6868 2905 6869 2906 6870 2907 6871 2908 6872 2909 6873 2910 6874 2911 6875 2912 6876 2913 6877 2914 6878 2915 6879 2916 6880 2917 6881 2918 6882 2919 6883 2920 6884 2921 6885 2922 6886 2923 6887 2924 6888 2925 6889 2926 6890 2927 6891 2928 6892 2929 6893 2930 6894 2931 6895 2932 6896 2933 6897 2934 6898 2935 6899 2936 6900 2937 6901 2938 6902 2908 6903 2908 6904 2908 6905 2908 6906 2908 6907 2908 6908 2908 6909 2908 6910 2908 6911 2908 6912 2908 6913 2908 6914 2908 6915 2908 6916 2908 6917 2939 6918 2940 6919 8 6920 8 6921 8 6922 8 6923 8 6924 8 6925 2941 6926 8 6927 8 6928 2942 6929 8 6930 2943 6931 2944 6932 2945 6933 2946 6934 2947 6935 2948 6936 8 6937 2949 6938 8 6939 8 6940 8 6941 8 6942 8 6943 2950 6944 2951 6945 8 6946 8 6947 2952 6948 8 6949 2953 6950 8 6951 8 6952 2954 6953 2955 6954 8 6955 2908 6956 2908 6957 2908 6958 2908 6959 2908 6960 2956 6961 2957 6962 2958 6963 2908 6964 2908 6965 2908 6966 2908 6967 2908 6968 2908 6969 2908 6970 2908 6971 2908 6972 2908 6973 2959 6974 8 6975 8 6976 8 6977 8 6978 8 6979 8 6980 8 6981 8 6982 8 6983 8 6984 8 6985 8 6986 8 6987 8 6988 8 6989 8 6990 8 6991 8 6992 8 6993 8 6994 8 6995 8 6996 8 6997 8 6998 8 6999 8 7000 8 7001 8 7002 8 7003 8 7004 8 7005 8 7006 8 7007 8 7008 8 7009 8 7010 8 7011 8 7012 8 7013 8 7014 8 7015 8 7016 8 7017 8 7018 8 7019 8 7020 8 7021 8 7022 8 7023 8 7024 8 7025 8 7026 8 7027 8 7028 8 7029 8 7030 8 7031 8 7032 8 7033 8 7034 8 7035 8 7036 8 7037 8 7038 8 7039 8 7040 8 7041 8 7042 8 7043 8 7044 8 7045 8 7046 8 7047 8 7048 8 7049 8 7050 8 7051 8 7052 8 7053 2960 7054 8 7055 8 7056 8 7057 2961 7058 8 7059 8 7060 2962 7061 2963 7062 2964 7063 8 7064 8 7065 2965 7066 2966 7067 2967 7068 2968 7069 8 7070 2969 7071 8 7072 2970 7073 2971 7074 2972 7075 2973 7076 2974 7077 2975 7078 2976 7079 2977 7080 2978 7081 2908 7082 2908 7083 2908 7084 2908 7085 2908 7086 2908 7087 2908 7088 2908 7089 2908 7090 2908 7091 2908 7092 2908 7093 2908 7094 2979 7095 2908 7096 2980 7097 8 7098 2981 7099 2982 7100 8 7101 2983 7102 8 7103 2984 7104 8 7105 8 7106 8 7107 2985 7108 2986 7109 2987 7110 8 7111 2988 7112 8 7113 8 7114 2989 7115 2990 7116 2991 7117 8 7118 8 7119 8 7120 2992 7121 2993 7122 8 7123 8 7124 8 7125 2994 7126 2995 7127 2996 7128 2997 7129 2998 7130 8 7131 2999 7132 8 7133 8 7134 3000 7135 8 7136 3001 7137 3002 7138 8 7139 8 7140 3003 7141 8 7142 8 7143 3004 7144 3005 7145 8 7146 8 7147 3006 7148 3007 7149 3008 7150 3009 7151 3010 7152 3011 7153 3012 7154 3013 7155 3014 7156 3015 7157 3016 7158 8 7159 3017 7160 3018 7161 3019 7162 3020 7163 3021 7164 3022 7165 3023 7166 3024 7167 3025 7168 3026 7169 8 7170 8 7171 3027 7172 8 7173 3028 7174 8 7175 8 7176 8 7177 3029 7178 8 7179 8 7180 3030 7181 3031 7182 3032 7183 3033 7184 3034 7185 8 7186 3035 7187 8 7188 8 7189 3036 7190 8 7191 3037 7192 3038 7193 8 7194 8 7195 3039 7196 8 7197 3040 7198 8 7199 3041 7200 3042 7201 3043 7202 2908 7203 2908 7204 3044 7205 3045 7206 2908 7207 2908 7208 2908 7209 2908 7210 8 7211 3046 7212 8 7213 3047 7214 8 7215 8 7216 2908 7217 2908 7218 2908 7219 2908 7220 3048 7221 3049 7222 8 7223 8 7224 8 7225 8 7226 8 7227 8 7228 3050 7229 3051 7230 3052 7231 8 7232 3053 7233 8 7234 3054 7235 3055 7236 3056 7237 3057 7238 3058 7239 3059 7240 8 7241 3060 7242 8 7243 8 7244 8 7245 8 7246 3061 7247 8 7248 8 7249 8 7250 3062 7251 3063 7252 3064 7253 8 7254 8 7255 3065 7256 8 7257 8 7258 3066 7259 3067 7260 3068 7261 3069 7262 3070 7263 3071 7264 8 7265 3072 7266 8 7267 3073 7268 3074 7269 3075 7270 3076 7271 3077 7272 8 7273 3078 7274 3079 7275 3080 7276 3081 7277 2908 7278 3082 7279 3083 7280 3084 7281 8 7282 2908 7283 2908 7284 2908 7285 8 7286 3085 7287 8 7288 3086 7289 8 7290 3087 7291 3088 7292 8 7293 3089 7294 3090 7295 8 7296 2908 7297 8 7298 2908 7299 2908 7300 2908 7301 3091 7302 3092 7303 8 7304 2908 7305 2908 7306 2908 7307 3093 7308 8 7309 8 7310 3094 7311 3095 7312 8 7313 3096 7314 8 7315 2908 7316 3097 7317 3098 7318 3099 7319 3100 7320 3101 7321 3102 7322 3103 7323 3104 7324 3105 7325 3106 7326 3107 7327 8 7328 3108 7329 8 7330 3109 7331 8 7332 8 7333 3110 7334 3111 7335 3112 7336 3113 7337 8 7338 8 7339 3114 7340 3115 7341 3116 7342 3117 7343 3118 7344 8 7345 3119 7346 8 7347 3120 7348 8 7349 8 7350 8 7351 8 7352 3121 7353 8 7354 3122 7355 3123 7356 3124 7357 3125 7358 3126 7359 3127 7360 8 7361 3128 7362 3129 7363 8 7364 3130 7365 8 7366 3131 7367 3132 7368 3133 7369 3134 7370 3135 7371 3136 7372 3137 7373 3138 7374 8 7375 8 7376 3139 7377 8 7378 8 7379 3140 7380 3141 7381 3142 7382 8 7383 8 7384 8 7385 8 7386 3143 7387 8 7388 8 7389 8 7390 8 7391 3144 7392 3145 7393 3146 7394 3147 7395 3140 7396 3148 7397 3149 7398 3150 7399 3151 7400 3152 7401 3153 7402 3154 7403 3155 7404 3156 7405 3157 7406 3158 7407 3159 7408 3160 7409 3161 7410 3162 7411 3163 7412 3164 7413 3165 7414 3166 7415 3167 7416 3168 7417 3169 7418 3170 7419 3171 7420 3172 7421 3173 7422 3174 7423 3175 7424 3176 7425 3177 7426 3140 7427 3140 7428 3140 7429 3140 7430 3140 7431 3140 7432 3140 7433 3140 7434 3140 7435 3140 7436 3140 7437 3140 7438 3140 7439 3140 7440 3140 7441 3178 7442 3179 7443 8 7444 8 7445 8 7446 8 7447 8 7448 8 7449 3180 7450 8 7451 8 7452 3181 7453 8 7454 3182 7455 3183 7456 3184 7457 3185 7458 3186 7459 3187 7460 8 7461 3188 7462 8 7463 8 7464 8 7465 8 7466 8 7467 3189 7468 3190 7469 8 7470 8 7471 3191 7472 8 7473 3192 7474 8 7475 8 7476 3193 7477 3194 7478 8 7479 3140 7480 3140 7481 3140 7482 3140 7483 3140 7484 3195 7485 3196 7486 3197 7487 3140 7488 3140 7489 3140 7490 3140 7491 3140 7492 3140 7493 3140 7494 3140 7495 3140 7496 3140 7497 3198 7498 3199 7499 3200 7500 3201 7501 3140 7502 3202 7503 8 7504 8 7505 8 7506 8 7507 8 7508 8 7509 8 7510 8 7511 8 7512 8 7513 8 7514 8 7515 8 7516 8 7517 8 7518 8 7519 8 7520 8 7521 8 7522 8 7523 8 7524 8 7525 8 7526 8 7527 8 7528 8 7529 8 7530 8 7531 8 7532 8 7533 8 7534 8 7535 8 7536 8 7537 8 7538 8 7539 8 7540 8 7541 8 7542 8 7543 8 7544 8 7545 8 7546 8 7547 8 7548 8 7549 8 7550 8 7551 8 7552 8 7553 8 7554 8 7555 8 7556 8 7557 8 7558 8 7559 8 7560 8 7561 8 7562 8 7563 8 7564 8 7565 8 7566 8 7567 8 7568 8 7569 8 7570 8 7571 8 7572 8 7573 8 7574 8 7575 8 7576 8 7577 8 7578 8 7579 8 7580 8 7581 8 7582 3203 7583 8 7584 8 7585 8 7586 3204 7587 8 7588 8 7589 3205 7590 3206 7591 3207 7592 8 7593 8 7594 3208 7595 3209 7596 3210 7597 3211 7598 8 7599 3212 7600 8 7601 3213 7602 3214 7603 3215 7604 3216 7605 3217 7606 3218 7607 3219 7608 3220 7609 3221 7610 3140 7611 3140 7612 3140 7613 3140 7614 3140 7615 3140 7616 3140 7617 3140 7618 3140 7619 3140 7620 3140 7621 3140 7622 3140 7623 3222 7624 3140 7625 3223 7626 8 7627 3224 7628 3225 7629 8 7630 3226 7631 8 7632 3227 7633 8 7634 8 7635 8 7636 3228 7637 3229 7638 3230 7639 8 7640 3231 7641 8 7642 8 7643 3232 7644 3233 7645 3234 7646 8 7647 8 7648 8 7649 3235 7650 3236 7651 8 7652 8 7653 8 7654 3237 7655 3238 7656 3239 7657 3240 7658 3241 7659 8 7660 3242 7661 8 7662 8 7663 3243 7664 8 7665 3244 7666 3245 7667 8 7668 8 7669 3246 7670 8 7671 8 7672 3247 7673 3248 7674 8 7675 8 7676 3249 7677 3250 7678 3251 7679 3252 7680 3253 7681 3254 7682 3255 7683 3256 7684 3257 7685 3258 7686 3259 7687 8 7688 3260 7689 3261 7690 3262 7691 3263 7692 3264 7693 3265 7694 3266 7695 3267 7696 3268 7697 3269 7698 3270 7699 3271 7700 3272 7701 3273 7702 3274 7703 3275 7704 3276 7705 3277 7706 3278 7707 8 7708 3279 7709 8 7710 3280 7711 8 7712 8 7713 3281 7714 3282 7715 3283 7716 3284 7717 8 7718 8 7719 3285 7720 3286 7721 3287 7722 3288 7723 3289 7724 8 7725 3290 7726 8 7727 3291 7728 8 7729 8 7730 8 7731 8 7732 3292 7733 8 7734 3293 7735 3294 7736 3295 7737 3296 7738 3297 7739 3298 7740 8 7741 3299 7742 3300 7743 8 7744 3301 7745 8 7746 3302 7747 3303 7748 3304 7749 3305 7750 3306 7751 3307 7752 3308 7753 8 7754 8 7755 3309 7756 8 7757 8 7758 3310 7759 3311 7760 3312 7761 8 7762 8 7763 8 7764 8 7765 3313 7766 8 7767 8 7768 8 7769 8 7770 3314 7771 3315 7772 3316 7773 3317 7774 3318 7775 3319 7776 3320 7777 3321 7778 3322 7779 3323 7780 3324 7781 3325 7782 3326 7783 3327 7784 3328 7785 3329 7786 3330 7787 3331 7788 3332 7789 3333 7790 3334 7791 3335 7792 3336 7793 3337 7794 3338 7795 3339 7796 3340 7797 3341 7798 3342 7799 3343 7800 3344 7801 3345 7802 3346 7803 3347 7804 3317 7805 3317 7806 3317 7807 3317 7808 3317 7809 3317 7810 3317 7811 3317 7812 3317 7813 3317 7814 3317 7815 3317 7816 3317 7817 3317 7818 3317 7819 3348 7820 3349 7821 8 7822 8 7823 8 7824 8 7825 8 7826 8 7827 3350 7828 8 7829 8 7830 3351 7831 8 7832 3352 7833 3353 7834 3354 7835 3355 7836 3356 7837 3357 7838 8 7839 3358 7840 8 7841 8 7842 8 7843 8 7844 8 7845 3359 7846 3360 7847 8 7848 8 7849 3361 7850 8 7851 3362 7852 8 7853 8 7854 3363 7855 3364 7856 8 7857 3317 7858 3317 7859 3317 7860 3317 7861 3317 7862 3365 7863 3366 7864 3367 7865 3317 7866 3317 7867 3317 7868 3317 7869 3317 7870 3317 7871 3317 7872 3317 7873 3317 7874 3317 7875 3368 7876 8 7877 8 7878 8 7879 8 7880 8 7881 8 7882 8 7883 8 7884 8 7885 8 7886 8 7887 8 7888 8 7889 8 7890 8 7891 8 7892 8 7893 8 7894 8 7895 8 7896 8 7897 8 7898 8 7899 8 7900 8 7901 8 7902 8 7903 8 7904 8 7905 8 7906 8 7907 8 7908 8 7909 8 7910 8 7911 8 7912 8 7913 8 7914 8 7915 8 7916 8 7917 8 7918 8 7919 8 7920 8 7921 8 7922 8 7923 8 7924 8 7925 8 7926 8 7927 8 7928 8 7929 8 7930 8 7931 8 7932 8 7933 8 7934 8 7935 8 7936 8 7937 8 7938 8 7939 8 7940 8 7941 8 7942 8 7943 8 7944 8 7945 8 7946 8 7947 8 7948 8 7949 8 7950 8 7951 8 7952 8 7953 8 7954 8 7955 3369 7956 8 7957 8 7958 8 7959 3370 7960 8 7961 8 7962 3371 7963 3372 7964 3373 7965 8 7966 8 7967 3374 7968 3375 7969 3376 7970 3377 7971 8 7972 3378 7973 8 7974 3379 7975 3380 7976 3381 7977 3382 7978 3383 7979 3384 7980 3385 7981 3386 7982 3387 7983 3317 7984 3317 7985 3317 7986 3317 7987 3317 7988 3317 7989 3317 7990 3317 7991 3317 7992 3317 7993 3317 7994 3317 7995 3317 7996 3388 7997 3317 7998 3389 7999 8 8000 3390 8001 3391 8002 8 8003 3392 8004 8 8005 3393 8006 8 8007 8 8008 8 8009 3394 8010 3395 8011 3396 8012 8 8013 3397 8014 8 8015 8 8016 3398 8017 3399 8018 3400 8019 8 8020 8 8021 8 8022 3401 8023 3402 8024 8 8025 8 8026 8 8027 3403 8028 3404 8029 3405 8030 3406 8031 3407 8032 8 8033 3408 8034 8 8035 8 8036 3409 8037 8 8038 3410 8039 3411 8040 8 8041 8 8042 3412 8043 8 8044 8 8045 3413 8046 3414 8047 8 8048 8 8049 3415 8050 3416 8051 3417 8052 3418 8053 3419 8054 3420 8055 3421 8056 3422 8057 3423 8058 3424 8059 3425 8060 8 8061 3426 8062 3427 8063 3428 8064 3429 8065 3430 8066 3431 8067 3432 8068 3433 8069 3434 8070 3435 8071 8 8072 8 8073 3436 8074 8 8075 3437 8076 8 8077 8 8078 8 8079 3438 8080 8 8081 8 8082 3439 8083 3440 8084 3441 8085 3442 8086 3443 8087 8 8088 3444 8089 8 8090 8 8091 3445 8092 8 8093 3446 8094 3447 8095 8 8096 8 8097 3448 8098 8 8099 3449 8100 8 8101 3450 8102 3451 8103 3452 8104 3317 8105 3317 8106 3453 8107 3454 8108 3317 8109 3317 8110 3317 8111 3317 8112 8 8113 3455 8114 8 8115 3456 8116 8 8117 8 8118 3317 8119 3317 8120 3317 8121 3317 8122 3457 8123 3458 8124 8 8125 8 8126 8 8127 8 8128 8 8129 8 8130 3459 8131 3460 8132 3461 8133 8 8134 3462 8135 8 8136 3463 8137 3464 8138 3465 8139 3466 8140 3467 8141 3468 8142 8 8143 3469 8144 8 8145 8 8146 8 8147 8 8148 3470 8149 8 8150 8 8151 8 8152 3471 8153 3472 8154 3473 8155 8 8156 8 8157 3474 8158 8 8159 8 8160 3475 8161 3476 8162 3477 8163 3478 8164 3479 8165 3480 8166 8 8167 3481 8168 8 8169 3482 8170 3483 8171 3484 8172 3485 8173 3486 8174 8 8175 3487 8176 3488 8177 3489 8178 3490 8179 3317 8180 3491 8181 3492 8182 3493 8183 8 8184 3317 8185 3317 8186 3317 8187 8 8188 3494 8189 8 8190 3495 8191 8 8192 3496 8193 3497 8194 8 8195 3498 8196 3499 8197 8 8198 3317 8199 8 8200 3317 8201 3317 8202 3317 8203 3500 8204 3501 8205 8 8206 3317 8207 3317 8208 3317 8209 3502 8210 8 8211 8 8212 3503 8213 3504 8214 8 8215 3505 8216 8 8217 3317 8218 3506 8219 8 8220 3507 8221 8 8222 8 8223 8 8224 3508 8225 8 8226 8 8227 3509 8228 3510 8229 3511 8230 3512 8231 3513 8232 8 8233 3514 8234 8 8235 8 8236 3515 8237 8 8238 3516 8239 3517 8240 8 8241 8 8242 3518 8243 8 8244 3519 8245 8 8246 3520 8247 3521 8248 8 8249 3522 8250 8 8251 3523 8252 8 8253 8 8254 8 8255 3524 8256 3525 8257 3526 8258 8 8259 3527 8260 8 8261 8 8262 3528 8263 3529 8264 3530 8265 8 8266 8 8267 8 8268 3531 8269 3532 8270 8 8271 8 8272 8 8273 3533 8274 3534 8275 3535 8276 3536 8277 3537 8278 8 8279 3538 8280 8 8281 8 8282 3539 8283 8 8284 3540 8285 3541 8286 8 8287 8 8288 3542 8289 8 8290 8 8291 3543 8292 3544 8293 8 8294 3545 8295 8 8296 3546 8297 3547 8298 3548 8299 3140 8300 3140 8301 3549 8302 3550 8303 3140 8304 3140 8305 3140 8306 3140 8307 8 8308 3551 8309 8 8310 3552 8311 8 8312 8 8313 3140 8314 3140 8315 3140 8316 3140 8317 3553 8318 8 8319 3554 8320 3555 8321 3556 8322 8 8323 8 8324 8 8325 3557 8326 3558 8327 3559 8328 8 8329 8 8330 8 8331 3560 8332 3561 8333 3562 8334 3563 8335 3564 8336 3565 8337 8 8338 3566 8339 3567 8340 3568 8341 3569 8342 3570 8343 8 8344 3571 8345 3572 8346 3200 8347 3573 8348 3574 8349 3575 8350 8 8351 3140 8352 3140 8353 3140 8354 8 8355 3576 8356 8 8357 3577 8358 8 8359 3578 8360 3579 8361 8 8362 3580 8363 3581 8364 8 8365 3140 8366 8 8367 3140 8368 3140 8369 3140 8370 3582 8371 3583 8372 8 8373 3140 8374 3140 8375 3140 8376 3584 8377 8 8378 8 8379 3585 8380 3586 8381 8 8382 3587 8383 8 8384 3140 8385 3588 8386 8 8387 3589 8388 8 8389 8 8390 8 8391 3590 8392 8 8393 8 8394 8 8395 3591 8396 3592 8397 3593 8398 3594 8399 3595 8400 8 8401 3596 8402 8 8403 8 8404 3597 8405 8 8406 3598 8407 3599 8408 8 8409 8 8410 3600 8411 8 8412 8 8413 3601 8414 8 8415 3602 8416 3603 8417 3604 8418 817 8419 817 8420 3605 8421 3606 8422 817 8423 817 8424 817 8425 817 8426 8 8427 3607 8428 8 8429 3608 8430 8 8431 8 8432 817 8433 817 8434 817 8435 817 8436 3609 8437 8 8438 3610 8439 3611 8440 3612 8441 8 8442 8 8443 8 8444 3613 8445 3614 8446 3615 8447 8 8448 8 8449 8 8450 3616 8451 3617 8452 3618 8453 3619 8454 3620 8455 3621 8456 8 8457 3622 8458 3623 8459 3624 8460 3625 8461 3626 8462 8 8463 3627 8464 3628 8465 2725 8466 3629 8467 3630 8468 3631 8469 8 8470 817 8471 817 8472 817 8473 8 8474 3632 8475 8 8476 3633 8477 8 8478 3634 8479 3635 8480 8 8481 3636 8482 3637 8483 8 8484 3140 8485 817 8486 8 8487 817 8488 817 8489 817 8490 3638 8491 3639 8492 8 8493 817 8494 817 8495 817 8496 3640 8497 8 8498 8 8499 3641 8500 3642 8501 8 8502 3643 8503 8 8504 817 8505 3644 8506 3645 8507 8 8508 3646 8509 8 8510 8 8511 3647 8512 3648 8513 8 8514 3649 8515 8 8516 8 8517 3650 8518 3651 8519 3652 8520 3653 8521 3654 8522 3655 8523 3656 8524 3657 8525 3658 8526 3659 8527 3660 8528 8 8529 3661 8530 3662 8531 3663 8532 3664 8533 3665 8534 3666 8535 3667 8536 3668 8537 3669 8538 3670 8539 8 8540 3671 8541 555 8542 3672 8543 8 8544 3673 8545 8 8546 8 8547 8 8548 3674 8549 8 8550 8 8551 3675 8552 3676 8553 3677 8554 3678 8555 3679 8556 8 8557 3680 8558 8 8559 8 8560 3681 8561 8 8562 3682 8563 3683 8564 8 8565 8 8566 3684 8567 8 8568 3685 8569 8 8570 3686 8571 3687 8572 3688 8573 662 8574 662 8575 3689 8576 3690 8577 662 8578 662 8579 662 8580 662 8581 8 8582 3691 8583 8 8584 3692 8585 8 8586 8 8587 662 8588 662 8589 662 8590 662 8591 3693 8592 8 8593 3694 8594 3695 8595 3696 8596 8 8597 8 8598 8 8599 3697 8600 3698 8601 3699 8602 8 8603 8 8604 8 8605 3700 8606 3701 8607 3702 8608 3703 8609 3704 8610 3705 8611 8 8612 3706 8613 3707 8614 3708 8615 3709 8616 3710 8617 8 8618 3711 8619 3712 8620 3713 8621 3714 8622 662 8623 3715 8624 3716 8625 3717 8626 8 8627 662 8628 662 8629 662 8630 8 8631 3718 8632 8 8633 3719 8634 8 8635 3720 8636 3721 8637 8 8638 3722 8639 3723 8640 8 8641 662 8642 8 8643 662 8644 662 8645 662 8646 3724 8647 3725 8648 8 8649 662 8650 662 8651 662 8652 3726 8653 8 8654 8 8655 3727 8656 3728 8657 8 8658 3729 8659 8 8660 662 8661 555 8662 8 8663 555 8664 555 8665 555 8666 3730 8667 8 8668 3731 8669 3732 8670 3733 8671 8 8672 8 8673 8 8674 3734 8675 3735 8676 3736 8677 8 8678 8 8679 8 8680 3737 8681 3738 8682 3739 8683 3740 8684 3741 8685 3742 8686 8 8687 3743 8688 3744 8689 3745 8690 3746 8691 3747 8692 8 8693 3748 8694 3749 8695 3750 8696 3751 8697 555 8698 3752 8699 3753 8700 3754 8701 8 8702 555 8703 555 8704 3755 8705 8 8706 3756 8707 8 8708 3757 8709 3758 8710 8 8711 3759 8712 3760 8713 8 8714 555 8715 3761 8716 3762 8717 8 8718 3763 8719 8 8720 8 8721 8 8722 8 8723 8 8724 8 8725 8 8726 8 8727 8 8728 8 8729 8 8730 8 8731 8 8732 8 8733 8 8734 8 8735 8 8736 8 8737 8 8738 8 8739 8 8740 8 8741 8 8742 8 8743 8 8744 8 8745 8 8746 8 8747 8 8748 8 8749 8 8750 8 8751 8 8752 8 8753 8 8754 8 8755 8 8756 8 8757 8 8758 8 8759 8 8760 8 8761 8 8762 8 8763 8 8764 8 8765 8 8766 8 8767 8 8768 8 8769 8 8770 8 8771 8 8772 8 8773 8 8774 8 8775 8 8776 8 8777 8 8778 8 8779 8 8780 8 8781 8 8782 8 8783 8 8784 8 8785 8 8786 8 8787 8 8788 8 8789 8 8790 8 8791 8 8792 8 8793 8 8794 8 8795 8 8796 8 8797 8 8798 3764 8799 8 8800 8 8801 8 8802 3765 8803 8 8804 8 8805 3766 8806 3767 8807 3768 8808 8 8809 8 8810 3769 8811 3770 8812 3771 8813 3772 8814 8 8815 3773 8816 8 8817 3774 8818 3775 8819 3776 8820 3777 8821 3778 8822 3779 8823 3780 8824 3781 8825 3782 8826 555 8827 555 8828 555 8829 555 8830 555 8831 555 8832 555 8833 555 8834 555 8835 555 8836 555 8837 555 8838 555 8839 3783 8840 555 8841 3784 8842 3785 8843 8 8844 3786 8845 8 8846 3787 8847 8 8848 8 8849 8 8850 3788 8851 3789 8852 3790 8853 8 8854 3791 8855 8 8856 8 8857 3792 8858 3793 8859 3794 8860 8 8861 8 8862 8 8863 3795 8864 3796 8865 8 8866 8 8867 8 8868 3797 8869 3798 8870 3799 8871 3800 8872 3801 8873 8 8874 3802 8875 8 8876 3803 8877 8 8878 3804 8879 8 8880 3805 8881 8 8882 8 8883 3806 8884 555 8885 3807 8886 3808 8887 3809 8888 3810 8889 3811 8890 3812 8891 3813 8892 3814 8893 3815 8894 3816 8895 3817 8896 8 8897 3818 8898 3819 8899 3820 8900 3821 8901 3822 8902 3823 8903 3824 8904 3825 8905 3826 8906 3827 8907 8 8908 3828 8909 8 8910 3829 8911 8 8912 8 8913 8 8914 3830 8915 8 8916 8 8917 3831 8918 3832 8919 3833 8920 3834 8921 3835 8922 8 8923 3836 8924 8 8925 8 8926 3837 8927 8 8928 3838 8929 3839 8930 8 8931 8 8932 3840 8933 8 8934 3841 8935 8 8936 3842 8937 3843 8938 3844 8939 555 8940 555 8941 555 8942 555 8943 3845 8944 8 8945 3846 8946 8 8947 8 8948 555 8949 555 8950 8 8951 8 8952 555 8953 8 8954 555 8955 555 8956 555 8957 555 8958 555 8959 3847 8960 8 8961 8 8962 3848 8963 3849 8964 8 8965 3850 8966 8 8967 555 8968 3851 8969 3852 8970 3853 8971 3854 8972 8 8973 3855 8974 8 8975 3856 8976 3857 8977 3858 8978 3859 8979 3860 8980 3861 8981 3862 8982 3863 8983 3864 8984 457 8985 457 8986 457 8987 457 8988 457 8989 457 8990 457 8991 457 8992 457 8993 457 8994 457 8995 457 8996 457 8997 3865 8998 457 8999 3866 9000 8 9001 3867 9002 3868 9003 8 9004 3869 9005 8 9006 3870 9007 8 9008 8 9009 8 9010 3871 9011 3872 9012 3873 9013 8 9014 3874 9015 8 9016 8 9017 3875 9018 3876 9019 3877 9020 8 9021 8 9022 8 9023 3878 9024 3879 9025 8 9026 8 9027 8 9028 3880 9029 3881 9030 3882 9031 3883 9032 3884 9033 8 9034 3885 9035 8 9036 8 9037 3886 9038 8 9039 3887 9040 3888 9041 8 9042 8 9043 3889 9044 8 9045 8 9046 3890 9047 3891 9048 8 9049 8 9050 3892 9051 3893 9052 3894 9053 3895 9054 3896 9055 3897 9056 3898 9057 3899 9058 3900 9059 3901 9060 3902 9061 8 9062 3903 9063 3904 9064 3905 9065 3906 9066 3907 9067 3908 9068 3909 9069 3910 9070 3911 9071 3912 9072 3913 9073 3914 9074 3915 9075 3916 9076 3917 9077 3918 9078 3919 9079 3920 9080 3921 9081 8 9082 3922 9083 8 9084 3923 9085 8 9086 8 9087 3924 9088 3925 9089 3926 9090 3927 9091 8 9092 8 9093 3928 9094 3929 9095 3930 9096 3931 9097 3932 9098 8 9099 3933 9100 8 9101 3934 9102 8 9103 8 9104 8 9105 8 9106 3935 9107 8 9108 3936 9109 3937 9110 3938 9111 3939 9112 3940 9113 3941 9114 8 9115 3942 9116 3943 9117 8 9118 3944 9119 8 9120 3945 9121 3946 9122 3947 9123 3948 9124 3949 9125 3950 9126 3951 9127 8 9128 8 9129 3952 9130 8 9131 8 9132 3953 9133 3954 9134 3955 9135 8 9136 8 9137 8 9138 8 9139 3956 9140 8 9141 8 9142 8 9143 8 9144 3957 9145 3958 9146 3959 9147 3960 9148 3961 9149 3962 9150 3963 9151 3964 9152 3965 9153 3966 9154 3967 9155 3968 9156 3969 9157 3970 9158 3971 9159 3972 9160 3973 9161 3974 9162 3975 9163 3976 9164 3977 9165 3978 9166 3979 9167 3980 9168 3981 9169 3982 9170 3983 9171 3984 9172 3985 9173 3986 9174 3987 9175 3988 9176 3989 9177 3990 9178 3960 9179 3960 9180 3960 9181 3960 9182 3960 9183 3960 9184 3960 9185 3960 9186 3960 9187 3960 9188 3960 9189 3960 9190 3960 9191 3960 9192 3960 9193 3991 9194 3992 9195 8 9196 8 9197 8 9198 8 9199 8 9200 8 9201 3993 9202 8 9203 8 9204 3994 9205 8 9206 3995 9207 3996 9208 3997 9209 3998 9210 3999 9211 4000 9212 8 9213 4001 9214 8 9215 8 9216 8 9217 8 9218 8 9219 4002 9220 4003 9221 8 9222 8 9223 4004 9224 8 9225 4005 9226 8 9227 8 9228 4006 9229 4007 9230 8 9231 3960 9232 3960 9233 3960 9234 3960 9235 3960 9236 4008 9237 4009 9238 4010 9239 3960 9240 3960 9241 3960 9242 3960 9243 3960 9244 3960 9245 3960 9246 3960 9247 3960 9248 3960 9249 4011 9250 8 9251 8 9252 8 9253 8 9254 8 9255 8 9256 8 9257 8 9258 8 9259 8 9260 8 9261 8 9262 8 9263 8 9264 8 9265 8 9266 8 9267 8 9268 8 9269 8 9270 8 9271 8 9272 8 9273 8 9274 8 9275 8 9276 8 9277 8 9278 8 9279 8 9280 8 9281 8 9282 8 9283 8 9284 8 9285 8 9286 8 9287 8 9288 8 9289 8 9290 8 9291 8 9292 8 9293 8 9294 8 9295 8 9296 8 9297 8 9298 8 9299 8 9300 8 9301 8 9302 8 9303 8 9304 8 9305 8 9306 8 9307 8 9308 8 9309 8 9310 8 9311 8 9312 8 9313 8 9314 8 9315 8 9316 8 9317 8 9318 8 9319 8 9320 8 9321 8 9322 8 9323 8 9324 8 9325 8 9326 8 9327 8 9328 8 9329 4012 9330 8 9331 8 9332 8 9333 4013 9334 8 9335 8 9336 4014 9337 4015 9338 4016 9339 8 9340 8 9341 4017 9342 4018 9343 4019 9344 4020 9345 8 9346 4021 9347 8 9348 4022 9349 4023 9350 4024 9351 4025 9352 4026 9353 4027 9354 4028 9355 4029 9356 4030 9357 3960 9358 3960 9359 3960 9360 3960 9361 3960 9362 3960 9363 3960 9364 3960 9365 3960 9366 3960 9367 3960 9368 3960 9369 3960 9370 4031 9371 3960 9372 4032 9373 8 9374 4033 9375 4034 9376 8 9377 4035 9378 8 9379 4036 9380 8 9381 8 9382 8 9383 4037 9384 4038 9385 4039 9386 8 9387 4040 9388 8 9389 8 9390 4041 9391 4042 9392 4043 9393 8 9394 8 9395 8 9396 4044 9397 4045 9398 8 9399 8 9400 8 9401 4046 9402 4047 9403 4048 9404 4049 9405 4050 9406 8 9407 4051 9408 8 9409 8 9410 4052 9411 8 9412 4053 9413 4054 9414 8 9415 8 9416 4055 9417 8 9418 8 9419 4056 9420 4057 9421 8 9422 8 9423 4058 9424 4059 9425 4060 9426 4061 9427 4062 9428 4063 9429 4064 9430 4065 9431 4066 9432 4067 9433 4068 9434 8 9435 4069 9436 4070 9437 4071 9438 4072 9439 4073 9440 4074 9441 4075 9442 4076 9443 4077 9444 4078 9445 8 9446 8 9447 4079 9448 8 9449 4080 9450 8 9451 8 9452 8 9453 4081 9454 8 9455 8 9456 4082 9457 4083 9458 4084 9459 4085 9460 4086 9461 8 9462 4087 9463 8 9464 8 9465 4088 9466 8 9467 4089 9468 4090 9469 8 9470 8 9471 4091 9472 8 9473 4092 9474 8 9475 4093 9476 4094 9477 4095 9478 3960 9479 3960 9480 4096 9481 4097 9482 4098 9483 4099 9484 4100 9485 4101 9486 4102 9487 4103 9488 4104 9489 4105 9490 8 9491 4106 9492 8 9493 4107 9494 8 9495 8 9496 4108 9497 4109 9498 4110 9499 4111 9500 8 9501 8 9502 4112 9503 4113 9504 4114 9505 4115 9506 4116 9507 8 9508 4117 9509 8 9510 4118 9511 8 9512 8 9513 8 9514 8 9515 4119 9516 8 9517 4120 9518 4121 9519 4122 9520 4123 9521 4124 9522 4125 9523 8 9524 4126 9525 4127 9526 8 9527 4128 9528 8 9529 4129 9530 4130 9531 4131 9532 4132 9533 4133 9534 4134 9535 4135 9536 8 9537 8 9538 4136 9539 8 9540 8 9541 4137 9542 4138 9543 4139 9544 8 9545 8 9546 8 9547 8 9548 4140 9549 8 9550 8 9551 8 9552 8 9553 4141 9554 4142 9555 4143 9556 4144 9557 4145 9558 4146 9559 4147 9560 4148 9561 4149 9562 4150 9563 4151 9564 4152 9565 4153 9566 4154 9567 4155 9568 4156 9569 4157 9570 4158 9571 4159 9572 4160 9573 4161 9574 4162 9575 4163 9576 4164 9577 4165 9578 4166 9579 4167 9580 4168 9581 4169 9582 4170 9583 4171 9584 4172 9585 4173 9586 4174 9587 4144 9588 4144 9589 4144 9590 4144 9591 4144 9592 4144 9593 4144 9594 4144 9595 4144 9596 4144 9597 4144 9598 4144 9599 4144 9600 4144 9601 4144 9602 4175 9603 4176 9604 8 9605 8 9606 8 9607 8 9608 8 9609 8 9610 4177 9611 8 9612 8 9613 4178 9614 8 9615 4179 9616 4180 9617 4181 9618 4182 9619 4183 9620 4184 9621 8 9622 4185 9623 8 9624 8 9625 8 9626 8 9627 8 9628 4186 9629 4187 9630 8 9631 8 9632 4188 9633 8 9634 4189 9635 8 9636 8 9637 4190 9638 4191 9639 8 9640 4144 9641 4144 9642 4144 9643 4144 9644 4144 9645 4192 9646 4193 9647 4194 9648 4144 9649 4144 9650 4144 9651 4144 9652 4144 9653 4144 9654 4144 9655 4144 9656 4144 9657 4144 9658 4195 9659 8 9660 8 9661 8 9662 8 9663 8 9664 8 9665 8 9666 8 9667 8 9668 8 9669 8 9670 8 9671 8 9672 8 9673 8 9674 8 9675 8 9676 8 9677 8 9678 8 9679 8 9680 8 9681 8 9682 8 9683 8 9684 8 9685 8 9686 8 9687 8 9688 8 9689 8 9690 8 9691 8 9692 8 9693 8 9694 8 9695 8 9696 8 9697 8 9698 8 9699 8 9700 8 9701 8 9702 8 9703 8 9704 8 9705 8 9706 8 9707 8 9708 8 9709 8 9710 8 9711 8 9712 8 9713 8 9714 8 9715 8 9716 8 9717 8 9718 8 9719 8 9720 8 9721 8 9722 8 9723 8 9724 8 9725 8 9726 8 9727 8 9728 8 9729 8 9730 8 9731 8 9732 8 9733 8 9734 8 9735 8 9736 8 9737 8 9738 4196 9739 8 9740 8 9741 8 9742 4197 9743 8 9744 8 9745 4198 9746 4199 9747 4200 9748 8 9749 8 9750 4201 9751 4202 9752 4203 9753 4204 9754 8 9755 4205 9756 8 9757 4206 9758 4207 9759 4208 9760 4209 9761 4210 9762 4211 9763 4212 9764 4213 9765 4214 9766 4144 9767 4144 9768 4144 9769 4144 9770 4144 9771 4144 9772 4144 9773 4144 9774 4144 9775 4144 9776 4144 9777 4144 9778 4144 9779 4215 9780 4144 9781 4216 9782 8 9783 4217 9784 4218 9785 8 9786 4219 9787 8 9788 4220 9789 8 9790 8 9791 8 9792 4221 9793 4222 9794 4223 9795 8 9796 4224 9797 8 9798 8 9799 4225 9800 4226 9801 4227 9802 8 9803 8 9804 8 9805 4228 9806 4229 9807 8 9808 8 9809 8 9810 4230 9811 4231 9812 4232 9813 4233 9814 4234 9815 8 9816 4235 9817 8 9818 8 9819 4236 9820 8 9821 4237 9822 4238 9823 8 9824 8 9825 4239 9826 8 9827 8 9828 4240 9829 4241 9830 8 9831 8 9832 4242 9833 4243 9834 4244 9835 4245 9836 4246 9837 4247 9838 4248 9839 4249 9840 4250 9841 4251 9842 4252 9843 8 9844 4253 9845 4254 9846 4255 9847 4256 9848 4257 9849 4258 9850 4259 9851 4260 9852 4261 9853 4262 9854 8 9855 4263 9856 3960 9857 4264 9858 8 9859 4265 9860 8 9861 8 9862 8 9863 4266 9864 8 9865 8 9866 4267 9867 4268 9868 4269 9869 4270 9870 4271 9871 8 9872 4272 9873 8 9874 8 9875 4273 9876 8 9877 4274 9878 4275 9879 8 9880 8 9881 4276 9882 8 9883 4277 9884 8 9885 4278 9886 4279 9887 4280 9888 4144 9889 4144 9890 4281 9891 4282 9892 4144 9893 4144 9894 4144 9895 4144 9896 8 9897 4283 9898 8 9899 4284 9900 8 9901 8 9902 4144 9903 4144 9904 4144 9905 4144 9906 4285 9907 4286 9908 8 9909 8 9910 8 9911 8 9912 8 9913 8 9914 4287 9915 4288 9916 4289 9917 8 9918 4290 9919 8 9920 4291 9921 4292 9922 4293 9923 4294 9924 4295 9925 4296 9926 8 9927 4297 9928 8 9929 8 9930 8 9931 8 9932 4298 9933 8 9934 8 9935 8 9936 4299 9937 4300 9938 4301 9939 8 9940 8 9941 4302 9942 8 9943 8 9944 4303 9945 4304 9946 4305 9947 4306 9948 4307 9949 4308 9950 8 9951 4309 9952 8 9953 4310 9954 4311 9955 4312 9956 4313 9957 4314 9958 8 9959 4315 9960 4316 9961 4317 9962 4318 9963 4144 9964 4319 9965 4320 9966 4321 9967 8 9968 4144 9969 4144 9970 4144 9971 8 9972 4322 9973 8 9974 4323 9975 8 9976 4324 9977 4325 9978 8 9979 4326 9980 4327 9981 8 9982 4144 9983 8 9984 4144 9985 4144 9986 4144 9987 4328 9988 4329 9989 8 9990 4144 9991 4144 9992 4144 9993 4330 9994 8 9995 8 9996 4331 9997 4332 9998 8 9999 4333 10000 8 10001 4144 10002 3960 10003 3960 10004 3960 10005 8 10006 4334 10007 8 10008 4335 10009 8 10010 8 10011 3960 10012 3960 10013 3960 10014 3960 10015 4336 10016 4337 10017 8 10018 8 10019 8 10020 8 10021 8 10022 8 10023 4338 10024 4339 10025 4340 10026 8 10027 4341 10028 8 10029 4342 10030 4343 10031 4344 10032 4345 10033 4346 10034 4347 10035 8 10036 4348 10037 8 10038 8 10039 8 10040 8 10041 4349 10042 8 10043 8 10044 8 10045 4350 10046 4351 10047 4352 10048 8 10049 8 10050 4353 10051 8 10052 8 10053 4354 10054 4355 10055 4356 10056 4357 10057 4358 10058 4359 10059 8 10060 4360 10061 8 10062 4361 10063 4362 10064 4363 10065 4364 10066 4365 10067 8 10068 4366 10069 4367 10070 4368 10071 4369 10072 3960 10073 4370 10074 4371 10075 4372 10076 8 10077 3960 10078 3960 10079 3960 10080 8 10081 4373 10082 8 10083 4374 10084 8 10085 4375 10086 4376 10087 8 10088 4377 10089 4378 10090 8 10091 3960 10092 8 10093 3960 10094 3960 10095 3960 10096 4379 10097 4380 10098 8 10099 3960 10100 3960 10101 3960 10102 4381 10103 8 10104 8 10105 4382 10106 4383 10107 8 10108 4384 10109 8 10110 3960 10111 4385 10112 353 10113 4386 10114 8 10115 4387 10116 8 10117 8 10118 8 10119 4388 10120 8 10121 8 10122 4389 10123 4390 10124 4391 10125 4392 10126 4393 10127 8 10128 4394 10129 8 10130 8 10131 4395 10132 8 10133 4396 10134 4397 10135 8 10136 8 10137 4398 10138 8 10139 4399 10140 8 10141 4400 10142 4401 10143 4402 10144 457 10145 457 10146 4403 10147 4404 10148 4405 10149 4406 10150 4407 10151 4408 10152 4409 10153 4410 10154 4411 10155 4412 10156 8 10157 4413 10158 8 10159 4414 10160 8 10161 8 10162 4415 10163 4416 10164 4417 10165 4418 10166 8 10167 8 10168 4419 10169 4420 10170 4421 10171 4422 10172 4423 10173 8 10174 4424 10175 8 10176 4425 10177 8 10178 8 10179 8 10180 8 10181 4426 10182 8 10183 4427 10184 4428 10185 4429 10186 4430 10187 4431 10188 4432 10189 8 10190 4433 10191 4434 10192 8 10193 4435 10194 8 10195 4436 10196 4437 10197 4438 10198 4439 10199 4440 10200 4441 10201 4442 10202 8 10203 8 10204 4443 10205 8 10206 8 10207 4444 10208 4445 10209 4446 10210 8 10211 8 10212 8 10213 8 10214 4447 10215 8 10216 8 10217 8 10218 8 10219 4448 10220 4449 10221 4450 10222 4451 10223 4452 10224 4453 10225 4454 10226 4455 10227 4456 10228 4457 10229 4458 10230 4459 10231 4460 10232 4461 10233 4462 10234 4463 10235 4464 10236 4465 10237 4466 10238 4467 10239 4468 10240 4469 10241 4470 10242 4471 10243 4472 10244 4473 10245 4474 10246 4475 10247 4476 10248 4477 10249 4478 10250 4479 10251 4480 10252 4481 10253 4451 10254 4451 10255 4451 10256 4451 10257 4451 10258 4451 10259 4451 10260 4451 10261 4451 10262 4451 10263 4451 10264 4451 10265 4451 10266 4451 10267 4451 10268 4482 10269 4483 10270 8 10271 8 10272 8 10273 8 10274 8 10275 8 10276 4484 10277 8 10278 8 10279 4485 10280 8 10281 4486 10282 4487 10283 4488 10284 4489 10285 4490 10286 4491 10287 8 10288 4492 10289 8 10290 8 10291 8 10292 8 10293 8 10294 4493 10295 4494 10296 8 10297 8 10298 4495 10299 8 10300 4496 10301 8 10302 8 10303 4497 10304 4498 10305 8 10306 4451 10307 4451 10308 4451 10309 4451 10310 4451 10311 4499 10312 4500 10313 4501 10314 4451 10315 4451 10316 4451 10317 4451 10318 4451 10319 4451 10320 4451 10321 4451 10322 4451 10323 4451 10324 4502 10325 8 10326 8 10327 8 10328 8 10329 8 10330 8 10331 8 10332 8 10333 8 10334 8 10335 8 10336 8 10337 8 10338 8 10339 8 10340 8 10341 8 10342 8 10343 8 10344 8 10345 8 10346 8 10347 8 10348 8 10349 8 10350 8 10351 8 10352 8 10353 8 10354 8 10355 8 10356 8 10357 8 10358 8 10359 8 10360 8 10361 8 10362 8 10363 8 10364 8 10365 8 10366 8 10367 8 10368 8 10369 8 10370 8 10371 8 10372 8 10373 8 10374 8 10375 8 10376 8 10377 8 10378 8 10379 8 10380 8 10381 8 10382 8 10383 8 10384 8 10385 8 10386 8 10387 8 10388 8 10389 8 10390 8 10391 8 10392 8 10393 8 10394 8 10395 8 10396 8 10397 8 10398 8 10399 8 10400 8 10401 8 10402 8 10403 8 10404 4503 10405 8 10406 8 10407 8 10408 4504 10409 8 10410 8 10411 4505 10412 4506 10413 4507 10414 8 10415 8 10416 4508 10417 4509 10418 4510 10419 4511 10420 8 10421 4512 10422 8 10423 4513 10424 4514 10425 4515 10426 4516 10427 4517 10428 4518 10429 4519 10430 4520 10431 4521 10432 4451 10433 4451 10434 4451 10435 4451 10436 4451 10437 4451 10438 4451 10439 4451 10440 4451 10441 4451 10442 4451 10443 4451 10444 4451 10445 4522 10446 4451 10447 4523 10448 8 10449 4524 10450 4525 10451 8 10452 4526 10453 8 10454 4527 10455 8 10456 8 10457 8 10458 4528 10459 4529 10460 4530 10461 8 10462 4531 10463 8 10464 8 10465 4532 10466 4533 10467 4534 10468 8 10469 8 10470 8 10471 4535 10472 4536 10473 8 10474 8 10475 8 10476 4537 10477 4538 10478 4539 10479 4540 10480 4541 10481 8 10482 4542 10483 8 10484 8 10485 4543 10486 8 10487 4544 10488 4545 10489 8 10490 8 10491 4546 10492 8 10493 8 10494 4547 10495 4548 10496 8 10497 8 10498 4549 10499 4550 10500 4551 10501 4552 10502 4553 10503 4554 10504 4555 10505 4556 10506 4557 10507 4558 10508 4559 10509 8 10510 4560 10511 4561 10512 4562 10513 4563 10514 4564 10515 4565 10516 4566 10517 4567 10518 4568 10519 4569 10520 8 10521 4570 10522 457 10523 4571 10524 8 10525 4572 10526 8 10527 8 10528 8 10529 4573 10530 8 10531 8 10532 4574 10533 4575 10534 4576 10535 4577 10536 4578 10537 8 10538 4579 10539 8 10540 8 10541 4580 10542 8 10543 4581 10544 4582 10545 8 10546 8 10547 4583 10548 8 10549 4584 10550 8 10551 4585 10552 4586 10553 4587 10554 4451 10555 4451 10556 4588 10557 4589 10558 4451 10559 4451 10560 4451 10561 4451 10562 8 10563 4590 10564 8 10565 4591 10566 8 10567 8 10568 4451 10569 4451 10570 4451 10571 4451 10572 4592 10573 8 10574 4593 10575 4594 10576 4595 10577 8 10578 8 10579 8 10580 4596 10581 4597 10582 4598 10583 8 10584 8 10585 8 10586 4599 10587 4600 10588 4601 10589 4602 10590 4603 10591 4604 10592 8 10593 4605 10594 4606 10595 4607 10596 4608 10597 4609 10598 8 10599 4610 10600 4611 10601 4612 10602 4613 10603 4451 10604 4614 10605 4615 10606 4616 10607 8 10608 4451 10609 4451 10610 4451 10611 8 10612 4617 10613 8 10614 4618 10615 8 10616 4619 10617 4620 10618 8 10619 4621 10620 4622 10621 8 10622 4451 10623 8 10624 4451 10625 4451 10626 4451 10627 4623 10628 4624 10629 8 10630 4451 10631 4451 10632 4451 10633 4625 10634 8 10635 8 10636 4626 10637 4627 10638 8 10639 4628 10640 8 10641 4451 10642 457 10643 457 10644 457 10645 8 10646 4629 10647 8 10648 4630 10649 8 10650 8 10651 457 10652 457 10653 457 10654 457 10655 4631 10656 8 10657 4632 10658 4633 10659 4634 10660 8 10661 8 10662 8 10663 4635 10664 4636 10665 4637 10666 8 10667 8 10668 8 10669 4638 10670 4639 10671 4640 10672 4641 10673 4642 10674 4643 10675 8 10676 4644 10677 4645 10678 4646 10679 4647 10680 4648 10681 8 10682 4649 10683 4650 10684 4651 10685 4652 10686 457 10687 4653 10688 4654 10689 4655 10690 8 10691 457 10692 457 10693 457 10694 8 10695 4656 10696 8 10697 4657 10698 8 10699 4658 10700 4659 10701 8 10702 4660 10703 4661 10704 8 10705 457 10706 8 10707 457 10708 457 10709 457 10710 4662 10711 4663 10712 8 10713 457 10714 353 10715 457 10716 457 10717 4664 10718 8 10719 8 10720 4665 10721 4666 10722 8 10723 4667 10724 8 10725 457 10726 353 10727 353 10728 353 10729 353 10730 353 10731 353 10732 353 10733 353 10734 4668 10735 8 10736 8 10737 8 10738 8 10739 8 10740 8 10741 8 10742 8 10743 8 10744 8 10745 8 10746 8 10747 8 10748 8 10749 8 10750 8 10751 8 10752 8 10753 8 10754 8 10755 8 10756 8 10757 8 10758 8 10759 8 10760 8 10761 8 10762 8 10763 8 10764 8 10765 8 10766 8 10767 8 10768 8 10769 8 10770 8 10771 8 10772 8 10773 8 10774 8 10775 8 10776 8 10777 8 10778 8 10779 8 10780 8 10781 8 10782 8 10783 8 10784 8 10785 8 10786 8 10787 8 10788 8 10789 8 10790 8 10791 8 10792 8 10793 8 10794 8 10795 8 10796 8 10797 8 10798 8 10799 8 10800 8 10801 8 10802 8 10803 8 10804 8 10805 8 10806 8 10807 8 10808 8 10809 8 10810 8 10811 8 10812 8 10813 8 10814 4669 10815 8 10816 8 10817 8 10818 4670 10819 8 10820 8 10821 4671 10822 4672 10823 4673 10824 8 10825 8 10826 4674 10827 4675 10828 4676 10829 4677 10830 8 10831 4678 10832 8 10833 4679 10834 4680 10835 4681 10836 4682 10837 4683 10838 4684 10839 4685 10840 4686 10841 4687 10842 353 10843 353 10844 353 10845 353 10846 353 10847 353 10848 353 10849 353 10850 353 10851 353 10852 353 10853 353 10854 353 10855 4688 10856 353 10857 4689 10858 8 10859 4690 10860 4691 10861 8 10862 4692 10863 8 10864 4693 10865 8 10866 8 10867 8 10868 4694 10869 4695 10870 4696 10871 8 10872 4697 10873 8 10874 8 10875 4698 10876 4699 10877 4700 10878 8 10879 8 10880 8 10881 4701 10882 4702 10883 8 10884 8 10885 8 10886 4703 10887 4704 10888 4705 10889 4706 10890 4707 10891 8 10892 4708 10893 8 10894 8 10895 4709 10896 8 10897 4710 10898 4711 10899 8 10900 8 10901 4712 10902 8 10903 8 10904 4713 10905 4714 10906 8 10907 8 10908 8 10909 4715 10910 4716 10911 4717 10912 4718 10913 4719 10914 4720 10915 4721 10916 4722 10917 4723 10918 4724 10919 4725 10920 8 10921 4726 10922 4727 10923 4728 10924 4729 10925 4730 10926 4731 10927 4732 10928 4733 10929 4734 10930 4735 10931 4736 10932 4737 10933 4738 10934 4739 10935 4740 10936 4741 10937 4742 10938 4743 10939 4744 10940 8 10941 4745 10942 8 10943 4746 10944 8 10945 8 10946 4747 10947 4748 10948 4749 10949 4750 10950 8 10951 8 10952 4751 10953 4752 10954 4753 10955 4754 10956 4755 10957 8 10958 4756 10959 8 10960 4757 10961 8 10962 8 10963 8 10964 8 10965 4758 10966 8 10967 4759 10968 4760 10969 4761 10970 4762 10971 4763 10972 4764 10973 8 10974 4765 10975 4766 10976 8 10977 4767 10978 8 10979 4768 10980 4769 10981 4770 10982 4771 10983 4772 10984 4773 10985 4774 10986 8 10987 8 10988 4775 10989 8 10990 8 10991 4776 10992 4777 10993 4778 10994 8 10995 8 10996 8 10997 8 10998 4779 10999 8 11000 8 11001 8 11002 8 11003 4780 11004 4781 11005 4782 11006 4783 11007 4784 11008 4785 11009 4786 11010 4787 11011 4788 11012 4789 11013 4790 11014 4791 11015 4792 11016 4793 11017 4794 11018 4795 11019 4796 11020 4797 11021 4798 11022 4799 11023 4800 11024 4801 11025 4802 11026 4803 11027 4804 11028 4805 11029 4806 11030 4807 11031 4808 11032 4809 11033 4810 11034 4811 11035 4812 11036 4813 11037 4783 11038 4783 11039 4783 11040 4783 11041 4783 11042 4783 11043 4783 11044 4783 11045 4783 11046 4783 11047 4783 11048 4783 11049 4783 11050 4783 11051 4783 11052 4814 11053 4815 11054 8 11055 8 11056 8 11057 8 11058 8 11059 8 11060 4816 11061 8 11062 8 11063 4817 11064 8 11065 4818 11066 4819 11067 4820 11068 4821 11069 4822 11070 4823 11071 8 11072 4824 11073 8 11074 8 11075 8 11076 8 11077 8 11078 4825 11079 4826 11080 8 11081 8 11082 4827 11083 8 11084 4828 11085 8 11086 8 11087 4829 11088 4830 11089 8 11090 4783 11091 4783 11092 4783 11093 4783 11094 4783 11095 4831 11096 4832 11097 4833 11098 4783 11099 4783 11100 4783 11101 4783 11102 4783 11103 4783 11104 4783 11105 4783 11106 4783 11107 4783 11108 4834 11109 8 11110 8 11111 8 11112 8 11113 8 11114 8 11115 8 11116 8 11117 8 11118 8 11119 8 11120 8 11121 8 11122 8 11123 8 11124 8 11125 8 11126 8 11127 8 11128 8 11129 8 11130 8 11131 8 11132 8 11133 8 11134 8 11135 8 11136 8 11137 8 11138 8 11139 8 11140 8 11141 8 11142 8 11143 8 11144 8 11145 8 11146 8 11147 8 11148 8 11149 8 11150 8 11151 8 11152 8 11153 8 11154 8 11155 8 11156 8 11157 8 11158 8 11159 8 11160 8 11161 8 11162 8 11163 8 11164 8 11165 8 11166 8 11167 8 11168 8 11169 8 11170 8 11171 8 11172 8 11173 8 11174 8 11175 8 11176 8 11177 8 11178 8 11179 8 11180 8 11181 8 11182 8 11183 8 11184 8 11185 8 11186 8 11187 8 11188 4835 11189 8 11190 8 11191 8 11192 4836 11193 8 11194 8 11195 4837 11196 4838 11197 4839 11198 8 11199 8 11200 4840 11201 4841 11202 4842 11203 4843 11204 8 11205 4844 11206 8 11207 4845 11208 4846 11209 4847 11210 4848 11211 4849 11212 4850 11213 4851 11214 4852 11215 4853 11216 4783 11217 4783 11218 4783 11219 4783 11220 4783 11221 4783 11222 4783 11223 4783 11224 4783 11225 4783 11226 4783 11227 4783 11228 4783 11229 4854 11230 4783 11231 4855 11232 8 11233 4856 11234 4857 11235 8 11236 4858 11237 8 11238 4859 11239 8 11240 8 11241 8 11242 4860 11243 4861 11244 4862 11245 8 11246 4863 11247 8 11248 8 11249 4864 11250 4865 11251 4866 11252 8 11253 8 11254 8 11255 4867 11256 4868 11257 8 11258 8 11259 8 11260 4869 11261 4870 11262 4871 11263 4872 11264 4873 11265 8 11266 4874 11267 8 11268 8 11269 4875 11270 8 11271 4876 11272 4877 11273 8 11274 8 11275 4878 11276 8 11277 8 11278 4879 11279 4880 11280 8 11281 8 11282 4881 11283 4882 11284 4883 11285 4884 11286 4885 11287 4886 11288 4887 11289 4888 11290 4889 11291 4890 11292 4891 11293 8 11294 4892 11295 4893 11296 4894 11297 4895 11298 4896 11299 4897 11300 4898 11301 4899 11302 4900 11303 4901 11304 8 11305 8 11306 4902 11307 8 11308 4903 11309 8 11310 8 11311 8 11312 4904 11313 8 11314 8 11315 4905 11316 4906 11317 4907 11318 4908 11319 4909 11320 8 11321 4910 11322 8 11323 8 11324 4911 11325 8 11326 4912 11327 4913 11328 8 11329 8 11330 4914 11331 8 11332 4915 11333 8 11334 4916 11335 4917 11336 4918 11337 4783 11338 4783 11339 4919 11340 4920 11341 4921 11342 4922 11343 4923 11344 4924 11345 4925 11346 4926 11347 4927 11348 4928 11349 8 11350 4929 11351 8 11352 4930 11353 8 11354 8 11355 4931 11356 4932 11357 4933 11358 4934 11359 8 11360 8 11361 4935 11362 4936 11363 4937 11364 4938 11365 4939 11366 8 11367 4940 11368 8 11369 4941 11370 8 11371 8 11372 8 11373 8 11374 4942 11375 8 11376 4943 11377 4944 11378 4945 11379 4946 11380 4947 11381 4948 11382 8 11383 4949 11384 4950 11385 8 11386 4951 11387 8 11388 4952 11389 4953 11390 4954 11391 4955 11392 4956 11393 4957 11394 4958 11395 8 11396 8 11397 4959 11398 8 11399 8 11400 4960 11401 4961 11402 4962 11403 8 11404 8 11405 8 11406 8 11407 4963 11408 8 11409 8 11410 8 11411 8 11412 4964 11413 4965 11414 4966 11415 4967 11416 4968 11417 4969 11418 4970 11419 4971 11420 4972 11421 4973 11422 4974 11423 4975 11424 4976 11425 4977 11426 4978 11427 4979 11428 4980 11429 4981 11430 4982 11431 4983 11432 4984 11433 4985 11434 4986 11435 4987 11436 4988 11437 4989 11438 4990 11439 4991 11440 4992 11441 4993 11442 4994 11443 4995 11444 4996 11445 4997 11446 4967 11447 4967 11448 4967 11449 4967 11450 4967 11451 4967 11452 4967 11453 4967 11454 4967 11455 4967 11456 4967 11457 4967 11458 4967 11459 4967 11460 4967 11461 4998 11462 4999 11463 8 11464 8 11465 8 11466 8 11467 8 11468 8 11469 5000 11470 8 11471 8 11472 5001 11473 8 11474 5002 11475 5003 11476 5004 11477 5005 11478 5006 11479 5007 11480 8 11481 5008 11482 8 11483 8 11484 8 11485 8 11486 8 11487 5009 11488 5010 11489 8 11490 8 11491 5011 11492 8 11493 5012 11494 8 11495 8 11496 5013 11497 5014 11498 8 11499 4967 11500 4967 11501 4967 11502 4967 11503 4967 11504 5015 11505 5016 11506 5017 11507 4967 11508 4967 11509 4967 11510 4967 11511 4967 11512 4967 11513 4967 11514 4967 11515 4967 11516 4967 11517 5018 11518 8 11519 8 11520 8 11521 8 11522 8 11523 8 11524 8 11525 8 11526 8 11527 8 11528 8 11529 8 11530 8 11531 8 11532 8 11533 8 11534 8 11535 8 11536 8 11537 8 11538 8 11539 8 11540 8 11541 8 11542 8 11543 8 11544 8 11545 8 11546 8 11547 8 11548 8 11549 8 11550 8 11551 8 11552 8 11553 8 11554 8 11555 8 11556 8 11557 8 11558 8 11559 8 11560 8 11561 8 11562 8 11563 8 11564 8 11565 8 11566 8 11567 8 11568 8 11569 8 11570 8 11571 8 11572 8 11573 8 11574 8 11575 8 11576 8 11577 8 11578 8 11579 8 11580 8 11581 8 11582 8 11583 8 11584 8 11585 8 11586 8 11587 8 11588 8 11589 8 11590 8 11591 8 11592 8 11593 8 11594 8 11595 8 11596 8 11597 5019 11598 8 11599 8 11600 8 11601 5020 11602 8 11603 8 11604 5021 11605 5022 11606 5023 11607 8 11608 8 11609 5024 11610 5025 11611 5026 11612 5027 11613 8 11614 5028 11615 8 11616 5029 11617 5030 11618 5031 11619 5032 11620 5033 11621 5034 11622 5035 11623 5036 11624 5037 11625 4967 11626 4967 11627 4967 11628 4967 11629 4967 11630 4967 11631 4967 11632 4967 11633 4967 11634 4967 11635 4967 11636 4967 11637 4967 11638 5038 11639 4967 11640 5039 11641 8 11642 5040 11643 5041 11644 8 11645 5042 11646 8 11647 5043 11648 8 11649 8 11650 8 11651 5044 11652 5045 11653 5046 11654 8 11655 5047 11656 8 11657 8 11658 5048 11659 5049 11660 5050 11661 8 11662 8 11663 8 11664 5051 11665 5052 11666 8 11667 8 11668 8 11669 5053 11670 5054 11671 5055 11672 5056 11673 5057 11674 8 11675 5058 11676 8 11677 8 11678 5059 11679 8 11680 5060 11681 5061 11682 8 11683 8 11684 5062 11685 8 11686 8 11687 5063 11688 5064 11689 8 11690 8 11691 5065 11692 5066 11693 5067 11694 5068 11695 5069 11696 5070 11697 5071 11698 5072 11699 5073 11700 5074 11701 5075 11702 8 11703 5076 11704 5077 11705 5078 11706 5079 11707 5080 11708 5081 11709 5082 11710 5083 11711 5084 11712 5085 11713 8 11714 5086 11715 4783 11716 5087 11717 8 11718 5088 11719 8 11720 8 11721 8 11722 5089 11723 8 11724 8 11725 5090 11726 5091 11727 5092 11728 5093 11729 5094 11730 8 11731 5095 11732 8 11733 8 11734 5096 11735 8 11736 5097 11737 5098 11738 8 11739 8 11740 5099 11741 8 11742 5100 11743 8 11744 5101 11745 5102 11746 5103 11747 4967 11748 4967 11749 5104 11750 5105 11751 4967 11752 4967 11753 4967 11754 4967 11755 8 11756 5106 11757 8 11758 5107 11759 8 11760 8 11761 4967 11762 4967 11763 4967 11764 4967 11765 5108 11766 5109 11767 8 11768 8 11769 8 11770 8 11771 8 11772 8 11773 5110 11774 5111 11775 5112 11776 8 11777 5113 11778 8 11779 5114 11780 5115 11781 5116 11782 5117 11783 5118 11784 5119 11785 8 11786 5120 11787 8 11788 8 11789 8 11790 8 11791 5121 11792 8 11793 8 11794 8 11795 5122 11796 5123 11797 5124 11798 8 11799 8 11800 5125 11801 8 11802 8 11803 5126 11804 5127 11805 5128 11806 5129 11807 5130 11808 5131 11809 8 11810 5132 11811 8 11812 5133 11813 5134 11814 5135 11815 5136 11816 5137 11817 8 11818 5138 11819 5139 11820 5140 11821 5141 11822 4967 11823 5142 11824 5143 11825 5144 11826 8 11827 4967 11828 4967 11829 4967 11830 8 11831 5145 11832 8 11833 5146 11834 8 11835 5147 11836 5148 11837 8 11838 5149 11839 5150 11840 8 11841 4967 11842 8 11843 4967 11844 4967 11845 4967 11846 5151 11847 5152 11848 8 11849 4967 11850 4967 11851 4967 11852 5153 11853 8 11854 8 11855 5154 11856 5155 11857 8 11858 5156 11859 8 11860 4967 11861 4783 11862 4783 11863 4783 11864 8 11865 5157 11866 8 11867 5158 11868 8 11869 8 11870 4783 11871 4783 11872 4783 11873 4783 11874 5159 11875 5160 11876 8 11877 8 11878 8 11879 8 11880 8 11881 8 11882 5161 11883 5162 11884 5163 11885 8 11886 5164 11887 8 11888 5165 11889 5166 11890 5167 11891 5168 11892 5169 11893 5170 11894 8 11895 5171 11896 8 11897 8 11898 8 11899 8 11900 5172 11901 8 11902 8 11903 8 11904 5173 11905 5174 11906 5175 11907 8 11908 8 11909 5176 11910 8 11911 8 11912 5177 11913 5178 11914 5179 11915 5180 11916 5181 11917 5182 11918 8 11919 5183 11920 8 11921 5184 11922 5185 11923 5186 11924 5187 11925 5188 11926 8 11927 5189 11928 5190 11929 5191 11930 5192 11931 4783 11932 5193 11933 5194 11934 5195 11935 8 11936 4783 11937 4783 11938 4783 11939 8 11940 5196 11941 8 11942 5197 11943 8 11944 5198 11945 5199 11946 8 11947 5200 11948 5201 11949 8 11950 4783 11951 8 11952 4783 11953 4783 11954 4783 11955 5202 11956 5203 11957 8 11958 4783 11959 4783 11960 4783 11961 5204 11962 8 11963 8 11964 5205 11965 5206 11966 8 11967 5207 11968 8 11969 4783 11970 5208 11971 8 11972 5209 11973 8 11974 8 11975 8 11976 5210 11977 8 11978 8 11979 5211 11980 5212 11981 5213 11982 5214 11983 5215 11984 8 11985 5216 11986 8 11987 8 11988 5217 11989 8 11990 5218 11991 5219 11992 8 11993 8 11994 5220 11995 8 11996 5221 11997 8 11998 5222 11999 5223 12000 5224 12001 353 12002 353 12003 5225 12004 5226 12005 5227 12006 5228 12007 5229 12008 5230 12009 5231 12010 5232 12011 5233 12012 5234 12013 8 12014 5235 12015 8 12016 5236 12017 8 12018 8 12019 5237 12020 5238 12021 5239 12022 5240 12023 8 12024 8 12025 5241 12026 5242 12027 5243 12028 5244 12029 5245 12030 8 12031 5246 12032 8 12033 5247 12034 8 12035 8 12036 8 12037 8 12038 5248 12039 8 12040 5249 12041 5250 12042 5251 12043 5252 12044 5253 12045 5254 12046 8 12047 5255 12048 5256 12049 8 12050 5257 12051 8 12052 5258 12053 5259 12054 5260 12055 5261 12056 5262 12057 5263 12058 5264 12059 8 12060 8 12061 5265 12062 8 12063 8 12064 5266 12065 5267 12066 5268 12067 8 12068 8 12069 8 12070 8 12071 5269 12072 8 12073 8 12074 8 12075 8 12076 5270 12077 5271 12078 5272 12079 5273 12080 5274 12081 5275 12082 5276 12083 5277 12084 5278 12085 5279 12086 5280 12087 5281 12088 5282 12089 5283 12090 5284 12091 5285 12092 5286 12093 5287 12094 5288 12095 5289 12096 5290 12097 5291 12098 5292 12099 5293 12100 5294 12101 5295 12102 5296 12103 5297 12104 5298 12105 5299 12106 5300 12107 5301 12108 5302 12109 5303 12110 5273 12111 5273 12112 5273 12113 5273 12114 5273 12115 5273 12116 5273 12117 5273 12118 5273 12119 5273 12120 5273 12121 5273 12122 5273 12123 5273 12124 5273 12125 5304 12126 5305 12127 8 12128 8 12129 8 12130 8 12131 8 12132 8 12133 5306 12134 8 12135 8 12136 5307 12137 8 12138 5308 12139 5309 12140 5310 12141 5311 12142 5312 12143 5313 12144 8 12145 5314 12146 8 12147 8 12148 8 12149 8 12150 8 12151 5315 12152 5316 12153 8 12154 8 12155 5317 12156 8 12157 5318 12158 8 12159 8 12160 5319 12161 5320 12162 8 12163 5273 12164 5273 12165 5273 12166 5273 12167 5273 12168 5321 12169 5322 12170 5323 12171 5273 12172 5273 12173 5273 12174 5273 12175 5273 12176 5273 12177 5273 12178 5273 12179 5273 12180 5273 12181 5324 12182 8 12183 8 12184 8 12185 8 12186 8 12187 8 12188 8 12189 8 12190 8 12191 8 12192 8 12193 8 12194 8 12195 8 12196 8 12197 8 12198 8 12199 8 12200 8 12201 8 12202 8 12203 8 12204 8 12205 8 12206 8 12207 8 12208 8 12209 8 12210 8 12211 8 12212 8 12213 8 12214 8 12215 8 12216 8 12217 8 12218 8 12219 8 12220 8 12221 8 12222 8 12223 8 12224 8 12225 8 12226 8 12227 8 12228 8 12229 8 12230 8 12231 8 12232 8 12233 8 12234 8 12235 8 12236 8 12237 8 12238 8 12239 8 12240 8 12241 8 12242 8 12243 8 12244 8 12245 8 12246 8 12247 8 12248 8 12249 8 12250 8 12251 8 12252 8 12253 8 12254 8 12255 8 12256 8 12257 8 12258 8 12259 8 12260 8 12261 5325 12262 8 12263 8 12264 8 12265 5326 12266 8 12267 8 12268 5327 12269 5328 12270 5329 12271 8 12272 8 12273 5330 12274 5331 12275 5332 12276 5333 12277 8 12278 5334 12279 8 12280 5335 12281 5336 12282 5337 12283 5338 12284 5339 12285 5340 12286 5341 12287 5342 12288 5343 12289 5273 12290 5273 12291 5273 12292 5273 12293 5273 12294 5273 12295 5273 12296 5273 12297 5273 12298 5273 12299 5273 12300 5273 12301 5273 12302 5344 12303 5273 12304 5345 12305 8 12306 5346 12307 5347 12308 8 12309 5348 12310 8 12311 5349 12312 8 12313 8 12314 8 12315 5350 12316 5351 12317 5352 12318 8 12319 5353 12320 8 12321 8 12322 5354 12323 5355 12324 5356 12325 8 12326 8 12327 8 12328 5357 12329 5358 12330 8 12331 8 12332 8 12333 5359 12334 5360 12335 5361 12336 5362 12337 5363 12338 8 12339 5364 12340 8 12341 8 12342 5365 12343 8 12344 5366 12345 5367 12346 8 12347 8 12348 5368 12349 8 12350 8 12351 5369 12352 5370 12353 8 12354 8 12355 5371 12356 5372 12357 5373 12358 5374 12359 5375 12360 5376 12361 5377 12362 5378 12363 5379 12364 5380 12365 5381 12366 8 12367 5382 12368 5383 12369 5384 12370 5385 12371 5386 12372 5387 12373 5388 12374 5389 12375 5390 12376 5391 12377 8 12378 5392 12379 353 12380 5393 12381 8 12382 5394 12383 8 12384 8 12385 8 12386 5395 12387 8 12388 8 12389 5396 12390 5397 12391 5398 12392 5399 12393 5400 12394 8 12395 5401 12396 8 12397 8 12398 5402 12399 8 12400 5403 12401 5404 12402 8 12403 8 12404 5405 12405 8 12406 5406 12407 8 12408 5407 12409 5408 12410 5409 12411 5273 12412 5273 12413 5410 12414 5411 12415 5273 12416 5273 12417 5273 12418 5273 12419 8 12420 5412 12421 8 12422 5413 12423 8 12424 8 12425 5273 12426 5273 12427 5273 12428 5273 12429 5414 12430 8 12431 5415 12432 5416 12433 5417 12434 8 12435 8 12436 8 12437 5418 12438 5419 12439 5420 12440 8 12441 8 12442 8 12443 5421 12444 5422 12445 5423 12446 5424 12447 5425 12448 5426 12449 8 12450 5427 12451 5428 12452 5429 12453 5430 12454 5431 12455 8 12456 5432 12457 5433 12458 5434 12459 5435 12460 5273 12461 5436 12462 5437 12463 5438 12464 8 12465 5273 12466 5273 12467 5273 12468 8 12469 5439 12470 8 12471 5440 12472 8 12473 5441 12474 5442 12475 8 12476 5443 12477 5444 12478 8 12479 5273 12480 8 12481 5273 12482 5273 12483 5273 12484 5445 12485 5446 12486 8 12487 5273 12488 5273 12489 5273 12490 5447 12491 8 12492 8 12493 5448 12494 5449 12495 8 12496 5450 12497 8 12498 5273 12499 353 12500 353 12501 353 12502 8 12503 5451 12504 8 12505 5452 12506 8 12507 8 12508 353 12509 353 12510 353 12511 353 12512 5453 12513 8 12514 5454 12515 5455 12516 5456 12517 8 12518 8 12519 8 12520 5457 12521 5458 12522 5459 12523 8 12524 8 12525 8 12526 5460 12527 5461 12528 5462 12529 5463 12530 5464 12531 5465 12532 8 12533 5466 12534 5467 12535 5468 12536 5469 12537 5470 12538 8 12539 5471 12540 5472 12541 5473 12542 5474 12543 353 12544 5475 12545 5476 12546 5477 12547 8 12548 353 12549 353 12550 353 12551 8 12552 5478 12553 8 12554 5479 12555 8 12556 5480 12557 5481 12558 8 12559 5482 12560 5483 12561 8 12562 353 12563 8 12564 353 12565 353 12566 353 12567 5484 12568 5485 12569 8 12570 353 12571 353 12572 353 12573 5486 12574 8 12575 8 12576 5487 12577 5488 12578 8 12579 5489 12580 8 12581 353 12582 8 12583 8 12584 8 12585 8 12586 5490 12587 5491 12588 8 12589 8 12590 5492 12591 8 12592 5493 12593 8 12594 8 12595 5494 12596 5495 12597 8 12598 271 12599 271 12600 271 12601 271 12602 271 12603 5496 12604 5497 12605 5498 12606 271 12607 271 12608 271 12609 271 12610 271 12611 271 12612 271 12613 271 12614 271 12615 271 12616 5499 12617 8 12618 8 12619 8 12620 8 12621 8 12622 8 12623 8 12624 8 12625 8 12626 8 12627 8 12628 8 12629 8 12630 8 12631 8 12632 8 12633 8 12634 8 12635 8 12636 8 12637 8 12638 8 12639 8 12640 8 12641 8 12642 8 12643 8 12644 8 12645 8 12646 8 12647 8 12648 8 12649 8 12650 8 12651 8 12652 8 12653 8 12654 8 12655 8 12656 8 12657 8 12658 8 12659 8 12660 8 12661 8 12662 8 12663 8 12664 8 12665 8 12666 8 12667 8 12668 8 12669 8 12670 8 12671 8 12672 8 12673 8 12674 8 12675 8 12676 8 12677 8 12678 8 12679 8 12680 8 12681 8 12682 8 12683 8 12684 8 12685 8 12686 8 12687 8 12688 8 12689 8 12690 8 12691 8 12692 8 12693 8 12694 8 12695 8 12696 5500 12697 8 12698 8 12699 8 12700 5501 12701 8 12702 8 12703 5502 12704 5503 12705 5504 12706 8 12707 8 12708 5505 12709 5506 12710 5507 12711 5508 12712 8 12713 5509 12714 8 12715 5510 12716 5511 12717 5512 12718 5513 12719 5514 12720 5515 12721 5516 12722 5517 12723 5518 12724 271 12725 271 12726 271 12727 271 12728 271 12729 271 12730 271 12731 271 12732 271 12733 271 12734 271 12735 271 12736 271 12737 5519 12738 271 12739 5520 12740 8 12741 5521 12742 5522 12743 8 12744 5523 12745 8 12746 5524 12747 8 12748 8 12749 8 12750 5525 12751 5526 12752 5527 12753 8 12754 5528 12755 8 12756 8 12757 5529 12758 5530 12759 5531 12760 8 12761 8 12762 8 12763 5532 12764 5533 12765 8 12766 8 12767 8 12768 5534 12769 5535 12770 5536 12771 5537 12772 5538 12773 8 12774 5539 12775 8 12776 8 12777 5540 12778 8 12779 5541 12780 5542 12781 8 12782 8 12783 5543 12784 8 12785 8 12786 5544 12787 5545 12788 8 12789 8 12790 5546 12791 5547 12792 5548 12793 5549 12794 5550 12795 5551 12796 5552 12797 5553 12798 5554 12799 5555 12800 5556 12801 8 12802 5557 12803 5558 12804 5559 12805 5560 12806 5561 12807 5562 12808 5563 12809 5564 12810 5565 12811 5566 12812 5567 12813 5568 12814 5569 12815 5570 12816 5571 12817 5572 12818 5573 12819 5574 12820 5575 12821 8 12822 5576 12823 8 12824 5577 12825 8 12826 8 12827 5578 12828 5579 12829 5580 12830 5581 12831 8 12832 8 12833 5582 12834 5583 12835 5584 12836 5585 12837 5586 12838 8 12839 5587 12840 8 12841 5588 12842 8 12843 8 12844 8 12845 8 12846 5589 12847 8 12848 5590 12849 5591 12850 5592 12851 5593 12852 5594 12853 5595 12854 8 12855 5596 12856 5597 12857 8 12858 5598 12859 8 12860 5599 12861 5600 12862 5601 12863 5602 12864 5603 12865 5604 12866 5605 12867 8 12868 8 12869 5606 12870 8 12871 8 12872 5607 12873 5608 12874 5609 12875 8 12876 8 12877 8 12878 8 12879 5610 12880 8 12881 8 12882 8 12883 8 12884 5611 12885 5612 12886 5613 12887 5614 12888 5615 12889 5616 12890 5617 12891 5618 12892 5619 12893 5620 12894 5621 12895 5622 12896 5623 12897 5624 12898 5625 12899 5626 12900 5627 12901 5628 12902 5629 12903 5630 12904 5631 12905 5632 12906 5633 12907 5634 12908 5635 12909 5636 12910 5637 12911 5638 12912 5639 12913 5640 12914 5641 12915 5642 12916 5643 12917 5644 12918 5614 12919 5614 12920 5614 12921 5614 12922 5614 12923 5614 12924 5614 12925 5614 12926 5614 12927 5614 12928 5614 12929 5614 12930 5614 12931 5614 12932 5614 12933 5645 12934 5646 12935 8 12936 8 12937 8 12938 8 12939 8 12940 8 12941 5647 12942 8 12943 8 12944 5648 12945 8 12946 5649 12947 5650 12948 5651 12949 5652 12950 5653 12951 5654 12952 8 12953 5655 12954 8 12955 8 12956 8 12957 8 12958 8 12959 5656 12960 5657 12961 8 12962 8 12963 5658 12964 8 12965 5659 12966 8 12967 8 12968 5660 12969 5661 12970 8 12971 5614 12972 5614 12973 5614 12974 5614 12975 5614 12976 5662 12977 5663 12978 5664 12979 5614 12980 5614 12981 5614 12982 5614 12983 5614 12984 5614 12985 5614 12986 5614 12987 5614 12988 5614 12989 5665 12990 8 12991 8 12992 8 12993 8 12994 8 12995 8 12996 8 12997 8 12998 8 12999 8 13000 8 13001 8 13002 8 13003 8 13004 8 13005 8 13006 8 13007 8 13008 8 13009 8 13010 8 13011 8 13012 8 13013 8 13014 8 13015 8 13016 8 13017 8 13018 8 13019 8 13020 8 13021 8 13022 8 13023 8 13024 8 13025 8 13026 8 13027 8 13028 8 13029 8 13030 8 13031 8 13032 8 13033 8 13034 8 13035 8 13036 8 13037 8 13038 8 13039 8 13040 8 13041 8 13042 8 13043 8 13044 8 13045 8 13046 8 13047 8 13048 8 13049 8 13050 8 13051 8 13052 8 13053 8 13054 8 13055 8 13056 8 13057 8 13058 8 13059 8 13060 8 13061 8 13062 8 13063 8 13064 8 13065 8 13066 8 13067 8 13068 8 13069 5666 13070 8 13071 8 13072 8 13073 5667 13074 8 13075 8 13076 5668 13077 5669 13078 5670 13079 8 13080 8 13081 5671 13082 5672 13083 5673 13084 5674 13085 8 13086 5675 13087 8 13088 5676 13089 5677 13090 5678 13091 5679 13092 5680 13093 5681 13094 5682 13095 5683 13096 5684 13097 5614 13098 5614 13099 5614 13100 5614 13101 5614 13102 5614 13103 5614 13104 5614 13105 5614 13106 5614 13107 5614 13108 5614 13109 5614 13110 5685 13111 5614 13112 5686 13113 8 13114 5687 13115 5688 13116 8 13117 5689 13118 8 13119 5690 13120 8 13121 8 13122 8 13123 5691 13124 5692 13125 5693 13126 8 13127 5694 13128 8 13129 8 13130 5695 13131 5696 13132 5697 13133 8 13134 8 13135 8 13136 5698 13137 5699 13138 8 13139 8 13140 8 13141 5700 13142 5701 13143 5702 13144 5703 13145 5704 13146 8 13147 5705 13148 8 13149 8 13150 5706 13151 8 13152 5707 13153 5708 13154 8 13155 8 13156 5709 13157 8 13158 8 13159 5710 13160 5711 13161 8 13162 8 13163 5712 13164 5713 13165 5714 13166 5715 13167 5716 13168 5717 13169 5718 13170 5719 13171 5720 13172 5721 13173 5722 13174 8 13175 5723 13176 5724 13177 5725 13178 5726 13179 5727 13180 5728 13181 5729 13182 5730 13183 5731 13184 5732 13185 8 13186 8 13187 5733 13188 8 13189 5734 13190 8 13191 8 13192 8 13193 5735 13194 8 13195 8 13196 5736 13197 5737 13198 5738 13199 5739 13200 5740 13201 8 13202 5741 13203 8 13204 8 13205 5742 13206 8 13207 5743 13208 5744 13209 8 13210 8 13211 5745 13212 8 13213 5746 13214 8 13215 5747 13216 5748 13217 5749 13218 5614 13219 5614 13220 5750 13221 5751 13222 5752 13223 5753 13224 5754 13225 5755 13226 5756 13227 5757 13228 5758 13229 5759 13230 8 13231 5760 13232 8 13233 5761 13234 8 13235 8 13236 5762 13237 5763 13238 5764 13239 5765 13240 8 13241 8 13242 5766 13243 5767 13244 5768 13245 5769 13246 5770 13247 8 13248 5771 13249 8 13250 5772 13251 8 13252 8 13253 8 13254 8 13255 5773 13256 8 13257 5774 13258 5775 13259 5776 13260 5777 13261 5778 13262 5779 13263 8 13264 5780 13265 5781 13266 8 13267 5782 13268 8 13269 5783 13270 5784 13271 5785 13272 5786 13273 5787 13274 5788 13275 5789 13276 8 13277 8 13278 5790 13279 8 13280 8 13281 5791 13282 5792 13283 5793 13284 8 13285 8 13286 8 13287 8 13288 5794 13289 8 13290 8 13291 8 13292 8 13293 5795 13294 5796 13295 5797 13296 5798 13297 5799 13298 5800 13299 5801 13300 5802 13301 5803 13302 5804 13303 5805 13304 5806 13305 5807 13306 5808 13307 5809 13308 5810 13309 5811 13310 5812 13311 5813 13312 5814 13313 5815 13314 5816 13315 5817 13316 5818 13317 5819 13318 5820 13319 5821 13320 5822 13321 5823 13322 5824 13323 5825 13324 5826 13325 5827 13326 5828 13327 5798 13328 5798 13329 5798 13330 5798 13331 5798 13332 5798 13333 5798 13334 5798 13335 5798 13336 5798 13337 5798 13338 5798 13339 5798 13340 5798 13341 5798 13342 5829 13343 5830 13344 8 13345 8 13346 8 13347 8 13348 8 13349 8 13350 5831 13351 8 13352 8 13353 5832 13354 8 13355 5833 13356 5834 13357 5835 13358 5836 13359 5837 13360 5838 13361 8 13362 5839 13363 8 13364 8 13365 8 13366 8 13367 8 13368 5840 13369 5841 13370 8 13371 8 13372 5842 13373 8 13374 5843 13375 8 13376 8 13377 5844 13378 5845 13379 8 13380 5798 13381 5798 13382 5798 13383 5798 13384 5798 13385 5846 13386 5847 13387 5848 13388 5798 13389 5798 13390 5798 13391 5798 13392 5798 13393 5798 13394 5798 13395 5798 13396 5798 13397 5798 13398 5849 13399 8 13400 8 13401 8 13402 8 13403 8 13404 8 13405 8 13406 8 13407 8 13408 8 13409 8 13410 8 13411 8 13412 8 13413 8 13414 8 13415 8 13416 8 13417 8 13418 8 13419 8 13420 8 13421 8 13422 8 13423 8 13424 8 13425 8 13426 8 13427 8 13428 8 13429 8 13430 8 13431 8 13432 8 13433 8 13434 8 13435 8 13436 8 13437 8 13438 8 13439 8 13440 8 13441 8 13442 8 13443 8 13444 8 13445 8 13446 8 13447 8 13448 8 13449 8 13450 8 13451 8 13452 8 13453 8 13454 8 13455 8 13456 8 13457 8 13458 8 13459 8 13460 8 13461 8 13462 8 13463 8 13464 8 13465 8 13466 8 13467 8 13468 8 13469 8 13470 8 13471 8 13472 8 13473 8 13474 8 13475 8 13476 8 13477 8 13478 5850 13479 8 13480 8 13481 8 13482 5851 13483 8 13484 8 13485 5852 13486 5853 13487 5854 13488 8 13489 8 13490 5855 13491 5856 13492 5857 13493 5858 13494 8 13495 5859 13496 8 13497 5860 13498 5861 13499 5862 13500 5863 13501 5864 13502 5865 13503 5866 13504 5867 13505 5868 13506 5798 13507 5798 13508 5798 13509 5798 13510 5798 13511 5798 13512 5798 13513 5798 13514 5798 13515 5798 13516 5798 13517 5798 13518 5798 13519 5869 13520 5798 13521 5870 13522 8 13523 5871 13524 5872 13525 8 13526 5873 13527 8 13528 5874 13529 8 13530 8 13531 8 13532 5875 13533 5876 13534 5877 13535 8 13536 5878 13537 8 13538 8 13539 5879 13540 5880 13541 5881 13542 8 13543 8 13544 8 13545 5882 13546 5883 13547 8 13548 8 13549 8 13550 5884 13551 5885 13552 5886 13553 5887 13554 5888 13555 8 13556 5889 13557 8 13558 8 13559 5890 13560 8 13561 5891 13562 5892 13563 8 13564 8 13565 5893 13566 8 13567 8 13568 5894 13569 5895 13570 8 13571 8 13572 5896 13573 5897 13574 5898 13575 5899 13576 5900 13577 5901 13578 5902 13579 5903 13580 5904 13581 5905 13582 5906 13583 8 13584 5907 13585 5908 13586 5909 13587 5910 13588 5911 13589 5912 13590 5913 13591 5914 13592 5915 13593 5916 13594 8 13595 5917 13596 5614 13597 5918 13598 8 13599 5919 13600 8 13601 8 13602 8 13603 5920 13604 8 13605 8 13606 5921 13607 5922 13608 5923 13609 5924 13610 5925 13611 8 13612 5926 13613 8 13614 8 13615 5927 13616 8 13617 5928 13618 5929 13619 8 13620 8 13621 5930 13622 8 13623 5931 13624 8 13625 5932 13626 5933 13627 5934 13628 5798 13629 5798 13630 5935 13631 5936 13632 5798 13633 5798 13634 5798 13635 5798 13636 8 13637 5937 13638 8 13639 5938 13640 8 13641 8 13642 5798 13643 5798 13644 5798 13645 5798 13646 5939 13647 5940 13648 8 13649 8 13650 8 13651 8 13652 8 13653 8 13654 5941 13655 5942 13656 5943 13657 8 13658 5944 13659 8 13660 5945 13661 5946 13662 5947 13663 5948 13664 5949 13665 5950 13666 8 13667 5951 13668 8 13669 8 13670 8 13671 8 13672 5952 13673 8 13674 8 13675 8 13676 5953 13677 5954 13678 5955 13679 8 13680 8 13681 5956 13682 8 13683 8 13684 5957 13685 5958 13686 5959 13687 5960 13688 5961 13689 5962 13690 8 13691 5963 13692 8 13693 5964 13694 5965 13695 5966 13696 5967 13697 5968 13698 8 13699 5969 13700 5970 13701 5971 13702 5972 13703 5798 13704 5973 13705 5974 13706 5975 13707 8 13708 5798 13709 5798 13710 5798 13711 8 13712 5976 13713 8 13714 5977 13715 8 13716 5978 13717 5979 13718 8 13719 5980 13720 5981 13721 8 13722 5798 13723 8 13724 5798 13725 5798 13726 5798 13727 5982 13728 5983 13729 8 13730 5798 13731 5798 13732 5798 13733 5984 13734 8 13735 8 13736 5985 13737 5986 13738 8 13739 5987 13740 8 13741 5798 13742 5614 13743 5614 13744 5614 13745 8 13746 5988 13747 8 13748 5989 13749 8 13750 8 13751 5614 13752 5614 13753 5614 13754 5614 13755 5990 13756 5991 13757 8 13758 8 13759 8 13760 8 13761 8 13762 8 13763 5992 13764 5993 13765 5994 13766 8 13767 5995 13768 8 13769 5996 13770 5997 13771 5998 13772 5999 13773 6000 13774 6001 13775 8 13776 6002 13777 8 13778 8 13779 8 13780 8 13781 6003 13782 8 13783 8 13784 8 13785 6004 13786 6005 13787 6006 13788 8 13789 8 13790 6007 13791 8 13792 8 13793 6008 13794 6009 13795 6010 13796 6011 13797 6012 13798 6013 13799 8 13800 6014 13801 8 13802 6015 13803 6016 13804 6017 13805 6018 13806 6019 13807 8 13808 6020 13809 6021 13810 6022 13811 6023 13812 5614 13813 6024 13814 6025 13815 6026 13816 8 13817 5614 13818 5614 13819 5614 13820 8 13821 6027 13822 8 13823 6028 13824 8 13825 6029 13826 6030 13827 8 13828 6031 13829 6032 13830 8 13831 5614 13832 8 13833 5614 13834 5614 13835 5614 13836 6033 13837 6034 13838 8 13839 5614 13840 5614 13841 5614 13842 6035 13843 8 13844 8 13845 6036 13846 6037 13847 8 13848 6038 13849 8 13850 5614 13851 8 13852 6039 13853 8 13854 6040 13855 8 13856 8 13857 8 13858 6041 13859 8 13860 8 13861 6042 13862 6043 13863 6044 13864 6045 13865 6046 13866 8 13867 6047 13868 8 13869 8 13870 6048 13871 8 13872 6049 13873 6050 13874 8 13875 8 13876 6051 13877 8 13878 6052 13879 8 13880 6053 13881 6054 13882 6055 13883 271 13884 271 13885 6056 13886 6057 13887 6058 13888 6059 13889 6060 13890 6061 13891 6062 13892 6063 13893 6064 13894 6065 13895 8 13896 6066 13897 8 13898 6067 13899 8 13900 8 13901 6068 13902 6069 13903 6070 13904 6071 13905 8 13906 8 13907 6072 13908 6073 13909 6074 13910 6075 13911 6076 13912 8 13913 6077 13914 8 13915 6078 13916 8 13917 8 13918 8 13919 8 13920 6079 13921 8 13922 6080 13923 6081 13924 6082 13925 6083 13926 6084 13927 6085 13928 8 13929 6086 13930 6087 13931 8 13932 6088 13933 8 13934 6089 13935 6090 13936 6091 13937 6092 13938 6093 13939 6094 13940 6095 13941 8 13942 8 13943 6096 13944 8 13945 8 13946 6097 13947 6098 13948 6099 13949 8 13950 8 13951 8 13952 8 13953 6100 13954 8 13955 8 13956 8 13957 8 13958 6101 13959 6102 13960 6103 13961 6104 13962 6105 13963 6106 13964 6107 13965 6108 13966 6109 13967 6110 13968 6111 13969 6112 13970 6113 13971 6114 13972 6115 13973 6116 13974 6117 13975 6118 13976 6119 13977 6120 13978 6121 13979 6122 13980 6123 13981 6124 13982 6125 13983 6126 13984 6127 13985 6128 13986 6129 13987 6130 13988 6131 13989 6132 13990 6133 13991 6134 13992 6104 13993 6104 13994 6104 13995 6104 13996 6104 13997 6104 13998 6104 13999 6104 14000 6104 14001 6104 14002 6104 14003 6104 14004 6104 14005 6104 14006 6104 14007 6135 14008 6136 14009 8 14010 8 14011 8 14012 8 14013 8 14014 8 14015 6137 14016 8 14017 8 14018 6138 14019 8 14020 6139 14021 6140 14022 6141 14023 6142 14024 6143 14025 6144 14026 8 14027 6145 14028 8 14029 8 14030 8 14031 8 14032 8 14033 6146 14034 6147 14035 8 14036 8 14037 6148 14038 8 14039 6149 14040 8 14041 8 14042 6150 14043 6151 14044 8 14045 6104 14046 6104 14047 6104 14048 6104 14049 6104 14050 6152 14051 6153 14052 6154 14053 6104 14054 6104 14055 6104 14056 6104 14057 6104 14058 6104 14059 6104 14060 6104 14061 6104 14062 6104 14063 6155 14064 8 14065 8 14066 8 14067 8 14068 8 14069 8 14070 8 14071 8 14072 8 14073 8 14074 8 14075 8 14076 8 14077 8 14078 8 14079 8 14080 8 14081 8 14082 8 14083 8 14084 8 14085 8 14086 8 14087 8 14088 8 14089 8 14090 8 14091 8 14092 8 14093 8 14094 8 14095 8 14096 8 14097 8 14098 8 14099 8 14100 8 14101 8 14102 8 14103 8 14104 8 14105 8 14106 8 14107 8 14108 8 14109 8 14110 8 14111 8 14112 8 14113 8 14114 8 14115 8 14116 8 14117 8 14118 8 14119 8 14120 8 14121 8 14122 8 14123 8 14124 8 14125 8 14126 8 14127 8 14128 8 14129 8 14130 8 14131 8 14132 8 14133 8 14134 8 14135 8 14136 8 14137 8 14138 8 14139 8 14140 8 14141 8 14142 8 14143 6156 14144 8 14145 8 14146 8 14147 6157 14148 8 14149 8 14150 6158 14151 6159 14152 6160 14153 8 14154 8 14155 6161 14156 6162 14157 6163 14158 6164 14159 8 14160 6165 14161 8 14162 6166 14163 6167 14164 6168 14165 6169 14166 6170 14167 6171 14168 6172 14169 6173 14170 6174 14171 6104 14172 6104 14173 6104 14174 6104 14175 6104 14176 6104 14177 6104 14178 6104 14179 6104 14180 6104 14181 6104 14182 6104 14183 6104 14184 6175 14185 6104 14186 6176 14187 8 14188 6177 14189 6178 14190 8 14191 6179 14192 8 14193 6180 14194 8 14195 8 14196 8 14197 6181 14198 6182 14199 6183 14200 8 14201 6184 14202 8 14203 8 14204 6185 14205 6186 14206 6187 14207 8 14208 8 14209 8 14210 6188 14211 6189 14212 8 14213 8 14214 8 14215 6190 14216 6191 14217 6192 14218 6193 14219 6194 14220 8 14221 6195 14222 8 14223 8 14224 6196 14225 8 14226 6197 14227 6198 14228 8 14229 8 14230 6199 14231 8 14232 8 14233 6200 14234 6201 14235 8 14236 8 14237 6202 14238 6203 14239 6204 14240 6205 14241 6206 14242 6207 14243 6208 14244 6209 14245 6210 14246 6211 14247 6212 14248 8 14249 6213 14250 6214 14251 6215 14252 6216 14253 6217 14254 6218 14255 6219 14256 6220 14257 6221 14258 6222 14259 8 14260 6223 14261 271 14262 6224 14263 8 14264 6225 14265 8 14266 8 14267 8 14268 6226 14269 8 14270 8 14271 6227 14272 6228 14273 6229 14274 6230 14275 6231 14276 8 14277 6232 14278 8 14279 8 14280 6233 14281 8 14282 6234 14283 6235 14284 8 14285 8 14286 6236 14287 8 14288 6237 14289 8 14290 6238 14291 6239 14292 6240 14293 6104 14294 6104 14295 6241 14296 6242 14297 6104 14298 6104 14299 6104 14300 6104 14301 8 14302 6243 14303 8 14304 6244 14305 8 14306 8 14307 6104 14308 6104 14309 6104 14310 6104 14311 6245 14312 8 14313 6246 14314 6247 14315 6248 14316 8 14317 8 14318 8 14319 6249 14320 6250 14321 6251 14322 8 14323 8 14324 8 14325 6252 14326 6253 14327 6254 14328 6255 14329 6256 14330 6257 14331 8 14332 6258 14333 6259 14334 6260 14335 6261 14336 6262 14337 8 14338 6263 14339 6264 14340 6265 14341 6266 14342 6104 14343 6267 14344 6268 14345 6269 14346 8 14347 6104 14348 6104 14349 6104 14350 8 14351 6270 14352 8 14353 6271 14354 8 14355 6272 14356 6273 14357 8 14358 6274 14359 6275 14360 8 14361 6104 14362 8 14363 6104 14364 6104 14365 6104 14366 6276 14367 6277 14368 8 14369 6104 14370 6104 14371 6104 14372 6278 14373 8 14374 8 14375 6279 14376 6280 14377 8 14378 6281 14379 8 14380 6104 14381 271 14382 271 14383 271 14384 8 14385 6282 14386 8 14387 6283 14388 8 14389 8 14390 271 14391 271 14392 271 14393 271 14394 6284 14395 8 14396 6285 14397 6286 14398 6287 14399 8 14400 8 14401 8 14402 6288 14403 6289 14404 6290 14405 8 14406 8 14407 8 14408 6291 14409 6292 14410 6293 14411 6294 14412 6295 14413 6296 14414 8 14415 6297 14416 6298 14417 6299 14418 6300 14419 6301 14420 8 14421 6302 14422 6303 14423 6304 14424 6305 14425 271 14426 6306 14427 6307 14428 6308 14429 8 14430 271 14431 271 14432 271 14433 8 14434 6309 14435 8 14436 6310 14437 8 14438 6311 14439 6312 14440 8 14441 6313 14442 6314 14443 8 14444 271 14445 8 14446 271 14447 271 14448 271 14449 6315 14450 6316 14451 8 14452 271 14453 271 14454 271 14455 6317 14456 8 14457 8 14458 6318 14459 6319 14460 8 14461 6320 14462 8 14463 271 14464 6321 14465 8 14466 8 14467 8 14468 8 14469 8 14470 6322 14471 6323 14472 8 14473 8 14474 6324 14475 8 14476 6325 14477 8 14478 8 14479 6326 14480 6327 14481 8 14482 183 14483 183 14484 183 14485 183 14486 183 14487 6328 14488 6329 14489 6330 14490 183 14491 183 14492 183 14493 183 14494 183 14495 183 14496 183 14497 183 14498 183 14499 183 14500 6331 14501 8 14502 8 14503 8 14504 8 14505 8 14506 8 14507 8 14508 8 14509 8 14510 8 14511 8 14512 8 14513 8 14514 8 14515 8 14516 8 14517 8 14518 8 14519 8 14520 8 14521 8 14522 8 14523 8 14524 8 14525 8 14526 8 14527 8 14528 8 14529 8 14530 8 14531 8 14532 8 14533 8 14534 8 14535 8 14536 8 14537 8 14538 8 14539 8 14540 8 14541 8 14542 8 14543 8 14544 8 14545 8 14546 8 14547 8 14548 8 14549 8 14550 8 14551 8 14552 8 14553 8 14554 8 14555 8 14556 8 14557 8 14558 8 14559 8 14560 8 14561 8 14562 8 14563 8 14564 8 14565 8 14566 8 14567 8 14568 8 14569 8 14570 8 14571 8 14572 8 14573 8 14574 8 14575 8 14576 8 14577 8 14578 8 14579 8 14580 6332 14581 8 14582 8 14583 8 14584 6333 14585 8 14586 8 14587 6334 14588 6335 14589 6336 14590 8 14591 8 14592 6337 14593 6338 14594 6339 14595 6340 14596 8 14597 6341 14598 8 14599 6342 14600 6343 14601 6344 14602 6345 14603 6346 14604 6347 14605 6348 14606 6349 14607 6350 14608 183 14609 183 14610 183 14611 183 14612 183 14613 183 14614 183 14615 183 14616 183 14617 183 14618 183 14619 183 14620 183 14621 6351 14622 183 14623 6352 14624 8 14625 6353 14626 6354 14627 8 14628 6355 14629 8 14630 6356 14631 8 14632 8 14633 8 14634 6357 14635 6358 14636 6359 14637 8 14638 6360 14639 8 14640 8 14641 6361 14642 6362 14643 6363 14644 8 14645 8 14646 8 14647 6364 14648 6365 14649 8 14650 8 14651 8 14652 6366 14653 6367 14654 6368 14655 6369 14656 6370 14657 8 14658 6371 14659 8 14660 8 14661 6372 14662 8 14663 6373 14664 6374 14665 8 14666 8 14667 6375 14668 8 14669 8 14670 6376 14671 6377 14672 8 14673 8 14674 6378 14675 6379 14676 6380 14677 6381 14678 6382 14679 6383 14680 6384 14681 6385 14682 6386 14683 6387 14684 6388 14685 8 14686 6389 14687 6390 14688 6391 14689 6392 14690 6393 14691 6394 14692 6395 14693 6396 14694 6397 14695 6398 14696 6399 14697 6400 14698 6401 14699 6402 14700 6403 14701 6404 14702 6405 14703 6406 14704 6407 14705 8 14706 6408 14707 8 14708 6409 14709 8 14710 8 14711 6410 14712 6411 14713 6412 14714 6413 14715 8 14716 8 14717 6414 14718 6415 14719 6416 14720 6417 14721 6418 14722 8 14723 6419 14724 8 14725 6420 14726 8 14727 8 14728 8 14729 8 14730 6421 14731 8 14732 6422 14733 6423 14734 6424 14735 6425 14736 6426 14737 6427 14738 8 14739 6428 14740 6429 14741 8 14742 6430 14743 8 14744 6431 14745 6432 14746 6433 14747 6434 14748 6435 14749 6436 14750 6437 14751 8 14752 8 14753 6438 14754 8 14755 8 14756 6439 14757 6440 14758 6441 14759 8 14760 8 14761 8 14762 8 14763 6442 14764 8 14765 8 14766 8 14767 8 14768 6443 14769 6444 14770 6445 14771 6446 14772 6447 14773 6448 14774 6449 14775 6450 14776 6451 14777 6452 14778 6453 14779 6454 14780 6455 14781 6456 14782 6457 14783 6458 14784 6459 14785 6460 14786 6461 14787 6462 14788 6463 14789 6464 14790 6465 14791 6466 14792 6467 14793 6468 14794 6469 14795 6470 14796 6471 14797 6472 14798 6473 14799 6474 14800 6475 14801 6476 14802 6446 14803 6446 14804 6446 14805 6446 14806 6446 14807 6446 14808 6446 14809 6446 14810 6446 14811 6446 14812 6446 14813 6446 14814 6446 14815 6446 14816 6446 14817 6477 14818 6478 14819 8 14820 8 14821 8 14822 8 14823 8 14824 8 14825 6479 14826 8 14827 8 14828 6480 14829 8 14830 6481 14831 6482 14832 6483 14833 6484 14834 6485 14835 6486 14836 8 14837 6487 14838 8 14839 8 14840 8 14841 8 14842 8 14843 6488 14844 6489 14845 8 14846 8 14847 6490 14848 8 14849 6491 14850 8 14851 8 14852 6492 14853 6493 14854 8 14855 6446 14856 6446 14857 6446 14858 6446 14859 6446 14860 6494 14861 6495 14862 6496 14863 6446 14864 6446 14865 6446 14866 6446 14867 6446 14868 6446 14869 6446 14870 6446 14871 6446 14872 6446 14873 6497 14874 8 14875 8 14876 8 14877 8 14878 8 14879 8 14880 8 14881 8 14882 8 14883 8 14884 8 14885 8 14886 8 14887 8 14888 8 14889 8 14890 8 14891 8 14892 8 14893 8 14894 8 14895 8 14896 8 14897 8 14898 8 14899 8 14900 8 14901 8 14902 8 14903 8 14904 8 14905 8 14906 8 14907 8 14908 8 14909 8 14910 8 14911 8 14912 8 14913 8 14914 8 14915 8 14916 8 14917 8 14918 8 14919 8 14920 8 14921 8 14922 8 14923 8 14924 8 14925 8 14926 8 14927 8 14928 8 14929 8 14930 8 14931 8 14932 8 14933 8 14934 8 14935 8 14936 8 14937 8 14938 8 14939 8 14940 8 14941 8 14942 8 14943 8 14944 8 14945 8 14946 8 14947 8 14948 8 14949 8 14950 8 14951 8 14952 8 14953 6498 14954 8 14955 8 14956 8 14957 6499 14958 8 14959 8 14960 6500 14961 6501 14962 6502 14963 8 14964 8 14965 6503 14966 6504 14967 6505 14968 6506 14969 8 14970 6507 14971 8 14972 6508 14973 6509 14974 6510 14975 6511 14976 6512 14977 6513 14978 6514 14979 6515 14980 6516 14981 6446 14982 6446 14983 6446 14984 6446 14985 6446 14986 6446 14987 6446 14988 6446 14989 6446 14990 6446 14991 6446 14992 6446 14993 6446 14994 6517 14995 6446 14996 6518 14997 8 14998 6519 14999 6520 15000 8 15001 6521 15002 8 15003 6522 15004 8 15005 8 15006 8 15007 6523 15008 6524 15009 6525 15010 8 15011 6526 15012 8 15013 8 15014 6527 15015 6528 15016 6529 15017 8 15018 8 15019 8 15020 6530 15021 6531 15022 8 15023 8 15024 8 15025 6532 15026 6533 15027 6534 15028 6535 15029 6536 15030 8 15031 6537 15032 8 15033 8 15034 6538 15035 8 15036 6539 15037 6540 15038 8 15039 8 15040 6541 15041 8 15042 8 15043 6542 15044 6543 15045 8 15046 8 15047 6544 15048 6545 15049 6546 15050 6547 15051 6548 15052 6549 15053 6550 15054 6551 15055 6552 15056 6553 15057 6554 15058 8 15059 6555 15060 6556 15061 6557 15062 6558 15063 6559 15064 6560 15065 6561 15066 6562 15067 6563 15068 6564 15069 8 15070 8 15071 6565 15072 8 15073 6566 15074 8 15075 8 15076 8 15077 6567 15078 8 15079 8 15080 6568 15081 6569 15082 6570 15083 6571 15084 6572 15085 8 15086 6573 15087 8 15088 8 15089 6574 15090 8 15091 6575 15092 6576 15093 8 15094 8 15095 6577 15096 8 15097 6578 15098 8 15099 6579 15100 6580 15101 6581 15102 6446 15103 6446 15104 6582 15105 6583 15106 6584 15107 6585 15108 6586 15109 6587 15110 6588 15111 6589 15112 6590 15113 6591 15114 8 15115 6592 15116 8 15117 6593 15118 8 15119 8 15120 6594 15121 6595 15122 6596 15123 6597 15124 8 15125 8 15126 6598 15127 6599 15128 6600 15129 6601 15130 6602 15131 8 15132 6603 15133 8 15134 6604 15135 8 15136 8 15137 8 15138 8 15139 6605 15140 8 15141 6606 15142 6607 15143 6608 15144 6609 15145 6610 15146 6611 15147 8 15148 6612 15149 6613 15150 8 15151 6614 15152 8 15153 6615 15154 6616 15155 6617 15156 6618 15157 6619 15158 6620 15159 6621 15160 8 15161 8 15162 6622 15163 8 15164 8 15165 6623 15166 6624 15167 6625 15168 8 15169 8 15170 8 15171 8 15172 6626 15173 8 15174 8 15175 8 15176 8 15177 6627 15178 6628 15179 6629 15180 6630 15181 6631 15182 6632 15183 6633 15184 6634 15185 6635 15186 6636 15187 6637 15188 6638 15189 6639 15190 6640 15191 6641 15192 6642 15193 6643 15194 6644 15195 6645 15196 6646 15197 6647 15198 6648 15199 6649 15200 6650 15201 6651 15202 6652 15203 6653 15204 6654 15205 6655 15206 6656 15207 6657 15208 6658 15209 6659 15210 6660 15211 6630 15212 6630 15213 6630 15214 6630 15215 6630 15216 6630 15217 6630 15218 6630 15219 6630 15220 6630 15221 6630 15222 6630 15223 6630 15224 6630 15225 6630 15226 6661 15227 6662 15228 8 15229 8 15230 8 15231 8 15232 8 15233 8 15234 6663 15235 8 15236 8 15237 6664 15238 8 15239 6665 15240 6666 15241 6667 15242 6668 15243 6669 15244 6670 15245 8 15246 6671 15247 8 15248 8 15249 8 15250 8 15251 8 15252 6672 15253 6673 15254 8 15255 8 15256 6674 15257 8 15258 6675 15259 8 15260 8 15261 6676 15262 6677 15263 8 15264 6630 15265 6630 15266 6630 15267 6630 15268 6630 15269 6678 15270 6679 15271 6680 15272 6630 15273 6630 15274 6630 15275 6630 15276 6630 15277 6630 15278 6630 15279 6630 15280 6630 15281 6630 15282 6681 15283 8 15284 8 15285 8 15286 8 15287 8 15288 8 15289 8 15290 8 15291 8 15292 8 15293 8 15294 8 15295 8 15296 8 15297 8 15298 8 15299 8 15300 8 15301 8 15302 8 15303 8 15304 8 15305 8 15306 8 15307 8 15308 8 15309 8 15310 8 15311 8 15312 8 15313 8 15314 8 15315 8 15316 8 15317 8 15318 8 15319 8 15320 8 15321 8 15322 8 15323 8 15324 8 15325 8 15326 8 15327 8 15328 8 15329 8 15330 8 15331 8 15332 8 15333 8 15334 8 15335 8 15336 8 15337 8 15338 8 15339 8 15340 8 15341 8 15342 8 15343 8 15344 8 15345 8 15346 8 15347 8 15348 8 15349 8 15350 8 15351 8 15352 8 15353 8 15354 8 15355 8 15356 8 15357 8 15358 8 15359 8 15360 8 15361 8 15362 6682 15363 8 15364 8 15365 8 15366 6683 15367 8 15368 8 15369 6684 15370 6685 15371 6686 15372 8 15373 8 15374 6687 15375 6688 15376 6689 15377 6690 15378 8 15379 6691 15380 8 15381 6692 15382 6693 15383 6694 15384 6695 15385 6696 15386 6697 15387 6698 15388 6699 15389 6700 15390 6630 15391 6630 15392 6630 15393 6630 15394 6630 15395 6630 15396 6630 15397 6630 15398 6630 15399 6630 15400 6630 15401 6630 15402 6630 15403 6701 15404 6630 15405 6702 15406 8 15407 6703 15408 6704 15409 8 15410 6705 15411 8 15412 6706 15413 8 15414 8 15415 8 15416 6707 15417 6708 15418 6709 15419 8 15420 6710 15421 8 15422 8 15423 6711 15424 6712 15425 6713 15426 8 15427 8 15428 8 15429 6714 15430 6715 15431 8 15432 8 15433 8 15434 6716 15435 6717 15436 6718 15437 6719 15438 6720 15439 8 15440 6721 15441 8 15442 8 15443 6722 15444 8 15445 6723 15446 6724 15447 8 15448 8 15449 6725 15450 8 15451 8 15452 6726 15453 6727 15454 8 15455 8 15456 6728 15457 6729 15458 6730 15459 6731 15460 6732 15461 6733 15462 6734 15463 6735 15464 6736 15465 6737 15466 6738 15467 8 15468 6739 15469 6740 15470 6741 15471 6742 15472 6743 15473 6744 15474 6745 15475 6746 15476 6747 15477 6748 15478 8 15479 6749 15480 6446 15481 6750 15482 8 15483 6751 15484 8 15485 8 15486 8 15487 6752 15488 8 15489 8 15490 6753 15491 6754 15492 6755 15493 6756 15494 6757 15495 8 15496 6758 15497 8 15498 8 15499 6759 15500 8 15501 6760 15502 6761 15503 8 15504 8 15505 6762 15506 8 15507 6763 15508 8 15509 6764 15510 6765 15511 6766 15512 6630 15513 6630 15514 6767 15515 6768 15516 6630 15517 6630 15518 6630 15519 6630 15520 8 15521 6769 15522 8 15523 6770 15524 8 15525 8 15526 6630 15527 6630 15528 6630 15529 6630 15530 6771 15531 6772 15532 8 15533 8 15534 8 15535 8 15536 8 15537 8 15538 6773 15539 6774 15540 6775 15541 8 15542 6776 15543 8 15544 6777 15545 6778 15546 6779 15547 6780 15548 6781 15549 6782 15550 8 15551 6783 15552 8 15553 8 15554 8 15555 8 15556 6784 15557 8 15558 8 15559 8 15560 6785 15561 6786 15562 6787 15563 8 15564 8 15565 6788 15566 8 15567 8 15568 6789 15569 6790 15570 6791 15571 6792 15572 6793 15573 6794 15574 8 15575 6795 15576 8 15577 6796 15578 6797 15579 6798 15580 6799 15581 6800 15582 8 15583 6801 15584 6802 15585 6803 15586 6804 15587 6630 15588 6805 15589 6806 15590 6807 15591 8 15592 6630 15593 6630 15594 6630 15595 8 15596 6808 15597 8 15598 6809 15599 8 15600 6810 15601 6811 15602 8 15603 6812 15604 6813 15605 8 15606 6630 15607 8 15608 6630 15609 6630 15610 6630 15611 6814 15612 6815 15613 8 15614 6630 15615 6630 15616 6630 15617 6816 15618 8 15619 8 15620 6817 15621 6818 15622 8 15623 6819 15624 8 15625 6630 15626 6446 15627 6446 15628 6446 15629 8 15630 6820 15631 8 15632 6821 15633 8 15634 8 15635 6446 15636 6446 15637 6446 15638 6446 15639 6822 15640 6823 15641 8 15642 8 15643 8 15644 8 15645 8 15646 8 15647 6824 15648 6825 15649 6826 15650 8 15651 6827 15652 8 15653 6828 15654 6829 15655 6830 15656 6831 15657 6832 15658 6833 15659 8 15660 6834 15661 8 15662 8 15663 8 15664 8 15665 6835 15666 8 15667 8 15668 8 15669 6836 15670 6837 15671 6838 15672 8 15673 8 15674 6839 15675 8 15676 8 15677 6840 15678 6841 15679 6842 15680 6843 15681 6844 15682 6845 15683 8 15684 6846 15685 8 15686 6847 15687 6848 15688 6849 15689 6850 15690 6851 15691 8 15692 6852 15693 6853 15694 6854 15695 6855 15696 6446 15697 6856 15698 6857 15699 6858 15700 8 15701 6446 15702 6446 15703 6446 15704 8 15705 6859 15706 8 15707 6860 15708 8 15709 6861 15710 6862 15711 8 15712 6863 15713 6864 15714 8 15715 6446 15716 8 15717 6446 15718 6446 15719 6446 15720 6865 15721 6866 15722 8 15723 6446 15724 6446 15725 6446 15726 6867 15727 8 15728 8 15729 6868 15730 6869 15731 8 15732 6870 15733 8 15734 6446 15735 6871 15736 8 15737 6872 15738 6873 15739 6874 15740 6875 15741 6876 15742 6877 15743 6878 15744 6879 15745 6880 15746 8 15747 6881 15748 6882 15749 6883 15750 6884 15751 6885 15752 6886 15753 6887 15754 8 15755 6888 15756 6889 15757 8 15758 6890 15759 8 15760 6891 15761 6892 15762 8 15763 6893 15764 8 15765 8 15766 6894 15767 6895 15768 6896 15769 8 15770 6897 15771 6898 15772 6899 15773 6900 15774 6901 15775 8 15776 6902 15777 6903 15778 6904 15779 6905 15780 6906 15781 6907 15782 6908 15783 6909 15784 6910 15785 6911 15786 6912 15787 6913 15788 6914 15789 8 15790 6915 15791 6916 15792 6917 15793 6918 15794 6919 15795 6920 15796 6921 15797 6922 15798 6923 15799 8 15800 6924 15801 8 15802 6925 15803 8 15804 6926 15805 8 15806 6927 15807 6928 15808 8 15809 8 15810 6929 15811 8 15812 8 15813 8 15814 8 15815 6930 15816 6931 15817 8 15818 6932 15819 6933 15820 6934 15821 6935 15822 6936 15823 8 15824 8 15825 8 15826 8 15827 8 15828 6937 15829 6938 15830 6939 15831 6940 15832 6941 15833 6942 15834 6943 15835 6944 15836 6945 15837 6946 15838 6947 15839 8 15840 8 15841 6948 15842 8 15843 6949 15844 6950 15845 6951 15846 6952 15847 6953 15848 6954 15849 8 15850 6955 15851 8 15852 8 15853 8 15854 8 15855 6956 15856 6957 15857 8 15858 6958 15859 6959 15860 8 15861 6960 15862 6961 15863 6962 15864 6963 15865 6964 15866 6965 15867 6966 15868 6967 15869 6968 15870 6969 15871 6970 15872 6971 15873 6972 15874 8 15875 6973 15876 6974 15877 6975 15878 6976 15879 6977 15880 6978 15881 6979 15882 6980 15883 6981 15884 8 15885 6982 15886 8 15887 6983 15888 8 15889 6984 15890 8 15891 6985 15892 6986 15893 8 15894 8 15895 6987 15896 8 15897 8 15898 8 15899 8 15900 6988 15901 6989 15902 8 15903 6990 15904 8 15905 6991 15906 6992 15907 6993 15908 6994 15909 6995 15910 6996 15911 6997 15912 6998 15913 6999 15914 7000 15915 7001 15916 8 15917 8 15918 7002 15919 7003 15920 7004 15921 7005 15922 7006 15923 7007 15924 7008 15925 8 15926 8 15927 7009 15928 7010 15929 8 15930 7011 15931 7012 15932 7013 15933 8 15934 7014 15935 7015 15936 7016 15937 7017 15938 8 15939 8 15940 7018 15941 8 15942 7019 15943 7020 15944 7021 15945 7022 15946 7023 15947 7024 15948 7025 15949 7026 15950 7027 15951 7028 15952 7029 15953 7030 15954 7031 15955 7032 15956 8 15957 7033 15958 7034 15959 7035 15960 7036 15961 7037 15962 7038 15963 7039 15964 7040 15965 7041 15966 8 15967 7042 15968 8 15969 7043 15970 8 15971 7044 15972 8 15973 7045 15974 7046 15975 8 15976 8 15977 7047 15978 8 15979 8 15980 8 15981 8 15982 7048 15983 7049 15984 8 15985 7050 15986 7051 15987 7052 15988 7053 15989 7054 15990 8 15991 7055 15992 7056 15993 7057 15994 7058 15995 7059 15996 7060 15997 7061 15998 7062 15999 7063 16000 8 16001 7064 16002 7065 16003 7066 16004 7067 16005 7068 16006 7069 16007 7070 16008 8 16009 7071 16010 7072 16011 8 16012 7073 16013 8 16014 7074 16015 7075 16016 8 16017 7076 16018 8 16019 8 16020 7077 16021 7078 16022 8 16023 7079 16024 7080 16025 7081 16026 8 16027 7082 16028 7083 16029 7084 16030 7085 16031 8 16032 8 16033 7086 16034 7087 16035 8 16036 7088 16037 7089 16038 8 16039 7090 16040 7091 16041 8 16042 7092 16043 8 16044 7093 16045 8 16046 8 16047 8 16048 7094 16049 8 16050 7095 16051 7096 16052 7097 16053 8 16054 7098 16055 8 16056 8 16057 7099 16058 7100 16059 7101 16060 8 16061 8 16062 7102 16063 7103 16064 7104 16065 8 16066 8 16067 8 16068 7105 16069 7106 16070 8 16071 8 16072 8 16073 7107 16074 7108 16075 7109 16076 7110 16077 7111 16078 8 16079 7112 16080 8 16081 8 16082 7113 16083 8 16084 7114 16085 7115 16086 8 16087 8 16088 7116 16089 8 16090 8 16091 7117 16092 7118 16093 8 16094 7119 16095 7120 16096 7121 16097 7122 16098 7123 16099 7124 16100 8 16101 7125 16102 7126 16103 8 16104 7127 16105 8 16106 8 16107 8 16108 8 16109 8 16110 7128 16111 8 16112 8 16113 8 16114 7129 16115 7130 16116 8 16117 7131 16118 7132 16119 7133 16120 8 16121 7134 16122 7135 16123 7136 16124 7137 16125 8 16126 7138 16127 7139 16128 7140 16129 7141 16130 7142 16131 8 16132 8 16133 8 16134 7143 16135 7144 16136 8 16137 7145 16138 7146 16139 7147 16140 7148 16141 7149 16142 7150 16143 7151 16144 7152 16145 7153 16146 7154 16147 7155 16148 7156 16149 7157 16150 8 16151 7158 16152 7159 16153 7160 16154 7161 16155 7162 16156 7163 16157 7164 16158 7165 16159 7166 16160 8 16161 7167 16162 8 16163 7168 16164 8 16165 7169 16166 8 16167 7170 16168 7171 16169 8 16170 8 16171 7172 16172 8 16173 8 16174 8 16175 8 16176 7173 16177 7174 16178 8 16179 7175 16180 7176 16181 7177 16182 7178 16183 7179 16184 8 16185 7180 16186 7181 16187 7182 16188 7183 16189 7184 16190 7185 16191 7186 16192 7187 16193 7188 16194 8 16195 7189 16196 7190 16197 7191 16198 7192 16199 7193 16200 7194 16201 7195 16202 8 16203 7196 16204 7197 16205 8 16206 7198 16207 8 16208 7199 16209 7200 16210 8 16211 7201 16212 8 16213 8 16214 7202 16215 7203 16216 8 16217 7204 16218 7205 16219 7206 16220 8 16221 7207 16222 7208 16223 7209 16224 7210 16225 8 16226 8 16227 7211 16228 7212 16229 8 16230 7213 16231 7214 16232 7215 16233 7216 16234 8 16235 7217 16236 7218 16237 7219 16238 7220 16239 7221 16240 7222 16241 8 16242 7223 16243 7224 16244 8 16245 7225 16246 7226 16247 7227 16248 7228 16249 7229 16250 7230 16251 8 16252 7231 16253 7232 16254 7233 16255 7234 16256 8 16257 7235 16258 7236 16259 7237 16260 8 16261 7238 16262 8 16263 7239 16264 7240 16265 7241 16266 7242 16267 7243 16268 8 16269 8 16270 8 16271 7244 16272 7245 16273 8 16274 7246 16275 7247 16276 7248 16277 7249 16278 7250 16279 7251 16280 7252 16281 7253 16282 7254 16283 7255 16284 7256 16285 7257 16286 7258 16287 8 16288 7259 16289 7260 16290 7261 16291 7262 16292 7263 16293 7264 16294 7265 16295 7266 16296 7267 16297 8 16298 7268 16299 8 16300 7269 16301 8 16302 7270 16303 8 16304 7271 16305 7272 16306 8 16307 8 16308 7273 16309 8 16310 8 16311 8 16312 8 16313 7274 16314 7275 16315 8 16316 7276 16317 7277 16318 7278 16319 7279 16320 7280 16321 8 16322 7281 16323 7282 16324 7283 16325 7284 16326 7285 16327 7286 16328 7287 16329 7288 16330 7289 16331 8 16332 7290 16333 7291 16334 7292 16335 7293 16336 7294 16337 7295 16338 7296 16339 8 16340 7297 16341 7298 16342 8 16343 7299 16344 8 16345 7300 16346 7301 16347 8 16348 7302 16349 8 16350 8 16351 7303 16352 7304 16353 8 16354 7305 16355 7306 16356 7307 16357 8 16358 7308 16359 7309 16360 7310 16361 7311 16362 8 16363 8 16364 7312 16365 7313 16366 8 16367 7314 16368 7315 16369 7316 16370 7317 16371 8 16372 7318 16373 7319 16374 7320 16375 7321 16376 7322 16377 7323 16378 8 16379 7324 16380 7325 16381 8 16382 7326 16383 7327 16384 7328 16385 7329 16386 7330 16387 7331 16388 8 16389 7332 16390 7333 16391 7334 16392 7335 16393 8 16394 7336 16395 7337 16396 7338 16397 8 16398 7339 16399 8 16400 7340 16401 7341 16402 7342 16403 8 16404 7343 16405 7344 16406 8 16407 8 16408 7345 16409 8 16410 8 16411 7346 16412 7347 16413 8 16414 8 16415 7348 16416 7349 16417 7350 16418 7351 16419 7352 16420 7353 16421 7354 16422 7355 16423 7356 16424 7357 16425 7358 16426 8 16427 7359 16428 8 16429 7360 16430 8 16431 8 16432 7361 16433 7362 16434 7363 16435 7364 16436 8 16437 8 16438 7365 16439 7366 16440 7367 16441 7368 16442 7369 16443 8 16444 7370 16445 8 16446 7371 16447 8 16448 8 16449 8 16450 8 16451 7372 16452 8 16453 7373 16454 7374 16455 7375 16456 7376 16457 7377 16458 7378 16459 8 16460 7379 16461 7380 16462 8 16463 7381 16464 8 16465 7382 16466 7383 16467 7384 16468 7385 16469 7386 16470 7387 16471 7388 16472 8 16473 8 16474 7389 16475 8 16476 8 16477 7390 16478 7391 16479 7392 16480 8 16481 8 16482 8 16483 8 16484 7393 16485 8 16486 8 16487 8 16488 8 16489 7394 16490 7395 16491 7396 16492 7390 16493 7397 16494 7398 16495 7399 16496 7400 16497 7401 16498 7402 16499 7403 16500 7404 16501 7405 16502 7406 16503 7407 16504 7408 16505 7409 16506 7410 16507 7411 16508 7412 16509 7413 16510 7414 16511 7415 16512 7416 16513 7417 16514 7418 16515 7419 16516 7420 16517 7421 16518 7422 16519 7423 16520 7424 16521 7425 16522 7426 16523 7390 16524 7390 16525 7390 16526 7390 16527 7390 16528 7390 16529 7390 16530 7390 16531 7390 16532 7390 16533 7390 16534 7390 16535 7390 16536 7390 16537 7390 16538 7427 16539 7428 16540 8 16541 8 16542 8 16543 8 16544 8 16545 8 16546 7429 16547 8 16548 8 16549 7430 16550 8 16551 7431 16552 7432 16553 7433 16554 7434 16555 7435 16556 7436 16557 8 16558 7437 16559 8 16560 8 16561 8 16562 8 16563 8 16564 7438 16565 7439 16566 8 16567 8 16568 7440 16569 8 16570 7441 16571 8 16572 8 16573 7442 16574 7443 16575 8 16576 7390 16577 7390 16578 7390 16579 7390 16580 7390 16581 7444 16582 7445 16583 7446 16584 7390 16585 7390 16586 7390 16587 7390 16588 7390 16589 7390 16590 7390 16591 7390 16592 7390 16593 7390 16594 7447 16595 8 16596 8 16597 8 16598 8 16599 8 16600 8 16601 8 16602 8 16603 8 16604 8 16605 8 16606 8 16607 8 16608 8 16609 8 16610 8 16611 8 16612 8 16613 8 16614 8 16615 8 16616 8 16617 8 16618 8 16619 8 16620 8 16621 8 16622 8 16623 8 16624 8 16625 8 16626 8 16627 8 16628 8 16629 8 16630 8 16631 8 16632 8 16633 8 16634 8 16635 8 16636 8 16637 8 16638 8 16639 8 16640 8 16641 8 16642 8 16643 8 16644 8 16645 8 16646 8 16647 8 16648 8 16649 8 16650 8 16651 8 16652 8 16653 8 16654 8 16655 8 16656 8 16657 8 16658 8 16659 8 16660 8 16661 8 16662 8 16663 8 16664 8 16665 8 16666 8 16667 8 16668 8 16669 8 16670 8 16671 8 16672 8 16673 8 16674 7448 16675 8 16676 8 16677 8 16678 7449 16679 8 16680 8 16681 7450 16682 7451 16683 7452 16684 8 16685 8 16686 7453 16687 7454 16688 7455 16689 7456 16690 8 16691 7457 16692 8 16693 7458 16694 7459 16695 7460 16696 7461 16697 7462 16698 7463 16699 7464 16700 7465 16701 7466 16702 7390 16703 7390 16704 7390 16705 7390 16706 7390 16707 7390 16708 7390 16709 7390 16710 7390 16711 7390 16712 7390 16713 7390 16714 7390 16715 7467 16716 7390 16717 7468 16718 8 16719 7469 16720 7470 16721 8 16722 7471 16723 8 16724 7472 16725 8 16726 8 16727 8 16728 7473 16729 7474 16730 7475 16731 8 16732 7476 16733 8 16734 8 16735 7477 16736 7478 16737 7479 16738 8 16739 8 16740 8 16741 7480 16742 7481 16743 8 16744 8 16745 8 16746 7482 16747 7483 16748 7484 16749 7485 16750 7486 16751 8 16752 7487 16753 8 16754 8 16755 7488 16756 8 16757 7489 16758 7490 16759 8 16760 8 16761 7491 16762 8 16763 8 16764 7492 16765 7493 16766 8 16767 8 16768 7494 16769 7495 16770 7496 16771 7497 16772 7498 16773 7499 16774 7500 16775 7501 16776 7502 16777 7503 16778 7504 16779 8 16780 7505 16781 7506 16782 7507 16783 7508 16784 7509 16785 7510 16786 7511 16787 7512 16788 7513 16789 7514 16790 8 16791 7515 16792 8 16793 7516 16794 8 16795 8 16796 8 16797 7517 16798 8 16799 8 16800 7518 16801 7519 16802 7520 16803 7521 16804 7522 16805 8 16806 7523 16807 8 16808 8 16809 7524 16810 8 16811 7525 16812 7526 16813 8 16814 8 16815 7527 16816 8 16817 7528 16818 8 16819 7529 16820 7530 16821 7531 16822 7390 16823 7390 16824 7532 16825 7533 16826 7534 16827 7535 16828 7536 16829 7537 16830 7538 16831 7539 16832 7540 16833 7541 16834 8 16835 7542 16836 8 16837 7543 16838 8 16839 8 16840 7544 16841 7545 16842 7546 16843 7547 16844 8 16845 8 16846 7548 16847 7549 16848 7550 16849 7551 16850 7552 16851 8 16852 7553 16853 8 16854 7554 16855 8 16856 8 16857 8 16858 8 16859 7555 16860 8 16861 7556 16862 7557 16863 7558 16864 7559 16865 7560 16866 7561 16867 8 16868 7562 16869 7563 16870 8 16871 7564 16872 8 16873 7565 16874 7566 16875 7567 16876 7568 16877 7569 16878 7570 16879 7571 16880 8 16881 8 16882 7572 16883 8 16884 8 16885 7573 16886 7574 16887 7575 16888 8 16889 8 16890 8 16891 8 16892 7576 16893 8 16894 8 16895 8 16896 8 16897 7577 16898 7578 16899 7579 16900 7580 16901 7581 16902 7582 16903 7583 16904 7584 16905 7585 16906 7586 16907 7587 16908 7588 16909 7589 16910 7590 16911 7591 16912 7592 16913 7593 16914 7594 16915 7595 16916 7596 16917 7597 16918 7598 16919 7599 16920 7600 16921 7601 16922 7602 16923 7603 16924 7604 16925 7605 16926 7606 16927 7607 16928 7608 16929 7609 16930 7610 16931 7580 16932 7580 16933 7580 16934 7580 16935 7580 16936 7580 16937 7580 16938 7580 16939 7580 16940 7580 16941 7580 16942 7580 16943 7580 16944 7580 16945 7580 16946 7611 16947 7612 16948 8 16949 8 16950 8 16951 8 16952 8 16953 8 16954 7613 16955 8 16956 8 16957 7614 16958 8 16959 7615 16960 7616 16961 7617 16962 7618 16963 7619 16964 7620 16965 8 16966 7621 16967 8 16968 8 16969 8 16970 8 16971 8 16972 7622 16973 7623 16974 8 16975 8 16976 7624 16977 8 16978 7625 16979 8 16980 8 16981 7626 16982 7627 16983 8 16984 7580 16985 7580 16986 7580 16987 7580 16988 7580 16989 7628 16990 7629 16991 7630 16992 7580 16993 7580 16994 7580 16995 7580 16996 7580 16997 7580 16998 7580 16999 7580 17000 7580 17001 7580 17002 7631 17003 8 17004 8 17005 8 17006 8 17007 8 17008 8 17009 8 17010 8 17011 8 17012 8 17013 8 17014 8 17015 8 17016 8 17017 8 17018 8 17019 8 17020 8 17021 8 17022 8 17023 8 17024 8 17025 8 17026 8 17027 8 17028 8 17029 8 17030 8 17031 8 17032 8 17033 8 17034 8 17035 8 17036 8 17037 8 17038 8 17039 8 17040 8 17041 8 17042 8 17043 8 17044 8 17045 8 17046 8 17047 8 17048 8 17049 8 17050 8 17051 8 17052 8 17053 8 17054 8 17055 8 17056 8 17057 8 17058 8 17059 8 17060 8 17061 8 17062 8 17063 8 17064 8 17065 8 17066 8 17067 8 17068 8 17069 8 17070 8 17071 8 17072 8 17073 8 17074 8 17075 8 17076 8 17077 8 17078 8 17079 8 17080 8 17081 8 17082 7632 17083 8 17084 8 17085 8 17086 7633 17087 8 17088 8 17089 7634 17090 7635 17091 7636 17092 8 17093 8 17094 7637 17095 7638 17096 7639 17097 7640 17098 8 17099 7641 17100 8 17101 7642 17102 7643 17103 7644 17104 7645 17105 7646 17106 7647 17107 7648 17108 7649 17109 7650 17110 7580 17111 7580 17112 7580 17113 7580 17114 7580 17115 7580 17116 7580 17117 7580 17118 7580 17119 7580 17120 7580 17121 7580 17122 7580 17123 7651 17124 7580 17125 7652 17126 8 17127 7653 17128 7654 17129 8 17130 7655 17131 8 17132 7656 17133 8 17134 8 17135 8 17136 7657 17137 7658 17138 7659 17139 8 17140 7660 17141 8 17142 8 17143 7661 17144 7662 17145 7663 17146 8 17147 8 17148 8 17149 7664 17150 7665 17151 8 17152 8 17153 8 17154 7666 17155 7667 17156 7668 17157 7669 17158 7670 17159 8 17160 7671 17161 8 17162 8 17163 7672 17164 8 17165 7673 17166 7674 17167 8 17168 8 17169 7675 17170 8 17171 8 17172 7676 17173 7677 17174 8 17175 8 17176 7678 17177 7679 17178 7680 17179 7681 17180 7682 17181 7683 17182 7684 17183 7685 17184 7686 17185 7687 17186 7688 17187 8 17188 7689 17189 7690 17190 7691 17191 7692 17192 7693 17193 7694 17194 7695 17195 7696 17196 7697 17197 7698 17198 8 17199 7699 17200 7390 17201 7700 17202 8 17203 7701 17204 8 17205 8 17206 8 17207 7702 17208 8 17209 8 17210 7703 17211 7704 17212 7705 17213 7706 17214 7707 17215 8 17216 7708 17217 8 17218 8 17219 7709 17220 8 17221 7710 17222 7711 17223 8 17224 8 17225 7712 17226 8 17227 7713 17228 8 17229 7714 17230 7715 17231 7716 17232 7580 17233 7580 17234 7717 17235 7718 17236 7580 17237 7580 17238 7580 17239 7580 17240 8 17241 7719 17242 8 17243 7720 17244 8 17245 8 17246 7580 17247 7580 17248 7580 17249 7580 17250 7721 17251 8 17252 7722 17253 7723 17254 7724 17255 8 17256 8 17257 8 17258 7725 17259 7726 17260 7727 17261 8 17262 8 17263 8 17264 7728 17265 7729 17266 7730 17267 7731 17268 7732 17269 7733 17270 8 17271 7734 17272 7735 17273 7736 17274 7737 17275 7738 17276 8 17277 7739 17278 7740 17279 7741 17280 7742 17281 7580 17282 7743 17283 7744 17284 7745 17285 8 17286 7580 17287 7580 17288 7580 17289 8 17290 7746 17291 8 17292 7747 17293 8 17294 7748 17295 7749 17296 8 17297 7750 17298 7751 17299 8 17300 7580 17301 8 17302 7580 17303 7580 17304 7580 17305 7752 17306 7753 17307 8 17308 7580 17309 7580 17310 7580 17311 7754 17312 8 17313 8 17314 7755 17315 7756 17316 8 17317 7757 17318 8 17319 7580 17320 7390 17321 7390 17322 7390 17323 8 17324 7758 17325 8 17326 7759 17327 8 17328 8 17329 7390 17330 7390 17331 7390 17332 7390 17333 7760 17334 8 17335 7761 17336 7762 17337 7763 17338 8 17339 8 17340 8 17341 7764 17342 7765 17343 7766 17344 8 17345 8 17346 8 17347 7767 17348 7768 17349 7769 17350 7770 17351 7771 17352 7772 17353 8 17354 7773 17355 7774 17356 7775 17357 7776 17358 7777 17359 8 17360 7778 17361 7779 17362 7780 17363 7781 17364 7390 17365 7782 17366 7783 17367 7784 17368 8 17369 7390 17370 7390 17371 7390 17372 8 17373 7785 17374 8 17375 7786 17376 8 17377 7787 17378 7788 17379 8 17380 7789 17381 7790 17382 8 17383 7390 17384 8 17385 7390 17386 7390 17387 7390 17388 7791 17389 7792 17390 8 17391 7390 17392 7390 17393 7390 17394 7793 17395 8 17396 8 17397 7794 17398 7795 17399 8 17400 7796 17401 8 17402 7390 17403 83 17404 7797 17405 8 17406 7798 17407 8 17408 7799 17409 8 17410 8 17411 7800 17412 8 17413 7801 17414 8 17415 8 17416 8 17417 7802 17418 7803 17419 8 17420 7804 17421 7805 17422 7806 17423 7807 17424 7808 17425 7809 17426 7810 17427 7811 17428 7812 17429 7813 17430 7814 17431 8 17432 7815 17433 8 17434 7816 17435 8 17436 8 17437 7817 17438 7818 17439 7819 17440 7820 17441 8 17442 8 17443 7821 17444 7822 17445 7823 17446 7824 17447 7825 17448 8 17449 7826 17450 8 17451 7827 17452 8 17453 8 17454 8 17455 8 17456 7828 17457 8 17458 7829 17459 7830 17460 7831 17461 7832 17462 7833 17463 7834 17464 8 17465 7835 17466 7836 17467 8 17468 7837 17469 8 17470 7838 17471 7839 17472 7840 17473 7841 17474 7842 17475 7843 17476 7844 17477 8 17478 8 17479 7845 17480 8 17481 8 17482 7846 17483 7847 17484 7848 17485 8 17486 8 17487 8 17488 8 17489 7849 17490 8 17491 8 17492 8 17493 8 17494 7850 17495 7851 17496 7852 17497 7853 17498 7854 17499 7855 17500 7856 17501 7857 17502 7858 17503 7859 17504 7860 17505 7861 17506 7862 17507 7863 17508 7864 17509 7865 17510 7866 17511 7867 17512 7868 17513 7869 17514 7870 17515 7871 17516 7872 17517 7873 17518 7874 17519 7875 17520 7876 17521 7877 17522 7878 17523 7879 17524 7880 17525 7881 17526 7882 17527 7883 17528 7853 17529 7853 17530 7853 17531 7853 17532 7853 17533 7853 17534 7853 17535 7853 17536 7853 17537 7853 17538 7853 17539 7853 17540 7853 17541 7853 17542 7853 17543 7884 17544 7885 17545 8 17546 8 17547 8 17548 8 17549 8 17550 8 17551 7886 17552 8 17553 8 17554 7887 17555 8 17556 7888 17557 7889 17558 7890 17559 7891 17560 7892 17561 7893 17562 8 17563 7894 17564 8 17565 8 17566 8 17567 8 17568 8 17569 7895 17570 7896 17571 8 17572 8 17573 7897 17574 8 17575 7898 17576 8 17577 8 17578 7899 17579 7900 17580 8 17581 7853 17582 7853 17583 7853 17584 7853 17585 7853 17586 7901 17587 7902 17588 7903 17589 7853 17590 7853 17591 7853 17592 7853 17593 7853 17594 7853 17595 7853 17596 7853 17597 7853 17598 7853 17599 7904 17600 8 17601 8 17602 8 17603 8 17604 8 17605 8 17606 8 17607 8 17608 8 17609 8 17610 8 17611 8 17612 8 17613 8 17614 8 17615 8 17616 8 17617 8 17618 8 17619 8 17620 8 17621 8 17622 8 17623 8 17624 8 17625 8 17626 8 17627 8 17628 8 17629 8 17630 8 17631 8 17632 8 17633 8 17634 8 17635 8 17636 8 17637 8 17638 8 17639 8 17640 8 17641 8 17642 8 17643 8 17644 8 17645 8 17646 8 17647 8 17648 8 17649 8 17650 8 17651 8 17652 8 17653 8 17654 8 17655 8 17656 8 17657 8 17658 8 17659 8 17660 8 17661 8 17662 8 17663 8 17664 8 17665 8 17666 8 17667 8 17668 8 17669 8 17670 8 17671 8 17672 8 17673 8 17674 8 17675 8 17676 8 17677 8 17678 8 17679 7905 17680 8 17681 8 17682 8 17683 7906 17684 8 17685 8 17686 7907 17687 7908 17688 7909 17689 8 17690 8 17691 7910 17692 7911 17693 7912 17694 7913 17695 8 17696 7914 17697 8 17698 7915 17699 7916 17700 7917 17701 7918 17702 7919 17703 7920 17704 7921 17705 7922 17706 7923 17707 7853 17708 7853 17709 7853 17710 7853 17711 7853 17712 7853 17713 7853 17714 7853 17715 7853 17716 7853 17717 7853 17718 7853 17719 7853 17720 7924 17721 7853 17722 7925 17723 8 17724 7926 17725 7927 17726 8 17727 7928 17728 8 17729 7929 17730 8 17731 8 17732 8 17733 7930 17734 7931 17735 7932 17736 8 17737 7933 17738 8 17739 8 17740 7934 17741 7935 17742 7936 17743 8 17744 8 17745 8 17746 7937 17747 7938 17748 8 17749 8 17750 8 17751 7939 17752 7940 17753 7941 17754 7942 17755 7943 17756 8 17757 7944 17758 8 17759 8 17760 7945 17761 8 17762 7946 17763 7947 17764 8 17765 8 17766 7948 17767 8 17768 8 17769 7949 17770 7950 17771 8 17772 8 17773 7951 17774 7952 17775 7953 17776 7954 17777 7955 17778 7956 17779 7957 17780 7958 17781 7959 17782 7960 17783 7961 17784 8 17785 7962 17786 7963 17787 7964 17788 7965 17789 7966 17790 7967 17791 7968 17792 7969 17793 7970 17794 7971 17795 7972 17796 7973 17797 7974 17798 7975 17799 7976 17800 7977 17801 7978 17802 7979 17803 7980 17804 8 17805 7981 17806 8 17807 7982 17808 8 17809 8 17810 7983 17811 7984 17812 7985 17813 7986 17814 8 17815 8 17816 7987 17817 7988 17818 7989 17819 7990 17820 7991 17821 8 17822 7992 17823 8 17824 7993 17825 8 17826 8 17827 8 17828 8 17829 7994 17830 8 17831 7995 17832 7996 17833 7997 17834 7998 17835 7999 17836 8000 17837 8 17838 8001 17839 8002 17840 8 17841 8003 17842 8 17843 8004 17844 8005 17845 8006 17846 8007 17847 8008 17848 8009 17849 8010 17850 8 17851 8 17852 8011 17853 8 17854 8 17855 8012 17856 8013 17857 8014 17858 8 17859 8 17860 8 17861 8 17862 8015 17863 8 17864 8 17865 8 17866 8 17867 8016 17868 8017 17869 8018 17870 8019 17871 8020 17872 8021 17873 8022 17874 8023 17875 8024 17876 8025 17877 8026 17878 8027 17879 8028 17880 8029 17881 8030 17882 8031 17883 8032 17884 8033 17885 8034 17886 8035 17887 8036 17888 8037 17889 8038 17890 8039 17891 8040 17892 8041 17893 8042 17894 8043 17895 8044 17896 8045 17897 8046 17898 8047 17899 8048 17900 8049 17901 8019 17902 8019 17903 8019 17904 8019 17905 8019 17906 8019 17907 8019 17908 8019 17909 8019 17910 8019 17911 8019 17912 8019 17913 8019 17914 8019 17915 8019 17916 8050 17917 8051 17918 8 17919 8 17920 8 17921 8 17922 8 17923 8 17924 8052 17925 8 17926 8 17927 8053 17928 8 17929 8054 17930 8055 17931 8056 17932 8057 17933 8058 17934 8059 17935 8 17936 8060 17937 8 17938 8 17939 8 17940 8 17941 8 17942 8061 17943 8062 17944 8 17945 8 17946 8063 17947 8 17948 8064 17949 8 17950 8 17951 8065 17952 8066 17953 8 17954 8019 17955 8019 17956 8019 17957 8019 17958 8019 17959 8067 17960 8068 17961 8069 17962 8019 17963 8019 17964 8019 17965 8019 17966 8019 17967 8019 17968 8019 17969 8019 17970 8019 17971 8019 17972 8070 17973 8 17974 8 17975 8 17976 8 17977 8 17978 8 17979 8 17980 8 17981 8 17982 8 17983 8 17984 8 17985 8 17986 8 17987 8 17988 8 17989 8 17990 8 17991 8 17992 8 17993 8 17994 8 17995 8 17996 8 17997 8 17998 8 17999 8 18000 8 18001 8 18002 8 18003 8 18004 8 18005 8 18006 8 18007 8 18008 8 18009 8 18010 8 18011 8 18012 8 18013 8 18014 8 18015 8 18016 8 18017 8 18018 8 18019 8 18020 8 18021 8 18022 8 18023 8 18024 8 18025 8 18026 8 18027 8 18028 8 18029 8 18030 8 18031 8 18032 8 18033 8 18034 8 18035 8 18036 8 18037 8 18038 8 18039 8 18040 8 18041 8 18042 8 18043 8 18044 8 18045 8 18046 8 18047 8 18048 8 18049 8 18050 8 18051 8 18052 8071 18053 8 18054 8 18055 8 18056 8072 18057 8 18058 8 18059 8073 18060 8074 18061 8075 18062 8 18063 8 18064 8076 18065 8077 18066 8078 18067 8079 18068 8 18069 8080 18070 8 18071 8081 18072 8082 18073 8083 18074 8084 18075 8085 18076 8086 18077 8087 18078 8088 18079 8089 18080 8019 18081 8019 18082 8019 18083 8019 18084 8019 18085 8019 18086 8019 18087 8019 18088 8019 18089 8019 18090 8019 18091 8019 18092 8019 18093 8090 18094 8019 18095 8091 18096 8 18097 8092 18098 8093 18099 8 18100 8094 18101 8 18102 8095 18103 8 18104 8 18105 8 18106 8096 18107 8097 18108 8098 18109 8 18110 8099 18111 8 18112 8 18113 8100 18114 8101 18115 8102 18116 8 18117 8 18118 8 18119 8103 18120 8104 18121 8 18122 8 18123 8 18124 8105 18125 8106 18126 8107 18127 8108 18128 8109 18129 8 18130 8110 18131 8 18132 8 18133 8111 18134 8 18135 8112 18136 8113 18137 8 18138 8 18139 8114 18140 8 18141 8 18142 8115 18143 8116 18144 8 18145 8 18146 8117 18147 8118 18148 8119 18149 8120 18150 8121 18151 8122 18152 8123 18153 8124 18154 8125 18155 8126 18156 8127 18157 8 18158 8128 18159 8129 18160 8130 18161 8131 18162 8132 18163 8133 18164 8134 18165 8135 18166 8136 18167 8137 18168 8 18169 8 18170 8138 18171 8 18172 8139 18173 8 18174 8 18175 8 18176 8140 18177 8 18178 8 18179 8141 18180 8142 18181 8143 18182 8144 18183 8145 18184 8 18185 8146 18186 8 18187 8 18188 8147 18189 8 18190 8148 18191 8149 18192 8 18193 8 18194 8150 18195 8 18196 8151 18197 8 18198 8152 18199 8153 18200 8154 18201 8019 18202 8019 18203 8155 18204 8156 18205 8157 18206 8158 18207 8159 18208 8160 18209 8161 18210 8162 18211 8163 18212 8164 18213 8 18214 8165 18215 8 18216 8166 18217 8 18218 8 18219 8167 18220 8168 18221 8169 18222 8170 18223 8 18224 8 18225 8171 18226 8172 18227 8173 18228 8174 18229 8175 18230 8 18231 8176 18232 8 18233 8177 18234 8 18235 8 18236 8 18237 8 18238 8178 18239 8 18240 8179 18241 8180 18242 8181 18243 8182 18244 8183 18245 8184 18246 8 18247 8185 18248 8186 18249 8 18250 8187 18251 8 18252 8188 18253 8189 18254 8190 18255 8191 18256 8192 18257 8193 18258 8194 18259 8 18260 8 18261 8195 18262 8 18263 8 18264 8196 18265 8197 18266 8198 18267 8 18268 8 18269 8 18270 8 18271 8199 18272 8 18273 8 18274 8 18275 8 18276 8200 18277 8201 18278 8202 18279 8203 18280 8204 18281 8205 18282 8206 18283 8207 18284 8208 18285 8209 18286 8210 18287 8211 18288 8212 18289 8213 18290 8214 18291 8215 18292 8216 18293 8217 18294 8218 18295 8219 18296 8220 18297 8221 18298 8222 18299 8223 18300 8224 18301 8225 18302 8226 18303 8227 18304 8228 18305 8229 18306 8230 18307 8231 18308 8232 18309 8233 18310 8203 18311 8203 18312 8203 18313 8203 18314 8203 18315 8203 18316 8203 18317 8203 18318 8203 18319 8203 18320 8203 18321 8203 18322 8203 18323 8203 18324 8203 18325 8234 18326 8235 18327 8 18328 8 18329 8 18330 8 18331 8 18332 8 18333 8236 18334 8 18335 8 18336 8237 18337 8 18338 8238 18339 8239 18340 8240 18341 8241 18342 8242 18343 8243 18344 8 18345 8244 18346 8 18347 8 18348 8 18349 8 18350 8 18351 8245 18352 8246 18353 8 18354 8 18355 8247 18356 8 18357 8248 18358 8 18359 8 18360 8249 18361 8250 18362 8 18363 8203 18364 8203 18365 8203 18366 8203 18367 8203 18368 8251 18369 8252 18370 8253 18371 8203 18372 8203 18373 8203 18374 8203 18375 8203 18376 8203 18377 8203 18378 8203 18379 8203 18380 8203 18381 8254 18382 8 18383 8 18384 8 18385 8 18386 8 18387 8 18388 8 18389 8 18390 8 18391 8 18392 8 18393 8 18394 8 18395 8 18396 8 18397 8 18398 8 18399 8 18400 8 18401 8 18402 8 18403 8 18404 8 18405 8 18406 8 18407 8 18408 8 18409 8 18410 8 18411 8 18412 8 18413 8 18414 8 18415 8 18416 8 18417 8 18418 8 18419 8 18420 8 18421 8 18422 8 18423 8 18424 8 18425 8 18426 8 18427 8 18428 8 18429 8 18430 8 18431 8 18432 8 18433 8 18434 8 18435 8 18436 8 18437 8 18438 8 18439 8 18440 8 18441 8 18442 8 18443 8 18444 8 18445 8 18446 8 18447 8 18448 8 18449 8 18450 8 18451 8 18452 8 18453 8 18454 8 18455 8 18456 8 18457 8 18458 8 18459 8 18460 8 18461 8255 18462 8 18463 8 18464 8 18465 8256 18466 8 18467 8 18468 8257 18469 8258 18470 8259 18471 8 18472 8 18473 8260 18474 8261 18475 8262 18476 8263 18477 8 18478 8264 18479 8 18480 8265 18481 8266 18482 8267 18483 8268 18484 8269 18485 8270 18486 8271 18487 8272 18488 8273 18489 8203 18490 8203 18491 8203 18492 8203 18493 8203 18494 8203 18495 8203 18496 8203 18497 8203 18498 8203 18499 8203 18500 8203 18501 8203 18502 8274 18503 8203 18504 8275 18505 8 18506 8276 18507 8277 18508 8 18509 8278 18510 8 18511 8279 18512 8 18513 8 18514 8 18515 8280 18516 8281 18517 8282 18518 8 18519 8283 18520 8 18521 8 18522 8284 18523 8285 18524 8286 18525 8 18526 8 18527 8 18528 8287 18529 8288 18530 8 18531 8 18532 8 18533 8289 18534 8290 18535 8291 18536 8292 18537 8293 18538 8 18539 8294 18540 8 18541 8 18542 8295 18543 8 18544 8296 18545 8297 18546 8 18547 8 18548 8298 18549 8 18550 8 18551 8299 18552 8300 18553 8 18554 8 18555 8301 18556 8302 18557 8303 18558 8304 18559 8305 18560 8306 18561 8307 18562 8308 18563 8309 18564 8310 18565 8311 18566 8 18567 8312 18568 8313 18569 8314 18570 8315 18571 8316 18572 8317 18573 8318 18574 8319 18575 8320 18576 8321 18577 8 18578 8322 18579 8019 18580 8323 18581 8 18582 8324 18583 8 18584 8 18585 8 18586 8325 18587 8 18588 8 18589 8326 18590 8327 18591 8328 18592 8329 18593 8330 18594 8 18595 8331 18596 8 18597 8 18598 8332 18599 8 18600 8333 18601 8334 18602 8 18603 8 18604 8335 18605 8 18606 8336 18607 8 18608 8337 18609 8338 18610 8339 18611 8203 18612 8203 18613 8340 18614 8341 18615 8203 18616 8203 18617 8203 18618 8203 18619 8 18620 8342 18621 8 18622 8343 18623 8 18624 8 18625 8203 18626 8203 18627 8203 18628 8203 18629 8344 18630 8345 18631 8 18632 8 18633 8 18634 8 18635 8 18636 8 18637 8346 18638 8347 18639 8348 18640 8 18641 8349 18642 8 18643 8350 18644 8351 18645 8352 18646 8353 18647 8354 18648 8355 18649 8 18650 8356 18651 8 18652 8 18653 8 18654 8 18655 8357 18656 8 18657 8 18658 8 18659 8358 18660 8359 18661 8360 18662 8 18663 8 18664 8361 18665 8 18666 8 18667 8362 18668 8363 18669 8364 18670 8365 18671 8366 18672 8367 18673 8 18674 8368 18675 8 18676 8369 18677 8370 18678 8371 18679 8372 18680 8373 18681 8 18682 8374 18683 8375 18684 8376 18685 8377 18686 8203 18687 8378 18688 8379 18689 8380 18690 8 18691 8203 18692 8203 18693 8203 18694 8 18695 8381 18696 8 18697 8382 18698 8 18699 8383 18700 8384 18701 8 18702 8385 18703 8386 18704 8 18705 8203 18706 8 18707 8203 18708 8203 18709 8203 18710 8387 18711 8388 18712 8 18713 8203 18714 8203 18715 8203 18716 8389 18717 8 18718 8 18719 8390 18720 8391 18721 8 18722 8392 18723 8 18724 8203 18725 8019 18726 8019 18727 8019 18728 8 18729 8393 18730 8 18731 8394 18732 8 18733 8 18734 8019 18735 8019 18736 8019 18737 8019 18738 8395 18739 8396 18740 8 18741 8 18742 8 18743 8 18744 8 18745 8 18746 8397 18747 8398 18748 8399 18749 8 18750 8400 18751 8 18752 8401 18753 8402 18754 8403 18755 8404 18756 8405 18757 8406 18758 8 18759 8407 18760 8 18761 8 18762 8 18763 8 18764 8408 18765 8 18766 8 18767 8 18768 8409 18769 8410 18770 8411 18771 8 18772 8 18773 8412 18774 8 18775 8 18776 8413 18777 8414 18778 8415 18779 8416 18780 8417 18781 8418 18782 8 18783 8419 18784 8 18785 8420 18786 8421 18787 8422 18788 8423 18789 8424 18790 8 18791 8425 18792 8426 18793 8427 18794 8428 18795 8019 18796 8429 18797 8430 18798 8431 18799 8 18800 8019 18801 8019 18802 8019 18803 8 18804 8432 18805 8 18806 8433 18807 8 18808 8434 18809 8435 18810 8 18811 8436 18812 8437 18813 8 18814 8019 18815 8 18816 8019 18817 8019 18818 8019 18819 8438 18820 8439 18821 8 18822 8019 18823 8019 18824 8019 18825 8440 18826 8 18827 8 18828 8441 18829 8442 18830 8 18831 8443 18832 8 18833 8019 18834 8444 18835 8445 18836 8446 18837 8 18838 8447 18839 8448 18840 8449 18841 8450 18842 8 18843 8451 18844 8452 18845 8 18846 8453 18847 8 18848 8 18849 8 18850 8454 18851 8455 18852 8456 18853 8 18854 8457 18855 8 18856 8 18857 8458 18858 8459 18859 8460 18860 8 18861 8 18862 8 18863 8461 18864 8462 18865 8463 18866 8464 18867 8 18868 8465 18869 8466 18870 8467 18871 8468 18872 8 18873 8469 18874 8470 18875 8 18876 8471 18877 8 18878 8 18879 8 18880 8472 18881 8473 18882 8474 18883 8 18884 8475 18885 8 18886 8 18887 8476 18888 8477 18889 8478 18890 8 18891 8 18892 8 18893 8479 18894 8480 18895 8 18896 8 18897 8481 18898 8 18899 8482 18900 8483 18901 8484 18902 8485 18903 8486 18904 8487 18905 8488 18906 8489 18907 8490 18908 8491 18909 8492 18910 8493 18911 8494 18912 8495 18913 8 18914 8496 18915 8497 18916 8498 18917 8499 18918 8500 18919 8501 18920 8502 18921 8503 18922 8504 18923 8 18924 8505 18925 8 18926 8506 18927 8 18928 8507 18929 8 18930 8508 18931 8509 18932 8 18933 8 18934 8510 18935 8 18936 8 18937 8 18938 8 18939 8511 18940 8512 18941 8 18942 8513 18943 8514 18944 8515 18945 8516 18946 8517 18947 8 18948 8518 18949 8519 18950 8520 18951 8521 18952 8522 18953 8523 18954 8524 18955 8525 18956 8526 18957 8 18958 8527 18959 8528 18960 8529 18961 8530 18962 8531 18963 8532 18964 8533 18965 8 18966 8534 18967 8535 18968 8 18969 8536 18970 8 18971 8537 18972 8538 18973 8 18974 8539 18975 8 18976 8 18977 8540 18978 8541 18979 8 18980 8542 18981 8543 18982 8544 18983 8 18984 8545 18985 8546 18986 8547 18987 8548 18988 8 18989 8 18990 8549 18991 8550 18992 8 18993 8551 18994 8552 18995 8553 18996 8554 18997 8 18998 8555 18999 8556 19000 8557 19001 8558 19002 8559 19003 8560 19004 8 19005 8561 19006 8562 19007 8 19008 8563 19009 8564 19010 8565 19011 8566 19012 8567 19013 8568 19014 8 19015 8569 19016 8570 19017 8571 19018 8572 19019 8 19020 8573 19021 8574 19022 8575 19023 8 19024 8576 19025 8 19026 8577 19027 8578 19028 8579 19029 8 19030 8580 19031 8581 19032 8 19033 8 19034 8582 19035 8 19036 8583 19037 8 19038 8584 19039 8 19040 8585 19041 8 19042 8586 19043 8587 19044 8588 19045 8589 19046 8590 19047 8591 19048 8592 19049 8593 19050 8594 19051 8 19052 8 19053 8595 19054 8596 19055 8 19056 8 19057 8597 19058 7390 19059 8598 19060 8599 19061 8600 19062 7390 19063 8601 19064 8602 19065 8603 19066 1098 19067 8604 19068 8605 19069 8606 19070 1098 19071 8607 19072 8608 19073 8 19074 8 19075 8609 19076 8 19077 8610 19078 8611 19079 8612 19080 8613 19081 8614 19082 8615 19083 8616 19084 8617 19085 8618 19086 8619 19087 8620 19088 8621 19089 8622 19090 8623 19091 8 19092 8624 19093 8625 19094 8626 19095 8627 19096 8628 19097 8629 19098 8630 19099 8631 19100 8632 19101 8 19102 8633 19103 8 19104 8634 19105 8 19106 8635 19107 8 19108 8636 19109 8637 19110 8 19111 8 19112 8638 19113 8 19114 8 19115 8 19116 8 19117 8639 19118 8640 19119 8 19120 8641 19121 8642 19122 8643 19123 8644 19124 8645 19125 8 19126 8646 19127 8647 19128 8648 19129 8649 19130 8650 19131 8651 19132 8652 19133 8653 19134 8654 19135 8 19136 8655 19137 8656 19138 8657 19139 8658 19140 8659 19141 8660 19142 8661 19143 8 19144 8662 19145 8663 19146 8 19147 8664 19148 8 19149 8665 19150 8666 19151 8 19152 8667 19153 8 19154 8 19155 8668 19156 8669 19157 8 19158 8670 19159 8671 19160 8672 19161 8 19162 8673 19163 8674 19164 8675 19165 8676 19166 8 19167 8 19168 8677 19169 8678 19170 8 19171 8679 19172 8680 19173 8681 19174 8682 19175 8 19176 8683 19177 8684 19178 8685 19179 8686 19180 8687 19181 8688 19182 8 19183 8689 19184 8690 19185 8 19186 8691 19187 8692 19188 8693 19189 8694 19190 8695 19191 8696 19192 8 19193 8697 19194 8698 19195 8699 19196 8700 19197 8 19198 8701 19199 8702 19200 8703 19201 8 19202 8704 19203 8 19204 8705 19205 8706 19206 8707 19207 8 19208 8708 19209 8709 19210 8 19211 8 19212 8710 19213 8 19214 8711 19215 8 19216 8712 19217 8 19218 8713 19219 8 19220 8714 19221 8715 19222 8716 19223 8717 19224 8718 19225 8719 19226 8720 19227 8721 19228 8722 19229 8 19230 8 19231 8723 19232 8724 19233 8 19234 8 19235 8725 19236 8726 19237 8727 19238 8728 19239 8729 19240 8730 19241 8731 19242 8732 19243 8 19244 8 19245 8733 19246 8734 19247 8 19248 8 19249 8735 19250 7390 19251 8736 19252 8 19253 8 19254 8 19255 8737 19256 8738 19257 8739 19258 8740 19259 8741 19260 8742 19261 8743 19262 8744 19263 8745 19264 8746 19265 8 19266 8747 19267 8748 19268 8749 19269 8750 19270 8751 19271 8752 19272 8753 19273 8 19274 8754 19275 8755 19276 8 19277 8756 19278 8757 19279 8758 19280 8759 19281 8 19282 8760 19283 8761 19284 8762 19285 8763 19286 8 19287 8764 19288 8765 19289 8766 19290 8767 19291 8768 19292 8 19293 8769 19294 8770 19295 8771 19296 8772 19297 8773 19298 8774 19299 8775 19300 8776 19301 8777 19302 8 19303 8778 19304 8779 19305 8780 19306 8781 19307 8782 19308 8783 19309 8784 19310 8 19311 8785 19312 8786 19313 8 19314 8787 19315 8 19316 8788 19317 8789 19318 8 19319 8790 19320 8 19321 8 19322 8791 19323 8792 19324 8 19325 8793 19326 8794 19327 8795 19328 8 19329 8796 19330 8797 19331 8798 19332 8799 19333 8 19334 8 19335 8800 19336 8801 19337 8 19338 8802 19339 8803 19340 8804 19341 8805 19342 8 19343 8806 19344 8807 19345 8808 19346 8809 19347 8810 19348 8811 19349 8 19350 8812 19351 8813 19352 8 19353 8814 19354 8815 19355 8816 19356 8817 19357 8818 19358 8819 19359 8 19360 8820 19361 8821 19362 8822 19363 8823 19364 8 19365 8824 19366 8825 19367 8826 19368 8 19369 8827 19370 8 19371 8828 19372 8829 19373 8830 19374 8 19375 8831 19376 8832 19377 8 19378 8 19379 8833 19380 8 19381 8834 19382 8 19383 8835 19384 8 19385 8836 19386 8 19387 8837 19388 8838 19389 8839 19390 8840 19391 8841 19392 8842 19393 8843 19394 8844 19395 8845 19396 8 19397 8 19398 8846 19399 8847 19400 8 19401 8 19402 8848 19403 8849 19404 8850 19405 8851 19406 8852 19407 8853 19408 8854 19409 8855 19410 8 19411 8 19412 8856 19413 8857 19414 8 19415 8858 19416 8 19417 8859 19418 8860 19419 8861 19420 8862 19421 8863 19422 8 19423 8864 19424 8865 19425 8 19426 8866 19427 8867 19428 8868 19429 8 19430 8869 19431 8 19432 8 19433 8870 19434 8 19435 8871 19436 8872 19437 8873 19438 8874 19439 8875 19440 8876 19441 8877 19442 8878 19443 8879 19444 8880 19445 8881 19446 8882 19447 8883 19448 8884 19449 8 19450 8885 19451 8886 19452 8887 19453 8888 19454 8889 19455 8890 19456 8891 19457 8892 19458 8893 19459 8 19460 8894 19461 8 19462 8895 19463 8 19464 8896 19465 8 19466 8897 19467 8898 19468 8 19469 8 19470 8899 19471 8 19472 8 19473 8 19474 8 19475 8900 19476 8901 19477 8 19478 8902 19479 8903 19480 8904 19481 8905 19482 8906 19483 8 19484 8907 19485 8908 19486 8909 19487 8910 19488 8911 19489 8912 19490 8913 19491 8914 19492 8915 19493 8 19494 8916 19495 8917 19496 8918 19497 8919 19498 8920 19499 8921 19500 8922 19501 8 19502 8923 19503 8924 19504 8 19505 8925 19506 8 19507 8926 19508 8927 19509 8 19510 8928 19511 8 19512 8 19513 8929 19514 8930 19515 8 19516 8931 19517 8932 19518 8933 19519 8 19520 8934 19521 8935 19522 8936 19523 8937 19524 8 19525 8 19526 8938 19527 8939 19528 8 19529 8940 19530 8941 19531 8942 19532 8943 19533 8 19534 8944 19535 8945 19536 8946 19537 8947 19538 8948 19539 8949 19540 8 19541 8950 19542 8951 19543 8 19544 8952 19545 8953 19546 8954 19547 8955 19548 8956 19549 8957 19550 8 19551 8958 19552 8959 19553 8960 19554 8961 19555 8 19556 8962 19557 8963 19558 8964 19559 8 19560 8965 19561 8 19562 8966 19563 8967 19564 8968 19565 8 19566 8969 19567 8970 19568 8 19569 8 19570 8971 19571 8 19572 8972 19573 8 19574 8973 19575 8 19576 8974 19577 8 19578 8975 19579 8976 19580 8977 19581 8978 19582 8979 19583 8980 19584 8981 19585 8982 19586 8983 19587 8 19588 8 19589 8984 19590 8985 19591 8 19592 8 19593 8986 19594 8987 19595 8988 19596 8989 19597 8990 19598 8991 19599 8992 19600 8993 19601 8 19602 8 19603 8994 19604 8995 19605 8 19606 8996 19607 8 19608 8997 19609 8998 19610 8999 19611 9000 19612 9001 19613 8 19614 9002 19615 9003 19616 9004 19617 9005 19618 8 19619 8 19620 9006 19621 9007 19622 8 19623 9008 19624 8 19625 9009 19626 8 19627 8 19628 9010 19629 9011 19630 9012 19631 9013 19632 9014 19633 9015 19634 9016 19635 9017 19636 9018 19637 9019 19638 9020 19639 8 19640 8 19641 9021 19642 9022 19643 9023 19644 9024 19645 9025 19646 9026 19647 9027 19648 8 19649 8 19650 9028 19651 9029 19652 8 19653 9030 19654 9031 19655 9032 19656 8 19657 9033 19658 9034 19659 9035 19660 9036 19661 8 19662 8 19663 9037 19664 9038 19665 8 19666 9039 19667 9040 19668 9041 19669 9042 19670 8 19671 9043 19672 9044 19673 9045 19674 9046 19675 9047 19676 9048 19677 8 19678 9049 19679 9050 19680 8 19681 9051 19682 9052 19683 9053 19684 9054 19685 9055 19686 9056 19687 8 19688 9057 19689 9058 19690 9059 19691 9060 19692 8 19693 9061 19694 9062 19695 9063 19696 9064 19697 8 19698 9065 19699 9066 19700 9067 19701 8 19702 9068 19703 8 19704 9069 19705 9070 19706 9071 19707 8 19708 9072 19709 9073 19710 8 19711 8 19712 9074 19713 8 19714 9075 19715 8 19716 9076 19717 9077 19718 9078 19719 9079 19720 9080 19721 9081 19722 8 19723 2690 19724 9082 19725 9083 19726 9084 19727 9085 19728 8 19729 6901 19730 9086 19731 9087 19732 8 19733 9088 19734 8 19735 9089 19736 8 19737 9090 19738 9091 19739 9092 19740 9093 19741 9094 19742 9095 19743 9096 19744 9097 19745 9098 19746 8 19747 8 19748 9099 19749 9100 19750 8 19751 8 19752 9101 19753 9102 19754 9103 19755 9104 19756 9105 19757 9106 19758 9107 19759 9108 19760 8 19761 8 19762 9109 19763 9110 19764 8 19765 9111 19766 9112 19767 9113 19768 9114 19769 9115 19770 9116 19771 9117 19772 8 19773 9118 19774 8 19775 8 19776 8 19777 8 19778 8 19779 8 19780 8 19781 9119 19782 8 19783 9120 19784 8 19785 8 19786 9121 19787 9122 19788 8 19789 8 19790 8 19791 8 19792 8 19793 8 19794 8 19795 8 19796 8 19797 8 19798 9123 19799 9124 19800 8 19801 8 19802 8 19803 8 19804 8 19805 8 19806 8 19807 8 19808 8 19809 8 19810 8 19811 9125 19812 8 19813 8 19814 8 19815 8 19816 8 19817 8 19818 8 19819 8 19820 9126 19821 8 19822 9127 19823 9128 19824 9129 19825 9130 19826 9131 19827 8 19828 9132 19829 9133 19830 9134 19831 9135 19832 8 19833 8 19834 9136 19835 9137 19836 8 19837 9138 19838 8 19839 9139 19840 9140 19841 9141 19842 9142 19843 9143 19844 9144 19845 8 19846 9145 19847 8 19848 9146 19849 9147 19850 8 19851 9148 19852 8 19853 9149 19854 8 19855 8 19856 8 19857 9150 19858 9151 19859 8 19860 9152 19861 9153 19862 9154 19863 8 19864 9155 19865 9156 19866 9157 19867 9158 19868 8 19869 8 19870 9159 19871 9160 19872 8 19873 9161 19874 9162 19875 9163 19876 9164 19877 8 19878 9165 19879 9166 19880 9167 19881 9168 19882 9169 19883 9170 19884 8 19885 9171 19886 9172 19887 8 19888 9173 19889 9174 19890 9175 19891 9176 19892 9177 19893 9178 19894 8 19895 9179 19896 9180 19897 9181 19898 9182 19899 8 19900 9183 19901 9184 19902 9185 19903 8 19904 9186 19905 8 19906 9187 19907 9188 19908 9189 19909 8 19910 9190 19911 9191 19912 8 19913 8 19914 9192 19915 8 19916 9193 19917 8 19918 9194 19919 9195 19920 9196 19921 8 19922 9197 19923 8 19924 9198 19925 8 19926 9199 19927 9200 19928 9201 19929 9202 19930 9203 19931 9204 19932 9205 19933 9206 19934 9207 19935 8 19936 8 19937 9208 19938 9209 19939 8 19940 8 19941 9210 19942 9211 19943 9212 19944 9213 19945 9214 19946 9215 19947 9216 19948 9217 19949 8 19950 8 19951 9218 19952 9219 19953 8 19954 9220 19955 8 19956 9221 19957 9222 19958 9223 19959 9224 19960 9225 19961 8 19962 9226 19963 9227 19964 9228 19965 9229 19966 8 19967 8 19968 9230 19969 9231 19970 8 19971 9232 19972 8 19973 9233 19974 8 19975 9234 19976 9235 19977 9236 19978 9237 19979 8 19980 8 19981 9238 19982 9239 19983 8 19984 9240 19985 8 19986 9241 19987 8 19988 8 19989 9242 19990 9243 19991 9244 19992 9245 19993 9246 19994 9247 19995 9248 19996 9249 19997 9250 19998 9251 19999 9252 20000 8 20001 8 20002 9253 20003 9254 20004 9255 20005 9256 20006 9257 20007 9258 20008 9259 20009 8 20010 8 20011 9260 20012 9261 20013 8 20014 9262 20015 9263 20016 9264 20017 8 20018 9265 20019 9266 20020 9267 20021 9268 20022 8 20023 8 20024 9269 20025 9270 20026 8 20027 9271 20028 9272 20029 9273 20030 9274 20031 8 20032 9275 20033 9276 20034 9277 20035 9278 20036 9279 20037 9280 20038 8 20039 9281 20040 9282 20041 8 20042 9283 20043 9284 20044 9285 20045 9286 20046 9287 20047 9288 20048 8 20049 9289 20050 9290 20051 9291 20052 9292 20053 8 20054 9293 20055 9294 20056 9295 20057 9296 20058 8 20059 9297 20060 9298 20061 9299 20062 8 20063 9300 20064 8 20065 9301 20066 9302 20067 9303 20068 8 20069 9304 20070 9305 20071 8 20072 8 20073 9306 20074 8 20075 9307 20076 8 20077 9308 20078 9309 20079 9310 20080 9311 20081 9312 20082 8 20083 9313 20084 9314 20085 9315 20086 8 20087 9316 20088 9317 20089 8 20090 9318 20091 8 20092 9319 20093 8 20094 9320 20095 9321 20096 9322 20097 9323 20098 9324 20099 9325 20100 9326 20101 9327 20102 9328 20103 8 20104 8 20105 9329 20106 9330 20107 8 20108 8 20109 9331 20110 9332 20111 9333 20112 9334 20113 9335 20114 9336 20115 9337 20116 9338 20117 8 20118 8 20119 9339 20120 9340 20121 8 20122 9341 20123 9342 20124 9343 20125 9344 20126 9345 20127 9346 20128 8 20129 8 20130 9347 20131 8 20132 9348 20133 9349 20134 9350 20135 9351 20136 9352 20137 8 20138 9353 20139 9354 20140 9355 20141 9356 20142 8 20143 8 20144 9357 20145 9358 20146 8 20147 9359 20148 8 20149 9360 20150 9361 20151 9362 20152 9363 20153 9364 20154 9365 20155 8 20156 9366 20157 8 20158 9367 20159 8 20160 9368 20161 9369 20162 9370 20163 9371 20164 9372 20165 8 20166 9373 20167 9374 20168 9375 20169 9376 20170 8 20171 8 20172 9377 20173 9378 20174 8 20175 9379 20176 8 20177 9380 20178 8 20179 8 20180 9381 20181 9382 20182 9383 20183 9384 20184 9385 20185 9386 20186 9387 20187 9388 20188 9389 20189 9390 20190 9391 20191 8 20192 8 20193 9392 20194 9393 20195 9394 20196 9395 20197 9396 20198 9397 20199 9398 20200 8 20201 8 20202 9399 20203 9400 20204 8 20205 9401 20206 9402 20207 9403 20208 8 20209 9404 20210 9405 20211 9406 20212 9407 20213 8 20214 8 20215 9408 20216 9409 20217 8 20218 9410 20219 9411 20220 9412 20221 9413 20222 8 20223 9414 20224 9415 20225 9416 20226 9417 20227 9418 20228 9419 20229 8 20230 9420 20231 9421 20232 8 20233 9422 20234 9423 20235 9424 20236 9425 20237 9426 20238 9427 20239 8 20240 9428 20241 9429 20242 9430 20243 9431 20244 8 20245 9432 20246 9433 20247 9434 20248 9435 20249 8 20250 9436 20251 9437 20252 9438 20253 8 20254 9439 20255 8 20256 9440 20257 9441 20258 9442 20259 8 20260 9443 20261 9444 20262 8 20263 8 20264 9445 20265 8 20266 9446 20267 8 20268 9447 20269 9448 20270 9449 20271 9450 20272 9451 20273 8 20274 9452 20275 9453 20276 9454 20277 8 20278 9455 20279 9456 20280 8 20281 9457 20282 8 20283 9458 20284 8 20285 9459 20286 9460 20287 9461 20288 9462 20289 9463 20290 9464 20291 9465 20292 9466 20293 9467 20294 8 20295 8 20296 9468 20297 9469 20298 8 20299 8 20300 9470 20301 9471 20302 9472 20303 9473 20304 9474 20305 9475 20306 9476 20307 9477 20308 8 20309 8 20310 9478 20311 9479 20312 8 20313 9480 20314 9481 20315 9482 20316 9483 20317 9484 20318 9485 20319 8 20320 8 20321 9486 20322 8 20323 9487 20324 9488 20325 9489 20326 9490 20327 9491 20328 8 20329 9492 20330 9493 20331 9494 20332 9495 20333 8 20334 8 20335 9496 20336 9497 20337 8 20338 9498 20339 8 20340 9499 20341 9500 20342 9501 20343 9502 20344 9503 20345 9504 20346 8 20347 9505 20348 8 20349 9506 20350 9507 20351 8 20352 8 20353 8 20354 9508 20355 9509 20356 9510 20357 9511 20358 9512 20359 8 20360 9513 20361 8 20362 8 20363 9514 20364 8 20365 9515 20366 9516 20367 8 20368 8 20369 9517 20370 8 20371 8 20372 9518 20373 8 20374 9519 20375 9520 20376 8 20377 8 20378 9521 20379 9522 20380 8 20381 8 20382 9523 20383 9524 20384 8 20385 8 20386 8 20387 9525 20388 9526 20389 9527 20390 9528 20391 9529 20392 8 20393 9530 20394 8 20395 8 20396 9531 20397 8 20398 9532 20399 9533 20400 8 20401 8 20402 9534 20403 8 20404 8 20405 9535 20406 8 20407 9536 20408 9537 20409 8 20410 8 20411 9538 20412 9539 20413 8 20414 9540 20415 8 20416 9541 20417 8 20418 8 20419 8 20420 9542 20421 8 20422 8 20423 9543 20424 9544 20425 9545 20426 9546 20427 9547 20428 8 20429 9548 20430 8 20431 8 20432 9549 20433 8 20434 9550 20435 9551 20436 8 20437 8 20438 9552 20439 8 20440 9553 20441 8 20442 9554 20443 9555 20444 9556 20445 7853 20446 7853 20447 9557 20448 9558 20449 9559 20450 9560 20451 9561 20452 9562 20453 9563 20454 9564 20455 9565 20456 9566 20457 8 20458 9567 20459 8 20460 9568 20461 8 20462 8 20463 9569 20464 9570 20465 9571 20466 9572 20467 8 20468 8 20469 9573 20470 9574 20471 9575 20472 9576 20473 9577 20474 8 20475 9578 20476 8 20477 9579 20478 8 20479 8 20480 8 20481 8 20482 9580 20483 8 20484 9581 20485 9582 20486 9583 20487 9584 20488 9585 20489 9586 20490 8 20491 9587 20492 9588 20493 8 20494 9589 20495 8 20496 9590 20497 9591 20498 9592 20499 9593 20500 9594 20501 9595 20502 9596 20503 8 20504 8 20505 9597 20506 8 20507 8 20508 9598 20509 9599 20510 9600 20511 8 20512 8 20513 8 20514 8 20515 9601 20516 8 20517 8 20518 8 20519 8 20520 9602 20521 9603 20522 9604 20523 9605 20524 9606 20525 9607 20526 9608 20527 9609 20528 9610 20529 9611 20530 9612 20531 9613 20532 9614 20533 9615 20534 9616 20535 9617 20536 9618 20537 9619 20538 9620 20539 9621 20540 9622 20541 9623 20542 9624 20543 9625 20544 9626 20545 9627 20546 9628 20547 9629 20548 9630 20549 9631 20550 9632 20551 9633 20552 9634 20553 9635 20554 9605 20555 9605 20556 9605 20557 9605 20558 9605 20559 9605 20560 9605 20561 9605 20562 9605 20563 9605 20564 9605 20565 9605 20566 9605 20567 9605 20568 9605 20569 9636 20570 9637 20571 8 20572 8 20573 8 20574 8 20575 8 20576 8 20577 9638 20578 8 20579 8 20580 9639 20581 8 20582 9640 20583 9641 20584 9642 20585 9643 20586 9644 20587 9645 20588 8 20589 9646 20590 8 20591 8 20592 8 20593 8 20594 8 20595 9647 20596 9648 20597 8 20598 8 20599 9649 20600 8 20601 9650 20602 8 20603 8 20604 9651 20605 9652 20606 8 20607 9605 20608 9605 20609 9605 20610 9605 20611 9605 20612 9653 20613 9654 20614 9655 20615 9605 20616 9605 20617 9605 20618 9605 20619 9605 20620 9605 20621 9605 20622 9605 20623 9605 20624 9605 20625 9656 20626 8 20627 8 20628 8 20629 8 20630 8 20631 8 20632 8 20633 8 20634 8 20635 8 20636 8 20637 8 20638 8 20639 8 20640 8 20641 8 20642 8 20643 8 20644 8 20645 8 20646 8 20647 8 20648 8 20649 8 20650 8 20651 8 20652 8 20653 8 20654 8 20655 8 20656 8 20657 8 20658 8 20659 8 20660 8 20661 8 20662 8 20663 8 20664 8 20665 8 20666 8 20667 8 20668 8 20669 8 20670 8 20671 8 20672 8 20673 8 20674 8 20675 8 20676 8 20677 8 20678 8 20679 8 20680 8 20681 8 20682 8 20683 8 20684 8 20685 8 20686 8 20687 8 20688 8 20689 8 20690 8 20691 8 20692 8 20693 8 20694 8 20695 8 20696 8 20697 8 20698 8 20699 8 20700 8 20701 8 20702 8 20703 8 20704 8 20705 9657 20706 8 20707 8 20708 8 20709 9658 20710 8 20711 8 20712 9659 20713 9660 20714 9661 20715 8 20716 8 20717 9662 20718 9663 20719 9664 20720 9665 20721 8 20722 9666 20723 8 20724 9667 20725 9668 20726 9669 20727 9670 20728 9671 20729 9672 20730 9673 20731 9674 20732 9675 20733 9605 20734 9605 20735 9605 20736 9605 20737 9605 20738 9605 20739 9605 20740 9605 20741 9605 20742 9605 20743 9605 20744 9605 20745 9605 20746 9676 20747 9605 20748 9677 20749 8 20750 9678 20751 9679 20752 8 20753 9680 20754 8 20755 9681 20756 8 20757 8 20758 8 20759 9682 20760 9683 20761 9684 20762 8 20763 9685 20764 8 20765 8 20766 9686 20767 9687 20768 9688 20769 8 20770 8 20771 8 20772 9689 20773 9690 20774 8 20775 8 20776 8 20777 9691 20778 9692 20779 9693 20780 9694 20781 9695 20782 8 20783 9696 20784 8 20785 8 20786 9697 20787 8 20788 9698 20789 9699 20790 8 20791 8 20792 9700 20793 8 20794 8 20795 9701 20796 9702 20797 8 20798 8 20799 9703 20800 9704 20801 9705 20802 9706 20803 9707 20804 9708 20805 9709 20806 9710 20807 9711 20808 9712 20809 9713 20810 8 20811 9714 20812 9715 20813 9716 20814 9717 20815 9718 20816 9719 20817 9720 20818 9721 20819 9722 20820 9723 20821 8 20822 9724 20823 7853 20824 9725 20825 8 20826 9726 20827 8 20828 8 20829 8 20830 9727 20831 8 20832 8 20833 9728 20834 9729 20835 9730 20836 9731 20837 9732 20838 8 20839 9733 20840 8 20841 8 20842 9734 20843 8 20844 9735 20845 9736 20846 8 20847 8 20848 9737 20849 8 20850 9738 20851 8 20852 9739 20853 9740 20854 9741 20855 9605 20856 9605 20857 9742 20858 9743 20859 9605 20860 9605 20861 9605 20862 9605 20863 8 20864 9744 20865 8 20866 9745 20867 8 20868 8 20869 9605 20870 9605 20871 9605 20872 9605 20873 9746 20874 8 20875 9747 20876 9748 20877 9749 20878 8 20879 8 20880 8 20881 9750 20882 9751 20883 9752 20884 8 20885 8 20886 8 20887 9753 20888 9754 20889 9755 20890 9756 20891 9757 20892 9758 20893 8 20894 9759 20895 9760 20896 9761 20897 9762 20898 9763 20899 8 20900 9764 20901 9765 20902 9766 20903 9767 20904 9605 20905 9768 20906 9769 20907 9770 20908 8 20909 9605 20910 9605 20911 9605 20912 8 20913 9771 20914 8 20915 9772 20916 8 20917 9773 20918 9774 20919 8 20920 9775 20921 9776 20922 8 20923 9605 20924 8 20925 9605 20926 9605 20927 9605 20928 9777 20929 9778 20930 8 20931 9605 20932 9605 20933 9605 20934 9779 20935 8 20936 8 20937 9780 20938 9781 20939 8 20940 9782 20941 8 20942 9605 20943 7853 20944 7853 20945 7853 20946 8 20947 9783 20948 8 20949 9784 20950 8 20951 8 20952 7853 20953 7853 20954 7853 20955 7853 20956 9785 20957 8 20958 9786 20959 9787 20960 9788 20961 8 20962 8 20963 8 20964 9789 20965 9790 20966 9791 20967 8 20968 8 20969 8 20970 9792 20971 9793 20972 9794 20973 9795 20974 9796 20975 9797 20976 8 20977 9798 20978 9799 20979 9800 20980 9801 20981 9802 20982 8 20983 9803 20984 9804 20985 9805 20986 9806 20987 7853 20988 9807 20989 9808 20990 9809 20991 8 20992 7853 20993 7853 20994 7853 20995 8 20996 9810 20997 8 20998 9811 20999 8 21000 9812 21001 9813 21002 8 21003 9814 21004 9815 21005 8 21006 7853 21007 8 21008 7853 21009 7853 21010 7853 21011 9816 21012 9817 21013 8 21014 7853 21015 7853 21016 7853 21017 9818 21018 8 21019 8 21020 9819 21021 9820 21022 8 21023 9821 21024 8 21025 7853 21026 9822 21027 9823 21028 9824 21029 9825 21030 9826 21031 9827 21032 9828 21033 9829 21034 8 21035 8 21036 9830 21037 9831 21038 8 21039 9832 21040 8 21041 9833 21042 9834 21043 9835 21044 9836 21045 9837 21046 8 21047 9838 21048 9839 21049 9840 21050 9841 21051 8 21052 8 21053 9842 21054 9843 21055 8 21056 9844 21057 8 21058 9845 21059 8 21060 8 21061 9846 21062 9847 21063 9848 21064 9849 21065 9850 21066 9851 21067 9852 21068 9853 21069 9854 21070 9855 21071 9856 21072 8 21073 8 21074 9857 21075 9858 21076 9859 21077 9860 21078 9861 21079 9862 21080 9863 21081 8 21082 8 21083 9864 21084 9865 21085 8 21086 9866 21087 9867 21088 9868 21089 8 21090 9869 21091 9870 21092 9871 21093 9872 21094 8 21095 8 21096 9873 21097 9874 21098 8 21099 9875 21100 9876 21101 9877 21102 9878 21103 8 21104 9879 21105 9880 21106 9881 21107 9882 21108 9883 21109 9884 21110 8 21111 9885 21112 9886 21113 8 21114 9887 21115 9888 21116 9889 21117 9890 21118 9891 21119 9892 21120 8 21121 9893 21122 9894 21123 9895 21124 9896 21125 8 21126 9897 21127 9898 21128 9899 21129 9900 21130 8 21131 9901 21132 9902 21133 9903 21134 8 21135 9904 21136 8 21137 9905 21138 9906 21139 9907 21140 8 21141 9908 21142 9909 21143 8 21144 8 21145 9910 21146 8 21147 9911 21148 8 21149 9912 21150 9913 21151 9914 21152 9915 21153 9916 21154 8 21155 9917 21156 9918 21157 9919 21158 8 21159 9920 21160 9921 21161 8 21162 9922 21163 8 21164 9923 21165 8 21166 9924 21167 9925 21168 9926 21169 9927 21170 9928 21171 9929 21172 9930 21173 9931 21174 9932 21175 8 21176 8 21177 9933 21178 9934 21179 8 21180 8 21181 9935 21182 9936 21183 9937 21184 9938 21185 9939 21186 9940 21187 9941 21188 9942 21189 8 21190 8 21191 9943 21192 9944 21193 8 21194 9945 21195 9946 21196 9947 21197 9948 21198 9949 21199 9950 21200 8 21201 9951 21202 8 21203 9952 21204 9953 21205 9954 21206 9955 21207 9956 21208 8 21209 9957 21210 9958 21211 9959 21212 9960 21213 8 21214 8 21215 9961 21216 9962 21217 8 21218 9963 21219 8 21220 9964 21221 9965 21222 9966 21223 9967 21224 9968 21225 9969 21226 8 21227 9970 21228 8 21229 9971 21230 9972 21231 9973 21232 9974 21233 9975 21234 9976 21235 9977 21236 9978 21237 9979 21238 9980 21239 8 21240 9981 21241 8 21242 9982 21243 8 21244 8 21245 9983 21246 9984 21247 9985 21248 9986 21249 8 21250 8 21251 9987 21252 9988 21253 9989 21254 9990 21255 9991 21256 8 21257 9992 21258 8 21259 9993 21260 8 21261 8 21262 8 21263 8 21264 9994 21265 8 21266 9995 21267 9996 21268 9997 21269 9998 21270 9999 21271 10000 21272 8 21273 10001 21274 10002 21275 8 21276 10003 21277 8 21278 10004 21279 10005 21280 10006 21281 10007 21282 10008 21283 10009 21284 10010 21285 8 21286 8 21287 10011 21288 8 21289 8 21290 10012 21291 10013 21292 10014 21293 8 21294 8 21295 8 21296 8 21297 10015 21298 8 21299 8 21300 8 21301 8 21302 10016 21303 10017 21304 10018 21305 10019 21306 10020 21307 10021 21308 10022 21309 10023 21310 10024 21311 10025 21312 10026 21313 10027 21314 10028 21315 10029 21316 10030 21317 10031 21318 10032 21319 10033 21320 10034 21321 10035 21322 10036 21323 10037 21324 10038 21325 10039 21326 10040 21327 10041 21328 10042 21329 10043 21330 10044 21331 10045 21332 10046 21333 10047 21334 10048 21335 10049 21336 10019 21337 10019 21338 10019 21339 10019 21340 10019 21341 10019 21342 10019 21343 10019 21344 10019 21345 10019 21346 10019 21347 10019 21348 10019 21349 10019 21350 10019 21351 10050 21352 10051 21353 8 21354 8 21355 8 21356 8 21357 8 21358 8 21359 10052 21360 8 21361 8 21362 10053 21363 8 21364 10054 21365 10055 21366 10056 21367 10057 21368 10058 21369 10059 21370 8 21371 10060 21372 8 21373 8 21374 8 21375 8 21376 8 21377 10061 21378 10062 21379 8 21380 8 21381 10063 21382 8 21383 10064 21384 8 21385 8 21386 10065 21387 10066 21388 8 21389 10019 21390 10019 21391 10019 21392 10019 21393 10019 21394 10067 21395 10068 21396 10069 21397 10019 21398 10019 21399 10019 21400 10019 21401 10019 21402 10019 21403 10019 21404 10019 21405 10019 21406 10019 21407 10070 21408 8 21409 8 21410 8 21411 8 21412 8 21413 8 21414 8 21415 8 21416 8 21417 8 21418 8 21419 8 21420 8 21421 8 21422 8 21423 8 21424 8 21425 8 21426 8 21427 8 21428 8 21429 8 21430 8 21431 8 21432 8 21433 8 21434 8 21435 8 21436 8 21437 8 21438 8 21439 8 21440 8 21441 8 21442 8 21443 8 21444 8 21445 8 21446 8 21447 8 21448 8 21449 8 21450 8 21451 8 21452 8 21453 8 21454 8 21455 8 21456 8 21457 8 21458 8 21459 8 21460 8 21461 8 21462 8 21463 8 21464 8 21465 8 21466 8 21467 8 21468 8 21469 8 21470 8 21471 8 21472 8 21473 8 21474 8 21475 8 21476 8 21477 8 21478 8 21479 8 21480 8 21481 8 21482 8 21483 8 21484 8 21485 8 21486 8 21487 10071 21488 8 21489 8 21490 8 21491 10072 21492 8 21493 8 21494 10073 21495 10074 21496 10075 21497 8 21498 8 21499 10076 21500 10077 21501 10078 21502 10079 21503 8 21504 10080 21505 8 21506 10081 21507 10082 21508 10083 21509 10084 21510 10085 21511 10086 21512 10087 21513 10088 21514 10089 21515 10019 21516 10019 21517 10019 21518 10019 21519 10019 21520 10019 21521 10019 21522 10019 21523 10019 21524 10019 21525 10019 21526 10019 21527 10019 21528 10090 21529 10019 21530 10091 21531 8 21532 10092 21533 10093 21534 8 21535 10094 21536 8 21537 10095 21538 8 21539 8 21540 8 21541 10096 21542 10097 21543 10098 21544 8 21545 10099 21546 8 21547 8 21548 10100 21549 10101 21550 10102 21551 8 21552 8 21553 8 21554 10103 21555 10104 21556 8 21557 8 21558 8 21559 10105 21560 10106 21561 10107 21562 10108 21563 10109 21564 8 21565 10110 21566 8 21567 8 21568 10111 21569 8 21570 10112 21571 10113 21572 8 21573 8 21574 10114 21575 8 21576 8 21577 10115 21578 10116 21579 8 21580 8 21581 10117 21582 10118 21583 10119 21584 10120 21585 10121 21586 10122 21587 10123 21588 10124 21589 10125 21590 10126 21591 10127 21592 8 21593 10128 21594 10129 21595 10130 21596 10131 21597 10132 21598 10133 21599 10134 21600 10135 21601 10136 21602 10137 21603 10138 21604 10139 21605 10140 21606 10141 21607 10142 21608 10143 21609 10144 21610 10145 21611 10146 21612 8 21613 10147 21614 8 21615 10148 21616 8 21617 8 21618 10149 21619 10150 21620 10151 21621 10152 21622 8 21623 8 21624 10153 21625 10154 21626 10155 21627 10156 21628 10157 21629 8 21630 10158 21631 8 21632 10159 21633 8 21634 8 21635 8 21636 8 21637 10160 21638 8 21639 10161 21640 10162 21641 10163 21642 10164 21643 10165 21644 10166 21645 8 21646 10167 21647 10168 21648 8 21649 10169 21650 8 21651 10170 21652 10171 21653 10172 21654 10173 21655 10174 21656 10175 21657 10176 21658 8 21659 8 21660 10177 21661 8 21662 8 21663 10178 21664 10179 21665 10180 21666 8 21667 8 21668 8 21669 8 21670 10181 21671 8 21672 8 21673 8 21674 8 21675 10182 21676 10183 21677 10184 21678 10185 21679 10186 21680 10187 21681 10188 21682 10189 21683 10190 21684 10191 21685 10192 21686 10193 21687 10194 21688 10195 21689 10196 21690 10197 21691 10198 21692 10199 21693 10200 21694 10201 21695 10202 21696 10203 21697 10204 21698 10205 21699 10206 21700 10207 21701 10208 21702 10209 21703 10210 21704 10211 21705 10212 21706 10213 21707 10214 21708 10215 21709 10185 21710 10185 21711 10185 21712 10185 21713 10185 21714 10185 21715 10185 21716 10185 21717 10185 21718 10185 21719 10185 21720 10185 21721 10185 21722 10185 21723 10185 21724 10216 21725 10217 21726 8 21727 8 21728 8 21729 8 21730 8 21731 8 21732 10218 21733 8 21734 8 21735 10219 21736 8 21737 10220 21738 10221 21739 10222 21740 10223 21741 10224 21742 10225 21743 8 21744 10226 21745 8 21746 8 21747 8 21748 8 21749 8 21750 10227 21751 10228 21752 8 21753 8 21754 10229 21755 8 21756 10230 21757 8 21758 8 21759 10231 21760 10232 21761 8 21762 10185 21763 10185 21764 10185 21765 10185 21766 10185 21767 10233 21768 10234 21769 10235 21770 10185 21771 10185 21772 10185 21773 10185 21774 10185 21775 10185 21776 10185 21777 10185 21778 10185 21779 10185 21780 10236 21781 8 21782 8 21783 8 21784 8 21785 8 21786 8 21787 8 21788 8 21789 8 21790 8 21791 8 21792 8 21793 8 21794 8 21795 8 21796 8 21797 8 21798 8 21799 8 21800 8 21801 8 21802 8 21803 8 21804 8 21805 8 21806 8 21807 8 21808 8 21809 8 21810 8 21811 8 21812 8 21813 8 21814 8 21815 8 21816 8 21817 8 21818 8 21819 8 21820 8 21821 8 21822 8 21823 8 21824 8 21825 8 21826 8 21827 8 21828 8 21829 8 21830 8 21831 8 21832 8 21833 8 21834 8 21835 8 21836 8 21837 8 21838 8 21839 8 21840 8 21841 8 21842 8 21843 8 21844 8 21845 8 21846 8 21847 8 21848 8 21849 8 21850 8 21851 8 21852 8 21853 8 21854 8 21855 8 21856 8 21857 8 21858 8 21859 8 21860 10237 21861 8 21862 8 21863 8 21864 10238 21865 8 21866 8 21867 10239 21868 10240 21869 10241 21870 8 21871 8 21872 10242 21873 10243 21874 10244 21875 10245 21876 8 21877 10246 21878 8 21879 10247 21880 10248 21881 10249 21882 10250 21883 10251 21884 10252 21885 10253 21886 10254 21887 10255 21888 10185 21889 10185 21890 10185 21891 10185 21892 10185 21893 10185 21894 10185 21895 10185 21896 10185 21897 10185 21898 10185 21899 10185 21900 10185 21901 10256 21902 10185 21903 10257 21904 8 21905 10258 21906 10259 21907 8 21908 10260 21909 8 21910 10261 21911 8 21912 8 21913 8 21914 10262 21915 10263 21916 10264 21917 8 21918 10265 21919 8 21920 8 21921 10266 21922 10267 21923 10268 21924 8 21925 8 21926 8 21927 10269 21928 10270 21929 8 21930 8 21931 8 21932 10271 21933 10272 21934 10273 21935 10274 21936 10275 21937 8 21938 10276 21939 8 21940 8 21941 10277 21942 8 21943 10278 21944 10279 21945 8 21946 8 21947 10280 21948 8 21949 8 21950 10281 21951 10282 21952 8 21953 8 21954 10283 21955 10284 21956 10285 21957 10286 21958 10287 21959 10288 21960 10289 21961 10290 21962 10291 21963 10292 21964 10293 21965 8 21966 10294 21967 10295 21968 10296 21969 10297 21970 10298 21971 10299 21972 10300 21973 10301 21974 10302 21975 10303 21976 8 21977 8 21978 10304 21979 8 21980 10305 21981 8 21982 8 21983 8 21984 10306 21985 8 21986 8 21987 10307 21988 10308 21989 10309 21990 10310 21991 10311 21992 8 21993 10312 21994 8 21995 8 21996 10313 21997 8 21998 10314 21999 10315 22000 8 22001 8 22002 10316 22003 8 22004 10317 22005 8 22006 10318 22007 10319 22008 10320 22009 10185 22010 10185 22011 10321 22012 10322 22013 10323 22014 10324 22015 10325 22016 10326 22017 10327 22018 10328 22019 10329 22020 10330 22021 8 22022 10331 22023 8 22024 10332 22025 8 22026 8 22027 10333 22028 10334 22029 10335 22030 10336 22031 8 22032 8 22033 10337 22034 10338 22035 10339 22036 10340 22037 10341 22038 8 22039 10342 22040 8 22041 10343 22042 8 22043 8 22044 8 22045 8 22046 10344 22047 8 22048 10345 22049 10346 22050 10347 22051 10348 22052 10349 22053 10350 22054 8 22055 10351 22056 10352 22057 8 22058 10353 22059 8 22060 10354 22061 10355 22062 10356 22063 10357 22064 10358 22065 10359 22066 10360 22067 8 22068 8 22069 10361 22070 8 22071 8 22072 10362 22073 10363 22074 10364 22075 8 22076 8 22077 8 22078 8 22079 10365 22080 8 22081 8 22082 8 22083 8 22084 10366 22085 10367 22086 10368 22087 10369 22088 10370 22089 10371 22090 10372 22091 10373 22092 10374 22093 10375 22094 10376 22095 10377 22096 10378 22097 10379 22098 10380 22099 10381 22100 10382 22101 10383 22102 10384 22103 10385 22104 10386 22105 10387 22106 10388 22107 10389 22108 10390 22109 10391 22110 10392 22111 10393 22112 10394 22113 10395 22114 10396 22115 10397 22116 10398 22117 10399 22118 10369 22119 10369 22120 10369 22121 10369 22122 10369 22123 10369 22124 10369 22125 10369 22126 10369 22127 10369 22128 10369 22129 10369 22130 10369 22131 10369 22132 10369 22133 10400 22134 10401 22135 8 22136 8 22137 8 22138 8 22139 8 22140 8 22141 10402 22142 8 22143 8 22144 10403 22145 8 22146 10404 22147 10405 22148 10406 22149 10407 22150 10408 22151 10409 22152 8 22153 10410 22154 8 22155 8 22156 8 22157 8 22158 8 22159 10411 22160 10412 22161 8 22162 8 22163 10413 22164 8 22165 10414 22166 8 22167 8 22168 10415 22169 10416 22170 8 22171 10369 22172 10369 22173 10369 22174 10369 22175 10369 22176 10417 22177 10418 22178 10419 22179 10369 22180 10369 22181 10369 22182 10369 22183 10369 22184 10369 22185 10369 22186 10369 22187 10369 22188 10369 22189 10420 22190 8 22191 8 22192 8 22193 8 22194 8 22195 8 22196 8 22197 8 22198 8 22199 8 22200 8 22201 8 22202 8 22203 8 22204 8 22205 8 22206 8 22207 8 22208 8 22209 8 22210 8 22211 8 22212 8 22213 8 22214 8 22215 8 22216 8 22217 8 22218 8 22219 8 22220 8 22221 8 22222 8 22223 8 22224 8 22225 8 22226 8 22227 8 22228 8 22229 8 22230 8 22231 8 22232 8 22233 8 22234 8 22235 8 22236 8 22237 8 22238 8 22239 8 22240 8 22241 8 22242 8 22243 8 22244 8 22245 8 22246 8 22247 8 22248 8 22249 8 22250 8 22251 8 22252 8 22253 8 22254 8 22255 8 22256 8 22257 8 22258 8 22259 8 22260 8 22261 8 22262 8 22263 8 22264 8 22265 8 22266 8 22267 8 22268 8 22269 10421 22270 8 22271 8 22272 8 22273 10422 22274 8 22275 8 22276 10423 22277 10424 22278 10425 22279 8 22280 8 22281 10426 22282 10427 22283 10428 22284 10429 22285 8 22286 10430 22287 8 22288 10431 22289 10432 22290 10433 22291 10434 22292 10435 22293 10436 22294 10437 22295 10438 22296 10439 22297 10369 22298 10369 22299 10369 22300 10369 22301 10369 22302 10369 22303 10369 22304 10369 22305 10369 22306 10369 22307 10369 22308 10369 22309 10369 22310 10440 22311 10369 22312 10441 22313 8 22314 10442 22315 10443 22316 8 22317 10444 22318 8 22319 10445 22320 8 22321 8 22322 8 22323 10446 22324 10447 22325 10448 22326 8 22327 10449 22328 8 22329 8 22330 10450 22331 10451 22332 10452 22333 8 22334 8 22335 8 22336 10453 22337 10454 22338 8 22339 8 22340 8 22341 10455 22342 10456 22343 10457 22344 10458 22345 10459 22346 8 22347 10460 22348 8 22349 8 22350 10461 22351 8 22352 10462 22353 10463 22354 8 22355 8 22356 10464 22357 8 22358 8 22359 10465 22360 10466 22361 8 22362 8 22363 10467 22364 10468 22365 10469 22366 10470 22367 10471 22368 10472 22369 10473 22370 10474 22371 10475 22372 10476 22373 10477 22374 8 22375 10478 22376 10479 22377 10480 22378 10481 22379 10482 22380 10483 22381 10484 22382 10485 22383 10486 22384 10487 22385 8 22386 10488 22387 10185 22388 10489 22389 8 22390 10490 22391 8 22392 8 22393 8 22394 10491 22395 8 22396 8 22397 10492 22398 10493 22399 10494 22400 10495 22401 10496 22402 8 22403 10497 22404 8 22405 8 22406 10498 22407 8 22408 10499 22409 10500 22410 8 22411 8 22412 10501 22413 8 22414 10502 22415 8 22416 10503 22417 10504 22418 10505 22419 10369 22420 10369 22421 10506 22422 10507 22423 10369 22424 10369 22425 10369 22426 10369 22427 8 22428 10508 22429 8 22430 10509 22431 8 22432 8 22433 10369 22434 10369 22435 10369 22436 10369 22437 10510 22438 10511 22439 8 22440 8 22441 8 22442 8 22443 8 22444 8 22445 10512 22446 10513 22447 10514 22448 8 22449 10515 22450 8 22451 10516 22452 10517 22453 10518 22454 10519 22455 10520 22456 10521 22457 8 22458 10522 22459 8 22460 8 22461 8 22462 8 22463 10523 22464 8 22465 8 22466 8 22467 10524 22468 10525 22469 10526 22470 8 22471 8 22472 10527 22473 8 22474 8 22475 10528 22476 10529 22477 10530 22478 10531 22479 10532 22480 10533 22481 8 22482 10534 22483 8 22484 10535 22485 10536 22486 10537 22487 10538 22488 10539 22489 8 22490 10540 22491 10541 22492 10542 22493 10543 22494 10369 22495 10544 22496 10545 22497 10546 22498 8 22499 10369 22500 10369 22501 10369 22502 8 22503 10547 22504 8 22505 10548 22506 8 22507 10549 22508 10550 22509 8 22510 10551 22511 10552 22512 8 22513 10369 22514 8 22515 10369 22516 10369 22517 10369 22518 10553 22519 10554 22520 8 22521 10369 22522 10369 22523 10369 22524 10555 22525 8 22526 8 22527 10556 22528 10557 22529 8 22530 10558 22531 8 22532 10369 22533 10185 22534 10185 22535 10185 22536 8 22537 10559 22538 8 22539 10560 22540 8 22541 8 22542 10185 22543 10185 22544 10185 22545 10185 22546 10561 22547 10562 22548 8 22549 8 22550 8 22551 8 22552 8 22553 8 22554 10563 22555 10564 22556 10565 22557 8 22558 10566 22559 8 22560 10567 22561 10568 22562 10569 22563 10570 22564 10571 22565 10572 22566 8 22567 10573 22568 8 22569 8 22570 8 22571 8 22572 10574 22573 8 22574 8 22575 8 22576 10575 22577 10576 22578 10577 22579 8 22580 8 22581 10578 22582 8 22583 8 22584 10579 22585 10580 22586 10581 22587 10582 22588 10583 22589 10584 22590 8 22591 10585 22592 8 22593 10586 22594 10587 22595 10588 22596 10589 22597 10590 22598 8 22599 10591 22600 10592 22601 10593 22602 10594 22603 10185 22604 10595 22605 10596 22606 10597 22607 8 22608 10185 22609 10185 22610 10185 22611 8 22612 10598 22613 8 22614 10599 22615 8 22616 10600 22617 10601 22618 8 22619 10602 22620 10603 22621 8 22622 10185 22623 8 22624 10185 22625 10185 22626 10185 22627 10604 22628 10605 22629 8 22630 10185 22631 10185 22632 10185 22633 10606 22634 8 22635 8 22636 10607 22637 10608 22638 8 22639 10609 22640 8 22641 10185 22642 10610 22643 7245 22644 10611 22645 8 22646 10612 22647 8 22648 8 22649 8 22650 10613 22651 8 22652 8 22653 10614 22654 10615 22655 10616 22656 10617 22657 10618 22658 8 22659 10619 22660 8 22661 8 22662 10620 22663 8 22664 10621 22665 10622 22666 8 22667 8 22668 10623 22669 8 22670 10624 22671 8 22672 10625 22673 10626 22674 10627 22675 10019 22676 10019 22677 10628 22678 10629 22679 10630 22680 10631 22681 10632 22682 10633 22683 10634 22684 10635 22685 10636 22686 10637 22687 8 22688 10638 22689 8 22690 10639 22691 8 22692 8 22693 10640 22694 10641 22695 10642 22696 10643 22697 8 22698 8 22699 10644 22700 10645 22701 10646 22702 10647 22703 10648 22704 8 22705 10649 22706 8 22707 10650 22708 8 22709 8 22710 8 22711 8 22712 10651 22713 8 22714 10652 22715 10653 22716 10654 22717 10655 22718 10656 22719 10657 22720 8 22721 10658 22722 10659 22723 8 22724 10660 22725 8 22726 10661 22727 10662 22728 10663 22729 10664 22730 10665 22731 10666 22732 10667 22733 8 22734 8 22735 10668 22736 8 22737 8 22738 10669 22739 10670 22740 10671 22741 8 22742 8 22743 8 22744 8 22745 10672 22746 8 22747 8 22748 8 22749 8 22750 10673 22751 10674 22752 10675 22753 10676 22754 10677 22755 10678 22756 10679 22757 10680 22758 10681 22759 10682 22760 10683 22761 10684 22762 10685 22763 10686 22764 10687 22765 10688 22766 10689 22767 10690 22768 10691 22769 10692 22770 10693 22771 10694 22772 10695 22773 10696 22774 10697 22775 10698 22776 10699 22777 10700 22778 10701 22779 10702 22780 10703 22781 10704 22782 10705 22783 10706 22784 10676 22785 10676 22786 10676 22787 10676 22788 10676 22789 10676 22790 10676 22791 10676 22792 10676 22793 10676 22794 10676 22795 10676 22796 10676 22797 10676 22798 10676 22799 10707 22800 10708 22801 8 22802 8 22803 8 22804 8 22805 8 22806 8 22807 10709 22808 8 22809 8 22810 10710 22811 8 22812 10711 22813 10712 22814 10713 22815 10714 22816 10715 22817 10716 22818 8 22819 10717 22820 8 22821 8 22822 8 22823 8 22824 8 22825 10718 22826 10719 22827 8 22828 8 22829 10720 22830 8 22831 10721 22832 8 22833 8 22834 10722 22835 10723 22836 8 22837 10676 22838 10676 22839 10676 22840 10676 22841 10676 22842 10724 22843 10725 22844 10726 22845 10676 22846 10676 22847 10676 22848 10676 22849 10676 22850 10676 22851 10676 22852 10676 22853 10676 22854 10676 22855 10727 22856 8 22857 8 22858 8 22859 8 22860 8 22861 8 22862 8 22863 8 22864 8 22865 8 22866 8 22867 8 22868 8 22869 8 22870 8 22871 8 22872 8 22873 8 22874 8 22875 8 22876 8 22877 8 22878 8 22879 8 22880 8 22881 8 22882 8 22883 8 22884 8 22885 8 22886 8 22887 8 22888 8 22889 8 22890 8 22891 8 22892 8 22893 8 22894 8 22895 8 22896 8 22897 8 22898 8 22899 8 22900 8 22901 8 22902 8 22903 8 22904 8 22905 8 22906 8 22907 8 22908 8 22909 8 22910 8 22911 8 22912 8 22913 8 22914 8 22915 8 22916 8 22917 8 22918 8 22919 8 22920 8 22921 8 22922 8 22923 8 22924 8 22925 8 22926 8 22927 8 22928 8 22929 8 22930 8 22931 8 22932 8 22933 8 22934 8 22935 10728 22936 8 22937 8 22938 8 22939 10729 22940 8 22941 8 22942 10730 22943 10731 22944 10732 22945 8 22946 8 22947 10733 22948 10734 22949 10735 22950 10736 22951 8 22952 10737 22953 8 22954 10738 22955 10739 22956 10740 22957 10741 22958 10742 22959 10743 22960 10744 22961 10745 22962 10746 22963 10676 22964 10676 22965 10676 22966 10676 22967 10676 22968 10676 22969 10676 22970 10676 22971 10676 22972 10676 22973 10676 22974 10676 22975 10676 22976 10747 22977 10676 22978 10748 22979 8 22980 10749 22981 10750 22982 8 22983 10751 22984 8 22985 10752 22986 8 22987 8 22988 8 22989 10753 22990 10754 22991 10755 22992 8 22993 10756 22994 8 22995 8 22996 10757 22997 10758 22998 10759 22999 8 23000 8 23001 8 23002 10760 23003 10761 23004 8 23005 8 23006 8 23007 10762 23008 10763 23009 10764 23010 10765 23011 10766 23012 8 23013 10767 23014 8 23015 8 23016 10768 23017 8 23018 10769 23019 10770 23020 8 23021 8 23022 10771 23023 8 23024 8 23025 10772 23026 10773 23027 8 23028 8 23029 10774 23030 10775 23031 10776 23032 10777 23033 10778 23034 10779 23035 10780 23036 10781 23037 10782 23038 10783 23039 10784 23040 8 23041 10785 23042 10786 23043 10787 23044 10788 23045 10789 23046 10790 23047 10791 23048 10792 23049 10793 23050 10794 23051 8 23052 10795 23053 10019 23054 10796 23055 8 23056 10797 23057 8 23058 8 23059 8 23060 10798 23061 8 23062 8 23063 10799 23064 10800 23065 10801 23066 10802 23067 10803 23068 8 23069 10804 23070 8 23071 8 23072 10805 23073 8 23074 10806 23075 10807 23076 8 23077 8 23078 10808 23079 8 23080 10809 23081 8 23082 10810 23083 10811 23084 10812 23085 10676 23086 10676 23087 10813 23088 10814 23089 10676 23090 10676 23091 10676 23092 10676 23093 8 23094 10815 23095 8 23096 10816 23097 8 23098 8 23099 10676 23100 10676 23101 10676 23102 10676 23103 10817 23104 8 23105 10818 23106 10819 23107 10820 23108 8 23109 8 23110 8 23111 10821 23112 10822 23113 10823 23114 8 23115 8 23116 8 23117 10824 23118 10825 23119 10826 23120 10827 23121 10828 23122 10829 23123 8 23124 10830 23125 10831 23126 10832 23127 10833 23128 10834 23129 8 23130 10835 23131 10836 23132 10837 23133 10838 23134 10676 23135 10839 23136 10840 23137 10841 23138 8 23139 10676 23140 10676 23141 10676 23142 8 23143 10842 23144 8 23145 10843 23146 8 23147 10844 23148 10845 23149 8 23150 10846 23151 10847 23152 8 23153 10676 23154 8 23155 10676 23156 10676 23157 10676 23158 10848 23159 10849 23160 8 23161 10676 23162 10676 23163 10676 23164 10850 23165 8 23166 8 23167 10851 23168 10852 23169 8 23170 10853 23171 8 23172 10676 23173 10019 23174 10019 23175 10019 23176 8 23177 10854 23178 8 23179 10855 23180 8 23181 8 23182 10019 23183 10019 23184 10019 23185 10019 23186 10856 23187 8 23188 10857 23189 10858 23190 10859 23191 8 23192 8 23193 8 23194 10860 23195 10861 23196 10862 23197 8 23198 8 23199 8 23200 10863 23201 10864 23202 10865 23203 10866 23204 10867 23205 10868 23206 8 23207 10869 23208 10870 23209 10871 23210 10872 23211 10873 23212 8 23213 10874 23214 10875 23215 10876 23216 10877 23217 10019 23218 10878 23219 10879 23220 10880 23221 8 23222 10019 23223 10019 23224 10019 23225 8 23226 10881 23227 8 23228 10882 23229 8 23230 10883 23231 10884 23232 8 23233 10885 23234 10886 23235 8 23236 10019 23237 8 23238 10019 23239 10019 23240 10019 23241 10887 23242 10888 23243 8 23244 10019 23245 10019 23246 10019 23247 10889 23248 8 23249 8 23250 10890 23251 10891 23252 8 23253 10892 23254 8 23255 10019 23256 10893 23257 10894 23258 8 23259 8 23260 10895 23261 8 23262 10896 23263 8 23264 10897 23265 8 23266 10898 23267 8 23268 10899 23269 10900 23270 10901 23271 10902 23272 10903 23273 10904 23274 10905 23275 10906 23276 10907 23277 8 23278 8 23279 10908 23280 10909 23281 8 23282 8 23283 10910 23284 10911 23285 10912 23286 10913 23287 10914 23288 10915 23289 10916 23290 10917 23291 8 23292 8 23293 10918 23294 10919 23295 8 23296 10920 23297 8 23298 10921 23299 10922 23300 10923 23301 10924 23302 10925 23303 8 23304 10926 23305 10927 23306 10928 23307 10929 23308 8 23309 8 23310 10930 23311 10931 23312 8 23313 10932 23314 8 23315 10933 23316 8 23317 8 23318 10934 23319 10935 23320 10936 23321 10937 23322 10938 23323 10939 23324 10940 23325 10941 23326 10942 23327 10943 23328 10944 23329 8 23330 8 23331 10945 23332 10946 23333 10947 23334 10948 23335 10949 23336 10950 23337 10951 23338 8 23339 8 23340 10952 23341 10953 23342 8 23343 10954 23344 10955 23345 10956 23346 8 23347 10957 23348 10958 23349 10959 23350 10960 23351 8 23352 8 23353 10961 23354 10962 23355 8 23356 10963 23357 10964 23358 10965 23359 10966 23360 8 23361 10967 23362 10968 23363 10969 23364 10970 23365 10971 23366 10972 23367 8 23368 10973 23369 10974 23370 8 23371 10975 23372 10976 23373 10977 23374 10978 23375 10979 23376 10980 23377 8 23378 10981 23379 10982 23380 10983 23381 10984 23382 8 23383 10985 23384 10986 23385 10987 23386 10988 23387 8 23388 10989 23389 10990 23390 10991 23391 8 23392 10992 23393 8 23394 10993 23395 10994 23396 10995 23397 8 23398 10996 23399 10997 23400 8 23401 8 23402 10998 23403 8 23404 10999 23405 8 23406 11000 23407 11001 23408 11002 23409 11003 23410 11004 23411 8 23412 11005 23413 11006 23414 11007 23415 8 23416 11008 23417 11009 23418 8 23419 11010 23420 8 23421 11011 23422 8 23423 11012 23424 11013 23425 11014 23426 11015 23427 11016 23428 11017 23429 11018 23430 11019 23431 11020 23432 8 23433 8 23434 11021 23435 11022 23436 8 23437 8 23438 11023 23439 11024 23440 11025 23441 11026 23442 11027 23443 11028 23444 11029 23445 11030 23446 8 23447 8 23448 11031 23449 11032 23450 8 23451 11033 23452 11034 23453 11035 23454 11036 23455 11037 23456 11038 23457 8 23458 11039 23459 8 23460 11040 23461 11041 23462 11042 23463 11043 23464 11044 23465 8 23466 11045 23467 11046 23468 11047 23469 11048 23470 8 23471 8 23472 11049 23473 11050 23474 8 23475 11051 23476 8 23477 11052 23478 11053 23479 11054 23480 11055 23481 11056 23482 11057 23483 8 23484 11058 23485 8 23486 11059 23487 11060 23488 7144 23489 11061 23490 8 23491 11062 23492 11063 23493 11064 23494 8 23495 11065 23496 11066 23497 8 23498 8 23499 11067 23500 8 23501 11068 23502 8 23503 11069 23504 8 23505 11070 23506 8 23507 11071 23508 11072 23509 11073 23510 11074 23511 11075 23512 11076 23513 11077 23514 11078 23515 11079 23516 8 23517 8 23518 11080 23519 11081 23520 8 23521 8 23522 11082 23523 11083 23524 11084 23525 11085 23526 11086 23527 11087 23528 11088 23529 11089 23530 8 23531 8 23532 11090 23533 11091 23534 8 23535 11092 23536 8 23537 11093 23538 11094 23539 11095 23540 11096 23541 11097 23542 8 23543 11098 23544 11099 23545 11100 23546 11101 23547 8 23548 8 23549 11102 23550 11103 23551 8 23552 11104 23553 8 23554 11105 23555 8 23556 8 23557 11106 23558 11107 23559 11108 23560 11109 23561 11110 23562 11111 23563 11112 23564 11113 23565 11114 23566 11115 23567 11116 23568 8 23569 8 23570 11117 23571 11118 23572 11119 23573 11120 23574 11121 23575 11122 23576 11123 23577 8 23578 8 23579 11124 23580 11125 23581 8 23582 11126 23583 11127 23584 11128 23585 8 23586 11129 23587 11130 23588 11131 23589 11132 23590 8 23591 8 23592 11133 23593 11134 23594 8 23595 11135 23596 11136 23597 11137 23598 11138 23599 8 23600 11139 23601 11140 23602 11141 23603 11142 23604 11143 23605 11144 23606 8 23607 11145 23608 11146 23609 8 23610 11147 23611 11148 23612 11149 23613 11150 23614 11151 23615 11152 23616 8 23617 11153 23618 11154 23619 11155 23620 11156 23621 8 23622 11157 23623 11158 23624 11159 23625 11160 23626 8 23627 11161 23628 11162 23629 11163 23630 8 23631 11164 23632 8 23633 11165 23634 11166 23635 11167 23636 8 23637 11168 23638 11169 23639 8 23640 8 23641 11170 23642 8 23643 11171 23644 8 23645 11172 23646 11173 23647 11174 23648 11175 23649 11176 23650 8 23651 11177 23652 11178 23653 11179 23654 8 23655 11180 23656 11181 23657 8 23658 11182 23659 8 23660 11183 23661 8 23662 11184 23663 11185 23664 11186 23665 11187 23666 11188 23667 11189 23668 11190 23669 11191 23670 11192 23671 8 23672 8 23673 11193 23674 11194 23675 8 23676 8 23677 11195 23678 11196 23679 11197 23680 11198 23681 11199 23682 11200 23683 11201 23684 11202 23685 8 23686 8 23687 11203 23688 11204 23689 8 23690 11205 23691 11206 23692 11207 23693 11208 23694 11209 23695 11210 23696 8 23697 8 23698 11211 23699 8 23700 11212 23701 11213 23702 11214 23703 11215 23704 11216 23705 8 23706 11217 23707 11218 23708 11219 23709 11220 23710 8 23711 8 23712 11221 23713 11222 23714 8 23715 11223 23716 8 23717 11224 23718 11225 23719 11226 23720 11227 23721 11228 23722 11229 23723 8 23724 11230 23725 8 23726 11231 23727 11232 23728 11233 23729 11234 23730 8 23731 11235 23732 11236 23733 11237 23734 11238 23735 11239 23736 11240 23737 8 23738 11241 23739 11242 23740 8 23741 11243 23742 11244 23743 11245 23744 11246 23745 11247 23746 11248 23747 8 23748 11249 23749 11250 23750 11251 23751 11252 23752 8 23753 11253 23754 11254 23755 11255 23756 11256 23757 8 23758 11257 23759 11258 23760 11259 23761 8 23762 11260 23763 8 23764 11261 23765 11262 23766 11263 23767 8 23768 11264 23769 11265 23770 8 23771 8 23772 11266 23773 8 23774 11267 23775 8 23776 11268 23777 11269 23778 11270 23779 11271 23780 11272 23781 8 23782 11273 23783 11274 23784 8 23785 11275 23786 8 23787 11276 23788 8 23789 11277 23790 11278 23791 11279 23792 11280 23793 11281 23794 11282 23795 11283 23796 11284 23797 11285 23798 8 23799 8 23800 11286 23801 11287 23802 8 23803 8 23804 11288 23805 11289 23806 11290 23807 11291 23808 11292 23809 11293 23810 11294 23811 11295 23812 8 23813 8 23814 11296 23815 11297 23816 8 23817 11298 23818 11299 23819 11300 23820 11301 23821 11302 23822 11303 23823 8 23824 8 23825 11304 23826 8 23827 11305 23828 11306 23829 11307 23830 11308 23831 11309 23832 8 23833 11310 23834 11311 23835 11312 23836 11313 23837 8 23838 8 23839 11314 23840 11315 23841 8 23842 11316 23843 8 23844 11317 23845 11318 23846 11319 23847 11320 23848 11321 23849 11322 23850 8 23851 11323 23852 8 23853 11324 23854 11325 23855 11326 23856 11327 23857 8 23858 8 23859 11328 23860 11329 23861 8 23862 11330 23863 11331 23864 11332 23865 11333 23866 8 23867 11334 23868 11335 23869 11336 23870 11337 23871 11338 23872 11339 23873 8 23874 11340 23875 11341 23876 8 23877 11342 23878 11343 23879 11344 23880 11345 23881 11346 23882 11347 23883 8 23884 11348 23885 11349 23886 11350 23887 11351 23888 8 23889 11352 23890 11353 23891 11354 23892 11355 23893 8 23894 11356 23895 11357 23896 11358 23897 8 23898 11359 23899 8 23900 11360 23901 11361 23902 11362 23903 8 23904 11363 23905 11364 23906 8 23907 8 23908 11365 23909 8 23910 11366 23911 8 23912 11367 23913 11368 23914 11369 23915 11370 23916 11371 23917 8 23918 11372 23919 11373 23920 11374 23921 8 23922 11375 23923 11376 23924 8 23925 11377 23926 8 23927 11378 23928 8 23929 11379 23930 11380 23931 11381 23932 11382 23933 11383 23934 11384 23935 11385 23936 11386 23937 11387 23938 8 23939 8 23940 11388 23941 11389 23942 8 23943 8 23944 11390 23945 11391 23946 11392 23947 11393 23948 11394 23949 11395 23950 11396 23951 11397 23952 8 23953 8 23954 11398 23955 11399 23956 8 23957 11400 23958 11401 23959 11402 23960 11403 23961 11404 23962 11405 23963 8 23964 11406 23965 8 23966 11407 23967 11408 23968 11409 23969 11410 23970 11411 23971 8 23972 11412 23973 11413 23974 11414 23975 11415 23976 8 23977 8 23978 11416 23979 11417 23980 8 23981 11418 23982 8 23983 11419 23984 11420 23985 11421 23986 11422 23987 11423 23988 11424 23989 8 23990 11425 23991 8 23992 11426 23993 11427 23994 11428 23995 8 23996 11429 23997 11430 23998 11431 23999 11432 24000 8 24001 8 24002 11433 24003 11434 24004 8 24005 11435 24006 11436 24007 11437 24008 11438 24009 8 24010 11439 24011 11440 24012 11441 24013 11442 24014 11443 24015 11444 24016 8 24017 11445 24018 11446 24019 8 24020 11447 24021 11448 24022 11449 24023 11450 24024 11451 24025 11452 24026 8 24027 11453 24028 11454 24029 11455 24030 11456 24031 8 24032 11457 24033 11458 24034 11459 24035 8 24036 11460 24037 8 24038 11461 24039 11462 24040 11463 24041 8 24042 11464 24043 11465 24044 8 24045 8 24046 11466 24047 8 24048 11467 24049 8 24050 11468 24051 8 24052 11469 24053 8 24054 11470 24055 11471 24056 11472 24057 11473 24058 11474 24059 11475 24060 11476 24061 11477 24062 11478 24063 8 24064 8 24065 11479 24066 11480 24067 8 24068 8 24069 11481 24070 11482 24071 11483 24072 11484 24073 11485 24074 11486 24075 11487 24076 11488 24077 8 24078 8 24079 11489 24080 11490 24081 8 24082 11491 24083 8 24084 11492 24085 11493 24086 11494 24087 11495 24088 11496 24089 8 24090 11497 24091 11498 24092 11499 24093 11500 24094 8 24095 8 24096 11501 24097 11502 24098 8 24099 11503 24100 8 24101 11504 24102 8 24103 11505 24104 8 24105 11506 24106 8 24107 8 24108 8 24109 11507 24110 8 24111 8 24112 11508 24113 11509 24114 11510 24115 11511 24116 11512 24117 8 24118 11513 24119 8 24120 8 24121 11514 24122 8 24123 11515 24124 11516 24125 8 24126 8 24127 11517 24128 8 24129 11518 24130 8 24131 11519 24132 11520 24133 11521 24134 183 24135 183 24136 11522 24137 11523 24138 11524 24139 11525 24140 11526 24141 11527 24142 11528 24143 11529 24144 11530 24145 11531 24146 8 24147 11532 24148 8 24149 11533 24150 8 24151 8 24152 11534 24153 11535 24154 11536 24155 11537 24156 8 24157 8 24158 11538 24159 11539 24160 11540 24161 11541 24162 11542 24163 8 24164 11543 24165 8 24166 11544 24167 8 24168 8 24169 8 24170 8 24171 11545 24172 8 24173 11546 24174 11547 24175 11548 24176 11549 24177 11550 24178 11551 24179 8 24180 11552 24181 11553 24182 8 24183 11554 24184 8 24185 11555 24186 11556 24187 11557 24188 11558 24189 11559 24190 11560 24191 11561 24192 8 24193 8 24194 11562 24195 8 24196 8 24197 11563 24198 11564 24199 11565 24200 8 24201 8 24202 8 24203 8 24204 11566 24205 8 24206 8 24207 8 24208 8 24209 11567 24210 11568 24211 11569 24212 11570 24213 11571 24214 11572 24215 11573 24216 11574 24217 11575 24218 11576 24219 11577 24220 11578 24221 11579 24222 11580 24223 11581 24224 11582 24225 11583 24226 11584 24227 11585 24228 11586 24229 11587 24230 11588 24231 11589 24232 11590 24233 11591 24234 11592 24235 11593 24236 11594 24237 11595 24238 11596 24239 11597 24240 11598 24241 11599 24242 11600 24243 11570 24244 11570 24245 11570 24246 11570 24247 11570 24248 11570 24249 11570 24250 11570 24251 11570 24252 11570 24253 11570 24254 11570 24255 11570 24256 11570 24257 11570 24258 11601 24259 11602 24260 8 24261 8 24262 8 24263 8 24264 8 24265 8 24266 11603 24267 8 24268 8 24269 11604 24270 8 24271 11605 24272 11606 24273 11607 24274 11608 24275 11609 24276 11610 24277 8 24278 11611 24279 8 24280 8 24281 8 24282 8 24283 8 24284 11612 24285 11613 24286 8 24287 8 24288 11614 24289 8 24290 11615 24291 8 24292 8 24293 11616 24294 11617 24295 8 24296 11570 24297 11570 24298 11570 24299 11570 24300 11570 24301 11618 24302 11619 24303 11620 24304 11570 24305 11570 24306 11570 24307 11570 24308 11570 24309 11570 24310 11570 24311 11570 24312 11570 24313 11570 24314 11621 24315 8 24316 8 24317 8 24318 8 24319 8 24320 8 24321 8 24322 8 24323 8 24324 8 24325 8 24326 8 24327 8 24328 8 24329 8 24330 8 24331 8 24332 8 24333 8 24334 8 24335 8 24336 8 24337 8 24338 8 24339 8 24340 8 24341 8 24342 8 24343 8 24344 8 24345 8 24346 8 24347 8 24348 8 24349 8 24350 8 24351 8 24352 8 24353 8 24354 8 24355 8 24356 8 24357 8 24358 8 24359 8 24360 8 24361 8 24362 8 24363 8 24364 8 24365 8 24366 8 24367 8 24368 8 24369 8 24370 8 24371 8 24372 8 24373 8 24374 8 24375 8 24376 8 24377 8 24378 8 24379 8 24380 8 24381 8 24382 8 24383 8 24384 8 24385 8 24386 8 24387 8 24388 8 24389 8 24390 8 24391 8 24392 8 24393 8 24394 11622 24395 8 24396 8 24397 8 24398 11623 24399 8 24400 8 24401 11624 24402 11625 24403 11626 24404 8 24405 8 24406 11627 24407 11628 24408 11629 24409 11630 24410 8 24411 11631 24412 8 24413 11632 24414 11633 24415 11634 24416 11635 24417 11636 24418 11637 24419 11638 24420 11639 24421 11640 24422 11570 24423 11570 24424 11570 24425 11570 24426 11570 24427 11570 24428 11570 24429 11570 24430 11570 24431 11570 24432 11570 24433 11570 24434 11570 24435 11641 24436 11570 24437 11642 24438 8 24439 11643 24440 11644 24441 8 24442 11645 24443 8 24444 11646 24445 8 24446 8 24447 8 24448 11647 24449 11648 24450 11649 24451 8 24452 11650 24453 8 24454 8 24455 11651 24456 11652 24457 11653 24458 8 24459 8 24460 8 24461 11654 24462 11655 24463 8 24464 8 24465 8 24466 11656 24467 11657 24468 11658 24469 11659 24470 11660 24471 8 24472 11661 24473 8 24474 8 24475 11662 24476 8 24477 11663 24478 11664 24479 8 24480 8 24481 11665 24482 8 24483 8 24484 11666 24485 11667 24486 8 24487 8 24488 11668 24489 11669 24490 11670 24491 11671 24492 11672 24493 11673 24494 11674 24495 11675 24496 11676 24497 11677 24498 11678 24499 8 24500 11679 24501 11680 24502 11681 24503 11682 24504 11683 24505 11684 24506 11685 24507 11686 24508 11687 24509 11688 24510 8 24511 11689 24512 183 24513 11690 24514 8 24515 11691 24516 8 24517 8 24518 8 24519 11692 24520 8 24521 8 24522 11693 24523 11694 24524 11695 24525 11696 24526 11697 24527 8 24528 11698 24529 8 24530 8 24531 11699 24532 8 24533 11700 24534 11701 24535 8 24536 8 24537 11702 24538 8 24539 11703 24540 8 24541 11704 24542 11705 24543 11706 24544 11570 24545 11570 24546 11707 24547 11708 24548 11570 24549 11570 24550 11570 24551 11570 24552 8 24553 11709 24554 8 24555 11710 24556 8 24557 8 24558 11570 24559 11570 24560 11570 24561 11570 24562 11711 24563 8 24564 11712 24565 11713 24566 11714 24567 8 24568 8 24569 8 24570 11715 24571 11716 24572 11717 24573 8 24574 8 24575 8 24576 11718 24577 11719 24578 11720 24579 11721 24580 11722 24581 11723 24582 8 24583 11724 24584 11725 24585 11726 24586 11727 24587 11728 24588 8 24589 11729 24590 11730 24591 11731 24592 11732 24593 11570 24594 11733 24595 11734 24596 11735 24597 8 24598 11570 24599 11570 24600 11570 24601 8 24602 11736 24603 8 24604 11737 24605 8 24606 11738 24607 11739 24608 8 24609 11740 24610 11741 24611 8 24612 11570 24613 8 24614 11570 24615 11570 24616 11570 24617 11742 24618 11743 24619 8 24620 11570 24621 11570 24622 11570 24623 11744 24624 8 24625 8 24626 11745 24627 11746 24628 8 24629 11747 24630 8 24631 11570 24632 183 24633 183 24634 183 24635 8 24636 11748 24637 8 24638 11749 24639 8 24640 8 24641 183 24642 183 24643 183 24644 183 24645 11750 24646 8 24647 11751 24648 11752 24649 11753 24650 8 24651 8 24652 8 24653 11754 24654 11755 24655 11756 24656 8 24657 8 24658 8 24659 11757 24660 11758 24661 11759 24662 11760 24663 11761 24664 11762 24665 8 24666 11763 24667 11764 24668 11765 24669 11766 24670 11767 24671 8 24672 11768 24673 11769 24674 11770 24675 11771 24676 183 24677 11772 24678 11773 24679 11774 24680 8 24681 183 24682 183 24683 183 24684 8 24685 11775 24686 8 24687 11776 24688 8 24689 11777 24690 11778 24691 8 24692 11779 24693 11780 24694 8 24695 183 24696 8 24697 183 24698 183 24699 183 24700 11781 24701 11782 24702 8 24703 183 24704 183 24705 183 24706 11783 24707 8 24708 8 24709 11784 24710 11785 24711 8 24712 11786 24713 8 24714 183 24715 8 24716 11787 24717 11788 24718 11789 24719 11790 24720 11791 24721 11792 24722 11793 24723 11794 24724 11795 24725 11796 24726 11797 24727 8 24728 8 24729 11798 24730 11799 24731 11800 24732 11801 24733 11802 24734 11803 24735 11804 24736 8 24737 8 24738 11805 24739 11806 24740 8 24741 11807 24742 11808 24743 11809 24744 8 24745 11810 24746 11811 24747 11812 24748 11813 24749 8 24750 8 24751 11814 24752 11815 24753 8 24754 11816 24755 11817 24756 11818 24757 11819 24758 8 24759 11820 24760 11821 24761 11822 24762 11823 24763 11824 24764 11825 24765 8 24766 11826 24767 11827 24768 8 24769 11828 24770 11829 24771 11830 24772 11831 24773 11832 24774 11833 24775 8 24776 11834 24777 11835 24778 11836 24779 11837 24780 8 24781 11838 24782 11839 24783 11840 24784 11841 24785 8 24786 11842 24787 11843 24788 11844 24789 8 24790 11845 24791 8 24792 11846 24793 11847 24794 11848 24795 8 24796 11849 24797 11850 24798 8 24799 8 24800 11851 24801 8 24802 11852 24803 8 24804 11853 24805 11854 24806 11855 24807 11856 24808 11857 24809 8 24810 11858 24811 11859 24812 11860 24813 8 24814 11861 24815 11862 24816 8 24817 11863 24818 8 24819 11864 24820 8 24821 11865 24822 11866 24823 11867 24824 11868 24825 11869 24826 11870 24827 11871 24828 11872 24829 11873 24830 8 24831 8 24832 11874 24833 11875 24834 8 24835 8 24836 11876 24837 11877 24838 11878 24839 11879 24840 11880 24841 11881 24842 11882 24843 11883 24844 8 24845 8 24846 11884 24847 11885 24848 8 24849 11886 24850 11887 24851 11888 24852 11889 24853 11890 24854 11891 24855 8 24856 11892 24857 8 24858 11893 24859 11894 24860 11895 24861 11896 24862 11897 24863 8 24864 11898 24865 11899 24866 11900 24867 11901 24868 8 24869 8 24870 11902 24871 11903 24872 8 24873 11904 24874 8 24875 11905 24876 11906 24877 11907 24878 11908 24879 11909 24880 11910 24881 8 24882 11911 24883 8 24884 11912 24885 8 24886 8 24887 8 24888 8 24889 8 24890 8 24891 8 24892 11913 24893 8 24894 11914 24895 8 24896 8 24897 11915 24898 11916 24899 8 24900 8 24901 8 24902 8 24903 8 24904 8 24905 8 24906 8 24907 8 24908 8 24909 11917 24910 11918 24911 8 24912 8 24913 8 24914 8 24915 8 24916 8 24917 8 24918 8 24919 8 24920 8 24921 8 24922 11919 24923 8 24924 8 24925 8 24926 8 24927 8 24928 8 24929 8 24930 8 24931 8 24932 8 24933 8 24934 11920 24935 8 24936 8 24937 8 24938 11921 24939 8 24940 11922 24941 11923 24942 11924 24943 8 24944 8 24945 8 24946 11925 24947 8 24948 8 24949 8 24950 8 24951 8 24952 8 24953 8 24954 8 24955 8 24956 8 24957 11926 24958 8 24959 11927 24960 8 24961 8 24962 11928 24963 11929 24964 11930 24965 8 24966 8 24967 8 24968 8 24969 8 24970 8 24971 8 24972 8 24973 11931 24974 11932 24975 8 24976 8 24977 8 24978 8 24979 8 24980 8 24981 8 24982 8 24983 8 24984 8 24985 11933 24986 8 24987 11934 24988 11935 24989 11936 24990 11937 24991 555 24992 11938 24993 8 24994 8 24995 8 24996 8 24997 8 24998 8 24999 8 25000 8 25001 8 25002 11939 25003 8 25004 8 25005 8 25006 8 25007 11940 25008 11941 25009 8 25010 11942 25011 11943 25012 8 25013 11944 25014 8 25015 11945 25016 8 25017 8 25018 8 25019 8 25020 8 25021 8 25022 8 25023 8 25024 8 25025 8 25026 8 25027 8 25028 8 25029 8 25030 8 25031 8 25032 8 25033 8 25034 8 25035 8 25036 8 25037 8 25038 8 25039 8 25040 8 25041 8 25042 8 25043 8 25044 8 25045 8 25046 8 25047 8 25048 8 25049 8 25050 8 25051 8 25052 8 25053 8 25054 8 25055 8 25056 8 25057 8 25058 8 25059 8 25060 8 25061 8 25062 8 25063 8 25064 8 25065 8 25066 8 25067 8 25068 8 25069 8 25070 8 25071 8 25072 8 25073 8 25074 8 25075 8 25076 8 25077 8 25078 8 25079 8 25080 8 25081 8 25082 8 25083 8 25084 8 25085 8 25086 8 25087 8 25088 8 25089 8 25090 8 25091 8 25092 8 25093 8 25094 8 25095 11946 25096 8 25097 8 25098 8 25099 11947 25100 8 25101 8 25102 11948 25103 11949 25104 11950 25105 8 25106 8 25107 11951 25108 11952 25109 11953 25110 11954 25111 8 25112 11955 25113 8 25114 11956 25115 11957 25116 11958 25117 11959 25118 11960 25119 11961 25120 11962 25121 11963 25122 11964 25123 83 25124 83 25125 83 25126 83 25127 83 25128 83 25129 83 25130 83 25131 83 25132 83 25133 83 25134 83 25135 83 25136 11965 25137 83 25138 11966 25139 8 25140 11967 25141 11968 25142 8 25143 11969 25144 8 25145 11970 25146 8 25147 8 25148 8 25149 11971 25150 11972 25151 11973 25152 8 25153 11974 25154 8 25155 8 25156 11975 25157 11976 25158 11977 25159 8 25160 8 25161 8 25162 11978 25163 11979 25164 8 25165 8 25166 8 25167 11980 25168 11981 25169 11982 25170 11983 25171 11984 25172 8 25173 11985 25174 8 25175 8 25176 11986 25177 8 25178 11987 25179 11988 25180 8 25181 8 25182 11989 25183 8 25184 8 25185 11990 25186 11991 25187 8 25188 8 25189 11992 25190 11993 25191 11994 25192 11995 25193 11996 25194 8 25195 11997 25196 11998 25197 11999 25198 12000 25199 12001 25200 12002 25201 12003 25202 12004 25203 12005 25204 8 25205 12006 25206 12007 25207 12008 25208 12009 25209 12010 25210 12011 25211 12012 25212 8 25213 12013 25214 12014 25215 8 25216 12015 25217 8 25218 12016 25219 12017 25220 8 25221 12018 25222 8 25223 8 25224 12019 25225 12020 25226 8 25227 12021 25228 12022 25229 12023 25230 8 25231 12024 25232 12025 25233 12026 25234 12027 25235 8 25236 8 25237 12028 25238 12029 25239 8 25240 12030 25241 12031 25242 12032 25243 12033 25244 8 25245 12034 25246 12035 25247 12036 25248 12037 25249 12038 25250 12039 25251 8 25252 12040 25253 12041 25254 8 25255 12042 25256 12043 25257 12044 25258 12045 25259 12046 25260 12047 25261 8 25262 12048 25263 12049 25264 12050 25265 12051 25266 8 25267 12052 25268 12053 25269 12054 25270 8 25271 12055 25272 8 25273 12056 25274 12057 25275 12058 25276 8 25277 12059 25278 12060 25279 8 25280 8 25281 12061 25282 8 25283 12062 25284 8 25285 12063 25286 8 25287 12064 25288 8 25289 12065 25290 12066 25291 12067 25292 12068 25293 12069 25294 12070 25295 12071 25296 12072 25297 12073 25298 8 25299 8 25300 12074 25301 12075 25302 8 25303 8 25304 12076 25305 12077 25306 12078 25307 12079 25308 12080 25309 12081 25310 12082 25311 12083 25312 8 25313 8 25314 12084 25315 12085 25316 8 25317 12086 25318 8 25319 12087 25320 12088 25321 12089 25322 12090 25323 12091 25324 8 25325 12092 25326 12093 25327 12094 25328 12095 25329 8 25330 8 25331 12096 25332 12097 25333 8 25334 12098 25335 8 25336 12099 25337 8 25338 8 25339 12100 25340 12101 25341 12102 25342 12103 25343 12104 25344 12105 25345 12106 25346 12107 25347 12108 25348 12109 25349 12110 25350 8 25351 12111 25352 12112 25353 12113 25354 12114 25355 12115 25356 12116 25357 12117 25358 12118 25359 12119 25360 12120 25361 12121 25362 12122 25363 12123 25364 12124 25365 12125 25366 12126 25367 12127 25368 12128 25369 12129 25370 8 25371 12130 25372 8 25373 12131 25374 8 25375 8 25376 12132 25377 12133 25378 12134 25379 12135 25380 8 25381 8 25382 12136 25383 12137 25384 12138 25385 12139 25386 12140 25387 8 25388 12141 25389 8 25390 12142 25391 8 25392 8 25393 8 25394 8 25395 12143 25396 8 25397 12144 25398 12145 25399 12146 25400 12147 25401 12148 25402 12149 25403 8 25404 12150 25405 12151 25406 8 25407 12152 25408 8 25409 12153 25410 12154 25411 12155 25412 12156 25413 12157 25414 12158 25415 12159 25416 8 25417 8 25418 12160 25419 8 25420 8 25421 12161 25422 12162 25423 12163 25424 8 25425 8 25426 8 25427 8 25428 12164 25429 8 25430 8 25431 8 25432 8 25433 12165 25434 12166 25435 12167 25436 12168 25437 12169 25438 12170 25439 12171 25440 12172 25441 12173 25442 12174 25443 12175 25444 12176 25445 12177 25446 12178 25447 12179 25448 12180 25449 12181 25450 12182 25451 12183 25452 12184 25453 12185 25454 12186 25455 12187 25456 12188 25457 12189 25458 12190 25459 12191 25460 12192 25461 12193 25462 12194 25463 12195 25464 12196 25465 12197 25466 12198 25467 12168 25468 12168 25469 12168 25470 12168 25471 12168 25472 12168 25473 12168 25474 12168 25475 12168 25476 12168 25477 12168 25478 12168 25479 12168 25480 12168 25481 12168 25482 12199 25483 12200 25484 8 25485 8 25486 8 25487 8 25488 8 25489 8 25490 12201 25491 8 25492 8 25493 12202 25494 8 25495 12203 25496 12204 25497 12205 25498 12206 25499 12207 25500 12208 25501 8 25502 12209 25503 8 25504 8 25505 8 25506 8 25507 8 25508 12210 25509 12211 25510 8 25511 8 25512 12212 25513 8 25514 12213 25515 8 25516 8 25517 12214 25518 12215 25519 8 25520 12168 25521 12168 25522 12168 25523 12168 25524 12168 25525 12216 25526 12217 25527 12218 25528 12168 25529 12168 25530 12168 25531 12168 25532 12168 25533 12168 25534 12168 25535 12168 25536 12168 25537 12168 25538 12219 25539 8 25540 8 25541 8 25542 8 25543 8 25544 8 25545 8 25546 8 25547 8 25548 8 25549 8 25550 8 25551 8 25552 8 25553 8 25554 8 25555 8 25556 8 25557 8 25558 8 25559 8 25560 8 25561 8 25562 8 25563 8 25564 8 25565 8 25566 8 25567 8 25568 8 25569 8 25570 8 25571 8 25572 8 25573 8 25574 8 25575 8 25576 8 25577 8 25578 8 25579 8 25580 8 25581 8 25582 8 25583 8 25584 8 25585 8 25586 8 25587 8 25588 8 25589 8 25590 8 25591 8 25592 8 25593 8 25594 8 25595 8 25596 8 25597 8 25598 8 25599 8 25600 8 25601 8 25602 8 25603 8 25604 8 25605 8 25606 8 25607 8 25608 8 25609 8 25610 8 25611 8 25612 8 25613 8 25614 8 25615 8 25616 8 25617 8 25618 12220 25619 8 25620 8 25621 8 25622 12221 25623 8 25624 8 25625 12222 25626 12223 25627 12224 25628 8 25629 8 25630 12225 25631 12226 25632 12227 25633 12228 25634 8 25635 12229 25636 8 25637 12230 25638 12231 25639 12232 25640 12233 25641 12234 25642 12235 25643 12236 25644 12237 25645 12238 25646 12168 25647 12168 25648 12168 25649 12168 25650 12168 25651 12168 25652 12168 25653 12168 25654 12168 25655 12168 25656 12168 25657 12168 25658 12168 25659 12239 25660 12168 25661 12240 25662 8 25663 12241 25664 12242 25665 8 25666 12243 25667 8 25668 12244 25669 8 25670 8 25671 8 25672 12245 25673 12246 25674 12247 25675 8 25676 12248 25677 8 25678 8 25679 12249 25680 12250 25681 12251 25682 8 25683 8 25684 8 25685 12252 25686 12253 25687 8 25688 8 25689 8 25690 12254 25691 12255 25692 12256 25693 12257 25694 12258 25695 8 25696 12259 25697 8 25698 8 25699 12260 25700 8 25701 12261 25702 12262 25703 8 25704 8 25705 12263 25706 8 25707 8 25708 12264 25709 12265 25710 8 25711 8 25712 12266 25713 12267 25714 12268 25715 12269 25716 12270 25717 12271 25718 12272 25719 12273 25720 12274 25721 12275 25722 12276 25723 8 25724 12277 25725 12278 25726 12279 25727 12280 25728 12281 25729 12282 25730 12283 25731 12284 25732 12285 25733 12286 25734 8 25735 8 25736 12287 25737 8 25738 12288 25739 8 25740 8 25741 8 25742 12289 25743 8 25744 8 25745 12290 25746 12291 25747 12292 25748 12293 25749 12294 25750 8 25751 12295 25752 8 25753 8 25754 12296 25755 8 25756 12297 25757 12298 25758 8 25759 8 25760 12299 25761 8 25762 12300 25763 8 25764 12301 25765 12302 25766 12303 25767 12168 25768 12168 25769 12304 25770 12305 25771 12306 25772 12307 25773 12308 25774 12309 25775 12310 25776 12311 25777 12312 25778 12313 25779 8 25780 12314 25781 8 25782 12315 25783 8 25784 8 25785 12316 25786 12317 25787 12318 25788 12319 25789 8 25790 8 25791 12320 25792 12321 25793 12322 25794 12323 25795 12324 25796 8 25797 12325 25798 8 25799 12326 25800 8 25801 8 25802 8 25803 8 25804 12327 25805 8 25806 12328 25807 12329 25808 12330 25809 12331 25810 12332 25811 12333 25812 8 25813 12334 25814 12335 25815 8 25816 12336 25817 8 25818 12337 25819 12338 25820 12339 25821 12340 25822 12341 25823 12342 25824 12343 25825 8 25826 8 25827 12344 25828 8 25829 8 25830 12345 25831 12346 25832 12347 25833 8 25834 8 25835 8 25836 8 25837 12348 25838 8 25839 8 25840 8 25841 8 25842 12349 25843 12350 25844 12351 25845 12352 25846 12353 25847 12354 25848 12355 25849 12356 25850 12357 25851 12358 25852 12359 25853 12360 25854 12361 25855 12362 25856 12363 25857 12364 25858 12365 25859 12366 25860 12367 25861 12368 25862 12369 25863 12370 25864 12371 25865 12372 25866 12373 25867 12374 25868 12375 25869 12376 25870 12377 25871 12378 25872 12379 25873 12380 25874 12381 25875 12382 25876 12352 25877 12352 25878 12352 25879 12352 25880 12352 25881 12352 25882 12352 25883 12352 25884 12352 25885 12352 25886 12352 25887 12352 25888 12352 25889 12352 25890 12352 25891 12383 25892 12384 25893 8 25894 8 25895 8 25896 8 25897 8 25898 8 25899 12385 25900 8 25901 8 25902 12386 25903 8 25904 12387 25905 12388 25906 12389 25907 12390 25908 12391 25909 12392 25910 8 25911 12393 25912 8 25913 8 25914 8 25915 8 25916 8 25917 12394 25918 12395 25919 8 25920 8 25921 12396 25922 8 25923 12397 25924 8 25925 8 25926 12398 25927 12399 25928 8 25929 12352 25930 12352 25931 12352 25932 12352 25933 12352 25934 12400 25935 12401 25936 12402 25937 12352 25938 12352 25939 12352 25940 12352 25941 12352 25942 12352 25943 12352 25944 12352 25945 12352 25946 12352 25947 12403 25948 8 25949 8 25950 8 25951 8 25952 8 25953 8 25954 8 25955 8 25956 8 25957 8 25958 8 25959 8 25960 8 25961 8 25962 8 25963 8 25964 8 25965 8 25966 8 25967 8 25968 8 25969 8 25970 8 25971 8 25972 8 25973 8 25974 8 25975 8 25976 8 25977 8 25978 8 25979 8 25980 8 25981 8 25982 8 25983 8 25984 8 25985 8 25986 8 25987 8 25988 8 25989 8 25990 8 25991 8 25992 8 25993 8 25994 8 25995 8 25996 8 25997 8 25998 8 25999 8 26000 8 26001 8 26002 8 26003 8 26004 8 26005 8 26006 8 26007 8 26008 8 26009 8 26010 8 26011 8 26012 8 26013 8 26014 8 26015 8 26016 8 26017 8 26018 8 26019 8 26020 8 26021 8 26022 8 26023 8 26024 8 26025 8 26026 8 26027 12404 26028 8 26029 8 26030 8 26031 12405 26032 8 26033 8 26034 12406 26035 12407 26036 12408 26037 8 26038 8 26039 12409 26040 12410 26041 12411 26042 12412 26043 8 26044 12413 26045 8 26046 12414 26047 12415 26048 12416 26049 12417 26050 12418 26051 12419 26052 12420 26053 12421 26054 12422 26055 12352 26056 12352 26057 12352 26058 12352 26059 12352 26060 12352 26061 12352 26062 12352 26063 12352 26064 12352 26065 12352 26066 12352 26067 12352 26068 12423 26069 12352 26070 12424 26071 8 26072 12425 26073 12426 26074 8 26075 12427 26076 8 26077 12428 26078 8 26079 8 26080 8 26081 12429 26082 12430 26083 12431 26084 8 26085 12432 26086 8 26087 8 26088 12433 26089 12434 26090 12435 26091 8 26092 8 26093 8 26094 12436 26095 12437 26096 8 26097 8 26098 8 26099 12438 26100 12439 26101 12440 26102 12441 26103 12442 26104 8 26105 12443 26106 8 26107 8 26108 12444 26109 8 26110 12445 26111 12446 26112 8 26113 8 26114 12447 26115 8 26116 8 26117 12448 26118 12449 26119 8 26120 8 26121 12450 26122 12451 26123 12452 26124 12453 26125 12454 26126 12455 26127 12456 26128 12457 26129 12458 26130 12459 26131 12460 26132 8 26133 12461 26134 12462 26135 12463 26136 12464 26137 12465 26138 12466 26139 12467 26140 12468 26141 12469 26142 12470 26143 8 26144 12471 26145 12168 26146 12472 26147 8 26148 12473 26149 8 26150 8 26151 8 26152 12474 26153 8 26154 8 26155 12475 26156 12476 26157 12477 26158 12478 26159 12479 26160 8 26161 12480 26162 8 26163 8 26164 12481 26165 8 26166 12482 26167 12483 26168 8 26169 8 26170 12484 26171 8 26172 12485 26173 8 26174 12486 26175 12487 26176 12488 26177 12352 26178 12352 26179 12489 26180 12490 26181 12352 26182 12352 26183 12352 26184 12352 26185 8 26186 12491 26187 8 26188 12492 26189 8 26190 8 26191 12352 26192 12352 26193 12352 26194 12352 26195 12493 26196 12494 26197 8 26198 8 26199 8 26200 8 26201 8 26202 8 26203 12495 26204 12496 26205 12497 26206 8 26207 12498 26208 8 26209 12499 26210 12500 26211 12501 26212 12502 26213 12503 26214 12504 26215 8 26216 12505 26217 8 26218 8 26219 8 26220 8 26221 12506 26222 8 26223 8 26224 8 26225 12507 26226 12508 26227 12509 26228 8 26229 8 26230 12510 26231 8 26232 8 26233 12511 26234 12512 26235 12513 26236 12514 26237 12515 26238 12516 26239 8 26240 12517 26241 8 26242 12518 26243 12519 26244 12520 26245 12521 26246 12522 26247 8 26248 12523 26249 12524 26250 12525 26251 12526 26252 12352 26253 12527 26254 12528 26255 12529 26256 8 26257 12352 26258 12352 26259 12352 26260 8 26261 12530 26262 8 26263 12531 26264 8 26265 12532 26266 12533 26267 8 26268 12534 26269 12535 26270 8 26271 12352 26272 8 26273 12352 26274 12352 26275 12352 26276 12536 26277 12537 26278 8 26279 12352 26280 12352 26281 12352 26282 12538 26283 8 26284 8 26285 12539 26286 12540 26287 8 26288 12541 26289 8 26290 12352 26291 12168 26292 12168 26293 12168 26294 8 26295 12542 26296 8 26297 12543 26298 8 26299 8 26300 12168 26301 12168 26302 12168 26303 12168 26304 12544 26305 12545 26306 8 26307 8 26308 8 26309 8 26310 8 26311 8 26312 12546 26313 12547 26314 12548 26315 8 26316 12549 26317 8 26318 12550 26319 12551 26320 12552 26321 12553 26322 12554 26323 12555 26324 8 26325 12556 26326 8 26327 8 26328 8 26329 8 26330 12557 26331 8 26332 8 26333 8 26334 12558 26335 12559 26336 12560 26337 8 26338 8 26339 12561 26340 8 26341 8 26342 12562 26343 12563 26344 12564 26345 12565 26346 12566 26347 12567 26348 8 26349 12568 26350 8 26351 12569 26352 12570 26353 12571 26354 12572 26355 12573 26356 8 26357 12574 26358 12575 26359 12576 26360 12577 26361 12168 26362 12578 26363 12579 26364 12580 26365 8 26366 12168 26367 12168 26368 12168 26369 8 26370 12581 26371 8 26372 12582 26373 8 26374 12583 26375 12584 26376 8 26377 12585 26378 12586 26379 8 26380 12168 26381 8 26382 12168 26383 12168 26384 12168 26385 12587 26386 12588 26387 8 26388 12168 26389 12168 26390 12168 26391 12589 26392 8 26393 8 26394 12590 26395 12591 26396 8 26397 12592 26398 8 26399 12168 26400 8 26401 12593 26402 8 26403 12594 26404 8 26405 8 26406 12595 26407 12596 26408 12597 26409 12598 26410 12599 26411 12600 26412 12601 26413 8 26414 8 26415 12602 26416 12603 26417 12604 26418 12605 26419 8 26420 8 26421 12606 26422 8 26423 12607 26424 8 26425 8 26426 8 26427 12608 26428 12609 26429 12610 26430 12611 26431 12612 26432 12613 26433 12614 26434 12615 26435 12616 26436 12617 26437 12618 26438 8 26439 12619 26440 12620 26441 12621 26442 12622 26443 12623 26444 12624 26445 12625 26446 8 26447 12626 26448 12627 26449 12628 26450 12629 26451 8 26452 12630 26453 12631 26454 12632 26455 12633 26456 12634 26457 8 26458 12635 26459 12636 26460 12637 26461 12638 26462 12639 26463 12640 26464 8 26465 12641 26466 12642 26467 12643 26468 12644 26469 12645 26470 12646 26471 8 26472 12647 26473 12648 26474 12649 26475 12650 26476 12651 26477 8 26478 12652 26479 8 26480 8 26481 12653 26482 8 26483 12654 26484 12655 26485 8 26486 8 26487 12656 26488 8 26489 12657 26490 12658 26491 8 26492 12659 26493 12660 26494 12661 26495 8 26496 12662 26497 8 26498 12663 26499 12664 26500 12665 26501 83 26502 83 26503 12666 26504 12667 26505 12668 26506 12669 26507 12670 26508 12671 26509 12672 26510 12673 26511 12674 26512 12675 26513 8 26514 12676 26515 8 26516 12677 26517 8 26518 8 26519 12678 26520 12679 26521 12680 26522 12681 26523 8 26524 8 26525 12682 26526 12683 26527 12684 26528 12685 26529 12686 26530 8 26531 12687 26532 8 26533 12688 26534 8 26535 8 26536 8 26537 8 26538 12689 26539 8 26540 12690 26541 12691 26542 12692 26543 12693 26544 12694 26545 12695 26546 8 26547 12696 26548 12697 26549 8 26550 12698 26551 8 26552 12699 26553 12700 26554 12701 26555 12702 26556 12703 26557 12704 26558 12705 26559 8 26560 8 26561 12706 26562 8 26563 8 26564 12707 26565 12708 26566 12709 26567 8 26568 8 26569 8 26570 8 26571 12710 26572 8 26573 8 26574 8 26575 8 26576 12711 26577 12712 26578 12713 26579 12714 26580 12715 26581 12716 26582 12717 26583 12718 26584 12719 26585 12720 26586 12721 26587 12722 26588 12723 26589 12724 26590 12725 26591 12726 26592 12727 26593 12728 26594 12729 26595 12730 26596 12731 26597 12732 26598 12733 26599 12734 26600 12735 26601 12736 26602 12737 26603 12738 26604 12739 26605 12740 26606 12741 26607 12742 26608 12743 26609 12744 26610 12714 26611 12714 26612 12714 26613 12714 26614 12714 26615 12714 26616 12714 26617 12714 26618 12714 26619 12714 26620 12714 26621 12714 26622 12714 26623 12714 26624 12714 26625 12745 26626 12746 26627 8 26628 8 26629 8 26630 8 26631 8 26632 8 26633 12747 26634 8 26635 8 26636 12748 26637 8 26638 12749 26639 12750 26640 12751 26641 12752 26642 12753 26643 12754 26644 8 26645 12755 26646 8 26647 8 26648 8 26649 8 26650 8 26651 12756 26652 12757 26653 8 26654 8 26655 12758 26656 8 26657 12759 26658 8 26659 8 26660 12760 26661 12761 26662 8 26663 12714 26664 12714 26665 12714 26666 12714 26667 12714 26668 12762 26669 12763 26670 12764 26671 12714 26672 12714 26673 12714 26674 12714 26675 12714 26676 12714 26677 12714 26678 12714 26679 12714 26680 12714 26681 12765 26682 8 26683 8 26684 8 26685 8 26686 8 26687 8 26688 8 26689 8 26690 8 26691 8 26692 8 26693 8 26694 8 26695 8 26696 8 26697 8 26698 8 26699 8 26700 8 26701 8 26702 8 26703 8 26704 8 26705 8 26706 8 26707 8 26708 8 26709 8 26710 8 26711 8 26712 8 26713 8 26714 8 26715 8 26716 8 26717 8 26718 8 26719 8 26720 8 26721 8 26722 8 26723 8 26724 8 26725 8 26726 8 26727 8 26728 8 26729 8 26730 8 26731 8 26732 8 26733 8 26734 8 26735 8 26736 8 26737 8 26738 8 26739 8 26740 8 26741 8 26742 8 26743 8 26744 8 26745 8 26746 8 26747 8 26748 8 26749 8 26750 8 26751 8 26752 8 26753 8 26754 8 26755 8 26756 8 26757 8 26758 8 26759 8 26760 8 26761 12766 26762 8 26763 8 26764 8 26765 12767 26766 8 26767 8 26768 12768 26769 12769 26770 12770 26771 8 26772 8 26773 12771 26774 12772 26775 12773 26776 12774 26777 8 26778 12775 26779 8 26780 12776 26781 12777 26782 12778 26783 12779 26784 12780 26785 12781 26786 12782 26787 12783 26788 12784 26789 12714 26790 12714 26791 12714 26792 12714 26793 12714 26794 12714 26795 12714 26796 12714 26797 12714 26798 12714 26799 12714 26800 12714 26801 12714 26802 12785 26803 12714 26804 12786 26805 8 26806 12787 26807 12788 26808 8 26809 12789 26810 8 26811 12790 26812 8 26813 8 26814 8 26815 12791 26816 12792 26817 12793 26818 8 26819 12794 26820 8 26821 8 26822 12795 26823 12796 26824 12797 26825 8 26826 8 26827 8 26828 12798 26829 12799 26830 8 26831 8 26832 8 26833 12800 26834 12801 26835 12802 26836 12803 26837 12804 26838 8 26839 12805 26840 8 26841 8 26842 12806 26843 8 26844 12807 26845 12808 26846 8 26847 8 26848 12809 26849 8 26850 8 26851 12810 26852 12811 26853 8 26854 8 26855 12812 26856 12813 26857 12814 26858 12815 26859 12816 26860 12817 26861 12818 26862 12819 26863 12820 26864 12821 26865 12822 26866 8 26867 12823 26868 12824 26869 12825 26870 12826 26871 12827 26872 12828 26873 12829 26874 12830 26875 12831 26876 12832 26877 8 26878 12833 26879 83 26880 12834 26881 8 26882 12835 26883 8 26884 8 26885 8 26886 12836 26887 8 26888 8 26889 12837 26890 12838 26891 12839 26892 12840 26893 12841 26894 8 26895 12842 26896 8 26897 8 26898 12843 26899 8 26900 12844 26901 12845 26902 8 26903 8 26904 12846 26905 8 26906 12847 26907 8 26908 12848 26909 12849 26910 12850 26911 12714 26912 12714 26913 12851 26914 12852 26915 12714 26916 12714 26917 12714 26918 12714 26919 8 26920 12853 26921 8 26922 12854 26923 8 26924 8 26925 12714 26926 12714 26927 12714 26928 12714 26929 12855 26930 8 26931 12856 26932 12857 26933 12858 26934 8 26935 8 26936 8 26937 12859 26938 12860 26939 12861 26940 8 26941 8 26942 8 26943 12862 26944 12863 26945 12864 26946 12865 26947 12866 26948 12867 26949 8 26950 12868 26951 12869 26952 12870 26953 12871 26954 12872 26955 8 26956 12873 26957 12874 26958 12875 26959 12876 26960 12714 26961 12877 26962 12878 26963 12879 26964 8 26965 12714 26966 12714 26967 12714 26968 8 26969 12880 26970 8 26971 12881 26972 8 26973 12882 26974 12883 26975 8 26976 12884 26977 12885 26978 8 26979 12714 26980 8 26981 12714 26982 12714 26983 12714 26984 12886 26985 12887 26986 8 26987 12714 26988 12714 26989 12714 26990 12888 26991 8 26992 8 26993 12889 26994 12890 26995 8 26996 12891 26997 8 26998 12714 26999 12892 27000 12893 27001 12894 27002 12895 27003 8 27004 8 27005 12896 27006 12897 27007 8 27008 8 27009 8 27010 8 27011 8 27012 8 27013 12898 27014 12899 27015 8 27016 8 27017 12900 27018 12901 27019 12902 27020 12903 27021 12904 27022 12905 27023 8 27024 8 27025 8 27026 8 27027 12906 27028 8 27029 8 27030 12907 27031 8 27032 12908 27033 12909 27034 8 27035 8 27036 12910 27037 8 27038 12911 27039 12912 27040 12913 27041 8 27042 8 27043 12914 27044 8 27045 8 27046 8 27047 8 27048 12915 27049 12916 27050 8 27051 8 27052 8 27053 12917 27054 8 27055 12918 27056 8 27057 12919 27058 12920 27059 8 27060 8 27061 83 27062 12921 27063 12922 27064 12923 27065 12924 27066 8 27067 12925 27068 12926 27069 12927 27070 12928 27071 12929 27072 12930 27073 8 27074 12931 27075 12932 27076 8 27077 12933 27078 12934 27079 12935 27080 12936 27081 12937 27082 12938 27083 8 27084 12939 27085 12940 27086 12941 27087 12942 27088 8 27089 83 27090 12943 27091 12944 27092 12945 27093 12946 27094 8 27095 12947 27096 12948 27097 12949 27098 8 27099 12950 27100 8 27101 12951 27102 12952 27103 12953 27104 8 27105 12954 27106 12955 27107 8 27108 83 27109 8 27110 12956 27111 8 27112 12957 27113 8 27114 12958 27115 8 27116 8 27117 12959 27118 8 27119 83 27120 83 27121 83 27122 83 27123 12960 27124 8 27125 12961 27126 12962 27127 12963 27128 8 27129 8 27130 8 27131 12964 27132 12965 27133 12966 27134 8 27135 8 27136 8 27137 12967 27138 12968 27139 12969 27140 12970 27141 12971 27142 12972 27143 8 27144 8 27145 8 27146 8 27147 8 27148 8 27149 8 27150 8 27151 8 27152 8 27153 8 27154 8 27155 8 27156 8 27157 8 27158 8 27159 8 27160 8 27161 8 27162 8 27163 8 27164 8 27165 8 27166 8 27167 8 27168 8 27169 8 27170 8 27171 8 27172 8 27173 8 27174 8 27175 8 27176 8 27177 8 27178 8 27179 8 27180 8 27181 8 27182 8 27183 8 27184 8 27185 8 27186 8 27187 8 27188 8 27189 8 27190 8 27191 8 27192 8 27193 8 27194 8 27195 8 27196 8 27197 8 27198 8 27199 8 27200 8 27201 8 27202 8 27203 8 27204 8 27205 8 27206 8 27207 8 27208 8 27209 8 27210 8 27211 8 27212 8 27213 8 27214 8 27215 8 27216 8 27217 8 27218 8 27219 8 27220 8 27221 8 27222 8 27223 8 27224 12973 27225 8 27226 8 27227 12974 27228 12975 27229 8 27230 12976 27231 12977 27232 12978 27233 12979 27234 12980 27235 12981 27236 12982 27237 12983 27238 12984 27239 12985 27240 12986 27241 12987 27242 12988 27243 12989 27244 8 27245 12990 27246 12991 27247 12992 27248 12993 27249 12994 27250 12995 27251 12996 27252 12997 27253 12998 27254 12999 27255 13000 27256 8 27257 13001 27258 8 27259 13002 27260 8 27261 13003 27262 8 27263 13004 27264 13005 27265 8 27266 8 27267 13006 27268 8 27269 8 27270 8 27271 8 27272 13007 27273 8 27274 13008 27275 8 27276 13009 27277 13010 27278 13011 27279 13012 27280 13013 27281 8 27282 13014 27283 13015 27284 13016 27285 13017 27286 13018 27287 13019 27288 13020 27289 13021 27290 13022 27291 8 27292 13023 27293 13024 27294 13025 27295 13026 27296 13027 27297 13028 27298 13029 27299 8 27300 13030 27301 13031 27302 8 27303 13032 27304 8 27305 13033 27306 13034 27307 8 27308 13035 27309 8 27310 8 27311 13036 27312 13037 27313 8 27314 13038 27315 13039 27316 13040 27317 8 27318 13041 27319 13042 27320 13043 27321 13044 27322 8 27323 8 27324 13045 27325 13046 27326 8 27327 13047 27328 13048 27329 13049 27330 13050 27331 8 27332 13051 27333 13052 27334 13053 27335 13054 27336 13055 27337 13056 27338 8 27339 13057 27340 13058 27341 8 27342 13059 27343 13060 27344 13061 27345 13062 27346 13063 27347 13064 27348 8 27349 13065 27350 13066 27351 13067 27352 13068 27353 8 27354 13069 27355 13070 27356 13071 27357 8 27358 13072 27359 8 27360 13073 27361 13074 27362 13075 27363 8 27364 13076 27365 13077 27366 8 27367 8 27368 13078 27369 8 27370 13079 27371 8 27372 13080 27373 8 27374 13081 27375 8 27376 13082 27377 13083 27378 13084 27379 13085 27380 13086 27381 13087 27382 13088 27383 13089 27384 13090 27385 8 27386 8 27387 13091 27388 13092 27389 8 27390 8 27391 13093 27392 13094 27393 13095 27394 13096 27395 13097 27396 13098 27397 13099 27398 13100 27399 8 27400 8 27401 13101 27402 13102 27403 8 27404 13103 27405 8 27406 13104 27407 13105 27408 13106 27409 13107 27410 13108 27411 8 27412 13109 27413 13110 27414 13111 27415 13112 27416 8 27417 8 27418 13113 27419 13114 27420 8 27421 13115 27422 8 27423 13116 27424 8 27425 8 27426 13117 27427 13118 27428 13119 27429 13120 27430 13121 27431 13122 27432 13123 27433 13124 27434 13125 27435 13126 27436 13127 27437 8 27438 8 27439 13128 27440 13129 27441 13130 27442 13131 27443 13132 27444 13133 27445 13134 27446 8 27447 8 27448 13135 27449 13136 27450 8 27451 13137 27452 13138 27453 13139 27454 8 27455 13140 27456 13141 27457 13142 27458 13143 27459 8 27460 8 27461 13144 27462 13145 27463 8 27464 13146 27465 13147 27466 13148 27467 13149 27468 13150 27469 8 27470 8 27471 8 27472 8 27473 13151 27474 13152 27475 13153 27476 13154 27477 8 27478 13155 27479 13156 27480 13157 27481 13158 27482 13159 27483 13160 27484 8 27485 13161 27486 13162 27487 8 27488 13163 27489 13164 27490 13165 27491 13166 27492 13167 27493 13168 27494 8 27495 13169 27496 13170 27497 13171 27498 13172 27499 8 27500 13173 27501 13174 27502 13175 27503 13176 27504 8 27505 13177 27506 13178 27507 13179 27508 8 27509 13180 27510 8 27511 13181 27512 13182 27513 13183 27514 8 27515 13184 27516 13185 27517 8 27518 8 27519 13186 27520 8 27521 13187 27522 8 27523 13188 27524 13189 27525 13190 27526 8 27527 8 27528 13191 27529 13192 27530 8 27531 13193 27532 13194 27533 13195 27534 13196 27535 13197 27536 8 27537 13198 27538 13199 27539 13200 27540 8 27541 13201 27542 13202 27543 8 27544 13203 27545 8 27546 13204 27547 8 27548 13205 27549 13206 27550 13207 27551 13208 27552 13209 27553 13210 27554 13211 27555 13212 27556 13213 27557 8 27558 8 27559 13214 27560 13215 27561 8 27562 8 27563 13216 27564 13217 27565 13218 27566 13219 27567 13220 27568 13221 27569 13222 27570 13223 27571 8 27572 8 27573 13224 27574 13225 27575 8 27576 13226 27577 13227 27578 13228 27579 13229 27580 13230 27581 13231 27582 8 27583 13232 27584 8 27585 13233 27586 13234 27587 13235 27588 13236 27589 13237 27590 8 27591 13238 27592 13239 27593 13240 27594 13241 27595 8 27596 8 27597 13242 27598 13243 27599 8 27600 13244 27601 8 27602 13245 27603 8 27604 13246 27605 13247 27606 8 27607 8 27608 13248 27609 13249 27610 8 27611 13250 27612 13251 27613 13252 27614 13253 27615 13254 27616 13255 27617 8 27618 13256 27619 8 27620 8 27621 13257 27622 13258 27623 8 27624 13259 27625 13260 27626 13261 27627 13262 27628 13263 27629 8 27630 13264 27631 13265 27632 13266 27633 13267 27634 83 27635 13268 27636 13269 27637 13270 27638 8 27639 83 27640 13271 27641 13272 27642 13273 27643 13274 27644 13275 27645 8 27646 83 27647 83 27648 8 27649 13276 27650 13277 27651 13278 27652 8 27653 13279 27654 13280 27655 13281 27656 13282 27657 13283 27658 13284 27659 13285 27660 13286 27661 13287 27662 13288 27663 8 27664 13289 27665 13290 27666 13291 27667 13292 27668 13293 27669 13294 27670 13295 27671 8 27672 13296 27673 13297 27674 8 27675 13298 27676 8 27677 13299 27678 13300 27679 8 27680 13301 27681 13302 27682 13303 27683 13304 27684 13305 27685 8 27686 8 27687 8 27688 13306 27689 13307 27690 8 27691 13308 27692 13309 27693 13310 27694 8 27695 13311 27696 13312 27697 13313 27698 13314 27699 8 27700 8 27701 13315 27702 13316 27703 8 27704 13317 27705 13318 27706 13319 27707 13320 27708 8 27709 13321 27710 13322 27711 13323 27712 13324 27713 13325 27714 13326 27715 8 27716 13327 27717 13328 27718 8 27719 13329 27720 13330 27721 13331 27722 13332 27723 13333 27724 13334 27725 8 27726 13335 27727 13336 27728 13337 27729 13338 27730 8 27731 13339 27732 13340 27733 13341 27734 8 27735 13342 27736 8 27737 13343 27738 13344 27739 13345 27740 8 27741 13346 27742 13347 27743 8 27744 8 27745 13348 27746 8 27747 13349 27748 8 27749 13350 27750 13351 27751 13352 27752 8 27753 13353 27754 8 27755 13354 27756 8 27757 13355 27758 13356 27759 13357 27760 13358 27761 13359 27762 13360 27763 13361 27764 13362 27765 13363 27766 8 27767 8 27768 13364 27769 13365 27770 8 27771 8 27772 13366 27773 13367 27774 13368 27775 13369 27776 13370 27777 13371 27778 13372 27779 13373 27780 8 27781 8 27782 13374 27783 13375 27784 8 27785 13376 27786 8 27787 13377 27788 13378 27789 13379 27790 13380 27791 13381 27792 8 27793 13382 27794 13383 27795 13384 27796 13385 27797 8 27798 8 27799 13386 27800 13387 27801 8 27802 13388 27803 8 27804 13389 27805 8 27806 13390 27807 13391 27808 8 27809 13392 27810 8 27811 13393 27812 8 27813 13394 27814 13395 27815 13396 27816 13397 27817 13398 27818 13399 27819 13400 27820 13401 27821 13402 27822 8 27823 8 27824 13403 27825 13404 27826 8 27827 8 27828 13405 27829 13406 27830 13407 27831 13408 27832 13409 27833 13410 27834 13411 27835 13412 27836 8 27837 8 27838 13413 27839 13414 27840 8 27841 13415 27842 8 27843 13416 27844 8 27845 13417 27846 13418 27847 8 27848 13419 27849 13420 27850 13421 27851 13422 27852 13423 27853 13424 27854 8 27855 13425 27856 13426 27857 8 27858 13427 27859 8 27860 83 27861 13428 27862 13429 27863 13430 27864 13431 27865 13432 27866 8 27867 13433 27868 13434 27869 13435 27870 13436 27871 8 27872 8 27873 13437 27874 13438 27875 8 27876 8 27877 13439 27878 8 27879 13440 27880 8 27881 83 27882 13441 27883 13442 27884 13443 27885 13444 27886 13445 27887 13446 27888 8 27889 13447 27890 8 27891 83 27892 13448 27893 13449 27894 8 27895 83 27896 13450 27897 13451 27898 8 27899 83 27900 83 27901 83 27902 13452 27903 8 27904 8 27905 13453 27906 13454 27907 8 27908 13455 27909 8 27910 83))
  "The generated goto-tables.")

(defvar
  phps-mode-parser--distinct-goto-tables
  #s(hash-table size 18699 test equal rehash-size 1.5 rehash-threshold 0.8125 data (0 ((top_statement_list 1)) 1 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 8) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS 16) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 19) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_ENUM 30) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_HALT_COMPILER 42) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INTERFACE 48) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 56) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 66) (T_START_HEREDOC 67) (T_STATIC 68) (T_STRING 69) (T_STRING_CAST 70) (T_SWITCH 71) (T_THROW 72) (T_TRAIT 73) (T_TRAIT_C 74) (T_TRY 75) (T_UNSET 76) (T_UNSET_CAST 77) (T_USE 78) (T_VARIABLE 79) (T_WHILE 80) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 85) (alt_if_stmt_without_else 86) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 89) (attributes 90) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 94) (class_modifier 95) (class_modifiers 96) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 100) (expr 101) (fn 102) (fully_dereferenceable 103) #'104 (function_call 105) (function_declaration_statement 106) (if_stmt 107) (if_stmt_without_else 108) (inline_function 109) (interface_declaration_statement 110) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 117) (static_member 118) (top_statement 119) (trait_declaration_statement 120) (variable 121) (variable_class_name 122) ("{" 123) ("~" 124)) 2 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27910) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 3 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 27908) (encaps_var 5158)) 4 (("$" 4) (T_VARIABLE 79) (simple_variable 27904) ("{" 27905)) 5 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27902) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27901) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27900) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8 nil 9 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27899) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 10 (("(" 27896)) 11 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27895) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 12 ((T_NAME_FULLY_QUALIFIED 6350) (T_NAME_QUALIFIED 6351) (T_NAME_RELATIVE 6352) (T_STATIC 6353) (T_STRING 6354) (attribute_decl 6355) (attribute_group 27892) (class_name 6357) (name 6358)) 13 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27891) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 14 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 27889) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 15 ((T_STRING 27882)) 16 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27881) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 17 ((T_STRING 27602) (const_decl 27603) (const_list 27879)) 18 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 27877) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 19 (("\"" 1489) ("$" 25138) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 25139) ("[" 1508) (array_object_dereferenceable 25140) (callable_expr 25141) (callable_variable 25142) (class_constant 1512) (class_name 25143) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 25144) (name 25145) (simple_variable 25146) (static_member 25147) (variable 27876) (variable_class_name 25149)) 20 (("(" 27867)) 21 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27861) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 22 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27860) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 23 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 27858) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 24 (("(" 27855)) 25 ((T_STRING 27848)) 26 (("(" 27845)) 27 (("(" 27841) (exit_expr 27842)) 28 (("(" 27828)) 29 (("(" 27813)) 30 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 27811) (simple_variable 17412)) 31 ((T_STRING 27809)) 32 (("(" 27806)) 33 (("(" 27649)) 34 (("\"" 1489) ("$" 25138) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 25139) ("[" 1508) (array_object_dereferenceable 25140) (callable_expr 25141) (callable_variable 25142) (class_constant 1512) (class_name 25143) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 25144) (name 25145) (simple_variable 25146) (static_member 25147) (variable 27648) (variable_class_name 25149)) 35 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27647) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 36 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27646) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 37 ((T_STRING 27640)) 38 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27639) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 39 (("(" 27635)) 40 (("(" 27630)) 41 (("(" 27624)) 42 ((T_ABSTRACT 27144) (T_ARRAY 27145) (T_AS 27146) (T_BREAK 27147) (T_CALLABLE 27148) (T_CASE 27149) (T_CATCH 27150) (T_CLASS 27151) (T_CLASS_C 27152) (T_CLONE 27153) (T_CONST 27154) (T_CONTINUE 27155) (T_DECLARE 27156) (T_DEFAULT 27157) (T_DIR 27158) (T_DO 27159) (T_ECHO 27160) (T_ELSE 27161) (T_ELSEIF 27162) (T_EMPTY 27163) (T_ENDDECLARE 27164) (T_ENDFOR 27165) (T_ENDFOREACH 27166) (T_ENDIF 27167) (T_ENDSWITCH 27168) (T_ENDWHILE 27169) (T_ENUM 27170) (T_EVAL 27171) (T_EXIT 27172) (T_EXTENDS 27173) (T_FILE 27174) (T_FINAL 27175) (T_FINALLY 27176) (T_FN 27177) (T_FOR 27178) (T_FOREACH 27179) (T_FUNCTION 27180) (T_FUNC_C 27181) (T_GLOBAL 27182) (T_GOTO 27183) (T_IF 27184) (T_IMPLEMENTS 27185) (T_INCLUDE 27186) (T_INCLUDE_ONCE 27187) (T_INSTANCEOF 27188) (T_INSTEADOF 27189) (T_INTERFACE 27190) (T_ISSET 27191) (T_LINE 27192) (T_LIST 27193) (T_LOGICAL_AND 27194) (T_LOGICAL_OR 27195) (T_LOGICAL_XOR 27196) (T_MATCH 27197) (T_METHOD_C 27198) (T_NAMESPACE 27199) (T_NAME_QUALIFIED 27200) (T_NEW 27201) (T_NS_C 27202) (T_PRINT 27203) (T_PRIVATE 27204) (T_PROTECTED 27205) (T_PUBLIC 27206) (T_READONLY 27207) (T_REQUIRE 27208) (T_REQUIRE_ONCE 27209) (T_RETURN 27210) (T_STATIC 27211) (T_STRING 27212) (T_SWITCH 27213) (T_THROW 27214) (T_TRAIT 27215) (T_TRAIT_C 27216) (T_TRY 27217) (T_UNSET 27218) (T_USE 27219) (T_VAR 27220) (T_WHILE 27221) (T_YIELD 27222) (identifier 27223) (namespace_declaration_name 27224) (reserved_non_modifiers 27225) (semi_reserved 27226) ("{" 27227)) 43 (("$" 25661) ("(" 26304) (T_ATTRIBUTE 2853) (T_CLASS 27123) (T_NAME_FULLY_QUALIFIED 26306) (T_NAME_QUALIFIED 26307) (T_NAME_RELATIVE 26308) (T_STATIC 26309) (T_STRING 26310) (T_VARIABLE 25662) (anonymous_class 27124) (attribute 2862) (attributes 27125) (class_name 26313) (class_name_reference 27126) (name 26315) (new_variable 26316) (simple_variable 26317)) 44 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27122) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 45 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27121) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 46 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27120) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 47 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27119) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 48 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 27117) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 49 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 27112) (T_END_HEREDOC 27113) (T_VARIABLE 2822) (encaps_list 27114) (encaps_var 2824)) 50 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 27110)) 51 ((":" 27109)) 52 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27108) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 53 (("(" 27090)) 54 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27089) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 55 ((T_STRING 27084)) 56 (("{" 27067)) 57 (("(" 27062)) 58 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27061) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 59 ((T_CONST 27008) (T_FUNCTION 27009) (T_NAME_FULLY_QUALIFIED 27010) (T_NAME_QUALIFIED 27011) (T_STRING 27012) (legacy_namespace_name 27013) (mixed_group_use_declaration 27014) (namespace_name 27015) (use_declaration 27016) (use_declarations 27017) (use_type 27018)) 60 (("(" 26999)) 61 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26564) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 62 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 26502) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 63 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 26498) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 64 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 26496) (encaps_list 2729) (encaps_var 2730)) 65 ((T_ELSE 26489) (T_ELSEIF 15767) (T_ENDIF 26490)) 66 ((T_NULLSAFE_OBJECT_OPERATOR 26472) (T_OBJECT_OPERATOR 26473) ("[" 26474) ("{" 26475)) 67 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 26415) (T_ENUM 26416) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 26417) (T_STATIC 15850) (T_TRAIT 26418) (attribute 15851) (attributed_statement 26419) (class_declaration_statement 26420) (class_modifier 95) (class_modifiers 26421) (enum_declaration_statement 26422) (fn 102) #'26423 (function_declaration_statement 26424) (inline_function 15853) (interface_declaration_statement 26425) (trait_declaration_statement 26426)) 68 (("(" 25099) (argument_list 26414)) 69 ((T_ABSTRACT 10) (T_CLASS 26406) (T_FINAL 34) (class_modifier 15842)) 70 ((T_PAAMAYIM_NEKUDOTAYIM 26401)) 71 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 26400) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 72 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 25351)) 73 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15826) (ampersand 15827) (returns_ref 25339)) 74 ((T_ELSE 25189) (T_ELSEIF 25190)) 75 (("(" 25099) (argument_list 25188)) 76 (("=" 25107) (T_AND_EQUAL 25108) (T_COALESCE_EQUAL 25109) (T_CONCAT_EQUAL 25110) (T_DEC 25111) (T_DIV_EQUAL 25112) (T_INC 25113) (T_MINUS_EQUAL 25114) (T_MOD_EQUAL 25115) (T_MUL_EQUAL 25116) (T_OR_EQUAL 25117) (T_PLUS_EQUAL 25118) (T_POW_EQUAL 25119) (T_SL_EQUAL 25120) (T_SR_EQUAL 25121) (T_XOR_EQUAL 25122)) 77 ((T_PAAMAYIM_NEKUDOTAYIM 25015)) 78 ((inner_statement_list 25013)) 79 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 131) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 80 ((T_NAME_FULLY_QUALIFIED 6350) (T_NAME_QUALIFIED 6351) (T_NAME_RELATIVE 6352) (T_STATIC 6353) (T_STRING 6354) (attribute_decl 6355) (attribute_group 25010) (class_name 6357) (name 6358)) 81 ((T_FN 35) (T_FUNCTION 126) (fn 102) #'132 (inline_function 16409)) 82 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 15850) (attribute 2654) (fn 102) #'132 (inline_function 15853)) 83 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 84 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 136)) 85 ((backup_doc_comment 137)) 86 (("(" 138)) 87 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 146)) 88 ((T_NAME_FULLY_QUALIFIED 6350) (T_NAME_QUALIFIED 6351) (T_NAME_RELATIVE 6352) (T_STATIC 6353) (T_STRING 6354) (attribute_decl 6355) (attribute_group 25007) (class_name 6357) (name 6358)) 89 ((T_ATTRIBUTE 139) (attribute 25005) (optional_property_modifiers 144) (parameter 25006)) 90 (("," 25002) (possible_comma 25003)) 91 (("?" 24946) (T_ARRAY 24947) (T_CALLABLE 24948) (T_NAME_FULLY_QUALIFIED 24949) (T_NAME_QUALIFIED 24950) (T_NAME_RELATIVE 24951) (T_PRIVATE 24952) (T_PROTECTED 24953) (T_PUBLIC 24954) (T_READONLY 24955) (T_STRING 24956) (intersection_type_without_static 24957) (name 24958) (optional_type_without_static 24959) (property_modifier 24960) (type_expr_without_static 24961) (type_without_static 24962) (union_type_without_static 24963)) 92 ((")" 147)) 93 ((T_USE 148) (lexical_vars 149)) 94 (("(" 24934)) 95 ((":" 150) (return_type 151)) 96 (("?" 24884) (T_ARRAY 24885) (T_CALLABLE 24886) (T_NAME_FULLY_QUALIFIED 24887) (T_NAME_QUALIFIED 24888) (T_NAME_RELATIVE 24889) (T_STATIC 24890) (T_STRING 24891) (intersection_type 24892) (name 24893) (type 24894) (type_expr 24895) (type_without_static 24896) (union_type 24897)) 97 ((backup_fn_flags 152)) 98 (("{" 153)) 99 ((inner_statement_list 154)) 100 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 196) ("~" 124)) 101 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 24882) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 102 ((T_STRING 24875)) 103 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 24873) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 104 (("(" 24864)) 105 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24858) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 106 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 24856) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 107 ((T_STRING 24849)) 108 (("(" 24836)) 109 (("(" 24821)) 110 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 24819) (simple_variable 17412)) 111 ((T_STRING 24817)) 112 (("(" 24814)) 113 (("(" 24810)) 114 ((T_STRING 24804)) 115 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 24802) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 116 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 24800)) 117 ((":" 24799)) 118 (("(" 24781)) 119 ((T_STRING 24776)) 120 (("{" 24759)) 121 (("(" 24754)) 122 (("(" 24745)) 123 ((T_ELSE 24738) (T_ELSEIF 15767) (T_ENDIF 24739)) 124 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 157) (T_ENUM 162) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 170) (T_STATIC 15850) (T_TRAIT 175) (attribute 15851) (attributed_statement 24737) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (enum_declaration_statement 185) (fn 102) #'187 (function_declaration_statement 188) (inline_function 15853) (interface_declaration_statement 192) (trait_declaration_statement 194)) 125 ((T_ABSTRACT 10) (T_CLASS 24729) (T_FINAL 34) (class_modifier 15842)) 126 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 24728) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 127 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15826) (ampersand 15827) (returns_ref 24716)) 128 ((T_ELSE 200) (T_ELSEIF 201)) 129 ((inner_statement_list 198)) 130 ((backup_fn_flags 197)) 131 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 199) ("~" 124)) 132 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 186) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24715) (static_member 118) (variable 121) (variable_class_name 122) ("{" 195) ("~" 124)) 133 (("(" 202)) 134 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 264) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 135 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24714) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 136 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 24712) (encaps_var 5158)) 137 (("$" 205) (T_VARIABLE 251) (simple_variable 24708) ("{" 24709)) 138 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24706) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 139 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24705) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 140 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24704) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 141 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24703) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 142 (("(" 24700)) 143 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24699) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 144 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24698) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 145 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24697) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 146 (("\"" 1489) ("$" 14623) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 14624) ("[" 1508) (array_object_dereferenceable 14625) (callable_expr 14626) (callable_variable 14627) (class_constant 1512) (class_name 14628) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 14629) (name 14630) (simple_variable 14631) (static_member 14632) (variable 24696) (variable_class_name 14634)) 147 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24695) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 148 (("(" 24692)) 149 (("(" 24689)) 150 (("(" 24685) (exit_expr 24686)) 151 (("\"" 1489) ("$" 14623) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 14624) ("[" 1508) (array_object_dereferenceable 14625) (callable_expr 14626) (callable_variable 14627) (class_constant 1512) (class_name 14628) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 14629) (name 14630) (simple_variable 14631) (static_member 14632) (variable 24684) (variable_class_name 14634)) 152 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24683) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 153 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24682) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 154 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24681) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 155 (("(" 24677)) 156 (("(" 24672)) 157 (("(" 24666)) 158 (("$" 14996) ("(" 15639) (T_ATTRIBUTE 2853) (T_CLASS 24645) (T_NAME_FULLY_QUALIFIED 15641) (T_NAME_QUALIFIED 15642) (T_NAME_RELATIVE 15643) (T_STATIC 15644) (T_STRING 15645) (T_VARIABLE 14997) (anonymous_class 24646) (attribute 2862) (attributes 24647) (class_name 15648) (class_name_reference 24648) (name 15650) (new_variable 15651) (simple_variable 15652)) 159 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24644) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 160 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24643) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 161 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24642) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 162 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24641) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 163 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 24636) (T_END_HEREDOC 24637) (T_VARIABLE 2822) (encaps_list 24638) (encaps_var 2824)) 164 ((T_FN 35) (T_FUNCTION 126) (fn 265) #'266 (inline_function 24635)) 165 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24634) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 166 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24633) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 167 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24632) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 168 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24197) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 169 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24135) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 170 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 24131) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 171 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 24129) (encaps_list 2729) (encaps_var 2730)) 172 ((T_NULLSAFE_OBJECT_OPERATOR 24112) (T_OBJECT_OPERATOR 24113) ("[" 24114) ("{" 24115)) 173 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 24109) (attribute 2654) (fn 265) #'266 (inline_function 24110)) 174 (("(" 14584) (argument_list 24108)) 175 ((T_PAAMAYIM_NEKUDOTAYIM 24103)) 176 (("%" 280) (")" 15735) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 177 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 14686)) 178 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 14674)) 179 (("(" 14584) (argument_list 14673)) 180 (("=" 14592) (T_AND_EQUAL 14593) (T_COALESCE_EQUAL 14594) (T_CONCAT_EQUAL 14595) (T_DEC 14596) (T_DIV_EQUAL 14597) (T_INC 14598) (T_MINUS_EQUAL 14599) (T_MOD_EQUAL 14600) (T_MUL_EQUAL 14601) (T_OR_EQUAL 14602) (T_PLUS_EQUAL 14603) (T_POW_EQUAL 14604) (T_SL_EQUAL 14605) (T_SR_EQUAL 14606) (T_XOR_EQUAL 14607)) 181 ((T_PAAMAYIM_NEKUDOTAYIM 14500)) 182 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 279) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 183 (("%" 280) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 184 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14499) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 185 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14498) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 186 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14497) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 187 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14496) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 188 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14495) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 189 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14494) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 190 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14493) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 191 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14492) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 192 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 14487) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 14488) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 193 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14486) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 194 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14485) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 195 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14484) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 196 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14483) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 197 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14482) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 198 (("$" 325) ("(" 326) (T_NAME_FULLY_QUALIFIED 327) (T_NAME_QUALIFIED 328) (T_NAME_RELATIVE 329) (T_STATIC 330) (T_STRING 331) (T_VARIABLE 332) (class_name 333) (class_name_reference 334) (name 335) (new_variable 336) (simple_variable 337)) 199 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 324) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 200 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 323) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 201 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 322) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 202 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 321) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 203 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 320) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 204 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 319) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 205 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 318) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 206 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 317) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 207 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 316) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 208 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 315) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 209 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 314) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 210 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 313) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 211 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 312) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 212 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 311) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 213 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 310) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 214 (("$" 325) (T_VARIABLE 332) (simple_variable 14478) ("{" 14479)) 215 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14476) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 216 ((T_PAAMAYIM_NEKUDOTAYIM 14474)) 217 ((T_NULLSAFE_OBJECT_OPERATOR 338) (T_OBJECT_OPERATOR 339) (T_PAAMAYIM_NEKUDOTAYIM 340) ("[" 341) ("{" 342)) 218 (("$" 325) (T_STRING 14467) (T_VARIABLE 332) (property_name 14473) (simple_variable 14469) ("{" 14470)) 219 (("$" 325) (T_STRING 14467) (T_VARIABLE 332) (property_name 14468) (simple_variable 14469) ("{" 14470)) 220 (("$" 325) (T_VARIABLE 332) (simple_variable 14466)) 221 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 14464) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 222 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 404) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 223 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14463) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 224 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 14461) (encaps_var 5158)) 225 (("$" 345) (T_VARIABLE 391) (simple_variable 14457) ("{" 14458)) 226 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14455) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 227 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14454) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 228 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14453) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 229 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14452) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 230 (("(" 14449)) 231 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14448) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 232 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14447) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 233 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14446) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 234 (("\"" 1489) ("$" 12739) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 12740) ("[" 1508) (array_object_dereferenceable 12741) (callable_expr 12742) (callable_variable 12743) (class_constant 1512) (class_name 12744) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 12745) (name 12746) (simple_variable 12747) (static_member 12748) (variable 14445) (variable_class_name 12750)) 235 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14444) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 236 (("(" 14441)) 237 (("(" 14438)) 238 (("(" 14434) (exit_expr 14435)) 239 (("\"" 1489) ("$" 12739) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 12740) ("[" 1508) (array_object_dereferenceable 12741) (callable_expr 12742) (callable_variable 12743) (class_constant 1512) (class_name 12744) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 12745) (name 12746) (simple_variable 12747) (static_member 12748) (variable 14433) (variable_class_name 12750)) 240 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14432) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 241 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14431) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 242 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14430) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 243 (("(" 14426)) 244 (("(" 14421)) 245 (("(" 14415)) 246 (("$" 13112) ("(" 13755) (T_ATTRIBUTE 2853) (T_CLASS 14394) (T_NAME_FULLY_QUALIFIED 13757) (T_NAME_QUALIFIED 13758) (T_NAME_RELATIVE 13759) (T_STATIC 13760) (T_STRING 13761) (T_VARIABLE 13113) (anonymous_class 14395) (attribute 2862) (attributes 14396) (class_name 13764) (class_name_reference 14397) (name 13766) (new_variable 13767) (simple_variable 13768)) 247 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14393) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 248 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14392) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 249 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14391) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 250 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14390) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 251 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 14385) (T_END_HEREDOC 14386) (T_VARIABLE 2822) (encaps_list 14387) (encaps_var 2824)) 252 ((T_FN 35) (T_FUNCTION 126) (fn 405) #'406 (inline_function 14384)) 253 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14383) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 254 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14382) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 255 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14381) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 256 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 13946) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 257 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13884) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 258 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 13880) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 259 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 13878) (encaps_list 2729) (encaps_var 2730)) 260 ((T_NULLSAFE_OBJECT_OPERATOR 13861) (T_OBJECT_OPERATOR 13862) ("[" 13863) ("{" 13864)) 261 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 13858) (attribute 2654) (fn 405) #'406 (inline_function 13859)) 262 (("(" 12700) (argument_list 13857)) 263 ((T_PAAMAYIM_NEKUDOTAYIM 13852)) 264 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13851)) 265 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 12802)) 266 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 12790)) 267 (("(" 12700) (argument_list 12789)) 268 (("=" 12708) (T_AND_EQUAL 12709) (T_COALESCE_EQUAL 12710) (T_CONCAT_EQUAL 12711) (T_DEC 12712) (T_DIV_EQUAL 12713) (T_INC 12714) (T_MINUS_EQUAL 12715) (T_MOD_EQUAL 12716) (T_MUL_EQUAL 12717) (T_OR_EQUAL 12718) (T_PLUS_EQUAL 12719) (T_POW_EQUAL 12720) (T_SL_EQUAL 12721) (T_SR_EQUAL 12722) (T_XOR_EQUAL 12723)) 269 ((T_PAAMAYIM_NEKUDOTAYIM 12616)) 270 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 419) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 271 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449)) 272 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12615) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 273 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12614) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 274 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12613) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 275 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12612) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 276 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12611) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 277 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12610) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 278 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12609) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 279 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12608) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 280 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 12603) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 12604) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 281 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12602) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 282 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12601) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 283 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12600) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 284 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12599) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 285 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12598) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 286 (("$" 465) ("(" 466) (T_NAME_FULLY_QUALIFIED 467) (T_NAME_QUALIFIED 468) (T_NAME_RELATIVE 469) (T_STATIC 470) (T_STRING 471) (T_VARIABLE 472) (class_name 473) (class_name_reference 474) (name 475) (new_variable 476) (simple_variable 477)) 287 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 464) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 288 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 463) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 289 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 462) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 290 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 461) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 291 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 460) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 292 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 459) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 293 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 458) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 294 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 457) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 295 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 456) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 296 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 455) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 297 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 454) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 298 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 453) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 299 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 452) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 300 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 451) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 301 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 450) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 302 (("$" 465) (T_VARIABLE 472) (simple_variable 12594) ("{" 12595)) 303 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 12592) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 304 ((T_PAAMAYIM_NEKUDOTAYIM 12590)) 305 ((T_NULLSAFE_OBJECT_OPERATOR 478) (T_OBJECT_OPERATOR 479) (T_PAAMAYIM_NEKUDOTAYIM 480) ("[" 481) ("{" 482)) 306 (("$" 465) (T_STRING 12583) (T_VARIABLE 472) (property_name 12589) (simple_variable 12585) ("{" 12586)) 307 (("$" 465) (T_STRING 12583) (T_VARIABLE 472) (property_name 12584) (simple_variable 12585) ("{" 12586)) 308 (("$" 465) (T_VARIABLE 472) (simple_variable 12582)) 309 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 555) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 310 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 483) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 311 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 484)) 312 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12581) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 313 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 12579) (encaps_var 5158)) 314 (("$" 487) (T_VARIABLE 533) (simple_variable 12575) ("{" 12576)) 315 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 12573) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 316 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12572) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 317 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12571) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 318 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12570) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 319 (("(" 12567)) 320 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12566) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 321 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12565) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 322 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12564) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 323 (("\"" 1489) ("$" 10857) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 10858) ("[" 1508) (array_object_dereferenceable 10859) (callable_expr 10860) (callable_variable 10861) (class_constant 1512) (class_name 10862) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 10863) (name 10864) (simple_variable 10865) (static_member 10866) (variable 12563) (variable_class_name 10868)) 324 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12562) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 325 (("(" 12559)) 326 (("(" 12556)) 327 (("(" 12552) (exit_expr 12553)) 328 (("\"" 1489) ("$" 10857) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 10858) ("[" 1508) (array_object_dereferenceable 10859) (callable_expr 10860) (callable_variable 10861) (class_constant 1512) (class_name 10862) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 10863) (name 10864) (simple_variable 10865) (static_member 10866) (variable 12551) (variable_class_name 10868)) 329 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12550) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 330 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12549) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 331 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12548) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 332 (("(" 12544)) 333 (("(" 12539)) 334 (("(" 12533)) 335 (("$" 11231) ("(" 11874) (T_ATTRIBUTE 2853) (T_CLASS 12512) (T_NAME_FULLY_QUALIFIED 11876) (T_NAME_QUALIFIED 11877) (T_NAME_RELATIVE 11878) (T_STATIC 11879) (T_STRING 11880) (T_VARIABLE 11232) (anonymous_class 12513) (attribute 2862) (attributes 12514) (class_name 11883) (class_name_reference 12515) (name 11885) (new_variable 11886) (simple_variable 11887)) 336 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12511) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 337 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12510) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 338 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12509) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 339 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12508) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 340 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 12503) (T_END_HEREDOC 12504) (T_VARIABLE 2822) (encaps_list 12505) (encaps_var 2824)) 341 ((T_FN 35) (T_FUNCTION 126) (fn 547) #'548 (inline_function 12502)) 342 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12501) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 343 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12500) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 344 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12499) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 345 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12064) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 346 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12002) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 347 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 11998) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 348 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 11996) (encaps_list 2729) (encaps_var 2730)) 349 ((T_NULLSAFE_OBJECT_OPERATOR 11979) (T_OBJECT_OPERATOR 11980) ("[" 11981) ("{" 11982)) 350 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 11976) (attribute 2654) (fn 547) #'548 (inline_function 11977)) 351 (("(" 10818) (argument_list 11975)) 352 ((T_PAAMAYIM_NEKUDOTAYIM 11970)) 353 (("%" 563) ("*" 564) ("+" 565) ("-" 566) ("." 567) ("/" 568) ("<" 569) (">" 570) ("?" 571) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 572) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 573) (T_BOOLEAN_AND 574) (T_BOOLEAN_OR 575) (T_COALESCE 576) (T_INSTANCEOF 577) (T_IS_EQUAL 578) (T_IS_GREATER_OR_EQUAL 579) (T_IS_IDENTICAL 580) (T_IS_NOT_EQUAL 581) (T_IS_NOT_IDENTICAL 582) (T_IS_SMALLER_OR_EQUAL 583) (T_LOGICAL_AND 584) (T_LOGICAL_OR 585) (T_LOGICAL_XOR 586) (T_POW 587) (T_SL 588) (T_SPACESHIP 589) (T_SR 590) ("^" 591) ("|" 592)) 354 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 10921)) 355 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 10909)) 356 (("(" 10818) (argument_list 10908)) 357 (("]" 10907)) 358 (("=" 10826) (T_AND_EQUAL 10827) (T_COALESCE_EQUAL 10828) (T_CONCAT_EQUAL 10829) (T_DEC 10830) (T_DIV_EQUAL 10831) (T_INC 10832) (T_MINUS_EQUAL 10833) (T_MOD_EQUAL 10834) (T_MUL_EQUAL 10835) (T_OR_EQUAL 10836) (T_PLUS_EQUAL 10837) (T_POW_EQUAL 10838) (T_SL_EQUAL 10839) (T_SR_EQUAL 10840) (T_XOR_EQUAL 10841)) 359 ((T_PAAMAYIM_NEKUDOTAYIM 10734)) 360 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 562) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 361 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10733) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 362 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10732) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 363 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10731) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 364 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10730) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 365 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10729) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 366 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10728) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 367 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10727) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 368 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10726) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 369 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 657) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 713) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 370 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 650) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 371 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 649) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 372 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 648) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 373 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 647) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 374 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 646) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 375 (("$" 608) ("(" 609) (T_NAME_FULLY_QUALIFIED 610) (T_NAME_QUALIFIED 611) (T_NAME_RELATIVE 612) (T_STATIC 613) (T_STRING 614) (T_VARIABLE 615) (class_name 616) (class_name_reference 617) (name 618) (new_variable 619) (simple_variable 620)) 376 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 607) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 377 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 606) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 378 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 605) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 379 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 604) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 380 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 603) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 381 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 602) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 382 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 601) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 383 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 600) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 384 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 599) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 385 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 598) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 386 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 597) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 387 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 596) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 388 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 595) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 389 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 594) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 390 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 593) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 391 (("$" 608) (T_VARIABLE 615) (simple_variable 642) ("{" 643)) 392 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 640) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 393 ((T_PAAMAYIM_NEKUDOTAYIM 638)) 394 ((T_NULLSAFE_OBJECT_OPERATOR 621) (T_OBJECT_OPERATOR 622) (T_PAAMAYIM_NEKUDOTAYIM 623) ("[" 624) ("{" 625)) 395 (("$" 608) (T_STRING 631) (T_VARIABLE 615) (property_name 637) (simple_variable 633) ("{" 634)) 396 (("$" 608) (T_STRING 631) (T_VARIABLE 615) (property_name 632) (simple_variable 633) ("{" 634)) 397 (("$" 608) (T_VARIABLE 615) (simple_variable 630)) 398 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 628) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 399 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 626) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 400 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 627)) 401 (("]" 629)) 402 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 635) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 403 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 636)) 404 (("$" 608) (T_VARIABLE 615) (simple_variable 639)) 405 (("%" 280) (")" 641) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 406 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 644) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 407 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 645)) 408 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10725) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 409 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 10723) (encaps_var 5158)) 410 (("$" 653) (T_VARIABLE 700) (simple_variable 10719) ("{" 10720)) 411 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10717) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 412 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10716) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 413 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10715) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 414 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10714) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 415 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10713) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 416 (("(" 10710)) 417 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10709) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 418 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10708) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 419 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10707) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 420 (("\"" 1489) ("$" 8999) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 9000) ("[" 1508) (array_object_dereferenceable 9001) (callable_expr 9002) (callable_variable 9003) (class_constant 1512) (class_name 9004) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 9005) (name 9006) (simple_variable 9007) (static_member 9008) (variable 10706) (variable_class_name 9010)) 421 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10705) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 422 (("(" 10702)) 423 (("(" 10699)) 424 (("(" 10695) (exit_expr 10696)) 425 (("\"" 1489) ("$" 8999) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 9000) ("[" 1508) (array_object_dereferenceable 9001) (callable_expr 9002) (callable_variable 9003) (class_constant 1512) (class_name 9004) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 9005) (name 9006) (simple_variable 9007) (static_member 9008) (variable 10694) (variable_class_name 9010)) 426 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10693) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 427 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10692) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 428 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10691) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 429 (("(" 10687)) 430 (("(" 10682)) 431 (("(" 10676)) 432 (("$" 9372) ("(" 10015) (T_ATTRIBUTE 2853) (T_CLASS 10655) (T_NAME_FULLY_QUALIFIED 10017) (T_NAME_QUALIFIED 10018) (T_NAME_RELATIVE 10019) (T_STATIC 10020) (T_STRING 10021) (T_VARIABLE 9373) (anonymous_class 10656) (attribute 2862) (attributes 10657) (class_name 10024) (class_name_reference 10658) (name 10026) (new_variable 10027) (simple_variable 10028)) 433 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10654) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 434 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10653) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 435 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10652) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 436 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10651) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 437 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 10646) (T_END_HEREDOC 10647) (T_VARIABLE 2822) (encaps_list 10648) (encaps_var 2824)) 438 ((T_FN 35) (T_FUNCTION 126) (fn 714) #'715 (inline_function 10645)) 439 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10644) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 440 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10643) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 441 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10642) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 442 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10207) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 443 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10145) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 444 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 10141) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 445 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 10139) (encaps_list 2729) (encaps_var 2730)) 446 ((T_NULLSAFE_OBJECT_OPERATOR 10122) (T_OBJECT_OPERATOR 10123) ("[" 10124) ("{" 10125)) 447 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 10119) (attribute 2654) (fn 714) #'715 (inline_function 10120)) 448 (("(" 917) (argument_list 10118)) 449 ((T_PAAMAYIM_NEKUDOTAYIM 10113)) 450 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 10111) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 451 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 9062)) 452 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 9050)) 453 (("(" 917) (argument_list 9049)) 454 (("=" 8968) (T_AND_EQUAL 8969) (T_COALESCE_EQUAL 8970) (T_CONCAT_EQUAL 8971) (T_DEC 8972) (T_DIV_EQUAL 8973) (T_INC 8974) (T_MINUS_EQUAL 8975) (T_MOD_EQUAL 8976) (T_MUL_EQUAL 8977) (T_OR_EQUAL 8978) (T_PLUS_EQUAL 8979) (T_POW_EQUAL 8980) (T_SL_EQUAL 8981) (T_SR_EQUAL 8982) (T_XOR_EQUAL 8983)) 455 ((T_PAAMAYIM_NEKUDOTAYIM 830)) 456 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 728) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 457 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 458 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 829) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 459 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 828) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 460 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 827) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 461 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 826) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 462 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 825) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 463 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 824) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 464 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 823) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 465 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 822) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 466 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 817) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 818) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 467 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 816) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 468 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 815) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 469 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 814) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 470 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 813) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 471 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 812) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 472 (("$" 774) ("(" 775) (T_NAME_FULLY_QUALIFIED 776) (T_NAME_QUALIFIED 777) (T_NAME_RELATIVE 778) (T_STATIC 779) (T_STRING 780) (T_VARIABLE 781) (class_name 782) (class_name_reference 783) (name 784) (new_variable 785) (simple_variable 786)) 473 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 773) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 474 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 772) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 475 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 771) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 476 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 770) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 477 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 769) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 478 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 768) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 479 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 767) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 480 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 766) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 481 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 765) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 482 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 764) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 483 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 763) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 484 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 762) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 485 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 761) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 486 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 760) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 487 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 759) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 488 (("$" 774) (T_VARIABLE 781) (simple_variable 808) ("{" 809)) 489 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 806) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 490 ((T_PAAMAYIM_NEKUDOTAYIM 804)) 491 ((T_NULLSAFE_OBJECT_OPERATOR 787) (T_OBJECT_OPERATOR 788) (T_PAAMAYIM_NEKUDOTAYIM 789) ("[" 790) ("{" 791)) 492 (("$" 774) (T_STRING 797) (T_VARIABLE 781) (property_name 803) (simple_variable 799) ("{" 800)) 493 (("$" 774) (T_STRING 797) (T_VARIABLE 781) (property_name 798) (simple_variable 799) ("{" 800)) 494 (("$" 774) (T_VARIABLE 781) (simple_variable 796)) 495 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 794) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 496 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 792) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 497 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 793)) 498 (("]" 795)) 499 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 801) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 500 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 802)) 501 (("$" 774) (T_VARIABLE 781) (simple_variable 805)) 502 (("%" 280) (")" 807) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 503 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 810) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 504 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 811)) 505 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 821) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 506 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 819) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 507 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 820) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 508 (("$" 653) (T_ABSTRACT 831) (T_ARRAY 832) (T_AS 833) (T_BREAK 834) (T_CALLABLE 835) (T_CASE 836) (T_CATCH 837) (T_CLASS 838) (T_CLASS_C 839) (T_CLONE 840) (T_CONST 841) (T_CONTINUE 842) (T_DECLARE 843) (T_DEFAULT 844) (T_DIR 845) (T_DO 846) (T_ECHO 847) (T_ELSE 848) (T_ELSEIF 849) (T_EMPTY 850) (T_ENDDECLARE 851) (T_ENDFOR 852) (T_ENDFOREACH 853) (T_ENDIF 854) (T_ENDSWITCH 855) (T_ENDWHILE 856) (T_ENUM 857) (T_EVAL 858) (T_EXIT 859) (T_EXTENDS 860) (T_FILE 861) (T_FINAL 862) (T_FINALLY 863) (T_FN 864) (T_FOR 865) (T_FOREACH 866) (T_FUNCTION 867) (T_FUNC_C 868) (T_GLOBAL 869) (T_GOTO 870) (T_IF 871) (T_IMPLEMENTS 872) (T_INCLUDE 873) (T_INCLUDE_ONCE 874) (T_INSTANCEOF 875) (T_INSTEADOF 876) (T_INTERFACE 877) (T_ISSET 878) (T_LINE 879) (T_LIST 880) (T_LOGICAL_AND 881) (T_LOGICAL_OR 882) (T_LOGICAL_XOR 883) (T_MATCH 884) (T_METHOD_C 885) (T_NAMESPACE 886) (T_NEW 887) (T_NS_C 888) (T_PRINT 889) (T_PRIVATE 890) (T_PROTECTED 891) (T_PUBLIC 892) (T_READONLY 893) (T_REQUIRE 894) (T_REQUIRE_ONCE 895) (T_RETURN 896) (T_STATIC 897) (T_STRING 898) (T_SWITCH 899) (T_THROW 900) (T_TRAIT 901) (T_TRAIT_C 902) (T_TRY 903) (T_UNSET 904) (T_USE 905) (T_VAR 906) (T_VARIABLE 700) (T_WHILE 907) (T_YIELD 908) (identifier 909) (member_name 910) (reserved_non_modifiers 911) (semi_reserved 912) (simple_variable 913) ("{" 914)) 509 (("(" 917) (argument_list 918)) 510 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 915) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 511 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 916)) 512 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 923) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 950) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 1045) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 513 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8967) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 514 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 8965) (encaps_var 5158)) 515 (("$" 921) (T_VARIABLE 1020) (simple_variable 8961) ("{" 8962)) 516 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8959) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 517 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8958) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 518 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8957) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 519 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8956) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 520 (("(" 8715)) 521 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8955) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 522 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8954) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 523 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8714) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 524 (("\"" 1489) ("$" 2851) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2860) ("[" 1508) (array_object_dereferenceable 8841) (callable_expr 8842) (callable_variable 8843) (class_constant 1512) (class_name 8844) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 8845) (name 8846) (simple_variable 8847) (static_member 8848) (variable 8953) (variable_class_name 8850)) 525 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8952) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 526 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8951) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 527 (("(" 8711)) 528 (("(" 8708)) 529 (("(" 8704) (exit_expr 8705)) 530 (("\"" 1489) ("$" 2851) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2860) ("[" 1508) (array_object_dereferenceable 8841) (callable_expr 8842) (callable_variable 8843) (class_constant 1512) (class_name 8844) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 8845) (name 8846) (simple_variable 8847) (static_member 8848) (variable 8950) (variable_class_name 8850)) 531 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8703) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 532 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8702) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 533 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8949) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 534 (("(" 8698)) 535 (("(" 8693)) 536 (("(" 8687)) 537 (("$" 2851) ("(" 2852) (T_ATTRIBUTE 2853) (T_CLASS 8666) (T_NAME_FULLY_QUALIFIED 2855) (T_NAME_QUALIFIED 2856) (T_NAME_RELATIVE 2857) (T_STATIC 2858) (T_STRING 2859) (T_VARIABLE 2860) (anonymous_class 8667) (attribute 2862) (attributes 8668) (class_name 2864) (class_name_reference 8669) (name 2866) (new_variable 2867) (simple_variable 2868)) 538 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8948) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 539 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8665) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 540 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8664) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 541 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8663) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 542 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 8943) (T_END_HEREDOC 8944) (T_VARIABLE 2822) (encaps_list 8945) (encaps_var 2824)) 543 ((T_FN 35) (T_FUNCTION 126) (fn 1036) #'1037 (inline_function 8662)) 544 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8942) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 545 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8661) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 546 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8941) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 547 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1243) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 548 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8940) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 549 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 8936) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 550 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 8934) (encaps_list 2729) (encaps_var 2730)) 551 ((T_NULLSAFE_OBJECT_OPERATOR 8917) (T_OBJECT_OPERATOR 8918) ("[" 8919) ("{" 8920)) 552 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 8914) (attribute 2654) (fn 1036) #'1037 (inline_function 8915)) 553 (("(" 8802) (argument_list 8913)) 554 ((T_PAAMAYIM_NEKUDOTAYIM 8908)) 555 (("%" 1081) ("*" 1082) ("+" 1083) ("-" 1084) ("." 1085) ("/" 1086) ("<" 1087) (">" 1088) ("?" 1089) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1090) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1091) (T_BOOLEAN_AND 1092) (T_BOOLEAN_OR 1093) (T_COALESCE 1094) (T_INSTANCEOF 1095) (T_IS_EQUAL 1096) (T_IS_GREATER_OR_EQUAL 1097) (T_IS_IDENTICAL 1098) (T_IS_NOT_EQUAL 1099) (T_IS_NOT_IDENTICAL 1100) (T_IS_SMALLER_OR_EQUAL 1101) (T_LOGICAL_AND 1102) (T_LOGICAL_OR 1103) (T_LOGICAL_XOR 1104) (T_POW 1105) (T_SL 1106) (T_SPACESHIP 1107) (T_SR 1108) ("^" 1109) ("|" 1110)) 556 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 8897)) 557 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 8885)) 558 ((":" 8883)) 559 (("(" 8802) (argument_list 8882)) 560 (("," 1449) (possible_comma 8880)) 561 (("=" 8810) (T_AND_EQUAL 8811) (T_COALESCE_EQUAL 8812) (T_CONCAT_EQUAL 8813) (T_DEC 8814) (T_DIV_EQUAL 8815) (T_INC 8816) (T_MINUS_EQUAL 8817) (T_MOD_EQUAL 8818) (T_MUL_EQUAL 8819) (T_OR_EQUAL 8820) (T_PLUS_EQUAL 8821) (T_POW_EQUAL 8822) (T_SL_EQUAL 8823) (T_SR_EQUAL 8824) (T_XOR_EQUAL 8825)) 562 ((T_PAAMAYIM_NEKUDOTAYIM 8718)) 563 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1080) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 564 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1181) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 565 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1180) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 566 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1179) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 567 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1178) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 568 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1177) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 569 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1176) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 570 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1175) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 571 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1174) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 572 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 1169) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 1170) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 573 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1168) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 574 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1167) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 575 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1166) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 576 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1165) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 577 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1164) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 578 (("$" 1126) ("(" 1127) (T_NAME_FULLY_QUALIFIED 1128) (T_NAME_QUALIFIED 1129) (T_NAME_RELATIVE 1130) (T_STATIC 1131) (T_STRING 1132) (T_VARIABLE 1133) (class_name 1134) (class_name_reference 1135) (name 1136) (new_variable 1137) (simple_variable 1138)) 579 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1125) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 580 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1124) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 581 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1123) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 582 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1122) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 583 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1121) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 584 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1120) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 585 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1119) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 586 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1118) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 587 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1117) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 588 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1116) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 589 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1115) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 590 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 591 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1113) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 592 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1112) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 593 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1111) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 594 (("$" 1126) (T_VARIABLE 1133) (simple_variable 1160) ("{" 1161)) 595 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 1158) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 596 ((T_PAAMAYIM_NEKUDOTAYIM 1156)) 597 ((T_NULLSAFE_OBJECT_OPERATOR 1139) (T_OBJECT_OPERATOR 1140) (T_PAAMAYIM_NEKUDOTAYIM 1141) ("[" 1142) ("{" 1143)) 598 (("$" 1126) (T_STRING 1149) (T_VARIABLE 1133) (property_name 1155) (simple_variable 1151) ("{" 1152)) 599 (("$" 1126) (T_STRING 1149) (T_VARIABLE 1133) (property_name 1150) (simple_variable 1151) ("{" 1152)) 600 (("$" 1126) (T_VARIABLE 1133) (simple_variable 1148)) 601 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 1146) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 602 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1144) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 603 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1145)) 604 (("]" 1147)) 605 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1153) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 606 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1154)) 607 (("$" 1126) (T_VARIABLE 1133) (simple_variable 1157)) 608 (("%" 280) (")" 1159) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 609 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1162) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 610 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1163)) 611 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1173) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 612 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 1171) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 613 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1172) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 614 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8660) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 615 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 8658) (encaps_var 5158)) 616 (("$" 1184) (T_VARIABLE 1230) (simple_variable 8654) ("{" 8655)) 617 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8652) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 618 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8651) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 619 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8650) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 620 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8649) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 621 (("(" 8646)) 622 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8645) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 623 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8644) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 624 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8643) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 625 (("\"" 1489) ("$" 1490) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 1507) ("[" 1508) (array_object_dereferenceable 1509) (callable_expr 1510) (callable_variable 1511) (class_constant 1512) (class_name 1513) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 1516) (name 1517) (simple_variable 1518) (static_member 1519) (variable 8642) (variable_class_name 1521)) 626 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8641) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 627 (("(" 8638)) 628 (("(" 8635)) 629 (("(" 8631) (exit_expr 8632)) 630 (("\"" 1489) ("$" 1490) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 1507) ("[" 1508) (array_object_dereferenceable 1509) (callable_expr 1510) (callable_variable 1511) (class_constant 1512) (class_name 1513) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 1516) (name 1517) (simple_variable 1518) (static_member 1519) (variable 8630) (variable_class_name 1521)) 631 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8629) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 632 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8628) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 633 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8627) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 634 (("(" 8623)) 635 (("(" 8618)) 636 (("(" 8612)) 637 (("$" 1490) ("(" 6503) (T_ATTRIBUTE 2853) (T_CLASS 8591) (T_NAME_FULLY_QUALIFIED 6505) (T_NAME_QUALIFIED 6506) (T_NAME_RELATIVE 6507) (T_STATIC 6508) (T_STRING 6509) (T_VARIABLE 1507) (anonymous_class 8592) (attribute 2862) (attributes 8593) (class_name 6512) (class_name_reference 8594) (name 6514) (new_variable 6515) (simple_variable 6516)) 638 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8590) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 639 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8589) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 640 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8588) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 641 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8587) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 642 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 8582) (T_END_HEREDOC 8583) (T_VARIABLE 2822) (encaps_list 8584) (encaps_var 2824)) 643 ((T_FN 35) (T_FUNCTION 126) (fn 1244) #'1245 (inline_function 8581)) 644 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8580) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 645 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8579) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 646 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8578) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 647 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8575) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 648 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8574) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 649 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 8570) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 650 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 8568) (encaps_list 2729) (encaps_var 2730)) 651 ((T_NULLSAFE_OBJECT_OPERATOR 8551) (T_OBJECT_OPERATOR 8552) ("[" 8553) ("{" 8554)) 652 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 8548) (attribute 2654) (fn 1244) #'1245 (inline_function 8549)) 653 (("(" 1444) (argument_list 8547)) 654 ((T_PAAMAYIM_NEKUDOTAYIM 8542)) 655 (("%" 1259) ("*" 1260) ("+" 1261) ("-" 1262) ("." 1263) ("/" 1264) ("<" 1265) (">" 1266) ("?" 1267) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1268) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1269) (T_BOOLEAN_AND 1270) (T_BOOLEAN_OR 1271) (T_COALESCE 1272) (T_DOUBLE_ARROW 8540) (T_INSTANCEOF 1273) (T_IS_EQUAL 1274) (T_IS_GREATER_OR_EQUAL 1275) (T_IS_IDENTICAL 1276) (T_IS_NOT_EQUAL 1277) (T_IS_NOT_IDENTICAL 1278) (T_IS_SMALLER_OR_EQUAL 1279) (T_LOGICAL_AND 1280) (T_LOGICAL_OR 1281) (T_LOGICAL_XOR 1282) (T_POW 1283) (T_SL 1284) (T_SPACESHIP 1285) (T_SR 1286) ("^" 1287) ("|" 1288)) 656 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 8529)) 657 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 8517)) 658 (("(" 1444) (argument_list 8516)) 659 (("=" 1456) (T_AND_EQUAL 1457) (T_COALESCE_EQUAL 1458) (T_CONCAT_EQUAL 1459) (T_DEC 1460) (T_DIV_EQUAL 1461) (T_INC 1462) (T_MINUS_EQUAL 1463) (T_MOD_EQUAL 1464) (T_MUL_EQUAL 1465) (T_OR_EQUAL 1466) (T_PLUS_EQUAL 1467) (T_POW_EQUAL 1468) (T_SL_EQUAL 1469) (T_SR_EQUAL 1470) (T_XOR_EQUAL 1471)) 660 ((T_PAAMAYIM_NEKUDOTAYIM 1360)) 661 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1258) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 662 (("%" 1259) ("*" 1260) ("+" 1261) ("-" 1262) ("." 1263) ("/" 1264) ("<" 1265) (">" 1266) ("?" 1267) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1268) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1269) (T_BOOLEAN_AND 1270) (T_BOOLEAN_OR 1271) (T_COALESCE 1272) (T_INSTANCEOF 1273) (T_IS_EQUAL 1274) (T_IS_GREATER_OR_EQUAL 1275) (T_IS_IDENTICAL 1276) (T_IS_NOT_EQUAL 1277) (T_IS_NOT_IDENTICAL 1278) (T_IS_SMALLER_OR_EQUAL 1279) (T_LOGICAL_AND 1280) (T_LOGICAL_OR 1281) (T_LOGICAL_XOR 1282) (T_POW 1283) (T_SL 1284) (T_SPACESHIP 1285) (T_SR 1286) ("^" 1287) ("|" 1288)) 663 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1359) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 664 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1358) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 665 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1357) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 666 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1356) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 667 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1355) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 668 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1354) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 669 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1353) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 670 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1352) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 671 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 1347) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 1348) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 672 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1346) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 673 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1345) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 674 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1344) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 675 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1343) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 676 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1342) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 677 (("$" 1304) ("(" 1305) (T_NAME_FULLY_QUALIFIED 1306) (T_NAME_QUALIFIED 1307) (T_NAME_RELATIVE 1308) (T_STATIC 1309) (T_STRING 1310) (T_VARIABLE 1311) (class_name 1312) (class_name_reference 1313) (name 1314) (new_variable 1315) (simple_variable 1316)) 678 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1303) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 679 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1302) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 680 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1301) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 681 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1300) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 682 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1299) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 683 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1298) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 684 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1297) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 685 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1296) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 686 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1295) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 687 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1294) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 688 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1293) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 689 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1292) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 690 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1291) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 691 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1290) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 692 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1289) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 693 (("$" 1304) (T_VARIABLE 1311) (simple_variable 1338) ("{" 1339)) 694 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 1336) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 695 ((T_PAAMAYIM_NEKUDOTAYIM 1334)) 696 ((T_NULLSAFE_OBJECT_OPERATOR 1317) (T_OBJECT_OPERATOR 1318) (T_PAAMAYIM_NEKUDOTAYIM 1319) ("[" 1320) ("{" 1321)) 697 (("$" 1304) (T_STRING 1327) (T_VARIABLE 1311) (property_name 1333) (simple_variable 1329) ("{" 1330)) 698 (("$" 1304) (T_STRING 1327) (T_VARIABLE 1311) (property_name 1328) (simple_variable 1329) ("{" 1330)) 699 (("$" 1304) (T_VARIABLE 1311) (simple_variable 1326)) 700 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 1324) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 701 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1322) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 702 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1323)) 703 (("]" 1325)) 704 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1331) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 705 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1332)) 706 (("$" 1304) (T_VARIABLE 1311) (simple_variable 1335)) 707 (("%" 280) (")" 1337) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 708 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1340) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 709 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1341)) 710 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1351) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 711 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 1349) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 712 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1350) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 713 (("$" 1184) (T_ABSTRACT 1361) (T_ARRAY 1362) (T_AS 1363) (T_BREAK 1364) (T_CALLABLE 1365) (T_CASE 1366) (T_CATCH 1367) (T_CLASS 1368) (T_CLASS_C 1369) (T_CLONE 1370) (T_CONST 1371) (T_CONTINUE 1372) (T_DECLARE 1373) (T_DEFAULT 1374) (T_DIR 1375) (T_DO 1376) (T_ECHO 1377) (T_ELSE 1378) (T_ELSEIF 1379) (T_EMPTY 1380) (T_ENDDECLARE 1381) (T_ENDFOR 1382) (T_ENDFOREACH 1383) (T_ENDIF 1384) (T_ENDSWITCH 1385) (T_ENDWHILE 1386) (T_ENUM 1387) (T_EVAL 1388) (T_EXIT 1389) (T_EXTENDS 1390) (T_FILE 1391) (T_FINAL 1392) (T_FINALLY 1393) (T_FN 1394) (T_FOR 1395) (T_FOREACH 1396) (T_FUNCTION 1397) (T_FUNC_C 1398) (T_GLOBAL 1399) (T_GOTO 1400) (T_IF 1401) (T_IMPLEMENTS 1402) (T_INCLUDE 1403) (T_INCLUDE_ONCE 1404) (T_INSTANCEOF 1405) (T_INSTEADOF 1406) (T_INTERFACE 1407) (T_ISSET 1408) (T_LINE 1409) (T_LIST 1410) (T_LOGICAL_AND 1411) (T_LOGICAL_OR 1412) (T_LOGICAL_XOR 1413) (T_MATCH 1414) (T_METHOD_C 1415) (T_NAMESPACE 1416) (T_NEW 1417) (T_NS_C 1418) (T_PRINT 1419) (T_PRIVATE 1420) (T_PROTECTED 1421) (T_PUBLIC 1422) (T_READONLY 1423) (T_REQUIRE 1424) (T_REQUIRE_ONCE 1425) (T_RETURN 1426) (T_STATIC 1427) (T_STRING 1428) (T_SWITCH 1429) (T_THROW 1430) (T_TRAIT 1431) (T_TRAIT_C 1432) (T_TRY 1433) (T_UNSET 1434) (T_USE 1435) (T_VAR 1436) (T_VARIABLE 1230) (T_WHILE 1437) (T_YIELD 1438) (identifier 1439) (member_name 1440) (reserved_non_modifiers 1441) (semi_reserved 1442) (simple_variable 1443) ("{" 914)) 714 (("(" 1444) (argument_list 1445)) 715 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 1446) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 1447) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 1448) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 716 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 1455) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 717 (("," 1449) (possible_comma 1450)) 718 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 1452) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1453) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 719 ((")" 1451)) 720 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 721 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (ampersand 1487) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1488) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 722 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1484) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 723 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1483) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 724 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1482) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 725 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1481) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 726 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1480) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 727 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1479) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 728 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1478) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 729 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1477) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 730 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1476) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 731 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1475) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 732 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1474) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 733 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1473) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 734 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 1472) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 735 (("\"" 1489) ("$" 1490) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 1507) ("[" 1508) (array_object_dereferenceable 1509) (callable_expr 1510) (callable_variable 1511) (class_constant 1512) (class_name 1513) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 1516) (name 1517) (simple_variable 1518) (static_member 1519) (variable 1520) (variable_class_name 1521)) 736 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 8514) (encaps_var 5158)) 737 (("$" 1490) (T_VARIABLE 1507) (simple_variable 8510) ("{" 8511)) 738 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8508) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 739 (("(" 8505)) 740 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 1695) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 741 ((T_NULLSAFE_OBJECT_OPERATOR 1621) (T_OBJECT_OPERATOR 1622) ("[" 1623) ("{" 1624)) 742 (("(" 1606) (argument_list 1620)) 743 ((T_PAAMAYIM_NEKUDOTAYIM 1615)) 744 (("(" 1606) (argument_list 1614)) 745 ((T_PAAMAYIM_NEKUDOTAYIM 1522)) 746 (("$" 1490) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 1507) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 1602) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 1605) ("{" 914)) 747 (("(" 1606) (argument_list 1607)) 748 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 1608) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 1609) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 1610) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 749 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 1613) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 750 (("," 1449) (possible_comma 1611)) 751 ((")" 1612)) 752 (("$" 1490) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 1507) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 1617) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 1618) ("{" 914)) 753 (("(" 1606) (argument_list 1619)) 754 (("$" 1490) (T_STRING 1629) (T_VARIABLE 1507) (property_name 1636) (simple_variable 1631) ("{" 1632)) 755 (("$" 1490) (T_STRING 1629) (T_VARIABLE 1507) (property_name 1630) (simple_variable 1631) ("{" 1632)) 756 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 1627) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 757 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1625) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 758 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1626)) 759 (("]" 1628)) 760 (("(" 1606) (argument_list 1635)) 761 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1633) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 762 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1634)) 763 (("(" 1606) (argument_list 1637)) 764 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8504) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 765 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 8502) (encaps_var 5158)) 766 (("$" 1640) (T_VARIABLE 1687) (simple_variable 8498) ("{" 8499)) 767 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8496) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 768 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8495) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 769 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8494) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 770 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8493) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 771 (("(" 8490)) 772 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8489) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 773 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8488) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 774 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8487) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 775 (("\"" 1489) ("$" 6721) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 6722) ("[" 1508) (array_object_dereferenceable 6723) (callable_expr 6724) (callable_variable 6725) (class_constant 1512) (class_name 6726) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 6727) (name 6728) (simple_variable 6729) (static_member 6730) (variable 8486) (variable_class_name 6732)) 776 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8485) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 777 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8484) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 778 (("(" 8481)) 779 (("(" 8478)) 780 (("(" 8474) (exit_expr 8475)) 781 (("\"" 1489) ("$" 6721) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 6722) ("[" 1508) (array_object_dereferenceable 6723) (callable_expr 6724) (callable_variable 6725) (class_constant 1512) (class_name 6726) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 6727) (name 6728) (simple_variable 6729) (static_member 6730) (variable 8473) (variable_class_name 6732)) 782 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8472) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 783 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8471) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 784 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8470) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 785 (("(" 8466)) 786 (("(" 8463)) 787 (("(" 8457)) 788 (("$" 7096) ("(" 7220) (T_ATTRIBUTE 2853) (T_CLASS 8436) (T_NAME_FULLY_QUALIFIED 7222) (T_NAME_QUALIFIED 7223) (T_NAME_RELATIVE 7224) (T_STATIC 7225) (T_STRING 7226) (T_VARIABLE 7097) (anonymous_class 8437) (attribute 2862) (attributes 8438) (class_name 7229) (class_name_reference 8439) (name 7231) (new_variable 7232) (simple_variable 7233)) 789 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8435) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 790 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8434) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 791 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8433) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 792 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8432) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 793 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 8427) (T_END_HEREDOC 8428) (T_VARIABLE 2822) (encaps_list 8429) (encaps_var 2824)) 794 ((T_FN 35) (T_FUNCTION 126) (fn 1704) #'1705 (inline_function 8426)) 795 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8425) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 796 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8424) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 797 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8423) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 798 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8420) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 799 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8419) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 800 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 8415) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 801 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 8413) (encaps_list 2729) (encaps_var 2730)) 802 (("\"" 1489) ("$" 8244) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 8245) ("[" 1508) (array_object_dereferenceable 8246) (callable_expr 8247) (callable_variable 8248) (class_constant 1512) (class_name 8249) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 8250) (name 8251) (simple_variable 8252) (static_member 8253) (variable 8412) (variable_class_name 8255)) 803 ((T_NULLSAFE_OBJECT_OPERATOR 8395) (T_OBJECT_OPERATOR 8396) ("[" 8397) ("{" 8398)) 804 (("]" 8394)) 805 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 8391) (attribute 2654) (fn 1704) #'1705 (inline_function 8392)) 806 (("(" 6682) (argument_list 8390)) 807 ((T_PAAMAYIM_NEKUDOTAYIM 8385)) 808 (("%" 1722) ("*" 1723) ("+" 1724) ("-" 1725) ("." 1726) ("/" 1727) ("<" 1728) (">" 1729) ("?" 1730) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1731) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1732) (T_BOOLEAN_AND 1733) (T_BOOLEAN_OR 1734) (T_COALESCE 1735) (T_DOUBLE_ARROW 7316) (T_INSTANCEOF 1736) (T_IS_EQUAL 1737) (T_IS_GREATER_OR_EQUAL 1738) (T_IS_IDENTICAL 1739) (T_IS_NOT_EQUAL 1740) (T_IS_NOT_IDENTICAL 1741) (T_IS_SMALLER_OR_EQUAL 1742) (T_LOGICAL_AND 1743) (T_LOGICAL_OR 1744) (T_LOGICAL_XOR 1745) (T_POW 1746) (T_SL 1747) (T_SPACESHIP 1748) (T_SR 1749) ("^" 1750) ("|" 1751)) 809 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 6786)) 810 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 6774)) 811 (("(" 6682) (argument_list 6773)) 812 (("," 6771)) 813 (("=" 6690) (T_AND_EQUAL 6691) (T_COALESCE_EQUAL 6692) (T_CONCAT_EQUAL 6693) (T_DEC 6694) (T_DIV_EQUAL 6695) (T_INC 6696) (T_MINUS_EQUAL 6697) (T_MOD_EQUAL 6698) (T_MUL_EQUAL 6699) (T_OR_EQUAL 6700) (T_PLUS_EQUAL 6701) (T_POW_EQUAL 6702) (T_SL_EQUAL 6703) (T_SR_EQUAL 6704) (T_XOR_EQUAL 6705)) 814 ((T_PAAMAYIM_NEKUDOTAYIM 6598)) 815 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1721) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 816 (("(" 1823)) 817 (("%" 1722) ("*" 1723) ("+" 1724) ("-" 1725) ("." 1726) ("/" 1727) ("<" 1728) (">" 1729) ("?" 1730) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1731) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1732) (T_BOOLEAN_AND 1733) (T_BOOLEAN_OR 1734) (T_COALESCE 1735) (T_INSTANCEOF 1736) (T_IS_EQUAL 1737) (T_IS_GREATER_OR_EQUAL 1738) (T_IS_IDENTICAL 1739) (T_IS_NOT_EQUAL 1740) (T_IS_NOT_IDENTICAL 1741) (T_IS_SMALLER_OR_EQUAL 1742) (T_LOGICAL_AND 1743) (T_LOGICAL_OR 1744) (T_LOGICAL_XOR 1745) (T_POW 1746) (T_SL 1747) (T_SPACESHIP 1748) (T_SR 1749) ("^" 1750) ("|" 1751)) 818 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1822) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 819 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1821) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 820 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1820) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 821 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1819) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 822 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1818) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 823 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1817) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 824 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1816) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 825 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1815) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 826 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 1810) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 1811) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 827 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1809) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 828 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1808) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 829 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1807) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 830 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1806) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 831 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1805) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 832 (("$" 1767) ("(" 1768) (T_NAME_FULLY_QUALIFIED 1769) (T_NAME_QUALIFIED 1770) (T_NAME_RELATIVE 1771) (T_STATIC 1772) (T_STRING 1773) (T_VARIABLE 1774) (class_name 1775) (class_name_reference 1776) (name 1777) (new_variable 1778) (simple_variable 1779)) 833 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1766) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 834 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1765) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 835 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1764) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 836 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1763) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 837 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1762) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 838 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1761) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 839 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1760) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 840 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1759) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 841 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1758) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 842 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1757) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 843 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1756) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 844 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1755) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 845 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1754) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 846 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1753) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 847 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1752) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 848 (("$" 1767) (T_VARIABLE 1774) (simple_variable 1801) ("{" 1802)) 849 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 1799) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 850 ((T_PAAMAYIM_NEKUDOTAYIM 1797)) 851 ((T_NULLSAFE_OBJECT_OPERATOR 1780) (T_OBJECT_OPERATOR 1781) (T_PAAMAYIM_NEKUDOTAYIM 1782) ("[" 1783) ("{" 1784)) 852 (("$" 1767) (T_STRING 1790) (T_VARIABLE 1774) (property_name 1796) (simple_variable 1792) ("{" 1793)) 853 (("$" 1767) (T_STRING 1790) (T_VARIABLE 1774) (property_name 1791) (simple_variable 1792) ("{" 1793)) 854 (("$" 1767) (T_VARIABLE 1774) (simple_variable 1789)) 855 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 1787) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 856 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1785) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 857 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1786)) 858 (("]" 1788)) 859 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1794) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 860 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1795)) 861 (("$" 1767) (T_VARIABLE 1774) (simple_variable 1798)) 862 (("%" 280) (")" 1800) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 863 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1803) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 864 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1804)) 865 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1814) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 866 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 1812) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 867 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1813) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 868 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 1881) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 869 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6597) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 870 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 6595) (encaps_var 5158)) 871 (("$" 1826) (T_VARIABLE 1873) (simple_variable 6591) ("{" 6592)) 872 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 6589) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 873 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6588) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 874 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6587) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 875 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6586) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 876 (("(" 6583)) 877 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6582) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 878 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6581) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 879 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6580) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 880 (("\"" 1489) ("$" 2137) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2138) ("[" 1508) (array_object_dereferenceable 2139) (callable_expr 2140) (callable_variable 2141) (class_constant 1512) (class_name 2142) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 2143) (name 2144) (simple_variable 2145) (static_member 2146) (variable 6579) (variable_class_name 2148)) 881 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6578) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 882 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 6577) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 883 (("(" 6574)) 884 (("(" 6571)) 885 (("(" 6567) (exit_expr 6568)) 886 (("\"" 1489) ("$" 2137) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2138) ("[" 1508) (array_object_dereferenceable 2139) (callable_expr 2140) (callable_variable 2141) (class_constant 1512) (class_name 2142) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 2143) (name 2144) (simple_variable 2145) (static_member 2146) (variable 6566) (variable_class_name 2148)) 887 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6565) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 888 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6564) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 889 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6563) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 890 (("(" 6559)) 891 (("(" 6556)) 892 (("(" 6550)) 893 (("$" 1490) ("(" 6503) (T_ATTRIBUTE 2853) (T_CLASS 6504) (T_NAME_FULLY_QUALIFIED 6505) (T_NAME_QUALIFIED 6506) (T_NAME_RELATIVE 6507) (T_STATIC 6508) (T_STRING 6509) (T_VARIABLE 1507) (anonymous_class 6510) (attribute 2862) (attributes 6511) (class_name 6512) (class_name_reference 6513) (name 6514) (new_variable 6515) (simple_variable 6516)) 894 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6502) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 895 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6501) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 896 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6500) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 897 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6499) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 898 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 6494) (T_END_HEREDOC 6495) (T_VARIABLE 2822) (encaps_list 6496) (encaps_var 2824)) 899 ((T_FN 35) (T_FUNCTION 126) (fn 1890) #'1891 (inline_function 6493)) 900 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6492) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 901 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6491) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 902 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6490) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 903 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6487) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 904 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6486) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 905 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 6482) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 906 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 6480) (encaps_list 2729) (encaps_var 2730)) 907 (("\"" 1489) ("$" 2674) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2675) ("[" 1508) (array_object_dereferenceable 2676) (callable_expr 2677) (callable_variable 2678) (class_constant 1512) (class_name 2679) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 2680) (name 2681) (simple_variable 2682) (static_member 2683) (variable 6479) (variable_class_name 2685)) 908 ((T_NULLSAFE_OBJECT_OPERATOR 6462) (T_OBJECT_OPERATOR 6463) ("[" 6464) ("{" 6465)) 909 ((")" 6459)) 910 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 6456) (attribute 2654) (fn 1890) #'1891 (inline_function 6457)) 911 (("(" 2098) (argument_list 6455)) 912 ((T_PAAMAYIM_NEKUDOTAYIM 6450)) 913 (("%" 1908) ("*" 1909) ("+" 1910) ("-" 1911) ("." 1912) ("/" 1913) ("<" 1914) (">" 1915) ("?" 1916) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1917) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1918) (T_BOOLEAN_AND 1919) (T_BOOLEAN_OR 1920) (T_COALESCE 1921) (T_DOUBLE_ARROW 2264) (T_INSTANCEOF 1922) (T_IS_EQUAL 1923) (T_IS_GREATER_OR_EQUAL 1924) (T_IS_IDENTICAL 1925) (T_IS_NOT_EQUAL 1926) (T_IS_NOT_IDENTICAL 1927) (T_IS_SMALLER_OR_EQUAL 1928) (T_LOGICAL_AND 1929) (T_LOGICAL_OR 1930) (T_LOGICAL_XOR 1931) (T_POW 1932) (T_SL 1933) (T_SPACESHIP 1934) (T_SR 1935) ("^" 1936) ("|" 1937)) 914 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 2202)) 915 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 2190)) 916 (("(" 2098) (argument_list 2189)) 917 (("," 2187)) 918 (("=" 2106) (T_AND_EQUAL 2107) (T_COALESCE_EQUAL 2108) (T_CONCAT_EQUAL 2109) (T_DEC 2110) (T_DIV_EQUAL 2111) (T_INC 2112) (T_MINUS_EQUAL 2113) (T_MOD_EQUAL 2114) (T_MUL_EQUAL 2115) (T_OR_EQUAL 2116) (T_PLUS_EQUAL 2117) (T_POW_EQUAL 2118) (T_SL_EQUAL 2119) (T_SR_EQUAL 2120) (T_XOR_EQUAL 2121)) 919 ((T_PAAMAYIM_NEKUDOTAYIM 2014)) 920 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1907) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 921 (("(" 2009)) 922 (("%" 1908) ("*" 1909) ("+" 1910) ("-" 1911) ("." 1912) ("/" 1913) ("<" 1914) (">" 1915) ("?" 1916) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1917) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1918) (T_BOOLEAN_AND 1919) (T_BOOLEAN_OR 1920) (T_COALESCE 1921) (T_INSTANCEOF 1922) (T_IS_EQUAL 1923) (T_IS_GREATER_OR_EQUAL 1924) (T_IS_IDENTICAL 1925) (T_IS_NOT_EQUAL 1926) (T_IS_NOT_IDENTICAL 1927) (T_IS_SMALLER_OR_EQUAL 1928) (T_LOGICAL_AND 1929) (T_LOGICAL_OR 1930) (T_LOGICAL_XOR 1931) (T_POW 1932) (T_SL 1933) (T_SPACESHIP 1934) (T_SR 1935) ("^" 1936) ("|" 1937)) 923 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2008) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 924 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2007) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 925 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2006) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 926 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2005) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 927 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2004) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 928 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2003) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 929 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2002) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 930 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2001) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 931 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 1996) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 1997) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 932 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1995) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 933 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1994) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 934 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1993) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 935 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1992) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 936 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1991) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 937 (("$" 1953) ("(" 1954) (T_NAME_FULLY_QUALIFIED 1955) (T_NAME_QUALIFIED 1956) (T_NAME_RELATIVE 1957) (T_STATIC 1958) (T_STRING 1959) (T_VARIABLE 1960) (class_name 1961) (class_name_reference 1962) (name 1963) (new_variable 1964) (simple_variable 1965)) 938 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1952) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 939 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1951) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 940 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1950) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 941 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1949) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 942 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1948) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 943 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1947) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 944 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1946) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 945 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1945) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 946 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1944) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 947 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1943) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 948 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1942) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 949 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1941) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 950 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1940) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 951 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1939) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 952 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1938) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 953 (("$" 1953) (T_VARIABLE 1960) (simple_variable 1987) ("{" 1988)) 954 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 1985) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 955 ((T_PAAMAYIM_NEKUDOTAYIM 1983)) 956 ((T_NULLSAFE_OBJECT_OPERATOR 1966) (T_OBJECT_OPERATOR 1967) (T_PAAMAYIM_NEKUDOTAYIM 1968) ("[" 1969) ("{" 1970)) 957 (("$" 1953) (T_STRING 1976) (T_VARIABLE 1960) (property_name 1982) (simple_variable 1978) ("{" 1979)) 958 (("$" 1953) (T_STRING 1976) (T_VARIABLE 1960) (property_name 1977) (simple_variable 1978) ("{" 1979)) 959 (("$" 1953) (T_VARIABLE 1960) (simple_variable 1975)) 960 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 1973) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 961 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1971) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 962 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1972)) 963 (("]" 1974)) 964 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1980) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 965 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1981)) 966 (("$" 1953) (T_VARIABLE 1960) (simple_variable 1984)) 967 (("%" 280) (")" 1986) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 968 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 1989) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 969 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 1990)) 970 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2000) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 971 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 1998) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 972 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1999) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 973 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 2010) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 974 ((")" 2011)) 975 (("=" 2012)) 976 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2013) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 977 (("$" 1826) (T_ABSTRACT 2015) (T_ARRAY 2016) (T_AS 2017) (T_BREAK 2018) (T_CALLABLE 2019) (T_CASE 2020) (T_CATCH 2021) (T_CLASS 2022) (T_CLASS_C 2023) (T_CLONE 2024) (T_CONST 2025) (T_CONTINUE 2026) (T_DECLARE 2027) (T_DEFAULT 2028) (T_DIR 2029) (T_DO 2030) (T_ECHO 2031) (T_ELSE 2032) (T_ELSEIF 2033) (T_EMPTY 2034) (T_ENDDECLARE 2035) (T_ENDFOR 2036) (T_ENDFOREACH 2037) (T_ENDIF 2038) (T_ENDSWITCH 2039) (T_ENDWHILE 2040) (T_ENUM 2041) (T_EVAL 2042) (T_EXIT 2043) (T_EXTENDS 2044) (T_FILE 2045) (T_FINAL 2046) (T_FINALLY 2047) (T_FN 2048) (T_FOR 2049) (T_FOREACH 2050) (T_FUNCTION 2051) (T_FUNC_C 2052) (T_GLOBAL 2053) (T_GOTO 2054) (T_IF 2055) (T_IMPLEMENTS 2056) (T_INCLUDE 2057) (T_INCLUDE_ONCE 2058) (T_INSTANCEOF 2059) (T_INSTEADOF 2060) (T_INTERFACE 2061) (T_ISSET 2062) (T_LINE 2063) (T_LIST 2064) (T_LOGICAL_AND 2065) (T_LOGICAL_OR 2066) (T_LOGICAL_XOR 2067) (T_MATCH 2068) (T_METHOD_C 2069) (T_NAMESPACE 2070) (T_NEW 2071) (T_NS_C 2072) (T_PRINT 2073) (T_PRIVATE 2074) (T_PROTECTED 2075) (T_PUBLIC 2076) (T_READONLY 2077) (T_REQUIRE 2078) (T_REQUIRE_ONCE 2079) (T_RETURN 2080) (T_STATIC 2081) (T_STRING 2082) (T_SWITCH 2083) (T_THROW 2084) (T_TRAIT 2085) (T_TRAIT_C 2086) (T_TRY 2087) (T_UNSET 2088) (T_USE 2089) (T_VAR 2090) (T_VARIABLE 1873) (T_WHILE 2091) (T_YIELD 2092) (identifier 2093) (member_name 2094) (reserved_non_modifiers 2095) (semi_reserved 2096) (simple_variable 2097) ("{" 914)) 978 (("(" 2098) (argument_list 2099)) 979 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2100) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 2101) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 2102) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 980 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2105) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 981 (("," 1449) (possible_comma 2103)) 982 ((")" 2104)) 983 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 2135) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2136) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 984 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2134) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 985 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2133) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 986 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2132) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 987 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2131) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 988 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2130) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 989 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2129) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 990 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2128) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 991 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2127) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 992 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2126) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 993 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2125) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 994 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2124) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 995 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2123) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 996 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2122) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 997 (("\"" 1489) ("$" 2137) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2138) ("[" 1508) (array_object_dereferenceable 2139) (callable_expr 2140) (callable_variable 2141) (class_constant 1512) (class_name 2142) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 2143) (name 2144) (simple_variable 2145) (static_member 2146) (variable 2147) (variable_class_name 2148)) 998 (("$" 2137) (T_VARIABLE 2138) (simple_variable 2183) ("{" 2184)) 999 ((T_NULLSAFE_OBJECT_OPERATOR 2166) (T_OBJECT_OPERATOR 2167) ("[" 2168) ("{" 2169)) 1000 (("(" 2152) (argument_list 2165)) 1001 ((T_PAAMAYIM_NEKUDOTAYIM 2161)) 1002 (("(" 2152) (argument_list 2160)) 1003 ((T_PAAMAYIM_NEKUDOTAYIM 2149)) 1004 (("$" 2137) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 2138) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 2150) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 2151) ("{" 914)) 1005 (("(" 2152) (argument_list 2153)) 1006 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2154) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 2155) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 2156) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1007 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2159) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1008 (("," 1449) (possible_comma 2157)) 1009 ((")" 2158)) 1010 (("$" 2137) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 2138) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 2162) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 2163) ("{" 914)) 1011 (("(" 2152) (argument_list 2164)) 1012 (("$" 2137) (T_STRING 2174) (T_VARIABLE 2138) (property_name 2181) (simple_variable 2176) ("{" 2177)) 1013 (("$" 2137) (T_STRING 2174) (T_VARIABLE 2138) (property_name 2175) (simple_variable 2176) ("{" 2177)) 1014 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 2172) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1015 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2170) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1016 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2171)) 1017 (("]" 2173)) 1018 (("(" 2152) (argument_list 2180)) 1019 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2178) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1020 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2179)) 1021 (("(" 2152) (argument_list 2182)) 1022 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2185) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1023 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2186)) 1024 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (possible_array_pair 2188) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 1025 ((backup_doc_comment 2191)) 1026 (("(" 2192)) 1027 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 2193)) 1028 ((")" 2194)) 1029 ((T_USE 148) (lexical_vars 2195)) 1030 ((":" 150) (return_type 2196)) 1031 ((backup_fn_flags 2197)) 1032 (("{" 2198)) 1033 ((inner_statement_list 2199)) 1034 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 2200) ("~" 124)) 1035 ((backup_fn_flags 2201)) 1036 ((backup_doc_comment 2203)) 1037 (("(" 2204)) 1038 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 2205)) 1039 ((")" 2206)) 1040 ((":" 2207) (return_type 2208)) 1041 (("?" 2214) (T_ARRAY 2215) (T_CALLABLE 2216) (T_NAME_FULLY_QUALIFIED 2217) (T_NAME_QUALIFIED 2218) (T_NAME_RELATIVE 2219) (T_STATIC 2220) (T_STRING 2221) (intersection_type 2222) (name 2223) (type 2224) (type_expr 2225) (type_without_static 2226) (union_type 2227)) 1042 ((T_DOUBLE_ARROW 2209)) 1043 ((backup_fn_flags 2210)) 1044 ((backup_lex_pos 2211)) 1045 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 2212) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 1046 (("%" 1259) ("*" 1260) ("+" 1261) ("-" 1262) ("." 1263) ("/" 1264) ("<" 1265) (">" 1266) ("?" 1267) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1268) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1269) (T_BOOLEAN_AND 1270) (T_BOOLEAN_OR 1271) (T_COALESCE 1272) (T_INSTANCEOF 1273) (T_IS_EQUAL 1274) (T_IS_GREATER_OR_EQUAL 1275) (T_IS_IDENTICAL 1276) (T_IS_NOT_EQUAL 1277) (T_IS_NOT_IDENTICAL 1278) (T_IS_SMALLER_OR_EQUAL 1279) (T_LOGICAL_AND 1280) (T_LOGICAL_OR 1281) (T_LOGICAL_XOR 1282) (T_POW 1283) (T_SL 1284) (T_SPACESHIP 1285) (T_SR 1286) ("^" 1287) (backup_fn_flags 2213) ("|" 1288)) 1047 ((T_ARRAY 2254) (T_CALLABLE 2255) (T_NAME_FULLY_QUALIFIED 2256) (T_NAME_QUALIFIED 2257) (T_NAME_RELATIVE 2258) (T_STATIC 2259) (T_STRING 2260) (name 2261) (type 2262) (type_without_static 2263)) 1048 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2252)) 1049 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2239) ("|" 2240)) 1050 (("|" 2228)) 1051 ((T_ARRAY 2229) (T_CALLABLE 2230) (T_NAME_FULLY_QUALIFIED 2231) (T_NAME_QUALIFIED 2232) (T_NAME_RELATIVE 2233) (T_STATIC 2234) (T_STRING 2235) (name 2236) (type 2237) (type_without_static 2238)) 1052 ((T_ARRAY 2242) (T_CALLABLE 2243) (T_NAME_FULLY_QUALIFIED 2244) (T_NAME_QUALIFIED 2245) (T_NAME_RELATIVE 2246) (T_STATIC 2247) (T_STRING 2248) (name 2249) (type 2250) (type_without_static 2251)) 1053 ((T_ARRAY 2229) (T_CALLABLE 2230) (T_NAME_FULLY_QUALIFIED 2231) (T_NAME_QUALIFIED 2232) (T_NAME_RELATIVE 2233) (T_STATIC 2234) (T_STRING 2235) (name 2236) (type 2241) (type_without_static 2238)) 1054 ((T_ARRAY 2242) (T_CALLABLE 2243) (T_NAME_FULLY_QUALIFIED 2244) (T_NAME_QUALIFIED 2245) (T_NAME_RELATIVE 2246) (T_STATIC 2247) (T_STRING 2248) (name 2249) (type 2253) (type_without_static 2251)) 1055 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2293) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (ampersand 2318) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2327) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1056 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 6449) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1057 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 6447) (encaps_var 5158)) 1058 (("$" 2267) (T_VARIABLE 2313) (simple_variable 6443) ("{" 6444)) 1059 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 6441) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 1060 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 6440) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1061 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 6439) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1062 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 6438) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1063 (("(" 6435)) 1064 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 6434) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1065 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 6433) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1066 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 6432) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1067 (("\"" 1489) ("$" 2573) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2574) ("[" 1508) (array_object_dereferenceable 2575) (callable_expr 2576) (callable_variable 2577) (class_constant 1512) (class_name 2578) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 2579) (name 2580) (simple_variable 2581) (static_member 2582) (variable 6431) (variable_class_name 2584)) 1068 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 6430) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1069 (("(" 6427)) 1070 (("(" 6424)) 1071 (("(" 6420) (exit_expr 6421)) 1072 (("\"" 1489) ("$" 2573) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2574) ("[" 1508) (array_object_dereferenceable 2575) (callable_expr 2576) (callable_variable 2577) (class_constant 1512) (class_name 2578) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 2579) (name 2580) (simple_variable 2581) (static_member 2582) (variable 6419) (variable_class_name 2584)) 1073 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 6418) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1074 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 6417) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1075 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 6416) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1076 (("(" 6412)) 1077 (("(" 6409)) 1078 (("(" 6403)) 1079 (("$" 2851) ("(" 2852) (T_ATTRIBUTE 2853) (T_CLASS 2854) (T_NAME_FULLY_QUALIFIED 2855) (T_NAME_QUALIFIED 2856) (T_NAME_RELATIVE 2857) (T_STATIC 2858) (T_STRING 2859) (T_VARIABLE 2860) (anonymous_class 2861) (attribute 2862) (attributes 2863) (class_name 2864) (class_name_reference 2865) (name 2866) (new_variable 2867) (simple_variable 2868)) 1080 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2850) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1081 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2849) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1082 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2848) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1083 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2847) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1084 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2820) (T_END_HEREDOC 2821) (T_VARIABLE 2822) (encaps_list 2823) (encaps_var 2824)) 1085 ((T_FN 35) (T_FUNCTION 126) (fn 2328) #'2329 (inline_function 2817)) 1086 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2816) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1087 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2815) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1088 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2814) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1089 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 2811) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 1090 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2810) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1091 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 2806) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 1092 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 2728) (encaps_list 2729) (encaps_var 2730)) 1093 (("\"" 1489) ("$" 2674) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2675) ("[" 1508) (array_object_dereferenceable 2676) (callable_expr 2677) (callable_variable 2678) (class_constant 1512) (class_name 2679) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 2680) (name 2681) (simple_variable 2682) (static_member 2683) (variable 2684) (variable_class_name 2685)) 1094 ((T_NULLSAFE_OBJECT_OPERATOR 2657) (T_OBJECT_OPERATOR 2658) ("[" 2659) ("{" 2660)) 1095 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 2653) (attribute 2654) (fn 2328) #'2329 (inline_function 2655)) 1096 (("(" 2534) (argument_list 2652)) 1097 ((T_PAAMAYIM_NEKUDOTAYIM 2647)) 1098 (("%" 2344) ("*" 2345) ("+" 2346) ("-" 2347) ("." 2348) ("/" 2349) ("<" 2350) (">" 2351) ("?" 2352) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2353) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2354) (T_BOOLEAN_AND 2355) (T_BOOLEAN_OR 2356) (T_COALESCE 2357) (T_INSTANCEOF 2358) (T_IS_EQUAL 2359) (T_IS_GREATER_OR_EQUAL 2360) (T_IS_IDENTICAL 2361) (T_IS_NOT_EQUAL 2362) (T_IS_NOT_IDENTICAL 2363) (T_IS_SMALLER_OR_EQUAL 2364) (T_LOGICAL_AND 2365) (T_LOGICAL_OR 2366) (T_LOGICAL_XOR 2367) (T_POW 2368) (T_SL 2369) (T_SPACESHIP 2370) (T_SR 2371) ("^" 2372) ("|" 2373)) 1099 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 2636)) 1100 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 2624)) 1101 (("(" 2534) (argument_list 2623)) 1102 (("=" 2542) (T_AND_EQUAL 2543) (T_COALESCE_EQUAL 2544) (T_CONCAT_EQUAL 2545) (T_DEC 2546) (T_DIV_EQUAL 2547) (T_INC 2548) (T_MINUS_EQUAL 2549) (T_MOD_EQUAL 2550) (T_MUL_EQUAL 2551) (T_OR_EQUAL 2552) (T_PLUS_EQUAL 2553) (T_POW_EQUAL 2554) (T_SL_EQUAL 2555) (T_SR_EQUAL 2556) (T_XOR_EQUAL 2557)) 1103 ((T_PAAMAYIM_NEKUDOTAYIM 2450)) 1104 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2343) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1105 (("(" 2445)) 1106 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2444) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1107 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2443) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1108 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2442) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1109 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2441) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1110 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2440) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1111 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2439) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1112 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2438) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1113 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2437) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1114 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 2432) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 2433) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 1115 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2431) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1116 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2430) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1117 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2429) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1118 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2428) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1119 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2427) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1120 (("$" 2389) ("(" 2390) (T_NAME_FULLY_QUALIFIED 2391) (T_NAME_QUALIFIED 2392) (T_NAME_RELATIVE 2393) (T_STATIC 2394) (T_STRING 2395) (T_VARIABLE 2396) (class_name 2397) (class_name_reference 2398) (name 2399) (new_variable 2400) (simple_variable 2401)) 1121 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2388) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1122 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2387) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1123 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2386) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1124 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2385) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1125 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2384) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1126 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2383) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1127 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2382) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1128 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2381) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1129 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2380) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1130 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2379) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1131 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2378) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1132 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2377) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1133 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2376) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1134 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2375) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1135 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2374) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1136 (("$" 2389) (T_VARIABLE 2396) (simple_variable 2423) ("{" 2424)) 1137 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 2421) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 1138 ((T_PAAMAYIM_NEKUDOTAYIM 2419)) 1139 ((T_NULLSAFE_OBJECT_OPERATOR 2402) (T_OBJECT_OPERATOR 2403) (T_PAAMAYIM_NEKUDOTAYIM 2404) ("[" 2405) ("{" 2406)) 1140 (("$" 2389) (T_STRING 2412) (T_VARIABLE 2396) (property_name 2418) (simple_variable 2414) ("{" 2415)) 1141 (("$" 2389) (T_STRING 2412) (T_VARIABLE 2396) (property_name 2413) (simple_variable 2414) ("{" 2415)) 1142 (("$" 2389) (T_VARIABLE 2396) (simple_variable 2411)) 1143 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 2409) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1144 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2407) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1145 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2408)) 1146 (("]" 2410)) 1147 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2416) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1148 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2417)) 1149 (("$" 2389) (T_VARIABLE 2396) (simple_variable 2420)) 1150 (("%" 280) (")" 2422) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 1151 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2425) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1152 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2426)) 1153 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2436) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1154 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 2434) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 1155 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2435) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1156 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 2446) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 1157 ((")" 2447)) 1158 (("=" 2448)) 1159 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2449) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1160 (("$" 2267) (T_ABSTRACT 2451) (T_ARRAY 2452) (T_AS 2453) (T_BREAK 2454) (T_CALLABLE 2455) (T_CASE 2456) (T_CATCH 2457) (T_CLASS 2458) (T_CLASS_C 2459) (T_CLONE 2460) (T_CONST 2461) (T_CONTINUE 2462) (T_DECLARE 2463) (T_DEFAULT 2464) (T_DIR 2465) (T_DO 2466) (T_ECHO 2467) (T_ELSE 2468) (T_ELSEIF 2469) (T_EMPTY 2470) (T_ENDDECLARE 2471) (T_ENDFOR 2472) (T_ENDFOREACH 2473) (T_ENDIF 2474) (T_ENDSWITCH 2475) (T_ENDWHILE 2476) (T_ENUM 2477) (T_EVAL 2478) (T_EXIT 2479) (T_EXTENDS 2480) (T_FILE 2481) (T_FINAL 2482) (T_FINALLY 2483) (T_FN 2484) (T_FOR 2485) (T_FOREACH 2486) (T_FUNCTION 2487) (T_FUNC_C 2488) (T_GLOBAL 2489) (T_GOTO 2490) (T_IF 2491) (T_IMPLEMENTS 2492) (T_INCLUDE 2493) (T_INCLUDE_ONCE 2494) (T_INSTANCEOF 2495) (T_INSTEADOF 2496) (T_INTERFACE 2497) (T_ISSET 2498) (T_LINE 2499) (T_LIST 2500) (T_LOGICAL_AND 2501) (T_LOGICAL_OR 2502) (T_LOGICAL_XOR 2503) (T_MATCH 2504) (T_METHOD_C 2505) (T_NAMESPACE 2506) (T_NEW 2507) (T_NS_C 2508) (T_PRINT 2509) (T_PRIVATE 2510) (T_PROTECTED 2511) (T_PUBLIC 2512) (T_READONLY 2513) (T_REQUIRE 2514) (T_REQUIRE_ONCE 2515) (T_RETURN 2516) (T_STATIC 2517) (T_STRING 2518) (T_SWITCH 2519) (T_THROW 2520) (T_TRAIT 2521) (T_TRAIT_C 2522) (T_TRY 2523) (T_UNSET 2524) (T_USE 2525) (T_VAR 2526) (T_VARIABLE 2313) (T_WHILE 2527) (T_YIELD 2528) (identifier 2529) (member_name 2530) (reserved_non_modifiers 2531) (semi_reserved 2532) (simple_variable 2533) ("{" 914)) 1161 (("(" 2534) (argument_list 2535)) 1162 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2536) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 2537) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 2538) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1163 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2541) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1164 (("," 1449) (possible_comma 2539)) 1165 ((")" 2540)) 1166 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (ampersand 2571) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2572) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1167 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2570) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1168 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2569) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1169 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2568) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1170 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2567) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1171 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2566) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1172 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2565) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1173 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2564) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1174 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2563) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1175 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2562) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1176 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2561) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1177 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2560) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1178 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2559) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1179 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2558) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1180 (("\"" 1489) ("$" 2573) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2574) ("[" 1508) (array_object_dereferenceable 2575) (callable_expr 2576) (callable_variable 2577) (class_constant 1512) (class_name 2578) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 2579) (name 2580) (simple_variable 2581) (static_member 2582) (variable 2583) (variable_class_name 2584)) 1181 (("$" 2573) (T_VARIABLE 2574) (simple_variable 2619) ("{" 2620)) 1182 ((T_NULLSAFE_OBJECT_OPERATOR 2602) (T_OBJECT_OPERATOR 2603) ("[" 2604) ("{" 2605)) 1183 (("(" 2588) (argument_list 2601)) 1184 ((T_PAAMAYIM_NEKUDOTAYIM 2597)) 1185 (("(" 2588) (argument_list 2596)) 1186 ((T_PAAMAYIM_NEKUDOTAYIM 2585)) 1187 (("$" 2573) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 2574) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 2586) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 2587) ("{" 914)) 1188 (("(" 2588) (argument_list 2589)) 1189 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2590) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 2591) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 2592) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1190 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2595) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1191 (("," 1449) (possible_comma 2593)) 1192 ((")" 2594)) 1193 (("$" 2573) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 2574) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 2598) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 2599) ("{" 914)) 1194 (("(" 2588) (argument_list 2600)) 1195 (("$" 2573) (T_STRING 2610) (T_VARIABLE 2574) (property_name 2617) (simple_variable 2612) ("{" 2613)) 1196 (("$" 2573) (T_STRING 2610) (T_VARIABLE 2574) (property_name 2611) (simple_variable 2612) ("{" 2613)) 1197 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 2608) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1198 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2606) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1199 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2607)) 1200 (("]" 2609)) 1201 (("(" 2588) (argument_list 2616)) 1202 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2614) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1203 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2615)) 1204 (("(" 2588) (argument_list 2618)) 1205 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2621) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1206 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2622)) 1207 ((backup_doc_comment 2625)) 1208 (("(" 2626)) 1209 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 2627)) 1210 ((")" 2628)) 1211 ((T_USE 148) (lexical_vars 2629)) 1212 ((":" 150) (return_type 2630)) 1213 ((backup_fn_flags 2631)) 1214 (("{" 2632)) 1215 ((inner_statement_list 2633)) 1216 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 2634) ("~" 124)) 1217 ((backup_fn_flags 2635)) 1218 ((backup_doc_comment 2637)) 1219 (("(" 2638)) 1220 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 2639)) 1221 ((")" 2640)) 1222 ((":" 2207) (return_type 2641)) 1223 ((T_DOUBLE_ARROW 2642)) 1224 ((backup_fn_flags 2643)) 1225 ((backup_lex_pos 2644)) 1226 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 2645) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1227 (("%" 1081) ("*" 1082) ("+" 1083) ("-" 1084) ("." 1085) ("/" 1086) ("<" 1087) (">" 1088) ("?" 1089) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1090) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1091) (T_BOOLEAN_AND 1092) (T_BOOLEAN_OR 1093) (T_COALESCE 1094) (T_INSTANCEOF 1095) (T_IS_EQUAL 1096) (T_IS_GREATER_OR_EQUAL 1097) (T_IS_IDENTICAL 1098) (T_IS_NOT_EQUAL 1099) (T_IS_NOT_IDENTICAL 1100) (T_IS_SMALLER_OR_EQUAL 1101) (T_LOGICAL_AND 1102) (T_LOGICAL_OR 1103) (T_LOGICAL_XOR 1104) (T_POW 1105) (T_SL 1106) (T_SPACESHIP 1107) (T_SR 1108) ("^" 1109) (backup_fn_flags 2646) ("|" 1110)) 1228 (("$" 2267) (T_ABSTRACT 2451) (T_ARRAY 2452) (T_AS 2453) (T_BREAK 2454) (T_CALLABLE 2455) (T_CASE 2456) (T_CATCH 2457) (T_CLASS 2458) (T_CLASS_C 2459) (T_CLONE 2460) (T_CONST 2461) (T_CONTINUE 2462) (T_DECLARE 2463) (T_DEFAULT 2464) (T_DIR 2465) (T_DO 2466) (T_ECHO 2467) (T_ELSE 2468) (T_ELSEIF 2469) (T_EMPTY 2470) (T_ENDDECLARE 2471) (T_ENDFOR 2472) (T_ENDFOREACH 2473) (T_ENDIF 2474) (T_ENDSWITCH 2475) (T_ENDWHILE 2476) (T_ENUM 2477) (T_EVAL 2478) (T_EXIT 2479) (T_EXTENDS 2480) (T_FILE 2481) (T_FINAL 2482) (T_FINALLY 2483) (T_FN 2484) (T_FOR 2485) (T_FOREACH 2486) (T_FUNCTION 2487) (T_FUNC_C 2488) (T_GLOBAL 2489) (T_GOTO 2490) (T_IF 2491) (T_IMPLEMENTS 2492) (T_INCLUDE 2493) (T_INCLUDE_ONCE 2494) (T_INSTANCEOF 2495) (T_INSTEADOF 2496) (T_INTERFACE 2497) (T_ISSET 2498) (T_LINE 2499) (T_LIST 2500) (T_LOGICAL_AND 2501) (T_LOGICAL_OR 2502) (T_LOGICAL_XOR 2503) (T_MATCH 2504) (T_METHOD_C 2505) (T_NAMESPACE 2506) (T_NEW 2507) (T_NS_C 2508) (T_PRINT 2509) (T_PRIVATE 2510) (T_PROTECTED 2511) (T_PUBLIC 2512) (T_READONLY 2513) (T_REQUIRE 2514) (T_REQUIRE_ONCE 2515) (T_RETURN 2516) (T_STATIC 2517) (T_STRING 2518) (T_SWITCH 2519) (T_THROW 2520) (T_TRAIT 2521) (T_TRAIT_C 2522) (T_TRY 2523) (T_UNSET 2524) (T_USE 2525) (T_VAR 2526) (T_VARIABLE 2313) (T_WHILE 2527) (T_YIELD 2528) (identifier 2648) (member_name 2649) (reserved_non_modifiers 2531) (semi_reserved 2532) (simple_variable 2650) ("{" 914)) 1229 (("(" 2534) (argument_list 2651)) 1230 ((T_FN 35) (T_FUNCTION 126) (fn 2328) #'2329 (inline_function 2656)) 1231 (("$" 2267) (T_STRING 2665) (T_VARIABLE 2313) (property_name 2672) (simple_variable 2667) ("{" 2668)) 1232 (("$" 2267) (T_STRING 2665) (T_VARIABLE 2313) (property_name 2666) (simple_variable 2667) ("{" 2668)) 1233 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 2663) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1234 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2661) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1235 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2662)) 1236 (("]" 2664)) 1237 (("(" 2534) (argument_list 2671)) 1238 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2669) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1239 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2670)) 1240 (("(" 2534) (argument_list 2673)) 1241 (("$" 2674) (T_VARIABLE 2675) (simple_variable 2720) ("{" 2721)) 1242 ((T_NULLSAFE_OBJECT_OPERATOR 2703) (T_OBJECT_OPERATOR 2704) ("[" 2705) ("{" 2706)) 1243 (("(" 2689) (argument_list 2702)) 1244 ((T_PAAMAYIM_NEKUDOTAYIM 2698)) 1245 (("(" 2689) (argument_list 2697)) 1246 ((T_PAAMAYIM_NEKUDOTAYIM 2686)) 1247 (("$" 2674) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 2675) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 2687) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 2688) ("{" 914)) 1248 (("(" 2689) (argument_list 2690)) 1249 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2691) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 2692) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 2693) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1250 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2696) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1251 (("," 1449) (possible_comma 2694)) 1252 ((")" 2695)) 1253 (("$" 2674) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 2675) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 2699) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 2700) ("{" 914)) 1254 (("(" 2689) (argument_list 2701)) 1255 (("$" 2674) (T_STRING 2711) (T_VARIABLE 2675) (property_name 2718) (simple_variable 2713) ("{" 2714)) 1256 (("$" 2674) (T_STRING 2711) (T_VARIABLE 2675) (property_name 2712) (simple_variable 2713) ("{" 2714)) 1257 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 2709) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1258 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2707) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1259 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2708)) 1260 (("]" 2710)) 1261 (("(" 2689) (argument_list 2717)) 1262 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2715) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1263 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2716)) 1264 (("(" 2689) (argument_list 2719)) 1265 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2722) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1266 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2723)) 1267 (("\"" 1489) ("$" 2755) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2756) ("[" 1508) (array_object_dereferenceable 2757) (callable_expr 2758) (callable_variable 2759) (class_constant 1512) (class_name 2760) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 2761) (name 2762) (simple_variable 2763) (static_member 2764) (variable 2765) (variable_class_name 2766)) 1268 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_STRING_VARNAME 2747) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2748) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1269 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_VARIABLE 2727) (encaps_var 2746)) 1270 ((T_NULLSAFE_OBJECT_OPERATOR 2734) (T_OBJECT_OPERATOR 2735) ("[" 2736)) 1271 (("`" 2733)) 1272 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2731) (T_VARIABLE 2727) (encaps_var 2732)) 1273 ((T_STRING 2745)) 1274 ((T_STRING 2744)) 1275 (("-" 2737) (T_NUM_STRING 2738) (T_STRING 2739) (T_VARIABLE 2740) (encaps_var_offset 2741)) 1276 ((T_NUM_STRING 2743)) 1277 (("]" 2742)) 1278 (("[" 2750) ("}" 2751)) 1279 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2749)) 1280 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 2752) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1281 (("%" 563) ("*" 564) ("+" 565) ("-" 566) ("." 567) ("/" 568) ("<" 569) (">" 570) ("?" 571) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 572) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 573) (T_BOOLEAN_AND 574) (T_BOOLEAN_OR 575) (T_COALESCE 576) (T_INSTANCEOF 577) (T_IS_EQUAL 578) (T_IS_GREATER_OR_EQUAL 579) (T_IS_IDENTICAL 580) (T_IS_NOT_EQUAL 581) (T_IS_NOT_IDENTICAL 582) (T_IS_SMALLER_OR_EQUAL 583) (T_LOGICAL_AND 584) (T_LOGICAL_OR 585) (T_LOGICAL_XOR 586) (T_POW 587) (T_SL 588) (T_SPACESHIP 589) (T_SR 590) ("]" 2753) ("^" 591) ("|" 592)) 1282 (("}" 2754)) 1283 (("$" 2755) (T_VARIABLE 2756) (simple_variable 2802) ("{" 2803)) 1284 ((T_NULLSAFE_OBJECT_OPERATOR 2785) (T_OBJECT_OPERATOR 2786) ("[" 2787) ("{" 2788)) 1285 (("(" 2770) (argument_list 2784)) 1286 ((T_PAAMAYIM_NEKUDOTAYIM 2780)) 1287 (("(" 2770) (argument_list 2779)) 1288 (("}" 2778)) 1289 ((T_PAAMAYIM_NEKUDOTAYIM 2767)) 1290 (("$" 2755) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 2756) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 2768) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 2769) ("{" 914)) 1291 (("(" 2770) (argument_list 2771)) 1292 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2772) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 2773) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 2774) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1293 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2777) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1294 (("," 1449) (possible_comma 2775)) 1295 ((")" 2776)) 1296 (("$" 2755) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 2756) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 2781) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 2782) ("{" 914)) 1297 (("(" 2770) (argument_list 2783)) 1298 (("$" 2755) (T_STRING 2793) (T_VARIABLE 2756) (property_name 2800) (simple_variable 2795) ("{" 2796)) 1299 (("$" 2755) (T_STRING 2793) (T_VARIABLE 2756) (property_name 2794) (simple_variable 2795) ("{" 2796)) 1300 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 2791) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1301 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2789) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1302 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2790)) 1303 (("]" 2792)) 1304 (("(" 2770) (argument_list 2799)) 1305 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2797) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1306 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2798)) 1307 (("(" 2770) (argument_list 2801)) 1308 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2804) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1309 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2805)) 1310 (("]" 2807)) 1311 (("=" 2808)) 1312 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2809) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1313 (("%" 1908) ("*" 1909) ("+" 1910) ("-" 1911) ("." 1912) ("/" 1913) ("<" 1914) (">" 1915) ("?" 1916) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1917) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1918) (T_BOOLEAN_AND 1919) (T_BOOLEAN_OR 1920) (T_COALESCE 1921) (T_DOUBLE_ARROW 2812) (T_INSTANCEOF 1922) (T_IS_EQUAL 1923) (T_IS_GREATER_OR_EQUAL 1924) (T_IS_IDENTICAL 1925) (T_IS_NOT_EQUAL 1926) (T_IS_NOT_IDENTICAL 1927) (T_IS_SMALLER_OR_EQUAL 1928) (T_LOGICAL_AND 1929) (T_LOGICAL_OR 1930) (T_LOGICAL_XOR 1931) (T_POW 1932) (T_SL 1933) (T_SPACESHIP 1934) (T_SR 1935) ("^" 1936) ("|" 1937)) 1314 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 2813) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 1315 (("\"" 1489) ("$" 2755) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2756) ("[" 1508) (array_object_dereferenceable 2757) (callable_expr 2758) (callable_variable 2759) (class_constant 1512) (class_name 2760) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 2761) (name 2762) (simple_variable 2763) (static_member 2764) (variable 2845) (variable_class_name 2766)) 1316 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_STRING_VARNAME 2837) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2838) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1317 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 2835) (T_VARIABLE 2822) (encaps_var 2836)) 1318 ((T_NULLSAFE_OBJECT_OPERATOR 2828) (T_OBJECT_OPERATOR 2829) ("[" 2830)) 1319 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 2826) (T_VARIABLE 2822) (encaps_var 2827)) 1320 ((T_STRING 2834)) 1321 ((T_STRING 2833)) 1322 (("-" 2737) (T_NUM_STRING 2738) (T_STRING 2739) (T_VARIABLE 2740) (encaps_var_offset 2831)) 1323 (("]" 2832)) 1324 (("[" 2840) ("}" 2841)) 1325 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2839)) 1326 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 2842) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1327 (("%" 563) ("*" 564) ("+" 565) ("-" 566) ("." 567) ("/" 568) ("<" 569) (">" 570) ("?" 571) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 572) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 573) (T_BOOLEAN_AND 574) (T_BOOLEAN_OR 575) (T_COALESCE 576) (T_INSTANCEOF 577) (T_IS_EQUAL 578) (T_IS_GREATER_OR_EQUAL 579) (T_IS_IDENTICAL 580) (T_IS_NOT_EQUAL 581) (T_IS_NOT_IDENTICAL 582) (T_IS_SMALLER_OR_EQUAL 583) (T_LOGICAL_AND 584) (T_LOGICAL_OR 585) (T_LOGICAL_XOR 586) (T_POW 587) (T_SL 588) (T_SPACESHIP 589) (T_SR 590) ("]" 2843) ("^" 591) ("|" 592)) 1328 (("}" 2844)) 1329 (("}" 2846)) 1330 (("$" 2851) (T_VARIABLE 2860) (simple_variable 6399) ("{" 6400)) 1331 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 6397) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 1332 ((T_NAME_FULLY_QUALIFIED 6350) (T_NAME_QUALIFIED 6351) (T_NAME_RELATIVE 6352) (T_STATIC 6353) (T_STRING 6354) (attribute_decl 6355) (attribute_group 6394) (class_name 6357) (name 6358)) 1333 (("(" 2899) (argument_list 2900) (ctor_arguments 2901)) 1334 ((T_ATTRIBUTE 2853) (T_CLASS 2854) (anonymous_class 2897) (attribute 2898)) 1335 ((T_PAAMAYIM_NEKUDOTAYIM 2895)) 1336 (("(" 2886) (argument_list 2887) (ctor_arguments 2888)) 1337 ((T_NULLSAFE_OBJECT_OPERATOR 2869) (T_OBJECT_OPERATOR 2870) (T_PAAMAYIM_NEKUDOTAYIM 2871) ("[" 2872) ("{" 2873)) 1338 (("$" 2851) (T_STRING 2879) (T_VARIABLE 2860) (property_name 2885) (simple_variable 2881) ("{" 2882)) 1339 (("$" 2851) (T_STRING 2879) (T_VARIABLE 2860) (property_name 2880) (simple_variable 2881) ("{" 2882)) 1340 (("$" 2851) (T_VARIABLE 2860) (simple_variable 2878)) 1341 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 2876) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1342 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2874) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1343 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2875)) 1344 (("]" 2877)) 1345 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 2883) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1346 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 2884)) 1347 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2889) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 2890) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 2891) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1348 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 2894) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1349 (("," 1449) (possible_comma 2892)) 1350 ((")" 2893)) 1351 (("$" 2851) (T_VARIABLE 2860) (simple_variable 2896)) 1352 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 6388) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 6389) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 6390) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1353 ((T_EXTENDS 2902) (extends_from 2903)) 1354 ((T_NAME_FULLY_QUALIFIED 6381) (T_NAME_QUALIFIED 6382) (T_NAME_RELATIVE 6383) (T_STATIC 6384) (T_STRING 6385) (class_name 6386) (name 6387)) 1355 ((T_IMPLEMENTS 2904) (implements_list 2905)) 1356 ((T_NAME_FULLY_QUALIFIED 6371) (T_NAME_QUALIFIED 6372) (T_NAME_RELATIVE 6373) (T_STATIC 6374) (T_STRING 6375) (class_name 6376) (class_name_list 6377) (name 6378)) 1357 ((backup_doc_comment 2906)) 1358 (("{" 2907)) 1359 ((class_statement_list 2908)) 1360 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 2929)) 1361 ((T_NAME_FULLY_QUALIFIED 6350) (T_NAME_QUALIFIED 6351) (T_NAME_RELATIVE 6352) (T_STATIC 6353) (T_STRING 6354) (attribute_decl 6355) (attribute_group 6356) (class_name 6357) (name 6358)) 1362 ((backup_doc_comment 6163)) 1363 ((T_NAME_FULLY_QUALIFIED 5868) (T_NAME_QUALIFIED 5869) (T_NAME_RELATIVE 5870) (T_STATIC 5871) (T_STRING 5872) (class_name 5873) (class_name_list 5874) (name 5875)) 1364 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_VAR 2919) (attribute 5866) (attributed_class_statement 5867) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928)) 1365 ((T_CONST 5626) (T_FUNCTION 5627) #'5628) 1366 ((T_ABSTRACT 2909) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (member_modifier 5625)) 1367 (("?" 2930) (T_ARRAY 2931) (T_CALLABLE 2932) (T_NAME_FULLY_QUALIFIED 2933) (T_NAME_QUALIFIED 2934) (T_NAME_RELATIVE 2935) (T_STRING 2936) (intersection_type_without_static 2937) (name 2938) (optional_type_without_static 2939) (type_expr_without_static 2940) (type_without_static 2941) (union_type_without_static 2942)) 1368 ((T_ARRAY 5617) (T_CALLABLE 5618) (T_NAME_FULLY_QUALIFIED 5619) (T_NAME_QUALIFIED 5620) (T_NAME_RELATIVE 5621) (T_STRING 5622) (name 5623) (type_without_static 5624)) 1369 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5615)) 1370 ((T_VARIABLE 2963) (property 2964) (property_list 2965)) 1371 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2952) ("|" 2953)) 1372 (("|" 2943)) 1373 ((T_ARRAY 2944) (T_CALLABLE 2945) (T_NAME_FULLY_QUALIFIED 2946) (T_NAME_QUALIFIED 2947) (T_NAME_RELATIVE 2948) (T_STRING 2949) (name 2950) (type_without_static 2951)) 1374 ((T_ARRAY 2955) (T_CALLABLE 2956) (T_NAME_FULLY_QUALIFIED 2957) (T_NAME_QUALIFIED 2958) (T_NAME_RELATIVE 2959) (T_STRING 2960) (name 2961) (type_without_static 2962)) 1375 ((T_ARRAY 2944) (T_CALLABLE 2945) (T_NAME_FULLY_QUALIFIED 2946) (T_NAME_QUALIFIED 2947) (T_NAME_RELATIVE 2948) (T_STRING 2949) (name 2950) (type_without_static 2954)) 1376 (("=" 2969) (backup_doc_comment 2970)) 1377 (("," 2966) (";" 2967)) 1378 ((T_VARIABLE 2963) (property 2968)) 1379 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3032) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1380 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5614) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1381 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 5612) (encaps_var 5158)) 1382 (("$" 2973) (T_VARIABLE 3019) (simple_variable 5608) ("{" 5609)) 1383 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5606) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 1384 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5605) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1385 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5604) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1386 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5603) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1387 (("(" 5600)) 1388 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5599) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1389 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5598) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1390 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5597) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1391 (("\"" 1489) ("$" 3272) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 3273) ("[" 1508) (array_object_dereferenceable 3274) (callable_expr 3275) (callable_variable 3276) (class_constant 1512) (class_name 3277) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 3278) (name 3279) (simple_variable 3280) (static_member 3281) (variable 5596) (variable_class_name 3283)) 1392 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5595) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1393 (("(" 5592)) 1394 (("(" 5589)) 1395 (("(" 5585) (exit_expr 5586)) 1396 (("\"" 1489) ("$" 3272) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 3273) ("[" 1508) (array_object_dereferenceable 3274) (callable_expr 3275) (callable_variable 3276) (class_constant 1512) (class_name 3277) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 3278) (name 3279) (simple_variable 3280) (static_member 3281) (variable 5584) (variable_class_name 3283)) 1397 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5583) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1398 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5582) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1399 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5581) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1400 (("(" 5577)) 1401 (("(" 5572)) 1402 (("(" 5566)) 1403 (("$" 3272) ("(" 5519) (T_ATTRIBUTE 2853) (T_CLASS 5520) (T_NAME_FULLY_QUALIFIED 5521) (T_NAME_QUALIFIED 5522) (T_NAME_RELATIVE 5523) (T_STATIC 5524) (T_STRING 5525) (T_VARIABLE 3273) (anonymous_class 5526) (attribute 2862) (attributes 5527) (class_name 5528) (class_name_reference 5529) (name 5530) (new_variable 5531) (simple_variable 5532)) 1404 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5518) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1405 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5517) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1406 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5516) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1407 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5515) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1408 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 5510) (T_END_HEREDOC 5511) (T_VARIABLE 2822) (encaps_list 5512) (encaps_var 2824)) 1409 ((T_FN 35) (T_FUNCTION 126) (fn 3033) #'3034 (inline_function 5509)) 1410 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5508) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1411 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5507) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1412 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5506) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1413 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3441) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1414 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3379) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1415 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 3375) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 1416 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 3373) (encaps_list 2729) (encaps_var 2730)) 1417 ((T_NULLSAFE_OBJECT_OPERATOR 3356) (T_OBJECT_OPERATOR 3357) ("[" 3358) ("{" 3359)) 1418 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 3353) (attribute 2654) (fn 3033) #'3034 (inline_function 3354)) 1419 (("(" 3233) (argument_list 3352)) 1420 ((T_PAAMAYIM_NEKUDOTAYIM 3347)) 1421 (("%" 3048) ("*" 3049) ("+" 3050) ("-" 3051) ("." 3052) ("/" 3053) ("<" 3054) (">" 3055) ("?" 3056) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3057) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3058) (T_BOOLEAN_AND 3059) (T_BOOLEAN_OR 3060) (T_COALESCE 3061) (T_INSTANCEOF 3062) (T_IS_EQUAL 3063) (T_IS_GREATER_OR_EQUAL 3064) (T_IS_IDENTICAL 3065) (T_IS_NOT_EQUAL 3066) (T_IS_NOT_IDENTICAL 3067) (T_IS_SMALLER_OR_EQUAL 3068) (T_LOGICAL_AND 3069) (T_LOGICAL_OR 3070) (T_LOGICAL_XOR 3071) (T_POW 3072) (T_SL 3073) (T_SPACESHIP 3074) (T_SR 3075) ("^" 3076) (backup_doc_comment 3346) ("|" 3077)) 1422 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 3335)) 1423 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 3323)) 1424 (("(" 3233) (argument_list 3322)) 1425 (("=" 3241) (T_AND_EQUAL 3242) (T_COALESCE_EQUAL 3243) (T_CONCAT_EQUAL 3244) (T_DEC 3245) (T_DIV_EQUAL 3246) (T_INC 3247) (T_MINUS_EQUAL 3248) (T_MOD_EQUAL 3249) (T_MUL_EQUAL 3250) (T_OR_EQUAL 3251) (T_PLUS_EQUAL 3252) (T_POW_EQUAL 3253) (T_SL_EQUAL 3254) (T_SR_EQUAL 3255) (T_XOR_EQUAL 3256)) 1426 ((T_PAAMAYIM_NEKUDOTAYIM 3149)) 1427 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3047) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1428 (("%" 3048) ("*" 3049) ("+" 3050) ("-" 3051) ("." 3052) ("/" 3053) ("<" 3054) (">" 3055) ("?" 3056) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3057) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3058) (T_BOOLEAN_AND 3059) (T_BOOLEAN_OR 3060) (T_COALESCE 3061) (T_INSTANCEOF 3062) (T_IS_EQUAL 3063) (T_IS_GREATER_OR_EQUAL 3064) (T_IS_IDENTICAL 3065) (T_IS_NOT_EQUAL 3066) (T_IS_NOT_IDENTICAL 3067) (T_IS_SMALLER_OR_EQUAL 3068) (T_LOGICAL_AND 3069) (T_LOGICAL_OR 3070) (T_LOGICAL_XOR 3071) (T_POW 3072) (T_SL 3073) (T_SPACESHIP 3074) (T_SR 3075) ("^" 3076) ("|" 3077)) 1429 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3148) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1430 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3147) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1431 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3146) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1432 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3145) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1433 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3144) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1434 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3143) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1435 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3142) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1436 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3141) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1437 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 3136) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 3137) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 1438 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3135) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1439 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3134) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1440 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3133) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1441 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3132) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1442 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3131) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1443 (("$" 3093) ("(" 3094) (T_NAME_FULLY_QUALIFIED 3095) (T_NAME_QUALIFIED 3096) (T_NAME_RELATIVE 3097) (T_STATIC 3098) (T_STRING 3099) (T_VARIABLE 3100) (class_name 3101) (class_name_reference 3102) (name 3103) (new_variable 3104) (simple_variable 3105)) 1444 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3092) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1445 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3091) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1446 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3090) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1447 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3089) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1448 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3088) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1449 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3087) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1450 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3086) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1451 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3085) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1452 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3084) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1453 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3083) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1454 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3082) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1455 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3081) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1456 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3080) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1457 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3079) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1458 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3078) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1459 (("$" 3093) (T_VARIABLE 3100) (simple_variable 3127) ("{" 3128)) 1460 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 3125) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 1461 ((T_PAAMAYIM_NEKUDOTAYIM 3123)) 1462 ((T_NULLSAFE_OBJECT_OPERATOR 3106) (T_OBJECT_OPERATOR 3107) (T_PAAMAYIM_NEKUDOTAYIM 3108) ("[" 3109) ("{" 3110)) 1463 (("$" 3093) (T_STRING 3116) (T_VARIABLE 3100) (property_name 3122) (simple_variable 3118) ("{" 3119)) 1464 (("$" 3093) (T_STRING 3116) (T_VARIABLE 3100) (property_name 3117) (simple_variable 3118) ("{" 3119)) 1465 (("$" 3093) (T_VARIABLE 3100) (simple_variable 3115)) 1466 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 3113) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1467 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3111) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1468 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3112)) 1469 (("]" 3114)) 1470 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3120) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1471 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3121)) 1472 (("$" 3093) (T_VARIABLE 3100) (simple_variable 3124)) 1473 (("%" 280) (")" 3126) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 1474 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3129) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1475 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3130)) 1476 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3140) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1477 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 3138) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 1478 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3139) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1479 (("$" 2973) (T_ABSTRACT 3150) (T_ARRAY 3151) (T_AS 3152) (T_BREAK 3153) (T_CALLABLE 3154) (T_CASE 3155) (T_CATCH 3156) (T_CLASS 3157) (T_CLASS_C 3158) (T_CLONE 3159) (T_CONST 3160) (T_CONTINUE 3161) (T_DECLARE 3162) (T_DEFAULT 3163) (T_DIR 3164) (T_DO 3165) (T_ECHO 3166) (T_ELSE 3167) (T_ELSEIF 3168) (T_EMPTY 3169) (T_ENDDECLARE 3170) (T_ENDFOR 3171) (T_ENDFOREACH 3172) (T_ENDIF 3173) (T_ENDSWITCH 3174) (T_ENDWHILE 3175) (T_ENUM 3176) (T_EVAL 3177) (T_EXIT 3178) (T_EXTENDS 3179) (T_FILE 3180) (T_FINAL 3181) (T_FINALLY 3182) (T_FN 3183) (T_FOR 3184) (T_FOREACH 3185) (T_FUNCTION 3186) (T_FUNC_C 3187) (T_GLOBAL 3188) (T_GOTO 3189) (T_IF 3190) (T_IMPLEMENTS 3191) (T_INCLUDE 3192) (T_INCLUDE_ONCE 3193) (T_INSTANCEOF 3194) (T_INSTEADOF 3195) (T_INTERFACE 3196) (T_ISSET 3197) (T_LINE 3198) (T_LIST 3199) (T_LOGICAL_AND 3200) (T_LOGICAL_OR 3201) (T_LOGICAL_XOR 3202) (T_MATCH 3203) (T_METHOD_C 3204) (T_NAMESPACE 3205) (T_NEW 3206) (T_NS_C 3207) (T_PRINT 3208) (T_PRIVATE 3209) (T_PROTECTED 3210) (T_PUBLIC 3211) (T_READONLY 3212) (T_REQUIRE 3213) (T_REQUIRE_ONCE 3214) (T_RETURN 3215) (T_STATIC 3216) (T_STRING 3217) (T_SWITCH 3218) (T_THROW 3219) (T_TRAIT 3220) (T_TRAIT_C 3221) (T_TRY 3222) (T_UNSET 3223) (T_USE 3224) (T_VAR 3225) (T_VARIABLE 3019) (T_WHILE 3226) (T_YIELD 3227) (identifier 3228) (member_name 3229) (reserved_non_modifiers 3230) (semi_reserved 3231) (simple_variable 3232) ("{" 914)) 1480 (("(" 3233) (argument_list 3234)) 1481 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 3235) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 3236) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 3237) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1482 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 3240) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1483 (("," 1449) (possible_comma 3238)) 1484 ((")" 3239)) 1485 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (ampersand 3270) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3271) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1486 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3269) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1487 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3268) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1488 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3267) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1489 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3266) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1490 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3265) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1491 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3264) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1492 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3263) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1493 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3262) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1494 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3261) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1495 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3260) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1496 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3259) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1497 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3258) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1498 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3257) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1499 (("\"" 1489) ("$" 3272) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 3273) ("[" 1508) (array_object_dereferenceable 3274) (callable_expr 3275) (callable_variable 3276) (class_constant 1512) (class_name 3277) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 3278) (name 3279) (simple_variable 3280) (static_member 3281) (variable 3282) (variable_class_name 3283)) 1500 (("$" 3272) (T_VARIABLE 3273) (simple_variable 3318) ("{" 3319)) 1501 ((T_NULLSAFE_OBJECT_OPERATOR 3301) (T_OBJECT_OPERATOR 3302) ("[" 3303) ("{" 3304)) 1502 (("(" 3287) (argument_list 3300)) 1503 ((T_PAAMAYIM_NEKUDOTAYIM 3296)) 1504 (("(" 3287) (argument_list 3295)) 1505 ((T_PAAMAYIM_NEKUDOTAYIM 3284)) 1506 (("$" 3272) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 3273) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 3285) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 3286) ("{" 914)) 1507 (("(" 3287) (argument_list 3288)) 1508 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 3289) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 3290) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 3291) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1509 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 3294) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1510 (("," 1449) (possible_comma 3292)) 1511 ((")" 3293)) 1512 (("$" 3272) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 3273) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 3297) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 3298) ("{" 914)) 1513 (("(" 3287) (argument_list 3299)) 1514 (("$" 3272) (T_STRING 3309) (T_VARIABLE 3273) (property_name 3316) (simple_variable 3311) ("{" 3312)) 1515 (("$" 3272) (T_STRING 3309) (T_VARIABLE 3273) (property_name 3310) (simple_variable 3311) ("{" 3312)) 1516 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 3307) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1517 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3305) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1518 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3306)) 1519 (("]" 3308)) 1520 (("(" 3287) (argument_list 3315)) 1521 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3313) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1522 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3314)) 1523 (("(" 3287) (argument_list 3317)) 1524 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3320) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1525 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3321)) 1526 ((backup_doc_comment 3324)) 1527 (("(" 3325)) 1528 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 3326)) 1529 ((")" 3327)) 1530 ((T_USE 148) (lexical_vars 3328)) 1531 ((":" 150) (return_type 3329)) 1532 ((backup_fn_flags 3330)) 1533 (("{" 3331)) 1534 ((inner_statement_list 3332)) 1535 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 3333) ("~" 124)) 1536 ((backup_fn_flags 3334)) 1537 ((backup_doc_comment 3336)) 1538 (("(" 3337)) 1539 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 3338)) 1540 ((")" 3339)) 1541 ((":" 2207) (return_type 3340)) 1542 ((T_DOUBLE_ARROW 3341)) 1543 ((backup_fn_flags 3342)) 1544 ((backup_lex_pos 3343)) 1545 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3344) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1546 (("%" 3048) ("*" 3049) ("+" 3050) ("-" 3051) ("." 3052) ("/" 3053) ("<" 3054) (">" 3055) ("?" 3056) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3057) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3058) (T_BOOLEAN_AND 3059) (T_BOOLEAN_OR 3060) (T_COALESCE 3061) (T_INSTANCEOF 3062) (T_IS_EQUAL 3063) (T_IS_GREATER_OR_EQUAL 3064) (T_IS_IDENTICAL 3065) (T_IS_NOT_EQUAL 3066) (T_IS_NOT_IDENTICAL 3067) (T_IS_SMALLER_OR_EQUAL 3068) (T_LOGICAL_AND 3069) (T_LOGICAL_OR 3070) (T_LOGICAL_XOR 3071) (T_POW 3072) (T_SL 3073) (T_SPACESHIP 3074) (T_SR 3075) ("^" 3076) (backup_fn_flags 3345) ("|" 3077)) 1547 (("$" 2973) (T_ABSTRACT 3150) (T_ARRAY 3151) (T_AS 3152) (T_BREAK 3153) (T_CALLABLE 3154) (T_CASE 3155) (T_CATCH 3156) (T_CLASS 3157) (T_CLASS_C 3158) (T_CLONE 3159) (T_CONST 3160) (T_CONTINUE 3161) (T_DECLARE 3162) (T_DEFAULT 3163) (T_DIR 3164) (T_DO 3165) (T_ECHO 3166) (T_ELSE 3167) (T_ELSEIF 3168) (T_EMPTY 3169) (T_ENDDECLARE 3170) (T_ENDFOR 3171) (T_ENDFOREACH 3172) (T_ENDIF 3173) (T_ENDSWITCH 3174) (T_ENDWHILE 3175) (T_ENUM 3176) (T_EVAL 3177) (T_EXIT 3178) (T_EXTENDS 3179) (T_FILE 3180) (T_FINAL 3181) (T_FINALLY 3182) (T_FN 3183) (T_FOR 3184) (T_FOREACH 3185) (T_FUNCTION 3186) (T_FUNC_C 3187) (T_GLOBAL 3188) (T_GOTO 3189) (T_IF 3190) (T_IMPLEMENTS 3191) (T_INCLUDE 3192) (T_INCLUDE_ONCE 3193) (T_INSTANCEOF 3194) (T_INSTEADOF 3195) (T_INTERFACE 3196) (T_ISSET 3197) (T_LINE 3198) (T_LIST 3199) (T_LOGICAL_AND 3200) (T_LOGICAL_OR 3201) (T_LOGICAL_XOR 3202) (T_MATCH 3203) (T_METHOD_C 3204) (T_NAMESPACE 3205) (T_NEW 3206) (T_NS_C 3207) (T_PRINT 3208) (T_PRIVATE 3209) (T_PROTECTED 3210) (T_PUBLIC 3211) (T_READONLY 3212) (T_REQUIRE 3213) (T_REQUIRE_ONCE 3214) (T_RETURN 3215) (T_STATIC 3216) (T_STRING 3217) (T_SWITCH 3218) (T_THROW 3219) (T_TRAIT 3220) (T_TRAIT_C 3221) (T_TRY 3222) (T_UNSET 3223) (T_USE 3224) (T_VAR 3225) (T_VARIABLE 3019) (T_WHILE 3226) (T_YIELD 3227) (identifier 3348) (member_name 3349) (reserved_non_modifiers 3230) (semi_reserved 3231) (simple_variable 3350) ("{" 914)) 1548 (("(" 3233) (argument_list 3351)) 1549 ((T_FN 35) (T_FUNCTION 126) (fn 3033) #'3034 (inline_function 3355)) 1550 (("$" 2973) (T_STRING 3364) (T_VARIABLE 3019) (property_name 3371) (simple_variable 3366) ("{" 3367)) 1551 (("$" 2973) (T_STRING 3364) (T_VARIABLE 3019) (property_name 3365) (simple_variable 3366) ("{" 3367)) 1552 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 3362) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1553 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3360) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1554 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3361)) 1555 (("]" 3363)) 1556 (("(" 3233) (argument_list 3370)) 1557 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3368) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1558 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3369)) 1559 (("(" 3233) (argument_list 3372)) 1560 (("`" 3374)) 1561 (("]" 3376)) 1562 (("=" 3377)) 1563 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3378) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1564 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 5505) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1565 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 5503) (encaps_var 5158)) 1566 (("$" 3382) (T_VARIABLE 3428) (simple_variable 5499) ("{" 5500)) 1567 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5497) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 1568 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 5496) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1569 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 5495) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1570 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 5494) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1571 (("(" 5491)) 1572 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 5490) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1573 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 5489) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1574 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 5488) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1575 (("\"" 1489) ("$" 3681) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 3682) ("[" 1508) (array_object_dereferenceable 3683) (callable_expr 3684) (callable_variable 3685) (class_constant 1512) (class_name 3686) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 3687) (name 3688) (simple_variable 3689) (static_member 3690) (variable 5487) (variable_class_name 3692)) 1576 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 5486) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1577 (("(" 5483)) 1578 (("(" 5480)) 1579 (("(" 5476) (exit_expr 5477)) 1580 (("\"" 1489) ("$" 3681) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 3682) ("[" 1508) (array_object_dereferenceable 3683) (callable_expr 3684) (callable_variable 3685) (class_constant 1512) (class_name 3686) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 3687) (name 3688) (simple_variable 3689) (static_member 3690) (variable 5475) (variable_class_name 3692)) 1581 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 5474) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1582 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 5473) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1583 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 5472) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1584 (("(" 5468)) 1585 (("(" 5463)) 1586 (("(" 3853)) 1587 (("$" 3681) ("(" 3806) (T_ATTRIBUTE 2853) (T_CLASS 3807) (T_NAME_FULLY_QUALIFIED 3808) (T_NAME_QUALIFIED 3809) (T_NAME_RELATIVE 3810) (T_STATIC 3811) (T_STRING 3812) (T_VARIABLE 3682) (anonymous_class 3813) (attribute 2862) (attributes 3814) (class_name 3815) (class_name_reference 3816) (name 3817) (new_variable 3818) (simple_variable 3819)) 1588 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3805) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1589 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3804) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1590 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3803) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1591 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3802) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1592 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 3797) (T_END_HEREDOC 3798) (T_VARIABLE 2822) (encaps_list 3799) (encaps_var 2824)) 1593 ((T_FN 35) (T_FUNCTION 126) (fn 3442) #'3443 (inline_function 3796)) 1594 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3795) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1595 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3794) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1596 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3793) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1597 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3790) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1598 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3789) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1599 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 3785) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 1600 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 3783) (encaps_list 2729) (encaps_var 2730)) 1601 ((T_NULLSAFE_OBJECT_OPERATOR 3766) (T_OBJECT_OPERATOR 3767) ("[" 3768) ("{" 3769)) 1602 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 3763) (attribute 2654) (fn 3442) #'3443 (inline_function 3764)) 1603 (("(" 3642) (argument_list 3762)) 1604 ((T_PAAMAYIM_NEKUDOTAYIM 3757)) 1605 (("%" 3457) ("*" 3458) ("+" 3459) ("-" 3460) ("." 3461) ("/" 3462) ("<" 3463) (">" 3464) ("?" 3465) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3466) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3467) (T_BOOLEAN_AND 3468) (T_BOOLEAN_OR 3469) (T_COALESCE 3470) (T_DOUBLE_ARROW 3755) (T_INSTANCEOF 3471) (T_IS_EQUAL 3472) (T_IS_GREATER_OR_EQUAL 3473) (T_IS_IDENTICAL 3474) (T_IS_NOT_EQUAL 3475) (T_IS_NOT_IDENTICAL 3476) (T_IS_SMALLER_OR_EQUAL 3477) (T_LOGICAL_AND 3478) (T_LOGICAL_OR 3479) (T_LOGICAL_XOR 3480) (T_POW 3481) (T_SL 3482) (T_SPACESHIP 3483) (T_SR 3484) ("^" 3485) ("|" 3486)) 1606 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 3744)) 1607 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 3732)) 1608 (("(" 3642) (argument_list 3731)) 1609 (("=" 3650) (T_AND_EQUAL 3651) (T_COALESCE_EQUAL 3652) (T_CONCAT_EQUAL 3653) (T_DEC 3654) (T_DIV_EQUAL 3655) (T_INC 3656) (T_MINUS_EQUAL 3657) (T_MOD_EQUAL 3658) (T_MUL_EQUAL 3659) (T_OR_EQUAL 3660) (T_PLUS_EQUAL 3661) (T_POW_EQUAL 3662) (T_SL_EQUAL 3663) (T_SR_EQUAL 3664) (T_XOR_EQUAL 3665)) 1610 ((T_PAAMAYIM_NEKUDOTAYIM 3558)) 1611 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3456) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1612 (("%" 3457) ("*" 3458) ("+" 3459) ("-" 3460) ("." 3461) ("/" 3462) ("<" 3463) (">" 3464) ("?" 3465) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3466) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3467) (T_BOOLEAN_AND 3468) (T_BOOLEAN_OR 3469) (T_COALESCE 3470) (T_INSTANCEOF 3471) (T_IS_EQUAL 3472) (T_IS_GREATER_OR_EQUAL 3473) (T_IS_IDENTICAL 3474) (T_IS_NOT_EQUAL 3475) (T_IS_NOT_IDENTICAL 3476) (T_IS_SMALLER_OR_EQUAL 3477) (T_LOGICAL_AND 3478) (T_LOGICAL_OR 3479) (T_LOGICAL_XOR 3480) (T_POW 3481) (T_SL 3482) (T_SPACESHIP 3483) (T_SR 3484) ("^" 3485) ("|" 3486)) 1613 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3557) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1614 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3556) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1615 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3555) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1616 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3554) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1617 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3553) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1618 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3552) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1619 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3551) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1620 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3550) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1621 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 3545) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 3546) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 1622 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3544) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1623 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3543) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1624 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3542) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1625 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3541) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1626 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3540) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1627 (("$" 3502) ("(" 3503) (T_NAME_FULLY_QUALIFIED 3504) (T_NAME_QUALIFIED 3505) (T_NAME_RELATIVE 3506) (T_STATIC 3507) (T_STRING 3508) (T_VARIABLE 3509) (class_name 3510) (class_name_reference 3511) (name 3512) (new_variable 3513) (simple_variable 3514)) 1628 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3501) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1629 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3500) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1630 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3499) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1631 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3498) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1632 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3497) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1633 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3496) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1634 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3495) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1635 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3494) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1636 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3493) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1637 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3492) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1638 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3491) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1639 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3490) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1640 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3489) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1641 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3488) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1642 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3487) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1643 (("$" 3502) (T_VARIABLE 3509) (simple_variable 3536) ("{" 3537)) 1644 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 3534) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 1645 ((T_PAAMAYIM_NEKUDOTAYIM 3532)) 1646 ((T_NULLSAFE_OBJECT_OPERATOR 3515) (T_OBJECT_OPERATOR 3516) (T_PAAMAYIM_NEKUDOTAYIM 3517) ("[" 3518) ("{" 3519)) 1647 (("$" 3502) (T_STRING 3525) (T_VARIABLE 3509) (property_name 3531) (simple_variable 3527) ("{" 3528)) 1648 (("$" 3502) (T_STRING 3525) (T_VARIABLE 3509) (property_name 3526) (simple_variable 3527) ("{" 3528)) 1649 (("$" 3502) (T_VARIABLE 3509) (simple_variable 3524)) 1650 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 3522) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1651 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3520) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1652 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3521)) 1653 (("]" 3523)) 1654 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3529) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1655 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3530)) 1656 (("$" 3502) (T_VARIABLE 3509) (simple_variable 3533)) 1657 (("%" 280) (")" 3535) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 1658 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3538) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1659 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3539)) 1660 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3549) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1661 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 3547) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 1662 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3548) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1663 (("$" 3382) (T_ABSTRACT 3559) (T_ARRAY 3560) (T_AS 3561) (T_BREAK 3562) (T_CALLABLE 3563) (T_CASE 3564) (T_CATCH 3565) (T_CLASS 3566) (T_CLASS_C 3567) (T_CLONE 3568) (T_CONST 3569) (T_CONTINUE 3570) (T_DECLARE 3571) (T_DEFAULT 3572) (T_DIR 3573) (T_DO 3574) (T_ECHO 3575) (T_ELSE 3576) (T_ELSEIF 3577) (T_EMPTY 3578) (T_ENDDECLARE 3579) (T_ENDFOR 3580) (T_ENDFOREACH 3581) (T_ENDIF 3582) (T_ENDSWITCH 3583) (T_ENDWHILE 3584) (T_ENUM 3585) (T_EVAL 3586) (T_EXIT 3587) (T_EXTENDS 3588) (T_FILE 3589) (T_FINAL 3590) (T_FINALLY 3591) (T_FN 3592) (T_FOR 3593) (T_FOREACH 3594) (T_FUNCTION 3595) (T_FUNC_C 3596) (T_GLOBAL 3597) (T_GOTO 3598) (T_IF 3599) (T_IMPLEMENTS 3600) (T_INCLUDE 3601) (T_INCLUDE_ONCE 3602) (T_INSTANCEOF 3603) (T_INSTEADOF 3604) (T_INTERFACE 3605) (T_ISSET 3606) (T_LINE 3607) (T_LIST 3608) (T_LOGICAL_AND 3609) (T_LOGICAL_OR 3610) (T_LOGICAL_XOR 3611) (T_MATCH 3612) (T_METHOD_C 3613) (T_NAMESPACE 3614) (T_NEW 3615) (T_NS_C 3616) (T_PRINT 3617) (T_PRIVATE 3618) (T_PROTECTED 3619) (T_PUBLIC 3620) (T_READONLY 3621) (T_REQUIRE 3622) (T_REQUIRE_ONCE 3623) (T_RETURN 3624) (T_STATIC 3625) (T_STRING 3626) (T_SWITCH 3627) (T_THROW 3628) (T_TRAIT 3629) (T_TRAIT_C 3630) (T_TRY 3631) (T_UNSET 3632) (T_USE 3633) (T_VAR 3634) (T_VARIABLE 3428) (T_WHILE 3635) (T_YIELD 3636) (identifier 3637) (member_name 3638) (reserved_non_modifiers 3639) (semi_reserved 3640) (simple_variable 3641) ("{" 914)) 1664 (("(" 3642) (argument_list 3643)) 1665 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 3644) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 3645) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 3646) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1666 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 3649) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1667 (("," 1449) (possible_comma 3647)) 1668 ((")" 3648)) 1669 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (ampersand 3679) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3680) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1670 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3678) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1671 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3677) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1672 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3676) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1673 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3675) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1674 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3674) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1675 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3673) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1676 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3672) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1677 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3671) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1678 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3670) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1679 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3669) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1680 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3668) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1681 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3667) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1682 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3666) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1683 (("\"" 1489) ("$" 3681) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 3682) ("[" 1508) (array_object_dereferenceable 3683) (callable_expr 3684) (callable_variable 3685) (class_constant 1512) (class_name 3686) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 3687) (name 3688) (simple_variable 3689) (static_member 3690) (variable 3691) (variable_class_name 3692)) 1684 (("$" 3681) (T_VARIABLE 3682) (simple_variable 3727) ("{" 3728)) 1685 ((T_NULLSAFE_OBJECT_OPERATOR 3710) (T_OBJECT_OPERATOR 3711) ("[" 3712) ("{" 3713)) 1686 (("(" 3696) (argument_list 3709)) 1687 ((T_PAAMAYIM_NEKUDOTAYIM 3705)) 1688 (("(" 3696) (argument_list 3704)) 1689 ((T_PAAMAYIM_NEKUDOTAYIM 3693)) 1690 (("$" 3681) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 3682) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 3694) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 3695) ("{" 914)) 1691 (("(" 3696) (argument_list 3697)) 1692 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 3698) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 3699) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 3700) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1693 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 3703) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1694 (("," 1449) (possible_comma 3701)) 1695 ((")" 3702)) 1696 (("$" 3681) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 3682) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 3706) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 3707) ("{" 914)) 1697 (("(" 3696) (argument_list 3708)) 1698 (("$" 3681) (T_STRING 3718) (T_VARIABLE 3682) (property_name 3725) (simple_variable 3720) ("{" 3721)) 1699 (("$" 3681) (T_STRING 3718) (T_VARIABLE 3682) (property_name 3719) (simple_variable 3720) ("{" 3721)) 1700 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 3716) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1701 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3714) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1702 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3715)) 1703 (("]" 3717)) 1704 (("(" 3696) (argument_list 3724)) 1705 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3722) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1706 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3723)) 1707 (("(" 3696) (argument_list 3726)) 1708 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3729) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1709 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3730)) 1710 ((backup_doc_comment 3733)) 1711 (("(" 3734)) 1712 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 3735)) 1713 ((")" 3736)) 1714 ((T_USE 148) (lexical_vars 3737)) 1715 ((":" 150) (return_type 3738)) 1716 ((backup_fn_flags 3739)) 1717 (("{" 3740)) 1718 ((inner_statement_list 3741)) 1719 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 3742) ("~" 124)) 1720 ((backup_fn_flags 3743)) 1721 ((backup_doc_comment 3745)) 1722 (("(" 3746)) 1723 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 3747)) 1724 ((")" 3748)) 1725 ((":" 2207) (return_type 3749)) 1726 ((T_DOUBLE_ARROW 3750)) 1727 ((backup_fn_flags 3751)) 1728 ((backup_lex_pos 3752)) 1729 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3753) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1730 (("%" 3457) ("*" 3458) ("+" 3459) ("-" 3460) ("." 3461) ("/" 3462) ("<" 3463) (">" 3464) ("?" 3465) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3466) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3467) (T_BOOLEAN_AND 3468) (T_BOOLEAN_OR 3469) (T_COALESCE 3470) (T_INSTANCEOF 3471) (T_IS_EQUAL 3472) (T_IS_GREATER_OR_EQUAL 3473) (T_IS_IDENTICAL 3474) (T_IS_NOT_EQUAL 3475) (T_IS_NOT_IDENTICAL 3476) (T_IS_SMALLER_OR_EQUAL 3477) (T_LOGICAL_AND 3478) (T_LOGICAL_OR 3479) (T_LOGICAL_XOR 3480) (T_POW 3481) (T_SL 3482) (T_SPACESHIP 3483) (T_SR 3484) ("^" 3485) (backup_fn_flags 3754) ("|" 3486)) 1731 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 3756) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 1732 (("$" 3382) (T_ABSTRACT 3559) (T_ARRAY 3560) (T_AS 3561) (T_BREAK 3562) (T_CALLABLE 3563) (T_CASE 3564) (T_CATCH 3565) (T_CLASS 3566) (T_CLASS_C 3567) (T_CLONE 3568) (T_CONST 3569) (T_CONTINUE 3570) (T_DECLARE 3571) (T_DEFAULT 3572) (T_DIR 3573) (T_DO 3574) (T_ECHO 3575) (T_ELSE 3576) (T_ELSEIF 3577) (T_EMPTY 3578) (T_ENDDECLARE 3579) (T_ENDFOR 3580) (T_ENDFOREACH 3581) (T_ENDIF 3582) (T_ENDSWITCH 3583) (T_ENDWHILE 3584) (T_ENUM 3585) (T_EVAL 3586) (T_EXIT 3587) (T_EXTENDS 3588) (T_FILE 3589) (T_FINAL 3590) (T_FINALLY 3591) (T_FN 3592) (T_FOR 3593) (T_FOREACH 3594) (T_FUNCTION 3595) (T_FUNC_C 3596) (T_GLOBAL 3597) (T_GOTO 3598) (T_IF 3599) (T_IMPLEMENTS 3600) (T_INCLUDE 3601) (T_INCLUDE_ONCE 3602) (T_INSTANCEOF 3603) (T_INSTEADOF 3604) (T_INTERFACE 3605) (T_ISSET 3606) (T_LINE 3607) (T_LIST 3608) (T_LOGICAL_AND 3609) (T_LOGICAL_OR 3610) (T_LOGICAL_XOR 3611) (T_MATCH 3612) (T_METHOD_C 3613) (T_NAMESPACE 3614) (T_NEW 3615) (T_NS_C 3616) (T_PRINT 3617) (T_PRIVATE 3618) (T_PROTECTED 3619) (T_PUBLIC 3620) (T_READONLY 3621) (T_REQUIRE 3622) (T_REQUIRE_ONCE 3623) (T_RETURN 3624) (T_STATIC 3625) (T_STRING 3626) (T_SWITCH 3627) (T_THROW 3628) (T_TRAIT 3629) (T_TRAIT_C 3630) (T_TRY 3631) (T_UNSET 3632) (T_USE 3633) (T_VAR 3634) (T_VARIABLE 3428) (T_WHILE 3635) (T_YIELD 3636) (identifier 3758) (member_name 3759) (reserved_non_modifiers 3639) (semi_reserved 3640) (simple_variable 3760) ("{" 914)) 1733 (("(" 3642) (argument_list 3761)) 1734 ((T_FN 35) (T_FUNCTION 126) (fn 3442) #'3443 (inline_function 3765)) 1735 (("$" 3382) (T_STRING 3774) (T_VARIABLE 3428) (property_name 3781) (simple_variable 3776) ("{" 3777)) 1736 (("$" 3382) (T_STRING 3774) (T_VARIABLE 3428) (property_name 3775) (simple_variable 3776) ("{" 3777)) 1737 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 3772) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1738 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3770) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1739 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3771)) 1740 (("]" 3773)) 1741 (("(" 3642) (argument_list 3780)) 1742 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3778) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1743 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3779)) 1744 (("(" 3642) (argument_list 3782)) 1745 (("`" 3784)) 1746 (("]" 3786)) 1747 (("=" 3787)) 1748 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3788) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1749 (("%" 3457) ("*" 3458) ("+" 3459) ("-" 3460) ("." 3461) ("/" 3462) ("<" 3463) (">" 3464) ("?" 3465) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3466) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3467) (T_BOOLEAN_AND 3468) (T_BOOLEAN_OR 3469) (T_COALESCE 3470) (T_DOUBLE_ARROW 3791) (T_INSTANCEOF 3471) (T_IS_EQUAL 3472) (T_IS_GREATER_OR_EQUAL 3473) (T_IS_IDENTICAL 3474) (T_IS_NOT_EQUAL 3475) (T_IS_NOT_IDENTICAL 3476) (T_IS_SMALLER_OR_EQUAL 3477) (T_LOGICAL_AND 3478) (T_LOGICAL_OR 3479) (T_LOGICAL_XOR 3480) (T_POW 3481) (T_SL 3482) (T_SPACESHIP 3483) (T_SR 3484) ("^" 3485) ("|" 3486)) 1750 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 3792) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 1751 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 3801) (T_VARIABLE 2822) (encaps_var 2836)) 1752 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 3800) (T_VARIABLE 2822) (encaps_var 2827)) 1753 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 3851) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 1754 (("(" 2899) (argument_list 2900) (ctor_arguments 3844)) 1755 ((T_ATTRIBUTE 2853) (T_CLASS 3807) (anonymous_class 3843) (attribute 2898)) 1756 ((T_PAAMAYIM_NEKUDOTAYIM 3841)) 1757 (("(" 3832) (argument_list 3833) (ctor_arguments 3834)) 1758 ((T_NULLSAFE_OBJECT_OPERATOR 3820) (T_OBJECT_OPERATOR 3821) (T_PAAMAYIM_NEKUDOTAYIM 3822) ("[" 3823) ("{" 3824)) 1759 (("$" 3681) (T_STRING 3718) (T_VARIABLE 3682) (property_name 3831) (simple_variable 3720) ("{" 3721)) 1760 (("$" 3681) (T_STRING 3718) (T_VARIABLE 3682) (property_name 3830) (simple_variable 3720) ("{" 3721)) 1761 (("$" 3681) (T_VARIABLE 3682) (simple_variable 3829)) 1762 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 3827) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1763 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 3825) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1764 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 3826)) 1765 (("]" 3828)) 1766 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 3835) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 3836) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 3837) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1767 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 3840) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1768 (("," 1449) (possible_comma 3838)) 1769 ((")" 3839)) 1770 (("$" 3681) (T_VARIABLE 3682) (simple_variable 3842)) 1771 ((T_EXTENDS 2902) (extends_from 3845)) 1772 ((T_IMPLEMENTS 2904) (implements_list 3846)) 1773 ((backup_doc_comment 3847)) 1774 (("{" 3848)) 1775 ((class_statement_list 3849)) 1776 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 3850)) 1777 (("%" 280) (")" 3852) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 1778 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 3854) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 1779 (("%" 280) (")" 3855) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 1780 (("{" 3856)) 1781 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 3928) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1782 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5462) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1783 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 5460) (encaps_var 5158)) 1784 (("$" 3859) (T_VARIABLE 3906) (simple_variable 5456) ("{" 5457)) 1785 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5454) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 1786 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5453) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1787 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5452) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1788 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5451) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1789 (("(" 5448)) 1790 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5447) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1791 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5446) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1792 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5445) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1793 (("\"" 1489) ("$" 4163) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 4164) ("[" 1508) (array_object_dereferenceable 4165) (callable_expr 4166) (callable_variable 4167) (class_constant 1512) (class_name 4168) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 4169) (name 4170) (simple_variable 4171) (static_member 4172) (variable 5444) (variable_class_name 4174)) 1794 (("," 5440) (possible_comma 5441)) 1795 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5439) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1796 (("(" 5436)) 1797 (("(" 5433)) 1798 (("(" 5429) (exit_expr 5430)) 1799 (("\"" 1489) ("$" 4163) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 4164) ("[" 1508) (array_object_dereferenceable 4165) (callable_expr 4166) (callable_variable 4167) (class_constant 1512) (class_name 4168) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 4169) (name 4170) (simple_variable 4171) (static_member 4172) (variable 5428) (variable_class_name 4174)) 1800 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5427) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1801 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5426) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1802 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5425) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1803 (("(" 5421)) 1804 (("(" 5416)) 1805 (("(" 5410)) 1806 (("$" 4163) ("(" 5363) (T_ATTRIBUTE 2853) (T_CLASS 5364) (T_NAME_FULLY_QUALIFIED 5365) (T_NAME_QUALIFIED 5366) (T_NAME_RELATIVE 5367) (T_STATIC 5368) (T_STRING 5369) (T_VARIABLE 4164) (anonymous_class 5370) (attribute 2862) (attributes 5371) (class_name 5372) (class_name_reference 5373) (name 5374) (new_variable 5375) (simple_variable 5376)) 1807 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5362) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1808 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5361) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1809 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5360) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1810 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5359) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1811 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 5354) (T_END_HEREDOC 5355) (T_VARIABLE 2822) (encaps_list 5356) (encaps_var 2824)) 1812 ((T_FN 35) (T_FUNCTION 126) (fn 3920) #'3921 (inline_function 5353)) 1813 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5352) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1814 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5351) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1815 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5350) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1816 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5347) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1817 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5346) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1818 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 5342) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 1819 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 5340) (encaps_list 2729) (encaps_var 2730)) 1820 ((T_NULLSAFE_OBJECT_OPERATOR 5323) (T_OBJECT_OPERATOR 5324) ("[" 5325) ("{" 5326)) 1821 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 5320) (attribute 2654) (fn 3920) #'3921 (inline_function 5321)) 1822 (("(" 4124) (argument_list 5319)) 1823 ((T_PAAMAYIM_NEKUDOTAYIM 5314)) 1824 (("%" 3939) ("*" 3940) ("+" 3941) ("-" 3942) ("." 3943) ("/" 3944) ("<" 3945) (">" 3946) ("?" 3947) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3948) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3949) (T_BOOLEAN_AND 3950) (T_BOOLEAN_OR 3951) (T_COALESCE 3952) (T_INSTANCEOF 3953) (T_IS_EQUAL 3954) (T_IS_GREATER_OR_EQUAL 3955) (T_IS_IDENTICAL 3956) (T_IS_NOT_EQUAL 3957) (T_IS_NOT_IDENTICAL 3958) (T_IS_SMALLER_OR_EQUAL 3959) (T_LOGICAL_AND 3960) (T_LOGICAL_OR 3961) (T_LOGICAL_XOR 3962) (T_POW 3963) (T_SL 3964) (T_SPACESHIP 3965) (T_SR 3966) ("^" 3967) ("|" 3968)) 1825 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 5303)) 1826 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 5291)) 1827 (("," 4218) (possible_comma 4219)) 1828 (("}" 4217)) 1829 (("(" 4124) (argument_list 4216)) 1830 (("," 4213) (possible_comma 4214)) 1831 (("=" 4132) (T_AND_EQUAL 4133) (T_COALESCE_EQUAL 4134) (T_CONCAT_EQUAL 4135) (T_DEC 4136) (T_DIV_EQUAL 4137) (T_INC 4138) (T_MINUS_EQUAL 4139) (T_MOD_EQUAL 4140) (T_MUL_EQUAL 4141) (T_OR_EQUAL 4142) (T_PLUS_EQUAL 4143) (T_POW_EQUAL 4144) (T_SL_EQUAL 4145) (T_SR_EQUAL 4146) (T_XOR_EQUAL 4147)) 1832 ((T_PAAMAYIM_NEKUDOTAYIM 4040)) 1833 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3938) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1834 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4039) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1835 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4038) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1836 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4037) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1837 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4036) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1838 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4035) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1839 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4034) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1840 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4033) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1841 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4032) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1842 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 4027) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 4028) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 1843 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4026) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1844 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4025) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1845 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4024) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1846 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4023) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1847 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4022) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1848 (("$" 3984) ("(" 3985) (T_NAME_FULLY_QUALIFIED 3986) (T_NAME_QUALIFIED 3987) (T_NAME_RELATIVE 3988) (T_STATIC 3989) (T_STRING 3990) (T_VARIABLE 3991) (class_name 3992) (class_name_reference 3993) (name 3994) (new_variable 3995) (simple_variable 3996)) 1849 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3983) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1850 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3982) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1851 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3981) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1852 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3980) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1853 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3979) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1854 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3978) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1855 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3977) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1856 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3976) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1857 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3975) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1858 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3974) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1859 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3973) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1860 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3972) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1861 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3971) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1862 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3970) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1863 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3969) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1864 (("$" 3984) (T_VARIABLE 3991) (simple_variable 4018) ("{" 4019)) 1865 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 4016) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 1866 ((T_PAAMAYIM_NEKUDOTAYIM 4014)) 1867 ((T_NULLSAFE_OBJECT_OPERATOR 3997) (T_OBJECT_OPERATOR 3998) (T_PAAMAYIM_NEKUDOTAYIM 3999) ("[" 4000) ("{" 4001)) 1868 (("$" 3984) (T_STRING 4007) (T_VARIABLE 3991) (property_name 4013) (simple_variable 4009) ("{" 4010)) 1869 (("$" 3984) (T_STRING 4007) (T_VARIABLE 3991) (property_name 4008) (simple_variable 4009) ("{" 4010)) 1870 (("$" 3984) (T_VARIABLE 3991) (simple_variable 4006)) 1871 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 4004) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1872 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4002) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1873 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4003)) 1874 (("]" 4005)) 1875 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4011) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1876 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4012)) 1877 (("$" 3984) (T_VARIABLE 3991) (simple_variable 4015)) 1878 (("%" 280) (")" 4017) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 1879 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4020) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1880 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4021)) 1881 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4031) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1882 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 4029) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 1883 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4030) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1884 (("$" 3859) (T_ABSTRACT 4041) (T_ARRAY 4042) (T_AS 4043) (T_BREAK 4044) (T_CALLABLE 4045) (T_CASE 4046) (T_CATCH 4047) (T_CLASS 4048) (T_CLASS_C 4049) (T_CLONE 4050) (T_CONST 4051) (T_CONTINUE 4052) (T_DECLARE 4053) (T_DEFAULT 4054) (T_DIR 4055) (T_DO 4056) (T_ECHO 4057) (T_ELSE 4058) (T_ELSEIF 4059) (T_EMPTY 4060) (T_ENDDECLARE 4061) (T_ENDFOR 4062) (T_ENDFOREACH 4063) (T_ENDIF 4064) (T_ENDSWITCH 4065) (T_ENDWHILE 4066) (T_ENUM 4067) (T_EVAL 4068) (T_EXIT 4069) (T_EXTENDS 4070) (T_FILE 4071) (T_FINAL 4072) (T_FINALLY 4073) (T_FN 4074) (T_FOR 4075) (T_FOREACH 4076) (T_FUNCTION 4077) (T_FUNC_C 4078) (T_GLOBAL 4079) (T_GOTO 4080) (T_IF 4081) (T_IMPLEMENTS 4082) (T_INCLUDE 4083) (T_INCLUDE_ONCE 4084) (T_INSTANCEOF 4085) (T_INSTEADOF 4086) (T_INTERFACE 4087) (T_ISSET 4088) (T_LINE 4089) (T_LIST 4090) (T_LOGICAL_AND 4091) (T_LOGICAL_OR 4092) (T_LOGICAL_XOR 4093) (T_MATCH 4094) (T_METHOD_C 4095) (T_NAMESPACE 4096) (T_NEW 4097) (T_NS_C 4098) (T_PRINT 4099) (T_PRIVATE 4100) (T_PROTECTED 4101) (T_PUBLIC 4102) (T_READONLY 4103) (T_REQUIRE 4104) (T_REQUIRE_ONCE 4105) (T_RETURN 4106) (T_STATIC 4107) (T_STRING 4108) (T_SWITCH 4109) (T_THROW 4110) (T_TRAIT 4111) (T_TRAIT_C 4112) (T_TRY 4113) (T_UNSET 4114) (T_USE 4115) (T_VAR 4116) (T_VARIABLE 3906) (T_WHILE 4117) (T_YIELD 4118) (identifier 4119) (member_name 4120) (reserved_non_modifiers 4121) (semi_reserved 4122) (simple_variable 4123) ("{" 914)) 1885 (("(" 4124) (argument_list 4125)) 1886 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 4126) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 4127) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 4128) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1887 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 4131) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1888 (("," 1449) (possible_comma 4129)) 1889 ((")" 4130)) 1890 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (ampersand 4161) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4162) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1891 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4160) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1892 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4159) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1893 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4158) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1894 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4157) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1895 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4156) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1896 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4155) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1897 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4154) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1898 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4153) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1899 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4152) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1900 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4151) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1901 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4150) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1902 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4149) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1903 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 4148) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1904 (("\"" 1489) ("$" 4163) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 4164) ("[" 1508) (array_object_dereferenceable 4165) (callable_expr 4166) (callable_variable 4167) (class_constant 1512) (class_name 4168) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 4169) (name 4170) (simple_variable 4171) (static_member 4172) (variable 4173) (variable_class_name 4174)) 1905 (("$" 4163) (T_VARIABLE 4164) (simple_variable 4209) ("{" 4210)) 1906 ((T_NULLSAFE_OBJECT_OPERATOR 4192) (T_OBJECT_OPERATOR 4193) ("[" 4194) ("{" 4195)) 1907 (("(" 4178) (argument_list 4191)) 1908 ((T_PAAMAYIM_NEKUDOTAYIM 4187)) 1909 (("(" 4178) (argument_list 4186)) 1910 ((T_PAAMAYIM_NEKUDOTAYIM 4175)) 1911 (("$" 4163) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 4164) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 4176) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 4177) ("{" 914)) 1912 (("(" 4178) (argument_list 4179)) 1913 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 4180) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 4181) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 4182) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1914 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 4185) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 1915 (("," 1449) (possible_comma 4183)) 1916 ((")" 4184)) 1917 (("$" 4163) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 4164) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 4188) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 4189) ("{" 914)) 1918 (("(" 4178) (argument_list 4190)) 1919 (("$" 4163) (T_STRING 4200) (T_VARIABLE 4164) (property_name 4207) (simple_variable 4202) ("{" 4203)) 1920 (("$" 4163) (T_STRING 4200) (T_VARIABLE 4164) (property_name 4201) (simple_variable 4202) ("{" 4203)) 1921 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 4198) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 1922 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4196) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1923 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4197)) 1924 (("]" 4199)) 1925 (("(" 4178) (argument_list 4206)) 1926 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4204) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1927 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4205)) 1928 (("(" 4178) (argument_list 4208)) 1929 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4211) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 1930 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4212)) 1931 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 4215) (match_arm_cond_list 3927) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1932 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5290) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 1933 ((T_DOUBLE_ARROW 4220)) 1934 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4282) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1935 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5289) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1936 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 5287) (encaps_var 5158)) 1937 (("$" 4223) (T_VARIABLE 4269) (simple_variable 5283) ("{" 5284)) 1938 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5281) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 1939 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5280) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1940 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5279) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1941 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5278) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1942 (("(" 5275)) 1943 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5274) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1944 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5273) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1945 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5272) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1946 (("\"" 1489) ("$" 4522) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 4523) ("[" 1508) (array_object_dereferenceable 4524) (callable_expr 4525) (callable_variable 4526) (class_constant 1512) (class_name 4527) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 4528) (name 4529) (simple_variable 4530) (static_member 4531) (variable 5271) (variable_class_name 4533)) 1947 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5270) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1948 (("(" 5267)) 1949 (("(" 5264)) 1950 (("(" 5260) (exit_expr 5261)) 1951 (("\"" 1489) ("$" 4522) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 4523) ("[" 1508) (array_object_dereferenceable 4524) (callable_expr 4525) (callable_variable 4526) (class_constant 1512) (class_name 4527) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 4528) (name 4529) (simple_variable 4530) (static_member 4531) (variable 5259) (variable_class_name 4533)) 1952 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5258) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1953 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5257) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1954 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5256) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1955 (("(" 5252)) 1956 (("(" 5247)) 1957 (("(" 5241)) 1958 (("$" 4522) ("(" 5194) (T_ATTRIBUTE 2853) (T_CLASS 5195) (T_NAME_FULLY_QUALIFIED 5196) (T_NAME_QUALIFIED 5197) (T_NAME_RELATIVE 5198) (T_STATIC 5199) (T_STRING 5200) (T_VARIABLE 4523) (anonymous_class 5201) (attribute 2862) (attributes 5202) (class_name 5203) (class_name_reference 5204) (name 5205) (new_variable 5206) (simple_variable 5207)) 1959 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5193) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1960 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5192) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1961 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5191) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1962 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5190) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1963 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 5185) (T_END_HEREDOC 5186) (T_VARIABLE 2822) (encaps_list 5187) (encaps_var 2824)) 1964 ((T_FN 35) (T_FUNCTION 126) (fn 4283) #'4284 (inline_function 5184)) 1965 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5183) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1966 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5182) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1967 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5181) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1968 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4690) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 1969 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4628) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1970 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 4624) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 1971 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 4622) (encaps_list 2729) (encaps_var 2730)) 1972 ((T_NULLSAFE_OBJECT_OPERATOR 4605) (T_OBJECT_OPERATOR 4606) ("[" 4607) ("{" 4608)) 1973 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 4602) (attribute 2654) (fn 4283) #'4284 (inline_function 4603)) 1974 (("(" 4483) (argument_list 4601)) 1975 ((T_PAAMAYIM_NEKUDOTAYIM 4596)) 1976 (("%" 4298) ("*" 4299) ("+" 4300) ("-" 4301) ("." 4302) ("/" 4303) ("<" 4304) (">" 4305) ("?" 4306) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4307) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4308) (T_BOOLEAN_AND 4309) (T_BOOLEAN_OR 4310) (T_COALESCE 4311) (T_INSTANCEOF 4312) (T_IS_EQUAL 4313) (T_IS_GREATER_OR_EQUAL 4314) (T_IS_IDENTICAL 4315) (T_IS_NOT_EQUAL 4316) (T_IS_NOT_IDENTICAL 4317) (T_IS_SMALLER_OR_EQUAL 4318) (T_LOGICAL_AND 4319) (T_LOGICAL_OR 4320) (T_LOGICAL_XOR 4321) (T_POW 4322) (T_SL 4323) (T_SPACESHIP 4324) (T_SR 4325) ("^" 4326) ("|" 4327)) 1977 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 4585)) 1978 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 4573)) 1979 (("(" 4483) (argument_list 4572)) 1980 (("=" 4491) (T_AND_EQUAL 4492) (T_COALESCE_EQUAL 4493) (T_CONCAT_EQUAL 4494) (T_DEC 4495) (T_DIV_EQUAL 4496) (T_INC 4497) (T_MINUS_EQUAL 4498) (T_MOD_EQUAL 4499) (T_MUL_EQUAL 4500) (T_OR_EQUAL 4501) (T_PLUS_EQUAL 4502) (T_POW_EQUAL 4503) (T_SL_EQUAL 4504) (T_SR_EQUAL 4505) (T_XOR_EQUAL 4506)) 1981 ((T_PAAMAYIM_NEKUDOTAYIM 4399)) 1982 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4297) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1983 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4398) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1984 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4397) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1985 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4396) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1986 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4395) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1987 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4394) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1988 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4393) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1989 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4392) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1990 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4391) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1991 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 4386) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 4387) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 1992 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4385) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1993 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4384) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1994 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4383) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1995 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4382) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1996 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4381) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1997 (("$" 4343) ("(" 4344) (T_NAME_FULLY_QUALIFIED 4345) (T_NAME_QUALIFIED 4346) (T_NAME_RELATIVE 4347) (T_STATIC 4348) (T_STRING 4349) (T_VARIABLE 4350) (class_name 4351) (class_name_reference 4352) (name 4353) (new_variable 4354) (simple_variable 4355)) 1998 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4342) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 1999 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4341) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2000 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4340) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2001 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4339) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2002 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4338) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2003 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4337) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2004 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4336) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2005 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4335) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2006 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4334) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2007 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4333) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2008 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4332) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2009 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4331) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2010 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4330) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2011 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4329) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2012 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4328) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2013 (("$" 4343) (T_VARIABLE 4350) (simple_variable 4377) ("{" 4378)) 2014 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 4375) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2015 ((T_PAAMAYIM_NEKUDOTAYIM 4373)) 2016 ((T_NULLSAFE_OBJECT_OPERATOR 4356) (T_OBJECT_OPERATOR 4357) (T_PAAMAYIM_NEKUDOTAYIM 4358) ("[" 4359) ("{" 4360)) 2017 (("$" 4343) (T_STRING 4366) (T_VARIABLE 4350) (property_name 4372) (simple_variable 4368) ("{" 4369)) 2018 (("$" 4343) (T_STRING 4366) (T_VARIABLE 4350) (property_name 4367) (simple_variable 4368) ("{" 4369)) 2019 (("$" 4343) (T_VARIABLE 4350) (simple_variable 4365)) 2020 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 4363) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2021 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4361) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2022 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4362)) 2023 (("]" 4364)) 2024 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4370) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2025 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4371)) 2026 (("$" 4343) (T_VARIABLE 4350) (simple_variable 4374)) 2027 (("%" 280) (")" 4376) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2028 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4379) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2029 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4380)) 2030 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4390) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2031 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 4388) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 2032 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4389) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2033 (("$" 4223) (T_ABSTRACT 4400) (T_ARRAY 4401) (T_AS 4402) (T_BREAK 4403) (T_CALLABLE 4404) (T_CASE 4405) (T_CATCH 4406) (T_CLASS 4407) (T_CLASS_C 4408) (T_CLONE 4409) (T_CONST 4410) (T_CONTINUE 4411) (T_DECLARE 4412) (T_DEFAULT 4413) (T_DIR 4414) (T_DO 4415) (T_ECHO 4416) (T_ELSE 4417) (T_ELSEIF 4418) (T_EMPTY 4419) (T_ENDDECLARE 4420) (T_ENDFOR 4421) (T_ENDFOREACH 4422) (T_ENDIF 4423) (T_ENDSWITCH 4424) (T_ENDWHILE 4425) (T_ENUM 4426) (T_EVAL 4427) (T_EXIT 4428) (T_EXTENDS 4429) (T_FILE 4430) (T_FINAL 4431) (T_FINALLY 4432) (T_FN 4433) (T_FOR 4434) (T_FOREACH 4435) (T_FUNCTION 4436) (T_FUNC_C 4437) (T_GLOBAL 4438) (T_GOTO 4439) (T_IF 4440) (T_IMPLEMENTS 4441) (T_INCLUDE 4442) (T_INCLUDE_ONCE 4443) (T_INSTANCEOF 4444) (T_INSTEADOF 4445) (T_INTERFACE 4446) (T_ISSET 4447) (T_LINE 4448) (T_LIST 4449) (T_LOGICAL_AND 4450) (T_LOGICAL_OR 4451) (T_LOGICAL_XOR 4452) (T_MATCH 4453) (T_METHOD_C 4454) (T_NAMESPACE 4455) (T_NEW 4456) (T_NS_C 4457) (T_PRINT 4458) (T_PRIVATE 4459) (T_PROTECTED 4460) (T_PUBLIC 4461) (T_READONLY 4462) (T_REQUIRE 4463) (T_REQUIRE_ONCE 4464) (T_RETURN 4465) (T_STATIC 4466) (T_STRING 4467) (T_SWITCH 4468) (T_THROW 4469) (T_TRAIT 4470) (T_TRAIT_C 4471) (T_TRY 4472) (T_UNSET 4473) (T_USE 4474) (T_VAR 4475) (T_VARIABLE 4269) (T_WHILE 4476) (T_YIELD 4477) (identifier 4478) (member_name 4479) (reserved_non_modifiers 4480) (semi_reserved 4481) (simple_variable 4482) ("{" 914)) 2034 (("(" 4483) (argument_list 4484)) 2035 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 4485) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 4486) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 4487) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2036 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 4490) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2037 (("," 1449) (possible_comma 4488)) 2038 ((")" 4489)) 2039 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (ampersand 4520) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4521) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2040 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4519) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2041 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4518) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2042 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4517) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2043 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4516) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2044 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4515) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2045 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4514) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2046 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4513) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2047 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4512) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2048 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4511) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2049 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4510) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2050 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4509) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2051 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4508) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2052 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4507) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2053 (("\"" 1489) ("$" 4522) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 4523) ("[" 1508) (array_object_dereferenceable 4524) (callable_expr 4525) (callable_variable 4526) (class_constant 1512) (class_name 4527) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 4528) (name 4529) (simple_variable 4530) (static_member 4531) (variable 4532) (variable_class_name 4533)) 2054 (("$" 4522) (T_VARIABLE 4523) (simple_variable 4568) ("{" 4569)) 2055 ((T_NULLSAFE_OBJECT_OPERATOR 4551) (T_OBJECT_OPERATOR 4552) ("[" 4553) ("{" 4554)) 2056 (("(" 4537) (argument_list 4550)) 2057 ((T_PAAMAYIM_NEKUDOTAYIM 4546)) 2058 (("(" 4537) (argument_list 4545)) 2059 ((T_PAAMAYIM_NEKUDOTAYIM 4534)) 2060 (("$" 4522) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 4523) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 4535) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 4536) ("{" 914)) 2061 (("(" 4537) (argument_list 4538)) 2062 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 4539) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 4540) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 4541) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2063 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 4544) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2064 (("," 1449) (possible_comma 4542)) 2065 ((")" 4543)) 2066 (("$" 4522) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 4523) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 4547) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 4548) ("{" 914)) 2067 (("(" 4537) (argument_list 4549)) 2068 (("$" 4522) (T_STRING 4559) (T_VARIABLE 4523) (property_name 4566) (simple_variable 4561) ("{" 4562)) 2069 (("$" 4522) (T_STRING 4559) (T_VARIABLE 4523) (property_name 4560) (simple_variable 4561) ("{" 4562)) 2070 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 4557) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2071 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4555) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2072 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4556)) 2073 (("]" 4558)) 2074 (("(" 4537) (argument_list 4565)) 2075 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4563) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2076 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4564)) 2077 (("(" 4537) (argument_list 4567)) 2078 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4570) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2079 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4571)) 2080 ((backup_doc_comment 4574)) 2081 (("(" 4575)) 2082 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 4576)) 2083 ((")" 4577)) 2084 ((T_USE 148) (lexical_vars 4578)) 2085 ((":" 150) (return_type 4579)) 2086 ((backup_fn_flags 4580)) 2087 (("{" 4581)) 2088 ((inner_statement_list 4582)) 2089 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 4583) ("~" 124)) 2090 ((backup_fn_flags 4584)) 2091 ((backup_doc_comment 4586)) 2092 (("(" 4587)) 2093 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 4588)) 2094 ((")" 4589)) 2095 ((":" 2207) (return_type 4590)) 2096 ((T_DOUBLE_ARROW 4591)) 2097 ((backup_fn_flags 4592)) 2098 ((backup_lex_pos 4593)) 2099 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4594) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2100 (("%" 4298) ("*" 4299) ("+" 4300) ("-" 4301) ("." 4302) ("/" 4303) ("<" 4304) (">" 4305) ("?" 4306) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4307) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4308) (T_BOOLEAN_AND 4309) (T_BOOLEAN_OR 4310) (T_COALESCE 4311) (T_INSTANCEOF 4312) (T_IS_EQUAL 4313) (T_IS_GREATER_OR_EQUAL 4314) (T_IS_IDENTICAL 4315) (T_IS_NOT_EQUAL 4316) (T_IS_NOT_IDENTICAL 4317) (T_IS_SMALLER_OR_EQUAL 4318) (T_LOGICAL_AND 4319) (T_LOGICAL_OR 4320) (T_LOGICAL_XOR 4321) (T_POW 4322) (T_SL 4323) (T_SPACESHIP 4324) (T_SR 4325) ("^" 4326) (backup_fn_flags 4595) ("|" 4327)) 2101 (("$" 4223) (T_ABSTRACT 4400) (T_ARRAY 4401) (T_AS 4402) (T_BREAK 4403) (T_CALLABLE 4404) (T_CASE 4405) (T_CATCH 4406) (T_CLASS 4407) (T_CLASS_C 4408) (T_CLONE 4409) (T_CONST 4410) (T_CONTINUE 4411) (T_DECLARE 4412) (T_DEFAULT 4413) (T_DIR 4414) (T_DO 4415) (T_ECHO 4416) (T_ELSE 4417) (T_ELSEIF 4418) (T_EMPTY 4419) (T_ENDDECLARE 4420) (T_ENDFOR 4421) (T_ENDFOREACH 4422) (T_ENDIF 4423) (T_ENDSWITCH 4424) (T_ENDWHILE 4425) (T_ENUM 4426) (T_EVAL 4427) (T_EXIT 4428) (T_EXTENDS 4429) (T_FILE 4430) (T_FINAL 4431) (T_FINALLY 4432) (T_FN 4433) (T_FOR 4434) (T_FOREACH 4435) (T_FUNCTION 4436) (T_FUNC_C 4437) (T_GLOBAL 4438) (T_GOTO 4439) (T_IF 4440) (T_IMPLEMENTS 4441) (T_INCLUDE 4442) (T_INCLUDE_ONCE 4443) (T_INSTANCEOF 4444) (T_INSTEADOF 4445) (T_INTERFACE 4446) (T_ISSET 4447) (T_LINE 4448) (T_LIST 4449) (T_LOGICAL_AND 4450) (T_LOGICAL_OR 4451) (T_LOGICAL_XOR 4452) (T_MATCH 4453) (T_METHOD_C 4454) (T_NAMESPACE 4455) (T_NEW 4456) (T_NS_C 4457) (T_PRINT 4458) (T_PRIVATE 4459) (T_PROTECTED 4460) (T_PUBLIC 4461) (T_READONLY 4462) (T_REQUIRE 4463) (T_REQUIRE_ONCE 4464) (T_RETURN 4465) (T_STATIC 4466) (T_STRING 4467) (T_SWITCH 4468) (T_THROW 4469) (T_TRAIT 4470) (T_TRAIT_C 4471) (T_TRY 4472) (T_UNSET 4473) (T_USE 4474) (T_VAR 4475) (T_VARIABLE 4269) (T_WHILE 4476) (T_YIELD 4477) (identifier 4597) (member_name 4598) (reserved_non_modifiers 4480) (semi_reserved 4481) (simple_variable 4599) ("{" 914)) 2102 (("(" 4483) (argument_list 4600)) 2103 ((T_FN 35) (T_FUNCTION 126) (fn 4283) #'4284 (inline_function 4604)) 2104 (("$" 4223) (T_STRING 4613) (T_VARIABLE 4269) (property_name 4620) (simple_variable 4615) ("{" 4616)) 2105 (("$" 4223) (T_STRING 4613) (T_VARIABLE 4269) (property_name 4614) (simple_variable 4615) ("{" 4616)) 2106 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 4611) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2107 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4609) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2108 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4610)) 2109 (("]" 4612)) 2110 (("(" 4483) (argument_list 4619)) 2111 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4617) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2112 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4618)) 2113 (("(" 4483) (argument_list 4621)) 2114 (("`" 4623)) 2115 (("]" 4625)) 2116 (("=" 4626)) 2117 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 4627) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2118 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5180) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2119 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 5157) (encaps_var 5158)) 2120 (("$" 4631) (T_VARIABLE 4677) (simple_variable 5149) ("{" 5150)) 2121 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5147) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2122 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5146) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2123 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5145) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2124 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5144) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2125 (("(" 5141)) 2126 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5140) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2127 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5139) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2128 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5138) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2129 (("\"" 1489) ("$" 4930) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 4931) ("[" 1508) (array_object_dereferenceable 4932) (callable_expr 4933) (callable_variable 4934) (class_constant 1512) (class_name 4935) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 4936) (name 4937) (simple_variable 4938) (static_member 4939) (variable 5137) (variable_class_name 4941)) 2130 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5136) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2131 (("(" 5133)) 2132 (("(" 5130)) 2133 (("(" 5125) (exit_expr 5126)) 2134 (("\"" 1489) ("$" 4930) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 4931) ("[" 1508) (array_object_dereferenceable 4932) (callable_expr 4933) (callable_variable 4934) (class_constant 1512) (class_name 4935) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 4936) (name 4937) (simple_variable 4938) (static_member 4939) (variable 5124) (variable_class_name 4941)) 2135 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5123) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2136 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5122) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2137 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5121) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2138 (("(" 5113)) 2139 (("(" 5108)) 2140 (("(" 5102)) 2141 (("$" 4930) ("(" 5055) (T_ATTRIBUTE 2853) (T_CLASS 5056) (T_NAME_FULLY_QUALIFIED 5057) (T_NAME_QUALIFIED 5058) (T_NAME_RELATIVE 5059) (T_STATIC 5060) (T_STRING 5061) (T_VARIABLE 4931) (anonymous_class 5062) (attribute 2862) (attributes 5063) (class_name 5064) (class_name_reference 5065) (name 5066) (new_variable 5067) (simple_variable 5068)) 2142 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5054) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2143 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5053) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2144 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5052) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2145 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5051) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2146 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 5046) (T_END_HEREDOC 5047) (T_VARIABLE 2822) (encaps_list 5048) (encaps_var 2824)) 2147 ((T_FN 35) (T_FUNCTION 126) (fn 4691) #'4692 (inline_function 5045)) 2148 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5044) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2149 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5043) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2150 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5042) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2151 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5039) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2152 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5038) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2153 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 5034) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2154 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 5032) (encaps_list 2729) (encaps_var 2730)) 2155 ((T_NULLSAFE_OBJECT_OPERATOR 5015) (T_OBJECT_OPERATOR 5016) ("[" 5017) ("{" 5018)) 2156 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 5012) (attribute 2654) (fn 4691) #'4692 (inline_function 5013)) 2157 (("(" 4891) (argument_list 5011)) 2158 ((T_PAAMAYIM_NEKUDOTAYIM 5006)) 2159 (("%" 4706) ("*" 4707) ("+" 4708) ("-" 4709) ("." 4710) ("/" 4711) ("<" 4712) (">" 4713) ("?" 4714) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4715) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4716) (T_BOOLEAN_AND 4717) (T_BOOLEAN_OR 4718) (T_COALESCE 4719) (T_DOUBLE_ARROW 5004) (T_INSTANCEOF 4720) (T_IS_EQUAL 4721) (T_IS_GREATER_OR_EQUAL 4722) (T_IS_IDENTICAL 4723) (T_IS_NOT_EQUAL 4724) (T_IS_NOT_IDENTICAL 4725) (T_IS_SMALLER_OR_EQUAL 4726) (T_LOGICAL_AND 4727) (T_LOGICAL_OR 4728) (T_LOGICAL_XOR 4729) (T_POW 4730) (T_SL 4731) (T_SPACESHIP 4732) (T_SR 4733) ("^" 4734) ("|" 4735)) 2160 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 4993)) 2161 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 4981)) 2162 (("(" 4891) (argument_list 4980)) 2163 (("=" 4899) (T_AND_EQUAL 4900) (T_COALESCE_EQUAL 4901) (T_CONCAT_EQUAL 4902) (T_DEC 4903) (T_DIV_EQUAL 4904) (T_INC 4905) (T_MINUS_EQUAL 4906) (T_MOD_EQUAL 4907) (T_MUL_EQUAL 4908) (T_OR_EQUAL 4909) (T_PLUS_EQUAL 4910) (T_POW_EQUAL 4911) (T_SL_EQUAL 4912) (T_SR_EQUAL 4913) (T_XOR_EQUAL 4914)) 2164 ((T_PAAMAYIM_NEKUDOTAYIM 4807)) 2165 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4705) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2166 (("%" 4706) ("*" 4707) ("+" 4708) ("-" 4709) ("." 4710) ("/" 4711) ("<" 4712) (">" 4713) ("?" 4714) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4715) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4716) (T_BOOLEAN_AND 4717) (T_BOOLEAN_OR 4718) (T_COALESCE 4719) (T_INSTANCEOF 4720) (T_IS_EQUAL 4721) (T_IS_GREATER_OR_EQUAL 4722) (T_IS_IDENTICAL 4723) (T_IS_NOT_EQUAL 4724) (T_IS_NOT_IDENTICAL 4725) (T_IS_SMALLER_OR_EQUAL 4726) (T_LOGICAL_AND 4727) (T_LOGICAL_OR 4728) (T_LOGICAL_XOR 4729) (T_POW 4730) (T_SL 4731) (T_SPACESHIP 4732) (T_SR 4733) ("^" 4734) ("|" 4735)) 2167 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4806) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2168 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4805) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2169 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4804) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2170 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4803) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2171 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4802) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2172 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4801) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2173 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4800) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2174 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4799) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2175 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 4794) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 4795) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 2176 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4793) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2177 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4792) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2178 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4791) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2179 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4790) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2180 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4789) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2181 (("$" 4751) ("(" 4752) (T_NAME_FULLY_QUALIFIED 4753) (T_NAME_QUALIFIED 4754) (T_NAME_RELATIVE 4755) (T_STATIC 4756) (T_STRING 4757) (T_VARIABLE 4758) (class_name 4759) (class_name_reference 4760) (name 4761) (new_variable 4762) (simple_variable 4763)) 2182 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4750) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2183 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4749) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2184 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4748) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2185 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4747) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2186 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4746) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2187 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4745) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2188 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4744) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2189 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4743) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2190 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4742) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2191 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4741) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2192 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4740) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2193 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4739) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2194 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4738) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2195 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4737) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2196 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4736) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2197 (("$" 4751) (T_VARIABLE 4758) (simple_variable 4785) ("{" 4786)) 2198 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 4783) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2199 ((T_PAAMAYIM_NEKUDOTAYIM 4781)) 2200 ((T_NULLSAFE_OBJECT_OPERATOR 4764) (T_OBJECT_OPERATOR 4765) (T_PAAMAYIM_NEKUDOTAYIM 4766) ("[" 4767) ("{" 4768)) 2201 (("$" 4751) (T_STRING 4774) (T_VARIABLE 4758) (property_name 4780) (simple_variable 4776) ("{" 4777)) 2202 (("$" 4751) (T_STRING 4774) (T_VARIABLE 4758) (property_name 4775) (simple_variable 4776) ("{" 4777)) 2203 (("$" 4751) (T_VARIABLE 4758) (simple_variable 4773)) 2204 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 4771) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2205 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4769) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2206 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4770)) 2207 (("]" 4772)) 2208 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4778) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2209 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4779)) 2210 (("$" 4751) (T_VARIABLE 4758) (simple_variable 4782)) 2211 (("%" 280) (")" 4784) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2212 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4787) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2213 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4788)) 2214 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4798) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2215 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 4796) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 2216 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4797) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2217 (("$" 4631) (T_ABSTRACT 4808) (T_ARRAY 4809) (T_AS 4810) (T_BREAK 4811) (T_CALLABLE 4812) (T_CASE 4813) (T_CATCH 4814) (T_CLASS 4815) (T_CLASS_C 4816) (T_CLONE 4817) (T_CONST 4818) (T_CONTINUE 4819) (T_DECLARE 4820) (T_DEFAULT 4821) (T_DIR 4822) (T_DO 4823) (T_ECHO 4824) (T_ELSE 4825) (T_ELSEIF 4826) (T_EMPTY 4827) (T_ENDDECLARE 4828) (T_ENDFOR 4829) (T_ENDFOREACH 4830) (T_ENDIF 4831) (T_ENDSWITCH 4832) (T_ENDWHILE 4833) (T_ENUM 4834) (T_EVAL 4835) (T_EXIT 4836) (T_EXTENDS 4837) (T_FILE 4838) (T_FINAL 4839) (T_FINALLY 4840) (T_FN 4841) (T_FOR 4842) (T_FOREACH 4843) (T_FUNCTION 4844) (T_FUNC_C 4845) (T_GLOBAL 4846) (T_GOTO 4847) (T_IF 4848) (T_IMPLEMENTS 4849) (T_INCLUDE 4850) (T_INCLUDE_ONCE 4851) (T_INSTANCEOF 4852) (T_INSTEADOF 4853) (T_INTERFACE 4854) (T_ISSET 4855) (T_LINE 4856) (T_LIST 4857) (T_LOGICAL_AND 4858) (T_LOGICAL_OR 4859) (T_LOGICAL_XOR 4860) (T_MATCH 4861) (T_METHOD_C 4862) (T_NAMESPACE 4863) (T_NEW 4864) (T_NS_C 4865) (T_PRINT 4866) (T_PRIVATE 4867) (T_PROTECTED 4868) (T_PUBLIC 4869) (T_READONLY 4870) (T_REQUIRE 4871) (T_REQUIRE_ONCE 4872) (T_RETURN 4873) (T_STATIC 4874) (T_STRING 4875) (T_SWITCH 4876) (T_THROW 4877) (T_TRAIT 4878) (T_TRAIT_C 4879) (T_TRY 4880) (T_UNSET 4881) (T_USE 4882) (T_VAR 4883) (T_VARIABLE 4677) (T_WHILE 4884) (T_YIELD 4885) (identifier 4886) (member_name 4887) (reserved_non_modifiers 4888) (semi_reserved 4889) (simple_variable 4890) ("{" 914)) 2218 (("(" 4891) (argument_list 4892)) 2219 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 4893) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 4894) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 4895) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2220 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 4898) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2221 (("," 1449) (possible_comma 4896)) 2222 ((")" 4897)) 2223 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (ampersand 4928) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4929) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2224 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4927) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2225 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4926) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2226 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4925) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2227 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4924) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2228 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4923) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2229 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4922) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2230 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4921) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2231 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4920) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2232 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4919) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2233 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4918) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2234 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4917) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2235 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4916) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2236 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 4915) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2237 (("\"" 1489) ("$" 4930) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 4931) ("[" 1508) (array_object_dereferenceable 4932) (callable_expr 4933) (callable_variable 4934) (class_constant 1512) (class_name 4935) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 4936) (name 4937) (simple_variable 4938) (static_member 4939) (variable 4940) (variable_class_name 4941)) 2238 (("$" 4930) (T_VARIABLE 4931) (simple_variable 4976) ("{" 4977)) 2239 ((T_NULLSAFE_OBJECT_OPERATOR 4959) (T_OBJECT_OPERATOR 4960) ("[" 4961) ("{" 4962)) 2240 (("(" 4945) (argument_list 4958)) 2241 ((T_PAAMAYIM_NEKUDOTAYIM 4954)) 2242 (("(" 4945) (argument_list 4953)) 2243 ((T_PAAMAYIM_NEKUDOTAYIM 4942)) 2244 (("$" 4930) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 4931) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 4943) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 4944) ("{" 914)) 2245 (("(" 4945) (argument_list 4946)) 2246 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 4947) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 4948) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 4949) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2247 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 4952) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2248 (("," 1449) (possible_comma 4950)) 2249 ((")" 4951)) 2250 (("$" 4930) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 4931) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 4955) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 4956) ("{" 914)) 2251 (("(" 4945) (argument_list 4957)) 2252 (("$" 4930) (T_STRING 4967) (T_VARIABLE 4931) (property_name 4974) (simple_variable 4969) ("{" 4970)) 2253 (("$" 4930) (T_STRING 4967) (T_VARIABLE 4931) (property_name 4968) (simple_variable 4969) ("{" 4970)) 2254 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 4965) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2255 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4963) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2256 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4964)) 2257 (("]" 4966)) 2258 (("(" 4945) (argument_list 4973)) 2259 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4971) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2260 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4972)) 2261 (("(" 4945) (argument_list 4975)) 2262 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 4978) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2263 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 4979)) 2264 ((backup_doc_comment 4982)) 2265 (("(" 4983)) 2266 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 4984)) 2267 ((")" 4985)) 2268 ((T_USE 148) (lexical_vars 4986)) 2269 ((":" 150) (return_type 4987)) 2270 ((backup_fn_flags 4988)) 2271 (("{" 4989)) 2272 ((inner_statement_list 4990)) 2273 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 4991) ("~" 124)) 2274 ((backup_fn_flags 4992)) 2275 ((backup_doc_comment 4994)) 2276 (("(" 4995)) 2277 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 4996)) 2278 ((")" 4997)) 2279 ((":" 2207) (return_type 4998)) 2280 ((T_DOUBLE_ARROW 4999)) 2281 ((backup_fn_flags 5000)) 2282 ((backup_lex_pos 5001)) 2283 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5002) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2284 (("%" 4706) ("*" 4707) ("+" 4708) ("-" 4709) ("." 4710) ("/" 4711) ("<" 4712) (">" 4713) ("?" 4714) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4715) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4716) (T_BOOLEAN_AND 4717) (T_BOOLEAN_OR 4718) (T_COALESCE 4719) (T_INSTANCEOF 4720) (T_IS_EQUAL 4721) (T_IS_GREATER_OR_EQUAL 4722) (T_IS_IDENTICAL 4723) (T_IS_NOT_EQUAL 4724) (T_IS_NOT_IDENTICAL 4725) (T_IS_SMALLER_OR_EQUAL 4726) (T_LOGICAL_AND 4727) (T_LOGICAL_OR 4728) (T_LOGICAL_XOR 4729) (T_POW 4730) (T_SL 4731) (T_SPACESHIP 4732) (T_SR 4733) ("^" 4734) (backup_fn_flags 5003) ("|" 4735)) 2285 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5005) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2286 (("$" 4631) (T_ABSTRACT 4808) (T_ARRAY 4809) (T_AS 4810) (T_BREAK 4811) (T_CALLABLE 4812) (T_CASE 4813) (T_CATCH 4814) (T_CLASS 4815) (T_CLASS_C 4816) (T_CLONE 4817) (T_CONST 4818) (T_CONTINUE 4819) (T_DECLARE 4820) (T_DEFAULT 4821) (T_DIR 4822) (T_DO 4823) (T_ECHO 4824) (T_ELSE 4825) (T_ELSEIF 4826) (T_EMPTY 4827) (T_ENDDECLARE 4828) (T_ENDFOR 4829) (T_ENDFOREACH 4830) (T_ENDIF 4831) (T_ENDSWITCH 4832) (T_ENDWHILE 4833) (T_ENUM 4834) (T_EVAL 4835) (T_EXIT 4836) (T_EXTENDS 4837) (T_FILE 4838) (T_FINAL 4839) (T_FINALLY 4840) (T_FN 4841) (T_FOR 4842) (T_FOREACH 4843) (T_FUNCTION 4844) (T_FUNC_C 4845) (T_GLOBAL 4846) (T_GOTO 4847) (T_IF 4848) (T_IMPLEMENTS 4849) (T_INCLUDE 4850) (T_INCLUDE_ONCE 4851) (T_INSTANCEOF 4852) (T_INSTEADOF 4853) (T_INTERFACE 4854) (T_ISSET 4855) (T_LINE 4856) (T_LIST 4857) (T_LOGICAL_AND 4858) (T_LOGICAL_OR 4859) (T_LOGICAL_XOR 4860) (T_MATCH 4861) (T_METHOD_C 4862) (T_NAMESPACE 4863) (T_NEW 4864) (T_NS_C 4865) (T_PRINT 4866) (T_PRIVATE 4867) (T_PROTECTED 4868) (T_PUBLIC 4869) (T_READONLY 4870) (T_REQUIRE 4871) (T_REQUIRE_ONCE 4872) (T_RETURN 4873) (T_STATIC 4874) (T_STRING 4875) (T_SWITCH 4876) (T_THROW 4877) (T_TRAIT 4878) (T_TRAIT_C 4879) (T_TRY 4880) (T_UNSET 4881) (T_USE 4882) (T_VAR 4883) (T_VARIABLE 4677) (T_WHILE 4884) (T_YIELD 4885) (identifier 5007) (member_name 5008) (reserved_non_modifiers 4888) (semi_reserved 4889) (simple_variable 5009) ("{" 914)) 2287 (("(" 4891) (argument_list 5010)) 2288 ((T_FN 35) (T_FUNCTION 126) (fn 4691) #'4692 (inline_function 5014)) 2289 (("$" 4631) (T_STRING 5023) (T_VARIABLE 4677) (property_name 5030) (simple_variable 5025) ("{" 5026)) 2290 (("$" 4631) (T_STRING 5023) (T_VARIABLE 4677) (property_name 5024) (simple_variable 5025) ("{" 5026)) 2291 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 5021) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2292 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5019) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2293 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5020)) 2294 (("]" 5022)) 2295 (("(" 4891) (argument_list 5029)) 2296 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5027) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2297 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5028)) 2298 (("(" 4891) (argument_list 5031)) 2299 (("`" 5033)) 2300 (("]" 5035)) 2301 (("=" 5036)) 2302 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5037) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2303 (("%" 4706) ("*" 4707) ("+" 4708) ("-" 4709) ("." 4710) ("/" 4711) ("<" 4712) (">" 4713) ("?" 4714) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4715) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4716) (T_BOOLEAN_AND 4717) (T_BOOLEAN_OR 4718) (T_COALESCE 4719) (T_DOUBLE_ARROW 5040) (T_INSTANCEOF 4720) (T_IS_EQUAL 4721) (T_IS_GREATER_OR_EQUAL 4722) (T_IS_IDENTICAL 4723) (T_IS_NOT_EQUAL 4724) (T_IS_NOT_IDENTICAL 4725) (T_IS_SMALLER_OR_EQUAL 4726) (T_LOGICAL_AND 4727) (T_LOGICAL_OR 4728) (T_LOGICAL_XOR 4729) (T_POW 4730) (T_SL 4731) (T_SPACESHIP 4732) (T_SR 4733) ("^" 4734) ("|" 4735)) 2304 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5041) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2305 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 5050) (T_VARIABLE 2822) (encaps_var 2836)) 2306 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 5049) (T_VARIABLE 2822) (encaps_var 2827)) 2307 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5100) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2308 (("(" 2899) (argument_list 2900) (ctor_arguments 5093)) 2309 ((T_ATTRIBUTE 2853) (T_CLASS 5056) (anonymous_class 5092) (attribute 2898)) 2310 ((T_PAAMAYIM_NEKUDOTAYIM 5090)) 2311 (("(" 5081) (argument_list 5082) (ctor_arguments 5083)) 2312 ((T_NULLSAFE_OBJECT_OPERATOR 5069) (T_OBJECT_OPERATOR 5070) (T_PAAMAYIM_NEKUDOTAYIM 5071) ("[" 5072) ("{" 5073)) 2313 (("$" 4930) (T_STRING 4967) (T_VARIABLE 4931) (property_name 5080) (simple_variable 4969) ("{" 4970)) 2314 (("$" 4930) (T_STRING 4967) (T_VARIABLE 4931) (property_name 5079) (simple_variable 4969) ("{" 4970)) 2315 (("$" 4930) (T_VARIABLE 4931) (simple_variable 5078)) 2316 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 5076) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2317 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5074) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2318 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5075)) 2319 (("]" 5077)) 2320 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 5084) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 5085) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 5086) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2321 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 5089) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2322 (("," 1449) (possible_comma 5087)) 2323 ((")" 5088)) 2324 (("$" 4930) (T_VARIABLE 4931) (simple_variable 5091)) 2325 ((T_EXTENDS 2902) (extends_from 5094)) 2326 ((T_IMPLEMENTS 2904) (implements_list 5095)) 2327 ((backup_doc_comment 5096)) 2328 (("{" 5097)) 2329 ((class_statement_list 5098)) 2330 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 5099)) 2331 (("%" 280) (")" 5101) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2332 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5103) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2333 (("%" 280) (")" 5104) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2334 (("{" 5105)) 2335 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 5106) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 2336 (("}" 5107)) 2337 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 5109) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2338 ((")" 5110)) 2339 (("=" 5111)) 2340 (("!" 4629) ("\"" 4630) ("$" 4631) ("(" 4632) ("+" 4633) ("-" 4634) ("@" 4635) (T_ARRAY 4636) (T_ARRAY_CAST 4637) (T_ATTRIBUTE 125) (T_BOOL_CAST 4638) (T_CLASS_C 4639) (T_CLONE 4640) (T_CONSTANT_ENCAPSED_STRING 4641) (T_DEC 4642) (T_DIR 4643) (T_DNUMBER 4644) (T_DOUBLE_CAST 4645) (T_EMPTY 4646) (T_EVAL 4647) (T_EXIT 4648) (T_FILE 4649) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4650) (T_INC 4651) (T_INCLUDE 4652) (T_INCLUDE_ONCE 4653) (T_INT_CAST 4654) (T_ISSET 4655) (T_LINE 4656) (T_LIST 4657) (T_LNUMBER 4658) (T_MATCH 4659) (T_METHOD_C 4660) (T_NAME_FULLY_QUALIFIED 4661) (T_NAME_QUALIFIED 4662) (T_NAME_RELATIVE 4663) (T_NEW 4664) (T_NS_C 4665) (T_OBJECT_CAST 4666) (T_PRINT 4667) (T_REQUIRE 4668) (T_REQUIRE_ONCE 4669) (T_START_HEREDOC 4670) (T_STATIC 4671) (T_STRING 4672) (T_STRING_CAST 4673) (T_THROW 4674) (T_TRAIT_C 4675) (T_UNSET_CAST 4676) (T_VARIABLE 4677) (T_YIELD 4678) (T_YIELD_FROM 4679) ("[" 4680) ("`" 4681) (array_object_dereferenceable 4682) (attribute 129) (attributes 4683) (callable_expr 4684) (callable_variable 4685) (class_constant 4686) (class_name 4687) (constant 4688) (dereferenceable_scalar 4689) (expr 5112) (fn 4691) (fully_dereferenceable 103) #'4692 (function_call 4693) (inline_function 4694) (internal_functions_in_yacc 4695) (match 4696) (name 4697) (new_expr 4698) (scalar 4699) (simple_variable 4700) (static_member 4701) (variable 4702) (variable_class_name 4703) ("~" 4704)) 2341 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 5116) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2342 (("," 5117) (possible_comma 5118)) 2343 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5120) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2344 ((")" 5119)) 2345 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 5128) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2346 ((")" 5129)) 2347 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5131) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2348 (("%" 280) (")" 5132) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2349 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5134) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2350 (("%" 280) (")" 5135) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2351 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 5142) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2352 ((")" 5143)) 2353 (("%" 280) (")" 5148) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2354 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5151) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2355 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5152)) 2356 (("\"" 1489) ("$" 2755) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2756) ("[" 1508) (array_object_dereferenceable 2757) (callable_expr 2758) (callable_variable 2759) (class_constant 1512) (class_name 2760) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 2761) (name 2762) (simple_variable 2763) (static_member 2764) (variable 5178) (variable_class_name 2766)) 2357 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_STRING_VARNAME 5170) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5171) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2358 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_VARIABLE 5156) (encaps_var 5169)) 2359 ((T_NULLSAFE_OBJECT_OPERATOR 5162) (T_OBJECT_OPERATOR 5163) ("[" 5164)) 2360 (("\"" 5159) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 2361 ((T_STRING 5168)) 2362 ((T_STRING 5167)) 2363 (("-" 2737) (T_NUM_STRING 2738) (T_STRING 2739) (T_VARIABLE 2740) (encaps_var_offset 5165)) 2364 (("]" 5166)) 2365 (("[" 5173) ("}" 5174)) 2366 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5172)) 2367 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 5175) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2368 (("%" 563) ("*" 564) ("+" 565) ("-" 566) ("." 567) ("/" 568) ("<" 569) (">" 570) ("?" 571) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 572) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 573) (T_BOOLEAN_AND 574) (T_BOOLEAN_OR 575) (T_COALESCE 576) (T_INSTANCEOF 577) (T_IS_EQUAL 578) (T_IS_GREATER_OR_EQUAL 579) (T_IS_IDENTICAL 580) (T_IS_NOT_EQUAL 581) (T_IS_NOT_IDENTICAL 582) (T_IS_SMALLER_OR_EQUAL 583) (T_LOGICAL_AND 584) (T_LOGICAL_OR 585) (T_LOGICAL_XOR 586) (T_POW 587) (T_SL 588) (T_SPACESHIP 589) (T_SR 590) ("]" 5176) ("^" 591) ("|" 592)) 2369 (("}" 5177)) 2370 (("}" 5179)) 2371 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 5189) (T_VARIABLE 2822) (encaps_var 2836)) 2372 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 5188) (T_VARIABLE 2822) (encaps_var 2827)) 2373 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5239) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2374 (("(" 2899) (argument_list 2900) (ctor_arguments 5232)) 2375 ((T_ATTRIBUTE 2853) (T_CLASS 5195) (anonymous_class 5231) (attribute 2898)) 2376 ((T_PAAMAYIM_NEKUDOTAYIM 5229)) 2377 (("(" 5220) (argument_list 5221) (ctor_arguments 5222)) 2378 ((T_NULLSAFE_OBJECT_OPERATOR 5208) (T_OBJECT_OPERATOR 5209) (T_PAAMAYIM_NEKUDOTAYIM 5210) ("[" 5211) ("{" 5212)) 2379 (("$" 4522) (T_STRING 4559) (T_VARIABLE 4523) (property_name 5219) (simple_variable 4561) ("{" 4562)) 2380 (("$" 4522) (T_STRING 4559) (T_VARIABLE 4523) (property_name 5218) (simple_variable 4561) ("{" 4562)) 2381 (("$" 4522) (T_VARIABLE 4523) (simple_variable 5217)) 2382 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 5215) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2383 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5213) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2384 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5214)) 2385 (("]" 5216)) 2386 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 5223) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 5224) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 5225) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2387 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 5228) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2388 (("," 1449) (possible_comma 5226)) 2389 ((")" 5227)) 2390 (("$" 4522) (T_VARIABLE 4523) (simple_variable 5230)) 2391 ((T_EXTENDS 2902) (extends_from 5233)) 2392 ((T_IMPLEMENTS 2904) (implements_list 5234)) 2393 ((backup_doc_comment 5235)) 2394 (("{" 5236)) 2395 ((class_statement_list 5237)) 2396 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 5238)) 2397 (("%" 280) (")" 5240) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2398 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5242) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2399 (("%" 280) (")" 5243) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2400 (("{" 5244)) 2401 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 5245) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 2402 (("}" 5246)) 2403 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 5248) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2404 ((")" 5249)) 2405 (("=" 5250)) 2406 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5251) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2407 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 5253) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2408 (("," 5117) (possible_comma 5254)) 2409 ((")" 5255)) 2410 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 5262) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2411 ((")" 5263)) 2412 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5265) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2413 (("%" 280) (")" 5266) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2414 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5268) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2415 (("%" 280) (")" 5269) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2416 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 5276) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2417 ((")" 5277)) 2418 (("%" 280) (")" 5282) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2419 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5285) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2420 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5286)) 2421 (("\"" 5288) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 2422 ((backup_doc_comment 5292)) 2423 (("(" 5293)) 2424 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 5294)) 2425 ((")" 5295)) 2426 ((T_USE 148) (lexical_vars 5296)) 2427 ((":" 150) (return_type 5297)) 2428 ((backup_fn_flags 5298)) 2429 (("{" 5299)) 2430 ((inner_statement_list 5300)) 2431 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 5301) ("~" 124)) 2432 ((backup_fn_flags 5302)) 2433 ((backup_doc_comment 5304)) 2434 (("(" 5305)) 2435 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 5306)) 2436 ((")" 5307)) 2437 ((":" 2207) (return_type 5308)) 2438 ((T_DOUBLE_ARROW 5309)) 2439 ((backup_fn_flags 5310)) 2440 ((backup_lex_pos 5311)) 2441 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5312) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 2442 (("%" 3939) ("*" 3940) ("+" 3941) ("-" 3942) ("." 3943) ("/" 3944) ("<" 3945) (">" 3946) ("?" 3947) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3948) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3949) (T_BOOLEAN_AND 3950) (T_BOOLEAN_OR 3951) (T_COALESCE 3952) (T_INSTANCEOF 3953) (T_IS_EQUAL 3954) (T_IS_GREATER_OR_EQUAL 3955) (T_IS_IDENTICAL 3956) (T_IS_NOT_EQUAL 3957) (T_IS_NOT_IDENTICAL 3958) (T_IS_SMALLER_OR_EQUAL 3959) (T_LOGICAL_AND 3960) (T_LOGICAL_OR 3961) (T_LOGICAL_XOR 3962) (T_POW 3963) (T_SL 3964) (T_SPACESHIP 3965) (T_SR 3966) ("^" 3967) (backup_fn_flags 5313) ("|" 3968)) 2443 (("$" 3859) (T_ABSTRACT 4041) (T_ARRAY 4042) (T_AS 4043) (T_BREAK 4044) (T_CALLABLE 4045) (T_CASE 4046) (T_CATCH 4047) (T_CLASS 4048) (T_CLASS_C 4049) (T_CLONE 4050) (T_CONST 4051) (T_CONTINUE 4052) (T_DECLARE 4053) (T_DEFAULT 4054) (T_DIR 4055) (T_DO 4056) (T_ECHO 4057) (T_ELSE 4058) (T_ELSEIF 4059) (T_EMPTY 4060) (T_ENDDECLARE 4061) (T_ENDFOR 4062) (T_ENDFOREACH 4063) (T_ENDIF 4064) (T_ENDSWITCH 4065) (T_ENDWHILE 4066) (T_ENUM 4067) (T_EVAL 4068) (T_EXIT 4069) (T_EXTENDS 4070) (T_FILE 4071) (T_FINAL 4072) (T_FINALLY 4073) (T_FN 4074) (T_FOR 4075) (T_FOREACH 4076) (T_FUNCTION 4077) (T_FUNC_C 4078) (T_GLOBAL 4079) (T_GOTO 4080) (T_IF 4081) (T_IMPLEMENTS 4082) (T_INCLUDE 4083) (T_INCLUDE_ONCE 4084) (T_INSTANCEOF 4085) (T_INSTEADOF 4086) (T_INTERFACE 4087) (T_ISSET 4088) (T_LINE 4089) (T_LIST 4090) (T_LOGICAL_AND 4091) (T_LOGICAL_OR 4092) (T_LOGICAL_XOR 4093) (T_MATCH 4094) (T_METHOD_C 4095) (T_NAMESPACE 4096) (T_NEW 4097) (T_NS_C 4098) (T_PRINT 4099) (T_PRIVATE 4100) (T_PROTECTED 4101) (T_PUBLIC 4102) (T_READONLY 4103) (T_REQUIRE 4104) (T_REQUIRE_ONCE 4105) (T_RETURN 4106) (T_STATIC 4107) (T_STRING 4108) (T_SWITCH 4109) (T_THROW 4110) (T_TRAIT 4111) (T_TRAIT_C 4112) (T_TRY 4113) (T_UNSET 4114) (T_USE 4115) (T_VAR 4116) (T_VARIABLE 3906) (T_WHILE 4117) (T_YIELD 4118) (identifier 5315) (member_name 5316) (reserved_non_modifiers 4121) (semi_reserved 4122) (simple_variable 5317) ("{" 914)) 2444 (("(" 4124) (argument_list 5318)) 2445 ((T_FN 35) (T_FUNCTION 126) (fn 3920) #'3921 (inline_function 5322)) 2446 (("$" 3859) (T_STRING 5331) (T_VARIABLE 3906) (property_name 5338) (simple_variable 5333) ("{" 5334)) 2447 (("$" 3859) (T_STRING 5331) (T_VARIABLE 3906) (property_name 5332) (simple_variable 5333) ("{" 5334)) 2448 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 5329) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2449 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5327) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2450 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5328)) 2451 (("]" 5330)) 2452 (("(" 4124) (argument_list 5337)) 2453 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5335) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2454 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5336)) 2455 (("(" 4124) (argument_list 5339)) 2456 (("`" 5341)) 2457 (("]" 5343)) 2458 (("=" 5344)) 2459 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5345) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 2460 (("%" 3939) ("*" 3940) ("+" 3941) ("-" 3942) ("." 3943) ("/" 3944) ("<" 3945) (">" 3946) ("?" 3947) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3948) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3949) (T_BOOLEAN_AND 3950) (T_BOOLEAN_OR 3951) (T_COALESCE 3952) (T_DOUBLE_ARROW 5348) (T_INSTANCEOF 3953) (T_IS_EQUAL 3954) (T_IS_GREATER_OR_EQUAL 3955) (T_IS_IDENTICAL 3956) (T_IS_NOT_EQUAL 3957) (T_IS_NOT_IDENTICAL 3958) (T_IS_SMALLER_OR_EQUAL 3959) (T_LOGICAL_AND 3960) (T_LOGICAL_OR 3961) (T_LOGICAL_XOR 3962) (T_POW 3963) (T_SL 3964) (T_SPACESHIP 3965) (T_SR 3966) ("^" 3967) ("|" 3968)) 2461 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5349) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 2462 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 5358) (T_VARIABLE 2822) (encaps_var 2836)) 2463 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 5357) (T_VARIABLE 2822) (encaps_var 2827)) 2464 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5408) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2465 (("(" 2899) (argument_list 2900) (ctor_arguments 5401)) 2466 ((T_ATTRIBUTE 2853) (T_CLASS 5364) (anonymous_class 5400) (attribute 2898)) 2467 ((T_PAAMAYIM_NEKUDOTAYIM 5398)) 2468 (("(" 5389) (argument_list 5390) (ctor_arguments 5391)) 2469 ((T_NULLSAFE_OBJECT_OPERATOR 5377) (T_OBJECT_OPERATOR 5378) (T_PAAMAYIM_NEKUDOTAYIM 5379) ("[" 5380) ("{" 5381)) 2470 (("$" 4163) (T_STRING 4200) (T_VARIABLE 4164) (property_name 5388) (simple_variable 4202) ("{" 4203)) 2471 (("$" 4163) (T_STRING 4200) (T_VARIABLE 4164) (property_name 5387) (simple_variable 4202) ("{" 4203)) 2472 (("$" 4163) (T_VARIABLE 4164) (simple_variable 5386)) 2473 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 5384) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2474 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5382) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2475 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5383)) 2476 (("]" 5385)) 2477 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 5392) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 5393) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 5394) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2478 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 5397) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2479 (("," 1449) (possible_comma 5395)) 2480 ((")" 5396)) 2481 (("$" 4163) (T_VARIABLE 4164) (simple_variable 5399)) 2482 ((T_EXTENDS 2902) (extends_from 5402)) 2483 ((T_IMPLEMENTS 2904) (implements_list 5403)) 2484 ((backup_doc_comment 5404)) 2485 (("{" 5405)) 2486 ((class_statement_list 5406)) 2487 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 5407)) 2488 (("%" 280) (")" 5409) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2489 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5411) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2490 (("%" 280) (")" 5412) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2491 (("{" 5413)) 2492 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 5414) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 2493 (("}" 5415)) 2494 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 5417) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2495 ((")" 5418)) 2496 (("=" 5419)) 2497 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 5420) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (name 3929) (new_expr 3930) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 2498 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 5422) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2499 (("," 5117) (possible_comma 5423)) 2500 ((")" 5424)) 2501 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 5431) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2502 ((")" 5432)) 2503 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5434) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2504 (("%" 280) (")" 5435) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2505 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5437) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2506 (("%" 280) (")" 5438) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2507 ((T_DOUBLE_ARROW 5442)) 2508 (("!" 4221) ("\"" 4222) ("$" 4223) ("(" 4224) ("+" 4225) ("-" 4226) ("@" 4227) (T_ARRAY 4228) (T_ARRAY_CAST 4229) (T_ATTRIBUTE 125) (T_BOOL_CAST 4230) (T_CLASS_C 4231) (T_CLONE 4232) (T_CONSTANT_ENCAPSED_STRING 4233) (T_DEC 4234) (T_DIR 4235) (T_DNUMBER 4236) (T_DOUBLE_CAST 4237) (T_EMPTY 4238) (T_EVAL 4239) (T_EXIT 4240) (T_FILE 4241) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 4242) (T_INC 4243) (T_INCLUDE 4244) (T_INCLUDE_ONCE 4245) (T_INT_CAST 4246) (T_ISSET 4247) (T_LINE 4248) (T_LIST 4249) (T_LNUMBER 4250) (T_MATCH 4251) (T_METHOD_C 4252) (T_NAME_FULLY_QUALIFIED 4253) (T_NAME_QUALIFIED 4254) (T_NAME_RELATIVE 4255) (T_NEW 4256) (T_NS_C 4257) (T_OBJECT_CAST 4258) (T_PRINT 4259) (T_REQUIRE 4260) (T_REQUIRE_ONCE 4261) (T_START_HEREDOC 4262) (T_STATIC 4263) (T_STRING 4264) (T_STRING_CAST 4265) (T_THROW 4266) (T_TRAIT_C 4267) (T_UNSET_CAST 4268) (T_VARIABLE 4269) (T_YIELD 4270) (T_YIELD_FROM 4271) ("[" 4272) ("`" 4273) (array_object_dereferenceable 4274) (attribute 129) (attributes 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 4278) (class_name 4279) (constant 4280) (dereferenceable_scalar 4281) (expr 5443) (fn 4283) (fully_dereferenceable 103) #'4284 (function_call 4285) (inline_function 4286) (internal_functions_in_yacc 4287) (match 4288) (name 4289) (new_expr 4290) (scalar 4291) (simple_variable 4292) (static_member 4293) (variable 4294) (variable_class_name 4295) ("~" 4296)) 2509 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 5449) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2510 ((")" 5450)) 2511 (("%" 280) (")" 5455) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2512 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5458) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2513 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5459)) 2514 (("\"" 5461) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 2515 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 5464) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2516 ((")" 5465)) 2517 (("=" 5466)) 2518 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 5467) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 2519 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 5469) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2520 (("," 5117) (possible_comma 5470)) 2521 ((")" 5471)) 2522 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 5478) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2523 ((")" 5479)) 2524 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5481) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2525 (("%" 280) (")" 5482) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2526 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5484) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2527 (("%" 280) (")" 5485) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2528 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 5492) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2529 ((")" 5493)) 2530 (("%" 280) (")" 5498) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2531 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5501) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2532 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5502)) 2533 (("\"" 5504) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 2534 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 5514) (T_VARIABLE 2822) (encaps_var 2836)) 2535 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 5513) (T_VARIABLE 2822) (encaps_var 2827)) 2536 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5564) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2537 (("(" 2899) (argument_list 2900) (ctor_arguments 5557)) 2538 ((T_ATTRIBUTE 2853) (T_CLASS 5520) (anonymous_class 5556) (attribute 2898)) 2539 ((T_PAAMAYIM_NEKUDOTAYIM 5554)) 2540 (("(" 5545) (argument_list 5546) (ctor_arguments 5547)) 2541 ((T_NULLSAFE_OBJECT_OPERATOR 5533) (T_OBJECT_OPERATOR 5534) (T_PAAMAYIM_NEKUDOTAYIM 5535) ("[" 5536) ("{" 5537)) 2542 (("$" 3272) (T_STRING 3309) (T_VARIABLE 3273) (property_name 5544) (simple_variable 3311) ("{" 3312)) 2543 (("$" 3272) (T_STRING 3309) (T_VARIABLE 3273) (property_name 5543) (simple_variable 3311) ("{" 3312)) 2544 (("$" 3272) (T_VARIABLE 3273) (simple_variable 5542)) 2545 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 5540) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2546 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5538) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2547 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5539)) 2548 (("]" 5541)) 2549 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 5548) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 5549) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 5550) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2550 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 5553) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2551 (("," 1449) (possible_comma 5551)) 2552 ((")" 5552)) 2553 (("$" 3272) (T_VARIABLE 3273) (simple_variable 5555)) 2554 ((T_EXTENDS 2902) (extends_from 5558)) 2555 ((T_IMPLEMENTS 2904) (implements_list 5559)) 2556 ((backup_doc_comment 5560)) 2557 (("{" 5561)) 2558 ((class_statement_list 5562)) 2559 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 5563)) 2560 (("%" 280) (")" 5565) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2561 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5567) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2562 (("%" 280) (")" 5568) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2563 (("{" 5569)) 2564 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 5570) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 2565 (("}" 5571)) 2566 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 5573) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2567 ((")" 5574)) 2568 (("=" 5575)) 2569 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5576) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 2570 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 5578) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2571 (("," 5117) (possible_comma 5579)) 2572 ((")" 5580)) 2573 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 5587) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2574 ((")" 5588)) 2575 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5590) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2576 (("%" 280) (")" 5591) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2577 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5593) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2578 (("%" 280) (")" 5594) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2579 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 5601) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2580 ((")" 5602)) 2581 (("%" 280) (")" 5607) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2582 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 5610) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2583 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 5611)) 2584 (("\"" 5613) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 2585 ((T_ARRAY 2955) (T_CALLABLE 2956) (T_NAME_FULLY_QUALIFIED 2957) (T_NAME_QUALIFIED 2958) (T_NAME_RELATIVE 2959) (T_STRING 2960) (name 2961) (type_without_static 5616)) 2586 ((T_ABSTRACT 5777) (T_ARRAY 5778) (T_AS 5779) (T_BREAK 5780) (T_CALLABLE 5781) (T_CASE 5782) (T_CATCH 5783) (T_CLASS 5784) (T_CLASS_C 5785) (T_CLONE 5786) (T_CONST 5787) (T_CONTINUE 5788) (T_DECLARE 5789) (T_DEFAULT 5790) (T_DIR 5791) (T_DO 5792) (T_ECHO 5793) (T_ELSE 5794) (T_ELSEIF 5795) (T_EMPTY 5796) (T_ENDDECLARE 5797) (T_ENDFOR 5798) (T_ENDFOREACH 5799) (T_ENDIF 5800) (T_ENDSWITCH 5801) (T_ENDWHILE 5802) (T_ENUM 5803) (T_EVAL 5804) (T_EXIT 5805) (T_EXTENDS 5806) (T_FILE 5807) (T_FINAL 5808) (T_FINALLY 5809) (T_FN 5810) (T_FOR 5811) (T_FOREACH 5812) (T_FUNCTION 5813) (T_FUNC_C 5814) (T_GLOBAL 5815) (T_GOTO 5816) (T_IF 5817) (T_IMPLEMENTS 5818) (T_INCLUDE 5819) (T_INCLUDE_ONCE 5820) (T_INSTANCEOF 5821) (T_INSTEADOF 5822) (T_INTERFACE 5823) (T_ISSET 5824) (T_LINE 5825) (T_LIST 5826) (T_LOGICAL_AND 5827) (T_LOGICAL_OR 5828) (T_LOGICAL_XOR 5829) (T_MATCH 5830) (T_METHOD_C 5831) (T_NAMESPACE 5832) (T_NEW 5833) (T_NS_C 5834) (T_PRINT 5835) (T_PRIVATE 5836) (T_PROTECTED 5837) (T_PUBLIC 5838) (T_READONLY 5839) (T_REQUIRE 5840) (T_REQUIRE_ONCE 5841) (T_RETURN 5842) (T_STATIC 5843) (T_STRING 5844) (T_SWITCH 5845) (T_THROW 5846) (T_TRAIT 5847) (T_TRAIT_C 5848) (T_TRY 5849) (T_UNSET 5850) (T_USE 5851) (T_VAR 5852) (T_WHILE 5853) (T_YIELD 5854) (class_const_decl 5855) (class_const_list 5856) (identifier 5857) (reserved_non_modifiers 5858) (semi_reserved 5859)) 2587 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5629) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5630) (ampersand 5631) (returns_ref 5632)) 2588 ((T_ABSTRACT 5633) (T_ARRAY 5634) (T_AS 5635) (T_BREAK 5636) (T_CALLABLE 5637) (T_CASE 5638) (T_CATCH 5639) (T_CLASS 5640) (T_CLASS_C 5641) (T_CLONE 5642) (T_CONST 5643) (T_CONTINUE 5644) (T_DECLARE 5645) (T_DEFAULT 5646) (T_DIR 5647) (T_DO 5648) (T_ECHO 5649) (T_ELSE 5650) (T_ELSEIF 5651) (T_EMPTY 5652) (T_ENDDECLARE 5653) (T_ENDFOR 5654) (T_ENDFOREACH 5655) (T_ENDIF 5656) (T_ENDSWITCH 5657) (T_ENDWHILE 5658) (T_ENUM 5659) (T_EVAL 5660) (T_EXIT 5661) (T_EXTENDS 5662) (T_FILE 5663) (T_FINAL 5664) (T_FINALLY 5665) (T_FN 5666) (T_FOR 5667) (T_FOREACH 5668) (T_FUNCTION 5669) (T_FUNC_C 5670) (T_GLOBAL 5671) (T_GOTO 5672) (T_IF 5673) (T_IMPLEMENTS 5674) (T_INCLUDE 5675) (T_INCLUDE_ONCE 5676) (T_INSTANCEOF 5677) (T_INSTEADOF 5678) (T_INTERFACE 5679) (T_ISSET 5680) (T_LINE 5681) (T_LIST 5682) (T_LOGICAL_AND 5683) (T_LOGICAL_OR 5684) (T_LOGICAL_XOR 5685) (T_MATCH 5686) (T_METHOD_C 5687) (T_NAMESPACE 5688) (T_NEW 5689) (T_NS_C 5690) (T_PRINT 5691) (T_PRIVATE 5692) (T_PROTECTED 5693) (T_PUBLIC 5694) (T_READONLY 5695) (T_REQUIRE 5696) (T_REQUIRE_ONCE 5697) (T_RETURN 5698) (T_STATIC 5699) (T_STRING 5700) (T_SWITCH 5701) (T_THROW 5702) (T_TRAIT 5703) (T_TRAIT_C 5704) (T_TRY 5705) (T_UNSET 5706) (T_USE 5707) (T_VAR 5708) (T_WHILE 5709) (T_YIELD 5710) (identifier 5711) (reserved_non_modifiers 5712) (semi_reserved 5713)) 2589 ((backup_doc_comment 5714)) 2590 (("(" 5715)) 2591 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 5716)) 2592 ((")" 5717)) 2593 ((":" 5718) (return_type 5719)) 2594 (("?" 5727) (T_ARRAY 5728) (T_CALLABLE 5729) (T_NAME_FULLY_QUALIFIED 5730) (T_NAME_QUALIFIED 5731) (T_NAME_RELATIVE 5732) (T_STATIC 5733) (T_STRING 5734) (intersection_type 5735) (name 5736) (type 5737) (type_expr 5738) (type_without_static 5739) (union_type 5740)) 2595 ((backup_fn_flags 5720)) 2596 ((";" 5721) (method_body 5722) ("{" 5723)) 2597 ((backup_fn_flags 5726)) 2598 ((inner_statement_list 5724)) 2599 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 5725) ("~" 124)) 2600 ((T_ARRAY 5767) (T_CALLABLE 5768) (T_NAME_FULLY_QUALIFIED 5769) (T_NAME_QUALIFIED 5770) (T_NAME_RELATIVE 5771) (T_STATIC 5772) (T_STRING 5773) (name 5774) (type 5775) (type_without_static 5776)) 2601 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5765)) 2602 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5752) ("|" 5753)) 2603 (("|" 5741)) 2604 ((T_ARRAY 5742) (T_CALLABLE 5743) (T_NAME_FULLY_QUALIFIED 5744) (T_NAME_QUALIFIED 5745) (T_NAME_RELATIVE 5746) (T_STATIC 5747) (T_STRING 5748) (name 5749) (type 5750) (type_without_static 5751)) 2605 ((T_ARRAY 5755) (T_CALLABLE 5756) (T_NAME_FULLY_QUALIFIED 5757) (T_NAME_QUALIFIED 5758) (T_NAME_RELATIVE 5759) (T_STATIC 5760) (T_STRING 5761) (name 5762) (type 5763) (type_without_static 5764)) 2606 ((T_ARRAY 5742) (T_CALLABLE 5743) (T_NAME_FULLY_QUALIFIED 5744) (T_NAME_QUALIFIED 5745) (T_NAME_RELATIVE 5746) (T_STATIC 5747) (T_STRING 5748) (name 5749) (type 5754) (type_without_static 5751)) 2607 ((T_ARRAY 5755) (T_CALLABLE 5756) (T_NAME_FULLY_QUALIFIED 5757) (T_NAME_QUALIFIED 5758) (T_NAME_RELATIVE 5759) (T_STATIC 5760) (T_STRING 5761) (name 5762) (type 5766) (type_without_static 5764)) 2608 (("," 5863) (";" 5864)) 2609 (("=" 5860)) 2610 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 5861) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 2611 (("%" 3048) ("*" 3049) ("+" 3050) ("-" 3051) ("." 3052) ("/" 3053) ("<" 3054) (">" 3055) ("?" 3056) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3057) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3058) (T_BOOLEAN_AND 3059) (T_BOOLEAN_OR 3060) (T_COALESCE 3061) (T_INSTANCEOF 3062) (T_IS_EQUAL 3063) (T_IS_GREATER_OR_EQUAL 3064) (T_IS_IDENTICAL 3065) (T_IS_NOT_EQUAL 3066) (T_IS_NOT_IDENTICAL 3067) (T_IS_SMALLER_OR_EQUAL 3068) (T_LOGICAL_AND 3069) (T_LOGICAL_OR 3070) (T_LOGICAL_XOR 3071) (T_POW 3072) (T_SL 3073) (T_SPACESHIP 3074) (T_SR 3075) ("^" 3076) (backup_doc_comment 5862) ("|" 3077)) 2612 ((T_ABSTRACT 5777) (T_ARRAY 5778) (T_AS 5779) (T_BREAK 5780) (T_CALLABLE 5781) (T_CASE 5782) (T_CATCH 5783) (T_CLASS 5784) (T_CLASS_C 5785) (T_CLONE 5786) (T_CONST 5787) (T_CONTINUE 5788) (T_DECLARE 5789) (T_DEFAULT 5790) (T_DIR 5791) (T_DO 5792) (T_ECHO 5793) (T_ELSE 5794) (T_ELSEIF 5795) (T_EMPTY 5796) (T_ENDDECLARE 5797) (T_ENDFOR 5798) (T_ENDFOREACH 5799) (T_ENDIF 5800) (T_ENDSWITCH 5801) (T_ENDWHILE 5802) (T_ENUM 5803) (T_EVAL 5804) (T_EXIT 5805) (T_EXTENDS 5806) (T_FILE 5807) (T_FINAL 5808) (T_FINALLY 5809) (T_FN 5810) (T_FOR 5811) (T_FOREACH 5812) (T_FUNCTION 5813) (T_FUNC_C 5814) (T_GLOBAL 5815) (T_GOTO 5816) (T_IF 5817) (T_IMPLEMENTS 5818) (T_INCLUDE 5819) (T_INCLUDE_ONCE 5820) (T_INSTANCEOF 5821) (T_INSTEADOF 5822) (T_INTERFACE 5823) (T_ISSET 5824) (T_LINE 5825) (T_LIST 5826) (T_LOGICAL_AND 5827) (T_LOGICAL_OR 5828) (T_LOGICAL_XOR 5829) (T_MATCH 5830) (T_METHOD_C 5831) (T_NAMESPACE 5832) (T_NEW 5833) (T_NS_C 5834) (T_PRINT 5835) (T_PRIVATE 5836) (T_PROTECTED 5837) (T_PUBLIC 5838) (T_READONLY 5839) (T_REQUIRE 5840) (T_REQUIRE_ONCE 5841) (T_RETURN 5842) (T_STATIC 5843) (T_STRING 5844) (T_SWITCH 5845) (T_THROW 5846) (T_TRAIT 5847) (T_TRAIT_C 5848) (T_TRY 5849) (T_UNSET 5850) (T_USE 5851) (T_VAR 5852) (T_WHILE 5853) (T_YIELD 5854) (class_const_decl 5865) (identifier 5857) (reserved_non_modifiers 5858) (semi_reserved 5859)) 2613 (("," 5876) (";" 5877) (trait_adaptations 5878) ("{" 5879)) 2614 ((T_NAME_FULLY_QUALIFIED 5868) (T_NAME_QUALIFIED 5869) (T_NAME_RELATIVE 5870) (T_STATIC 5871) (T_STRING 5872) (class_name 6162) (name 5875)) 2615 ((T_ABSTRACT 5880) (T_ARRAY 5881) (T_AS 5882) (T_BREAK 5883) (T_CALLABLE 5884) (T_CASE 5885) (T_CATCH 5886) (T_CLASS 5887) (T_CLASS_C 5888) (T_CLONE 5889) (T_CONST 5890) (T_CONTINUE 5891) (T_DECLARE 5892) (T_DEFAULT 5893) (T_DIR 5894) (T_DO 5895) (T_ECHO 5896) (T_ELSE 5897) (T_ELSEIF 5898) (T_EMPTY 5899) (T_ENDDECLARE 5900) (T_ENDFOR 5901) (T_ENDFOREACH 5902) (T_ENDIF 5903) (T_ENDSWITCH 5904) (T_ENDWHILE 5905) (T_ENUM 5906) (T_EVAL 5907) (T_EXIT 5908) (T_EXTENDS 5909) (T_FILE 5910) (T_FINAL 5911) (T_FINALLY 5912) (T_FN 5913) (T_FOR 5914) (T_FOREACH 5915) (T_FUNCTION 5916) (T_FUNC_C 5917) (T_GLOBAL 5918) (T_GOTO 5919) (T_IF 5920) (T_IMPLEMENTS 5921) (T_INCLUDE 5922) (T_INCLUDE_ONCE 5923) (T_INSTANCEOF 5924) (T_INSTEADOF 5925) (T_INTERFACE 5926) (T_ISSET 5927) (T_LINE 5928) (T_LIST 5929) (T_LOGICAL_AND 5930) (T_LOGICAL_OR 5931) (T_LOGICAL_XOR 5932) (T_MATCH 5933) (T_METHOD_C 5934) (T_NAMESPACE 5935) (T_NAME_FULLY_QUALIFIED 5936) (T_NAME_QUALIFIED 5937) (T_NAME_RELATIVE 5938) (T_NEW 5939) (T_NS_C 5940) (T_PRINT 5941) (T_PRIVATE 5942) (T_PROTECTED 5943) (T_PUBLIC 5944) (T_READONLY 5945) (T_REQUIRE 5946) (T_REQUIRE_ONCE 5947) (T_RETURN 5948) (T_STATIC 5949) (T_STRING 5950) (T_SWITCH 5951) (T_THROW 5952) (T_TRAIT 5953) (T_TRAIT_C 5954) (T_TRY 5955) (T_UNSET 5956) (T_USE 5957) (T_VAR 5958) (T_WHILE 5959) (T_YIELD 5960) (absolute_trait_method_reference 5961) (class_name 5962) (identifier 5963) (name 5964) (reserved_non_modifiers 5965) (semi_reserved 5966) (trait_adaptation 5967) (trait_adaptation_list 5968) (trait_alias 5969) (trait_method_reference 5970) (trait_precedence 5971) ("}" 5972)) 2616 ((T_INSTEADOF 6151)) 2617 ((T_PAAMAYIM_NEKUDOTAYIM 6069)) 2618 ((T_ABSTRACT 5880) (T_ARRAY 5881) (T_AS 5882) (T_BREAK 5883) (T_CALLABLE 5884) (T_CASE 5885) (T_CATCH 5886) (T_CLASS 5887) (T_CLASS_C 5888) (T_CLONE 5889) (T_CONST 5890) (T_CONTINUE 5891) (T_DECLARE 5892) (T_DEFAULT 5893) (T_DIR 5894) (T_DO 5895) (T_ECHO 5896) (T_ELSE 5897) (T_ELSEIF 5898) (T_EMPTY 5899) (T_ENDDECLARE 5900) (T_ENDFOR 5901) (T_ENDFOREACH 5902) (T_ENDIF 5903) (T_ENDSWITCH 5904) (T_ENDWHILE 5905) (T_ENUM 5906) (T_EVAL 5907) (T_EXIT 5908) (T_EXTENDS 5909) (T_FILE 5910) (T_FINAL 5911) (T_FINALLY 5912) (T_FN 5913) (T_FOR 5914) (T_FOREACH 5915) (T_FUNCTION 5916) (T_FUNC_C 5917) (T_GLOBAL 5918) (T_GOTO 5919) (T_IF 5920) (T_IMPLEMENTS 5921) (T_INCLUDE 5922) (T_INCLUDE_ONCE 5923) (T_INSTANCEOF 5924) (T_INSTEADOF 5925) (T_INTERFACE 5926) (T_ISSET 5927) (T_LINE 5928) (T_LIST 5929) (T_LOGICAL_AND 5930) (T_LOGICAL_OR 5931) (T_LOGICAL_XOR 5932) (T_MATCH 5933) (T_METHOD_C 5934) (T_NAMESPACE 5935) (T_NAME_FULLY_QUALIFIED 5936) (T_NAME_QUALIFIED 5937) (T_NAME_RELATIVE 5938) (T_NEW 5939) (T_NS_C 5940) (T_PRINT 5941) (T_PRIVATE 5942) (T_PROTECTED 5943) (T_PUBLIC 5944) (T_READONLY 5945) (T_REQUIRE 5946) (T_REQUIRE_ONCE 5947) (T_RETURN 5948) (T_STATIC 5949) (T_STRING 5950) (T_SWITCH 5951) (T_THROW 5952) (T_TRAIT 5953) (T_TRAIT_C 5954) (T_TRY 5955) (T_UNSET 5956) (T_USE 5957) (T_VAR 5958) (T_WHILE 5959) (T_YIELD 5960) (absolute_trait_method_reference 5961) (class_name 5962) (identifier 5963) (name 5964) (reserved_non_modifiers 5965) (semi_reserved 5966) (trait_adaptation 6067) (trait_alias 5969) (trait_method_reference 5970) (trait_precedence 5971) ("}" 6068)) 2619 ((";" 6066)) 2620 ((T_AS 5974)) 2621 ((";" 5973)) 2622 ((T_ABSTRACT 5975) (T_ARRAY 5976) (T_AS 5977) (T_BREAK 5978) (T_CALLABLE 5979) (T_CASE 5980) (T_CATCH 5981) (T_CLASS 5982) (T_CLASS_C 5983) (T_CLONE 5984) (T_CONST 5985) (T_CONTINUE 5986) (T_DECLARE 5987) (T_DEFAULT 5988) (T_DIR 5989) (T_DO 5990) (T_ECHO 5991) (T_ELSE 5992) (T_ELSEIF 5993) (T_EMPTY 5994) (T_ENDDECLARE 5995) (T_ENDFOR 5996) (T_ENDFOREACH 5997) (T_ENDIF 5998) (T_ENDSWITCH 5999) (T_ENDWHILE 6000) (T_ENUM 6001) (T_EVAL 6002) (T_EXIT 6003) (T_EXTENDS 6004) (T_FILE 6005) (T_FINAL 6006) (T_FINALLY 6007) (T_FN 6008) (T_FOR 6009) (T_FOREACH 6010) (T_FUNCTION 6011) (T_FUNC_C 6012) (T_GLOBAL 6013) (T_GOTO 6014) (T_IF 6015) (T_IMPLEMENTS 6016) (T_INCLUDE 6017) (T_INCLUDE_ONCE 6018) (T_INSTANCEOF 6019) (T_INSTEADOF 6020) (T_INTERFACE 6021) (T_ISSET 6022) (T_LINE 6023) (T_LIST 6024) (T_LOGICAL_AND 6025) (T_LOGICAL_OR 6026) (T_LOGICAL_XOR 6027) (T_MATCH 6028) (T_METHOD_C 6029) (T_NAMESPACE 6030) (T_NEW 6031) (T_NS_C 6032) (T_PRINT 6033) (T_PRIVATE 6034) (T_PROTECTED 6035) (T_PUBLIC 6036) (T_READONLY 6037) (T_REQUIRE 6038) (T_REQUIRE_ONCE 6039) (T_RETURN 6040) (T_STATIC 6041) (T_STRING 6042) (T_SWITCH 6043) (T_THROW 6044) (T_TRAIT 6045) (T_TRAIT_C 6046) (T_TRY 6047) (T_UNSET 6048) (T_USE 6049) (T_VAR 6050) (T_WHILE 6051) (T_YIELD 6052) (member_modifier 6053) (reserved_non_modifiers 6054)) 2623 ((T_ABSTRACT 6055) (T_ARRAY 5976) (T_AS 5977) (T_BREAK 5978) (T_CALLABLE 5979) (T_CASE 5980) (T_CATCH 5981) (T_CLASS 5982) (T_CLASS_C 5983) (T_CLONE 5984) (T_CONST 5985) (T_CONTINUE 5986) (T_DECLARE 5987) (T_DEFAULT 5988) (T_DIR 5989) (T_DO 5990) (T_ECHO 5991) (T_ELSE 5992) (T_ELSEIF 5993) (T_EMPTY 5994) (T_ENDDECLARE 5995) (T_ENDFOR 5996) (T_ENDFOREACH 5997) (T_ENDIF 5998) (T_ENDSWITCH 5999) (T_ENDWHILE 6000) (T_ENUM 6001) (T_EVAL 6002) (T_EXIT 6003) (T_EXTENDS 6004) (T_FILE 6005) (T_FINAL 6056) (T_FINALLY 6007) (T_FN 6008) (T_FOR 6009) (T_FOREACH 6010) (T_FUNCTION 6011) (T_FUNC_C 6012) (T_GLOBAL 6013) (T_GOTO 6014) (T_IF 6015) (T_IMPLEMENTS 6016) (T_INCLUDE 6017) (T_INCLUDE_ONCE 6018) (T_INSTANCEOF 6019) (T_INSTEADOF 6020) (T_INTERFACE 6021) (T_ISSET 6022) (T_LINE 6023) (T_LIST 6024) (T_LOGICAL_AND 6025) (T_LOGICAL_OR 6026) (T_LOGICAL_XOR 6027) (T_MATCH 6028) (T_METHOD_C 6029) (T_NAMESPACE 6030) (T_NEW 6031) (T_NS_C 6032) (T_PRINT 6033) (T_PRIVATE 6057) (T_PROTECTED 6058) (T_PUBLIC 6059) (T_READONLY 6060) (T_REQUIRE 6038) (T_REQUIRE_ONCE 6039) (T_RETURN 6040) (T_STATIC 6061) (T_STRING 6062) (T_SWITCH 6043) (T_THROW 6044) (T_TRAIT 6045) (T_TRAIT_C 6046) (T_TRY 6047) (T_UNSET 6048) (T_USE 6049) (T_VAR 6050) (T_WHILE 6051) (T_YIELD 6052) (identifier 6063) (reserved_non_modifiers 6064) (semi_reserved 6065)) 2624 ((T_ABSTRACT 6070) (T_ARRAY 6071) (T_AS 6072) (T_BREAK 6073) (T_CALLABLE 6074) (T_CASE 6075) (T_CATCH 6076) (T_CLASS 6077) (T_CLASS_C 6078) (T_CLONE 6079) (T_CONST 6080) (T_CONTINUE 6081) (T_DECLARE 6082) (T_DEFAULT 6083) (T_DIR 6084) (T_DO 6085) (T_ECHO 6086) (T_ELSE 6087) (T_ELSEIF 6088) (T_EMPTY 6089) (T_ENDDECLARE 6090) (T_ENDFOR 6091) (T_ENDFOREACH 6092) (T_ENDIF 6093) (T_ENDSWITCH 6094) (T_ENDWHILE 6095) (T_ENUM 6096) (T_EVAL 6097) (T_EXIT 6098) (T_EXTENDS 6099) (T_FILE 6100) (T_FINAL 6101) (T_FINALLY 6102) (T_FN 6103) (T_FOR 6104) (T_FOREACH 6105) (T_FUNCTION 6106) (T_FUNC_C 6107) (T_GLOBAL 6108) (T_GOTO 6109) (T_IF 6110) (T_IMPLEMENTS 6111) (T_INCLUDE 6112) (T_INCLUDE_ONCE 6113) (T_INSTANCEOF 6114) (T_INSTEADOF 6115) (T_INTERFACE 6116) (T_ISSET 6117) (T_LINE 6118) (T_LIST 6119) (T_LOGICAL_AND 6120) (T_LOGICAL_OR 6121) (T_LOGICAL_XOR 6122) (T_MATCH 6123) (T_METHOD_C 6124) (T_NAMESPACE 6125) (T_NEW 6126) (T_NS_C 6127) (T_PRINT 6128) (T_PRIVATE 6129) (T_PROTECTED 6130) (T_PUBLIC 6131) (T_READONLY 6132) (T_REQUIRE 6133) (T_REQUIRE_ONCE 6134) (T_RETURN 6135) (T_STATIC 6136) (T_STRING 6137) (T_SWITCH 6138) (T_THROW 6139) (T_TRAIT 6140) (T_TRAIT_C 6141) (T_TRY 6142) (T_UNSET 6143) (T_USE 6144) (T_VAR 6145) (T_WHILE 6146) (T_YIELD 6147) (identifier 6148) (reserved_non_modifiers 6149) (semi_reserved 6150)) 2625 ((T_NAME_FULLY_QUALIFIED 6152) (T_NAME_QUALIFIED 6153) (T_NAME_RELATIVE 6154) (T_STATIC 6155) (T_STRING 6156) (class_name 6157) (class_name_list 6158) (name 6159)) 2626 (("," 6160)) 2627 ((T_NAME_FULLY_QUALIFIED 6152) (T_NAME_QUALIFIED 6153) (T_NAME_RELATIVE 6154) (T_STATIC 6155) (T_STRING 6156) (class_name 6161) (name 6159)) 2628 ((T_ABSTRACT 6164) (T_ARRAY 6165) (T_AS 6166) (T_BREAK 6167) (T_CALLABLE 6168) (T_CASE 6169) (T_CATCH 6170) (T_CLASS 6171) (T_CLASS_C 6172) (T_CLONE 6173) (T_CONST 6174) (T_CONTINUE 6175) (T_DECLARE 6176) (T_DEFAULT 6177) (T_DIR 6178) (T_DO 6179) (T_ECHO 6180) (T_ELSE 6181) (T_ELSEIF 6182) (T_EMPTY 6183) (T_ENDDECLARE 6184) (T_ENDFOR 6185) (T_ENDFOREACH 6186) (T_ENDIF 6187) (T_ENDSWITCH 6188) (T_ENDWHILE 6189) (T_ENUM 6190) (T_EVAL 6191) (T_EXIT 6192) (T_EXTENDS 6193) (T_FILE 6194) (T_FINAL 6195) (T_FINALLY 6196) (T_FN 6197) (T_FOR 6198) (T_FOREACH 6199) (T_FUNCTION 6200) (T_FUNC_C 6201) (T_GLOBAL 6202) (T_GOTO 6203) (T_IF 6204) (T_IMPLEMENTS 6205) (T_INCLUDE 6206) (T_INCLUDE_ONCE 6207) (T_INSTANCEOF 6208) (T_INSTEADOF 6209) (T_INTERFACE 6210) (T_ISSET 6211) (T_LINE 6212) (T_LIST 6213) (T_LOGICAL_AND 6214) (T_LOGICAL_OR 6215) (T_LOGICAL_XOR 6216) (T_MATCH 6217) (T_METHOD_C 6218) (T_NAMESPACE 6219) (T_NEW 6220) (T_NS_C 6221) (T_PRINT 6222) (T_PRIVATE 6223) (T_PROTECTED 6224) (T_PUBLIC 6225) (T_READONLY 6226) (T_REQUIRE 6227) (T_REQUIRE_ONCE 6228) (T_RETURN 6229) (T_STATIC 6230) (T_STRING 6231) (T_SWITCH 6232) (T_THROW 6233) (T_TRAIT 6234) (T_TRAIT_C 6235) (T_TRY 6236) (T_UNSET 6237) (T_USE 6238) (T_VAR 6239) (T_WHILE 6240) (T_YIELD 6241) (identifier 6242) (reserved_non_modifiers 6243) (semi_reserved 6244)) 2629 (("=" 6245) (enum_case_expr 6246)) 2630 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6248) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2631 ((";" 6247)) 2632 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6349) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2633 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6348) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2634 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6347) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2635 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6346) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2636 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6345) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2637 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6344) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2638 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6343) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2639 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6342) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2640 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 6337) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 6338) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 2641 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6336) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2642 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6335) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2643 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6334) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2644 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6333) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2645 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6332) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2646 (("$" 6294) ("(" 6295) (T_NAME_FULLY_QUALIFIED 6296) (T_NAME_QUALIFIED 6297) (T_NAME_RELATIVE 6298) (T_STATIC 6299) (T_STRING 6300) (T_VARIABLE 6301) (class_name 6302) (class_name_reference 6303) (name 6304) (new_variable 6305) (simple_variable 6306)) 2647 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6293) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2648 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6292) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2649 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6291) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2650 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6290) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2651 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6289) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2652 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6288) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2653 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6287) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2654 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6286) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2655 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6285) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2656 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6284) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2657 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6283) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2658 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6282) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2659 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6281) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2660 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6280) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2661 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6279) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2662 (("$" 6294) (T_VARIABLE 6301) (simple_variable 6328) ("{" 6329)) 2663 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 6326) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2664 ((T_PAAMAYIM_NEKUDOTAYIM 6324)) 2665 ((T_NULLSAFE_OBJECT_OPERATOR 6307) (T_OBJECT_OPERATOR 6308) (T_PAAMAYIM_NEKUDOTAYIM 6309) ("[" 6310) ("{" 6311)) 2666 (("$" 6294) (T_STRING 6317) (T_VARIABLE 6301) (property_name 6323) (simple_variable 6319) ("{" 6320)) 2667 (("$" 6294) (T_STRING 6317) (T_VARIABLE 6301) (property_name 6318) (simple_variable 6319) ("{" 6320)) 2668 (("$" 6294) (T_VARIABLE 6301) (simple_variable 6316)) 2669 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 6314) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2670 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6312) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2671 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6313)) 2672 (("]" 6315)) 2673 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6321) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2674 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6322)) 2675 (("$" 6294) (T_VARIABLE 6301) (simple_variable 6325)) 2676 (("%" 280) (")" 6327) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2677 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6330) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2678 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6331)) 2679 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6341) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2680 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 6339) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 2681 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 6340) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 2682 (("," 6367) (possible_comma 6368)) 2683 (("(" 6359) (argument_list 6360)) 2684 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 6361) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 6362) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 6363) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2685 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 6366) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2686 (("," 1449) (possible_comma 6364)) 2687 ((")" 6365)) 2688 ((T_NAME_FULLY_QUALIFIED 6350) (T_NAME_QUALIFIED 6351) (T_NAME_RELATIVE 6352) (T_STATIC 6353) (T_STRING 6354) (attribute_decl 6370) (class_name 6357) (name 6358)) 2689 (("]" 6369)) 2690 (("," 6379)) 2691 ((T_NAME_FULLY_QUALIFIED 6371) (T_NAME_QUALIFIED 6372) (T_NAME_RELATIVE 6373) (T_STATIC 6374) (T_STRING 6375) (class_name 6380) (name 6378)) 2692 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 6393) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2693 (("," 1449) (possible_comma 6391)) 2694 ((")" 6392)) 2695 (("," 6367) (possible_comma 6395)) 2696 (("]" 6396)) 2697 (("%" 280) (")" 6398) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2698 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6401) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2699 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6402)) 2700 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 6404) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2701 (("%" 280) (")" 6405) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2702 (("{" 6406)) 2703 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 6407) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 2704 (("}" 6408)) 2705 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 6410) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2706 ((")" 6411)) 2707 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 6413) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2708 (("," 5117) (possible_comma 6414)) 2709 ((")" 6415)) 2710 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 6422) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2711 ((")" 6423)) 2712 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 6425) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2713 (("%" 280) (")" 6426) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2714 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 6428) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2715 (("%" 280) (")" 6429) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2716 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 6436) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2717 ((")" 6437)) 2718 (("%" 280) (")" 6442) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2719 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6445) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2720 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6446)) 2721 (("\"" 6448) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 2722 (("$" 1826) (T_ABSTRACT 2015) (T_ARRAY 2016) (T_AS 2017) (T_BREAK 2018) (T_CALLABLE 2019) (T_CASE 2020) (T_CATCH 2021) (T_CLASS 2022) (T_CLASS_C 2023) (T_CLONE 2024) (T_CONST 2025) (T_CONTINUE 2026) (T_DECLARE 2027) (T_DEFAULT 2028) (T_DIR 2029) (T_DO 2030) (T_ECHO 2031) (T_ELSE 2032) (T_ELSEIF 2033) (T_EMPTY 2034) (T_ENDDECLARE 2035) (T_ENDFOR 2036) (T_ENDFOREACH 2037) (T_ENDIF 2038) (T_ENDSWITCH 2039) (T_ENDWHILE 2040) (T_ENUM 2041) (T_EVAL 2042) (T_EXIT 2043) (T_EXTENDS 2044) (T_FILE 2045) (T_FINAL 2046) (T_FINALLY 2047) (T_FN 2048) (T_FOR 2049) (T_FOREACH 2050) (T_FUNCTION 2051) (T_FUNC_C 2052) (T_GLOBAL 2053) (T_GOTO 2054) (T_IF 2055) (T_IMPLEMENTS 2056) (T_INCLUDE 2057) (T_INCLUDE_ONCE 2058) (T_INSTANCEOF 2059) (T_INSTEADOF 2060) (T_INTERFACE 2061) (T_ISSET 2062) (T_LINE 2063) (T_LIST 2064) (T_LOGICAL_AND 2065) (T_LOGICAL_OR 2066) (T_LOGICAL_XOR 2067) (T_MATCH 2068) (T_METHOD_C 2069) (T_NAMESPACE 2070) (T_NEW 2071) (T_NS_C 2072) (T_PRINT 2073) (T_PRIVATE 2074) (T_PROTECTED 2075) (T_PUBLIC 2076) (T_READONLY 2077) (T_REQUIRE 2078) (T_REQUIRE_ONCE 2079) (T_RETURN 2080) (T_STATIC 2081) (T_STRING 2082) (T_SWITCH 2083) (T_THROW 2084) (T_TRAIT 2085) (T_TRAIT_C 2086) (T_TRY 2087) (T_UNSET 2088) (T_USE 2089) (T_VAR 2090) (T_VARIABLE 1873) (T_WHILE 2091) (T_YIELD 2092) (identifier 6451) (member_name 6452) (reserved_non_modifiers 2095) (semi_reserved 2096) (simple_variable 6453) ("{" 914)) 2723 (("(" 2098) (argument_list 6454)) 2724 ((T_FN 35) (T_FUNCTION 126) (fn 1890) #'1891 (inline_function 6458)) 2725 (("=" 6460)) 2726 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6461) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2727 (("$" 1826) (T_STRING 6470) (T_VARIABLE 1873) (property_name 6477) (simple_variable 6472) ("{" 6473)) 2728 (("$" 1826) (T_STRING 6470) (T_VARIABLE 1873) (property_name 6471) (simple_variable 6472) ("{" 6473)) 2729 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 6468) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2730 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6466) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2731 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6467)) 2732 (("]" 6469)) 2733 (("(" 2098) (argument_list 6476)) 2734 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6474) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2735 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6475)) 2736 (("(" 2098) (argument_list 6478)) 2737 (("`" 6481)) 2738 (("]" 6483)) 2739 (("=" 6484)) 2740 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6485) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2741 (("%" 1908) ("*" 1909) ("+" 1910) ("-" 1911) ("." 1912) ("/" 1913) ("<" 1914) (">" 1915) ("?" 1916) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1917) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1918) (T_BOOLEAN_AND 1919) (T_BOOLEAN_OR 1920) (T_COALESCE 1921) (T_DOUBLE_ARROW 6488) (T_INSTANCEOF 1922) (T_IS_EQUAL 1923) (T_IS_GREATER_OR_EQUAL 1924) (T_IS_IDENTICAL 1925) (T_IS_NOT_EQUAL 1926) (T_IS_NOT_IDENTICAL 1927) (T_IS_SMALLER_OR_EQUAL 1928) (T_LOGICAL_AND 1929) (T_LOGICAL_OR 1930) (T_LOGICAL_XOR 1931) (T_POW 1932) (T_SL 1933) (T_SPACESHIP 1934) (T_SR 1935) ("^" 1936) ("|" 1937)) 2742 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1906) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (array_object_dereferenceable 1879) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 6489) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2743 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 6498) (T_VARIABLE 2822) (encaps_var 2836)) 2744 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 6497) (T_VARIABLE 2822) (encaps_var 2827)) 2745 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 6548) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2746 (("(" 2899) (argument_list 2900) (ctor_arguments 6541)) 2747 ((T_ATTRIBUTE 2853) (T_CLASS 6504) (anonymous_class 6540) (attribute 2898)) 2748 ((T_PAAMAYIM_NEKUDOTAYIM 6538)) 2749 (("(" 6529) (argument_list 6530) (ctor_arguments 6531)) 2750 ((T_NULLSAFE_OBJECT_OPERATOR 6517) (T_OBJECT_OPERATOR 6518) (T_PAAMAYIM_NEKUDOTAYIM 6519) ("[" 6520) ("{" 6521)) 2751 (("$" 1490) (T_STRING 1629) (T_VARIABLE 1507) (property_name 6528) (simple_variable 1631) ("{" 1632)) 2752 (("$" 1490) (T_STRING 1629) (T_VARIABLE 1507) (property_name 6527) (simple_variable 1631) ("{" 1632)) 2753 (("$" 1490) (T_VARIABLE 1507) (simple_variable 6526)) 2754 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 6524) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2755 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6522) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2756 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6523)) 2757 (("]" 6525)) 2758 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 6532) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 6533) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 6534) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2759 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 6537) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2760 (("," 1449) (possible_comma 6535)) 2761 ((")" 6536)) 2762 (("$" 1490) (T_VARIABLE 1507) (simple_variable 6539)) 2763 ((T_EXTENDS 2902) (extends_from 6542)) 2764 ((T_IMPLEMENTS 2904) (implements_list 6543)) 2765 ((backup_doc_comment 6544)) 2766 (("{" 6545)) 2767 ((class_statement_list 6546)) 2768 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 6547)) 2769 (("%" 280) (")" 6549) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2770 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 6551) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2771 (("%" 280) (")" 6552) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2772 (("{" 6553)) 2773 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 6554) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 2774 (("}" 6555)) 2775 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 6557) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2776 ((")" 6558)) 2777 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 6560) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2778 (("," 5117) (possible_comma 6561)) 2779 ((")" 6562)) 2780 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 6569) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2781 ((")" 6570)) 2782 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 6572) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2783 (("%" 280) (")" 6573) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2784 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 6575) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2785 (("%" 280) (")" 6576) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2786 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 6584) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 2787 ((")" 6585)) 2788 (("%" 280) (")" 6590) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2789 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6593) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2790 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6594)) 2791 (("\"" 6596) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 2792 (("$" 1640) (T_ABSTRACT 6599) (T_ARRAY 6600) (T_AS 6601) (T_BREAK 6602) (T_CALLABLE 6603) (T_CASE 6604) (T_CATCH 6605) (T_CLASS 6606) (T_CLASS_C 6607) (T_CLONE 6608) (T_CONST 6609) (T_CONTINUE 6610) (T_DECLARE 6611) (T_DEFAULT 6612) (T_DIR 6613) (T_DO 6614) (T_ECHO 6615) (T_ELSE 6616) (T_ELSEIF 6617) (T_EMPTY 6618) (T_ENDDECLARE 6619) (T_ENDFOR 6620) (T_ENDFOREACH 6621) (T_ENDIF 6622) (T_ENDSWITCH 6623) (T_ENDWHILE 6624) (T_ENUM 6625) (T_EVAL 6626) (T_EXIT 6627) (T_EXTENDS 6628) (T_FILE 6629) (T_FINAL 6630) (T_FINALLY 6631) (T_FN 6632) (T_FOR 6633) (T_FOREACH 6634) (T_FUNCTION 6635) (T_FUNC_C 6636) (T_GLOBAL 6637) (T_GOTO 6638) (T_IF 6639) (T_IMPLEMENTS 6640) (T_INCLUDE 6641) (T_INCLUDE_ONCE 6642) (T_INSTANCEOF 6643) (T_INSTEADOF 6644) (T_INTERFACE 6645) (T_ISSET 6646) (T_LINE 6647) (T_LIST 6648) (T_LOGICAL_AND 6649) (T_LOGICAL_OR 6650) (T_LOGICAL_XOR 6651) (T_MATCH 6652) (T_METHOD_C 6653) (T_NAMESPACE 6654) (T_NEW 6655) (T_NS_C 6656) (T_PRINT 6657) (T_PRIVATE 6658) (T_PROTECTED 6659) (T_PUBLIC 6660) (T_READONLY 6661) (T_REQUIRE 6662) (T_REQUIRE_ONCE 6663) (T_RETURN 6664) (T_STATIC 6665) (T_STRING 6666) (T_SWITCH 6667) (T_THROW 6668) (T_TRAIT 6669) (T_TRAIT_C 6670) (T_TRY 6671) (T_UNSET 6672) (T_USE 6673) (T_VAR 6674) (T_VARIABLE 1687) (T_WHILE 6675) (T_YIELD 6676) (identifier 6677) (member_name 6678) (reserved_non_modifiers 6679) (semi_reserved 6680) (simple_variable 6681) ("{" 914)) 2793 (("(" 6682) (argument_list 6683)) 2794 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 6684) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 6685) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 6686) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2795 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 6689) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2796 (("," 1449) (possible_comma 6687)) 2797 ((")" 6688)) 2798 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 6719) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6720) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2799 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6718) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2800 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6717) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2801 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6716) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2802 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6715) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2803 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6714) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2804 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6713) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2805 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6712) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2806 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6711) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2807 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6710) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2808 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6709) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2809 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6708) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2810 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6707) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2811 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 6706) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2812 (("\"" 1489) ("$" 6721) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 6722) ("[" 1508) (array_object_dereferenceable 6723) (callable_expr 6724) (callable_variable 6725) (class_constant 1512) (class_name 6726) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 6727) (name 6728) (simple_variable 6729) (static_member 6730) (variable 6731) (variable_class_name 6732)) 2813 (("$" 6721) (T_VARIABLE 6722) (simple_variable 6767) ("{" 6768)) 2814 ((T_NULLSAFE_OBJECT_OPERATOR 6750) (T_OBJECT_OPERATOR 6751) ("[" 6752) ("{" 6753)) 2815 (("(" 6736) (argument_list 6749)) 2816 ((T_PAAMAYIM_NEKUDOTAYIM 6745)) 2817 (("(" 6736) (argument_list 6744)) 2818 ((T_PAAMAYIM_NEKUDOTAYIM 6733)) 2819 (("$" 6721) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 6722) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 6734) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 6735) ("{" 914)) 2820 (("(" 6736) (argument_list 6737)) 2821 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 6738) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 6739) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 6740) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2822 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 6743) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2823 (("," 1449) (possible_comma 6741)) 2824 ((")" 6742)) 2825 (("$" 6721) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 6722) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 6746) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 6747) ("{" 914)) 2826 (("(" 6736) (argument_list 6748)) 2827 (("$" 6721) (T_STRING 6758) (T_VARIABLE 6722) (property_name 6765) (simple_variable 6760) ("{" 6761)) 2828 (("$" 6721) (T_STRING 6758) (T_VARIABLE 6722) (property_name 6759) (simple_variable 6760) ("{" 6761)) 2829 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 6756) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2830 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6754) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2831 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6755)) 2832 (("]" 6757)) 2833 (("(" 6736) (argument_list 6764)) 2834 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6762) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2835 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6763)) 2836 (("(" 6736) (argument_list 6766)) 2837 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6769) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2838 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6770)) 2839 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (possible_array_pair 6772) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2840 ((backup_doc_comment 6775)) 2841 (("(" 6776)) 2842 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 6777)) 2843 ((")" 6778)) 2844 ((T_USE 148) (lexical_vars 6779)) 2845 ((":" 150) (return_type 6780)) 2846 ((backup_fn_flags 6781)) 2847 (("{" 6782)) 2848 ((inner_statement_list 6783)) 2849 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 6784) ("~" 124)) 2850 ((backup_fn_flags 6785)) 2851 ((backup_doc_comment 6787)) 2852 (("(" 6788)) 2853 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 6789)) 2854 ((")" 6790)) 2855 ((":" 2207) (return_type 6791)) 2856 ((T_DOUBLE_ARROW 6792)) 2857 ((backup_fn_flags 6793)) 2858 ((backup_lex_pos 6794)) 2859 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6856) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2860 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7315) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2861 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 7313) (encaps_var 5158)) 2862 (("$" 6797) (T_VARIABLE 6843) (simple_variable 7309) ("{" 7310)) 2863 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 7307) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2864 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7306) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2865 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7305) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2866 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7304) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2867 (("(" 7301)) 2868 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7300) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2869 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7299) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2870 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7298) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2871 (("\"" 1489) ("$" 7096) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 7097) ("[" 1508) (array_object_dereferenceable 7098) (callable_expr 7099) (callable_variable 7100) (class_constant 1512) (class_name 7101) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 7102) (name 7103) (simple_variable 7104) (static_member 7105) (variable 7297) (variable_class_name 7107)) 2872 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7296) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2873 (("(" 7293)) 2874 (("(" 7290)) 2875 (("(" 7286) (exit_expr 7287)) 2876 (("\"" 1489) ("$" 7096) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 7097) ("[" 1508) (array_object_dereferenceable 7098) (callable_expr 7099) (callable_variable 7100) (class_constant 1512) (class_name 7101) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 7102) (name 7103) (simple_variable 7104) (static_member 7105) (variable 7285) (variable_class_name 7107)) 2877 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7284) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2878 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7283) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2879 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7282) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2880 (("(" 7278)) 2881 (("(" 7273)) 2882 (("(" 7267)) 2883 (("$" 7096) ("(" 7220) (T_ATTRIBUTE 2853) (T_CLASS 7221) (T_NAME_FULLY_QUALIFIED 7222) (T_NAME_QUALIFIED 7223) (T_NAME_RELATIVE 7224) (T_STATIC 7225) (T_STRING 7226) (T_VARIABLE 7097) (anonymous_class 7227) (attribute 2862) (attributes 7228) (class_name 7229) (class_name_reference 7230) (name 7231) (new_variable 7232) (simple_variable 7233)) 2884 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7219) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2885 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7218) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2886 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7217) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2887 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7216) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2888 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 7211) (T_END_HEREDOC 7212) (T_VARIABLE 2822) (encaps_list 7213) (encaps_var 2824)) 2889 ((T_FN 35) (T_FUNCTION 126) (fn 6857) #'6858 (inline_function 7210)) 2890 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7209) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2891 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7208) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2892 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7207) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2893 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7204) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2894 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7203) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2895 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 7199) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 2896 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 7197) (encaps_list 2729) (encaps_var 2730)) 2897 ((T_NULLSAFE_OBJECT_OPERATOR 7180) (T_OBJECT_OPERATOR 7181) ("[" 7182) ("{" 7183)) 2898 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 7177) (attribute 2654) (fn 6857) #'6858 (inline_function 7178)) 2899 (("(" 7057) (argument_list 7176)) 2900 ((T_PAAMAYIM_NEKUDOTAYIM 7171)) 2901 (("%" 6872) ("*" 6873) ("+" 6874) ("-" 6875) ("." 6876) ("/" 6877) ("<" 6878) (">" 6879) ("?" 6880) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6881) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6882) (T_BOOLEAN_AND 6883) (T_BOOLEAN_OR 6884) (T_COALESCE 6885) (T_INSTANCEOF 6886) (T_IS_EQUAL 6887) (T_IS_GREATER_OR_EQUAL 6888) (T_IS_IDENTICAL 6889) (T_IS_NOT_EQUAL 6890) (T_IS_NOT_IDENTICAL 6891) (T_IS_SMALLER_OR_EQUAL 6892) (T_LOGICAL_AND 6893) (T_LOGICAL_OR 6894) (T_LOGICAL_XOR 6895) (T_POW 6896) (T_SL 6897) (T_SPACESHIP 6898) (T_SR 6899) ("^" 6900) (backup_fn_flags 7170) ("|" 6901)) 2902 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 7159)) 2903 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 7147)) 2904 (("(" 7057) (argument_list 7146)) 2905 (("=" 7065) (T_AND_EQUAL 7066) (T_COALESCE_EQUAL 7067) (T_CONCAT_EQUAL 7068) (T_DEC 7069) (T_DIV_EQUAL 7070) (T_INC 7071) (T_MINUS_EQUAL 7072) (T_MOD_EQUAL 7073) (T_MUL_EQUAL 7074) (T_OR_EQUAL 7075) (T_PLUS_EQUAL 7076) (T_POW_EQUAL 7077) (T_SL_EQUAL 7078) (T_SR_EQUAL 7079) (T_XOR_EQUAL 7080)) 2906 ((T_PAAMAYIM_NEKUDOTAYIM 6973)) 2907 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6871) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2908 (("%" 6872) ("*" 6873) ("+" 6874) ("-" 6875) ("." 6876) ("/" 6877) ("<" 6878) (">" 6879) ("?" 6880) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6881) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6882) (T_BOOLEAN_AND 6883) (T_BOOLEAN_OR 6884) (T_COALESCE 6885) (T_INSTANCEOF 6886) (T_IS_EQUAL 6887) (T_IS_GREATER_OR_EQUAL 6888) (T_IS_IDENTICAL 6889) (T_IS_NOT_EQUAL 6890) (T_IS_NOT_IDENTICAL 6891) (T_IS_SMALLER_OR_EQUAL 6892) (T_LOGICAL_AND 6893) (T_LOGICAL_OR 6894) (T_LOGICAL_XOR 6895) (T_POW 6896) (T_SL 6897) (T_SPACESHIP 6898) (T_SR 6899) ("^" 6900) ("|" 6901)) 2909 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6972) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2910 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6971) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2911 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6970) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2912 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6969) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2913 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6968) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2914 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6967) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2915 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6966) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2916 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6965) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2917 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 6960) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 6961) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 2918 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6959) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2919 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6958) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2920 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6957) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2921 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6956) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2922 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6955) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2923 (("$" 6917) ("(" 6918) (T_NAME_FULLY_QUALIFIED 6919) (T_NAME_QUALIFIED 6920) (T_NAME_RELATIVE 6921) (T_STATIC 6922) (T_STRING 6923) (T_VARIABLE 6924) (class_name 6925) (class_name_reference 6926) (name 6927) (new_variable 6928) (simple_variable 6929)) 2924 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6916) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2925 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6915) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2926 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6914) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2927 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6913) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2928 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6912) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2929 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6911) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2930 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6910) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2931 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6909) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2932 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6908) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2933 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6907) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2934 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6906) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2935 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6905) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2936 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6904) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2937 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6903) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2938 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6902) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2939 (("$" 6917) (T_VARIABLE 6924) (simple_variable 6951) ("{" 6952)) 2940 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 6949) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 2941 ((T_PAAMAYIM_NEKUDOTAYIM 6947)) 2942 ((T_NULLSAFE_OBJECT_OPERATOR 6930) (T_OBJECT_OPERATOR 6931) (T_PAAMAYIM_NEKUDOTAYIM 6932) ("[" 6933) ("{" 6934)) 2943 (("$" 6917) (T_STRING 6940) (T_VARIABLE 6924) (property_name 6946) (simple_variable 6942) ("{" 6943)) 2944 (("$" 6917) (T_STRING 6940) (T_VARIABLE 6924) (property_name 6941) (simple_variable 6942) ("{" 6943)) 2945 (("$" 6917) (T_VARIABLE 6924) (simple_variable 6939)) 2946 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 6937) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2947 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6935) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2948 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6936)) 2949 (("]" 6938)) 2950 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6944) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2951 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6945)) 2952 (("$" 6917) (T_VARIABLE 6924) (simple_variable 6948)) 2953 (("%" 280) (")" 6950) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 2954 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 6953) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2955 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 6954)) 2956 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6964) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2957 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 6962) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 2958 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 6963) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2959 (("$" 6797) (T_ABSTRACT 6974) (T_ARRAY 6975) (T_AS 6976) (T_BREAK 6977) (T_CALLABLE 6978) (T_CASE 6979) (T_CATCH 6980) (T_CLASS 6981) (T_CLASS_C 6982) (T_CLONE 6983) (T_CONST 6984) (T_CONTINUE 6985) (T_DECLARE 6986) (T_DEFAULT 6987) (T_DIR 6988) (T_DO 6989) (T_ECHO 6990) (T_ELSE 6991) (T_ELSEIF 6992) (T_EMPTY 6993) (T_ENDDECLARE 6994) (T_ENDFOR 6995) (T_ENDFOREACH 6996) (T_ENDIF 6997) (T_ENDSWITCH 6998) (T_ENDWHILE 6999) (T_ENUM 7000) (T_EVAL 7001) (T_EXIT 7002) (T_EXTENDS 7003) (T_FILE 7004) (T_FINAL 7005) (T_FINALLY 7006) (T_FN 7007) (T_FOR 7008) (T_FOREACH 7009) (T_FUNCTION 7010) (T_FUNC_C 7011) (T_GLOBAL 7012) (T_GOTO 7013) (T_IF 7014) (T_IMPLEMENTS 7015) (T_INCLUDE 7016) (T_INCLUDE_ONCE 7017) (T_INSTANCEOF 7018) (T_INSTEADOF 7019) (T_INTERFACE 7020) (T_ISSET 7021) (T_LINE 7022) (T_LIST 7023) (T_LOGICAL_AND 7024) (T_LOGICAL_OR 7025) (T_LOGICAL_XOR 7026) (T_MATCH 7027) (T_METHOD_C 7028) (T_NAMESPACE 7029) (T_NEW 7030) (T_NS_C 7031) (T_PRINT 7032) (T_PRIVATE 7033) (T_PROTECTED 7034) (T_PUBLIC 7035) (T_READONLY 7036) (T_REQUIRE 7037) (T_REQUIRE_ONCE 7038) (T_RETURN 7039) (T_STATIC 7040) (T_STRING 7041) (T_SWITCH 7042) (T_THROW 7043) (T_TRAIT 7044) (T_TRAIT_C 7045) (T_TRY 7046) (T_UNSET 7047) (T_USE 7048) (T_VAR 7049) (T_VARIABLE 6843) (T_WHILE 7050) (T_YIELD 7051) (identifier 7052) (member_name 7053) (reserved_non_modifiers 7054) (semi_reserved 7055) (simple_variable 7056) ("{" 914)) 2960 (("(" 7057) (argument_list 7058)) 2961 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 7059) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 7060) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 7061) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2962 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 7064) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2963 (("," 1449) (possible_comma 7062)) 2964 ((")" 7063)) 2965 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (ampersand 7094) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7095) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2966 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7093) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2967 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7092) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2968 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7091) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2969 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7090) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2970 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7089) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2971 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7088) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2972 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7087) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2973 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7086) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2974 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7085) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2975 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7084) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2976 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7083) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2977 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7082) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2978 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7081) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 2979 (("\"" 1489) ("$" 7096) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 7097) ("[" 1508) (array_object_dereferenceable 7098) (callable_expr 7099) (callable_variable 7100) (class_constant 1512) (class_name 7101) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 7102) (name 7103) (simple_variable 7104) (static_member 7105) (variable 7106) (variable_class_name 7107)) 2980 (("$" 7096) (T_VARIABLE 7097) (simple_variable 7142) ("{" 7143)) 2981 ((T_NULLSAFE_OBJECT_OPERATOR 7125) (T_OBJECT_OPERATOR 7126) ("[" 7127) ("{" 7128)) 2982 (("(" 7111) (argument_list 7124)) 2983 ((T_PAAMAYIM_NEKUDOTAYIM 7120)) 2984 (("(" 7111) (argument_list 7119)) 2985 ((T_PAAMAYIM_NEKUDOTAYIM 7108)) 2986 (("$" 7096) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 7097) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 7109) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 7110) ("{" 914)) 2987 (("(" 7111) (argument_list 7112)) 2988 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 7113) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 7114) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 7115) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2989 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 7118) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 2990 (("," 1449) (possible_comma 7116)) 2991 ((")" 7117)) 2992 (("$" 7096) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 7097) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 7121) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 7122) ("{" 914)) 2993 (("(" 7111) (argument_list 7123)) 2994 (("$" 7096) (T_STRING 7133) (T_VARIABLE 7097) (property_name 7140) (simple_variable 7135) ("{" 7136)) 2995 (("$" 7096) (T_STRING 7133) (T_VARIABLE 7097) (property_name 7134) (simple_variable 7135) ("{" 7136)) 2996 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 7131) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 2997 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7129) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 2998 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7130)) 2999 (("]" 7132)) 3000 (("(" 7111) (argument_list 7139)) 3001 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7137) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3002 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7138)) 3003 (("(" 7111) (argument_list 7141)) 3004 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7144) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3005 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7145)) 3006 ((backup_doc_comment 7148)) 3007 (("(" 7149)) 3008 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 7150)) 3009 ((")" 7151)) 3010 ((T_USE 148) (lexical_vars 7152)) 3011 ((":" 150) (return_type 7153)) 3012 ((backup_fn_flags 7154)) 3013 (("{" 7155)) 3014 ((inner_statement_list 7156)) 3015 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 7157) ("~" 124)) 3016 ((backup_fn_flags 7158)) 3017 ((backup_doc_comment 7160)) 3018 (("(" 7161)) 3019 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 7162)) 3020 ((")" 7163)) 3021 ((":" 2207) (return_type 7164)) 3022 ((T_DOUBLE_ARROW 7165)) 3023 ((backup_fn_flags 7166)) 3024 ((backup_lex_pos 7167)) 3025 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7168) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 3026 (("%" 6872) ("*" 6873) ("+" 6874) ("-" 6875) ("." 6876) ("/" 6877) ("<" 6878) (">" 6879) ("?" 6880) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6881) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6882) (T_BOOLEAN_AND 6883) (T_BOOLEAN_OR 6884) (T_COALESCE 6885) (T_INSTANCEOF 6886) (T_IS_EQUAL 6887) (T_IS_GREATER_OR_EQUAL 6888) (T_IS_IDENTICAL 6889) (T_IS_NOT_EQUAL 6890) (T_IS_NOT_IDENTICAL 6891) (T_IS_SMALLER_OR_EQUAL 6892) (T_LOGICAL_AND 6893) (T_LOGICAL_OR 6894) (T_LOGICAL_XOR 6895) (T_POW 6896) (T_SL 6897) (T_SPACESHIP 6898) (T_SR 6899) ("^" 6900) (backup_fn_flags 7169) ("|" 6901)) 3027 (("$" 6797) (T_ABSTRACT 6974) (T_ARRAY 6975) (T_AS 6976) (T_BREAK 6977) (T_CALLABLE 6978) (T_CASE 6979) (T_CATCH 6980) (T_CLASS 6981) (T_CLASS_C 6982) (T_CLONE 6983) (T_CONST 6984) (T_CONTINUE 6985) (T_DECLARE 6986) (T_DEFAULT 6987) (T_DIR 6988) (T_DO 6989) (T_ECHO 6990) (T_ELSE 6991) (T_ELSEIF 6992) (T_EMPTY 6993) (T_ENDDECLARE 6994) (T_ENDFOR 6995) (T_ENDFOREACH 6996) (T_ENDIF 6997) (T_ENDSWITCH 6998) (T_ENDWHILE 6999) (T_ENUM 7000) (T_EVAL 7001) (T_EXIT 7002) (T_EXTENDS 7003) (T_FILE 7004) (T_FINAL 7005) (T_FINALLY 7006) (T_FN 7007) (T_FOR 7008) (T_FOREACH 7009) (T_FUNCTION 7010) (T_FUNC_C 7011) (T_GLOBAL 7012) (T_GOTO 7013) (T_IF 7014) (T_IMPLEMENTS 7015) (T_INCLUDE 7016) (T_INCLUDE_ONCE 7017) (T_INSTANCEOF 7018) (T_INSTEADOF 7019) (T_INTERFACE 7020) (T_ISSET 7021) (T_LINE 7022) (T_LIST 7023) (T_LOGICAL_AND 7024) (T_LOGICAL_OR 7025) (T_LOGICAL_XOR 7026) (T_MATCH 7027) (T_METHOD_C 7028) (T_NAMESPACE 7029) (T_NEW 7030) (T_NS_C 7031) (T_PRINT 7032) (T_PRIVATE 7033) (T_PROTECTED 7034) (T_PUBLIC 7035) (T_READONLY 7036) (T_REQUIRE 7037) (T_REQUIRE_ONCE 7038) (T_RETURN 7039) (T_STATIC 7040) (T_STRING 7041) (T_SWITCH 7042) (T_THROW 7043) (T_TRAIT 7044) (T_TRAIT_C 7045) (T_TRY 7046) (T_UNSET 7047) (T_USE 7048) (T_VAR 7049) (T_VARIABLE 6843) (T_WHILE 7050) (T_YIELD 7051) (identifier 7172) (member_name 7173) (reserved_non_modifiers 7054) (semi_reserved 7055) (simple_variable 7174) ("{" 914)) 3028 (("(" 7057) (argument_list 7175)) 3029 ((T_FN 35) (T_FUNCTION 126) (fn 6857) #'6858 (inline_function 7179)) 3030 (("$" 6797) (T_STRING 7188) (T_VARIABLE 6843) (property_name 7195) (simple_variable 7190) ("{" 7191)) 3031 (("$" 6797) (T_STRING 7188) (T_VARIABLE 6843) (property_name 7189) (simple_variable 7190) ("{" 7191)) 3032 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 7186) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3033 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7184) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3034 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7185)) 3035 (("]" 7187)) 3036 (("(" 7057) (argument_list 7194)) 3037 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7192) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3038 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7193)) 3039 (("(" 7057) (argument_list 7196)) 3040 (("`" 7198)) 3041 (("]" 7200)) 3042 (("=" 7201)) 3043 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7202) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 3044 (("%" 6872) ("*" 6873) ("+" 6874) ("-" 6875) ("." 6876) ("/" 6877) ("<" 6878) (">" 6879) ("?" 6880) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6881) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6882) (T_BOOLEAN_AND 6883) (T_BOOLEAN_OR 6884) (T_COALESCE 6885) (T_DOUBLE_ARROW 7205) (T_INSTANCEOF 6886) (T_IS_EQUAL 6887) (T_IS_GREATER_OR_EQUAL 6888) (T_IS_IDENTICAL 6889) (T_IS_NOT_EQUAL 6890) (T_IS_NOT_IDENTICAL 6891) (T_IS_SMALLER_OR_EQUAL 6892) (T_LOGICAL_AND 6893) (T_LOGICAL_OR 6894) (T_LOGICAL_XOR 6895) (T_POW 6896) (T_SL 6897) (T_SPACESHIP 6898) (T_SR 6899) ("^" 6900) ("|" 6901)) 3045 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7206) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 3046 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 7215) (T_VARIABLE 2822) (encaps_var 2836)) 3047 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 7214) (T_VARIABLE 2822) (encaps_var 2827)) 3048 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 7265) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3049 (("(" 2899) (argument_list 2900) (ctor_arguments 7258)) 3050 ((T_ATTRIBUTE 2853) (T_CLASS 7221) (anonymous_class 7257) (attribute 2898)) 3051 ((T_PAAMAYIM_NEKUDOTAYIM 7255)) 3052 (("(" 7246) (argument_list 7247) (ctor_arguments 7248)) 3053 ((T_NULLSAFE_OBJECT_OPERATOR 7234) (T_OBJECT_OPERATOR 7235) (T_PAAMAYIM_NEKUDOTAYIM 7236) ("[" 7237) ("{" 7238)) 3054 (("$" 7096) (T_STRING 7133) (T_VARIABLE 7097) (property_name 7245) (simple_variable 7135) ("{" 7136)) 3055 (("$" 7096) (T_STRING 7133) (T_VARIABLE 7097) (property_name 7244) (simple_variable 7135) ("{" 7136)) 3056 (("$" 7096) (T_VARIABLE 7097) (simple_variable 7243)) 3057 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 7241) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3058 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7239) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3059 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7240)) 3060 (("]" 7242)) 3061 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 7249) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 7250) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 7251) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3062 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 7254) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3063 (("," 1449) (possible_comma 7252)) 3064 ((")" 7253)) 3065 (("$" 7096) (T_VARIABLE 7097) (simple_variable 7256)) 3066 ((T_EXTENDS 2902) (extends_from 7259)) 3067 ((T_IMPLEMENTS 2904) (implements_list 7260)) 3068 ((backup_doc_comment 7261)) 3069 (("{" 7262)) 3070 ((class_statement_list 7263)) 3071 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 7264)) 3072 (("%" 280) (")" 7266) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3073 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 7268) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3074 (("%" 280) (")" 7269) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3075 (("{" 7270)) 3076 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 7271) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 3077 (("}" 7272)) 3078 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 7274) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3079 ((")" 7275)) 3080 (("=" 7276)) 3081 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 7277) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 3082 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 7279) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3083 (("," 5117) (possible_comma 7280)) 3084 ((")" 7281)) 3085 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 7288) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3086 ((")" 7289)) 3087 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 7291) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3088 (("%" 280) (")" 7292) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3089 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 7294) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3090 (("%" 280) (")" 7295) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3091 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 7302) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3092 ((")" 7303)) 3093 (("%" 280) (")" 7308) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3094 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7311) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3095 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7312)) 3096 (("\"" 7314) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 3097 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7345) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (ampersand 7370) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7379) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3098 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8384) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3099 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 8382) (encaps_var 5158)) 3100 (("$" 7319) (T_VARIABLE 7365) (simple_variable 8378) ("{" 8379)) 3101 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8376) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3102 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8375) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3103 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8374) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3104 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8373) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3105 (("(" 8370)) 3106 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8369) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3107 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8368) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3108 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8367) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3109 (("\"" 1489) ("$" 7625) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 7626) ("[" 1508) (array_object_dereferenceable 7627) (callable_expr 7628) (callable_variable 7629) (class_constant 1512) (class_name 7630) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 7631) (name 7632) (simple_variable 7633) (static_member 7634) (variable 8366) (variable_class_name 7636)) 3110 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8365) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3111 (("(" 8362)) 3112 (("(" 8359)) 3113 (("(" 8355) (exit_expr 8356)) 3114 (("\"" 1489) ("$" 7625) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 7626) ("[" 1508) (array_object_dereferenceable 7627) (callable_expr 7628) (callable_variable 7629) (class_constant 1512) (class_name 7630) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 7631) (name 7632) (simple_variable 7633) (static_member 7634) (variable 8354) (variable_class_name 7636)) 3115 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8353) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3116 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8352) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3117 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8351) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3118 (("(" 8347)) 3119 (("(" 8344)) 3120 (("(" 8338)) 3121 (("$" 7998) ("(" 8122) (T_ATTRIBUTE 2853) (T_CLASS 8317) (T_NAME_FULLY_QUALIFIED 8124) (T_NAME_QUALIFIED 8125) (T_NAME_RELATIVE 8126) (T_STATIC 8127) (T_STRING 8128) (T_VARIABLE 7999) (anonymous_class 8318) (attribute 2862) (attributes 8319) (class_name 8131) (class_name_reference 8320) (name 8133) (new_variable 8134) (simple_variable 8135)) 3122 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8316) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3123 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8315) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3124 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8314) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3125 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8313) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3126 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 8308) (T_END_HEREDOC 8309) (T_VARIABLE 2822) (encaps_list 8310) (encaps_var 2824)) 3127 ((T_FN 35) (T_FUNCTION 126) (fn 7380) #'7381 (inline_function 8307)) 3128 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8306) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3129 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8305) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3130 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8304) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3131 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8301) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 3132 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8300) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3133 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 8296) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 3134 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 8294) (encaps_list 2729) (encaps_var 2730)) 3135 (("\"" 1489) ("$" 8244) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 8245) ("[" 1508) (array_object_dereferenceable 8246) (callable_expr 8247) (callable_variable 8248) (class_constant 1512) (class_name 8249) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 8250) (name 8251) (simple_variable 8252) (static_member 8253) (variable 8254) (variable_class_name 8255)) 3136 ((T_NULLSAFE_OBJECT_OPERATOR 8227) (T_OBJECT_OPERATOR 8228) ("[" 8229) ("{" 8230)) 3137 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 8224) (attribute 2654) (fn 7380) #'7381 (inline_function 8225)) 3138 (("(" 7586) (argument_list 8223)) 3139 ((T_PAAMAYIM_NEKUDOTAYIM 8218)) 3140 (("%" 7396) ("*" 7397) ("+" 7398) ("-" 7399) ("." 7400) ("/" 7401) ("<" 7402) (">" 7403) ("?" 7404) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 7405) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7406) (T_BOOLEAN_AND 7407) (T_BOOLEAN_OR 7408) (T_COALESCE 7409) (T_INSTANCEOF 7410) (T_IS_EQUAL 7411) (T_IS_GREATER_OR_EQUAL 7412) (T_IS_IDENTICAL 7413) (T_IS_NOT_EQUAL 7414) (T_IS_NOT_IDENTICAL 7415) (T_IS_SMALLER_OR_EQUAL 7416) (T_LOGICAL_AND 7417) (T_LOGICAL_OR 7418) (T_LOGICAL_XOR 7419) (T_POW 7420) (T_SL 7421) (T_SPACESHIP 7422) (T_SR 7423) ("^" 7424) ("|" 7425)) 3141 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 7688)) 3142 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 7676)) 3143 (("(" 7586) (argument_list 7675)) 3144 (("=" 7594) (T_AND_EQUAL 7595) (T_COALESCE_EQUAL 7596) (T_CONCAT_EQUAL 7597) (T_DEC 7598) (T_DIV_EQUAL 7599) (T_INC 7600) (T_MINUS_EQUAL 7601) (T_MOD_EQUAL 7602) (T_MUL_EQUAL 7603) (T_OR_EQUAL 7604) (T_PLUS_EQUAL 7605) (T_POW_EQUAL 7606) (T_SL_EQUAL 7607) (T_SR_EQUAL 7608) (T_XOR_EQUAL 7609)) 3145 ((T_PAAMAYIM_NEKUDOTAYIM 7502)) 3146 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7395) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3147 (("(" 7497)) 3148 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7496) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3149 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7495) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3150 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7494) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3151 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7493) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3152 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7492) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3153 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7491) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3154 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7490) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3155 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7489) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3156 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 7484) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 7485) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3157 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7483) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3158 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7482) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3159 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7481) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3160 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7480) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3161 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7479) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3162 (("$" 7441) ("(" 7442) (T_NAME_FULLY_QUALIFIED 7443) (T_NAME_QUALIFIED 7444) (T_NAME_RELATIVE 7445) (T_STATIC 7446) (T_STRING 7447) (T_VARIABLE 7448) (class_name 7449) (class_name_reference 7450) (name 7451) (new_variable 7452) (simple_variable 7453)) 3163 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7440) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3164 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7439) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3165 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7438) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3166 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7437) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3167 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7436) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3168 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7435) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3169 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7434) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3170 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7433) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3171 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7432) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3172 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7431) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3173 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7430) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3174 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7429) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3175 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7428) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3176 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7427) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3177 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7426) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3178 (("$" 7441) (T_VARIABLE 7448) (simple_variable 7475) ("{" 7476)) 3179 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 7473) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3180 ((T_PAAMAYIM_NEKUDOTAYIM 7471)) 3181 ((T_NULLSAFE_OBJECT_OPERATOR 7454) (T_OBJECT_OPERATOR 7455) (T_PAAMAYIM_NEKUDOTAYIM 7456) ("[" 7457) ("{" 7458)) 3182 (("$" 7441) (T_STRING 7464) (T_VARIABLE 7448) (property_name 7470) (simple_variable 7466) ("{" 7467)) 3183 (("$" 7441) (T_STRING 7464) (T_VARIABLE 7448) (property_name 7465) (simple_variable 7466) ("{" 7467)) 3184 (("$" 7441) (T_VARIABLE 7448) (simple_variable 7463)) 3185 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 7461) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3186 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7459) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3187 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7460)) 3188 (("]" 7462)) 3189 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7468) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3190 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7469)) 3191 (("$" 7441) (T_VARIABLE 7448) (simple_variable 7472)) 3192 (("%" 280) (")" 7474) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3193 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7477) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3194 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7478)) 3195 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7488) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3196 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 7486) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 3197 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7487) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3198 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 7498) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3199 ((")" 7499)) 3200 (("=" 7500)) 3201 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7501) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3202 (("$" 7319) (T_ABSTRACT 7503) (T_ARRAY 7504) (T_AS 7505) (T_BREAK 7506) (T_CALLABLE 7507) (T_CASE 7508) (T_CATCH 7509) (T_CLASS 7510) (T_CLASS_C 7511) (T_CLONE 7512) (T_CONST 7513) (T_CONTINUE 7514) (T_DECLARE 7515) (T_DEFAULT 7516) (T_DIR 7517) (T_DO 7518) (T_ECHO 7519) (T_ELSE 7520) (T_ELSEIF 7521) (T_EMPTY 7522) (T_ENDDECLARE 7523) (T_ENDFOR 7524) (T_ENDFOREACH 7525) (T_ENDIF 7526) (T_ENDSWITCH 7527) (T_ENDWHILE 7528) (T_ENUM 7529) (T_EVAL 7530) (T_EXIT 7531) (T_EXTENDS 7532) (T_FILE 7533) (T_FINAL 7534) (T_FINALLY 7535) (T_FN 7536) (T_FOR 7537) (T_FOREACH 7538) (T_FUNCTION 7539) (T_FUNC_C 7540) (T_GLOBAL 7541) (T_GOTO 7542) (T_IF 7543) (T_IMPLEMENTS 7544) (T_INCLUDE 7545) (T_INCLUDE_ONCE 7546) (T_INSTANCEOF 7547) (T_INSTEADOF 7548) (T_INTERFACE 7549) (T_ISSET 7550) (T_LINE 7551) (T_LIST 7552) (T_LOGICAL_AND 7553) (T_LOGICAL_OR 7554) (T_LOGICAL_XOR 7555) (T_MATCH 7556) (T_METHOD_C 7557) (T_NAMESPACE 7558) (T_NEW 7559) (T_NS_C 7560) (T_PRINT 7561) (T_PRIVATE 7562) (T_PROTECTED 7563) (T_PUBLIC 7564) (T_READONLY 7565) (T_REQUIRE 7566) (T_REQUIRE_ONCE 7567) (T_RETURN 7568) (T_STATIC 7569) (T_STRING 7570) (T_SWITCH 7571) (T_THROW 7572) (T_TRAIT 7573) (T_TRAIT_C 7574) (T_TRY 7575) (T_UNSET 7576) (T_USE 7577) (T_VAR 7578) (T_VARIABLE 7365) (T_WHILE 7579) (T_YIELD 7580) (identifier 7581) (member_name 7582) (reserved_non_modifiers 7583) (semi_reserved 7584) (simple_variable 7585) ("{" 914)) 3203 (("(" 7586) (argument_list 7587)) 3204 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 7588) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 7589) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 7590) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3205 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 7593) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3206 (("," 1449) (possible_comma 7591)) 3207 ((")" 7592)) 3208 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (ampersand 7623) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7624) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3209 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7622) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3210 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7621) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3211 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7620) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3212 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7619) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3213 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7618) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3214 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7617) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3215 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7616) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3216 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7615) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3217 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7614) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3218 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7613) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3219 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7612) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3220 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7611) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3221 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 7610) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3222 (("\"" 1489) ("$" 7625) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 7626) ("[" 1508) (array_object_dereferenceable 7627) (callable_expr 7628) (callable_variable 7629) (class_constant 1512) (class_name 7630) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 7631) (name 7632) (simple_variable 7633) (static_member 7634) (variable 7635) (variable_class_name 7636)) 3223 (("$" 7625) (T_VARIABLE 7626) (simple_variable 7671) ("{" 7672)) 3224 ((T_NULLSAFE_OBJECT_OPERATOR 7654) (T_OBJECT_OPERATOR 7655) ("[" 7656) ("{" 7657)) 3225 (("(" 7640) (argument_list 7653)) 3226 ((T_PAAMAYIM_NEKUDOTAYIM 7649)) 3227 (("(" 7640) (argument_list 7648)) 3228 ((T_PAAMAYIM_NEKUDOTAYIM 7637)) 3229 (("$" 7625) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 7626) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 7638) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 7639) ("{" 914)) 3230 (("(" 7640) (argument_list 7641)) 3231 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 7642) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 7643) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 7644) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3232 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 7647) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3233 (("," 1449) (possible_comma 7645)) 3234 ((")" 7646)) 3235 (("$" 7625) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 7626) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 7650) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 7651) ("{" 914)) 3236 (("(" 7640) (argument_list 7652)) 3237 (("$" 7625) (T_STRING 7662) (T_VARIABLE 7626) (property_name 7669) (simple_variable 7664) ("{" 7665)) 3238 (("$" 7625) (T_STRING 7662) (T_VARIABLE 7626) (property_name 7663) (simple_variable 7664) ("{" 7665)) 3239 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 7660) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3240 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7658) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3241 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7659)) 3242 (("]" 7661)) 3243 (("(" 7640) (argument_list 7668)) 3244 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7666) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3245 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7667)) 3246 (("(" 7640) (argument_list 7670)) 3247 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7673) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3248 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7674)) 3249 ((backup_doc_comment 7677)) 3250 (("(" 7678)) 3251 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 7679)) 3252 ((")" 7680)) 3253 ((T_USE 148) (lexical_vars 7681)) 3254 ((":" 150) (return_type 7682)) 3255 ((backup_fn_flags 7683)) 3256 (("{" 7684)) 3257 ((inner_statement_list 7685)) 3258 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 7686) ("~" 124)) 3259 ((backup_fn_flags 7687)) 3260 ((backup_doc_comment 7689)) 3261 (("(" 7690)) 3262 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 7691)) 3263 ((")" 7692)) 3264 ((":" 2207) (return_type 7693)) 3265 ((T_DOUBLE_ARROW 7694)) 3266 ((backup_fn_flags 7695)) 3267 ((backup_lex_pos 7696)) 3268 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7758) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3269 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8217) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3270 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 8215) (encaps_var 5158)) 3271 (("$" 7699) (T_VARIABLE 7745) (simple_variable 8211) ("{" 8212)) 3272 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8209) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3273 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8208) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3274 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8207) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3275 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8206) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3276 (("(" 8203)) 3277 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8202) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3278 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8201) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3279 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8200) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3280 (("\"" 1489) ("$" 7998) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 7999) ("[" 1508) (array_object_dereferenceable 8000) (callable_expr 8001) (callable_variable 8002) (class_constant 1512) (class_name 8003) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 8004) (name 8005) (simple_variable 8006) (static_member 8007) (variable 8199) (variable_class_name 8009)) 3281 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8198) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3282 (("(" 8195)) 3283 (("(" 8192)) 3284 (("(" 8188) (exit_expr 8189)) 3285 (("\"" 1489) ("$" 7998) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 7999) ("[" 1508) (array_object_dereferenceable 8000) (callable_expr 8001) (callable_variable 8002) (class_constant 1512) (class_name 8003) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 8004) (name 8005) (simple_variable 8006) (static_member 8007) (variable 8187) (variable_class_name 8009)) 3286 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8186) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3287 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8185) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3288 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8184) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3289 (("(" 8180)) 3290 (("(" 8175)) 3291 (("(" 8169)) 3292 (("$" 7998) ("(" 8122) (T_ATTRIBUTE 2853) (T_CLASS 8123) (T_NAME_FULLY_QUALIFIED 8124) (T_NAME_QUALIFIED 8125) (T_NAME_RELATIVE 8126) (T_STATIC 8127) (T_STRING 8128) (T_VARIABLE 7999) (anonymous_class 8129) (attribute 2862) (attributes 8130) (class_name 8131) (class_name_reference 8132) (name 8133) (new_variable 8134) (simple_variable 8135)) 3293 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8121) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3294 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8120) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3295 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8119) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3296 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8118) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3297 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 8113) (T_END_HEREDOC 8114) (T_VARIABLE 2822) (encaps_list 8115) (encaps_var 2824)) 3298 ((T_FN 35) (T_FUNCTION 126) (fn 7759) #'7760 (inline_function 8112)) 3299 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8111) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3300 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8110) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3301 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8109) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3302 (("!" 6795) ("\"" 6796) ("$" 6797) ("(" 6798) ("+" 6799) ("-" 6800) ("@" 6801) (T_ARRAY 6802) (T_ARRAY_CAST 6803) (T_ATTRIBUTE 125) (T_BOOL_CAST 6804) (T_CLASS_C 6805) (T_CLONE 6806) (T_CONSTANT_ENCAPSED_STRING 6807) (T_DEC 6808) (T_DIR 6809) (T_DNUMBER 6810) (T_DOUBLE_CAST 6811) (T_EMPTY 6812) (T_EVAL 6813) (T_EXIT 6814) (T_FILE 6815) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 6816) (T_INC 6817) (T_INCLUDE 6818) (T_INCLUDE_ONCE 6819) (T_INT_CAST 6820) (T_ISSET 6821) (T_LINE 6822) (T_LIST 6823) (T_LNUMBER 6824) (T_MATCH 6825) (T_METHOD_C 6826) (T_NAME_FULLY_QUALIFIED 6827) (T_NAME_QUALIFIED 6828) (T_NAME_RELATIVE 6829) (T_NEW 6830) (T_NS_C 6831) (T_OBJECT_CAST 6832) (T_PRINT 6833) (T_REQUIRE 6834) (T_REQUIRE_ONCE 6835) (T_START_HEREDOC 6836) (T_STATIC 6837) (T_STRING 6838) (T_STRING_CAST 6839) (T_THROW 6840) (T_TRAIT_C 6841) (T_UNSET_CAST 6842) (T_VARIABLE 6843) (T_YIELD 6844) (T_YIELD_FROM 6845) ("[" 6846) ("`" 6847) (array_object_dereferenceable 6848) (attribute 129) (attributes 6849) (callable_expr 6850) (callable_variable 6851) (class_constant 6852) (class_name 6853) (constant 6854) (dereferenceable_scalar 6855) (expr 8106) (fn 6857) (fully_dereferenceable 103) #'6858 (function_call 6859) (inline_function 6860) (internal_functions_in_yacc 6861) (match 6862) (name 6863) (new_expr 6864) (scalar 6865) (simple_variable 6866) (static_member 6867) (variable 6868) (variable_class_name 6869) ("~" 6870)) 3303 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8105) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3304 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 8101) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 3305 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 8099) (encaps_list 2729) (encaps_var 2730)) 3306 ((T_NULLSAFE_OBJECT_OPERATOR 8082) (T_OBJECT_OPERATOR 8083) ("[" 8084) ("{" 8085)) 3307 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 8079) (attribute 2654) (fn 7759) #'7760 (inline_function 8080)) 3308 (("(" 7959) (argument_list 8078)) 3309 ((T_PAAMAYIM_NEKUDOTAYIM 8073)) 3310 (("%" 7774) ("*" 7775) ("+" 7776) ("-" 7777) ("." 7778) ("/" 7779) ("<" 7780) (">" 7781) ("?" 7782) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 7783) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7784) (T_BOOLEAN_AND 7785) (T_BOOLEAN_OR 7786) (T_COALESCE 7787) (T_INSTANCEOF 7788) (T_IS_EQUAL 7789) (T_IS_GREATER_OR_EQUAL 7790) (T_IS_IDENTICAL 7791) (T_IS_NOT_EQUAL 7792) (T_IS_NOT_IDENTICAL 7793) (T_IS_SMALLER_OR_EQUAL 7794) (T_LOGICAL_AND 7795) (T_LOGICAL_OR 7796) (T_LOGICAL_XOR 7797) (T_POW 7798) (T_SL 7799) (T_SPACESHIP 7800) (T_SR 7801) ("^" 7802) (backup_fn_flags 8072) ("|" 7803)) 3311 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 8061)) 3312 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 8049)) 3313 (("(" 7959) (argument_list 8048)) 3314 (("=" 7967) (T_AND_EQUAL 7968) (T_COALESCE_EQUAL 7969) (T_CONCAT_EQUAL 7970) (T_DEC 7971) (T_DIV_EQUAL 7972) (T_INC 7973) (T_MINUS_EQUAL 7974) (T_MOD_EQUAL 7975) (T_MUL_EQUAL 7976) (T_OR_EQUAL 7977) (T_PLUS_EQUAL 7978) (T_POW_EQUAL 7979) (T_SL_EQUAL 7980) (T_SR_EQUAL 7981) (T_XOR_EQUAL 7982)) 3315 ((T_PAAMAYIM_NEKUDOTAYIM 7875)) 3316 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7773) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3317 (("%" 7774) ("*" 7775) ("+" 7776) ("-" 7777) ("." 7778) ("/" 7779) ("<" 7780) (">" 7781) ("?" 7782) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 7783) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7784) (T_BOOLEAN_AND 7785) (T_BOOLEAN_OR 7786) (T_COALESCE 7787) (T_INSTANCEOF 7788) (T_IS_EQUAL 7789) (T_IS_GREATER_OR_EQUAL 7790) (T_IS_IDENTICAL 7791) (T_IS_NOT_EQUAL 7792) (T_IS_NOT_IDENTICAL 7793) (T_IS_SMALLER_OR_EQUAL 7794) (T_LOGICAL_AND 7795) (T_LOGICAL_OR 7796) (T_LOGICAL_XOR 7797) (T_POW 7798) (T_SL 7799) (T_SPACESHIP 7800) (T_SR 7801) ("^" 7802) ("|" 7803)) 3318 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7874) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3319 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7873) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3320 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7872) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3321 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7871) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3322 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7870) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3323 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7869) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3324 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7868) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3325 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7867) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3326 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 7862) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 7863) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3327 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7861) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3328 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7860) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3329 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7859) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3330 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7858) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3331 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7857) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3332 (("$" 7819) ("(" 7820) (T_NAME_FULLY_QUALIFIED 7821) (T_NAME_QUALIFIED 7822) (T_NAME_RELATIVE 7823) (T_STATIC 7824) (T_STRING 7825) (T_VARIABLE 7826) (class_name 7827) (class_name_reference 7828) (name 7829) (new_variable 7830) (simple_variable 7831)) 3333 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7818) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3334 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7817) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3335 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7816) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3336 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7815) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3337 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7814) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3338 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7813) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3339 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7812) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3340 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7811) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3341 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7810) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3342 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7809) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3343 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7808) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3344 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7807) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3345 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7806) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3346 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7805) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3347 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7804) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3348 (("$" 7819) (T_VARIABLE 7826) (simple_variable 7853) ("{" 7854)) 3349 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 7851) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3350 ((T_PAAMAYIM_NEKUDOTAYIM 7849)) 3351 ((T_NULLSAFE_OBJECT_OPERATOR 7832) (T_OBJECT_OPERATOR 7833) (T_PAAMAYIM_NEKUDOTAYIM 7834) ("[" 7835) ("{" 7836)) 3352 (("$" 7819) (T_STRING 7842) (T_VARIABLE 7826) (property_name 7848) (simple_variable 7844) ("{" 7845)) 3353 (("$" 7819) (T_STRING 7842) (T_VARIABLE 7826) (property_name 7843) (simple_variable 7844) ("{" 7845)) 3354 (("$" 7819) (T_VARIABLE 7826) (simple_variable 7841)) 3355 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 7839) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3356 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7837) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3357 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7838)) 3358 (("]" 7840)) 3359 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7846) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3360 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7847)) 3361 (("$" 7819) (T_VARIABLE 7826) (simple_variable 7850)) 3362 (("%" 280) (")" 7852) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3363 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 7855) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3364 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 7856)) 3365 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7866) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3366 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 7864) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 3367 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7865) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3368 (("$" 7699) (T_ABSTRACT 7876) (T_ARRAY 7877) (T_AS 7878) (T_BREAK 7879) (T_CALLABLE 7880) (T_CASE 7881) (T_CATCH 7882) (T_CLASS 7883) (T_CLASS_C 7884) (T_CLONE 7885) (T_CONST 7886) (T_CONTINUE 7887) (T_DECLARE 7888) (T_DEFAULT 7889) (T_DIR 7890) (T_DO 7891) (T_ECHO 7892) (T_ELSE 7893) (T_ELSEIF 7894) (T_EMPTY 7895) (T_ENDDECLARE 7896) (T_ENDFOR 7897) (T_ENDFOREACH 7898) (T_ENDIF 7899) (T_ENDSWITCH 7900) (T_ENDWHILE 7901) (T_ENUM 7902) (T_EVAL 7903) (T_EXIT 7904) (T_EXTENDS 7905) (T_FILE 7906) (T_FINAL 7907) (T_FINALLY 7908) (T_FN 7909) (T_FOR 7910) (T_FOREACH 7911) (T_FUNCTION 7912) (T_FUNC_C 7913) (T_GLOBAL 7914) (T_GOTO 7915) (T_IF 7916) (T_IMPLEMENTS 7917) (T_INCLUDE 7918) (T_INCLUDE_ONCE 7919) (T_INSTANCEOF 7920) (T_INSTEADOF 7921) (T_INTERFACE 7922) (T_ISSET 7923) (T_LINE 7924) (T_LIST 7925) (T_LOGICAL_AND 7926) (T_LOGICAL_OR 7927) (T_LOGICAL_XOR 7928) (T_MATCH 7929) (T_METHOD_C 7930) (T_NAMESPACE 7931) (T_NEW 7932) (T_NS_C 7933) (T_PRINT 7934) (T_PRIVATE 7935) (T_PROTECTED 7936) (T_PUBLIC 7937) (T_READONLY 7938) (T_REQUIRE 7939) (T_REQUIRE_ONCE 7940) (T_RETURN 7941) (T_STATIC 7942) (T_STRING 7943) (T_SWITCH 7944) (T_THROW 7945) (T_TRAIT 7946) (T_TRAIT_C 7947) (T_TRY 7948) (T_UNSET 7949) (T_USE 7950) (T_VAR 7951) (T_VARIABLE 7745) (T_WHILE 7952) (T_YIELD 7953) (identifier 7954) (member_name 7955) (reserved_non_modifiers 7956) (semi_reserved 7957) (simple_variable 7958) ("{" 914)) 3369 (("(" 7959) (argument_list 7960)) 3370 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 7961) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 7962) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 7963) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3371 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 7966) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3372 (("," 1449) (possible_comma 7964)) 3373 ((")" 7965)) 3374 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (ampersand 7996) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7997) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3375 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7995) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3376 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7994) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3377 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7993) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3378 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7992) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3379 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7991) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3380 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7990) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3381 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7989) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3382 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7988) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3383 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7987) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3384 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7986) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3385 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7985) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3386 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7984) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3387 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 7983) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3388 (("\"" 1489) ("$" 7998) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 7999) ("[" 1508) (array_object_dereferenceable 8000) (callable_expr 8001) (callable_variable 8002) (class_constant 1512) (class_name 8003) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 8004) (name 8005) (simple_variable 8006) (static_member 8007) (variable 8008) (variable_class_name 8009)) 3389 (("$" 7998) (T_VARIABLE 7999) (simple_variable 8044) ("{" 8045)) 3390 ((T_NULLSAFE_OBJECT_OPERATOR 8027) (T_OBJECT_OPERATOR 8028) ("[" 8029) ("{" 8030)) 3391 (("(" 8013) (argument_list 8026)) 3392 ((T_PAAMAYIM_NEKUDOTAYIM 8022)) 3393 (("(" 8013) (argument_list 8021)) 3394 ((T_PAAMAYIM_NEKUDOTAYIM 8010)) 3395 (("$" 7998) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 7999) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 8011) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 8012) ("{" 914)) 3396 (("(" 8013) (argument_list 8014)) 3397 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8015) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 8016) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 8017) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3398 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8020) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3399 (("," 1449) (possible_comma 8018)) 3400 ((")" 8019)) 3401 (("$" 7998) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 7999) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 8023) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 8024) ("{" 914)) 3402 (("(" 8013) (argument_list 8025)) 3403 (("$" 7998) (T_STRING 8035) (T_VARIABLE 7999) (property_name 8042) (simple_variable 8037) ("{" 8038)) 3404 (("$" 7998) (T_STRING 8035) (T_VARIABLE 7999) (property_name 8036) (simple_variable 8037) ("{" 8038)) 3405 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 8033) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3406 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8031) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3407 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8032)) 3408 (("]" 8034)) 3409 (("(" 8013) (argument_list 8041)) 3410 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8039) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3411 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8040)) 3412 (("(" 8013) (argument_list 8043)) 3413 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8046) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3414 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8047)) 3415 ((backup_doc_comment 8050)) 3416 (("(" 8051)) 3417 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 8052)) 3418 ((")" 8053)) 3419 ((T_USE 148) (lexical_vars 8054)) 3420 ((":" 150) (return_type 8055)) 3421 ((backup_fn_flags 8056)) 3422 (("{" 8057)) 3423 ((inner_statement_list 8058)) 3424 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 8059) ("~" 124)) 3425 ((backup_fn_flags 8060)) 3426 ((backup_doc_comment 8062)) 3427 (("(" 8063)) 3428 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 8064)) 3429 ((")" 8065)) 3430 ((":" 2207) (return_type 8066)) 3431 ((T_DOUBLE_ARROW 8067)) 3432 ((backup_fn_flags 8068)) 3433 ((backup_lex_pos 8069)) 3434 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8070) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3435 (("%" 7774) ("*" 7775) ("+" 7776) ("-" 7777) ("." 7778) ("/" 7779) ("<" 7780) (">" 7781) ("?" 7782) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 7783) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7784) (T_BOOLEAN_AND 7785) (T_BOOLEAN_OR 7786) (T_COALESCE 7787) (T_INSTANCEOF 7788) (T_IS_EQUAL 7789) (T_IS_GREATER_OR_EQUAL 7790) (T_IS_IDENTICAL 7791) (T_IS_NOT_EQUAL 7792) (T_IS_NOT_IDENTICAL 7793) (T_IS_SMALLER_OR_EQUAL 7794) (T_LOGICAL_AND 7795) (T_LOGICAL_OR 7796) (T_LOGICAL_XOR 7797) (T_POW 7798) (T_SL 7799) (T_SPACESHIP 7800) (T_SR 7801) ("^" 7802) (backup_fn_flags 8071) ("|" 7803)) 3436 (("$" 7699) (T_ABSTRACT 7876) (T_ARRAY 7877) (T_AS 7878) (T_BREAK 7879) (T_CALLABLE 7880) (T_CASE 7881) (T_CATCH 7882) (T_CLASS 7883) (T_CLASS_C 7884) (T_CLONE 7885) (T_CONST 7886) (T_CONTINUE 7887) (T_DECLARE 7888) (T_DEFAULT 7889) (T_DIR 7890) (T_DO 7891) (T_ECHO 7892) (T_ELSE 7893) (T_ELSEIF 7894) (T_EMPTY 7895) (T_ENDDECLARE 7896) (T_ENDFOR 7897) (T_ENDFOREACH 7898) (T_ENDIF 7899) (T_ENDSWITCH 7900) (T_ENDWHILE 7901) (T_ENUM 7902) (T_EVAL 7903) (T_EXIT 7904) (T_EXTENDS 7905) (T_FILE 7906) (T_FINAL 7907) (T_FINALLY 7908) (T_FN 7909) (T_FOR 7910) (T_FOREACH 7911) (T_FUNCTION 7912) (T_FUNC_C 7913) (T_GLOBAL 7914) (T_GOTO 7915) (T_IF 7916) (T_IMPLEMENTS 7917) (T_INCLUDE 7918) (T_INCLUDE_ONCE 7919) (T_INSTANCEOF 7920) (T_INSTEADOF 7921) (T_INTERFACE 7922) (T_ISSET 7923) (T_LINE 7924) (T_LIST 7925) (T_LOGICAL_AND 7926) (T_LOGICAL_OR 7927) (T_LOGICAL_XOR 7928) (T_MATCH 7929) (T_METHOD_C 7930) (T_NAMESPACE 7931) (T_NEW 7932) (T_NS_C 7933) (T_PRINT 7934) (T_PRIVATE 7935) (T_PROTECTED 7936) (T_PUBLIC 7937) (T_READONLY 7938) (T_REQUIRE 7939) (T_REQUIRE_ONCE 7940) (T_RETURN 7941) (T_STATIC 7942) (T_STRING 7943) (T_SWITCH 7944) (T_THROW 7945) (T_TRAIT 7946) (T_TRAIT_C 7947) (T_TRY 7948) (T_UNSET 7949) (T_USE 7950) (T_VAR 7951) (T_VARIABLE 7745) (T_WHILE 7952) (T_YIELD 7953) (identifier 8074) (member_name 8075) (reserved_non_modifiers 7956) (semi_reserved 7957) (simple_variable 8076) ("{" 914)) 3437 (("(" 7959) (argument_list 8077)) 3438 ((T_FN 35) (T_FUNCTION 126) (fn 7759) #'7760 (inline_function 8081)) 3439 (("$" 7699) (T_STRING 8090) (T_VARIABLE 7745) (property_name 8097) (simple_variable 8092) ("{" 8093)) 3440 (("$" 7699) (T_STRING 8090) (T_VARIABLE 7745) (property_name 8091) (simple_variable 8092) ("{" 8093)) 3441 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 8088) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3442 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8086) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3443 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8087)) 3444 (("]" 8089)) 3445 (("(" 7959) (argument_list 8096)) 3446 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8094) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3447 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8095)) 3448 (("(" 7959) (argument_list 8098)) 3449 (("`" 8100)) 3450 (("]" 8102)) 3451 (("=" 8103)) 3452 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8104) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3453 (("%" 6872) ("*" 6873) ("+" 6874) ("-" 6875) ("." 6876) ("/" 6877) ("<" 6878) (">" 6879) ("?" 6880) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6881) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6882) (T_BOOLEAN_AND 6883) (T_BOOLEAN_OR 6884) (T_COALESCE 6885) (T_DOUBLE_ARROW 8107) (T_INSTANCEOF 6886) (T_IS_EQUAL 6887) (T_IS_GREATER_OR_EQUAL 6888) (T_IS_IDENTICAL 6889) (T_IS_NOT_EQUAL 6890) (T_IS_NOT_IDENTICAL 6891) (T_IS_SMALLER_OR_EQUAL 6892) (T_LOGICAL_AND 6893) (T_LOGICAL_OR 6894) (T_LOGICAL_XOR 6895) (T_POW 6896) (T_SL 6897) (T_SPACESHIP 6898) (T_SR 6899) ("^" 6900) ("|" 6901)) 3454 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8108) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3455 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 8117) (T_VARIABLE 2822) (encaps_var 2836)) 3456 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 8116) (T_VARIABLE 2822) (encaps_var 2827)) 3457 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8167) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3458 (("(" 2899) (argument_list 2900) (ctor_arguments 8160)) 3459 ((T_ATTRIBUTE 2853) (T_CLASS 8123) (anonymous_class 8159) (attribute 2898)) 3460 ((T_PAAMAYIM_NEKUDOTAYIM 8157)) 3461 (("(" 8148) (argument_list 8149) (ctor_arguments 8150)) 3462 ((T_NULLSAFE_OBJECT_OPERATOR 8136) (T_OBJECT_OPERATOR 8137) (T_PAAMAYIM_NEKUDOTAYIM 8138) ("[" 8139) ("{" 8140)) 3463 (("$" 7998) (T_STRING 8035) (T_VARIABLE 7999) (property_name 8147) (simple_variable 8037) ("{" 8038)) 3464 (("$" 7998) (T_STRING 8035) (T_VARIABLE 7999) (property_name 8146) (simple_variable 8037) ("{" 8038)) 3465 (("$" 7998) (T_VARIABLE 7999) (simple_variable 8145)) 3466 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 8143) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3467 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8141) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3468 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8142)) 3469 (("]" 8144)) 3470 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8151) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 8152) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 8153) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3471 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8156) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3472 (("," 1449) (possible_comma 8154)) 3473 ((")" 8155)) 3474 (("$" 7998) (T_VARIABLE 7999) (simple_variable 8158)) 3475 ((T_EXTENDS 2902) (extends_from 8161)) 3476 ((T_IMPLEMENTS 2904) (implements_list 8162)) 3477 ((backup_doc_comment 8163)) 3478 (("{" 8164)) 3479 ((class_statement_list 8165)) 3480 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 8166)) 3481 (("%" 280) (")" 8168) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3482 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8170) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3483 (("%" 280) (")" 8171) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3484 (("{" 8172)) 3485 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 8173) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 3486 (("}" 8174)) 3487 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 8176) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3488 ((")" 8177)) 3489 (("=" 8178)) 3490 (("!" 7697) ("\"" 7698) ("$" 7699) ("(" 7700) ("+" 7701) ("-" 7702) ("@" 7703) (T_ARRAY 7704) (T_ARRAY_CAST 7705) (T_ATTRIBUTE 125) (T_BOOL_CAST 7706) (T_CLASS_C 7707) (T_CLONE 7708) (T_CONSTANT_ENCAPSED_STRING 7709) (T_DEC 7710) (T_DIR 7711) (T_DNUMBER 7712) (T_DOUBLE_CAST 7713) (T_EMPTY 7714) (T_EVAL 7715) (T_EXIT 7716) (T_FILE 7717) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7718) (T_INC 7719) (T_INCLUDE 7720) (T_INCLUDE_ONCE 7721) (T_INT_CAST 7722) (T_ISSET 7723) (T_LINE 7724) (T_LIST 7725) (T_LNUMBER 7726) (T_MATCH 7727) (T_METHOD_C 7728) (T_NAME_FULLY_QUALIFIED 7729) (T_NAME_QUALIFIED 7730) (T_NAME_RELATIVE 7731) (T_NEW 7732) (T_NS_C 7733) (T_OBJECT_CAST 7734) (T_PRINT 7735) (T_REQUIRE 7736) (T_REQUIRE_ONCE 7737) (T_START_HEREDOC 7738) (T_STATIC 7739) (T_STRING 7740) (T_STRING_CAST 7741) (T_THROW 7742) (T_TRAIT_C 7743) (T_UNSET_CAST 7744) (T_VARIABLE 7745) (T_YIELD 7746) (T_YIELD_FROM 7747) ("[" 7748) ("`" 7749) (array_object_dereferenceable 7750) (attribute 129) (attributes 7751) (callable_expr 7752) (callable_variable 7753) (class_constant 7754) (class_name 7755) (constant 7756) (dereferenceable_scalar 7757) (expr 8179) (fn 7759) (fully_dereferenceable 103) #'7760 (function_call 7761) (inline_function 7762) (internal_functions_in_yacc 7763) (match 7764) (name 7765) (new_expr 7766) (scalar 7767) (simple_variable 7768) (static_member 7769) (variable 7770) (variable_class_name 7771) ("~" 7772)) 3491 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 8181) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3492 (("," 5117) (possible_comma 8182)) 3493 ((")" 8183)) 3494 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 8190) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3495 ((")" 8191)) 3496 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8193) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3497 (("%" 280) (")" 8194) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3498 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8196) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3499 (("%" 280) (")" 8197) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3500 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 8204) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3501 ((")" 8205)) 3502 (("%" 280) (")" 8210) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3503 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8213) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3504 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8214)) 3505 (("\"" 8216) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 3506 (("$" 7319) (T_ABSTRACT 7503) (T_ARRAY 7504) (T_AS 7505) (T_BREAK 7506) (T_CALLABLE 7507) (T_CASE 7508) (T_CATCH 7509) (T_CLASS 7510) (T_CLASS_C 7511) (T_CLONE 7512) (T_CONST 7513) (T_CONTINUE 7514) (T_DECLARE 7515) (T_DEFAULT 7516) (T_DIR 7517) (T_DO 7518) (T_ECHO 7519) (T_ELSE 7520) (T_ELSEIF 7521) (T_EMPTY 7522) (T_ENDDECLARE 7523) (T_ENDFOR 7524) (T_ENDFOREACH 7525) (T_ENDIF 7526) (T_ENDSWITCH 7527) (T_ENDWHILE 7528) (T_ENUM 7529) (T_EVAL 7530) (T_EXIT 7531) (T_EXTENDS 7532) (T_FILE 7533) (T_FINAL 7534) (T_FINALLY 7535) (T_FN 7536) (T_FOR 7537) (T_FOREACH 7538) (T_FUNCTION 7539) (T_FUNC_C 7540) (T_GLOBAL 7541) (T_GOTO 7542) (T_IF 7543) (T_IMPLEMENTS 7544) (T_INCLUDE 7545) (T_INCLUDE_ONCE 7546) (T_INSTANCEOF 7547) (T_INSTEADOF 7548) (T_INTERFACE 7549) (T_ISSET 7550) (T_LINE 7551) (T_LIST 7552) (T_LOGICAL_AND 7553) (T_LOGICAL_OR 7554) (T_LOGICAL_XOR 7555) (T_MATCH 7556) (T_METHOD_C 7557) (T_NAMESPACE 7558) (T_NEW 7559) (T_NS_C 7560) (T_PRINT 7561) (T_PRIVATE 7562) (T_PROTECTED 7563) (T_PUBLIC 7564) (T_READONLY 7565) (T_REQUIRE 7566) (T_REQUIRE_ONCE 7567) (T_RETURN 7568) (T_STATIC 7569) (T_STRING 7570) (T_SWITCH 7571) (T_THROW 7572) (T_TRAIT 7573) (T_TRAIT_C 7574) (T_TRY 7575) (T_UNSET 7576) (T_USE 7577) (T_VAR 7578) (T_VARIABLE 7365) (T_WHILE 7579) (T_YIELD 7580) (identifier 8219) (member_name 8220) (reserved_non_modifiers 7583) (semi_reserved 7584) (simple_variable 8221) ("{" 914)) 3507 (("(" 7586) (argument_list 8222)) 3508 ((T_FN 35) (T_FUNCTION 126) (fn 7380) #'7381 (inline_function 8226)) 3509 (("$" 7319) (T_STRING 8235) (T_VARIABLE 7365) (property_name 8242) (simple_variable 8237) ("{" 8238)) 3510 (("$" 7319) (T_STRING 8235) (T_VARIABLE 7365) (property_name 8236) (simple_variable 8237) ("{" 8238)) 3511 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 8233) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3512 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8231) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3513 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8232)) 3514 (("]" 8234)) 3515 (("(" 7586) (argument_list 8241)) 3516 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8239) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3517 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8240)) 3518 (("(" 7586) (argument_list 8243)) 3519 (("$" 8244) (T_VARIABLE 8245) (simple_variable 8290) ("{" 8291)) 3520 ((T_NULLSAFE_OBJECT_OPERATOR 8273) (T_OBJECT_OPERATOR 8274) ("[" 8275) ("{" 8276)) 3521 (("(" 8259) (argument_list 8272)) 3522 ((T_PAAMAYIM_NEKUDOTAYIM 8268)) 3523 (("(" 8259) (argument_list 8267)) 3524 ((T_PAAMAYIM_NEKUDOTAYIM 8256)) 3525 (("$" 8244) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 8245) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 8257) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 8258) ("{" 914)) 3526 (("(" 8259) (argument_list 8260)) 3527 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8261) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 8262) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 8263) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3528 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8266) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3529 (("," 1449) (possible_comma 8264)) 3530 ((")" 8265)) 3531 (("$" 8244) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 8245) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 8269) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 8270) ("{" 914)) 3532 (("(" 8259) (argument_list 8271)) 3533 (("$" 8244) (T_STRING 8281) (T_VARIABLE 8245) (property_name 8288) (simple_variable 8283) ("{" 8284)) 3534 (("$" 8244) (T_STRING 8281) (T_VARIABLE 8245) (property_name 8282) (simple_variable 8283) ("{" 8284)) 3535 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 8279) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3536 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8277) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3537 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8278)) 3538 (("]" 8280)) 3539 (("(" 8259) (argument_list 8287)) 3540 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8285) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3541 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8286)) 3542 (("(" 8259) (argument_list 8289)) 3543 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8292) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3544 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8293)) 3545 (("`" 8295)) 3546 (("]" 8297)) 3547 (("=" 8298)) 3548 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8299) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3549 (("%" 1722) ("*" 1723) ("+" 1724) ("-" 1725) ("." 1726) ("/" 1727) ("<" 1728) (">" 1729) ("?" 1730) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1731) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1732) (T_BOOLEAN_AND 1733) (T_BOOLEAN_OR 1734) (T_COALESCE 1735) (T_DOUBLE_ARROW 8302) (T_INSTANCEOF 1736) (T_IS_EQUAL 1737) (T_IS_GREATER_OR_EQUAL 1738) (T_IS_IDENTICAL 1739) (T_IS_NOT_EQUAL 1740) (T_IS_NOT_IDENTICAL 1741) (T_IS_SMALLER_OR_EQUAL 1742) (T_LOGICAL_AND 1743) (T_LOGICAL_OR 1744) (T_LOGICAL_XOR 1745) (T_POW 1746) (T_SL 1747) (T_SPACESHIP 1748) (T_SR 1749) ("^" 1750) ("|" 1751)) 3550 (("!" 7317) ("\"" 7318) ("$" 7319) ("(" 7320) ("+" 7321) ("-" 7322) ("@" 7323) (T_ARRAY 7324) (T_ARRAY_CAST 7325) (T_ATTRIBUTE 125) (T_BOOL_CAST 7326) (T_CLASS_C 7327) (T_CLONE 7328) (T_CONSTANT_ENCAPSED_STRING 7329) (T_DEC 7330) (T_DIR 7331) (T_DNUMBER 7332) (T_DOUBLE_CAST 7333) (T_EMPTY 7334) (T_EVAL 7335) (T_EXIT 7336) (T_FILE 7337) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 7338) (T_INC 7339) (T_INCLUDE 7340) (T_INCLUDE_ONCE 7341) (T_INT_CAST 7342) (T_ISSET 7343) (T_LINE 7344) (T_LIST 7394) (T_LNUMBER 7346) (T_MATCH 7347) (T_METHOD_C 7348) (T_NAME_FULLY_QUALIFIED 7349) (T_NAME_QUALIFIED 7350) (T_NAME_RELATIVE 7351) (T_NEW 7352) (T_NS_C 7353) (T_OBJECT_CAST 7354) (T_PRINT 7355) (T_REQUIRE 7356) (T_REQUIRE_ONCE 7357) (T_START_HEREDOC 7358) (T_STATIC 7359) (T_STRING 7360) (T_STRING_CAST 7361) (T_THROW 7362) (T_TRAIT_C 7363) (T_UNSET_CAST 7364) (T_VARIABLE 7365) (T_YIELD 7366) (T_YIELD_FROM 7367) ("[" 7368) ("`" 7369) (array_object_dereferenceable 7371) (attribute 129) (attributes 7372) (callable_expr 7373) (callable_variable 7374) (class_constant 7375) (class_name 7376) (constant 7377) (dereferenceable_scalar 7378) (expr 8303) (fn 7380) (fully_dereferenceable 103) #'7381 (function_call 7382) (inline_function 7383) (internal_functions_in_yacc 7384) (match 7385) (name 7386) (new_expr 7387) (scalar 7388) (simple_variable 7389) (static_member 7390) (variable 7391) (variable_class_name 7392) ("~" 7393)) 3551 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 8312) (T_VARIABLE 2822) (encaps_var 2836)) 3552 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 8311) (T_VARIABLE 2822) (encaps_var 2827)) 3553 (("(" 2899) (argument_list 2900) (ctor_arguments 8331)) 3554 ((T_ATTRIBUTE 2853) (T_CLASS 8317) (anonymous_class 8330) (attribute 2898)) 3555 (("(" 8321) (argument_list 8322) (ctor_arguments 8323)) 3556 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8324) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 8325) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 8326) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3557 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8329) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3558 (("," 1449) (possible_comma 8327)) 3559 ((")" 8328)) 3560 ((T_EXTENDS 2902) (extends_from 8332)) 3561 ((T_IMPLEMENTS 2904) (implements_list 8333)) 3562 ((backup_doc_comment 8334)) 3563 (("{" 8335)) 3564 ((class_statement_list 8336)) 3565 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 8337)) 3566 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8339) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3567 (("%" 280) (")" 8340) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3568 (("{" 8341)) 3569 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 8342) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 3570 (("}" 8343)) 3571 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 8345) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3572 ((")" 8346)) 3573 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 8348) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3574 (("," 5117) (possible_comma 8349)) 3575 ((")" 8350)) 3576 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 8357) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3577 ((")" 8358)) 3578 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8360) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3579 (("%" 280) (")" 8361) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3580 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8363) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3581 (("%" 280) (")" 8364) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3582 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 8371) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3583 ((")" 8372)) 3584 (("%" 280) (")" 8377) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3585 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8380) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3586 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8381)) 3587 (("\"" 8383) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 3588 (("$" 1640) (T_ABSTRACT 6599) (T_ARRAY 6600) (T_AS 6601) (T_BREAK 6602) (T_CALLABLE 6603) (T_CASE 6604) (T_CATCH 6605) (T_CLASS 6606) (T_CLASS_C 6607) (T_CLONE 6608) (T_CONST 6609) (T_CONTINUE 6610) (T_DECLARE 6611) (T_DEFAULT 6612) (T_DIR 6613) (T_DO 6614) (T_ECHO 6615) (T_ELSE 6616) (T_ELSEIF 6617) (T_EMPTY 6618) (T_ENDDECLARE 6619) (T_ENDFOR 6620) (T_ENDFOREACH 6621) (T_ENDIF 6622) (T_ENDSWITCH 6623) (T_ENDWHILE 6624) (T_ENUM 6625) (T_EVAL 6626) (T_EXIT 6627) (T_EXTENDS 6628) (T_FILE 6629) (T_FINAL 6630) (T_FINALLY 6631) (T_FN 6632) (T_FOR 6633) (T_FOREACH 6634) (T_FUNCTION 6635) (T_FUNC_C 6636) (T_GLOBAL 6637) (T_GOTO 6638) (T_IF 6639) (T_IMPLEMENTS 6640) (T_INCLUDE 6641) (T_INCLUDE_ONCE 6642) (T_INSTANCEOF 6643) (T_INSTEADOF 6644) (T_INTERFACE 6645) (T_ISSET 6646) (T_LINE 6647) (T_LIST 6648) (T_LOGICAL_AND 6649) (T_LOGICAL_OR 6650) (T_LOGICAL_XOR 6651) (T_MATCH 6652) (T_METHOD_C 6653) (T_NAMESPACE 6654) (T_NEW 6655) (T_NS_C 6656) (T_PRINT 6657) (T_PRIVATE 6658) (T_PROTECTED 6659) (T_PUBLIC 6660) (T_READONLY 6661) (T_REQUIRE 6662) (T_REQUIRE_ONCE 6663) (T_RETURN 6664) (T_STATIC 6665) (T_STRING 6666) (T_SWITCH 6667) (T_THROW 6668) (T_TRAIT 6669) (T_TRAIT_C 6670) (T_TRY 6671) (T_UNSET 6672) (T_USE 6673) (T_VAR 6674) (T_VARIABLE 1687) (T_WHILE 6675) (T_YIELD 6676) (identifier 8386) (member_name 8387) (reserved_non_modifiers 6679) (semi_reserved 6680) (simple_variable 8388) ("{" 914)) 3589 (("(" 6682) (argument_list 8389)) 3590 ((T_FN 35) (T_FUNCTION 126) (fn 1704) #'1705 (inline_function 8393)) 3591 (("$" 1640) (T_STRING 8403) (T_VARIABLE 1687) (property_name 8410) (simple_variable 8405) ("{" 8406)) 3592 (("$" 1640) (T_STRING 8403) (T_VARIABLE 1687) (property_name 8404) (simple_variable 8405) ("{" 8406)) 3593 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 8401) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3594 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8399) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3595 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8400)) 3596 (("]" 8402)) 3597 (("(" 6682) (argument_list 8409)) 3598 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8407) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3599 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8408)) 3600 (("(" 6682) (argument_list 8411)) 3601 (("`" 8414)) 3602 (("]" 8416)) 3603 (("=" 8417)) 3604 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8418) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 3605 (("%" 1722) ("*" 1723) ("+" 1724) ("-" 1725) ("." 1726) ("/" 1727) ("<" 1728) (">" 1729) ("?" 1730) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1731) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1732) (T_BOOLEAN_AND 1733) (T_BOOLEAN_OR 1734) (T_COALESCE 1735) (T_DOUBLE_ARROW 8421) (T_INSTANCEOF 1736) (T_IS_EQUAL 1737) (T_IS_GREATER_OR_EQUAL 1738) (T_IS_IDENTICAL 1739) (T_IS_NOT_EQUAL 1740) (T_IS_NOT_IDENTICAL 1741) (T_IS_SMALLER_OR_EQUAL 1742) (T_LOGICAL_AND 1743) (T_LOGICAL_OR 1744) (T_LOGICAL_XOR 1745) (T_POW 1746) (T_SL 1747) (T_SPACESHIP 1748) (T_SR 1749) ("^" 1750) ("|" 1751)) 3606 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1720) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (array_object_dereferenceable 1693) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 8422) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 3607 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 8431) (T_VARIABLE 2822) (encaps_var 2836)) 3608 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 8430) (T_VARIABLE 2822) (encaps_var 2827)) 3609 (("(" 2899) (argument_list 2900) (ctor_arguments 8450)) 3610 ((T_ATTRIBUTE 2853) (T_CLASS 8436) (anonymous_class 8449) (attribute 2898)) 3611 (("(" 8440) (argument_list 8441) (ctor_arguments 8442)) 3612 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8443) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 8444) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 8445) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3613 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8448) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3614 (("," 1449) (possible_comma 8446)) 3615 ((")" 8447)) 3616 ((T_EXTENDS 2902) (extends_from 8451)) 3617 ((T_IMPLEMENTS 2904) (implements_list 8452)) 3618 ((backup_doc_comment 8453)) 3619 (("{" 8454)) 3620 ((class_statement_list 8455)) 3621 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 8456)) 3622 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8458) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3623 (("%" 280) (")" 8459) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3624 (("{" 8460)) 3625 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 8461) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 3626 (("}" 8462)) 3627 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 8464) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3628 ((")" 8465)) 3629 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 8467) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3630 (("," 5117) (possible_comma 8468)) 3631 ((")" 8469)) 3632 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 8476) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3633 ((")" 8477)) 3634 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8479) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3635 (("%" 280) (")" 8480) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3636 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8482) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3637 (("%" 280) (")" 8483) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3638 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 8491) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3639 ((")" 8492)) 3640 (("%" 280) (")" 8497) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3641 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8500) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3642 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8501)) 3643 (("\"" 8503) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 3644 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 8506) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3645 ((")" 8507)) 3646 (("%" 280) (")" 8509) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3647 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8512) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3648 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8513)) 3649 (("\"" 8515) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 3650 ((backup_doc_comment 8518)) 3651 (("(" 8519)) 3652 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 8520)) 3653 ((")" 8521)) 3654 ((T_USE 148) (lexical_vars 8522)) 3655 ((":" 150) (return_type 8523)) 3656 ((backup_fn_flags 8524)) 3657 (("{" 8525)) 3658 ((inner_statement_list 8526)) 3659 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 8527) ("~" 124)) 3660 ((backup_fn_flags 8528)) 3661 ((backup_doc_comment 8530)) 3662 (("(" 8531)) 3663 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 8532)) 3664 ((")" 8533)) 3665 ((":" 2207) (return_type 8534)) 3666 ((T_DOUBLE_ARROW 8535)) 3667 ((backup_fn_flags 8536)) 3668 ((backup_lex_pos 8537)) 3669 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8538) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 3670 (("%" 1259) ("*" 1260) ("+" 1261) ("-" 1262) ("." 1263) ("/" 1264) ("<" 1265) (">" 1266) ("?" 1267) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1268) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1269) (T_BOOLEAN_AND 1270) (T_BOOLEAN_OR 1271) (T_COALESCE 1272) (T_INSTANCEOF 1273) (T_IS_EQUAL 1274) (T_IS_GREATER_OR_EQUAL 1275) (T_IS_IDENTICAL 1276) (T_IS_NOT_EQUAL 1277) (T_IS_NOT_IDENTICAL 1278) (T_IS_SMALLER_OR_EQUAL 1279) (T_LOGICAL_AND 1280) (T_LOGICAL_OR 1281) (T_LOGICAL_XOR 1282) (T_POW 1283) (T_SL 1284) (T_SPACESHIP 1285) (T_SR 1286) ("^" 1287) (backup_fn_flags 8539) ("|" 1288)) 3671 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8541) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3672 (("$" 1184) (T_ABSTRACT 1361) (T_ARRAY 1362) (T_AS 1363) (T_BREAK 1364) (T_CALLABLE 1365) (T_CASE 1366) (T_CATCH 1367) (T_CLASS 1368) (T_CLASS_C 1369) (T_CLONE 1370) (T_CONST 1371) (T_CONTINUE 1372) (T_DECLARE 1373) (T_DEFAULT 1374) (T_DIR 1375) (T_DO 1376) (T_ECHO 1377) (T_ELSE 1378) (T_ELSEIF 1379) (T_EMPTY 1380) (T_ENDDECLARE 1381) (T_ENDFOR 1382) (T_ENDFOREACH 1383) (T_ENDIF 1384) (T_ENDSWITCH 1385) (T_ENDWHILE 1386) (T_ENUM 1387) (T_EVAL 1388) (T_EXIT 1389) (T_EXTENDS 1390) (T_FILE 1391) (T_FINAL 1392) (T_FINALLY 1393) (T_FN 1394) (T_FOR 1395) (T_FOREACH 1396) (T_FUNCTION 1397) (T_FUNC_C 1398) (T_GLOBAL 1399) (T_GOTO 1400) (T_IF 1401) (T_IMPLEMENTS 1402) (T_INCLUDE 1403) (T_INCLUDE_ONCE 1404) (T_INSTANCEOF 1405) (T_INSTEADOF 1406) (T_INTERFACE 1407) (T_ISSET 1408) (T_LINE 1409) (T_LIST 1410) (T_LOGICAL_AND 1411) (T_LOGICAL_OR 1412) (T_LOGICAL_XOR 1413) (T_MATCH 1414) (T_METHOD_C 1415) (T_NAMESPACE 1416) (T_NEW 1417) (T_NS_C 1418) (T_PRINT 1419) (T_PRIVATE 1420) (T_PROTECTED 1421) (T_PUBLIC 1422) (T_READONLY 1423) (T_REQUIRE 1424) (T_REQUIRE_ONCE 1425) (T_RETURN 1426) (T_STATIC 1427) (T_STRING 1428) (T_SWITCH 1429) (T_THROW 1430) (T_TRAIT 1431) (T_TRAIT_C 1432) (T_TRY 1433) (T_UNSET 1434) (T_USE 1435) (T_VAR 1436) (T_VARIABLE 1230) (T_WHILE 1437) (T_YIELD 1438) (identifier 8543) (member_name 8544) (reserved_non_modifiers 1441) (semi_reserved 1442) (simple_variable 8545) ("{" 914)) 3673 (("(" 1444) (argument_list 8546)) 3674 ((T_FN 35) (T_FUNCTION 126) (fn 1244) #'1245 (inline_function 8550)) 3675 (("$" 1184) (T_STRING 8559) (T_VARIABLE 1230) (property_name 8566) (simple_variable 8561) ("{" 8562)) 3676 (("$" 1184) (T_STRING 8559) (T_VARIABLE 1230) (property_name 8560) (simple_variable 8561) ("{" 8562)) 3677 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 8557) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3678 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8555) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3679 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8556)) 3680 (("]" 8558)) 3681 (("(" 1444) (argument_list 8565)) 3682 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8563) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3683 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8564)) 3684 (("(" 1444) (argument_list 8567)) 3685 (("`" 8569)) 3686 (("]" 8571)) 3687 (("=" 8572)) 3688 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8573) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 3689 (("%" 1259) ("*" 1260) ("+" 1261) ("-" 1262) ("." 1263) ("/" 1264) ("<" 1265) (">" 1266) ("?" 1267) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1268) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1269) (T_BOOLEAN_AND 1270) (T_BOOLEAN_OR 1271) (T_COALESCE 1272) (T_DOUBLE_ARROW 8576) (T_INSTANCEOF 1273) (T_IS_EQUAL 1274) (T_IS_GREATER_OR_EQUAL 1275) (T_IS_IDENTICAL 1276) (T_IS_NOT_EQUAL 1277) (T_IS_NOT_IDENTICAL 1278) (T_IS_SMALLER_OR_EQUAL 1279) (T_LOGICAL_AND 1280) (T_LOGICAL_OR 1281) (T_LOGICAL_XOR 1282) (T_POW 1283) (T_SL 1284) (T_SPACESHIP 1285) (T_SR 1286) ("^" 1287) ("|" 1288)) 3690 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8577) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 3691 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 8586) (T_VARIABLE 2822) (encaps_var 2836)) 3692 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 8585) (T_VARIABLE 2822) (encaps_var 2827)) 3693 (("(" 2899) (argument_list 2900) (ctor_arguments 8605)) 3694 ((T_ATTRIBUTE 2853) (T_CLASS 8591) (anonymous_class 8604) (attribute 2898)) 3695 (("(" 8595) (argument_list 8596) (ctor_arguments 8597)) 3696 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8598) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 8599) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 8600) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3697 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8603) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3698 (("," 1449) (possible_comma 8601)) 3699 ((")" 8602)) 3700 ((T_EXTENDS 2902) (extends_from 8606)) 3701 ((T_IMPLEMENTS 2904) (implements_list 8607)) 3702 ((backup_doc_comment 8608)) 3703 (("{" 8609)) 3704 ((class_statement_list 8610)) 3705 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 8611)) 3706 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8613) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3707 (("%" 280) (")" 8614) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3708 (("{" 8615)) 3709 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 8616) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 3710 (("}" 8617)) 3711 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 8619) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3712 ((")" 8620)) 3713 (("=" 8621)) 3714 (("!" 1182) ("\"" 1183) ("$" 1184) ("(" 1185) ("+" 1186) ("-" 1187) ("@" 1188) (T_ARRAY 1189) (T_ARRAY_CAST 1190) (T_ATTRIBUTE 125) (T_BOOL_CAST 1191) (T_CLASS_C 1192) (T_CLONE 1193) (T_CONSTANT_ENCAPSED_STRING 1194) (T_DEC 1195) (T_DIR 1196) (T_DNUMBER 1197) (T_DOUBLE_CAST 1198) (T_EMPTY 1199) (T_EVAL 1200) (T_EXIT 1201) (T_FILE 1202) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1203) (T_INC 1204) (T_INCLUDE 1205) (T_INCLUDE_ONCE 1206) (T_INT_CAST 1207) (T_ISSET 1208) (T_LINE 1209) (T_LIST 1210) (T_LNUMBER 1211) (T_MATCH 1212) (T_METHOD_C 1213) (T_NAME_FULLY_QUALIFIED 1214) (T_NAME_QUALIFIED 1215) (T_NAME_RELATIVE 1216) (T_NEW 1217) (T_NS_C 1218) (T_OBJECT_CAST 1219) (T_PRINT 1220) (T_REQUIRE 1221) (T_REQUIRE_ONCE 1222) (T_START_HEREDOC 1223) (T_STATIC 1224) (T_STRING 1225) (T_STRING_CAST 1226) (T_THROW 1227) (T_TRAIT_C 1228) (T_UNSET_CAST 1229) (T_VARIABLE 1230) (T_YIELD 1231) (T_YIELD_FROM 1232) ("[" 1233) ("`" 1234) (array_object_dereferenceable 1235) (attribute 129) (attributes 1236) (callable_expr 1237) (callable_variable 1238) (class_constant 1239) (class_name 1240) (constant 1241) (dereferenceable_scalar 1242) (expr 8622) (fn 1244) (fully_dereferenceable 103) #'1245 (function_call 1246) (inline_function 1247) (internal_functions_in_yacc 1248) (match 1249) (name 1250) (new_expr 1251) (scalar 1252) (simple_variable 1253) (static_member 1254) (variable 1255) (variable_class_name 1256) ("~" 1257)) 3715 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 8624) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3716 (("," 5117) (possible_comma 8625)) 3717 ((")" 8626)) 3718 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 8633) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3719 ((")" 8634)) 3720 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8636) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3721 (("%" 280) (")" 8637) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3722 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8639) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3723 (("%" 280) (")" 8640) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3724 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 8647) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3725 ((")" 8648)) 3726 (("%" 280) (")" 8653) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3727 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8656) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3728 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8657)) 3729 (("\"" 8659) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 3730 (("(" 2899) (argument_list 2900) (ctor_arguments 8680)) 3731 ((T_ATTRIBUTE 2853) (T_CLASS 8666) (anonymous_class 8679) (attribute 2898)) 3732 (("(" 8670) (argument_list 8671) (ctor_arguments 8672)) 3733 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8673) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 8674) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 8675) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3734 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8678) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3735 (("," 1449) (possible_comma 8676)) 3736 ((")" 8677)) 3737 ((T_EXTENDS 2902) (extends_from 8681)) 3738 ((T_IMPLEMENTS 2904) (implements_list 8682)) 3739 ((backup_doc_comment 8683)) 3740 (("{" 8684)) 3741 ((class_statement_list 8685)) 3742 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 8686)) 3743 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8688) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3744 (("%" 280) (")" 8689) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3745 (("{" 8690)) 3746 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 8691) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 3747 (("}" 8692)) 3748 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 8694) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3749 ((")" 8695)) 3750 (("=" 8696)) 3751 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8697) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3752 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 8699) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3753 (("," 5117) (possible_comma 8700)) 3754 ((")" 8701)) 3755 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 8706) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3756 ((")" 8707)) 3757 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8709) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3758 (("%" 280) (")" 8710) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3759 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 8712) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3760 (("%" 280) (")" 8713) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3761 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 8716) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 3762 ((")" 8717)) 3763 (("$" 921) (T_ABSTRACT 8719) (T_ARRAY 8720) (T_AS 8721) (T_BREAK 8722) (T_CALLABLE 8723) (T_CASE 8724) (T_CATCH 8725) (T_CLASS 8726) (T_CLASS_C 8727) (T_CLONE 8728) (T_CONST 8729) (T_CONTINUE 8730) (T_DECLARE 8731) (T_DEFAULT 8732) (T_DIR 8733) (T_DO 8734) (T_ECHO 8735) (T_ELSE 8736) (T_ELSEIF 8737) (T_EMPTY 8738) (T_ENDDECLARE 8739) (T_ENDFOR 8740) (T_ENDFOREACH 8741) (T_ENDIF 8742) (T_ENDSWITCH 8743) (T_ENDWHILE 8744) (T_ENUM 8745) (T_EVAL 8746) (T_EXIT 8747) (T_EXTENDS 8748) (T_FILE 8749) (T_FINAL 8750) (T_FINALLY 8751) (T_FN 8752) (T_FOR 8753) (T_FOREACH 8754) (T_FUNCTION 8755) (T_FUNC_C 8756) (T_GLOBAL 8757) (T_GOTO 8758) (T_IF 8759) (T_IMPLEMENTS 8760) (T_INCLUDE 8761) (T_INCLUDE_ONCE 8762) (T_INSTANCEOF 8763) (T_INSTEADOF 8764) (T_INTERFACE 8765) (T_ISSET 8766) (T_LINE 8767) (T_LIST 8768) (T_LOGICAL_AND 8769) (T_LOGICAL_OR 8770) (T_LOGICAL_XOR 8771) (T_MATCH 8772) (T_METHOD_C 8773) (T_NAMESPACE 8774) (T_NEW 8775) (T_NS_C 8776) (T_PRINT 8777) (T_PRIVATE 8778) (T_PROTECTED 8779) (T_PUBLIC 8780) (T_READONLY 8781) (T_REQUIRE 8782) (T_REQUIRE_ONCE 8783) (T_RETURN 8784) (T_STATIC 8785) (T_STRING 8786) (T_SWITCH 8787) (T_THROW 8788) (T_TRAIT 8789) (T_TRAIT_C 8790) (T_TRY 8791) (T_UNSET 8792) (T_USE 8793) (T_VAR 8794) (T_VARIABLE 1020) (T_WHILE 8795) (T_YIELD 8796) (identifier 8797) (member_name 8798) (reserved_non_modifiers 8799) (semi_reserved 8800) (simple_variable 8801) ("{" 914)) 3764 (("(" 8802) (argument_list 8803)) 3765 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8804) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 8805) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 8806) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3766 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8809) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3767 (("," 1449) (possible_comma 8807)) 3768 ((")" 8808)) 3769 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (ampersand 8839) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8840) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3770 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8838) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3771 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8837) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3772 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8836) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3773 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8835) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3774 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8834) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3775 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8833) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3776 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8832) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3777 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8831) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3778 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8830) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3779 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8829) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3780 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8828) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3781 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8827) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3782 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8826) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3783 (("\"" 1489) ("$" 2851) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 2860) ("[" 1508) (array_object_dereferenceable 8841) (callable_expr 8842) (callable_variable 8843) (class_constant 1512) (class_name 8844) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 8845) (name 8846) (simple_variable 8847) (static_member 8848) (variable 8849) (variable_class_name 8850)) 3784 ((T_NULLSAFE_OBJECT_OPERATOR 8868) (T_OBJECT_OPERATOR 8869) ("[" 8870) ("{" 8871)) 3785 (("(" 8854) (argument_list 8867)) 3786 ((T_PAAMAYIM_NEKUDOTAYIM 8863)) 3787 (("(" 8854) (argument_list 8862)) 3788 ((T_PAAMAYIM_NEKUDOTAYIM 8851)) 3789 (("$" 2851) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 2860) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 8852) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 8853) ("{" 914)) 3790 (("(" 8854) (argument_list 8855)) 3791 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8856) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 8857) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 8858) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3792 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 8861) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3793 (("," 1449) (possible_comma 8859)) 3794 ((")" 8860)) 3795 (("$" 2851) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 2860) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 8864) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 8865) ("{" 914)) 3796 (("(" 8854) (argument_list 8866)) 3797 (("$" 2851) (T_STRING 2879) (T_VARIABLE 2860) (property_name 8878) (simple_variable 2881) ("{" 2882)) 3798 (("$" 2851) (T_STRING 2879) (T_VARIABLE 2860) (property_name 8876) (simple_variable 2881) ("{" 2882)) 3799 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 8874) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3800 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8872) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3801 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8873)) 3802 (("]" 8875)) 3803 (("(" 8854) (argument_list 8877)) 3804 (("(" 8854) (argument_list 8879)) 3805 ((")" 8881)) 3806 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8884) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3807 ((backup_doc_comment 8886)) 3808 (("(" 8887)) 3809 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 8888)) 3810 ((")" 8889)) 3811 ((T_USE 148) (lexical_vars 8890)) 3812 ((":" 150) (return_type 8891)) 3813 ((backup_fn_flags 8892)) 3814 (("{" 8893)) 3815 ((inner_statement_list 8894)) 3816 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 8895) ("~" 124)) 3817 ((backup_fn_flags 8896)) 3818 ((backup_doc_comment 8898)) 3819 (("(" 8899)) 3820 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 8900)) 3821 ((")" 8901)) 3822 ((":" 2207) (return_type 8902)) 3823 ((T_DOUBLE_ARROW 8903)) 3824 ((backup_fn_flags 8904)) 3825 ((backup_lex_pos 8905)) 3826 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8906) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3827 (("%" 1081) ("*" 1082) ("+" 1083) ("-" 1084) ("." 1085) ("/" 1086) ("<" 1087) (">" 1088) ("?" 1089) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1090) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1091) (T_BOOLEAN_AND 1092) (T_BOOLEAN_OR 1093) (T_COALESCE 1094) (T_INSTANCEOF 1095) (T_IS_EQUAL 1096) (T_IS_GREATER_OR_EQUAL 1097) (T_IS_IDENTICAL 1098) (T_IS_NOT_EQUAL 1099) (T_IS_NOT_IDENTICAL 1100) (T_IS_SMALLER_OR_EQUAL 1101) (T_LOGICAL_AND 1102) (T_LOGICAL_OR 1103) (T_LOGICAL_XOR 1104) (T_POW 1105) (T_SL 1106) (T_SPACESHIP 1107) (T_SR 1108) ("^" 1109) (backup_fn_flags 8907) ("|" 1110)) 3828 (("$" 921) (T_ABSTRACT 8719) (T_ARRAY 8720) (T_AS 8721) (T_BREAK 8722) (T_CALLABLE 8723) (T_CASE 8724) (T_CATCH 8725) (T_CLASS 8726) (T_CLASS_C 8727) (T_CLONE 8728) (T_CONST 8729) (T_CONTINUE 8730) (T_DECLARE 8731) (T_DEFAULT 8732) (T_DIR 8733) (T_DO 8734) (T_ECHO 8735) (T_ELSE 8736) (T_ELSEIF 8737) (T_EMPTY 8738) (T_ENDDECLARE 8739) (T_ENDFOR 8740) (T_ENDFOREACH 8741) (T_ENDIF 8742) (T_ENDSWITCH 8743) (T_ENDWHILE 8744) (T_ENUM 8745) (T_EVAL 8746) (T_EXIT 8747) (T_EXTENDS 8748) (T_FILE 8749) (T_FINAL 8750) (T_FINALLY 8751) (T_FN 8752) (T_FOR 8753) (T_FOREACH 8754) (T_FUNCTION 8755) (T_FUNC_C 8756) (T_GLOBAL 8757) (T_GOTO 8758) (T_IF 8759) (T_IMPLEMENTS 8760) (T_INCLUDE 8761) (T_INCLUDE_ONCE 8762) (T_INSTANCEOF 8763) (T_INSTEADOF 8764) (T_INTERFACE 8765) (T_ISSET 8766) (T_LINE 8767) (T_LIST 8768) (T_LOGICAL_AND 8769) (T_LOGICAL_OR 8770) (T_LOGICAL_XOR 8771) (T_MATCH 8772) (T_METHOD_C 8773) (T_NAMESPACE 8774) (T_NEW 8775) (T_NS_C 8776) (T_PRINT 8777) (T_PRIVATE 8778) (T_PROTECTED 8779) (T_PUBLIC 8780) (T_READONLY 8781) (T_REQUIRE 8782) (T_REQUIRE_ONCE 8783) (T_RETURN 8784) (T_STATIC 8785) (T_STRING 8786) (T_SWITCH 8787) (T_THROW 8788) (T_TRAIT 8789) (T_TRAIT_C 8790) (T_TRY 8791) (T_UNSET 8792) (T_USE 8793) (T_VAR 8794) (T_VARIABLE 1020) (T_WHILE 8795) (T_YIELD 8796) (identifier 8909) (member_name 8910) (reserved_non_modifiers 8799) (semi_reserved 8800) (simple_variable 8911) ("{" 914)) 3829 (("(" 8802) (argument_list 8912)) 3830 ((T_FN 35) (T_FUNCTION 126) (fn 1036) #'1037 (inline_function 8916)) 3831 (("$" 921) (T_STRING 8925) (T_VARIABLE 1020) (property_name 8932) (simple_variable 8927) ("{" 8928)) 3832 (("$" 921) (T_STRING 8925) (T_VARIABLE 1020) (property_name 8926) (simple_variable 8927) ("{" 8928)) 3833 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 8923) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3834 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8921) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3835 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8922)) 3836 (("]" 8924)) 3837 (("(" 8802) (argument_list 8931)) 3838 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8929) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3839 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8930)) 3840 (("(" 8802) (argument_list 8933)) 3841 (("`" 8935)) 3842 (("]" 8937)) 3843 (("=" 8938)) 3844 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 8939) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3845 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 8947) (T_VARIABLE 2822) (encaps_var 2836)) 3846 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 8946) (T_VARIABLE 2822) (encaps_var 2827)) 3847 (("%" 280) (")" 8960) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 3848 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 8963) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3849 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 8964)) 3850 (("\"" 8966) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 3851 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (ampersand 8997) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8998) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3852 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8996) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3853 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8995) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3854 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8994) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3855 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8993) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3856 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8992) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3857 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8991) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3858 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8990) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3859 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8989) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3860 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8988) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3861 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8987) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3862 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8986) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3863 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8985) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3864 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 8984) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3865 (("\"" 1489) ("$" 8999) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 9000) ("[" 1508) (array_object_dereferenceable 9001) (callable_expr 9002) (callable_variable 9003) (class_constant 1512) (class_name 9004) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 9005) (name 9006) (simple_variable 9007) (static_member 9008) (variable 9009) (variable_class_name 9010)) 3866 (("$" 8999) (T_VARIABLE 9000) (simple_variable 9045) ("{" 9046)) 3867 ((T_NULLSAFE_OBJECT_OPERATOR 9028) (T_OBJECT_OPERATOR 9029) ("[" 9030) ("{" 9031)) 3868 (("(" 9014) (argument_list 9027)) 3869 ((T_PAAMAYIM_NEKUDOTAYIM 9023)) 3870 (("(" 9014) (argument_list 9022)) 3871 ((T_PAAMAYIM_NEKUDOTAYIM 9011)) 3872 (("$" 8999) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 9000) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 9012) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 9013) ("{" 914)) 3873 (("(" 9014) (argument_list 9015)) 3874 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 9016) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 9017) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 9018) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3875 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 9021) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 3876 (("," 1449) (possible_comma 9019)) 3877 ((")" 9020)) 3878 (("$" 8999) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 9000) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 9024) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 9025) ("{" 914)) 3879 (("(" 9014) (argument_list 9026)) 3880 (("$" 8999) (T_STRING 9036) (T_VARIABLE 9000) (property_name 9043) (simple_variable 9038) ("{" 9039)) 3881 (("$" 8999) (T_STRING 9036) (T_VARIABLE 9000) (property_name 9037) (simple_variable 9038) ("{" 9039)) 3882 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 9034) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3883 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9032) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3884 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9033)) 3885 (("]" 9035)) 3886 (("(" 9014) (argument_list 9042)) 3887 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9040) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3888 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9041)) 3889 (("(" 9014) (argument_list 9044)) 3890 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9047) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 3891 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9048)) 3892 ((backup_doc_comment 9051)) 3893 (("(" 9052)) 3894 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 9053)) 3895 ((")" 9054)) 3896 ((T_USE 148) (lexical_vars 9055)) 3897 ((":" 150) (return_type 9056)) 3898 ((backup_fn_flags 9057)) 3899 (("{" 9058)) 3900 ((inner_statement_list 9059)) 3901 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 9060) ("~" 124)) 3902 ((backup_fn_flags 9061)) 3903 ((backup_doc_comment 9063)) 3904 (("(" 9064)) 3905 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 9065)) 3906 ((")" 9066)) 3907 ((":" 2207) (return_type 9067)) 3908 ((T_DOUBLE_ARROW 9068)) 3909 ((backup_fn_flags 9069)) 3910 ((backup_lex_pos 9070)) 3911 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9132) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3912 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10110) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3913 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 10108) (encaps_var 5158)) 3914 (("$" 9073) (T_VARIABLE 9119) (simple_variable 10104) ("{" 10105)) 3915 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10102) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3916 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10101) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3917 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10100) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3918 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10099) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3919 (("(" 10096)) 3920 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10095) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3921 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10094) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3922 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10093) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3923 (("\"" 1489) ("$" 9372) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 9373) ("[" 1508) (array_object_dereferenceable 9374) (callable_expr 9375) (callable_variable 9376) (class_constant 1512) (class_name 9377) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 9378) (name 9379) (simple_variable 9380) (static_member 9381) (variable 10092) (variable_class_name 9383)) 3924 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10091) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3925 (("(" 10088)) 3926 (("(" 10085)) 3927 (("(" 10081) (exit_expr 10082)) 3928 (("\"" 1489) ("$" 9372) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 9373) ("[" 1508) (array_object_dereferenceable 9374) (callable_expr 9375) (callable_variable 9376) (class_constant 1512) (class_name 9377) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 9378) (name 9379) (simple_variable 9380) (static_member 9381) (variable 10080) (variable_class_name 9383)) 3929 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10079) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3930 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10078) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3931 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10077) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3932 (("(" 10073)) 3933 (("(" 10068)) 3934 (("(" 10062)) 3935 (("$" 9372) ("(" 10015) (T_ATTRIBUTE 2853) (T_CLASS 10016) (T_NAME_FULLY_QUALIFIED 10017) (T_NAME_QUALIFIED 10018) (T_NAME_RELATIVE 10019) (T_STATIC 10020) (T_STRING 10021) (T_VARIABLE 9373) (anonymous_class 10022) (attribute 2862) (attributes 10023) (class_name 10024) (class_name_reference 10025) (name 10026) (new_variable 10027) (simple_variable 10028)) 3936 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10014) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3937 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10013) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3938 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10012) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3939 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10011) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3940 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 10006) (T_END_HEREDOC 10007) (T_VARIABLE 2822) (encaps_list 10008) (encaps_var 2824)) 3941 ((T_FN 35) (T_FUNCTION 126) (fn 9133) #'9134 (inline_function 10005)) 3942 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10004) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3943 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10003) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3944 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10002) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3945 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9541) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 3946 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9479) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3947 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 9475) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 3948 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 9473) (encaps_list 2729) (encaps_var 2730)) 3949 ((T_NULLSAFE_OBJECT_OPERATOR 9456) (T_OBJECT_OPERATOR 9457) ("[" 9458) ("{" 9459)) 3950 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 9453) (attribute 2654) (fn 9133) #'9134 (inline_function 9454)) 3951 (("(" 9333) (argument_list 9452)) 3952 ((T_PAAMAYIM_NEKUDOTAYIM 9447)) 3953 (("%" 9148) ("*" 9149) ("+" 9150) ("-" 9151) ("." 9152) ("/" 9153) ("<" 9154) (">" 9155) ("?" 9156) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9157) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9158) (T_BOOLEAN_AND 9159) (T_BOOLEAN_OR 9160) (T_COALESCE 9161) (T_INSTANCEOF 9162) (T_IS_EQUAL 9163) (T_IS_GREATER_OR_EQUAL 9164) (T_IS_IDENTICAL 9165) (T_IS_NOT_EQUAL 9166) (T_IS_NOT_IDENTICAL 9167) (T_IS_SMALLER_OR_EQUAL 9168) (T_LOGICAL_AND 9169) (T_LOGICAL_OR 9170) (T_LOGICAL_XOR 9171) (T_POW 9172) (T_SL 9173) (T_SPACESHIP 9174) (T_SR 9175) ("^" 9176) (backup_fn_flags 9446) ("|" 9177)) 3954 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 9435)) 3955 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 9423)) 3956 (("(" 9333) (argument_list 9422)) 3957 (("=" 9341) (T_AND_EQUAL 9342) (T_COALESCE_EQUAL 9343) (T_CONCAT_EQUAL 9344) (T_DEC 9345) (T_DIV_EQUAL 9346) (T_INC 9347) (T_MINUS_EQUAL 9348) (T_MOD_EQUAL 9349) (T_MUL_EQUAL 9350) (T_OR_EQUAL 9351) (T_PLUS_EQUAL 9352) (T_POW_EQUAL 9353) (T_SL_EQUAL 9354) (T_SR_EQUAL 9355) (T_XOR_EQUAL 9356)) 3958 ((T_PAAMAYIM_NEKUDOTAYIM 9249)) 3959 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9147) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3960 (("%" 9148) ("*" 9149) ("+" 9150) ("-" 9151) ("." 9152) ("/" 9153) ("<" 9154) (">" 9155) ("?" 9156) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9157) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9158) (T_BOOLEAN_AND 9159) (T_BOOLEAN_OR 9160) (T_COALESCE 9161) (T_INSTANCEOF 9162) (T_IS_EQUAL 9163) (T_IS_GREATER_OR_EQUAL 9164) (T_IS_IDENTICAL 9165) (T_IS_NOT_EQUAL 9166) (T_IS_NOT_IDENTICAL 9167) (T_IS_SMALLER_OR_EQUAL 9168) (T_LOGICAL_AND 9169) (T_LOGICAL_OR 9170) (T_LOGICAL_XOR 9171) (T_POW 9172) (T_SL 9173) (T_SPACESHIP 9174) (T_SR 9175) ("^" 9176) ("|" 9177)) 3961 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9248) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3962 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9247) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3963 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9246) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3964 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9245) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3965 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9244) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3966 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9243) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3967 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9242) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3968 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9241) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3969 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 9236) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 9237) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 3970 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9235) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3971 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9234) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3972 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9233) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3973 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9232) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3974 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9231) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3975 (("$" 9193) ("(" 9194) (T_NAME_FULLY_QUALIFIED 9195) (T_NAME_QUALIFIED 9196) (T_NAME_RELATIVE 9197) (T_STATIC 9198) (T_STRING 9199) (T_VARIABLE 9200) (class_name 9201) (class_name_reference 9202) (name 9203) (new_variable 9204) (simple_variable 9205)) 3976 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9192) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3977 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9191) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3978 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9190) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3979 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9189) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3980 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9188) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3981 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9187) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3982 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9186) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3983 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9185) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3984 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9184) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3985 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9183) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3986 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9182) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3987 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9181) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3988 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9180) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3989 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9179) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3990 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9178) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 3991 (("$" 9193) (T_VARIABLE 9200) (simple_variable 9227) ("{" 9228)) 3992 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 9225) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 3993 ((T_PAAMAYIM_NEKUDOTAYIM 9223)) 3994 ((T_NULLSAFE_OBJECT_OPERATOR 9206) (T_OBJECT_OPERATOR 9207) (T_PAAMAYIM_NEKUDOTAYIM 9208) ("[" 9209) ("{" 9210)) 3995 (("$" 9193) (T_STRING 9216) (T_VARIABLE 9200) (property_name 9222) (simple_variable 9218) ("{" 9219)) 3996 (("$" 9193) (T_STRING 9216) (T_VARIABLE 9200) (property_name 9217) (simple_variable 9218) ("{" 9219)) 3997 (("$" 9193) (T_VARIABLE 9200) (simple_variable 9215)) 3998 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 9213) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 3999 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9211) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4000 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9212)) 4001 (("]" 9214)) 4002 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9220) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4003 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9221)) 4004 (("$" 9193) (T_VARIABLE 9200) (simple_variable 9224)) 4005 (("%" 280) (")" 9226) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4006 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9229) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4007 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9230)) 4008 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9240) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4009 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 9238) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 4010 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9239) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4011 (("$" 9073) (T_ABSTRACT 9250) (T_ARRAY 9251) (T_AS 9252) (T_BREAK 9253) (T_CALLABLE 9254) (T_CASE 9255) (T_CATCH 9256) (T_CLASS 9257) (T_CLASS_C 9258) (T_CLONE 9259) (T_CONST 9260) (T_CONTINUE 9261) (T_DECLARE 9262) (T_DEFAULT 9263) (T_DIR 9264) (T_DO 9265) (T_ECHO 9266) (T_ELSE 9267) (T_ELSEIF 9268) (T_EMPTY 9269) (T_ENDDECLARE 9270) (T_ENDFOR 9271) (T_ENDFOREACH 9272) (T_ENDIF 9273) (T_ENDSWITCH 9274) (T_ENDWHILE 9275) (T_ENUM 9276) (T_EVAL 9277) (T_EXIT 9278) (T_EXTENDS 9279) (T_FILE 9280) (T_FINAL 9281) (T_FINALLY 9282) (T_FN 9283) (T_FOR 9284) (T_FOREACH 9285) (T_FUNCTION 9286) (T_FUNC_C 9287) (T_GLOBAL 9288) (T_GOTO 9289) (T_IF 9290) (T_IMPLEMENTS 9291) (T_INCLUDE 9292) (T_INCLUDE_ONCE 9293) (T_INSTANCEOF 9294) (T_INSTEADOF 9295) (T_INTERFACE 9296) (T_ISSET 9297) (T_LINE 9298) (T_LIST 9299) (T_LOGICAL_AND 9300) (T_LOGICAL_OR 9301) (T_LOGICAL_XOR 9302) (T_MATCH 9303) (T_METHOD_C 9304) (T_NAMESPACE 9305) (T_NEW 9306) (T_NS_C 9307) (T_PRINT 9308) (T_PRIVATE 9309) (T_PROTECTED 9310) (T_PUBLIC 9311) (T_READONLY 9312) (T_REQUIRE 9313) (T_REQUIRE_ONCE 9314) (T_RETURN 9315) (T_STATIC 9316) (T_STRING 9317) (T_SWITCH 9318) (T_THROW 9319) (T_TRAIT 9320) (T_TRAIT_C 9321) (T_TRY 9322) (T_UNSET 9323) (T_USE 9324) (T_VAR 9325) (T_VARIABLE 9119) (T_WHILE 9326) (T_YIELD 9327) (identifier 9328) (member_name 9329) (reserved_non_modifiers 9330) (semi_reserved 9331) (simple_variable 9332) ("{" 914)) 4012 (("(" 9333) (argument_list 9334)) 4013 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 9335) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 9336) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 9337) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4014 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 9340) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4015 (("," 1449) (possible_comma 9338)) 4016 ((")" 9339)) 4017 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (ampersand 9370) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9371) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4018 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9369) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4019 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9368) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4020 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9367) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4021 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9366) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4022 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9365) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4023 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9364) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4024 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9363) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4025 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9362) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4026 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9361) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4027 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9360) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4028 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9359) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4029 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9358) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4030 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9357) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4031 (("\"" 1489) ("$" 9372) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 9373) ("[" 1508) (array_object_dereferenceable 9374) (callable_expr 9375) (callable_variable 9376) (class_constant 1512) (class_name 9377) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 9378) (name 9379) (simple_variable 9380) (static_member 9381) (variable 9382) (variable_class_name 9383)) 4032 (("$" 9372) (T_VARIABLE 9373) (simple_variable 9418) ("{" 9419)) 4033 ((T_NULLSAFE_OBJECT_OPERATOR 9401) (T_OBJECT_OPERATOR 9402) ("[" 9403) ("{" 9404)) 4034 (("(" 9387) (argument_list 9400)) 4035 ((T_PAAMAYIM_NEKUDOTAYIM 9396)) 4036 (("(" 9387) (argument_list 9395)) 4037 ((T_PAAMAYIM_NEKUDOTAYIM 9384)) 4038 (("$" 9372) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 9373) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 9385) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 9386) ("{" 914)) 4039 (("(" 9387) (argument_list 9388)) 4040 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 9389) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 9390) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 9391) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4041 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 9394) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4042 (("," 1449) (possible_comma 9392)) 4043 ((")" 9393)) 4044 (("$" 9372) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 9373) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 9397) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 9398) ("{" 914)) 4045 (("(" 9387) (argument_list 9399)) 4046 (("$" 9372) (T_STRING 9409) (T_VARIABLE 9373) (property_name 9416) (simple_variable 9411) ("{" 9412)) 4047 (("$" 9372) (T_STRING 9409) (T_VARIABLE 9373) (property_name 9410) (simple_variable 9411) ("{" 9412)) 4048 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 9407) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4049 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9405) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4050 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9406)) 4051 (("]" 9408)) 4052 (("(" 9387) (argument_list 9415)) 4053 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9413) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4054 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9414)) 4055 (("(" 9387) (argument_list 9417)) 4056 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9420) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4057 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9421)) 4058 ((backup_doc_comment 9424)) 4059 (("(" 9425)) 4060 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 9426)) 4061 ((")" 9427)) 4062 ((T_USE 148) (lexical_vars 9428)) 4063 ((":" 150) (return_type 9429)) 4064 ((backup_fn_flags 9430)) 4065 (("{" 9431)) 4066 ((inner_statement_list 9432)) 4067 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 9433) ("~" 124)) 4068 ((backup_fn_flags 9434)) 4069 ((backup_doc_comment 9436)) 4070 (("(" 9437)) 4071 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 9438)) 4072 ((")" 9439)) 4073 ((":" 2207) (return_type 9440)) 4074 ((T_DOUBLE_ARROW 9441)) 4075 ((backup_fn_flags 9442)) 4076 ((backup_lex_pos 9443)) 4077 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9444) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4078 (("%" 9148) ("*" 9149) ("+" 9150) ("-" 9151) ("." 9152) ("/" 9153) ("<" 9154) (">" 9155) ("?" 9156) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9157) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9158) (T_BOOLEAN_AND 9159) (T_BOOLEAN_OR 9160) (T_COALESCE 9161) (T_INSTANCEOF 9162) (T_IS_EQUAL 9163) (T_IS_GREATER_OR_EQUAL 9164) (T_IS_IDENTICAL 9165) (T_IS_NOT_EQUAL 9166) (T_IS_NOT_IDENTICAL 9167) (T_IS_SMALLER_OR_EQUAL 9168) (T_LOGICAL_AND 9169) (T_LOGICAL_OR 9170) (T_LOGICAL_XOR 9171) (T_POW 9172) (T_SL 9173) (T_SPACESHIP 9174) (T_SR 9175) ("^" 9176) (backup_fn_flags 9445) ("|" 9177)) 4079 (("$" 9073) (T_ABSTRACT 9250) (T_ARRAY 9251) (T_AS 9252) (T_BREAK 9253) (T_CALLABLE 9254) (T_CASE 9255) (T_CATCH 9256) (T_CLASS 9257) (T_CLASS_C 9258) (T_CLONE 9259) (T_CONST 9260) (T_CONTINUE 9261) (T_DECLARE 9262) (T_DEFAULT 9263) (T_DIR 9264) (T_DO 9265) (T_ECHO 9266) (T_ELSE 9267) (T_ELSEIF 9268) (T_EMPTY 9269) (T_ENDDECLARE 9270) (T_ENDFOR 9271) (T_ENDFOREACH 9272) (T_ENDIF 9273) (T_ENDSWITCH 9274) (T_ENDWHILE 9275) (T_ENUM 9276) (T_EVAL 9277) (T_EXIT 9278) (T_EXTENDS 9279) (T_FILE 9280) (T_FINAL 9281) (T_FINALLY 9282) (T_FN 9283) (T_FOR 9284) (T_FOREACH 9285) (T_FUNCTION 9286) (T_FUNC_C 9287) (T_GLOBAL 9288) (T_GOTO 9289) (T_IF 9290) (T_IMPLEMENTS 9291) (T_INCLUDE 9292) (T_INCLUDE_ONCE 9293) (T_INSTANCEOF 9294) (T_INSTEADOF 9295) (T_INTERFACE 9296) (T_ISSET 9297) (T_LINE 9298) (T_LIST 9299) (T_LOGICAL_AND 9300) (T_LOGICAL_OR 9301) (T_LOGICAL_XOR 9302) (T_MATCH 9303) (T_METHOD_C 9304) (T_NAMESPACE 9305) (T_NEW 9306) (T_NS_C 9307) (T_PRINT 9308) (T_PRIVATE 9309) (T_PROTECTED 9310) (T_PUBLIC 9311) (T_READONLY 9312) (T_REQUIRE 9313) (T_REQUIRE_ONCE 9314) (T_RETURN 9315) (T_STATIC 9316) (T_STRING 9317) (T_SWITCH 9318) (T_THROW 9319) (T_TRAIT 9320) (T_TRAIT_C 9321) (T_TRY 9322) (T_UNSET 9323) (T_USE 9324) (T_VAR 9325) (T_VARIABLE 9119) (T_WHILE 9326) (T_YIELD 9327) (identifier 9448) (member_name 9449) (reserved_non_modifiers 9330) (semi_reserved 9331) (simple_variable 9450) ("{" 914)) 4080 (("(" 9333) (argument_list 9451)) 4081 ((T_FN 35) (T_FUNCTION 126) (fn 9133) #'9134 (inline_function 9455)) 4082 (("$" 9073) (T_STRING 9464) (T_VARIABLE 9119) (property_name 9471) (simple_variable 9466) ("{" 9467)) 4083 (("$" 9073) (T_STRING 9464) (T_VARIABLE 9119) (property_name 9465) (simple_variable 9466) ("{" 9467)) 4084 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 9462) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4085 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9460) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4086 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9461)) 4087 (("]" 9463)) 4088 (("(" 9333) (argument_list 9470)) 4089 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9468) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4090 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9469)) 4091 (("(" 9333) (argument_list 9472)) 4092 (("`" 9474)) 4093 (("]" 9476)) 4094 (("=" 9477)) 4095 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9478) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4096 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 10001) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4097 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 9999) (encaps_var 5158)) 4098 (("$" 9482) (T_VARIABLE 9528) (simple_variable 9995) ("{" 9996)) 4099 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 9993) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4100 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9992) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4101 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9991) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4102 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9990) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4103 (("(" 9987)) 4104 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9986) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4105 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9985) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4106 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9984) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4107 (("\"" 1489) ("$" 9781) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 9782) ("[" 1508) (array_object_dereferenceable 9783) (callable_expr 9784) (callable_variable 9785) (class_constant 1512) (class_name 9786) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 9787) (name 9788) (simple_variable 9789) (static_member 9790) (variable 9983) (variable_class_name 9792)) 4108 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9982) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4109 (("(" 9979)) 4110 (("(" 9976)) 4111 (("(" 9972) (exit_expr 9973)) 4112 (("\"" 1489) ("$" 9781) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 9782) ("[" 1508) (array_object_dereferenceable 9783) (callable_expr 9784) (callable_variable 9785) (class_constant 1512) (class_name 9786) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 9787) (name 9788) (simple_variable 9789) (static_member 9790) (variable 9971) (variable_class_name 9792)) 4113 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9970) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4114 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9969) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4115 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9968) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4116 (("(" 9964)) 4117 (("(" 9959)) 4118 (("(" 9953)) 4119 (("$" 9781) ("(" 9906) (T_ATTRIBUTE 2853) (T_CLASS 9907) (T_NAME_FULLY_QUALIFIED 9908) (T_NAME_QUALIFIED 9909) (T_NAME_RELATIVE 9910) (T_STATIC 9911) (T_STRING 9912) (T_VARIABLE 9782) (anonymous_class 9913) (attribute 2862) (attributes 9914) (class_name 9915) (class_name_reference 9916) (name 9917) (new_variable 9918) (simple_variable 9919)) 4120 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9905) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4121 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9904) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4122 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9903) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4123 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9902) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4124 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 9897) (T_END_HEREDOC 9898) (T_VARIABLE 2822) (encaps_list 9899) (encaps_var 2824)) 4125 ((T_FN 35) (T_FUNCTION 126) (fn 9542) #'9543 (inline_function 9896)) 4126 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9895) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4127 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9894) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4128 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9893) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4129 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9890) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4130 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9889) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4131 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 9885) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 4132 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 9883) (encaps_list 2729) (encaps_var 2730)) 4133 ((T_NULLSAFE_OBJECT_OPERATOR 9866) (T_OBJECT_OPERATOR 9867) ("[" 9868) ("{" 9869)) 4134 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 9863) (attribute 2654) (fn 9542) #'9543 (inline_function 9864)) 4135 (("(" 9742) (argument_list 9862)) 4136 ((T_PAAMAYIM_NEKUDOTAYIM 9857)) 4137 (("%" 9557) ("*" 9558) ("+" 9559) ("-" 9560) ("." 9561) ("/" 9562) ("<" 9563) (">" 9564) ("?" 9565) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9566) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9567) (T_BOOLEAN_AND 9568) (T_BOOLEAN_OR 9569) (T_COALESCE 9570) (T_DOUBLE_ARROW 9855) (T_INSTANCEOF 9571) (T_IS_EQUAL 9572) (T_IS_GREATER_OR_EQUAL 9573) (T_IS_IDENTICAL 9574) (T_IS_NOT_EQUAL 9575) (T_IS_NOT_IDENTICAL 9576) (T_IS_SMALLER_OR_EQUAL 9577) (T_LOGICAL_AND 9578) (T_LOGICAL_OR 9579) (T_LOGICAL_XOR 9580) (T_POW 9581) (T_SL 9582) (T_SPACESHIP 9583) (T_SR 9584) ("^" 9585) ("|" 9586)) 4138 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 9844)) 4139 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 9832)) 4140 (("(" 9742) (argument_list 9831)) 4141 (("=" 9750) (T_AND_EQUAL 9751) (T_COALESCE_EQUAL 9752) (T_CONCAT_EQUAL 9753) (T_DEC 9754) (T_DIV_EQUAL 9755) (T_INC 9756) (T_MINUS_EQUAL 9757) (T_MOD_EQUAL 9758) (T_MUL_EQUAL 9759) (T_OR_EQUAL 9760) (T_PLUS_EQUAL 9761) (T_POW_EQUAL 9762) (T_SL_EQUAL 9763) (T_SR_EQUAL 9764) (T_XOR_EQUAL 9765)) 4142 ((T_PAAMAYIM_NEKUDOTAYIM 9658)) 4143 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9556) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4144 (("%" 9557) ("*" 9558) ("+" 9559) ("-" 9560) ("." 9561) ("/" 9562) ("<" 9563) (">" 9564) ("?" 9565) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9566) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9567) (T_BOOLEAN_AND 9568) (T_BOOLEAN_OR 9569) (T_COALESCE 9570) (T_INSTANCEOF 9571) (T_IS_EQUAL 9572) (T_IS_GREATER_OR_EQUAL 9573) (T_IS_IDENTICAL 9574) (T_IS_NOT_EQUAL 9575) (T_IS_NOT_IDENTICAL 9576) (T_IS_SMALLER_OR_EQUAL 9577) (T_LOGICAL_AND 9578) (T_LOGICAL_OR 9579) (T_LOGICAL_XOR 9580) (T_POW 9581) (T_SL 9582) (T_SPACESHIP 9583) (T_SR 9584) ("^" 9585) ("|" 9586)) 4145 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9657) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4146 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9656) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4147 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9655) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4148 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9654) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4149 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9653) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4150 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9652) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4151 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9651) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4152 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9650) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4153 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 9645) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 9646) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 4154 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9644) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4155 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9643) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4156 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9642) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4157 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9641) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4158 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9640) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4159 (("$" 9602) ("(" 9603) (T_NAME_FULLY_QUALIFIED 9604) (T_NAME_QUALIFIED 9605) (T_NAME_RELATIVE 9606) (T_STATIC 9607) (T_STRING 9608) (T_VARIABLE 9609) (class_name 9610) (class_name_reference 9611) (name 9612) (new_variable 9613) (simple_variable 9614)) 4160 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9601) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4161 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9600) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4162 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9599) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4163 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9598) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4164 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9597) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4165 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9596) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4166 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9595) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4167 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9594) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4168 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9593) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4169 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9592) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4170 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9591) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4171 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9590) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4172 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9589) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4173 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9588) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4174 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9587) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4175 (("$" 9602) (T_VARIABLE 9609) (simple_variable 9636) ("{" 9637)) 4176 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 9634) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4177 ((T_PAAMAYIM_NEKUDOTAYIM 9632)) 4178 ((T_NULLSAFE_OBJECT_OPERATOR 9615) (T_OBJECT_OPERATOR 9616) (T_PAAMAYIM_NEKUDOTAYIM 9617) ("[" 9618) ("{" 9619)) 4179 (("$" 9602) (T_STRING 9625) (T_VARIABLE 9609) (property_name 9631) (simple_variable 9627) ("{" 9628)) 4180 (("$" 9602) (T_STRING 9625) (T_VARIABLE 9609) (property_name 9626) (simple_variable 9627) ("{" 9628)) 4181 (("$" 9602) (T_VARIABLE 9609) (simple_variable 9624)) 4182 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 9622) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4183 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9620) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4184 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9621)) 4185 (("]" 9623)) 4186 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9629) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4187 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9630)) 4188 (("$" 9602) (T_VARIABLE 9609) (simple_variable 9633)) 4189 (("%" 280) (")" 9635) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4190 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9638) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4191 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9639)) 4192 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9649) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4193 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 9647) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 4194 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9648) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4195 (("$" 9482) (T_ABSTRACT 9659) (T_ARRAY 9660) (T_AS 9661) (T_BREAK 9662) (T_CALLABLE 9663) (T_CASE 9664) (T_CATCH 9665) (T_CLASS 9666) (T_CLASS_C 9667) (T_CLONE 9668) (T_CONST 9669) (T_CONTINUE 9670) (T_DECLARE 9671) (T_DEFAULT 9672) (T_DIR 9673) (T_DO 9674) (T_ECHO 9675) (T_ELSE 9676) (T_ELSEIF 9677) (T_EMPTY 9678) (T_ENDDECLARE 9679) (T_ENDFOR 9680) (T_ENDFOREACH 9681) (T_ENDIF 9682) (T_ENDSWITCH 9683) (T_ENDWHILE 9684) (T_ENUM 9685) (T_EVAL 9686) (T_EXIT 9687) (T_EXTENDS 9688) (T_FILE 9689) (T_FINAL 9690) (T_FINALLY 9691) (T_FN 9692) (T_FOR 9693) (T_FOREACH 9694) (T_FUNCTION 9695) (T_FUNC_C 9696) (T_GLOBAL 9697) (T_GOTO 9698) (T_IF 9699) (T_IMPLEMENTS 9700) (T_INCLUDE 9701) (T_INCLUDE_ONCE 9702) (T_INSTANCEOF 9703) (T_INSTEADOF 9704) (T_INTERFACE 9705) (T_ISSET 9706) (T_LINE 9707) (T_LIST 9708) (T_LOGICAL_AND 9709) (T_LOGICAL_OR 9710) (T_LOGICAL_XOR 9711) (T_MATCH 9712) (T_METHOD_C 9713) (T_NAMESPACE 9714) (T_NEW 9715) (T_NS_C 9716) (T_PRINT 9717) (T_PRIVATE 9718) (T_PROTECTED 9719) (T_PUBLIC 9720) (T_READONLY 9721) (T_REQUIRE 9722) (T_REQUIRE_ONCE 9723) (T_RETURN 9724) (T_STATIC 9725) (T_STRING 9726) (T_SWITCH 9727) (T_THROW 9728) (T_TRAIT 9729) (T_TRAIT_C 9730) (T_TRY 9731) (T_UNSET 9732) (T_USE 9733) (T_VAR 9734) (T_VARIABLE 9528) (T_WHILE 9735) (T_YIELD 9736) (identifier 9737) (member_name 9738) (reserved_non_modifiers 9739) (semi_reserved 9740) (simple_variable 9741) ("{" 914)) 4196 (("(" 9742) (argument_list 9743)) 4197 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 9744) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 9745) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 9746) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4198 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 9749) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4199 (("," 1449) (possible_comma 9747)) 4200 ((")" 9748)) 4201 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (ampersand 9779) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9780) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4202 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9778) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4203 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9777) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4204 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9776) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4205 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9775) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4206 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9774) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4207 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9773) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4208 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9772) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4209 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9771) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4210 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9770) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4211 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9769) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4212 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9768) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4213 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9767) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4214 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9766) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4215 (("\"" 1489) ("$" 9781) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 9782) ("[" 1508) (array_object_dereferenceable 9783) (callable_expr 9784) (callable_variable 9785) (class_constant 1512) (class_name 9786) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 9787) (name 9788) (simple_variable 9789) (static_member 9790) (variable 9791) (variable_class_name 9792)) 4216 (("$" 9781) (T_VARIABLE 9782) (simple_variable 9827) ("{" 9828)) 4217 ((T_NULLSAFE_OBJECT_OPERATOR 9810) (T_OBJECT_OPERATOR 9811) ("[" 9812) ("{" 9813)) 4218 (("(" 9796) (argument_list 9809)) 4219 ((T_PAAMAYIM_NEKUDOTAYIM 9805)) 4220 (("(" 9796) (argument_list 9804)) 4221 ((T_PAAMAYIM_NEKUDOTAYIM 9793)) 4222 (("$" 9781) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 9782) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 9794) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 9795) ("{" 914)) 4223 (("(" 9796) (argument_list 9797)) 4224 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 9798) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 9799) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 9800) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4225 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 9803) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4226 (("," 1449) (possible_comma 9801)) 4227 ((")" 9802)) 4228 (("$" 9781) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 9782) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 9806) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 9807) ("{" 914)) 4229 (("(" 9796) (argument_list 9808)) 4230 (("$" 9781) (T_STRING 9818) (T_VARIABLE 9782) (property_name 9825) (simple_variable 9820) ("{" 9821)) 4231 (("$" 9781) (T_STRING 9818) (T_VARIABLE 9782) (property_name 9819) (simple_variable 9820) ("{" 9821)) 4232 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 9816) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4233 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9814) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4234 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9815)) 4235 (("]" 9817)) 4236 (("(" 9796) (argument_list 9824)) 4237 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9822) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4238 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9823)) 4239 (("(" 9796) (argument_list 9826)) 4240 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9829) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4241 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9830)) 4242 ((backup_doc_comment 9833)) 4243 (("(" 9834)) 4244 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 9835)) 4245 ((")" 9836)) 4246 ((T_USE 148) (lexical_vars 9837)) 4247 ((":" 150) (return_type 9838)) 4248 ((backup_fn_flags 9839)) 4249 (("{" 9840)) 4250 ((inner_statement_list 9841)) 4251 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 9842) ("~" 124)) 4252 ((backup_fn_flags 9843)) 4253 ((backup_doc_comment 9845)) 4254 (("(" 9846)) 4255 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 9847)) 4256 ((")" 9848)) 4257 ((":" 2207) (return_type 9849)) 4258 ((T_DOUBLE_ARROW 9850)) 4259 ((backup_fn_flags 9851)) 4260 ((backup_lex_pos 9852)) 4261 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9853) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4262 (("%" 9557) ("*" 9558) ("+" 9559) ("-" 9560) ("." 9561) ("/" 9562) ("<" 9563) (">" 9564) ("?" 9565) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9566) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9567) (T_BOOLEAN_AND 9568) (T_BOOLEAN_OR 9569) (T_COALESCE 9570) (T_INSTANCEOF 9571) (T_IS_EQUAL 9572) (T_IS_GREATER_OR_EQUAL 9573) (T_IS_IDENTICAL 9574) (T_IS_NOT_EQUAL 9575) (T_IS_NOT_IDENTICAL 9576) (T_IS_SMALLER_OR_EQUAL 9577) (T_LOGICAL_AND 9578) (T_LOGICAL_OR 9579) (T_LOGICAL_XOR 9580) (T_POW 9581) (T_SL 9582) (T_SPACESHIP 9583) (T_SR 9584) ("^" 9585) (backup_fn_flags 9854) ("|" 9586)) 4263 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 9856) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4264 (("$" 9482) (T_ABSTRACT 9659) (T_ARRAY 9660) (T_AS 9661) (T_BREAK 9662) (T_CALLABLE 9663) (T_CASE 9664) (T_CATCH 9665) (T_CLASS 9666) (T_CLASS_C 9667) (T_CLONE 9668) (T_CONST 9669) (T_CONTINUE 9670) (T_DECLARE 9671) (T_DEFAULT 9672) (T_DIR 9673) (T_DO 9674) (T_ECHO 9675) (T_ELSE 9676) (T_ELSEIF 9677) (T_EMPTY 9678) (T_ENDDECLARE 9679) (T_ENDFOR 9680) (T_ENDFOREACH 9681) (T_ENDIF 9682) (T_ENDSWITCH 9683) (T_ENDWHILE 9684) (T_ENUM 9685) (T_EVAL 9686) (T_EXIT 9687) (T_EXTENDS 9688) (T_FILE 9689) (T_FINAL 9690) (T_FINALLY 9691) (T_FN 9692) (T_FOR 9693) (T_FOREACH 9694) (T_FUNCTION 9695) (T_FUNC_C 9696) (T_GLOBAL 9697) (T_GOTO 9698) (T_IF 9699) (T_IMPLEMENTS 9700) (T_INCLUDE 9701) (T_INCLUDE_ONCE 9702) (T_INSTANCEOF 9703) (T_INSTEADOF 9704) (T_INTERFACE 9705) (T_ISSET 9706) (T_LINE 9707) (T_LIST 9708) (T_LOGICAL_AND 9709) (T_LOGICAL_OR 9710) (T_LOGICAL_XOR 9711) (T_MATCH 9712) (T_METHOD_C 9713) (T_NAMESPACE 9714) (T_NEW 9715) (T_NS_C 9716) (T_PRINT 9717) (T_PRIVATE 9718) (T_PROTECTED 9719) (T_PUBLIC 9720) (T_READONLY 9721) (T_REQUIRE 9722) (T_REQUIRE_ONCE 9723) (T_RETURN 9724) (T_STATIC 9725) (T_STRING 9726) (T_SWITCH 9727) (T_THROW 9728) (T_TRAIT 9729) (T_TRAIT_C 9730) (T_TRY 9731) (T_UNSET 9732) (T_USE 9733) (T_VAR 9734) (T_VARIABLE 9528) (T_WHILE 9735) (T_YIELD 9736) (identifier 9858) (member_name 9859) (reserved_non_modifiers 9739) (semi_reserved 9740) (simple_variable 9860) ("{" 914)) 4265 (("(" 9742) (argument_list 9861)) 4266 ((T_FN 35) (T_FUNCTION 126) (fn 9542) #'9543 (inline_function 9865)) 4267 (("$" 9482) (T_STRING 9874) (T_VARIABLE 9528) (property_name 9881) (simple_variable 9876) ("{" 9877)) 4268 (("$" 9482) (T_STRING 9874) (T_VARIABLE 9528) (property_name 9875) (simple_variable 9876) ("{" 9877)) 4269 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 9872) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4270 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9870) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4271 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9871)) 4272 (("]" 9873)) 4273 (("(" 9742) (argument_list 9880)) 4274 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9878) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4275 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9879)) 4276 (("(" 9742) (argument_list 9882)) 4277 (("`" 9884)) 4278 (("]" 9886)) 4279 (("=" 9887)) 4280 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9888) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4281 (("%" 9557) ("*" 9558) ("+" 9559) ("-" 9560) ("." 9561) ("/" 9562) ("<" 9563) (">" 9564) ("?" 9565) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9566) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9567) (T_BOOLEAN_AND 9568) (T_BOOLEAN_OR 9569) (T_COALESCE 9570) (T_DOUBLE_ARROW 9891) (T_INSTANCEOF 9571) (T_IS_EQUAL 9572) (T_IS_GREATER_OR_EQUAL 9573) (T_IS_IDENTICAL 9574) (T_IS_NOT_EQUAL 9575) (T_IS_NOT_IDENTICAL 9576) (T_IS_SMALLER_OR_EQUAL 9577) (T_LOGICAL_AND 9578) (T_LOGICAL_OR 9579) (T_LOGICAL_XOR 9580) (T_POW 9581) (T_SL 9582) (T_SPACESHIP 9583) (T_SR 9584) ("^" 9585) ("|" 9586)) 4282 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9892) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4283 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 9901) (T_VARIABLE 2822) (encaps_var 2836)) 4284 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 9900) (T_VARIABLE 2822) (encaps_var 2827)) 4285 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 9951) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4286 (("(" 2899) (argument_list 2900) (ctor_arguments 9944)) 4287 ((T_ATTRIBUTE 2853) (T_CLASS 9907) (anonymous_class 9943) (attribute 2898)) 4288 ((T_PAAMAYIM_NEKUDOTAYIM 9941)) 4289 (("(" 9932) (argument_list 9933) (ctor_arguments 9934)) 4290 ((T_NULLSAFE_OBJECT_OPERATOR 9920) (T_OBJECT_OPERATOR 9921) (T_PAAMAYIM_NEKUDOTAYIM 9922) ("[" 9923) ("{" 9924)) 4291 (("$" 9781) (T_STRING 9818) (T_VARIABLE 9782) (property_name 9931) (simple_variable 9820) ("{" 9821)) 4292 (("$" 9781) (T_STRING 9818) (T_VARIABLE 9782) (property_name 9930) (simple_variable 9820) ("{" 9821)) 4293 (("$" 9781) (T_VARIABLE 9782) (simple_variable 9929)) 4294 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 9927) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4295 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9925) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4296 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9926)) 4297 (("]" 9928)) 4298 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 9935) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 9936) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 9937) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4299 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 9940) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4300 (("," 1449) (possible_comma 9938)) 4301 ((")" 9939)) 4302 (("$" 9781) (T_VARIABLE 9782) (simple_variable 9942)) 4303 ((T_EXTENDS 2902) (extends_from 9945)) 4304 ((T_IMPLEMENTS 2904) (implements_list 9946)) 4305 ((backup_doc_comment 9947)) 4306 (("{" 9948)) 4307 ((class_statement_list 9949)) 4308 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 9950)) 4309 (("%" 280) (")" 9952) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4310 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 9954) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4311 (("%" 280) (")" 9955) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4312 (("{" 9956)) 4313 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 9957) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 4314 (("}" 9958)) 4315 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 9960) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 4316 ((")" 9961)) 4317 (("=" 9962)) 4318 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 9963) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4319 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 9965) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4320 (("," 5117) (possible_comma 9966)) 4321 ((")" 9967)) 4322 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 9974) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4323 ((")" 9975)) 4324 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 9977) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4325 (("%" 280) (")" 9978) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4326 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 9980) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4327 (("%" 280) (")" 9981) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4328 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 9988) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 4329 ((")" 9989)) 4330 (("%" 280) (")" 9994) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4331 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 9997) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4332 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 9998)) 4333 (("\"" 10000) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 4334 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 10010) (T_VARIABLE 2822) (encaps_var 2836)) 4335 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 10009) (T_VARIABLE 2822) (encaps_var 2827)) 4336 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10060) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4337 (("(" 2899) (argument_list 2900) (ctor_arguments 10053)) 4338 ((T_ATTRIBUTE 2853) (T_CLASS 10016) (anonymous_class 10052) (attribute 2898)) 4339 ((T_PAAMAYIM_NEKUDOTAYIM 10050)) 4340 (("(" 10041) (argument_list 10042) (ctor_arguments 10043)) 4341 ((T_NULLSAFE_OBJECT_OPERATOR 10029) (T_OBJECT_OPERATOR 10030) (T_PAAMAYIM_NEKUDOTAYIM 10031) ("[" 10032) ("{" 10033)) 4342 (("$" 9372) (T_STRING 9409) (T_VARIABLE 9373) (property_name 10040) (simple_variable 9411) ("{" 9412)) 4343 (("$" 9372) (T_STRING 9409) (T_VARIABLE 9373) (property_name 10039) (simple_variable 9411) ("{" 9412)) 4344 (("$" 9372) (T_VARIABLE 9373) (simple_variable 10038)) 4345 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 10036) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4346 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10034) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4347 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10035)) 4348 (("]" 10037)) 4349 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10044) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 10045) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 10046) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4350 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10049) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4351 (("," 1449) (possible_comma 10047)) 4352 ((")" 10048)) 4353 (("$" 9372) (T_VARIABLE 9373) (simple_variable 10051)) 4354 ((T_EXTENDS 2902) (extends_from 10054)) 4355 ((T_IMPLEMENTS 2904) (implements_list 10055)) 4356 ((backup_doc_comment 10056)) 4357 (("{" 10057)) 4358 ((class_statement_list 10058)) 4359 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 10059)) 4360 (("%" 280) (")" 10061) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4361 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10063) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4362 (("%" 280) (")" 10064) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4363 (("{" 10065)) 4364 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 10066) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 4365 (("}" 10067)) 4366 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 10069) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 4367 ((")" 10070)) 4368 (("=" 10071)) 4369 (("!" 9071) ("\"" 9072) ("$" 9073) ("(" 9074) ("+" 9075) ("-" 9076) ("@" 9077) (T_ARRAY 9078) (T_ARRAY_CAST 9079) (T_ATTRIBUTE 125) (T_BOOL_CAST 9080) (T_CLASS_C 9081) (T_CLONE 9082) (T_CONSTANT_ENCAPSED_STRING 9083) (T_DEC 9084) (T_DIR 9085) (T_DNUMBER 9086) (T_DOUBLE_CAST 9087) (T_EMPTY 9088) (T_EVAL 9089) (T_EXIT 9090) (T_FILE 9091) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9092) (T_INC 9093) (T_INCLUDE 9094) (T_INCLUDE_ONCE 9095) (T_INT_CAST 9096) (T_ISSET 9097) (T_LINE 9098) (T_LIST 9099) (T_LNUMBER 9100) (T_MATCH 9101) (T_METHOD_C 9102) (T_NAME_FULLY_QUALIFIED 9103) (T_NAME_QUALIFIED 9104) (T_NAME_RELATIVE 9105) (T_NEW 9106) (T_NS_C 9107) (T_OBJECT_CAST 9108) (T_PRINT 9109) (T_REQUIRE 9110) (T_REQUIRE_ONCE 9111) (T_START_HEREDOC 9112) (T_STATIC 9113) (T_STRING 9114) (T_STRING_CAST 9115) (T_THROW 9116) (T_TRAIT_C 9117) (T_UNSET_CAST 9118) (T_VARIABLE 9119) (T_YIELD 9120) (T_YIELD_FROM 9121) ("[" 9122) ("`" 9123) (array_object_dereferenceable 9124) (attribute 129) (attributes 9125) (callable_expr 9126) (callable_variable 9127) (class_constant 9128) (class_name 9129) (constant 9130) (dereferenceable_scalar 9131) (expr 10072) (fn 9133) (fully_dereferenceable 103) #'9134 (function_call 9135) (inline_function 9136) (internal_functions_in_yacc 9137) (match 9138) (name 9139) (new_expr 9140) (scalar 9141) (simple_variable 9142) (static_member 9143) (variable 9144) (variable_class_name 9145) ("~" 9146)) 4370 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 10074) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4371 (("," 5117) (possible_comma 10075)) 4372 ((")" 10076)) 4373 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 10083) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4374 ((")" 10084)) 4375 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10086) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4376 (("%" 280) (")" 10087) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4377 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10089) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4378 (("%" 280) (")" 10090) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4379 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 10097) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 4380 ((")" 10098)) 4381 (("%" 280) (")" 10103) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4382 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10106) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4383 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10107)) 4384 (("\"" 10109) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 4385 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10112) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4386 (("$" 653) (T_ABSTRACT 831) (T_ARRAY 832) (T_AS 833) (T_BREAK 834) (T_CALLABLE 835) (T_CASE 836) (T_CATCH 837) (T_CLASS 838) (T_CLASS_C 839) (T_CLONE 840) (T_CONST 841) (T_CONTINUE 842) (T_DECLARE 843) (T_DEFAULT 844) (T_DIR 845) (T_DO 846) (T_ECHO 847) (T_ELSE 848) (T_ELSEIF 849) (T_EMPTY 850) (T_ENDDECLARE 851) (T_ENDFOR 852) (T_ENDFOREACH 853) (T_ENDIF 854) (T_ENDSWITCH 855) (T_ENDWHILE 856) (T_ENUM 857) (T_EVAL 858) (T_EXIT 859) (T_EXTENDS 860) (T_FILE 861) (T_FINAL 862) (T_FINALLY 863) (T_FN 864) (T_FOR 865) (T_FOREACH 866) (T_FUNCTION 867) (T_FUNC_C 868) (T_GLOBAL 869) (T_GOTO 870) (T_IF 871) (T_IMPLEMENTS 872) (T_INCLUDE 873) (T_INCLUDE_ONCE 874) (T_INSTANCEOF 875) (T_INSTEADOF 876) (T_INTERFACE 877) (T_ISSET 878) (T_LINE 879) (T_LIST 880) (T_LOGICAL_AND 881) (T_LOGICAL_OR 882) (T_LOGICAL_XOR 883) (T_MATCH 884) (T_METHOD_C 885) (T_NAMESPACE 886) (T_NEW 887) (T_NS_C 888) (T_PRINT 889) (T_PRIVATE 890) (T_PROTECTED 891) (T_PUBLIC 892) (T_READONLY 893) (T_REQUIRE 894) (T_REQUIRE_ONCE 895) (T_RETURN 896) (T_STATIC 897) (T_STRING 898) (T_SWITCH 899) (T_THROW 900) (T_TRAIT 901) (T_TRAIT_C 902) (T_TRY 903) (T_UNSET 904) (T_USE 905) (T_VAR 906) (T_VARIABLE 700) (T_WHILE 907) (T_YIELD 908) (identifier 10114) (member_name 10115) (reserved_non_modifiers 911) (semi_reserved 912) (simple_variable 10116) ("{" 914)) 4387 (("(" 917) (argument_list 10117)) 4388 ((T_FN 35) (T_FUNCTION 126) (fn 714) #'715 (inline_function 10121)) 4389 (("$" 653) (T_STRING 10130) (T_VARIABLE 700) (property_name 10137) (simple_variable 10132) ("{" 10133)) 4390 (("$" 653) (T_STRING 10130) (T_VARIABLE 700) (property_name 10131) (simple_variable 10132) ("{" 10133)) 4391 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 10128) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4392 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10126) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4393 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10127)) 4394 (("]" 10129)) 4395 (("(" 917) (argument_list 10136)) 4396 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10134) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4397 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10135)) 4398 (("(" 917) (argument_list 10138)) 4399 (("`" 10140)) 4400 (("]" 10142)) 4401 (("=" 10143)) 4402 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10144) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 4403 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10641) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4404 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 10639) (encaps_var 5158)) 4405 (("$" 10148) (T_VARIABLE 10194) (simple_variable 10635) ("{" 10636)) 4406 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10633) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4407 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10632) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4408 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10631) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4409 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10630) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4410 (("(" 10627)) 4411 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10626) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4412 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10625) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4413 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10624) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4414 (("\"" 1489) ("$" 10447) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 10448) ("[" 1508) (array_object_dereferenceable 10449) (callable_expr 10450) (callable_variable 10451) (class_constant 1512) (class_name 10452) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 10453) (name 10454) (simple_variable 10455) (static_member 10456) (variable 10623) (variable_class_name 10458)) 4415 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10622) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4416 (("(" 10619)) 4417 (("(" 10616)) 4418 (("(" 10612) (exit_expr 10613)) 4419 (("\"" 1489) ("$" 10447) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 10448) ("[" 1508) (array_object_dereferenceable 10449) (callable_expr 10450) (callable_variable 10451) (class_constant 1512) (class_name 10452) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 10453) (name 10454) (simple_variable 10455) (static_member 10456) (variable 10611) (variable_class_name 10458)) 4420 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10610) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4421 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10609) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4422 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10608) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4423 (("(" 10604)) 4424 (("(" 10599)) 4425 (("(" 10593)) 4426 (("$" 9781) ("(" 9906) (T_ATTRIBUTE 2853) (T_CLASS 10572) (T_NAME_FULLY_QUALIFIED 9908) (T_NAME_QUALIFIED 9909) (T_NAME_RELATIVE 9910) (T_STATIC 9911) (T_STRING 9912) (T_VARIABLE 9782) (anonymous_class 10573) (attribute 2862) (attributes 10574) (class_name 9915) (class_name_reference 10575) (name 9917) (new_variable 9918) (simple_variable 9919)) 4427 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10571) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4428 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10570) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4429 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10569) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4430 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10568) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4431 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 10563) (T_END_HEREDOC 10564) (T_VARIABLE 2822) (encaps_list 10565) (encaps_var 2824)) 4432 ((T_FN 35) (T_FUNCTION 126) (fn 10208) #'10209 (inline_function 10562)) 4433 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10561) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4434 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10560) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4435 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10559) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4436 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10556) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4437 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10555) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4438 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 10551) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 4439 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 10549) (encaps_list 2729) (encaps_var 2730)) 4440 ((T_NULLSAFE_OBJECT_OPERATOR 10532) (T_OBJECT_OPERATOR 10533) ("[" 10534) ("{" 10535)) 4441 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 10529) (attribute 2654) (fn 10208) #'10209 (inline_function 10530)) 4442 (("(" 10408) (argument_list 10528)) 4443 ((T_PAAMAYIM_NEKUDOTAYIM 10523)) 4444 (("%" 10223) ("*" 10224) ("+" 10225) ("-" 10226) ("." 10227) ("/" 10228) ("<" 10229) (">" 10230) ("?" 10231) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10232) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10233) (T_BOOLEAN_AND 10234) (T_BOOLEAN_OR 10235) (T_COALESCE 10236) (T_DOUBLE_ARROW 10521) (T_INSTANCEOF 10237) (T_IS_EQUAL 10238) (T_IS_GREATER_OR_EQUAL 10239) (T_IS_IDENTICAL 10240) (T_IS_NOT_EQUAL 10241) (T_IS_NOT_IDENTICAL 10242) (T_IS_SMALLER_OR_EQUAL 10243) (T_LOGICAL_AND 10244) (T_LOGICAL_OR 10245) (T_LOGICAL_XOR 10246) (T_POW 10247) (T_SL 10248) (T_SPACESHIP 10249) (T_SR 10250) ("^" 10251) ("|" 10252)) 4445 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 10510)) 4446 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 10498)) 4447 (("(" 10408) (argument_list 10497)) 4448 (("=" 10416) (T_AND_EQUAL 10417) (T_COALESCE_EQUAL 10418) (T_CONCAT_EQUAL 10419) (T_DEC 10420) (T_DIV_EQUAL 10421) (T_INC 10422) (T_MINUS_EQUAL 10423) (T_MOD_EQUAL 10424) (T_MUL_EQUAL 10425) (T_OR_EQUAL 10426) (T_PLUS_EQUAL 10427) (T_POW_EQUAL 10428) (T_SL_EQUAL 10429) (T_SR_EQUAL 10430) (T_XOR_EQUAL 10431)) 4449 ((T_PAAMAYIM_NEKUDOTAYIM 10324)) 4450 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10222) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4451 (("%" 10223) ("*" 10224) ("+" 10225) ("-" 10226) ("." 10227) ("/" 10228) ("<" 10229) (">" 10230) ("?" 10231) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10232) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10233) (T_BOOLEAN_AND 10234) (T_BOOLEAN_OR 10235) (T_COALESCE 10236) (T_INSTANCEOF 10237) (T_IS_EQUAL 10238) (T_IS_GREATER_OR_EQUAL 10239) (T_IS_IDENTICAL 10240) (T_IS_NOT_EQUAL 10241) (T_IS_NOT_IDENTICAL 10242) (T_IS_SMALLER_OR_EQUAL 10243) (T_LOGICAL_AND 10244) (T_LOGICAL_OR 10245) (T_LOGICAL_XOR 10246) (T_POW 10247) (T_SL 10248) (T_SPACESHIP 10249) (T_SR 10250) ("^" 10251) ("|" 10252)) 4452 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10323) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4453 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10322) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4454 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10321) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4455 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10320) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4456 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10319) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4457 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10318) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4458 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10317) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4459 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10316) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4460 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 10311) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10312) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 4461 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10310) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4462 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10309) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4463 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10308) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4464 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10307) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4465 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10306) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4466 (("$" 10268) ("(" 10269) (T_NAME_FULLY_QUALIFIED 10270) (T_NAME_QUALIFIED 10271) (T_NAME_RELATIVE 10272) (T_STATIC 10273) (T_STRING 10274) (T_VARIABLE 10275) (class_name 10276) (class_name_reference 10277) (name 10278) (new_variable 10279) (simple_variable 10280)) 4467 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10267) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4468 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10266) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4469 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10265) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4470 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10264) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4471 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10263) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4472 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10262) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4473 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10261) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4474 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10260) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4475 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10259) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4476 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10258) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4477 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10257) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4478 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10256) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4479 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10255) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4480 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10254) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4481 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10253) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4482 (("$" 10268) (T_VARIABLE 10275) (simple_variable 10302) ("{" 10303)) 4483 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10300) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4484 ((T_PAAMAYIM_NEKUDOTAYIM 10298)) 4485 ((T_NULLSAFE_OBJECT_OPERATOR 10281) (T_OBJECT_OPERATOR 10282) (T_PAAMAYIM_NEKUDOTAYIM 10283) ("[" 10284) ("{" 10285)) 4486 (("$" 10268) (T_STRING 10291) (T_VARIABLE 10275) (property_name 10297) (simple_variable 10293) ("{" 10294)) 4487 (("$" 10268) (T_STRING 10291) (T_VARIABLE 10275) (property_name 10292) (simple_variable 10293) ("{" 10294)) 4488 (("$" 10268) (T_VARIABLE 10275) (simple_variable 10290)) 4489 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 10288) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4490 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10286) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4491 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10287)) 4492 (("]" 10289)) 4493 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10295) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4494 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10296)) 4495 (("$" 10268) (T_VARIABLE 10275) (simple_variable 10299)) 4496 (("%" 280) (")" 10301) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4497 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10304) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4498 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10305)) 4499 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10315) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4500 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 10313) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 4501 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10314) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4502 (("$" 10148) (T_ABSTRACT 10325) (T_ARRAY 10326) (T_AS 10327) (T_BREAK 10328) (T_CALLABLE 10329) (T_CASE 10330) (T_CATCH 10331) (T_CLASS 10332) (T_CLASS_C 10333) (T_CLONE 10334) (T_CONST 10335) (T_CONTINUE 10336) (T_DECLARE 10337) (T_DEFAULT 10338) (T_DIR 10339) (T_DO 10340) (T_ECHO 10341) (T_ELSE 10342) (T_ELSEIF 10343) (T_EMPTY 10344) (T_ENDDECLARE 10345) (T_ENDFOR 10346) (T_ENDFOREACH 10347) (T_ENDIF 10348) (T_ENDSWITCH 10349) (T_ENDWHILE 10350) (T_ENUM 10351) (T_EVAL 10352) (T_EXIT 10353) (T_EXTENDS 10354) (T_FILE 10355) (T_FINAL 10356) (T_FINALLY 10357) (T_FN 10358) (T_FOR 10359) (T_FOREACH 10360) (T_FUNCTION 10361) (T_FUNC_C 10362) (T_GLOBAL 10363) (T_GOTO 10364) (T_IF 10365) (T_IMPLEMENTS 10366) (T_INCLUDE 10367) (T_INCLUDE_ONCE 10368) (T_INSTANCEOF 10369) (T_INSTEADOF 10370) (T_INTERFACE 10371) (T_ISSET 10372) (T_LINE 10373) (T_LIST 10374) (T_LOGICAL_AND 10375) (T_LOGICAL_OR 10376) (T_LOGICAL_XOR 10377) (T_MATCH 10378) (T_METHOD_C 10379) (T_NAMESPACE 10380) (T_NEW 10381) (T_NS_C 10382) (T_PRINT 10383) (T_PRIVATE 10384) (T_PROTECTED 10385) (T_PUBLIC 10386) (T_READONLY 10387) (T_REQUIRE 10388) (T_REQUIRE_ONCE 10389) (T_RETURN 10390) (T_STATIC 10391) (T_STRING 10392) (T_SWITCH 10393) (T_THROW 10394) (T_TRAIT 10395) (T_TRAIT_C 10396) (T_TRY 10397) (T_UNSET 10398) (T_USE 10399) (T_VAR 10400) (T_VARIABLE 10194) (T_WHILE 10401) (T_YIELD 10402) (identifier 10403) (member_name 10404) (reserved_non_modifiers 10405) (semi_reserved 10406) (simple_variable 10407) ("{" 914)) 4503 (("(" 10408) (argument_list 10409)) 4504 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10410) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 10411) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 10412) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4505 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10415) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4506 (("," 1449) (possible_comma 10413)) 4507 ((")" 10414)) 4508 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (ampersand 10445) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10446) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4509 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10444) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4510 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10443) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4511 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10442) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4512 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10441) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4513 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10440) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4514 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10439) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4515 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10438) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4516 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10437) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4517 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10436) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4518 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10435) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4519 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10434) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4520 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10433) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4521 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10432) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4522 (("\"" 1489) ("$" 10447) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 10448) ("[" 1508) (array_object_dereferenceable 10449) (callable_expr 10450) (callable_variable 10451) (class_constant 1512) (class_name 10452) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 10453) (name 10454) (simple_variable 10455) (static_member 10456) (variable 10457) (variable_class_name 10458)) 4523 (("$" 10447) (T_VARIABLE 10448) (simple_variable 10493) ("{" 10494)) 4524 ((T_NULLSAFE_OBJECT_OPERATOR 10476) (T_OBJECT_OPERATOR 10477) ("[" 10478) ("{" 10479)) 4525 (("(" 10462) (argument_list 10475)) 4526 ((T_PAAMAYIM_NEKUDOTAYIM 10471)) 4527 (("(" 10462) (argument_list 10470)) 4528 ((T_PAAMAYIM_NEKUDOTAYIM 10459)) 4529 (("$" 10447) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 10448) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 10460) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 10461) ("{" 914)) 4530 (("(" 10462) (argument_list 10463)) 4531 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10464) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 10465) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 10466) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4532 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10469) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4533 (("," 1449) (possible_comma 10467)) 4534 ((")" 10468)) 4535 (("$" 10447) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 10448) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 10472) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 10473) ("{" 914)) 4536 (("(" 10462) (argument_list 10474)) 4537 (("$" 10447) (T_STRING 10484) (T_VARIABLE 10448) (property_name 10491) (simple_variable 10486) ("{" 10487)) 4538 (("$" 10447) (T_STRING 10484) (T_VARIABLE 10448) (property_name 10485) (simple_variable 10486) ("{" 10487)) 4539 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 10482) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4540 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10480) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4541 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10481)) 4542 (("]" 10483)) 4543 (("(" 10462) (argument_list 10490)) 4544 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10488) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4545 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10489)) 4546 (("(" 10462) (argument_list 10492)) 4547 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10495) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4548 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10496)) 4549 ((backup_doc_comment 10499)) 4550 (("(" 10500)) 4551 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 10501)) 4552 ((")" 10502)) 4553 ((T_USE 148) (lexical_vars 10503)) 4554 ((":" 150) (return_type 10504)) 4555 ((backup_fn_flags 10505)) 4556 (("{" 10506)) 4557 ((inner_statement_list 10507)) 4558 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 10508) ("~" 124)) 4559 ((backup_fn_flags 10509)) 4560 ((backup_doc_comment 10511)) 4561 (("(" 10512)) 4562 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 10513)) 4563 ((")" 10514)) 4564 ((":" 2207) (return_type 10515)) 4565 ((T_DOUBLE_ARROW 10516)) 4566 ((backup_fn_flags 10517)) 4567 ((backup_lex_pos 10518)) 4568 (("!" 9480) ("\"" 9481) ("$" 9482) ("(" 9483) ("+" 9484) ("-" 9485) ("@" 9486) (T_ARRAY 9487) (T_ARRAY_CAST 9488) (T_ATTRIBUTE 125) (T_BOOL_CAST 9489) (T_CLASS_C 9490) (T_CLONE 9491) (T_CONSTANT_ENCAPSED_STRING 9492) (T_DEC 9493) (T_DIR 9494) (T_DNUMBER 9495) (T_DOUBLE_CAST 9496) (T_EMPTY 9497) (T_EVAL 9498) (T_EXIT 9499) (T_FILE 9500) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 9501) (T_INC 9502) (T_INCLUDE 9503) (T_INCLUDE_ONCE 9504) (T_INT_CAST 9505) (T_ISSET 9506) (T_LINE 9507) (T_LIST 9508) (T_LNUMBER 9509) (T_MATCH 9510) (T_METHOD_C 9511) (T_NAME_FULLY_QUALIFIED 9512) (T_NAME_QUALIFIED 9513) (T_NAME_RELATIVE 9514) (T_NEW 9515) (T_NS_C 9516) (T_OBJECT_CAST 9517) (T_PRINT 9518) (T_REQUIRE 9519) (T_REQUIRE_ONCE 9520) (T_START_HEREDOC 9521) (T_STATIC 9522) (T_STRING 9523) (T_STRING_CAST 9524) (T_THROW 9525) (T_TRAIT_C 9526) (T_UNSET_CAST 9527) (T_VARIABLE 9528) (T_YIELD 9529) (T_YIELD_FROM 9530) ("[" 9531) ("`" 9532) (array_object_dereferenceable 9533) (attribute 129) (attributes 9534) (callable_expr 9535) (callable_variable 9536) (class_constant 9537) (class_name 9538) (constant 9539) (dereferenceable_scalar 9540) (expr 10519) (fn 9542) (fully_dereferenceable 103) #'9543 (function_call 9544) (inline_function 9545) (internal_functions_in_yacc 9546) (match 9547) (name 9548) (new_expr 9549) (scalar 9550) (simple_variable 9551) (static_member 9552) (variable 9553) (variable_class_name 9554) ("~" 9555)) 4569 (("%" 9557) ("*" 9558) ("+" 9559) ("-" 9560) ("." 9561) ("/" 9562) ("<" 9563) (">" 9564) ("?" 9565) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9566) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9567) (T_BOOLEAN_AND 9568) (T_BOOLEAN_OR 9569) (T_COALESCE 9570) (T_INSTANCEOF 9571) (T_IS_EQUAL 9572) (T_IS_GREATER_OR_EQUAL 9573) (T_IS_IDENTICAL 9574) (T_IS_NOT_EQUAL 9575) (T_IS_NOT_IDENTICAL 9576) (T_IS_SMALLER_OR_EQUAL 9577) (T_LOGICAL_AND 9578) (T_LOGICAL_OR 9579) (T_LOGICAL_XOR 9580) (T_POW 9581) (T_SL 9582) (T_SPACESHIP 9583) (T_SR 9584) ("^" 9585) (backup_fn_flags 10520) ("|" 9586)) 4570 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10522) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 4571 (("$" 10148) (T_ABSTRACT 10325) (T_ARRAY 10326) (T_AS 10327) (T_BREAK 10328) (T_CALLABLE 10329) (T_CASE 10330) (T_CATCH 10331) (T_CLASS 10332) (T_CLASS_C 10333) (T_CLONE 10334) (T_CONST 10335) (T_CONTINUE 10336) (T_DECLARE 10337) (T_DEFAULT 10338) (T_DIR 10339) (T_DO 10340) (T_ECHO 10341) (T_ELSE 10342) (T_ELSEIF 10343) (T_EMPTY 10344) (T_ENDDECLARE 10345) (T_ENDFOR 10346) (T_ENDFOREACH 10347) (T_ENDIF 10348) (T_ENDSWITCH 10349) (T_ENDWHILE 10350) (T_ENUM 10351) (T_EVAL 10352) (T_EXIT 10353) (T_EXTENDS 10354) (T_FILE 10355) (T_FINAL 10356) (T_FINALLY 10357) (T_FN 10358) (T_FOR 10359) (T_FOREACH 10360) (T_FUNCTION 10361) (T_FUNC_C 10362) (T_GLOBAL 10363) (T_GOTO 10364) (T_IF 10365) (T_IMPLEMENTS 10366) (T_INCLUDE 10367) (T_INCLUDE_ONCE 10368) (T_INSTANCEOF 10369) (T_INSTEADOF 10370) (T_INTERFACE 10371) (T_ISSET 10372) (T_LINE 10373) (T_LIST 10374) (T_LOGICAL_AND 10375) (T_LOGICAL_OR 10376) (T_LOGICAL_XOR 10377) (T_MATCH 10378) (T_METHOD_C 10379) (T_NAMESPACE 10380) (T_NEW 10381) (T_NS_C 10382) (T_PRINT 10383) (T_PRIVATE 10384) (T_PROTECTED 10385) (T_PUBLIC 10386) (T_READONLY 10387) (T_REQUIRE 10388) (T_REQUIRE_ONCE 10389) (T_RETURN 10390) (T_STATIC 10391) (T_STRING 10392) (T_SWITCH 10393) (T_THROW 10394) (T_TRAIT 10395) (T_TRAIT_C 10396) (T_TRY 10397) (T_UNSET 10398) (T_USE 10399) (T_VAR 10400) (T_VARIABLE 10194) (T_WHILE 10401) (T_YIELD 10402) (identifier 10524) (member_name 10525) (reserved_non_modifiers 10405) (semi_reserved 10406) (simple_variable 10526) ("{" 914)) 4572 (("(" 10408) (argument_list 10527)) 4573 ((T_FN 35) (T_FUNCTION 126) (fn 10208) #'10209 (inline_function 10531)) 4574 (("$" 10148) (T_STRING 10540) (T_VARIABLE 10194) (property_name 10547) (simple_variable 10542) ("{" 10543)) 4575 (("$" 10148) (T_STRING 10540) (T_VARIABLE 10194) (property_name 10541) (simple_variable 10542) ("{" 10543)) 4576 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 10538) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4577 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10536) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4578 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10537)) 4579 (("]" 10539)) 4580 (("(" 10408) (argument_list 10546)) 4581 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10544) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4582 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10545)) 4583 (("(" 10408) (argument_list 10548)) 4584 (("`" 10550)) 4585 (("]" 10552)) 4586 (("=" 10553)) 4587 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10554) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4588 (("%" 10223) ("*" 10224) ("+" 10225) ("-" 10226) ("." 10227) ("/" 10228) ("<" 10229) (">" 10230) ("?" 10231) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10232) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10233) (T_BOOLEAN_AND 10234) (T_BOOLEAN_OR 10235) (T_COALESCE 10236) (T_DOUBLE_ARROW 10557) (T_INSTANCEOF 10237) (T_IS_EQUAL 10238) (T_IS_GREATER_OR_EQUAL 10239) (T_IS_IDENTICAL 10240) (T_IS_NOT_EQUAL 10241) (T_IS_NOT_IDENTICAL 10242) (T_IS_SMALLER_OR_EQUAL 10243) (T_LOGICAL_AND 10244) (T_LOGICAL_OR 10245) (T_LOGICAL_XOR 10246) (T_POW 10247) (T_SL 10248) (T_SPACESHIP 10249) (T_SR 10250) ("^" 10251) ("|" 10252)) 4589 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10558) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4590 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 10567) (T_VARIABLE 2822) (encaps_var 2836)) 4591 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 10566) (T_VARIABLE 2822) (encaps_var 2827)) 4592 (("(" 2899) (argument_list 2900) (ctor_arguments 10586)) 4593 ((T_ATTRIBUTE 2853) (T_CLASS 10572) (anonymous_class 10585) (attribute 2898)) 4594 (("(" 10576) (argument_list 10577) (ctor_arguments 10578)) 4595 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10579) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 10580) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 10581) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4596 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10584) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4597 (("," 1449) (possible_comma 10582)) 4598 ((")" 10583)) 4599 ((T_EXTENDS 2902) (extends_from 10587)) 4600 ((T_IMPLEMENTS 2904) (implements_list 10588)) 4601 ((backup_doc_comment 10589)) 4602 (("{" 10590)) 4603 ((class_statement_list 10591)) 4604 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 10592)) 4605 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10594) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4606 (("%" 280) (")" 10595) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4607 (("{" 10596)) 4608 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 10597) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 4609 (("}" 10598)) 4610 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 10600) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 4611 ((")" 10601)) 4612 (("=" 10602)) 4613 (("!" 10146) ("\"" 10147) ("$" 10148) ("(" 10149) ("+" 10150) ("-" 10151) ("@" 10152) (T_ARRAY 10153) (T_ARRAY_CAST 10154) (T_ATTRIBUTE 125) (T_BOOL_CAST 10155) (T_CLASS_C 10156) (T_CLONE 10157) (T_CONSTANT_ENCAPSED_STRING 10158) (T_DEC 10159) (T_DIR 10160) (T_DNUMBER 10161) (T_DOUBLE_CAST 10162) (T_EMPTY 10163) (T_EVAL 10164) (T_EXIT 10165) (T_FILE 10166) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10167) (T_INC 10168) (T_INCLUDE 10169) (T_INCLUDE_ONCE 10170) (T_INT_CAST 10171) (T_ISSET 10172) (T_LINE 10173) (T_LIST 10174) (T_LNUMBER 10175) (T_MATCH 10176) (T_METHOD_C 10177) (T_NAME_FULLY_QUALIFIED 10178) (T_NAME_QUALIFIED 10179) (T_NAME_RELATIVE 10180) (T_NEW 10181) (T_NS_C 10182) (T_OBJECT_CAST 10183) (T_PRINT 10184) (T_REQUIRE 10185) (T_REQUIRE_ONCE 10186) (T_START_HEREDOC 10187) (T_STATIC 10188) (T_STRING 10189) (T_STRING_CAST 10190) (T_THROW 10191) (T_TRAIT_C 10192) (T_UNSET_CAST 10193) (T_VARIABLE 10194) (T_YIELD 10195) (T_YIELD_FROM 10196) ("[" 10197) ("`" 10198) (array_object_dereferenceable 10199) (attribute 129) (attributes 10200) (callable_expr 10201) (callable_variable 10202) (class_constant 10203) (class_name 10204) (constant 10205) (dereferenceable_scalar 10206) (expr 10603) (fn 10208) (fully_dereferenceable 103) #'10209 (function_call 10210) (inline_function 10211) (internal_functions_in_yacc 10212) (match 10213) (name 10214) (new_expr 10215) (scalar 10216) (simple_variable 10217) (static_member 10218) (variable 10219) (variable_class_name 10220) ("~" 10221)) 4614 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 10605) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4615 (("," 5117) (possible_comma 10606)) 4616 ((")" 10607)) 4617 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 10614) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4618 ((")" 10615)) 4619 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10617) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4620 (("%" 280) (")" 10618) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4621 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10620) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4622 (("%" 280) (")" 10621) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4623 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 10628) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 4624 ((")" 10629)) 4625 (("%" 280) (")" 10634) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4626 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10637) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4627 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10638)) 4628 (("\"" 10640) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 4629 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 10650) (T_VARIABLE 2822) (encaps_var 2836)) 4630 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 10649) (T_VARIABLE 2822) (encaps_var 2827)) 4631 (("(" 2899) (argument_list 2900) (ctor_arguments 10669)) 4632 ((T_ATTRIBUTE 2853) (T_CLASS 10655) (anonymous_class 10668) (attribute 2898)) 4633 (("(" 10659) (argument_list 10660) (ctor_arguments 10661)) 4634 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10662) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 10663) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 10664) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4635 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10667) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4636 (("," 1449) (possible_comma 10665)) 4637 ((")" 10666)) 4638 ((T_EXTENDS 2902) (extends_from 10670)) 4639 ((T_IMPLEMENTS 2904) (implements_list 10671)) 4640 ((backup_doc_comment 10672)) 4641 (("{" 10673)) 4642 ((class_statement_list 10674)) 4643 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 10675)) 4644 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10677) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4645 (("%" 280) (")" 10678) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4646 (("{" 10679)) 4647 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 10680) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 4648 (("}" 10681)) 4649 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 10683) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 4650 ((")" 10684)) 4651 (("=" 10685)) 4652 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 10686) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 4653 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 10688) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4654 (("," 5117) (possible_comma 10689)) 4655 ((")" 10690)) 4656 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 10697) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4657 ((")" 10698)) 4658 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10700) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4659 (("%" 280) (")" 10701) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4660 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 10703) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4661 (("%" 280) (")" 10704) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4662 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 10711) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 4663 ((")" 10712)) 4664 (("%" 280) (")" 10718) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4665 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10721) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4666 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10722)) 4667 (("\"" 10724) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 4668 (("$" 487) (T_ABSTRACT 10735) (T_ARRAY 10736) (T_AS 10737) (T_BREAK 10738) (T_CALLABLE 10739) (T_CASE 10740) (T_CATCH 10741) (T_CLASS 10742) (T_CLASS_C 10743) (T_CLONE 10744) (T_CONST 10745) (T_CONTINUE 10746) (T_DECLARE 10747) (T_DEFAULT 10748) (T_DIR 10749) (T_DO 10750) (T_ECHO 10751) (T_ELSE 10752) (T_ELSEIF 10753) (T_EMPTY 10754) (T_ENDDECLARE 10755) (T_ENDFOR 10756) (T_ENDFOREACH 10757) (T_ENDIF 10758) (T_ENDSWITCH 10759) (T_ENDWHILE 10760) (T_ENUM 10761) (T_EVAL 10762) (T_EXIT 10763) (T_EXTENDS 10764) (T_FILE 10765) (T_FINAL 10766) (T_FINALLY 10767) (T_FN 10768) (T_FOR 10769) (T_FOREACH 10770) (T_FUNCTION 10771) (T_FUNC_C 10772) (T_GLOBAL 10773) (T_GOTO 10774) (T_IF 10775) (T_IMPLEMENTS 10776) (T_INCLUDE 10777) (T_INCLUDE_ONCE 10778) (T_INSTANCEOF 10779) (T_INSTEADOF 10780) (T_INTERFACE 10781) (T_ISSET 10782) (T_LINE 10783) (T_LIST 10784) (T_LOGICAL_AND 10785) (T_LOGICAL_OR 10786) (T_LOGICAL_XOR 10787) (T_MATCH 10788) (T_METHOD_C 10789) (T_NAMESPACE 10790) (T_NEW 10791) (T_NS_C 10792) (T_PRINT 10793) (T_PRIVATE 10794) (T_PROTECTED 10795) (T_PUBLIC 10796) (T_READONLY 10797) (T_REQUIRE 10798) (T_REQUIRE_ONCE 10799) (T_RETURN 10800) (T_STATIC 10801) (T_STRING 10802) (T_SWITCH 10803) (T_THROW 10804) (T_TRAIT 10805) (T_TRAIT_C 10806) (T_TRY 10807) (T_UNSET 10808) (T_USE 10809) (T_VAR 10810) (T_VARIABLE 533) (T_WHILE 10811) (T_YIELD 10812) (identifier 10813) (member_name 10814) (reserved_non_modifiers 10815) (semi_reserved 10816) (simple_variable 10817) ("{" 914)) 4669 (("(" 10818) (argument_list 10819)) 4670 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10820) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 10821) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 10822) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4671 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10825) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4672 (("," 1449) (possible_comma 10823)) 4673 ((")" 10824)) 4674 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (ampersand 10855) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10856) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4675 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10854) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4676 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10853) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4677 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10852) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4678 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10851) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4679 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10850) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4680 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10849) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4681 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10848) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4682 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10847) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4683 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10846) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4684 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10845) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4685 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10844) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4686 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10843) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4687 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 10842) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4688 (("\"" 1489) ("$" 10857) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 10858) ("[" 1508) (array_object_dereferenceable 10859) (callable_expr 10860) (callable_variable 10861) (class_constant 1512) (class_name 10862) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 10863) (name 10864) (simple_variable 10865) (static_member 10866) (variable 10867) (variable_class_name 10868)) 4689 (("$" 10857) (T_VARIABLE 10858) (simple_variable 10903) ("{" 10904)) 4690 ((T_NULLSAFE_OBJECT_OPERATOR 10886) (T_OBJECT_OPERATOR 10887) ("[" 10888) ("{" 10889)) 4691 (("(" 10872) (argument_list 10885)) 4692 ((T_PAAMAYIM_NEKUDOTAYIM 10881)) 4693 (("(" 10872) (argument_list 10880)) 4694 ((T_PAAMAYIM_NEKUDOTAYIM 10869)) 4695 (("$" 10857) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 10858) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 10870) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 10871) ("{" 914)) 4696 (("(" 10872) (argument_list 10873)) 4697 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10874) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 10875) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 10876) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4698 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 10879) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4699 (("," 1449) (possible_comma 10877)) 4700 ((")" 10878)) 4701 (("$" 10857) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 10858) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 10882) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 10883) ("{" 914)) 4702 (("(" 10872) (argument_list 10884)) 4703 (("$" 10857) (T_STRING 10894) (T_VARIABLE 10858) (property_name 10901) (simple_variable 10896) ("{" 10897)) 4704 (("$" 10857) (T_STRING 10894) (T_VARIABLE 10858) (property_name 10895) (simple_variable 10896) ("{" 10897)) 4705 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 10892) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4706 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10890) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4707 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10891)) 4708 (("]" 10893)) 4709 (("(" 10872) (argument_list 10900)) 4710 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10898) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4711 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10899)) 4712 (("(" 10872) (argument_list 10902)) 4713 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 10905) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4714 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 10906)) 4715 ((backup_doc_comment 10910)) 4716 (("(" 10911)) 4717 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 10912)) 4718 ((")" 10913)) 4719 ((T_USE 148) (lexical_vars 10914)) 4720 ((":" 150) (return_type 10915)) 4721 ((backup_fn_flags 10916)) 4722 (("{" 10917)) 4723 ((inner_statement_list 10918)) 4724 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 10919) ("~" 124)) 4725 ((backup_fn_flags 10920)) 4726 ((backup_doc_comment 10922)) 4727 (("(" 10923)) 4728 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 10924)) 4729 ((")" 10925)) 4730 ((":" 2207) (return_type 10926)) 4731 ((T_DOUBLE_ARROW 10927)) 4732 ((backup_fn_flags 10928)) 4733 ((backup_lex_pos 10929)) 4734 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 10991) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4735 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11969) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4736 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 11967) (encaps_var 5158)) 4737 (("$" 10932) (T_VARIABLE 10978) (simple_variable 11963) ("{" 11964)) 4738 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 11961) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4739 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11960) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4740 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11959) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4741 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11958) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4742 (("(" 11955)) 4743 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11954) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4744 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11953) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4745 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11952) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4746 (("\"" 1489) ("$" 11231) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 11232) ("[" 1508) (array_object_dereferenceable 11233) (callable_expr 11234) (callable_variable 11235) (class_constant 1512) (class_name 11236) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 11237) (name 11238) (simple_variable 11239) (static_member 11240) (variable 11951) (variable_class_name 11242)) 4747 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11950) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4748 (("(" 11947)) 4749 (("(" 11944)) 4750 (("(" 11940) (exit_expr 11941)) 4751 (("\"" 1489) ("$" 11231) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 11232) ("[" 1508) (array_object_dereferenceable 11233) (callable_expr 11234) (callable_variable 11235) (class_constant 1512) (class_name 11236) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 11237) (name 11238) (simple_variable 11239) (static_member 11240) (variable 11939) (variable_class_name 11242)) 4752 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11938) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4753 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11937) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4754 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11936) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4755 (("(" 11932)) 4756 (("(" 11927)) 4757 (("(" 11921)) 4758 (("$" 11231) ("(" 11874) (T_ATTRIBUTE 2853) (T_CLASS 11875) (T_NAME_FULLY_QUALIFIED 11876) (T_NAME_QUALIFIED 11877) (T_NAME_RELATIVE 11878) (T_STATIC 11879) (T_STRING 11880) (T_VARIABLE 11232) (anonymous_class 11881) (attribute 2862) (attributes 11882) (class_name 11883) (class_name_reference 11884) (name 11885) (new_variable 11886) (simple_variable 11887)) 4759 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11873) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4760 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11872) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4761 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11871) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4762 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11870) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4763 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 11865) (T_END_HEREDOC 11866) (T_VARIABLE 2822) (encaps_list 11867) (encaps_var 2824)) 4764 ((T_FN 35) (T_FUNCTION 126) (fn 10992) #'10993 (inline_function 11864)) 4765 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11863) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4766 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11862) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4767 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11861) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4768 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11400) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4769 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11338) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4770 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 11334) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 4771 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 11332) (encaps_list 2729) (encaps_var 2730)) 4772 ((T_NULLSAFE_OBJECT_OPERATOR 11315) (T_OBJECT_OPERATOR 11316) ("[" 11317) ("{" 11318)) 4773 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 11312) (attribute 2654) (fn 10992) #'10993 (inline_function 11313)) 4774 (("(" 11192) (argument_list 11311)) 4775 ((T_PAAMAYIM_NEKUDOTAYIM 11306)) 4776 (("%" 11007) ("*" 11008) ("+" 11009) ("-" 11010) ("." 11011) ("/" 11012) ("<" 11013) (">" 11014) ("?" 11015) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11016) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11017) (T_BOOLEAN_AND 11018) (T_BOOLEAN_OR 11019) (T_COALESCE 11020) (T_INSTANCEOF 11021) (T_IS_EQUAL 11022) (T_IS_GREATER_OR_EQUAL 11023) (T_IS_IDENTICAL 11024) (T_IS_NOT_EQUAL 11025) (T_IS_NOT_IDENTICAL 11026) (T_IS_SMALLER_OR_EQUAL 11027) (T_LOGICAL_AND 11028) (T_LOGICAL_OR 11029) (T_LOGICAL_XOR 11030) (T_POW 11031) (T_SL 11032) (T_SPACESHIP 11033) (T_SR 11034) ("^" 11035) (backup_fn_flags 11305) ("|" 11036)) 4777 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 11294)) 4778 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 11282)) 4779 (("(" 11192) (argument_list 11281)) 4780 (("=" 11200) (T_AND_EQUAL 11201) (T_COALESCE_EQUAL 11202) (T_CONCAT_EQUAL 11203) (T_DEC 11204) (T_DIV_EQUAL 11205) (T_INC 11206) (T_MINUS_EQUAL 11207) (T_MOD_EQUAL 11208) (T_MUL_EQUAL 11209) (T_OR_EQUAL 11210) (T_PLUS_EQUAL 11211) (T_POW_EQUAL 11212) (T_SL_EQUAL 11213) (T_SR_EQUAL 11214) (T_XOR_EQUAL 11215)) 4781 ((T_PAAMAYIM_NEKUDOTAYIM 11108)) 4782 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11006) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4783 (("%" 11007) ("*" 11008) ("+" 11009) ("-" 11010) ("." 11011) ("/" 11012) ("<" 11013) (">" 11014) ("?" 11015) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11016) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11017) (T_BOOLEAN_AND 11018) (T_BOOLEAN_OR 11019) (T_COALESCE 11020) (T_INSTANCEOF 11021) (T_IS_EQUAL 11022) (T_IS_GREATER_OR_EQUAL 11023) (T_IS_IDENTICAL 11024) (T_IS_NOT_EQUAL 11025) (T_IS_NOT_IDENTICAL 11026) (T_IS_SMALLER_OR_EQUAL 11027) (T_LOGICAL_AND 11028) (T_LOGICAL_OR 11029) (T_LOGICAL_XOR 11030) (T_POW 11031) (T_SL 11032) (T_SPACESHIP 11033) (T_SR 11034) ("^" 11035) ("|" 11036)) 4784 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11107) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4785 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11106) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4786 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11105) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4787 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11104) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4788 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11103) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4789 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11102) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4790 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11101) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4791 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11100) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4792 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 11095) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 11096) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 4793 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11094) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4794 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11093) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4795 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11092) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4796 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11091) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4797 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11090) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4798 (("$" 11052) ("(" 11053) (T_NAME_FULLY_QUALIFIED 11054) (T_NAME_QUALIFIED 11055) (T_NAME_RELATIVE 11056) (T_STATIC 11057) (T_STRING 11058) (T_VARIABLE 11059) (class_name 11060) (class_name_reference 11061) (name 11062) (new_variable 11063) (simple_variable 11064)) 4799 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11051) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4800 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11050) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4801 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11049) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4802 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11048) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4803 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11047) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4804 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11046) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4805 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11045) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4806 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11044) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4807 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11043) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4808 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11042) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4809 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11041) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4810 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11040) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4811 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11039) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4812 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11038) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4813 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11037) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4814 (("$" 11052) (T_VARIABLE 11059) (simple_variable 11086) ("{" 11087)) 4815 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 11084) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4816 ((T_PAAMAYIM_NEKUDOTAYIM 11082)) 4817 ((T_NULLSAFE_OBJECT_OPERATOR 11065) (T_OBJECT_OPERATOR 11066) (T_PAAMAYIM_NEKUDOTAYIM 11067) ("[" 11068) ("{" 11069)) 4818 (("$" 11052) (T_STRING 11075) (T_VARIABLE 11059) (property_name 11081) (simple_variable 11077) ("{" 11078)) 4819 (("$" 11052) (T_STRING 11075) (T_VARIABLE 11059) (property_name 11076) (simple_variable 11077) ("{" 11078)) 4820 (("$" 11052) (T_VARIABLE 11059) (simple_variable 11074)) 4821 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 11072) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4822 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11070) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4823 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11071)) 4824 (("]" 11073)) 4825 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11079) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4826 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11080)) 4827 (("$" 11052) (T_VARIABLE 11059) (simple_variable 11083)) 4828 (("%" 280) (")" 11085) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 4829 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11088) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4830 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11089)) 4831 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11099) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4832 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 11097) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 4833 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11098) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4834 (("$" 10932) (T_ABSTRACT 11109) (T_ARRAY 11110) (T_AS 11111) (T_BREAK 11112) (T_CALLABLE 11113) (T_CASE 11114) (T_CATCH 11115) (T_CLASS 11116) (T_CLASS_C 11117) (T_CLONE 11118) (T_CONST 11119) (T_CONTINUE 11120) (T_DECLARE 11121) (T_DEFAULT 11122) (T_DIR 11123) (T_DO 11124) (T_ECHO 11125) (T_ELSE 11126) (T_ELSEIF 11127) (T_EMPTY 11128) (T_ENDDECLARE 11129) (T_ENDFOR 11130) (T_ENDFOREACH 11131) (T_ENDIF 11132) (T_ENDSWITCH 11133) (T_ENDWHILE 11134) (T_ENUM 11135) (T_EVAL 11136) (T_EXIT 11137) (T_EXTENDS 11138) (T_FILE 11139) (T_FINAL 11140) (T_FINALLY 11141) (T_FN 11142) (T_FOR 11143) (T_FOREACH 11144) (T_FUNCTION 11145) (T_FUNC_C 11146) (T_GLOBAL 11147) (T_GOTO 11148) (T_IF 11149) (T_IMPLEMENTS 11150) (T_INCLUDE 11151) (T_INCLUDE_ONCE 11152) (T_INSTANCEOF 11153) (T_INSTEADOF 11154) (T_INTERFACE 11155) (T_ISSET 11156) (T_LINE 11157) (T_LIST 11158) (T_LOGICAL_AND 11159) (T_LOGICAL_OR 11160) (T_LOGICAL_XOR 11161) (T_MATCH 11162) (T_METHOD_C 11163) (T_NAMESPACE 11164) (T_NEW 11165) (T_NS_C 11166) (T_PRINT 11167) (T_PRIVATE 11168) (T_PROTECTED 11169) (T_PUBLIC 11170) (T_READONLY 11171) (T_REQUIRE 11172) (T_REQUIRE_ONCE 11173) (T_RETURN 11174) (T_STATIC 11175) (T_STRING 11176) (T_SWITCH 11177) (T_THROW 11178) (T_TRAIT 11179) (T_TRAIT_C 11180) (T_TRY 11181) (T_UNSET 11182) (T_USE 11183) (T_VAR 11184) (T_VARIABLE 10978) (T_WHILE 11185) (T_YIELD 11186) (identifier 11187) (member_name 11188) (reserved_non_modifiers 11189) (semi_reserved 11190) (simple_variable 11191) ("{" 914)) 4835 (("(" 11192) (argument_list 11193)) 4836 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 11194) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 11195) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 11196) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4837 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 11199) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4838 (("," 1449) (possible_comma 11197)) 4839 ((")" 11198)) 4840 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (ampersand 11229) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11230) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4841 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11228) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4842 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11227) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4843 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11226) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4844 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11225) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4845 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11224) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4846 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11223) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4847 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11222) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4848 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11221) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4849 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11220) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4850 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11219) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4851 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11218) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4852 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11217) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4853 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11216) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4854 (("\"" 1489) ("$" 11231) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 11232) ("[" 1508) (array_object_dereferenceable 11233) (callable_expr 11234) (callable_variable 11235) (class_constant 1512) (class_name 11236) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 11237) (name 11238) (simple_variable 11239) (static_member 11240) (variable 11241) (variable_class_name 11242)) 4855 (("$" 11231) (T_VARIABLE 11232) (simple_variable 11277) ("{" 11278)) 4856 ((T_NULLSAFE_OBJECT_OPERATOR 11260) (T_OBJECT_OPERATOR 11261) ("[" 11262) ("{" 11263)) 4857 (("(" 11246) (argument_list 11259)) 4858 ((T_PAAMAYIM_NEKUDOTAYIM 11255)) 4859 (("(" 11246) (argument_list 11254)) 4860 ((T_PAAMAYIM_NEKUDOTAYIM 11243)) 4861 (("$" 11231) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 11232) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 11244) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 11245) ("{" 914)) 4862 (("(" 11246) (argument_list 11247)) 4863 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 11248) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 11249) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 11250) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4864 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 11253) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 4865 (("," 1449) (possible_comma 11251)) 4866 ((")" 11252)) 4867 (("$" 11231) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 11232) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 11256) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 11257) ("{" 914)) 4868 (("(" 11246) (argument_list 11258)) 4869 (("$" 11231) (T_STRING 11268) (T_VARIABLE 11232) (property_name 11275) (simple_variable 11270) ("{" 11271)) 4870 (("$" 11231) (T_STRING 11268) (T_VARIABLE 11232) (property_name 11269) (simple_variable 11270) ("{" 11271)) 4871 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 11266) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4872 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11264) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4873 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11265)) 4874 (("]" 11267)) 4875 (("(" 11246) (argument_list 11274)) 4876 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11272) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4877 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11273)) 4878 (("(" 11246) (argument_list 11276)) 4879 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11279) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4880 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11280)) 4881 ((backup_doc_comment 11283)) 4882 (("(" 11284)) 4883 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 11285)) 4884 ((")" 11286)) 4885 ((T_USE 148) (lexical_vars 11287)) 4886 ((":" 150) (return_type 11288)) 4887 ((backup_fn_flags 11289)) 4888 (("{" 11290)) 4889 ((inner_statement_list 11291)) 4890 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 11292) ("~" 124)) 4891 ((backup_fn_flags 11293)) 4892 ((backup_doc_comment 11295)) 4893 (("(" 11296)) 4894 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 11297)) 4895 ((")" 11298)) 4896 ((":" 2207) (return_type 11299)) 4897 ((T_DOUBLE_ARROW 11300)) 4898 ((backup_fn_flags 11301)) 4899 ((backup_lex_pos 11302)) 4900 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11303) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4901 (("%" 11007) ("*" 11008) ("+" 11009) ("-" 11010) ("." 11011) ("/" 11012) ("<" 11013) (">" 11014) ("?" 11015) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11016) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11017) (T_BOOLEAN_AND 11018) (T_BOOLEAN_OR 11019) (T_COALESCE 11020) (T_INSTANCEOF 11021) (T_IS_EQUAL 11022) (T_IS_GREATER_OR_EQUAL 11023) (T_IS_IDENTICAL 11024) (T_IS_NOT_EQUAL 11025) (T_IS_NOT_IDENTICAL 11026) (T_IS_SMALLER_OR_EQUAL 11027) (T_LOGICAL_AND 11028) (T_LOGICAL_OR 11029) (T_LOGICAL_XOR 11030) (T_POW 11031) (T_SL 11032) (T_SPACESHIP 11033) (T_SR 11034) ("^" 11035) (backup_fn_flags 11304) ("|" 11036)) 4902 (("$" 10932) (T_ABSTRACT 11109) (T_ARRAY 11110) (T_AS 11111) (T_BREAK 11112) (T_CALLABLE 11113) (T_CASE 11114) (T_CATCH 11115) (T_CLASS 11116) (T_CLASS_C 11117) (T_CLONE 11118) (T_CONST 11119) (T_CONTINUE 11120) (T_DECLARE 11121) (T_DEFAULT 11122) (T_DIR 11123) (T_DO 11124) (T_ECHO 11125) (T_ELSE 11126) (T_ELSEIF 11127) (T_EMPTY 11128) (T_ENDDECLARE 11129) (T_ENDFOR 11130) (T_ENDFOREACH 11131) (T_ENDIF 11132) (T_ENDSWITCH 11133) (T_ENDWHILE 11134) (T_ENUM 11135) (T_EVAL 11136) (T_EXIT 11137) (T_EXTENDS 11138) (T_FILE 11139) (T_FINAL 11140) (T_FINALLY 11141) (T_FN 11142) (T_FOR 11143) (T_FOREACH 11144) (T_FUNCTION 11145) (T_FUNC_C 11146) (T_GLOBAL 11147) (T_GOTO 11148) (T_IF 11149) (T_IMPLEMENTS 11150) (T_INCLUDE 11151) (T_INCLUDE_ONCE 11152) (T_INSTANCEOF 11153) (T_INSTEADOF 11154) (T_INTERFACE 11155) (T_ISSET 11156) (T_LINE 11157) (T_LIST 11158) (T_LOGICAL_AND 11159) (T_LOGICAL_OR 11160) (T_LOGICAL_XOR 11161) (T_MATCH 11162) (T_METHOD_C 11163) (T_NAMESPACE 11164) (T_NEW 11165) (T_NS_C 11166) (T_PRINT 11167) (T_PRIVATE 11168) (T_PROTECTED 11169) (T_PUBLIC 11170) (T_READONLY 11171) (T_REQUIRE 11172) (T_REQUIRE_ONCE 11173) (T_RETURN 11174) (T_STATIC 11175) (T_STRING 11176) (T_SWITCH 11177) (T_THROW 11178) (T_TRAIT 11179) (T_TRAIT_C 11180) (T_TRY 11181) (T_UNSET 11182) (T_USE 11183) (T_VAR 11184) (T_VARIABLE 10978) (T_WHILE 11185) (T_YIELD 11186) (identifier 11307) (member_name 11308) (reserved_non_modifiers 11189) (semi_reserved 11190) (simple_variable 11309) ("{" 914)) 4903 (("(" 11192) (argument_list 11310)) 4904 ((T_FN 35) (T_FUNCTION 126) (fn 10992) #'10993 (inline_function 11314)) 4905 (("$" 10932) (T_STRING 11323) (T_VARIABLE 10978) (property_name 11330) (simple_variable 11325) ("{" 11326)) 4906 (("$" 10932) (T_STRING 11323) (T_VARIABLE 10978) (property_name 11324) (simple_variable 11325) ("{" 11326)) 4907 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 11321) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 4908 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11319) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4909 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11320)) 4910 (("]" 11322)) 4911 (("(" 11192) (argument_list 11329)) 4912 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11327) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 4913 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11328)) 4914 (("(" 11192) (argument_list 11331)) 4915 (("`" 11333)) 4916 (("]" 11335)) 4917 (("=" 11336)) 4918 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11337) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 4919 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11860) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4920 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 11858) (encaps_var 5158)) 4921 (("$" 11341) (T_VARIABLE 11387) (simple_variable 11854) ("{" 11855)) 4922 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 11852) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 4923 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11851) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4924 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11850) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4925 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11849) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4926 (("(" 11846)) 4927 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11845) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4928 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11844) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4929 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11843) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4930 (("\"" 1489) ("$" 11640) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 11641) ("[" 1508) (array_object_dereferenceable 11642) (callable_expr 11643) (callable_variable 11644) (class_constant 1512) (class_name 11645) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 11646) (name 11647) (simple_variable 11648) (static_member 11649) (variable 11842) (variable_class_name 11651)) 4931 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11841) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4932 (("(" 11838)) 4933 (("(" 11835)) 4934 (("(" 11831) (exit_expr 11832)) 4935 (("\"" 1489) ("$" 11640) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 11641) ("[" 1508) (array_object_dereferenceable 11642) (callable_expr 11643) (callable_variable 11644) (class_constant 1512) (class_name 11645) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 11646) (name 11647) (simple_variable 11648) (static_member 11649) (variable 11830) (variable_class_name 11651)) 4936 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11829) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4937 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11828) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4938 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11827) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4939 (("(" 11823)) 4940 (("(" 11818)) 4941 (("(" 11812)) 4942 (("$" 11640) ("(" 11765) (T_ATTRIBUTE 2853) (T_CLASS 11766) (T_NAME_FULLY_QUALIFIED 11767) (T_NAME_QUALIFIED 11768) (T_NAME_RELATIVE 11769) (T_STATIC 11770) (T_STRING 11771) (T_VARIABLE 11641) (anonymous_class 11772) (attribute 2862) (attributes 11773) (class_name 11774) (class_name_reference 11775) (name 11776) (new_variable 11777) (simple_variable 11778)) 4943 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11764) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4944 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11763) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4945 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11762) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4946 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11761) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4947 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 11756) (T_END_HEREDOC 11757) (T_VARIABLE 2822) (encaps_list 11758) (encaps_var 2824)) 4948 ((T_FN 35) (T_FUNCTION 126) (fn 11401) #'11402 (inline_function 11755)) 4949 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11754) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4950 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11753) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4951 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11752) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4952 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11749) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4953 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11748) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4954 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 11744) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 4955 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 11742) (encaps_list 2729) (encaps_var 2730)) 4956 ((T_NULLSAFE_OBJECT_OPERATOR 11725) (T_OBJECT_OPERATOR 11726) ("[" 11727) ("{" 11728)) 4957 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 11722) (attribute 2654) (fn 11401) #'11402 (inline_function 11723)) 4958 (("(" 11601) (argument_list 11721)) 4959 ((T_PAAMAYIM_NEKUDOTAYIM 11716)) 4960 (("%" 11416) ("*" 11417) ("+" 11418) ("-" 11419) ("." 11420) ("/" 11421) ("<" 11422) (">" 11423) ("?" 11424) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11425) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11426) (T_BOOLEAN_AND 11427) (T_BOOLEAN_OR 11428) (T_COALESCE 11429) (T_DOUBLE_ARROW 11714) (T_INSTANCEOF 11430) (T_IS_EQUAL 11431) (T_IS_GREATER_OR_EQUAL 11432) (T_IS_IDENTICAL 11433) (T_IS_NOT_EQUAL 11434) (T_IS_NOT_IDENTICAL 11435) (T_IS_SMALLER_OR_EQUAL 11436) (T_LOGICAL_AND 11437) (T_LOGICAL_OR 11438) (T_LOGICAL_XOR 11439) (T_POW 11440) (T_SL 11441) (T_SPACESHIP 11442) (T_SR 11443) ("^" 11444) ("|" 11445)) 4961 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 11703)) 4962 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 11691)) 4963 (("(" 11601) (argument_list 11690)) 4964 (("=" 11609) (T_AND_EQUAL 11610) (T_COALESCE_EQUAL 11611) (T_CONCAT_EQUAL 11612) (T_DEC 11613) (T_DIV_EQUAL 11614) (T_INC 11615) (T_MINUS_EQUAL 11616) (T_MOD_EQUAL 11617) (T_MUL_EQUAL 11618) (T_OR_EQUAL 11619) (T_PLUS_EQUAL 11620) (T_POW_EQUAL 11621) (T_SL_EQUAL 11622) (T_SR_EQUAL 11623) (T_XOR_EQUAL 11624)) 4965 ((T_PAAMAYIM_NEKUDOTAYIM 11517)) 4966 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11415) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4967 (("%" 11416) ("*" 11417) ("+" 11418) ("-" 11419) ("." 11420) ("/" 11421) ("<" 11422) (">" 11423) ("?" 11424) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11425) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11426) (T_BOOLEAN_AND 11427) (T_BOOLEAN_OR 11428) (T_COALESCE 11429) (T_INSTANCEOF 11430) (T_IS_EQUAL 11431) (T_IS_GREATER_OR_EQUAL 11432) (T_IS_IDENTICAL 11433) (T_IS_NOT_EQUAL 11434) (T_IS_NOT_IDENTICAL 11435) (T_IS_SMALLER_OR_EQUAL 11436) (T_LOGICAL_AND 11437) (T_LOGICAL_OR 11438) (T_LOGICAL_XOR 11439) (T_POW 11440) (T_SL 11441) (T_SPACESHIP 11442) (T_SR 11443) ("^" 11444) ("|" 11445)) 4968 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11516) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4969 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11515) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4970 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11514) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4971 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11513) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4972 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11512) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4973 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11511) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4974 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11510) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4975 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11509) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4976 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 11504) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 11505) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 4977 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11503) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4978 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11502) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4979 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11501) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4980 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11500) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4981 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11499) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4982 (("$" 11461) ("(" 11462) (T_NAME_FULLY_QUALIFIED 11463) (T_NAME_QUALIFIED 11464) (T_NAME_RELATIVE 11465) (T_STATIC 11466) (T_STRING 11467) (T_VARIABLE 11468) (class_name 11469) (class_name_reference 11470) (name 11471) (new_variable 11472) (simple_variable 11473)) 4983 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11460) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4984 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11459) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4985 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11458) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4986 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11457) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4987 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11456) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4988 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11455) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4989 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11454) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4990 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11453) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4991 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11452) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4992 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11451) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4993 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11450) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4994 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11449) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4995 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11448) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4996 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11447) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4997 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11446) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 4998 (("$" 11461) (T_VARIABLE 11468) (simple_variable 11495) ("{" 11496)) 4999 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 11493) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5000 ((T_PAAMAYIM_NEKUDOTAYIM 11491)) 5001 ((T_NULLSAFE_OBJECT_OPERATOR 11474) (T_OBJECT_OPERATOR 11475) (T_PAAMAYIM_NEKUDOTAYIM 11476) ("[" 11477) ("{" 11478)) 5002 (("$" 11461) (T_STRING 11484) (T_VARIABLE 11468) (property_name 11490) (simple_variable 11486) ("{" 11487)) 5003 (("$" 11461) (T_STRING 11484) (T_VARIABLE 11468) (property_name 11485) (simple_variable 11486) ("{" 11487)) 5004 (("$" 11461) (T_VARIABLE 11468) (simple_variable 11483)) 5005 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 11481) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5006 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11479) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5007 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11480)) 5008 (("]" 11482)) 5009 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11488) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5010 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11489)) 5011 (("$" 11461) (T_VARIABLE 11468) (simple_variable 11492)) 5012 (("%" 280) (")" 11494) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5013 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11497) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5014 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11498)) 5015 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11508) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5016 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 11506) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 5017 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11507) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5018 (("$" 11341) (T_ABSTRACT 11518) (T_ARRAY 11519) (T_AS 11520) (T_BREAK 11521) (T_CALLABLE 11522) (T_CASE 11523) (T_CATCH 11524) (T_CLASS 11525) (T_CLASS_C 11526) (T_CLONE 11527) (T_CONST 11528) (T_CONTINUE 11529) (T_DECLARE 11530) (T_DEFAULT 11531) (T_DIR 11532) (T_DO 11533) (T_ECHO 11534) (T_ELSE 11535) (T_ELSEIF 11536) (T_EMPTY 11537) (T_ENDDECLARE 11538) (T_ENDFOR 11539) (T_ENDFOREACH 11540) (T_ENDIF 11541) (T_ENDSWITCH 11542) (T_ENDWHILE 11543) (T_ENUM 11544) (T_EVAL 11545) (T_EXIT 11546) (T_EXTENDS 11547) (T_FILE 11548) (T_FINAL 11549) (T_FINALLY 11550) (T_FN 11551) (T_FOR 11552) (T_FOREACH 11553) (T_FUNCTION 11554) (T_FUNC_C 11555) (T_GLOBAL 11556) (T_GOTO 11557) (T_IF 11558) (T_IMPLEMENTS 11559) (T_INCLUDE 11560) (T_INCLUDE_ONCE 11561) (T_INSTANCEOF 11562) (T_INSTEADOF 11563) (T_INTERFACE 11564) (T_ISSET 11565) (T_LINE 11566) (T_LIST 11567) (T_LOGICAL_AND 11568) (T_LOGICAL_OR 11569) (T_LOGICAL_XOR 11570) (T_MATCH 11571) (T_METHOD_C 11572) (T_NAMESPACE 11573) (T_NEW 11574) (T_NS_C 11575) (T_PRINT 11576) (T_PRIVATE 11577) (T_PROTECTED 11578) (T_PUBLIC 11579) (T_READONLY 11580) (T_REQUIRE 11581) (T_REQUIRE_ONCE 11582) (T_RETURN 11583) (T_STATIC 11584) (T_STRING 11585) (T_SWITCH 11586) (T_THROW 11587) (T_TRAIT 11588) (T_TRAIT_C 11589) (T_TRY 11590) (T_UNSET 11591) (T_USE 11592) (T_VAR 11593) (T_VARIABLE 11387) (T_WHILE 11594) (T_YIELD 11595) (identifier 11596) (member_name 11597) (reserved_non_modifiers 11598) (semi_reserved 11599) (simple_variable 11600) ("{" 914)) 5019 (("(" 11601) (argument_list 11602)) 5020 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 11603) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 11604) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 11605) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5021 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 11608) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5022 (("," 1449) (possible_comma 11606)) 5023 ((")" 11607)) 5024 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (ampersand 11638) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11639) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5025 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11637) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5026 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11636) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5027 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11635) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5028 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11634) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5029 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11633) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5030 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11632) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5031 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11631) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5032 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11630) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5033 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11629) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5034 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11628) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5035 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11627) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5036 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11626) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5037 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11625) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5038 (("\"" 1489) ("$" 11640) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 11641) ("[" 1508) (array_object_dereferenceable 11642) (callable_expr 11643) (callable_variable 11644) (class_constant 1512) (class_name 11645) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 11646) (name 11647) (simple_variable 11648) (static_member 11649) (variable 11650) (variable_class_name 11651)) 5039 (("$" 11640) (T_VARIABLE 11641) (simple_variable 11686) ("{" 11687)) 5040 ((T_NULLSAFE_OBJECT_OPERATOR 11669) (T_OBJECT_OPERATOR 11670) ("[" 11671) ("{" 11672)) 5041 (("(" 11655) (argument_list 11668)) 5042 ((T_PAAMAYIM_NEKUDOTAYIM 11664)) 5043 (("(" 11655) (argument_list 11663)) 5044 ((T_PAAMAYIM_NEKUDOTAYIM 11652)) 5045 (("$" 11640) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 11641) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 11653) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 11654) ("{" 914)) 5046 (("(" 11655) (argument_list 11656)) 5047 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 11657) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 11658) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 11659) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5048 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 11662) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5049 (("," 1449) (possible_comma 11660)) 5050 ((")" 11661)) 5051 (("$" 11640) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 11641) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 11665) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 11666) ("{" 914)) 5052 (("(" 11655) (argument_list 11667)) 5053 (("$" 11640) (T_STRING 11677) (T_VARIABLE 11641) (property_name 11684) (simple_variable 11679) ("{" 11680)) 5054 (("$" 11640) (T_STRING 11677) (T_VARIABLE 11641) (property_name 11678) (simple_variable 11679) ("{" 11680)) 5055 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 11675) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5056 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11673) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5057 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11674)) 5058 (("]" 11676)) 5059 (("(" 11655) (argument_list 11683)) 5060 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11681) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5061 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11682)) 5062 (("(" 11655) (argument_list 11685)) 5063 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11688) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5064 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11689)) 5065 ((backup_doc_comment 11692)) 5066 (("(" 11693)) 5067 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 11694)) 5068 ((")" 11695)) 5069 ((T_USE 148) (lexical_vars 11696)) 5070 ((":" 150) (return_type 11697)) 5071 ((backup_fn_flags 11698)) 5072 (("{" 11699)) 5073 ((inner_statement_list 11700)) 5074 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 11701) ("~" 124)) 5075 ((backup_fn_flags 11702)) 5076 ((backup_doc_comment 11704)) 5077 (("(" 11705)) 5078 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 11706)) 5079 ((")" 11707)) 5080 ((":" 2207) (return_type 11708)) 5081 ((T_DOUBLE_ARROW 11709)) 5082 ((backup_fn_flags 11710)) 5083 ((backup_lex_pos 11711)) 5084 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11712) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5085 (("%" 11416) ("*" 11417) ("+" 11418) ("-" 11419) ("." 11420) ("/" 11421) ("<" 11422) (">" 11423) ("?" 11424) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11425) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11426) (T_BOOLEAN_AND 11427) (T_BOOLEAN_OR 11428) (T_COALESCE 11429) (T_INSTANCEOF 11430) (T_IS_EQUAL 11431) (T_IS_GREATER_OR_EQUAL 11432) (T_IS_IDENTICAL 11433) (T_IS_NOT_EQUAL 11434) (T_IS_NOT_IDENTICAL 11435) (T_IS_SMALLER_OR_EQUAL 11436) (T_LOGICAL_AND 11437) (T_LOGICAL_OR 11438) (T_LOGICAL_XOR 11439) (T_POW 11440) (T_SL 11441) (T_SPACESHIP 11442) (T_SR 11443) ("^" 11444) (backup_fn_flags 11713) ("|" 11445)) 5086 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11715) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 5087 (("$" 11341) (T_ABSTRACT 11518) (T_ARRAY 11519) (T_AS 11520) (T_BREAK 11521) (T_CALLABLE 11522) (T_CASE 11523) (T_CATCH 11524) (T_CLASS 11525) (T_CLASS_C 11526) (T_CLONE 11527) (T_CONST 11528) (T_CONTINUE 11529) (T_DECLARE 11530) (T_DEFAULT 11531) (T_DIR 11532) (T_DO 11533) (T_ECHO 11534) (T_ELSE 11535) (T_ELSEIF 11536) (T_EMPTY 11537) (T_ENDDECLARE 11538) (T_ENDFOR 11539) (T_ENDFOREACH 11540) (T_ENDIF 11541) (T_ENDSWITCH 11542) (T_ENDWHILE 11543) (T_ENUM 11544) (T_EVAL 11545) (T_EXIT 11546) (T_EXTENDS 11547) (T_FILE 11548) (T_FINAL 11549) (T_FINALLY 11550) (T_FN 11551) (T_FOR 11552) (T_FOREACH 11553) (T_FUNCTION 11554) (T_FUNC_C 11555) (T_GLOBAL 11556) (T_GOTO 11557) (T_IF 11558) (T_IMPLEMENTS 11559) (T_INCLUDE 11560) (T_INCLUDE_ONCE 11561) (T_INSTANCEOF 11562) (T_INSTEADOF 11563) (T_INTERFACE 11564) (T_ISSET 11565) (T_LINE 11566) (T_LIST 11567) (T_LOGICAL_AND 11568) (T_LOGICAL_OR 11569) (T_LOGICAL_XOR 11570) (T_MATCH 11571) (T_METHOD_C 11572) (T_NAMESPACE 11573) (T_NEW 11574) (T_NS_C 11575) (T_PRINT 11576) (T_PRIVATE 11577) (T_PROTECTED 11578) (T_PUBLIC 11579) (T_READONLY 11580) (T_REQUIRE 11581) (T_REQUIRE_ONCE 11582) (T_RETURN 11583) (T_STATIC 11584) (T_STRING 11585) (T_SWITCH 11586) (T_THROW 11587) (T_TRAIT 11588) (T_TRAIT_C 11589) (T_TRY 11590) (T_UNSET 11591) (T_USE 11592) (T_VAR 11593) (T_VARIABLE 11387) (T_WHILE 11594) (T_YIELD 11595) (identifier 11717) (member_name 11718) (reserved_non_modifiers 11598) (semi_reserved 11599) (simple_variable 11719) ("{" 914)) 5088 (("(" 11601) (argument_list 11720)) 5089 ((T_FN 35) (T_FUNCTION 126) (fn 11401) #'11402 (inline_function 11724)) 5090 (("$" 11341) (T_STRING 11733) (T_VARIABLE 11387) (property_name 11740) (simple_variable 11735) ("{" 11736)) 5091 (("$" 11341) (T_STRING 11733) (T_VARIABLE 11387) (property_name 11734) (simple_variable 11735) ("{" 11736)) 5092 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 11731) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5093 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11729) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5094 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11730)) 5095 (("]" 11732)) 5096 (("(" 11601) (argument_list 11739)) 5097 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11737) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5098 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11738)) 5099 (("(" 11601) (argument_list 11741)) 5100 (("`" 11743)) 5101 (("]" 11745)) 5102 (("=" 11746)) 5103 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11747) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5104 (("%" 11416) ("*" 11417) ("+" 11418) ("-" 11419) ("." 11420) ("/" 11421) ("<" 11422) (">" 11423) ("?" 11424) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11425) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11426) (T_BOOLEAN_AND 11427) (T_BOOLEAN_OR 11428) (T_COALESCE 11429) (T_DOUBLE_ARROW 11750) (T_INSTANCEOF 11430) (T_IS_EQUAL 11431) (T_IS_GREATER_OR_EQUAL 11432) (T_IS_IDENTICAL 11433) (T_IS_NOT_EQUAL 11434) (T_IS_NOT_IDENTICAL 11435) (T_IS_SMALLER_OR_EQUAL 11436) (T_LOGICAL_AND 11437) (T_LOGICAL_OR 11438) (T_LOGICAL_XOR 11439) (T_POW 11440) (T_SL 11441) (T_SPACESHIP 11442) (T_SR 11443) ("^" 11444) ("|" 11445)) 5105 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11751) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5106 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 11760) (T_VARIABLE 2822) (encaps_var 2836)) 5107 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 11759) (T_VARIABLE 2822) (encaps_var 2827)) 5108 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 11810) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5109 (("(" 2899) (argument_list 2900) (ctor_arguments 11803)) 5110 ((T_ATTRIBUTE 2853) (T_CLASS 11766) (anonymous_class 11802) (attribute 2898)) 5111 ((T_PAAMAYIM_NEKUDOTAYIM 11800)) 5112 (("(" 11791) (argument_list 11792) (ctor_arguments 11793)) 5113 ((T_NULLSAFE_OBJECT_OPERATOR 11779) (T_OBJECT_OPERATOR 11780) (T_PAAMAYIM_NEKUDOTAYIM 11781) ("[" 11782) ("{" 11783)) 5114 (("$" 11640) (T_STRING 11677) (T_VARIABLE 11641) (property_name 11790) (simple_variable 11679) ("{" 11680)) 5115 (("$" 11640) (T_STRING 11677) (T_VARIABLE 11641) (property_name 11789) (simple_variable 11679) ("{" 11680)) 5116 (("$" 11640) (T_VARIABLE 11641) (simple_variable 11788)) 5117 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 11786) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5118 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11784) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5119 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11785)) 5120 (("]" 11787)) 5121 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 11794) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 11795) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 11796) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5122 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 11799) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5123 (("," 1449) (possible_comma 11797)) 5124 ((")" 11798)) 5125 (("$" 11640) (T_VARIABLE 11641) (simple_variable 11801)) 5126 ((T_EXTENDS 2902) (extends_from 11804)) 5127 ((T_IMPLEMENTS 2904) (implements_list 11805)) 5128 ((backup_doc_comment 11806)) 5129 (("{" 11807)) 5130 ((class_statement_list 11808)) 5131 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 11809)) 5132 (("%" 280) (")" 11811) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5133 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 11813) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5134 (("%" 280) (")" 11814) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5135 (("{" 11815)) 5136 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 11816) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 5137 (("}" 11817)) 5138 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 11819) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 5139 ((")" 11820)) 5140 (("=" 11821)) 5141 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 11822) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5142 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 11824) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5143 (("," 5117) (possible_comma 11825)) 5144 ((")" 11826)) 5145 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 11833) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5146 ((")" 11834)) 5147 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 11836) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5148 (("%" 280) (")" 11837) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5149 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 11839) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5150 (("%" 280) (")" 11840) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5151 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 11847) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 5152 ((")" 11848)) 5153 (("%" 280) (")" 11853) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5154 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11856) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5155 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11857)) 5156 (("\"" 11859) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 5157 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 11869) (T_VARIABLE 2822) (encaps_var 2836)) 5158 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 11868) (T_VARIABLE 2822) (encaps_var 2827)) 5159 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 11919) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5160 (("(" 2899) (argument_list 2900) (ctor_arguments 11912)) 5161 ((T_ATTRIBUTE 2853) (T_CLASS 11875) (anonymous_class 11911) (attribute 2898)) 5162 ((T_PAAMAYIM_NEKUDOTAYIM 11909)) 5163 (("(" 11900) (argument_list 11901) (ctor_arguments 11902)) 5164 ((T_NULLSAFE_OBJECT_OPERATOR 11888) (T_OBJECT_OPERATOR 11889) (T_PAAMAYIM_NEKUDOTAYIM 11890) ("[" 11891) ("{" 11892)) 5165 (("$" 11231) (T_STRING 11268) (T_VARIABLE 11232) (property_name 11899) (simple_variable 11270) ("{" 11271)) 5166 (("$" 11231) (T_STRING 11268) (T_VARIABLE 11232) (property_name 11898) (simple_variable 11270) ("{" 11271)) 5167 (("$" 11231) (T_VARIABLE 11232) (simple_variable 11897)) 5168 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 11895) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5169 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11893) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5170 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11894)) 5171 (("]" 11896)) 5172 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 11903) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 11904) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 11905) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5173 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 11908) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5174 (("," 1449) (possible_comma 11906)) 5175 ((")" 11907)) 5176 (("$" 11231) (T_VARIABLE 11232) (simple_variable 11910)) 5177 ((T_EXTENDS 2902) (extends_from 11913)) 5178 ((T_IMPLEMENTS 2904) (implements_list 11914)) 5179 ((backup_doc_comment 11915)) 5180 (("{" 11916)) 5181 ((class_statement_list 11917)) 5182 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 11918)) 5183 (("%" 280) (")" 11920) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5184 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 11922) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5185 (("%" 280) (")" 11923) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5186 (("{" 11924)) 5187 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 11925) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 5188 (("}" 11926)) 5189 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 11928) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 5190 ((")" 11929)) 5191 (("=" 11930)) 5192 (("!" 10930) ("\"" 10931) ("$" 10932) ("(" 10933) ("+" 10934) ("-" 10935) ("@" 10936) (T_ARRAY 10937) (T_ARRAY_CAST 10938) (T_ATTRIBUTE 125) (T_BOOL_CAST 10939) (T_CLASS_C 10940) (T_CLONE 10941) (T_CONSTANT_ENCAPSED_STRING 10942) (T_DEC 10943) (T_DIR 10944) (T_DNUMBER 10945) (T_DOUBLE_CAST 10946) (T_EMPTY 10947) (T_EVAL 10948) (T_EXIT 10949) (T_FILE 10950) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 10951) (T_INC 10952) (T_INCLUDE 10953) (T_INCLUDE_ONCE 10954) (T_INT_CAST 10955) (T_ISSET 10956) (T_LINE 10957) (T_LIST 10958) (T_LNUMBER 10959) (T_MATCH 10960) (T_METHOD_C 10961) (T_NAME_FULLY_QUALIFIED 10962) (T_NAME_QUALIFIED 10963) (T_NAME_RELATIVE 10964) (T_NEW 10965) (T_NS_C 10966) (T_OBJECT_CAST 10967) (T_PRINT 10968) (T_REQUIRE 10969) (T_REQUIRE_ONCE 10970) (T_START_HEREDOC 10971) (T_STATIC 10972) (T_STRING 10973) (T_STRING_CAST 10974) (T_THROW 10975) (T_TRAIT_C 10976) (T_UNSET_CAST 10977) (T_VARIABLE 10978) (T_YIELD 10979) (T_YIELD_FROM 10980) ("[" 10981) ("`" 10982) (array_object_dereferenceable 10983) (attribute 129) (attributes 10984) (callable_expr 10985) (callable_variable 10986) (class_constant 10987) (class_name 10988) (constant 10989) (dereferenceable_scalar 10990) (expr 11931) (fn 10992) (fully_dereferenceable 103) #'10993 (function_call 10994) (inline_function 10995) (internal_functions_in_yacc 10996) (match 10997) (name 10998) (new_expr 10999) (scalar 11000) (simple_variable 11001) (static_member 11002) (variable 11003) (variable_class_name 11004) ("~" 11005)) 5193 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 11933) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5194 (("," 5117) (possible_comma 11934)) 5195 ((")" 11935)) 5196 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 11942) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5197 ((")" 11943)) 5198 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 11945) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5199 (("%" 280) (")" 11946) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5200 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 11948) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5201 (("%" 280) (")" 11949) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5202 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 11956) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 5203 ((")" 11957)) 5204 (("%" 280) (")" 11962) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5205 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11965) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5206 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11966)) 5207 (("\"" 11968) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 5208 (("$" 487) (T_ABSTRACT 10735) (T_ARRAY 10736) (T_AS 10737) (T_BREAK 10738) (T_CALLABLE 10739) (T_CASE 10740) (T_CATCH 10741) (T_CLASS 10742) (T_CLASS_C 10743) (T_CLONE 10744) (T_CONST 10745) (T_CONTINUE 10746) (T_DECLARE 10747) (T_DEFAULT 10748) (T_DIR 10749) (T_DO 10750) (T_ECHO 10751) (T_ELSE 10752) (T_ELSEIF 10753) (T_EMPTY 10754) (T_ENDDECLARE 10755) (T_ENDFOR 10756) (T_ENDFOREACH 10757) (T_ENDIF 10758) (T_ENDSWITCH 10759) (T_ENDWHILE 10760) (T_ENUM 10761) (T_EVAL 10762) (T_EXIT 10763) (T_EXTENDS 10764) (T_FILE 10765) (T_FINAL 10766) (T_FINALLY 10767) (T_FN 10768) (T_FOR 10769) (T_FOREACH 10770) (T_FUNCTION 10771) (T_FUNC_C 10772) (T_GLOBAL 10773) (T_GOTO 10774) (T_IF 10775) (T_IMPLEMENTS 10776) (T_INCLUDE 10777) (T_INCLUDE_ONCE 10778) (T_INSTANCEOF 10779) (T_INSTEADOF 10780) (T_INTERFACE 10781) (T_ISSET 10782) (T_LINE 10783) (T_LIST 10784) (T_LOGICAL_AND 10785) (T_LOGICAL_OR 10786) (T_LOGICAL_XOR 10787) (T_MATCH 10788) (T_METHOD_C 10789) (T_NAMESPACE 10790) (T_NEW 10791) (T_NS_C 10792) (T_PRINT 10793) (T_PRIVATE 10794) (T_PROTECTED 10795) (T_PUBLIC 10796) (T_READONLY 10797) (T_REQUIRE 10798) (T_REQUIRE_ONCE 10799) (T_RETURN 10800) (T_STATIC 10801) (T_STRING 10802) (T_SWITCH 10803) (T_THROW 10804) (T_TRAIT 10805) (T_TRAIT_C 10806) (T_TRY 10807) (T_UNSET 10808) (T_USE 10809) (T_VAR 10810) (T_VARIABLE 533) (T_WHILE 10811) (T_YIELD 10812) (identifier 11971) (member_name 11972) (reserved_non_modifiers 10815) (semi_reserved 10816) (simple_variable 11973) ("{" 914)) 5209 (("(" 10818) (argument_list 11974)) 5210 ((T_FN 35) (T_FUNCTION 126) (fn 547) #'548 (inline_function 11978)) 5211 (("$" 487) (T_STRING 11987) (T_VARIABLE 533) (property_name 11994) (simple_variable 11989) ("{" 11990)) 5212 (("$" 487) (T_STRING 11987) (T_VARIABLE 533) (property_name 11988) (simple_variable 11989) ("{" 11990)) 5213 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 11985) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5214 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11983) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5215 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11984)) 5216 (("]" 11986)) 5217 (("(" 10818) (argument_list 11993)) 5218 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 11991) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5219 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 11992)) 5220 (("(" 10818) (argument_list 11995)) 5221 (("`" 11997)) 5222 (("]" 11999)) 5223 (("=" 12000)) 5224 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12001) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5225 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12498) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5226 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 12496) (encaps_var 5158)) 5227 (("$" 12005) (T_VARIABLE 12051) (simple_variable 12492) ("{" 12493)) 5228 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 12490) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5229 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12489) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5230 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12488) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5231 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12487) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5232 (("(" 12484)) 5233 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12483) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5234 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12482) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5235 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12481) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5236 (("\"" 1489) ("$" 12304) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 12305) ("[" 1508) (array_object_dereferenceable 12306) (callable_expr 12307) (callable_variable 12308) (class_constant 1512) (class_name 12309) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 12310) (name 12311) (simple_variable 12312) (static_member 12313) (variable 12480) (variable_class_name 12315)) 5237 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12479) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5238 (("(" 12476)) 5239 (("(" 12473)) 5240 (("(" 12469) (exit_expr 12470)) 5241 (("\"" 1489) ("$" 12304) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 12305) ("[" 1508) (array_object_dereferenceable 12306) (callable_expr 12307) (callable_variable 12308) (class_constant 1512) (class_name 12309) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 12310) (name 12311) (simple_variable 12312) (static_member 12313) (variable 12468) (variable_class_name 12315)) 5242 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12467) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5243 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12466) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5244 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12465) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5245 (("(" 12461)) 5246 (("(" 12456)) 5247 (("(" 12450)) 5248 (("$" 11640) ("(" 11765) (T_ATTRIBUTE 2853) (T_CLASS 12429) (T_NAME_FULLY_QUALIFIED 11767) (T_NAME_QUALIFIED 11768) (T_NAME_RELATIVE 11769) (T_STATIC 11770) (T_STRING 11771) (T_VARIABLE 11641) (anonymous_class 12430) (attribute 2862) (attributes 12431) (class_name 11774) (class_name_reference 12432) (name 11776) (new_variable 11777) (simple_variable 11778)) 5249 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12428) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5250 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12427) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5251 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12426) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5252 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12425) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5253 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 12420) (T_END_HEREDOC 12421) (T_VARIABLE 2822) (encaps_list 12422) (encaps_var 2824)) 5254 ((T_FN 35) (T_FUNCTION 126) (fn 12065) #'12066 (inline_function 12419)) 5255 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12418) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5256 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12417) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5257 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12416) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5258 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12413) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5259 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12412) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5260 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 12408) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 5261 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 12406) (encaps_list 2729) (encaps_var 2730)) 5262 ((T_NULLSAFE_OBJECT_OPERATOR 12389) (T_OBJECT_OPERATOR 12390) ("[" 12391) ("{" 12392)) 5263 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 12386) (attribute 2654) (fn 12065) #'12066 (inline_function 12387)) 5264 (("(" 12265) (argument_list 12385)) 5265 ((T_PAAMAYIM_NEKUDOTAYIM 12380)) 5266 (("%" 12080) ("*" 12081) ("+" 12082) ("-" 12083) ("." 12084) ("/" 12085) ("<" 12086) (">" 12087) ("?" 12088) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12089) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12090) (T_BOOLEAN_AND 12091) (T_BOOLEAN_OR 12092) (T_COALESCE 12093) (T_DOUBLE_ARROW 12378) (T_INSTANCEOF 12094) (T_IS_EQUAL 12095) (T_IS_GREATER_OR_EQUAL 12096) (T_IS_IDENTICAL 12097) (T_IS_NOT_EQUAL 12098) (T_IS_NOT_IDENTICAL 12099) (T_IS_SMALLER_OR_EQUAL 12100) (T_LOGICAL_AND 12101) (T_LOGICAL_OR 12102) (T_LOGICAL_XOR 12103) (T_POW 12104) (T_SL 12105) (T_SPACESHIP 12106) (T_SR 12107) ("^" 12108) ("|" 12109)) 5267 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 12367)) 5268 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 12355)) 5269 (("(" 12265) (argument_list 12354)) 5270 (("=" 12273) (T_AND_EQUAL 12274) (T_COALESCE_EQUAL 12275) (T_CONCAT_EQUAL 12276) (T_DEC 12277) (T_DIV_EQUAL 12278) (T_INC 12279) (T_MINUS_EQUAL 12280) (T_MOD_EQUAL 12281) (T_MUL_EQUAL 12282) (T_OR_EQUAL 12283) (T_PLUS_EQUAL 12284) (T_POW_EQUAL 12285) (T_SL_EQUAL 12286) (T_SR_EQUAL 12287) (T_XOR_EQUAL 12288)) 5271 ((T_PAAMAYIM_NEKUDOTAYIM 12181)) 5272 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12079) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5273 (("%" 12080) ("*" 12081) ("+" 12082) ("-" 12083) ("." 12084) ("/" 12085) ("<" 12086) (">" 12087) ("?" 12088) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12089) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12090) (T_BOOLEAN_AND 12091) (T_BOOLEAN_OR 12092) (T_COALESCE 12093) (T_INSTANCEOF 12094) (T_IS_EQUAL 12095) (T_IS_GREATER_OR_EQUAL 12096) (T_IS_IDENTICAL 12097) (T_IS_NOT_EQUAL 12098) (T_IS_NOT_IDENTICAL 12099) (T_IS_SMALLER_OR_EQUAL 12100) (T_LOGICAL_AND 12101) (T_LOGICAL_OR 12102) (T_LOGICAL_XOR 12103) (T_POW 12104) (T_SL 12105) (T_SPACESHIP 12106) (T_SR 12107) ("^" 12108) ("|" 12109)) 5274 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12180) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5275 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12179) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5276 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12178) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5277 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12177) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5278 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12176) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5279 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12175) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5280 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12174) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5281 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12173) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5282 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 12168) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 12169) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 5283 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12167) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5284 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12166) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5285 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12165) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5286 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12164) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5287 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12163) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5288 (("$" 12125) ("(" 12126) (T_NAME_FULLY_QUALIFIED 12127) (T_NAME_QUALIFIED 12128) (T_NAME_RELATIVE 12129) (T_STATIC 12130) (T_STRING 12131) (T_VARIABLE 12132) (class_name 12133) (class_name_reference 12134) (name 12135) (new_variable 12136) (simple_variable 12137)) 5289 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12124) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5290 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12123) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5291 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12122) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5292 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12121) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5293 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12120) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5294 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12119) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5295 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12118) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5296 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12117) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5297 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12116) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5298 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12115) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5299 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12114) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5300 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12113) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5301 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12112) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5302 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12111) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5303 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12110) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5304 (("$" 12125) (T_VARIABLE 12132) (simple_variable 12159) ("{" 12160)) 5305 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 12157) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5306 ((T_PAAMAYIM_NEKUDOTAYIM 12155)) 5307 ((T_NULLSAFE_OBJECT_OPERATOR 12138) (T_OBJECT_OPERATOR 12139) (T_PAAMAYIM_NEKUDOTAYIM 12140) ("[" 12141) ("{" 12142)) 5308 (("$" 12125) (T_STRING 12148) (T_VARIABLE 12132) (property_name 12154) (simple_variable 12150) ("{" 12151)) 5309 (("$" 12125) (T_STRING 12148) (T_VARIABLE 12132) (property_name 12149) (simple_variable 12150) ("{" 12151)) 5310 (("$" 12125) (T_VARIABLE 12132) (simple_variable 12147)) 5311 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 12145) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5312 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12143) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5313 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12144)) 5314 (("]" 12146)) 5315 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12152) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5316 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12153)) 5317 (("$" 12125) (T_VARIABLE 12132) (simple_variable 12156)) 5318 (("%" 280) (")" 12158) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5319 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12161) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5320 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12162)) 5321 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12172) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5322 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 12170) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 5323 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12171) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5324 (("$" 12005) (T_ABSTRACT 12182) (T_ARRAY 12183) (T_AS 12184) (T_BREAK 12185) (T_CALLABLE 12186) (T_CASE 12187) (T_CATCH 12188) (T_CLASS 12189) (T_CLASS_C 12190) (T_CLONE 12191) (T_CONST 12192) (T_CONTINUE 12193) (T_DECLARE 12194) (T_DEFAULT 12195) (T_DIR 12196) (T_DO 12197) (T_ECHO 12198) (T_ELSE 12199) (T_ELSEIF 12200) (T_EMPTY 12201) (T_ENDDECLARE 12202) (T_ENDFOR 12203) (T_ENDFOREACH 12204) (T_ENDIF 12205) (T_ENDSWITCH 12206) (T_ENDWHILE 12207) (T_ENUM 12208) (T_EVAL 12209) (T_EXIT 12210) (T_EXTENDS 12211) (T_FILE 12212) (T_FINAL 12213) (T_FINALLY 12214) (T_FN 12215) (T_FOR 12216) (T_FOREACH 12217) (T_FUNCTION 12218) (T_FUNC_C 12219) (T_GLOBAL 12220) (T_GOTO 12221) (T_IF 12222) (T_IMPLEMENTS 12223) (T_INCLUDE 12224) (T_INCLUDE_ONCE 12225) (T_INSTANCEOF 12226) (T_INSTEADOF 12227) (T_INTERFACE 12228) (T_ISSET 12229) (T_LINE 12230) (T_LIST 12231) (T_LOGICAL_AND 12232) (T_LOGICAL_OR 12233) (T_LOGICAL_XOR 12234) (T_MATCH 12235) (T_METHOD_C 12236) (T_NAMESPACE 12237) (T_NEW 12238) (T_NS_C 12239) (T_PRINT 12240) (T_PRIVATE 12241) (T_PROTECTED 12242) (T_PUBLIC 12243) (T_READONLY 12244) (T_REQUIRE 12245) (T_REQUIRE_ONCE 12246) (T_RETURN 12247) (T_STATIC 12248) (T_STRING 12249) (T_SWITCH 12250) (T_THROW 12251) (T_TRAIT 12252) (T_TRAIT_C 12253) (T_TRY 12254) (T_UNSET 12255) (T_USE 12256) (T_VAR 12257) (T_VARIABLE 12051) (T_WHILE 12258) (T_YIELD 12259) (identifier 12260) (member_name 12261) (reserved_non_modifiers 12262) (semi_reserved 12263) (simple_variable 12264) ("{" 914)) 5325 (("(" 12265) (argument_list 12266)) 5326 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 12267) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 12268) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 12269) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5327 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 12272) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5328 (("," 1449) (possible_comma 12270)) 5329 ((")" 12271)) 5330 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (ampersand 12302) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12303) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5331 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12301) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5332 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12300) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5333 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12299) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5334 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12298) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5335 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12297) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5336 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12296) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5337 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12295) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5338 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12294) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5339 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12293) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5340 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12292) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5341 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12291) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5342 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12290) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5343 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12289) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5344 (("\"" 1489) ("$" 12304) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 12305) ("[" 1508) (array_object_dereferenceable 12306) (callable_expr 12307) (callable_variable 12308) (class_constant 1512) (class_name 12309) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 12310) (name 12311) (simple_variable 12312) (static_member 12313) (variable 12314) (variable_class_name 12315)) 5345 (("$" 12304) (T_VARIABLE 12305) (simple_variable 12350) ("{" 12351)) 5346 ((T_NULLSAFE_OBJECT_OPERATOR 12333) (T_OBJECT_OPERATOR 12334) ("[" 12335) ("{" 12336)) 5347 (("(" 12319) (argument_list 12332)) 5348 ((T_PAAMAYIM_NEKUDOTAYIM 12328)) 5349 (("(" 12319) (argument_list 12327)) 5350 ((T_PAAMAYIM_NEKUDOTAYIM 12316)) 5351 (("$" 12304) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 12305) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 12317) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 12318) ("{" 914)) 5352 (("(" 12319) (argument_list 12320)) 5353 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 12321) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 12322) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 12323) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5354 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 12326) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5355 (("," 1449) (possible_comma 12324)) 5356 ((")" 12325)) 5357 (("$" 12304) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 12305) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 12329) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 12330) ("{" 914)) 5358 (("(" 12319) (argument_list 12331)) 5359 (("$" 12304) (T_STRING 12341) (T_VARIABLE 12305) (property_name 12348) (simple_variable 12343) ("{" 12344)) 5360 (("$" 12304) (T_STRING 12341) (T_VARIABLE 12305) (property_name 12342) (simple_variable 12343) ("{" 12344)) 5361 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 12339) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5362 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12337) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5363 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12338)) 5364 (("]" 12340)) 5365 (("(" 12319) (argument_list 12347)) 5366 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12345) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5367 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12346)) 5368 (("(" 12319) (argument_list 12349)) 5369 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12352) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5370 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12353)) 5371 ((backup_doc_comment 12356)) 5372 (("(" 12357)) 5373 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 12358)) 5374 ((")" 12359)) 5375 ((T_USE 148) (lexical_vars 12360)) 5376 ((":" 150) (return_type 12361)) 5377 ((backup_fn_flags 12362)) 5378 (("{" 12363)) 5379 ((inner_statement_list 12364)) 5380 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 12365) ("~" 124)) 5381 ((backup_fn_flags 12366)) 5382 ((backup_doc_comment 12368)) 5383 (("(" 12369)) 5384 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 12370)) 5385 ((")" 12371)) 5386 ((":" 2207) (return_type 12372)) 5387 ((T_DOUBLE_ARROW 12373)) 5388 ((backup_fn_flags 12374)) 5389 ((backup_lex_pos 12375)) 5390 (("!" 11339) ("\"" 11340) ("$" 11341) ("(" 11342) ("+" 11343) ("-" 11344) ("@" 11345) (T_ARRAY 11346) (T_ARRAY_CAST 11347) (T_ATTRIBUTE 125) (T_BOOL_CAST 11348) (T_CLASS_C 11349) (T_CLONE 11350) (T_CONSTANT_ENCAPSED_STRING 11351) (T_DEC 11352) (T_DIR 11353) (T_DNUMBER 11354) (T_DOUBLE_CAST 11355) (T_EMPTY 11356) (T_EVAL 11357) (T_EXIT 11358) (T_FILE 11359) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 11360) (T_INC 11361) (T_INCLUDE 11362) (T_INCLUDE_ONCE 11363) (T_INT_CAST 11364) (T_ISSET 11365) (T_LINE 11366) (T_LIST 11367) (T_LNUMBER 11368) (T_MATCH 11369) (T_METHOD_C 11370) (T_NAME_FULLY_QUALIFIED 11371) (T_NAME_QUALIFIED 11372) (T_NAME_RELATIVE 11373) (T_NEW 11374) (T_NS_C 11375) (T_OBJECT_CAST 11376) (T_PRINT 11377) (T_REQUIRE 11378) (T_REQUIRE_ONCE 11379) (T_START_HEREDOC 11380) (T_STATIC 11381) (T_STRING 11382) (T_STRING_CAST 11383) (T_THROW 11384) (T_TRAIT_C 11385) (T_UNSET_CAST 11386) (T_VARIABLE 11387) (T_YIELD 11388) (T_YIELD_FROM 11389) ("[" 11390) ("`" 11391) (array_object_dereferenceable 11392) (attribute 129) (attributes 11393) (callable_expr 11394) (callable_variable 11395) (class_constant 11396) (class_name 11397) (constant 11398) (dereferenceable_scalar 11399) (expr 12376) (fn 11401) (fully_dereferenceable 103) #'11402 (function_call 11403) (inline_function 11404) (internal_functions_in_yacc 11405) (match 11406) (name 11407) (new_expr 11408) (scalar 11409) (simple_variable 11410) (static_member 11411) (variable 11412) (variable_class_name 11413) ("~" 11414)) 5391 (("%" 11416) ("*" 11417) ("+" 11418) ("-" 11419) ("." 11420) ("/" 11421) ("<" 11422) (">" 11423) ("?" 11424) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11425) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11426) (T_BOOLEAN_AND 11427) (T_BOOLEAN_OR 11428) (T_COALESCE 11429) (T_INSTANCEOF 11430) (T_IS_EQUAL 11431) (T_IS_GREATER_OR_EQUAL 11432) (T_IS_IDENTICAL 11433) (T_IS_NOT_EQUAL 11434) (T_IS_NOT_IDENTICAL 11435) (T_IS_SMALLER_OR_EQUAL 11436) (T_LOGICAL_AND 11437) (T_LOGICAL_OR 11438) (T_LOGICAL_XOR 11439) (T_POW 11440) (T_SL 11441) (T_SPACESHIP 11442) (T_SR 11443) ("^" 11444) (backup_fn_flags 12377) ("|" 11445)) 5392 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12379) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5393 (("$" 12005) (T_ABSTRACT 12182) (T_ARRAY 12183) (T_AS 12184) (T_BREAK 12185) (T_CALLABLE 12186) (T_CASE 12187) (T_CATCH 12188) (T_CLASS 12189) (T_CLASS_C 12190) (T_CLONE 12191) (T_CONST 12192) (T_CONTINUE 12193) (T_DECLARE 12194) (T_DEFAULT 12195) (T_DIR 12196) (T_DO 12197) (T_ECHO 12198) (T_ELSE 12199) (T_ELSEIF 12200) (T_EMPTY 12201) (T_ENDDECLARE 12202) (T_ENDFOR 12203) (T_ENDFOREACH 12204) (T_ENDIF 12205) (T_ENDSWITCH 12206) (T_ENDWHILE 12207) (T_ENUM 12208) (T_EVAL 12209) (T_EXIT 12210) (T_EXTENDS 12211) (T_FILE 12212) (T_FINAL 12213) (T_FINALLY 12214) (T_FN 12215) (T_FOR 12216) (T_FOREACH 12217) (T_FUNCTION 12218) (T_FUNC_C 12219) (T_GLOBAL 12220) (T_GOTO 12221) (T_IF 12222) (T_IMPLEMENTS 12223) (T_INCLUDE 12224) (T_INCLUDE_ONCE 12225) (T_INSTANCEOF 12226) (T_INSTEADOF 12227) (T_INTERFACE 12228) (T_ISSET 12229) (T_LINE 12230) (T_LIST 12231) (T_LOGICAL_AND 12232) (T_LOGICAL_OR 12233) (T_LOGICAL_XOR 12234) (T_MATCH 12235) (T_METHOD_C 12236) (T_NAMESPACE 12237) (T_NEW 12238) (T_NS_C 12239) (T_PRINT 12240) (T_PRIVATE 12241) (T_PROTECTED 12242) (T_PUBLIC 12243) (T_READONLY 12244) (T_REQUIRE 12245) (T_REQUIRE_ONCE 12246) (T_RETURN 12247) (T_STATIC 12248) (T_STRING 12249) (T_SWITCH 12250) (T_THROW 12251) (T_TRAIT 12252) (T_TRAIT_C 12253) (T_TRY 12254) (T_UNSET 12255) (T_USE 12256) (T_VAR 12257) (T_VARIABLE 12051) (T_WHILE 12258) (T_YIELD 12259) (identifier 12381) (member_name 12382) (reserved_non_modifiers 12262) (semi_reserved 12263) (simple_variable 12383) ("{" 914)) 5394 (("(" 12265) (argument_list 12384)) 5395 ((T_FN 35) (T_FUNCTION 126) (fn 12065) #'12066 (inline_function 12388)) 5396 (("$" 12005) (T_STRING 12397) (T_VARIABLE 12051) (property_name 12404) (simple_variable 12399) ("{" 12400)) 5397 (("$" 12005) (T_STRING 12397) (T_VARIABLE 12051) (property_name 12398) (simple_variable 12399) ("{" 12400)) 5398 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 12395) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5399 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12393) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5400 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12394)) 5401 (("]" 12396)) 5402 (("(" 12265) (argument_list 12403)) 5403 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12401) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5404 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12402)) 5405 (("(" 12265) (argument_list 12405)) 5406 (("`" 12407)) 5407 (("]" 12409)) 5408 (("=" 12410)) 5409 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12411) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5410 (("%" 12080) ("*" 12081) ("+" 12082) ("-" 12083) ("." 12084) ("/" 12085) ("<" 12086) (">" 12087) ("?" 12088) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12089) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12090) (T_BOOLEAN_AND 12091) (T_BOOLEAN_OR 12092) (T_COALESCE 12093) (T_DOUBLE_ARROW 12414) (T_INSTANCEOF 12094) (T_IS_EQUAL 12095) (T_IS_GREATER_OR_EQUAL 12096) (T_IS_IDENTICAL 12097) (T_IS_NOT_EQUAL 12098) (T_IS_NOT_IDENTICAL 12099) (T_IS_SMALLER_OR_EQUAL 12100) (T_LOGICAL_AND 12101) (T_LOGICAL_OR 12102) (T_LOGICAL_XOR 12103) (T_POW 12104) (T_SL 12105) (T_SPACESHIP 12106) (T_SR 12107) ("^" 12108) ("|" 12109)) 5411 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12415) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5412 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 12424) (T_VARIABLE 2822) (encaps_var 2836)) 5413 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 12423) (T_VARIABLE 2822) (encaps_var 2827)) 5414 (("(" 2899) (argument_list 2900) (ctor_arguments 12443)) 5415 ((T_ATTRIBUTE 2853) (T_CLASS 12429) (anonymous_class 12442) (attribute 2898)) 5416 (("(" 12433) (argument_list 12434) (ctor_arguments 12435)) 5417 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 12436) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 12437) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 12438) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5418 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 12441) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5419 (("," 1449) (possible_comma 12439)) 5420 ((")" 12440)) 5421 ((T_EXTENDS 2902) (extends_from 12444)) 5422 ((T_IMPLEMENTS 2904) (implements_list 12445)) 5423 ((backup_doc_comment 12446)) 5424 (("{" 12447)) 5425 ((class_statement_list 12448)) 5426 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 12449)) 5427 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 12451) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5428 (("%" 280) (")" 12452) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5429 (("{" 12453)) 5430 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 12454) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 5431 (("}" 12455)) 5432 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 12457) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 5433 ((")" 12458)) 5434 (("=" 12459)) 5435 (("!" 12003) ("\"" 12004) ("$" 12005) ("(" 12006) ("+" 12007) ("-" 12008) ("@" 12009) (T_ARRAY 12010) (T_ARRAY_CAST 12011) (T_ATTRIBUTE 125) (T_BOOL_CAST 12012) (T_CLASS_C 12013) (T_CLONE 12014) (T_CONSTANT_ENCAPSED_STRING 12015) (T_DEC 12016) (T_DIR 12017) (T_DNUMBER 12018) (T_DOUBLE_CAST 12019) (T_EMPTY 12020) (T_EVAL 12021) (T_EXIT 12022) (T_FILE 12023) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12024) (T_INC 12025) (T_INCLUDE 12026) (T_INCLUDE_ONCE 12027) (T_INT_CAST 12028) (T_ISSET 12029) (T_LINE 12030) (T_LIST 12031) (T_LNUMBER 12032) (T_MATCH 12033) (T_METHOD_C 12034) (T_NAME_FULLY_QUALIFIED 12035) (T_NAME_QUALIFIED 12036) (T_NAME_RELATIVE 12037) (T_NEW 12038) (T_NS_C 12039) (T_OBJECT_CAST 12040) (T_PRINT 12041) (T_REQUIRE 12042) (T_REQUIRE_ONCE 12043) (T_START_HEREDOC 12044) (T_STATIC 12045) (T_STRING 12046) (T_STRING_CAST 12047) (T_THROW 12048) (T_TRAIT_C 12049) (T_UNSET_CAST 12050) (T_VARIABLE 12051) (T_YIELD 12052) (T_YIELD_FROM 12053) ("[" 12054) ("`" 12055) (array_object_dereferenceable 12056) (attribute 129) (attributes 12057) (callable_expr 12058) (callable_variable 12059) (class_constant 12060) (class_name 12061) (constant 12062) (dereferenceable_scalar 12063) (expr 12460) (fn 12065) (fully_dereferenceable 103) #'12066 (function_call 12067) (inline_function 12068) (internal_functions_in_yacc 12069) (match 12070) (name 12071) (new_expr 12072) (scalar 12073) (simple_variable 12074) (static_member 12075) (variable 12076) (variable_class_name 12077) ("~" 12078)) 5436 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 12462) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5437 (("," 5117) (possible_comma 12463)) 5438 ((")" 12464)) 5439 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 12471) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5440 ((")" 12472)) 5441 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 12474) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5442 (("%" 280) (")" 12475) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5443 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 12477) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5444 (("%" 280) (")" 12478) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5445 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 12485) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 5446 ((")" 12486)) 5447 (("%" 280) (")" 12491) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5448 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12494) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5449 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12495)) 5450 (("\"" 12497) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 5451 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 12507) (T_VARIABLE 2822) (encaps_var 2836)) 5452 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 12506) (T_VARIABLE 2822) (encaps_var 2827)) 5453 (("(" 2899) (argument_list 2900) (ctor_arguments 12526)) 5454 ((T_ATTRIBUTE 2853) (T_CLASS 12512) (anonymous_class 12525) (attribute 2898)) 5455 (("(" 12516) (argument_list 12517) (ctor_arguments 12518)) 5456 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 12519) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 12520) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 12521) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5457 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 12524) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5458 (("," 1449) (possible_comma 12522)) 5459 ((")" 12523)) 5460 ((T_EXTENDS 2902) (extends_from 12527)) 5461 ((T_IMPLEMENTS 2904) (implements_list 12528)) 5462 ((backup_doc_comment 12529)) 5463 (("{" 12530)) 5464 ((class_statement_list 12531)) 5465 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 12532)) 5466 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 12534) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5467 (("%" 280) (")" 12535) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5468 (("{" 12536)) 5469 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 12537) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 5470 (("}" 12538)) 5471 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 12540) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 5472 ((")" 12541)) 5473 (("=" 12542)) 5474 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 12543) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5475 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 12545) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5476 (("," 5117) (possible_comma 12546)) 5477 ((")" 12547)) 5478 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 12554) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5479 ((")" 12555)) 5480 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 12557) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5481 (("%" 280) (")" 12558) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5482 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 12560) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5483 (("%" 280) (")" 12561) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5484 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 12568) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 5485 ((")" 12569)) 5486 (("%" 280) (")" 12574) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5487 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12577) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5488 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12578)) 5489 (("\"" 12580) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 5490 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12587) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5491 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12588)) 5492 (("$" 465) (T_VARIABLE 472) (simple_variable 12591)) 5493 (("%" 280) (")" 12593) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5494 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12596) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5495 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12597)) 5496 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12607) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5497 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 12605) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 5498 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12606) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5499 (("$" 345) (T_ABSTRACT 12617) (T_ARRAY 12618) (T_AS 12619) (T_BREAK 12620) (T_CALLABLE 12621) (T_CASE 12622) (T_CATCH 12623) (T_CLASS 12624) (T_CLASS_C 12625) (T_CLONE 12626) (T_CONST 12627) (T_CONTINUE 12628) (T_DECLARE 12629) (T_DEFAULT 12630) (T_DIR 12631) (T_DO 12632) (T_ECHO 12633) (T_ELSE 12634) (T_ELSEIF 12635) (T_EMPTY 12636) (T_ENDDECLARE 12637) (T_ENDFOR 12638) (T_ENDFOREACH 12639) (T_ENDIF 12640) (T_ENDSWITCH 12641) (T_ENDWHILE 12642) (T_ENUM 12643) (T_EVAL 12644) (T_EXIT 12645) (T_EXTENDS 12646) (T_FILE 12647) (T_FINAL 12648) (T_FINALLY 12649) (T_FN 12650) (T_FOR 12651) (T_FOREACH 12652) (T_FUNCTION 12653) (T_FUNC_C 12654) (T_GLOBAL 12655) (T_GOTO 12656) (T_IF 12657) (T_IMPLEMENTS 12658) (T_INCLUDE 12659) (T_INCLUDE_ONCE 12660) (T_INSTANCEOF 12661) (T_INSTEADOF 12662) (T_INTERFACE 12663) (T_ISSET 12664) (T_LINE 12665) (T_LIST 12666) (T_LOGICAL_AND 12667) (T_LOGICAL_OR 12668) (T_LOGICAL_XOR 12669) (T_MATCH 12670) (T_METHOD_C 12671) (T_NAMESPACE 12672) (T_NEW 12673) (T_NS_C 12674) (T_PRINT 12675) (T_PRIVATE 12676) (T_PROTECTED 12677) (T_PUBLIC 12678) (T_READONLY 12679) (T_REQUIRE 12680) (T_REQUIRE_ONCE 12681) (T_RETURN 12682) (T_STATIC 12683) (T_STRING 12684) (T_SWITCH 12685) (T_THROW 12686) (T_TRAIT 12687) (T_TRAIT_C 12688) (T_TRY 12689) (T_UNSET 12690) (T_USE 12691) (T_VAR 12692) (T_VARIABLE 391) (T_WHILE 12693) (T_YIELD 12694) (identifier 12695) (member_name 12696) (reserved_non_modifiers 12697) (semi_reserved 12698) (simple_variable 12699) ("{" 914)) 5500 (("(" 12700) (argument_list 12701)) 5501 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 12702) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 12703) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 12704) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5502 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 12707) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5503 (("," 1449) (possible_comma 12705)) 5504 ((")" 12706)) 5505 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (ampersand 12737) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12738) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5506 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12736) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5507 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12735) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5508 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12734) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5509 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12733) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5510 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12732) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5511 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12731) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5512 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12730) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5513 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12729) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5514 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12728) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5515 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12727) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5516 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12726) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5517 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12725) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5518 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12724) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5519 (("\"" 1489) ("$" 12739) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 12740) ("[" 1508) (array_object_dereferenceable 12741) (callable_expr 12742) (callable_variable 12743) (class_constant 1512) (class_name 12744) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 12745) (name 12746) (simple_variable 12747) (static_member 12748) (variable 12749) (variable_class_name 12750)) 5520 (("$" 12739) (T_VARIABLE 12740) (simple_variable 12785) ("{" 12786)) 5521 ((T_NULLSAFE_OBJECT_OPERATOR 12768) (T_OBJECT_OPERATOR 12769) ("[" 12770) ("{" 12771)) 5522 (("(" 12754) (argument_list 12767)) 5523 ((T_PAAMAYIM_NEKUDOTAYIM 12763)) 5524 (("(" 12754) (argument_list 12762)) 5525 ((T_PAAMAYIM_NEKUDOTAYIM 12751)) 5526 (("$" 12739) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 12740) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 12752) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 12753) ("{" 914)) 5527 (("(" 12754) (argument_list 12755)) 5528 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 12756) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 12757) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 12758) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5529 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 12761) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5530 (("," 1449) (possible_comma 12759)) 5531 ((")" 12760)) 5532 (("$" 12739) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 12740) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 12764) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 12765) ("{" 914)) 5533 (("(" 12754) (argument_list 12766)) 5534 (("$" 12739) (T_STRING 12776) (T_VARIABLE 12740) (property_name 12783) (simple_variable 12778) ("{" 12779)) 5535 (("$" 12739) (T_STRING 12776) (T_VARIABLE 12740) (property_name 12777) (simple_variable 12778) ("{" 12779)) 5536 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 12774) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5537 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12772) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5538 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12773)) 5539 (("]" 12775)) 5540 (("(" 12754) (argument_list 12782)) 5541 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12780) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5542 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12781)) 5543 (("(" 12754) (argument_list 12784)) 5544 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12787) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5545 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12788)) 5546 ((backup_doc_comment 12791)) 5547 (("(" 12792)) 5548 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 12793)) 5549 ((")" 12794)) 5550 ((T_USE 148) (lexical_vars 12795)) 5551 ((":" 150) (return_type 12796)) 5552 ((backup_fn_flags 12797)) 5553 (("{" 12798)) 5554 ((inner_statement_list 12799)) 5555 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 12800) ("~" 124)) 5556 ((backup_fn_flags 12801)) 5557 ((backup_doc_comment 12803)) 5558 (("(" 12804)) 5559 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 12805)) 5560 ((")" 12806)) 5561 ((":" 2207) (return_type 12807)) 5562 ((T_DOUBLE_ARROW 12808)) 5563 ((backup_fn_flags 12809)) 5564 ((backup_lex_pos 12810)) 5565 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12872) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5566 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13850) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5567 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 13848) (encaps_var 5158)) 5568 (("$" 12813) (T_VARIABLE 12859) (simple_variable 13844) ("{" 13845)) 5569 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 13842) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5570 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13841) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5571 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13840) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5572 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13839) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5573 (("(" 13836)) 5574 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13835) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5575 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13834) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5576 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13833) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5577 (("\"" 1489) ("$" 13112) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 13113) ("[" 1508) (array_object_dereferenceable 13114) (callable_expr 13115) (callable_variable 13116) (class_constant 1512) (class_name 13117) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 13118) (name 13119) (simple_variable 13120) (static_member 13121) (variable 13832) (variable_class_name 13123)) 5578 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13831) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5579 (("(" 13828)) 5580 (("(" 13825)) 5581 (("(" 13821) (exit_expr 13822)) 5582 (("\"" 1489) ("$" 13112) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 13113) ("[" 1508) (array_object_dereferenceable 13114) (callable_expr 13115) (callable_variable 13116) (class_constant 1512) (class_name 13117) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 13118) (name 13119) (simple_variable 13120) (static_member 13121) (variable 13820) (variable_class_name 13123)) 5583 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13819) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5584 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13818) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5585 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13817) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5586 (("(" 13813)) 5587 (("(" 13808)) 5588 (("(" 13802)) 5589 (("$" 13112) ("(" 13755) (T_ATTRIBUTE 2853) (T_CLASS 13756) (T_NAME_FULLY_QUALIFIED 13757) (T_NAME_QUALIFIED 13758) (T_NAME_RELATIVE 13759) (T_STATIC 13760) (T_STRING 13761) (T_VARIABLE 13113) (anonymous_class 13762) (attribute 2862) (attributes 13763) (class_name 13764) (class_name_reference 13765) (name 13766) (new_variable 13767) (simple_variable 13768)) 5590 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13754) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5591 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13753) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5592 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13752) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5593 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13751) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5594 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 13746) (T_END_HEREDOC 13747) (T_VARIABLE 2822) (encaps_list 13748) (encaps_var 2824)) 5595 ((T_FN 35) (T_FUNCTION 126) (fn 12873) #'12874 (inline_function 13745)) 5596 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13744) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5597 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13743) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5598 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13742) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5599 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13281) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5600 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13219) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5601 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 13215) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 5602 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 13213) (encaps_list 2729) (encaps_var 2730)) 5603 ((T_NULLSAFE_OBJECT_OPERATOR 13196) (T_OBJECT_OPERATOR 13197) ("[" 13198) ("{" 13199)) 5604 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 13193) (attribute 2654) (fn 12873) #'12874 (inline_function 13194)) 5605 (("(" 13073) (argument_list 13192)) 5606 ((T_PAAMAYIM_NEKUDOTAYIM 13187)) 5607 (("%" 12888) ("*" 12889) ("+" 12890) ("-" 12891) ("." 12892) ("/" 12893) ("<" 12894) (">" 12895) ("?" 12896) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12897) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12898) (T_BOOLEAN_AND 12899) (T_BOOLEAN_OR 12900) (T_COALESCE 12901) (T_INSTANCEOF 12902) (T_IS_EQUAL 12903) (T_IS_GREATER_OR_EQUAL 12904) (T_IS_IDENTICAL 12905) (T_IS_NOT_EQUAL 12906) (T_IS_NOT_IDENTICAL 12907) (T_IS_SMALLER_OR_EQUAL 12908) (T_LOGICAL_AND 12909) (T_LOGICAL_OR 12910) (T_LOGICAL_XOR 12911) (T_POW 12912) (T_SL 12913) (T_SPACESHIP 12914) (T_SR 12915) ("^" 12916) (backup_fn_flags 13186) ("|" 12917)) 5608 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 13175)) 5609 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 13163)) 5610 (("(" 13073) (argument_list 13162)) 5611 (("=" 13081) (T_AND_EQUAL 13082) (T_COALESCE_EQUAL 13083) (T_CONCAT_EQUAL 13084) (T_DEC 13085) (T_DIV_EQUAL 13086) (T_INC 13087) (T_MINUS_EQUAL 13088) (T_MOD_EQUAL 13089) (T_MUL_EQUAL 13090) (T_OR_EQUAL 13091) (T_PLUS_EQUAL 13092) (T_POW_EQUAL 13093) (T_SL_EQUAL 13094) (T_SR_EQUAL 13095) (T_XOR_EQUAL 13096)) 5612 ((T_PAAMAYIM_NEKUDOTAYIM 12989)) 5613 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12887) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5614 (("%" 12888) ("*" 12889) ("+" 12890) ("-" 12891) ("." 12892) ("/" 12893) ("<" 12894) (">" 12895) ("?" 12896) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12897) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12898) (T_BOOLEAN_AND 12899) (T_BOOLEAN_OR 12900) (T_COALESCE 12901) (T_INSTANCEOF 12902) (T_IS_EQUAL 12903) (T_IS_GREATER_OR_EQUAL 12904) (T_IS_IDENTICAL 12905) (T_IS_NOT_EQUAL 12906) (T_IS_NOT_IDENTICAL 12907) (T_IS_SMALLER_OR_EQUAL 12908) (T_LOGICAL_AND 12909) (T_LOGICAL_OR 12910) (T_LOGICAL_XOR 12911) (T_POW 12912) (T_SL 12913) (T_SPACESHIP 12914) (T_SR 12915) ("^" 12916) ("|" 12917)) 5615 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12988) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5616 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12987) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5617 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12986) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5618 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12985) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5619 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12984) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5620 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12983) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5621 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12982) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5622 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12981) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5623 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 12976) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 12977) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 5624 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12975) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5625 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12974) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5626 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12973) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5627 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12972) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5628 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12971) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5629 (("$" 12933) ("(" 12934) (T_NAME_FULLY_QUALIFIED 12935) (T_NAME_QUALIFIED 12936) (T_NAME_RELATIVE 12937) (T_STATIC 12938) (T_STRING 12939) (T_VARIABLE 12940) (class_name 12941) (class_name_reference 12942) (name 12943) (new_variable 12944) (simple_variable 12945)) 5630 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12932) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5631 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12931) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5632 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12930) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5633 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12929) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5634 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12928) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5635 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12927) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5636 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12926) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5637 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12925) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5638 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12924) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5639 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12923) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5640 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12922) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5641 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12921) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5642 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12920) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5643 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12919) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5644 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12918) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5645 (("$" 12933) (T_VARIABLE 12940) (simple_variable 12967) ("{" 12968)) 5646 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 12965) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5647 ((T_PAAMAYIM_NEKUDOTAYIM 12963)) 5648 ((T_NULLSAFE_OBJECT_OPERATOR 12946) (T_OBJECT_OPERATOR 12947) (T_PAAMAYIM_NEKUDOTAYIM 12948) ("[" 12949) ("{" 12950)) 5649 (("$" 12933) (T_STRING 12956) (T_VARIABLE 12940) (property_name 12962) (simple_variable 12958) ("{" 12959)) 5650 (("$" 12933) (T_STRING 12956) (T_VARIABLE 12940) (property_name 12957) (simple_variable 12958) ("{" 12959)) 5651 (("$" 12933) (T_VARIABLE 12940) (simple_variable 12955)) 5652 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 12953) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5653 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12951) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5654 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12952)) 5655 (("]" 12954)) 5656 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12960) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5657 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12961)) 5658 (("$" 12933) (T_VARIABLE 12940) (simple_variable 12964)) 5659 (("%" 280) (")" 12966) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5660 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 12969) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5661 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 12970)) 5662 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12980) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5663 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 12978) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 5664 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 12979) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5665 (("$" 12813) (T_ABSTRACT 12990) (T_ARRAY 12991) (T_AS 12992) (T_BREAK 12993) (T_CALLABLE 12994) (T_CASE 12995) (T_CATCH 12996) (T_CLASS 12997) (T_CLASS_C 12998) (T_CLONE 12999) (T_CONST 13000) (T_CONTINUE 13001) (T_DECLARE 13002) (T_DEFAULT 13003) (T_DIR 13004) (T_DO 13005) (T_ECHO 13006) (T_ELSE 13007) (T_ELSEIF 13008) (T_EMPTY 13009) (T_ENDDECLARE 13010) (T_ENDFOR 13011) (T_ENDFOREACH 13012) (T_ENDIF 13013) (T_ENDSWITCH 13014) (T_ENDWHILE 13015) (T_ENUM 13016) (T_EVAL 13017) (T_EXIT 13018) (T_EXTENDS 13019) (T_FILE 13020) (T_FINAL 13021) (T_FINALLY 13022) (T_FN 13023) (T_FOR 13024) (T_FOREACH 13025) (T_FUNCTION 13026) (T_FUNC_C 13027) (T_GLOBAL 13028) (T_GOTO 13029) (T_IF 13030) (T_IMPLEMENTS 13031) (T_INCLUDE 13032) (T_INCLUDE_ONCE 13033) (T_INSTANCEOF 13034) (T_INSTEADOF 13035) (T_INTERFACE 13036) (T_ISSET 13037) (T_LINE 13038) (T_LIST 13039) (T_LOGICAL_AND 13040) (T_LOGICAL_OR 13041) (T_LOGICAL_XOR 13042) (T_MATCH 13043) (T_METHOD_C 13044) (T_NAMESPACE 13045) (T_NEW 13046) (T_NS_C 13047) (T_PRINT 13048) (T_PRIVATE 13049) (T_PROTECTED 13050) (T_PUBLIC 13051) (T_READONLY 13052) (T_REQUIRE 13053) (T_REQUIRE_ONCE 13054) (T_RETURN 13055) (T_STATIC 13056) (T_STRING 13057) (T_SWITCH 13058) (T_THROW 13059) (T_TRAIT 13060) (T_TRAIT_C 13061) (T_TRY 13062) (T_UNSET 13063) (T_USE 13064) (T_VAR 13065) (T_VARIABLE 12859) (T_WHILE 13066) (T_YIELD 13067) (identifier 13068) (member_name 13069) (reserved_non_modifiers 13070) (semi_reserved 13071) (simple_variable 13072) ("{" 914)) 5666 (("(" 13073) (argument_list 13074)) 5667 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 13075) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 13076) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 13077) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5668 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 13080) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5669 (("," 1449) (possible_comma 13078)) 5670 ((")" 13079)) 5671 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (ampersand 13110) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13111) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5672 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13109) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5673 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13108) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5674 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13107) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5675 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13106) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5676 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13105) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5677 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13104) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5678 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13103) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5679 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13102) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5680 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13101) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5681 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13100) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5682 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13099) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5683 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13098) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5684 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13097) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5685 (("\"" 1489) ("$" 13112) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 13113) ("[" 1508) (array_object_dereferenceable 13114) (callable_expr 13115) (callable_variable 13116) (class_constant 1512) (class_name 13117) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 13118) (name 13119) (simple_variable 13120) (static_member 13121) (variable 13122) (variable_class_name 13123)) 5686 (("$" 13112) (T_VARIABLE 13113) (simple_variable 13158) ("{" 13159)) 5687 ((T_NULLSAFE_OBJECT_OPERATOR 13141) (T_OBJECT_OPERATOR 13142) ("[" 13143) ("{" 13144)) 5688 (("(" 13127) (argument_list 13140)) 5689 ((T_PAAMAYIM_NEKUDOTAYIM 13136)) 5690 (("(" 13127) (argument_list 13135)) 5691 ((T_PAAMAYIM_NEKUDOTAYIM 13124)) 5692 (("$" 13112) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 13113) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 13125) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 13126) ("{" 914)) 5693 (("(" 13127) (argument_list 13128)) 5694 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 13129) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 13130) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 13131) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5695 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 13134) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5696 (("," 1449) (possible_comma 13132)) 5697 ((")" 13133)) 5698 (("$" 13112) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 13113) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 13137) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 13138) ("{" 914)) 5699 (("(" 13127) (argument_list 13139)) 5700 (("$" 13112) (T_STRING 13149) (T_VARIABLE 13113) (property_name 13156) (simple_variable 13151) ("{" 13152)) 5701 (("$" 13112) (T_STRING 13149) (T_VARIABLE 13113) (property_name 13150) (simple_variable 13151) ("{" 13152)) 5702 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 13147) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5703 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13145) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5704 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13146)) 5705 (("]" 13148)) 5706 (("(" 13127) (argument_list 13155)) 5707 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13153) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5708 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13154)) 5709 (("(" 13127) (argument_list 13157)) 5710 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13160) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5711 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13161)) 5712 ((backup_doc_comment 13164)) 5713 (("(" 13165)) 5714 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 13166)) 5715 ((")" 13167)) 5716 ((T_USE 148) (lexical_vars 13168)) 5717 ((":" 150) (return_type 13169)) 5718 ((backup_fn_flags 13170)) 5719 (("{" 13171)) 5720 ((inner_statement_list 13172)) 5721 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 13173) ("~" 124)) 5722 ((backup_fn_flags 13174)) 5723 ((backup_doc_comment 13176)) 5724 (("(" 13177)) 5725 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 13178)) 5726 ((")" 13179)) 5727 ((":" 2207) (return_type 13180)) 5728 ((T_DOUBLE_ARROW 13181)) 5729 ((backup_fn_flags 13182)) 5730 ((backup_lex_pos 13183)) 5731 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13184) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5732 (("%" 12888) ("*" 12889) ("+" 12890) ("-" 12891) ("." 12892) ("/" 12893) ("<" 12894) (">" 12895) ("?" 12896) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12897) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12898) (T_BOOLEAN_AND 12899) (T_BOOLEAN_OR 12900) (T_COALESCE 12901) (T_INSTANCEOF 12902) (T_IS_EQUAL 12903) (T_IS_GREATER_OR_EQUAL 12904) (T_IS_IDENTICAL 12905) (T_IS_NOT_EQUAL 12906) (T_IS_NOT_IDENTICAL 12907) (T_IS_SMALLER_OR_EQUAL 12908) (T_LOGICAL_AND 12909) (T_LOGICAL_OR 12910) (T_LOGICAL_XOR 12911) (T_POW 12912) (T_SL 12913) (T_SPACESHIP 12914) (T_SR 12915) ("^" 12916) (backup_fn_flags 13185) ("|" 12917)) 5733 (("$" 12813) (T_ABSTRACT 12990) (T_ARRAY 12991) (T_AS 12992) (T_BREAK 12993) (T_CALLABLE 12994) (T_CASE 12995) (T_CATCH 12996) (T_CLASS 12997) (T_CLASS_C 12998) (T_CLONE 12999) (T_CONST 13000) (T_CONTINUE 13001) (T_DECLARE 13002) (T_DEFAULT 13003) (T_DIR 13004) (T_DO 13005) (T_ECHO 13006) (T_ELSE 13007) (T_ELSEIF 13008) (T_EMPTY 13009) (T_ENDDECLARE 13010) (T_ENDFOR 13011) (T_ENDFOREACH 13012) (T_ENDIF 13013) (T_ENDSWITCH 13014) (T_ENDWHILE 13015) (T_ENUM 13016) (T_EVAL 13017) (T_EXIT 13018) (T_EXTENDS 13019) (T_FILE 13020) (T_FINAL 13021) (T_FINALLY 13022) (T_FN 13023) (T_FOR 13024) (T_FOREACH 13025) (T_FUNCTION 13026) (T_FUNC_C 13027) (T_GLOBAL 13028) (T_GOTO 13029) (T_IF 13030) (T_IMPLEMENTS 13031) (T_INCLUDE 13032) (T_INCLUDE_ONCE 13033) (T_INSTANCEOF 13034) (T_INSTEADOF 13035) (T_INTERFACE 13036) (T_ISSET 13037) (T_LINE 13038) (T_LIST 13039) (T_LOGICAL_AND 13040) (T_LOGICAL_OR 13041) (T_LOGICAL_XOR 13042) (T_MATCH 13043) (T_METHOD_C 13044) (T_NAMESPACE 13045) (T_NEW 13046) (T_NS_C 13047) (T_PRINT 13048) (T_PRIVATE 13049) (T_PROTECTED 13050) (T_PUBLIC 13051) (T_READONLY 13052) (T_REQUIRE 13053) (T_REQUIRE_ONCE 13054) (T_RETURN 13055) (T_STATIC 13056) (T_STRING 13057) (T_SWITCH 13058) (T_THROW 13059) (T_TRAIT 13060) (T_TRAIT_C 13061) (T_TRY 13062) (T_UNSET 13063) (T_USE 13064) (T_VAR 13065) (T_VARIABLE 12859) (T_WHILE 13066) (T_YIELD 13067) (identifier 13188) (member_name 13189) (reserved_non_modifiers 13070) (semi_reserved 13071) (simple_variable 13190) ("{" 914)) 5734 (("(" 13073) (argument_list 13191)) 5735 ((T_FN 35) (T_FUNCTION 126) (fn 12873) #'12874 (inline_function 13195)) 5736 (("$" 12813) (T_STRING 13204) (T_VARIABLE 12859) (property_name 13211) (simple_variable 13206) ("{" 13207)) 5737 (("$" 12813) (T_STRING 13204) (T_VARIABLE 12859) (property_name 13205) (simple_variable 13206) ("{" 13207)) 5738 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 13202) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5739 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13200) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5740 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13201)) 5741 (("]" 13203)) 5742 (("(" 13073) (argument_list 13210)) 5743 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13208) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5744 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13209)) 5745 (("(" 13073) (argument_list 13212)) 5746 (("`" 13214)) 5747 (("]" 13216)) 5748 (("=" 13217)) 5749 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13218) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5750 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13741) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5751 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 13739) (encaps_var 5158)) 5752 (("$" 13222) (T_VARIABLE 13268) (simple_variable 13735) ("{" 13736)) 5753 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 13733) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5754 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13732) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5755 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13731) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5756 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13730) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5757 (("(" 13727)) 5758 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13726) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5759 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13725) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5760 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13724) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5761 (("\"" 1489) ("$" 13521) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 13522) ("[" 1508) (array_object_dereferenceable 13523) (callable_expr 13524) (callable_variable 13525) (class_constant 1512) (class_name 13526) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 13527) (name 13528) (simple_variable 13529) (static_member 13530) (variable 13723) (variable_class_name 13532)) 5762 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13722) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5763 (("(" 13719)) 5764 (("(" 13716)) 5765 (("(" 13712) (exit_expr 13713)) 5766 (("\"" 1489) ("$" 13521) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 13522) ("[" 1508) (array_object_dereferenceable 13523) (callable_expr 13524) (callable_variable 13525) (class_constant 1512) (class_name 13526) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 13527) (name 13528) (simple_variable 13529) (static_member 13530) (variable 13711) (variable_class_name 13532)) 5767 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13710) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5768 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13709) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5769 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13708) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5770 (("(" 13704)) 5771 (("(" 13699)) 5772 (("(" 13693)) 5773 (("$" 13521) ("(" 13646) (T_ATTRIBUTE 2853) (T_CLASS 13647) (T_NAME_FULLY_QUALIFIED 13648) (T_NAME_QUALIFIED 13649) (T_NAME_RELATIVE 13650) (T_STATIC 13651) (T_STRING 13652) (T_VARIABLE 13522) (anonymous_class 13653) (attribute 2862) (attributes 13654) (class_name 13655) (class_name_reference 13656) (name 13657) (new_variable 13658) (simple_variable 13659)) 5774 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13645) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5775 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13644) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5776 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13643) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5777 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13642) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5778 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 13637) (T_END_HEREDOC 13638) (T_VARIABLE 2822) (encaps_list 13639) (encaps_var 2824)) 5779 ((T_FN 35) (T_FUNCTION 126) (fn 13282) #'13283 (inline_function 13636)) 5780 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13635) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5781 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13634) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5782 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13633) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5783 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13630) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5784 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13629) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5785 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 13625) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 5786 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 13623) (encaps_list 2729) (encaps_var 2730)) 5787 ((T_NULLSAFE_OBJECT_OPERATOR 13606) (T_OBJECT_OPERATOR 13607) ("[" 13608) ("{" 13609)) 5788 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 13603) (attribute 2654) (fn 13282) #'13283 (inline_function 13604)) 5789 (("(" 13482) (argument_list 13602)) 5790 ((T_PAAMAYIM_NEKUDOTAYIM 13597)) 5791 (("%" 13297) ("*" 13298) ("+" 13299) ("-" 13300) ("." 13301) ("/" 13302) ("<" 13303) (">" 13304) ("?" 13305) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13306) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13307) (T_BOOLEAN_AND 13308) (T_BOOLEAN_OR 13309) (T_COALESCE 13310) (T_DOUBLE_ARROW 13595) (T_INSTANCEOF 13311) (T_IS_EQUAL 13312) (T_IS_GREATER_OR_EQUAL 13313) (T_IS_IDENTICAL 13314) (T_IS_NOT_EQUAL 13315) (T_IS_NOT_IDENTICAL 13316) (T_IS_SMALLER_OR_EQUAL 13317) (T_LOGICAL_AND 13318) (T_LOGICAL_OR 13319) (T_LOGICAL_XOR 13320) (T_POW 13321) (T_SL 13322) (T_SPACESHIP 13323) (T_SR 13324) ("^" 13325) ("|" 13326)) 5792 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 13584)) 5793 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 13572)) 5794 (("(" 13482) (argument_list 13571)) 5795 (("=" 13490) (T_AND_EQUAL 13491) (T_COALESCE_EQUAL 13492) (T_CONCAT_EQUAL 13493) (T_DEC 13494) (T_DIV_EQUAL 13495) (T_INC 13496) (T_MINUS_EQUAL 13497) (T_MOD_EQUAL 13498) (T_MUL_EQUAL 13499) (T_OR_EQUAL 13500) (T_PLUS_EQUAL 13501) (T_POW_EQUAL 13502) (T_SL_EQUAL 13503) (T_SR_EQUAL 13504) (T_XOR_EQUAL 13505)) 5796 ((T_PAAMAYIM_NEKUDOTAYIM 13398)) 5797 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13296) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5798 (("%" 13297) ("*" 13298) ("+" 13299) ("-" 13300) ("." 13301) ("/" 13302) ("<" 13303) (">" 13304) ("?" 13305) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13306) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13307) (T_BOOLEAN_AND 13308) (T_BOOLEAN_OR 13309) (T_COALESCE 13310) (T_INSTANCEOF 13311) (T_IS_EQUAL 13312) (T_IS_GREATER_OR_EQUAL 13313) (T_IS_IDENTICAL 13314) (T_IS_NOT_EQUAL 13315) (T_IS_NOT_IDENTICAL 13316) (T_IS_SMALLER_OR_EQUAL 13317) (T_LOGICAL_AND 13318) (T_LOGICAL_OR 13319) (T_LOGICAL_XOR 13320) (T_POW 13321) (T_SL 13322) (T_SPACESHIP 13323) (T_SR 13324) ("^" 13325) ("|" 13326)) 5799 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13397) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5800 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13396) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5801 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13395) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5802 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13394) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5803 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13393) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5804 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13392) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5805 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13391) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5806 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13390) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5807 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 13385) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 13386) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 5808 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13384) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5809 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13383) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5810 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13382) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5811 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13381) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5812 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13380) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5813 (("$" 13342) ("(" 13343) (T_NAME_FULLY_QUALIFIED 13344) (T_NAME_QUALIFIED 13345) (T_NAME_RELATIVE 13346) (T_STATIC 13347) (T_STRING 13348) (T_VARIABLE 13349) (class_name 13350) (class_name_reference 13351) (name 13352) (new_variable 13353) (simple_variable 13354)) 5814 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13341) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5815 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13340) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5816 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13339) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5817 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13338) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5818 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13337) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5819 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13336) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5820 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13335) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5821 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13334) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5822 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13333) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5823 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13332) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5824 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13331) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5825 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13330) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5826 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13329) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5827 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13328) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5828 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13327) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5829 (("$" 13342) (T_VARIABLE 13349) (simple_variable 13376) ("{" 13377)) 5830 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 13374) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5831 ((T_PAAMAYIM_NEKUDOTAYIM 13372)) 5832 ((T_NULLSAFE_OBJECT_OPERATOR 13355) (T_OBJECT_OPERATOR 13356) (T_PAAMAYIM_NEKUDOTAYIM 13357) ("[" 13358) ("{" 13359)) 5833 (("$" 13342) (T_STRING 13365) (T_VARIABLE 13349) (property_name 13371) (simple_variable 13367) ("{" 13368)) 5834 (("$" 13342) (T_STRING 13365) (T_VARIABLE 13349) (property_name 13366) (simple_variable 13367) ("{" 13368)) 5835 (("$" 13342) (T_VARIABLE 13349) (simple_variable 13364)) 5836 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 13362) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5837 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13360) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5838 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13361)) 5839 (("]" 13363)) 5840 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13369) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5841 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13370)) 5842 (("$" 13342) (T_VARIABLE 13349) (simple_variable 13373)) 5843 (("%" 280) (")" 13375) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5844 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13378) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5845 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13379)) 5846 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13389) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5847 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 13387) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 5848 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13388) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5849 (("$" 13222) (T_ABSTRACT 13399) (T_ARRAY 13400) (T_AS 13401) (T_BREAK 13402) (T_CALLABLE 13403) (T_CASE 13404) (T_CATCH 13405) (T_CLASS 13406) (T_CLASS_C 13407) (T_CLONE 13408) (T_CONST 13409) (T_CONTINUE 13410) (T_DECLARE 13411) (T_DEFAULT 13412) (T_DIR 13413) (T_DO 13414) (T_ECHO 13415) (T_ELSE 13416) (T_ELSEIF 13417) (T_EMPTY 13418) (T_ENDDECLARE 13419) (T_ENDFOR 13420) (T_ENDFOREACH 13421) (T_ENDIF 13422) (T_ENDSWITCH 13423) (T_ENDWHILE 13424) (T_ENUM 13425) (T_EVAL 13426) (T_EXIT 13427) (T_EXTENDS 13428) (T_FILE 13429) (T_FINAL 13430) (T_FINALLY 13431) (T_FN 13432) (T_FOR 13433) (T_FOREACH 13434) (T_FUNCTION 13435) (T_FUNC_C 13436) (T_GLOBAL 13437) (T_GOTO 13438) (T_IF 13439) (T_IMPLEMENTS 13440) (T_INCLUDE 13441) (T_INCLUDE_ONCE 13442) (T_INSTANCEOF 13443) (T_INSTEADOF 13444) (T_INTERFACE 13445) (T_ISSET 13446) (T_LINE 13447) (T_LIST 13448) (T_LOGICAL_AND 13449) (T_LOGICAL_OR 13450) (T_LOGICAL_XOR 13451) (T_MATCH 13452) (T_METHOD_C 13453) (T_NAMESPACE 13454) (T_NEW 13455) (T_NS_C 13456) (T_PRINT 13457) (T_PRIVATE 13458) (T_PROTECTED 13459) (T_PUBLIC 13460) (T_READONLY 13461) (T_REQUIRE 13462) (T_REQUIRE_ONCE 13463) (T_RETURN 13464) (T_STATIC 13465) (T_STRING 13466) (T_SWITCH 13467) (T_THROW 13468) (T_TRAIT 13469) (T_TRAIT_C 13470) (T_TRY 13471) (T_UNSET 13472) (T_USE 13473) (T_VAR 13474) (T_VARIABLE 13268) (T_WHILE 13475) (T_YIELD 13476) (identifier 13477) (member_name 13478) (reserved_non_modifiers 13479) (semi_reserved 13480) (simple_variable 13481) ("{" 914)) 5850 (("(" 13482) (argument_list 13483)) 5851 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 13484) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 13485) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 13486) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5852 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 13489) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5853 (("," 1449) (possible_comma 13487)) 5854 ((")" 13488)) 5855 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (ampersand 13519) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13520) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5856 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13518) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5857 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13517) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5858 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13516) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5859 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13515) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5860 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13514) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5861 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13513) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5862 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13512) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5863 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13511) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5864 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13510) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5865 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13509) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5866 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13508) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5867 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13507) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5868 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13506) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5869 (("\"" 1489) ("$" 13521) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 13522) ("[" 1508) (array_object_dereferenceable 13523) (callable_expr 13524) (callable_variable 13525) (class_constant 1512) (class_name 13526) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 13527) (name 13528) (simple_variable 13529) (static_member 13530) (variable 13531) (variable_class_name 13532)) 5870 (("$" 13521) (T_VARIABLE 13522) (simple_variable 13567) ("{" 13568)) 5871 ((T_NULLSAFE_OBJECT_OPERATOR 13550) (T_OBJECT_OPERATOR 13551) ("[" 13552) ("{" 13553)) 5872 (("(" 13536) (argument_list 13549)) 5873 ((T_PAAMAYIM_NEKUDOTAYIM 13545)) 5874 (("(" 13536) (argument_list 13544)) 5875 ((T_PAAMAYIM_NEKUDOTAYIM 13533)) 5876 (("$" 13521) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 13522) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 13534) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 13535) ("{" 914)) 5877 (("(" 13536) (argument_list 13537)) 5878 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 13538) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 13539) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 13540) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5879 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 13543) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5880 (("," 1449) (possible_comma 13541)) 5881 ((")" 13542)) 5882 (("$" 13521) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 13522) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 13546) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 13547) ("{" 914)) 5883 (("(" 13536) (argument_list 13548)) 5884 (("$" 13521) (T_STRING 13558) (T_VARIABLE 13522) (property_name 13565) (simple_variable 13560) ("{" 13561)) 5885 (("$" 13521) (T_STRING 13558) (T_VARIABLE 13522) (property_name 13559) (simple_variable 13560) ("{" 13561)) 5886 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 13556) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5887 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13554) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5888 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13555)) 5889 (("]" 13557)) 5890 (("(" 13536) (argument_list 13564)) 5891 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13562) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5892 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13563)) 5893 (("(" 13536) (argument_list 13566)) 5894 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13569) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5895 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13570)) 5896 ((backup_doc_comment 13573)) 5897 (("(" 13574)) 5898 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 13575)) 5899 ((")" 13576)) 5900 ((T_USE 148) (lexical_vars 13577)) 5901 ((":" 150) (return_type 13578)) 5902 ((backup_fn_flags 13579)) 5903 (("{" 13580)) 5904 ((inner_statement_list 13581)) 5905 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 13582) ("~" 124)) 5906 ((backup_fn_flags 13583)) 5907 ((backup_doc_comment 13585)) 5908 (("(" 13586)) 5909 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 13587)) 5910 ((")" 13588)) 5911 ((":" 2207) (return_type 13589)) 5912 ((T_DOUBLE_ARROW 13590)) 5913 ((backup_fn_flags 13591)) 5914 ((backup_lex_pos 13592)) 5915 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13593) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5916 (("%" 13297) ("*" 13298) ("+" 13299) ("-" 13300) ("." 13301) ("/" 13302) ("<" 13303) (">" 13304) ("?" 13305) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13306) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13307) (T_BOOLEAN_AND 13308) (T_BOOLEAN_OR 13309) (T_COALESCE 13310) (T_INSTANCEOF 13311) (T_IS_EQUAL 13312) (T_IS_GREATER_OR_EQUAL 13313) (T_IS_IDENTICAL 13314) (T_IS_NOT_EQUAL 13315) (T_IS_NOT_IDENTICAL 13316) (T_IS_SMALLER_OR_EQUAL 13317) (T_LOGICAL_AND 13318) (T_LOGICAL_OR 13319) (T_LOGICAL_XOR 13320) (T_POW 13321) (T_SL 13322) (T_SPACESHIP 13323) (T_SR 13324) ("^" 13325) (backup_fn_flags 13594) ("|" 13326)) 5917 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13596) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 5918 (("$" 13222) (T_ABSTRACT 13399) (T_ARRAY 13400) (T_AS 13401) (T_BREAK 13402) (T_CALLABLE 13403) (T_CASE 13404) (T_CATCH 13405) (T_CLASS 13406) (T_CLASS_C 13407) (T_CLONE 13408) (T_CONST 13409) (T_CONTINUE 13410) (T_DECLARE 13411) (T_DEFAULT 13412) (T_DIR 13413) (T_DO 13414) (T_ECHO 13415) (T_ELSE 13416) (T_ELSEIF 13417) (T_EMPTY 13418) (T_ENDDECLARE 13419) (T_ENDFOR 13420) (T_ENDFOREACH 13421) (T_ENDIF 13422) (T_ENDSWITCH 13423) (T_ENDWHILE 13424) (T_ENUM 13425) (T_EVAL 13426) (T_EXIT 13427) (T_EXTENDS 13428) (T_FILE 13429) (T_FINAL 13430) (T_FINALLY 13431) (T_FN 13432) (T_FOR 13433) (T_FOREACH 13434) (T_FUNCTION 13435) (T_FUNC_C 13436) (T_GLOBAL 13437) (T_GOTO 13438) (T_IF 13439) (T_IMPLEMENTS 13440) (T_INCLUDE 13441) (T_INCLUDE_ONCE 13442) (T_INSTANCEOF 13443) (T_INSTEADOF 13444) (T_INTERFACE 13445) (T_ISSET 13446) (T_LINE 13447) (T_LIST 13448) (T_LOGICAL_AND 13449) (T_LOGICAL_OR 13450) (T_LOGICAL_XOR 13451) (T_MATCH 13452) (T_METHOD_C 13453) (T_NAMESPACE 13454) (T_NEW 13455) (T_NS_C 13456) (T_PRINT 13457) (T_PRIVATE 13458) (T_PROTECTED 13459) (T_PUBLIC 13460) (T_READONLY 13461) (T_REQUIRE 13462) (T_REQUIRE_ONCE 13463) (T_RETURN 13464) (T_STATIC 13465) (T_STRING 13466) (T_SWITCH 13467) (T_THROW 13468) (T_TRAIT 13469) (T_TRAIT_C 13470) (T_TRY 13471) (T_UNSET 13472) (T_USE 13473) (T_VAR 13474) (T_VARIABLE 13268) (T_WHILE 13475) (T_YIELD 13476) (identifier 13598) (member_name 13599) (reserved_non_modifiers 13479) (semi_reserved 13480) (simple_variable 13600) ("{" 914)) 5919 (("(" 13482) (argument_list 13601)) 5920 ((T_FN 35) (T_FUNCTION 126) (fn 13282) #'13283 (inline_function 13605)) 5921 (("$" 13222) (T_STRING 13614) (T_VARIABLE 13268) (property_name 13621) (simple_variable 13616) ("{" 13617)) 5922 (("$" 13222) (T_STRING 13614) (T_VARIABLE 13268) (property_name 13615) (simple_variable 13616) ("{" 13617)) 5923 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 13612) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5924 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13610) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5925 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13611)) 5926 (("]" 13613)) 5927 (("(" 13482) (argument_list 13620)) 5928 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13618) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5929 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13619)) 5930 (("(" 13482) (argument_list 13622)) 5931 (("`" 13624)) 5932 (("]" 13626)) 5933 (("=" 13627)) 5934 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13628) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5935 (("%" 13297) ("*" 13298) ("+" 13299) ("-" 13300) ("." 13301) ("/" 13302) ("<" 13303) (">" 13304) ("?" 13305) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13306) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13307) (T_BOOLEAN_AND 13308) (T_BOOLEAN_OR 13309) (T_COALESCE 13310) (T_DOUBLE_ARROW 13631) (T_INSTANCEOF 13311) (T_IS_EQUAL 13312) (T_IS_GREATER_OR_EQUAL 13313) (T_IS_IDENTICAL 13314) (T_IS_NOT_EQUAL 13315) (T_IS_NOT_IDENTICAL 13316) (T_IS_SMALLER_OR_EQUAL 13317) (T_LOGICAL_AND 13318) (T_LOGICAL_OR 13319) (T_LOGICAL_XOR 13320) (T_POW 13321) (T_SL 13322) (T_SPACESHIP 13323) (T_SR 13324) ("^" 13325) ("|" 13326)) 5936 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13632) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5937 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 13641) (T_VARIABLE 2822) (encaps_var 2836)) 5938 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 13640) (T_VARIABLE 2822) (encaps_var 2827)) 5939 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 13691) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5940 (("(" 2899) (argument_list 2900) (ctor_arguments 13684)) 5941 ((T_ATTRIBUTE 2853) (T_CLASS 13647) (anonymous_class 13683) (attribute 2898)) 5942 ((T_PAAMAYIM_NEKUDOTAYIM 13681)) 5943 (("(" 13672) (argument_list 13673) (ctor_arguments 13674)) 5944 ((T_NULLSAFE_OBJECT_OPERATOR 13660) (T_OBJECT_OPERATOR 13661) (T_PAAMAYIM_NEKUDOTAYIM 13662) ("[" 13663) ("{" 13664)) 5945 (("$" 13521) (T_STRING 13558) (T_VARIABLE 13522) (property_name 13671) (simple_variable 13560) ("{" 13561)) 5946 (("$" 13521) (T_STRING 13558) (T_VARIABLE 13522) (property_name 13670) (simple_variable 13560) ("{" 13561)) 5947 (("$" 13521) (T_VARIABLE 13522) (simple_variable 13669)) 5948 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 13667) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 5949 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13665) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5950 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13666)) 5951 (("]" 13668)) 5952 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 13675) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 13676) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 13677) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5953 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 13680) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5954 (("," 1449) (possible_comma 13678)) 5955 ((")" 13679)) 5956 (("$" 13521) (T_VARIABLE 13522) (simple_variable 13682)) 5957 ((T_EXTENDS 2902) (extends_from 13685)) 5958 ((T_IMPLEMENTS 2904) (implements_list 13686)) 5959 ((backup_doc_comment 13687)) 5960 (("{" 13688)) 5961 ((class_statement_list 13689)) 5962 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 13690)) 5963 (("%" 280) (")" 13692) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5964 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 13694) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5965 (("%" 280) (")" 13695) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5966 (("{" 13696)) 5967 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 13697) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 5968 (("}" 13698)) 5969 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 13700) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 5970 ((")" 13701)) 5971 (("=" 13702)) 5972 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 13703) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 5973 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 13705) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 5974 (("," 5117) (possible_comma 13706)) 5975 ((")" 13707)) 5976 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 13714) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5977 ((")" 13715)) 5978 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 13717) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5979 (("%" 280) (")" 13718) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5980 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 13720) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5981 (("%" 280) (")" 13721) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5982 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 13728) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 5983 ((")" 13729)) 5984 (("%" 280) (")" 13734) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 5985 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13737) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 5986 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13738)) 5987 (("\"" 13740) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 5988 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 13750) (T_VARIABLE 2822) (encaps_var 2836)) 5989 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 13749) (T_VARIABLE 2822) (encaps_var 2827)) 5990 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 13800) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 5991 (("(" 2899) (argument_list 2900) (ctor_arguments 13793)) 5992 ((T_ATTRIBUTE 2853) (T_CLASS 13756) (anonymous_class 13792) (attribute 2898)) 5993 ((T_PAAMAYIM_NEKUDOTAYIM 13790)) 5994 (("(" 13781) (argument_list 13782) (ctor_arguments 13783)) 5995 ((T_NULLSAFE_OBJECT_OPERATOR 13769) (T_OBJECT_OPERATOR 13770) (T_PAAMAYIM_NEKUDOTAYIM 13771) ("[" 13772) ("{" 13773)) 5996 (("$" 13112) (T_STRING 13149) (T_VARIABLE 13113) (property_name 13780) (simple_variable 13151) ("{" 13152)) 5997 (("$" 13112) (T_STRING 13149) (T_VARIABLE 13113) (property_name 13779) (simple_variable 13151) ("{" 13152)) 5998 (("$" 13112) (T_VARIABLE 13113) (simple_variable 13778)) 5999 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 13776) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6000 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13774) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6001 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13775)) 6002 (("]" 13777)) 6003 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 13784) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 13785) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 13786) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6004 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 13789) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6005 (("," 1449) (possible_comma 13787)) 6006 ((")" 13788)) 6007 (("$" 13112) (T_VARIABLE 13113) (simple_variable 13791)) 6008 ((T_EXTENDS 2902) (extends_from 13794)) 6009 ((T_IMPLEMENTS 2904) (implements_list 13795)) 6010 ((backup_doc_comment 13796)) 6011 (("{" 13797)) 6012 ((class_statement_list 13798)) 6013 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 13799)) 6014 (("%" 280) (")" 13801) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6015 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 13803) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6016 (("%" 280) (")" 13804) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6017 (("{" 13805)) 6018 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 13806) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 6019 (("}" 13807)) 6020 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 13809) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 6021 ((")" 13810)) 6022 (("=" 13811)) 6023 (("!" 12811) ("\"" 12812) ("$" 12813) ("(" 12814) ("+" 12815) ("-" 12816) ("@" 12817) (T_ARRAY 12818) (T_ARRAY_CAST 12819) (T_ATTRIBUTE 125) (T_BOOL_CAST 12820) (T_CLASS_C 12821) (T_CLONE 12822) (T_CONSTANT_ENCAPSED_STRING 12823) (T_DEC 12824) (T_DIR 12825) (T_DNUMBER 12826) (T_DOUBLE_CAST 12827) (T_EMPTY 12828) (T_EVAL 12829) (T_EXIT 12830) (T_FILE 12831) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 12832) (T_INC 12833) (T_INCLUDE 12834) (T_INCLUDE_ONCE 12835) (T_INT_CAST 12836) (T_ISSET 12837) (T_LINE 12838) (T_LIST 12839) (T_LNUMBER 12840) (T_MATCH 12841) (T_METHOD_C 12842) (T_NAME_FULLY_QUALIFIED 12843) (T_NAME_QUALIFIED 12844) (T_NAME_RELATIVE 12845) (T_NEW 12846) (T_NS_C 12847) (T_OBJECT_CAST 12848) (T_PRINT 12849) (T_REQUIRE 12850) (T_REQUIRE_ONCE 12851) (T_START_HEREDOC 12852) (T_STATIC 12853) (T_STRING 12854) (T_STRING_CAST 12855) (T_THROW 12856) (T_TRAIT_C 12857) (T_UNSET_CAST 12858) (T_VARIABLE 12859) (T_YIELD 12860) (T_YIELD_FROM 12861) ("[" 12862) ("`" 12863) (array_object_dereferenceable 12864) (attribute 129) (attributes 12865) (callable_expr 12866) (callable_variable 12867) (class_constant 12868) (class_name 12869) (constant 12870) (dereferenceable_scalar 12871) (expr 13812) (fn 12873) (fully_dereferenceable 103) #'12874 (function_call 12875) (inline_function 12876) (internal_functions_in_yacc 12877) (match 12878) (name 12879) (new_expr 12880) (scalar 12881) (simple_variable 12882) (static_member 12883) (variable 12884) (variable_class_name 12885) ("~" 12886)) 6024 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 13814) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6025 (("," 5117) (possible_comma 13815)) 6026 ((")" 13816)) 6027 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 13823) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6028 ((")" 13824)) 6029 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 13826) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6030 (("%" 280) (")" 13827) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6031 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 13829) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6032 (("%" 280) (")" 13830) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6033 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 13837) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 6034 ((")" 13838)) 6035 (("%" 280) (")" 13843) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6036 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13846) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6037 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13847)) 6038 (("\"" 13849) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 6039 (("$" 345) (T_ABSTRACT 12617) (T_ARRAY 12618) (T_AS 12619) (T_BREAK 12620) (T_CALLABLE 12621) (T_CASE 12622) (T_CATCH 12623) (T_CLASS 12624) (T_CLASS_C 12625) (T_CLONE 12626) (T_CONST 12627) (T_CONTINUE 12628) (T_DECLARE 12629) (T_DEFAULT 12630) (T_DIR 12631) (T_DO 12632) (T_ECHO 12633) (T_ELSE 12634) (T_ELSEIF 12635) (T_EMPTY 12636) (T_ENDDECLARE 12637) (T_ENDFOR 12638) (T_ENDFOREACH 12639) (T_ENDIF 12640) (T_ENDSWITCH 12641) (T_ENDWHILE 12642) (T_ENUM 12643) (T_EVAL 12644) (T_EXIT 12645) (T_EXTENDS 12646) (T_FILE 12647) (T_FINAL 12648) (T_FINALLY 12649) (T_FN 12650) (T_FOR 12651) (T_FOREACH 12652) (T_FUNCTION 12653) (T_FUNC_C 12654) (T_GLOBAL 12655) (T_GOTO 12656) (T_IF 12657) (T_IMPLEMENTS 12658) (T_INCLUDE 12659) (T_INCLUDE_ONCE 12660) (T_INSTANCEOF 12661) (T_INSTEADOF 12662) (T_INTERFACE 12663) (T_ISSET 12664) (T_LINE 12665) (T_LIST 12666) (T_LOGICAL_AND 12667) (T_LOGICAL_OR 12668) (T_LOGICAL_XOR 12669) (T_MATCH 12670) (T_METHOD_C 12671) (T_NAMESPACE 12672) (T_NEW 12673) (T_NS_C 12674) (T_PRINT 12675) (T_PRIVATE 12676) (T_PROTECTED 12677) (T_PUBLIC 12678) (T_READONLY 12679) (T_REQUIRE 12680) (T_REQUIRE_ONCE 12681) (T_RETURN 12682) (T_STATIC 12683) (T_STRING 12684) (T_SWITCH 12685) (T_THROW 12686) (T_TRAIT 12687) (T_TRAIT_C 12688) (T_TRY 12689) (T_UNSET 12690) (T_USE 12691) (T_VAR 12692) (T_VARIABLE 391) (T_WHILE 12693) (T_YIELD 12694) (identifier 13853) (member_name 13854) (reserved_non_modifiers 12697) (semi_reserved 12698) (simple_variable 13855) ("{" 914)) 6040 (("(" 12700) (argument_list 13856)) 6041 ((T_FN 35) (T_FUNCTION 126) (fn 405) #'406 (inline_function 13860)) 6042 (("$" 345) (T_STRING 13869) (T_VARIABLE 391) (property_name 13876) (simple_variable 13871) ("{" 13872)) 6043 (("$" 345) (T_STRING 13869) (T_VARIABLE 391) (property_name 13870) (simple_variable 13871) ("{" 13872)) 6044 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 13867) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6045 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13865) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6046 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13866)) 6047 (("]" 13868)) 6048 (("(" 12700) (argument_list 13875)) 6049 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13873) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6050 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 13874)) 6051 (("(" 12700) (argument_list 13877)) 6052 (("`" 13879)) 6053 (("]" 13881)) 6054 (("=" 13882)) 6055 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 13883) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6056 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14380) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6057 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 14378) (encaps_var 5158)) 6058 (("$" 13887) (T_VARIABLE 13933) (simple_variable 14374) ("{" 14375)) 6059 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14372) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6060 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14371) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6061 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14370) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6062 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14369) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6063 (("(" 14366)) 6064 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14365) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6065 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14364) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6066 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14363) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6067 (("\"" 1489) ("$" 14186) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 14187) ("[" 1508) (array_object_dereferenceable 14188) (callable_expr 14189) (callable_variable 14190) (class_constant 1512) (class_name 14191) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 14192) (name 14193) (simple_variable 14194) (static_member 14195) (variable 14362) (variable_class_name 14197)) 6068 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14361) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6069 (("(" 14358)) 6070 (("(" 14355)) 6071 (("(" 14351) (exit_expr 14352)) 6072 (("\"" 1489) ("$" 14186) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 14187) ("[" 1508) (array_object_dereferenceable 14188) (callable_expr 14189) (callable_variable 14190) (class_constant 1512) (class_name 14191) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 14192) (name 14193) (simple_variable 14194) (static_member 14195) (variable 14350) (variable_class_name 14197)) 6073 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14349) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6074 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14348) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6075 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14347) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6076 (("(" 14343)) 6077 (("(" 14338)) 6078 (("(" 14332)) 6079 (("$" 13521) ("(" 13646) (T_ATTRIBUTE 2853) (T_CLASS 14311) (T_NAME_FULLY_QUALIFIED 13648) (T_NAME_QUALIFIED 13649) (T_NAME_RELATIVE 13650) (T_STATIC 13651) (T_STRING 13652) (T_VARIABLE 13522) (anonymous_class 14312) (attribute 2862) (attributes 14313) (class_name 13655) (class_name_reference 14314) (name 13657) (new_variable 13658) (simple_variable 13659)) 6080 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14310) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6081 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14309) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6082 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14308) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6083 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14307) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6084 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 14302) (T_END_HEREDOC 14303) (T_VARIABLE 2822) (encaps_list 14304) (encaps_var 2824)) 6085 ((T_FN 35) (T_FUNCTION 126) (fn 13947) #'13948 (inline_function 14301)) 6086 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14300) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6087 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14299) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6088 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14298) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6089 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14295) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6090 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14294) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6091 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 14290) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 6092 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 14288) (encaps_list 2729) (encaps_var 2730)) 6093 ((T_NULLSAFE_OBJECT_OPERATOR 14271) (T_OBJECT_OPERATOR 14272) ("[" 14273) ("{" 14274)) 6094 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 14268) (attribute 2654) (fn 13947) #'13948 (inline_function 14269)) 6095 (("(" 14147) (argument_list 14267)) 6096 ((T_PAAMAYIM_NEKUDOTAYIM 14262)) 6097 (("%" 13962) ("*" 13963) ("+" 13964) ("-" 13965) ("." 13966) ("/" 13967) ("<" 13968) (">" 13969) ("?" 13970) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13971) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13972) (T_BOOLEAN_AND 13973) (T_BOOLEAN_OR 13974) (T_COALESCE 13975) (T_DOUBLE_ARROW 14260) (T_INSTANCEOF 13976) (T_IS_EQUAL 13977) (T_IS_GREATER_OR_EQUAL 13978) (T_IS_IDENTICAL 13979) (T_IS_NOT_EQUAL 13980) (T_IS_NOT_IDENTICAL 13981) (T_IS_SMALLER_OR_EQUAL 13982) (T_LOGICAL_AND 13983) (T_LOGICAL_OR 13984) (T_LOGICAL_XOR 13985) (T_POW 13986) (T_SL 13987) (T_SPACESHIP 13988) (T_SR 13989) ("^" 13990) ("|" 13991)) 6098 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 14249)) 6099 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 14237)) 6100 (("(" 14147) (argument_list 14236)) 6101 (("=" 14155) (T_AND_EQUAL 14156) (T_COALESCE_EQUAL 14157) (T_CONCAT_EQUAL 14158) (T_DEC 14159) (T_DIV_EQUAL 14160) (T_INC 14161) (T_MINUS_EQUAL 14162) (T_MOD_EQUAL 14163) (T_MUL_EQUAL 14164) (T_OR_EQUAL 14165) (T_PLUS_EQUAL 14166) (T_POW_EQUAL 14167) (T_SL_EQUAL 14168) (T_SR_EQUAL 14169) (T_XOR_EQUAL 14170)) 6102 ((T_PAAMAYIM_NEKUDOTAYIM 14063)) 6103 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 13961) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6104 (("%" 13962) ("*" 13963) ("+" 13964) ("-" 13965) ("." 13966) ("/" 13967) ("<" 13968) (">" 13969) ("?" 13970) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13971) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13972) (T_BOOLEAN_AND 13973) (T_BOOLEAN_OR 13974) (T_COALESCE 13975) (T_INSTANCEOF 13976) (T_IS_EQUAL 13977) (T_IS_GREATER_OR_EQUAL 13978) (T_IS_IDENTICAL 13979) (T_IS_NOT_EQUAL 13980) (T_IS_NOT_IDENTICAL 13981) (T_IS_SMALLER_OR_EQUAL 13982) (T_LOGICAL_AND 13983) (T_LOGICAL_OR 13984) (T_LOGICAL_XOR 13985) (T_POW 13986) (T_SL 13987) (T_SPACESHIP 13988) (T_SR 13989) ("^" 13990) ("|" 13991)) 6105 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14062) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6106 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14061) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6107 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14060) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6108 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14059) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6109 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14058) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6110 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14057) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6111 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14056) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6112 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14055) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6113 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 14050) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 14051) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 6114 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14049) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6115 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14048) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6116 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14047) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6117 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14046) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6118 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14045) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6119 (("$" 14007) ("(" 14008) (T_NAME_FULLY_QUALIFIED 14009) (T_NAME_QUALIFIED 14010) (T_NAME_RELATIVE 14011) (T_STATIC 14012) (T_STRING 14013) (T_VARIABLE 14014) (class_name 14015) (class_name_reference 14016) (name 14017) (new_variable 14018) (simple_variable 14019)) 6120 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14006) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6121 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14005) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6122 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14004) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6123 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14003) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6124 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14002) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6125 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14001) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6126 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14000) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6127 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 13999) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6128 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 13998) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6129 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 13997) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6130 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 13996) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6131 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 13995) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6132 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 13994) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6133 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 13993) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6134 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 13992) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6135 (("$" 14007) (T_VARIABLE 14014) (simple_variable 14041) ("{" 14042)) 6136 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14039) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6137 ((T_PAAMAYIM_NEKUDOTAYIM 14037)) 6138 ((T_NULLSAFE_OBJECT_OPERATOR 14020) (T_OBJECT_OPERATOR 14021) (T_PAAMAYIM_NEKUDOTAYIM 14022) ("[" 14023) ("{" 14024)) 6139 (("$" 14007) (T_STRING 14030) (T_VARIABLE 14014) (property_name 14036) (simple_variable 14032) ("{" 14033)) 6140 (("$" 14007) (T_STRING 14030) (T_VARIABLE 14014) (property_name 14031) (simple_variable 14032) ("{" 14033)) 6141 (("$" 14007) (T_VARIABLE 14014) (simple_variable 14029)) 6142 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 14027) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6143 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14025) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6144 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14026)) 6145 (("]" 14028)) 6146 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14034) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6147 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14035)) 6148 (("$" 14007) (T_VARIABLE 14014) (simple_variable 14038)) 6149 (("%" 280) (")" 14040) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6150 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14043) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6151 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14044)) 6152 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14054) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6153 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 14052) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 6154 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14053) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6155 (("$" 13887) (T_ABSTRACT 14064) (T_ARRAY 14065) (T_AS 14066) (T_BREAK 14067) (T_CALLABLE 14068) (T_CASE 14069) (T_CATCH 14070) (T_CLASS 14071) (T_CLASS_C 14072) (T_CLONE 14073) (T_CONST 14074) (T_CONTINUE 14075) (T_DECLARE 14076) (T_DEFAULT 14077) (T_DIR 14078) (T_DO 14079) (T_ECHO 14080) (T_ELSE 14081) (T_ELSEIF 14082) (T_EMPTY 14083) (T_ENDDECLARE 14084) (T_ENDFOR 14085) (T_ENDFOREACH 14086) (T_ENDIF 14087) (T_ENDSWITCH 14088) (T_ENDWHILE 14089) (T_ENUM 14090) (T_EVAL 14091) (T_EXIT 14092) (T_EXTENDS 14093) (T_FILE 14094) (T_FINAL 14095) (T_FINALLY 14096) (T_FN 14097) (T_FOR 14098) (T_FOREACH 14099) (T_FUNCTION 14100) (T_FUNC_C 14101) (T_GLOBAL 14102) (T_GOTO 14103) (T_IF 14104) (T_IMPLEMENTS 14105) (T_INCLUDE 14106) (T_INCLUDE_ONCE 14107) (T_INSTANCEOF 14108) (T_INSTEADOF 14109) (T_INTERFACE 14110) (T_ISSET 14111) (T_LINE 14112) (T_LIST 14113) (T_LOGICAL_AND 14114) (T_LOGICAL_OR 14115) (T_LOGICAL_XOR 14116) (T_MATCH 14117) (T_METHOD_C 14118) (T_NAMESPACE 14119) (T_NEW 14120) (T_NS_C 14121) (T_PRINT 14122) (T_PRIVATE 14123) (T_PROTECTED 14124) (T_PUBLIC 14125) (T_READONLY 14126) (T_REQUIRE 14127) (T_REQUIRE_ONCE 14128) (T_RETURN 14129) (T_STATIC 14130) (T_STRING 14131) (T_SWITCH 14132) (T_THROW 14133) (T_TRAIT 14134) (T_TRAIT_C 14135) (T_TRY 14136) (T_UNSET 14137) (T_USE 14138) (T_VAR 14139) (T_VARIABLE 13933) (T_WHILE 14140) (T_YIELD 14141) (identifier 14142) (member_name 14143) (reserved_non_modifiers 14144) (semi_reserved 14145) (simple_variable 14146) ("{" 914)) 6156 (("(" 14147) (argument_list 14148)) 6157 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14149) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 14150) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 14151) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6158 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14154) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6159 (("," 1449) (possible_comma 14152)) 6160 ((")" 14153)) 6161 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (ampersand 14184) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14185) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6162 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14183) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6163 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14182) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6164 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14181) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6165 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14180) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6166 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14179) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6167 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14178) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6168 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14177) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6169 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14176) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6170 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14175) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6171 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14174) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6172 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14173) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6173 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14172) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6174 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14171) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6175 (("\"" 1489) ("$" 14186) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 14187) ("[" 1508) (array_object_dereferenceable 14188) (callable_expr 14189) (callable_variable 14190) (class_constant 1512) (class_name 14191) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 14192) (name 14193) (simple_variable 14194) (static_member 14195) (variable 14196) (variable_class_name 14197)) 6176 (("$" 14186) (T_VARIABLE 14187) (simple_variable 14232) ("{" 14233)) 6177 ((T_NULLSAFE_OBJECT_OPERATOR 14215) (T_OBJECT_OPERATOR 14216) ("[" 14217) ("{" 14218)) 6178 (("(" 14201) (argument_list 14214)) 6179 ((T_PAAMAYIM_NEKUDOTAYIM 14210)) 6180 (("(" 14201) (argument_list 14209)) 6181 ((T_PAAMAYIM_NEKUDOTAYIM 14198)) 6182 (("$" 14186) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 14187) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 14199) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 14200) ("{" 914)) 6183 (("(" 14201) (argument_list 14202)) 6184 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14203) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 14204) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 14205) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6185 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14208) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6186 (("," 1449) (possible_comma 14206)) 6187 ((")" 14207)) 6188 (("$" 14186) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 14187) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 14211) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 14212) ("{" 914)) 6189 (("(" 14201) (argument_list 14213)) 6190 (("$" 14186) (T_STRING 14223) (T_VARIABLE 14187) (property_name 14230) (simple_variable 14225) ("{" 14226)) 6191 (("$" 14186) (T_STRING 14223) (T_VARIABLE 14187) (property_name 14224) (simple_variable 14225) ("{" 14226)) 6192 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 14221) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6193 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14219) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6194 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14220)) 6195 (("]" 14222)) 6196 (("(" 14201) (argument_list 14229)) 6197 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14227) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6198 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14228)) 6199 (("(" 14201) (argument_list 14231)) 6200 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14234) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6201 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14235)) 6202 ((backup_doc_comment 14238)) 6203 (("(" 14239)) 6204 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 14240)) 6205 ((")" 14241)) 6206 ((T_USE 148) (lexical_vars 14242)) 6207 ((":" 150) (return_type 14243)) 6208 ((backup_fn_flags 14244)) 6209 (("{" 14245)) 6210 ((inner_statement_list 14246)) 6211 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 14247) ("~" 124)) 6212 ((backup_fn_flags 14248)) 6213 ((backup_doc_comment 14250)) 6214 (("(" 14251)) 6215 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 14252)) 6216 ((")" 14253)) 6217 ((":" 2207) (return_type 14254)) 6218 ((T_DOUBLE_ARROW 14255)) 6219 ((backup_fn_flags 14256)) 6220 ((backup_lex_pos 14257)) 6221 (("!" 13220) ("\"" 13221) ("$" 13222) ("(" 13223) ("+" 13224) ("-" 13225) ("@" 13226) (T_ARRAY 13227) (T_ARRAY_CAST 13228) (T_ATTRIBUTE 125) (T_BOOL_CAST 13229) (T_CLASS_C 13230) (T_CLONE 13231) (T_CONSTANT_ENCAPSED_STRING 13232) (T_DEC 13233) (T_DIR 13234) (T_DNUMBER 13235) (T_DOUBLE_CAST 13236) (T_EMPTY 13237) (T_EVAL 13238) (T_EXIT 13239) (T_FILE 13240) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13241) (T_INC 13242) (T_INCLUDE 13243) (T_INCLUDE_ONCE 13244) (T_INT_CAST 13245) (T_ISSET 13246) (T_LINE 13247) (T_LIST 13248) (T_LNUMBER 13249) (T_MATCH 13250) (T_METHOD_C 13251) (T_NAME_FULLY_QUALIFIED 13252) (T_NAME_QUALIFIED 13253) (T_NAME_RELATIVE 13254) (T_NEW 13255) (T_NS_C 13256) (T_OBJECT_CAST 13257) (T_PRINT 13258) (T_REQUIRE 13259) (T_REQUIRE_ONCE 13260) (T_START_HEREDOC 13261) (T_STATIC 13262) (T_STRING 13263) (T_STRING_CAST 13264) (T_THROW 13265) (T_TRAIT_C 13266) (T_UNSET_CAST 13267) (T_VARIABLE 13268) (T_YIELD 13269) (T_YIELD_FROM 13270) ("[" 13271) ("`" 13272) (array_object_dereferenceable 13273) (attribute 129) (attributes 13274) (callable_expr 13275) (callable_variable 13276) (class_constant 13277) (class_name 13278) (constant 13279) (dereferenceable_scalar 13280) (expr 14258) (fn 13282) (fully_dereferenceable 103) #'13283 (function_call 13284) (inline_function 13285) (internal_functions_in_yacc 13286) (match 13287) (name 13288) (new_expr 13289) (scalar 13290) (simple_variable 13291) (static_member 13292) (variable 13293) (variable_class_name 13294) ("~" 13295)) 6222 (("%" 13297) ("*" 13298) ("+" 13299) ("-" 13300) ("." 13301) ("/" 13302) ("<" 13303) (">" 13304) ("?" 13305) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13306) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13307) (T_BOOLEAN_AND 13308) (T_BOOLEAN_OR 13309) (T_COALESCE 13310) (T_INSTANCEOF 13311) (T_IS_EQUAL 13312) (T_IS_GREATER_OR_EQUAL 13313) (T_IS_IDENTICAL 13314) (T_IS_NOT_EQUAL 13315) (T_IS_NOT_IDENTICAL 13316) (T_IS_SMALLER_OR_EQUAL 13317) (T_LOGICAL_AND 13318) (T_LOGICAL_OR 13319) (T_LOGICAL_XOR 13320) (T_POW 13321) (T_SL 13322) (T_SPACESHIP 13323) (T_SR 13324) ("^" 13325) (backup_fn_flags 14259) ("|" 13326)) 6223 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14261) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6224 (("$" 13887) (T_ABSTRACT 14064) (T_ARRAY 14065) (T_AS 14066) (T_BREAK 14067) (T_CALLABLE 14068) (T_CASE 14069) (T_CATCH 14070) (T_CLASS 14071) (T_CLASS_C 14072) (T_CLONE 14073) (T_CONST 14074) (T_CONTINUE 14075) (T_DECLARE 14076) (T_DEFAULT 14077) (T_DIR 14078) (T_DO 14079) (T_ECHO 14080) (T_ELSE 14081) (T_ELSEIF 14082) (T_EMPTY 14083) (T_ENDDECLARE 14084) (T_ENDFOR 14085) (T_ENDFOREACH 14086) (T_ENDIF 14087) (T_ENDSWITCH 14088) (T_ENDWHILE 14089) (T_ENUM 14090) (T_EVAL 14091) (T_EXIT 14092) (T_EXTENDS 14093) (T_FILE 14094) (T_FINAL 14095) (T_FINALLY 14096) (T_FN 14097) (T_FOR 14098) (T_FOREACH 14099) (T_FUNCTION 14100) (T_FUNC_C 14101) (T_GLOBAL 14102) (T_GOTO 14103) (T_IF 14104) (T_IMPLEMENTS 14105) (T_INCLUDE 14106) (T_INCLUDE_ONCE 14107) (T_INSTANCEOF 14108) (T_INSTEADOF 14109) (T_INTERFACE 14110) (T_ISSET 14111) (T_LINE 14112) (T_LIST 14113) (T_LOGICAL_AND 14114) (T_LOGICAL_OR 14115) (T_LOGICAL_XOR 14116) (T_MATCH 14117) (T_METHOD_C 14118) (T_NAMESPACE 14119) (T_NEW 14120) (T_NS_C 14121) (T_PRINT 14122) (T_PRIVATE 14123) (T_PROTECTED 14124) (T_PUBLIC 14125) (T_READONLY 14126) (T_REQUIRE 14127) (T_REQUIRE_ONCE 14128) (T_RETURN 14129) (T_STATIC 14130) (T_STRING 14131) (T_SWITCH 14132) (T_THROW 14133) (T_TRAIT 14134) (T_TRAIT_C 14135) (T_TRY 14136) (T_UNSET 14137) (T_USE 14138) (T_VAR 14139) (T_VARIABLE 13933) (T_WHILE 14140) (T_YIELD 14141) (identifier 14263) (member_name 14264) (reserved_non_modifiers 14144) (semi_reserved 14145) (simple_variable 14265) ("{" 914)) 6225 (("(" 14147) (argument_list 14266)) 6226 ((T_FN 35) (T_FUNCTION 126) (fn 13947) #'13948 (inline_function 14270)) 6227 (("$" 13887) (T_STRING 14279) (T_VARIABLE 13933) (property_name 14286) (simple_variable 14281) ("{" 14282)) 6228 (("$" 13887) (T_STRING 14279) (T_VARIABLE 13933) (property_name 14280) (simple_variable 14281) ("{" 14282)) 6229 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 14277) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6230 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14275) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6231 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14276)) 6232 (("]" 14278)) 6233 (("(" 14147) (argument_list 14285)) 6234 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14283) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6235 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14284)) 6236 (("(" 14147) (argument_list 14287)) 6237 (("`" 14289)) 6238 (("]" 14291)) 6239 (("=" 14292)) 6240 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14293) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6241 (("%" 13962) ("*" 13963) ("+" 13964) ("-" 13965) ("." 13966) ("/" 13967) ("<" 13968) (">" 13969) ("?" 13970) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13971) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13972) (T_BOOLEAN_AND 13973) (T_BOOLEAN_OR 13974) (T_COALESCE 13975) (T_DOUBLE_ARROW 14296) (T_INSTANCEOF 13976) (T_IS_EQUAL 13977) (T_IS_GREATER_OR_EQUAL 13978) (T_IS_IDENTICAL 13979) (T_IS_NOT_EQUAL 13980) (T_IS_NOT_IDENTICAL 13981) (T_IS_SMALLER_OR_EQUAL 13982) (T_LOGICAL_AND 13983) (T_LOGICAL_OR 13984) (T_LOGICAL_XOR 13985) (T_POW 13986) (T_SL 13987) (T_SPACESHIP 13988) (T_SR 13989) ("^" 13990) ("|" 13991)) 6242 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14297) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6243 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 14306) (T_VARIABLE 2822) (encaps_var 2836)) 6244 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 14305) (T_VARIABLE 2822) (encaps_var 2827)) 6245 (("(" 2899) (argument_list 2900) (ctor_arguments 14325)) 6246 ((T_ATTRIBUTE 2853) (T_CLASS 14311) (anonymous_class 14324) (attribute 2898)) 6247 (("(" 14315) (argument_list 14316) (ctor_arguments 14317)) 6248 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14318) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 14319) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 14320) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6249 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14323) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6250 (("," 1449) (possible_comma 14321)) 6251 ((")" 14322)) 6252 ((T_EXTENDS 2902) (extends_from 14326)) 6253 ((T_IMPLEMENTS 2904) (implements_list 14327)) 6254 ((backup_doc_comment 14328)) 6255 (("{" 14329)) 6256 ((class_statement_list 14330)) 6257 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 14331)) 6258 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14333) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6259 (("%" 280) (")" 14334) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6260 (("{" 14335)) 6261 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 14336) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 6262 (("}" 14337)) 6263 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 14339) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 6264 ((")" 14340)) 6265 (("=" 14341)) 6266 (("!" 13885) ("\"" 13886) ("$" 13887) ("(" 13888) ("+" 13889) ("-" 13890) ("@" 13891) (T_ARRAY 13892) (T_ARRAY_CAST 13893) (T_ATTRIBUTE 125) (T_BOOL_CAST 13894) (T_CLASS_C 13895) (T_CLONE 13896) (T_CONSTANT_ENCAPSED_STRING 13897) (T_DEC 13898) (T_DIR 13899) (T_DNUMBER 13900) (T_DOUBLE_CAST 13901) (T_EMPTY 13902) (T_EVAL 13903) (T_EXIT 13904) (T_FILE 13905) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 13906) (T_INC 13907) (T_INCLUDE 13908) (T_INCLUDE_ONCE 13909) (T_INT_CAST 13910) (T_ISSET 13911) (T_LINE 13912) (T_LIST 13913) (T_LNUMBER 13914) (T_MATCH 13915) (T_METHOD_C 13916) (T_NAME_FULLY_QUALIFIED 13917) (T_NAME_QUALIFIED 13918) (T_NAME_RELATIVE 13919) (T_NEW 13920) (T_NS_C 13921) (T_OBJECT_CAST 13922) (T_PRINT 13923) (T_REQUIRE 13924) (T_REQUIRE_ONCE 13925) (T_START_HEREDOC 13926) (T_STATIC 13927) (T_STRING 13928) (T_STRING_CAST 13929) (T_THROW 13930) (T_TRAIT_C 13931) (T_UNSET_CAST 13932) (T_VARIABLE 13933) (T_YIELD 13934) (T_YIELD_FROM 13935) ("[" 13936) ("`" 13937) (array_object_dereferenceable 13938) (attribute 129) (attributes 13939) (callable_expr 13940) (callable_variable 13941) (class_constant 13942) (class_name 13943) (constant 13944) (dereferenceable_scalar 13945) (expr 14342) (fn 13947) (fully_dereferenceable 103) #'13948 (function_call 13949) (inline_function 13950) (internal_functions_in_yacc 13951) (match 13952) (name 13953) (new_expr 13954) (scalar 13955) (simple_variable 13956) (static_member 13957) (variable 13958) (variable_class_name 13959) ("~" 13960)) 6267 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 14344) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6268 (("," 5117) (possible_comma 14345)) 6269 ((")" 14346)) 6270 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 14353) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6271 ((")" 14354)) 6272 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14356) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6273 (("%" 280) (")" 14357) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6274 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14359) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6275 (("%" 280) (")" 14360) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6276 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 14367) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 6277 ((")" 14368)) 6278 (("%" 280) (")" 14373) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6279 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14376) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6280 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14377)) 6281 (("\"" 14379) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 6282 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 14389) (T_VARIABLE 2822) (encaps_var 2836)) 6283 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 14388) (T_VARIABLE 2822) (encaps_var 2827)) 6284 (("(" 2899) (argument_list 2900) (ctor_arguments 14408)) 6285 ((T_ATTRIBUTE 2853) (T_CLASS 14394) (anonymous_class 14407) (attribute 2898)) 6286 (("(" 14398) (argument_list 14399) (ctor_arguments 14400)) 6287 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14401) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 14402) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 14403) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6288 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14406) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6289 (("," 1449) (possible_comma 14404)) 6290 ((")" 14405)) 6291 ((T_EXTENDS 2902) (extends_from 14409)) 6292 ((T_IMPLEMENTS 2904) (implements_list 14410)) 6293 ((backup_doc_comment 14411)) 6294 (("{" 14412)) 6295 ((class_statement_list 14413)) 6296 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 14414)) 6297 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14416) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6298 (("%" 280) (")" 14417) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6299 (("{" 14418)) 6300 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 14419) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 6301 (("}" 14420)) 6302 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 14422) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 6303 ((")" 14423)) 6304 (("=" 14424)) 6305 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14425) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6306 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 14427) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6307 (("," 5117) (possible_comma 14428)) 6308 ((")" 14429)) 6309 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 14436) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6310 ((")" 14437)) 6311 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14439) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6312 (("%" 280) (")" 14440) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6313 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14442) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6314 (("%" 280) (")" 14443) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6315 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 14450) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 6316 ((")" 14451)) 6317 (("%" 280) (")" 14456) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6318 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14459) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6319 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14460)) 6320 (("\"" 14462) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 6321 (("]" 14465)) 6322 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14471) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6323 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14472)) 6324 (("$" 325) (T_VARIABLE 332) (simple_variable 14475)) 6325 (("%" 280) (")" 14477) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6326 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14480) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6327 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14481)) 6328 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14491) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6329 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 14489) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 6330 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14490) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6331 (("$" 205) (T_ABSTRACT 14501) (T_ARRAY 14502) (T_AS 14503) (T_BREAK 14504) (T_CALLABLE 14505) (T_CASE 14506) (T_CATCH 14507) (T_CLASS 14508) (T_CLASS_C 14509) (T_CLONE 14510) (T_CONST 14511) (T_CONTINUE 14512) (T_DECLARE 14513) (T_DEFAULT 14514) (T_DIR 14515) (T_DO 14516) (T_ECHO 14517) (T_ELSE 14518) (T_ELSEIF 14519) (T_EMPTY 14520) (T_ENDDECLARE 14521) (T_ENDFOR 14522) (T_ENDFOREACH 14523) (T_ENDIF 14524) (T_ENDSWITCH 14525) (T_ENDWHILE 14526) (T_ENUM 14527) (T_EVAL 14528) (T_EXIT 14529) (T_EXTENDS 14530) (T_FILE 14531) (T_FINAL 14532) (T_FINALLY 14533) (T_FN 14534) (T_FOR 14535) (T_FOREACH 14536) (T_FUNCTION 14537) (T_FUNC_C 14538) (T_GLOBAL 14539) (T_GOTO 14540) (T_IF 14541) (T_IMPLEMENTS 14542) (T_INCLUDE 14543) (T_INCLUDE_ONCE 14544) (T_INSTANCEOF 14545) (T_INSTEADOF 14546) (T_INTERFACE 14547) (T_ISSET 14548) (T_LINE 14549) (T_LIST 14550) (T_LOGICAL_AND 14551) (T_LOGICAL_OR 14552) (T_LOGICAL_XOR 14553) (T_MATCH 14554) (T_METHOD_C 14555) (T_NAMESPACE 14556) (T_NEW 14557) (T_NS_C 14558) (T_PRINT 14559) (T_PRIVATE 14560) (T_PROTECTED 14561) (T_PUBLIC 14562) (T_READONLY 14563) (T_REQUIRE 14564) (T_REQUIRE_ONCE 14565) (T_RETURN 14566) (T_STATIC 14567) (T_STRING 14568) (T_SWITCH 14569) (T_THROW 14570) (T_TRAIT 14571) (T_TRAIT_C 14572) (T_TRY 14573) (T_UNSET 14574) (T_USE 14575) (T_VAR 14576) (T_VARIABLE 251) (T_WHILE 14577) (T_YIELD 14578) (identifier 14579) (member_name 14580) (reserved_non_modifiers 14581) (semi_reserved 14582) (simple_variable 14583) ("{" 914)) 6332 (("(" 14584) (argument_list 14585)) 6333 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14586) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 14587) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 14588) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6334 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14591) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6335 (("," 1449) (possible_comma 14589)) 6336 ((")" 14590)) 6337 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (ampersand 14621) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14622) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6338 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14620) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6339 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14619) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6340 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14618) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6341 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14617) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6342 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14616) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6343 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14615) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6344 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14614) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6345 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14613) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6346 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14612) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6347 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14611) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6348 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14610) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6349 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14609) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6350 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14608) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6351 (("\"" 1489) ("$" 14623) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 14624) ("[" 1508) (array_object_dereferenceable 14625) (callable_expr 14626) (callable_variable 14627) (class_constant 1512) (class_name 14628) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 14629) (name 14630) (simple_variable 14631) (static_member 14632) (variable 14633) (variable_class_name 14634)) 6352 (("$" 14623) (T_VARIABLE 14624) (simple_variable 14669) ("{" 14670)) 6353 ((T_NULLSAFE_OBJECT_OPERATOR 14652) (T_OBJECT_OPERATOR 14653) ("[" 14654) ("{" 14655)) 6354 (("(" 14638) (argument_list 14651)) 6355 ((T_PAAMAYIM_NEKUDOTAYIM 14647)) 6356 (("(" 14638) (argument_list 14646)) 6357 ((T_PAAMAYIM_NEKUDOTAYIM 14635)) 6358 (("$" 14623) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 14624) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 14636) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 14637) ("{" 914)) 6359 (("(" 14638) (argument_list 14639)) 6360 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14640) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 14641) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 14642) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6361 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14645) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6362 (("," 1449) (possible_comma 14643)) 6363 ((")" 14644)) 6364 (("$" 14623) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 14624) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 14648) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 14649) ("{" 914)) 6365 (("(" 14638) (argument_list 14650)) 6366 (("$" 14623) (T_STRING 14660) (T_VARIABLE 14624) (property_name 14667) (simple_variable 14662) ("{" 14663)) 6367 (("$" 14623) (T_STRING 14660) (T_VARIABLE 14624) (property_name 14661) (simple_variable 14662) ("{" 14663)) 6368 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 14658) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6369 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14656) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6370 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14657)) 6371 (("]" 14659)) 6372 (("(" 14638) (argument_list 14666)) 6373 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14664) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6374 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14665)) 6375 (("(" 14638) (argument_list 14668)) 6376 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14671) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6377 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14672)) 6378 ((backup_doc_comment 14675)) 6379 (("(" 14676)) 6380 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 14677)) 6381 ((")" 14678)) 6382 ((T_USE 148) (lexical_vars 14679)) 6383 ((":" 150) (return_type 14680)) 6384 ((backup_fn_flags 14681)) 6385 (("{" 14682)) 6386 ((inner_statement_list 14683)) 6387 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 14684) ("~" 124)) 6388 ((backup_fn_flags 14685)) 6389 ((backup_doc_comment 14687)) 6390 (("(" 14688)) 6391 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 14689)) 6392 ((")" 14690)) 6393 ((":" 2207) (return_type 14691)) 6394 ((T_DOUBLE_ARROW 14692)) 6395 ((backup_fn_flags 14693)) 6396 ((backup_lex_pos 14694)) 6397 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14756) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6398 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15734) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6399 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 15732) (encaps_var 5158)) 6400 (("$" 14697) (T_VARIABLE 14743) (simple_variable 15728) ("{" 15729)) 6401 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15726) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6402 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15725) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6403 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15724) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6404 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15723) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6405 (("(" 15720)) 6406 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15719) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6407 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15718) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6408 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15717) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6409 (("\"" 1489) ("$" 14996) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 14997) ("[" 1508) (array_object_dereferenceable 14998) (callable_expr 14999) (callable_variable 15000) (class_constant 1512) (class_name 15001) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 15002) (name 15003) (simple_variable 15004) (static_member 15005) (variable 15716) (variable_class_name 15007)) 6410 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15715) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6411 (("(" 15712)) 6412 (("(" 15709)) 6413 (("(" 15705) (exit_expr 15706)) 6414 (("\"" 1489) ("$" 14996) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 14997) ("[" 1508) (array_object_dereferenceable 14998) (callable_expr 14999) (callable_variable 15000) (class_constant 1512) (class_name 15001) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 15002) (name 15003) (simple_variable 15004) (static_member 15005) (variable 15704) (variable_class_name 15007)) 6415 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15703) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6416 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15702) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6417 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15701) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6418 (("(" 15697)) 6419 (("(" 15692)) 6420 (("(" 15686)) 6421 (("$" 14996) ("(" 15639) (T_ATTRIBUTE 2853) (T_CLASS 15640) (T_NAME_FULLY_QUALIFIED 15641) (T_NAME_QUALIFIED 15642) (T_NAME_RELATIVE 15643) (T_STATIC 15644) (T_STRING 15645) (T_VARIABLE 14997) (anonymous_class 15646) (attribute 2862) (attributes 15647) (class_name 15648) (class_name_reference 15649) (name 15650) (new_variable 15651) (simple_variable 15652)) 6422 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15638) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6423 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15637) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6424 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15636) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6425 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15635) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6426 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 15630) (T_END_HEREDOC 15631) (T_VARIABLE 2822) (encaps_list 15632) (encaps_var 2824)) 6427 ((T_FN 35) (T_FUNCTION 126) (fn 14757) #'14758 (inline_function 15629)) 6428 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15628) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6429 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15627) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6430 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15626) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6431 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15165) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6432 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15103) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6433 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 15099) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 6434 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 15097) (encaps_list 2729) (encaps_var 2730)) 6435 ((T_NULLSAFE_OBJECT_OPERATOR 15080) (T_OBJECT_OPERATOR 15081) ("[" 15082) ("{" 15083)) 6436 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 15077) (attribute 2654) (fn 14757) #'14758 (inline_function 15078)) 6437 (("(" 14957) (argument_list 15076)) 6438 ((T_PAAMAYIM_NEKUDOTAYIM 15071)) 6439 (("%" 14772) ("*" 14773) ("+" 14774) ("-" 14775) ("." 14776) ("/" 14777) ("<" 14778) (">" 14779) ("?" 14780) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14781) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14782) (T_BOOLEAN_AND 14783) (T_BOOLEAN_OR 14784) (T_COALESCE 14785) (T_INSTANCEOF 14786) (T_IS_EQUAL 14787) (T_IS_GREATER_OR_EQUAL 14788) (T_IS_IDENTICAL 14789) (T_IS_NOT_EQUAL 14790) (T_IS_NOT_IDENTICAL 14791) (T_IS_SMALLER_OR_EQUAL 14792) (T_LOGICAL_AND 14793) (T_LOGICAL_OR 14794) (T_LOGICAL_XOR 14795) (T_POW 14796) (T_SL 14797) (T_SPACESHIP 14798) (T_SR 14799) ("^" 14800) (backup_fn_flags 15070) ("|" 14801)) 6440 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 15059)) 6441 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 15047)) 6442 (("(" 14957) (argument_list 15046)) 6443 (("=" 14965) (T_AND_EQUAL 14966) (T_COALESCE_EQUAL 14967) (T_CONCAT_EQUAL 14968) (T_DEC 14969) (T_DIV_EQUAL 14970) (T_INC 14971) (T_MINUS_EQUAL 14972) (T_MOD_EQUAL 14973) (T_MUL_EQUAL 14974) (T_OR_EQUAL 14975) (T_PLUS_EQUAL 14976) (T_POW_EQUAL 14977) (T_SL_EQUAL 14978) (T_SR_EQUAL 14979) (T_XOR_EQUAL 14980)) 6444 ((T_PAAMAYIM_NEKUDOTAYIM 14873)) 6445 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14771) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6446 (("%" 14772) ("*" 14773) ("+" 14774) ("-" 14775) ("." 14776) ("/" 14777) ("<" 14778) (">" 14779) ("?" 14780) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14781) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14782) (T_BOOLEAN_AND 14783) (T_BOOLEAN_OR 14784) (T_COALESCE 14785) (T_INSTANCEOF 14786) (T_IS_EQUAL 14787) (T_IS_GREATER_OR_EQUAL 14788) (T_IS_IDENTICAL 14789) (T_IS_NOT_EQUAL 14790) (T_IS_NOT_IDENTICAL 14791) (T_IS_SMALLER_OR_EQUAL 14792) (T_LOGICAL_AND 14793) (T_LOGICAL_OR 14794) (T_LOGICAL_XOR 14795) (T_POW 14796) (T_SL 14797) (T_SPACESHIP 14798) (T_SR 14799) ("^" 14800) ("|" 14801)) 6447 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14872) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6448 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14871) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6449 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14870) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6450 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14869) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6451 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14868) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6452 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14867) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6453 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14866) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6454 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14865) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6455 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 14860) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 14861) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 6456 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14859) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6457 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14858) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6458 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14857) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6459 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14856) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6460 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14855) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6461 (("$" 14817) ("(" 14818) (T_NAME_FULLY_QUALIFIED 14819) (T_NAME_QUALIFIED 14820) (T_NAME_RELATIVE 14821) (T_STATIC 14822) (T_STRING 14823) (T_VARIABLE 14824) (class_name 14825) (class_name_reference 14826) (name 14827) (new_variable 14828) (simple_variable 14829)) 6462 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14816) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6463 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14815) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6464 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14814) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6465 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14813) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6466 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14812) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6467 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14811) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6468 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14810) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6469 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14809) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6470 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14808) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6471 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14807) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6472 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14806) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6473 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14805) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6474 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14804) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6475 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14803) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6476 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14802) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6477 (("$" 14817) (T_VARIABLE 14824) (simple_variable 14851) ("{" 14852)) 6478 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 14849) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6479 ((T_PAAMAYIM_NEKUDOTAYIM 14847)) 6480 ((T_NULLSAFE_OBJECT_OPERATOR 14830) (T_OBJECT_OPERATOR 14831) (T_PAAMAYIM_NEKUDOTAYIM 14832) ("[" 14833) ("{" 14834)) 6481 (("$" 14817) (T_STRING 14840) (T_VARIABLE 14824) (property_name 14846) (simple_variable 14842) ("{" 14843)) 6482 (("$" 14817) (T_STRING 14840) (T_VARIABLE 14824) (property_name 14841) (simple_variable 14842) ("{" 14843)) 6483 (("$" 14817) (T_VARIABLE 14824) (simple_variable 14839)) 6484 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 14837) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6485 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14835) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6486 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14836)) 6487 (("]" 14838)) 6488 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14844) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6489 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14845)) 6490 (("$" 14817) (T_VARIABLE 14824) (simple_variable 14848)) 6491 (("%" 280) (")" 14850) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6492 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 14853) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6493 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 14854)) 6494 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14864) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6495 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 14862) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 6496 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14863) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6497 (("$" 14697) (T_ABSTRACT 14874) (T_ARRAY 14875) (T_AS 14876) (T_BREAK 14877) (T_CALLABLE 14878) (T_CASE 14879) (T_CATCH 14880) (T_CLASS 14881) (T_CLASS_C 14882) (T_CLONE 14883) (T_CONST 14884) (T_CONTINUE 14885) (T_DECLARE 14886) (T_DEFAULT 14887) (T_DIR 14888) (T_DO 14889) (T_ECHO 14890) (T_ELSE 14891) (T_ELSEIF 14892) (T_EMPTY 14893) (T_ENDDECLARE 14894) (T_ENDFOR 14895) (T_ENDFOREACH 14896) (T_ENDIF 14897) (T_ENDSWITCH 14898) (T_ENDWHILE 14899) (T_ENUM 14900) (T_EVAL 14901) (T_EXIT 14902) (T_EXTENDS 14903) (T_FILE 14904) (T_FINAL 14905) (T_FINALLY 14906) (T_FN 14907) (T_FOR 14908) (T_FOREACH 14909) (T_FUNCTION 14910) (T_FUNC_C 14911) (T_GLOBAL 14912) (T_GOTO 14913) (T_IF 14914) (T_IMPLEMENTS 14915) (T_INCLUDE 14916) (T_INCLUDE_ONCE 14917) (T_INSTANCEOF 14918) (T_INSTEADOF 14919) (T_INTERFACE 14920) (T_ISSET 14921) (T_LINE 14922) (T_LIST 14923) (T_LOGICAL_AND 14924) (T_LOGICAL_OR 14925) (T_LOGICAL_XOR 14926) (T_MATCH 14927) (T_METHOD_C 14928) (T_NAMESPACE 14929) (T_NEW 14930) (T_NS_C 14931) (T_PRINT 14932) (T_PRIVATE 14933) (T_PROTECTED 14934) (T_PUBLIC 14935) (T_READONLY 14936) (T_REQUIRE 14937) (T_REQUIRE_ONCE 14938) (T_RETURN 14939) (T_STATIC 14940) (T_STRING 14941) (T_SWITCH 14942) (T_THROW 14943) (T_TRAIT 14944) (T_TRAIT_C 14945) (T_TRY 14946) (T_UNSET 14947) (T_USE 14948) (T_VAR 14949) (T_VARIABLE 14743) (T_WHILE 14950) (T_YIELD 14951) (identifier 14952) (member_name 14953) (reserved_non_modifiers 14954) (semi_reserved 14955) (simple_variable 14956) ("{" 914)) 6498 (("(" 14957) (argument_list 14958)) 6499 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14959) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 14960) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 14961) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6500 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 14964) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6501 (("," 1449) (possible_comma 14962)) 6502 ((")" 14963)) 6503 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (ampersand 14994) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14995) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6504 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14993) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6505 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14992) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6506 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14991) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6507 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14990) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6508 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14989) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6509 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14988) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6510 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14987) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6511 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14986) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6512 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14985) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6513 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14984) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6514 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14983) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6515 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14982) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6516 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 14981) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6517 (("\"" 1489) ("$" 14996) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 14997) ("[" 1508) (array_object_dereferenceable 14998) (callable_expr 14999) (callable_variable 15000) (class_constant 1512) (class_name 15001) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 15002) (name 15003) (simple_variable 15004) (static_member 15005) (variable 15006) (variable_class_name 15007)) 6518 (("$" 14996) (T_VARIABLE 14997) (simple_variable 15042) ("{" 15043)) 6519 ((T_NULLSAFE_OBJECT_OPERATOR 15025) (T_OBJECT_OPERATOR 15026) ("[" 15027) ("{" 15028)) 6520 (("(" 15011) (argument_list 15024)) 6521 ((T_PAAMAYIM_NEKUDOTAYIM 15020)) 6522 (("(" 15011) (argument_list 15019)) 6523 ((T_PAAMAYIM_NEKUDOTAYIM 15008)) 6524 (("$" 14996) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 14997) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 15009) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 15010) ("{" 914)) 6525 (("(" 15011) (argument_list 15012)) 6526 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 15013) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 15014) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 15015) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6527 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 15018) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6528 (("," 1449) (possible_comma 15016)) 6529 ((")" 15017)) 6530 (("$" 14996) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 14997) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 15021) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 15022) ("{" 914)) 6531 (("(" 15011) (argument_list 15023)) 6532 (("$" 14996) (T_STRING 15033) (T_VARIABLE 14997) (property_name 15040) (simple_variable 15035) ("{" 15036)) 6533 (("$" 14996) (T_STRING 15033) (T_VARIABLE 14997) (property_name 15034) (simple_variable 15035) ("{" 15036)) 6534 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 15031) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6535 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15029) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6536 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15030)) 6537 (("]" 15032)) 6538 (("(" 15011) (argument_list 15039)) 6539 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15037) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6540 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15038)) 6541 (("(" 15011) (argument_list 15041)) 6542 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15044) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6543 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15045)) 6544 ((backup_doc_comment 15048)) 6545 (("(" 15049)) 6546 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 15050)) 6547 ((")" 15051)) 6548 ((T_USE 148) (lexical_vars 15052)) 6549 ((":" 150) (return_type 15053)) 6550 ((backup_fn_flags 15054)) 6551 (("{" 15055)) 6552 ((inner_statement_list 15056)) 6553 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 15057) ("~" 124)) 6554 ((backup_fn_flags 15058)) 6555 ((backup_doc_comment 15060)) 6556 (("(" 15061)) 6557 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 15062)) 6558 ((")" 15063)) 6559 ((":" 2207) (return_type 15064)) 6560 ((T_DOUBLE_ARROW 15065)) 6561 ((backup_fn_flags 15066)) 6562 ((backup_lex_pos 15067)) 6563 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15068) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6564 (("%" 14772) ("*" 14773) ("+" 14774) ("-" 14775) ("." 14776) ("/" 14777) ("<" 14778) (">" 14779) ("?" 14780) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14781) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14782) (T_BOOLEAN_AND 14783) (T_BOOLEAN_OR 14784) (T_COALESCE 14785) (T_INSTANCEOF 14786) (T_IS_EQUAL 14787) (T_IS_GREATER_OR_EQUAL 14788) (T_IS_IDENTICAL 14789) (T_IS_NOT_EQUAL 14790) (T_IS_NOT_IDENTICAL 14791) (T_IS_SMALLER_OR_EQUAL 14792) (T_LOGICAL_AND 14793) (T_LOGICAL_OR 14794) (T_LOGICAL_XOR 14795) (T_POW 14796) (T_SL 14797) (T_SPACESHIP 14798) (T_SR 14799) ("^" 14800) (backup_fn_flags 15069) ("|" 14801)) 6565 (("$" 14697) (T_ABSTRACT 14874) (T_ARRAY 14875) (T_AS 14876) (T_BREAK 14877) (T_CALLABLE 14878) (T_CASE 14879) (T_CATCH 14880) (T_CLASS 14881) (T_CLASS_C 14882) (T_CLONE 14883) (T_CONST 14884) (T_CONTINUE 14885) (T_DECLARE 14886) (T_DEFAULT 14887) (T_DIR 14888) (T_DO 14889) (T_ECHO 14890) (T_ELSE 14891) (T_ELSEIF 14892) (T_EMPTY 14893) (T_ENDDECLARE 14894) (T_ENDFOR 14895) (T_ENDFOREACH 14896) (T_ENDIF 14897) (T_ENDSWITCH 14898) (T_ENDWHILE 14899) (T_ENUM 14900) (T_EVAL 14901) (T_EXIT 14902) (T_EXTENDS 14903) (T_FILE 14904) (T_FINAL 14905) (T_FINALLY 14906) (T_FN 14907) (T_FOR 14908) (T_FOREACH 14909) (T_FUNCTION 14910) (T_FUNC_C 14911) (T_GLOBAL 14912) (T_GOTO 14913) (T_IF 14914) (T_IMPLEMENTS 14915) (T_INCLUDE 14916) (T_INCLUDE_ONCE 14917) (T_INSTANCEOF 14918) (T_INSTEADOF 14919) (T_INTERFACE 14920) (T_ISSET 14921) (T_LINE 14922) (T_LIST 14923) (T_LOGICAL_AND 14924) (T_LOGICAL_OR 14925) (T_LOGICAL_XOR 14926) (T_MATCH 14927) (T_METHOD_C 14928) (T_NAMESPACE 14929) (T_NEW 14930) (T_NS_C 14931) (T_PRINT 14932) (T_PRIVATE 14933) (T_PROTECTED 14934) (T_PUBLIC 14935) (T_READONLY 14936) (T_REQUIRE 14937) (T_REQUIRE_ONCE 14938) (T_RETURN 14939) (T_STATIC 14940) (T_STRING 14941) (T_SWITCH 14942) (T_THROW 14943) (T_TRAIT 14944) (T_TRAIT_C 14945) (T_TRY 14946) (T_UNSET 14947) (T_USE 14948) (T_VAR 14949) (T_VARIABLE 14743) (T_WHILE 14950) (T_YIELD 14951) (identifier 15072) (member_name 15073) (reserved_non_modifiers 14954) (semi_reserved 14955) (simple_variable 15074) ("{" 914)) 6566 (("(" 14957) (argument_list 15075)) 6567 ((T_FN 35) (T_FUNCTION 126) (fn 14757) #'14758 (inline_function 15079)) 6568 (("$" 14697) (T_STRING 15088) (T_VARIABLE 14743) (property_name 15095) (simple_variable 15090) ("{" 15091)) 6569 (("$" 14697) (T_STRING 15088) (T_VARIABLE 14743) (property_name 15089) (simple_variable 15090) ("{" 15091)) 6570 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 15086) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6571 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15084) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6572 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15085)) 6573 (("]" 15087)) 6574 (("(" 14957) (argument_list 15094)) 6575 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15092) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6576 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15093)) 6577 (("(" 14957) (argument_list 15096)) 6578 (("`" 15098)) 6579 (("]" 15100)) 6580 (("=" 15101)) 6581 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15102) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6582 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15625) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6583 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 15623) (encaps_var 5158)) 6584 (("$" 15106) (T_VARIABLE 15152) (simple_variable 15619) ("{" 15620)) 6585 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15617) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6586 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15616) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6587 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15615) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6588 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15614) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6589 (("(" 15611)) 6590 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15610) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6591 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15609) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6592 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15608) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6593 (("\"" 1489) ("$" 15405) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 15406) ("[" 1508) (array_object_dereferenceable 15407) (callable_expr 15408) (callable_variable 15409) (class_constant 1512) (class_name 15410) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 15411) (name 15412) (simple_variable 15413) (static_member 15414) (variable 15607) (variable_class_name 15416)) 6594 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15606) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6595 (("(" 15603)) 6596 (("(" 15600)) 6597 (("(" 15596) (exit_expr 15597)) 6598 (("\"" 1489) ("$" 15405) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 15406) ("[" 1508) (array_object_dereferenceable 15407) (callable_expr 15408) (callable_variable 15409) (class_constant 1512) (class_name 15410) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 15411) (name 15412) (simple_variable 15413) (static_member 15414) (variable 15595) (variable_class_name 15416)) 6599 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15594) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6600 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15593) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6601 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15592) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6602 (("(" 15588)) 6603 (("(" 15583)) 6604 (("(" 15577)) 6605 (("$" 15405) ("(" 15530) (T_ATTRIBUTE 2853) (T_CLASS 15531) (T_NAME_FULLY_QUALIFIED 15532) (T_NAME_QUALIFIED 15533) (T_NAME_RELATIVE 15534) (T_STATIC 15535) (T_STRING 15536) (T_VARIABLE 15406) (anonymous_class 15537) (attribute 2862) (attributes 15538) (class_name 15539) (class_name_reference 15540) (name 15541) (new_variable 15542) (simple_variable 15543)) 6606 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15529) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6607 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15528) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6608 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15527) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6609 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15526) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6610 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 15521) (T_END_HEREDOC 15522) (T_VARIABLE 2822) (encaps_list 15523) (encaps_var 2824)) 6611 ((T_FN 35) (T_FUNCTION 126) (fn 15166) #'15167 (inline_function 15520)) 6612 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15519) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6613 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15518) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6614 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15517) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6615 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15514) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6616 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15513) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6617 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 15509) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 6618 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 15507) (encaps_list 2729) (encaps_var 2730)) 6619 ((T_NULLSAFE_OBJECT_OPERATOR 15490) (T_OBJECT_OPERATOR 15491) ("[" 15492) ("{" 15493)) 6620 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 15487) (attribute 2654) (fn 15166) #'15167 (inline_function 15488)) 6621 (("(" 15366) (argument_list 15486)) 6622 ((T_PAAMAYIM_NEKUDOTAYIM 15481)) 6623 (("%" 15181) ("*" 15182) ("+" 15183) ("-" 15184) ("." 15185) ("/" 15186) ("<" 15187) (">" 15188) ("?" 15189) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15190) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15191) (T_BOOLEAN_AND 15192) (T_BOOLEAN_OR 15193) (T_COALESCE 15194) (T_DOUBLE_ARROW 15479) (T_INSTANCEOF 15195) (T_IS_EQUAL 15196) (T_IS_GREATER_OR_EQUAL 15197) (T_IS_IDENTICAL 15198) (T_IS_NOT_EQUAL 15199) (T_IS_NOT_IDENTICAL 15200) (T_IS_SMALLER_OR_EQUAL 15201) (T_LOGICAL_AND 15202) (T_LOGICAL_OR 15203) (T_LOGICAL_XOR 15204) (T_POW 15205) (T_SL 15206) (T_SPACESHIP 15207) (T_SR 15208) ("^" 15209) ("|" 15210)) 6624 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 15468)) 6625 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 15456)) 6626 (("(" 15366) (argument_list 15455)) 6627 (("=" 15374) (T_AND_EQUAL 15375) (T_COALESCE_EQUAL 15376) (T_CONCAT_EQUAL 15377) (T_DEC 15378) (T_DIV_EQUAL 15379) (T_INC 15380) (T_MINUS_EQUAL 15381) (T_MOD_EQUAL 15382) (T_MUL_EQUAL 15383) (T_OR_EQUAL 15384) (T_PLUS_EQUAL 15385) (T_POW_EQUAL 15386) (T_SL_EQUAL 15387) (T_SR_EQUAL 15388) (T_XOR_EQUAL 15389)) 6628 ((T_PAAMAYIM_NEKUDOTAYIM 15282)) 6629 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15180) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6630 (("%" 15181) ("*" 15182) ("+" 15183) ("-" 15184) ("." 15185) ("/" 15186) ("<" 15187) (">" 15188) ("?" 15189) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15190) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15191) (T_BOOLEAN_AND 15192) (T_BOOLEAN_OR 15193) (T_COALESCE 15194) (T_INSTANCEOF 15195) (T_IS_EQUAL 15196) (T_IS_GREATER_OR_EQUAL 15197) (T_IS_IDENTICAL 15198) (T_IS_NOT_EQUAL 15199) (T_IS_NOT_IDENTICAL 15200) (T_IS_SMALLER_OR_EQUAL 15201) (T_LOGICAL_AND 15202) (T_LOGICAL_OR 15203) (T_LOGICAL_XOR 15204) (T_POW 15205) (T_SL 15206) (T_SPACESHIP 15207) (T_SR 15208) ("^" 15209) ("|" 15210)) 6631 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15281) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6632 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15280) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6633 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15279) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6634 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15278) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6635 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15277) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6636 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15276) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6637 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15275) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6638 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15274) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6639 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 15269) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 15270) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 6640 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15268) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6641 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15267) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6642 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15266) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6643 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15265) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6644 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15264) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6645 (("$" 15226) ("(" 15227) (T_NAME_FULLY_QUALIFIED 15228) (T_NAME_QUALIFIED 15229) (T_NAME_RELATIVE 15230) (T_STATIC 15231) (T_STRING 15232) (T_VARIABLE 15233) (class_name 15234) (class_name_reference 15235) (name 15236) (new_variable 15237) (simple_variable 15238)) 6646 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15225) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6647 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15224) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6648 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15223) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6649 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15222) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6650 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15221) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6651 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15220) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6652 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15219) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6653 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15218) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6654 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15217) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6655 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15216) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6656 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15215) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6657 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15214) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6658 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15213) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6659 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15212) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6660 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15211) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6661 (("$" 15226) (T_VARIABLE 15233) (simple_variable 15260) ("{" 15261)) 6662 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15258) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6663 ((T_PAAMAYIM_NEKUDOTAYIM 15256)) 6664 ((T_NULLSAFE_OBJECT_OPERATOR 15239) (T_OBJECT_OPERATOR 15240) (T_PAAMAYIM_NEKUDOTAYIM 15241) ("[" 15242) ("{" 15243)) 6665 (("$" 15226) (T_STRING 15249) (T_VARIABLE 15233) (property_name 15255) (simple_variable 15251) ("{" 15252)) 6666 (("$" 15226) (T_STRING 15249) (T_VARIABLE 15233) (property_name 15250) (simple_variable 15251) ("{" 15252)) 6667 (("$" 15226) (T_VARIABLE 15233) (simple_variable 15248)) 6668 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 15246) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6669 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15244) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6670 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15245)) 6671 (("]" 15247)) 6672 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15253) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6673 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15254)) 6674 (("$" 15226) (T_VARIABLE 15233) (simple_variable 15257)) 6675 (("%" 280) (")" 15259) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6676 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15262) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6677 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15263)) 6678 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15273) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6679 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 15271) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 6680 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15272) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6681 (("$" 15106) (T_ABSTRACT 15283) (T_ARRAY 15284) (T_AS 15285) (T_BREAK 15286) (T_CALLABLE 15287) (T_CASE 15288) (T_CATCH 15289) (T_CLASS 15290) (T_CLASS_C 15291) (T_CLONE 15292) (T_CONST 15293) (T_CONTINUE 15294) (T_DECLARE 15295) (T_DEFAULT 15296) (T_DIR 15297) (T_DO 15298) (T_ECHO 15299) (T_ELSE 15300) (T_ELSEIF 15301) (T_EMPTY 15302) (T_ENDDECLARE 15303) (T_ENDFOR 15304) (T_ENDFOREACH 15305) (T_ENDIF 15306) (T_ENDSWITCH 15307) (T_ENDWHILE 15308) (T_ENUM 15309) (T_EVAL 15310) (T_EXIT 15311) (T_EXTENDS 15312) (T_FILE 15313) (T_FINAL 15314) (T_FINALLY 15315) (T_FN 15316) (T_FOR 15317) (T_FOREACH 15318) (T_FUNCTION 15319) (T_FUNC_C 15320) (T_GLOBAL 15321) (T_GOTO 15322) (T_IF 15323) (T_IMPLEMENTS 15324) (T_INCLUDE 15325) (T_INCLUDE_ONCE 15326) (T_INSTANCEOF 15327) (T_INSTEADOF 15328) (T_INTERFACE 15329) (T_ISSET 15330) (T_LINE 15331) (T_LIST 15332) (T_LOGICAL_AND 15333) (T_LOGICAL_OR 15334) (T_LOGICAL_XOR 15335) (T_MATCH 15336) (T_METHOD_C 15337) (T_NAMESPACE 15338) (T_NEW 15339) (T_NS_C 15340) (T_PRINT 15341) (T_PRIVATE 15342) (T_PROTECTED 15343) (T_PUBLIC 15344) (T_READONLY 15345) (T_REQUIRE 15346) (T_REQUIRE_ONCE 15347) (T_RETURN 15348) (T_STATIC 15349) (T_STRING 15350) (T_SWITCH 15351) (T_THROW 15352) (T_TRAIT 15353) (T_TRAIT_C 15354) (T_TRY 15355) (T_UNSET 15356) (T_USE 15357) (T_VAR 15358) (T_VARIABLE 15152) (T_WHILE 15359) (T_YIELD 15360) (identifier 15361) (member_name 15362) (reserved_non_modifiers 15363) (semi_reserved 15364) (simple_variable 15365) ("{" 914)) 6682 (("(" 15366) (argument_list 15367)) 6683 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 15368) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 15369) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 15370) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6684 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 15373) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6685 (("," 1449) (possible_comma 15371)) 6686 ((")" 15372)) 6687 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (ampersand 15403) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15404) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6688 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15402) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6689 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15401) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6690 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15400) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6691 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15399) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6692 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15398) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6693 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15397) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6694 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15396) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6695 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15395) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6696 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15394) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6697 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15393) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6698 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15392) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6699 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15391) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6700 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15390) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6701 (("\"" 1489) ("$" 15405) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 15406) ("[" 1508) (array_object_dereferenceable 15407) (callable_expr 15408) (callable_variable 15409) (class_constant 1512) (class_name 15410) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 15411) (name 15412) (simple_variable 15413) (static_member 15414) (variable 15415) (variable_class_name 15416)) 6702 (("$" 15405) (T_VARIABLE 15406) (simple_variable 15451) ("{" 15452)) 6703 ((T_NULLSAFE_OBJECT_OPERATOR 15434) (T_OBJECT_OPERATOR 15435) ("[" 15436) ("{" 15437)) 6704 (("(" 15420) (argument_list 15433)) 6705 ((T_PAAMAYIM_NEKUDOTAYIM 15429)) 6706 (("(" 15420) (argument_list 15428)) 6707 ((T_PAAMAYIM_NEKUDOTAYIM 15417)) 6708 (("$" 15405) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 15406) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 15418) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 15419) ("{" 914)) 6709 (("(" 15420) (argument_list 15421)) 6710 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 15422) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 15423) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 15424) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6711 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 15427) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6712 (("," 1449) (possible_comma 15425)) 6713 ((")" 15426)) 6714 (("$" 15405) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 15406) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 15430) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 15431) ("{" 914)) 6715 (("(" 15420) (argument_list 15432)) 6716 (("$" 15405) (T_STRING 15442) (T_VARIABLE 15406) (property_name 15449) (simple_variable 15444) ("{" 15445)) 6717 (("$" 15405) (T_STRING 15442) (T_VARIABLE 15406) (property_name 15443) (simple_variable 15444) ("{" 15445)) 6718 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 15440) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6719 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15438) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6720 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15439)) 6721 (("]" 15441)) 6722 (("(" 15420) (argument_list 15448)) 6723 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15446) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6724 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15447)) 6725 (("(" 15420) (argument_list 15450)) 6726 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15453) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6727 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15454)) 6728 ((backup_doc_comment 15457)) 6729 (("(" 15458)) 6730 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 15459)) 6731 ((")" 15460)) 6732 ((T_USE 148) (lexical_vars 15461)) 6733 ((":" 150) (return_type 15462)) 6734 ((backup_fn_flags 15463)) 6735 (("{" 15464)) 6736 ((inner_statement_list 15465)) 6737 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 15466) ("~" 124)) 6738 ((backup_fn_flags 15467)) 6739 ((backup_doc_comment 15469)) 6740 (("(" 15470)) 6741 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 15471)) 6742 ((")" 15472)) 6743 ((":" 2207) (return_type 15473)) 6744 ((T_DOUBLE_ARROW 15474)) 6745 ((backup_fn_flags 15475)) 6746 ((backup_lex_pos 15476)) 6747 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15477) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6748 (("%" 15181) ("*" 15182) ("+" 15183) ("-" 15184) ("." 15185) ("/" 15186) ("<" 15187) (">" 15188) ("?" 15189) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15190) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15191) (T_BOOLEAN_AND 15192) (T_BOOLEAN_OR 15193) (T_COALESCE 15194) (T_INSTANCEOF 15195) (T_IS_EQUAL 15196) (T_IS_GREATER_OR_EQUAL 15197) (T_IS_IDENTICAL 15198) (T_IS_NOT_EQUAL 15199) (T_IS_NOT_IDENTICAL 15200) (T_IS_SMALLER_OR_EQUAL 15201) (T_LOGICAL_AND 15202) (T_LOGICAL_OR 15203) (T_LOGICAL_XOR 15204) (T_POW 15205) (T_SL 15206) (T_SPACESHIP 15207) (T_SR 15208) ("^" 15209) (backup_fn_flags 15478) ("|" 15210)) 6749 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15480) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6750 (("$" 15106) (T_ABSTRACT 15283) (T_ARRAY 15284) (T_AS 15285) (T_BREAK 15286) (T_CALLABLE 15287) (T_CASE 15288) (T_CATCH 15289) (T_CLASS 15290) (T_CLASS_C 15291) (T_CLONE 15292) (T_CONST 15293) (T_CONTINUE 15294) (T_DECLARE 15295) (T_DEFAULT 15296) (T_DIR 15297) (T_DO 15298) (T_ECHO 15299) (T_ELSE 15300) (T_ELSEIF 15301) (T_EMPTY 15302) (T_ENDDECLARE 15303) (T_ENDFOR 15304) (T_ENDFOREACH 15305) (T_ENDIF 15306) (T_ENDSWITCH 15307) (T_ENDWHILE 15308) (T_ENUM 15309) (T_EVAL 15310) (T_EXIT 15311) (T_EXTENDS 15312) (T_FILE 15313) (T_FINAL 15314) (T_FINALLY 15315) (T_FN 15316) (T_FOR 15317) (T_FOREACH 15318) (T_FUNCTION 15319) (T_FUNC_C 15320) (T_GLOBAL 15321) (T_GOTO 15322) (T_IF 15323) (T_IMPLEMENTS 15324) (T_INCLUDE 15325) (T_INCLUDE_ONCE 15326) (T_INSTANCEOF 15327) (T_INSTEADOF 15328) (T_INTERFACE 15329) (T_ISSET 15330) (T_LINE 15331) (T_LIST 15332) (T_LOGICAL_AND 15333) (T_LOGICAL_OR 15334) (T_LOGICAL_XOR 15335) (T_MATCH 15336) (T_METHOD_C 15337) (T_NAMESPACE 15338) (T_NEW 15339) (T_NS_C 15340) (T_PRINT 15341) (T_PRIVATE 15342) (T_PROTECTED 15343) (T_PUBLIC 15344) (T_READONLY 15345) (T_REQUIRE 15346) (T_REQUIRE_ONCE 15347) (T_RETURN 15348) (T_STATIC 15349) (T_STRING 15350) (T_SWITCH 15351) (T_THROW 15352) (T_TRAIT 15353) (T_TRAIT_C 15354) (T_TRY 15355) (T_UNSET 15356) (T_USE 15357) (T_VAR 15358) (T_VARIABLE 15152) (T_WHILE 15359) (T_YIELD 15360) (identifier 15482) (member_name 15483) (reserved_non_modifiers 15363) (semi_reserved 15364) (simple_variable 15484) ("{" 914)) 6751 (("(" 15366) (argument_list 15485)) 6752 ((T_FN 35) (T_FUNCTION 126) (fn 15166) #'15167 (inline_function 15489)) 6753 (("$" 15106) (T_STRING 15498) (T_VARIABLE 15152) (property_name 15505) (simple_variable 15500) ("{" 15501)) 6754 (("$" 15106) (T_STRING 15498) (T_VARIABLE 15152) (property_name 15499) (simple_variable 15500) ("{" 15501)) 6755 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 15496) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6756 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15494) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6757 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15495)) 6758 (("]" 15497)) 6759 (("(" 15366) (argument_list 15504)) 6760 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15502) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6761 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15503)) 6762 (("(" 15366) (argument_list 15506)) 6763 (("`" 15508)) 6764 (("]" 15510)) 6765 (("=" 15511)) 6766 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15512) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6767 (("%" 15181) ("*" 15182) ("+" 15183) ("-" 15184) ("." 15185) ("/" 15186) ("<" 15187) (">" 15188) ("?" 15189) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15190) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15191) (T_BOOLEAN_AND 15192) (T_BOOLEAN_OR 15193) (T_COALESCE 15194) (T_DOUBLE_ARROW 15515) (T_INSTANCEOF 15195) (T_IS_EQUAL 15196) (T_IS_GREATER_OR_EQUAL 15197) (T_IS_IDENTICAL 15198) (T_IS_NOT_EQUAL 15199) (T_IS_NOT_IDENTICAL 15200) (T_IS_SMALLER_OR_EQUAL 15201) (T_LOGICAL_AND 15202) (T_LOGICAL_OR 15203) (T_LOGICAL_XOR 15204) (T_POW 15205) (T_SL 15206) (T_SPACESHIP 15207) (T_SR 15208) ("^" 15209) ("|" 15210)) 6768 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15516) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6769 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 15525) (T_VARIABLE 2822) (encaps_var 2836)) 6770 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 15524) (T_VARIABLE 2822) (encaps_var 2827)) 6771 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15575) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6772 (("(" 2899) (argument_list 2900) (ctor_arguments 15568)) 6773 ((T_ATTRIBUTE 2853) (T_CLASS 15531) (anonymous_class 15567) (attribute 2898)) 6774 ((T_PAAMAYIM_NEKUDOTAYIM 15565)) 6775 (("(" 15556) (argument_list 15557) (ctor_arguments 15558)) 6776 ((T_NULLSAFE_OBJECT_OPERATOR 15544) (T_OBJECT_OPERATOR 15545) (T_PAAMAYIM_NEKUDOTAYIM 15546) ("[" 15547) ("{" 15548)) 6777 (("$" 15405) (T_STRING 15442) (T_VARIABLE 15406) (property_name 15555) (simple_variable 15444) ("{" 15445)) 6778 (("$" 15405) (T_STRING 15442) (T_VARIABLE 15406) (property_name 15554) (simple_variable 15444) ("{" 15445)) 6779 (("$" 15405) (T_VARIABLE 15406) (simple_variable 15553)) 6780 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 15551) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6781 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15549) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6782 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15550)) 6783 (("]" 15552)) 6784 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 15559) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 15560) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 15561) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6785 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 15564) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6786 (("," 1449) (possible_comma 15562)) 6787 ((")" 15563)) 6788 (("$" 15405) (T_VARIABLE 15406) (simple_variable 15566)) 6789 ((T_EXTENDS 2902) (extends_from 15569)) 6790 ((T_IMPLEMENTS 2904) (implements_list 15570)) 6791 ((backup_doc_comment 15571)) 6792 (("{" 15572)) 6793 ((class_statement_list 15573)) 6794 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 15574)) 6795 (("%" 280) (")" 15576) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6796 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15578) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6797 (("%" 280) (")" 15579) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6798 (("{" 15580)) 6799 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 15581) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 6800 (("}" 15582)) 6801 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 15584) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 6802 ((")" 15585)) 6803 (("=" 15586)) 6804 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 15587) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 6805 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 15589) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6806 (("," 5117) (possible_comma 15590)) 6807 ((")" 15591)) 6808 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 15598) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6809 ((")" 15599)) 6810 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15601) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6811 (("%" 280) (")" 15602) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6812 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15604) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6813 (("%" 280) (")" 15605) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6814 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 15612) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 6815 ((")" 15613)) 6816 (("%" 280) (")" 15618) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6817 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15621) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6818 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15622)) 6819 (("\"" 15624) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 6820 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 15634) (T_VARIABLE 2822) (encaps_var 2836)) 6821 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 15633) (T_VARIABLE 2822) (encaps_var 2827)) 6822 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15684) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6823 (("(" 2899) (argument_list 2900) (ctor_arguments 15677)) 6824 ((T_ATTRIBUTE 2853) (T_CLASS 15640) (anonymous_class 15676) (attribute 2898)) 6825 ((T_PAAMAYIM_NEKUDOTAYIM 15674)) 6826 (("(" 15665) (argument_list 15666) (ctor_arguments 15667)) 6827 ((T_NULLSAFE_OBJECT_OPERATOR 15653) (T_OBJECT_OPERATOR 15654) (T_PAAMAYIM_NEKUDOTAYIM 15655) ("[" 15656) ("{" 15657)) 6828 (("$" 14996) (T_STRING 15033) (T_VARIABLE 14997) (property_name 15664) (simple_variable 15035) ("{" 15036)) 6829 (("$" 14996) (T_STRING 15033) (T_VARIABLE 14997) (property_name 15663) (simple_variable 15035) ("{" 15036)) 6830 (("$" 14996) (T_VARIABLE 14997) (simple_variable 15662)) 6831 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 15660) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 6832 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15658) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6833 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15659)) 6834 (("]" 15661)) 6835 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 15668) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 15669) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 15670) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6836 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 15673) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6837 (("," 1449) (possible_comma 15671)) 6838 ((")" 15672)) 6839 (("$" 14996) (T_VARIABLE 14997) (simple_variable 15675)) 6840 ((T_EXTENDS 2902) (extends_from 15678)) 6841 ((T_IMPLEMENTS 2904) (implements_list 15679)) 6842 ((backup_doc_comment 15680)) 6843 (("{" 15681)) 6844 ((class_statement_list 15682)) 6845 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 15683)) 6846 (("%" 280) (")" 15685) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6847 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15687) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6848 (("%" 280) (")" 15688) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6849 (("{" 15689)) 6850 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 15690) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 6851 (("}" 15691)) 6852 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 15693) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 6853 ((")" 15694)) 6854 (("=" 15695)) 6855 (("!" 14695) ("\"" 14696) ("$" 14697) ("(" 14698) ("+" 14699) ("-" 14700) ("@" 14701) (T_ARRAY 14702) (T_ARRAY_CAST 14703) (T_ATTRIBUTE 125) (T_BOOL_CAST 14704) (T_CLASS_C 14705) (T_CLONE 14706) (T_CONSTANT_ENCAPSED_STRING 14707) (T_DEC 14708) (T_DIR 14709) (T_DNUMBER 14710) (T_DOUBLE_CAST 14711) (T_EMPTY 14712) (T_EVAL 14713) (T_EXIT 14714) (T_FILE 14715) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 14716) (T_INC 14717) (T_INCLUDE 14718) (T_INCLUDE_ONCE 14719) (T_INT_CAST 14720) (T_ISSET 14721) (T_LINE 14722) (T_LIST 14723) (T_LNUMBER 14724) (T_MATCH 14725) (T_METHOD_C 14726) (T_NAME_FULLY_QUALIFIED 14727) (T_NAME_QUALIFIED 14728) (T_NAME_RELATIVE 14729) (T_NEW 14730) (T_NS_C 14731) (T_OBJECT_CAST 14732) (T_PRINT 14733) (T_REQUIRE 14734) (T_REQUIRE_ONCE 14735) (T_START_HEREDOC 14736) (T_STATIC 14737) (T_STRING 14738) (T_STRING_CAST 14739) (T_THROW 14740) (T_TRAIT_C 14741) (T_UNSET_CAST 14742) (T_VARIABLE 14743) (T_YIELD 14744) (T_YIELD_FROM 14745) ("[" 14746) ("`" 14747) (array_object_dereferenceable 14748) (attribute 129) (attributes 14749) (callable_expr 14750) (callable_variable 14751) (class_constant 14752) (class_name 14753) (constant 14754) (dereferenceable_scalar 14755) (expr 15696) (fn 14757) (fully_dereferenceable 103) #'14758 (function_call 14759) (inline_function 14760) (internal_functions_in_yacc 14761) (match 14762) (name 14763) (new_expr 14764) (scalar 14765) (simple_variable 14766) (static_member 14767) (variable 14768) (variable_class_name 14769) ("~" 14770)) 6856 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 15698) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 6857 (("," 5117) (possible_comma 15699)) 6858 ((")" 15700)) 6859 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 15707) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6860 ((")" 15708)) 6861 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15710) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6862 (("%" 280) (")" 15711) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6863 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15713) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6864 (("%" 280) (")" 15714) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6865 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 15721) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 6866 ((")" 15722)) 6867 (("%" 280) (")" 15727) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6868 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 15730) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 6869 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 15731)) 6870 (("\"" 15733) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 6871 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15736) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15737) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15738) (T_DEC 22) (T_DECLARE 15739) (T_DIR 24) (T_DNUMBER 25) (T_DO 15740) (T_DOUBLE_CAST 27) (T_ECHO 15741) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15742) (T_FOREACH 15743) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15744) (T_GOTO 15745) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15746) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15747) (T_START_HEREDOC 67) (T_STATIC 15748) (T_STRING 15749) (T_STRING_CAST 70) (T_SWITCH 15750) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15751) (T_UNSET 15752) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15753) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15754) (alt_if_stmt_without_else 15755) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15756) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15757) (if_stmt_without_else 15758) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15759) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15760) ("~" 124)) 6872 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 24101) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 6873 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 24099) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 6874 (("(" 24090)) 6875 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24084) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 6876 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 24082) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 6877 (("(" 24069)) 6878 (("(" 24054)) 6879 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 24052) (simple_variable 17412)) 6880 ((T_STRING 24050)) 6881 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 24048) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 6882 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 24046)) 6883 ((":" 24045)) 6884 (("(" 24027)) 6885 (("{" 24010)) 6886 (("(" 24005)) 6887 (("(" 23996)) 6888 ((T_ELSE 15766) (T_ELSEIF 15767) (T_ENDIF 15768)) 6889 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 15765) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 6890 ((T_ELSE 15763) (T_ELSEIF 201)) 6891 ((inner_statement_list 15761)) 6892 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 15762) ("~" 124)) 6893 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15736) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15737) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15738) (T_DEC 22) (T_DECLARE 15739) (T_DIR 24) (T_DNUMBER 25) (T_DO 15740) (T_DOUBLE_CAST 27) (T_ECHO 15741) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15742) (T_FOREACH 15743) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15744) (T_GOTO 15745) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15746) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15747) (T_START_HEREDOC 67) (T_STATIC 15748) (T_STRING 15749) (T_STRING_CAST 70) (T_SWITCH 15750) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15751) (T_UNSET 15752) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15753) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15754) (alt_if_stmt_without_else 15755) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15756) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15757) (if_stmt_without_else 15758) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15764) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15760) ("~" 124)) 6894 ((":" 23992)) 6895 (("(" 15770)) 6896 ((";" 15769)) 6897 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15771) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6898 (("%" 280) (")" 15772) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6899 ((":" 15773)) 6900 ((inner_statement_list 15774)) 6901 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15775) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15776) (T_CLASS 15777) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15778) (T_DEC 22) (T_DECLARE 15779) (T_DIR 24) (T_DNUMBER 25) (T_DO 15780) (T_DOUBLE_CAST 27) (T_ECHO 15781) (T_EMPTY 29) (T_ENUM 15782) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15783) (T_FOREACH 15784) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15785) (T_GOTO 15786) (T_HALT_COMPILER 15787) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15789) (T_INTERFACE 15790) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15791) (T_START_HEREDOC 67) (T_STATIC 15792) (T_STRING 15793) (T_STRING_CAST 70) (T_SWITCH 15794) (T_THROW 72) (T_TRAIT 15795) (T_TRAIT_C 74) (T_TRY 15796) (T_UNSET 15797) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15798) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15799) (alt_if_stmt_without_else 15800) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15801) (attributes 15802) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15803) (class_modifier 95) (class_modifiers 15804) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15805) (expr 15806) (fn 102) (fully_dereferenceable 103) #'15807 (function_call 105) (function_declaration_statement 15808) (if_stmt 15809) (if_stmt_without_else 15810) (inline_function 109) (inner_statement 15811) (interface_declaration_statement 15812) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15813) (static_member 118) (trait_declaration_statement 15814) (variable 121) (variable_class_name 122) ("{" 15815) ("~" 124)) 6902 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23990) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 6903 ((T_STRING 23983)) 6904 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23981) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 6905 (("(" 23972)) 6906 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23966) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 6907 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 23964) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 6908 ((T_STRING 23957)) 6909 (("(" 23944)) 6910 (("(" 23929)) 6911 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 23927) (simple_variable 17412)) 6912 ((T_STRING 23925)) 6913 (("(" 23922)) 6914 (("(" 23918)) 6915 ((T_STRING 23912)) 6916 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23910) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 6917 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 23908)) 6918 ((":" 23907)) 6919 (("(" 23889)) 6920 ((T_STRING 23884)) 6921 (("{" 23867)) 6922 (("(" 23862)) 6923 (("(" 23853)) 6924 ((T_ELSE 15855) (T_ELSEIF 15767) (T_ENDIF 15856)) 6925 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 15777) (T_ENUM 15782) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 15790) (T_STATIC 15850) (T_TRAIT 15795) (attribute 15851) (attributed_statement 15852) (class_declaration_statement 15803) (class_modifier 95) (class_modifiers 15804) (enum_declaration_statement 15805) (fn 102) #'15807 (function_declaration_statement 15808) (inline_function 15853) (interface_declaration_statement 15812) (trait_declaration_statement 15814)) 6926 ((T_ABSTRACT 10) (T_CLASS 15841) (T_FINAL 34) (class_modifier 15842)) 6927 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 15840) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 6928 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15826) (ampersand 15827) (returns_ref 15828)) 6929 ((T_ELSE 15818) (T_ELSEIF 15819)) 6930 ((inner_statement_list 15816)) 6931 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 15817) ("~" 124)) 6932 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15775) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15776) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15778) (T_DEC 22) (T_DECLARE 15779) (T_DIR 24) (T_DNUMBER 25) (T_DO 15780) (T_DOUBLE_CAST 27) (T_ECHO 15781) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15783) (T_FOREACH 15784) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15785) (T_GOTO 15786) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15789) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15791) (T_START_HEREDOC 67) (T_STATIC 15792) (T_STRING 15793) (T_STRING_CAST 70) (T_SWITCH 15794) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15796) (T_UNSET 15797) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15798) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15799) (alt_if_stmt_without_else 15800) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15806) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15809) (if_stmt_without_else 15810) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15824) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15815) ("~" 124)) 6933 (("(" 15820)) 6934 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15821) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 6935 (("%" 280) (")" 15822) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 6936 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15775) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15776) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15778) (T_DEC 22) (T_DECLARE 15779) (T_DIR 24) (T_DNUMBER 25) (T_DO 15780) (T_DOUBLE_CAST 27) (T_ECHO 15781) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15783) (T_FOREACH 15784) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15785) (T_GOTO 15786) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15789) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15791) (T_START_HEREDOC 67) (T_STATIC 15792) (T_STRING 15793) (T_STRING_CAST 70) (T_SWITCH 15794) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15796) (T_UNSET 15797) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15798) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15799) (alt_if_stmt_without_else 15800) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15806) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15809) (if_stmt_without_else 15810) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15823) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15815) ("~" 124)) 6937 ((T_STRING 15829) (backup_doc_comment 137)) 6938 ((backup_doc_comment 15830)) 6939 (("(" 15831)) 6940 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 15832)) 6941 ((")" 15833)) 6942 ((":" 150) (return_type 15834)) 6943 ((backup_fn_flags 15835)) 6944 (("{" 15836)) 6945 ((inner_statement_list 15837)) 6946 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 15838) ("~" 124)) 6947 ((backup_fn_flags 15839)) 6948 ((T_STRING 15843)) 6949 ((T_EXTENDS 2902) (extends_from 15844)) 6950 ((T_IMPLEMENTS 2904) (implements_list 15845)) 6951 ((backup_doc_comment 15846)) 6952 (("{" 15847)) 6953 ((class_statement_list 15848)) 6954 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 15849)) 6955 ((T_FN 35) (T_FUNCTION 126) (fn 102) #'132 (inline_function 15854)) 6956 ((":" 15858)) 6957 ((";" 15857)) 6958 ((inner_statement_list 15859)) 6959 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 15867) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 6960 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23851) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 6961 ((T_STRING 23844)) 6962 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23842) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 6963 (("(" 23833)) 6964 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23827) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 6965 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 23825) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 6966 ((";" 23824)) 6967 ((T_STRING 23817)) 6968 (("(" 23804)) 6969 (("(" 23789)) 6970 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 23787) (simple_variable 17412)) 6971 ((T_STRING 23785)) 6972 (("(" 23782)) 6973 ((T_STRING 23776)) 6974 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23774) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 6975 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 23772)) 6976 ((":" 23771)) 6977 (("(" 23753)) 6978 ((T_STRING 23748)) 6979 (("{" 23731)) 6980 (("(" 23726)) 6981 (("(" 15934)) 6982 ((T_ELSE 15927) (T_ELSEIF 15767) (T_ENDIF 15928)) 6983 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 15862) (T_ENUM 15868) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 15875) (T_STATIC 15850) (T_TRAIT 15880) (attribute 15851) (attributed_statement 15926) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (enum_declaration_statement 15890) (fn 102) #'15892 (function_declaration_statement 15893) (inline_function 15853) (interface_declaration_statement 15897) (trait_declaration_statement 15899)) 6984 ((T_ABSTRACT 10) (T_CLASS 15918) (T_FINAL 34) (class_modifier 15842)) 6985 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 15917) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 6986 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15826) (ampersand 15827) (returns_ref 15905)) 6987 ((T_ELSE 15903) (T_ELSEIF 15819)) 6988 ((inner_statement_list 15901)) 6989 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 15902) ("~" 124)) 6990 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15891) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15904) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 6991 ((T_STRING 15906) (backup_doc_comment 137)) 6992 ((backup_doc_comment 15907)) 6993 (("(" 15908)) 6994 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 15909)) 6995 ((")" 15910)) 6996 ((":" 150) (return_type 15911)) 6997 ((backup_fn_flags 15912)) 6998 (("{" 15913)) 6999 ((inner_statement_list 15914)) 7000 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 15915) ("~" 124)) 7001 ((backup_fn_flags 15916)) 7002 ((T_STRING 15919)) 7003 ((T_EXTENDS 2902) (extends_from 15920)) 7004 ((T_IMPLEMENTS 2904) (implements_list 15921)) 7005 ((backup_doc_comment 15922)) 7006 (("{" 15923)) 7007 ((class_statement_list 15924)) 7008 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 15925)) 7009 ((":" 15930)) 7010 ((";" 15929)) 7011 ((inner_statement_list 15931)) 7012 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 15932) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 7013 ((";" 15933)) 7014 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15935) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7015 (("%" 280) (")" 15936) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7016 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 15937) (";" 15860) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15891) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15938) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 15939) ("{" 15900) ("~" 124)) 7017 ((inner_statement_list 15940)) 7018 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 15948) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 7019 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23724) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7020 ((T_STRING 23717)) 7021 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23715) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7022 (("(" 23706)) 7023 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23700) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 7024 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 23698) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7025 ((";" 23697)) 7026 ((T_STRING 23690)) 7027 (("(" 23677)) 7028 (("(" 23662)) 7029 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 23660) (simple_variable 17412)) 7030 ((T_STRING 23658)) 7031 (("(" 23655)) 7032 (("(" 23651)) 7033 ((T_STRING 23645)) 7034 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23643) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7035 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 23641)) 7036 ((":" 23640)) 7037 (("(" 23622)) 7038 ((T_STRING 23617)) 7039 (("{" 23600)) 7040 (("(" 23595)) 7041 (("(" 23586)) 7042 ((T_ELSE 23579) (T_ELSEIF 15767) (T_ENDIF 23580)) 7043 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 15943) (T_ENUM 15949) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 15957) (T_STATIC 15850) (T_TRAIT 15962) (attribute 15851) (attributed_statement 23578) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (enum_declaration_statement 15972) (fn 102) #'15974 (function_declaration_statement 15975) (inline_function 15853) (interface_declaration_statement 15979) (trait_declaration_statement 15981)) 7044 ((T_ABSTRACT 10) (T_CLASS 23570) (T_FINAL 34) (class_modifier 15842)) 7045 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 23569) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 7046 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15826) (ampersand 15827) (returns_ref 23557)) 7047 ((T_ELSE 15985) (T_ELSEIF 15986)) 7048 ((inner_statement_list 15983)) 7049 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 15984) ("~" 124)) 7050 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15973) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23556) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 7051 (("(" 15987)) 7052 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 15988) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7053 (("%" 280) (")" 15989) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7054 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15990) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15991) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15992) (T_DEC 22) (T_DECLARE 15993) (T_DIR 24) (T_DNUMBER 25) (T_DO 15994) (T_DOUBLE_CAST 27) (T_ECHO 15995) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15996) (T_FOREACH 15997) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15998) (T_GOTO 15999) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16000) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16001) (T_START_HEREDOC 67) (T_STATIC 16002) (T_STRING 16003) (T_STRING_CAST 70) (T_SWITCH 16004) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16005) (T_UNSET 16006) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16007) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16008) (alt_if_stmt_without_else 16009) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16010) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16011) (if_stmt_without_else 16012) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 16013) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16014) ("~" 124)) 7055 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23554) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7056 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23552) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7057 (("(" 23543)) 7058 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23537) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 7059 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 23535) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7060 (("(" 23522)) 7061 (("(" 23507)) 7062 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 23505) (simple_variable 17412)) 7063 ((T_STRING 23503)) 7064 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23501) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7065 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 23499)) 7066 ((":" 23498)) 7067 (("(" 16121)) 7068 (("{" 16094)) 7069 (("(" 16036)) 7070 (("(" 16027)) 7071 ((T_ELSE 16020) (T_ELSEIF 15767) (T_ENDIF 16021)) 7072 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 16019) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 7073 ((T_ELSE 16017) (T_ELSEIF 15986)) 7074 ((inner_statement_list 16015)) 7075 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16016) ("~" 124)) 7076 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15990) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15991) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15992) (T_DEC 22) (T_DECLARE 15993) (T_DIR 24) (T_DNUMBER 25) (T_DO 15994) (T_DOUBLE_CAST 27) (T_ECHO 15995) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15996) (T_FOREACH 15997) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15998) (T_GOTO 15999) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16000) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16001) (T_START_HEREDOC 67) (T_STATIC 16002) (T_STRING 16003) (T_STRING_CAST 70) (T_SWITCH 16004) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16005) (T_UNSET 16006) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16007) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16008) (alt_if_stmt_without_else 16009) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16010) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16011) (if_stmt_without_else 16012) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 16018) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16014) ("~" 124)) 7077 ((":" 16023)) 7078 ((";" 16022)) 7079 ((inner_statement_list 16024)) 7080 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 16025) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 7081 ((";" 16026)) 7082 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 16028) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7083 (("%" 280) (")" 16029) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7084 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 16030) (";" 15990) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15991) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15992) (T_DEC 22) (T_DECLARE 15993) (T_DIR 24) (T_DNUMBER 25) (T_DO 15994) (T_DOUBLE_CAST 27) (T_ECHO 15995) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15996) (T_FOREACH 15997) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15998) (T_GOTO 15999) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16000) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16001) (T_START_HEREDOC 67) (T_STATIC 16002) (T_STRING 16003) (T_STRING_CAST 70) (T_SWITCH 16004) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16005) (T_UNSET 16006) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16007) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16008) (alt_if_stmt_without_else 16009) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16010) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16011) (if_stmt_without_else 16012) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 16031) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 16032) ("{" 16014) ("~" 124)) 7085 ((inner_statement_list 16033)) 7086 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 16034) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 7087 ((";" 16035)) 7088 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 16048) (variable 16049) (variable_class_name 16050)) 7089 (("$" 16037) (T_VARIABLE 16038) (simple_variable 16090) ("{" 16091)) 7090 ((T_NULLSAFE_OBJECT_OPERATOR 16073) (T_OBJECT_OPERATOR 16074) ("[" 16075) ("{" 16076)) 7091 (("(" 16054) (argument_list 16072)) 7092 ((T_PAAMAYIM_NEKUDOTAYIM 16068)) 7093 (("(" 16054) (argument_list 16067)) 7094 (("," 16062) (possible_comma 16063)) 7095 ((T_PAAMAYIM_NEKUDOTAYIM 16051)) 7096 (("$" 16037) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 16038) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 16052) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 16053) ("{" 914)) 7097 (("(" 16054) (argument_list 16055)) 7098 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 16056) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 16057) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 16058) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7099 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 16061) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7100 (("," 1449) (possible_comma 16059)) 7101 ((")" 16060)) 7102 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16066) (variable 16049) (variable_class_name 16050)) 7103 ((")" 16064)) 7104 ((";" 16065)) 7105 (("$" 16037) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 16038) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 16069) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 16070) ("{" 914)) 7106 (("(" 16054) (argument_list 16071)) 7107 (("$" 16037) (T_STRING 16081) (T_VARIABLE 16038) (property_name 16088) (simple_variable 16083) ("{" 16084)) 7108 (("$" 16037) (T_STRING 16081) (T_VARIABLE 16038) (property_name 16082) (simple_variable 16083) ("{" 16084)) 7109 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 16079) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 7110 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16077) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7111 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16078)) 7112 (("]" 16080)) 7113 (("(" 16054) (argument_list 16087)) 7114 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16085) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7115 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16086)) 7116 (("(" 16054) (argument_list 16089)) 7117 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16092) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7118 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16093)) 7119 ((inner_statement_list 16095)) 7120 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16096) ("~" 124)) 7121 ((catch_list 16097)) 7122 ((T_CATCH 16098) (T_FINALLY 16099) (finally_statement 16100)) 7123 (("(" 16104)) 7124 (("{" 16101)) 7125 ((inner_statement_list 16102)) 7126 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16103) ("~" 124)) 7127 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 16110) (class_name 16111) (name 16112)) 7128 ((T_VARIABLE 16113) (optional_variable 16114) ("|" 16115)) 7129 ((")" 16117)) 7130 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (class_name 16116) (name 16112)) 7131 (("{" 16118)) 7132 ((inner_statement_list 16119)) 7133 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16120) ("~" 124)) 7134 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 16122) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7135 (("%" 280) (")" 16123) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7136 ((":" 16124) (switch_case_list 16125) ("{" 16126)) 7137 ((";" 23491) (case_list 23492)) 7138 ((";" 16127) (case_list 16128)) 7139 ((case_list 23489)) 7140 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 16131)) 7141 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23486) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 7142 ((":" 16132) (";" 16133) (case_separator 16134)) 7143 ((inner_statement_list 16135)) 7144 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 16136) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 16137) (T_CLASS 16138) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16139) (T_DEC 22) (T_DECLARE 16140) (T_DIR 24) (T_DNUMBER 25) (T_DO 16141) (T_DOUBLE_CAST 27) (T_ECHO 16142) (T_EMPTY 29) (T_ENUM 16143) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 16144) (T_FOREACH 16145) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 16146) (T_GOTO 16147) (T_HALT_COMPILER 16148) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16150) (T_INTERFACE 16151) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16152) (T_START_HEREDOC 67) (T_STATIC 16153) (T_STRING 16154) (T_STRING_CAST 70) (T_SWITCH 16155) (T_THROW 72) (T_TRAIT 16156) (T_TRAIT_C 74) (T_TRY 16157) (T_UNSET 16158) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16159) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16160) (alt_if_stmt_without_else 16161) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 16162) (attributes 16163) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 16164) (class_modifier 95) (class_modifiers 16165) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 16166) (expr 16167) (fn 102) (fully_dereferenceable 103) #'16168 (function_call 105) (function_declaration_statement 16169) (if_stmt 16170) (if_stmt_without_else 16171) (inline_function 109) (inner_statement 16172) (interface_declaration_statement 16173) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 16174) (static_member 118) (trait_declaration_statement 16175) (variable 121) (variable_class_name 122) ("{" 16176) ("~" 124)) 7145 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23484) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7146 ((T_STRING 23477)) 7147 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23475) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7148 (("(" 23466)) 7149 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23460) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 7150 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 23458) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7151 ((T_STRING 23451)) 7152 (("(" 23438)) 7153 (("(" 23423)) 7154 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 23421) (simple_variable 17412)) 7155 ((T_STRING 23419)) 7156 (("(" 23416)) 7157 (("(" 23412)) 7158 ((T_STRING 23406)) 7159 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23404) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7160 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 23402)) 7161 ((":" 23401)) 7162 (("(" 23383)) 7163 ((T_STRING 23378)) 7164 (("{" 23361)) 7165 (("(" 23356)) 7166 (("(" 23347)) 7167 ((T_ELSE 23340) (T_ELSEIF 15767) (T_ENDIF 23341)) 7168 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 16138) (T_ENUM 16143) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 16151) (T_STATIC 15850) (T_TRAIT 16156) (attribute 15851) (attributed_statement 23339) (class_declaration_statement 16164) (class_modifier 95) (class_modifiers 16165) (enum_declaration_statement 16166) (fn 102) #'16168 (function_declaration_statement 16169) (inline_function 15853) (interface_declaration_statement 16173) (trait_declaration_statement 16175)) 7169 ((T_ABSTRACT 10) (T_CLASS 23331) (T_FINAL 34) (class_modifier 15842)) 7170 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 23330) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 7171 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15826) (ampersand 15827) (returns_ref 23318)) 7172 ((T_ELSE 16179) (T_ELSEIF 16180)) 7173 ((inner_statement_list 16177)) 7174 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16178) ("~" 124)) 7175 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 16136) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16137) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16139) (T_DEC 22) (T_DECLARE 16140) (T_DIR 24) (T_DNUMBER 25) (T_DO 16141) (T_DOUBLE_CAST 27) (T_ECHO 16142) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16144) (T_FOREACH 16145) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16146) (T_GOTO 16147) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16150) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16152) (T_START_HEREDOC 67) (T_STATIC 16153) (T_STRING 16154) (T_STRING_CAST 70) (T_SWITCH 16155) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16157) (T_UNSET 16158) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16159) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16160) (alt_if_stmt_without_else 16161) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16167) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16170) (if_stmt_without_else 16171) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23317) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16176) ("~" 124)) 7176 (("(" 16181)) 7177 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 16182) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7178 (("%" 280) (")" 16183) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7179 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 16184) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16185) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16186) (T_DEC 22) (T_DECLARE 16187) (T_DIR 24) (T_DNUMBER 25) (T_DO 16188) (T_DOUBLE_CAST 27) (T_ECHO 16189) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16190) (T_FOREACH 16191) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16192) (T_GOTO 16193) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16194) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16195) (T_START_HEREDOC 67) (T_STATIC 16196) (T_STRING 16197) (T_STRING_CAST 70) (T_SWITCH 16198) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16199) (T_UNSET 16200) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16201) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16202) (alt_if_stmt_without_else 16203) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16204) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16205) (if_stmt_without_else 16206) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 16207) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16208) ("~" 124)) 7180 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23315) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7181 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23313) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7182 (("(" 23304)) 7183 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23298) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 7184 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 23296) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7185 (("(" 23283)) 7186 (("(" 23268)) 7187 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 23266) (simple_variable 17412)) 7188 ((T_STRING 23264)) 7189 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 23262) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7190 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 23260)) 7191 ((":" 23259)) 7192 (("(" 16252)) 7193 (("{" 16235)) 7194 (("(" 16230)) 7195 (("(" 16221)) 7196 ((T_ELSE 16214) (T_ELSEIF 15767) (T_ENDIF 16215)) 7197 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 16213) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 7198 ((T_ELSE 16211) (T_ELSEIF 16180)) 7199 ((inner_statement_list 16209)) 7200 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16210) ("~" 124)) 7201 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 16184) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16185) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16186) (T_DEC 22) (T_DECLARE 16187) (T_DIR 24) (T_DNUMBER 25) (T_DO 16188) (T_DOUBLE_CAST 27) (T_ECHO 16189) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16190) (T_FOREACH 16191) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16192) (T_GOTO 16193) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16194) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16195) (T_START_HEREDOC 67) (T_STATIC 16196) (T_STRING 16197) (T_STRING_CAST 70) (T_SWITCH 16198) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16199) (T_UNSET 16200) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16201) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16202) (alt_if_stmt_without_else 16203) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16204) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16205) (if_stmt_without_else 16206) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 16212) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16208) ("~" 124)) 7202 ((":" 16217)) 7203 ((";" 16216)) 7204 ((inner_statement_list 16218)) 7205 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 16219) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 7206 ((";" 16220)) 7207 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 16222) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7208 (("%" 280) (")" 16223) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7209 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 16224) (";" 16184) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16185) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16186) (T_DEC 22) (T_DECLARE 16187) (T_DIR 24) (T_DNUMBER 25) (T_DO 16188) (T_DOUBLE_CAST 27) (T_ECHO 16189) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16190) (T_FOREACH 16191) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16192) (T_GOTO 16193) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16194) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16195) (T_START_HEREDOC 67) (T_STATIC 16196) (T_STRING 16197) (T_STRING_CAST 70) (T_SWITCH 16198) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16199) (T_UNSET 16200) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16201) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16202) (alt_if_stmt_without_else 16203) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16204) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16205) (if_stmt_without_else 16206) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 16225) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 16226) ("{" 16208) ("~" 124)) 7210 ((inner_statement_list 16227)) 7211 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 16228) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 7212 ((";" 16229)) 7213 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 16231) (variable 16049) (variable_class_name 16050)) 7214 (("," 16062) (possible_comma 16232)) 7215 ((")" 16233)) 7216 ((";" 16234)) 7217 ((inner_statement_list 16236)) 7218 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16237) ("~" 124)) 7219 ((catch_list 16238)) 7220 ((T_CATCH 16239) (T_FINALLY 16240) (finally_statement 16241)) 7221 (("(" 16245)) 7222 (("{" 16242)) 7223 ((inner_statement_list 16243)) 7224 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16244) ("~" 124)) 7225 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 16246) (class_name 16111) (name 16112)) 7226 ((T_VARIABLE 16113) (optional_variable 16247) ("|" 16115)) 7227 ((")" 16248)) 7228 (("{" 16249)) 7229 ((inner_statement_list 16250)) 7230 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16251) ("~" 124)) 7231 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 16253) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7232 (("%" 280) (")" 16254) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7233 ((":" 16255) (switch_case_list 16256) ("{" 16257)) 7234 ((";" 16263) (case_list 16264)) 7235 ((";" 16258) (case_list 16259)) 7236 ((case_list 16261)) 7237 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 16260)) 7238 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 16262)) 7239 ((case_list 23256)) 7240 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 16267)) 7241 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21290) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 7242 ((":" 16269) (";" 16270) (case_separator 16271)) 7243 ((";" 16268)) 7244 ((inner_statement_list 16272)) 7245 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 16273) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 16274) (T_CLASS 16275) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16276) (T_DEC 22) (T_DECLARE 16277) (T_DIR 24) (T_DNUMBER 25) (T_DO 16278) (T_DOUBLE_CAST 27) (T_ECHO 16279) (T_EMPTY 29) (T_ENUM 16280) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 16281) (T_FOREACH 16282) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 16283) (T_GOTO 16284) (T_HALT_COMPILER 16285) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16287) (T_INTERFACE 16288) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16289) (T_START_HEREDOC 67) (T_STATIC 16290) (T_STRING 16291) (T_STRING_CAST 70) (T_SWITCH 16292) (T_THROW 72) (T_TRAIT 16293) (T_TRAIT_C 74) (T_TRY 16294) (T_UNSET 16295) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16296) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16297) (alt_if_stmt_without_else 16298) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 16299) (attributes 16300) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 16301) (class_modifier 95) (class_modifiers 16302) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 16303) (expr 16304) (fn 102) (fully_dereferenceable 103) #'16305 (function_call 105) (function_declaration_statement 16306) (if_stmt 16307) (if_stmt_without_else 16308) (inline_function 109) (inner_statement 16309) (interface_declaration_statement 16310) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 16311) (static_member 118) (trait_declaration_statement 16312) (variable 121) (variable_class_name 122) ("{" 16313) ("~" 124)) 7246 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 21227) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7247 ((T_STRING 21220)) 7248 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 21218) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7249 (("(" 21209)) 7250 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 21203) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 7251 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 21201) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7252 ((T_STRING 21194)) 7253 (("(" 21181)) 7254 (("(" 21166)) 7255 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 21164) (simple_variable 17412)) 7256 ((T_STRING 21162)) 7257 (("(" 21159)) 7258 (("(" 21155)) 7259 ((T_STRING 21149)) 7260 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 21147) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7261 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 21145)) 7262 ((":" 21144)) 7263 (("(" 21126)) 7264 ((T_STRING 21121)) 7265 (("{" 21104)) 7266 (("(" 21099)) 7267 (("(" 21090)) 7268 ((T_ELSE 21083) (T_ELSEIF 15767) (T_ENDIF 21084)) 7269 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 16275) (T_ENUM 16280) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 16288) (T_STATIC 15850) (T_TRAIT 16293) (attribute 15851) (attributed_statement 21082) (class_declaration_statement 16301) (class_modifier 95) (class_modifiers 16302) (enum_declaration_statement 16303) (fn 102) #'16305 (function_declaration_statement 16306) (inline_function 15853) (interface_declaration_statement 16310) (trait_declaration_statement 16312)) 7270 ((T_ABSTRACT 10) (T_CLASS 21074) (T_FINAL 34) (class_modifier 15842)) 7271 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 21073) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 7272 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15826) (ampersand 15827) (returns_ref 21061)) 7273 ((T_ELSE 16316) (T_ELSEIF 16317)) 7274 ((inner_statement_list 16314)) 7275 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16315) ("~" 124)) 7276 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 16273) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16274) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16276) (T_DEC 22) (T_DECLARE 16277) (T_DIR 24) (T_DNUMBER 25) (T_DO 16278) (T_DOUBLE_CAST 27) (T_ECHO 16279) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16281) (T_FOREACH 16282) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16283) (T_GOTO 16284) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16287) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16289) (T_START_HEREDOC 67) (T_STATIC 16290) (T_STRING 16291) (T_STRING_CAST 70) (T_SWITCH 16292) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16294) (T_UNSET 16295) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16296) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16297) (alt_if_stmt_without_else 16298) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16304) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16307) (if_stmt_without_else 16308) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 21060) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16313) ("~" 124)) 7277 (("(" 16318)) 7278 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 16319) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7279 (("%" 280) (")" 16320) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7280 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 16321) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16322) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16323) (T_DEC 22) (T_DECLARE 16324) (T_DIR 24) (T_DNUMBER 25) (T_DO 16325) (T_DOUBLE_CAST 27) (T_ECHO 16326) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16327) (T_FOREACH 16328) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16329) (T_GOTO 16330) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16331) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16332) (T_START_HEREDOC 67) (T_STATIC 16333) (T_STRING 16334) (T_STRING_CAST 70) (T_SWITCH 16335) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16336) (T_UNSET 16337) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16338) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16339) (alt_if_stmt_without_else 16340) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16341) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16342) (if_stmt_without_else 16343) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 16344) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16345) ("~" 124)) 7281 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 21058) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7282 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 21056) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7283 (("(" 21047)) 7284 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 21041) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 7285 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 21039) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7286 (("(" 21026)) 7287 (("(" 17420)) 7288 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 17411) (simple_variable 17412)) 7289 ((T_STRING 17406)) 7290 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 17404) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 7291 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 16411)) 7292 ((":" 16407)) 7293 (("(" 16389)) 7294 (("{" 16372)) 7295 (("(" 16367)) 7296 (("(" 16358)) 7297 ((T_ELSE 16351) (T_ELSEIF 15767) (T_ENDIF 16352)) 7298 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 16350) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 7299 ((T_ELSE 16348) (T_ELSEIF 16317)) 7300 ((inner_statement_list 16346)) 7301 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16347) ("~" 124)) 7302 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 16321) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16322) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16323) (T_DEC 22) (T_DECLARE 16324) (T_DIR 24) (T_DNUMBER 25) (T_DO 16325) (T_DOUBLE_CAST 27) (T_ECHO 16326) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16327) (T_FOREACH 16328) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16329) (T_GOTO 16330) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16331) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16332) (T_START_HEREDOC 67) (T_STATIC 16333) (T_STRING 16334) (T_STRING_CAST 70) (T_SWITCH 16335) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16336) (T_UNSET 16337) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16338) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16339) (alt_if_stmt_without_else 16340) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16341) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16342) (if_stmt_without_else 16343) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 16349) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16345) ("~" 124)) 7303 ((":" 16354)) 7304 ((";" 16353)) 7305 ((inner_statement_list 16355)) 7306 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 16356) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 7307 ((";" 16357)) 7308 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 16359) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7309 (("%" 280) (")" 16360) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7310 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 16361) (";" 16321) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16322) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16323) (T_DEC 22) (T_DECLARE 16324) (T_DIR 24) (T_DNUMBER 25) (T_DO 16325) (T_DOUBLE_CAST 27) (T_ECHO 16326) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16327) (T_FOREACH 16328) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16329) (T_GOTO 16330) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16331) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16332) (T_START_HEREDOC 67) (T_STATIC 16333) (T_STRING 16334) (T_STRING_CAST 70) (T_SWITCH 16335) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16336) (T_UNSET 16337) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16338) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16339) (alt_if_stmt_without_else 16340) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16341) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16342) (if_stmt_without_else 16343) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 16362) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 16363) ("{" 16345) ("~" 124)) 7311 ((inner_statement_list 16364)) 7312 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 16365) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 7313 ((";" 16366)) 7314 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 16368) (variable 16049) (variable_class_name 16050)) 7315 (("," 16062) (possible_comma 16369)) 7316 ((")" 16370)) 7317 ((";" 16371)) 7318 ((inner_statement_list 16373)) 7319 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16374) ("~" 124)) 7320 ((catch_list 16375)) 7321 ((T_CATCH 16376) (T_FINALLY 16377) (finally_statement 16378)) 7322 (("(" 16382)) 7323 (("{" 16379)) 7324 ((inner_statement_list 16380)) 7325 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16381) ("~" 124)) 7326 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 16383) (class_name 16111) (name 16112)) 7327 ((T_VARIABLE 16113) (optional_variable 16384) ("|" 16115)) 7328 ((")" 16385)) 7329 (("{" 16386)) 7330 ((inner_statement_list 16387)) 7331 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16388) ("~" 124)) 7332 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 16390) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7333 (("%" 280) (")" 16391) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7334 ((":" 16392) (switch_case_list 16393) ("{" 16394)) 7335 ((";" 16400) (case_list 16401)) 7336 ((";" 16395) (case_list 16396)) 7337 ((case_list 16398)) 7338 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 16397)) 7339 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 16399)) 7340 ((case_list 16404)) 7341 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 16402)) 7342 ((";" 16403)) 7343 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 16405)) 7344 ((";" 16406)) 7345 (("=" 16415)) 7346 (("," 16412) (";" 16413)) 7347 ((T_VARIABLE 16408) (static_var 16414)) 7348 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16477) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7349 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17402) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7350 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 17400) (encaps_var 5158)) 7351 (("$" 16418) (T_VARIABLE 16464) (simple_variable 17396) ("{" 17397)) 7352 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 17394) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7353 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17393) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7354 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17392) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7355 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17391) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7356 (("(" 17388)) 7357 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17387) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7358 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17386) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7359 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17385) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7360 (("\"" 1489) ("$" 16717) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16718) ("[" 1508) (array_object_dereferenceable 16719) (callable_expr 16720) (callable_variable 16721) (class_constant 1512) (class_name 16722) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16723) (name 16724) (simple_variable 16725) (static_member 16726) (variable 17384) (variable_class_name 16728)) 7361 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17383) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7362 (("(" 17380)) 7363 (("(" 17377)) 7364 (("(" 17373) (exit_expr 17374)) 7365 (("\"" 1489) ("$" 16717) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16718) ("[" 1508) (array_object_dereferenceable 16719) (callable_expr 16720) (callable_variable 16721) (class_constant 1512) (class_name 16722) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16723) (name 16724) (simple_variable 16725) (static_member 16726) (variable 17372) (variable_class_name 16728)) 7366 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17371) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7367 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17370) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7368 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17369) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7369 (("(" 17365)) 7370 (("(" 17360)) 7371 (("(" 17354)) 7372 (("$" 3272) ("(" 5519) (T_ATTRIBUTE 2853) (T_CLASS 17333) (T_NAME_FULLY_QUALIFIED 5521) (T_NAME_QUALIFIED 5522) (T_NAME_RELATIVE 5523) (T_STATIC 5524) (T_STRING 5525) (T_VARIABLE 3273) (anonymous_class 17334) (attribute 2862) (attributes 17335) (class_name 5528) (class_name_reference 17336) (name 5530) (new_variable 5531) (simple_variable 5532)) 7373 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17332) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7374 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17331) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7375 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17330) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7376 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17329) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7377 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 17324) (T_END_HEREDOC 17325) (T_VARIABLE 2822) (encaps_list 17326) (encaps_var 2824)) 7378 ((T_FN 35) (T_FUNCTION 126) (fn 16478) #'16479 (inline_function 17323)) 7379 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17322) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7380 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17321) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7381 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17320) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7382 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16885) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7383 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16823) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7384 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 16819) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 7385 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 16817) (encaps_list 2729) (encaps_var 2730)) 7386 ((T_NULLSAFE_OBJECT_OPERATOR 16800) (T_OBJECT_OPERATOR 16801) ("[" 16802) ("{" 16803)) 7387 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 16797) (attribute 2654) (fn 16478) #'16479 (inline_function 16798)) 7388 (("(" 16678) (argument_list 16796)) 7389 ((T_PAAMAYIM_NEKUDOTAYIM 16791)) 7390 (("%" 16493) ("*" 16494) ("+" 16495) ("-" 16496) ("." 16497) ("/" 16498) ("<" 16499) (">" 16500) ("?" 16501) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16502) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16503) (T_BOOLEAN_AND 16504) (T_BOOLEAN_OR 16505) (T_COALESCE 16506) (T_INSTANCEOF 16507) (T_IS_EQUAL 16508) (T_IS_GREATER_OR_EQUAL 16509) (T_IS_IDENTICAL 16510) (T_IS_NOT_EQUAL 16511) (T_IS_NOT_IDENTICAL 16512) (T_IS_SMALLER_OR_EQUAL 16513) (T_LOGICAL_AND 16514) (T_LOGICAL_OR 16515) (T_LOGICAL_XOR 16516) (T_POW 16517) (T_SL 16518) (T_SPACESHIP 16519) (T_SR 16520) ("^" 16521) ("|" 16522)) 7391 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 16780)) 7392 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 16768)) 7393 (("(" 16678) (argument_list 16767)) 7394 (("=" 16686) (T_AND_EQUAL 16687) (T_COALESCE_EQUAL 16688) (T_CONCAT_EQUAL 16689) (T_DEC 16690) (T_DIV_EQUAL 16691) (T_INC 16692) (T_MINUS_EQUAL 16693) (T_MOD_EQUAL 16694) (T_MUL_EQUAL 16695) (T_OR_EQUAL 16696) (T_PLUS_EQUAL 16697) (T_POW_EQUAL 16698) (T_SL_EQUAL 16699) (T_SR_EQUAL 16700) (T_XOR_EQUAL 16701)) 7395 ((T_PAAMAYIM_NEKUDOTAYIM 16594)) 7396 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16492) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7397 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16593) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7398 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16592) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7399 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16591) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7400 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16590) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7401 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16589) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7402 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16588) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7403 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16587) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7404 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16586) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7405 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 16581) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 16582) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 7406 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16580) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7407 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16579) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7408 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16578) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7409 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16577) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7410 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16576) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7411 (("$" 16538) ("(" 16539) (T_NAME_FULLY_QUALIFIED 16540) (T_NAME_QUALIFIED 16541) (T_NAME_RELATIVE 16542) (T_STATIC 16543) (T_STRING 16544) (T_VARIABLE 16545) (class_name 16546) (class_name_reference 16547) (name 16548) (new_variable 16549) (simple_variable 16550)) 7412 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16537) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7413 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16536) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7414 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16535) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7415 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16534) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7416 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16533) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7417 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16532) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7418 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16531) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7419 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16530) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7420 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16529) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7421 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16528) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7422 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16527) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7423 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16526) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7424 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16525) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7425 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16524) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7426 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16523) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7427 (("$" 16538) (T_VARIABLE 16545) (simple_variable 16572) ("{" 16573)) 7428 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 16570) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7429 ((T_PAAMAYIM_NEKUDOTAYIM 16568)) 7430 ((T_NULLSAFE_OBJECT_OPERATOR 16551) (T_OBJECT_OPERATOR 16552) (T_PAAMAYIM_NEKUDOTAYIM 16553) ("[" 16554) ("{" 16555)) 7431 (("$" 16538) (T_STRING 16561) (T_VARIABLE 16545) (property_name 16567) (simple_variable 16563) ("{" 16564)) 7432 (("$" 16538) (T_STRING 16561) (T_VARIABLE 16545) (property_name 16562) (simple_variable 16563) ("{" 16564)) 7433 (("$" 16538) (T_VARIABLE 16545) (simple_variable 16560)) 7434 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 16558) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 7435 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16556) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7436 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16557)) 7437 (("]" 16559)) 7438 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16565) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7439 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16566)) 7440 (("$" 16538) (T_VARIABLE 16545) (simple_variable 16569)) 7441 (("%" 280) (")" 16571) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7442 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16574) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7443 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16575)) 7444 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16585) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7445 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 16583) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 7446 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16584) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7447 (("$" 16418) (T_ABSTRACT 16595) (T_ARRAY 16596) (T_AS 16597) (T_BREAK 16598) (T_CALLABLE 16599) (T_CASE 16600) (T_CATCH 16601) (T_CLASS 16602) (T_CLASS_C 16603) (T_CLONE 16604) (T_CONST 16605) (T_CONTINUE 16606) (T_DECLARE 16607) (T_DEFAULT 16608) (T_DIR 16609) (T_DO 16610) (T_ECHO 16611) (T_ELSE 16612) (T_ELSEIF 16613) (T_EMPTY 16614) (T_ENDDECLARE 16615) (T_ENDFOR 16616) (T_ENDFOREACH 16617) (T_ENDIF 16618) (T_ENDSWITCH 16619) (T_ENDWHILE 16620) (T_ENUM 16621) (T_EVAL 16622) (T_EXIT 16623) (T_EXTENDS 16624) (T_FILE 16625) (T_FINAL 16626) (T_FINALLY 16627) (T_FN 16628) (T_FOR 16629) (T_FOREACH 16630) (T_FUNCTION 16631) (T_FUNC_C 16632) (T_GLOBAL 16633) (T_GOTO 16634) (T_IF 16635) (T_IMPLEMENTS 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INSTANCEOF 16639) (T_INSTEADOF 16640) (T_INTERFACE 16641) (T_ISSET 16642) (T_LINE 16643) (T_LIST 16644) (T_LOGICAL_AND 16645) (T_LOGICAL_OR 16646) (T_LOGICAL_XOR 16647) (T_MATCH 16648) (T_METHOD_C 16649) (T_NAMESPACE 16650) (T_NEW 16651) (T_NS_C 16652) (T_PRINT 16653) (T_PRIVATE 16654) (T_PROTECTED 16655) (T_PUBLIC 16656) (T_READONLY 16657) (T_REQUIRE 16658) (T_REQUIRE_ONCE 16659) (T_RETURN 16660) (T_STATIC 16661) (T_STRING 16662) (T_SWITCH 16663) (T_THROW 16664) (T_TRAIT 16665) (T_TRAIT_C 16666) (T_TRY 16667) (T_UNSET 16668) (T_USE 16669) (T_VAR 16670) (T_VARIABLE 16464) (T_WHILE 16671) (T_YIELD 16672) (identifier 16673) (member_name 16674) (reserved_non_modifiers 16675) (semi_reserved 16676) (simple_variable 16677) ("{" 914)) 7448 (("(" 16678) (argument_list 16679)) 7449 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 16680) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 16681) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 16682) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7450 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 16685) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7451 (("," 1449) (possible_comma 16683)) 7452 ((")" 16684)) 7453 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (ampersand 16715) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16716) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7454 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16714) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7455 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16713) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7456 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16712) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7457 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16711) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7458 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16710) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7459 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16709) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7460 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16708) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7461 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16707) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7462 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16706) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7463 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16705) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7464 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16704) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7465 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16703) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7466 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16702) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7467 (("\"" 1489) ("$" 16717) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16718) ("[" 1508) (array_object_dereferenceable 16719) (callable_expr 16720) (callable_variable 16721) (class_constant 1512) (class_name 16722) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16723) (name 16724) (simple_variable 16725) (static_member 16726) (variable 16727) (variable_class_name 16728)) 7468 (("$" 16717) (T_VARIABLE 16718) (simple_variable 16763) ("{" 16764)) 7469 ((T_NULLSAFE_OBJECT_OPERATOR 16746) (T_OBJECT_OPERATOR 16747) ("[" 16748) ("{" 16749)) 7470 (("(" 16732) (argument_list 16745)) 7471 ((T_PAAMAYIM_NEKUDOTAYIM 16741)) 7472 (("(" 16732) (argument_list 16740)) 7473 ((T_PAAMAYIM_NEKUDOTAYIM 16729)) 7474 (("$" 16717) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 16718) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 16730) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 16731) ("{" 914)) 7475 (("(" 16732) (argument_list 16733)) 7476 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 16734) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 16735) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 16736) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7477 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 16739) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7478 (("," 1449) (possible_comma 16737)) 7479 ((")" 16738)) 7480 (("$" 16717) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 16718) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 16742) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 16743) ("{" 914)) 7481 (("(" 16732) (argument_list 16744)) 7482 (("$" 16717) (T_STRING 16754) (T_VARIABLE 16718) (property_name 16761) (simple_variable 16756) ("{" 16757)) 7483 (("$" 16717) (T_STRING 16754) (T_VARIABLE 16718) (property_name 16755) (simple_variable 16756) ("{" 16757)) 7484 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 16752) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 7485 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16750) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7486 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16751)) 7487 (("]" 16753)) 7488 (("(" 16732) (argument_list 16760)) 7489 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16758) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7490 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16759)) 7491 (("(" 16732) (argument_list 16762)) 7492 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16765) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7493 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16766)) 7494 ((backup_doc_comment 16769)) 7495 (("(" 16770)) 7496 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 16771)) 7497 ((")" 16772)) 7498 ((T_USE 148) (lexical_vars 16773)) 7499 ((":" 150) (return_type 16774)) 7500 ((backup_fn_flags 16775)) 7501 (("{" 16776)) 7502 ((inner_statement_list 16777)) 7503 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 16778) ("~" 124)) 7504 ((backup_fn_flags 16779)) 7505 ((backup_doc_comment 16781)) 7506 (("(" 16782)) 7507 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 16783)) 7508 ((")" 16784)) 7509 ((":" 2207) (return_type 16785)) 7510 ((T_DOUBLE_ARROW 16786)) 7511 ((backup_fn_flags 16787)) 7512 ((backup_lex_pos 16788)) 7513 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 16789) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 7514 (("%" 3048) ("*" 3049) ("+" 3050) ("-" 3051) ("." 3052) ("/" 3053) ("<" 3054) (">" 3055) ("?" 3056) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3057) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3058) (T_BOOLEAN_AND 3059) (T_BOOLEAN_OR 3060) (T_COALESCE 3061) (T_INSTANCEOF 3062) (T_IS_EQUAL 3063) (T_IS_GREATER_OR_EQUAL 3064) (T_IS_IDENTICAL 3065) (T_IS_NOT_EQUAL 3066) (T_IS_NOT_IDENTICAL 3067) (T_IS_SMALLER_OR_EQUAL 3068) (T_LOGICAL_AND 3069) (T_LOGICAL_OR 3070) (T_LOGICAL_XOR 3071) (T_POW 3072) (T_SL 3073) (T_SPACESHIP 3074) (T_SR 3075) ("^" 3076) (backup_fn_flags 16790) ("|" 3077)) 7515 (("$" 16418) (T_ABSTRACT 16595) (T_ARRAY 16596) (T_AS 16597) (T_BREAK 16598) (T_CALLABLE 16599) (T_CASE 16600) (T_CATCH 16601) (T_CLASS 16602) (T_CLASS_C 16603) (T_CLONE 16604) (T_CONST 16605) (T_CONTINUE 16606) (T_DECLARE 16607) (T_DEFAULT 16608) (T_DIR 16609) (T_DO 16610) (T_ECHO 16611) (T_ELSE 16612) (T_ELSEIF 16613) (T_EMPTY 16614) (T_ENDDECLARE 16615) (T_ENDFOR 16616) (T_ENDFOREACH 16617) (T_ENDIF 16618) (T_ENDSWITCH 16619) (T_ENDWHILE 16620) (T_ENUM 16621) (T_EVAL 16622) (T_EXIT 16623) (T_EXTENDS 16624) (T_FILE 16625) (T_FINAL 16626) (T_FINALLY 16627) (T_FN 16628) (T_FOR 16629) (T_FOREACH 16630) (T_FUNCTION 16631) (T_FUNC_C 16632) (T_GLOBAL 16633) (T_GOTO 16634) (T_IF 16635) (T_IMPLEMENTS 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INSTANCEOF 16639) (T_INSTEADOF 16640) (T_INTERFACE 16641) (T_ISSET 16642) (T_LINE 16643) (T_LIST 16644) (T_LOGICAL_AND 16645) (T_LOGICAL_OR 16646) (T_LOGICAL_XOR 16647) (T_MATCH 16648) (T_METHOD_C 16649) (T_NAMESPACE 16650) (T_NEW 16651) (T_NS_C 16652) (T_PRINT 16653) (T_PRIVATE 16654) (T_PROTECTED 16655) (T_PUBLIC 16656) (T_READONLY 16657) (T_REQUIRE 16658) (T_REQUIRE_ONCE 16659) (T_RETURN 16660) (T_STATIC 16661) (T_STRING 16662) (T_SWITCH 16663) (T_THROW 16664) (T_TRAIT 16665) (T_TRAIT_C 16666) (T_TRY 16667) (T_UNSET 16668) (T_USE 16669) (T_VAR 16670) (T_VARIABLE 16464) (T_WHILE 16671) (T_YIELD 16672) (identifier 16792) (member_name 16793) (reserved_non_modifiers 16675) (semi_reserved 16676) (simple_variable 16794) ("{" 914)) 7516 (("(" 16678) (argument_list 16795)) 7517 ((T_FN 35) (T_FUNCTION 126) (fn 16478) #'16479 (inline_function 16799)) 7518 (("$" 16418) (T_STRING 16808) (T_VARIABLE 16464) (property_name 16815) (simple_variable 16810) ("{" 16811)) 7519 (("$" 16418) (T_STRING 16808) (T_VARIABLE 16464) (property_name 16809) (simple_variable 16810) ("{" 16811)) 7520 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 16806) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 7521 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16804) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7522 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16805)) 7523 (("]" 16807)) 7524 (("(" 16678) (argument_list 16814)) 7525 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16812) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7526 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16813)) 7527 (("(" 16678) (argument_list 16816)) 7528 (("`" 16818)) 7529 (("]" 16820)) 7530 (("=" 16821)) 7531 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 16822) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7532 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17319) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7533 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 17317) (encaps_var 5158)) 7534 (("$" 16826) (T_VARIABLE 16872) (simple_variable 17313) ("{" 17314)) 7535 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 17311) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7536 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17310) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7537 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17309) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7538 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17308) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7539 (("(" 17305)) 7540 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17304) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7541 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17303) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7542 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17302) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7543 (("\"" 1489) ("$" 17125) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 17126) ("[" 1508) (array_object_dereferenceable 17127) (callable_expr 17128) (callable_variable 17129) (class_constant 1512) (class_name 17130) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 17131) (name 17132) (simple_variable 17133) (static_member 17134) (variable 17301) (variable_class_name 17136)) 7544 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17300) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7545 (("(" 17297)) 7546 (("(" 17294)) 7547 (("(" 17290) (exit_expr 17291)) 7548 (("\"" 1489) ("$" 17125) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 17126) ("[" 1508) (array_object_dereferenceable 17127) (callable_expr 17128) (callable_variable 17129) (class_constant 1512) (class_name 17130) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 17131) (name 17132) (simple_variable 17133) (static_member 17134) (variable 17289) (variable_class_name 17136)) 7549 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17288) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7550 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17287) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7551 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17286) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7552 (("(" 17282)) 7553 (("(" 17277)) 7554 (("(" 17271)) 7555 (("$" 3681) ("(" 3806) (T_ATTRIBUTE 2853) (T_CLASS 17250) (T_NAME_FULLY_QUALIFIED 3808) (T_NAME_QUALIFIED 3809) (T_NAME_RELATIVE 3810) (T_STATIC 3811) (T_STRING 3812) (T_VARIABLE 3682) (anonymous_class 17251) (attribute 2862) (attributes 17252) (class_name 3815) (class_name_reference 17253) (name 3817) (new_variable 3818) (simple_variable 3819)) 7556 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17249) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7557 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17248) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7558 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17247) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7559 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17246) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7560 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 17241) (T_END_HEREDOC 17242) (T_VARIABLE 2822) (encaps_list 17243) (encaps_var 2824)) 7561 ((T_FN 35) (T_FUNCTION 126) (fn 16886) #'16887 (inline_function 17240)) 7562 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17239) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7563 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17238) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7564 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17237) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7565 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17234) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7566 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17233) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7567 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 17229) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 7568 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 17227) (encaps_list 2729) (encaps_var 2730)) 7569 ((T_NULLSAFE_OBJECT_OPERATOR 17210) (T_OBJECT_OPERATOR 17211) ("[" 17212) ("{" 17213)) 7570 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 17207) (attribute 2654) (fn 16886) #'16887 (inline_function 17208)) 7571 (("(" 17086) (argument_list 17206)) 7572 ((T_PAAMAYIM_NEKUDOTAYIM 17201)) 7573 (("%" 16901) ("*" 16902) ("+" 16903) ("-" 16904) ("." 16905) ("/" 16906) ("<" 16907) (">" 16908) ("?" 16909) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16910) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16911) (T_BOOLEAN_AND 16912) (T_BOOLEAN_OR 16913) (T_COALESCE 16914) (T_DOUBLE_ARROW 17199) (T_INSTANCEOF 16915) (T_IS_EQUAL 16916) (T_IS_GREATER_OR_EQUAL 16917) (T_IS_IDENTICAL 16918) (T_IS_NOT_EQUAL 16919) (T_IS_NOT_IDENTICAL 16920) (T_IS_SMALLER_OR_EQUAL 16921) (T_LOGICAL_AND 16922) (T_LOGICAL_OR 16923) (T_LOGICAL_XOR 16924) (T_POW 16925) (T_SL 16926) (T_SPACESHIP 16927) (T_SR 16928) ("^" 16929) ("|" 16930)) 7574 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 17188)) 7575 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 17176)) 7576 (("(" 17086) (argument_list 17175)) 7577 (("=" 17094) (T_AND_EQUAL 17095) (T_COALESCE_EQUAL 17096) (T_CONCAT_EQUAL 17097) (T_DEC 17098) (T_DIV_EQUAL 17099) (T_INC 17100) (T_MINUS_EQUAL 17101) (T_MOD_EQUAL 17102) (T_MUL_EQUAL 17103) (T_OR_EQUAL 17104) (T_PLUS_EQUAL 17105) (T_POW_EQUAL 17106) (T_SL_EQUAL 17107) (T_SR_EQUAL 17108) (T_XOR_EQUAL 17109)) 7578 ((T_PAAMAYIM_NEKUDOTAYIM 17002)) 7579 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16900) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7580 (("%" 16901) ("*" 16902) ("+" 16903) ("-" 16904) ("." 16905) ("/" 16906) ("<" 16907) (">" 16908) ("?" 16909) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16910) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16911) (T_BOOLEAN_AND 16912) (T_BOOLEAN_OR 16913) (T_COALESCE 16914) (T_INSTANCEOF 16915) (T_IS_EQUAL 16916) (T_IS_GREATER_OR_EQUAL 16917) (T_IS_IDENTICAL 16918) (T_IS_NOT_EQUAL 16919) (T_IS_NOT_IDENTICAL 16920) (T_IS_SMALLER_OR_EQUAL 16921) (T_LOGICAL_AND 16922) (T_LOGICAL_OR 16923) (T_LOGICAL_XOR 16924) (T_POW 16925) (T_SL 16926) (T_SPACESHIP 16927) (T_SR 16928) ("^" 16929) ("|" 16930)) 7581 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17001) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7582 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17000) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7583 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16999) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7584 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16998) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7585 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16997) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7586 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16996) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7587 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16995) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7588 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16994) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7589 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 16989) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 16990) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 7590 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16988) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7591 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16987) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7592 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16986) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7593 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16985) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7594 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16984) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7595 (("$" 16946) ("(" 16947) (T_NAME_FULLY_QUALIFIED 16948) (T_NAME_QUALIFIED 16949) (T_NAME_RELATIVE 16950) (T_STATIC 16951) (T_STRING 16952) (T_VARIABLE 16953) (class_name 16954) (class_name_reference 16955) (name 16956) (new_variable 16957) (simple_variable 16958)) 7596 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16945) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7597 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16944) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7598 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16943) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7599 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16942) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7600 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16941) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7601 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16940) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7602 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16939) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7603 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16938) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7604 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16937) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7605 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16936) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7606 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16935) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7607 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16934) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7608 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16933) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7609 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16932) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7610 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16931) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7611 (("$" 16946) (T_VARIABLE 16953) (simple_variable 16980) ("{" 16981)) 7612 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 16978) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7613 ((T_PAAMAYIM_NEKUDOTAYIM 16976)) 7614 ((T_NULLSAFE_OBJECT_OPERATOR 16959) (T_OBJECT_OPERATOR 16960) (T_PAAMAYIM_NEKUDOTAYIM 16961) ("[" 16962) ("{" 16963)) 7615 (("$" 16946) (T_STRING 16969) (T_VARIABLE 16953) (property_name 16975) (simple_variable 16971) ("{" 16972)) 7616 (("$" 16946) (T_STRING 16969) (T_VARIABLE 16953) (property_name 16970) (simple_variable 16971) ("{" 16972)) 7617 (("$" 16946) (T_VARIABLE 16953) (simple_variable 16968)) 7618 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 16966) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 7619 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16964) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7620 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16965)) 7621 (("]" 16967)) 7622 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16973) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7623 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16974)) 7624 (("$" 16946) (T_VARIABLE 16953) (simple_variable 16977)) 7625 (("%" 280) (")" 16979) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7626 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 16982) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7627 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 16983)) 7628 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16993) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7629 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 16991) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 7630 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 16992) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7631 (("$" 16826) (T_ABSTRACT 17003) (T_ARRAY 17004) (T_AS 17005) (T_BREAK 17006) (T_CALLABLE 17007) (T_CASE 17008) (T_CATCH 17009) (T_CLASS 17010) (T_CLASS_C 17011) (T_CLONE 17012) (T_CONST 17013) (T_CONTINUE 17014) (T_DECLARE 17015) (T_DEFAULT 17016) (T_DIR 17017) (T_DO 17018) (T_ECHO 17019) (T_ELSE 17020) (T_ELSEIF 17021) (T_EMPTY 17022) (T_ENDDECLARE 17023) (T_ENDFOR 17024) (T_ENDFOREACH 17025) (T_ENDIF 17026) (T_ENDSWITCH 17027) (T_ENDWHILE 17028) (T_ENUM 17029) (T_EVAL 17030) (T_EXIT 17031) (T_EXTENDS 17032) (T_FILE 17033) (T_FINAL 17034) (T_FINALLY 17035) (T_FN 17036) (T_FOR 17037) (T_FOREACH 17038) (T_FUNCTION 17039) (T_FUNC_C 17040) (T_GLOBAL 17041) (T_GOTO 17042) (T_IF 17043) (T_IMPLEMENTS 17044) (T_INCLUDE 17045) (T_INCLUDE_ONCE 17046) (T_INSTANCEOF 17047) (T_INSTEADOF 17048) (T_INTERFACE 17049) (T_ISSET 17050) (T_LINE 17051) (T_LIST 17052) (T_LOGICAL_AND 17053) (T_LOGICAL_OR 17054) (T_LOGICAL_XOR 17055) (T_MATCH 17056) (T_METHOD_C 17057) (T_NAMESPACE 17058) (T_NEW 17059) (T_NS_C 17060) (T_PRINT 17061) (T_PRIVATE 17062) (T_PROTECTED 17063) (T_PUBLIC 17064) (T_READONLY 17065) (T_REQUIRE 17066) (T_REQUIRE_ONCE 17067) (T_RETURN 17068) (T_STATIC 17069) (T_STRING 17070) (T_SWITCH 17071) (T_THROW 17072) (T_TRAIT 17073) (T_TRAIT_C 17074) (T_TRY 17075) (T_UNSET 17076) (T_USE 17077) (T_VAR 17078) (T_VARIABLE 16872) (T_WHILE 17079) (T_YIELD 17080) (identifier 17081) (member_name 17082) (reserved_non_modifiers 17083) (semi_reserved 17084) (simple_variable 17085) ("{" 914)) 7632 (("(" 17086) (argument_list 17087)) 7633 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 17088) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 17089) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 17090) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7634 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 17093) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7635 (("," 1449) (possible_comma 17091)) 7636 ((")" 17092)) 7637 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (ampersand 17123) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17124) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7638 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17122) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7639 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17121) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7640 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17120) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7641 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17119) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7642 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17118) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7643 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17117) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7644 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17116) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7645 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17115) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7646 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17114) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7647 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17113) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7648 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17112) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7649 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17111) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7650 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17110) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7651 (("\"" 1489) ("$" 17125) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 17126) ("[" 1508) (array_object_dereferenceable 17127) (callable_expr 17128) (callable_variable 17129) (class_constant 1512) (class_name 17130) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 17131) (name 17132) (simple_variable 17133) (static_member 17134) (variable 17135) (variable_class_name 17136)) 7652 (("$" 17125) (T_VARIABLE 17126) (simple_variable 17171) ("{" 17172)) 7653 ((T_NULLSAFE_OBJECT_OPERATOR 17154) (T_OBJECT_OPERATOR 17155) ("[" 17156) ("{" 17157)) 7654 (("(" 17140) (argument_list 17153)) 7655 ((T_PAAMAYIM_NEKUDOTAYIM 17149)) 7656 (("(" 17140) (argument_list 17148)) 7657 ((T_PAAMAYIM_NEKUDOTAYIM 17137)) 7658 (("$" 17125) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 17126) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 17138) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 17139) ("{" 914)) 7659 (("(" 17140) (argument_list 17141)) 7660 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 17142) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 17143) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 17144) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7661 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 17147) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7662 (("," 1449) (possible_comma 17145)) 7663 ((")" 17146)) 7664 (("$" 17125) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 17126) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 17150) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 17151) ("{" 914)) 7665 (("(" 17140) (argument_list 17152)) 7666 (("$" 17125) (T_STRING 17162) (T_VARIABLE 17126) (property_name 17169) (simple_variable 17164) ("{" 17165)) 7667 (("$" 17125) (T_STRING 17162) (T_VARIABLE 17126) (property_name 17163) (simple_variable 17164) ("{" 17165)) 7668 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 17160) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 7669 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17158) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7670 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17159)) 7671 (("]" 17161)) 7672 (("(" 17140) (argument_list 17168)) 7673 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17166) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7674 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17167)) 7675 (("(" 17140) (argument_list 17170)) 7676 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17173) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7677 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17174)) 7678 ((backup_doc_comment 17177)) 7679 (("(" 17178)) 7680 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 17179)) 7681 ((")" 17180)) 7682 ((T_USE 148) (lexical_vars 17181)) 7683 ((":" 150) (return_type 17182)) 7684 ((backup_fn_flags 17183)) 7685 (("{" 17184)) 7686 ((inner_statement_list 17185)) 7687 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 17186) ("~" 124)) 7688 ((backup_fn_flags 17187)) 7689 ((backup_doc_comment 17189)) 7690 (("(" 17190)) 7691 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 17191)) 7692 ((")" 17192)) 7693 ((":" 2207) (return_type 17193)) 7694 ((T_DOUBLE_ARROW 17194)) 7695 ((backup_fn_flags 17195)) 7696 ((backup_lex_pos 17196)) 7697 (("!" 3380) ("\"" 3381) ("$" 3382) ("(" 3383) ("+" 3384) ("-" 3385) ("@" 3386) (T_ARRAY 3387) (T_ARRAY_CAST 3388) (T_ATTRIBUTE 125) (T_BOOL_CAST 3389) (T_CLASS_C 3390) (T_CLONE 3391) (T_CONSTANT_ENCAPSED_STRING 3392) (T_DEC 3393) (T_DIR 3394) (T_DNUMBER 3395) (T_DOUBLE_CAST 3396) (T_EMPTY 3397) (T_EVAL 3398) (T_EXIT 3399) (T_FILE 3400) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3401) (T_INC 3402) (T_INCLUDE 3403) (T_INCLUDE_ONCE 3404) (T_INT_CAST 3405) (T_ISSET 3406) (T_LINE 3407) (T_LIST 3408) (T_LNUMBER 3409) (T_MATCH 3410) (T_METHOD_C 3411) (T_NAME_FULLY_QUALIFIED 3412) (T_NAME_QUALIFIED 3413) (T_NAME_RELATIVE 3414) (T_NEW 3415) (T_NS_C 3416) (T_OBJECT_CAST 3417) (T_PRINT 3418) (T_REQUIRE 3419) (T_REQUIRE_ONCE 3420) (T_START_HEREDOC 3421) (T_STATIC 3422) (T_STRING 3423) (T_STRING_CAST 3424) (T_THROW 3425) (T_TRAIT_C 3426) (T_UNSET_CAST 3427) (T_VARIABLE 3428) (T_YIELD 3429) (T_YIELD_FROM 3430) ("[" 3431) ("`" 3432) (array_object_dereferenceable 3433) (attribute 129) (attributes 3434) (callable_expr 3435) (callable_variable 3436) (class_constant 3437) (class_name 3438) (constant 3439) (dereferenceable_scalar 3440) (expr 17197) (fn 3442) (fully_dereferenceable 103) #'3443 (function_call 3444) (inline_function 3445) (internal_functions_in_yacc 3446) (match 3447) (name 3448) (new_expr 3449) (scalar 3450) (simple_variable 3451) (static_member 3452) (variable 3453) (variable_class_name 3454) ("~" 3455)) 7698 (("%" 3457) ("*" 3458) ("+" 3459) ("-" 3460) ("." 3461) ("/" 3462) ("<" 3463) (">" 3464) ("?" 3465) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3466) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3467) (T_BOOLEAN_AND 3468) (T_BOOLEAN_OR 3469) (T_COALESCE 3470) (T_INSTANCEOF 3471) (T_IS_EQUAL 3472) (T_IS_GREATER_OR_EQUAL 3473) (T_IS_IDENTICAL 3474) (T_IS_NOT_EQUAL 3475) (T_IS_NOT_IDENTICAL 3476) (T_IS_SMALLER_OR_EQUAL 3477) (T_LOGICAL_AND 3478) (T_LOGICAL_OR 3479) (T_LOGICAL_XOR 3480) (T_POW 3481) (T_SL 3482) (T_SPACESHIP 3483) (T_SR 3484) ("^" 3485) (backup_fn_flags 17198) ("|" 3486)) 7699 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17200) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7700 (("$" 16826) (T_ABSTRACT 17003) (T_ARRAY 17004) (T_AS 17005) (T_BREAK 17006) (T_CALLABLE 17007) (T_CASE 17008) (T_CATCH 17009) (T_CLASS 17010) (T_CLASS_C 17011) (T_CLONE 17012) (T_CONST 17013) (T_CONTINUE 17014) (T_DECLARE 17015) (T_DEFAULT 17016) (T_DIR 17017) (T_DO 17018) (T_ECHO 17019) (T_ELSE 17020) (T_ELSEIF 17021) (T_EMPTY 17022) (T_ENDDECLARE 17023) (T_ENDFOR 17024) (T_ENDFOREACH 17025) (T_ENDIF 17026) (T_ENDSWITCH 17027) (T_ENDWHILE 17028) (T_ENUM 17029) (T_EVAL 17030) (T_EXIT 17031) (T_EXTENDS 17032) (T_FILE 17033) (T_FINAL 17034) (T_FINALLY 17035) (T_FN 17036) (T_FOR 17037) (T_FOREACH 17038) (T_FUNCTION 17039) (T_FUNC_C 17040) (T_GLOBAL 17041) (T_GOTO 17042) (T_IF 17043) (T_IMPLEMENTS 17044) (T_INCLUDE 17045) (T_INCLUDE_ONCE 17046) (T_INSTANCEOF 17047) (T_INSTEADOF 17048) (T_INTERFACE 17049) (T_ISSET 17050) (T_LINE 17051) (T_LIST 17052) (T_LOGICAL_AND 17053) (T_LOGICAL_OR 17054) (T_LOGICAL_XOR 17055) (T_MATCH 17056) (T_METHOD_C 17057) (T_NAMESPACE 17058) (T_NEW 17059) (T_NS_C 17060) (T_PRINT 17061) (T_PRIVATE 17062) (T_PROTECTED 17063) (T_PUBLIC 17064) (T_READONLY 17065) (T_REQUIRE 17066) (T_REQUIRE_ONCE 17067) (T_RETURN 17068) (T_STATIC 17069) (T_STRING 17070) (T_SWITCH 17071) (T_THROW 17072) (T_TRAIT 17073) (T_TRAIT_C 17074) (T_TRY 17075) (T_UNSET 17076) (T_USE 17077) (T_VAR 17078) (T_VARIABLE 16872) (T_WHILE 17079) (T_YIELD 17080) (identifier 17202) (member_name 17203) (reserved_non_modifiers 17083) (semi_reserved 17084) (simple_variable 17204) ("{" 914)) 7701 (("(" 17086) (argument_list 17205)) 7702 ((T_FN 35) (T_FUNCTION 126) (fn 16886) #'16887 (inline_function 17209)) 7703 (("$" 16826) (T_STRING 17218) (T_VARIABLE 16872) (property_name 17225) (simple_variable 17220) ("{" 17221)) 7704 (("$" 16826) (T_STRING 17218) (T_VARIABLE 16872) (property_name 17219) (simple_variable 17220) ("{" 17221)) 7705 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 17216) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 7706 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17214) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7707 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17215)) 7708 (("]" 17217)) 7709 (("(" 17086) (argument_list 17224)) 7710 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17222) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7711 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17223)) 7712 (("(" 17086) (argument_list 17226)) 7713 (("`" 17228)) 7714 (("]" 17230)) 7715 (("=" 17231)) 7716 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17232) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7717 (("%" 16901) ("*" 16902) ("+" 16903) ("-" 16904) ("." 16905) ("/" 16906) ("<" 16907) (">" 16908) ("?" 16909) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16910) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16911) (T_BOOLEAN_AND 16912) (T_BOOLEAN_OR 16913) (T_COALESCE 16914) (T_DOUBLE_ARROW 17235) (T_INSTANCEOF 16915) (T_IS_EQUAL 16916) (T_IS_GREATER_OR_EQUAL 16917) (T_IS_IDENTICAL 16918) (T_IS_NOT_EQUAL 16919) (T_IS_NOT_IDENTICAL 16920) (T_IS_SMALLER_OR_EQUAL 16921) (T_LOGICAL_AND 16922) (T_LOGICAL_OR 16923) (T_LOGICAL_XOR 16924) (T_POW 16925) (T_SL 16926) (T_SPACESHIP 16927) (T_SR 16928) ("^" 16929) ("|" 16930)) 7718 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17236) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7719 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 17245) (T_VARIABLE 2822) (encaps_var 2836)) 7720 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 17244) (T_VARIABLE 2822) (encaps_var 2827)) 7721 (("(" 2899) (argument_list 2900) (ctor_arguments 17264)) 7722 ((T_ATTRIBUTE 2853) (T_CLASS 17250) (anonymous_class 17263) (attribute 2898)) 7723 (("(" 17254) (argument_list 17255) (ctor_arguments 17256)) 7724 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 17257) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 17258) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 17259) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7725 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 17262) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7726 (("," 1449) (possible_comma 17260)) 7727 ((")" 17261)) 7728 ((T_EXTENDS 2902) (extends_from 17265)) 7729 ((T_IMPLEMENTS 2904) (implements_list 17266)) 7730 ((backup_doc_comment 17267)) 7731 (("{" 17268)) 7732 ((class_statement_list 17269)) 7733 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 17270)) 7734 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 17272) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7735 (("%" 280) (")" 17273) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7736 (("{" 17274)) 7737 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 17275) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 7738 (("}" 17276)) 7739 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 17278) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 7740 ((")" 17279)) 7741 (("=" 17280)) 7742 (("!" 16824) ("\"" 16825) ("$" 16826) ("(" 16827) ("+" 16828) ("-" 16829) ("@" 16830) (T_ARRAY 16831) (T_ARRAY_CAST 16832) (T_ATTRIBUTE 125) (T_BOOL_CAST 16833) (T_CLASS_C 16834) (T_CLONE 16835) (T_CONSTANT_ENCAPSED_STRING 16836) (T_DEC 16837) (T_DIR 16838) (T_DNUMBER 16839) (T_DOUBLE_CAST 16840) (T_EMPTY 16841) (T_EVAL 16842) (T_EXIT 16843) (T_FILE 16844) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16845) (T_INC 16846) (T_INCLUDE 16847) (T_INCLUDE_ONCE 16848) (T_INT_CAST 16849) (T_ISSET 16850) (T_LINE 16851) (T_LIST 16852) (T_LNUMBER 16853) (T_MATCH 16854) (T_METHOD_C 16855) (T_NAME_FULLY_QUALIFIED 16856) (T_NAME_QUALIFIED 16857) (T_NAME_RELATIVE 16858) (T_NEW 16859) (T_NS_C 16860) (T_OBJECT_CAST 16861) (T_PRINT 16862) (T_REQUIRE 16863) (T_REQUIRE_ONCE 16864) (T_START_HEREDOC 16865) (T_STATIC 16866) (T_STRING 16867) (T_STRING_CAST 16868) (T_THROW 16869) (T_TRAIT_C 16870) (T_UNSET_CAST 16871) (T_VARIABLE 16872) (T_YIELD 16873) (T_YIELD_FROM 16874) ("[" 16875) ("`" 16876) (array_object_dereferenceable 16877) (attribute 129) (attributes 16878) (callable_expr 16879) (callable_variable 16880) (class_constant 16881) (class_name 16882) (constant 16883) (dereferenceable_scalar 16884) (expr 17281) (fn 16886) (fully_dereferenceable 103) #'16887 (function_call 16888) (inline_function 16889) (internal_functions_in_yacc 16890) (match 16891) (name 16892) (new_expr 16893) (scalar 16894) (simple_variable 16895) (static_member 16896) (variable 16897) (variable_class_name 16898) ("~" 16899)) 7743 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 17283) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7744 (("," 5117) (possible_comma 17284)) 7745 ((")" 17285)) 7746 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 17292) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7747 ((")" 17293)) 7748 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 17295) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7749 (("%" 280) (")" 17296) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7750 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 17298) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7751 (("%" 280) (")" 17299) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7752 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 17306) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 7753 ((")" 17307)) 7754 (("%" 280) (")" 17312) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7755 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17315) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7756 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17316)) 7757 (("\"" 17318) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 7758 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 17328) (T_VARIABLE 2822) (encaps_var 2836)) 7759 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 17327) (T_VARIABLE 2822) (encaps_var 2827)) 7760 (("(" 2899) (argument_list 2900) (ctor_arguments 17347)) 7761 ((T_ATTRIBUTE 2853) (T_CLASS 17333) (anonymous_class 17346) (attribute 2898)) 7762 (("(" 17337) (argument_list 17338) (ctor_arguments 17339)) 7763 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 17340) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 17341) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 17342) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7764 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 17345) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7765 (("," 1449) (possible_comma 17343)) 7766 ((")" 17344)) 7767 ((T_EXTENDS 2902) (extends_from 17348)) 7768 ((T_IMPLEMENTS 2904) (implements_list 17349)) 7769 ((backup_doc_comment 17350)) 7770 (("{" 17351)) 7771 ((class_statement_list 17352)) 7772 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 17353)) 7773 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 17355) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7774 (("%" 280) (")" 17356) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7775 (("{" 17357)) 7776 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 17358) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 7777 (("}" 17359)) 7778 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 17361) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 7779 ((")" 17362)) 7780 (("=" 17363)) 7781 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 17364) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 7782 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 17366) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7783 (("," 5117) (possible_comma 17367)) 7784 ((")" 17368)) 7785 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 17375) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7786 ((")" 17376)) 7787 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 17378) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7788 (("%" 280) (")" 17379) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7789 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 17381) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7790 (("%" 280) (")" 17382) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7791 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 17389) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 7792 ((")" 17390)) 7793 (("%" 280) (")" 17395) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7794 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17398) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7795 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17399)) 7796 (("\"" 17401) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 7797 ((";" 17405)) 7798 ((";" 17407)) 7799 (("$" 17408) (T_VARIABLE 17409) (simple_variable 17416) ("{" 17417)) 7800 (("," 17413) (";" 17414)) 7801 (("$" 17408) (T_VARIABLE 17409) (global_var 17415) (simple_variable 17412)) 7802 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17418) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7803 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17419)) 7804 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17482) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7805 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 21025) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7806 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 21023) (encaps_var 5158)) 7807 (("$" 17423) (T_VARIABLE 17469) (simple_variable 21019) ("{" 21020)) 7808 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 21017) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7809 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 21016) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7810 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 21015) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7811 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 21014) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7812 (("(" 21011)) 7813 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 21010) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7814 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 21009) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7815 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 21008) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7816 (("\"" 1489) ("$" 17722) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 17723) ("[" 1508) (array_object_dereferenceable 17724) (callable_expr 17725) (callable_variable 17726) (class_constant 1512) (class_name 17727) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 17728) (name 17729) (simple_variable 17730) (static_member 17731) (variable 21007) (variable_class_name 17733)) 7817 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 21006) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7818 (("(" 21003)) 7819 (("(" 21000)) 7820 (("(" 20996) (exit_expr 20997)) 7821 (("\"" 1489) ("$" 17722) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 17723) ("[" 1508) (array_object_dereferenceable 17724) (callable_expr 17725) (callable_variable 17726) (class_constant 1512) (class_name 17727) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 17728) (name 17729) (simple_variable 17730) (static_member 17731) (variable 20995) (variable_class_name 17733)) 7822 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20994) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7823 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20993) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7824 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20992) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7825 (("(" 20988)) 7826 (("(" 20983)) 7827 (("(" 20977)) 7828 (("$" 18095) ("(" 18738) (T_ATTRIBUTE 2853) (T_CLASS 20956) (T_NAME_FULLY_QUALIFIED 18740) (T_NAME_QUALIFIED 18741) (T_NAME_RELATIVE 18742) (T_STATIC 18743) (T_STRING 18744) (T_VARIABLE 18096) (anonymous_class 20957) (attribute 2862) (attributes 20958) (class_name 18747) (class_name_reference 20959) (name 18749) (new_variable 18750) (simple_variable 18751)) 7829 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20955) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7830 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20954) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7831 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20953) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7832 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20952) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7833 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 20947) (T_END_HEREDOC 20948) (T_VARIABLE 2822) (encaps_list 20949) (encaps_var 2824)) 7834 ((T_FN 35) (T_FUNCTION 126) (fn 17483) #'17484 (inline_function 20946)) 7835 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20945) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7836 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20944) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7837 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20943) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7838 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20508) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 7839 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20446) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7840 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 20442) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 7841 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 20440) (encaps_list 2729) (encaps_var 2730)) 7842 ((T_NULLSAFE_OBJECT_OPERATOR 20423) (T_OBJECT_OPERATOR 20424) ("[" 20425) ("{" 20426)) 7843 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 20420) (attribute 2654) (fn 17483) #'17484 (inline_function 20421)) 7844 (("(" 17683) (argument_list 20419)) 7845 ((T_PAAMAYIM_NEKUDOTAYIM 20414)) 7846 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 18834) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 7847 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 17785)) 7848 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 17773)) 7849 (("(" 17683) (argument_list 17772)) 7850 (("=" 17691) (T_AND_EQUAL 17692) (T_COALESCE_EQUAL 17693) (T_CONCAT_EQUAL 17694) (T_DEC 17695) (T_DIV_EQUAL 17696) (T_INC 17697) (T_MINUS_EQUAL 17698) (T_MOD_EQUAL 17699) (T_MUL_EQUAL 17700) (T_OR_EQUAL 17701) (T_PLUS_EQUAL 17702) (T_POW_EQUAL 17703) (T_SL_EQUAL 17704) (T_SR_EQUAL 17705) (T_XOR_EQUAL 17706)) 7851 ((T_PAAMAYIM_NEKUDOTAYIM 17599)) 7852 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17497) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7853 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 7854 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17598) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7855 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17597) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7856 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17596) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7857 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17595) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7858 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17594) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7859 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17593) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7860 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17592) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7861 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17591) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7862 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 17586) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 17587) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 7863 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17585) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7864 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17584) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7865 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17583) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7866 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17582) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7867 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17581) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7868 (("$" 17543) ("(" 17544) (T_NAME_FULLY_QUALIFIED 17545) (T_NAME_QUALIFIED 17546) (T_NAME_RELATIVE 17547) (T_STATIC 17548) (T_STRING 17549) (T_VARIABLE 17550) (class_name 17551) (class_name_reference 17552) (name 17553) (new_variable 17554) (simple_variable 17555)) 7869 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17542) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7870 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17541) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7871 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17540) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7872 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17539) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7873 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17538) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7874 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17537) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7875 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17536) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7876 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17535) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7877 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17534) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7878 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17533) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7879 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17532) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7880 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17531) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7881 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17530) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7882 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17529) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7883 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17528) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7884 (("$" 17543) (T_VARIABLE 17550) (simple_variable 17577) ("{" 17578)) 7885 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 17575) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7886 ((T_PAAMAYIM_NEKUDOTAYIM 17573)) 7887 ((T_NULLSAFE_OBJECT_OPERATOR 17556) (T_OBJECT_OPERATOR 17557) (T_PAAMAYIM_NEKUDOTAYIM 17558) ("[" 17559) ("{" 17560)) 7888 (("$" 17543) (T_STRING 17566) (T_VARIABLE 17550) (property_name 17572) (simple_variable 17568) ("{" 17569)) 7889 (("$" 17543) (T_STRING 17566) (T_VARIABLE 17550) (property_name 17567) (simple_variable 17568) ("{" 17569)) 7890 (("$" 17543) (T_VARIABLE 17550) (simple_variable 17565)) 7891 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 17563) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 7892 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17561) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7893 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17562)) 7894 (("]" 17564)) 7895 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17570) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7896 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17571)) 7897 (("$" 17543) (T_VARIABLE 17550) (simple_variable 17574)) 7898 (("%" 280) (")" 17576) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 7899 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17579) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7900 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17580)) 7901 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17590) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7902 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 17588) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 7903 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17589) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7904 (("$" 17423) (T_ABSTRACT 17600) (T_ARRAY 17601) (T_AS 17602) (T_BREAK 17603) (T_CALLABLE 17604) (T_CASE 17605) (T_CATCH 17606) (T_CLASS 17607) (T_CLASS_C 17608) (T_CLONE 17609) (T_CONST 17610) (T_CONTINUE 17611) (T_DECLARE 17612) (T_DEFAULT 17613) (T_DIR 17614) (T_DO 17615) (T_ECHO 17616) (T_ELSE 17617) (T_ELSEIF 17618) (T_EMPTY 17619) (T_ENDDECLARE 17620) (T_ENDFOR 17621) (T_ENDFOREACH 17622) (T_ENDIF 17623) (T_ENDSWITCH 17624) (T_ENDWHILE 17625) (T_ENUM 17626) (T_EVAL 17627) (T_EXIT 17628) (T_EXTENDS 17629) (T_FILE 17630) (T_FINAL 17631) (T_FINALLY 17632) (T_FN 17633) (T_FOR 17634) (T_FOREACH 17635) (T_FUNCTION 17636) (T_FUNC_C 17637) (T_GLOBAL 17638) (T_GOTO 17639) (T_IF 17640) (T_IMPLEMENTS 17641) (T_INCLUDE 17642) (T_INCLUDE_ONCE 17643) (T_INSTANCEOF 17644) (T_INSTEADOF 17645) (T_INTERFACE 17646) (T_ISSET 17647) (T_LINE 17648) (T_LIST 17649) (T_LOGICAL_AND 17650) (T_LOGICAL_OR 17651) (T_LOGICAL_XOR 17652) (T_MATCH 17653) (T_METHOD_C 17654) (T_NAMESPACE 17655) (T_NEW 17656) (T_NS_C 17657) (T_PRINT 17658) (T_PRIVATE 17659) (T_PROTECTED 17660) (T_PUBLIC 17661) (T_READONLY 17662) (T_REQUIRE 17663) (T_REQUIRE_ONCE 17664) (T_RETURN 17665) (T_STATIC 17666) (T_STRING 17667) (T_SWITCH 17668) (T_THROW 17669) (T_TRAIT 17670) (T_TRAIT_C 17671) (T_TRY 17672) (T_UNSET 17673) (T_USE 17674) (T_VAR 17675) (T_VARIABLE 17469) (T_WHILE 17676) (T_YIELD 17677) (identifier 17678) (member_name 17679) (reserved_non_modifiers 17680) (semi_reserved 17681) (simple_variable 17682) ("{" 914)) 7905 (("(" 17683) (argument_list 17684)) 7906 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 17685) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 17686) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 17687) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7907 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 17690) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7908 (("," 1449) (possible_comma 17688)) 7909 ((")" 17689)) 7910 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (ampersand 17720) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17721) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7911 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17719) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7912 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17718) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7913 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17717) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7914 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17716) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7915 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17715) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7916 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17714) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7917 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17713) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7918 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17712) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7919 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17711) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7920 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17710) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7921 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17709) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7922 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17708) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7923 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 17707) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 7924 (("\"" 1489) ("$" 17722) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 17723) ("[" 1508) (array_object_dereferenceable 17724) (callable_expr 17725) (callable_variable 17726) (class_constant 1512) (class_name 17727) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 17728) (name 17729) (simple_variable 17730) (static_member 17731) (variable 17732) (variable_class_name 17733)) 7925 (("$" 17722) (T_VARIABLE 17723) (simple_variable 17768) ("{" 17769)) 7926 ((T_NULLSAFE_OBJECT_OPERATOR 17751) (T_OBJECT_OPERATOR 17752) ("[" 17753) ("{" 17754)) 7927 (("(" 17737) (argument_list 17750)) 7928 ((T_PAAMAYIM_NEKUDOTAYIM 17746)) 7929 (("(" 17737) (argument_list 17745)) 7930 ((T_PAAMAYIM_NEKUDOTAYIM 17734)) 7931 (("$" 17722) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 17723) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 17735) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 17736) ("{" 914)) 7932 (("(" 17737) (argument_list 17738)) 7933 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 17739) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 17740) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 17741) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7934 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 17744) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 7935 (("," 1449) (possible_comma 17742)) 7936 ((")" 17743)) 7937 (("$" 17722) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 17723) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 17747) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 17748) ("{" 914)) 7938 (("(" 17737) (argument_list 17749)) 7939 (("$" 17722) (T_STRING 17759) (T_VARIABLE 17723) (property_name 17766) (simple_variable 17761) ("{" 17762)) 7940 (("$" 17722) (T_STRING 17759) (T_VARIABLE 17723) (property_name 17760) (simple_variable 17761) ("{" 17762)) 7941 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 17757) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 7942 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17755) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7943 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17756)) 7944 (("]" 17758)) 7945 (("(" 17737) (argument_list 17765)) 7946 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17763) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7947 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17764)) 7948 (("(" 17737) (argument_list 17767)) 7949 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17770) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 7950 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17771)) 7951 ((backup_doc_comment 17774)) 7952 (("(" 17775)) 7953 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 17776)) 7954 ((")" 17777)) 7955 ((T_USE 148) (lexical_vars 17778)) 7956 ((":" 150) (return_type 17779)) 7957 ((backup_fn_flags 17780)) 7958 (("{" 17781)) 7959 ((inner_statement_list 17782)) 7960 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 17783) ("~" 124)) 7961 ((backup_fn_flags 17784)) 7962 ((backup_doc_comment 17786)) 7963 (("(" 17787)) 7964 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 17788)) 7965 ((")" 17789)) 7966 ((":" 2207) (return_type 17790)) 7967 ((T_DOUBLE_ARROW 17791)) 7968 ((backup_fn_flags 17792)) 7969 ((backup_lex_pos 17793)) 7970 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17855) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7971 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18833) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7972 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 18831) (encaps_var 5158)) 7973 (("$" 17796) (T_VARIABLE 17842) (simple_variable 18827) ("{" 18828)) 7974 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18825) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 7975 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18824) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7976 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18823) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7977 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18822) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7978 (("(" 18819)) 7979 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18818) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7980 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18817) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7981 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18816) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7982 (("\"" 1489) ("$" 18095) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18096) ("[" 1508) (array_object_dereferenceable 18097) (callable_expr 18098) (callable_variable 18099) (class_constant 1512) (class_name 18100) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 18101) (name 18102) (simple_variable 18103) (static_member 18104) (variable 18815) (variable_class_name 18106)) 7983 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18814) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7984 (("(" 18811)) 7985 (("(" 18808)) 7986 (("(" 18804) (exit_expr 18805)) 7987 (("\"" 1489) ("$" 18095) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18096) ("[" 1508) (array_object_dereferenceable 18097) (callable_expr 18098) (callable_variable 18099) (class_constant 1512) (class_name 18100) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 18101) (name 18102) (simple_variable 18103) (static_member 18104) (variable 18803) (variable_class_name 18106)) 7988 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18802) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7989 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18801) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7990 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18800) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7991 (("(" 18796)) 7992 (("(" 18791)) 7993 (("(" 18785)) 7994 (("$" 18095) ("(" 18738) (T_ATTRIBUTE 2853) (T_CLASS 18739) (T_NAME_FULLY_QUALIFIED 18740) (T_NAME_QUALIFIED 18741) (T_NAME_RELATIVE 18742) (T_STATIC 18743) (T_STRING 18744) (T_VARIABLE 18096) (anonymous_class 18745) (attribute 2862) (attributes 18746) (class_name 18747) (class_name_reference 18748) (name 18749) (new_variable 18750) (simple_variable 18751)) 7995 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18737) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7996 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18736) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7997 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18735) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7998 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18734) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 7999 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 18729) (T_END_HEREDOC 18730) (T_VARIABLE 2822) (encaps_list 18731) (encaps_var 2824)) 8000 ((T_FN 35) (T_FUNCTION 126) (fn 17856) #'17857 (inline_function 18728)) 8001 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18727) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8002 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18726) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8003 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18725) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8004 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18264) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8005 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18202) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8006 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 18198) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 8007 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 18196) (encaps_list 2729) (encaps_var 2730)) 8008 ((T_NULLSAFE_OBJECT_OPERATOR 18179) (T_OBJECT_OPERATOR 18180) ("[" 18181) ("{" 18182)) 8009 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 18176) (attribute 2654) (fn 17856) #'17857 (inline_function 18177)) 8010 (("(" 18056) (argument_list 18175)) 8011 ((T_PAAMAYIM_NEKUDOTAYIM 18170)) 8012 (("%" 17871) ("*" 17872) ("+" 17873) ("-" 17874) ("." 17875) ("/" 17876) ("<" 17877) (">" 17878) ("?" 17879) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17880) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17881) (T_BOOLEAN_AND 17882) (T_BOOLEAN_OR 17883) (T_COALESCE 17884) (T_INSTANCEOF 17885) (T_IS_EQUAL 17886) (T_IS_GREATER_OR_EQUAL 17887) (T_IS_IDENTICAL 17888) (T_IS_NOT_EQUAL 17889) (T_IS_NOT_IDENTICAL 17890) (T_IS_SMALLER_OR_EQUAL 17891) (T_LOGICAL_AND 17892) (T_LOGICAL_OR 17893) (T_LOGICAL_XOR 17894) (T_POW 17895) (T_SL 17896) (T_SPACESHIP 17897) (T_SR 17898) ("^" 17899) (backup_fn_flags 18169) ("|" 17900)) 8013 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 18158)) 8014 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 18146)) 8015 (("(" 18056) (argument_list 18145)) 8016 (("=" 18064) (T_AND_EQUAL 18065) (T_COALESCE_EQUAL 18066) (T_CONCAT_EQUAL 18067) (T_DEC 18068) (T_DIV_EQUAL 18069) (T_INC 18070) (T_MINUS_EQUAL 18071) (T_MOD_EQUAL 18072) (T_MUL_EQUAL 18073) (T_OR_EQUAL 18074) (T_PLUS_EQUAL 18075) (T_POW_EQUAL 18076) (T_SL_EQUAL 18077) (T_SR_EQUAL 18078) (T_XOR_EQUAL 18079)) 8017 ((T_PAAMAYIM_NEKUDOTAYIM 17972)) 8018 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17870) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8019 (("%" 17871) ("*" 17872) ("+" 17873) ("-" 17874) ("." 17875) ("/" 17876) ("<" 17877) (">" 17878) ("?" 17879) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17880) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17881) (T_BOOLEAN_AND 17882) (T_BOOLEAN_OR 17883) (T_COALESCE 17884) (T_INSTANCEOF 17885) (T_IS_EQUAL 17886) (T_IS_GREATER_OR_EQUAL 17887) (T_IS_IDENTICAL 17888) (T_IS_NOT_EQUAL 17889) (T_IS_NOT_IDENTICAL 17890) (T_IS_SMALLER_OR_EQUAL 17891) (T_LOGICAL_AND 17892) (T_LOGICAL_OR 17893) (T_LOGICAL_XOR 17894) (T_POW 17895) (T_SL 17896) (T_SPACESHIP 17897) (T_SR 17898) ("^" 17899) ("|" 17900)) 8020 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17971) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8021 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17970) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8022 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17969) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8023 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17968) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8024 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17967) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8025 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17966) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8026 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17965) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8027 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17964) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8028 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 17959) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 17960) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 8029 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17958) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8030 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17957) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8031 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17956) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8032 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17955) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8033 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17954) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8034 (("$" 17916) ("(" 17917) (T_NAME_FULLY_QUALIFIED 17918) (T_NAME_QUALIFIED 17919) (T_NAME_RELATIVE 17920) (T_STATIC 17921) (T_STRING 17922) (T_VARIABLE 17923) (class_name 17924) (class_name_reference 17925) (name 17926) (new_variable 17927) (simple_variable 17928)) 8035 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17915) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8036 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17914) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8037 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17913) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8038 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17912) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8039 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17911) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8040 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17910) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8041 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17909) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8042 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17908) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8043 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17907) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8044 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17906) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8045 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17905) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8046 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17904) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8047 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17903) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8048 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17902) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8049 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17901) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8050 (("$" 17916) (T_VARIABLE 17923) (simple_variable 17950) ("{" 17951)) 8051 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 17948) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8052 ((T_PAAMAYIM_NEKUDOTAYIM 17946)) 8053 ((T_NULLSAFE_OBJECT_OPERATOR 17929) (T_OBJECT_OPERATOR 17930) (T_PAAMAYIM_NEKUDOTAYIM 17931) ("[" 17932) ("{" 17933)) 8054 (("$" 17916) (T_STRING 17939) (T_VARIABLE 17923) (property_name 17945) (simple_variable 17941) ("{" 17942)) 8055 (("$" 17916) (T_STRING 17939) (T_VARIABLE 17923) (property_name 17940) (simple_variable 17941) ("{" 17942)) 8056 (("$" 17916) (T_VARIABLE 17923) (simple_variable 17938)) 8057 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 17936) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 8058 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17934) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8059 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17935)) 8060 (("]" 17937)) 8061 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17943) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8062 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17944)) 8063 (("$" 17916) (T_VARIABLE 17923) (simple_variable 17947)) 8064 (("%" 280) (")" 17949) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8065 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 17952) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8066 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 17953)) 8067 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17963) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8068 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 17961) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 8069 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 17962) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8070 (("$" 17796) (T_ABSTRACT 17973) (T_ARRAY 17974) (T_AS 17975) (T_BREAK 17976) (T_CALLABLE 17977) (T_CASE 17978) (T_CATCH 17979) (T_CLASS 17980) (T_CLASS_C 17981) (T_CLONE 17982) (T_CONST 17983) (T_CONTINUE 17984) (T_DECLARE 17985) (T_DEFAULT 17986) (T_DIR 17987) (T_DO 17988) (T_ECHO 17989) (T_ELSE 17990) (T_ELSEIF 17991) (T_EMPTY 17992) (T_ENDDECLARE 17993) (T_ENDFOR 17994) (T_ENDFOREACH 17995) (T_ENDIF 17996) (T_ENDSWITCH 17997) (T_ENDWHILE 17998) (T_ENUM 17999) (T_EVAL 18000) (T_EXIT 18001) (T_EXTENDS 18002) (T_FILE 18003) (T_FINAL 18004) (T_FINALLY 18005) (T_FN 18006) (T_FOR 18007) (T_FOREACH 18008) (T_FUNCTION 18009) (T_FUNC_C 18010) (T_GLOBAL 18011) (T_GOTO 18012) (T_IF 18013) (T_IMPLEMENTS 18014) (T_INCLUDE 18015) (T_INCLUDE_ONCE 18016) (T_INSTANCEOF 18017) (T_INSTEADOF 18018) (T_INTERFACE 18019) (T_ISSET 18020) (T_LINE 18021) (T_LIST 18022) (T_LOGICAL_AND 18023) (T_LOGICAL_OR 18024) (T_LOGICAL_XOR 18025) (T_MATCH 18026) (T_METHOD_C 18027) (T_NAMESPACE 18028) (T_NEW 18029) (T_NS_C 18030) (T_PRINT 18031) (T_PRIVATE 18032) (T_PROTECTED 18033) (T_PUBLIC 18034) (T_READONLY 18035) (T_REQUIRE 18036) (T_REQUIRE_ONCE 18037) (T_RETURN 18038) (T_STATIC 18039) (T_STRING 18040) (T_SWITCH 18041) (T_THROW 18042) (T_TRAIT 18043) (T_TRAIT_C 18044) (T_TRY 18045) (T_UNSET 18046) (T_USE 18047) (T_VAR 18048) (T_VARIABLE 17842) (T_WHILE 18049) (T_YIELD 18050) (identifier 18051) (member_name 18052) (reserved_non_modifiers 18053) (semi_reserved 18054) (simple_variable 18055) ("{" 914)) 8071 (("(" 18056) (argument_list 18057)) 8072 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18058) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 18059) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 18060) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8073 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18063) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8074 (("," 1449) (possible_comma 18061)) 8075 ((")" 18062)) 8076 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (ampersand 18093) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18094) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8077 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18092) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8078 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18091) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8079 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18090) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8080 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18089) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8081 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18088) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8082 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18087) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8083 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18086) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8084 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18085) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8085 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18084) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8086 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18083) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8087 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18082) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8088 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18081) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8089 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18080) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8090 (("\"" 1489) ("$" 18095) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18096) ("[" 1508) (array_object_dereferenceable 18097) (callable_expr 18098) (callable_variable 18099) (class_constant 1512) (class_name 18100) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 18101) (name 18102) (simple_variable 18103) (static_member 18104) (variable 18105) (variable_class_name 18106)) 8091 (("$" 18095) (T_VARIABLE 18096) (simple_variable 18141) ("{" 18142)) 8092 ((T_NULLSAFE_OBJECT_OPERATOR 18124) (T_OBJECT_OPERATOR 18125) ("[" 18126) ("{" 18127)) 8093 (("(" 18110) (argument_list 18123)) 8094 ((T_PAAMAYIM_NEKUDOTAYIM 18119)) 8095 (("(" 18110) (argument_list 18118)) 8096 ((T_PAAMAYIM_NEKUDOTAYIM 18107)) 8097 (("$" 18095) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 18096) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 18108) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 18109) ("{" 914)) 8098 (("(" 18110) (argument_list 18111)) 8099 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18112) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 18113) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 18114) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8100 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18117) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8101 (("," 1449) (possible_comma 18115)) 8102 ((")" 18116)) 8103 (("$" 18095) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 18096) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 18120) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 18121) ("{" 914)) 8104 (("(" 18110) (argument_list 18122)) 8105 (("$" 18095) (T_STRING 18132) (T_VARIABLE 18096) (property_name 18139) (simple_variable 18134) ("{" 18135)) 8106 (("$" 18095) (T_STRING 18132) (T_VARIABLE 18096) (property_name 18133) (simple_variable 18134) ("{" 18135)) 8107 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 18130) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 8108 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18128) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8109 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18129)) 8110 (("]" 18131)) 8111 (("(" 18110) (argument_list 18138)) 8112 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18136) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8113 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18137)) 8114 (("(" 18110) (argument_list 18140)) 8115 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18143) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8116 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18144)) 8117 ((backup_doc_comment 18147)) 8118 (("(" 18148)) 8119 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 18149)) 8120 ((")" 18150)) 8121 ((T_USE 148) (lexical_vars 18151)) 8122 ((":" 150) (return_type 18152)) 8123 ((backup_fn_flags 18153)) 8124 (("{" 18154)) 8125 ((inner_statement_list 18155)) 8126 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 18156) ("~" 124)) 8127 ((backup_fn_flags 18157)) 8128 ((backup_doc_comment 18159)) 8129 (("(" 18160)) 8130 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 18161)) 8131 ((")" 18162)) 8132 ((":" 2207) (return_type 18163)) 8133 ((T_DOUBLE_ARROW 18164)) 8134 ((backup_fn_flags 18165)) 8135 ((backup_lex_pos 18166)) 8136 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18167) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8137 (("%" 17871) ("*" 17872) ("+" 17873) ("-" 17874) ("." 17875) ("/" 17876) ("<" 17877) (">" 17878) ("?" 17879) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17880) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17881) (T_BOOLEAN_AND 17882) (T_BOOLEAN_OR 17883) (T_COALESCE 17884) (T_INSTANCEOF 17885) (T_IS_EQUAL 17886) (T_IS_GREATER_OR_EQUAL 17887) (T_IS_IDENTICAL 17888) (T_IS_NOT_EQUAL 17889) (T_IS_NOT_IDENTICAL 17890) (T_IS_SMALLER_OR_EQUAL 17891) (T_LOGICAL_AND 17892) (T_LOGICAL_OR 17893) (T_LOGICAL_XOR 17894) (T_POW 17895) (T_SL 17896) (T_SPACESHIP 17897) (T_SR 17898) ("^" 17899) (backup_fn_flags 18168) ("|" 17900)) 8138 (("$" 17796) (T_ABSTRACT 17973) (T_ARRAY 17974) (T_AS 17975) (T_BREAK 17976) (T_CALLABLE 17977) (T_CASE 17978) (T_CATCH 17979) (T_CLASS 17980) (T_CLASS_C 17981) (T_CLONE 17982) (T_CONST 17983) (T_CONTINUE 17984) (T_DECLARE 17985) (T_DEFAULT 17986) (T_DIR 17987) (T_DO 17988) (T_ECHO 17989) (T_ELSE 17990) (T_ELSEIF 17991) (T_EMPTY 17992) (T_ENDDECLARE 17993) (T_ENDFOR 17994) (T_ENDFOREACH 17995) (T_ENDIF 17996) (T_ENDSWITCH 17997) (T_ENDWHILE 17998) (T_ENUM 17999) (T_EVAL 18000) (T_EXIT 18001) (T_EXTENDS 18002) (T_FILE 18003) (T_FINAL 18004) (T_FINALLY 18005) (T_FN 18006) (T_FOR 18007) (T_FOREACH 18008) (T_FUNCTION 18009) (T_FUNC_C 18010) (T_GLOBAL 18011) (T_GOTO 18012) (T_IF 18013) (T_IMPLEMENTS 18014) (T_INCLUDE 18015) (T_INCLUDE_ONCE 18016) (T_INSTANCEOF 18017) (T_INSTEADOF 18018) (T_INTERFACE 18019) (T_ISSET 18020) (T_LINE 18021) (T_LIST 18022) (T_LOGICAL_AND 18023) (T_LOGICAL_OR 18024) (T_LOGICAL_XOR 18025) (T_MATCH 18026) (T_METHOD_C 18027) (T_NAMESPACE 18028) (T_NEW 18029) (T_NS_C 18030) (T_PRINT 18031) (T_PRIVATE 18032) (T_PROTECTED 18033) (T_PUBLIC 18034) (T_READONLY 18035) (T_REQUIRE 18036) (T_REQUIRE_ONCE 18037) (T_RETURN 18038) (T_STATIC 18039) (T_STRING 18040) (T_SWITCH 18041) (T_THROW 18042) (T_TRAIT 18043) (T_TRAIT_C 18044) (T_TRY 18045) (T_UNSET 18046) (T_USE 18047) (T_VAR 18048) (T_VARIABLE 17842) (T_WHILE 18049) (T_YIELD 18050) (identifier 18171) (member_name 18172) (reserved_non_modifiers 18053) (semi_reserved 18054) (simple_variable 18173) ("{" 914)) 8139 (("(" 18056) (argument_list 18174)) 8140 ((T_FN 35) (T_FUNCTION 126) (fn 17856) #'17857 (inline_function 18178)) 8141 (("$" 17796) (T_STRING 18187) (T_VARIABLE 17842) (property_name 18194) (simple_variable 18189) ("{" 18190)) 8142 (("$" 17796) (T_STRING 18187) (T_VARIABLE 17842) (property_name 18188) (simple_variable 18189) ("{" 18190)) 8143 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 18185) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 8144 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18183) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8145 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18184)) 8146 (("]" 18186)) 8147 (("(" 18056) (argument_list 18193)) 8148 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18191) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8149 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18192)) 8150 (("(" 18056) (argument_list 18195)) 8151 (("`" 18197)) 8152 (("]" 18199)) 8153 (("=" 18200)) 8154 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18201) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8155 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18724) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8156 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 18722) (encaps_var 5158)) 8157 (("$" 18205) (T_VARIABLE 18251) (simple_variable 18718) ("{" 18719)) 8158 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18716) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8159 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18715) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8160 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18714) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8161 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18713) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8162 (("(" 18710)) 8163 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18709) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8164 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18708) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8165 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18707) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8166 (("\"" 1489) ("$" 18504) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18505) ("[" 1508) (array_object_dereferenceable 18506) (callable_expr 18507) (callable_variable 18508) (class_constant 1512) (class_name 18509) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 18510) (name 18511) (simple_variable 18512) (static_member 18513) (variable 18706) (variable_class_name 18515)) 8167 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18705) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8168 (("(" 18702)) 8169 (("(" 18699)) 8170 (("(" 18695) (exit_expr 18696)) 8171 (("\"" 1489) ("$" 18504) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18505) ("[" 1508) (array_object_dereferenceable 18506) (callable_expr 18507) (callable_variable 18508) (class_constant 1512) (class_name 18509) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 18510) (name 18511) (simple_variable 18512) (static_member 18513) (variable 18694) (variable_class_name 18515)) 8172 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18693) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8173 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18692) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8174 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18691) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8175 (("(" 18687)) 8176 (("(" 18682)) 8177 (("(" 18676)) 8178 (("$" 18504) ("(" 18629) (T_ATTRIBUTE 2853) (T_CLASS 18630) (T_NAME_FULLY_QUALIFIED 18631) (T_NAME_QUALIFIED 18632) (T_NAME_RELATIVE 18633) (T_STATIC 18634) (T_STRING 18635) (T_VARIABLE 18505) (anonymous_class 18636) (attribute 2862) (attributes 18637) (class_name 18638) (class_name_reference 18639) (name 18640) (new_variable 18641) (simple_variable 18642)) 8179 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18628) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8180 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18627) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8181 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18626) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8182 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18625) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8183 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 18620) (T_END_HEREDOC 18621) (T_VARIABLE 2822) (encaps_list 18622) (encaps_var 2824)) 8184 ((T_FN 35) (T_FUNCTION 126) (fn 18265) #'18266 (inline_function 18619)) 8185 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18618) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8186 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18617) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8187 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18616) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8188 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18613) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8189 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18612) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8190 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 18608) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 8191 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 18606) (encaps_list 2729) (encaps_var 2730)) 8192 ((T_NULLSAFE_OBJECT_OPERATOR 18589) (T_OBJECT_OPERATOR 18590) ("[" 18591) ("{" 18592)) 8193 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 18586) (attribute 2654) (fn 18265) #'18266 (inline_function 18587)) 8194 (("(" 18465) (argument_list 18585)) 8195 ((T_PAAMAYIM_NEKUDOTAYIM 18580)) 8196 (("%" 18280) ("*" 18281) ("+" 18282) ("-" 18283) ("." 18284) ("/" 18285) ("<" 18286) (">" 18287) ("?" 18288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18290) (T_BOOLEAN_AND 18291) (T_BOOLEAN_OR 18292) (T_COALESCE 18293) (T_DOUBLE_ARROW 18578) (T_INSTANCEOF 18294) (T_IS_EQUAL 18295) (T_IS_GREATER_OR_EQUAL 18296) (T_IS_IDENTICAL 18297) (T_IS_NOT_EQUAL 18298) (T_IS_NOT_IDENTICAL 18299) (T_IS_SMALLER_OR_EQUAL 18300) (T_LOGICAL_AND 18301) (T_LOGICAL_OR 18302) (T_LOGICAL_XOR 18303) (T_POW 18304) (T_SL 18305) (T_SPACESHIP 18306) (T_SR 18307) ("^" 18308) ("|" 18309)) 8197 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 18567)) 8198 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 18555)) 8199 (("(" 18465) (argument_list 18554)) 8200 (("=" 18473) (T_AND_EQUAL 18474) (T_COALESCE_EQUAL 18475) (T_CONCAT_EQUAL 18476) (T_DEC 18477) (T_DIV_EQUAL 18478) (T_INC 18479) (T_MINUS_EQUAL 18480) (T_MOD_EQUAL 18481) (T_MUL_EQUAL 18482) (T_OR_EQUAL 18483) (T_PLUS_EQUAL 18484) (T_POW_EQUAL 18485) (T_SL_EQUAL 18486) (T_SR_EQUAL 18487) (T_XOR_EQUAL 18488)) 8201 ((T_PAAMAYIM_NEKUDOTAYIM 18381)) 8202 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18279) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8203 (("%" 18280) ("*" 18281) ("+" 18282) ("-" 18283) ("." 18284) ("/" 18285) ("<" 18286) (">" 18287) ("?" 18288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18290) (T_BOOLEAN_AND 18291) (T_BOOLEAN_OR 18292) (T_COALESCE 18293) (T_INSTANCEOF 18294) (T_IS_EQUAL 18295) (T_IS_GREATER_OR_EQUAL 18296) (T_IS_IDENTICAL 18297) (T_IS_NOT_EQUAL 18298) (T_IS_NOT_IDENTICAL 18299) (T_IS_SMALLER_OR_EQUAL 18300) (T_LOGICAL_AND 18301) (T_LOGICAL_OR 18302) (T_LOGICAL_XOR 18303) (T_POW 18304) (T_SL 18305) (T_SPACESHIP 18306) (T_SR 18307) ("^" 18308) ("|" 18309)) 8204 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18380) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8205 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18379) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8206 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18378) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8207 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18377) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8208 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18376) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8209 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18375) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8210 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18374) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8211 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18373) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8212 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 18368) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 18369) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 8213 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18367) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8214 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18366) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8215 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18365) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8216 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18364) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8217 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18363) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8218 (("$" 18325) ("(" 18326) (T_NAME_FULLY_QUALIFIED 18327) (T_NAME_QUALIFIED 18328) (T_NAME_RELATIVE 18329) (T_STATIC 18330) (T_STRING 18331) (T_VARIABLE 18332) (class_name 18333) (class_name_reference 18334) (name 18335) (new_variable 18336) (simple_variable 18337)) 8219 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18324) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8220 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18323) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8221 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18322) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8222 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18321) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8223 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18320) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8224 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18319) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8225 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18318) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8226 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18317) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8227 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18316) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8228 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18315) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8229 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18314) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8230 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18313) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8231 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18312) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8232 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18311) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8233 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18310) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8234 (("$" 18325) (T_VARIABLE 18332) (simple_variable 18359) ("{" 18360)) 8235 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18357) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8236 ((T_PAAMAYIM_NEKUDOTAYIM 18355)) 8237 ((T_NULLSAFE_OBJECT_OPERATOR 18338) (T_OBJECT_OPERATOR 18339) (T_PAAMAYIM_NEKUDOTAYIM 18340) ("[" 18341) ("{" 18342)) 8238 (("$" 18325) (T_STRING 18348) (T_VARIABLE 18332) (property_name 18354) (simple_variable 18350) ("{" 18351)) 8239 (("$" 18325) (T_STRING 18348) (T_VARIABLE 18332) (property_name 18349) (simple_variable 18350) ("{" 18351)) 8240 (("$" 18325) (T_VARIABLE 18332) (simple_variable 18347)) 8241 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 18345) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 8242 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18343) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8243 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18344)) 8244 (("]" 18346)) 8245 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18352) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8246 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18353)) 8247 (("$" 18325) (T_VARIABLE 18332) (simple_variable 18356)) 8248 (("%" 280) (")" 18358) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8249 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18361) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8250 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18362)) 8251 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18372) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8252 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 18370) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 8253 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18371) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8254 (("$" 18205) (T_ABSTRACT 18382) (T_ARRAY 18383) (T_AS 18384) (T_BREAK 18385) (T_CALLABLE 18386) (T_CASE 18387) (T_CATCH 18388) (T_CLASS 18389) (T_CLASS_C 18390) (T_CLONE 18391) (T_CONST 18392) (T_CONTINUE 18393) (T_DECLARE 18394) (T_DEFAULT 18395) (T_DIR 18396) (T_DO 18397) (T_ECHO 18398) (T_ELSE 18399) (T_ELSEIF 18400) (T_EMPTY 18401) (T_ENDDECLARE 18402) (T_ENDFOR 18403) (T_ENDFOREACH 18404) (T_ENDIF 18405) (T_ENDSWITCH 18406) (T_ENDWHILE 18407) (T_ENUM 18408) (T_EVAL 18409) (T_EXIT 18410) (T_EXTENDS 18411) (T_FILE 18412) (T_FINAL 18413) (T_FINALLY 18414) (T_FN 18415) (T_FOR 18416) (T_FOREACH 18417) (T_FUNCTION 18418) (T_FUNC_C 18419) (T_GLOBAL 18420) (T_GOTO 18421) (T_IF 18422) (T_IMPLEMENTS 18423) (T_INCLUDE 18424) (T_INCLUDE_ONCE 18425) (T_INSTANCEOF 18426) (T_INSTEADOF 18427) (T_INTERFACE 18428) (T_ISSET 18429) (T_LINE 18430) (T_LIST 18431) (T_LOGICAL_AND 18432) (T_LOGICAL_OR 18433) (T_LOGICAL_XOR 18434) (T_MATCH 18435) (T_METHOD_C 18436) (T_NAMESPACE 18437) (T_NEW 18438) (T_NS_C 18439) (T_PRINT 18440) (T_PRIVATE 18441) (T_PROTECTED 18442) (T_PUBLIC 18443) (T_READONLY 18444) (T_REQUIRE 18445) (T_REQUIRE_ONCE 18446) (T_RETURN 18447) (T_STATIC 18448) (T_STRING 18449) (T_SWITCH 18450) (T_THROW 18451) (T_TRAIT 18452) (T_TRAIT_C 18453) (T_TRY 18454) (T_UNSET 18455) (T_USE 18456) (T_VAR 18457) (T_VARIABLE 18251) (T_WHILE 18458) (T_YIELD 18459) (identifier 18460) (member_name 18461) (reserved_non_modifiers 18462) (semi_reserved 18463) (simple_variable 18464) ("{" 914)) 8255 (("(" 18465) (argument_list 18466)) 8256 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18467) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 18468) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 18469) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8257 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18472) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8258 (("," 1449) (possible_comma 18470)) 8259 ((")" 18471)) 8260 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (ampersand 18502) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18503) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8261 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18501) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8262 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18500) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8263 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18499) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8264 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18498) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8265 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18497) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8266 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18496) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8267 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18495) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8268 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18494) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8269 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18493) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8270 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18492) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8271 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18491) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8272 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18490) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8273 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18489) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8274 (("\"" 1489) ("$" 18504) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18505) ("[" 1508) (array_object_dereferenceable 18506) (callable_expr 18507) (callable_variable 18508) (class_constant 1512) (class_name 18509) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 18510) (name 18511) (simple_variable 18512) (static_member 18513) (variable 18514) (variable_class_name 18515)) 8275 (("$" 18504) (T_VARIABLE 18505) (simple_variable 18550) ("{" 18551)) 8276 ((T_NULLSAFE_OBJECT_OPERATOR 18533) (T_OBJECT_OPERATOR 18534) ("[" 18535) ("{" 18536)) 8277 (("(" 18519) (argument_list 18532)) 8278 ((T_PAAMAYIM_NEKUDOTAYIM 18528)) 8279 (("(" 18519) (argument_list 18527)) 8280 ((T_PAAMAYIM_NEKUDOTAYIM 18516)) 8281 (("$" 18504) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 18505) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 18517) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 18518) ("{" 914)) 8282 (("(" 18519) (argument_list 18520)) 8283 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18521) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 18522) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 18523) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8284 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18526) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8285 (("," 1449) (possible_comma 18524)) 8286 ((")" 18525)) 8287 (("$" 18504) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 18505) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 18529) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 18530) ("{" 914)) 8288 (("(" 18519) (argument_list 18531)) 8289 (("$" 18504) (T_STRING 18541) (T_VARIABLE 18505) (property_name 18548) (simple_variable 18543) ("{" 18544)) 8290 (("$" 18504) (T_STRING 18541) (T_VARIABLE 18505) (property_name 18542) (simple_variable 18543) ("{" 18544)) 8291 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 18539) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 8292 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18537) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8293 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18538)) 8294 (("]" 18540)) 8295 (("(" 18519) (argument_list 18547)) 8296 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18545) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8297 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18546)) 8298 (("(" 18519) (argument_list 18549)) 8299 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18552) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8300 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18553)) 8301 ((backup_doc_comment 18556)) 8302 (("(" 18557)) 8303 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 18558)) 8304 ((")" 18559)) 8305 ((T_USE 148) (lexical_vars 18560)) 8306 ((":" 150) (return_type 18561)) 8307 ((backup_fn_flags 18562)) 8308 (("{" 18563)) 8309 ((inner_statement_list 18564)) 8310 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 18565) ("~" 124)) 8311 ((backup_fn_flags 18566)) 8312 ((backup_doc_comment 18568)) 8313 (("(" 18569)) 8314 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 18570)) 8315 ((")" 18571)) 8316 ((":" 2207) (return_type 18572)) 8317 ((T_DOUBLE_ARROW 18573)) 8318 ((backup_fn_flags 18574)) 8319 ((backup_lex_pos 18575)) 8320 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18576) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8321 (("%" 18280) ("*" 18281) ("+" 18282) ("-" 18283) ("." 18284) ("/" 18285) ("<" 18286) (">" 18287) ("?" 18288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18290) (T_BOOLEAN_AND 18291) (T_BOOLEAN_OR 18292) (T_COALESCE 18293) (T_INSTANCEOF 18294) (T_IS_EQUAL 18295) (T_IS_GREATER_OR_EQUAL 18296) (T_IS_IDENTICAL 18297) (T_IS_NOT_EQUAL 18298) (T_IS_NOT_IDENTICAL 18299) (T_IS_SMALLER_OR_EQUAL 18300) (T_LOGICAL_AND 18301) (T_LOGICAL_OR 18302) (T_LOGICAL_XOR 18303) (T_POW 18304) (T_SL 18305) (T_SPACESHIP 18306) (T_SR 18307) ("^" 18308) (backup_fn_flags 18577) ("|" 18309)) 8322 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18579) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8323 (("$" 18205) (T_ABSTRACT 18382) (T_ARRAY 18383) (T_AS 18384) (T_BREAK 18385) (T_CALLABLE 18386) (T_CASE 18387) (T_CATCH 18388) (T_CLASS 18389) (T_CLASS_C 18390) (T_CLONE 18391) (T_CONST 18392) (T_CONTINUE 18393) (T_DECLARE 18394) (T_DEFAULT 18395) (T_DIR 18396) (T_DO 18397) (T_ECHO 18398) (T_ELSE 18399) (T_ELSEIF 18400) (T_EMPTY 18401) (T_ENDDECLARE 18402) (T_ENDFOR 18403) (T_ENDFOREACH 18404) (T_ENDIF 18405) (T_ENDSWITCH 18406) (T_ENDWHILE 18407) (T_ENUM 18408) (T_EVAL 18409) (T_EXIT 18410) (T_EXTENDS 18411) (T_FILE 18412) (T_FINAL 18413) (T_FINALLY 18414) (T_FN 18415) (T_FOR 18416) (T_FOREACH 18417) (T_FUNCTION 18418) (T_FUNC_C 18419) (T_GLOBAL 18420) (T_GOTO 18421) (T_IF 18422) (T_IMPLEMENTS 18423) (T_INCLUDE 18424) (T_INCLUDE_ONCE 18425) (T_INSTANCEOF 18426) (T_INSTEADOF 18427) (T_INTERFACE 18428) (T_ISSET 18429) (T_LINE 18430) (T_LIST 18431) (T_LOGICAL_AND 18432) (T_LOGICAL_OR 18433) (T_LOGICAL_XOR 18434) (T_MATCH 18435) (T_METHOD_C 18436) (T_NAMESPACE 18437) (T_NEW 18438) (T_NS_C 18439) (T_PRINT 18440) (T_PRIVATE 18441) (T_PROTECTED 18442) (T_PUBLIC 18443) (T_READONLY 18444) (T_REQUIRE 18445) (T_REQUIRE_ONCE 18446) (T_RETURN 18447) (T_STATIC 18448) (T_STRING 18449) (T_SWITCH 18450) (T_THROW 18451) (T_TRAIT 18452) (T_TRAIT_C 18453) (T_TRY 18454) (T_UNSET 18455) (T_USE 18456) (T_VAR 18457) (T_VARIABLE 18251) (T_WHILE 18458) (T_YIELD 18459) (identifier 18581) (member_name 18582) (reserved_non_modifiers 18462) (semi_reserved 18463) (simple_variable 18583) ("{" 914)) 8324 (("(" 18465) (argument_list 18584)) 8325 ((T_FN 35) (T_FUNCTION 126) (fn 18265) #'18266 (inline_function 18588)) 8326 (("$" 18205) (T_STRING 18597) (T_VARIABLE 18251) (property_name 18604) (simple_variable 18599) ("{" 18600)) 8327 (("$" 18205) (T_STRING 18597) (T_VARIABLE 18251) (property_name 18598) (simple_variable 18599) ("{" 18600)) 8328 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 18595) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 8329 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18593) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8330 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18594)) 8331 (("]" 18596)) 8332 (("(" 18465) (argument_list 18603)) 8333 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18601) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8334 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18602)) 8335 (("(" 18465) (argument_list 18605)) 8336 (("`" 18607)) 8337 (("]" 18609)) 8338 (("=" 18610)) 8339 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18611) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8340 (("%" 18280) ("*" 18281) ("+" 18282) ("-" 18283) ("." 18284) ("/" 18285) ("<" 18286) (">" 18287) ("?" 18288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18290) (T_BOOLEAN_AND 18291) (T_BOOLEAN_OR 18292) (T_COALESCE 18293) (T_DOUBLE_ARROW 18614) (T_INSTANCEOF 18294) (T_IS_EQUAL 18295) (T_IS_GREATER_OR_EQUAL 18296) (T_IS_IDENTICAL 18297) (T_IS_NOT_EQUAL 18298) (T_IS_NOT_IDENTICAL 18299) (T_IS_SMALLER_OR_EQUAL 18300) (T_LOGICAL_AND 18301) (T_LOGICAL_OR 18302) (T_LOGICAL_XOR 18303) (T_POW 18304) (T_SL 18305) (T_SPACESHIP 18306) (T_SR 18307) ("^" 18308) ("|" 18309)) 8341 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18615) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8342 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 18624) (T_VARIABLE 2822) (encaps_var 2836)) 8343 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 18623) (T_VARIABLE 2822) (encaps_var 2827)) 8344 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18674) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8345 (("(" 2899) (argument_list 2900) (ctor_arguments 18667)) 8346 ((T_ATTRIBUTE 2853) (T_CLASS 18630) (anonymous_class 18666) (attribute 2898)) 8347 ((T_PAAMAYIM_NEKUDOTAYIM 18664)) 8348 (("(" 18655) (argument_list 18656) (ctor_arguments 18657)) 8349 ((T_NULLSAFE_OBJECT_OPERATOR 18643) (T_OBJECT_OPERATOR 18644) (T_PAAMAYIM_NEKUDOTAYIM 18645) ("[" 18646) ("{" 18647)) 8350 (("$" 18504) (T_STRING 18541) (T_VARIABLE 18505) (property_name 18654) (simple_variable 18543) ("{" 18544)) 8351 (("$" 18504) (T_STRING 18541) (T_VARIABLE 18505) (property_name 18653) (simple_variable 18543) ("{" 18544)) 8352 (("$" 18504) (T_VARIABLE 18505) (simple_variable 18652)) 8353 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 18650) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 8354 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18648) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8355 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18649)) 8356 (("]" 18651)) 8357 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18658) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 18659) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 18660) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8358 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18663) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8359 (("," 1449) (possible_comma 18661)) 8360 ((")" 18662)) 8361 (("$" 18504) (T_VARIABLE 18505) (simple_variable 18665)) 8362 ((T_EXTENDS 2902) (extends_from 18668)) 8363 ((T_IMPLEMENTS 2904) (implements_list 18669)) 8364 ((backup_doc_comment 18670)) 8365 (("{" 18671)) 8366 ((class_statement_list 18672)) 8367 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 18673)) 8368 (("%" 280) (")" 18675) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8369 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18677) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8370 (("%" 280) (")" 18678) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8371 (("{" 18679)) 8372 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 18680) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 8373 (("}" 18681)) 8374 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 18683) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 8375 ((")" 18684)) 8376 (("=" 18685)) 8377 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 18686) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 8378 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 18688) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8379 (("," 5117) (possible_comma 18689)) 8380 ((")" 18690)) 8381 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 18697) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8382 ((")" 18698)) 8383 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18700) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8384 (("%" 280) (")" 18701) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8385 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18703) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8386 (("%" 280) (")" 18704) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8387 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 18711) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 8388 ((")" 18712)) 8389 (("%" 280) (")" 18717) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8390 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18720) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8391 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18721)) 8392 (("\"" 18723) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 8393 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 18733) (T_VARIABLE 2822) (encaps_var 2836)) 8394 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 18732) (T_VARIABLE 2822) (encaps_var 2827)) 8395 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18783) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8396 (("(" 2899) (argument_list 2900) (ctor_arguments 18776)) 8397 ((T_ATTRIBUTE 2853) (T_CLASS 18739) (anonymous_class 18775) (attribute 2898)) 8398 ((T_PAAMAYIM_NEKUDOTAYIM 18773)) 8399 (("(" 18764) (argument_list 18765) (ctor_arguments 18766)) 8400 ((T_NULLSAFE_OBJECT_OPERATOR 18752) (T_OBJECT_OPERATOR 18753) (T_PAAMAYIM_NEKUDOTAYIM 18754) ("[" 18755) ("{" 18756)) 8401 (("$" 18095) (T_STRING 18132) (T_VARIABLE 18096) (property_name 18763) (simple_variable 18134) ("{" 18135)) 8402 (("$" 18095) (T_STRING 18132) (T_VARIABLE 18096) (property_name 18762) (simple_variable 18134) ("{" 18135)) 8403 (("$" 18095) (T_VARIABLE 18096) (simple_variable 18761)) 8404 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 18759) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 8405 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18757) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8406 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18758)) 8407 (("]" 18760)) 8408 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18767) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 18768) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 18769) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8409 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18772) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8410 (("," 1449) (possible_comma 18770)) 8411 ((")" 18771)) 8412 (("$" 18095) (T_VARIABLE 18096) (simple_variable 18774)) 8413 ((T_EXTENDS 2902) (extends_from 18777)) 8414 ((T_IMPLEMENTS 2904) (implements_list 18778)) 8415 ((backup_doc_comment 18779)) 8416 (("{" 18780)) 8417 ((class_statement_list 18781)) 8418 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 18782)) 8419 (("%" 280) (")" 18784) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8420 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18786) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8421 (("%" 280) (")" 18787) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8422 (("{" 18788)) 8423 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 18789) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 8424 (("}" 18790)) 8425 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 18792) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 8426 ((")" 18793)) 8427 (("=" 18794)) 8428 (("!" 17794) ("\"" 17795) ("$" 17796) ("(" 17797) ("+" 17798) ("-" 17799) ("@" 17800) (T_ARRAY 17801) (T_ARRAY_CAST 17802) (T_ATTRIBUTE 125) (T_BOOL_CAST 17803) (T_CLASS_C 17804) (T_CLONE 17805) (T_CONSTANT_ENCAPSED_STRING 17806) (T_DEC 17807) (T_DIR 17808) (T_DNUMBER 17809) (T_DOUBLE_CAST 17810) (T_EMPTY 17811) (T_EVAL 17812) (T_EXIT 17813) (T_FILE 17814) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17815) (T_INC 17816) (T_INCLUDE 17817) (T_INCLUDE_ONCE 17818) (T_INT_CAST 17819) (T_ISSET 17820) (T_LINE 17821) (T_LIST 17822) (T_LNUMBER 17823) (T_MATCH 17824) (T_METHOD_C 17825) (T_NAME_FULLY_QUALIFIED 17826) (T_NAME_QUALIFIED 17827) (T_NAME_RELATIVE 17828) (T_NEW 17829) (T_NS_C 17830) (T_OBJECT_CAST 17831) (T_PRINT 17832) (T_REQUIRE 17833) (T_REQUIRE_ONCE 17834) (T_START_HEREDOC 17835) (T_STATIC 17836) (T_STRING 17837) (T_STRING_CAST 17838) (T_THROW 17839) (T_TRAIT_C 17840) (T_UNSET_CAST 17841) (T_VARIABLE 17842) (T_YIELD 17843) (T_YIELD_FROM 17844) ("[" 17845) ("`" 17846) (array_object_dereferenceable 17847) (attribute 129) (attributes 17848) (callable_expr 17849) (callable_variable 17850) (class_constant 17851) (class_name 17852) (constant 17853) (dereferenceable_scalar 17854) (expr 18795) (fn 17856) (fully_dereferenceable 103) #'17857 (function_call 17858) (inline_function 17859) (internal_functions_in_yacc 17860) (match 17861) (name 17862) (new_expr 17863) (scalar 17864) (simple_variable 17865) (static_member 17866) (variable 17867) (variable_class_name 17868) ("~" 17869)) 8429 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 18797) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8430 (("," 5117) (possible_comma 18798)) 8431 ((")" 18799)) 8432 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 18806) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8433 ((")" 18807)) 8434 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18809) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8435 (("%" 280) (")" 18810) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8436 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18812) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8437 (("%" 280) (")" 18813) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8438 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 18820) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 8439 ((")" 18821)) 8440 (("%" 280) (")" 18826) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8441 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 18829) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 8442 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 18830)) 8443 (("\"" 18832) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 8444 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 18844) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 8445 (("$" 18835) (T_VARIABLE 18837) (simple_variable 20410) ("{" 20411)) 8446 (("(" 20407)) 8447 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 20405) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 8448 (("\"" 1489) ("$" 18835) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 1508) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 20404) (variable_class_name 18850)) 8449 ((T_NULLSAFE_OBJECT_OPERATOR 20387) (T_OBJECT_OPERATOR 20388) ("[" 20389) ("{" 20390)) 8450 (("(" 18854) (argument_list 20386)) 8451 ((T_PAAMAYIM_NEKUDOTAYIM 20382)) 8452 ((")" 18863) (T_DOUBLE_ARROW 18864)) 8453 (("(" 18854) (argument_list 18862)) 8454 ((T_PAAMAYIM_NEKUDOTAYIM 18851)) 8455 (("$" 18835) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 18837) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 18852) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 18853) ("{" 914)) 8456 (("(" 18854) (argument_list 18855)) 8457 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18856) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 18857) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 18858) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8458 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18861) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8459 (("," 1449) (possible_comma 18859)) 8460 ((")" 18860)) 8461 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 18894) (";" 16321) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16322) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16323) (T_DEC 22) (T_DECLARE 16324) (T_DIR 24) (T_DNUMBER 25) (T_DO 16325) (T_DOUBLE_CAST 27) (T_ECHO 16326) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16327) (T_FOREACH 16328) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16329) (T_GOTO 16330) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16331) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16332) (T_START_HEREDOC 67) (T_STATIC 16333) (T_STRING 16334) (T_STRING_CAST 70) (T_SWITCH 16335) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16336) (T_UNSET 16337) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16338) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16339) (alt_if_stmt_without_else 16340) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16341) (fn 102) (foreach_statement 20381) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16342) (if_stmt_without_else 16343) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18896) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16345) ("~" 124)) 8462 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 18874) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 8463 (("$" 18865) (T_VARIABLE 18867) (simple_variable 20377) ("{" 20378)) 8464 (("(" 20374)) 8465 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 20372) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 8466 (("\"" 1489) ("$" 18865) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 1508) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 20371) (variable_class_name 18880)) 8467 ((T_NULLSAFE_OBJECT_OPERATOR 20354) (T_OBJECT_OPERATOR 20355) ("[" 20356) ("{" 20357)) 8468 (("(" 18884) (argument_list 20353)) 8469 ((T_PAAMAYIM_NEKUDOTAYIM 20349)) 8470 ((")" 18893)) 8471 (("(" 18884) (argument_list 18892)) 8472 ((T_PAAMAYIM_NEKUDOTAYIM 18881)) 8473 (("$" 18865) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 18867) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 18882) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 18883) ("{" 914)) 8474 (("(" 18884) (argument_list 18885)) 8475 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18886) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 18887) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 18888) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8476 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 18891) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 8477 (("," 1449) (possible_comma 18889)) 8478 ((")" 18890)) 8479 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 18894) (";" 16321) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16322) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16323) (T_DEC 22) (T_DECLARE 16324) (T_DIR 24) (T_DNUMBER 25) (T_DO 16325) (T_DOUBLE_CAST 27) (T_ECHO 16326) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16327) (T_FOREACH 16328) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16329) (T_GOTO 16330) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16331) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16332) (T_START_HEREDOC 67) (T_STATIC 16333) (T_STRING 16334) (T_STRING_CAST 70) (T_SWITCH 16335) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16336) (T_UNSET 16337) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16338) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16339) (alt_if_stmt_without_else 16340) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16341) (fn 102) (foreach_statement 18895) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16342) (if_stmt_without_else 16343) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18896) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16345) ("~" 124)) 8480 ((inner_statement_list 18897)) 8481 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 18905) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 8482 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 20347) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8483 ((T_STRING 20340)) 8484 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 20338) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8485 (("(" 20329)) 8486 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20323) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 8487 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 20321) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8488 ((";" 20320)) 8489 ((T_STRING 20313)) 8490 (("(" 20300)) 8491 (("(" 20285)) 8492 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 20283) (simple_variable 17412)) 8493 ((T_STRING 20281)) 8494 (("(" 20278)) 8495 (("(" 20274)) 8496 ((T_STRING 20268)) 8497 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 20266) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8498 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 20264)) 8499 ((":" 20263)) 8500 (("(" 20245)) 8501 ((T_STRING 20240)) 8502 (("{" 20223)) 8503 (("(" 20218)) 8504 (("(" 20209)) 8505 ((T_ELSE 20202) (T_ELSEIF 15767) (T_ENDIF 20203)) 8506 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 18900) (T_ENUM 18906) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 18914) (T_STATIC 15850) (T_TRAIT 18919) (attribute 15851) (attributed_statement 20201) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (enum_declaration_statement 18929) (fn 102) #'18931 (function_declaration_statement 18932) (inline_function 15853) (interface_declaration_statement 18936) (trait_declaration_statement 18938)) 8507 ((T_ABSTRACT 10) (T_CLASS 20193) (T_FINAL 34) (class_modifier 15842)) 8508 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 20192) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 8509 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15826) (ampersand 15827) (returns_ref 20180)) 8510 ((T_ELSE 18942) (T_ELSEIF 18943)) 8511 ((inner_statement_list 18940)) 8512 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 18941) ("~" 124)) 8513 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 18930) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20179) (static_member 118) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 8514 (("(" 18944)) 8515 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18945) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8516 (("%" 280) (")" 18946) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8517 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18947) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18948) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18949) (T_DEC 22) (T_DECLARE 18950) (T_DIR 24) (T_DNUMBER 25) (T_DO 18951) (T_DOUBLE_CAST 27) (T_ECHO 18952) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18953) (T_FOREACH 18954) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18955) (T_GOTO 18956) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18958) (T_START_HEREDOC 67) (T_STATIC 18959) (T_STRING 18960) (T_STRING_CAST 70) (T_SWITCH 18961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18962) (T_UNSET 18963) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18964) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18965) (alt_if_stmt_without_else 18966) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 18967) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18968) (if_stmt_without_else 18969) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18970) (static_member 118) (variable 121) (variable_class_name 122) ("{" 18971) ("~" 124)) 8518 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 20177) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8519 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 20175) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8520 (("(" 20166)) 8521 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20160) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 8522 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 20158) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8523 (("(" 19057)) 8524 (("(" 19042)) 8525 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 19040) (simple_variable 17412)) 8526 ((T_STRING 19038)) 8527 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19036) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8528 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 19034)) 8529 ((":" 19033)) 8530 (("(" 19015)) 8531 (("{" 18998)) 8532 (("(" 18993)) 8533 (("(" 18984)) 8534 ((T_ELSE 18977) (T_ELSEIF 15767) (T_ENDIF 18978)) 8535 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 18976) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 8536 ((T_ELSE 18974) (T_ELSEIF 18943)) 8537 ((inner_statement_list 18972)) 8538 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 18973) ("~" 124)) 8539 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18947) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18948) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18949) (T_DEC 22) (T_DECLARE 18950) (T_DIR 24) (T_DNUMBER 25) (T_DO 18951) (T_DOUBLE_CAST 27) (T_ECHO 18952) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18953) (T_FOREACH 18954) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18955) (T_GOTO 18956) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18958) (T_START_HEREDOC 67) (T_STATIC 18959) (T_STRING 18960) (T_STRING_CAST 70) (T_SWITCH 18961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18962) (T_UNSET 18963) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18964) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18965) (alt_if_stmt_without_else 18966) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 18967) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18968) (if_stmt_without_else 18969) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18975) (static_member 118) (variable 121) (variable_class_name 122) ("{" 18971) ("~" 124)) 8540 ((":" 18980)) 8541 ((";" 18979)) 8542 ((inner_statement_list 18981)) 8543 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 18982) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 8544 ((";" 18983)) 8545 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 18985) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8546 (("%" 280) (")" 18986) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8547 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 18987) (";" 18947) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18948) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18949) (T_DEC 22) (T_DECLARE 18950) (T_DIR 24) (T_DNUMBER 25) (T_DO 18951) (T_DOUBLE_CAST 27) (T_ECHO 18952) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18953) (T_FOREACH 18954) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18955) (T_GOTO 18956) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18958) (T_START_HEREDOC 67) (T_STATIC 18959) (T_STRING 18960) (T_STRING_CAST 70) (T_SWITCH 18961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18962) (T_UNSET 18963) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18964) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18965) (alt_if_stmt_without_else 18966) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 18967) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18968) (if_stmt_without_else 18969) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18988) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 18989) ("{" 18971) ("~" 124)) 8548 ((inner_statement_list 18990)) 8549 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 18991) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 8550 ((";" 18992)) 8551 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 18994) (variable 16049) (variable_class_name 16050)) 8552 (("," 16062) (possible_comma 18995)) 8553 ((")" 18996)) 8554 ((";" 18997)) 8555 ((inner_statement_list 18999)) 8556 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19000) ("~" 124)) 8557 ((catch_list 19001)) 8558 ((T_CATCH 19002) (T_FINALLY 19003) (finally_statement 19004)) 8559 (("(" 19008)) 8560 (("{" 19005)) 8561 ((inner_statement_list 19006)) 8562 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19007) ("~" 124)) 8563 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 19009) (class_name 16111) (name 16112)) 8564 ((T_VARIABLE 16113) (optional_variable 19010) ("|" 16115)) 8565 ((")" 19011)) 8566 (("{" 19012)) 8567 ((inner_statement_list 19013)) 8568 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19014) ("~" 124)) 8569 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19016) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8570 (("%" 280) (")" 19017) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8571 ((":" 19018) (switch_case_list 19019) ("{" 19020)) 8572 ((";" 19026) (case_list 19027)) 8573 ((";" 19021) (case_list 19022)) 8574 ((case_list 19024)) 8575 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 19023)) 8576 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 19025)) 8577 ((case_list 19030)) 8578 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 19028)) 8579 ((";" 19029)) 8580 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 19031)) 8581 ((";" 19032)) 8582 (("," 16412) (";" 19035)) 8583 ((";" 19037)) 8584 ((";" 19039)) 8585 (("," 17413) (";" 19041)) 8586 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 19043) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 8587 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 19044) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 8588 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 19045) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 8589 ((")" 19046) (T_DOUBLE_ARROW 19047)) 8590 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19050) (";" 18947) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18948) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18949) (T_DEC 22) (T_DECLARE 18950) (T_DIR 24) (T_DNUMBER 25) (T_DO 18951) (T_DOUBLE_CAST 27) (T_ECHO 18952) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18953) (T_FOREACH 18954) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18955) (T_GOTO 18956) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18958) (T_START_HEREDOC 67) (T_STATIC 18959) (T_STRING 18960) (T_STRING_CAST 70) (T_SWITCH 18961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18962) (T_UNSET 18963) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18964) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18965) (alt_if_stmt_without_else 18966) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 18967) (fn 102) (foreach_statement 19056) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18968) (if_stmt_without_else 18969) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19052) (static_member 118) (variable 121) (variable_class_name 122) ("{" 18971) ("~" 124)) 8591 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 19048) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 8592 ((")" 19049)) 8593 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19050) (";" 18947) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18948) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18949) (T_DEC 22) (T_DECLARE 18950) (T_DIR 24) (T_DNUMBER 25) (T_DO 18951) (T_DOUBLE_CAST 27) (T_ECHO 18952) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18953) (T_FOREACH 18954) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18955) (T_GOTO 18956) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18958) (T_START_HEREDOC 67) (T_STATIC 18959) (T_STRING 18960) (T_STRING_CAST 70) (T_SWITCH 18961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18962) (T_UNSET 18963) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18964) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18965) (alt_if_stmt_without_else 18966) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 18967) (fn 102) (foreach_statement 19051) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18968) (if_stmt_without_else 18969) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19052) (static_member 118) (variable 121) (variable_class_name 122) ("{" 18971) ("~" 124)) 8594 ((inner_statement_list 19053)) 8595 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 19054) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 8596 ((";" 19055)) 8597 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 19059) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8598 ((";" 19063)) 8599 (("," 19061)) 8600 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19062) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8601 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 19064) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8602 ((";" 19065)) 8603 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 19067) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 8604 ((")" 19071)) 8605 (("," 19069)) 8606 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19070) (fn 2328) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 8607 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19072) (";" 18947) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18948) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18949) (T_DEC 22) (T_DECLARE 18950) (T_DIR 24) (T_DNUMBER 25) (T_DO 18951) (T_DOUBLE_CAST 27) (T_ECHO 18952) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18953) (T_FOREACH 18954) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18955) (T_GOTO 18956) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18958) (T_START_HEREDOC 67) (T_STATIC 18959) (T_STRING 18960) (T_STRING_CAST 70) (T_SWITCH 18961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18962) (T_UNSET 18963) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18964) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18965) (alt_if_stmt_without_else 18966) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 18967) (fn 102) (for_statement 19073) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18968) (if_stmt_without_else 18969) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19074) (static_member 118) (variable 121) (variable_class_name 122) ("{" 18971) ("~" 124)) 8608 ((inner_statement_list 19075)) 8609 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 19083) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 8610 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 20156) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8611 ((T_STRING 20149)) 8612 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 20147) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8613 (("(" 20138)) 8614 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20132) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 8615 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 20130) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8616 ((";" 20129)) 8617 ((T_STRING 20122)) 8618 (("(" 20109)) 8619 (("(" 20094)) 8620 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 20092) (simple_variable 17412)) 8621 ((T_STRING 20090)) 8622 (("(" 20087)) 8623 (("(" 20083)) 8624 ((T_STRING 20077)) 8625 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 20075) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8626 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 20073)) 8627 ((":" 20072)) 8628 (("(" 20054)) 8629 ((T_STRING 20049)) 8630 (("{" 20032)) 8631 (("(" 20027)) 8632 (("(" 20018)) 8633 ((T_ELSE 20011) (T_ELSEIF 15767) (T_ENDIF 20012)) 8634 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 19078) (T_ENUM 19084) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 19092) (T_STATIC 15850) (T_TRAIT 19097) (attribute 15851) (attributed_statement 20010) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (enum_declaration_statement 19107) (fn 102) #'19109 (function_declaration_statement 19110) (inline_function 15853) (interface_declaration_statement 19114) (trait_declaration_statement 19116)) 8635 ((T_ABSTRACT 10) (T_CLASS 20002) (T_FINAL 34) (class_modifier 15842)) 8636 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 20001) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 8637 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15826) (ampersand 15827) (returns_ref 19989)) 8638 ((T_ELSE 19120) (T_ELSEIF 19121)) 8639 ((inner_statement_list 19118)) 8640 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19119) ("~" 124)) 8641 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19108) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19988) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 8642 (("(" 19122)) 8643 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19123) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8644 (("%" 280) (")" 19124) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8645 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19125) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19126) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19127) (T_DEC 22) (T_DECLARE 19128) (T_DIR 24) (T_DNUMBER 25) (T_DO 19129) (T_DOUBLE_CAST 27) (T_ECHO 19130) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19131) (T_FOREACH 19132) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19133) (T_GOTO 19134) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19135) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19136) (T_START_HEREDOC 67) (T_STATIC 19137) (T_STRING 19138) (T_STRING_CAST 70) (T_SWITCH 19139) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19140) (T_UNSET 19141) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19142) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19143) (alt_if_stmt_without_else 19144) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19145) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19146) (if_stmt_without_else 19147) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19148) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19149) ("~" 124)) 8646 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19986) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8647 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19984) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8648 (("(" 19975)) 8649 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19278) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 8650 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 19249) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8651 (("(" 19235)) 8652 (("(" 19220)) 8653 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 19218) (simple_variable 17412)) 8654 ((T_STRING 19216)) 8655 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19214) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8656 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 19212)) 8657 ((":" 19211)) 8658 (("(" 19193)) 8659 (("{" 19176)) 8660 (("(" 19171)) 8661 (("(" 19162)) 8662 ((T_ELSE 19155) (T_ELSEIF 15767) (T_ENDIF 19156)) 8663 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 19154) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 8664 ((T_ELSE 19152) (T_ELSEIF 19121)) 8665 ((inner_statement_list 19150)) 8666 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19151) ("~" 124)) 8667 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19125) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19126) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19127) (T_DEC 22) (T_DECLARE 19128) (T_DIR 24) (T_DNUMBER 25) (T_DO 19129) (T_DOUBLE_CAST 27) (T_ECHO 19130) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19131) (T_FOREACH 19132) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19133) (T_GOTO 19134) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19135) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19136) (T_START_HEREDOC 67) (T_STATIC 19137) (T_STRING 19138) (T_STRING_CAST 70) (T_SWITCH 19139) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19140) (T_UNSET 19141) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19142) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19143) (alt_if_stmt_without_else 19144) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19145) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19146) (if_stmt_without_else 19147) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19153) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19149) ("~" 124)) 8668 ((":" 19158)) 8669 ((";" 19157)) 8670 ((inner_statement_list 19159)) 8671 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 19160) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 8672 ((";" 19161)) 8673 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19163) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8674 (("%" 280) (")" 19164) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8675 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19165) (";" 19125) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19126) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19127) (T_DEC 22) (T_DECLARE 19128) (T_DIR 24) (T_DNUMBER 25) (T_DO 19129) (T_DOUBLE_CAST 27) (T_ECHO 19130) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19131) (T_FOREACH 19132) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19133) (T_GOTO 19134) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19135) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19136) (T_START_HEREDOC 67) (T_STATIC 19137) (T_STRING 19138) (T_STRING_CAST 70) (T_SWITCH 19139) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19140) (T_UNSET 19141) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19142) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19143) (alt_if_stmt_without_else 19144) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19145) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19146) (if_stmt_without_else 19147) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19166) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 19167) ("{" 19149) ("~" 124)) 8676 ((inner_statement_list 19168)) 8677 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 19169) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 8678 ((";" 19170)) 8679 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 19172) (variable 16049) (variable_class_name 16050)) 8680 (("," 16062) (possible_comma 19173)) 8681 ((")" 19174)) 8682 ((";" 19175)) 8683 ((inner_statement_list 19177)) 8684 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19178) ("~" 124)) 8685 ((catch_list 19179)) 8686 ((T_CATCH 19180) (T_FINALLY 19181) (finally_statement 19182)) 8687 (("(" 19186)) 8688 (("{" 19183)) 8689 ((inner_statement_list 19184)) 8690 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19185) ("~" 124)) 8691 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 19187) (class_name 16111) (name 16112)) 8692 ((T_VARIABLE 16113) (optional_variable 19188) ("|" 16115)) 8693 ((")" 19189)) 8694 (("{" 19190)) 8695 ((inner_statement_list 19191)) 8696 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19192) ("~" 124)) 8697 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19194) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8698 (("%" 280) (")" 19195) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8699 ((":" 19196) (switch_case_list 19197) ("{" 19198)) 8700 ((";" 19204) (case_list 19205)) 8701 ((";" 19199) (case_list 19200)) 8702 ((case_list 19202)) 8703 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 19201)) 8704 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 19203)) 8705 ((case_list 19208)) 8706 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 19206)) 8707 ((";" 19207)) 8708 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 19209)) 8709 ((";" 19210)) 8710 (("," 16412) (";" 19213)) 8711 ((";" 19215)) 8712 ((";" 19217)) 8713 (("," 17413) (";" 19219)) 8714 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 19221) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 8715 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 19222) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 8716 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 19223) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 8717 ((")" 19224) (T_DOUBLE_ARROW 19225)) 8718 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19228) (";" 19125) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19126) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19127) (T_DEC 22) (T_DECLARE 19128) (T_DIR 24) (T_DNUMBER 25) (T_DO 19129) (T_DOUBLE_CAST 27) (T_ECHO 19130) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19131) (T_FOREACH 19132) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19133) (T_GOTO 19134) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19135) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19136) (T_START_HEREDOC 67) (T_STATIC 19137) (T_STRING 19138) (T_STRING_CAST 70) (T_SWITCH 19139) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19140) (T_UNSET 19141) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19142) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19143) (alt_if_stmt_without_else 19144) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19145) (fn 102) (foreach_statement 19234) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19146) (if_stmt_without_else 19147) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19230) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19149) ("~" 124)) 8719 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 19226) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 8720 ((")" 19227)) 8721 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19228) (";" 19125) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19126) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19127) (T_DEC 22) (T_DECLARE 19128) (T_DIR 24) (T_DNUMBER 25) (T_DO 19129) (T_DOUBLE_CAST 27) (T_ECHO 19130) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19131) (T_FOREACH 19132) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19133) (T_GOTO 19134) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19135) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19136) (T_START_HEREDOC 67) (T_STATIC 19137) (T_STRING 19138) (T_STRING_CAST 70) (T_SWITCH 19139) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19140) (T_UNSET 19141) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19142) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19143) (alt_if_stmt_without_else 19144) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19145) (fn 102) (foreach_statement 19229) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19146) (if_stmt_without_else 19147) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19230) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19149) ("~" 124)) 8722 ((inner_statement_list 19231)) 8723 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 19232) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 8724 ((";" 19233)) 8725 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 19236) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8726 ((";" 19237)) 8727 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 19238) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8728 ((";" 19239)) 8729 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 19240) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 8730 ((")" 19241)) 8731 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19242) (";" 19125) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19126) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19127) (T_DEC 22) (T_DECLARE 19128) (T_DIR 24) (T_DNUMBER 25) (T_DO 19129) (T_DOUBLE_CAST 27) (T_ECHO 19130) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19131) (T_FOREACH 19132) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19133) (T_GOTO 19134) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19135) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19136) (T_START_HEREDOC 67) (T_STATIC 19137) (T_STRING 19138) (T_STRING_CAST 70) (T_SWITCH 19139) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19140) (T_UNSET 19141) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19142) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19143) (alt_if_stmt_without_else 19144) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19145) (fn 102) (for_statement 19243) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19146) (if_stmt_without_else 19147) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19244) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19149) ("~" 124)) 8732 ((inner_statement_list 19245)) 8733 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 19246) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 8734 ((";" 19247)) 8735 (("," 19251) (";" 19252)) 8736 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19253) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8737 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19973) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8738 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19971) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8739 (("(" 19962)) 8740 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19956) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 8741 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 19954) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8742 (("(" 19941)) 8743 (("(" 19926)) 8744 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 19924) (simple_variable 17412)) 8745 ((T_STRING 19922)) 8746 (("(" 19918)) 8747 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19916) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8748 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 19914)) 8749 ((":" 19913)) 8750 (("(" 19895)) 8751 (("{" 19878)) 8752 (("(" 19873)) 8753 (("(" 19864)) 8754 ((T_ELSE 19857) (T_ELSEIF 15767) (T_ENDIF 19858)) 8755 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 19856) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 8756 ((T_ELSE 19287) (T_ELSEIF 19288)) 8757 ((T_WHILE 19282)) 8758 ((inner_statement_list 19280)) 8759 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19281) ("~" 124)) 8760 (("(" 19283)) 8761 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19284) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8762 (("%" 280) (")" 19285) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8763 ((";" 19286)) 8764 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19855) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 8765 (("(" 19289)) 8766 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19290) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8767 (("%" 280) (")" 19291) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8768 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19292) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19293) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19294) (T_DEC 22) (T_DECLARE 19295) (T_DIR 24) (T_DNUMBER 25) (T_DO 19296) (T_DOUBLE_CAST 27) (T_ECHO 19297) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19298) (T_FOREACH 19299) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19300) (T_GOTO 19301) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19302) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19303) (T_START_HEREDOC 67) (T_STATIC 19304) (T_STRING 19305) (T_STRING_CAST 70) (T_SWITCH 19306) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19307) (T_UNSET 19308) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19309) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19310) (alt_if_stmt_without_else 19311) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19312) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19313) (if_stmt_without_else 19314) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19315) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19316) ("~" 124)) 8769 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19853) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8770 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19851) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8771 (("(" 19423)) 8772 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19417) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 8773 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 19415) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8774 (("(" 19402)) 8775 (("(" 19387)) 8776 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 19385) (simple_variable 17412)) 8777 ((T_STRING 19383)) 8778 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19381) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8779 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 19379)) 8780 ((":" 19378)) 8781 (("(" 19360)) 8782 (("{" 19343)) 8783 (("(" 19338)) 8784 (("(" 19329)) 8785 ((T_ELSE 19322) (T_ELSEIF 15767) (T_ENDIF 19323)) 8786 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 19321) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 8787 ((T_ELSE 19319) (T_ELSEIF 19288)) 8788 ((inner_statement_list 19317)) 8789 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19318) ("~" 124)) 8790 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19292) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19293) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19294) (T_DEC 22) (T_DECLARE 19295) (T_DIR 24) (T_DNUMBER 25) (T_DO 19296) (T_DOUBLE_CAST 27) (T_ECHO 19297) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19298) (T_FOREACH 19299) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19300) (T_GOTO 19301) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19302) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19303) (T_START_HEREDOC 67) (T_STATIC 19304) (T_STRING 19305) (T_STRING_CAST 70) (T_SWITCH 19306) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19307) (T_UNSET 19308) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19309) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19310) (alt_if_stmt_without_else 19311) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19312) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19313) (if_stmt_without_else 19314) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19320) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19316) ("~" 124)) 8791 ((":" 19325)) 8792 ((";" 19324)) 8793 ((inner_statement_list 19326)) 8794 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 19327) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 8795 ((";" 19328)) 8796 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19330) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8797 (("%" 280) (")" 19331) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8798 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19332) (";" 19292) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19293) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19294) (T_DEC 22) (T_DECLARE 19295) (T_DIR 24) (T_DNUMBER 25) (T_DO 19296) (T_DOUBLE_CAST 27) (T_ECHO 19297) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19298) (T_FOREACH 19299) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19300) (T_GOTO 19301) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19302) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19303) (T_START_HEREDOC 67) (T_STATIC 19304) (T_STRING 19305) (T_STRING_CAST 70) (T_SWITCH 19306) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19307) (T_UNSET 19308) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19309) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19310) (alt_if_stmt_without_else 19311) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19312) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19313) (if_stmt_without_else 19314) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19333) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 19334) ("{" 19316) ("~" 124)) 8799 ((inner_statement_list 19335)) 8800 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 19336) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 8801 ((";" 19337)) 8802 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 19339) (variable 16049) (variable_class_name 16050)) 8803 (("," 16062) (possible_comma 19340)) 8804 ((")" 19341)) 8805 ((";" 19342)) 8806 ((inner_statement_list 19344)) 8807 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19345) ("~" 124)) 8808 ((catch_list 19346)) 8809 ((T_CATCH 19347) (T_FINALLY 19348) (finally_statement 19349)) 8810 (("(" 19353)) 8811 (("{" 19350)) 8812 ((inner_statement_list 19351)) 8813 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19352) ("~" 124)) 8814 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 19354) (class_name 16111) (name 16112)) 8815 ((T_VARIABLE 16113) (optional_variable 19355) ("|" 16115)) 8816 ((")" 19356)) 8817 (("{" 19357)) 8818 ((inner_statement_list 19358)) 8819 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19359) ("~" 124)) 8820 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19361) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8821 (("%" 280) (")" 19362) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8822 ((":" 19363) (switch_case_list 19364) ("{" 19365)) 8823 ((";" 19371) (case_list 19372)) 8824 ((";" 19366) (case_list 19367)) 8825 ((case_list 19369)) 8826 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 19368)) 8827 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 19370)) 8828 ((case_list 19375)) 8829 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 19373)) 8830 ((";" 19374)) 8831 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 19376)) 8832 ((";" 19377)) 8833 (("," 16412) (";" 19380)) 8834 ((";" 19382)) 8835 ((";" 19384)) 8836 (("," 17413) (";" 19386)) 8837 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 19388) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 8838 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 19389) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 8839 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 19390) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 8840 ((")" 19391) (T_DOUBLE_ARROW 19392)) 8841 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19395) (";" 19292) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19293) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19294) (T_DEC 22) (T_DECLARE 19295) (T_DIR 24) (T_DNUMBER 25) (T_DO 19296) (T_DOUBLE_CAST 27) (T_ECHO 19297) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19298) (T_FOREACH 19299) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19300) (T_GOTO 19301) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19302) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19303) (T_START_HEREDOC 67) (T_STATIC 19304) (T_STRING 19305) (T_STRING_CAST 70) (T_SWITCH 19306) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19307) (T_UNSET 19308) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19309) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19310) (alt_if_stmt_without_else 19311) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19312) (fn 102) (foreach_statement 19401) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19313) (if_stmt_without_else 19314) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19397) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19316) ("~" 124)) 8842 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 19393) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 8843 ((")" 19394)) 8844 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19395) (";" 19292) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19293) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19294) (T_DEC 22) (T_DECLARE 19295) (T_DIR 24) (T_DNUMBER 25) (T_DO 19296) (T_DOUBLE_CAST 27) (T_ECHO 19297) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19298) (T_FOREACH 19299) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19300) (T_GOTO 19301) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19302) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19303) (T_START_HEREDOC 67) (T_STATIC 19304) (T_STRING 19305) (T_STRING_CAST 70) (T_SWITCH 19306) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19307) (T_UNSET 19308) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19309) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19310) (alt_if_stmt_without_else 19311) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19312) (fn 102) (foreach_statement 19396) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19313) (if_stmt_without_else 19314) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19397) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19316) ("~" 124)) 8845 ((inner_statement_list 19398)) 8846 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 19399) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 8847 ((";" 19400)) 8848 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 19403) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8849 ((";" 19404)) 8850 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 19405) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8851 ((";" 19406)) 8852 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 19407) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 8853 ((")" 19408)) 8854 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19409) (";" 19292) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19293) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19294) (T_DEC 22) (T_DECLARE 19295) (T_DIR 24) (T_DNUMBER 25) (T_DO 19296) (T_DOUBLE_CAST 27) (T_ECHO 19297) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19298) (T_FOREACH 19299) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19300) (T_GOTO 19301) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19302) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19303) (T_START_HEREDOC 67) (T_STATIC 19304) (T_STRING 19305) (T_STRING_CAST 70) (T_SWITCH 19306) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19307) (T_UNSET 19308) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19309) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19310) (alt_if_stmt_without_else 19311) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19312) (fn 102) (for_statement 19410) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19313) (if_stmt_without_else 19314) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19411) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19316) ("~" 124)) 8855 ((inner_statement_list 19412)) 8856 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 19413) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 8857 ((";" 19414)) 8858 (("," 19251) (";" 19416)) 8859 ((T_WHILE 19418)) 8860 (("(" 19419)) 8861 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19420) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8862 (("%" 280) (")" 19421) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8863 ((";" 19422)) 8864 ((T_STRING 19424) (const_decl 19425) (const_list 19426)) 8865 (("=" 19848)) 8866 ((")" 19427) ("," 19428)) 8867 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19430) (";" 19292) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19293) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19294) (T_DEC 22) (T_DECLARE 19295) (T_DIR 24) (T_DNUMBER 25) (T_DO 19296) (T_DOUBLE_CAST 27) (T_ECHO 19297) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19298) (T_FOREACH 19299) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19300) (T_GOTO 19301) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19302) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19303) (T_START_HEREDOC 67) (T_STATIC 19304) (T_STRING 19305) (T_STRING_CAST 70) (T_SWITCH 19306) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19307) (T_UNSET 19308) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19309) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19310) (alt_if_stmt_without_else 19311) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 19431) (dereferenceable_scalar 99) (expr 19312) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19313) (if_stmt_without_else 19314) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19432) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19316) ("~" 124)) 8868 ((T_STRING 19424) (const_decl 19429)) 8869 ((inner_statement_list 19433)) 8870 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 19441) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 8871 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19846) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8872 ((T_STRING 19839)) 8873 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19837) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8874 (("(" 19828)) 8875 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19822) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 8876 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 19820) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8877 ((";" 19819)) 8878 ((T_STRING 19765)) 8879 (("(" 19752)) 8880 (("(" 19737)) 8881 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 19735) (simple_variable 17412)) 8882 ((T_STRING 19733)) 8883 (("(" 19730)) 8884 (("(" 19724)) 8885 ((T_STRING 19716)) 8886 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19714) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8887 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 19712)) 8888 ((":" 19711)) 8889 (("(" 19693)) 8890 ((T_STRING 19688)) 8891 (("{" 19671)) 8892 (("(" 19666)) 8893 (("(" 19657)) 8894 ((T_ELSE 19650) (T_ELSEIF 15767) (T_ENDIF 19651)) 8895 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 19436) (T_ENUM 19442) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 19450) (T_STATIC 15850) (T_TRAIT 19455) (attribute 15851) (attributed_statement 19649) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (enum_declaration_statement 19465) (fn 102) #'19467 (function_declaration_statement 19468) (inline_function 15853) (interface_declaration_statement 19472) (trait_declaration_statement 19474)) 8896 ((T_ABSTRACT 10) (T_CLASS 19641) (T_FINAL 34) (class_modifier 15842)) 8897 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 19640) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 8898 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15826) (ampersand 15827) (returns_ref 19628)) 8899 ((T_ELSE 19478) (T_ELSEIF 19479)) 8900 ((inner_statement_list 19476)) 8901 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19477) ("~" 124)) 8902 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19466) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19627) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 8903 (("(" 19480)) 8904 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19481) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8905 (("%" 280) (")" 19482) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8906 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19483) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19484) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19485) (T_DEC 22) (T_DECLARE 19486) (T_DIR 24) (T_DNUMBER 25) (T_DO 19487) (T_DOUBLE_CAST 27) (T_ECHO 19488) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19489) (T_FOREACH 19490) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19491) (T_GOTO 19492) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19493) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19494) (T_START_HEREDOC 67) (T_STATIC 19495) (T_STRING 19496) (T_STRING_CAST 70) (T_SWITCH 19497) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19498) (T_UNSET 19499) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19500) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19501) (alt_if_stmt_without_else 19502) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19503) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19504) (if_stmt_without_else 19505) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19506) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19507) ("~" 124)) 8907 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19625) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8908 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19623) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8909 (("(" 19614)) 8910 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19608) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 8911 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 19606) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8912 (("(" 19593)) 8913 (("(" 19578)) 8914 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 19576) (simple_variable 17412)) 8915 ((T_STRING 19574)) 8916 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 19572) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 8917 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 19570)) 8918 ((":" 19569)) 8919 (("(" 19551)) 8920 (("{" 19534)) 8921 (("(" 19529)) 8922 (("(" 19520)) 8923 ((T_ELSE 19513) (T_ELSEIF 15767) (T_ENDIF 19514)) 8924 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 19512) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 8925 ((T_ELSE 19510) (T_ELSEIF 19479)) 8926 ((inner_statement_list 19508)) 8927 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19509) ("~" 124)) 8928 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19483) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19484) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19485) (T_DEC 22) (T_DECLARE 19486) (T_DIR 24) (T_DNUMBER 25) (T_DO 19487) (T_DOUBLE_CAST 27) (T_ECHO 19488) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19489) (T_FOREACH 19490) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19491) (T_GOTO 19492) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19493) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19494) (T_START_HEREDOC 67) (T_STATIC 19495) (T_STRING 19496) (T_STRING_CAST 70) (T_SWITCH 19497) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19498) (T_UNSET 19499) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19500) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19501) (alt_if_stmt_without_else 19502) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19503) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19504) (if_stmt_without_else 19505) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19511) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19507) ("~" 124)) 8929 ((":" 19516)) 8930 ((";" 19515)) 8931 ((inner_statement_list 19517)) 8932 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 19518) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 8933 ((";" 19519)) 8934 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19521) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8935 (("%" 280) (")" 19522) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8936 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19523) (";" 19483) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19484) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19485) (T_DEC 22) (T_DECLARE 19486) (T_DIR 24) (T_DNUMBER 25) (T_DO 19487) (T_DOUBLE_CAST 27) (T_ECHO 19488) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19489) (T_FOREACH 19490) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19491) (T_GOTO 19492) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19493) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19494) (T_START_HEREDOC 67) (T_STATIC 19495) (T_STRING 19496) (T_STRING_CAST 70) (T_SWITCH 19497) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19498) (T_UNSET 19499) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19500) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19501) (alt_if_stmt_without_else 19502) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19503) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19504) (if_stmt_without_else 19505) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19524) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 19525) ("{" 19507) ("~" 124)) 8937 ((inner_statement_list 19526)) 8938 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 19527) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 8939 ((";" 19528)) 8940 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 19530) (variable 16049) (variable_class_name 16050)) 8941 (("," 16062) (possible_comma 19531)) 8942 ((")" 19532)) 8943 ((";" 19533)) 8944 ((inner_statement_list 19535)) 8945 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19536) ("~" 124)) 8946 ((catch_list 19537)) 8947 ((T_CATCH 19538) (T_FINALLY 19539) (finally_statement 19540)) 8948 (("(" 19544)) 8949 (("{" 19541)) 8950 ((inner_statement_list 19542)) 8951 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19543) ("~" 124)) 8952 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 19545) (class_name 16111) (name 16112)) 8953 ((T_VARIABLE 16113) (optional_variable 19546) ("|" 16115)) 8954 ((")" 19547)) 8955 (("{" 19548)) 8956 ((inner_statement_list 19549)) 8957 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19550) ("~" 124)) 8958 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19552) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 8959 (("%" 280) (")" 19553) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 8960 ((":" 19554) (switch_case_list 19555) ("{" 19556)) 8961 ((";" 19562) (case_list 19563)) 8962 ((";" 19557) (case_list 19558)) 8963 ((case_list 19560)) 8964 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 19559)) 8965 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 19561)) 8966 ((case_list 19566)) 8967 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 19564)) 8968 ((";" 19565)) 8969 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 19567)) 8970 ((";" 19568)) 8971 (("," 16412) (";" 19571)) 8972 ((";" 19573)) 8973 ((";" 19575)) 8974 (("," 17413) (";" 19577)) 8975 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 19579) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 8976 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 19580) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 8977 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 19581) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 8978 ((")" 19582) (T_DOUBLE_ARROW 19583)) 8979 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19586) (";" 19483) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19484) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19485) (T_DEC 22) (T_DECLARE 19486) (T_DIR 24) (T_DNUMBER 25) (T_DO 19487) (T_DOUBLE_CAST 27) (T_ECHO 19488) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19489) (T_FOREACH 19490) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19491) (T_GOTO 19492) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19493) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19494) (T_START_HEREDOC 67) (T_STATIC 19495) (T_STRING 19496) (T_STRING_CAST 70) (T_SWITCH 19497) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19498) (T_UNSET 19499) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19500) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19501) (alt_if_stmt_without_else 19502) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19503) (fn 102) (foreach_statement 19592) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19504) (if_stmt_without_else 19505) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19588) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19507) ("~" 124)) 8980 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 19584) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 8981 ((")" 19585)) 8982 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19586) (";" 19483) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19484) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19485) (T_DEC 22) (T_DECLARE 19486) (T_DIR 24) (T_DNUMBER 25) (T_DO 19487) (T_DOUBLE_CAST 27) (T_ECHO 19488) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19489) (T_FOREACH 19490) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19491) (T_GOTO 19492) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19493) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19494) (T_START_HEREDOC 67) (T_STATIC 19495) (T_STRING 19496) (T_STRING_CAST 70) (T_SWITCH 19497) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19498) (T_UNSET 19499) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19500) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19501) (alt_if_stmt_without_else 19502) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19503) (fn 102) (foreach_statement 19587) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19504) (if_stmt_without_else 19505) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19588) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19507) ("~" 124)) 8983 ((inner_statement_list 19589)) 8984 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 19590) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 8985 ((";" 19591)) 8986 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 19594) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8987 ((";" 19595)) 8988 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 19596) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 8989 ((";" 19597)) 8990 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 19598) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 8991 ((")" 19599)) 8992 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19600) (";" 19483) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19484) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19485) (T_DEC 22) (T_DECLARE 19486) (T_DIR 24) (T_DNUMBER 25) (T_DO 19487) (T_DOUBLE_CAST 27) (T_ECHO 19488) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19489) (T_FOREACH 19490) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19491) (T_GOTO 19492) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19493) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19494) (T_START_HEREDOC 67) (T_STATIC 19495) (T_STRING 19496) (T_STRING_CAST 70) (T_SWITCH 19497) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19498) (T_UNSET 19499) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19500) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19501) (alt_if_stmt_without_else 19502) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19503) (fn 102) (for_statement 19601) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19504) (if_stmt_without_else 19505) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19602) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19507) ("~" 124)) 8993 ((inner_statement_list 19603)) 8994 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 19604) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 8995 ((";" 19605)) 8996 (("," 19251) (";" 19607)) 8997 ((T_WHILE 19609)) 8998 (("(" 19610)) 8999 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19611) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9000 (("%" 280) (")" 19612) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9001 ((";" 19613)) 9002 ((T_STRING 19424) (const_decl 19425) (const_list 19615)) 9003 ((")" 19616) ("," 19428)) 9004 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19617) (";" 19483) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19484) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19485) (T_DEC 22) (T_DECLARE 19486) (T_DIR 24) (T_DNUMBER 25) (T_DO 19487) (T_DOUBLE_CAST 27) (T_ECHO 19488) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19489) (T_FOREACH 19490) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19491) (T_GOTO 19492) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19493) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19494) (T_START_HEREDOC 67) (T_STATIC 19495) (T_STRING 19496) (T_STRING_CAST 70) (T_SWITCH 19497) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19498) (T_UNSET 19499) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19500) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19501) (alt_if_stmt_without_else 19502) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 19618) (dereferenceable_scalar 99) (expr 19503) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19504) (if_stmt_without_else 19505) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19619) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19507) ("~" 124)) 9005 ((inner_statement_list 19620)) 9006 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 19621) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9007 ((";" 19622)) 9008 ((";" 19624)) 9009 ((";" 19626)) 9010 ((T_STRING 19629) (backup_doc_comment 137)) 9011 ((backup_doc_comment 19630)) 9012 (("(" 19631)) 9013 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 19632)) 9014 ((")" 19633)) 9015 ((":" 150) (return_type 19634)) 9016 ((backup_fn_flags 19635)) 9017 (("{" 19636)) 9018 ((inner_statement_list 19637)) 9019 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19638) ("~" 124)) 9020 ((backup_fn_flags 19639)) 9021 ((T_STRING 19642)) 9022 ((T_EXTENDS 2902) (extends_from 19643)) 9023 ((T_IMPLEMENTS 2904) (implements_list 19644)) 9024 ((backup_doc_comment 19645)) 9025 (("{" 19646)) 9026 ((class_statement_list 19647)) 9027 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 19648)) 9028 ((":" 19653)) 9029 ((";" 19652)) 9030 ((inner_statement_list 19654)) 9031 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 19655) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 9032 ((";" 19656)) 9033 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19658) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9034 (("%" 280) (")" 19659) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9035 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19660) (";" 19434) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19466) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19661) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 19662) ("{" 19475) ("~" 124)) 9036 ((inner_statement_list 19663)) 9037 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 19664) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 9038 ((";" 19665)) 9039 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 19667) (variable 16049) (variable_class_name 16050)) 9040 (("," 16062) (possible_comma 19668)) 9041 ((")" 19669)) 9042 ((";" 19670)) 9043 ((inner_statement_list 19672)) 9044 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19673) ("~" 124)) 9045 ((catch_list 19674)) 9046 ((T_CATCH 19675) (T_FINALLY 19676) (finally_statement 19677)) 9047 (("(" 19681)) 9048 (("{" 19678)) 9049 ((inner_statement_list 19679)) 9050 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19680) ("~" 124)) 9051 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 19682) (class_name 16111) (name 16112)) 9052 ((T_VARIABLE 16113) (optional_variable 19683) ("|" 16115)) 9053 ((")" 19684)) 9054 (("{" 19685)) 9055 ((inner_statement_list 19686)) 9056 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19687) ("~" 124)) 9057 ((backup_doc_comment 19689)) 9058 (("{" 19690)) 9059 ((class_statement_list 19691)) 9060 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 19692)) 9061 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19694) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9062 (("%" 280) (")" 19695) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9063 ((":" 19696) (switch_case_list 19697) ("{" 19698)) 9064 ((";" 19704) (case_list 19705)) 9065 ((";" 19699) (case_list 19700)) 9066 ((case_list 19702)) 9067 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 19701)) 9068 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 19703)) 9069 ((case_list 19708)) 9070 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 19706)) 9071 ((";" 19707)) 9072 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 19709)) 9073 ((";" 19710)) 9074 (("," 16412) (";" 19713)) 9075 ((";" 19715)) 9076 ((T_EXTENDS 19717) (interface_extends_list 19718)) 9077 ((T_NAME_FULLY_QUALIFIED 6371) (T_NAME_QUALIFIED 6372) (T_NAME_RELATIVE 6373) (T_STATIC 6374) (T_STRING 6375) (class_name 6376) (class_name_list 19723) (name 6378)) 9078 ((backup_doc_comment 19719)) 9079 (("{" 19720)) 9080 ((class_statement_list 19721)) 9081 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 19722)) 9082 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19725) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9083 (("%" 280) (")" 19726) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9084 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19727) (";" 19483) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19484) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19485) (T_DEC 22) (T_DECLARE 19486) (T_DIR 24) (T_DNUMBER 25) (T_DO 19487) (T_DOUBLE_CAST 27) (T_ECHO 19488) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19489) (T_FOREACH 19490) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19491) (T_GOTO 19492) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19493) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19494) (T_START_HEREDOC 67) (T_STATIC 19495) (T_STRING 19496) (T_STRING_CAST 70) (T_SWITCH 19497) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19498) (T_UNSET 19499) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19500) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19501) (alt_if_stmt_without_else 19502) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19503) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19504) (if_stmt_without_else 19505) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19728) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19507) ("~" 124)) 9085 ((inner_statement_list 19729)) 9086 ((")" 19731)) 9087 ((";" 19732)) 9088 ((";" 19734)) 9089 (("," 17413) (";" 19736)) 9090 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 19738) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 9091 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 19739) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 9092 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 19740) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 9093 ((")" 19741) (T_DOUBLE_ARROW 19742)) 9094 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19745) (";" 19434) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19466) (fn 102) (foreach_statement 19751) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19747) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9095 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 19743) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 9096 ((")" 19744)) 9097 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19745) (";" 19434) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19466) (fn 102) (foreach_statement 19746) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19747) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9098 ((inner_statement_list 19748)) 9099 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 19749) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 9100 ((";" 19750)) 9101 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 19753) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 9102 ((";" 19754)) 9103 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 19755) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 9104 ((";" 19756)) 9105 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 19757) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 9106 ((")" 19758)) 9107 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19759) (";" 19434) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19466) (fn 102) (for_statement 19760) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19761) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9108 ((inner_statement_list 19762)) 9109 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 19763) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 9110 ((";" 19764)) 9111 ((":" 19766) (enum_backing_type 19767)) 9112 (("?" 19773) (T_ARRAY 19774) (T_CALLABLE 19775) (T_NAME_FULLY_QUALIFIED 19776) (T_NAME_QUALIFIED 19777) (T_NAME_RELATIVE 19778) (T_STATIC 19779) (T_STRING 19780) (intersection_type 19781) (name 19782) (type 19783) (type_expr 19784) (type_without_static 19785) (union_type 19786)) 9113 ((T_IMPLEMENTS 2904) (implements_list 19768)) 9114 ((backup_doc_comment 19769)) 9115 (("{" 19770)) 9116 ((class_statement_list 19771)) 9117 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 19772)) 9118 ((T_ARRAY 19813) (T_CALLABLE 19814) (T_NAME_FULLY_QUALIFIED 6381) (T_NAME_QUALIFIED 6382) (T_NAME_RELATIVE 6383) (T_STATIC 19815) (T_STRING 6385) (name 19816) (type 19817) (type_without_static 19818)) 9119 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19811)) 9120 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19798) ("|" 19799)) 9121 (("|" 19787)) 9122 ((T_ARRAY 19788) (T_CALLABLE 19789) (T_NAME_FULLY_QUALIFIED 19790) (T_NAME_QUALIFIED 19791) (T_NAME_RELATIVE 19792) (T_STATIC 19793) (T_STRING 19794) (name 19795) (type 19796) (type_without_static 19797)) 9123 ((T_ARRAY 19801) (T_CALLABLE 19802) (T_NAME_FULLY_QUALIFIED 19803) (T_NAME_QUALIFIED 19804) (T_NAME_RELATIVE 19805) (T_STATIC 19806) (T_STRING 19807) (name 19808) (type 19809) (type_without_static 19810)) 9124 ((T_ARRAY 19788) (T_CALLABLE 19789) (T_NAME_FULLY_QUALIFIED 19790) (T_NAME_QUALIFIED 19791) (T_NAME_RELATIVE 19792) (T_STATIC 19793) (T_STRING 19794) (name 19795) (type 19800) (type_without_static 19797)) 9125 ((T_ARRAY 19801) (T_CALLABLE 19802) (T_NAME_FULLY_QUALIFIED 19803) (T_NAME_QUALIFIED 19804) (T_NAME_RELATIVE 19805) (T_STATIC 19806) (T_STRING 19807) (name 19808) (type 19812) (type_without_static 19810)) 9126 (("," 19251) (";" 19821)) 9127 ((T_WHILE 19823)) 9128 (("(" 19824)) 9129 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19825) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9130 (("%" 280) (")" 19826) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9131 ((";" 19827)) 9132 ((T_STRING 19424) (const_decl 19425) (const_list 19829)) 9133 ((")" 19830) ("," 19428)) 9134 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19831) (";" 19434) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 19832) (dereferenceable_scalar 99) (expr 19466) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19833) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9135 ((inner_statement_list 19834)) 9136 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 19835) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9137 ((";" 19836)) 9138 ((";" 19838)) 9139 ((T_EXTENDS 2902) (extends_from 19840)) 9140 ((T_IMPLEMENTS 2904) (implements_list 19841)) 9141 ((backup_doc_comment 19842)) 9142 (("{" 19843)) 9143 ((class_statement_list 19844)) 9144 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 19845)) 9145 ((";" 19847)) 9146 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 19849) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 9147 (("%" 1081) ("*" 1082) ("+" 1083) ("-" 1084) ("." 1085) ("/" 1086) ("<" 1087) (">" 1088) ("?" 1089) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1090) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1091) (T_BOOLEAN_AND 1092) (T_BOOLEAN_OR 1093) (T_COALESCE 1094) (T_INSTANCEOF 1095) (T_IS_EQUAL 1096) (T_IS_GREATER_OR_EQUAL 1097) (T_IS_IDENTICAL 1098) (T_IS_NOT_EQUAL 1099) (T_IS_NOT_IDENTICAL 1100) (T_IS_SMALLER_OR_EQUAL 1101) (T_LOGICAL_AND 1102) (T_LOGICAL_OR 1103) (T_LOGICAL_XOR 1104) (T_POW 1105) (T_SL 1106) (T_SPACESHIP 1107) (T_SR 1108) ("^" 1109) (backup_doc_comment 19850) ("|" 1110)) 9148 ((";" 19852)) 9149 ((";" 19854)) 9150 ((":" 19860)) 9151 ((";" 19859)) 9152 ((inner_statement_list 19861)) 9153 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 19862) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 9154 ((";" 19863)) 9155 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19865) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9156 (("%" 280) (")" 19866) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9157 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19867) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19868) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 19869) ("{" 19279) ("~" 124)) 9158 ((inner_statement_list 19870)) 9159 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 19871) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 9160 ((";" 19872)) 9161 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 19874) (variable 16049) (variable_class_name 16050)) 9162 (("," 16062) (possible_comma 19875)) 9163 ((")" 19876)) 9164 ((";" 19877)) 9165 ((inner_statement_list 19879)) 9166 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19880) ("~" 124)) 9167 ((catch_list 19881)) 9168 ((T_CATCH 19882) (T_FINALLY 19883) (finally_statement 19884)) 9169 (("(" 19888)) 9170 (("{" 19885)) 9171 ((inner_statement_list 19886)) 9172 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19887) ("~" 124)) 9173 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 19889) (class_name 16111) (name 16112)) 9174 ((T_VARIABLE 16113) (optional_variable 19890) ("|" 16115)) 9175 ((")" 19891)) 9176 (("{" 19892)) 9177 ((inner_statement_list 19893)) 9178 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19894) ("~" 124)) 9179 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19896) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9180 (("%" 280) (")" 19897) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9181 ((":" 19898) (switch_case_list 19899) ("{" 19900)) 9182 ((";" 19906) (case_list 19907)) 9183 ((";" 19901) (case_list 19902)) 9184 ((case_list 19904)) 9185 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 19903)) 9186 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 19905)) 9187 ((case_list 19910)) 9188 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 19908)) 9189 ((";" 19909)) 9190 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 19911)) 9191 ((";" 19912)) 9192 (("," 16412) (";" 19915)) 9193 ((";" 19917)) 9194 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19919) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9195 (("%" 280) (")" 19920) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9196 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19727) (";" 19292) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19293) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19294) (T_DEC 22) (T_DECLARE 19295) (T_DIR 24) (T_DNUMBER 25) (T_DO 19296) (T_DOUBLE_CAST 27) (T_ECHO 19297) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19298) (T_FOREACH 19299) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19300) (T_GOTO 19301) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19302) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19303) (T_START_HEREDOC 67) (T_STATIC 19304) (T_STRING 19305) (T_STRING_CAST 70) (T_SWITCH 19306) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19307) (T_UNSET 19308) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19309) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19310) (alt_if_stmt_without_else 19311) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19312) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19313) (if_stmt_without_else 19314) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19921) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19316) ("~" 124)) 9197 ((";" 19923)) 9198 (("," 17413) (";" 19925)) 9199 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 19927) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 9200 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 19928) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 9201 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 19929) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 9202 ((")" 19930) (T_DOUBLE_ARROW 19931)) 9203 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19934) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (foreach_statement 19940) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19936) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 9204 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 19932) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 9205 ((")" 19933)) 9206 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19934) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (foreach_statement 19935) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19936) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 9207 ((inner_statement_list 19937)) 9208 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 19938) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 9209 ((";" 19939)) 9210 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 19942) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 9211 ((";" 19943)) 9212 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 19944) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 9213 ((";" 19945)) 9214 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 19946) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 9215 ((")" 19947)) 9216 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19948) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (for_statement 19949) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19950) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 9217 ((inner_statement_list 19951)) 9218 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 19952) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 9219 ((";" 19953)) 9220 (("," 19251) (";" 19955)) 9221 ((T_WHILE 19957)) 9222 (("(" 19958)) 9223 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 19959) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9224 (("%" 280) (")" 19960) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9225 ((";" 19961)) 9226 ((T_STRING 19424) (const_decl 19425) (const_list 19963)) 9227 ((")" 19964) ("," 19428)) 9228 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19965) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 19966) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19967) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 9229 ((inner_statement_list 19968)) 9230 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 19969) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9231 ((";" 19970)) 9232 ((";" 19972)) 9233 ((";" 19974)) 9234 ((T_STRING 19424) (const_decl 19425) (const_list 19976)) 9235 ((")" 19977) ("," 19428)) 9236 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19978) (";" 19125) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19126) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19127) (T_DEC 22) (T_DECLARE 19128) (T_DIR 24) (T_DNUMBER 25) (T_DO 19129) (T_DOUBLE_CAST 27) (T_ECHO 19130) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19131) (T_FOREACH 19132) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19133) (T_GOTO 19134) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19135) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19136) (T_START_HEREDOC 67) (T_STATIC 19137) (T_STRING 19138) (T_STRING_CAST 70) (T_SWITCH 19139) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19140) (T_UNSET 19141) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19142) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19143) (alt_if_stmt_without_else 19144) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 19979) (dereferenceable_scalar 99) (expr 19145) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19146) (if_stmt_without_else 19147) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19980) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19149) ("~" 124)) 9237 ((inner_statement_list 19981)) 9238 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 19982) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9239 ((";" 19983)) 9240 ((";" 19985)) 9241 ((";" 19987)) 9242 ((T_STRING 19990) (backup_doc_comment 137)) 9243 ((backup_doc_comment 19991)) 9244 (("(" 19992)) 9245 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 19993)) 9246 ((")" 19994)) 9247 ((":" 150) (return_type 19995)) 9248 ((backup_fn_flags 19996)) 9249 (("{" 19997)) 9250 ((inner_statement_list 19998)) 9251 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 19999) ("~" 124)) 9252 ((backup_fn_flags 20000)) 9253 ((T_STRING 20003)) 9254 ((T_EXTENDS 2902) (extends_from 20004)) 9255 ((T_IMPLEMENTS 2904) (implements_list 20005)) 9256 ((backup_doc_comment 20006)) 9257 (("{" 20007)) 9258 ((class_statement_list 20008)) 9259 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 20009)) 9260 ((":" 20014)) 9261 ((";" 20013)) 9262 ((inner_statement_list 20015)) 9263 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 20016) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 9264 ((";" 20017)) 9265 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20019) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9266 (("%" 280) (")" 20020) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9267 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20021) (";" 19076) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19108) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20022) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 20023) ("{" 19117) ("~" 124)) 9268 ((inner_statement_list 20024)) 9269 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 20025) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 9270 ((";" 20026)) 9271 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 20028) (variable 16049) (variable_class_name 16050)) 9272 (("," 16062) (possible_comma 20029)) 9273 ((")" 20030)) 9274 ((";" 20031)) 9275 ((inner_statement_list 20033)) 9276 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 20034) ("~" 124)) 9277 ((catch_list 20035)) 9278 ((T_CATCH 20036) (T_FINALLY 20037) (finally_statement 20038)) 9279 (("(" 20042)) 9280 (("{" 20039)) 9281 ((inner_statement_list 20040)) 9282 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 20041) ("~" 124)) 9283 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 20043) (class_name 16111) (name 16112)) 9284 ((T_VARIABLE 16113) (optional_variable 20044) ("|" 16115)) 9285 ((")" 20045)) 9286 (("{" 20046)) 9287 ((inner_statement_list 20047)) 9288 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 20048) ("~" 124)) 9289 ((backup_doc_comment 20050)) 9290 (("{" 20051)) 9291 ((class_statement_list 20052)) 9292 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 20053)) 9293 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20055) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9294 (("%" 280) (")" 20056) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9295 ((":" 20057) (switch_case_list 20058) ("{" 20059)) 9296 ((";" 20065) (case_list 20066)) 9297 ((";" 20060) (case_list 20061)) 9298 ((case_list 20063)) 9299 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 20062)) 9300 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 20064)) 9301 ((case_list 20069)) 9302 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 20067)) 9303 ((";" 20068)) 9304 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 20070)) 9305 ((";" 20071)) 9306 (("," 16412) (";" 20074)) 9307 ((";" 20076)) 9308 ((T_EXTENDS 19717) (interface_extends_list 20078)) 9309 ((backup_doc_comment 20079)) 9310 (("{" 20080)) 9311 ((class_statement_list 20081)) 9312 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 20082)) 9313 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20084) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9314 (("%" 280) (")" 20085) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9315 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19727) (";" 19125) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19126) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19127) (T_DEC 22) (T_DECLARE 19128) (T_DIR 24) (T_DNUMBER 25) (T_DO 19129) (T_DOUBLE_CAST 27) (T_ECHO 19130) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19131) (T_FOREACH 19132) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19133) (T_GOTO 19134) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19135) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19136) (T_START_HEREDOC 67) (T_STATIC 19137) (T_STRING 19138) (T_STRING_CAST 70) (T_SWITCH 19139) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19140) (T_UNSET 19141) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19142) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19143) (alt_if_stmt_without_else 19144) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19145) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19146) (if_stmt_without_else 19147) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20086) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19149) ("~" 124)) 9316 ((")" 20088)) 9317 ((";" 20089)) 9318 ((";" 20091)) 9319 (("," 17413) (";" 20093)) 9320 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20095) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 9321 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 20096) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 9322 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 20097) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 9323 ((")" 20098) (T_DOUBLE_ARROW 20099)) 9324 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20102) (";" 19076) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19108) (fn 102) (foreach_statement 20108) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20104) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 9325 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 20100) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 9326 ((")" 20101)) 9327 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20102) (";" 19076) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19108) (fn 102) (foreach_statement 20103) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20104) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 9328 ((inner_statement_list 20105)) 9329 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 20106) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 9330 ((";" 20107)) 9331 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 20110) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 9332 ((";" 20111)) 9333 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 20112) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 9334 ((";" 20113)) 9335 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 20114) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 9336 ((")" 20115)) 9337 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20116) (";" 19076) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19108) (fn 102) (for_statement 20117) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20118) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 9338 ((inner_statement_list 20119)) 9339 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 20120) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 9340 ((";" 20121)) 9341 ((":" 19766) (enum_backing_type 20123)) 9342 ((T_IMPLEMENTS 2904) (implements_list 20124)) 9343 ((backup_doc_comment 20125)) 9344 (("{" 20126)) 9345 ((class_statement_list 20127)) 9346 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 20128)) 9347 (("," 19251) (";" 20131)) 9348 ((T_WHILE 20133)) 9349 (("(" 20134)) 9350 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20135) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9351 (("%" 280) (")" 20136) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9352 ((";" 20137)) 9353 ((T_STRING 19424) (const_decl 19425) (const_list 20139)) 9354 ((")" 20140) ("," 19428)) 9355 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20141) (";" 19076) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 20142) (dereferenceable_scalar 99) (expr 19108) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20143) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 9356 ((inner_statement_list 20144)) 9357 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 20145) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9358 ((";" 20146)) 9359 ((";" 20148)) 9360 ((T_EXTENDS 2902) (extends_from 20150)) 9361 ((T_IMPLEMENTS 2904) (implements_list 20151)) 9362 ((backup_doc_comment 20152)) 9363 (("{" 20153)) 9364 ((class_statement_list 20154)) 9365 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 20155)) 9366 ((";" 20157)) 9367 (("," 19251) (";" 20159)) 9368 ((T_WHILE 20161)) 9369 (("(" 20162)) 9370 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20163) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9371 (("%" 280) (")" 20164) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9372 ((";" 20165)) 9373 ((T_STRING 19424) (const_decl 19425) (const_list 20167)) 9374 ((")" 20168) ("," 19428)) 9375 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20169) (";" 18947) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18948) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18949) (T_DEC 22) (T_DECLARE 18950) (T_DIR 24) (T_DNUMBER 25) (T_DO 18951) (T_DOUBLE_CAST 27) (T_ECHO 18952) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18953) (T_FOREACH 18954) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18955) (T_GOTO 18956) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18958) (T_START_HEREDOC 67) (T_STATIC 18959) (T_STRING 18960) (T_STRING_CAST 70) (T_SWITCH 18961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18962) (T_UNSET 18963) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18964) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18965) (alt_if_stmt_without_else 18966) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 20170) (dereferenceable_scalar 99) (expr 18967) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18968) (if_stmt_without_else 18969) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20171) (static_member 118) (variable 121) (variable_class_name 122) ("{" 18971) ("~" 124)) 9376 ((inner_statement_list 20172)) 9377 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 20173) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9378 ((";" 20174)) 9379 ((";" 20176)) 9380 ((";" 20178)) 9381 ((T_STRING 20181) (backup_doc_comment 137)) 9382 ((backup_doc_comment 20182)) 9383 (("(" 20183)) 9384 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 20184)) 9385 ((")" 20185)) 9386 ((":" 150) (return_type 20186)) 9387 ((backup_fn_flags 20187)) 9388 (("{" 20188)) 9389 ((inner_statement_list 20189)) 9390 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 20190) ("~" 124)) 9391 ((backup_fn_flags 20191)) 9392 ((T_STRING 20194)) 9393 ((T_EXTENDS 2902) (extends_from 20195)) 9394 ((T_IMPLEMENTS 2904) (implements_list 20196)) 9395 ((backup_doc_comment 20197)) 9396 (("{" 20198)) 9397 ((class_statement_list 20199)) 9398 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 20200)) 9399 ((":" 20205)) 9400 ((";" 20204)) 9401 ((inner_statement_list 20206)) 9402 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 20207) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 9403 ((";" 20208)) 9404 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20210) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9405 (("%" 280) (")" 20211) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9406 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20212) (";" 18898) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 18930) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20213) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 20214) ("{" 18939) ("~" 124)) 9407 ((inner_statement_list 20215)) 9408 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 20216) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 9409 ((";" 20217)) 9410 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 20219) (variable 16049) (variable_class_name 16050)) 9411 (("," 16062) (possible_comma 20220)) 9412 ((")" 20221)) 9413 ((";" 20222)) 9414 ((inner_statement_list 20224)) 9415 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 20225) ("~" 124)) 9416 ((catch_list 20226)) 9417 ((T_CATCH 20227) (T_FINALLY 20228) (finally_statement 20229)) 9418 (("(" 20233)) 9419 (("{" 20230)) 9420 ((inner_statement_list 20231)) 9421 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 20232) ("~" 124)) 9422 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 20234) (class_name 16111) (name 16112)) 9423 ((T_VARIABLE 16113) (optional_variable 20235) ("|" 16115)) 9424 ((")" 20236)) 9425 (("{" 20237)) 9426 ((inner_statement_list 20238)) 9427 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 20239) ("~" 124)) 9428 ((backup_doc_comment 20241)) 9429 (("{" 20242)) 9430 ((class_statement_list 20243)) 9431 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 20244)) 9432 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20246) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9433 (("%" 280) (")" 20247) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9434 ((":" 20248) (switch_case_list 20249) ("{" 20250)) 9435 ((";" 20256) (case_list 20257)) 9436 ((";" 20251) (case_list 20252)) 9437 ((case_list 20254)) 9438 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 20253)) 9439 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 20255)) 9440 ((case_list 20260)) 9441 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 20258)) 9442 ((";" 20259)) 9443 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 20261)) 9444 ((";" 20262)) 9445 (("," 16412) (";" 20265)) 9446 ((";" 20267)) 9447 ((T_EXTENDS 19717) (interface_extends_list 20269)) 9448 ((backup_doc_comment 20270)) 9449 (("{" 20271)) 9450 ((class_statement_list 20272)) 9451 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 20273)) 9452 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20275) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9453 (("%" 280) (")" 20276) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9454 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19727) (";" 18947) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18948) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18949) (T_DEC 22) (T_DECLARE 18950) (T_DIR 24) (T_DNUMBER 25) (T_DO 18951) (T_DOUBLE_CAST 27) (T_ECHO 18952) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18953) (T_FOREACH 18954) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18955) (T_GOTO 18956) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18958) (T_START_HEREDOC 67) (T_STATIC 18959) (T_STRING 18960) (T_STRING_CAST 70) (T_SWITCH 18961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18962) (T_UNSET 18963) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18964) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18965) (alt_if_stmt_without_else 18966) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 18967) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18968) (if_stmt_without_else 18969) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20277) (static_member 118) (variable 121) (variable_class_name 122) ("{" 18971) ("~" 124)) 9455 ((")" 20279)) 9456 ((";" 20280)) 9457 ((";" 20282)) 9458 (("," 17413) (";" 20284)) 9459 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20286) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 9460 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 20287) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 9461 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 20288) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 9462 ((")" 20289) (T_DOUBLE_ARROW 20290)) 9463 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20293) (";" 18898) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 18930) (fn 102) (foreach_statement 20299) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20295) (static_member 118) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 9464 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 20291) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 9465 ((")" 20292)) 9466 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20293) (";" 18898) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 18930) (fn 102) (foreach_statement 20294) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20295) (static_member 118) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 9467 ((inner_statement_list 20296)) 9468 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 20297) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 9469 ((";" 20298)) 9470 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 20301) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 9471 ((";" 20302)) 9472 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 20303) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 9473 ((";" 20304)) 9474 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 20305) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 9475 ((")" 20306)) 9476 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20307) (";" 18898) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 18930) (fn 102) (for_statement 20308) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20309) (static_member 118) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 9477 ((inner_statement_list 20310)) 9478 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 20311) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 9479 ((";" 20312)) 9480 ((":" 19766) (enum_backing_type 20314)) 9481 ((T_IMPLEMENTS 2904) (implements_list 20315)) 9482 ((backup_doc_comment 20316)) 9483 (("{" 20317)) 9484 ((class_statement_list 20318)) 9485 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 20319)) 9486 (("," 19251) (";" 20322)) 9487 ((T_WHILE 20324)) 9488 (("(" 20325)) 9489 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20326) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9490 (("%" 280) (")" 20327) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9491 ((";" 20328)) 9492 ((T_STRING 19424) (const_decl 19425) (const_list 20330)) 9493 ((")" 20331) ("," 19428)) 9494 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20332) (";" 18898) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 20333) (dereferenceable_scalar 99) (expr 18930) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 20334) (static_member 118) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 9495 ((inner_statement_list 20335)) 9496 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 20336) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9497 ((";" 20337)) 9498 ((";" 20339)) 9499 ((T_EXTENDS 2902) (extends_from 20341)) 9500 ((T_IMPLEMENTS 2904) (implements_list 20342)) 9501 ((backup_doc_comment 20343)) 9502 (("{" 20344)) 9503 ((class_statement_list 20345)) 9504 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 20346)) 9505 ((";" 20348)) 9506 (("$" 18865) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 18867) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 20350) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 20351) ("{" 914)) 9507 (("(" 18884) (argument_list 20352)) 9508 (("$" 18865) (T_STRING 20362) (T_VARIABLE 18867) (property_name 20369) (simple_variable 20364) ("{" 20365)) 9509 (("$" 18865) (T_STRING 20362) (T_VARIABLE 18867) (property_name 20363) (simple_variable 20364) ("{" 20365)) 9510 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 20360) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 9511 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20358) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9512 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20359)) 9513 (("]" 20361)) 9514 (("(" 18884) (argument_list 20368)) 9515 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20366) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9516 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20367)) 9517 (("(" 18884) (argument_list 20370)) 9518 (("]" 20373)) 9519 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 20375) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 9520 ((")" 20376)) 9521 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20379) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9522 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20380)) 9523 (("$" 18835) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 18837) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 20383) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 20384) ("{" 914)) 9524 (("(" 18854) (argument_list 20385)) 9525 (("$" 18835) (T_STRING 20395) (T_VARIABLE 18837) (property_name 20402) (simple_variable 20397) ("{" 20398)) 9526 (("$" 18835) (T_STRING 20395) (T_VARIABLE 18837) (property_name 20396) (simple_variable 20397) ("{" 20398)) 9527 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 20393) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 9528 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20391) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9529 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20392)) 9530 (("]" 20394)) 9531 (("(" 18854) (argument_list 20401)) 9532 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20399) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9533 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20400)) 9534 (("(" 18854) (argument_list 20403)) 9535 (("]" 20406)) 9536 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 20408) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 9537 ((")" 20409)) 9538 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20412) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9539 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20413)) 9540 (("$" 17423) (T_ABSTRACT 17600) (T_ARRAY 17601) (T_AS 17602) (T_BREAK 17603) (T_CALLABLE 17604) (T_CASE 17605) (T_CATCH 17606) (T_CLASS 17607) (T_CLASS_C 17608) (T_CLONE 17609) (T_CONST 17610) (T_CONTINUE 17611) (T_DECLARE 17612) (T_DEFAULT 17613) (T_DIR 17614) (T_DO 17615) (T_ECHO 17616) (T_ELSE 17617) (T_ELSEIF 17618) (T_EMPTY 17619) (T_ENDDECLARE 17620) (T_ENDFOR 17621) (T_ENDFOREACH 17622) (T_ENDIF 17623) (T_ENDSWITCH 17624) (T_ENDWHILE 17625) (T_ENUM 17626) (T_EVAL 17627) (T_EXIT 17628) (T_EXTENDS 17629) (T_FILE 17630) (T_FINAL 17631) (T_FINALLY 17632) (T_FN 17633) (T_FOR 17634) (T_FOREACH 17635) (T_FUNCTION 17636) (T_FUNC_C 17637) (T_GLOBAL 17638) (T_GOTO 17639) (T_IF 17640) (T_IMPLEMENTS 17641) (T_INCLUDE 17642) (T_INCLUDE_ONCE 17643) (T_INSTANCEOF 17644) (T_INSTEADOF 17645) (T_INTERFACE 17646) (T_ISSET 17647) (T_LINE 17648) (T_LIST 17649) (T_LOGICAL_AND 17650) (T_LOGICAL_OR 17651) (T_LOGICAL_XOR 17652) (T_MATCH 17653) (T_METHOD_C 17654) (T_NAMESPACE 17655) (T_NEW 17656) (T_NS_C 17657) (T_PRINT 17658) (T_PRIVATE 17659) (T_PROTECTED 17660) (T_PUBLIC 17661) (T_READONLY 17662) (T_REQUIRE 17663) (T_REQUIRE_ONCE 17664) (T_RETURN 17665) (T_STATIC 17666) (T_STRING 17667) (T_SWITCH 17668) (T_THROW 17669) (T_TRAIT 17670) (T_TRAIT_C 17671) (T_TRY 17672) (T_UNSET 17673) (T_USE 17674) (T_VAR 17675) (T_VARIABLE 17469) (T_WHILE 17676) (T_YIELD 17677) (identifier 20415) (member_name 20416) (reserved_non_modifiers 17680) (semi_reserved 17681) (simple_variable 20417) ("{" 914)) 9541 (("(" 17683) (argument_list 20418)) 9542 ((T_FN 35) (T_FUNCTION 126) (fn 17483) #'17484 (inline_function 20422)) 9543 (("$" 17423) (T_STRING 20431) (T_VARIABLE 17469) (property_name 20438) (simple_variable 20433) ("{" 20434)) 9544 (("$" 17423) (T_STRING 20431) (T_VARIABLE 17469) (property_name 20432) (simple_variable 20433) ("{" 20434)) 9545 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 20429) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 9546 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20427) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9547 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20428)) 9548 (("]" 20430)) 9549 (("(" 17683) (argument_list 20437)) 9550 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20435) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9551 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20436)) 9552 (("(" 17683) (argument_list 20439)) 9553 (("`" 20441)) 9554 (("]" 20443)) 9555 (("=" 20444)) 9556 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20445) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 9557 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20942) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9558 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 20940) (encaps_var 5158)) 9559 (("$" 20449) (T_VARIABLE 20495) (simple_variable 20936) ("{" 20937)) 9560 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20934) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9561 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20933) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9562 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20932) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9563 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20931) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9564 (("(" 20928)) 9565 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20927) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9566 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20926) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9567 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20925) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9568 (("\"" 1489) ("$" 20748) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 20749) ("[" 1508) (array_object_dereferenceable 20750) (callable_expr 20751) (callable_variable 20752) (class_constant 1512) (class_name 20753) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 20754) (name 20755) (simple_variable 20756) (static_member 20757) (variable 20924) (variable_class_name 20759)) 9569 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20923) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9570 (("(" 20920)) 9571 (("(" 20917)) 9572 (("(" 20913) (exit_expr 20914)) 9573 (("\"" 1489) ("$" 20748) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 20749) ("[" 1508) (array_object_dereferenceable 20750) (callable_expr 20751) (callable_variable 20752) (class_constant 1512) (class_name 20753) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 20754) (name 20755) (simple_variable 20756) (static_member 20757) (variable 20912) (variable_class_name 20759)) 9574 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20911) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9575 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20910) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9576 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20909) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9577 (("(" 20905)) 9578 (("(" 20900)) 9579 (("(" 20894)) 9580 (("$" 18504) ("(" 18629) (T_ATTRIBUTE 2853) (T_CLASS 20873) (T_NAME_FULLY_QUALIFIED 18631) (T_NAME_QUALIFIED 18632) (T_NAME_RELATIVE 18633) (T_STATIC 18634) (T_STRING 18635) (T_VARIABLE 18505) (anonymous_class 20874) (attribute 2862) (attributes 20875) (class_name 18638) (class_name_reference 20876) (name 18640) (new_variable 18641) (simple_variable 18642)) 9581 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20872) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9582 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20871) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9583 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20870) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9584 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20869) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9585 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 20864) (T_END_HEREDOC 20865) (T_VARIABLE 2822) (encaps_list 20866) (encaps_var 2824)) 9586 ((T_FN 35) (T_FUNCTION 126) (fn 20509) #'20510 (inline_function 20863)) 9587 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20862) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9588 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20861) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9589 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20860) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9590 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20857) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9591 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20856) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9592 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 20852) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 9593 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 20850) (encaps_list 2729) (encaps_var 2730)) 9594 ((T_NULLSAFE_OBJECT_OPERATOR 20833) (T_OBJECT_OPERATOR 20834) ("[" 20835) ("{" 20836)) 9595 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 20830) (attribute 2654) (fn 20509) #'20510 (inline_function 20831)) 9596 (("(" 20709) (argument_list 20829)) 9597 ((T_PAAMAYIM_NEKUDOTAYIM 20824)) 9598 (("%" 20524) ("*" 20525) ("+" 20526) ("-" 20527) ("." 20528) ("/" 20529) ("<" 20530) (">" 20531) ("?" 20532) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 20533) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 20534) (T_BOOLEAN_AND 20535) (T_BOOLEAN_OR 20536) (T_COALESCE 20537) (T_DOUBLE_ARROW 20822) (T_INSTANCEOF 20538) (T_IS_EQUAL 20539) (T_IS_GREATER_OR_EQUAL 20540) (T_IS_IDENTICAL 20541) (T_IS_NOT_EQUAL 20542) (T_IS_NOT_IDENTICAL 20543) (T_IS_SMALLER_OR_EQUAL 20544) (T_LOGICAL_AND 20545) (T_LOGICAL_OR 20546) (T_LOGICAL_XOR 20547) (T_POW 20548) (T_SL 20549) (T_SPACESHIP 20550) (T_SR 20551) ("^" 20552) ("|" 20553)) 9599 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 20811)) 9600 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 20799)) 9601 (("(" 20709) (argument_list 20798)) 9602 (("=" 20717) (T_AND_EQUAL 20718) (T_COALESCE_EQUAL 20719) (T_CONCAT_EQUAL 20720) (T_DEC 20721) (T_DIV_EQUAL 20722) (T_INC 20723) (T_MINUS_EQUAL 20724) (T_MOD_EQUAL 20725) (T_MUL_EQUAL 20726) (T_OR_EQUAL 20727) (T_PLUS_EQUAL 20728) (T_POW_EQUAL 20729) (T_SL_EQUAL 20730) (T_SR_EQUAL 20731) (T_XOR_EQUAL 20732)) 9603 ((T_PAAMAYIM_NEKUDOTAYIM 20625)) 9604 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20523) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9605 (("%" 20524) ("*" 20525) ("+" 20526) ("-" 20527) ("." 20528) ("/" 20529) ("<" 20530) (">" 20531) ("?" 20532) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 20533) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 20534) (T_BOOLEAN_AND 20535) (T_BOOLEAN_OR 20536) (T_COALESCE 20537) (T_INSTANCEOF 20538) (T_IS_EQUAL 20539) (T_IS_GREATER_OR_EQUAL 20540) (T_IS_IDENTICAL 20541) (T_IS_NOT_EQUAL 20542) (T_IS_NOT_IDENTICAL 20543) (T_IS_SMALLER_OR_EQUAL 20544) (T_LOGICAL_AND 20545) (T_LOGICAL_OR 20546) (T_LOGICAL_XOR 20547) (T_POW 20548) (T_SL 20549) (T_SPACESHIP 20550) (T_SR 20551) ("^" 20552) ("|" 20553)) 9606 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20624) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9607 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20623) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9608 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20622) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9609 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20621) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9610 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20620) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9611 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20619) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9612 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20618) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9613 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20617) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9614 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 20612) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 20613) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 9615 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20611) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9616 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20610) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9617 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20609) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9618 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20608) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9619 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20607) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9620 (("$" 20569) ("(" 20570) (T_NAME_FULLY_QUALIFIED 20571) (T_NAME_QUALIFIED 20572) (T_NAME_RELATIVE 20573) (T_STATIC 20574) (T_STRING 20575) (T_VARIABLE 20576) (class_name 20577) (class_name_reference 20578) (name 20579) (new_variable 20580) (simple_variable 20581)) 9621 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20568) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9622 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20567) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9623 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20566) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9624 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20565) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9625 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20564) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9626 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20563) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9627 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20562) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9628 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20561) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9629 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20560) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9630 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20559) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9631 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20558) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9632 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20557) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9633 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20556) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9634 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20555) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9635 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20554) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9636 (("$" 20569) (T_VARIABLE 20576) (simple_variable 20603) ("{" 20604)) 9637 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20601) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9638 ((T_PAAMAYIM_NEKUDOTAYIM 20599)) 9639 ((T_NULLSAFE_OBJECT_OPERATOR 20582) (T_OBJECT_OPERATOR 20583) (T_PAAMAYIM_NEKUDOTAYIM 20584) ("[" 20585) ("{" 20586)) 9640 (("$" 20569) (T_STRING 20592) (T_VARIABLE 20576) (property_name 20598) (simple_variable 20594) ("{" 20595)) 9641 (("$" 20569) (T_STRING 20592) (T_VARIABLE 20576) (property_name 20593) (simple_variable 20594) ("{" 20595)) 9642 (("$" 20569) (T_VARIABLE 20576) (simple_variable 20591)) 9643 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 20589) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 9644 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20587) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9645 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20588)) 9646 (("]" 20590)) 9647 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20596) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9648 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20597)) 9649 (("$" 20569) (T_VARIABLE 20576) (simple_variable 20600)) 9650 (("%" 280) (")" 20602) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9651 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20605) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9652 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20606)) 9653 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20616) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9654 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 20614) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 9655 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20615) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9656 (("$" 20449) (T_ABSTRACT 20626) (T_ARRAY 20627) (T_AS 20628) (T_BREAK 20629) (T_CALLABLE 20630) (T_CASE 20631) (T_CATCH 20632) (T_CLASS 20633) (T_CLASS_C 20634) (T_CLONE 20635) (T_CONST 20636) (T_CONTINUE 20637) (T_DECLARE 20638) (T_DEFAULT 20639) (T_DIR 20640) (T_DO 20641) (T_ECHO 20642) (T_ELSE 20643) (T_ELSEIF 20644) (T_EMPTY 20645) (T_ENDDECLARE 20646) (T_ENDFOR 20647) (T_ENDFOREACH 20648) (T_ENDIF 20649) (T_ENDSWITCH 20650) (T_ENDWHILE 20651) (T_ENUM 20652) (T_EVAL 20653) (T_EXIT 20654) (T_EXTENDS 20655) (T_FILE 20656) (T_FINAL 20657) (T_FINALLY 20658) (T_FN 20659) (T_FOR 20660) (T_FOREACH 20661) (T_FUNCTION 20662) (T_FUNC_C 20663) (T_GLOBAL 20664) (T_GOTO 20665) (T_IF 20666) (T_IMPLEMENTS 20667) (T_INCLUDE 20668) (T_INCLUDE_ONCE 20669) (T_INSTANCEOF 20670) (T_INSTEADOF 20671) (T_INTERFACE 20672) (T_ISSET 20673) (T_LINE 20674) (T_LIST 20675) (T_LOGICAL_AND 20676) (T_LOGICAL_OR 20677) (T_LOGICAL_XOR 20678) (T_MATCH 20679) (T_METHOD_C 20680) (T_NAMESPACE 20681) (T_NEW 20682) (T_NS_C 20683) (T_PRINT 20684) (T_PRIVATE 20685) (T_PROTECTED 20686) (T_PUBLIC 20687) (T_READONLY 20688) (T_REQUIRE 20689) (T_REQUIRE_ONCE 20690) (T_RETURN 20691) (T_STATIC 20692) (T_STRING 20693) (T_SWITCH 20694) (T_THROW 20695) (T_TRAIT 20696) (T_TRAIT_C 20697) (T_TRY 20698) (T_UNSET 20699) (T_USE 20700) (T_VAR 20701) (T_VARIABLE 20495) (T_WHILE 20702) (T_YIELD 20703) (identifier 20704) (member_name 20705) (reserved_non_modifiers 20706) (semi_reserved 20707) (simple_variable 20708) ("{" 914)) 9657 (("(" 20709) (argument_list 20710)) 9658 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 20711) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 20712) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 20713) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 9659 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 20716) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 9660 (("," 1449) (possible_comma 20714)) 9661 ((")" 20715)) 9662 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (ampersand 20746) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20747) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9663 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20745) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9664 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20744) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9665 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20743) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9666 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20742) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9667 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20741) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9668 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20740) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9669 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20739) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9670 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20738) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9671 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20737) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9672 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20736) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9673 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20735) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9674 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20734) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9675 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20733) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9676 (("\"" 1489) ("$" 20748) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 20749) ("[" 1508) (array_object_dereferenceable 20750) (callable_expr 20751) (callable_variable 20752) (class_constant 1512) (class_name 20753) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 20754) (name 20755) (simple_variable 20756) (static_member 20757) (variable 20758) (variable_class_name 20759)) 9677 (("$" 20748) (T_VARIABLE 20749) (simple_variable 20794) ("{" 20795)) 9678 ((T_NULLSAFE_OBJECT_OPERATOR 20777) (T_OBJECT_OPERATOR 20778) ("[" 20779) ("{" 20780)) 9679 (("(" 20763) (argument_list 20776)) 9680 ((T_PAAMAYIM_NEKUDOTAYIM 20772)) 9681 (("(" 20763) (argument_list 20771)) 9682 ((T_PAAMAYIM_NEKUDOTAYIM 20760)) 9683 (("$" 20748) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 20749) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 20761) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 20762) ("{" 914)) 9684 (("(" 20763) (argument_list 20764)) 9685 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 20765) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 20766) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 20767) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 9686 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 20770) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 9687 (("," 1449) (possible_comma 20768)) 9688 ((")" 20769)) 9689 (("$" 20748) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 20749) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 20773) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 20774) ("{" 914)) 9690 (("(" 20763) (argument_list 20775)) 9691 (("$" 20748) (T_STRING 20785) (T_VARIABLE 20749) (property_name 20792) (simple_variable 20787) ("{" 20788)) 9692 (("$" 20748) (T_STRING 20785) (T_VARIABLE 20749) (property_name 20786) (simple_variable 20787) ("{" 20788)) 9693 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 20783) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 9694 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20781) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9695 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20782)) 9696 (("]" 20784)) 9697 (("(" 20763) (argument_list 20791)) 9698 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20789) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9699 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20790)) 9700 (("(" 20763) (argument_list 20793)) 9701 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20796) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9702 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20797)) 9703 ((backup_doc_comment 20800)) 9704 (("(" 20801)) 9705 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 20802)) 9706 ((")" 20803)) 9707 ((T_USE 148) (lexical_vars 20804)) 9708 ((":" 150) (return_type 20805)) 9709 ((backup_fn_flags 20806)) 9710 (("{" 20807)) 9711 ((inner_statement_list 20808)) 9712 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 20809) ("~" 124)) 9713 ((backup_fn_flags 20810)) 9714 ((backup_doc_comment 20812)) 9715 (("(" 20813)) 9716 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 20814)) 9717 ((")" 20815)) 9718 ((":" 2207) (return_type 20816)) 9719 ((T_DOUBLE_ARROW 20817)) 9720 ((backup_fn_flags 20818)) 9721 ((backup_lex_pos 20819)) 9722 (("!" 18203) ("\"" 18204) ("$" 18205) ("(" 18206) ("+" 18207) ("-" 18208) ("@" 18209) (T_ARRAY 18210) (T_ARRAY_CAST 18211) (T_ATTRIBUTE 125) (T_BOOL_CAST 18212) (T_CLASS_C 18213) (T_CLONE 18214) (T_CONSTANT_ENCAPSED_STRING 18215) (T_DEC 18216) (T_DIR 18217) (T_DNUMBER 18218) (T_DOUBLE_CAST 18219) (T_EMPTY 18220) (T_EVAL 18221) (T_EXIT 18222) (T_FILE 18223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 18224) (T_INC 18225) (T_INCLUDE 18226) (T_INCLUDE_ONCE 18227) (T_INT_CAST 18228) (T_ISSET 18229) (T_LINE 18230) (T_LIST 18231) (T_LNUMBER 18232) (T_MATCH 18233) (T_METHOD_C 18234) (T_NAME_FULLY_QUALIFIED 18235) (T_NAME_QUALIFIED 18236) (T_NAME_RELATIVE 18237) (T_NEW 18238) (T_NS_C 18239) (T_OBJECT_CAST 18240) (T_PRINT 18241) (T_REQUIRE 18242) (T_REQUIRE_ONCE 18243) (T_START_HEREDOC 18244) (T_STATIC 18245) (T_STRING 18246) (T_STRING_CAST 18247) (T_THROW 18248) (T_TRAIT_C 18249) (T_UNSET_CAST 18250) (T_VARIABLE 18251) (T_YIELD 18252) (T_YIELD_FROM 18253) ("[" 18254) ("`" 18255) (array_object_dereferenceable 18256) (attribute 129) (attributes 18257) (callable_expr 18258) (callable_variable 18259) (class_constant 18260) (class_name 18261) (constant 18262) (dereferenceable_scalar 18263) (expr 20820) (fn 18265) (fully_dereferenceable 103) #'18266 (function_call 18267) (inline_function 18268) (internal_functions_in_yacc 18269) (match 18270) (name 18271) (new_expr 18272) (scalar 18273) (simple_variable 18274) (static_member 18275) (variable 18276) (variable_class_name 18277) ("~" 18278)) 9723 (("%" 18280) ("*" 18281) ("+" 18282) ("-" 18283) ("." 18284) ("/" 18285) ("<" 18286) (">" 18287) ("?" 18288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18290) (T_BOOLEAN_AND 18291) (T_BOOLEAN_OR 18292) (T_COALESCE 18293) (T_INSTANCEOF 18294) (T_IS_EQUAL 18295) (T_IS_GREATER_OR_EQUAL 18296) (T_IS_IDENTICAL 18297) (T_IS_NOT_EQUAL 18298) (T_IS_NOT_IDENTICAL 18299) (T_IS_SMALLER_OR_EQUAL 18300) (T_LOGICAL_AND 18301) (T_LOGICAL_OR 18302) (T_LOGICAL_XOR 18303) (T_POW 18304) (T_SL 18305) (T_SPACESHIP 18306) (T_SR 18307) ("^" 18308) (backup_fn_flags 20821) ("|" 18309)) 9724 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20823) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 9725 (("$" 20449) (T_ABSTRACT 20626) (T_ARRAY 20627) (T_AS 20628) (T_BREAK 20629) (T_CALLABLE 20630) (T_CASE 20631) (T_CATCH 20632) (T_CLASS 20633) (T_CLASS_C 20634) (T_CLONE 20635) (T_CONST 20636) (T_CONTINUE 20637) (T_DECLARE 20638) (T_DEFAULT 20639) (T_DIR 20640) (T_DO 20641) (T_ECHO 20642) (T_ELSE 20643) (T_ELSEIF 20644) (T_EMPTY 20645) (T_ENDDECLARE 20646) (T_ENDFOR 20647) (T_ENDFOREACH 20648) (T_ENDIF 20649) (T_ENDSWITCH 20650) (T_ENDWHILE 20651) (T_ENUM 20652) (T_EVAL 20653) (T_EXIT 20654) (T_EXTENDS 20655) (T_FILE 20656) (T_FINAL 20657) (T_FINALLY 20658) (T_FN 20659) (T_FOR 20660) (T_FOREACH 20661) (T_FUNCTION 20662) (T_FUNC_C 20663) (T_GLOBAL 20664) (T_GOTO 20665) (T_IF 20666) (T_IMPLEMENTS 20667) (T_INCLUDE 20668) (T_INCLUDE_ONCE 20669) (T_INSTANCEOF 20670) (T_INSTEADOF 20671) (T_INTERFACE 20672) (T_ISSET 20673) (T_LINE 20674) (T_LIST 20675) (T_LOGICAL_AND 20676) (T_LOGICAL_OR 20677) (T_LOGICAL_XOR 20678) (T_MATCH 20679) (T_METHOD_C 20680) (T_NAMESPACE 20681) (T_NEW 20682) (T_NS_C 20683) (T_PRINT 20684) (T_PRIVATE 20685) (T_PROTECTED 20686) (T_PUBLIC 20687) (T_READONLY 20688) (T_REQUIRE 20689) (T_REQUIRE_ONCE 20690) (T_RETURN 20691) (T_STATIC 20692) (T_STRING 20693) (T_SWITCH 20694) (T_THROW 20695) (T_TRAIT 20696) (T_TRAIT_C 20697) (T_TRY 20698) (T_UNSET 20699) (T_USE 20700) (T_VAR 20701) (T_VARIABLE 20495) (T_WHILE 20702) (T_YIELD 20703) (identifier 20825) (member_name 20826) (reserved_non_modifiers 20706) (semi_reserved 20707) (simple_variable 20827) ("{" 914)) 9726 (("(" 20709) (argument_list 20828)) 9727 ((T_FN 35) (T_FUNCTION 126) (fn 20509) #'20510 (inline_function 20832)) 9728 (("$" 20449) (T_STRING 20841) (T_VARIABLE 20495) (property_name 20848) (simple_variable 20843) ("{" 20844)) 9729 (("$" 20449) (T_STRING 20841) (T_VARIABLE 20495) (property_name 20842) (simple_variable 20843) ("{" 20844)) 9730 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 20839) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 9731 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20837) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9732 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20838)) 9733 (("]" 20840)) 9734 (("(" 20709) (argument_list 20847)) 9735 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20845) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9736 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20846)) 9737 (("(" 20709) (argument_list 20849)) 9738 (("`" 20851)) 9739 (("]" 20853)) 9740 (("=" 20854)) 9741 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20855) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9742 (("%" 20524) ("*" 20525) ("+" 20526) ("-" 20527) ("." 20528) ("/" 20529) ("<" 20530) (">" 20531) ("?" 20532) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 20533) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 20534) (T_BOOLEAN_AND 20535) (T_BOOLEAN_OR 20536) (T_COALESCE 20537) (T_DOUBLE_ARROW 20858) (T_INSTANCEOF 20538) (T_IS_EQUAL 20539) (T_IS_GREATER_OR_EQUAL 20540) (T_IS_IDENTICAL 20541) (T_IS_NOT_EQUAL 20542) (T_IS_NOT_IDENTICAL 20543) (T_IS_SMALLER_OR_EQUAL 20544) (T_LOGICAL_AND 20545) (T_LOGICAL_OR 20546) (T_LOGICAL_XOR 20547) (T_POW 20548) (T_SL 20549) (T_SPACESHIP 20550) (T_SR 20551) ("^" 20552) ("|" 20553)) 9743 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20859) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9744 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 20868) (T_VARIABLE 2822) (encaps_var 2836)) 9745 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 20867) (T_VARIABLE 2822) (encaps_var 2827)) 9746 (("(" 2899) (argument_list 2900) (ctor_arguments 20887)) 9747 ((T_ATTRIBUTE 2853) (T_CLASS 20873) (anonymous_class 20886) (attribute 2898)) 9748 (("(" 20877) (argument_list 20878) (ctor_arguments 20879)) 9749 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 20880) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 20881) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 20882) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 9750 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 20885) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 9751 (("," 1449) (possible_comma 20883)) 9752 ((")" 20884)) 9753 ((T_EXTENDS 2902) (extends_from 20888)) 9754 ((T_IMPLEMENTS 2904) (implements_list 20889)) 9755 ((backup_doc_comment 20890)) 9756 (("{" 20891)) 9757 ((class_statement_list 20892)) 9758 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 20893)) 9759 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20895) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9760 (("%" 280) (")" 20896) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9761 (("{" 20897)) 9762 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 20898) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 9763 (("}" 20899)) 9764 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 20901) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 9765 ((")" 20902)) 9766 (("=" 20903)) 9767 (("!" 20447) ("\"" 20448) ("$" 20449) ("(" 20450) ("+" 20451) ("-" 20452) ("@" 20453) (T_ARRAY 20454) (T_ARRAY_CAST 20455) (T_ATTRIBUTE 125) (T_BOOL_CAST 20456) (T_CLASS_C 20457) (T_CLONE 20458) (T_CONSTANT_ENCAPSED_STRING 20459) (T_DEC 20460) (T_DIR 20461) (T_DNUMBER 20462) (T_DOUBLE_CAST 20463) (T_EMPTY 20464) (T_EVAL 20465) (T_EXIT 20466) (T_FILE 20467) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 20468) (T_INC 20469) (T_INCLUDE 20470) (T_INCLUDE_ONCE 20471) (T_INT_CAST 20472) (T_ISSET 20473) (T_LINE 20474) (T_LIST 20475) (T_LNUMBER 20476) (T_MATCH 20477) (T_METHOD_C 20478) (T_NAME_FULLY_QUALIFIED 20479) (T_NAME_QUALIFIED 20480) (T_NAME_RELATIVE 20481) (T_NEW 20482) (T_NS_C 20483) (T_OBJECT_CAST 20484) (T_PRINT 20485) (T_REQUIRE 20486) (T_REQUIRE_ONCE 20487) (T_START_HEREDOC 20488) (T_STATIC 20489) (T_STRING 20490) (T_STRING_CAST 20491) (T_THROW 20492) (T_TRAIT_C 20493) (T_UNSET_CAST 20494) (T_VARIABLE 20495) (T_YIELD 20496) (T_YIELD_FROM 20497) ("[" 20498) ("`" 20499) (array_object_dereferenceable 20500) (attribute 129) (attributes 20501) (callable_expr 20502) (callable_variable 20503) (class_constant 20504) (class_name 20505) (constant 20506) (dereferenceable_scalar 20507) (expr 20904) (fn 20509) (fully_dereferenceable 103) #'20510 (function_call 20511) (inline_function 20512) (internal_functions_in_yacc 20513) (match 20514) (name 20515) (new_expr 20516) (scalar 20517) (simple_variable 20518) (static_member 20519) (variable 20520) (variable_class_name 20521) ("~" 20522)) 9768 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 20906) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 9769 (("," 5117) (possible_comma 20907)) 9770 ((")" 20908)) 9771 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 20915) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9772 ((")" 20916)) 9773 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20918) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9774 (("%" 280) (")" 20919) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9775 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20921) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9776 (("%" 280) (")" 20922) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9777 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 20929) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 9778 ((")" 20930)) 9779 (("%" 280) (")" 20935) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9780 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 20938) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9781 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 20939)) 9782 (("\"" 20941) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 9783 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 20951) (T_VARIABLE 2822) (encaps_var 2836)) 9784 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 20950) (T_VARIABLE 2822) (encaps_var 2827)) 9785 (("(" 2899) (argument_list 2900) (ctor_arguments 20970)) 9786 ((T_ATTRIBUTE 2853) (T_CLASS 20956) (anonymous_class 20969) (attribute 2898)) 9787 (("(" 20960) (argument_list 20961) (ctor_arguments 20962)) 9788 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 20963) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 20964) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 20965) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 9789 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 20968) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 9790 (("," 1449) (possible_comma 20966)) 9791 ((")" 20967)) 9792 ((T_EXTENDS 2902) (extends_from 20971)) 9793 ((T_IMPLEMENTS 2904) (implements_list 20972)) 9794 ((backup_doc_comment 20973)) 9795 (("{" 20974)) 9796 ((class_statement_list 20975)) 9797 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 20976)) 9798 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 20978) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9799 (("%" 280) (")" 20979) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9800 (("{" 20980)) 9801 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 20981) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 9802 (("}" 20982)) 9803 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 20984) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 9804 ((")" 20985)) 9805 (("=" 20986)) 9806 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 20987) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 9807 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 20989) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 9808 (("," 5117) (possible_comma 20990)) 9809 ((")" 20991)) 9810 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 20998) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9811 ((")" 20999)) 9812 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 21001) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9813 (("%" 280) (")" 21002) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9814 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 21004) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9815 (("%" 280) (")" 21005) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9816 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 21012) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 9817 ((")" 21013)) 9818 (("%" 280) (")" 21018) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9819 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21021) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 9820 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21022)) 9821 (("\"" 21024) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 9822 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 21027) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 9823 ((";" 21028)) 9824 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 21029) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 9825 ((";" 21030)) 9826 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 21031) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 9827 ((")" 21032)) 9828 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21033) (";" 16321) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16322) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16323) (T_DEC 22) (T_DECLARE 16324) (T_DIR 24) (T_DNUMBER 25) (T_DO 16325) (T_DOUBLE_CAST 27) (T_ECHO 16326) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16327) (T_FOREACH 16328) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16329) (T_GOTO 16330) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16331) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16332) (T_START_HEREDOC 67) (T_STATIC 16333) (T_STRING 16334) (T_STRING_CAST 70) (T_SWITCH 16335) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16336) (T_UNSET 16337) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16338) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16339) (alt_if_stmt_without_else 16340) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16341) (fn 102) (for_statement 21034) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16342) (if_stmt_without_else 16343) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 21035) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16345) ("~" 124)) 9829 ((inner_statement_list 21036)) 9830 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 21037) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 9831 ((";" 21038)) 9832 (("," 19251) (";" 21040)) 9833 ((T_WHILE 21042)) 9834 (("(" 21043)) 9835 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 21044) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9836 (("%" 280) (")" 21045) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9837 ((";" 21046)) 9838 ((T_STRING 19424) (const_decl 19425) (const_list 21048)) 9839 ((")" 21049) ("," 19428)) 9840 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21050) (";" 16321) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16322) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16323) (T_DEC 22) (T_DECLARE 16324) (T_DIR 24) (T_DNUMBER 25) (T_DO 16325) (T_DOUBLE_CAST 27) (T_ECHO 16326) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16327) (T_FOREACH 16328) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16329) (T_GOTO 16330) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16331) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16332) (T_START_HEREDOC 67) (T_STATIC 16333) (T_STRING 16334) (T_STRING_CAST 70) (T_SWITCH 16335) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16336) (T_UNSET 16337) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16338) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16339) (alt_if_stmt_without_else 16340) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 21051) (dereferenceable_scalar 99) (expr 16341) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16342) (if_stmt_without_else 16343) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 21052) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16345) ("~" 124)) 9841 ((inner_statement_list 21053)) 9842 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 21054) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9843 ((";" 21055)) 9844 ((";" 21057)) 9845 ((";" 21059)) 9846 ((T_STRING 21062) (backup_doc_comment 137)) 9847 ((backup_doc_comment 21063)) 9848 (("(" 21064)) 9849 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 21065)) 9850 ((")" 21066)) 9851 ((":" 150) (return_type 21067)) 9852 ((backup_fn_flags 21068)) 9853 (("{" 21069)) 9854 ((inner_statement_list 21070)) 9855 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 21071) ("~" 124)) 9856 ((backup_fn_flags 21072)) 9857 ((T_STRING 21075)) 9858 ((T_EXTENDS 2902) (extends_from 21076)) 9859 ((T_IMPLEMENTS 2904) (implements_list 21077)) 9860 ((backup_doc_comment 21078)) 9861 (("{" 21079)) 9862 ((class_statement_list 21080)) 9863 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 21081)) 9864 ((":" 21086)) 9865 ((";" 21085)) 9866 ((inner_statement_list 21087)) 9867 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 21088) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 9868 ((";" 21089)) 9869 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 21091) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9870 (("%" 280) (")" 21092) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9871 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21093) (";" 16273) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16274) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16276) (T_DEC 22) (T_DECLARE 16277) (T_DIR 24) (T_DNUMBER 25) (T_DO 16278) (T_DOUBLE_CAST 27) (T_ECHO 16279) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16281) (T_FOREACH 16282) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16283) (T_GOTO 16284) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16287) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16289) (T_START_HEREDOC 67) (T_STATIC 16290) (T_STRING 16291) (T_STRING_CAST 70) (T_SWITCH 16292) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16294) (T_UNSET 16295) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16296) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16297) (alt_if_stmt_without_else 16298) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16304) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16307) (if_stmt_without_else 16308) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 21094) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 21095) ("{" 16313) ("~" 124)) 9872 ((inner_statement_list 21096)) 9873 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 21097) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 9874 ((";" 21098)) 9875 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 21100) (variable 16049) (variable_class_name 16050)) 9876 (("," 16062) (possible_comma 21101)) 9877 ((")" 21102)) 9878 ((";" 21103)) 9879 ((inner_statement_list 21105)) 9880 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 21106) ("~" 124)) 9881 ((catch_list 21107)) 9882 ((T_CATCH 21108) (T_FINALLY 21109) (finally_statement 21110)) 9883 (("(" 21114)) 9884 (("{" 21111)) 9885 ((inner_statement_list 21112)) 9886 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 21113) ("~" 124)) 9887 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 21115) (class_name 16111) (name 16112)) 9888 ((T_VARIABLE 16113) (optional_variable 21116) ("|" 16115)) 9889 ((")" 21117)) 9890 (("{" 21118)) 9891 ((inner_statement_list 21119)) 9892 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 21120) ("~" 124)) 9893 ((backup_doc_comment 21122)) 9894 (("{" 21123)) 9895 ((class_statement_list 21124)) 9896 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 21125)) 9897 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 21127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9898 (("%" 280) (")" 21128) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9899 ((":" 21129) (switch_case_list 21130) ("{" 21131)) 9900 ((";" 21137) (case_list 21138)) 9901 ((";" 21132) (case_list 21133)) 9902 ((case_list 21135)) 9903 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 21134)) 9904 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 21136)) 9905 ((case_list 21141)) 9906 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 21139)) 9907 ((";" 21140)) 9908 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 21142)) 9909 ((";" 21143)) 9910 (("," 16412) (";" 21146)) 9911 ((";" 21148)) 9912 ((T_EXTENDS 19717) (interface_extends_list 21150)) 9913 ((backup_doc_comment 21151)) 9914 (("{" 21152)) 9915 ((class_statement_list 21153)) 9916 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 21154)) 9917 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 21156) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9918 (("%" 280) (")" 21157) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9919 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19727) (";" 16321) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16322) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16323) (T_DEC 22) (T_DECLARE 16324) (T_DIR 24) (T_DNUMBER 25) (T_DO 16325) (T_DOUBLE_CAST 27) (T_ECHO 16326) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16327) (T_FOREACH 16328) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16329) (T_GOTO 16330) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16331) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16332) (T_START_HEREDOC 67) (T_STATIC 16333) (T_STRING 16334) (T_STRING_CAST 70) (T_SWITCH 16335) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16336) (T_UNSET 16337) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16338) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16339) (alt_if_stmt_without_else 16340) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16341) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16342) (if_stmt_without_else 16343) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 21158) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16345) ("~" 124)) 9920 ((")" 21160)) 9921 ((";" 21161)) 9922 ((";" 21163)) 9923 (("," 17413) (";" 21165)) 9924 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 21167) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 9925 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 21168) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 9926 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 21169) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 9927 ((")" 21170) (T_DOUBLE_ARROW 21171)) 9928 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21174) (";" 16273) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16274) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16276) (T_DEC 22) (T_DECLARE 16277) (T_DIR 24) (T_DNUMBER 25) (T_DO 16278) (T_DOUBLE_CAST 27) (T_ECHO 16279) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16281) (T_FOREACH 16282) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16283) (T_GOTO 16284) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16287) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16289) (T_START_HEREDOC 67) (T_STATIC 16290) (T_STRING 16291) (T_STRING_CAST 70) (T_SWITCH 16292) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16294) (T_UNSET 16295) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16296) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16297) (alt_if_stmt_without_else 16298) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16304) (fn 102) (foreach_statement 21180) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16307) (if_stmt_without_else 16308) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 21176) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16313) ("~" 124)) 9929 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 21172) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 9930 ((")" 21173)) 9931 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21174) (";" 16273) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16274) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16276) (T_DEC 22) (T_DECLARE 16277) (T_DIR 24) (T_DNUMBER 25) (T_DO 16278) (T_DOUBLE_CAST 27) (T_ECHO 16279) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16281) (T_FOREACH 16282) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16283) (T_GOTO 16284) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16287) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16289) (T_START_HEREDOC 67) (T_STATIC 16290) (T_STRING 16291) (T_STRING_CAST 70) (T_SWITCH 16292) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16294) (T_UNSET 16295) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16296) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16297) (alt_if_stmt_without_else 16298) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16304) (fn 102) (foreach_statement 21175) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16307) (if_stmt_without_else 16308) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 21176) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16313) ("~" 124)) 9932 ((inner_statement_list 21177)) 9933 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 21178) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 9934 ((";" 21179)) 9935 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 21182) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 9936 ((";" 21183)) 9937 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 21184) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 9938 ((";" 21185)) 9939 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 21186) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 9940 ((")" 21187)) 9941 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21188) (";" 16273) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16274) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16276) (T_DEC 22) (T_DECLARE 16277) (T_DIR 24) (T_DNUMBER 25) (T_DO 16278) (T_DOUBLE_CAST 27) (T_ECHO 16279) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16281) (T_FOREACH 16282) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16283) (T_GOTO 16284) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16287) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16289) (T_START_HEREDOC 67) (T_STATIC 16290) (T_STRING 16291) (T_STRING_CAST 70) (T_SWITCH 16292) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16294) (T_UNSET 16295) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16296) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16297) (alt_if_stmt_without_else 16298) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16304) (fn 102) (for_statement 21189) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16307) (if_stmt_without_else 16308) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 21190) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16313) ("~" 124)) 9942 ((inner_statement_list 21191)) 9943 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 21192) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 9944 ((";" 21193)) 9945 ((":" 19766) (enum_backing_type 21195)) 9946 ((T_IMPLEMENTS 2904) (implements_list 21196)) 9947 ((backup_doc_comment 21197)) 9948 (("{" 21198)) 9949 ((class_statement_list 21199)) 9950 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 21200)) 9951 (("," 19251) (";" 21202)) 9952 ((T_WHILE 21204)) 9953 (("(" 21205)) 9954 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 21206) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9955 (("%" 280) (")" 21207) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 9956 ((";" 21208)) 9957 ((T_STRING 19424) (const_decl 19425) (const_list 21210)) 9958 ((")" 21211) ("," 19428)) 9959 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21212) (";" 16273) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16274) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16276) (T_DEC 22) (T_DECLARE 16277) (T_DIR 24) (T_DNUMBER 25) (T_DO 16278) (T_DOUBLE_CAST 27) (T_ECHO 16279) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16281) (T_FOREACH 16282) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16283) (T_GOTO 16284) (T_IF 16286) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16287) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16289) (T_START_HEREDOC 67) (T_STATIC 16290) (T_STRING 16291) (T_STRING_CAST 70) (T_SWITCH 16292) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16294) (T_UNSET 16295) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16296) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16297) (alt_if_stmt_without_else 16298) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 21213) (dereferenceable_scalar 99) (expr 16304) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16307) (if_stmt_without_else 16308) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 21214) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16313) ("~" 124)) 9960 ((inner_statement_list 21215)) 9961 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 21216) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 9962 ((";" 21217)) 9963 ((";" 21219)) 9964 ((T_EXTENDS 2902) (extends_from 21221)) 9965 ((T_IMPLEMENTS 2904) (implements_list 21222)) 9966 ((backup_doc_comment 21223)) 9967 (("{" 21224)) 9968 ((class_statement_list 21225)) 9969 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 21226)) 9970 ((";" 21228)) 9971 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23255) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9972 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 23253) (encaps_var 5158)) 9973 (("$" 21231) (T_VARIABLE 21277) (simple_variable 23249) ("{" 23250)) 9974 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23247) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 9975 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23246) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9976 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23245) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9977 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23244) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9978 (("(" 23241)) 9979 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23240) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9980 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23239) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9981 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23238) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9982 (("\"" 1489) ("$" 21530) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 21531) ("[" 1508) (array_object_dereferenceable 21532) (callable_expr 21533) (callable_variable 21534) (class_constant 1512) (class_name 21535) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 21536) (name 21537) (simple_variable 21538) (static_member 21539) (variable 23237) (variable_class_name 21541)) 9983 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23236) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9984 (("(" 23233)) 9985 (("(" 23230)) 9986 (("(" 23226) (exit_expr 23227)) 9987 (("\"" 1489) ("$" 21530) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 21531) ("[" 1508) (array_object_dereferenceable 21532) (callable_expr 21533) (callable_variable 21534) (class_constant 1512) (class_name 21535) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 21536) (name 21537) (simple_variable 21538) (static_member 21539) (variable 23225) (variable_class_name 21541)) 9988 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23224) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9989 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23223) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9990 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23222) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9991 (("(" 23218)) 9992 (("(" 23213)) 9993 (("(" 23207)) 9994 (("$" 21903) ("(" 22546) (T_ATTRIBUTE 2853) (T_CLASS 23186) (T_NAME_FULLY_QUALIFIED 22548) (T_NAME_QUALIFIED 22549) (T_NAME_RELATIVE 22550) (T_STATIC 22551) (T_STRING 22552) (T_VARIABLE 21904) (anonymous_class 23187) (attribute 2862) (attributes 23188) (class_name 22555) (class_name_reference 23189) (name 22557) (new_variable 22558) (simple_variable 22559)) 9995 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23185) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9996 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23184) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9997 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23183) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9998 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23182) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 9999 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 23177) (T_END_HEREDOC 23178) (T_VARIABLE 2822) (encaps_list 23179) (encaps_var 2824)) 10000 ((T_FN 35) (T_FUNCTION 126) (fn 21291) #'21292 (inline_function 23176)) 10001 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23175) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10002 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23174) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10003 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23173) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10004 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22738) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10005 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 22676) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10006 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 22672) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 10007 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 22670) (encaps_list 2729) (encaps_var 2730)) 10008 ((T_NULLSAFE_OBJECT_OPERATOR 22653) (T_OBJECT_OPERATOR 22654) ("[" 22655) ("{" 22656)) 10009 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 22650) (attribute 2654) (fn 21291) #'21292 (inline_function 22651)) 10010 (("(" 21491) (argument_list 22649)) 10011 ((T_PAAMAYIM_NEKUDOTAYIM 22644)) 10012 (("%" 21306) ("*" 21307) ("+" 21308) ("-" 21309) ("." 21310) ("/" 21311) (":" 16269) (";" 16270) ("<" 21312) (">" 21313) ("?" 21314) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 21315) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 21316) (T_BOOLEAN_AND 21317) (T_BOOLEAN_OR 21318) (T_COALESCE 21319) (T_INSTANCEOF 21320) (T_IS_EQUAL 21321) (T_IS_GREATER_OR_EQUAL 21322) (T_IS_IDENTICAL 21323) (T_IS_NOT_EQUAL 21324) (T_IS_NOT_IDENTICAL 21325) (T_IS_SMALLER_OR_EQUAL 21326) (T_LOGICAL_AND 21327) (T_LOGICAL_OR 21328) (T_LOGICAL_XOR 21329) (T_POW 21330) (T_SL 21331) (T_SPACESHIP 21332) (T_SR 21333) ("^" 21334) (case_separator 22642) ("|" 21335)) 10013 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 21593)) 10014 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 21581)) 10015 (("(" 21491) (argument_list 21580)) 10016 (("=" 21499) (T_AND_EQUAL 21500) (T_COALESCE_EQUAL 21501) (T_CONCAT_EQUAL 21502) (T_DEC 21503) (T_DIV_EQUAL 21504) (T_INC 21505) (T_MINUS_EQUAL 21506) (T_MOD_EQUAL 21507) (T_MUL_EQUAL 21508) (T_OR_EQUAL 21509) (T_PLUS_EQUAL 21510) (T_POW_EQUAL 21511) (T_SL_EQUAL 21512) (T_SR_EQUAL 21513) (T_XOR_EQUAL 21514)) 10017 ((T_PAAMAYIM_NEKUDOTAYIM 21407)) 10018 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21305) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10019 (("%" 21306) ("*" 21307) ("+" 21308) ("-" 21309) ("." 21310) ("/" 21311) ("<" 21312) (">" 21313) ("?" 21314) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 21315) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 21316) (T_BOOLEAN_AND 21317) (T_BOOLEAN_OR 21318) (T_COALESCE 21319) (T_INSTANCEOF 21320) (T_IS_EQUAL 21321) (T_IS_GREATER_OR_EQUAL 21322) (T_IS_IDENTICAL 21323) (T_IS_NOT_EQUAL 21324) (T_IS_NOT_IDENTICAL 21325) (T_IS_SMALLER_OR_EQUAL 21326) (T_LOGICAL_AND 21327) (T_LOGICAL_OR 21328) (T_LOGICAL_XOR 21329) (T_POW 21330) (T_SL 21331) (T_SPACESHIP 21332) (T_SR 21333) ("^" 21334) ("|" 21335)) 10020 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21406) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10021 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21405) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10022 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21404) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10023 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21403) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10024 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21402) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10025 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21401) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10026 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21400) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10027 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21399) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10028 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 21394) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 21395) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 10029 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21393) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10030 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21392) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10031 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21391) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10032 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21390) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10033 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21389) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10034 (("$" 21351) ("(" 21352) (T_NAME_FULLY_QUALIFIED 21353) (T_NAME_QUALIFIED 21354) (T_NAME_RELATIVE 21355) (T_STATIC 21356) (T_STRING 21357) (T_VARIABLE 21358) (class_name 21359) (class_name_reference 21360) (name 21361) (new_variable 21362) (simple_variable 21363)) 10035 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21350) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10036 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21349) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10037 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21348) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10038 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21347) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10039 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21346) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10040 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21345) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10041 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21344) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10042 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21343) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10043 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21342) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10044 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21341) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10045 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21340) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10046 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21339) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10047 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21338) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10048 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21337) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10049 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21336) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10050 (("$" 21351) (T_VARIABLE 21358) (simple_variable 21385) ("{" 21386)) 10051 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 21383) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10052 ((T_PAAMAYIM_NEKUDOTAYIM 21381)) 10053 ((T_NULLSAFE_OBJECT_OPERATOR 21364) (T_OBJECT_OPERATOR 21365) (T_PAAMAYIM_NEKUDOTAYIM 21366) ("[" 21367) ("{" 21368)) 10054 (("$" 21351) (T_STRING 21374) (T_VARIABLE 21358) (property_name 21380) (simple_variable 21376) ("{" 21377)) 10055 (("$" 21351) (T_STRING 21374) (T_VARIABLE 21358) (property_name 21375) (simple_variable 21376) ("{" 21377)) 10056 (("$" 21351) (T_VARIABLE 21358) (simple_variable 21373)) 10057 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 21371) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10058 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21369) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10059 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21370)) 10060 (("]" 21372)) 10061 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21378) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10062 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21379)) 10063 (("$" 21351) (T_VARIABLE 21358) (simple_variable 21382)) 10064 (("%" 280) (")" 21384) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10065 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21387) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10066 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21388)) 10067 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21398) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10068 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 21396) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 10069 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21397) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10070 (("$" 21231) (T_ABSTRACT 21408) (T_ARRAY 21409) (T_AS 21410) (T_BREAK 21411) (T_CALLABLE 21412) (T_CASE 21413) (T_CATCH 21414) (T_CLASS 21415) (T_CLASS_C 21416) (T_CLONE 21417) (T_CONST 21418) (T_CONTINUE 21419) (T_DECLARE 21420) (T_DEFAULT 21421) (T_DIR 21422) (T_DO 21423) (T_ECHO 21424) (T_ELSE 21425) (T_ELSEIF 21426) (T_EMPTY 21427) (T_ENDDECLARE 21428) (T_ENDFOR 21429) (T_ENDFOREACH 21430) (T_ENDIF 21431) (T_ENDSWITCH 21432) (T_ENDWHILE 21433) (T_ENUM 21434) (T_EVAL 21435) (T_EXIT 21436) (T_EXTENDS 21437) (T_FILE 21438) (T_FINAL 21439) (T_FINALLY 21440) (T_FN 21441) (T_FOR 21442) (T_FOREACH 21443) (T_FUNCTION 21444) (T_FUNC_C 21445) (T_GLOBAL 21446) (T_GOTO 21447) (T_IF 21448) (T_IMPLEMENTS 21449) (T_INCLUDE 21450) (T_INCLUDE_ONCE 21451) (T_INSTANCEOF 21452) (T_INSTEADOF 21453) (T_INTERFACE 21454) (T_ISSET 21455) (T_LINE 21456) (T_LIST 21457) (T_LOGICAL_AND 21458) (T_LOGICAL_OR 21459) (T_LOGICAL_XOR 21460) (T_MATCH 21461) (T_METHOD_C 21462) (T_NAMESPACE 21463) (T_NEW 21464) (T_NS_C 21465) (T_PRINT 21466) (T_PRIVATE 21467) (T_PROTECTED 21468) (T_PUBLIC 21469) (T_READONLY 21470) (T_REQUIRE 21471) (T_REQUIRE_ONCE 21472) (T_RETURN 21473) (T_STATIC 21474) (T_STRING 21475) (T_SWITCH 21476) (T_THROW 21477) (T_TRAIT 21478) (T_TRAIT_C 21479) (T_TRY 21480) (T_UNSET 21481) (T_USE 21482) (T_VAR 21483) (T_VARIABLE 21277) (T_WHILE 21484) (T_YIELD 21485) (identifier 21486) (member_name 21487) (reserved_non_modifiers 21488) (semi_reserved 21489) (simple_variable 21490) ("{" 914)) 10071 (("(" 21491) (argument_list 21492)) 10072 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 21493) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 21494) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 21495) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10073 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 21498) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10074 (("," 1449) (possible_comma 21496)) 10075 ((")" 21497)) 10076 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (ampersand 21528) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21529) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10077 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21527) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10078 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21526) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10079 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21525) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10080 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21524) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10081 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21523) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10082 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21522) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10083 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21521) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10084 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21520) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10085 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21519) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10086 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21518) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10087 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21517) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10088 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21516) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10089 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 21515) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10090 (("\"" 1489) ("$" 21530) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 21531) ("[" 1508) (array_object_dereferenceable 21532) (callable_expr 21533) (callable_variable 21534) (class_constant 1512) (class_name 21535) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 21536) (name 21537) (simple_variable 21538) (static_member 21539) (variable 21540) (variable_class_name 21541)) 10091 (("$" 21530) (T_VARIABLE 21531) (simple_variable 21576) ("{" 21577)) 10092 ((T_NULLSAFE_OBJECT_OPERATOR 21559) (T_OBJECT_OPERATOR 21560) ("[" 21561) ("{" 21562)) 10093 (("(" 21545) (argument_list 21558)) 10094 ((T_PAAMAYIM_NEKUDOTAYIM 21554)) 10095 (("(" 21545) (argument_list 21553)) 10096 ((T_PAAMAYIM_NEKUDOTAYIM 21542)) 10097 (("$" 21530) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 21531) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 21543) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 21544) ("{" 914)) 10098 (("(" 21545) (argument_list 21546)) 10099 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 21547) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 21548) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 21549) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10100 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 21552) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10101 (("," 1449) (possible_comma 21550)) 10102 ((")" 21551)) 10103 (("$" 21530) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 21531) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 21555) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 21556) ("{" 914)) 10104 (("(" 21545) (argument_list 21557)) 10105 (("$" 21530) (T_STRING 21567) (T_VARIABLE 21531) (property_name 21574) (simple_variable 21569) ("{" 21570)) 10106 (("$" 21530) (T_STRING 21567) (T_VARIABLE 21531) (property_name 21568) (simple_variable 21569) ("{" 21570)) 10107 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 21565) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10108 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21563) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10109 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21564)) 10110 (("]" 21566)) 10111 (("(" 21545) (argument_list 21573)) 10112 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21571) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10113 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21572)) 10114 (("(" 21545) (argument_list 21575)) 10115 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21578) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10116 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21579)) 10117 ((backup_doc_comment 21582)) 10118 (("(" 21583)) 10119 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 21584)) 10120 ((")" 21585)) 10121 ((T_USE 148) (lexical_vars 21586)) 10122 ((":" 150) (return_type 21587)) 10123 ((backup_fn_flags 21588)) 10124 (("{" 21589)) 10125 ((inner_statement_list 21590)) 10126 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 21591) ("~" 124)) 10127 ((backup_fn_flags 21592)) 10128 ((backup_doc_comment 21594)) 10129 (("(" 21595)) 10130 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 21596)) 10131 ((")" 21597)) 10132 ((":" 2207) (return_type 21598)) 10133 ((T_DOUBLE_ARROW 21599)) 10134 ((backup_fn_flags 21600)) 10135 ((backup_lex_pos 21601)) 10136 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21663) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10137 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22641) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10138 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 22639) (encaps_var 5158)) 10139 (("$" 21604) (T_VARIABLE 21650) (simple_variable 22635) ("{" 22636)) 10140 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 22633) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10141 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22632) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10142 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22631) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10143 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22630) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10144 (("(" 22627)) 10145 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22626) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10146 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22625) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10147 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22624) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10148 (("\"" 1489) ("$" 21903) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 21904) ("[" 1508) (array_object_dereferenceable 21905) (callable_expr 21906) (callable_variable 21907) (class_constant 1512) (class_name 21908) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 21909) (name 21910) (simple_variable 21911) (static_member 21912) (variable 22623) (variable_class_name 21914)) 10149 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22622) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10150 (("(" 22619)) 10151 (("(" 22616)) 10152 (("(" 22612) (exit_expr 22613)) 10153 (("\"" 1489) ("$" 21903) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 21904) ("[" 1508) (array_object_dereferenceable 21905) (callable_expr 21906) (callable_variable 21907) (class_constant 1512) (class_name 21908) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 21909) (name 21910) (simple_variable 21911) (static_member 21912) (variable 22611) (variable_class_name 21914)) 10154 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22610) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10155 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22609) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10156 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22608) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10157 (("(" 22604)) 10158 (("(" 22599)) 10159 (("(" 22593)) 10160 (("$" 21903) ("(" 22546) (T_ATTRIBUTE 2853) (T_CLASS 22547) (T_NAME_FULLY_QUALIFIED 22548) (T_NAME_QUALIFIED 22549) (T_NAME_RELATIVE 22550) (T_STATIC 22551) (T_STRING 22552) (T_VARIABLE 21904) (anonymous_class 22553) (attribute 2862) (attributes 22554) (class_name 22555) (class_name_reference 22556) (name 22557) (new_variable 22558) (simple_variable 22559)) 10161 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22545) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10162 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22544) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10163 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22543) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10164 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22542) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10165 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 22537) (T_END_HEREDOC 22538) (T_VARIABLE 2822) (encaps_list 22539) (encaps_var 2824)) 10166 ((T_FN 35) (T_FUNCTION 126) (fn 21664) #'21665 (inline_function 22536)) 10167 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22535) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10168 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22534) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10169 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22533) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10170 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22072) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10171 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22010) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10172 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 22006) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 10173 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 22004) (encaps_list 2729) (encaps_var 2730)) 10174 ((T_NULLSAFE_OBJECT_OPERATOR 21987) (T_OBJECT_OPERATOR 21988) ("[" 21989) ("{" 21990)) 10175 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 21984) (attribute 2654) (fn 21664) #'21665 (inline_function 21985)) 10176 (("(" 21864) (argument_list 21983)) 10177 ((T_PAAMAYIM_NEKUDOTAYIM 21978)) 10178 (("%" 21679) ("*" 21680) ("+" 21681) ("-" 21682) ("." 21683) ("/" 21684) ("<" 21685) (">" 21686) ("?" 21687) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 21688) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 21689) (T_BOOLEAN_AND 21690) (T_BOOLEAN_OR 21691) (T_COALESCE 21692) (T_INSTANCEOF 21693) (T_IS_EQUAL 21694) (T_IS_GREATER_OR_EQUAL 21695) (T_IS_IDENTICAL 21696) (T_IS_NOT_EQUAL 21697) (T_IS_NOT_IDENTICAL 21698) (T_IS_SMALLER_OR_EQUAL 21699) (T_LOGICAL_AND 21700) (T_LOGICAL_OR 21701) (T_LOGICAL_XOR 21702) (T_POW 21703) (T_SL 21704) (T_SPACESHIP 21705) (T_SR 21706) ("^" 21707) (backup_fn_flags 21977) ("|" 21708)) 10179 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 21966)) 10180 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 21954)) 10181 (("(" 21864) (argument_list 21953)) 10182 (("=" 21872) (T_AND_EQUAL 21873) (T_COALESCE_EQUAL 21874) (T_CONCAT_EQUAL 21875) (T_DEC 21876) (T_DIV_EQUAL 21877) (T_INC 21878) (T_MINUS_EQUAL 21879) (T_MOD_EQUAL 21880) (T_MUL_EQUAL 21881) (T_OR_EQUAL 21882) (T_PLUS_EQUAL 21883) (T_POW_EQUAL 21884) (T_SL_EQUAL 21885) (T_SR_EQUAL 21886) (T_XOR_EQUAL 21887)) 10183 ((T_PAAMAYIM_NEKUDOTAYIM 21780)) 10184 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21678) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10185 (("%" 21679) ("*" 21680) ("+" 21681) ("-" 21682) ("." 21683) ("/" 21684) ("<" 21685) (">" 21686) ("?" 21687) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 21688) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 21689) (T_BOOLEAN_AND 21690) (T_BOOLEAN_OR 21691) (T_COALESCE 21692) (T_INSTANCEOF 21693) (T_IS_EQUAL 21694) (T_IS_GREATER_OR_EQUAL 21695) (T_IS_IDENTICAL 21696) (T_IS_NOT_EQUAL 21697) (T_IS_NOT_IDENTICAL 21698) (T_IS_SMALLER_OR_EQUAL 21699) (T_LOGICAL_AND 21700) (T_LOGICAL_OR 21701) (T_LOGICAL_XOR 21702) (T_POW 21703) (T_SL 21704) (T_SPACESHIP 21705) (T_SR 21706) ("^" 21707) ("|" 21708)) 10186 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21779) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10187 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21778) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10188 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21777) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10189 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21776) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10190 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21775) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10191 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21774) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10192 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21773) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10193 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21772) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10194 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 21767) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 21768) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 10195 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21766) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10196 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21765) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10197 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21764) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10198 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21763) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10199 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21762) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10200 (("$" 21724) ("(" 21725) (T_NAME_FULLY_QUALIFIED 21726) (T_NAME_QUALIFIED 21727) (T_NAME_RELATIVE 21728) (T_STATIC 21729) (T_STRING 21730) (T_VARIABLE 21731) (class_name 21732) (class_name_reference 21733) (name 21734) (new_variable 21735) (simple_variable 21736)) 10201 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21723) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10202 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21722) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10203 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21721) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10204 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21720) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10205 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21719) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10206 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21718) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10207 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21717) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10208 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21716) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10209 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21715) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10210 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21714) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10211 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21713) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10212 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21712) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10213 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21711) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10214 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21710) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10215 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21709) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10216 (("$" 21724) (T_VARIABLE 21731) (simple_variable 21758) ("{" 21759)) 10217 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 21756) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10218 ((T_PAAMAYIM_NEKUDOTAYIM 21754)) 10219 ((T_NULLSAFE_OBJECT_OPERATOR 21737) (T_OBJECT_OPERATOR 21738) (T_PAAMAYIM_NEKUDOTAYIM 21739) ("[" 21740) ("{" 21741)) 10220 (("$" 21724) (T_STRING 21747) (T_VARIABLE 21731) (property_name 21753) (simple_variable 21749) ("{" 21750)) 10221 (("$" 21724) (T_STRING 21747) (T_VARIABLE 21731) (property_name 21748) (simple_variable 21749) ("{" 21750)) 10222 (("$" 21724) (T_VARIABLE 21731) (simple_variable 21746)) 10223 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 21744) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10224 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21742) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10225 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21743)) 10226 (("]" 21745)) 10227 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21751) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10228 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21752)) 10229 (("$" 21724) (T_VARIABLE 21731) (simple_variable 21755)) 10230 (("%" 280) (")" 21757) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10231 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21760) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10232 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21761)) 10233 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21771) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10234 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 21769) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 10235 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21770) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10236 (("$" 21604) (T_ABSTRACT 21781) (T_ARRAY 21782) (T_AS 21783) (T_BREAK 21784) (T_CALLABLE 21785) (T_CASE 21786) (T_CATCH 21787) (T_CLASS 21788) (T_CLASS_C 21789) (T_CLONE 21790) (T_CONST 21791) (T_CONTINUE 21792) (T_DECLARE 21793) (T_DEFAULT 21794) (T_DIR 21795) (T_DO 21796) (T_ECHO 21797) (T_ELSE 21798) (T_ELSEIF 21799) (T_EMPTY 21800) (T_ENDDECLARE 21801) (T_ENDFOR 21802) (T_ENDFOREACH 21803) (T_ENDIF 21804) (T_ENDSWITCH 21805) (T_ENDWHILE 21806) (T_ENUM 21807) (T_EVAL 21808) (T_EXIT 21809) (T_EXTENDS 21810) (T_FILE 21811) (T_FINAL 21812) (T_FINALLY 21813) (T_FN 21814) (T_FOR 21815) (T_FOREACH 21816) (T_FUNCTION 21817) (T_FUNC_C 21818) (T_GLOBAL 21819) (T_GOTO 21820) (T_IF 21821) (T_IMPLEMENTS 21822) (T_INCLUDE 21823) (T_INCLUDE_ONCE 21824) (T_INSTANCEOF 21825) (T_INSTEADOF 21826) (T_INTERFACE 21827) (T_ISSET 21828) (T_LINE 21829) (T_LIST 21830) (T_LOGICAL_AND 21831) (T_LOGICAL_OR 21832) (T_LOGICAL_XOR 21833) (T_MATCH 21834) (T_METHOD_C 21835) (T_NAMESPACE 21836) (T_NEW 21837) (T_NS_C 21838) (T_PRINT 21839) (T_PRIVATE 21840) (T_PROTECTED 21841) (T_PUBLIC 21842) (T_READONLY 21843) (T_REQUIRE 21844) (T_REQUIRE_ONCE 21845) (T_RETURN 21846) (T_STATIC 21847) (T_STRING 21848) (T_SWITCH 21849) (T_THROW 21850) (T_TRAIT 21851) (T_TRAIT_C 21852) (T_TRY 21853) (T_UNSET 21854) (T_USE 21855) (T_VAR 21856) (T_VARIABLE 21650) (T_WHILE 21857) (T_YIELD 21858) (identifier 21859) (member_name 21860) (reserved_non_modifiers 21861) (semi_reserved 21862) (simple_variable 21863) ("{" 914)) 10237 (("(" 21864) (argument_list 21865)) 10238 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 21866) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 21867) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 21868) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10239 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 21871) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10240 (("," 1449) (possible_comma 21869)) 10241 ((")" 21870)) 10242 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (ampersand 21901) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21902) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10243 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21900) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10244 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21899) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10245 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21898) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10246 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21897) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10247 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21896) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10248 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21895) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10249 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21894) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10250 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21893) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10251 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21892) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10252 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21891) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10253 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21890) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10254 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21889) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10255 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21888) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10256 (("\"" 1489) ("$" 21903) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 21904) ("[" 1508) (array_object_dereferenceable 21905) (callable_expr 21906) (callable_variable 21907) (class_constant 1512) (class_name 21908) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 21909) (name 21910) (simple_variable 21911) (static_member 21912) (variable 21913) (variable_class_name 21914)) 10257 (("$" 21903) (T_VARIABLE 21904) (simple_variable 21949) ("{" 21950)) 10258 ((T_NULLSAFE_OBJECT_OPERATOR 21932) (T_OBJECT_OPERATOR 21933) ("[" 21934) ("{" 21935)) 10259 (("(" 21918) (argument_list 21931)) 10260 ((T_PAAMAYIM_NEKUDOTAYIM 21927)) 10261 (("(" 21918) (argument_list 21926)) 10262 ((T_PAAMAYIM_NEKUDOTAYIM 21915)) 10263 (("$" 21903) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 21904) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 21916) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 21917) ("{" 914)) 10264 (("(" 21918) (argument_list 21919)) 10265 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 21920) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 21921) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 21922) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10266 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 21925) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10267 (("," 1449) (possible_comma 21923)) 10268 ((")" 21924)) 10269 (("$" 21903) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 21904) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 21928) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 21929) ("{" 914)) 10270 (("(" 21918) (argument_list 21930)) 10271 (("$" 21903) (T_STRING 21940) (T_VARIABLE 21904) (property_name 21947) (simple_variable 21942) ("{" 21943)) 10272 (("$" 21903) (T_STRING 21940) (T_VARIABLE 21904) (property_name 21941) (simple_variable 21942) ("{" 21943)) 10273 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 21938) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10274 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21936) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10275 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21937)) 10276 (("]" 21939)) 10277 (("(" 21918) (argument_list 21946)) 10278 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21944) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10279 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21945)) 10280 (("(" 21918) (argument_list 21948)) 10281 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21951) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10282 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21952)) 10283 ((backup_doc_comment 21955)) 10284 (("(" 21956)) 10285 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 21957)) 10286 ((")" 21958)) 10287 ((T_USE 148) (lexical_vars 21959)) 10288 ((":" 150) (return_type 21960)) 10289 ((backup_fn_flags 21961)) 10290 (("{" 21962)) 10291 ((inner_statement_list 21963)) 10292 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 21964) ("~" 124)) 10293 ((backup_fn_flags 21965)) 10294 ((backup_doc_comment 21967)) 10295 (("(" 21968)) 10296 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 21969)) 10297 ((")" 21970)) 10298 ((":" 2207) (return_type 21971)) 10299 ((T_DOUBLE_ARROW 21972)) 10300 ((backup_fn_flags 21973)) 10301 ((backup_lex_pos 21974)) 10302 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 21975) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10303 (("%" 21679) ("*" 21680) ("+" 21681) ("-" 21682) ("." 21683) ("/" 21684) ("<" 21685) (">" 21686) ("?" 21687) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 21688) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 21689) (T_BOOLEAN_AND 21690) (T_BOOLEAN_OR 21691) (T_COALESCE 21692) (T_INSTANCEOF 21693) (T_IS_EQUAL 21694) (T_IS_GREATER_OR_EQUAL 21695) (T_IS_IDENTICAL 21696) (T_IS_NOT_EQUAL 21697) (T_IS_NOT_IDENTICAL 21698) (T_IS_SMALLER_OR_EQUAL 21699) (T_LOGICAL_AND 21700) (T_LOGICAL_OR 21701) (T_LOGICAL_XOR 21702) (T_POW 21703) (T_SL 21704) (T_SPACESHIP 21705) (T_SR 21706) ("^" 21707) (backup_fn_flags 21976) ("|" 21708)) 10304 (("$" 21604) (T_ABSTRACT 21781) (T_ARRAY 21782) (T_AS 21783) (T_BREAK 21784) (T_CALLABLE 21785) (T_CASE 21786) (T_CATCH 21787) (T_CLASS 21788) (T_CLASS_C 21789) (T_CLONE 21790) (T_CONST 21791) (T_CONTINUE 21792) (T_DECLARE 21793) (T_DEFAULT 21794) (T_DIR 21795) (T_DO 21796) (T_ECHO 21797) (T_ELSE 21798) (T_ELSEIF 21799) (T_EMPTY 21800) (T_ENDDECLARE 21801) (T_ENDFOR 21802) (T_ENDFOREACH 21803) (T_ENDIF 21804) (T_ENDSWITCH 21805) (T_ENDWHILE 21806) (T_ENUM 21807) (T_EVAL 21808) (T_EXIT 21809) (T_EXTENDS 21810) (T_FILE 21811) (T_FINAL 21812) (T_FINALLY 21813) (T_FN 21814) (T_FOR 21815) (T_FOREACH 21816) (T_FUNCTION 21817) (T_FUNC_C 21818) (T_GLOBAL 21819) (T_GOTO 21820) (T_IF 21821) (T_IMPLEMENTS 21822) (T_INCLUDE 21823) (T_INCLUDE_ONCE 21824) (T_INSTANCEOF 21825) (T_INSTEADOF 21826) (T_INTERFACE 21827) (T_ISSET 21828) (T_LINE 21829) (T_LIST 21830) (T_LOGICAL_AND 21831) (T_LOGICAL_OR 21832) (T_LOGICAL_XOR 21833) (T_MATCH 21834) (T_METHOD_C 21835) (T_NAMESPACE 21836) (T_NEW 21837) (T_NS_C 21838) (T_PRINT 21839) (T_PRIVATE 21840) (T_PROTECTED 21841) (T_PUBLIC 21842) (T_READONLY 21843) (T_REQUIRE 21844) (T_REQUIRE_ONCE 21845) (T_RETURN 21846) (T_STATIC 21847) (T_STRING 21848) (T_SWITCH 21849) (T_THROW 21850) (T_TRAIT 21851) (T_TRAIT_C 21852) (T_TRY 21853) (T_UNSET 21854) (T_USE 21855) (T_VAR 21856) (T_VARIABLE 21650) (T_WHILE 21857) (T_YIELD 21858) (identifier 21979) (member_name 21980) (reserved_non_modifiers 21861) (semi_reserved 21862) (simple_variable 21981) ("{" 914)) 10305 (("(" 21864) (argument_list 21982)) 10306 ((T_FN 35) (T_FUNCTION 126) (fn 21664) #'21665 (inline_function 21986)) 10307 (("$" 21604) (T_STRING 21995) (T_VARIABLE 21650) (property_name 22002) (simple_variable 21997) ("{" 21998)) 10308 (("$" 21604) (T_STRING 21995) (T_VARIABLE 21650) (property_name 21996) (simple_variable 21997) ("{" 21998)) 10309 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 21993) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10310 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21991) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10311 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 21992)) 10312 (("]" 21994)) 10313 (("(" 21864) (argument_list 22001)) 10314 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 21999) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10315 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22000)) 10316 (("(" 21864) (argument_list 22003)) 10317 (("`" 22005)) 10318 (("]" 22007)) 10319 (("=" 22008)) 10320 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22009) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10321 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22532) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10322 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 22530) (encaps_var 5158)) 10323 (("$" 22013) (T_VARIABLE 22059) (simple_variable 22526) ("{" 22527)) 10324 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 22524) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10325 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22523) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10326 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22522) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10327 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22521) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10328 (("(" 22518)) 10329 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22517) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10330 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22516) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10331 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22515) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10332 (("\"" 1489) ("$" 22312) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 22313) ("[" 1508) (array_object_dereferenceable 22314) (callable_expr 22315) (callable_variable 22316) (class_constant 1512) (class_name 22317) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 22318) (name 22319) (simple_variable 22320) (static_member 22321) (variable 22514) (variable_class_name 22323)) 10333 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22513) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10334 (("(" 22510)) 10335 (("(" 22507)) 10336 (("(" 22503) (exit_expr 22504)) 10337 (("\"" 1489) ("$" 22312) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 22313) ("[" 1508) (array_object_dereferenceable 22314) (callable_expr 22315) (callable_variable 22316) (class_constant 1512) (class_name 22317) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 22318) (name 22319) (simple_variable 22320) (static_member 22321) (variable 22502) (variable_class_name 22323)) 10338 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22501) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10339 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22500) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10340 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22499) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10341 (("(" 22495)) 10342 (("(" 22490)) 10343 (("(" 22484)) 10344 (("$" 22312) ("(" 22437) (T_ATTRIBUTE 2853) (T_CLASS 22438) (T_NAME_FULLY_QUALIFIED 22439) (T_NAME_QUALIFIED 22440) (T_NAME_RELATIVE 22441) (T_STATIC 22442) (T_STRING 22443) (T_VARIABLE 22313) (anonymous_class 22444) (attribute 2862) (attributes 22445) (class_name 22446) (class_name_reference 22447) (name 22448) (new_variable 22449) (simple_variable 22450)) 10345 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22436) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10346 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22435) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10347 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22434) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10348 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22433) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10349 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 22428) (T_END_HEREDOC 22429) (T_VARIABLE 2822) (encaps_list 22430) (encaps_var 2824)) 10350 ((T_FN 35) (T_FUNCTION 126) (fn 22073) #'22074 (inline_function 22427)) 10351 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22426) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10352 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22425) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10353 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22424) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10354 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22421) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10355 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22420) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10356 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 22416) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 10357 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 22414) (encaps_list 2729) (encaps_var 2730)) 10358 ((T_NULLSAFE_OBJECT_OPERATOR 22397) (T_OBJECT_OPERATOR 22398) ("[" 22399) ("{" 22400)) 10359 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 22394) (attribute 2654) (fn 22073) #'22074 (inline_function 22395)) 10360 (("(" 22273) (argument_list 22393)) 10361 ((T_PAAMAYIM_NEKUDOTAYIM 22388)) 10362 (("%" 22088) ("*" 22089) ("+" 22090) ("-" 22091) ("." 22092) ("/" 22093) ("<" 22094) (">" 22095) ("?" 22096) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22097) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22098) (T_BOOLEAN_AND 22099) (T_BOOLEAN_OR 22100) (T_COALESCE 22101) (T_DOUBLE_ARROW 22386) (T_INSTANCEOF 22102) (T_IS_EQUAL 22103) (T_IS_GREATER_OR_EQUAL 22104) (T_IS_IDENTICAL 22105) (T_IS_NOT_EQUAL 22106) (T_IS_NOT_IDENTICAL 22107) (T_IS_SMALLER_OR_EQUAL 22108) (T_LOGICAL_AND 22109) (T_LOGICAL_OR 22110) (T_LOGICAL_XOR 22111) (T_POW 22112) (T_SL 22113) (T_SPACESHIP 22114) (T_SR 22115) ("^" 22116) ("|" 22117)) 10363 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 22375)) 10364 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 22363)) 10365 (("(" 22273) (argument_list 22362)) 10366 (("=" 22281) (T_AND_EQUAL 22282) (T_COALESCE_EQUAL 22283) (T_CONCAT_EQUAL 22284) (T_DEC 22285) (T_DIV_EQUAL 22286) (T_INC 22287) (T_MINUS_EQUAL 22288) (T_MOD_EQUAL 22289) (T_MUL_EQUAL 22290) (T_OR_EQUAL 22291) (T_PLUS_EQUAL 22292) (T_POW_EQUAL 22293) (T_SL_EQUAL 22294) (T_SR_EQUAL 22295) (T_XOR_EQUAL 22296)) 10367 ((T_PAAMAYIM_NEKUDOTAYIM 22189)) 10368 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22087) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10369 (("%" 22088) ("*" 22089) ("+" 22090) ("-" 22091) ("." 22092) ("/" 22093) ("<" 22094) (">" 22095) ("?" 22096) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22097) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22098) (T_BOOLEAN_AND 22099) (T_BOOLEAN_OR 22100) (T_COALESCE 22101) (T_INSTANCEOF 22102) (T_IS_EQUAL 22103) (T_IS_GREATER_OR_EQUAL 22104) (T_IS_IDENTICAL 22105) (T_IS_NOT_EQUAL 22106) (T_IS_NOT_IDENTICAL 22107) (T_IS_SMALLER_OR_EQUAL 22108) (T_LOGICAL_AND 22109) (T_LOGICAL_OR 22110) (T_LOGICAL_XOR 22111) (T_POW 22112) (T_SL 22113) (T_SPACESHIP 22114) (T_SR 22115) ("^" 22116) ("|" 22117)) 10370 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22188) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10371 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22187) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10372 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22186) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10373 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22185) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10374 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22184) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10375 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22183) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10376 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22182) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10377 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22181) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10378 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 22176) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 22177) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 10379 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22175) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10380 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22174) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10381 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22173) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10382 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22172) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10383 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22171) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10384 (("$" 22133) ("(" 22134) (T_NAME_FULLY_QUALIFIED 22135) (T_NAME_QUALIFIED 22136) (T_NAME_RELATIVE 22137) (T_STATIC 22138) (T_STRING 22139) (T_VARIABLE 22140) (class_name 22141) (class_name_reference 22142) (name 22143) (new_variable 22144) (simple_variable 22145)) 10385 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22132) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10386 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22131) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10387 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22130) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10388 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22129) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10389 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22128) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10390 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22127) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10391 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22126) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10392 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22125) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10393 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22124) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10394 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22123) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10395 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22122) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10396 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22121) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10397 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22120) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10398 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22119) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10399 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22118) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10400 (("$" 22133) (T_VARIABLE 22140) (simple_variable 22167) ("{" 22168)) 10401 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 22165) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10402 ((T_PAAMAYIM_NEKUDOTAYIM 22163)) 10403 ((T_NULLSAFE_OBJECT_OPERATOR 22146) (T_OBJECT_OPERATOR 22147) (T_PAAMAYIM_NEKUDOTAYIM 22148) ("[" 22149) ("{" 22150)) 10404 (("$" 22133) (T_STRING 22156) (T_VARIABLE 22140) (property_name 22162) (simple_variable 22158) ("{" 22159)) 10405 (("$" 22133) (T_STRING 22156) (T_VARIABLE 22140) (property_name 22157) (simple_variable 22158) ("{" 22159)) 10406 (("$" 22133) (T_VARIABLE 22140) (simple_variable 22155)) 10407 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 22153) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10408 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22151) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10409 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22152)) 10410 (("]" 22154)) 10411 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22160) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10412 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22161)) 10413 (("$" 22133) (T_VARIABLE 22140) (simple_variable 22164)) 10414 (("%" 280) (")" 22166) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10415 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22169) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10416 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22170)) 10417 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22180) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10418 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 22178) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 10419 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22179) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10420 (("$" 22013) (T_ABSTRACT 22190) (T_ARRAY 22191) (T_AS 22192) (T_BREAK 22193) (T_CALLABLE 22194) (T_CASE 22195) (T_CATCH 22196) (T_CLASS 22197) (T_CLASS_C 22198) (T_CLONE 22199) (T_CONST 22200) (T_CONTINUE 22201) (T_DECLARE 22202) (T_DEFAULT 22203) (T_DIR 22204) (T_DO 22205) (T_ECHO 22206) (T_ELSE 22207) (T_ELSEIF 22208) (T_EMPTY 22209) (T_ENDDECLARE 22210) (T_ENDFOR 22211) (T_ENDFOREACH 22212) (T_ENDIF 22213) (T_ENDSWITCH 22214) (T_ENDWHILE 22215) (T_ENUM 22216) (T_EVAL 22217) (T_EXIT 22218) (T_EXTENDS 22219) (T_FILE 22220) (T_FINAL 22221) (T_FINALLY 22222) (T_FN 22223) (T_FOR 22224) (T_FOREACH 22225) (T_FUNCTION 22226) (T_FUNC_C 22227) (T_GLOBAL 22228) (T_GOTO 22229) (T_IF 22230) (T_IMPLEMENTS 22231) (T_INCLUDE 22232) (T_INCLUDE_ONCE 22233) (T_INSTANCEOF 22234) (T_INSTEADOF 22235) (T_INTERFACE 22236) (T_ISSET 22237) (T_LINE 22238) (T_LIST 22239) (T_LOGICAL_AND 22240) (T_LOGICAL_OR 22241) (T_LOGICAL_XOR 22242) (T_MATCH 22243) (T_METHOD_C 22244) (T_NAMESPACE 22245) (T_NEW 22246) (T_NS_C 22247) (T_PRINT 22248) (T_PRIVATE 22249) (T_PROTECTED 22250) (T_PUBLIC 22251) (T_READONLY 22252) (T_REQUIRE 22253) (T_REQUIRE_ONCE 22254) (T_RETURN 22255) (T_STATIC 22256) (T_STRING 22257) (T_SWITCH 22258) (T_THROW 22259) (T_TRAIT 22260) (T_TRAIT_C 22261) (T_TRY 22262) (T_UNSET 22263) (T_USE 22264) (T_VAR 22265) (T_VARIABLE 22059) (T_WHILE 22266) (T_YIELD 22267) (identifier 22268) (member_name 22269) (reserved_non_modifiers 22270) (semi_reserved 22271) (simple_variable 22272) ("{" 914)) 10421 (("(" 22273) (argument_list 22274)) 10422 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 22275) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 22276) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 22277) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10423 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 22280) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10424 (("," 1449) (possible_comma 22278)) 10425 ((")" 22279)) 10426 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (ampersand 22310) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22311) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10427 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22309) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10428 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22308) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10429 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22307) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10430 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22306) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10431 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22305) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10432 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22304) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10433 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22303) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10434 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22302) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10435 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22301) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10436 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22300) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10437 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22299) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10438 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22298) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10439 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22297) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10440 (("\"" 1489) ("$" 22312) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 22313) ("[" 1508) (array_object_dereferenceable 22314) (callable_expr 22315) (callable_variable 22316) (class_constant 1512) (class_name 22317) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 22318) (name 22319) (simple_variable 22320) (static_member 22321) (variable 22322) (variable_class_name 22323)) 10441 (("$" 22312) (T_VARIABLE 22313) (simple_variable 22358) ("{" 22359)) 10442 ((T_NULLSAFE_OBJECT_OPERATOR 22341) (T_OBJECT_OPERATOR 22342) ("[" 22343) ("{" 22344)) 10443 (("(" 22327) (argument_list 22340)) 10444 ((T_PAAMAYIM_NEKUDOTAYIM 22336)) 10445 (("(" 22327) (argument_list 22335)) 10446 ((T_PAAMAYIM_NEKUDOTAYIM 22324)) 10447 (("$" 22312) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 22313) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 22325) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 22326) ("{" 914)) 10448 (("(" 22327) (argument_list 22328)) 10449 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 22329) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 22330) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 22331) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10450 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 22334) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10451 (("," 1449) (possible_comma 22332)) 10452 ((")" 22333)) 10453 (("$" 22312) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 22313) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 22337) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 22338) ("{" 914)) 10454 (("(" 22327) (argument_list 22339)) 10455 (("$" 22312) (T_STRING 22349) (T_VARIABLE 22313) (property_name 22356) (simple_variable 22351) ("{" 22352)) 10456 (("$" 22312) (T_STRING 22349) (T_VARIABLE 22313) (property_name 22350) (simple_variable 22351) ("{" 22352)) 10457 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 22347) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10458 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22345) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10459 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22346)) 10460 (("]" 22348)) 10461 (("(" 22327) (argument_list 22355)) 10462 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22353) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10463 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22354)) 10464 (("(" 22327) (argument_list 22357)) 10465 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22360) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10466 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22361)) 10467 ((backup_doc_comment 22364)) 10468 (("(" 22365)) 10469 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 22366)) 10470 ((")" 22367)) 10471 ((T_USE 148) (lexical_vars 22368)) 10472 ((":" 150) (return_type 22369)) 10473 ((backup_fn_flags 22370)) 10474 (("{" 22371)) 10475 ((inner_statement_list 22372)) 10476 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 22373) ("~" 124)) 10477 ((backup_fn_flags 22374)) 10478 ((backup_doc_comment 22376)) 10479 (("(" 22377)) 10480 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 22378)) 10481 ((")" 22379)) 10482 ((":" 2207) (return_type 22380)) 10483 ((T_DOUBLE_ARROW 22381)) 10484 ((backup_fn_flags 22382)) 10485 ((backup_lex_pos 22383)) 10486 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22384) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10487 (("%" 22088) ("*" 22089) ("+" 22090) ("-" 22091) ("." 22092) ("/" 22093) ("<" 22094) (">" 22095) ("?" 22096) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22097) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22098) (T_BOOLEAN_AND 22099) (T_BOOLEAN_OR 22100) (T_COALESCE 22101) (T_INSTANCEOF 22102) (T_IS_EQUAL 22103) (T_IS_GREATER_OR_EQUAL 22104) (T_IS_IDENTICAL 22105) (T_IS_NOT_EQUAL 22106) (T_IS_NOT_IDENTICAL 22107) (T_IS_SMALLER_OR_EQUAL 22108) (T_LOGICAL_AND 22109) (T_LOGICAL_OR 22110) (T_LOGICAL_XOR 22111) (T_POW 22112) (T_SL 22113) (T_SPACESHIP 22114) (T_SR 22115) ("^" 22116) (backup_fn_flags 22385) ("|" 22117)) 10488 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22387) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10489 (("$" 22013) (T_ABSTRACT 22190) (T_ARRAY 22191) (T_AS 22192) (T_BREAK 22193) (T_CALLABLE 22194) (T_CASE 22195) (T_CATCH 22196) (T_CLASS 22197) (T_CLASS_C 22198) (T_CLONE 22199) (T_CONST 22200) (T_CONTINUE 22201) (T_DECLARE 22202) (T_DEFAULT 22203) (T_DIR 22204) (T_DO 22205) (T_ECHO 22206) (T_ELSE 22207) (T_ELSEIF 22208) (T_EMPTY 22209) (T_ENDDECLARE 22210) (T_ENDFOR 22211) (T_ENDFOREACH 22212) (T_ENDIF 22213) (T_ENDSWITCH 22214) (T_ENDWHILE 22215) (T_ENUM 22216) (T_EVAL 22217) (T_EXIT 22218) (T_EXTENDS 22219) (T_FILE 22220) (T_FINAL 22221) (T_FINALLY 22222) (T_FN 22223) (T_FOR 22224) (T_FOREACH 22225) (T_FUNCTION 22226) (T_FUNC_C 22227) (T_GLOBAL 22228) (T_GOTO 22229) (T_IF 22230) (T_IMPLEMENTS 22231) (T_INCLUDE 22232) (T_INCLUDE_ONCE 22233) (T_INSTANCEOF 22234) (T_INSTEADOF 22235) (T_INTERFACE 22236) (T_ISSET 22237) (T_LINE 22238) (T_LIST 22239) (T_LOGICAL_AND 22240) (T_LOGICAL_OR 22241) (T_LOGICAL_XOR 22242) (T_MATCH 22243) (T_METHOD_C 22244) (T_NAMESPACE 22245) (T_NEW 22246) (T_NS_C 22247) (T_PRINT 22248) (T_PRIVATE 22249) (T_PROTECTED 22250) (T_PUBLIC 22251) (T_READONLY 22252) (T_REQUIRE 22253) (T_REQUIRE_ONCE 22254) (T_RETURN 22255) (T_STATIC 22256) (T_STRING 22257) (T_SWITCH 22258) (T_THROW 22259) (T_TRAIT 22260) (T_TRAIT_C 22261) (T_TRY 22262) (T_UNSET 22263) (T_USE 22264) (T_VAR 22265) (T_VARIABLE 22059) (T_WHILE 22266) (T_YIELD 22267) (identifier 22389) (member_name 22390) (reserved_non_modifiers 22270) (semi_reserved 22271) (simple_variable 22391) ("{" 914)) 10490 (("(" 22273) (argument_list 22392)) 10491 ((T_FN 35) (T_FUNCTION 126) (fn 22073) #'22074 (inline_function 22396)) 10492 (("$" 22013) (T_STRING 22405) (T_VARIABLE 22059) (property_name 22412) (simple_variable 22407) ("{" 22408)) 10493 (("$" 22013) (T_STRING 22405) (T_VARIABLE 22059) (property_name 22406) (simple_variable 22407) ("{" 22408)) 10494 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 22403) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10495 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22401) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10496 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22402)) 10497 (("]" 22404)) 10498 (("(" 22273) (argument_list 22411)) 10499 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22409) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10500 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22410)) 10501 (("(" 22273) (argument_list 22413)) 10502 (("`" 22415)) 10503 (("]" 22417)) 10504 (("=" 22418)) 10505 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22419) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10506 (("%" 22088) ("*" 22089) ("+" 22090) ("-" 22091) ("." 22092) ("/" 22093) ("<" 22094) (">" 22095) ("?" 22096) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22097) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22098) (T_BOOLEAN_AND 22099) (T_BOOLEAN_OR 22100) (T_COALESCE 22101) (T_DOUBLE_ARROW 22422) (T_INSTANCEOF 22102) (T_IS_EQUAL 22103) (T_IS_GREATER_OR_EQUAL 22104) (T_IS_IDENTICAL 22105) (T_IS_NOT_EQUAL 22106) (T_IS_NOT_IDENTICAL 22107) (T_IS_SMALLER_OR_EQUAL 22108) (T_LOGICAL_AND 22109) (T_LOGICAL_OR 22110) (T_LOGICAL_XOR 22111) (T_POW 22112) (T_SL 22113) (T_SPACESHIP 22114) (T_SR 22115) ("^" 22116) ("|" 22117)) 10507 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22423) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10508 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 22432) (T_VARIABLE 2822) (encaps_var 2836)) 10509 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 22431) (T_VARIABLE 2822) (encaps_var 2827)) 10510 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 22482) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10511 (("(" 2899) (argument_list 2900) (ctor_arguments 22475)) 10512 ((T_ATTRIBUTE 2853) (T_CLASS 22438) (anonymous_class 22474) (attribute 2898)) 10513 ((T_PAAMAYIM_NEKUDOTAYIM 22472)) 10514 (("(" 22463) (argument_list 22464) (ctor_arguments 22465)) 10515 ((T_NULLSAFE_OBJECT_OPERATOR 22451) (T_OBJECT_OPERATOR 22452) (T_PAAMAYIM_NEKUDOTAYIM 22453) ("[" 22454) ("{" 22455)) 10516 (("$" 22312) (T_STRING 22349) (T_VARIABLE 22313) (property_name 22462) (simple_variable 22351) ("{" 22352)) 10517 (("$" 22312) (T_STRING 22349) (T_VARIABLE 22313) (property_name 22461) (simple_variable 22351) ("{" 22352)) 10518 (("$" 22312) (T_VARIABLE 22313) (simple_variable 22460)) 10519 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 22458) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10520 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22456) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10521 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22457)) 10522 (("]" 22459)) 10523 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 22466) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 22467) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 22468) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10524 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 22471) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10525 (("," 1449) (possible_comma 22469)) 10526 ((")" 22470)) 10527 (("$" 22312) (T_VARIABLE 22313) (simple_variable 22473)) 10528 ((T_EXTENDS 2902) (extends_from 22476)) 10529 ((T_IMPLEMENTS 2904) (implements_list 22477)) 10530 ((backup_doc_comment 22478)) 10531 (("{" 22479)) 10532 ((class_statement_list 22480)) 10533 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 22481)) 10534 (("%" 280) (")" 22483) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10535 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 22485) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10536 (("%" 280) (")" 22486) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10537 (("{" 22487)) 10538 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 22488) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 10539 (("}" 22489)) 10540 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 22491) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 10541 ((")" 22492)) 10542 (("=" 22493)) 10543 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 22494) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10544 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 22496) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10545 (("," 5117) (possible_comma 22497)) 10546 ((")" 22498)) 10547 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 22505) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10548 ((")" 22506)) 10549 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 22508) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10550 (("%" 280) (")" 22509) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10551 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 22511) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10552 (("%" 280) (")" 22512) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10553 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 22519) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 10554 ((")" 22520)) 10555 (("%" 280) (")" 22525) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10556 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22528) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10557 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22529)) 10558 (("\"" 22531) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 10559 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 22541) (T_VARIABLE 2822) (encaps_var 2836)) 10560 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 22540) (T_VARIABLE 2822) (encaps_var 2827)) 10561 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 22591) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10562 (("(" 2899) (argument_list 2900) (ctor_arguments 22584)) 10563 ((T_ATTRIBUTE 2853) (T_CLASS 22547) (anonymous_class 22583) (attribute 2898)) 10564 ((T_PAAMAYIM_NEKUDOTAYIM 22581)) 10565 (("(" 22572) (argument_list 22573) (ctor_arguments 22574)) 10566 ((T_NULLSAFE_OBJECT_OPERATOR 22560) (T_OBJECT_OPERATOR 22561) (T_PAAMAYIM_NEKUDOTAYIM 22562) ("[" 22563) ("{" 22564)) 10567 (("$" 21903) (T_STRING 21940) (T_VARIABLE 21904) (property_name 22571) (simple_variable 21942) ("{" 21943)) 10568 (("$" 21903) (T_STRING 21940) (T_VARIABLE 21904) (property_name 22570) (simple_variable 21942) ("{" 21943)) 10569 (("$" 21903) (T_VARIABLE 21904) (simple_variable 22569)) 10570 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 22567) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10571 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22565) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10572 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22566)) 10573 (("]" 22568)) 10574 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 22575) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 22576) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 22577) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10575 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 22580) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10576 (("," 1449) (possible_comma 22578)) 10577 ((")" 22579)) 10578 (("$" 21903) (T_VARIABLE 21904) (simple_variable 22582)) 10579 ((T_EXTENDS 2902) (extends_from 22585)) 10580 ((T_IMPLEMENTS 2904) (implements_list 22586)) 10581 ((backup_doc_comment 22587)) 10582 (("{" 22588)) 10583 ((class_statement_list 22589)) 10584 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 22590)) 10585 (("%" 280) (")" 22592) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10586 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 22594) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10587 (("%" 280) (")" 22595) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10588 (("{" 22596)) 10589 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 22597) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 10590 (("}" 22598)) 10591 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 22600) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 10592 ((")" 22601)) 10593 (("=" 22602)) 10594 (("!" 21602) ("\"" 21603) ("$" 21604) ("(" 21605) ("+" 21606) ("-" 21607) ("@" 21608) (T_ARRAY 21609) (T_ARRAY_CAST 21610) (T_ATTRIBUTE 125) (T_BOOL_CAST 21611) (T_CLASS_C 21612) (T_CLONE 21613) (T_CONSTANT_ENCAPSED_STRING 21614) (T_DEC 21615) (T_DIR 21616) (T_DNUMBER 21617) (T_DOUBLE_CAST 21618) (T_EMPTY 21619) (T_EVAL 21620) (T_EXIT 21621) (T_FILE 21622) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21623) (T_INC 21624) (T_INCLUDE 21625) (T_INCLUDE_ONCE 21626) (T_INT_CAST 21627) (T_ISSET 21628) (T_LINE 21629) (T_LIST 21630) (T_LNUMBER 21631) (T_MATCH 21632) (T_METHOD_C 21633) (T_NAME_FULLY_QUALIFIED 21634) (T_NAME_QUALIFIED 21635) (T_NAME_RELATIVE 21636) (T_NEW 21637) (T_NS_C 21638) (T_OBJECT_CAST 21639) (T_PRINT 21640) (T_REQUIRE 21641) (T_REQUIRE_ONCE 21642) (T_START_HEREDOC 21643) (T_STATIC 21644) (T_STRING 21645) (T_STRING_CAST 21646) (T_THROW 21647) (T_TRAIT_C 21648) (T_UNSET_CAST 21649) (T_VARIABLE 21650) (T_YIELD 21651) (T_YIELD_FROM 21652) ("[" 21653) ("`" 21654) (array_object_dereferenceable 21655) (attribute 129) (attributes 21656) (callable_expr 21657) (callable_variable 21658) (class_constant 21659) (class_name 21660) (constant 21661) (dereferenceable_scalar 21662) (expr 22603) (fn 21664) (fully_dereferenceable 103) #'21665 (function_call 21666) (inline_function 21667) (internal_functions_in_yacc 21668) (match 21669) (name 21670) (new_expr 21671) (scalar 21672) (simple_variable 21673) (static_member 21674) (variable 21675) (variable_class_name 21676) ("~" 21677)) 10595 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 22605) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10596 (("," 5117) (possible_comma 22606)) 10597 ((")" 22607)) 10598 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 22614) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10599 ((")" 22615)) 10600 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 22617) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10601 (("%" 280) (")" 22618) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10602 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 22620) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10603 (("%" 280) (")" 22621) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10604 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 22628) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 10605 ((")" 22629)) 10606 (("%" 280) (")" 22634) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10607 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22637) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10608 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22638)) 10609 (("\"" 22640) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 10610 ((inner_statement_list 22643)) 10611 (("$" 21231) (T_ABSTRACT 21408) (T_ARRAY 21409) (T_AS 21410) (T_BREAK 21411) (T_CALLABLE 21412) (T_CASE 21413) (T_CATCH 21414) (T_CLASS 21415) (T_CLASS_C 21416) (T_CLONE 21417) (T_CONST 21418) (T_CONTINUE 21419) (T_DECLARE 21420) (T_DEFAULT 21421) (T_DIR 21422) (T_DO 21423) (T_ECHO 21424) (T_ELSE 21425) (T_ELSEIF 21426) (T_EMPTY 21427) (T_ENDDECLARE 21428) (T_ENDFOR 21429) (T_ENDFOREACH 21430) (T_ENDIF 21431) (T_ENDSWITCH 21432) (T_ENDWHILE 21433) (T_ENUM 21434) (T_EVAL 21435) (T_EXIT 21436) (T_EXTENDS 21437) (T_FILE 21438) (T_FINAL 21439) (T_FINALLY 21440) (T_FN 21441) (T_FOR 21442) (T_FOREACH 21443) (T_FUNCTION 21444) (T_FUNC_C 21445) (T_GLOBAL 21446) (T_GOTO 21447) (T_IF 21448) (T_IMPLEMENTS 21449) (T_INCLUDE 21450) (T_INCLUDE_ONCE 21451) (T_INSTANCEOF 21452) (T_INSTEADOF 21453) (T_INTERFACE 21454) (T_ISSET 21455) (T_LINE 21456) (T_LIST 21457) (T_LOGICAL_AND 21458) (T_LOGICAL_OR 21459) (T_LOGICAL_XOR 21460) (T_MATCH 21461) (T_METHOD_C 21462) (T_NAMESPACE 21463) (T_NEW 21464) (T_NS_C 21465) (T_PRINT 21466) (T_PRIVATE 21467) (T_PROTECTED 21468) (T_PUBLIC 21469) (T_READONLY 21470) (T_REQUIRE 21471) (T_REQUIRE_ONCE 21472) (T_RETURN 21473) (T_STATIC 21474) (T_STRING 21475) (T_SWITCH 21476) (T_THROW 21477) (T_TRAIT 21478) (T_TRAIT_C 21479) (T_TRY 21480) (T_UNSET 21481) (T_USE 21482) (T_VAR 21483) (T_VARIABLE 21277) (T_WHILE 21484) (T_YIELD 21485) (identifier 22645) (member_name 22646) (reserved_non_modifiers 21488) (semi_reserved 21489) (simple_variable 22647) ("{" 914)) 10612 (("(" 21491) (argument_list 22648)) 10613 ((T_FN 35) (T_FUNCTION 126) (fn 21291) #'21292 (inline_function 22652)) 10614 (("$" 21231) (T_STRING 22661) (T_VARIABLE 21277) (property_name 22668) (simple_variable 22663) ("{" 22664)) 10615 (("$" 21231) (T_STRING 22661) (T_VARIABLE 21277) (property_name 22662) (simple_variable 22663) ("{" 22664)) 10616 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 22659) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10617 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22657) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10618 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22658)) 10619 (("]" 22660)) 10620 (("(" 21491) (argument_list 22667)) 10621 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22665) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10622 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22666)) 10623 (("(" 21491) (argument_list 22669)) 10624 (("`" 22671)) 10625 (("]" 22673)) 10626 (("=" 22674)) 10627 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 22675) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10628 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23172) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10629 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 23170) (encaps_var 5158)) 10630 (("$" 22679) (T_VARIABLE 22725) (simple_variable 23166) ("{" 23167)) 10631 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23164) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10632 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23163) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10633 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23162) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10634 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23161) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10635 (("(" 23158)) 10636 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23157) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10637 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23156) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10638 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23155) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10639 (("\"" 1489) ("$" 22978) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 22979) ("[" 1508) (array_object_dereferenceable 22980) (callable_expr 22981) (callable_variable 22982) (class_constant 1512) (class_name 22983) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 22984) (name 22985) (simple_variable 22986) (static_member 22987) (variable 23154) (variable_class_name 22989)) 10640 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23153) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10641 (("(" 23150)) 10642 (("(" 23147)) 10643 (("(" 23143) (exit_expr 23144)) 10644 (("\"" 1489) ("$" 22978) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 22979) ("[" 1508) (array_object_dereferenceable 22980) (callable_expr 22981) (callable_variable 22982) (class_constant 1512) (class_name 22983) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 22984) (name 22985) (simple_variable 22986) (static_member 22987) (variable 23142) (variable_class_name 22989)) 10645 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23141) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10646 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23140) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10647 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23139) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10648 (("(" 23135)) 10649 (("(" 23130)) 10650 (("(" 23124)) 10651 (("$" 22312) ("(" 22437) (T_ATTRIBUTE 2853) (T_CLASS 23103) (T_NAME_FULLY_QUALIFIED 22439) (T_NAME_QUALIFIED 22440) (T_NAME_RELATIVE 22441) (T_STATIC 22442) (T_STRING 22443) (T_VARIABLE 22313) (anonymous_class 23104) (attribute 2862) (attributes 23105) (class_name 22446) (class_name_reference 23106) (name 22448) (new_variable 22449) (simple_variable 22450)) 10652 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23102) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10653 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23101) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10654 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23100) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10655 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23099) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10656 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 23094) (T_END_HEREDOC 23095) (T_VARIABLE 2822) (encaps_list 23096) (encaps_var 2824)) 10657 ((T_FN 35) (T_FUNCTION 126) (fn 22739) #'22740 (inline_function 23093)) 10658 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23092) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10659 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23091) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10660 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23090) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10661 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23087) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10662 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23086) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10663 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 23082) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 10664 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 23080) (encaps_list 2729) (encaps_var 2730)) 10665 ((T_NULLSAFE_OBJECT_OPERATOR 23063) (T_OBJECT_OPERATOR 23064) ("[" 23065) ("{" 23066)) 10666 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 23060) (attribute 2654) (fn 22739) #'22740 (inline_function 23061)) 10667 (("(" 22939) (argument_list 23059)) 10668 ((T_PAAMAYIM_NEKUDOTAYIM 23054)) 10669 (("%" 22754) ("*" 22755) ("+" 22756) ("-" 22757) ("." 22758) ("/" 22759) ("<" 22760) (">" 22761) ("?" 22762) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22763) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22764) (T_BOOLEAN_AND 22765) (T_BOOLEAN_OR 22766) (T_COALESCE 22767) (T_DOUBLE_ARROW 23052) (T_INSTANCEOF 22768) (T_IS_EQUAL 22769) (T_IS_GREATER_OR_EQUAL 22770) (T_IS_IDENTICAL 22771) (T_IS_NOT_EQUAL 22772) (T_IS_NOT_IDENTICAL 22773) (T_IS_SMALLER_OR_EQUAL 22774) (T_LOGICAL_AND 22775) (T_LOGICAL_OR 22776) (T_LOGICAL_XOR 22777) (T_POW 22778) (T_SL 22779) (T_SPACESHIP 22780) (T_SR 22781) ("^" 22782) ("|" 22783)) 10670 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 23041)) 10671 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 23029)) 10672 (("(" 22939) (argument_list 23028)) 10673 (("=" 22947) (T_AND_EQUAL 22948) (T_COALESCE_EQUAL 22949) (T_CONCAT_EQUAL 22950) (T_DEC 22951) (T_DIV_EQUAL 22952) (T_INC 22953) (T_MINUS_EQUAL 22954) (T_MOD_EQUAL 22955) (T_MUL_EQUAL 22956) (T_OR_EQUAL 22957) (T_PLUS_EQUAL 22958) (T_POW_EQUAL 22959) (T_SL_EQUAL 22960) (T_SR_EQUAL 22961) (T_XOR_EQUAL 22962)) 10674 ((T_PAAMAYIM_NEKUDOTAYIM 22855)) 10675 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22753) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10676 (("%" 22754) ("*" 22755) ("+" 22756) ("-" 22757) ("." 22758) ("/" 22759) ("<" 22760) (">" 22761) ("?" 22762) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22763) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22764) (T_BOOLEAN_AND 22765) (T_BOOLEAN_OR 22766) (T_COALESCE 22767) (T_INSTANCEOF 22768) (T_IS_EQUAL 22769) (T_IS_GREATER_OR_EQUAL 22770) (T_IS_IDENTICAL 22771) (T_IS_NOT_EQUAL 22772) (T_IS_NOT_IDENTICAL 22773) (T_IS_SMALLER_OR_EQUAL 22774) (T_LOGICAL_AND 22775) (T_LOGICAL_OR 22776) (T_LOGICAL_XOR 22777) (T_POW 22778) (T_SL 22779) (T_SPACESHIP 22780) (T_SR 22781) ("^" 22782) ("|" 22783)) 10677 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22854) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10678 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22853) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10679 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22852) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10680 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22851) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10681 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22850) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10682 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22849) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10683 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22848) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10684 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22847) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10685 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 22842) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 22843) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 10686 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22841) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10687 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22840) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10688 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22839) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10689 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22838) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10690 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22837) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10691 (("$" 22799) ("(" 22800) (T_NAME_FULLY_QUALIFIED 22801) (T_NAME_QUALIFIED 22802) (T_NAME_RELATIVE 22803) (T_STATIC 22804) (T_STRING 22805) (T_VARIABLE 22806) (class_name 22807) (class_name_reference 22808) (name 22809) (new_variable 22810) (simple_variable 22811)) 10692 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22798) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10693 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22797) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10694 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22796) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10695 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22795) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10696 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22794) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10697 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22793) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10698 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22792) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10699 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22791) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10700 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22790) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10701 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22789) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10702 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22788) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10703 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22787) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10704 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22786) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10705 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22785) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10706 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22784) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10707 (("$" 22799) (T_VARIABLE 22806) (simple_variable 22833) ("{" 22834)) 10708 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 22831) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10709 ((T_PAAMAYIM_NEKUDOTAYIM 22829)) 10710 ((T_NULLSAFE_OBJECT_OPERATOR 22812) (T_OBJECT_OPERATOR 22813) (T_PAAMAYIM_NEKUDOTAYIM 22814) ("[" 22815) ("{" 22816)) 10711 (("$" 22799) (T_STRING 22822) (T_VARIABLE 22806) (property_name 22828) (simple_variable 22824) ("{" 22825)) 10712 (("$" 22799) (T_STRING 22822) (T_VARIABLE 22806) (property_name 22823) (simple_variable 22824) ("{" 22825)) 10713 (("$" 22799) (T_VARIABLE 22806) (simple_variable 22821)) 10714 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 22819) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10715 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22817) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10716 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22818)) 10717 (("]" 22820)) 10718 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22826) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10719 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22827)) 10720 (("$" 22799) (T_VARIABLE 22806) (simple_variable 22830)) 10721 (("%" 280) (")" 22832) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10722 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 22835) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10723 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 22836)) 10724 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22846) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10725 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 22844) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 10726 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22845) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10727 (("$" 22679) (T_ABSTRACT 22856) (T_ARRAY 22857) (T_AS 22858) (T_BREAK 22859) (T_CALLABLE 22860) (T_CASE 22861) (T_CATCH 22862) (T_CLASS 22863) (T_CLASS_C 22864) (T_CLONE 22865) (T_CONST 22866) (T_CONTINUE 22867) (T_DECLARE 22868) (T_DEFAULT 22869) (T_DIR 22870) (T_DO 22871) (T_ECHO 22872) (T_ELSE 22873) (T_ELSEIF 22874) (T_EMPTY 22875) (T_ENDDECLARE 22876) (T_ENDFOR 22877) (T_ENDFOREACH 22878) (T_ENDIF 22879) (T_ENDSWITCH 22880) (T_ENDWHILE 22881) (T_ENUM 22882) (T_EVAL 22883) (T_EXIT 22884) (T_EXTENDS 22885) (T_FILE 22886) (T_FINAL 22887) (T_FINALLY 22888) (T_FN 22889) (T_FOR 22890) (T_FOREACH 22891) (T_FUNCTION 22892) (T_FUNC_C 22893) (T_GLOBAL 22894) (T_GOTO 22895) (T_IF 22896) (T_IMPLEMENTS 22897) (T_INCLUDE 22898) (T_INCLUDE_ONCE 22899) (T_INSTANCEOF 22900) (T_INSTEADOF 22901) (T_INTERFACE 22902) (T_ISSET 22903) (T_LINE 22904) (T_LIST 22905) (T_LOGICAL_AND 22906) (T_LOGICAL_OR 22907) (T_LOGICAL_XOR 22908) (T_MATCH 22909) (T_METHOD_C 22910) (T_NAMESPACE 22911) (T_NEW 22912) (T_NS_C 22913) (T_PRINT 22914) (T_PRIVATE 22915) (T_PROTECTED 22916) (T_PUBLIC 22917) (T_READONLY 22918) (T_REQUIRE 22919) (T_REQUIRE_ONCE 22920) (T_RETURN 22921) (T_STATIC 22922) (T_STRING 22923) (T_SWITCH 22924) (T_THROW 22925) (T_TRAIT 22926) (T_TRAIT_C 22927) (T_TRY 22928) (T_UNSET 22929) (T_USE 22930) (T_VAR 22931) (T_VARIABLE 22725) (T_WHILE 22932) (T_YIELD 22933) (identifier 22934) (member_name 22935) (reserved_non_modifiers 22936) (semi_reserved 22937) (simple_variable 22938) ("{" 914)) 10728 (("(" 22939) (argument_list 22940)) 10729 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 22941) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 22942) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 22943) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10730 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 22946) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10731 (("," 1449) (possible_comma 22944)) 10732 ((")" 22945)) 10733 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (ampersand 22976) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22977) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10734 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22975) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10735 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22974) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10736 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22973) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10737 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22972) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10738 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22971) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10739 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22970) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10740 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22969) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10741 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22968) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10742 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22967) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10743 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22966) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10744 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22965) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10745 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22964) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10746 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 22963) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10747 (("\"" 1489) ("$" 22978) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 22979) ("[" 1508) (array_object_dereferenceable 22980) (callable_expr 22981) (callable_variable 22982) (class_constant 1512) (class_name 22983) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 22984) (name 22985) (simple_variable 22986) (static_member 22987) (variable 22988) (variable_class_name 22989)) 10748 (("$" 22978) (T_VARIABLE 22979) (simple_variable 23024) ("{" 23025)) 10749 ((T_NULLSAFE_OBJECT_OPERATOR 23007) (T_OBJECT_OPERATOR 23008) ("[" 23009) ("{" 23010)) 10750 (("(" 22993) (argument_list 23006)) 10751 ((T_PAAMAYIM_NEKUDOTAYIM 23002)) 10752 (("(" 22993) (argument_list 23001)) 10753 ((T_PAAMAYIM_NEKUDOTAYIM 22990)) 10754 (("$" 22978) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 22979) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 22991) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 22992) ("{" 914)) 10755 (("(" 22993) (argument_list 22994)) 10756 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 22995) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 22996) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 22997) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10757 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 23000) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10758 (("," 1449) (possible_comma 22998)) 10759 ((")" 22999)) 10760 (("$" 22978) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 22979) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 23003) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 23004) ("{" 914)) 10761 (("(" 22993) (argument_list 23005)) 10762 (("$" 22978) (T_STRING 23015) (T_VARIABLE 22979) (property_name 23022) (simple_variable 23017) ("{" 23018)) 10763 (("$" 22978) (T_STRING 23015) (T_VARIABLE 22979) (property_name 23016) (simple_variable 23017) ("{" 23018)) 10764 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 23013) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10765 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 23011) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10766 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 23012)) 10767 (("]" 23014)) 10768 (("(" 22993) (argument_list 23021)) 10769 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 23019) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10770 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 23020)) 10771 (("(" 22993) (argument_list 23023)) 10772 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 23026) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10773 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 23027)) 10774 ((backup_doc_comment 23030)) 10775 (("(" 23031)) 10776 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 23032)) 10777 ((")" 23033)) 10778 ((T_USE 148) (lexical_vars 23034)) 10779 ((":" 150) (return_type 23035)) 10780 ((backup_fn_flags 23036)) 10781 (("{" 23037)) 10782 ((inner_statement_list 23038)) 10783 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23039) ("~" 124)) 10784 ((backup_fn_flags 23040)) 10785 ((backup_doc_comment 23042)) 10786 (("(" 23043)) 10787 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 23044)) 10788 ((")" 23045)) 10789 ((":" 2207) (return_type 23046)) 10790 ((T_DOUBLE_ARROW 23047)) 10791 ((backup_fn_flags 23048)) 10792 ((backup_lex_pos 23049)) 10793 (("!" 22011) ("\"" 22012) ("$" 22013) ("(" 22014) ("+" 22015) ("-" 22016) ("@" 22017) (T_ARRAY 22018) (T_ARRAY_CAST 22019) (T_ATTRIBUTE 125) (T_BOOL_CAST 22020) (T_CLASS_C 22021) (T_CLONE 22022) (T_CONSTANT_ENCAPSED_STRING 22023) (T_DEC 22024) (T_DIR 22025) (T_DNUMBER 22026) (T_DOUBLE_CAST 22027) (T_EMPTY 22028) (T_EVAL 22029) (T_EXIT 22030) (T_FILE 22031) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22032) (T_INC 22033) (T_INCLUDE 22034) (T_INCLUDE_ONCE 22035) (T_INT_CAST 22036) (T_ISSET 22037) (T_LINE 22038) (T_LIST 22039) (T_LNUMBER 22040) (T_MATCH 22041) (T_METHOD_C 22042) (T_NAME_FULLY_QUALIFIED 22043) (T_NAME_QUALIFIED 22044) (T_NAME_RELATIVE 22045) (T_NEW 22046) (T_NS_C 22047) (T_OBJECT_CAST 22048) (T_PRINT 22049) (T_REQUIRE 22050) (T_REQUIRE_ONCE 22051) (T_START_HEREDOC 22052) (T_STATIC 22053) (T_STRING 22054) (T_STRING_CAST 22055) (T_THROW 22056) (T_TRAIT_C 22057) (T_UNSET_CAST 22058) (T_VARIABLE 22059) (T_YIELD 22060) (T_YIELD_FROM 22061) ("[" 22062) ("`" 22063) (array_object_dereferenceable 22064) (attribute 129) (attributes 22065) (callable_expr 22066) (callable_variable 22067) (class_constant 22068) (class_name 22069) (constant 22070) (dereferenceable_scalar 22071) (expr 23050) (fn 22073) (fully_dereferenceable 103) #'22074 (function_call 22075) (inline_function 22076) (internal_functions_in_yacc 22077) (match 22078) (name 22079) (new_expr 22080) (scalar 22081) (simple_variable 22082) (static_member 22083) (variable 22084) (variable_class_name 22085) ("~" 22086)) 10794 (("%" 22088) ("*" 22089) ("+" 22090) ("-" 22091) ("." 22092) ("/" 22093) ("<" 22094) (">" 22095) ("?" 22096) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22097) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22098) (T_BOOLEAN_AND 22099) (T_BOOLEAN_OR 22100) (T_COALESCE 22101) (T_INSTANCEOF 22102) (T_IS_EQUAL 22103) (T_IS_GREATER_OR_EQUAL 22104) (T_IS_IDENTICAL 22105) (T_IS_NOT_EQUAL 22106) (T_IS_NOT_IDENTICAL 22107) (T_IS_SMALLER_OR_EQUAL 22108) (T_LOGICAL_AND 22109) (T_LOGICAL_OR 22110) (T_LOGICAL_XOR 22111) (T_POW 22112) (T_SL 22113) (T_SPACESHIP 22114) (T_SR 22115) ("^" 22116) (backup_fn_flags 23051) ("|" 22117)) 10795 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23053) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10796 (("$" 22679) (T_ABSTRACT 22856) (T_ARRAY 22857) (T_AS 22858) (T_BREAK 22859) (T_CALLABLE 22860) (T_CASE 22861) (T_CATCH 22862) (T_CLASS 22863) (T_CLASS_C 22864) (T_CLONE 22865) (T_CONST 22866) (T_CONTINUE 22867) (T_DECLARE 22868) (T_DEFAULT 22869) (T_DIR 22870) (T_DO 22871) (T_ECHO 22872) (T_ELSE 22873) (T_ELSEIF 22874) (T_EMPTY 22875) (T_ENDDECLARE 22876) (T_ENDFOR 22877) (T_ENDFOREACH 22878) (T_ENDIF 22879) (T_ENDSWITCH 22880) (T_ENDWHILE 22881) (T_ENUM 22882) (T_EVAL 22883) (T_EXIT 22884) (T_EXTENDS 22885) (T_FILE 22886) (T_FINAL 22887) (T_FINALLY 22888) (T_FN 22889) (T_FOR 22890) (T_FOREACH 22891) (T_FUNCTION 22892) (T_FUNC_C 22893) (T_GLOBAL 22894) (T_GOTO 22895) (T_IF 22896) (T_IMPLEMENTS 22897) (T_INCLUDE 22898) (T_INCLUDE_ONCE 22899) (T_INSTANCEOF 22900) (T_INSTEADOF 22901) (T_INTERFACE 22902) (T_ISSET 22903) (T_LINE 22904) (T_LIST 22905) (T_LOGICAL_AND 22906) (T_LOGICAL_OR 22907) (T_LOGICAL_XOR 22908) (T_MATCH 22909) (T_METHOD_C 22910) (T_NAMESPACE 22911) (T_NEW 22912) (T_NS_C 22913) (T_PRINT 22914) (T_PRIVATE 22915) (T_PROTECTED 22916) (T_PUBLIC 22917) (T_READONLY 22918) (T_REQUIRE 22919) (T_REQUIRE_ONCE 22920) (T_RETURN 22921) (T_STATIC 22922) (T_STRING 22923) (T_SWITCH 22924) (T_THROW 22925) (T_TRAIT 22926) (T_TRAIT_C 22927) (T_TRY 22928) (T_UNSET 22929) (T_USE 22930) (T_VAR 22931) (T_VARIABLE 22725) (T_WHILE 22932) (T_YIELD 22933) (identifier 23055) (member_name 23056) (reserved_non_modifiers 22936) (semi_reserved 22937) (simple_variable 23057) ("{" 914)) 10797 (("(" 22939) (argument_list 23058)) 10798 ((T_FN 35) (T_FUNCTION 126) (fn 22739) #'22740 (inline_function 23062)) 10799 (("$" 22679) (T_STRING 23071) (T_VARIABLE 22725) (property_name 23078) (simple_variable 23073) ("{" 23074)) 10800 (("$" 22679) (T_STRING 23071) (T_VARIABLE 22725) (property_name 23072) (simple_variable 23073) ("{" 23074)) 10801 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 23069) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 10802 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 23067) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10803 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 23068)) 10804 (("]" 23070)) 10805 (("(" 22939) (argument_list 23077)) 10806 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 23075) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10807 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 23076)) 10808 (("(" 22939) (argument_list 23079)) 10809 (("`" 23081)) 10810 (("]" 23083)) 10811 (("=" 23084)) 10812 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23085) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10813 (("%" 22754) ("*" 22755) ("+" 22756) ("-" 22757) ("." 22758) ("/" 22759) ("<" 22760) (">" 22761) ("?" 22762) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22763) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22764) (T_BOOLEAN_AND 22765) (T_BOOLEAN_OR 22766) (T_COALESCE 22767) (T_DOUBLE_ARROW 23088) (T_INSTANCEOF 22768) (T_IS_EQUAL 22769) (T_IS_GREATER_OR_EQUAL 22770) (T_IS_IDENTICAL 22771) (T_IS_NOT_EQUAL 22772) (T_IS_NOT_IDENTICAL 22773) (T_IS_SMALLER_OR_EQUAL 22774) (T_LOGICAL_AND 22775) (T_LOGICAL_OR 22776) (T_LOGICAL_XOR 22777) (T_POW 22778) (T_SL 22779) (T_SPACESHIP 22780) (T_SR 22781) ("^" 22782) ("|" 22783)) 10814 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23089) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10815 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 23098) (T_VARIABLE 2822) (encaps_var 2836)) 10816 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 23097) (T_VARIABLE 2822) (encaps_var 2827)) 10817 (("(" 2899) (argument_list 2900) (ctor_arguments 23117)) 10818 ((T_ATTRIBUTE 2853) (T_CLASS 23103) (anonymous_class 23116) (attribute 2898)) 10819 (("(" 23107) (argument_list 23108) (ctor_arguments 23109)) 10820 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 23110) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 23111) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 23112) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10821 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 23115) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10822 (("," 1449) (possible_comma 23113)) 10823 ((")" 23114)) 10824 ((T_EXTENDS 2902) (extends_from 23118)) 10825 ((T_IMPLEMENTS 2904) (implements_list 23119)) 10826 ((backup_doc_comment 23120)) 10827 (("{" 23121)) 10828 ((class_statement_list 23122)) 10829 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23123)) 10830 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23125) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10831 (("%" 280) (")" 23126) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10832 (("{" 23127)) 10833 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 23128) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 10834 (("}" 23129)) 10835 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 23131) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 10836 ((")" 23132)) 10837 (("=" 23133)) 10838 (("!" 22677) ("\"" 22678) ("$" 22679) ("(" 22680) ("+" 22681) ("-" 22682) ("@" 22683) (T_ARRAY 22684) (T_ARRAY_CAST 22685) (T_ATTRIBUTE 125) (T_BOOL_CAST 22686) (T_CLASS_C 22687) (T_CLONE 22688) (T_CONSTANT_ENCAPSED_STRING 22689) (T_DEC 22690) (T_DIR 22691) (T_DNUMBER 22692) (T_DOUBLE_CAST 22693) (T_EMPTY 22694) (T_EVAL 22695) (T_EXIT 22696) (T_FILE 22697) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 22698) (T_INC 22699) (T_INCLUDE 22700) (T_INCLUDE_ONCE 22701) (T_INT_CAST 22702) (T_ISSET 22703) (T_LINE 22704) (T_LIST 22705) (T_LNUMBER 22706) (T_MATCH 22707) (T_METHOD_C 22708) (T_NAME_FULLY_QUALIFIED 22709) (T_NAME_QUALIFIED 22710) (T_NAME_RELATIVE 22711) (T_NEW 22712) (T_NS_C 22713) (T_OBJECT_CAST 22714) (T_PRINT 22715) (T_REQUIRE 22716) (T_REQUIRE_ONCE 22717) (T_START_HEREDOC 22718) (T_STATIC 22719) (T_STRING 22720) (T_STRING_CAST 22721) (T_THROW 22722) (T_TRAIT_C 22723) (T_UNSET_CAST 22724) (T_VARIABLE 22725) (T_YIELD 22726) (T_YIELD_FROM 22727) ("[" 22728) ("`" 22729) (array_object_dereferenceable 22730) (attribute 129) (attributes 22731) (callable_expr 22732) (callable_variable 22733) (class_constant 22734) (class_name 22735) (constant 22736) (dereferenceable_scalar 22737) (expr 23134) (fn 22739) (fully_dereferenceable 103) #'22740 (function_call 22741) (inline_function 22742) (internal_functions_in_yacc 22743) (match 22744) (name 22745) (new_expr 22746) (scalar 22747) (simple_variable 22748) (static_member 22749) (variable 22750) (variable_class_name 22751) ("~" 22752)) 10839 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 23136) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10840 (("," 5117) (possible_comma 23137)) 10841 ((")" 23138)) 10842 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 23145) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10843 ((")" 23146)) 10844 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23148) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10845 (("%" 280) (")" 23149) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10846 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23151) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10847 (("%" 280) (")" 23152) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10848 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 23159) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 10849 ((")" 23160)) 10850 (("%" 280) (")" 23165) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10851 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 23168) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10852 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 23169)) 10853 (("\"" 23171) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 10854 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 23181) (T_VARIABLE 2822) (encaps_var 2836)) 10855 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 23180) (T_VARIABLE 2822) (encaps_var 2827)) 10856 (("(" 2899) (argument_list 2900) (ctor_arguments 23200)) 10857 ((T_ATTRIBUTE 2853) (T_CLASS 23186) (anonymous_class 23199) (attribute 2898)) 10858 (("(" 23190) (argument_list 23191) (ctor_arguments 23192)) 10859 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 23193) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 23194) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 23195) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10860 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 23198) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10861 (("," 1449) (possible_comma 23196)) 10862 ((")" 23197)) 10863 ((T_EXTENDS 2902) (extends_from 23201)) 10864 ((T_IMPLEMENTS 2904) (implements_list 23202)) 10865 ((backup_doc_comment 23203)) 10866 (("{" 23204)) 10867 ((class_statement_list 23205)) 10868 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23206)) 10869 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23208) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10870 (("%" 280) (")" 23209) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10871 (("{" 23210)) 10872 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 23211) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 10873 (("}" 23212)) 10874 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 23214) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 10875 ((")" 23215)) 10876 (("=" 23216)) 10877 (("!" 21229) ("\"" 21230) ("$" 21231) ("(" 21232) ("+" 21233) ("-" 21234) ("@" 21235) (T_ARRAY 21236) (T_ARRAY_CAST 21237) (T_ATTRIBUTE 125) (T_BOOL_CAST 21238) (T_CLASS_C 21239) (T_CLONE 21240) (T_CONSTANT_ENCAPSED_STRING 21241) (T_DEC 21242) (T_DIR 21243) (T_DNUMBER 21244) (T_DOUBLE_CAST 21245) (T_EMPTY 21246) (T_EVAL 21247) (T_EXIT 21248) (T_FILE 21249) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 21250) (T_INC 21251) (T_INCLUDE 21252) (T_INCLUDE_ONCE 21253) (T_INT_CAST 21254) (T_ISSET 21255) (T_LINE 21256) (T_LIST 21257) (T_LNUMBER 21258) (T_MATCH 21259) (T_METHOD_C 21260) (T_NAME_FULLY_QUALIFIED 21261) (T_NAME_QUALIFIED 21262) (T_NAME_RELATIVE 21263) (T_NEW 21264) (T_NS_C 21265) (T_OBJECT_CAST 21266) (T_PRINT 21267) (T_REQUIRE 21268) (T_REQUIRE_ONCE 21269) (T_START_HEREDOC 21270) (T_STATIC 21271) (T_STRING 21272) (T_STRING_CAST 21273) (T_THROW 21274) (T_TRAIT_C 21275) (T_UNSET_CAST 21276) (T_VARIABLE 21277) (T_YIELD 21278) (T_YIELD_FROM 21279) ("[" 21280) ("`" 21281) (array_object_dereferenceable 21282) (attribute 129) (attributes 21283) (callable_expr 21284) (callable_variable 21285) (class_constant 21286) (class_name 21287) (constant 21288) (dereferenceable_scalar 21289) (expr 23217) (fn 21291) (fully_dereferenceable 103) #'21292 (function_call 21293) (inline_function 21294) (internal_functions_in_yacc 21295) (match 21296) (name 21297) (new_expr 21298) (scalar 21299) (simple_variable 21300) (static_member 21301) (variable 21302) (variable_class_name 21303) ("~" 21304)) 10878 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 23219) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 10879 (("," 5117) (possible_comma 23220)) 10880 ((")" 23221)) 10881 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 23228) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10882 ((")" 23229)) 10883 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23231) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10884 (("%" 280) (")" 23232) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10885 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23234) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10886 (("%" 280) (")" 23235) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10887 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 23242) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 10888 ((")" 23243)) 10889 (("%" 280) (")" 23248) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10890 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 23251) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 10891 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 23252)) 10892 (("\"" 23254) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 10893 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 23257)) 10894 ((";" 23258)) 10895 (("," 16412) (";" 23261)) 10896 ((";" 23263)) 10897 ((";" 23265)) 10898 (("," 17413) (";" 23267)) 10899 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 23269) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 10900 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 23270) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 10901 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 23271) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 10902 ((")" 23272) (T_DOUBLE_ARROW 23273)) 10903 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23276) (";" 16184) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16185) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16186) (T_DEC 22) (T_DECLARE 16187) (T_DIR 24) (T_DNUMBER 25) (T_DO 16188) (T_DOUBLE_CAST 27) (T_ECHO 16189) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16190) (T_FOREACH 16191) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16192) (T_GOTO 16193) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16194) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16195) (T_START_HEREDOC 67) (T_STATIC 16196) (T_STRING 16197) (T_STRING_CAST 70) (T_SWITCH 16198) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16199) (T_UNSET 16200) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16201) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16202) (alt_if_stmt_without_else 16203) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16204) (fn 102) (foreach_statement 23282) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16205) (if_stmt_without_else 16206) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23278) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16208) ("~" 124)) 10904 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 23274) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 10905 ((")" 23275)) 10906 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23276) (";" 16184) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16185) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16186) (T_DEC 22) (T_DECLARE 16187) (T_DIR 24) (T_DNUMBER 25) (T_DO 16188) (T_DOUBLE_CAST 27) (T_ECHO 16189) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16190) (T_FOREACH 16191) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16192) (T_GOTO 16193) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16194) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16195) (T_START_HEREDOC 67) (T_STATIC 16196) (T_STRING 16197) (T_STRING_CAST 70) (T_SWITCH 16198) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16199) (T_UNSET 16200) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16201) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16202) (alt_if_stmt_without_else 16203) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16204) (fn 102) (foreach_statement 23277) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16205) (if_stmt_without_else 16206) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23278) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16208) ("~" 124)) 10907 ((inner_statement_list 23279)) 10908 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 23280) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 10909 ((";" 23281)) 10910 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 23284) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 10911 ((";" 23285)) 10912 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 23286) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 10913 ((";" 23287)) 10914 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 23288) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 10915 ((")" 23289)) 10916 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23290) (";" 16184) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16185) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16186) (T_DEC 22) (T_DECLARE 16187) (T_DIR 24) (T_DNUMBER 25) (T_DO 16188) (T_DOUBLE_CAST 27) (T_ECHO 16189) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16190) (T_FOREACH 16191) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16192) (T_GOTO 16193) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16194) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16195) (T_START_HEREDOC 67) (T_STATIC 16196) (T_STRING 16197) (T_STRING_CAST 70) (T_SWITCH 16198) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16199) (T_UNSET 16200) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16201) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16202) (alt_if_stmt_without_else 16203) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16204) (fn 102) (for_statement 23291) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16205) (if_stmt_without_else 16206) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23292) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16208) ("~" 124)) 10917 ((inner_statement_list 23293)) 10918 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 23294) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 10919 ((";" 23295)) 10920 (("," 19251) (";" 23297)) 10921 ((T_WHILE 23299)) 10922 (("(" 23300)) 10923 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23301) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10924 (("%" 280) (")" 23302) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10925 ((";" 23303)) 10926 ((T_STRING 19424) (const_decl 19425) (const_list 23305)) 10927 ((")" 23306) ("," 19428)) 10928 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23307) (";" 16184) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16185) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16186) (T_DEC 22) (T_DECLARE 16187) (T_DIR 24) (T_DNUMBER 25) (T_DO 16188) (T_DOUBLE_CAST 27) (T_ECHO 16189) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16190) (T_FOREACH 16191) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16192) (T_GOTO 16193) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16194) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16195) (T_START_HEREDOC 67) (T_STATIC 16196) (T_STRING 16197) (T_STRING_CAST 70) (T_SWITCH 16198) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16199) (T_UNSET 16200) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16201) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16202) (alt_if_stmt_without_else 16203) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 23308) (dereferenceable_scalar 99) (expr 16204) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16205) (if_stmt_without_else 16206) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23309) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16208) ("~" 124)) 10929 ((inner_statement_list 23310)) 10930 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 23311) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 10931 ((";" 23312)) 10932 ((";" 23314)) 10933 ((";" 23316)) 10934 ((T_STRING 23319) (backup_doc_comment 137)) 10935 ((backup_doc_comment 23320)) 10936 (("(" 23321)) 10937 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 23322)) 10938 ((")" 23323)) 10939 ((":" 150) (return_type 23324)) 10940 ((backup_fn_flags 23325)) 10941 (("{" 23326)) 10942 ((inner_statement_list 23327)) 10943 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23328) ("~" 124)) 10944 ((backup_fn_flags 23329)) 10945 ((T_STRING 23332)) 10946 ((T_EXTENDS 2902) (extends_from 23333)) 10947 ((T_IMPLEMENTS 2904) (implements_list 23334)) 10948 ((backup_doc_comment 23335)) 10949 (("{" 23336)) 10950 ((class_statement_list 23337)) 10951 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23338)) 10952 ((":" 23343)) 10953 ((";" 23342)) 10954 ((inner_statement_list 23344)) 10955 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 23345) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 10956 ((";" 23346)) 10957 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23348) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10958 (("%" 280) (")" 23349) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10959 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23350) (";" 16136) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16137) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16139) (T_DEC 22) (T_DECLARE 16140) (T_DIR 24) (T_DNUMBER 25) (T_DO 16141) (T_DOUBLE_CAST 27) (T_ECHO 16142) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16144) (T_FOREACH 16145) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16146) (T_GOTO 16147) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16150) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16152) (T_START_HEREDOC 67) (T_STATIC 16153) (T_STRING 16154) (T_STRING_CAST 70) (T_SWITCH 16155) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16157) (T_UNSET 16158) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16159) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16160) (alt_if_stmt_without_else 16161) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16167) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16170) (if_stmt_without_else 16171) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23351) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 23352) ("{" 16176) ("~" 124)) 10960 ((inner_statement_list 23353)) 10961 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 23354) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 10962 ((";" 23355)) 10963 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 23357) (variable 16049) (variable_class_name 16050)) 10964 (("," 16062) (possible_comma 23358)) 10965 ((")" 23359)) 10966 ((";" 23360)) 10967 ((inner_statement_list 23362)) 10968 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23363) ("~" 124)) 10969 ((catch_list 23364)) 10970 ((T_CATCH 23365) (T_FINALLY 23366) (finally_statement 23367)) 10971 (("(" 23371)) 10972 (("{" 23368)) 10973 ((inner_statement_list 23369)) 10974 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23370) ("~" 124)) 10975 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 23372) (class_name 16111) (name 16112)) 10976 ((T_VARIABLE 16113) (optional_variable 23373) ("|" 16115)) 10977 ((")" 23374)) 10978 (("{" 23375)) 10979 ((inner_statement_list 23376)) 10980 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23377) ("~" 124)) 10981 ((backup_doc_comment 23379)) 10982 (("{" 23380)) 10983 ((class_statement_list 23381)) 10984 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23382)) 10985 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23384) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 10986 (("%" 280) (")" 23385) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 10987 ((":" 23386) (switch_case_list 23387) ("{" 23388)) 10988 ((";" 23394) (case_list 23395)) 10989 ((";" 23389) (case_list 23390)) 10990 ((case_list 23392)) 10991 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 23391)) 10992 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 23393)) 10993 ((case_list 23398)) 10994 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 23396)) 10995 ((";" 23397)) 10996 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 23399)) 10997 ((";" 23400)) 10998 (("," 16412) (";" 23403)) 10999 ((";" 23405)) 11000 ((T_EXTENDS 19717) (interface_extends_list 23407)) 11001 ((backup_doc_comment 23408)) 11002 (("{" 23409)) 11003 ((class_statement_list 23410)) 11004 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23411)) 11005 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23413) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11006 (("%" 280) (")" 23414) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11007 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19727) (";" 16184) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16185) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16186) (T_DEC 22) (T_DECLARE 16187) (T_DIR 24) (T_DNUMBER 25) (T_DO 16188) (T_DOUBLE_CAST 27) (T_ECHO 16189) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16190) (T_FOREACH 16191) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16192) (T_GOTO 16193) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16194) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16195) (T_START_HEREDOC 67) (T_STATIC 16196) (T_STRING 16197) (T_STRING_CAST 70) (T_SWITCH 16198) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16199) (T_UNSET 16200) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16201) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16202) (alt_if_stmt_without_else 16203) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16204) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16205) (if_stmt_without_else 16206) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23415) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16208) ("~" 124)) 11008 ((")" 23417)) 11009 ((";" 23418)) 11010 ((";" 23420)) 11011 (("," 17413) (";" 23422)) 11012 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 23424) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 11013 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 23425) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 11014 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 23426) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 11015 ((")" 23427) (T_DOUBLE_ARROW 23428)) 11016 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23431) (";" 16136) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16137) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16139) (T_DEC 22) (T_DECLARE 16140) (T_DIR 24) (T_DNUMBER 25) (T_DO 16141) (T_DOUBLE_CAST 27) (T_ECHO 16142) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16144) (T_FOREACH 16145) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16146) (T_GOTO 16147) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16150) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16152) (T_START_HEREDOC 67) (T_STATIC 16153) (T_STRING 16154) (T_STRING_CAST 70) (T_SWITCH 16155) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16157) (T_UNSET 16158) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16159) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16160) (alt_if_stmt_without_else 16161) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16167) (fn 102) (foreach_statement 23437) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16170) (if_stmt_without_else 16171) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23433) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16176) ("~" 124)) 11017 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 23429) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 11018 ((")" 23430)) 11019 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23431) (";" 16136) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16137) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16139) (T_DEC 22) (T_DECLARE 16140) (T_DIR 24) (T_DNUMBER 25) (T_DO 16141) (T_DOUBLE_CAST 27) (T_ECHO 16142) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16144) (T_FOREACH 16145) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16146) (T_GOTO 16147) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16150) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16152) (T_START_HEREDOC 67) (T_STATIC 16153) (T_STRING 16154) (T_STRING_CAST 70) (T_SWITCH 16155) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16157) (T_UNSET 16158) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16159) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16160) (alt_if_stmt_without_else 16161) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16167) (fn 102) (foreach_statement 23432) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16170) (if_stmt_without_else 16171) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23433) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16176) ("~" 124)) 11020 ((inner_statement_list 23434)) 11021 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 23435) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 11022 ((";" 23436)) 11023 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 23439) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11024 ((";" 23440)) 11025 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 23441) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11026 ((";" 23442)) 11027 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 23443) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 11028 ((")" 23444)) 11029 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23445) (";" 16136) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16137) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16139) (T_DEC 22) (T_DECLARE 16140) (T_DIR 24) (T_DNUMBER 25) (T_DO 16141) (T_DOUBLE_CAST 27) (T_ECHO 16142) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16144) (T_FOREACH 16145) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16146) (T_GOTO 16147) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16150) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16152) (T_START_HEREDOC 67) (T_STATIC 16153) (T_STRING 16154) (T_STRING_CAST 70) (T_SWITCH 16155) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16157) (T_UNSET 16158) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16159) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16160) (alt_if_stmt_without_else 16161) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16167) (fn 102) (for_statement 23446) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16170) (if_stmt_without_else 16171) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23447) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16176) ("~" 124)) 11030 ((inner_statement_list 23448)) 11031 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 23449) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 11032 ((";" 23450)) 11033 ((":" 19766) (enum_backing_type 23452)) 11034 ((T_IMPLEMENTS 2904) (implements_list 23453)) 11035 ((backup_doc_comment 23454)) 11036 (("{" 23455)) 11037 ((class_statement_list 23456)) 11038 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23457)) 11039 (("," 19251) (";" 23459)) 11040 ((T_WHILE 23461)) 11041 (("(" 23462)) 11042 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23463) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11043 (("%" 280) (")" 23464) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11044 ((";" 23465)) 11045 ((T_STRING 19424) (const_decl 19425) (const_list 23467)) 11046 ((")" 23468) ("," 19428)) 11047 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23469) (";" 16136) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 16137) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 16139) (T_DEC 22) (T_DECLARE 16140) (T_DIR 24) (T_DNUMBER 25) (T_DO 16141) (T_DOUBLE_CAST 27) (T_ECHO 16142) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 16144) (T_FOREACH 16145) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 16146) (T_GOTO 16147) (T_IF 16149) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16150) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16152) (T_START_HEREDOC 67) (T_STATIC 16153) (T_STRING 16154) (T_STRING_CAST 70) (T_SWITCH 16155) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16157) (T_UNSET 16158) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16159) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16160) (alt_if_stmt_without_else 16161) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 23470) (dereferenceable_scalar 99) (expr 16167) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16170) (if_stmt_without_else 16171) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23471) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16176) ("~" 124)) 11048 ((inner_statement_list 23472)) 11049 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 23473) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 11050 ((";" 23474)) 11051 ((";" 23476)) 11052 ((T_EXTENDS 2902) (extends_from 23478)) 11053 ((T_IMPLEMENTS 2904) (implements_list 23479)) 11054 ((backup_doc_comment 23480)) 11055 (("{" 23481)) 11056 ((class_statement_list 23482)) 11057 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23483)) 11058 ((";" 23485)) 11059 (("%" 21306) ("*" 21307) ("+" 21308) ("-" 21309) ("." 21310) ("/" 21311) (":" 16132) (";" 16133) ("<" 21312) (">" 21313) ("?" 21314) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 21315) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 21316) (T_BOOLEAN_AND 21317) (T_BOOLEAN_OR 21318) (T_COALESCE 21319) (T_INSTANCEOF 21320) (T_IS_EQUAL 21321) (T_IS_GREATER_OR_EQUAL 21322) (T_IS_IDENTICAL 21323) (T_IS_NOT_EQUAL 21324) (T_IS_NOT_IDENTICAL 21325) (T_IS_SMALLER_OR_EQUAL 21326) (T_LOGICAL_AND 21327) (T_LOGICAL_OR 21328) (T_LOGICAL_XOR 21329) (T_POW 21330) (T_SL 21331) (T_SPACESHIP 21332) (T_SR 21333) ("^" 21334) (case_separator 23487) ("|" 21335)) 11060 ((inner_statement_list 23488)) 11061 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 23490)) 11062 ((case_list 23495)) 11063 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 23493)) 11064 ((";" 23494)) 11065 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 23496)) 11066 ((";" 23497)) 11067 (("," 16412) (";" 23500)) 11068 ((";" 23502)) 11069 ((";" 23504)) 11070 (("," 17413) (";" 23506)) 11071 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 23508) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 11072 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 23509) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 11073 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 23510) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 11074 ((")" 23511) (T_DOUBLE_ARROW 23512)) 11075 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23515) (";" 15990) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15991) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15992) (T_DEC 22) (T_DECLARE 15993) (T_DIR 24) (T_DNUMBER 25) (T_DO 15994) (T_DOUBLE_CAST 27) (T_ECHO 15995) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15996) (T_FOREACH 15997) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15998) (T_GOTO 15999) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16000) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16001) (T_START_HEREDOC 67) (T_STATIC 16002) (T_STRING 16003) (T_STRING_CAST 70) (T_SWITCH 16004) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16005) (T_UNSET 16006) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16007) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16008) (alt_if_stmt_without_else 16009) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16010) (fn 102) (foreach_statement 23521) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16011) (if_stmt_without_else 16012) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23517) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16014) ("~" 124)) 11076 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 23513) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 11077 ((")" 23514)) 11078 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23515) (";" 15990) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15991) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15992) (T_DEC 22) (T_DECLARE 15993) (T_DIR 24) (T_DNUMBER 25) (T_DO 15994) (T_DOUBLE_CAST 27) (T_ECHO 15995) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15996) (T_FOREACH 15997) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15998) (T_GOTO 15999) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16000) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16001) (T_START_HEREDOC 67) (T_STATIC 16002) (T_STRING 16003) (T_STRING_CAST 70) (T_SWITCH 16004) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16005) (T_UNSET 16006) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16007) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16008) (alt_if_stmt_without_else 16009) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16010) (fn 102) (foreach_statement 23516) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16011) (if_stmt_without_else 16012) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23517) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16014) ("~" 124)) 11079 ((inner_statement_list 23518)) 11080 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 23519) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 11081 ((";" 23520)) 11082 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 23523) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11083 ((";" 23524)) 11084 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 23525) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11085 ((";" 23526)) 11086 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 23527) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 11087 ((")" 23528)) 11088 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23529) (";" 15990) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15991) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15992) (T_DEC 22) (T_DECLARE 15993) (T_DIR 24) (T_DNUMBER 25) (T_DO 15994) (T_DOUBLE_CAST 27) (T_ECHO 15995) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15996) (T_FOREACH 15997) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15998) (T_GOTO 15999) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16000) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16001) (T_START_HEREDOC 67) (T_STATIC 16002) (T_STRING 16003) (T_STRING_CAST 70) (T_SWITCH 16004) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16005) (T_UNSET 16006) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16007) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16008) (alt_if_stmt_without_else 16009) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16010) (fn 102) (for_statement 23530) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16011) (if_stmt_without_else 16012) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23531) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16014) ("~" 124)) 11089 ((inner_statement_list 23532)) 11090 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 23533) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 11091 ((";" 23534)) 11092 (("," 19251) (";" 23536)) 11093 ((T_WHILE 23538)) 11094 (("(" 23539)) 11095 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23540) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11096 (("%" 280) (")" 23541) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11097 ((";" 23542)) 11098 ((T_STRING 19424) (const_decl 19425) (const_list 23544)) 11099 ((")" 23545) ("," 19428)) 11100 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23546) (";" 15990) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15991) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15992) (T_DEC 22) (T_DECLARE 15993) (T_DIR 24) (T_DNUMBER 25) (T_DO 15994) (T_DOUBLE_CAST 27) (T_ECHO 15995) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15996) (T_FOREACH 15997) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15998) (T_GOTO 15999) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16000) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16001) (T_START_HEREDOC 67) (T_STATIC 16002) (T_STRING 16003) (T_STRING_CAST 70) (T_SWITCH 16004) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16005) (T_UNSET 16006) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16007) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16008) (alt_if_stmt_without_else 16009) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 23547) (dereferenceable_scalar 99) (expr 16010) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16011) (if_stmt_without_else 16012) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23548) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16014) ("~" 124)) 11101 ((inner_statement_list 23549)) 11102 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 23550) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 11103 ((";" 23551)) 11104 ((";" 23553)) 11105 ((";" 23555)) 11106 ((T_STRING 23558) (backup_doc_comment 137)) 11107 ((backup_doc_comment 23559)) 11108 (("(" 23560)) 11109 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 23561)) 11110 ((")" 23562)) 11111 ((":" 150) (return_type 23563)) 11112 ((backup_fn_flags 23564)) 11113 (("{" 23565)) 11114 ((inner_statement_list 23566)) 11115 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23567) ("~" 124)) 11116 ((backup_fn_flags 23568)) 11117 ((T_STRING 23571)) 11118 ((T_EXTENDS 2902) (extends_from 23572)) 11119 ((T_IMPLEMENTS 2904) (implements_list 23573)) 11120 ((backup_doc_comment 23574)) 11121 (("{" 23575)) 11122 ((class_statement_list 23576)) 11123 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23577)) 11124 ((":" 23582)) 11125 ((";" 23581)) 11126 ((inner_statement_list 23583)) 11127 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 23584) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 11128 ((";" 23585)) 11129 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23587) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11130 (("%" 280) (")" 23588) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11131 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23589) (";" 15941) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15973) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23590) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 23591) ("{" 15982) ("~" 124)) 11132 ((inner_statement_list 23592)) 11133 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 23593) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 11134 ((";" 23594)) 11135 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 23596) (variable 16049) (variable_class_name 16050)) 11136 (("," 16062) (possible_comma 23597)) 11137 ((")" 23598)) 11138 ((";" 23599)) 11139 ((inner_statement_list 23601)) 11140 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23602) ("~" 124)) 11141 ((catch_list 23603)) 11142 ((T_CATCH 23604) (T_FINALLY 23605) (finally_statement 23606)) 11143 (("(" 23610)) 11144 (("{" 23607)) 11145 ((inner_statement_list 23608)) 11146 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23609) ("~" 124)) 11147 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 23611) (class_name 16111) (name 16112)) 11148 ((T_VARIABLE 16113) (optional_variable 23612) ("|" 16115)) 11149 ((")" 23613)) 11150 (("{" 23614)) 11151 ((inner_statement_list 23615)) 11152 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23616) ("~" 124)) 11153 ((backup_doc_comment 23618)) 11154 (("{" 23619)) 11155 ((class_statement_list 23620)) 11156 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23621)) 11157 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23623) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11158 (("%" 280) (")" 23624) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11159 ((":" 23625) (switch_case_list 23626) ("{" 23627)) 11160 ((";" 23633) (case_list 23634)) 11161 ((";" 23628) (case_list 23629)) 11162 ((case_list 23631)) 11163 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 23630)) 11164 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 23632)) 11165 ((case_list 23637)) 11166 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 23635)) 11167 ((";" 23636)) 11168 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 23638)) 11169 ((";" 23639)) 11170 (("," 16412) (";" 23642)) 11171 ((";" 23644)) 11172 ((T_EXTENDS 19717) (interface_extends_list 23646)) 11173 ((backup_doc_comment 23647)) 11174 (("{" 23648)) 11175 ((class_statement_list 23649)) 11176 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23650)) 11177 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23652) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11178 (("%" 280) (")" 23653) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11179 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19727) (";" 15990) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15991) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15992) (T_DEC 22) (T_DECLARE 15993) (T_DIR 24) (T_DNUMBER 25) (T_DO 15994) (T_DOUBLE_CAST 27) (T_ECHO 15995) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15996) (T_FOREACH 15997) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15998) (T_GOTO 15999) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 16000) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 16001) (T_START_HEREDOC 67) (T_STATIC 16002) (T_STRING 16003) (T_STRING_CAST 70) (T_SWITCH 16004) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 16005) (T_UNSET 16006) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 16007) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 16008) (alt_if_stmt_without_else 16009) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 16010) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 16011) (if_stmt_without_else 16012) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23654) (static_member 118) (variable 121) (variable_class_name 122) ("{" 16014) ("~" 124)) 11180 ((")" 23656)) 11181 ((";" 23657)) 11182 ((";" 23659)) 11183 (("," 17413) (";" 23661)) 11184 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 23663) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 11185 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 23664) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 11186 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 23665) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 11187 ((")" 23666) (T_DOUBLE_ARROW 23667)) 11188 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23670) (";" 15941) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15973) (fn 102) (foreach_statement 23676) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23672) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 11189 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 23668) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 11190 ((")" 23669)) 11191 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23670) (";" 15941) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15973) (fn 102) (foreach_statement 23671) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23672) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 11192 ((inner_statement_list 23673)) 11193 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 23674) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 11194 ((";" 23675)) 11195 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 23678) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11196 ((";" 23679)) 11197 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 23680) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11198 ((";" 23681)) 11199 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 23682) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 11200 ((")" 23683)) 11201 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23684) (";" 15941) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15973) (fn 102) (for_statement 23685) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23686) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 11202 ((inner_statement_list 23687)) 11203 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 23688) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 11204 ((";" 23689)) 11205 ((":" 19766) (enum_backing_type 23691)) 11206 ((T_IMPLEMENTS 2904) (implements_list 23692)) 11207 ((backup_doc_comment 23693)) 11208 (("{" 23694)) 11209 ((class_statement_list 23695)) 11210 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23696)) 11211 (("," 19251) (";" 23699)) 11212 ((T_WHILE 23701)) 11213 (("(" 23702)) 11214 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23703) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11215 (("%" 280) (")" 23704) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11216 ((";" 23705)) 11217 ((T_STRING 19424) (const_decl 19425) (const_list 23707)) 11218 ((")" 23708) ("," 19428)) 11219 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23709) (";" 15941) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 23710) (dereferenceable_scalar 99) (expr 15973) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23711) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 11220 ((inner_statement_list 23712)) 11221 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 23713) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 11222 ((";" 23714)) 11223 ((";" 23716)) 11224 ((T_EXTENDS 2902) (extends_from 23718)) 11225 ((T_IMPLEMENTS 2904) (implements_list 23719)) 11226 ((backup_doc_comment 23720)) 11227 (("{" 23721)) 11228 ((class_statement_list 23722)) 11229 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23723)) 11230 ((";" 23725)) 11231 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 23727) (variable 16049) (variable_class_name 16050)) 11232 (("," 16062) (possible_comma 23728)) 11233 ((")" 23729)) 11234 ((";" 23730)) 11235 ((inner_statement_list 23732)) 11236 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23733) ("~" 124)) 11237 ((catch_list 23734)) 11238 ((T_CATCH 23735) (T_FINALLY 23736) (finally_statement 23737)) 11239 (("(" 23741)) 11240 (("{" 23738)) 11241 ((inner_statement_list 23739)) 11242 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23740) ("~" 124)) 11243 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 23742) (class_name 16111) (name 16112)) 11244 ((T_VARIABLE 16113) (optional_variable 23743) ("|" 16115)) 11245 ((")" 23744)) 11246 (("{" 23745)) 11247 ((inner_statement_list 23746)) 11248 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23747) ("~" 124)) 11249 ((backup_doc_comment 23749)) 11250 (("{" 23750)) 11251 ((class_statement_list 23751)) 11252 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23752)) 11253 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23754) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11254 (("%" 280) (")" 23755) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11255 ((":" 23756) (switch_case_list 23757) ("{" 23758)) 11256 ((";" 23764) (case_list 23765)) 11257 ((";" 23759) (case_list 23760)) 11258 ((case_list 23762)) 11259 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 23761)) 11260 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 23763)) 11261 ((case_list 23768)) 11262 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 23766)) 11263 ((";" 23767)) 11264 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 23769)) 11265 ((";" 23770)) 11266 (("," 16412) (";" 23773)) 11267 ((";" 23775)) 11268 ((T_EXTENDS 19717) (interface_extends_list 23777)) 11269 ((backup_doc_comment 23778)) 11270 (("{" 23779)) 11271 ((class_statement_list 23780)) 11272 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23781)) 11273 ((")" 23783)) 11274 ((";" 23784)) 11275 ((";" 23786)) 11276 (("," 17413) (";" 23788)) 11277 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 23790) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 11278 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 23791) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 11279 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 23792) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 11280 ((")" 23793) (T_DOUBLE_ARROW 23794)) 11281 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23797) (";" 15860) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15891) (fn 102) (foreach_statement 23803) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23799) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 11282 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 23795) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 11283 ((")" 23796)) 11284 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23797) (";" 15860) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15891) (fn 102) (foreach_statement 23798) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23799) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 11285 ((inner_statement_list 23800)) 11286 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 23801) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 11287 ((";" 23802)) 11288 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 23805) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11289 ((";" 23806)) 11290 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 23807) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11291 ((";" 23808)) 11292 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 23809) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 11293 ((")" 23810)) 11294 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23811) (";" 15860) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15891) (fn 102) (for_statement 23812) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23813) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 11295 ((inner_statement_list 23814)) 11296 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 23815) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 11297 ((";" 23816)) 11298 ((":" 19766) (enum_backing_type 23818)) 11299 ((T_IMPLEMENTS 2904) (implements_list 23819)) 11300 ((backup_doc_comment 23820)) 11301 (("{" 23821)) 11302 ((class_statement_list 23822)) 11303 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23823)) 11304 (("," 19251) (";" 23826)) 11305 ((T_WHILE 23828)) 11306 (("(" 23829)) 11307 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23830) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11308 (("%" 280) (")" 23831) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11309 ((";" 23832)) 11310 ((T_STRING 19424) (const_decl 19425) (const_list 23834)) 11311 ((")" 23835) ("," 19428)) 11312 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23836) (";" 15860) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 23837) (dereferenceable_scalar 99) (expr 15891) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23838) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 11313 ((inner_statement_list 23839)) 11314 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 23840) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 11315 ((";" 23841)) 11316 ((";" 23843)) 11317 ((T_EXTENDS 2902) (extends_from 23845)) 11318 ((T_IMPLEMENTS 2904) (implements_list 23846)) 11319 ((backup_doc_comment 23847)) 11320 (("{" 23848)) 11321 ((class_statement_list 23849)) 11322 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23850)) 11323 ((";" 23852)) 11324 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23854) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11325 (("%" 280) (")" 23855) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11326 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23856) (";" 15775) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15776) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15778) (T_DEC 22) (T_DECLARE 15779) (T_DIR 24) (T_DNUMBER 25) (T_DO 15780) (T_DOUBLE_CAST 27) (T_ECHO 15781) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15783) (T_FOREACH 15784) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15785) (T_GOTO 15786) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15789) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15791) (T_START_HEREDOC 67) (T_STATIC 15792) (T_STRING 15793) (T_STRING_CAST 70) (T_SWITCH 15794) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15796) (T_UNSET 15797) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15798) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15799) (alt_if_stmt_without_else 15800) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15806) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15809) (if_stmt_without_else 15810) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23857) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 23858) ("{" 15815) ("~" 124)) 11327 ((inner_statement_list 23859)) 11328 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 23860) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 11329 ((";" 23861)) 11330 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 23863) (variable 16049) (variable_class_name 16050)) 11331 (("," 16062) (possible_comma 23864)) 11332 ((")" 23865)) 11333 ((";" 23866)) 11334 ((inner_statement_list 23868)) 11335 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23869) ("~" 124)) 11336 ((catch_list 23870)) 11337 ((T_CATCH 23871) (T_FINALLY 23872) (finally_statement 23873)) 11338 (("(" 23877)) 11339 (("{" 23874)) 11340 ((inner_statement_list 23875)) 11341 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23876) ("~" 124)) 11342 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 23878) (class_name 16111) (name 16112)) 11343 ((T_VARIABLE 16113) (optional_variable 23879) ("|" 16115)) 11344 ((")" 23880)) 11345 (("{" 23881)) 11346 ((inner_statement_list 23882)) 11347 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 23883) ("~" 124)) 11348 ((backup_doc_comment 23885)) 11349 (("{" 23886)) 11350 ((class_statement_list 23887)) 11351 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23888)) 11352 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23890) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11353 (("%" 280) (")" 23891) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11354 ((":" 23892) (switch_case_list 23893) ("{" 23894)) 11355 ((";" 23900) (case_list 23901)) 11356 ((";" 23895) (case_list 23896)) 11357 ((case_list 23898)) 11358 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 23897)) 11359 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 23899)) 11360 ((case_list 23904)) 11361 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 23902)) 11362 ((";" 23903)) 11363 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 23905)) 11364 ((";" 23906)) 11365 (("," 16412) (";" 23909)) 11366 ((";" 23911)) 11367 ((T_EXTENDS 19717) (interface_extends_list 23913)) 11368 ((backup_doc_comment 23914)) 11369 (("{" 23915)) 11370 ((class_statement_list 23916)) 11371 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23917)) 11372 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23919) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11373 (("%" 280) (")" 23920) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11374 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19727) (";" 15775) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15776) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15778) (T_DEC 22) (T_DECLARE 15779) (T_DIR 24) (T_DNUMBER 25) (T_DO 15780) (T_DOUBLE_CAST 27) (T_ECHO 15781) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15783) (T_FOREACH 15784) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15785) (T_GOTO 15786) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15789) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15791) (T_START_HEREDOC 67) (T_STATIC 15792) (T_STRING 15793) (T_STRING_CAST 70) (T_SWITCH 15794) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15796) (T_UNSET 15797) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15798) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15799) (alt_if_stmt_without_else 15800) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15806) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15809) (if_stmt_without_else 15810) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23921) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15815) ("~" 124)) 11375 ((")" 23923)) 11376 ((";" 23924)) 11377 ((";" 23926)) 11378 (("," 17413) (";" 23928)) 11379 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 23930) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 11380 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 23931) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 11381 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 23932) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 11382 ((")" 23933) (T_DOUBLE_ARROW 23934)) 11383 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23937) (";" 15775) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15776) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15778) (T_DEC 22) (T_DECLARE 15779) (T_DIR 24) (T_DNUMBER 25) (T_DO 15780) (T_DOUBLE_CAST 27) (T_ECHO 15781) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15783) (T_FOREACH 15784) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15785) (T_GOTO 15786) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15789) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15791) (T_START_HEREDOC 67) (T_STATIC 15792) (T_STRING 15793) (T_STRING_CAST 70) (T_SWITCH 15794) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15796) (T_UNSET 15797) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15798) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15799) (alt_if_stmt_without_else 15800) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15806) (fn 102) (foreach_statement 23943) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15809) (if_stmt_without_else 15810) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23939) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15815) ("~" 124)) 11384 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 23935) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 11385 ((")" 23936)) 11386 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23937) (";" 15775) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15776) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15778) (T_DEC 22) (T_DECLARE 15779) (T_DIR 24) (T_DNUMBER 25) (T_DO 15780) (T_DOUBLE_CAST 27) (T_ECHO 15781) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15783) (T_FOREACH 15784) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15785) (T_GOTO 15786) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15789) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15791) (T_START_HEREDOC 67) (T_STATIC 15792) (T_STRING 15793) (T_STRING_CAST 70) (T_SWITCH 15794) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15796) (T_UNSET 15797) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15798) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15799) (alt_if_stmt_without_else 15800) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15806) (fn 102) (foreach_statement 23938) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15809) (if_stmt_without_else 15810) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23939) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15815) ("~" 124)) 11387 ((inner_statement_list 23940)) 11388 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 23941) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 11389 ((";" 23942)) 11390 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 23945) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11391 ((";" 23946)) 11392 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 23947) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11393 ((";" 23948)) 11394 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 23949) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 11395 ((")" 23950)) 11396 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23951) (";" 15775) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15776) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15778) (T_DEC 22) (T_DECLARE 15779) (T_DIR 24) (T_DNUMBER 25) (T_DO 15780) (T_DOUBLE_CAST 27) (T_ECHO 15781) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15783) (T_FOREACH 15784) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15785) (T_GOTO 15786) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15789) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15791) (T_START_HEREDOC 67) (T_STATIC 15792) (T_STRING 15793) (T_STRING_CAST 70) (T_SWITCH 15794) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15796) (T_UNSET 15797) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15798) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15799) (alt_if_stmt_without_else 15800) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15806) (fn 102) (for_statement 23952) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15809) (if_stmt_without_else 15810) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23953) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15815) ("~" 124)) 11397 ((inner_statement_list 23954)) 11398 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 23955) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 11399 ((";" 23956)) 11400 ((":" 19766) (enum_backing_type 23958)) 11401 ((T_IMPLEMENTS 2904) (implements_list 23959)) 11402 ((backup_doc_comment 23960)) 11403 (("{" 23961)) 11404 ((class_statement_list 23962)) 11405 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23963)) 11406 (("," 19251) (";" 23965)) 11407 ((T_WHILE 23967)) 11408 (("(" 23968)) 11409 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23969) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11410 (("%" 280) (")" 23970) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11411 ((";" 23971)) 11412 ((T_STRING 19424) (const_decl 19425) (const_list 23973)) 11413 ((")" 23974) ("," 19428)) 11414 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23975) (";" 15775) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15776) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15778) (T_DEC 22) (T_DECLARE 15779) (T_DIR 24) (T_DNUMBER 25) (T_DO 15780) (T_DOUBLE_CAST 27) (T_ECHO 15781) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15783) (T_FOREACH 15784) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15785) (T_GOTO 15786) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15789) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15791) (T_START_HEREDOC 67) (T_STATIC 15792) (T_STRING 15793) (T_STRING_CAST 70) (T_SWITCH 15794) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15796) (T_UNSET 15797) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15798) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15799) (alt_if_stmt_without_else 15800) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 23976) (dereferenceable_scalar 99) (expr 15806) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15809) (if_stmt_without_else 15810) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 23977) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15815) ("~" 124)) 11415 ((inner_statement_list 23978)) 11416 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 23979) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 11417 ((";" 23980)) 11418 ((";" 23982)) 11419 ((T_EXTENDS 2902) (extends_from 23984)) 11420 ((T_IMPLEMENTS 2904) (implements_list 23985)) 11421 ((backup_doc_comment 23986)) 11422 (("{" 23987)) 11423 ((class_statement_list 23988)) 11424 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 23989)) 11425 ((";" 23991)) 11426 ((inner_statement_list 23993)) 11427 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 23994) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 11428 ((";" 23995)) 11429 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 23997) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11430 (("%" 280) (")" 23998) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11431 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23999) (";" 15736) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15737) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15738) (T_DEC 22) (T_DECLARE 15739) (T_DIR 24) (T_DNUMBER 25) (T_DO 15740) (T_DOUBLE_CAST 27) (T_ECHO 15741) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15742) (T_FOREACH 15743) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15744) (T_GOTO 15745) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15746) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15747) (T_START_HEREDOC 67) (T_STATIC 15748) (T_STRING 15749) (T_STRING_CAST 70) (T_SWITCH 15750) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15751) (T_UNSET 15752) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15753) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15754) (alt_if_stmt_without_else 15755) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15756) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15757) (if_stmt_without_else 15758) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24000) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 24001) ("{" 15760) ("~" 124)) 11432 ((inner_statement_list 24002)) 11433 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 24003) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 11434 ((";" 24004)) 11435 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 24006) (variable 16049) (variable_class_name 16050)) 11436 (("," 16062) (possible_comma 24007)) 11437 ((")" 24008)) 11438 ((";" 24009)) 11439 ((inner_statement_list 24011)) 11440 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 24012) ("~" 124)) 11441 ((catch_list 24013)) 11442 ((T_CATCH 24014) (T_FINALLY 24015) (finally_statement 24016)) 11443 (("(" 24020)) 11444 (("{" 24017)) 11445 ((inner_statement_list 24018)) 11446 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 24019) ("~" 124)) 11447 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 24021) (class_name 16111) (name 16112)) 11448 ((T_VARIABLE 16113) (optional_variable 24022) ("|" 16115)) 11449 ((")" 24023)) 11450 (("{" 24024)) 11451 ((inner_statement_list 24025)) 11452 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 24026) ("~" 124)) 11453 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24028) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11454 (("%" 280) (")" 24029) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11455 ((":" 24030) (switch_case_list 24031) ("{" 24032)) 11456 ((";" 24038) (case_list 24039)) 11457 ((";" 24033) (case_list 24034)) 11458 ((case_list 24036)) 11459 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 24035)) 11460 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 24037)) 11461 ((case_list 24042)) 11462 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 24040)) 11463 ((";" 24041)) 11464 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 24043)) 11465 ((";" 24044)) 11466 (("," 16412) (";" 24047)) 11467 ((";" 24049)) 11468 ((";" 24051)) 11469 (("," 17413) (";" 24053)) 11470 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 24055) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 11471 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 24056) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 11472 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 24057) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 11473 ((")" 24058) (T_DOUBLE_ARROW 24059)) 11474 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24062) (";" 15736) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15737) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15738) (T_DEC 22) (T_DECLARE 15739) (T_DIR 24) (T_DNUMBER 25) (T_DO 15740) (T_DOUBLE_CAST 27) (T_ECHO 15741) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15742) (T_FOREACH 15743) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15744) (T_GOTO 15745) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15746) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15747) (T_START_HEREDOC 67) (T_STATIC 15748) (T_STRING 15749) (T_STRING_CAST 70) (T_SWITCH 15750) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15751) (T_UNSET 15752) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15753) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15754) (alt_if_stmt_without_else 15755) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15756) (fn 102) (foreach_statement 24068) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15757) (if_stmt_without_else 15758) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24064) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15760) ("~" 124)) 11475 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 24060) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 11476 ((")" 24061)) 11477 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24062) (";" 15736) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15737) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15738) (T_DEC 22) (T_DECLARE 15739) (T_DIR 24) (T_DNUMBER 25) (T_DO 15740) (T_DOUBLE_CAST 27) (T_ECHO 15741) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15742) (T_FOREACH 15743) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15744) (T_GOTO 15745) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15746) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15747) (T_START_HEREDOC 67) (T_STATIC 15748) (T_STRING 15749) (T_STRING_CAST 70) (T_SWITCH 15750) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15751) (T_UNSET 15752) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15753) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15754) (alt_if_stmt_without_else 15755) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15756) (fn 102) (foreach_statement 24063) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15757) (if_stmt_without_else 15758) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24064) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15760) ("~" 124)) 11478 ((inner_statement_list 24065)) 11479 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 24066) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 11480 ((";" 24067)) 11481 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 24070) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11482 ((";" 24071)) 11483 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 24072) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11484 ((";" 24073)) 11485 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 24074) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 11486 ((")" 24075)) 11487 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24076) (";" 15736) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15737) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15738) (T_DEC 22) (T_DECLARE 15739) (T_DIR 24) (T_DNUMBER 25) (T_DO 15740) (T_DOUBLE_CAST 27) (T_ECHO 15741) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15742) (T_FOREACH 15743) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15744) (T_GOTO 15745) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15746) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15747) (T_START_HEREDOC 67) (T_STATIC 15748) (T_STRING 15749) (T_STRING_CAST 70) (T_SWITCH 15750) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15751) (T_UNSET 15752) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15753) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15754) (alt_if_stmt_without_else 15755) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15756) (fn 102) (for_statement 24077) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15757) (if_stmt_without_else 15758) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24078) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15760) ("~" 124)) 11488 ((inner_statement_list 24079)) 11489 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 24080) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 11490 ((";" 24081)) 11491 (("," 19251) (";" 24083)) 11492 ((T_WHILE 24085)) 11493 (("(" 24086)) 11494 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24087) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11495 (("%" 280) (")" 24088) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11496 ((";" 24089)) 11497 ((T_STRING 19424) (const_decl 19425) (const_list 24091)) 11498 ((")" 24092) ("," 19428)) 11499 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24093) (";" 15736) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15737) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15738) (T_DEC 22) (T_DECLARE 15739) (T_DIR 24) (T_DNUMBER 25) (T_DO 15740) (T_DOUBLE_CAST 27) (T_ECHO 15741) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15742) (T_FOREACH 15743) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15744) (T_GOTO 15745) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15746) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15747) (T_START_HEREDOC 67) (T_STATIC 15748) (T_STRING 15749) (T_STRING_CAST 70) (T_SWITCH 15750) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15751) (T_UNSET 15752) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15753) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15754) (alt_if_stmt_without_else 15755) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 24094) (dereferenceable_scalar 99) (expr 15756) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15757) (if_stmt_without_else 15758) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24095) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15760) ("~" 124)) 11500 ((inner_statement_list 24096)) 11501 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 24097) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 11502 ((";" 24098)) 11503 ((";" 24100)) 11504 ((";" 24102)) 11505 (("$" 205) (T_ABSTRACT 14501) (T_ARRAY 14502) (T_AS 14503) (T_BREAK 14504) (T_CALLABLE 14505) (T_CASE 14506) (T_CATCH 14507) (T_CLASS 14508) (T_CLASS_C 14509) (T_CLONE 14510) (T_CONST 14511) (T_CONTINUE 14512) (T_DECLARE 14513) (T_DEFAULT 14514) (T_DIR 14515) (T_DO 14516) (T_ECHO 14517) (T_ELSE 14518) (T_ELSEIF 14519) (T_EMPTY 14520) (T_ENDDECLARE 14521) (T_ENDFOR 14522) (T_ENDFOREACH 14523) (T_ENDIF 14524) (T_ENDSWITCH 14525) (T_ENDWHILE 14526) (T_ENUM 14527) (T_EVAL 14528) (T_EXIT 14529) (T_EXTENDS 14530) (T_FILE 14531) (T_FINAL 14532) (T_FINALLY 14533) (T_FN 14534) (T_FOR 14535) (T_FOREACH 14536) (T_FUNCTION 14537) (T_FUNC_C 14538) (T_GLOBAL 14539) (T_GOTO 14540) (T_IF 14541) (T_IMPLEMENTS 14542) (T_INCLUDE 14543) (T_INCLUDE_ONCE 14544) (T_INSTANCEOF 14545) (T_INSTEADOF 14546) (T_INTERFACE 14547) (T_ISSET 14548) (T_LINE 14549) (T_LIST 14550) (T_LOGICAL_AND 14551) (T_LOGICAL_OR 14552) (T_LOGICAL_XOR 14553) (T_MATCH 14554) (T_METHOD_C 14555) (T_NAMESPACE 14556) (T_NEW 14557) (T_NS_C 14558) (T_PRINT 14559) (T_PRIVATE 14560) (T_PROTECTED 14561) (T_PUBLIC 14562) (T_READONLY 14563) (T_REQUIRE 14564) (T_REQUIRE_ONCE 14565) (T_RETURN 14566) (T_STATIC 14567) (T_STRING 14568) (T_SWITCH 14569) (T_THROW 14570) (T_TRAIT 14571) (T_TRAIT_C 14572) (T_TRY 14573) (T_UNSET 14574) (T_USE 14575) (T_VAR 14576) (T_VARIABLE 251) (T_WHILE 14577) (T_YIELD 14578) (identifier 24104) (member_name 24105) (reserved_non_modifiers 14581) (semi_reserved 14582) (simple_variable 24106) ("{" 914)) 11506 (("(" 14584) (argument_list 24107)) 11507 ((T_FN 35) (T_FUNCTION 126) (fn 265) #'266 (inline_function 24111)) 11508 (("$" 205) (T_STRING 24120) (T_VARIABLE 251) (property_name 24127) (simple_variable 24122) ("{" 24123)) 11509 (("$" 205) (T_STRING 24120) (T_VARIABLE 251) (property_name 24121) (simple_variable 24122) ("{" 24123)) 11510 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 24118) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 11511 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 24116) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11512 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 24117)) 11513 (("]" 24119)) 11514 (("(" 14584) (argument_list 24126)) 11515 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 24124) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11516 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 24125)) 11517 (("(" 14584) (argument_list 24128)) 11518 (("`" 24130)) 11519 (("]" 24132)) 11520 (("=" 24133)) 11521 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24134) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11522 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24631) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11523 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 24629) (encaps_var 5158)) 11524 (("$" 24138) (T_VARIABLE 24184) (simple_variable 24625) ("{" 24626)) 11525 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24623) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11526 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24622) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11527 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24621) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11528 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24620) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11529 (("(" 24617)) 11530 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24616) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11531 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24615) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11532 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24614) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11533 (("\"" 1489) ("$" 24437) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 24438) ("[" 1508) (array_object_dereferenceable 24439) (callable_expr 24440) (callable_variable 24441) (class_constant 1512) (class_name 24442) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 24443) (name 24444) (simple_variable 24445) (static_member 24446) (variable 24613) (variable_class_name 24448)) 11534 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24612) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11535 (("(" 24609)) 11536 (("(" 24606)) 11537 (("(" 24602) (exit_expr 24603)) 11538 (("\"" 1489) ("$" 24437) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 24438) ("[" 1508) (array_object_dereferenceable 24439) (callable_expr 24440) (callable_variable 24441) (class_constant 1512) (class_name 24442) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 24443) (name 24444) (simple_variable 24445) (static_member 24446) (variable 24601) (variable_class_name 24448)) 11539 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24600) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11540 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24599) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11541 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24598) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11542 (("(" 24594)) 11543 (("(" 24589)) 11544 (("(" 24583)) 11545 (("$" 15405) ("(" 15530) (T_ATTRIBUTE 2853) (T_CLASS 24562) (T_NAME_FULLY_QUALIFIED 15532) (T_NAME_QUALIFIED 15533) (T_NAME_RELATIVE 15534) (T_STATIC 15535) (T_STRING 15536) (T_VARIABLE 15406) (anonymous_class 24563) (attribute 2862) (attributes 24564) (class_name 15539) (class_name_reference 24565) (name 15541) (new_variable 15542) (simple_variable 15543)) 11546 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24561) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11547 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24560) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11548 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24559) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11549 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24558) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11550 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 24553) (T_END_HEREDOC 24554) (T_VARIABLE 2822) (encaps_list 24555) (encaps_var 2824)) 11551 ((T_FN 35) (T_FUNCTION 126) (fn 24198) #'24199 (inline_function 24552)) 11552 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24551) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11553 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24550) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11554 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24549) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11555 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24546) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11556 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24545) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11557 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 24541) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 11558 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 24539) (encaps_list 2729) (encaps_var 2730)) 11559 ((T_NULLSAFE_OBJECT_OPERATOR 24522) (T_OBJECT_OPERATOR 24523) ("[" 24524) ("{" 24525)) 11560 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 24519) (attribute 2654) (fn 24198) #'24199 (inline_function 24520)) 11561 (("(" 24398) (argument_list 24518)) 11562 ((T_PAAMAYIM_NEKUDOTAYIM 24513)) 11563 (("%" 24213) ("*" 24214) ("+" 24215) ("-" 24216) ("." 24217) ("/" 24218) ("<" 24219) (">" 24220) ("?" 24221) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 24222) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24223) (T_BOOLEAN_AND 24224) (T_BOOLEAN_OR 24225) (T_COALESCE 24226) (T_DOUBLE_ARROW 24511) (T_INSTANCEOF 24227) (T_IS_EQUAL 24228) (T_IS_GREATER_OR_EQUAL 24229) (T_IS_IDENTICAL 24230) (T_IS_NOT_EQUAL 24231) (T_IS_NOT_IDENTICAL 24232) (T_IS_SMALLER_OR_EQUAL 24233) (T_LOGICAL_AND 24234) (T_LOGICAL_OR 24235) (T_LOGICAL_XOR 24236) (T_POW 24237) (T_SL 24238) (T_SPACESHIP 24239) (T_SR 24240) ("^" 24241) ("|" 24242)) 11564 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 24500)) 11565 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 24488)) 11566 (("(" 24398) (argument_list 24487)) 11567 (("=" 24406) (T_AND_EQUAL 24407) (T_COALESCE_EQUAL 24408) (T_CONCAT_EQUAL 24409) (T_DEC 24410) (T_DIV_EQUAL 24411) (T_INC 24412) (T_MINUS_EQUAL 24413) (T_MOD_EQUAL 24414) (T_MUL_EQUAL 24415) (T_OR_EQUAL 24416) (T_PLUS_EQUAL 24417) (T_POW_EQUAL 24418) (T_SL_EQUAL 24419) (T_SR_EQUAL 24420) (T_XOR_EQUAL 24421)) 11568 ((T_PAAMAYIM_NEKUDOTAYIM 24314)) 11569 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24212) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11570 (("%" 24213) ("*" 24214) ("+" 24215) ("-" 24216) ("." 24217) ("/" 24218) ("<" 24219) (">" 24220) ("?" 24221) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 24222) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24223) (T_BOOLEAN_AND 24224) (T_BOOLEAN_OR 24225) (T_COALESCE 24226) (T_INSTANCEOF 24227) (T_IS_EQUAL 24228) (T_IS_GREATER_OR_EQUAL 24229) (T_IS_IDENTICAL 24230) (T_IS_NOT_EQUAL 24231) (T_IS_NOT_IDENTICAL 24232) (T_IS_SMALLER_OR_EQUAL 24233) (T_LOGICAL_AND 24234) (T_LOGICAL_OR 24235) (T_LOGICAL_XOR 24236) (T_POW 24237) (T_SL 24238) (T_SPACESHIP 24239) (T_SR 24240) ("^" 24241) ("|" 24242)) 11571 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24313) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11572 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24312) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11573 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24311) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11574 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24310) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11575 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24309) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11576 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24308) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11577 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24307) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11578 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24306) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11579 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 24301) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 24302) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 11580 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24300) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11581 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24299) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11582 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24298) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11583 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24297) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11584 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24296) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11585 (("$" 24258) ("(" 24259) (T_NAME_FULLY_QUALIFIED 24260) (T_NAME_QUALIFIED 24261) (T_NAME_RELATIVE 24262) (T_STATIC 24263) (T_STRING 24264) (T_VARIABLE 24265) (class_name 24266) (class_name_reference 24267) (name 24268) (new_variable 24269) (simple_variable 24270)) 11586 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24257) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11587 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24256) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11588 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24255) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11589 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24254) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11590 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24253) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11591 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24252) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11592 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24251) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11593 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24250) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11594 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24249) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11595 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24248) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11596 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24247) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11597 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24246) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11598 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24245) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11599 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24244) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11600 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24243) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11601 (("$" 24258) (T_VARIABLE 24265) (simple_variable 24292) ("{" 24293)) 11602 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24290) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11603 ((T_PAAMAYIM_NEKUDOTAYIM 24288)) 11604 ((T_NULLSAFE_OBJECT_OPERATOR 24271) (T_OBJECT_OPERATOR 24272) (T_PAAMAYIM_NEKUDOTAYIM 24273) ("[" 24274) ("{" 24275)) 11605 (("$" 24258) (T_STRING 24281) (T_VARIABLE 24265) (property_name 24287) (simple_variable 24283) ("{" 24284)) 11606 (("$" 24258) (T_STRING 24281) (T_VARIABLE 24265) (property_name 24282) (simple_variable 24283) ("{" 24284)) 11607 (("$" 24258) (T_VARIABLE 24265) (simple_variable 24280)) 11608 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 24278) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 11609 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 24276) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11610 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 24277)) 11611 (("]" 24279)) 11612 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 24285) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11613 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 24286)) 11614 (("$" 24258) (T_VARIABLE 24265) (simple_variable 24289)) 11615 (("%" 280) (")" 24291) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11616 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 24294) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11617 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 24295)) 11618 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24305) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11619 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 24303) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 11620 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24304) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11621 (("$" 24138) (T_ABSTRACT 24315) (T_ARRAY 24316) (T_AS 24317) (T_BREAK 24318) (T_CALLABLE 24319) (T_CASE 24320) (T_CATCH 24321) (T_CLASS 24322) (T_CLASS_C 24323) (T_CLONE 24324) (T_CONST 24325) (T_CONTINUE 24326) (T_DECLARE 24327) (T_DEFAULT 24328) (T_DIR 24329) (T_DO 24330) (T_ECHO 24331) (T_ELSE 24332) (T_ELSEIF 24333) (T_EMPTY 24334) (T_ENDDECLARE 24335) (T_ENDFOR 24336) (T_ENDFOREACH 24337) (T_ENDIF 24338) (T_ENDSWITCH 24339) (T_ENDWHILE 24340) (T_ENUM 24341) (T_EVAL 24342) (T_EXIT 24343) (T_EXTENDS 24344) (T_FILE 24345) (T_FINAL 24346) (T_FINALLY 24347) (T_FN 24348) (T_FOR 24349) (T_FOREACH 24350) (T_FUNCTION 24351) (T_FUNC_C 24352) (T_GLOBAL 24353) (T_GOTO 24354) (T_IF 24355) (T_IMPLEMENTS 24356) (T_INCLUDE 24357) (T_INCLUDE_ONCE 24358) (T_INSTANCEOF 24359) (T_INSTEADOF 24360) (T_INTERFACE 24361) (T_ISSET 24362) (T_LINE 24363) (T_LIST 24364) (T_LOGICAL_AND 24365) (T_LOGICAL_OR 24366) (T_LOGICAL_XOR 24367) (T_MATCH 24368) (T_METHOD_C 24369) (T_NAMESPACE 24370) (T_NEW 24371) (T_NS_C 24372) (T_PRINT 24373) (T_PRIVATE 24374) (T_PROTECTED 24375) (T_PUBLIC 24376) (T_READONLY 24377) (T_REQUIRE 24378) (T_REQUIRE_ONCE 24379) (T_RETURN 24380) (T_STATIC 24381) (T_STRING 24382) (T_SWITCH 24383) (T_THROW 24384) (T_TRAIT 24385) (T_TRAIT_C 24386) (T_TRY 24387) (T_UNSET 24388) (T_USE 24389) (T_VAR 24390) (T_VARIABLE 24184) (T_WHILE 24391) (T_YIELD 24392) (identifier 24393) (member_name 24394) (reserved_non_modifiers 24395) (semi_reserved 24396) (simple_variable 24397) ("{" 914)) 11622 (("(" 24398) (argument_list 24399)) 11623 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 24400) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 24401) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 24402) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11624 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 24405) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11625 (("," 1449) (possible_comma 24403)) 11626 ((")" 24404)) 11627 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (ampersand 24435) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24436) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11628 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24434) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11629 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24433) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11630 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24432) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11631 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24431) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11632 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24430) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11633 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24429) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11634 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24428) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11635 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24427) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11636 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24426) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11637 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24425) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11638 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24424) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11639 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24423) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11640 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24422) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11641 (("\"" 1489) ("$" 24437) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 24438) ("[" 1508) (array_object_dereferenceable 24439) (callable_expr 24440) (callable_variable 24441) (class_constant 1512) (class_name 24442) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 24443) (name 24444) (simple_variable 24445) (static_member 24446) (variable 24447) (variable_class_name 24448)) 11642 (("$" 24437) (T_VARIABLE 24438) (simple_variable 24483) ("{" 24484)) 11643 ((T_NULLSAFE_OBJECT_OPERATOR 24466) (T_OBJECT_OPERATOR 24467) ("[" 24468) ("{" 24469)) 11644 (("(" 24452) (argument_list 24465)) 11645 ((T_PAAMAYIM_NEKUDOTAYIM 24461)) 11646 (("(" 24452) (argument_list 24460)) 11647 ((T_PAAMAYIM_NEKUDOTAYIM 24449)) 11648 (("$" 24437) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 24438) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 24450) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 24451) ("{" 914)) 11649 (("(" 24452) (argument_list 24453)) 11650 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 24454) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 24455) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 24456) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11651 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 24459) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11652 (("," 1449) (possible_comma 24457)) 11653 ((")" 24458)) 11654 (("$" 24437) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 24438) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 24462) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 24463) ("{" 914)) 11655 (("(" 24452) (argument_list 24464)) 11656 (("$" 24437) (T_STRING 24474) (T_VARIABLE 24438) (property_name 24481) (simple_variable 24476) ("{" 24477)) 11657 (("$" 24437) (T_STRING 24474) (T_VARIABLE 24438) (property_name 24475) (simple_variable 24476) ("{" 24477)) 11658 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 24472) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 11659 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 24470) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11660 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 24471)) 11661 (("]" 24473)) 11662 (("(" 24452) (argument_list 24480)) 11663 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 24478) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11664 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 24479)) 11665 (("(" 24452) (argument_list 24482)) 11666 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 24485) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11667 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 24486)) 11668 ((backup_doc_comment 24489)) 11669 (("(" 24490)) 11670 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 24491)) 11671 ((")" 24492)) 11672 ((T_USE 148) (lexical_vars 24493)) 11673 ((":" 150) (return_type 24494)) 11674 ((backup_fn_flags 24495)) 11675 (("{" 24496)) 11676 ((inner_statement_list 24497)) 11677 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 24498) ("~" 124)) 11678 ((backup_fn_flags 24499)) 11679 ((backup_doc_comment 24501)) 11680 (("(" 24502)) 11681 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 24503)) 11682 ((")" 24504)) 11683 ((":" 2207) (return_type 24505)) 11684 ((T_DOUBLE_ARROW 24506)) 11685 ((backup_fn_flags 24507)) 11686 ((backup_lex_pos 24508)) 11687 (("!" 15104) ("\"" 15105) ("$" 15106) ("(" 15107) ("+" 15108) ("-" 15109) ("@" 15110) (T_ARRAY 15111) (T_ARRAY_CAST 15112) (T_ATTRIBUTE 125) (T_BOOL_CAST 15113) (T_CLASS_C 15114) (T_CLONE 15115) (T_CONSTANT_ENCAPSED_STRING 15116) (T_DEC 15117) (T_DIR 15118) (T_DNUMBER 15119) (T_DOUBLE_CAST 15120) (T_EMPTY 15121) (T_EVAL 15122) (T_EXIT 15123) (T_FILE 15124) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 15125) (T_INC 15126) (T_INCLUDE 15127) (T_INCLUDE_ONCE 15128) (T_INT_CAST 15129) (T_ISSET 15130) (T_LINE 15131) (T_LIST 15132) (T_LNUMBER 15133) (T_MATCH 15134) (T_METHOD_C 15135) (T_NAME_FULLY_QUALIFIED 15136) (T_NAME_QUALIFIED 15137) (T_NAME_RELATIVE 15138) (T_NEW 15139) (T_NS_C 15140) (T_OBJECT_CAST 15141) (T_PRINT 15142) (T_REQUIRE 15143) (T_REQUIRE_ONCE 15144) (T_START_HEREDOC 15145) (T_STATIC 15146) (T_STRING 15147) (T_STRING_CAST 15148) (T_THROW 15149) (T_TRAIT_C 15150) (T_UNSET_CAST 15151) (T_VARIABLE 15152) (T_YIELD 15153) (T_YIELD_FROM 15154) ("[" 15155) ("`" 15156) (array_object_dereferenceable 15157) (attribute 129) (attributes 15158) (callable_expr 15159) (callable_variable 15160) (class_constant 15161) (class_name 15162) (constant 15163) (dereferenceable_scalar 15164) (expr 24509) (fn 15166) (fully_dereferenceable 103) #'15167 (function_call 15168) (inline_function 15169) (internal_functions_in_yacc 15170) (match 15171) (name 15172) (new_expr 15173) (scalar 15174) (simple_variable 15175) (static_member 15176) (variable 15177) (variable_class_name 15178) ("~" 15179)) 11688 (("%" 15181) ("*" 15182) ("+" 15183) ("-" 15184) ("." 15185) ("/" 15186) ("<" 15187) (">" 15188) ("?" 15189) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15190) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15191) (T_BOOLEAN_AND 15192) (T_BOOLEAN_OR 15193) (T_COALESCE 15194) (T_INSTANCEOF 15195) (T_IS_EQUAL 15196) (T_IS_GREATER_OR_EQUAL 15197) (T_IS_IDENTICAL 15198) (T_IS_NOT_EQUAL 15199) (T_IS_NOT_IDENTICAL 15200) (T_IS_SMALLER_OR_EQUAL 15201) (T_LOGICAL_AND 15202) (T_LOGICAL_OR 15203) (T_LOGICAL_XOR 15204) (T_POW 15205) (T_SL 15206) (T_SPACESHIP 15207) (T_SR 15208) ("^" 15209) (backup_fn_flags 24510) ("|" 15210)) 11689 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24512) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11690 (("$" 24138) (T_ABSTRACT 24315) (T_ARRAY 24316) (T_AS 24317) (T_BREAK 24318) (T_CALLABLE 24319) (T_CASE 24320) (T_CATCH 24321) (T_CLASS 24322) (T_CLASS_C 24323) (T_CLONE 24324) (T_CONST 24325) (T_CONTINUE 24326) (T_DECLARE 24327) (T_DEFAULT 24328) (T_DIR 24329) (T_DO 24330) (T_ECHO 24331) (T_ELSE 24332) (T_ELSEIF 24333) (T_EMPTY 24334) (T_ENDDECLARE 24335) (T_ENDFOR 24336) (T_ENDFOREACH 24337) (T_ENDIF 24338) (T_ENDSWITCH 24339) (T_ENDWHILE 24340) (T_ENUM 24341) (T_EVAL 24342) (T_EXIT 24343) (T_EXTENDS 24344) (T_FILE 24345) (T_FINAL 24346) (T_FINALLY 24347) (T_FN 24348) (T_FOR 24349) (T_FOREACH 24350) (T_FUNCTION 24351) (T_FUNC_C 24352) (T_GLOBAL 24353) (T_GOTO 24354) (T_IF 24355) (T_IMPLEMENTS 24356) (T_INCLUDE 24357) (T_INCLUDE_ONCE 24358) (T_INSTANCEOF 24359) (T_INSTEADOF 24360) (T_INTERFACE 24361) (T_ISSET 24362) (T_LINE 24363) (T_LIST 24364) (T_LOGICAL_AND 24365) (T_LOGICAL_OR 24366) (T_LOGICAL_XOR 24367) (T_MATCH 24368) (T_METHOD_C 24369) (T_NAMESPACE 24370) (T_NEW 24371) (T_NS_C 24372) (T_PRINT 24373) (T_PRIVATE 24374) (T_PROTECTED 24375) (T_PUBLIC 24376) (T_READONLY 24377) (T_REQUIRE 24378) (T_REQUIRE_ONCE 24379) (T_RETURN 24380) (T_STATIC 24381) (T_STRING 24382) (T_SWITCH 24383) (T_THROW 24384) (T_TRAIT 24385) (T_TRAIT_C 24386) (T_TRY 24387) (T_UNSET 24388) (T_USE 24389) (T_VAR 24390) (T_VARIABLE 24184) (T_WHILE 24391) (T_YIELD 24392) (identifier 24514) (member_name 24515) (reserved_non_modifiers 24395) (semi_reserved 24396) (simple_variable 24516) ("{" 914)) 11691 (("(" 24398) (argument_list 24517)) 11692 ((T_FN 35) (T_FUNCTION 126) (fn 24198) #'24199 (inline_function 24521)) 11693 (("$" 24138) (T_STRING 24530) (T_VARIABLE 24184) (property_name 24537) (simple_variable 24532) ("{" 24533)) 11694 (("$" 24138) (T_STRING 24530) (T_VARIABLE 24184) (property_name 24531) (simple_variable 24532) ("{" 24533)) 11695 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 24528) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 11696 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 24526) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11697 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 24527)) 11698 (("]" 24529)) 11699 (("(" 24398) (argument_list 24536)) 11700 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 24534) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11701 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 24535)) 11702 (("(" 24398) (argument_list 24538)) 11703 (("`" 24540)) 11704 (("]" 24542)) 11705 (("=" 24543)) 11706 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24544) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11707 (("%" 24213) ("*" 24214) ("+" 24215) ("-" 24216) ("." 24217) ("/" 24218) ("<" 24219) (">" 24220) ("?" 24221) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 24222) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24223) (T_BOOLEAN_AND 24224) (T_BOOLEAN_OR 24225) (T_COALESCE 24226) (T_DOUBLE_ARROW 24547) (T_INSTANCEOF 24227) (T_IS_EQUAL 24228) (T_IS_GREATER_OR_EQUAL 24229) (T_IS_IDENTICAL 24230) (T_IS_NOT_EQUAL 24231) (T_IS_NOT_IDENTICAL 24232) (T_IS_SMALLER_OR_EQUAL 24233) (T_LOGICAL_AND 24234) (T_LOGICAL_OR 24235) (T_LOGICAL_XOR 24236) (T_POW 24237) (T_SL 24238) (T_SPACESHIP 24239) (T_SR 24240) ("^" 24241) ("|" 24242)) 11708 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24548) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11709 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 24557) (T_VARIABLE 2822) (encaps_var 2836)) 11710 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 24556) (T_VARIABLE 2822) (encaps_var 2827)) 11711 (("(" 2899) (argument_list 2900) (ctor_arguments 24576)) 11712 ((T_ATTRIBUTE 2853) (T_CLASS 24562) (anonymous_class 24575) (attribute 2898)) 11713 (("(" 24566) (argument_list 24567) (ctor_arguments 24568)) 11714 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 24569) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 24570) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 24571) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11715 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 24574) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11716 (("," 1449) (possible_comma 24572)) 11717 ((")" 24573)) 11718 ((T_EXTENDS 2902) (extends_from 24577)) 11719 ((T_IMPLEMENTS 2904) (implements_list 24578)) 11720 ((backup_doc_comment 24579)) 11721 (("{" 24580)) 11722 ((class_statement_list 24581)) 11723 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 24582)) 11724 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24584) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11725 (("%" 280) (")" 24585) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11726 (("{" 24586)) 11727 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 24587) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 11728 (("}" 24588)) 11729 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 24590) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 11730 ((")" 24591)) 11731 (("=" 24592)) 11732 (("!" 24136) ("\"" 24137) ("$" 24138) ("(" 24139) ("+" 24140) ("-" 24141) ("@" 24142) (T_ARRAY 24143) (T_ARRAY_CAST 24144) (T_ATTRIBUTE 125) (T_BOOL_CAST 24145) (T_CLASS_C 24146) (T_CLONE 24147) (T_CONSTANT_ENCAPSED_STRING 24148) (T_DEC 24149) (T_DIR 24150) (T_DNUMBER 24151) (T_DOUBLE_CAST 24152) (T_EMPTY 24153) (T_EVAL 24154) (T_EXIT 24155) (T_FILE 24156) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 24157) (T_INC 24158) (T_INCLUDE 24159) (T_INCLUDE_ONCE 24160) (T_INT_CAST 24161) (T_ISSET 24162) (T_LINE 24163) (T_LIST 24164) (T_LNUMBER 24165) (T_MATCH 24166) (T_METHOD_C 24167) (T_NAME_FULLY_QUALIFIED 24168) (T_NAME_QUALIFIED 24169) (T_NAME_RELATIVE 24170) (T_NEW 24171) (T_NS_C 24172) (T_OBJECT_CAST 24173) (T_PRINT 24174) (T_REQUIRE 24175) (T_REQUIRE_ONCE 24176) (T_START_HEREDOC 24177) (T_STATIC 24178) (T_STRING 24179) (T_STRING_CAST 24180) (T_THROW 24181) (T_TRAIT_C 24182) (T_UNSET_CAST 24183) (T_VARIABLE 24184) (T_YIELD 24185) (T_YIELD_FROM 24186) ("[" 24187) ("`" 24188) (array_object_dereferenceable 24189) (attribute 129) (attributes 24190) (callable_expr 24191) (callable_variable 24192) (class_constant 24193) (class_name 24194) (constant 24195) (dereferenceable_scalar 24196) (expr 24593) (fn 24198) (fully_dereferenceable 103) #'24199 (function_call 24200) (inline_function 24201) (internal_functions_in_yacc 24202) (match 24203) (name 24204) (new_expr 24205) (scalar 24206) (simple_variable 24207) (static_member 24208) (variable 24209) (variable_class_name 24210) ("~" 24211)) 11733 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 24595) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11734 (("," 5117) (possible_comma 24596)) 11735 ((")" 24597)) 11736 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 24604) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11737 ((")" 24605)) 11738 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24607) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11739 (("%" 280) (")" 24608) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11740 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24610) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11741 (("%" 280) (")" 24611) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11742 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 24618) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 11743 ((")" 24619)) 11744 (("%" 280) (")" 24624) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11745 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 24627) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11746 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 24628)) 11747 (("\"" 24630) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 11748 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 24640) (T_VARIABLE 2822) (encaps_var 2836)) 11749 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 24639) (T_VARIABLE 2822) (encaps_var 2827)) 11750 (("(" 2899) (argument_list 2900) (ctor_arguments 24659)) 11751 ((T_ATTRIBUTE 2853) (T_CLASS 24645) (anonymous_class 24658) (attribute 2898)) 11752 (("(" 24649) (argument_list 24650) (ctor_arguments 24651)) 11753 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 24652) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 24653) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 24654) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11754 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 24657) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11755 (("," 1449) (possible_comma 24655)) 11756 ((")" 24656)) 11757 ((T_EXTENDS 2902) (extends_from 24660)) 11758 ((T_IMPLEMENTS 2904) (implements_list 24661)) 11759 ((backup_doc_comment 24662)) 11760 (("{" 24663)) 11761 ((class_statement_list 24664)) 11762 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 24665)) 11763 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24667) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11764 (("%" 280) (")" 24668) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11765 (("{" 24669)) 11766 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 24670) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 11767 (("}" 24671)) 11768 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 24673) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 11769 ((")" 24674)) 11770 (("=" 24675)) 11771 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24676) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11772 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 24678) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11773 (("," 5117) (possible_comma 24679)) 11774 ((")" 24680)) 11775 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 24687) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11776 ((")" 24688)) 11777 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24690) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11778 (("%" 280) (")" 24691) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11779 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24693) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11780 (("%" 280) (")" 24694) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11781 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 24701) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 11782 ((")" 24702)) 11783 (("%" 280) (")" 24707) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11784 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 24710) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11785 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 24711)) 11786 (("\"" 24713) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 11787 ((T_STRING 24717) (backup_doc_comment 137)) 11788 ((backup_doc_comment 24718)) 11789 (("(" 24719)) 11790 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 24720)) 11791 ((")" 24721)) 11792 ((":" 150) (return_type 24722)) 11793 ((backup_fn_flags 24723)) 11794 (("{" 24724)) 11795 ((inner_statement_list 24725)) 11796 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 24726) ("~" 124)) 11797 ((backup_fn_flags 24727)) 11798 ((T_STRING 24730)) 11799 ((T_EXTENDS 2902) (extends_from 24731)) 11800 ((T_IMPLEMENTS 2904) (implements_list 24732)) 11801 ((backup_doc_comment 24733)) 11802 (("{" 24734)) 11803 ((class_statement_list 24735)) 11804 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 24736)) 11805 ((":" 24741)) 11806 ((";" 24740)) 11807 ((inner_statement_list 24742)) 11808 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 24743) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 11809 ((";" 24744)) 11810 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24746) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11811 (("%" 280) (")" 24747) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11812 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24748) (";" 155) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 186) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24749) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 24750) ("{" 195) ("~" 124)) 11813 ((inner_statement_list 24751)) 11814 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 24752) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 11815 ((";" 24753)) 11816 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 24755) (variable 16049) (variable_class_name 16050)) 11817 (("," 16062) (possible_comma 24756)) 11818 ((")" 24757)) 11819 ((";" 24758)) 11820 ((inner_statement_list 24760)) 11821 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 24761) ("~" 124)) 11822 ((catch_list 24762)) 11823 ((T_CATCH 24763) (T_FINALLY 24764) (finally_statement 24765)) 11824 (("(" 24769)) 11825 (("{" 24766)) 11826 ((inner_statement_list 24767)) 11827 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 24768) ("~" 124)) 11828 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 24770) (class_name 16111) (name 16112)) 11829 ((T_VARIABLE 16113) (optional_variable 24771) ("|" 16115)) 11830 ((")" 24772)) 11831 (("{" 24773)) 11832 ((inner_statement_list 24774)) 11833 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 24775) ("~" 124)) 11834 ((backup_doc_comment 24777)) 11835 (("{" 24778)) 11836 ((class_statement_list 24779)) 11837 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 24780)) 11838 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24782) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11839 (("%" 280) (")" 24783) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11840 ((":" 24784) (switch_case_list 24785) ("{" 24786)) 11841 ((";" 24792) (case_list 24793)) 11842 ((";" 24787) (case_list 24788)) 11843 ((case_list 24790)) 11844 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 24789)) 11845 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 24791)) 11846 ((case_list 24796)) 11847 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 24794)) 11848 ((";" 24795)) 11849 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 24797)) 11850 ((";" 24798)) 11851 (("," 16412) (";" 24801)) 11852 ((";" 24803)) 11853 ((T_EXTENDS 19717) (interface_extends_list 24805)) 11854 ((backup_doc_comment 24806)) 11855 (("{" 24807)) 11856 ((class_statement_list 24808)) 11857 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 24809)) 11858 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24811) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11859 (("%" 280) (")" 24812) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11860 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19727) (";" 15736) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15737) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15738) (T_DEC 22) (T_DECLARE 15739) (T_DIR 24) (T_DNUMBER 25) (T_DO 15740) (T_DOUBLE_CAST 27) (T_ECHO 15741) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 15742) (T_FOREACH 15743) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 15744) (T_GOTO 15745) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15746) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15747) (T_START_HEREDOC 67) (T_STATIC 15748) (T_STRING 15749) (T_STRING_CAST 70) (T_SWITCH 15750) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 15751) (T_UNSET 15752) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15753) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15754) (alt_if_stmt_without_else 15755) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 15756) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 15757) (if_stmt_without_else 15758) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24813) (static_member 118) (variable 121) (variable_class_name 122) ("{" 15760) ("~" 124)) 11861 ((")" 24815)) 11862 ((";" 24816)) 11863 ((";" 24818)) 11864 (("," 17413) (";" 24820)) 11865 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 24822) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 11866 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 24823) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 11867 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 24824) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 11868 ((")" 24825) (T_DOUBLE_ARROW 24826)) 11869 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24829) (";" 155) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 186) (fn 102) (foreach_statement 24835) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24831) (static_member 118) (variable 121) (variable_class_name 122) ("{" 195) ("~" 124)) 11870 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 24827) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 11871 ((")" 24828)) 11872 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24829) (";" 155) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 186) (fn 102) (foreach_statement 24830) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24831) (static_member 118) (variable 121) (variable_class_name 122) ("{" 195) ("~" 124)) 11873 ((inner_statement_list 24832)) 11874 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 24833) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 11875 ((";" 24834)) 11876 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 24837) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11877 ((";" 24838)) 11878 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 24839) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 11879 ((";" 24840)) 11880 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 24841) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 11881 ((")" 24842)) 11882 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24843) (";" 155) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 186) (fn 102) (for_statement 24844) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24845) (static_member 118) (variable 121) (variable_class_name 122) ("{" 195) ("~" 124)) 11883 ((inner_statement_list 24846)) 11884 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 24847) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 11885 ((";" 24848)) 11886 ((":" 19766) (enum_backing_type 24850)) 11887 ((T_IMPLEMENTS 2904) (implements_list 24851)) 11888 ((backup_doc_comment 24852)) 11889 (("{" 24853)) 11890 ((class_statement_list 24854)) 11891 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 24855)) 11892 (("," 19251) (";" 24857)) 11893 ((T_WHILE 24859)) 11894 (("(" 24860)) 11895 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 24861) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11896 (("%" 280) (")" 24862) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11897 ((";" 24863)) 11898 ((T_STRING 19424) (const_decl 19425) (const_list 24865)) 11899 ((")" 24866) ("," 19428)) 11900 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24867) (";" 155) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 24868) (dereferenceable_scalar 99) (expr 186) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 24869) (static_member 118) (variable 121) (variable_class_name 122) ("{" 195) ("~" 124)) 11901 ((inner_statement_list 24870)) 11902 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 24871) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 11903 ((";" 24872)) 11904 ((";" 24874)) 11905 ((T_EXTENDS 2902) (extends_from 24876)) 11906 ((T_IMPLEMENTS 2904) (implements_list 24877)) 11907 ((backup_doc_comment 24878)) 11908 (("{" 24879)) 11909 ((class_statement_list 24880)) 11910 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 24881)) 11911 ((";" 24883)) 11912 ((T_ARRAY 24924) (T_CALLABLE 24925) (T_NAME_FULLY_QUALIFIED 24926) (T_NAME_QUALIFIED 24927) (T_NAME_RELATIVE 24928) (T_STATIC 24929) (T_STRING 24930) (name 24931) (type 24932) (type_without_static 24933)) 11913 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24922)) 11914 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24909) ("|" 24910)) 11915 (("|" 24898)) 11916 ((T_ARRAY 24899) (T_CALLABLE 24900) (T_NAME_FULLY_QUALIFIED 24901) (T_NAME_QUALIFIED 24902) (T_NAME_RELATIVE 24903) (T_STATIC 24904) (T_STRING 24905) (name 24906) (type 24907) (type_without_static 24908)) 11917 ((T_ARRAY 24912) (T_CALLABLE 24913) (T_NAME_FULLY_QUALIFIED 24914) (T_NAME_QUALIFIED 24915) (T_NAME_RELATIVE 24916) (T_STATIC 24917) (T_STRING 24918) (name 24919) (type 24920) (type_without_static 24921)) 11918 ((T_ARRAY 24899) (T_CALLABLE 24900) (T_NAME_FULLY_QUALIFIED 24901) (T_NAME_QUALIFIED 24902) (T_NAME_RELATIVE 24903) (T_STATIC 24904) (T_STRING 24905) (name 24906) (type 24911) (type_without_static 24908)) 11919 ((T_ARRAY 24912) (T_CALLABLE 24913) (T_NAME_FULLY_QUALIFIED 24914) (T_NAME_QUALIFIED 24915) (T_NAME_RELATIVE 24916) (T_STATIC 24917) (T_STRING 24918) (name 24919) (type 24923) (type_without_static 24921)) 11920 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 24935) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24936) (T_VARIABLE 24937) (ampersand 24938) (lexical_var 24939) (lexical_var_list 24940)) 11921 ((T_VARIABLE 24945)) 11922 (("," 24941) (possible_comma 24942)) 11923 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 24935) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24936) (T_VARIABLE 24937) (ampersand 24938) (lexical_var 24944)) 11924 ((")" 24943)) 11925 ((T_ARRAY 24994) (T_CALLABLE 24995) (T_NAME_FULLY_QUALIFIED 24996) (T_NAME_QUALIFIED 24997) (T_NAME_RELATIVE 24998) (T_STRING 24999) (name 25000) (type_without_static 25001)) 11926 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24992)) 11927 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 24984) (is_reference 24985)) 11928 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24973) ("|" 24974)) 11929 (("|" 24964)) 11930 ((T_ARRAY 24965) (T_CALLABLE 24966) (T_NAME_FULLY_QUALIFIED 24967) (T_NAME_QUALIFIED 24968) (T_NAME_RELATIVE 24969) (T_STRING 24970) (name 24971) (type_without_static 24972)) 11931 ((T_ARRAY 24976) (T_CALLABLE 24977) (T_NAME_FULLY_QUALIFIED 24978) (T_NAME_QUALIFIED 24979) (T_NAME_RELATIVE 24980) (T_STRING 24981) (name 24982) (type_without_static 24983)) 11932 ((T_ARRAY 24965) (T_CALLABLE 24966) (T_NAME_FULLY_QUALIFIED 24967) (T_NAME_QUALIFIED 24968) (T_NAME_RELATIVE 24969) (T_STRING 24970) (name 24971) (type_without_static 24975)) 11933 ((T_ELLIPSIS 24986) (is_variadic 24987)) 11934 ((T_VARIABLE 24988)) 11935 ((backup_doc_comment 24989)) 11936 (("=" 24990)) 11937 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 24991) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11938 ((T_ARRAY 24976) (T_CALLABLE 24977) (T_NAME_FULLY_QUALIFIED 24978) (T_NAME_QUALIFIED 24979) (T_NAME_RELATIVE 24980) (T_STRING 24981) (name 24982) (type_without_static 24993)) 11939 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 25004) (attributes 142) (optional_property_modifiers 144) (parameter 145)) 11940 (("," 6367) (possible_comma 25008)) 11941 (("]" 25009)) 11942 (("," 6367) (possible_comma 25011)) 11943 (("]" 25012)) 11944 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 25014) ("~" 124)) 11945 (("$" 4) (T_ABSTRACT 25016) (T_ARRAY 25017) (T_AS 25018) (T_BREAK 25019) (T_CALLABLE 25020) (T_CASE 25021) (T_CATCH 25022) (T_CLASS 25023) (T_CLASS_C 25024) (T_CLONE 25025) (T_CONST 25026) (T_CONTINUE 25027) (T_DECLARE 25028) (T_DEFAULT 25029) (T_DIR 25030) (T_DO 25031) (T_ECHO 25032) (T_ELSE 25033) (T_ELSEIF 25034) (T_EMPTY 25035) (T_ENDDECLARE 25036) (T_ENDFOR 25037) (T_ENDFOREACH 25038) (T_ENDIF 25039) (T_ENDSWITCH 25040) (T_ENDWHILE 25041) (T_ENUM 25042) (T_EVAL 25043) (T_EXIT 25044) (T_EXTENDS 25045) (T_FILE 25046) (T_FINAL 25047) (T_FINALLY 25048) (T_FN 25049) (T_FOR 25050) (T_FOREACH 25051) (T_FUNCTION 25052) (T_FUNC_C 25053) (T_GLOBAL 25054) (T_GOTO 25055) (T_IF 25056) (T_IMPLEMENTS 25057) (T_INCLUDE 25058) (T_INCLUDE_ONCE 25059) (T_INSTANCEOF 25060) (T_INSTEADOF 25061) (T_INTERFACE 25062) (T_ISSET 25063) (T_LINE 25064) (T_LIST 25065) (T_LOGICAL_AND 25066) (T_LOGICAL_OR 25067) (T_LOGICAL_XOR 25068) (T_MATCH 25069) (T_METHOD_C 25070) (T_NAMESPACE 25071) (T_NEW 25072) (T_NS_C 25073) (T_PRINT 25074) (T_PRIVATE 25075) (T_PROTECTED 25076) (T_PUBLIC 25077) (T_READONLY 25078) (T_REQUIRE 25079) (T_REQUIRE_ONCE 25080) (T_RETURN 25081) (T_STATIC 25082) (T_STRING 25083) (T_SWITCH 25084) (T_THROW 25085) (T_TRAIT 25086) (T_TRAIT_C 25087) (T_TRY 25088) (T_UNSET 25089) (T_USE 25090) (T_VAR 25091) (T_VARIABLE 79) (T_WHILE 25092) (T_YIELD 25093) (identifier 25094) (member_name 25095) (reserved_non_modifiers 25096) (semi_reserved 25097) (simple_variable 25098) ("{" 914)) 11946 (("(" 25099) (argument_list 25100)) 11947 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 25101) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 25102) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 25103) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11948 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 25106) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11949 (("," 1449) (possible_comma 25104)) 11950 ((")" 25105)) 11951 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (ampersand 25136) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25137) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11952 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25135) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11953 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25134) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11954 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25133) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11955 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25132) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11956 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25131) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11957 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25130) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11958 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25129) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11959 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25128) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11960 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25127) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11961 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25126) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11962 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25125) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11963 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25124) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11964 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25123) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11965 (("\"" 1489) ("$" 25138) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 25139) ("[" 1508) (array_object_dereferenceable 25140) (callable_expr 25141) (callable_variable 25142) (class_constant 1512) (class_name 25143) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 25144) (name 25145) (simple_variable 25146) (static_member 25147) (variable 25148) (variable_class_name 25149)) 11966 (("$" 25138) (T_VARIABLE 25139) (simple_variable 25184) ("{" 25185)) 11967 ((T_NULLSAFE_OBJECT_OPERATOR 25167) (T_OBJECT_OPERATOR 25168) ("[" 25169) ("{" 25170)) 11968 (("(" 25153) (argument_list 25166)) 11969 ((T_PAAMAYIM_NEKUDOTAYIM 25162)) 11970 (("(" 25153) (argument_list 25161)) 11971 ((T_PAAMAYIM_NEKUDOTAYIM 25150)) 11972 (("$" 25138) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 25139) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 25151) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 25152) ("{" 914)) 11973 (("(" 25153) (argument_list 25154)) 11974 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 25155) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 25156) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 25157) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11975 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 25160) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 11976 (("," 1449) (possible_comma 25158)) 11977 ((")" 25159)) 11978 (("$" 25138) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 25139) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 25163) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 25164) ("{" 914)) 11979 (("(" 25153) (argument_list 25165)) 11980 (("$" 25138) (T_STRING 25175) (T_VARIABLE 25139) (property_name 25182) (simple_variable 25177) ("{" 25178)) 11981 (("$" 25138) (T_STRING 25175) (T_VARIABLE 25139) (property_name 25176) (simple_variable 25177) ("{" 25178)) 11982 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 25173) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 11983 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25171) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11984 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25172)) 11985 (("]" 25174)) 11986 (("(" 25153) (argument_list 25181)) 11987 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25179) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11988 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25180)) 11989 (("(" 25153) (argument_list 25183)) 11990 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25186) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 11991 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25187)) 11992 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 66) (T_START_HEREDOC 67) (T_STATIC 68) (T_STRING 69) (T_STRING_CAST 70) (T_SWITCH 71) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 75) (T_UNSET 76) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 80) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 85) (alt_if_stmt_without_else 86) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 101) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 107) (if_stmt_without_else 108) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 25338) (static_member 118) (variable 121) (variable_class_name 122) ("{" 123) ("~" 124)) 11993 (("(" 25191)) 11994 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 25192) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 11995 (("%" 280) (")" 25193) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 11996 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 25194) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 25195) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 25196) (T_DEC 22) (T_DECLARE 25197) (T_DIR 24) (T_DNUMBER 25) (T_DO 25198) (T_DOUBLE_CAST 27) (T_ECHO 25199) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 25200) (T_FOREACH 25201) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 25202) (T_GOTO 25203) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 25204) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 25205) (T_START_HEREDOC 67) (T_STATIC 25206) (T_STRING 25207) (T_STRING_CAST 70) (T_SWITCH 25208) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 25209) (T_UNSET 25210) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 25211) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 25212) (alt_if_stmt_without_else 25213) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25214) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 25215) (if_stmt_without_else 25216) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 25217) (static_member 118) (variable 121) (variable_class_name 122) ("{" 25218) ("~" 124)) 11997 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 25336) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11998 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 25334) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 11999 (("(" 25325)) 12000 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 25319) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 12001 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 25317) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 12002 (("(" 25304)) 12003 (("(" 25289)) 12004 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 25287) (simple_variable 17412)) 12005 ((T_STRING 25285)) 12006 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 25283) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 12007 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 25281)) 12008 ((":" 25280)) 12009 (("(" 25262)) 12010 (("{" 25245)) 12011 (("(" 25240)) 12012 (("(" 25231)) 12013 ((T_ELSE 25224) (T_ELSEIF 15767) (T_ENDIF 25225)) 12014 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 25223) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 12015 ((T_ELSE 25221) (T_ELSEIF 25190)) 12016 ((inner_statement_list 25219)) 12017 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 25220) ("~" 124)) 12018 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 25194) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 25195) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 25196) (T_DEC 22) (T_DECLARE 25197) (T_DIR 24) (T_DNUMBER 25) (T_DO 25198) (T_DOUBLE_CAST 27) (T_ECHO 25199) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 25200) (T_FOREACH 25201) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 25202) (T_GOTO 25203) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 25204) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 25205) (T_START_HEREDOC 67) (T_STATIC 25206) (T_STRING 25207) (T_STRING_CAST 70) (T_SWITCH 25208) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 25209) (T_UNSET 25210) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 25211) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 25212) (alt_if_stmt_without_else 25213) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25214) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 25215) (if_stmt_without_else 25216) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 25222) (static_member 118) (variable 121) (variable_class_name 122) ("{" 25218) ("~" 124)) 12019 ((":" 25227)) 12020 ((";" 25226)) 12021 ((inner_statement_list 25228)) 12022 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 25229) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 12023 ((";" 25230)) 12024 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 25232) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12025 (("%" 280) (")" 25233) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12026 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25234) (";" 25194) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 25195) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 25196) (T_DEC 22) (T_DECLARE 25197) (T_DIR 24) (T_DNUMBER 25) (T_DO 25198) (T_DOUBLE_CAST 27) (T_ECHO 25199) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 25200) (T_FOREACH 25201) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 25202) (T_GOTO 25203) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 25204) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 25205) (T_START_HEREDOC 67) (T_STATIC 25206) (T_STRING 25207) (T_STRING_CAST 70) (T_SWITCH 25208) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 25209) (T_UNSET 25210) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 25211) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 25212) (alt_if_stmt_without_else 25213) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25214) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 25215) (if_stmt_without_else 25216) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 25235) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 25236) ("{" 25218) ("~" 124)) 12027 ((inner_statement_list 25237)) 12028 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 25238) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 12029 ((";" 25239)) 12030 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 25241) (variable 16049) (variable_class_name 16050)) 12031 (("," 16062) (possible_comma 25242)) 12032 ((")" 25243)) 12033 ((";" 25244)) 12034 ((inner_statement_list 25246)) 12035 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 25247) ("~" 124)) 12036 ((catch_list 25248)) 12037 ((T_CATCH 25249) (T_FINALLY 25250) (finally_statement 25251)) 12038 (("(" 25255)) 12039 (("{" 25252)) 12040 ((inner_statement_list 25253)) 12041 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 25254) ("~" 124)) 12042 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 25256) (class_name 16111) (name 16112)) 12043 ((T_VARIABLE 16113) (optional_variable 25257) ("|" 16115)) 12044 ((")" 25258)) 12045 (("{" 25259)) 12046 ((inner_statement_list 25260)) 12047 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 25261) ("~" 124)) 12048 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 25263) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12049 (("%" 280) (")" 25264) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12050 ((":" 25265) (switch_case_list 25266) ("{" 25267)) 12051 ((";" 25273) (case_list 25274)) 12052 ((";" 25268) (case_list 25269)) 12053 ((case_list 25271)) 12054 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 25270)) 12055 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 25272)) 12056 ((case_list 25277)) 12057 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 25275)) 12058 ((";" 25276)) 12059 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 25278)) 12060 ((";" 25279)) 12061 (("," 16412) (";" 25282)) 12062 ((";" 25284)) 12063 ((";" 25286)) 12064 (("," 17413) (";" 25288)) 12065 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 25290) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 12066 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 25291) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 12067 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 25292) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 12068 ((")" 25293) (T_DOUBLE_ARROW 25294)) 12069 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25297) (";" 25194) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 25195) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 25196) (T_DEC 22) (T_DECLARE 25197) (T_DIR 24) (T_DNUMBER 25) (T_DO 25198) (T_DOUBLE_CAST 27) (T_ECHO 25199) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 25200) (T_FOREACH 25201) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 25202) (T_GOTO 25203) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 25204) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 25205) (T_START_HEREDOC 67) (T_STATIC 25206) (T_STRING 25207) (T_STRING_CAST 70) (T_SWITCH 25208) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 25209) (T_UNSET 25210) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 25211) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 25212) (alt_if_stmt_without_else 25213) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25214) (fn 102) (foreach_statement 25303) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 25215) (if_stmt_without_else 25216) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 25299) (static_member 118) (variable 121) (variable_class_name 122) ("{" 25218) ("~" 124)) 12070 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 25295) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 12071 ((")" 25296)) 12072 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25297) (";" 25194) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 25195) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 25196) (T_DEC 22) (T_DECLARE 25197) (T_DIR 24) (T_DNUMBER 25) (T_DO 25198) (T_DOUBLE_CAST 27) (T_ECHO 25199) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 25200) (T_FOREACH 25201) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 25202) (T_GOTO 25203) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 25204) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 25205) (T_START_HEREDOC 67) (T_STATIC 25206) (T_STRING 25207) (T_STRING_CAST 70) (T_SWITCH 25208) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 25209) (T_UNSET 25210) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 25211) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 25212) (alt_if_stmt_without_else 25213) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25214) (fn 102) (foreach_statement 25298) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 25215) (if_stmt_without_else 25216) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 25299) (static_member 118) (variable 121) (variable_class_name 122) ("{" 25218) ("~" 124)) 12073 ((inner_statement_list 25300)) 12074 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 25301) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 12075 ((";" 25302)) 12076 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 25305) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 12077 ((";" 25306)) 12078 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 25307) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 12079 ((";" 25308)) 12080 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 25309) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 12081 ((")" 25310)) 12082 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25311) (";" 25194) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 25195) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 25196) (T_DEC 22) (T_DECLARE 25197) (T_DIR 24) (T_DNUMBER 25) (T_DO 25198) (T_DOUBLE_CAST 27) (T_ECHO 25199) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 25200) (T_FOREACH 25201) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 25202) (T_GOTO 25203) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 25204) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 25205) (T_START_HEREDOC 67) (T_STATIC 25206) (T_STRING 25207) (T_STRING_CAST 70) (T_SWITCH 25208) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 25209) (T_UNSET 25210) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 25211) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 25212) (alt_if_stmt_without_else 25213) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25214) (fn 102) (for_statement 25312) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 25215) (if_stmt_without_else 25216) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 25313) (static_member 118) (variable 121) (variable_class_name 122) ("{" 25218) ("~" 124)) 12083 ((inner_statement_list 25314)) 12084 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 25315) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 12085 ((";" 25316)) 12086 (("," 19251) (";" 25318)) 12087 ((T_WHILE 25320)) 12088 (("(" 25321)) 12089 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 25322) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12090 (("%" 280) (")" 25323) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12091 ((";" 25324)) 12092 ((T_STRING 19424) (const_decl 19425) (const_list 25326)) 12093 ((")" 25327) ("," 19428)) 12094 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 25328) (";" 25194) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 25195) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 25196) (T_DEC 22) (T_DECLARE 25197) (T_DIR 24) (T_DNUMBER 25) (T_DO 25198) (T_DOUBLE_CAST 27) (T_ECHO 25199) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 25200) (T_FOREACH 25201) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 25202) (T_GOTO 25203) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 25204) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 25205) (T_START_HEREDOC 67) (T_STATIC 25206) (T_STRING 25207) (T_STRING_CAST 70) (T_SWITCH 25208) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 25209) (T_UNSET 25210) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 25211) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 25212) (alt_if_stmt_without_else 25213) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 25329) (dereferenceable_scalar 99) (expr 25214) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 25215) (if_stmt_without_else 25216) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 25330) (static_member 118) (variable 121) (variable_class_name 122) ("{" 25218) ("~" 124)) 12095 ((inner_statement_list 25331)) 12096 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 25332) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 12097 ((";" 25333)) 12098 ((";" 25335)) 12099 ((";" 25337)) 12100 ((T_STRING 25340) (backup_doc_comment 137)) 12101 ((backup_doc_comment 25341)) 12102 (("(" 25342)) 12103 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 25343)) 12104 ((")" 25344)) 12105 ((":" 150) (return_type 25345)) 12106 ((backup_fn_flags 25346)) 12107 (("{" 25347)) 12108 ((inner_statement_list 25348)) 12109 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 25349) ("~" 124)) 12110 ((backup_fn_flags 25350)) 12111 ((backup_doc_comment 25352)) 12112 (("(" 25353)) 12113 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 25354)) 12114 ((")" 25355)) 12115 ((":" 2207) (return_type 25356)) 12116 ((T_DOUBLE_ARROW 25357)) 12117 ((backup_fn_flags 25358)) 12118 ((backup_lex_pos 25359)) 12119 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25421) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12120 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26399) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12121 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 26397) (encaps_var 5158)) 12122 (("$" 25362) (T_VARIABLE 25408) (simple_variable 26393) ("{" 26394)) 12123 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26391) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12124 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26390) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12125 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26389) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12126 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26388) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12127 (("(" 26385)) 12128 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26384) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12129 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26383) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12130 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26382) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12131 (("\"" 1489) ("$" 25661) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 25662) ("[" 1508) (array_object_dereferenceable 25663) (callable_expr 25664) (callable_variable 25665) (class_constant 1512) (class_name 25666) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 25667) (name 25668) (simple_variable 25669) (static_member 25670) (variable 26381) (variable_class_name 25672)) 12132 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26380) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12133 (("(" 26377)) 12134 (("(" 26374)) 12135 (("(" 26370) (exit_expr 26371)) 12136 (("\"" 1489) ("$" 25661) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 25662) ("[" 1508) (array_object_dereferenceable 25663) (callable_expr 25664) (callable_variable 25665) (class_constant 1512) (class_name 25666) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 25667) (name 25668) (simple_variable 25669) (static_member 25670) (variable 26369) (variable_class_name 25672)) 12137 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26368) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12138 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26367) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12139 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26366) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12140 (("(" 26362)) 12141 (("(" 26357)) 12142 (("(" 26351)) 12143 (("$" 25661) ("(" 26304) (T_ATTRIBUTE 2853) (T_CLASS 26305) (T_NAME_FULLY_QUALIFIED 26306) (T_NAME_QUALIFIED 26307) (T_NAME_RELATIVE 26308) (T_STATIC 26309) (T_STRING 26310) (T_VARIABLE 25662) (anonymous_class 26311) (attribute 2862) (attributes 26312) (class_name 26313) (class_name_reference 26314) (name 26315) (new_variable 26316) (simple_variable 26317)) 12144 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26303) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12145 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26302) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12146 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26301) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12147 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26300) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12148 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 26295) (T_END_HEREDOC 26296) (T_VARIABLE 2822) (encaps_list 26297) (encaps_var 2824)) 12149 ((T_FN 35) (T_FUNCTION 126) (fn 25422) #'25423 (inline_function 26294)) 12150 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26293) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12151 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26292) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12152 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26291) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12153 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25830) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12154 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25768) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12155 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 25764) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 12156 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 25762) (encaps_list 2729) (encaps_var 2730)) 12157 ((T_NULLSAFE_OBJECT_OPERATOR 25745) (T_OBJECT_OPERATOR 25746) ("[" 25747) ("{" 25748)) 12158 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 25742) (attribute 2654) (fn 25422) #'25423 (inline_function 25743)) 12159 (("(" 25622) (argument_list 25741)) 12160 ((T_PAAMAYIM_NEKUDOTAYIM 25736)) 12161 (("%" 25437) ("*" 25438) ("+" 25439) ("-" 25440) ("." 25441) ("/" 25442) ("<" 25443) (">" 25444) ("?" 25445) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25446) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25447) (T_BOOLEAN_AND 25448) (T_BOOLEAN_OR 25449) (T_COALESCE 25450) (T_INSTANCEOF 25451) (T_IS_EQUAL 25452) (T_IS_GREATER_OR_EQUAL 25453) (T_IS_IDENTICAL 25454) (T_IS_NOT_EQUAL 25455) (T_IS_NOT_IDENTICAL 25456) (T_IS_SMALLER_OR_EQUAL 25457) (T_LOGICAL_AND 25458) (T_LOGICAL_OR 25459) (T_LOGICAL_XOR 25460) (T_POW 25461) (T_SL 25462) (T_SPACESHIP 25463) (T_SR 25464) ("^" 25465) (backup_fn_flags 25735) ("|" 25466)) 12162 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 25724)) 12163 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 25712)) 12164 (("(" 25622) (argument_list 25711)) 12165 (("=" 25630) (T_AND_EQUAL 25631) (T_COALESCE_EQUAL 25632) (T_CONCAT_EQUAL 25633) (T_DEC 25634) (T_DIV_EQUAL 25635) (T_INC 25636) (T_MINUS_EQUAL 25637) (T_MOD_EQUAL 25638) (T_MUL_EQUAL 25639) (T_OR_EQUAL 25640) (T_PLUS_EQUAL 25641) (T_POW_EQUAL 25642) (T_SL_EQUAL 25643) (T_SR_EQUAL 25644) (T_XOR_EQUAL 25645)) 12166 ((T_PAAMAYIM_NEKUDOTAYIM 25538)) 12167 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25436) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12168 (("%" 25437) ("*" 25438) ("+" 25439) ("-" 25440) ("." 25441) ("/" 25442) ("<" 25443) (">" 25444) ("?" 25445) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25446) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25447) (T_BOOLEAN_AND 25448) (T_BOOLEAN_OR 25449) (T_COALESCE 25450) (T_INSTANCEOF 25451) (T_IS_EQUAL 25452) (T_IS_GREATER_OR_EQUAL 25453) (T_IS_IDENTICAL 25454) (T_IS_NOT_EQUAL 25455) (T_IS_NOT_IDENTICAL 25456) (T_IS_SMALLER_OR_EQUAL 25457) (T_LOGICAL_AND 25458) (T_LOGICAL_OR 25459) (T_LOGICAL_XOR 25460) (T_POW 25461) (T_SL 25462) (T_SPACESHIP 25463) (T_SR 25464) ("^" 25465) ("|" 25466)) 12169 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25537) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12170 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25536) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12171 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25535) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12172 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25534) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12173 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25533) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12174 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25532) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12175 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25531) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12176 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25530) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12177 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 25525) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 25526) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 12178 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25524) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12179 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25523) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12180 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25522) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12181 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25521) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12182 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25520) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12183 (("$" 25482) ("(" 25483) (T_NAME_FULLY_QUALIFIED 25484) (T_NAME_QUALIFIED 25485) (T_NAME_RELATIVE 25486) (T_STATIC 25487) (T_STRING 25488) (T_VARIABLE 25489) (class_name 25490) (class_name_reference 25491) (name 25492) (new_variable 25493) (simple_variable 25494)) 12184 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25481) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12185 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25480) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12186 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25479) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12187 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25478) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12188 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25477) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12189 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25476) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12190 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25475) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12191 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25474) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12192 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25473) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12193 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25472) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12194 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25471) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12195 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25470) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12196 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25469) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12197 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25468) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12198 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25467) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12199 (("$" 25482) (T_VARIABLE 25489) (simple_variable 25516) ("{" 25517)) 12200 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 25514) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12201 ((T_PAAMAYIM_NEKUDOTAYIM 25512)) 12202 ((T_NULLSAFE_OBJECT_OPERATOR 25495) (T_OBJECT_OPERATOR 25496) (T_PAAMAYIM_NEKUDOTAYIM 25497) ("[" 25498) ("{" 25499)) 12203 (("$" 25482) (T_STRING 25505) (T_VARIABLE 25489) (property_name 25511) (simple_variable 25507) ("{" 25508)) 12204 (("$" 25482) (T_STRING 25505) (T_VARIABLE 25489) (property_name 25506) (simple_variable 25507) ("{" 25508)) 12205 (("$" 25482) (T_VARIABLE 25489) (simple_variable 25504)) 12206 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 25502) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 12207 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25500) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12208 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25501)) 12209 (("]" 25503)) 12210 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25509) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12211 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25510)) 12212 (("$" 25482) (T_VARIABLE 25489) (simple_variable 25513)) 12213 (("%" 280) (")" 25515) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12214 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25518) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12215 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25519)) 12216 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25529) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12217 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 25527) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 12218 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25528) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12219 (("$" 25362) (T_ABSTRACT 25539) (T_ARRAY 25540) (T_AS 25541) (T_BREAK 25542) (T_CALLABLE 25543) (T_CASE 25544) (T_CATCH 25545) (T_CLASS 25546) (T_CLASS_C 25547) (T_CLONE 25548) (T_CONST 25549) (T_CONTINUE 25550) (T_DECLARE 25551) (T_DEFAULT 25552) (T_DIR 25553) (T_DO 25554) (T_ECHO 25555) (T_ELSE 25556) (T_ELSEIF 25557) (T_EMPTY 25558) (T_ENDDECLARE 25559) (T_ENDFOR 25560) (T_ENDFOREACH 25561) (T_ENDIF 25562) (T_ENDSWITCH 25563) (T_ENDWHILE 25564) (T_ENUM 25565) (T_EVAL 25566) (T_EXIT 25567) (T_EXTENDS 25568) (T_FILE 25569) (T_FINAL 25570) (T_FINALLY 25571) (T_FN 25572) (T_FOR 25573) (T_FOREACH 25574) (T_FUNCTION 25575) (T_FUNC_C 25576) (T_GLOBAL 25577) (T_GOTO 25578) (T_IF 25579) (T_IMPLEMENTS 25580) (T_INCLUDE 25581) (T_INCLUDE_ONCE 25582) (T_INSTANCEOF 25583) (T_INSTEADOF 25584) (T_INTERFACE 25585) (T_ISSET 25586) (T_LINE 25587) (T_LIST 25588) (T_LOGICAL_AND 25589) (T_LOGICAL_OR 25590) (T_LOGICAL_XOR 25591) (T_MATCH 25592) (T_METHOD_C 25593) (T_NAMESPACE 25594) (T_NEW 25595) (T_NS_C 25596) (T_PRINT 25597) (T_PRIVATE 25598) (T_PROTECTED 25599) (T_PUBLIC 25600) (T_READONLY 25601) (T_REQUIRE 25602) (T_REQUIRE_ONCE 25603) (T_RETURN 25604) (T_STATIC 25605) (T_STRING 25606) (T_SWITCH 25607) (T_THROW 25608) (T_TRAIT 25609) (T_TRAIT_C 25610) (T_TRY 25611) (T_UNSET 25612) (T_USE 25613) (T_VAR 25614) (T_VARIABLE 25408) (T_WHILE 25615) (T_YIELD 25616) (identifier 25617) (member_name 25618) (reserved_non_modifiers 25619) (semi_reserved 25620) (simple_variable 25621) ("{" 914)) 12220 (("(" 25622) (argument_list 25623)) 12221 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 25624) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 25625) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 25626) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12222 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 25629) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12223 (("," 1449) (possible_comma 25627)) 12224 ((")" 25628)) 12225 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (ampersand 25659) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25660) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12226 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25658) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12227 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25657) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12228 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25656) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12229 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25655) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12230 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25654) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12231 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25653) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12232 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25652) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12233 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25651) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12234 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25650) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12235 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25649) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12236 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25648) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12237 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25647) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12238 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25646) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12239 (("\"" 1489) ("$" 25661) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 25662) ("[" 1508) (array_object_dereferenceable 25663) (callable_expr 25664) (callable_variable 25665) (class_constant 1512) (class_name 25666) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 25667) (name 25668) (simple_variable 25669) (static_member 25670) (variable 25671) (variable_class_name 25672)) 12240 (("$" 25661) (T_VARIABLE 25662) (simple_variable 25707) ("{" 25708)) 12241 ((T_NULLSAFE_OBJECT_OPERATOR 25690) (T_OBJECT_OPERATOR 25691) ("[" 25692) ("{" 25693)) 12242 (("(" 25676) (argument_list 25689)) 12243 ((T_PAAMAYIM_NEKUDOTAYIM 25685)) 12244 (("(" 25676) (argument_list 25684)) 12245 ((T_PAAMAYIM_NEKUDOTAYIM 25673)) 12246 (("$" 25661) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 25662) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 25674) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 25675) ("{" 914)) 12247 (("(" 25676) (argument_list 25677)) 12248 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 25678) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 25679) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 25680) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12249 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 25683) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12250 (("," 1449) (possible_comma 25681)) 12251 ((")" 25682)) 12252 (("$" 25661) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 25662) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 25686) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 25687) ("{" 914)) 12253 (("(" 25676) (argument_list 25688)) 12254 (("$" 25661) (T_STRING 25698) (T_VARIABLE 25662) (property_name 25705) (simple_variable 25700) ("{" 25701)) 12255 (("$" 25661) (T_STRING 25698) (T_VARIABLE 25662) (property_name 25699) (simple_variable 25700) ("{" 25701)) 12256 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 25696) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 12257 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25694) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12258 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25695)) 12259 (("]" 25697)) 12260 (("(" 25676) (argument_list 25704)) 12261 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25702) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12262 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25703)) 12263 (("(" 25676) (argument_list 25706)) 12264 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25709) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12265 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25710)) 12266 ((backup_doc_comment 25713)) 12267 (("(" 25714)) 12268 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 25715)) 12269 ((")" 25716)) 12270 ((T_USE 148) (lexical_vars 25717)) 12271 ((":" 150) (return_type 25718)) 12272 ((backup_fn_flags 25719)) 12273 (("{" 25720)) 12274 ((inner_statement_list 25721)) 12275 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 25722) ("~" 124)) 12276 ((backup_fn_flags 25723)) 12277 ((backup_doc_comment 25725)) 12278 (("(" 25726)) 12279 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 25727)) 12280 ((")" 25728)) 12281 ((":" 2207) (return_type 25729)) 12282 ((T_DOUBLE_ARROW 25730)) 12283 ((backup_fn_flags 25731)) 12284 ((backup_lex_pos 25732)) 12285 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25733) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12286 (("%" 25437) ("*" 25438) ("+" 25439) ("-" 25440) ("." 25441) ("/" 25442) ("<" 25443) (">" 25444) ("?" 25445) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25446) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25447) (T_BOOLEAN_AND 25448) (T_BOOLEAN_OR 25449) (T_COALESCE 25450) (T_INSTANCEOF 25451) (T_IS_EQUAL 25452) (T_IS_GREATER_OR_EQUAL 25453) (T_IS_IDENTICAL 25454) (T_IS_NOT_EQUAL 25455) (T_IS_NOT_IDENTICAL 25456) (T_IS_SMALLER_OR_EQUAL 25457) (T_LOGICAL_AND 25458) (T_LOGICAL_OR 25459) (T_LOGICAL_XOR 25460) (T_POW 25461) (T_SL 25462) (T_SPACESHIP 25463) (T_SR 25464) ("^" 25465) (backup_fn_flags 25734) ("|" 25466)) 12287 (("$" 25362) (T_ABSTRACT 25539) (T_ARRAY 25540) (T_AS 25541) (T_BREAK 25542) (T_CALLABLE 25543) (T_CASE 25544) (T_CATCH 25545) (T_CLASS 25546) (T_CLASS_C 25547) (T_CLONE 25548) (T_CONST 25549) (T_CONTINUE 25550) (T_DECLARE 25551) (T_DEFAULT 25552) (T_DIR 25553) (T_DO 25554) (T_ECHO 25555) (T_ELSE 25556) (T_ELSEIF 25557) (T_EMPTY 25558) (T_ENDDECLARE 25559) (T_ENDFOR 25560) (T_ENDFOREACH 25561) (T_ENDIF 25562) (T_ENDSWITCH 25563) (T_ENDWHILE 25564) (T_ENUM 25565) (T_EVAL 25566) (T_EXIT 25567) (T_EXTENDS 25568) (T_FILE 25569) (T_FINAL 25570) (T_FINALLY 25571) (T_FN 25572) (T_FOR 25573) (T_FOREACH 25574) (T_FUNCTION 25575) (T_FUNC_C 25576) (T_GLOBAL 25577) (T_GOTO 25578) (T_IF 25579) (T_IMPLEMENTS 25580) (T_INCLUDE 25581) (T_INCLUDE_ONCE 25582) (T_INSTANCEOF 25583) (T_INSTEADOF 25584) (T_INTERFACE 25585) (T_ISSET 25586) (T_LINE 25587) (T_LIST 25588) (T_LOGICAL_AND 25589) (T_LOGICAL_OR 25590) (T_LOGICAL_XOR 25591) (T_MATCH 25592) (T_METHOD_C 25593) (T_NAMESPACE 25594) (T_NEW 25595) (T_NS_C 25596) (T_PRINT 25597) (T_PRIVATE 25598) (T_PROTECTED 25599) (T_PUBLIC 25600) (T_READONLY 25601) (T_REQUIRE 25602) (T_REQUIRE_ONCE 25603) (T_RETURN 25604) (T_STATIC 25605) (T_STRING 25606) (T_SWITCH 25607) (T_THROW 25608) (T_TRAIT 25609) (T_TRAIT_C 25610) (T_TRY 25611) (T_UNSET 25612) (T_USE 25613) (T_VAR 25614) (T_VARIABLE 25408) (T_WHILE 25615) (T_YIELD 25616) (identifier 25737) (member_name 25738) (reserved_non_modifiers 25619) (semi_reserved 25620) (simple_variable 25739) ("{" 914)) 12288 (("(" 25622) (argument_list 25740)) 12289 ((T_FN 35) (T_FUNCTION 126) (fn 25422) #'25423 (inline_function 25744)) 12290 (("$" 25362) (T_STRING 25753) (T_VARIABLE 25408) (property_name 25760) (simple_variable 25755) ("{" 25756)) 12291 (("$" 25362) (T_STRING 25753) (T_VARIABLE 25408) (property_name 25754) (simple_variable 25755) ("{" 25756)) 12292 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 25751) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 12293 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25749) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12294 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25750)) 12295 (("]" 25752)) 12296 (("(" 25622) (argument_list 25759)) 12297 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25757) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12298 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25758)) 12299 (("(" 25622) (argument_list 25761)) 12300 (("`" 25763)) 12301 (("]" 25765)) 12302 (("=" 25766)) 12303 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 25767) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12304 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26290) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12305 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 26288) (encaps_var 5158)) 12306 (("$" 25771) (T_VARIABLE 25817) (simple_variable 26284) ("{" 26285)) 12307 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26282) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12308 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26281) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12309 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26280) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12310 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26279) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12311 (("(" 26276)) 12312 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26275) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12313 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26274) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12314 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26273) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12315 (("\"" 1489) ("$" 26070) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 26071) ("[" 1508) (array_object_dereferenceable 26072) (callable_expr 26073) (callable_variable 26074) (class_constant 1512) (class_name 26075) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 26076) (name 26077) (simple_variable 26078) (static_member 26079) (variable 26272) (variable_class_name 26081)) 12316 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26271) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12317 (("(" 26268)) 12318 (("(" 26265)) 12319 (("(" 26261) (exit_expr 26262)) 12320 (("\"" 1489) ("$" 26070) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 26071) ("[" 1508) (array_object_dereferenceable 26072) (callable_expr 26073) (callable_variable 26074) (class_constant 1512) (class_name 26075) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 26076) (name 26077) (simple_variable 26078) (static_member 26079) (variable 26260) (variable_class_name 26081)) 12321 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26259) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12322 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26258) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12323 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26257) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12324 (("(" 26253)) 12325 (("(" 26248)) 12326 (("(" 26242)) 12327 (("$" 26070) ("(" 26195) (T_ATTRIBUTE 2853) (T_CLASS 26196) (T_NAME_FULLY_QUALIFIED 26197) (T_NAME_QUALIFIED 26198) (T_NAME_RELATIVE 26199) (T_STATIC 26200) (T_STRING 26201) (T_VARIABLE 26071) (anonymous_class 26202) (attribute 2862) (attributes 26203) (class_name 26204) (class_name_reference 26205) (name 26206) (new_variable 26207) (simple_variable 26208)) 12328 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26194) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12329 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26193) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12330 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26192) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12331 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26191) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12332 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 26186) (T_END_HEREDOC 26187) (T_VARIABLE 2822) (encaps_list 26188) (encaps_var 2824)) 12333 ((T_FN 35) (T_FUNCTION 126) (fn 25831) #'25832 (inline_function 26185)) 12334 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26184) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12335 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26183) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12336 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26182) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12337 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26179) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12338 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26178) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12339 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 26174) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 12340 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 26172) (encaps_list 2729) (encaps_var 2730)) 12341 ((T_NULLSAFE_OBJECT_OPERATOR 26155) (T_OBJECT_OPERATOR 26156) ("[" 26157) ("{" 26158)) 12342 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 26152) (attribute 2654) (fn 25831) #'25832 (inline_function 26153)) 12343 (("(" 26031) (argument_list 26151)) 12344 ((T_PAAMAYIM_NEKUDOTAYIM 26146)) 12345 (("%" 25846) ("*" 25847) ("+" 25848) ("-" 25849) ("." 25850) ("/" 25851) ("<" 25852) (">" 25853) ("?" 25854) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25855) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25856) (T_BOOLEAN_AND 25857) (T_BOOLEAN_OR 25858) (T_COALESCE 25859) (T_DOUBLE_ARROW 26144) (T_INSTANCEOF 25860) (T_IS_EQUAL 25861) (T_IS_GREATER_OR_EQUAL 25862) (T_IS_IDENTICAL 25863) (T_IS_NOT_EQUAL 25864) (T_IS_NOT_IDENTICAL 25865) (T_IS_SMALLER_OR_EQUAL 25866) (T_LOGICAL_AND 25867) (T_LOGICAL_OR 25868) (T_LOGICAL_XOR 25869) (T_POW 25870) (T_SL 25871) (T_SPACESHIP 25872) (T_SR 25873) ("^" 25874) ("|" 25875)) 12346 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 26133)) 12347 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 26121)) 12348 (("(" 26031) (argument_list 26120)) 12349 (("=" 26039) (T_AND_EQUAL 26040) (T_COALESCE_EQUAL 26041) (T_CONCAT_EQUAL 26042) (T_DEC 26043) (T_DIV_EQUAL 26044) (T_INC 26045) (T_MINUS_EQUAL 26046) (T_MOD_EQUAL 26047) (T_MUL_EQUAL 26048) (T_OR_EQUAL 26049) (T_PLUS_EQUAL 26050) (T_POW_EQUAL 26051) (T_SL_EQUAL 26052) (T_SR_EQUAL 26053) (T_XOR_EQUAL 26054)) 12350 ((T_PAAMAYIM_NEKUDOTAYIM 25947)) 12351 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25845) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12352 (("%" 25846) ("*" 25847) ("+" 25848) ("-" 25849) ("." 25850) ("/" 25851) ("<" 25852) (">" 25853) ("?" 25854) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25855) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25856) (T_BOOLEAN_AND 25857) (T_BOOLEAN_OR 25858) (T_COALESCE 25859) (T_INSTANCEOF 25860) (T_IS_EQUAL 25861) (T_IS_GREATER_OR_EQUAL 25862) (T_IS_IDENTICAL 25863) (T_IS_NOT_EQUAL 25864) (T_IS_NOT_IDENTICAL 25865) (T_IS_SMALLER_OR_EQUAL 25866) (T_LOGICAL_AND 25867) (T_LOGICAL_OR 25868) (T_LOGICAL_XOR 25869) (T_POW 25870) (T_SL 25871) (T_SPACESHIP 25872) (T_SR 25873) ("^" 25874) ("|" 25875)) 12353 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25946) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12354 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25945) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12355 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25944) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12356 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25943) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12357 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25942) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12358 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25941) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12359 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25940) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12360 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25939) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12361 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 25934) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 25935) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 12362 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25933) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12363 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25932) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12364 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25931) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12365 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25930) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12366 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25929) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12367 (("$" 25891) ("(" 25892) (T_NAME_FULLY_QUALIFIED 25893) (T_NAME_QUALIFIED 25894) (T_NAME_RELATIVE 25895) (T_STATIC 25896) (T_STRING 25897) (T_VARIABLE 25898) (class_name 25899) (class_name_reference 25900) (name 25901) (new_variable 25902) (simple_variable 25903)) 12368 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25890) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12369 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25889) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12370 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25888) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12371 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25887) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12372 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25886) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12373 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25885) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12374 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25884) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12375 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25883) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12376 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25882) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12377 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25881) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12378 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25880) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12379 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25879) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12380 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25878) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12381 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25877) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12382 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25876) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12383 (("$" 25891) (T_VARIABLE 25898) (simple_variable 25925) ("{" 25926)) 12384 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 25923) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12385 ((T_PAAMAYIM_NEKUDOTAYIM 25921)) 12386 ((T_NULLSAFE_OBJECT_OPERATOR 25904) (T_OBJECT_OPERATOR 25905) (T_PAAMAYIM_NEKUDOTAYIM 25906) ("[" 25907) ("{" 25908)) 12387 (("$" 25891) (T_STRING 25914) (T_VARIABLE 25898) (property_name 25920) (simple_variable 25916) ("{" 25917)) 12388 (("$" 25891) (T_STRING 25914) (T_VARIABLE 25898) (property_name 25915) (simple_variable 25916) ("{" 25917)) 12389 (("$" 25891) (T_VARIABLE 25898) (simple_variable 25913)) 12390 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 25911) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 12391 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25909) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12392 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25910)) 12393 (("]" 25912)) 12394 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25918) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12395 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25919)) 12396 (("$" 25891) (T_VARIABLE 25898) (simple_variable 25922)) 12397 (("%" 280) (")" 25924) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12398 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 25927) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12399 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 25928)) 12400 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25938) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12401 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 25936) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 12402 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 25937) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12403 (("$" 25771) (T_ABSTRACT 25948) (T_ARRAY 25949) (T_AS 25950) (T_BREAK 25951) (T_CALLABLE 25952) (T_CASE 25953) (T_CATCH 25954) (T_CLASS 25955) (T_CLASS_C 25956) (T_CLONE 25957) (T_CONST 25958) (T_CONTINUE 25959) (T_DECLARE 25960) (T_DEFAULT 25961) (T_DIR 25962) (T_DO 25963) (T_ECHO 25964) (T_ELSE 25965) (T_ELSEIF 25966) (T_EMPTY 25967) (T_ENDDECLARE 25968) (T_ENDFOR 25969) (T_ENDFOREACH 25970) (T_ENDIF 25971) (T_ENDSWITCH 25972) (T_ENDWHILE 25973) (T_ENUM 25974) (T_EVAL 25975) (T_EXIT 25976) (T_EXTENDS 25977) (T_FILE 25978) (T_FINAL 25979) (T_FINALLY 25980) (T_FN 25981) (T_FOR 25982) (T_FOREACH 25983) (T_FUNCTION 25984) (T_FUNC_C 25985) (T_GLOBAL 25986) (T_GOTO 25987) (T_IF 25988) (T_IMPLEMENTS 25989) (T_INCLUDE 25990) (T_INCLUDE_ONCE 25991) (T_INSTANCEOF 25992) (T_INSTEADOF 25993) (T_INTERFACE 25994) (T_ISSET 25995) (T_LINE 25996) (T_LIST 25997) (T_LOGICAL_AND 25998) (T_LOGICAL_OR 25999) (T_LOGICAL_XOR 26000) (T_MATCH 26001) (T_METHOD_C 26002) (T_NAMESPACE 26003) (T_NEW 26004) (T_NS_C 26005) (T_PRINT 26006) (T_PRIVATE 26007) (T_PROTECTED 26008) (T_PUBLIC 26009) (T_READONLY 26010) (T_REQUIRE 26011) (T_REQUIRE_ONCE 26012) (T_RETURN 26013) (T_STATIC 26014) (T_STRING 26015) (T_SWITCH 26016) (T_THROW 26017) (T_TRAIT 26018) (T_TRAIT_C 26019) (T_TRY 26020) (T_UNSET 26021) (T_USE 26022) (T_VAR 26023) (T_VARIABLE 25817) (T_WHILE 26024) (T_YIELD 26025) (identifier 26026) (member_name 26027) (reserved_non_modifiers 26028) (semi_reserved 26029) (simple_variable 26030) ("{" 914)) 12404 (("(" 26031) (argument_list 26032)) 12405 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26033) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 26034) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 26035) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12406 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26038) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12407 (("," 1449) (possible_comma 26036)) 12408 ((")" 26037)) 12409 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (ampersand 26068) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26069) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12410 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26067) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12411 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26066) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12412 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26065) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12413 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26064) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12414 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26063) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12415 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26062) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12416 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26061) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12417 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26060) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12418 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26059) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12419 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26058) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12420 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26057) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12421 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26056) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12422 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26055) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12423 (("\"" 1489) ("$" 26070) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 26071) ("[" 1508) (array_object_dereferenceable 26072) (callable_expr 26073) (callable_variable 26074) (class_constant 1512) (class_name 26075) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 26076) (name 26077) (simple_variable 26078) (static_member 26079) (variable 26080) (variable_class_name 26081)) 12424 (("$" 26070) (T_VARIABLE 26071) (simple_variable 26116) ("{" 26117)) 12425 ((T_NULLSAFE_OBJECT_OPERATOR 26099) (T_OBJECT_OPERATOR 26100) ("[" 26101) ("{" 26102)) 12426 (("(" 26085) (argument_list 26098)) 12427 ((T_PAAMAYIM_NEKUDOTAYIM 26094)) 12428 (("(" 26085) (argument_list 26093)) 12429 ((T_PAAMAYIM_NEKUDOTAYIM 26082)) 12430 (("$" 26070) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 26071) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 26083) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 26084) ("{" 914)) 12431 (("(" 26085) (argument_list 26086)) 12432 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26087) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 26088) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 26089) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12433 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26092) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12434 (("," 1449) (possible_comma 26090)) 12435 ((")" 26091)) 12436 (("$" 26070) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 26071) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 26095) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 26096) ("{" 914)) 12437 (("(" 26085) (argument_list 26097)) 12438 (("$" 26070) (T_STRING 26107) (T_VARIABLE 26071) (property_name 26114) (simple_variable 26109) ("{" 26110)) 12439 (("$" 26070) (T_STRING 26107) (T_VARIABLE 26071) (property_name 26108) (simple_variable 26109) ("{" 26110)) 12440 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 26105) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 12441 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26103) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12442 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26104)) 12443 (("]" 26106)) 12444 (("(" 26085) (argument_list 26113)) 12445 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26111) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12446 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26112)) 12447 (("(" 26085) (argument_list 26115)) 12448 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26118) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12449 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26119)) 12450 ((backup_doc_comment 26122)) 12451 (("(" 26123)) 12452 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 26124)) 12453 ((")" 26125)) 12454 ((T_USE 148) (lexical_vars 26126)) 12455 ((":" 150) (return_type 26127)) 12456 ((backup_fn_flags 26128)) 12457 (("{" 26129)) 12458 ((inner_statement_list 26130)) 12459 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 26131) ("~" 124)) 12460 ((backup_fn_flags 26132)) 12461 ((backup_doc_comment 26134)) 12462 (("(" 26135)) 12463 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 26136)) 12464 ((")" 26137)) 12465 ((":" 2207) (return_type 26138)) 12466 ((T_DOUBLE_ARROW 26139)) 12467 ((backup_fn_flags 26140)) 12468 ((backup_lex_pos 26141)) 12469 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26142) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12470 (("%" 25846) ("*" 25847) ("+" 25848) ("-" 25849) ("." 25850) ("/" 25851) ("<" 25852) (">" 25853) ("?" 25854) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25855) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25856) (T_BOOLEAN_AND 25857) (T_BOOLEAN_OR 25858) (T_COALESCE 25859) (T_INSTANCEOF 25860) (T_IS_EQUAL 25861) (T_IS_GREATER_OR_EQUAL 25862) (T_IS_IDENTICAL 25863) (T_IS_NOT_EQUAL 25864) (T_IS_NOT_IDENTICAL 25865) (T_IS_SMALLER_OR_EQUAL 25866) (T_LOGICAL_AND 25867) (T_LOGICAL_OR 25868) (T_LOGICAL_XOR 25869) (T_POW 25870) (T_SL 25871) (T_SPACESHIP 25872) (T_SR 25873) ("^" 25874) (backup_fn_flags 26143) ("|" 25875)) 12471 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26145) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12472 (("$" 25771) (T_ABSTRACT 25948) (T_ARRAY 25949) (T_AS 25950) (T_BREAK 25951) (T_CALLABLE 25952) (T_CASE 25953) (T_CATCH 25954) (T_CLASS 25955) (T_CLASS_C 25956) (T_CLONE 25957) (T_CONST 25958) (T_CONTINUE 25959) (T_DECLARE 25960) (T_DEFAULT 25961) (T_DIR 25962) (T_DO 25963) (T_ECHO 25964) (T_ELSE 25965) (T_ELSEIF 25966) (T_EMPTY 25967) (T_ENDDECLARE 25968) (T_ENDFOR 25969) (T_ENDFOREACH 25970) (T_ENDIF 25971) (T_ENDSWITCH 25972) (T_ENDWHILE 25973) (T_ENUM 25974) (T_EVAL 25975) (T_EXIT 25976) (T_EXTENDS 25977) (T_FILE 25978) (T_FINAL 25979) (T_FINALLY 25980) (T_FN 25981) (T_FOR 25982) (T_FOREACH 25983) (T_FUNCTION 25984) (T_FUNC_C 25985) (T_GLOBAL 25986) (T_GOTO 25987) (T_IF 25988) (T_IMPLEMENTS 25989) (T_INCLUDE 25990) (T_INCLUDE_ONCE 25991) (T_INSTANCEOF 25992) (T_INSTEADOF 25993) (T_INTERFACE 25994) (T_ISSET 25995) (T_LINE 25996) (T_LIST 25997) (T_LOGICAL_AND 25998) (T_LOGICAL_OR 25999) (T_LOGICAL_XOR 26000) (T_MATCH 26001) (T_METHOD_C 26002) (T_NAMESPACE 26003) (T_NEW 26004) (T_NS_C 26005) (T_PRINT 26006) (T_PRIVATE 26007) (T_PROTECTED 26008) (T_PUBLIC 26009) (T_READONLY 26010) (T_REQUIRE 26011) (T_REQUIRE_ONCE 26012) (T_RETURN 26013) (T_STATIC 26014) (T_STRING 26015) (T_SWITCH 26016) (T_THROW 26017) (T_TRAIT 26018) (T_TRAIT_C 26019) (T_TRY 26020) (T_UNSET 26021) (T_USE 26022) (T_VAR 26023) (T_VARIABLE 25817) (T_WHILE 26024) (T_YIELD 26025) (identifier 26147) (member_name 26148) (reserved_non_modifiers 26028) (semi_reserved 26029) (simple_variable 26149) ("{" 914)) 12473 (("(" 26031) (argument_list 26150)) 12474 ((T_FN 35) (T_FUNCTION 126) (fn 25831) #'25832 (inline_function 26154)) 12475 (("$" 25771) (T_STRING 26163) (T_VARIABLE 25817) (property_name 26170) (simple_variable 26165) ("{" 26166)) 12476 (("$" 25771) (T_STRING 26163) (T_VARIABLE 25817) (property_name 26164) (simple_variable 26165) ("{" 26166)) 12477 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 26161) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 12478 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26159) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12479 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26160)) 12480 (("]" 26162)) 12481 (("(" 26031) (argument_list 26169)) 12482 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26167) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12483 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26168)) 12484 (("(" 26031) (argument_list 26171)) 12485 (("`" 26173)) 12486 (("]" 26175)) 12487 (("=" 26176)) 12488 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26177) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12489 (("%" 25846) ("*" 25847) ("+" 25848) ("-" 25849) ("." 25850) ("/" 25851) ("<" 25852) (">" 25853) ("?" 25854) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25855) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25856) (T_BOOLEAN_AND 25857) (T_BOOLEAN_OR 25858) (T_COALESCE 25859) (T_DOUBLE_ARROW 26180) (T_INSTANCEOF 25860) (T_IS_EQUAL 25861) (T_IS_GREATER_OR_EQUAL 25862) (T_IS_IDENTICAL 25863) (T_IS_NOT_EQUAL 25864) (T_IS_NOT_IDENTICAL 25865) (T_IS_SMALLER_OR_EQUAL 25866) (T_LOGICAL_AND 25867) (T_LOGICAL_OR 25868) (T_LOGICAL_XOR 25869) (T_POW 25870) (T_SL 25871) (T_SPACESHIP 25872) (T_SR 25873) ("^" 25874) ("|" 25875)) 12490 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26181) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12491 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 26190) (T_VARIABLE 2822) (encaps_var 2836)) 12492 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 26189) (T_VARIABLE 2822) (encaps_var 2827)) 12493 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26240) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12494 (("(" 2899) (argument_list 2900) (ctor_arguments 26233)) 12495 ((T_ATTRIBUTE 2853) (T_CLASS 26196) (anonymous_class 26232) (attribute 2898)) 12496 ((T_PAAMAYIM_NEKUDOTAYIM 26230)) 12497 (("(" 26221) (argument_list 26222) (ctor_arguments 26223)) 12498 ((T_NULLSAFE_OBJECT_OPERATOR 26209) (T_OBJECT_OPERATOR 26210) (T_PAAMAYIM_NEKUDOTAYIM 26211) ("[" 26212) ("{" 26213)) 12499 (("$" 26070) (T_STRING 26107) (T_VARIABLE 26071) (property_name 26220) (simple_variable 26109) ("{" 26110)) 12500 (("$" 26070) (T_STRING 26107) (T_VARIABLE 26071) (property_name 26219) (simple_variable 26109) ("{" 26110)) 12501 (("$" 26070) (T_VARIABLE 26071) (simple_variable 26218)) 12502 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 26216) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 12503 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26214) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12504 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26215)) 12505 (("]" 26217)) 12506 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26224) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 26225) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 26226) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12507 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26229) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12508 (("," 1449) (possible_comma 26227)) 12509 ((")" 26228)) 12510 (("$" 26070) (T_VARIABLE 26071) (simple_variable 26231)) 12511 ((T_EXTENDS 2902) (extends_from 26234)) 12512 ((T_IMPLEMENTS 2904) (implements_list 26235)) 12513 ((backup_doc_comment 26236)) 12514 (("{" 26237)) 12515 ((class_statement_list 26238)) 12516 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 26239)) 12517 (("%" 280) (")" 26241) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12518 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26243) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12519 (("%" 280) (")" 26244) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12520 (("{" 26245)) 12521 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 26246) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 12522 (("}" 26247)) 12523 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 26249) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 12524 ((")" 26250)) 12525 (("=" 26251)) 12526 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26252) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12527 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 26254) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12528 (("," 5117) (possible_comma 26255)) 12529 ((")" 26256)) 12530 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 26263) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12531 ((")" 26264)) 12532 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26266) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12533 (("%" 280) (")" 26267) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12534 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26269) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12535 (("%" 280) (")" 26270) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12536 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 26277) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 12537 ((")" 26278)) 12538 (("%" 280) (")" 26283) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12539 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26286) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12540 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26287)) 12541 (("\"" 26289) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 12542 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 26299) (T_VARIABLE 2822) (encaps_var 2836)) 12543 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 26298) (T_VARIABLE 2822) (encaps_var 2827)) 12544 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26349) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12545 (("(" 2899) (argument_list 2900) (ctor_arguments 26342)) 12546 ((T_ATTRIBUTE 2853) (T_CLASS 26305) (anonymous_class 26341) (attribute 2898)) 12547 ((T_PAAMAYIM_NEKUDOTAYIM 26339)) 12548 (("(" 26330) (argument_list 26331) (ctor_arguments 26332)) 12549 ((T_NULLSAFE_OBJECT_OPERATOR 26318) (T_OBJECT_OPERATOR 26319) (T_PAAMAYIM_NEKUDOTAYIM 26320) ("[" 26321) ("{" 26322)) 12550 (("$" 25661) (T_STRING 25698) (T_VARIABLE 25662) (property_name 26329) (simple_variable 25700) ("{" 25701)) 12551 (("$" 25661) (T_STRING 25698) (T_VARIABLE 25662) (property_name 26328) (simple_variable 25700) ("{" 25701)) 12552 (("$" 25661) (T_VARIABLE 25662) (simple_variable 26327)) 12553 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 26325) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 12554 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26323) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12555 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26324)) 12556 (("]" 26326)) 12557 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26333) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 26334) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 26335) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12558 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26338) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12559 (("," 1449) (possible_comma 26336)) 12560 ((")" 26337)) 12561 (("$" 25661) (T_VARIABLE 25662) (simple_variable 26340)) 12562 ((T_EXTENDS 2902) (extends_from 26343)) 12563 ((T_IMPLEMENTS 2904) (implements_list 26344)) 12564 ((backup_doc_comment 26345)) 12565 (("{" 26346)) 12566 ((class_statement_list 26347)) 12567 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 26348)) 12568 (("%" 280) (")" 26350) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12569 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26352) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12570 (("%" 280) (")" 26353) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12571 (("{" 26354)) 12572 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 26355) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 12573 (("}" 26356)) 12574 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 26358) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 12575 ((")" 26359)) 12576 (("=" 26360)) 12577 (("!" 25360) ("\"" 25361) ("$" 25362) ("(" 25363) ("+" 25364) ("-" 25365) ("@" 25366) (T_ARRAY 25367) (T_ARRAY_CAST 25368) (T_ATTRIBUTE 125) (T_BOOL_CAST 25369) (T_CLASS_C 25370) (T_CLONE 25371) (T_CONSTANT_ENCAPSED_STRING 25372) (T_DEC 25373) (T_DIR 25374) (T_DNUMBER 25375) (T_DOUBLE_CAST 25376) (T_EMPTY 25377) (T_EVAL 25378) (T_EXIT 25379) (T_FILE 25380) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25381) (T_INC 25382) (T_INCLUDE 25383) (T_INCLUDE_ONCE 25384) (T_INT_CAST 25385) (T_ISSET 25386) (T_LINE 25387) (T_LIST 25388) (T_LNUMBER 25389) (T_MATCH 25390) (T_METHOD_C 25391) (T_NAME_FULLY_QUALIFIED 25392) (T_NAME_QUALIFIED 25393) (T_NAME_RELATIVE 25394) (T_NEW 25395) (T_NS_C 25396) (T_OBJECT_CAST 25397) (T_PRINT 25398) (T_REQUIRE 25399) (T_REQUIRE_ONCE 25400) (T_START_HEREDOC 25401) (T_STATIC 25402) (T_STRING 25403) (T_STRING_CAST 25404) (T_THROW 25405) (T_TRAIT_C 25406) (T_UNSET_CAST 25407) (T_VARIABLE 25408) (T_YIELD 25409) (T_YIELD_FROM 25410) ("[" 25411) ("`" 25412) (array_object_dereferenceable 25413) (attribute 129) (attributes 25414) (callable_expr 25415) (callable_variable 25416) (class_constant 25417) (class_name 25418) (constant 25419) (dereferenceable_scalar 25420) (expr 26361) (fn 25422) (fully_dereferenceable 103) #'25423 (function_call 25424) (inline_function 25425) (internal_functions_in_yacc 25426) (match 25427) (name 25428) (new_expr 25429) (scalar 25430) (simple_variable 25431) (static_member 25432) (variable 25433) (variable_class_name 25434) ("~" 25435)) 12578 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 26363) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12579 (("," 5117) (possible_comma 26364)) 12580 ((")" 26365)) 12581 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 26372) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12582 ((")" 26373)) 12583 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26375) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12584 (("%" 280) (")" 26376) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12585 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26378) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12586 (("%" 280) (")" 26379) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12587 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 26386) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 12588 ((")" 26387)) 12589 (("%" 280) (")" 26392) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12590 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26395) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12591 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26396)) 12592 (("\"" 26398) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 12593 (("$" 4) (T_ABSTRACT 25016) (T_ARRAY 25017) (T_AS 25018) (T_BREAK 25019) (T_CALLABLE 25020) (T_CASE 25021) (T_CATCH 25022) (T_CLASS 25023) (T_CLASS_C 25024) (T_CLONE 25025) (T_CONST 25026) (T_CONTINUE 25027) (T_DECLARE 25028) (T_DEFAULT 25029) (T_DIR 25030) (T_DO 25031) (T_ECHO 25032) (T_ELSE 25033) (T_ELSEIF 25034) (T_EMPTY 25035) (T_ENDDECLARE 25036) (T_ENDFOR 25037) (T_ENDFOREACH 25038) (T_ENDIF 25039) (T_ENDSWITCH 25040) (T_ENDWHILE 25041) (T_ENUM 25042) (T_EVAL 25043) (T_EXIT 25044) (T_EXTENDS 25045) (T_FILE 25046) (T_FINAL 25047) (T_FINALLY 25048) (T_FN 25049) (T_FOR 25050) (T_FOREACH 25051) (T_FUNCTION 25052) (T_FUNC_C 25053) (T_GLOBAL 25054) (T_GOTO 25055) (T_IF 25056) (T_IMPLEMENTS 25057) (T_INCLUDE 25058) (T_INCLUDE_ONCE 25059) (T_INSTANCEOF 25060) (T_INSTEADOF 25061) (T_INTERFACE 25062) (T_ISSET 25063) (T_LINE 25064) (T_LIST 25065) (T_LOGICAL_AND 25066) (T_LOGICAL_OR 25067) (T_LOGICAL_XOR 25068) (T_MATCH 25069) (T_METHOD_C 25070) (T_NAMESPACE 25071) (T_NEW 25072) (T_NS_C 25073) (T_PRINT 25074) (T_PRIVATE 25075) (T_PROTECTED 25076) (T_PUBLIC 25077) (T_READONLY 25078) (T_REQUIRE 25079) (T_REQUIRE_ONCE 25080) (T_RETURN 25081) (T_STATIC 25082) (T_STRING 25083) (T_SWITCH 25084) (T_THROW 25085) (T_TRAIT 25086) (T_TRAIT_C 25087) (T_TRY 25088) (T_UNSET 25089) (T_USE 25090) (T_VAR 25091) (T_VARIABLE 79) (T_WHILE 25092) (T_YIELD 25093) (identifier 26402) (member_name 26403) (reserved_non_modifiers 25096) (semi_reserved 25097) (simple_variable 26404) ("{" 914)) 12594 (("(" 25099) (argument_list 26405)) 12595 ((T_STRING 26407)) 12596 ((T_EXTENDS 2902) (extends_from 26408)) 12597 ((T_IMPLEMENTS 2904) (implements_list 26409)) 12598 ((backup_doc_comment 26410)) 12599 (("{" 26411)) 12600 ((class_statement_list 26412)) 12601 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 26413)) 12602 ((T_STRING 26465)) 12603 ((T_STRING 26458)) 12604 ((T_STRING 26452)) 12605 ((T_STRING 26447)) 12606 ((T_ABSTRACT 10) (T_CLASS 26439) (T_FINAL 34) (class_modifier 15842)) 12607 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15826) (ampersand 15827) (returns_ref 26427)) 12608 ((T_STRING 26428) (backup_doc_comment 137)) 12609 ((backup_doc_comment 26429)) 12610 (("(" 26430)) 12611 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 26431)) 12612 ((")" 26432)) 12613 ((":" 150) (return_type 26433)) 12614 ((backup_fn_flags 26434)) 12615 (("{" 26435)) 12616 ((inner_statement_list 26436)) 12617 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 26437) ("~" 124)) 12618 ((backup_fn_flags 26438)) 12619 ((T_STRING 26440)) 12620 ((T_EXTENDS 2902) (extends_from 26441)) 12621 ((T_IMPLEMENTS 2904) (implements_list 26442)) 12622 ((backup_doc_comment 26443)) 12623 (("{" 26444)) 12624 ((class_statement_list 26445)) 12625 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 26446)) 12626 ((backup_doc_comment 26448)) 12627 (("{" 26449)) 12628 ((class_statement_list 26450)) 12629 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 26451)) 12630 ((T_EXTENDS 19717) (interface_extends_list 26453)) 12631 ((backup_doc_comment 26454)) 12632 (("{" 26455)) 12633 ((class_statement_list 26456)) 12634 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 26457)) 12635 ((":" 19766) (enum_backing_type 26459)) 12636 ((T_IMPLEMENTS 2904) (implements_list 26460)) 12637 ((backup_doc_comment 26461)) 12638 (("{" 26462)) 12639 ((class_statement_list 26463)) 12640 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 26464)) 12641 ((T_EXTENDS 2902) (extends_from 26466)) 12642 ((T_IMPLEMENTS 2904) (implements_list 26467)) 12643 ((backup_doc_comment 26468)) 12644 (("{" 26469)) 12645 ((class_statement_list 26470)) 12646 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 26471)) 12647 (("$" 4) (T_STRING 26480) (T_VARIABLE 79) (property_name 26487) (simple_variable 26482) ("{" 26483)) 12648 (("$" 4) (T_STRING 26480) (T_VARIABLE 79) (property_name 26481) (simple_variable 26482) ("{" 26483)) 12649 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 26478) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 12650 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26476) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12651 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26477)) 12652 (("]" 26479)) 12653 (("(" 25099) (argument_list 26486)) 12654 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26484) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12655 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26485)) 12656 (("(" 25099) (argument_list 26488)) 12657 ((":" 26492)) 12658 ((";" 26491)) 12659 ((inner_statement_list 26493)) 12660 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 26494) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 12661 ((";" 26495)) 12662 (("`" 26497)) 12663 (("]" 26499)) 12664 (("=" 26500)) 12665 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 26501) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 12666 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26998) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12667 ((T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5155) (T_VARIABLE 5156) (encaps_list 26996) (encaps_var 5158)) 12668 (("$" 26505) (T_VARIABLE 26551) (simple_variable 26992) ("{" 26993)) 12669 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26990) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12670 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26989) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12671 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26988) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12672 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26987) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12673 (("(" 26984)) 12674 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26983) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12675 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26982) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12676 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26981) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12677 (("\"" 1489) ("$" 26804) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 26805) ("[" 1508) (array_object_dereferenceable 26806) (callable_expr 26807) (callable_variable 26808) (class_constant 1512) (class_name 26809) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 26810) (name 26811) (simple_variable 26812) (static_member 26813) (variable 26980) (variable_class_name 26815)) 12678 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26979) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12679 (("(" 26976)) 12680 (("(" 26973)) 12681 (("(" 26969) (exit_expr 26970)) 12682 (("\"" 1489) ("$" 26804) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 26805) ("[" 1508) (array_object_dereferenceable 26806) (callable_expr 26807) (callable_variable 26808) (class_constant 1512) (class_name 26809) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 26810) (name 26811) (simple_variable 26812) (static_member 26813) (variable 26968) (variable_class_name 26815)) 12683 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26967) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12684 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26966) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12685 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26965) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12686 (("(" 26961)) 12687 (("(" 26956)) 12688 (("(" 26950)) 12689 (("$" 26070) ("(" 26195) (T_ATTRIBUTE 2853) (T_CLASS 26929) (T_NAME_FULLY_QUALIFIED 26197) (T_NAME_QUALIFIED 26198) (T_NAME_RELATIVE 26199) (T_STATIC 26200) (T_STRING 26201) (T_VARIABLE 26071) (anonymous_class 26930) (attribute 2862) (attributes 26931) (class_name 26204) (class_name_reference 26932) (name 26206) (new_variable 26207) (simple_variable 26208)) 12690 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26928) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12691 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26927) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12692 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26926) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12693 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26925) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12694 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 26920) (T_END_HEREDOC 26921) (T_VARIABLE 2822) (encaps_list 26922) (encaps_var 2824)) 12695 ((T_FN 35) (T_FUNCTION 126) (fn 26565) #'26566 (inline_function 26919)) 12696 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26918) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12697 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26917) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12698 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26916) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12699 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26913) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12700 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26912) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12701 (("!" 1638) ("\"" 1639) ("$" 1640) ("(" 1641) ("+" 1642) ("-" 1643) ("@" 1644) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1645) (T_ARRAY_CAST 1646) (T_ATTRIBUTE 125) (T_BOOL_CAST 1647) (T_CLASS_C 1648) (T_CLONE 1649) (T_CONSTANT_ENCAPSED_STRING 1650) (T_DEC 1651) (T_DIR 1652) (T_DNUMBER 1653) (T_DOUBLE_CAST 1654) (T_ELLIPSIS 1655) (T_EMPTY 1656) (T_EVAL 1657) (T_EXIT 1658) (T_FILE 1659) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1660) (T_INC 1661) (T_INCLUDE 1662) (T_INCLUDE_ONCE 1663) (T_INT_CAST 1664) (T_ISSET 1665) (T_LINE 1666) (T_LIST 1667) (T_LNUMBER 1668) (T_MATCH 1669) (T_METHOD_C 1670) (T_NAME_FULLY_QUALIFIED 1671) (T_NAME_QUALIFIED 1672) (T_NAME_RELATIVE 1673) (T_NEW 1674) (T_NS_C 1675) (T_OBJECT_CAST 1676) (T_PRINT 1677) (T_REQUIRE 1678) (T_REQUIRE_ONCE 1679) (T_START_HEREDOC 1680) (T_STATIC 1681) (T_STRING 1682) (T_STRING_CAST 1683) (T_THROW 1684) (T_TRAIT_C 1685) (T_UNSET_CAST 1686) (T_VARIABLE 1687) (T_YIELD 1688) (T_YIELD_FROM 1689) ("[" 1690) ("`" 1691) (ampersand 1692) (array_object_dereferenceable 1693) (array_pair 1694) (array_pair_list 26908) (attribute 129) (attributes 1696) (callable_expr 1697) (callable_variable 1698) (class_constant 1699) (class_name 1700) (constant 1701) (dereferenceable_scalar 1702) (expr 1703) (fn 1704) (fully_dereferenceable 103) #'1705 (function_call 1706) (inline_function 1707) (internal_functions_in_yacc 1708) (match 1709) (name 1710) (new_expr 1711) (non_empty_array_pair_list 1712) (possible_array_pair 1713) (scalar 1714) (simple_variable 1715) (static_member 1716) (variable 1717) (variable_class_name 1718) ("~" 1719)) 12702 ((T_CURLY_OPEN 2724) (T_DOLLAR_OPEN_CURLY_BRACES 2725) (T_ENCAPSED_AND_WHITESPACE 2726) (T_VARIABLE 2727) (backticks_expr 26906) (encaps_list 2729) (encaps_var 2730)) 12703 ((T_NULLSAFE_OBJECT_OPERATOR 26889) (T_OBJECT_OPERATOR 26890) ("[" 26891) ("{" 26892)) 12704 ((T_ATTRIBUTE 125) (T_FN 35) (T_FUNCTION 126) (T_STATIC 26886) (attribute 2654) (fn 26565) #'26566 (inline_function 26887)) 12705 (("(" 26765) (argument_list 26885)) 12706 ((T_PAAMAYIM_NEKUDOTAYIM 26880)) 12707 (("%" 26580) ("*" 26581) ("+" 26582) ("-" 26583) ("." 26584) ("/" 26585) ("<" 26586) (">" 26587) ("?" 26588) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 26589) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26590) (T_BOOLEAN_AND 26591) (T_BOOLEAN_OR 26592) (T_COALESCE 26593) (T_DOUBLE_ARROW 26878) (T_INSTANCEOF 26594) (T_IS_EQUAL 26595) (T_IS_GREATER_OR_EQUAL 26596) (T_IS_IDENTICAL 26597) (T_IS_NOT_EQUAL 26598) (T_IS_NOT_IDENTICAL 26599) (T_IS_SMALLER_OR_EQUAL 26600) (T_LOGICAL_AND 26601) (T_LOGICAL_OR 26602) (T_LOGICAL_XOR 26603) (T_POW 26604) (T_SL 26605) (T_SPACESHIP 26606) (T_SR 26607) ("^" 26608) ("|" 26609)) 12708 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 26867)) 12709 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 133) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 134) (ampersand 135) (returns_ref 26855)) 12710 (("(" 26765) (argument_list 26854)) 12711 (("=" 26773) (T_AND_EQUAL 26774) (T_COALESCE_EQUAL 26775) (T_CONCAT_EQUAL 26776) (T_DEC 26777) (T_DIV_EQUAL 26778) (T_INC 26779) (T_MINUS_EQUAL 26780) (T_MOD_EQUAL 26781) (T_MUL_EQUAL 26782) (T_OR_EQUAL 26783) (T_PLUS_EQUAL 26784) (T_POW_EQUAL 26785) (T_SL_EQUAL 26786) (T_SR_EQUAL 26787) (T_XOR_EQUAL 26788)) 12712 ((T_PAAMAYIM_NEKUDOTAYIM 26681)) 12713 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26579) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12714 (("%" 26580) ("*" 26581) ("+" 26582) ("-" 26583) ("." 26584) ("/" 26585) ("<" 26586) (">" 26587) ("?" 26588) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 26589) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26590) (T_BOOLEAN_AND 26591) (T_BOOLEAN_OR 26592) (T_COALESCE 26593) (T_INSTANCEOF 26594) (T_IS_EQUAL 26595) (T_IS_GREATER_OR_EQUAL 26596) (T_IS_IDENTICAL 26597) (T_IS_NOT_EQUAL 26598) (T_IS_NOT_IDENTICAL 26599) (T_IS_SMALLER_OR_EQUAL 26600) (T_LOGICAL_AND 26601) (T_LOGICAL_OR 26602) (T_LOGICAL_XOR 26603) (T_POW 26604) (T_SL 26605) (T_SPACESHIP 26606) (T_SR 26607) ("^" 26608) ("|" 26609)) 12715 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26680) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12716 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26679) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12717 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26678) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12718 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26677) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12719 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26676) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12720 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26675) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12721 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26674) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12722 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26673) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12723 (("!" 651) ("\"" 652) ("$" 653) ("(" 654) ("+" 655) ("-" 656) (":" 26668) ("@" 658) (T_ARRAY 659) (T_ARRAY_CAST 660) (T_ATTRIBUTE 125) (T_BOOL_CAST 661) (T_CLASS_C 662) (T_CLONE 663) (T_CONSTANT_ENCAPSED_STRING 664) (T_DEC 665) (T_DIR 666) (T_DNUMBER 667) (T_DOUBLE_CAST 668) (T_EMPTY 669) (T_EVAL 670) (T_EXIT 671) (T_FILE 672) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 673) (T_INC 674) (T_INCLUDE 675) (T_INCLUDE_ONCE 676) (T_INT_CAST 677) (T_ISSET 678) (T_LINE 679) (T_LIST 680) (T_LNUMBER 681) (T_MATCH 682) (T_METHOD_C 683) (T_NAME_FULLY_QUALIFIED 684) (T_NAME_QUALIFIED 685) (T_NAME_RELATIVE 686) (T_NEW 687) (T_NS_C 688) (T_OBJECT_CAST 689) (T_PRINT 690) (T_REQUIRE 691) (T_REQUIRE_ONCE 692) (T_START_HEREDOC 693) (T_STATIC 694) (T_STRING 695) (T_STRING_CAST 696) (T_THROW 697) (T_TRAIT_C 698) (T_UNSET_CAST 699) (T_VARIABLE 700) (T_YIELD 701) (T_YIELD_FROM 702) ("[" 703) ("`" 704) (array_object_dereferenceable 705) (attribute 129) (attributes 706) (callable_expr 707) (callable_variable 708) (class_constant 709) (class_name 710) (constant 711) (dereferenceable_scalar 712) (expr 26669) (fn 714) (fully_dereferenceable 103) #'715 (function_call 716) (inline_function 717) (internal_functions_in_yacc 718) (match 719) (name 720) (new_expr 721) (scalar 722) (simple_variable 723) (static_member 724) (variable 725) (variable_class_name 726) ("~" 727)) 12724 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26667) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12725 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26666) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12726 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26665) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12727 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26664) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12728 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26663) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12729 (("$" 26625) ("(" 26626) (T_NAME_FULLY_QUALIFIED 26627) (T_NAME_QUALIFIED 26628) (T_NAME_RELATIVE 26629) (T_STATIC 26630) (T_STRING 26631) (T_VARIABLE 26632) (class_name 26633) (class_name_reference 26634) (name 26635) (new_variable 26636) (simple_variable 26637)) 12730 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26624) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12731 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26623) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12732 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26622) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12733 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26621) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12734 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26620) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12735 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26619) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12736 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26618) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12737 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26617) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12738 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26616) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12739 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26615) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12740 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26614) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12741 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26613) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12742 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26612) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12743 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26611) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12744 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26610) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12745 (("$" 26625) (T_VARIABLE 26632) (simple_variable 26659) ("{" 26660)) 12746 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26657) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12747 ((T_PAAMAYIM_NEKUDOTAYIM 26655)) 12748 ((T_NULLSAFE_OBJECT_OPERATOR 26638) (T_OBJECT_OPERATOR 26639) (T_PAAMAYIM_NEKUDOTAYIM 26640) ("[" 26641) ("{" 26642)) 12749 (("$" 26625) (T_STRING 26648) (T_VARIABLE 26632) (property_name 26654) (simple_variable 26650) ("{" 26651)) 12750 (("$" 26625) (T_STRING 26648) (T_VARIABLE 26632) (property_name 26649) (simple_variable 26650) ("{" 26651)) 12751 (("$" 26625) (T_VARIABLE 26632) (simple_variable 26647)) 12752 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 26645) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 12753 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26643) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12754 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26644)) 12755 (("]" 26646)) 12756 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26652) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12757 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26653)) 12758 (("$" 26625) (T_VARIABLE 26632) (simple_variable 26656)) 12759 (("%" 280) (")" 26658) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12760 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26661) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12761 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26662)) 12762 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26672) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12763 (("%" 729) ("*" 730) ("+" 731) ("-" 732) ("." 733) ("/" 734) (":" 26670) ("<" 735) (">" 736) ("?" 737) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 738) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 739) (T_BOOLEAN_AND 740) (T_BOOLEAN_OR 741) (T_COALESCE 742) (T_INSTANCEOF 743) (T_IS_EQUAL 744) (T_IS_GREATER_OR_EQUAL 745) (T_IS_IDENTICAL 746) (T_IS_NOT_EQUAL 747) (T_IS_NOT_IDENTICAL 748) (T_IS_SMALLER_OR_EQUAL 749) (T_LOGICAL_AND 750) (T_LOGICAL_OR 751) (T_LOGICAL_XOR 752) (T_POW 753) (T_SL 754) (T_SPACESHIP 755) (T_SR 756) ("^" 757) ("|" 758)) 12764 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26671) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12765 (("$" 26505) (T_ABSTRACT 26682) (T_ARRAY 26683) (T_AS 26684) (T_BREAK 26685) (T_CALLABLE 26686) (T_CASE 26687) (T_CATCH 26688) (T_CLASS 26689) (T_CLASS_C 26690) (T_CLONE 26691) (T_CONST 26692) (T_CONTINUE 26693) (T_DECLARE 26694) (T_DEFAULT 26695) (T_DIR 26696) (T_DO 26697) (T_ECHO 26698) (T_ELSE 26699) (T_ELSEIF 26700) (T_EMPTY 26701) (T_ENDDECLARE 26702) (T_ENDFOR 26703) (T_ENDFOREACH 26704) (T_ENDIF 26705) (T_ENDSWITCH 26706) (T_ENDWHILE 26707) (T_ENUM 26708) (T_EVAL 26709) (T_EXIT 26710) (T_EXTENDS 26711) (T_FILE 26712) (T_FINAL 26713) (T_FINALLY 26714) (T_FN 26715) (T_FOR 26716) (T_FOREACH 26717) (T_FUNCTION 26718) (T_FUNC_C 26719) (T_GLOBAL 26720) (T_GOTO 26721) (T_IF 26722) (T_IMPLEMENTS 26723) (T_INCLUDE 26724) (T_INCLUDE_ONCE 26725) (T_INSTANCEOF 26726) (T_INSTEADOF 26727) (T_INTERFACE 26728) (T_ISSET 26729) (T_LINE 26730) (T_LIST 26731) (T_LOGICAL_AND 26732) (T_LOGICAL_OR 26733) (T_LOGICAL_XOR 26734) (T_MATCH 26735) (T_METHOD_C 26736) (T_NAMESPACE 26737) (T_NEW 26738) (T_NS_C 26739) (T_PRINT 26740) (T_PRIVATE 26741) (T_PROTECTED 26742) (T_PUBLIC 26743) (T_READONLY 26744) (T_REQUIRE 26745) (T_REQUIRE_ONCE 26746) (T_RETURN 26747) (T_STATIC 26748) (T_STRING 26749) (T_SWITCH 26750) (T_THROW 26751) (T_TRAIT 26752) (T_TRAIT_C 26753) (T_TRY 26754) (T_UNSET 26755) (T_USE 26756) (T_VAR 26757) (T_VARIABLE 26551) (T_WHILE 26758) (T_YIELD 26759) (identifier 26760) (member_name 26761) (reserved_non_modifiers 26762) (semi_reserved 26763) (simple_variable 26764) ("{" 914)) 12766 (("(" 26765) (argument_list 26766)) 12767 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26767) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 26768) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 26769) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12768 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26772) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12769 (("," 1449) (possible_comma 26770)) 12770 ((")" 26771)) 12771 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (ampersand 26802) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26803) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12772 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26801) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12773 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26800) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12774 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26799) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12775 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26798) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12776 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26797) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12777 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26796) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12778 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26795) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12779 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26794) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12780 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26793) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12781 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26792) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12782 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26791) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12783 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26790) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12784 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26789) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12785 (("\"" 1489) ("$" 26804) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 26805) ("[" 1508) (array_object_dereferenceable 26806) (callable_expr 26807) (callable_variable 26808) (class_constant 1512) (class_name 26809) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 26810) (name 26811) (simple_variable 26812) (static_member 26813) (variable 26814) (variable_class_name 26815)) 12786 (("$" 26804) (T_VARIABLE 26805) (simple_variable 26850) ("{" 26851)) 12787 ((T_NULLSAFE_OBJECT_OPERATOR 26833) (T_OBJECT_OPERATOR 26834) ("[" 26835) ("{" 26836)) 12788 (("(" 26819) (argument_list 26832)) 12789 ((T_PAAMAYIM_NEKUDOTAYIM 26828)) 12790 (("(" 26819) (argument_list 26827)) 12791 ((T_PAAMAYIM_NEKUDOTAYIM 26816)) 12792 (("$" 26804) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 26805) (T_WHILE 1599) (T_YIELD 1600) (identifier 1601) (member_name 26817) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 26818) ("{" 914)) 12793 (("(" 26819) (argument_list 26820)) 12794 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26821) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 26822) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 26823) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12795 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26826) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12796 (("," 1449) (possible_comma 26824)) 12797 ((")" 26825)) 12798 (("$" 26804) (T_ABSTRACT 1523) (T_ARRAY 1524) (T_AS 1525) (T_BREAK 1526) (T_CALLABLE 1527) (T_CASE 1528) (T_CATCH 1529) (T_CLASS 1530) (T_CLASS_C 1531) (T_CLONE 1532) (T_CONST 1533) (T_CONTINUE 1534) (T_DECLARE 1535) (T_DEFAULT 1536) (T_DIR 1537) (T_DO 1538) (T_ECHO 1539) (T_ELSE 1540) (T_ELSEIF 1541) (T_EMPTY 1542) (T_ENDDECLARE 1543) (T_ENDFOR 1544) (T_ENDFOREACH 1545) (T_ENDIF 1546) (T_ENDSWITCH 1547) (T_ENDWHILE 1548) (T_ENUM 1549) (T_EVAL 1550) (T_EXIT 1551) (T_EXTENDS 1552) (T_FILE 1553) (T_FINAL 1554) (T_FINALLY 1555) (T_FN 1556) (T_FOR 1557) (T_FOREACH 1558) (T_FUNCTION 1559) (T_FUNC_C 1560) (T_GLOBAL 1561) (T_GOTO 1562) (T_IF 1563) (T_IMPLEMENTS 1564) (T_INCLUDE 1565) (T_INCLUDE_ONCE 1566) (T_INSTANCEOF 1567) (T_INSTEADOF 1568) (T_INTERFACE 1569) (T_ISSET 1570) (T_LINE 1571) (T_LIST 1572) (T_LOGICAL_AND 1573) (T_LOGICAL_OR 1574) (T_LOGICAL_XOR 1575) (T_MATCH 1576) (T_METHOD_C 1577) (T_NAMESPACE 1578) (T_NEW 1579) (T_NS_C 1580) (T_PRINT 1581) (T_PRIVATE 1582) (T_PROTECTED 1583) (T_PUBLIC 1584) (T_READONLY 1585) (T_REQUIRE 1586) (T_REQUIRE_ONCE 1587) (T_RETURN 1588) (T_STATIC 1589) (T_STRING 1590) (T_SWITCH 1591) (T_THROW 1592) (T_TRAIT 1593) (T_TRAIT_C 1594) (T_TRY 1595) (T_UNSET 1596) (T_USE 1597) (T_VAR 1598) (T_VARIABLE 26805) (T_WHILE 1599) (T_YIELD 1600) (identifier 1616) (member_name 26829) (reserved_non_modifiers 1603) (semi_reserved 1604) (simple_variable 26830) ("{" 914)) 12799 (("(" 26819) (argument_list 26831)) 12800 (("$" 26804) (T_STRING 26841) (T_VARIABLE 26805) (property_name 26848) (simple_variable 26843) ("{" 26844)) 12801 (("$" 26804) (T_STRING 26841) (T_VARIABLE 26805) (property_name 26842) (simple_variable 26843) ("{" 26844)) 12802 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 26839) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 12803 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26837) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12804 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26838)) 12805 (("]" 26840)) 12806 (("(" 26819) (argument_list 26847)) 12807 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26845) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12808 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26846)) 12809 (("(" 26819) (argument_list 26849)) 12810 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26852) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12811 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26853)) 12812 ((backup_doc_comment 26856)) 12813 (("(" 26857)) 12814 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 26858)) 12815 ((")" 26859)) 12816 ((T_USE 148) (lexical_vars 26860)) 12817 ((":" 150) (return_type 26861)) 12818 ((backup_fn_flags 26862)) 12819 (("{" 26863)) 12820 ((inner_statement_list 26864)) 12821 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 26865) ("~" 124)) 12822 ((backup_fn_flags 26866)) 12823 ((backup_doc_comment 26868)) 12824 (("(" 26869)) 12825 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 26870)) 12826 ((")" 26871)) 12827 ((":" 2207) (return_type 26872)) 12828 ((T_DOUBLE_ARROW 26873)) 12829 ((backup_fn_flags 26874)) 12830 ((backup_lex_pos 26875)) 12831 (("!" 25769) ("\"" 25770) ("$" 25771) ("(" 25772) ("+" 25773) ("-" 25774) ("@" 25775) (T_ARRAY 25776) (T_ARRAY_CAST 25777) (T_ATTRIBUTE 125) (T_BOOL_CAST 25778) (T_CLASS_C 25779) (T_CLONE 25780) (T_CONSTANT_ENCAPSED_STRING 25781) (T_DEC 25782) (T_DIR 25783) (T_DNUMBER 25784) (T_DOUBLE_CAST 25785) (T_EMPTY 25786) (T_EVAL 25787) (T_EXIT 25788) (T_FILE 25789) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 25790) (T_INC 25791) (T_INCLUDE 25792) (T_INCLUDE_ONCE 25793) (T_INT_CAST 25794) (T_ISSET 25795) (T_LINE 25796) (T_LIST 25797) (T_LNUMBER 25798) (T_MATCH 25799) (T_METHOD_C 25800) (T_NAME_FULLY_QUALIFIED 25801) (T_NAME_QUALIFIED 25802) (T_NAME_RELATIVE 25803) (T_NEW 25804) (T_NS_C 25805) (T_OBJECT_CAST 25806) (T_PRINT 25807) (T_REQUIRE 25808) (T_REQUIRE_ONCE 25809) (T_START_HEREDOC 25810) (T_STATIC 25811) (T_STRING 25812) (T_STRING_CAST 25813) (T_THROW 25814) (T_TRAIT_C 25815) (T_UNSET_CAST 25816) (T_VARIABLE 25817) (T_YIELD 25818) (T_YIELD_FROM 25819) ("[" 25820) ("`" 25821) (array_object_dereferenceable 25822) (attribute 129) (attributes 25823) (callable_expr 25824) (callable_variable 25825) (class_constant 25826) (class_name 25827) (constant 25828) (dereferenceable_scalar 25829) (expr 26876) (fn 25831) (fully_dereferenceable 103) #'25832 (function_call 25833) (inline_function 25834) (internal_functions_in_yacc 25835) (match 25836) (name 25837) (new_expr 25838) (scalar 25839) (simple_variable 25840) (static_member 25841) (variable 25842) (variable_class_name 25843) ("~" 25844)) 12832 (("%" 25846) ("*" 25847) ("+" 25848) ("-" 25849) ("." 25850) ("/" 25851) ("<" 25852) (">" 25853) ("?" 25854) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25855) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25856) (T_BOOLEAN_AND 25857) (T_BOOLEAN_OR 25858) (T_COALESCE 25859) (T_INSTANCEOF 25860) (T_IS_EQUAL 25861) (T_IS_GREATER_OR_EQUAL 25862) (T_IS_IDENTICAL 25863) (T_IS_NOT_EQUAL 25864) (T_IS_NOT_IDENTICAL 25865) (T_IS_SMALLER_OR_EQUAL 25866) (T_LOGICAL_AND 25867) (T_LOGICAL_OR 25868) (T_LOGICAL_XOR 25869) (T_POW 25870) (T_SL 25871) (T_SPACESHIP 25872) (T_SR 25873) ("^" 25874) (backup_fn_flags 26877) ("|" 25875)) 12833 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 26879) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 12834 (("$" 26505) (T_ABSTRACT 26682) (T_ARRAY 26683) (T_AS 26684) (T_BREAK 26685) (T_CALLABLE 26686) (T_CASE 26687) (T_CATCH 26688) (T_CLASS 26689) (T_CLASS_C 26690) (T_CLONE 26691) (T_CONST 26692) (T_CONTINUE 26693) (T_DECLARE 26694) (T_DEFAULT 26695) (T_DIR 26696) (T_DO 26697) (T_ECHO 26698) (T_ELSE 26699) (T_ELSEIF 26700) (T_EMPTY 26701) (T_ENDDECLARE 26702) (T_ENDFOR 26703) (T_ENDFOREACH 26704) (T_ENDIF 26705) (T_ENDSWITCH 26706) (T_ENDWHILE 26707) (T_ENUM 26708) (T_EVAL 26709) (T_EXIT 26710) (T_EXTENDS 26711) (T_FILE 26712) (T_FINAL 26713) (T_FINALLY 26714) (T_FN 26715) (T_FOR 26716) (T_FOREACH 26717) (T_FUNCTION 26718) (T_FUNC_C 26719) (T_GLOBAL 26720) (T_GOTO 26721) (T_IF 26722) (T_IMPLEMENTS 26723) (T_INCLUDE 26724) (T_INCLUDE_ONCE 26725) (T_INSTANCEOF 26726) (T_INSTEADOF 26727) (T_INTERFACE 26728) (T_ISSET 26729) (T_LINE 26730) (T_LIST 26731) (T_LOGICAL_AND 26732) (T_LOGICAL_OR 26733) (T_LOGICAL_XOR 26734) (T_MATCH 26735) (T_METHOD_C 26736) (T_NAMESPACE 26737) (T_NEW 26738) (T_NS_C 26739) (T_PRINT 26740) (T_PRIVATE 26741) (T_PROTECTED 26742) (T_PUBLIC 26743) (T_READONLY 26744) (T_REQUIRE 26745) (T_REQUIRE_ONCE 26746) (T_RETURN 26747) (T_STATIC 26748) (T_STRING 26749) (T_SWITCH 26750) (T_THROW 26751) (T_TRAIT 26752) (T_TRAIT_C 26753) (T_TRY 26754) (T_UNSET 26755) (T_USE 26756) (T_VAR 26757) (T_VARIABLE 26551) (T_WHILE 26758) (T_YIELD 26759) (identifier 26881) (member_name 26882) (reserved_non_modifiers 26762) (semi_reserved 26763) (simple_variable 26883) ("{" 914)) 12835 (("(" 26765) (argument_list 26884)) 12836 ((T_FN 35) (T_FUNCTION 126) (fn 26565) #'26566 (inline_function 26888)) 12837 (("$" 26505) (T_STRING 26897) (T_VARIABLE 26551) (property_name 26904) (simple_variable 26899) ("{" 26900)) 12838 (("$" 26505) (T_STRING 26897) (T_VARIABLE 26551) (property_name 26898) (simple_variable 26899) ("{" 26900)) 12839 (("!" 485) ("\"" 486) ("$" 487) ("(" 488) ("+" 489) ("-" 490) ("@" 491) (T_ARRAY 492) (T_ARRAY_CAST 493) (T_ATTRIBUTE 125) (T_BOOL_CAST 494) (T_CLASS_C 495) (T_CLONE 496) (T_CONSTANT_ENCAPSED_STRING 497) (T_DEC 498) (T_DIR 499) (T_DNUMBER 500) (T_DOUBLE_CAST 501) (T_EMPTY 502) (T_EVAL 503) (T_EXIT 504) (T_FILE 505) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 506) (T_INC 507) (T_INCLUDE 508) (T_INCLUDE_ONCE 509) (T_INT_CAST 510) (T_ISSET 511) (T_LINE 512) (T_LIST 513) (T_LNUMBER 514) (T_MATCH 515) (T_METHOD_C 516) (T_NAME_FULLY_QUALIFIED 517) (T_NAME_QUALIFIED 518) (T_NAME_RELATIVE 519) (T_NEW 520) (T_NS_C 521) (T_OBJECT_CAST 522) (T_PRINT 523) (T_REQUIRE 524) (T_REQUIRE_ONCE 525) (T_START_HEREDOC 526) (T_STATIC 527) (T_STRING 528) (T_STRING_CAST 529) (T_THROW 530) (T_TRAIT_C 531) (T_UNSET_CAST 532) (T_VARIABLE 533) (T_YIELD 534) (T_YIELD_FROM 535) ("[" 536) ("`" 537) (array_object_dereferenceable 538) (attribute 129) (attributes 539) (callable_expr 540) (callable_variable 541) (class_constant 542) (class_name 543) (constant 544) (dereferenceable_scalar 545) (expr 546) (fn 547) (fully_dereferenceable 103) #'548 (function_call 549) (inline_function 550) (internal_functions_in_yacc 551) (match 552) (name 553) (new_expr 554) (optional_expr 26895) (scalar 556) (simple_variable 557) (static_member 558) (variable 559) (variable_class_name 560) ("~" 561)) 12840 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26893) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12841 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26894)) 12842 (("]" 26896)) 12843 (("(" 26765) (argument_list 26903)) 12844 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26901) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12845 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26902)) 12846 (("(" 26765) (argument_list 26905)) 12847 (("`" 26907)) 12848 (("]" 26909)) 12849 (("=" 26910)) 12850 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26911) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12851 (("%" 26580) ("*" 26581) ("+" 26582) ("-" 26583) ("." 26584) ("/" 26585) ("<" 26586) (">" 26587) ("?" 26588) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 26589) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26590) (T_BOOLEAN_AND 26591) (T_BOOLEAN_OR 26592) (T_COALESCE 26593) (T_DOUBLE_ARROW 26914) (T_INSTANCEOF 26594) (T_IS_EQUAL 26595) (T_IS_GREATER_OR_EQUAL 26596) (T_IS_IDENTICAL 26597) (T_IS_NOT_EQUAL 26598) (T_IS_NOT_IDENTICAL 26599) (T_IS_SMALLER_OR_EQUAL 26600) (T_LOGICAL_AND 26601) (T_LOGICAL_OR 26602) (T_LOGICAL_XOR 26603) (T_POW 26604) (T_SL 26605) (T_SPACESHIP 26606) (T_SR 26607) ("^" 26608) ("|" 26609)) 12852 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26915) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12853 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 26924) (T_VARIABLE 2822) (encaps_var 2836)) 12854 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 26923) (T_VARIABLE 2822) (encaps_var 2827)) 12855 (("(" 2899) (argument_list 2900) (ctor_arguments 26943)) 12856 ((T_ATTRIBUTE 2853) (T_CLASS 26929) (anonymous_class 26942) (attribute 2898)) 12857 (("(" 26933) (argument_list 26934) (ctor_arguments 26935)) 12858 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26936) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 26937) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 26938) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12859 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 26941) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12860 (("," 1449) (possible_comma 26939)) 12861 ((")" 26940)) 12862 ((T_EXTENDS 2902) (extends_from 26944)) 12863 ((T_IMPLEMENTS 2904) (implements_list 26945)) 12864 ((backup_doc_comment 26946)) 12865 (("{" 26947)) 12866 ((class_statement_list 26948)) 12867 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 26949)) 12868 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26951) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12869 (("%" 280) (")" 26952) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12870 (("{" 26953)) 12871 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 26954) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 12872 (("}" 26955)) 12873 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 26957) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 12874 ((")" 26958)) 12875 (("=" 26959)) 12876 (("!" 26503) ("\"" 26504) ("$" 26505) ("(" 26506) ("+" 26507) ("-" 26508) ("@" 26509) (T_ARRAY 26510) (T_ARRAY_CAST 26511) (T_ATTRIBUTE 125) (T_BOOL_CAST 26512) (T_CLASS_C 26513) (T_CLONE 26514) (T_CONSTANT_ENCAPSED_STRING 26515) (T_DEC 26516) (T_DIR 26517) (T_DNUMBER 26518) (T_DOUBLE_CAST 26519) (T_EMPTY 26520) (T_EVAL 26521) (T_EXIT 26522) (T_FILE 26523) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 26524) (T_INC 26525) (T_INCLUDE 26526) (T_INCLUDE_ONCE 26527) (T_INT_CAST 26528) (T_ISSET 26529) (T_LINE 26530) (T_LIST 26531) (T_LNUMBER 26532) (T_MATCH 26533) (T_METHOD_C 26534) (T_NAME_FULLY_QUALIFIED 26535) (T_NAME_QUALIFIED 26536) (T_NAME_RELATIVE 26537) (T_NEW 26538) (T_NS_C 26539) (T_OBJECT_CAST 26540) (T_PRINT 26541) (T_REQUIRE 26542) (T_REQUIRE_ONCE 26543) (T_START_HEREDOC 26544) (T_STATIC 26545) (T_STRING 26546) (T_STRING_CAST 26547) (T_THROW 26548) (T_TRAIT_C 26549) (T_UNSET_CAST 26550) (T_VARIABLE 26551) (T_YIELD 26552) (T_YIELD_FROM 26553) ("[" 26554) ("`" 26555) (array_object_dereferenceable 26556) (attribute 129) (attributes 26557) (callable_expr 26558) (callable_variable 26559) (class_constant 26560) (class_name 26561) (constant 26562) (dereferenceable_scalar 26563) (expr 26960) (fn 26565) (fully_dereferenceable 103) #'26566 (function_call 26567) (inline_function 26568) (internal_functions_in_yacc 26569) (match 26570) (name 26571) (new_expr 26572) (scalar 26573) (simple_variable 26574) (static_member 26575) (variable 26576) (variable_class_name 26577) ("~" 26578)) 12877 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 26962) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12878 (("," 5117) (possible_comma 26963)) 12879 ((")" 26964)) 12880 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 26971) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12881 ((")" 26972)) 12882 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26974) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12883 (("%" 280) (")" 26975) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12884 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 26977) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12885 (("%" 280) (")" 26978) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12886 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 26985) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 12887 ((")" 26986)) 12888 (("%" 280) (")" 26991) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12889 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 26994) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 12890 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 26995)) 12891 (("\"" 26997) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161)) 12892 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27000) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12893 (("%" 280) (")" 27001) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12894 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27002) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 66) (T_START_HEREDOC 67) (T_STATIC 68) (T_STRING 69) (T_STRING_CAST 70) (T_SWITCH 71) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 75) (T_UNSET 76) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 80) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 85) (alt_if_stmt_without_else 86) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 101) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 107) (if_stmt_without_else 108) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27003) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 27004) ("{" 123) ("~" 124)) 12895 ((inner_statement_list 27005)) 12896 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 27006) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 12897 ((";" 27007)) 12898 ((T_AS 27030) (T_NS_SEPARATOR 27048)) 12899 ((";" 27047)) 12900 (("," 27022) (";" 27046)) 12901 ((T_NAME_FULLY_QUALIFIED 27010) (T_NAME_QUALIFIED 27011) (T_STRING 27012) (group_use_declaration 27019) (legacy_namespace_name 27020) (namespace_name 27015) (use_declaration 27016) (use_declarations 27021)) 12902 ((";" 27045)) 12903 ((T_AS 27030) (T_NS_SEPARATOR 27032)) 12904 (("," 27022) (";" 27023)) 12905 ((T_NAME_FULLY_QUALIFIED 27024) (T_NAME_QUALIFIED 27025) (T_STRING 27026) (legacy_namespace_name 27027) (namespace_name 27028) (use_declaration 27029)) 12906 ((T_AS 27030)) 12907 ((T_STRING 27031)) 12908 (("{" 27033)) 12909 ((T_NAME_QUALIFIED 27034) (T_STRING 27035) (namespace_name 27036) (unprefixed_use_declaration 27037) (unprefixed_use_declarations 27038)) 12910 ((T_AS 27043)) 12911 (("," 27039) (possible_comma 27040)) 12912 ((T_NAME_QUALIFIED 27034) (T_STRING 27035) (namespace_name 27036) (unprefixed_use_declaration 27042)) 12913 (("}" 27041)) 12914 ((T_STRING 27044)) 12915 (("{" 27049)) 12916 ((T_CONST 27050) (T_FUNCTION 27051) (T_NAME_QUALIFIED 27034) (T_STRING 27035) (inline_use_declaration 27052) (inline_use_declarations 27053) (namespace_name 27036) (unprefixed_use_declaration 27054) (use_type 27055)) 12917 (("," 27057) (possible_comma 27058)) 12918 ((T_NAME_QUALIFIED 27034) (T_STRING 27035) (namespace_name 27036) (unprefixed_use_declaration 27056)) 12919 ((T_CONST 27050) (T_FUNCTION 27051) (T_NAME_QUALIFIED 27034) (T_STRING 27035) (inline_use_declaration 27060) (namespace_name 27036) (unprefixed_use_declaration 27054) (use_type 27055)) 12920 (("}" 27059)) 12921 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 27063) (variable 16049) (variable_class_name 16050)) 12922 (("," 16062) (possible_comma 27064)) 12923 ((")" 27065)) 12924 ((";" 27066)) 12925 ((inner_statement_list 27068)) 12926 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27069) ("~" 124)) 12927 ((catch_list 27070)) 12928 ((T_CATCH 27071) (T_FINALLY 27072) (finally_statement 27073)) 12929 (("(" 27077)) 12930 (("{" 27074)) 12931 ((inner_statement_list 27075)) 12932 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27076) ("~" 124)) 12933 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 27078) (class_name 16111) (name 16112)) 12934 ((T_VARIABLE 16113) (optional_variable 27079) ("|" 16115)) 12935 ((")" 27080)) 12936 (("{" 27081)) 12937 ((inner_statement_list 27082)) 12938 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27083) ("~" 124)) 12939 ((backup_doc_comment 27085)) 12940 (("{" 27086)) 12941 ((class_statement_list 27087)) 12942 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 27088)) 12943 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27091) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 12944 (("%" 280) (")" 27092) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 12945 ((":" 27093) (switch_case_list 27094) ("{" 27095)) 12946 ((";" 27101) (case_list 27102)) 12947 ((";" 27096) (case_list 27097)) 12948 ((case_list 27099)) 12949 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 27098)) 12950 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 27100)) 12951 ((case_list 27105)) 12952 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 27103)) 12953 ((";" 27104)) 12954 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 27106)) 12955 ((";" 27107)) 12956 (("," 16412) (";" 27111)) 12957 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_END_HEREDOC 27116) (T_VARIABLE 2822) (encaps_var 2836)) 12958 ((T_CURLY_OPEN 2818) (T_DOLLAR_OPEN_CURLY_BRACES 2819) (T_ENCAPSED_AND_WHITESPACE 2825) (T_END_HEREDOC 27115) (T_VARIABLE 2822) (encaps_var 2827)) 12959 ((";" 27118)) 12960 (("(" 2899) (argument_list 2900) (ctor_arguments 27137)) 12961 ((T_ATTRIBUTE 2853) (T_CLASS 27123) (anonymous_class 27136) (attribute 2898)) 12962 (("(" 27127) (argument_list 27128) (ctor_arguments 27129)) 12963 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 27130) ("+" 924) ("-" 925) ("@" 926) (T_ABSTRACT 927) (T_ARRAY 928) (T_ARRAY_CAST 929) (T_AS 930) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_BREAK 932) (T_CALLABLE 933) (T_CASE 934) (T_CATCH 935) (T_CLASS 936) (T_CLASS_C 937) (T_CLONE 938) (T_CONST 939) (T_CONSTANT_ENCAPSED_STRING 940) (T_CONTINUE 941) (T_DEC 942) (T_DECLARE 943) (T_DEFAULT 944) (T_DIR 945) (T_DNUMBER 946) (T_DO 947) (T_DOUBLE_CAST 948) (T_ECHO 949) (T_ELLIPSIS 27131) (T_ELSE 951) (T_ELSEIF 952) (T_EMPTY 953) (T_ENDDECLARE 954) (T_ENDFOR 955) (T_ENDFOREACH 956) (T_ENDIF 957) (T_ENDSWITCH 958) (T_ENDWHILE 959) (T_ENUM 960) (T_EVAL 961) (T_EXIT 962) (T_EXTENDS 963) (T_FILE 964) (T_FINAL 965) (T_FINALLY 966) (T_FN 967) (T_FOR 968) (T_FOREACH 969) (T_FUNCTION 970) (T_FUNC_C 971) (T_GLOBAL 972) (T_GOTO 973) (T_IF 974) (T_IMPLEMENTS 975) (T_INC 976) (T_INCLUDE 977) (T_INCLUDE_ONCE 978) (T_INSTANCEOF 979) (T_INSTEADOF 980) (T_INTERFACE 981) (T_INT_CAST 982) (T_ISSET 983) (T_LINE 984) (T_LIST 985) (T_LNUMBER 986) (T_LOGICAL_AND 987) (T_LOGICAL_OR 988) (T_LOGICAL_XOR 989) (T_MATCH 990) (T_METHOD_C 991) (T_NAMESPACE 992) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 996) (T_NS_C 997) (T_OBJECT_CAST 998) (T_PRINT 999) (T_PRIVATE 1000) (T_PROTECTED 1001) (T_PUBLIC 1002) (T_READONLY 1003) (T_REQUIRE 1004) (T_REQUIRE_ONCE 1005) (T_RETURN 1006) (T_START_HEREDOC 1007) (T_STATIC 1008) (T_STRING 1009) (T_STRING_CAST 1010) (T_SWITCH 1011) (T_THROW 1012) (T_TRAIT 1013) (T_TRAIT_C 1014) (T_TRY 1015) (T_UNSET 1016) (T_UNSET_CAST 1017) (T_USE 1018) (T_VAR 1019) (T_VARIABLE 1020) (T_WHILE 1021) (T_YIELD 1022) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (argument 1026) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1035) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (identifier 1039) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (non_empty_argument_list 27132) (reserved_non_modifiers 1046) (scalar 1047) (semi_reserved 1048) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12964 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) (")" 27135) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 1454) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 12965 (("," 1449) (possible_comma 27133)) 12966 ((")" 27134)) 12967 ((T_EXTENDS 2902) (extends_from 27138)) 12968 ((T_IMPLEMENTS 2904) (implements_list 27139)) 12969 ((backup_doc_comment 27140)) 12970 (("{" 27141)) 12971 ((class_statement_list 27142)) 12972 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 27143)) 12973 ((";" 27620) ("{" 27621)) 12974 ((top_statement_list 27228)) 12975 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 27229) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 27230) (T_CLASS 27231) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 27232) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27233) (T_DEC 22) (T_DECLARE 27234) (T_DIR 24) (T_DNUMBER 25) (T_DO 27235) (T_DOUBLE_CAST 27) (T_ECHO 27236) (T_EMPTY 29) (T_ENUM 27237) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 27238) (T_FOREACH 27239) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 27240) (T_GOTO 27241) (T_HALT_COMPILER 27242) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27244) (T_INTERFACE 27245) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 27246) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27247) (T_START_HEREDOC 67) (T_STATIC 27248) (T_STRING 27249) (T_STRING_CAST 70) (T_SWITCH 27250) (T_THROW 72) (T_TRAIT 27251) (T_TRAIT_C 74) (T_TRY 27252) (T_UNSET 27253) (T_UNSET_CAST 77) (T_USE 27254) (T_VARIABLE 79) (T_WHILE 27255) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27256) (alt_if_stmt_without_else 27257) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 27258) (attributes 27259) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 27260) (class_modifier 95) (class_modifiers 27261) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 27262) (expr 27263) (fn 102) (fully_dereferenceable 103) #'27264 (function_call 105) (function_declaration_statement 27265) (if_stmt 27266) (if_stmt_without_else 27267) (inline_function 109) (interface_declaration_statement 27268) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27269) (static_member 118) (top_statement 27270) (trait_declaration_statement 27271) (variable 121) (variable_class_name 122) ("{" 27272) ("}" 27273) ("~" 124)) 12976 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 27618) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 12977 ((T_STRING 27611)) 12978 ((T_STRING 27602) (const_decl 27603) (const_list 27604)) 12979 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 27600) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 12980 (("(" 27591)) 12981 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27585) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 12982 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 27583) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 12983 ((T_STRING 27576)) 12984 (("(" 27563)) 12985 (("(" 27548)) 12986 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 27546) (simple_variable 17412)) 12987 ((T_STRING 27544)) 12988 (("(" 27541)) 12989 (("(" 27537)) 12990 ((T_STRING 27531)) 12991 ((T_ABSTRACT 27144) (T_ARRAY 27145) (T_AS 27146) (T_BREAK 27147) (T_CALLABLE 27148) (T_CASE 27149) (T_CATCH 27150) (T_CLASS 27151) (T_CLASS_C 27152) (T_CLONE 27153) (T_CONST 27154) (T_CONTINUE 27155) (T_DECLARE 27156) (T_DEFAULT 27157) (T_DIR 27158) (T_DO 27159) (T_ECHO 27160) (T_ELSE 27161) (T_ELSEIF 27162) (T_EMPTY 27163) (T_ENDDECLARE 27164) (T_ENDFOR 27165) (T_ENDFOREACH 27166) (T_ENDIF 27167) (T_ENDSWITCH 27168) (T_ENDWHILE 27169) (T_ENUM 27170) (T_EVAL 27171) (T_EXIT 27172) (T_EXTENDS 27173) (T_FILE 27174) (T_FINAL 27175) (T_FINALLY 27176) (T_FN 27177) (T_FOR 27178) (T_FOREACH 27179) (T_FUNCTION 27180) (T_FUNC_C 27181) (T_GLOBAL 27182) (T_GOTO 27183) (T_IF 27184) (T_IMPLEMENTS 27185) (T_INCLUDE 27186) (T_INCLUDE_ONCE 27187) (T_INSTANCEOF 27188) (T_INSTEADOF 27189) (T_INTERFACE 27190) (T_ISSET 27191) (T_LINE 27192) (T_LIST 27193) (T_LOGICAL_AND 27194) (T_LOGICAL_OR 27195) (T_LOGICAL_XOR 27196) (T_MATCH 27197) (T_METHOD_C 27198) (T_NAMESPACE 27199) (T_NAME_QUALIFIED 27200) (T_NEW 27201) (T_NS_C 27202) (T_PRINT 27203) (T_PRIVATE 27204) (T_PROTECTED 27205) (T_PUBLIC 27206) (T_READONLY 27207) (T_REQUIRE 27208) (T_REQUIRE_ONCE 27209) (T_RETURN 27210) (T_STATIC 27211) (T_STRING 27212) (T_SWITCH 27213) (T_THROW 27214) (T_TRAIT 27215) (T_TRAIT_C 27216) (T_TRY 27217) (T_UNSET 27218) (T_USE 27219) (T_VAR 27220) (T_WHILE 27221) (T_YIELD 27222) (identifier 27223) (namespace_declaration_name 27523) (reserved_non_modifiers 27225) (semi_reserved 27226) ("{" 27524)) 12992 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 27521) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 12993 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 27519)) 12994 ((":" 27518)) 12995 (("(" 27500)) 12996 ((T_STRING 27495)) 12997 (("{" 27478)) 12998 (("(" 27473)) 12999 ((T_CONST 27008) (T_FUNCTION 27009) (T_NAME_FULLY_QUALIFIED 27010) (T_NAME_QUALIFIED 27011) (T_STRING 27012) (legacy_namespace_name 27013) (mixed_group_use_declaration 27464) (namespace_name 27015) (use_declaration 27016) (use_declarations 27465) (use_type 27466)) 13000 (("(" 27455)) 13001 ((T_ELSE 27448) (T_ELSEIF 15767) (T_ENDIF 27449)) 13002 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 27231) (T_ENUM 27237) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 27245) (T_STATIC 15850) (T_TRAIT 27251) (attribute 15851) (attributed_statement 27447) (class_declaration_statement 27260) (class_modifier 95) (class_modifiers 27261) (enum_declaration_statement 27262) (fn 102) #'27264 (function_declaration_statement 27265) (inline_function 15853) (interface_declaration_statement 27268) (trait_declaration_statement 27271)) 13003 ((T_ABSTRACT 10) (T_CLASS 27439) (T_FINAL 34) (class_modifier 15842)) 13004 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 27438) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 13005 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15826) (ampersand 15827) (returns_ref 27426)) 13006 ((T_ELSE 27276) (T_ELSEIF 27277)) 13007 ((inner_statement_list 27274)) 13008 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27275) ("~" 124)) 13009 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 27229) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27230) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27233) (T_DEC 22) (T_DECLARE 27234) (T_DIR 24) (T_DNUMBER 25) (T_DO 27235) (T_DOUBLE_CAST 27) (T_ECHO 27236) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27238) (T_FOREACH 27239) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27240) (T_GOTO 27241) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27247) (T_START_HEREDOC 67) (T_STATIC 27248) (T_STRING 27249) (T_STRING_CAST 70) (T_SWITCH 27250) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27252) (T_UNSET 27253) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27255) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27256) (alt_if_stmt_without_else 27257) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27263) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27266) (if_stmt_without_else 27267) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27425) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27272) ("~" 124)) 13010 (("(" 27278)) 13011 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27279) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13012 (("%" 280) (")" 27280) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13013 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 27281) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27282) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27283) (T_DEC 22) (T_DECLARE 27284) (T_DIR 24) (T_DNUMBER 25) (T_DO 27285) (T_DOUBLE_CAST 27) (T_ECHO 27286) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27287) (T_FOREACH 27288) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27289) (T_GOTO 27290) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27291) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27292) (T_START_HEREDOC 67) (T_STATIC 27293) (T_STRING 27294) (T_STRING_CAST 70) (T_SWITCH 27295) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27296) (T_UNSET 27297) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27298) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27299) (alt_if_stmt_without_else 27300) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27301) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27302) (if_stmt_without_else 27303) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27304) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27305) ("~" 124)) 13014 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 27423) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 13015 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 27421) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 13016 (("(" 27412)) 13017 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27406) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 13018 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 27404) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 13019 (("(" 27391)) 13020 (("(" 27376)) 13021 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 27374) (simple_variable 17412)) 13022 ((T_STRING 27372)) 13023 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 27370) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 13024 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 27368)) 13025 ((":" 27367)) 13026 (("(" 27349)) 13027 (("{" 27332)) 13028 (("(" 27327)) 13029 (("(" 27318)) 13030 ((T_ELSE 27311) (T_ELSEIF 15767) (T_ENDIF 27312)) 13031 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 27310) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 13032 ((T_ELSE 27308) (T_ELSEIF 27277)) 13033 ((inner_statement_list 27306)) 13034 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27307) ("~" 124)) 13035 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 27281) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27282) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27283) (T_DEC 22) (T_DECLARE 27284) (T_DIR 24) (T_DNUMBER 25) (T_DO 27285) (T_DOUBLE_CAST 27) (T_ECHO 27286) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27287) (T_FOREACH 27288) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27289) (T_GOTO 27290) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27291) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27292) (T_START_HEREDOC 67) (T_STATIC 27293) (T_STRING 27294) (T_STRING_CAST 70) (T_SWITCH 27295) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27296) (T_UNSET 27297) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27298) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27299) (alt_if_stmt_without_else 27300) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27301) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27302) (if_stmt_without_else 27303) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27309) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27305) ("~" 124)) 13036 ((":" 27314)) 13037 ((";" 27313)) 13038 ((inner_statement_list 27315)) 13039 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 27316) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 13040 ((";" 27317)) 13041 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27319) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13042 (("%" 280) (")" 27320) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13043 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27321) (";" 27281) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27282) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27283) (T_DEC 22) (T_DECLARE 27284) (T_DIR 24) (T_DNUMBER 25) (T_DO 27285) (T_DOUBLE_CAST 27) (T_ECHO 27286) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27287) (T_FOREACH 27288) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27289) (T_GOTO 27290) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27291) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27292) (T_START_HEREDOC 67) (T_STATIC 27293) (T_STRING 27294) (T_STRING_CAST 70) (T_SWITCH 27295) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27296) (T_UNSET 27297) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27298) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27299) (alt_if_stmt_without_else 27300) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27301) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27302) (if_stmt_without_else 27303) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27322) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 27323) ("{" 27305) ("~" 124)) 13044 ((inner_statement_list 27324)) 13045 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 27325) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 13046 ((";" 27326)) 13047 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 27328) (variable 16049) (variable_class_name 16050)) 13048 (("," 16062) (possible_comma 27329)) 13049 ((")" 27330)) 13050 ((";" 27331)) 13051 ((inner_statement_list 27333)) 13052 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27334) ("~" 124)) 13053 ((catch_list 27335)) 13054 ((T_CATCH 27336) (T_FINALLY 27337) (finally_statement 27338)) 13055 (("(" 27342)) 13056 (("{" 27339)) 13057 ((inner_statement_list 27340)) 13058 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27341) ("~" 124)) 13059 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 27343) (class_name 16111) (name 16112)) 13060 ((T_VARIABLE 16113) (optional_variable 27344) ("|" 16115)) 13061 ((")" 27345)) 13062 (("{" 27346)) 13063 ((inner_statement_list 27347)) 13064 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27348) ("~" 124)) 13065 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27350) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13066 (("%" 280) (")" 27351) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13067 ((":" 27352) (switch_case_list 27353) ("{" 27354)) 13068 ((";" 27360) (case_list 27361)) 13069 ((";" 27355) (case_list 27356)) 13070 ((case_list 27358)) 13071 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 27357)) 13072 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 27359)) 13073 ((case_list 27364)) 13074 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 27362)) 13075 ((";" 27363)) 13076 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 27365)) 13077 ((";" 27366)) 13078 (("," 16412) (";" 27369)) 13079 ((";" 27371)) 13080 ((";" 27373)) 13081 (("," 17413) (";" 27375)) 13082 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 27377) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 13083 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 27378) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 13084 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 27379) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 13085 ((")" 27380) (T_DOUBLE_ARROW 27381)) 13086 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27384) (";" 27281) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27282) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27283) (T_DEC 22) (T_DECLARE 27284) (T_DIR 24) (T_DNUMBER 25) (T_DO 27285) (T_DOUBLE_CAST 27) (T_ECHO 27286) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27287) (T_FOREACH 27288) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27289) (T_GOTO 27290) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27291) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27292) (T_START_HEREDOC 67) (T_STATIC 27293) (T_STRING 27294) (T_STRING_CAST 70) (T_SWITCH 27295) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27296) (T_UNSET 27297) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27298) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27299) (alt_if_stmt_without_else 27300) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27301) (fn 102) (foreach_statement 27390) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27302) (if_stmt_without_else 27303) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27386) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27305) ("~" 124)) 13087 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 27382) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 13088 ((")" 27383)) 13089 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27384) (";" 27281) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27282) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27283) (T_DEC 22) (T_DECLARE 27284) (T_DIR 24) (T_DNUMBER 25) (T_DO 27285) (T_DOUBLE_CAST 27) (T_ECHO 27286) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27287) (T_FOREACH 27288) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27289) (T_GOTO 27290) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27291) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27292) (T_START_HEREDOC 67) (T_STATIC 27293) (T_STRING 27294) (T_STRING_CAST 70) (T_SWITCH 27295) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27296) (T_UNSET 27297) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27298) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27299) (alt_if_stmt_without_else 27300) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27301) (fn 102) (foreach_statement 27385) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27302) (if_stmt_without_else 27303) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27386) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27305) ("~" 124)) 13090 ((inner_statement_list 27387)) 13091 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 27388) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 13092 ((";" 27389)) 13093 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 27392) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 13094 ((";" 27393)) 13095 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 27394) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 13096 ((";" 27395)) 13097 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 27396) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 13098 ((")" 27397)) 13099 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27398) (";" 27281) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27282) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27283) (T_DEC 22) (T_DECLARE 27284) (T_DIR 24) (T_DNUMBER 25) (T_DO 27285) (T_DOUBLE_CAST 27) (T_ECHO 27286) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27287) (T_FOREACH 27288) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27289) (T_GOTO 27290) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27291) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27292) (T_START_HEREDOC 67) (T_STATIC 27293) (T_STRING 27294) (T_STRING_CAST 70) (T_SWITCH 27295) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27296) (T_UNSET 27297) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27298) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27299) (alt_if_stmt_without_else 27300) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27301) (fn 102) (for_statement 27399) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27302) (if_stmt_without_else 27303) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27400) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27305) ("~" 124)) 13100 ((inner_statement_list 27401)) 13101 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 27402) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 13102 ((";" 27403)) 13103 (("," 19251) (";" 27405)) 13104 ((T_WHILE 27407)) 13105 (("(" 27408)) 13106 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27409) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13107 (("%" 280) (")" 27410) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13108 ((";" 27411)) 13109 ((T_STRING 19424) (const_decl 19425) (const_list 27413)) 13110 ((")" 27414) ("," 19428)) 13111 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27415) (";" 27281) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27282) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27283) (T_DEC 22) (T_DECLARE 27284) (T_DIR 24) (T_DNUMBER 25) (T_DO 27285) (T_DOUBLE_CAST 27) (T_ECHO 27286) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27287) (T_FOREACH 27288) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27289) (T_GOTO 27290) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27291) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27292) (T_START_HEREDOC 67) (T_STATIC 27293) (T_STRING 27294) (T_STRING_CAST 70) (T_SWITCH 27295) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27296) (T_UNSET 27297) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27298) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27299) (alt_if_stmt_without_else 27300) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 27416) (dereferenceable_scalar 99) (expr 27301) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27302) (if_stmt_without_else 27303) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27417) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27305) ("~" 124)) 13112 ((inner_statement_list 27418)) 13113 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 27419) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 13114 ((";" 27420)) 13115 ((";" 27422)) 13116 ((";" 27424)) 13117 ((T_STRING 27427) (backup_doc_comment 137)) 13118 ((backup_doc_comment 27428)) 13119 (("(" 27429)) 13120 ((T_ATTRIBUTE 139) (attribute 140) (attributed_parameter 141) (attributes 142) (non_empty_parameter_list 143) (optional_property_modifiers 144) (parameter 145) (parameter_list 27430)) 13121 ((")" 27431)) 13122 ((":" 150) (return_type 27432)) 13123 ((backup_fn_flags 27433)) 13124 (("{" 27434)) 13125 ((inner_statement_list 27435)) 13126 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27436) ("~" 124)) 13127 ((backup_fn_flags 27437)) 13128 ((T_STRING 27440)) 13129 ((T_EXTENDS 2902) (extends_from 27441)) 13130 ((T_IMPLEMENTS 2904) (implements_list 27442)) 13131 ((backup_doc_comment 27443)) 13132 (("{" 27444)) 13133 ((class_statement_list 27445)) 13134 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 27446)) 13135 ((":" 27451)) 13136 ((";" 27450)) 13137 ((inner_statement_list 27452)) 13138 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 27453) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 13139 ((";" 27454)) 13140 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27456) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13141 (("%" 280) (")" 27457) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13142 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27458) (";" 27229) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27230) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27233) (T_DEC 22) (T_DECLARE 27234) (T_DIR 24) (T_DNUMBER 25) (T_DO 27235) (T_DOUBLE_CAST 27) (T_ECHO 27236) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27238) (T_FOREACH 27239) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27240) (T_GOTO 27241) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27247) (T_START_HEREDOC 67) (T_STATIC 27248) (T_STRING 27249) (T_STRING_CAST 70) (T_SWITCH 27250) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27252) (T_UNSET 27253) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27255) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27256) (alt_if_stmt_without_else 27257) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27263) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27266) (if_stmt_without_else 27267) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27459) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 27460) ("{" 27272) ("~" 124)) 13143 ((inner_statement_list 27461)) 13144 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 27462) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 13145 ((";" 27463)) 13146 ((";" 27472)) 13147 (("," 27022) (";" 27471)) 13148 ((T_NAME_FULLY_QUALIFIED 27010) (T_NAME_QUALIFIED 27011) (T_STRING 27012) (group_use_declaration 27467) (legacy_namespace_name 27020) (namespace_name 27015) (use_declaration 27016) (use_declarations 27468)) 13149 ((";" 27470)) 13150 (("," 27022) (";" 27469)) 13151 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 27474) (variable 16049) (variable_class_name 16050)) 13152 (("," 16062) (possible_comma 27475)) 13153 ((")" 27476)) 13154 ((";" 27477)) 13155 ((inner_statement_list 27479)) 13156 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27480) ("~" 124)) 13157 ((catch_list 27481)) 13158 ((T_CATCH 27482) (T_FINALLY 27483) (finally_statement 27484)) 13159 (("(" 27488)) 13160 (("{" 27485)) 13161 ((inner_statement_list 27486)) 13162 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27487) ("~" 124)) 13163 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 27489) (class_name 16111) (name 16112)) 13164 ((T_VARIABLE 16113) (optional_variable 27490) ("|" 16115)) 13165 ((")" 27491)) 13166 (("{" 27492)) 13167 ((inner_statement_list 27493)) 13168 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27494) ("~" 124)) 13169 ((backup_doc_comment 27496)) 13170 (("{" 27497)) 13171 ((class_statement_list 27498)) 13172 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 27499)) 13173 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27501) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13174 (("%" 280) (")" 27502) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13175 ((":" 27503) (switch_case_list 27504) ("{" 27505)) 13176 ((";" 27511) (case_list 27512)) 13177 ((";" 27506) (case_list 27507)) 13178 ((case_list 27509)) 13179 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 27508)) 13180 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 27510)) 13181 ((case_list 27515)) 13182 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 27513)) 13183 ((";" 27514)) 13184 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 27516)) 13185 ((";" 27517)) 13186 (("," 16412) (";" 27520)) 13187 ((";" 27522)) 13188 ((";" 27527) ("{" 27528)) 13189 ((top_statement_list 27525)) 13190 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 27229) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 27230) (T_CLASS 27231) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 27232) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27233) (T_DEC 22) (T_DECLARE 27234) (T_DIR 24) (T_DNUMBER 25) (T_DO 27235) (T_DOUBLE_CAST 27) (T_ECHO 27236) (T_EMPTY 29) (T_ENUM 27237) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 27238) (T_FOREACH 27239) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 27240) (T_GOTO 27241) (T_HALT_COMPILER 27242) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27244) (T_INTERFACE 27245) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 27246) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27247) (T_START_HEREDOC 67) (T_STATIC 27248) (T_STRING 27249) (T_STRING_CAST 70) (T_SWITCH 27250) (T_THROW 72) (T_TRAIT 27251) (T_TRAIT_C 74) (T_TRY 27252) (T_UNSET 27253) (T_UNSET_CAST 77) (T_USE 27254) (T_VARIABLE 79) (T_WHILE 27255) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27256) (alt_if_stmt_without_else 27257) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 27258) (attributes 27259) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 27260) (class_modifier 95) (class_modifiers 27261) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 27262) (expr 27263) (fn 102) (fully_dereferenceable 103) #'27264 (function_call 105) (function_declaration_statement 27265) (if_stmt 27266) (if_stmt_without_else 27267) (inline_function 109) (interface_declaration_statement 27268) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27269) (static_member 118) (top_statement 27270) (trait_declaration_statement 27271) (variable 121) (variable_class_name 122) ("{" 27272) ("}" 27526) ("~" 124)) 13191 ((top_statement_list 27529)) 13192 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 27229) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 27230) (T_CLASS 27231) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 27232) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27233) (T_DEC 22) (T_DECLARE 27234) (T_DIR 24) (T_DNUMBER 25) (T_DO 27235) (T_DOUBLE_CAST 27) (T_ECHO 27236) (T_EMPTY 29) (T_ENUM 27237) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 27238) (T_FOREACH 27239) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 27240) (T_GOTO 27241) (T_HALT_COMPILER 27242) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27244) (T_INTERFACE 27245) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 27246) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27247) (T_START_HEREDOC 67) (T_STATIC 27248) (T_STRING 27249) (T_STRING_CAST 70) (T_SWITCH 27250) (T_THROW 72) (T_TRAIT 27251) (T_TRAIT_C 74) (T_TRY 27252) (T_UNSET 27253) (T_UNSET_CAST 77) (T_USE 27254) (T_VARIABLE 79) (T_WHILE 27255) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27256) (alt_if_stmt_without_else 27257) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 27258) (attributes 27259) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 27260) (class_modifier 95) (class_modifiers 27261) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 27262) (expr 27263) (fn 102) (fully_dereferenceable 103) #'27264 (function_call 105) (function_declaration_statement 27265) (if_stmt 27266) (if_stmt_without_else 27267) (inline_function 109) (interface_declaration_statement 27268) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27269) (static_member 118) (top_statement 27270) (trait_declaration_statement 27271) (variable 121) (variable_class_name 122) ("{" 27272) ("}" 27530) ("~" 124)) 13193 ((T_EXTENDS 19717) (interface_extends_list 27532)) 13194 ((backup_doc_comment 27533)) 13195 (("{" 27534)) 13196 ((class_statement_list 27535)) 13197 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 27536)) 13198 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27538) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13199 (("%" 280) (")" 27539) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13200 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19727) (";" 27281) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27282) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27283) (T_DEC 22) (T_DECLARE 27284) (T_DIR 24) (T_DNUMBER 25) (T_DO 27285) (T_DOUBLE_CAST 27) (T_ECHO 27286) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27287) (T_FOREACH 27288) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27289) (T_GOTO 27290) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27291) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27292) (T_START_HEREDOC 67) (T_STATIC 27293) (T_STRING 27294) (T_STRING_CAST 70) (T_SWITCH 27295) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27296) (T_UNSET 27297) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27298) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27299) (alt_if_stmt_without_else 27300) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27301) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27302) (if_stmt_without_else 27303) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27540) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27305) ("~" 124)) 13201 ((")" 27542)) 13202 ((";" 27543)) 13203 ((";" 27545)) 13204 (("," 17413) (";" 27547)) 13205 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 27549) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 13206 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 27550) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 13207 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 27551) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 13208 ((")" 27552) (T_DOUBLE_ARROW 27553)) 13209 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27556) (";" 27229) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27230) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27233) (T_DEC 22) (T_DECLARE 27234) (T_DIR 24) (T_DNUMBER 25) (T_DO 27235) (T_DOUBLE_CAST 27) (T_ECHO 27236) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27238) (T_FOREACH 27239) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27240) (T_GOTO 27241) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27247) (T_START_HEREDOC 67) (T_STATIC 27248) (T_STRING 27249) (T_STRING_CAST 70) (T_SWITCH 27250) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27252) (T_UNSET 27253) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27255) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27256) (alt_if_stmt_without_else 27257) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27263) (fn 102) (foreach_statement 27562) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27266) (if_stmt_without_else 27267) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27558) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27272) ("~" 124)) 13210 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 27554) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 13211 ((")" 27555)) 13212 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27556) (";" 27229) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27230) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27233) (T_DEC 22) (T_DECLARE 27234) (T_DIR 24) (T_DNUMBER 25) (T_DO 27235) (T_DOUBLE_CAST 27) (T_ECHO 27236) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27238) (T_FOREACH 27239) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27240) (T_GOTO 27241) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27247) (T_START_HEREDOC 67) (T_STATIC 27248) (T_STRING 27249) (T_STRING_CAST 70) (T_SWITCH 27250) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27252) (T_UNSET 27253) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27255) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27256) (alt_if_stmt_without_else 27257) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27263) (fn 102) (foreach_statement 27557) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27266) (if_stmt_without_else 27267) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27558) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27272) ("~" 124)) 13213 ((inner_statement_list 27559)) 13214 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 27560) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 13215 ((";" 27561)) 13216 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 27564) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 13217 ((";" 27565)) 13218 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 27566) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 13219 ((";" 27567)) 13220 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 27568) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 13221 ((")" 27569)) 13222 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27570) (";" 27229) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27230) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27233) (T_DEC 22) (T_DECLARE 27234) (T_DIR 24) (T_DNUMBER 25) (T_DO 27235) (T_DOUBLE_CAST 27) (T_ECHO 27236) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27238) (T_FOREACH 27239) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27240) (T_GOTO 27241) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27247) (T_START_HEREDOC 67) (T_STATIC 27248) (T_STRING 27249) (T_STRING_CAST 70) (T_SWITCH 27250) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27252) (T_UNSET 27253) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27255) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27256) (alt_if_stmt_without_else 27257) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27263) (fn 102) (for_statement 27571) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27266) (if_stmt_without_else 27267) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27572) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27272) ("~" 124)) 13223 ((inner_statement_list 27573)) 13224 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 27574) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 13225 ((";" 27575)) 13226 ((":" 19766) (enum_backing_type 27577)) 13227 ((T_IMPLEMENTS 2904) (implements_list 27578)) 13228 ((backup_doc_comment 27579)) 13229 (("{" 27580)) 13230 ((class_statement_list 27581)) 13231 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 27582)) 13232 (("," 19251) (";" 27584)) 13233 ((T_WHILE 27586)) 13234 (("(" 27587)) 13235 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27588) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13236 (("%" 280) (")" 27589) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13237 ((";" 27590)) 13238 ((T_STRING 19424) (const_decl 19425) (const_list 27592)) 13239 ((")" 27593) ("," 19428)) 13240 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27594) (";" 27229) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27230) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27233) (T_DEC 22) (T_DECLARE 27234) (T_DIR 24) (T_DNUMBER 25) (T_DO 27235) (T_DOUBLE_CAST 27) (T_ECHO 27236) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27238) (T_FOREACH 27239) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27240) (T_GOTO 27241) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27247) (T_START_HEREDOC 67) (T_STATIC 27248) (T_STRING 27249) (T_STRING_CAST 70) (T_SWITCH 27250) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27252) (T_UNSET 27253) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27255) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27256) (alt_if_stmt_without_else 27257) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 27595) (dereferenceable_scalar 99) (expr 27263) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27266) (if_stmt_without_else 27267) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27596) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27272) ("~" 124)) 13241 ((inner_statement_list 27597)) 13242 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 27598) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 13243 ((";" 27599)) 13244 ((";" 27601)) 13245 (("=" 27608)) 13246 (("," 27605) (";" 27606)) 13247 ((T_STRING 27602) (const_decl 27607)) 13248 (("!" 2971) ("\"" 2972) ("$" 2973) ("(" 2974) ("+" 2975) ("-" 2976) ("@" 2977) (T_ARRAY 2978) (T_ARRAY_CAST 2979) (T_ATTRIBUTE 125) (T_BOOL_CAST 2980) (T_CLASS_C 2981) (T_CLONE 2982) (T_CONSTANT_ENCAPSED_STRING 2983) (T_DEC 2984) (T_DIR 2985) (T_DNUMBER 2986) (T_DOUBLE_CAST 2987) (T_EMPTY 2988) (T_EVAL 2989) (T_EXIT 2990) (T_FILE 2991) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2992) (T_INC 2993) (T_INCLUDE 2994) (T_INCLUDE_ONCE 2995) (T_INT_CAST 2996) (T_ISSET 2997) (T_LINE 2998) (T_LIST 2999) (T_LNUMBER 3000) (T_MATCH 3001) (T_METHOD_C 3002) (T_NAME_FULLY_QUALIFIED 3003) (T_NAME_QUALIFIED 3004) (T_NAME_RELATIVE 3005) (T_NEW 3006) (T_NS_C 3007) (T_OBJECT_CAST 3008) (T_PRINT 3009) (T_REQUIRE 3010) (T_REQUIRE_ONCE 3011) (T_START_HEREDOC 3012) (T_STATIC 3013) (T_STRING 3014) (T_STRING_CAST 3015) (T_THROW 3016) (T_TRAIT_C 3017) (T_UNSET_CAST 3018) (T_VARIABLE 3019) (T_YIELD 3020) (T_YIELD_FROM 3021) ("[" 3022) ("`" 3023) (array_object_dereferenceable 3024) (attribute 129) (attributes 3025) (callable_expr 3026) (callable_variable 3027) (class_constant 3028) (class_name 3029) (constant 3030) (dereferenceable_scalar 3031) (expr 27609) (fn 3033) (fully_dereferenceable 103) #'3034 (function_call 3035) (inline_function 3036) (internal_functions_in_yacc 3037) (match 3038) (name 3039) (new_expr 3040) (scalar 3041) (simple_variable 3042) (static_member 3043) (variable 3044) (variable_class_name 3045) ("~" 3046)) 13249 (("%" 3048) ("*" 3049) ("+" 3050) ("-" 3051) ("." 3052) ("/" 3053) ("<" 3054) (">" 3055) ("?" 3056) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3057) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3058) (T_BOOLEAN_AND 3059) (T_BOOLEAN_OR 3060) (T_COALESCE 3061) (T_INSTANCEOF 3062) (T_IS_EQUAL 3063) (T_IS_GREATER_OR_EQUAL 3064) (T_IS_IDENTICAL 3065) (T_IS_NOT_EQUAL 3066) (T_IS_NOT_IDENTICAL 3067) (T_IS_SMALLER_OR_EQUAL 3068) (T_LOGICAL_AND 3069) (T_LOGICAL_OR 3070) (T_LOGICAL_XOR 3071) (T_POW 3072) (T_SL 3073) (T_SPACESHIP 3074) (T_SR 3075) ("^" 3076) (backup_doc_comment 27610) ("|" 3077)) 13250 ((T_EXTENDS 2902) (extends_from 27612)) 13251 ((T_IMPLEMENTS 2904) (implements_list 27613)) 13252 ((backup_doc_comment 27614)) 13253 (("{" 27615)) 13254 ((class_statement_list 27616)) 13255 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 27617)) 13256 ((";" 27619)) 13257 ((top_statement_list 27622)) 13258 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 27229) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 27230) (T_CLASS 27231) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 27232) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27233) (T_DEC 22) (T_DECLARE 27234) (T_DIR 24) (T_DNUMBER 25) (T_DO 27235) (T_DOUBLE_CAST 27) (T_ECHO 27236) (T_EMPTY 29) (T_ENUM 27237) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 27238) (T_FOREACH 27239) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 27240) (T_GOTO 27241) (T_HALT_COMPILER 27242) (T_IF 27243) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27244) (T_INTERFACE 27245) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 27246) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27247) (T_START_HEREDOC 67) (T_STATIC 27248) (T_STRING 27249) (T_STRING_CAST 70) (T_SWITCH 27250) (T_THROW 72) (T_TRAIT 27251) (T_TRAIT_C 74) (T_TRY 27252) (T_UNSET 27253) (T_UNSET_CAST 77) (T_USE 27254) (T_VARIABLE 79) (T_WHILE 27255) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27256) (alt_if_stmt_without_else 27257) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 27258) (attributes 27259) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 27260) (class_modifier 95) (class_modifiers 27261) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 27262) (expr 27263) (fn 102) (fully_dereferenceable 103) #'27264 (function_call 105) (function_declaration_statement 27265) (if_stmt 27266) (if_stmt_without_else 27267) (inline_function 109) (interface_declaration_statement 27268) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27269) (static_member 118) (top_statement 27270) (trait_declaration_statement 27271) (variable 121) (variable_class_name 122) ("{" 27272) ("}" 27623) ("~" 124)) 13259 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27625) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13260 (("%" 280) (")" 27626) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13261 (("{" 27627)) 13262 (("!" 3857) ("\"" 3858) ("$" 3859) ("(" 3860) ("+" 3861) ("-" 3862) ("@" 3863) (T_ARRAY 3864) (T_ARRAY_CAST 3865) (T_ATTRIBUTE 125) (T_BOOL_CAST 3866) (T_CLASS_C 3867) (T_CLONE 3868) (T_CONSTANT_ENCAPSED_STRING 3869) (T_DEC 3870) (T_DEFAULT 3871) (T_DIR 3872) (T_DNUMBER 3873) (T_DOUBLE_CAST 3874) (T_EMPTY 3875) (T_EVAL 3876) (T_EXIT 3877) (T_FILE 3878) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 3879) (T_INC 3880) (T_INCLUDE 3881) (T_INCLUDE_ONCE 3882) (T_INT_CAST 3883) (T_ISSET 3884) (T_LINE 3885) (T_LIST 3886) (T_LNUMBER 3887) (T_MATCH 3888) (T_METHOD_C 3889) (T_NAME_FULLY_QUALIFIED 3890) (T_NAME_QUALIFIED 3891) (T_NAME_RELATIVE 3892) (T_NEW 3893) (T_NS_C 3894) (T_OBJECT_CAST 3895) (T_PRINT 3896) (T_REQUIRE 3897) (T_REQUIRE_ONCE 3898) (T_START_HEREDOC 3899) (T_STATIC 3900) (T_STRING 3901) (T_STRING_CAST 3902) (T_THROW 3903) (T_TRAIT_C 3904) (T_UNSET_CAST 3905) (T_VARIABLE 3906) (T_YIELD 3907) (T_YIELD_FROM 3908) ("[" 3909) ("`" 3910) (array_object_dereferenceable 3911) (attribute 129) (attributes 3912) (callable_expr 3913) (callable_variable 3914) (class_constant 3915) (class_name 3916) (constant 3917) (dereferenceable_scalar 3918) (expr 3919) (fn 3920) (fully_dereferenceable 103) #'3921 (function_call 3922) (inline_function 3923) (internal_functions_in_yacc 3924) (match 3925) (match_arm 3926) (match_arm_cond_list 3927) (match_arm_list 27628) (name 3929) (new_expr 3930) (non_empty_match_arm_list 3931) (scalar 3932) (simple_variable 3933) (static_member 3934) (variable 3935) (variable_class_name 3936) ("~" 3937)) 13263 (("}" 27629)) 13264 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 27631) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 13265 ((")" 27632)) 13266 (("=" 27633)) 13267 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27634) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 13268 (("!" 919) ("\"" 920) ("$" 921) ("(" 922) ("+" 924) ("-" 925) ("@" 926) (T_ARRAY 1054) (T_ARRAY_CAST 929) (T_ATTRIBUTE 125) (T_BOOL_CAST 931) (T_CLASS_C 1055) (T_CLONE 1056) (T_CONSTANT_ENCAPSED_STRING 940) (T_DEC 942) (T_DIR 1057) (T_DNUMBER 946) (T_DOUBLE_CAST 948) (T_EMPTY 1058) (T_EVAL 1059) (T_EXIT 1060) (T_FILE 1061) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1062) (T_INC 976) (T_INCLUDE 1063) (T_INCLUDE_ONCE 1064) (T_INT_CAST 982) (T_ISSET 1065) (T_LINE 1066) (T_LIST 1067) (T_LNUMBER 986) (T_MATCH 1068) (T_METHOD_C 1069) (T_NAME_FULLY_QUALIFIED 993) (T_NAME_QUALIFIED 994) (T_NAME_RELATIVE 995) (T_NEW 1070) (T_NS_C 1071) (T_OBJECT_CAST 998) (T_PRINT 1072) (T_REQUIRE 1073) (T_REQUIRE_ONCE 1074) (T_START_HEREDOC 1007) (T_STATIC 1075) (T_STRING 1076) (T_STRING_CAST 1010) (T_THROW 1077) (T_TRAIT_C 1078) (T_UNSET_CAST 1017) (T_VARIABLE 1020) (T_YIELD 1079) (T_YIELD_FROM 1023) ("[" 1024) ("`" 1025) (array_object_dereferenceable 1027) (attribute 129) (attributes 1028) (callable_expr 1029) (callable_variable 1030) (class_constant 1031) (class_name 1032) (constant 1033) (dereferenceable_scalar 1034) (expr 5114) (fn 1036) (fully_dereferenceable 103) #'1037 (function_call 1038) (inline_function 1040) (internal_functions_in_yacc 1041) (isset_variable 5115) (isset_variables 27636) (match 1042) (name 1043) (new_expr 1044) (scalar 1047) (simple_variable 1049) (static_member 1050) (variable 1051) (variable_class_name 1052) ("~" 1053)) 13269 (("," 5117) (possible_comma 27637)) 13270 ((")" 27638)) 13271 ((T_EXTENDS 19717) (interface_extends_list 27641)) 13272 ((backup_doc_comment 27642)) 13273 (("{" 27643)) 13274 ((class_statement_list 27644)) 13275 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 27645)) 13276 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27650) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13277 (("%" 280) (")" 27651) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13278 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19727) (";" 27652) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27653) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27654) (T_DEC 22) (T_DECLARE 27655) (T_DIR 24) (T_DNUMBER 25) (T_DO 27656) (T_DOUBLE_CAST 27) (T_ECHO 27657) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27658) (T_FOREACH 27659) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27660) (T_GOTO 27661) (T_IF 27662) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27663) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27664) (T_START_HEREDOC 67) (T_STATIC 27665) (T_STRING 27666) (T_STRING_CAST 70) (T_SWITCH 27667) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27668) (T_UNSET 27669) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27670) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27671) (alt_if_stmt_without_else 27672) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27673) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27674) (if_stmt_without_else 27675) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27676) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27677) ("~" 124)) 13279 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 27804) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 13280 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 27802) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 13281 (("(" 27793)) 13282 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19254) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 19255) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19256) (T_DEC 22) (T_DECLARE 19257) (T_DIR 24) (T_DNUMBER 25) (T_DO 19258) (T_DOUBLE_CAST 27) (T_ECHO 19259) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 19260) (T_FOREACH 19261) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 19262) (T_GOTO 19263) (T_IF 19264) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19265) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19266) (T_START_HEREDOC 67) (T_STATIC 19267) (T_STRING 19268) (T_STRING_CAST 70) (T_SWITCH 19269) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 19270) (T_UNSET 19271) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19272) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19273) (alt_if_stmt_without_else 19274) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 19275) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 19276) (if_stmt_without_else 19277) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27787) (static_member 118) (variable 121) (variable_class_name 122) ("{" 19279) ("~" 124)) 13283 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (echo_expr 19248) (echo_expr_list 27785) (expr 19250) (fn 16478) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 13284 (("(" 27772)) 13285 (("(" 27757)) 13286 (("$" 17408) (T_VARIABLE 17409) (global_var 17410) (global_var_list 27755) (simple_variable 17412)) 13287 ((T_STRING 27753)) 13288 (("(" 27749)) 13289 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_START_HEREDOC 67) (T_STATIC 127) (T_STRING 128) (T_STRING_CAST 70) (T_THROW 72) (T_TRAIT_C 74) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 17403) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (optional_expr 27747) (scalar 115) (simple_variable 116) (static_member 118) (variable 121) (variable_class_name 122) ("~" 124)) 13290 ((T_FN 35) (T_FUNCTION 126) (T_VARIABLE 16408) (fn 102) #'132 (inline_function 16409) (static_var 16410) (static_var_list 27745)) 13291 ((":" 27744)) 13292 (("(" 27726)) 13293 (("{" 27709)) 13294 (("(" 27704)) 13295 (("(" 27695)) 13296 ((T_ELSE 27688) (T_ELSEIF 15767) (T_ENDIF 27689)) 13297 (("%" 6249) ("*" 6250) ("+" 6251) ("-" 6252) ("." 6253) ("/" 6254) (";" 27687) ("<" 6255) (">" 6256) ("?" 6257) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6258) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6259) (T_BOOLEAN_AND 6260) (T_BOOLEAN_OR 6261) (T_COALESCE 6262) (T_INSTANCEOF 6263) (T_IS_EQUAL 6264) (T_IS_GREATER_OR_EQUAL 6265) (T_IS_IDENTICAL 6266) (T_IS_NOT_EQUAL 6267) (T_IS_NOT_IDENTICAL 6268) (T_IS_SMALLER_OR_EQUAL 6269) (T_LOGICAL_AND 6270) (T_LOGICAL_OR 6271) (T_LOGICAL_XOR 6272) (T_POW 6273) (T_SL 6274) (T_SPACESHIP 6275) (T_SR 6276) ("^" 6277) ("|" 6278)) 13298 ((T_ELSE 27680) (T_ELSEIF 27681)) 13299 ((inner_statement_list 27678)) 13300 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27679) ("~" 124)) 13301 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 27652) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27653) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27654) (T_DEC 22) (T_DECLARE 27655) (T_DIR 24) (T_DNUMBER 25) (T_DO 27656) (T_DOUBLE_CAST 27) (T_ECHO 27657) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27658) (T_FOREACH 27659) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27660) (T_GOTO 27661) (T_IF 27662) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27663) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27664) (T_START_HEREDOC 67) (T_STATIC 27665) (T_STRING 27666) (T_STRING_CAST 70) (T_SWITCH 27667) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27668) (T_UNSET 27669) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27670) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27671) (alt_if_stmt_without_else 27672) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27673) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27674) (if_stmt_without_else 27675) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27686) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27677) ("~" 124)) 13302 (("(" 27682)) 13303 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27683) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13304 (("%" 280) (")" 27684) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13305 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 27652) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27653) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27654) (T_DEC 22) (T_DECLARE 27655) (T_DIR 24) (T_DNUMBER 25) (T_DO 27656) (T_DOUBLE_CAST 27) (T_ECHO 27657) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27658) (T_FOREACH 27659) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27660) (T_GOTO 27661) (T_IF 27662) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27663) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27664) (T_START_HEREDOC 67) (T_STATIC 27665) (T_STRING 27666) (T_STRING_CAST 70) (T_SWITCH 27667) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27668) (T_UNSET 27669) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27670) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27671) (alt_if_stmt_without_else 27672) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27673) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27674) (if_stmt_without_else 27675) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27685) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27677) ("~" 124)) 13306 ((":" 27691)) 13307 ((";" 27690)) 13308 ((inner_statement_list 27692)) 13309 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15860) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15861) (T_CLASS 15862) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15863) (T_DEC 22) (T_DECLARE 15864) (T_DIR 24) (T_DNUMBER 25) (T_DO 15865) (T_DOUBLE_CAST 27) (T_ECHO 15866) (T_EMPTY 29) (T_ENDIF 27693) (T_ENUM 15868) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15869) (T_FOREACH 15870) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15871) (T_GOTO 15872) (T_HALT_COMPILER 15873) (T_IF 15788) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15874) (T_INTERFACE 15875) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15876) (T_START_HEREDOC 67) (T_STATIC 15877) (T_STRING 15878) (T_STRING_CAST 70) (T_SWITCH 15879) (T_THROW 72) (T_TRAIT 15880) (T_TRAIT_C 74) (T_TRY 15881) (T_UNSET 15882) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15883) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15884) (alt_if_stmt_without_else 15885) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15886) (attributes 15887) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15888) (class_modifier 95) (class_modifiers 15889) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15890) (expr 15891) (fn 102) (fully_dereferenceable 103) #'15892 (function_call 105) (function_declaration_statement 15893) (if_stmt 15894) (if_stmt_without_else 15895) (inline_function 109) (inner_statement 15896) (interface_declaration_statement 15897) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15898) (static_member 118) (trait_declaration_statement 15899) (variable 121) (variable_class_name 122) ("{" 15900) ("~" 124)) 13310 ((";" 27694)) 13311 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27696) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13312 (("%" 280) (")" 27697) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13313 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27698) (";" 27652) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27653) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27654) (T_DEC 22) (T_DECLARE 27655) (T_DIR 24) (T_DNUMBER 25) (T_DO 27656) (T_DOUBLE_CAST 27) (T_ECHO 27657) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27658) (T_FOREACH 27659) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27660) (T_GOTO 27661) (T_IF 27662) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27663) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27664) (T_START_HEREDOC 67) (T_STATIC 27665) (T_STRING 27666) (T_STRING_CAST 70) (T_SWITCH 27667) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27668) (T_UNSET 27669) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27670) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27671) (alt_if_stmt_without_else 27672) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27673) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27674) (if_stmt_without_else 27675) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27699) (static_member 118) (variable 121) (variable_class_name 122) (while_statement 27700) ("{" 27677) ("~" 124)) 13314 ((inner_statement_list 27701)) 13315 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 15941) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15942) (T_CLASS 15943) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 15944) (T_DEC 22) (T_DECLARE 15945) (T_DIR 24) (T_DNUMBER 25) (T_DO 15946) (T_DOUBLE_CAST 27) (T_ECHO 15947) (T_EMPTY 29) (T_ENDWHILE 27702) (T_ENUM 15949) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 15950) (T_FOREACH 15951) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 15952) (T_GOTO 15953) (T_HALT_COMPILER 15954) (T_IF 15955) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 15956) (T_INTERFACE 15957) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 15958) (T_START_HEREDOC 67) (T_STATIC 15959) (T_STRING 15960) (T_STRING_CAST 70) (T_SWITCH 15961) (T_THROW 72) (T_TRAIT 15962) (T_TRAIT_C 74) (T_TRY 15963) (T_UNSET 15964) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 15965) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 15966) (alt_if_stmt_without_else 15967) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 15968) (attributes 15969) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 15970) (class_modifier 95) (class_modifiers 15971) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 15972) (expr 15973) (fn 102) (fully_dereferenceable 103) #'15974 (function_call 105) (function_declaration_statement 15975) (if_stmt 15976) (if_stmt_without_else 15977) (inline_function 109) (inner_statement 15978) (interface_declaration_statement 15979) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 15980) (static_member 118) (trait_declaration_statement 15981) (variable 121) (variable_class_name 122) ("{" 15982) ("~" 124)) 13316 ((";" 27703)) 13317 (("\"" 1489) ("$" 16037) ("(" 1491) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 16038) ("[" 1508) (array_object_dereferenceable 16039) (callable_expr 16040) (callable_variable 16041) (class_constant 1512) (class_name 16042) (constant 1514) (dereferenceable_scalar 1515) (fully_dereferenceable 103) (function_call 16043) (name 16044) (simple_variable 16045) (static_member 16046) (unset_variable 16047) (unset_variables 27705) (variable 16049) (variable_class_name 16050)) 13318 (("," 16062) (possible_comma 27706)) 13319 ((")" 27707)) 13320 ((";" 27708)) 13321 ((inner_statement_list 27710)) 13322 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27711) ("~" 124)) 13323 ((catch_list 27712)) 13324 ((T_CATCH 27713) (T_FINALLY 27714) (finally_statement 27715)) 13325 (("(" 27719)) 13326 (("{" 27716)) 13327 ((inner_statement_list 27717)) 13328 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27718) ("~" 124)) 13329 ((T_NAME_FULLY_QUALIFIED 16105) (T_NAME_QUALIFIED 16106) (T_NAME_RELATIVE 16107) (T_STATIC 16108) (T_STRING 16109) (catch_name_list 27720) (class_name 16111) (name 16112)) 13330 ((T_VARIABLE 16113) (optional_variable 27721) ("|" 16115)) 13331 ((")" 27722)) 13332 (("{" 27723)) 13333 ((inner_statement_list 27724)) 13334 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 155) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 156) (T_CLASS 157) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 158) (T_DEC 22) (T_DECLARE 159) (T_DIR 24) (T_DNUMBER 25) (T_DO 160) (T_DOUBLE_CAST 27) (T_ECHO 161) (T_EMPTY 29) (T_ENUM 162) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 163) (T_FOREACH 164) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 165) (T_GOTO 166) (T_HALT_COMPILER 167) (T_IF 168) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 169) (T_INTERFACE 170) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 171) (T_START_HEREDOC 67) (T_STATIC 172) (T_STRING 173) (T_STRING_CAST 70) (T_SWITCH 174) (T_THROW 72) (T_TRAIT 175) (T_TRAIT_C 74) (T_TRY 176) (T_UNSET 177) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 178) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 179) (alt_if_stmt_without_else 180) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 181) (attributes 182) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 183) (class_modifier 95) (class_modifiers 184) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 185) (expr 186) (fn 102) (fully_dereferenceable 103) #'187 (function_call 105) (function_declaration_statement 188) (if_stmt 189) (if_stmt_without_else 190) (inline_function 109) (inner_statement 191) (interface_declaration_statement 192) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 193) (static_member 118) (trait_declaration_statement 194) (variable 121) (variable_class_name 122) ("{" 195) ("}" 27725) ("~" 124)) 13335 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27727) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13336 (("%" 280) (")" 27728) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13337 ((":" 27729) (switch_case_list 27730) ("{" 27731)) 13338 ((";" 27737) (case_list 27738)) 13339 ((";" 27732) (case_list 27733)) 13340 ((case_list 27735)) 13341 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 27734)) 13342 ((T_CASE 16129) (T_DEFAULT 16130) ("}" 27736)) 13343 ((case_list 27741)) 13344 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 27739)) 13345 ((";" 27740)) 13346 ((T_CASE 16265) (T_DEFAULT 16266) (T_ENDSWITCH 27742)) 13347 ((";" 27743)) 13348 (("," 16412) (";" 27746)) 13349 ((";" 27748)) 13350 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27750) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13351 (("%" 280) (")" 27751) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13352 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 19727) (";" 25194) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 25195) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 25196) (T_DEC 22) (T_DECLARE 25197) (T_DIR 24) (T_DNUMBER 25) (T_DO 25198) (T_DOUBLE_CAST 27) (T_ECHO 25199) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 25200) (T_FOREACH 25201) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 25202) (T_GOTO 25203) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 25204) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 25205) (T_START_HEREDOC 67) (T_STATIC 25206) (T_STRING 25207) (T_STRING_CAST 70) (T_SWITCH 25208) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 25209) (T_UNSET 25210) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 25211) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 25212) (alt_if_stmt_without_else 25213) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 25214) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 25215) (if_stmt_without_else 25216) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27752) (static_member 118) (variable 121) (variable_class_name 122) ("{" 25218) ("~" 124)) 13353 ((";" 27754)) 13354 (("," 17413) (";" 27756)) 13355 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 27758) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 13356 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 27759) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 13357 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 27760) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 13358 ((")" 27761) (T_DOUBLE_ARROW 27762)) 13359 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27765) (";" 27652) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27653) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27654) (T_DEC 22) (T_DECLARE 27655) (T_DIR 24) (T_DNUMBER 25) (T_DO 27656) (T_DOUBLE_CAST 27) (T_ECHO 27657) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27658) (T_FOREACH 27659) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27660) (T_GOTO 27661) (T_IF 27662) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27663) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27664) (T_START_HEREDOC 67) (T_STATIC 27665) (T_STRING 27666) (T_STRING_CAST 70) (T_SWITCH 27667) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27668) (T_UNSET 27669) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27670) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27671) (alt_if_stmt_without_else 27672) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27673) (fn 102) (foreach_statement 27771) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27674) (if_stmt_without_else 27675) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27767) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27677) ("~" 124)) 13360 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 27763) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 13361 ((")" 27764)) 13362 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27765) (";" 27652) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27653) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27654) (T_DEC 22) (T_DECLARE 27655) (T_DIR 24) (T_DNUMBER 25) (T_DO 27656) (T_DOUBLE_CAST 27) (T_ECHO 27657) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27658) (T_FOREACH 27659) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27660) (T_GOTO 27661) (T_IF 27662) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27663) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27664) (T_START_HEREDOC 67) (T_STATIC 27665) (T_STRING 27666) (T_STRING_CAST 70) (T_SWITCH 27667) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27668) (T_UNSET 27669) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27670) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27671) (alt_if_stmt_without_else 27672) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27673) (fn 102) (foreach_statement 27766) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27674) (if_stmt_without_else 27675) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27767) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27677) ("~" 124)) 13363 ((inner_statement_list 27768)) 13364 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 27769) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 13365 ((";" 27770)) 13366 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 27773) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 13367 ((";" 27774)) 13368 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 27775) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 13369 ((";" 27776)) 13370 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 27777) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 13371 ((")" 27778)) 13372 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27779) (";" 27652) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27653) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27654) (T_DEC 22) (T_DECLARE 27655) (T_DIR 24) (T_DNUMBER 25) (T_DO 27656) (T_DOUBLE_CAST 27) (T_ECHO 27657) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27658) (T_FOREACH 27659) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27660) (T_GOTO 27661) (T_IF 27662) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27663) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27664) (T_START_HEREDOC 67) (T_STATIC 27665) (T_STRING 27666) (T_STRING_CAST 70) (T_SWITCH 27667) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27668) (T_UNSET 27669) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27670) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27671) (alt_if_stmt_without_else 27672) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 27673) (fn 102) (for_statement 27780) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27674) (if_stmt_without_else 27675) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27781) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27677) ("~" 124)) 13373 ((inner_statement_list 27782)) 13374 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 27783) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 13375 ((";" 27784)) 13376 (("," 19251) (";" 27786)) 13377 ((T_WHILE 27788)) 13378 (("(" 27789)) 13379 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27790) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13380 (("%" 280) (")" 27791) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13381 ((";" 27792)) 13382 ((T_STRING 19424) (const_decl 19425) (const_list 27794)) 13383 ((")" 27795) ("," 19428)) 13384 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27796) (";" 27652) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 27653) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 27654) (T_DEC 22) (T_DECLARE 27655) (T_DIR 24) (T_DNUMBER 25) (T_DO 27656) (T_DOUBLE_CAST 27) (T_ECHO 27657) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 27658) (T_FOREACH 27659) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 27660) (T_GOTO 27661) (T_IF 27662) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 27663) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 27664) (T_START_HEREDOC 67) (T_STATIC 27665) (T_STRING 27666) (T_STRING_CAST 70) (T_SWITCH 27667) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 27668) (T_UNSET 27669) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 27670) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 27671) (alt_if_stmt_without_else 27672) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 27797) (dereferenceable_scalar 99) (expr 27673) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 27674) (if_stmt_without_else 27675) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27798) (static_member 118) (variable 121) (variable_class_name 122) ("{" 27677) ("~" 124)) 13385 ((inner_statement_list 27799)) 13386 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 27800) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 13387 ((";" 27801)) 13388 ((";" 27803)) 13389 ((";" 27805)) 13390 ((")" 27807)) 13391 ((";" 27808)) 13392 ((";" 27810)) 13393 (("," 17413) (";" 27812)) 13394 (("!" 17421) ("\"" 17422) ("$" 17423) ("(" 17424) ("+" 17425) ("-" 17426) ("@" 17427) (T_ARRAY 17428) (T_ARRAY_CAST 17429) (T_ATTRIBUTE 125) (T_BOOL_CAST 17430) (T_CLASS_C 17431) (T_CLONE 17432) (T_CONSTANT_ENCAPSED_STRING 17433) (T_DEC 17434) (T_DIR 17435) (T_DNUMBER 17436) (T_DOUBLE_CAST 17437) (T_EMPTY 17438) (T_EVAL 17439) (T_EXIT 17440) (T_FILE 17441) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 17442) (T_INC 17443) (T_INCLUDE 17444) (T_INCLUDE_ONCE 17445) (T_INT_CAST 17446) (T_ISSET 17447) (T_LINE 17448) (T_LIST 17449) (T_LNUMBER 17450) (T_MATCH 17451) (T_METHOD_C 17452) (T_NAME_FULLY_QUALIFIED 17453) (T_NAME_QUALIFIED 17454) (T_NAME_RELATIVE 17455) (T_NEW 17456) (T_NS_C 17457) (T_OBJECT_CAST 17458) (T_PRINT 17459) (T_REQUIRE 17460) (T_REQUIRE_ONCE 17461) (T_START_HEREDOC 17462) (T_STATIC 17463) (T_STRING 17464) (T_STRING_CAST 17465) (T_THROW 17466) (T_TRAIT_C 17467) (T_UNSET_CAST 17468) (T_VARIABLE 17469) (T_YIELD 17470) (T_YIELD_FROM 17471) ("[" 17472) ("`" 17473) (array_object_dereferenceable 17474) (attribute 129) (attributes 17475) (callable_expr 17476) (callable_variable 17477) (class_constant 17478) (class_name 17479) (constant 17480) (dereferenceable_scalar 17481) (expr 27814) (fn 17483) (fully_dereferenceable 103) #'17484 (function_call 17485) (inline_function 17486) (internal_functions_in_yacc 17487) (match 17488) (name 17489) (new_expr 17490) (scalar 17491) (simple_variable 17492) (static_member 17493) (variable 17494) (variable_class_name 17495) ("~" 17496)) 13395 (("%" 17498) ("*" 17499) ("+" 17500) ("-" 17501) ("." 17502) ("/" 17503) ("<" 17504) (">" 17505) ("?" 17506) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17507) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17508) (T_AS 27815) (T_BOOLEAN_AND 17509) (T_BOOLEAN_OR 17510) (T_COALESCE 17511) (T_INSTANCEOF 17512) (T_IS_EQUAL 17513) (T_IS_GREATER_OR_EQUAL 17514) (T_IS_IDENTICAL 17515) (T_IS_NOT_EQUAL 17516) (T_IS_NOT_IDENTICAL 17517) (T_IS_SMALLER_OR_EQUAL 17518) (T_LOGICAL_AND 17519) (T_LOGICAL_OR 17520) (T_LOGICAL_XOR 17521) (T_POW 17522) (T_SL 17523) (T_SPACESHIP 17524) (T_SR 17525) ("^" 17526) ("|" 17527)) 13396 (("\"" 1489) ("$" 18835) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18836) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18837) ("[" 18838) (ampersand 18839) (array_object_dereferenceable 18840) (callable_expr 18841) (callable_variable 18842) (class_constant 1512) (class_name 18843) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 27816) (fully_dereferenceable 103) (function_call 18845) (name 18846) (simple_variable 18847) (static_member 18848) (variable 18849) (variable_class_name 18850)) 13397 ((")" 27817) (T_DOUBLE_ARROW 27818)) 13398 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27821) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 66) (T_START_HEREDOC 67) (T_STATIC 68) (T_STRING 69) (T_STRING_CAST 70) (T_SWITCH 71) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 75) (T_UNSET 76) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 80) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 85) (alt_if_stmt_without_else 86) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 101) (fn 102) (foreach_statement 27827) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 107) (if_stmt_without_else 108) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27823) (static_member 118) (variable 121) (variable_class_name 122) ("{" 123) ("~" 124)) 13399 (("\"" 1489) ("$" 18865) ("(" 1491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1492) (T_CLASS_C 1493) (T_CONSTANT_ENCAPSED_STRING 1494) (T_DIR 1495) (T_FILE 1496) (T_FUNC_C 1497) (T_LINE 1498) (T_LIST 18866) (T_METHOD_C 1499) (T_NAME_FULLY_QUALIFIED 1500) (T_NAME_QUALIFIED 1501) (T_NAME_RELATIVE 1502) (T_NS_C 1503) (T_STATIC 1504) (T_STRING 1505) (T_TRAIT_C 1506) (T_VARIABLE 18867) ("[" 18868) (ampersand 18869) (array_object_dereferenceable 18870) (callable_expr 18871) (callable_variable 18872) (class_constant 1512) (class_name 18873) (constant 1514) (dereferenceable_scalar 1515) (foreach_variable 27819) (fully_dereferenceable 103) (function_call 18875) (name 18876) (simple_variable 18877) (static_member 18878) (variable 18879) (variable_class_name 18880)) 13400 ((")" 27820)) 13401 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27821) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 66) (T_START_HEREDOC 67) (T_STATIC 68) (T_STRING 69) (T_STRING_CAST 70) (T_SWITCH 71) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 75) (T_UNSET 76) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 80) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 85) (alt_if_stmt_without_else 86) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 101) (fn 102) (foreach_statement 27822) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 107) (if_stmt_without_else 108) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27823) (static_member 118) (variable 121) (variable_class_name 122) ("{" 123) ("~" 124)) 13402 ((inner_statement_list 27824)) 13403 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18898) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 18899) (T_CLASS 18900) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18901) (T_DEC 22) (T_DECLARE 18902) (T_DIR 24) (T_DNUMBER 25) (T_DO 18903) (T_DOUBLE_CAST 27) (T_ECHO 18904) (T_EMPTY 29) (T_ENDFOREACH 27825) (T_ENUM 18906) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18907) (T_FOREACH 18908) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18909) (T_GOTO 18910) (T_HALT_COMPILER 18911) (T_IF 18912) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18913) (T_INTERFACE 18914) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 18915) (T_START_HEREDOC 67) (T_STATIC 18916) (T_STRING 18917) (T_STRING_CAST 70) (T_SWITCH 18918) (T_THROW 72) (T_TRAIT 18919) (T_TRAIT_C 74) (T_TRY 18920) (T_UNSET 18921) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 18922) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 18923) (alt_if_stmt_without_else 18924) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 18925) (attributes 18926) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 18927) (class_modifier 95) (class_modifiers 18928) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 18929) (expr 18930) (fn 102) (fully_dereferenceable 103) #'18931 (function_call 105) (function_declaration_statement 18932) (if_stmt 18933) (if_stmt_without_else 18934) (inline_function 109) (inner_statement 18935) (interface_declaration_statement 18936) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 18937) (static_member 118) (trait_declaration_statement 18938) (variable 121) (variable_class_name 122) ("{" 18939) ("~" 124)) 13404 ((";" 27826)) 13405 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 27829) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 13406 ((";" 27830)) 13407 (("!" 16416) ("\"" 16417) ("$" 16418) ("(" 16419) ("+" 16420) ("-" 16421) ("@" 16422) (T_ARRAY 16423) (T_ARRAY_CAST 16424) (T_ATTRIBUTE 125) (T_BOOL_CAST 16425) (T_CLASS_C 16426) (T_CLONE 16427) (T_CONSTANT_ENCAPSED_STRING 16428) (T_DEC 16429) (T_DIR 16430) (T_DNUMBER 16431) (T_DOUBLE_CAST 16432) (T_EMPTY 16433) (T_EVAL 16434) (T_EXIT 16435) (T_FILE 16436) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 16437) (T_INC 16438) (T_INCLUDE 16439) (T_INCLUDE_ONCE 16440) (T_INT_CAST 16441) (T_ISSET 16442) (T_LINE 16443) (T_LIST 16444) (T_LNUMBER 16445) (T_MATCH 16446) (T_METHOD_C 16447) (T_NAME_FULLY_QUALIFIED 16448) (T_NAME_QUALIFIED 16449) (T_NAME_RELATIVE 16450) (T_NEW 16451) (T_NS_C 16452) (T_OBJECT_CAST 16453) (T_PRINT 16454) (T_REQUIRE 16455) (T_REQUIRE_ONCE 16456) (T_START_HEREDOC 16457) (T_STATIC 16458) (T_STRING 16459) (T_STRING_CAST 16460) (T_THROW 16461) (T_TRAIT_C 16462) (T_UNSET_CAST 16463) (T_VARIABLE 16464) (T_YIELD 16465) (T_YIELD_FROM 16466) ("[" 16467) ("`" 16468) (array_object_dereferenceable 16469) (attribute 129) (attributes 16470) (callable_expr 16471) (callable_variable 16472) (class_constant 16473) (class_name 16474) (constant 16475) (dereferenceable_scalar 16476) (expr 19058) (fn 16478) (for_exprs 27831) (fully_dereferenceable 103) #'16479 (function_call 16480) (inline_function 16481) (internal_functions_in_yacc 16482) (match 16483) (name 16484) (new_expr 16485) (non_empty_for_exprs 19060) (scalar 16486) (simple_variable 16487) (static_member 16488) (variable 16489) (variable_class_name 16490) ("~" 16491)) 13408 ((";" 27832)) 13409 (("!" 2265) ("\"" 2266) ("$" 2267) ("(" 2268) ("+" 2269) ("-" 2270) ("@" 2271) (T_ARRAY 2272) (T_ARRAY_CAST 2273) (T_ATTRIBUTE 125) (T_BOOL_CAST 2274) (T_CLASS_C 2275) (T_CLONE 2276) (T_CONSTANT_ENCAPSED_STRING 2277) (T_DEC 2278) (T_DIR 2279) (T_DNUMBER 2280) (T_DOUBLE_CAST 2281) (T_EMPTY 2282) (T_EVAL 2283) (T_EXIT 2284) (T_FILE 2285) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 2286) (T_INC 2287) (T_INCLUDE 2288) (T_INCLUDE_ONCE 2289) (T_INT_CAST 2290) (T_ISSET 2291) (T_LINE 2292) (T_LIST 2342) (T_LNUMBER 2294) (T_MATCH 2295) (T_METHOD_C 2296) (T_NAME_FULLY_QUALIFIED 2297) (T_NAME_QUALIFIED 2298) (T_NAME_RELATIVE 2299) (T_NEW 2300) (T_NS_C 2301) (T_OBJECT_CAST 2302) (T_PRINT 2303) (T_REQUIRE 2304) (T_REQUIRE_ONCE 2305) (T_START_HEREDOC 2306) (T_STATIC 2307) (T_STRING 2308) (T_STRING_CAST 2309) (T_THROW 2310) (T_TRAIT_C 2311) (T_UNSET_CAST 2312) (T_VARIABLE 2313) (T_YIELD 2314) (T_YIELD_FROM 2315) ("[" 2316) ("`" 2317) (array_object_dereferenceable 2319) (attribute 129) (attributes 2320) (callable_expr 2321) (callable_variable 2322) (class_constant 2323) (class_name 2324) (constant 2325) (dereferenceable_scalar 2326) (expr 19066) (fn 2328) (for_exprs 27833) (fully_dereferenceable 103) #'2329 (function_call 2330) (inline_function 2331) (internal_functions_in_yacc 2332) (match 2333) (name 2334) (new_expr 2335) (non_empty_for_exprs 19068) (scalar 2336) (simple_variable 2337) (static_member 2338) (variable 2339) (variable_class_name 2340) ("~" 2341)) 13410 ((")" 27834)) 13411 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27835) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 66) (T_START_HEREDOC 67) (T_STATIC 68) (T_STRING 69) (T_STRING_CAST 70) (T_SWITCH 71) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 75) (T_UNSET 76) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 80) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 85) (alt_if_stmt_without_else 86) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (dereferenceable_scalar 99) (expr 101) (fn 102) (for_statement 27836) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 107) (if_stmt_without_else 108) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27837) (static_member 118) (variable 121) (variable_class_name 122) ("{" 123) ("~" 124)) 13412 ((inner_statement_list 27838)) 13413 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19076) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19077) (T_CLASS 19078) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19079) (T_DEC 22) (T_DECLARE 19080) (T_DIR 24) (T_DNUMBER 25) (T_DO 19081) (T_DOUBLE_CAST 27) (T_ECHO 19082) (T_EMPTY 29) (T_ENDFOR 27839) (T_ENUM 19084) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19085) (T_FOREACH 19086) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19087) (T_GOTO 19088) (T_HALT_COMPILER 19089) (T_IF 19090) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19091) (T_INTERFACE 19092) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19093) (T_START_HEREDOC 67) (T_STATIC 19094) (T_STRING 19095) (T_STRING_CAST 70) (T_SWITCH 19096) (T_THROW 72) (T_TRAIT 19097) (T_TRAIT_C 74) (T_TRY 19098) (T_UNSET 19099) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19100) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19101) (alt_if_stmt_without_else 19102) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19103) (attributes 19104) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19105) (class_modifier 95) (class_modifiers 19106) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19107) (expr 19108) (fn 102) (fully_dereferenceable 103) #'19109 (function_call 105) (function_declaration_statement 19110) (if_stmt 19111) (if_stmt_without_else 19112) (inline_function 109) (inner_statement 19113) (interface_declaration_statement 19114) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19115) (static_member 118) (trait_declaration_statement 19116) (variable 121) (variable_class_name 122) ("{" 19117) ("~" 124)) 13414 ((";" 27840)) 13415 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 5127) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (optional_expr 27843) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13416 ((")" 27844)) 13417 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27846) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13418 (("%" 280) (")" 27847) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13419 ((":" 19766) (enum_backing_type 27849)) 13420 ((T_IMPLEMENTS 2904) (implements_list 27850)) 13421 ((backup_doc_comment 27851)) 13422 (("{" 27852)) 13423 ((class_statement_list 27853)) 13424 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 27854)) 13425 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27856) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13426 (("%" 280) (")" 27857) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13427 (("," 19251) (";" 27859)) 13428 ((T_WHILE 27862)) 13429 (("(" 27863)) 13430 (("!" 203) ("\"" 204) ("$" 205) ("(" 206) ("+" 207) ("-" 208) ("@" 209) (T_ARRAY 210) (T_ARRAY_CAST 211) (T_ATTRIBUTE 125) (T_BOOL_CAST 212) (T_CLASS_C 213) (T_CLONE 214) (T_CONSTANT_ENCAPSED_STRING 215) (T_DEC 216) (T_DIR 217) (T_DNUMBER 218) (T_DOUBLE_CAST 219) (T_EMPTY 220) (T_EVAL 221) (T_EXIT 222) (T_FILE 223) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 224) (T_INC 225) (T_INCLUDE 226) (T_INCLUDE_ONCE 227) (T_INT_CAST 228) (T_ISSET 229) (T_LINE 230) (T_LIST 231) (T_LNUMBER 232) (T_MATCH 233) (T_METHOD_C 234) (T_NAME_FULLY_QUALIFIED 235) (T_NAME_QUALIFIED 236) (T_NAME_RELATIVE 237) (T_NEW 238) (T_NS_C 239) (T_OBJECT_CAST 240) (T_PRINT 241) (T_REQUIRE 242) (T_REQUIRE_ONCE 243) (T_START_HEREDOC 244) (T_STATIC 245) (T_STRING 246) (T_STRING_CAST 247) (T_THROW 248) (T_TRAIT_C 249) (T_UNSET_CAST 250) (T_VARIABLE 251) (T_YIELD 252) (T_YIELD_FROM 253) ("[" 254) ("`" 255) (array_object_dereferenceable 256) (attribute 129) (attributes 257) (callable_expr 258) (callable_variable 259) (class_constant 260) (class_name 261) (constant 262) (dereferenceable_scalar 263) (expr 27864) (fn 265) (fully_dereferenceable 103) #'266 (function_call 267) (inline_function 268) (internal_functions_in_yacc 269) (match 270) (name 271) (new_expr 272) (scalar 273) (simple_variable 274) (static_member 275) (variable 276) (variable_class_name 277) ("~" 278)) 13431 (("%" 280) (")" 27865) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13432 ((";" 27866)) 13433 ((T_STRING 19424) (const_decl 19425) (const_list 27868)) 13434 ((")" 27869) ("," 19428)) 13435 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27870) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 125) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 126) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 66) (T_START_HEREDOC 67) (T_STATIC 68) (T_STRING 69) (T_STRING_CAST 70) (T_SWITCH 71) (T_THROW 72) (T_TRAIT_C 74) (T_TRY 75) (T_UNSET 76) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 80) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 85) (alt_if_stmt_without_else 86) (array_object_dereferenceable 87) (attribute 129) (attributes 130) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_name 97) (constant 98) (declare_statement 27871) (dereferenceable_scalar 99) (expr 101) (fn 102) (fully_dereferenceable 103) #'132 (function_call 105) (if_stmt 107) (if_stmt_without_else 108) (inline_function 109) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 27872) (static_member 118) (variable 121) (variable_class_name 122) ("{" 123) ("~" 124)) 13436 ((inner_statement_list 27873)) 13437 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 19434) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 19435) (T_CLASS 19436) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 19437) (T_DEC 22) (T_DECLARE 19438) (T_DIR 24) (T_DNUMBER 25) (T_DO 19439) (T_DOUBLE_CAST 27) (T_ECHO 19440) (T_EMPTY 29) (T_ENDDECLARE 27874) (T_ENUM 19442) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 19443) (T_FOREACH 19444) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 19445) (T_GOTO 19446) (T_HALT_COMPILER 19447) (T_IF 19448) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 19449) (T_INTERFACE 19450) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_REQUIRE 64) (T_REQUIRE_ONCE 65) (T_RETURN 19451) (T_START_HEREDOC 67) (T_STATIC 19452) (T_STRING 19453) (T_STRING_CAST 70) (T_SWITCH 19454) (T_THROW 72) (T_TRAIT 19455) (T_TRAIT_C 74) (T_TRY 19456) (T_UNSET 19457) (T_UNSET_CAST 77) (T_VARIABLE 79) (T_WHILE 19458) (T_YIELD 81) (T_YIELD_FROM 82) ("[" 83) ("`" 84) (alt_if_stmt 19459) (alt_if_stmt_without_else 19460) (array_object_dereferenceable 87) (attribute 88) (attributed_statement 19461) (attributes 19462) (callable_expr 91) (callable_variable 92) (class_constant 93) (class_declaration_statement 19463) (class_modifier 95) (class_modifiers 19464) (class_name 97) (constant 98) (dereferenceable_scalar 99) (enum_declaration_statement 19465) (expr 19466) (fn 102) (fully_dereferenceable 103) #'19467 (function_call 105) (function_declaration_statement 19468) (if_stmt 19469) (if_stmt_without_else 19470) (inline_function 109) (inner_statement 19471) (interface_declaration_statement 19472) (internal_functions_in_yacc 111) (match 112) (name 113) (new_expr 114) (scalar 115) (simple_variable 116) (statement 19473) (static_member 118) (trait_declaration_statement 19474) (variable 121) (variable_class_name 122) ("{" 19475) ("~" 124)) 13438 ((";" 27875)) 13439 ((";" 27878)) 13440 (("," 27605) (";" 27880)) 13441 ((T_EXTENDS 2902) (extends_from 27883)) 13442 ((T_IMPLEMENTS 2904) (implements_list 27884)) 13443 ((backup_doc_comment 27885)) 13444 (("{" 27886)) 13445 ((class_statement_list 27887)) 13446 ((T_ABSTRACT 2909) (T_ATTRIBUTE 2910) (T_CASE 2911) (T_FINAL 2912) (T_PRIVATE 2913) (T_PROTECTED 2914) (T_PUBLIC 2915) (T_READONLY 2916) (T_STATIC 2917) (T_USE 2918) (T_VAR 2919) (attribute 2920) (attributed_class_statement 2921) (attributes 2922) (class_statement 2923) (enum_case 2924) (member_modifier 2925) (method_modifiers 2926) (non_empty_member_modifiers 2927) (variable_modifiers 2928) ("}" 27888)) 13447 ((";" 27890)) 13448 (("," 6367) (possible_comma 27893)) 13449 (("]" 27894)) 13450 (("!" 1824) ("\"" 1825) ("$" 1826) ("(" 1827) ("+" 1828) ("-" 1829) ("@" 1830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1485) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1486) (T_ARRAY 1831) (T_ARRAY_CAST 1832) (T_ATTRIBUTE 125) (T_BOOL_CAST 1833) (T_CLASS_C 1834) (T_CLONE 1835) (T_CONSTANT_ENCAPSED_STRING 1836) (T_DEC 1837) (T_DIR 1838) (T_DNUMBER 1839) (T_DOUBLE_CAST 1840) (T_ELLIPSIS 1841) (T_EMPTY 1842) (T_EVAL 1843) (T_EXIT 1844) (T_FILE 1845) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 1846) (T_INC 1847) (T_INCLUDE 1848) (T_INCLUDE_ONCE 1849) (T_INT_CAST 1850) (T_ISSET 1851) (T_LINE 1852) (T_LIST 1853) (T_LNUMBER 1854) (T_MATCH 1855) (T_METHOD_C 1856) (T_NAME_FULLY_QUALIFIED 1857) (T_NAME_QUALIFIED 1858) (T_NAME_RELATIVE 1859) (T_NEW 1860) (T_NS_C 1861) (T_OBJECT_CAST 1862) (T_PRINT 1863) (T_REQUIRE 1864) (T_REQUIRE_ONCE 1865) (T_START_HEREDOC 1866) (T_STATIC 1867) (T_STRING 1868) (T_STRING_CAST 1869) (T_THROW 1870) (T_TRAIT_C 1871) (T_UNSET_CAST 1872) (T_VARIABLE 1873) (T_YIELD 1874) (T_YIELD_FROM 1875) ("[" 1876) ("`" 1877) (ampersand 1878) (array_object_dereferenceable 1879) (array_pair 1880) (array_pair_list 27897) (attribute 129) (attributes 1882) (callable_expr 1883) (callable_variable 1884) (class_constant 1885) (class_name 1886) (constant 1887) (dereferenceable_scalar 1888) (expr 1889) (fn 1890) (fully_dereferenceable 103) #'1891 (function_call 1892) (inline_function 1893) (internal_functions_in_yacc 1894) (match 1895) (name 1896) (new_expr 1897) (non_empty_array_pair_list 1898) (possible_array_pair 1899) (scalar 1900) (simple_variable 1901) (static_member 1902) (variable 1903) (variable_class_name 1904) ("~" 1905)) 13451 ((")" 27898)) 13452 (("%" 280) (")" 27903) ("*" 281) ("+" 282) ("-" 283) ("." 284) ("/" 285) ("<" 286) (">" 287) ("?" 288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 290) (T_BOOLEAN_AND 291) (T_BOOLEAN_OR 292) (T_COALESCE 293) (T_INSTANCEOF 294) (T_IS_EQUAL 295) (T_IS_GREATER_OR_EQUAL 296) (T_IS_IDENTICAL 297) (T_IS_NOT_EQUAL 298) (T_IS_NOT_IDENTICAL 299) (T_IS_SMALLER_OR_EQUAL 300) (T_LOGICAL_AND 301) (T_LOGICAL_OR 302) (T_LOGICAL_XOR 303) (T_POW 304) (T_SL 305) (T_SPACESHIP 306) (T_SR 307) ("^" 308) ("|" 309)) 13453 (("!" 343) ("\"" 344) ("$" 345) ("(" 346) ("+" 347) ("-" 348) ("@" 349) (T_ARRAY 350) (T_ARRAY_CAST 351) (T_ATTRIBUTE 125) (T_BOOL_CAST 352) (T_CLASS_C 353) (T_CLONE 354) (T_CONSTANT_ENCAPSED_STRING 355) (T_DEC 356) (T_DIR 357) (T_DNUMBER 358) (T_DOUBLE_CAST 359) (T_EMPTY 360) (T_EVAL 361) (T_EXIT 362) (T_FILE 363) (T_FN 35) (T_FUNCTION 126) (T_FUNC_C 364) (T_INC 365) (T_INCLUDE 366) (T_INCLUDE_ONCE 367) (T_INT_CAST 368) (T_ISSET 369) (T_LINE 370) (T_LIST 371) (T_LNUMBER 372) (T_MATCH 373) (T_METHOD_C 374) (T_NAME_FULLY_QUALIFIED 375) (T_NAME_QUALIFIED 376) (T_NAME_RELATIVE 377) (T_NEW 378) (T_NS_C 379) (T_OBJECT_CAST 380) (T_PRINT 381) (T_REQUIRE 382) (T_REQUIRE_ONCE 383) (T_START_HEREDOC 384) (T_STATIC 385) (T_STRING 386) (T_STRING_CAST 387) (T_THROW 388) (T_TRAIT_C 389) (T_UNSET_CAST 390) (T_VARIABLE 391) (T_YIELD 392) (T_YIELD_FROM 393) ("[" 394) ("`" 395) (array_object_dereferenceable 396) (attribute 129) (attributes 397) (callable_expr 398) (callable_variable 399) (class_constant 400) (class_name 401) (constant 402) (dereferenceable_scalar 403) (expr 27906) (fn 405) (fully_dereferenceable 103) #'406 (function_call 407) (inline_function 408) (internal_functions_in_yacc 409) (match 410) (name 411) (new_expr 412) (scalar 413) (simple_variable 414) (static_member 415) (variable 416) (variable_class_name 417) ("~" 418)) 13454 (("%" 420) ("*" 421) ("+" 422) ("-" 423) ("." 424) ("/" 425) ("<" 426) (">" 427) ("?" 428) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 429) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 430) (T_BOOLEAN_AND 431) (T_BOOLEAN_OR 432) (T_COALESCE 433) (T_INSTANCEOF 434) (T_IS_EQUAL 435) (T_IS_GREATER_OR_EQUAL 436) (T_IS_IDENTICAL 437) (T_IS_NOT_EQUAL 438) (T_IS_NOT_IDENTICAL 439) (T_IS_SMALLER_OR_EQUAL 440) (T_LOGICAL_AND 441) (T_LOGICAL_OR 442) (T_LOGICAL_XOR 443) (T_POW 444) (T_SL 445) (T_SPACESHIP 446) (T_SR 447) ("^" 448) ("|" 449) ("}" 27907)) 13455 (("\"" 27909) (T_CURLY_OPEN 5153) (T_DOLLAR_OPEN_CURLY_BRACES 5154) (T_ENCAPSED_AND_WHITESPACE 5160) (T_VARIABLE 5156) (encaps_var 5161))))
  "The generated distinct goto-tables.")

(defvar
  phps-mode-parser--table-productions-number-reverse
  #s(hash-table size 730 test equal rehash-size 1.5 rehash-threshold 0.8125 data (0 ((start) (top_statement_list)) 1 ((reserved_non_modifiers) (T_INCLUDE)) 2 ((reserved_non_modifiers) (T_INCLUDE_ONCE)) 3 ((reserved_non_modifiers) (T_EVAL)) 4 ((reserved_non_modifiers) (T_REQUIRE)) 5 ((reserved_non_modifiers) (T_REQUIRE_ONCE)) 6 ((reserved_non_modifiers) (T_LOGICAL_OR)) 7 ((reserved_non_modifiers) (T_LOGICAL_XOR)) 8 ((reserved_non_modifiers) (T_LOGICAL_AND)) 9 ((reserved_non_modifiers) (T_INSTANCEOF)) 10 ((reserved_non_modifiers) (T_NEW)) 11 ((reserved_non_modifiers) (T_CLONE)) 12 ((reserved_non_modifiers) (T_EXIT)) 13 ((reserved_non_modifiers) (T_IF)) 14 ((reserved_non_modifiers) (T_ELSEIF)) 15 ((reserved_non_modifiers) (T_ELSE)) 16 ((reserved_non_modifiers) (T_ENDIF)) 17 ((reserved_non_modifiers) (T_ECHO)) 18 ((reserved_non_modifiers) (T_DO)) 19 ((reserved_non_modifiers) (T_WHILE)) 20 ((reserved_non_modifiers) (T_ENDWHILE)) 21 ((reserved_non_modifiers) (T_FOR)) 22 ((reserved_non_modifiers) (T_ENDFOR)) 23 ((reserved_non_modifiers) (T_FOREACH)) 24 ((reserved_non_modifiers) (T_ENDFOREACH)) 25 ((reserved_non_modifiers) (T_DECLARE)) 26 ((reserved_non_modifiers) (T_ENDDECLARE)) 27 ((reserved_non_modifiers) (T_AS)) 28 ((reserved_non_modifiers) (T_TRY)) 29 ((reserved_non_modifiers) (T_CATCH)) 30 ((reserved_non_modifiers) (T_FINALLY)) 31 ((reserved_non_modifiers) (T_THROW)) 32 ((reserved_non_modifiers) (T_USE)) 33 ((reserved_non_modifiers) (T_INSTEADOF)) 34 ((reserved_non_modifiers) (T_GLOBAL)) 35 ((reserved_non_modifiers) (T_VAR)) 36 ((reserved_non_modifiers) (T_UNSET)) 37 ((reserved_non_modifiers) (T_ISSET)) 38 ((reserved_non_modifiers) (T_EMPTY)) 39 ((reserved_non_modifiers) (T_CONTINUE)) 40 ((reserved_non_modifiers) (T_GOTO)) 41 ((reserved_non_modifiers) (T_FUNCTION)) 42 ((reserved_non_modifiers) (T_CONST)) 43 ((reserved_non_modifiers) (T_RETURN)) 44 ((reserved_non_modifiers) (T_PRINT)) 45 ((reserved_non_modifiers) (T_YIELD)) 46 ((reserved_non_modifiers) (T_LIST)) 47 ((reserved_non_modifiers) (T_SWITCH)) 48 ((reserved_non_modifiers) (T_ENDSWITCH)) 49 ((reserved_non_modifiers) (T_CASE)) 50 ((reserved_non_modifiers) (T_DEFAULT)) 51 ((reserved_non_modifiers) (T_BREAK)) 52 ((reserved_non_modifiers) (T_ARRAY)) 53 ((reserved_non_modifiers) (T_CALLABLE)) 54 ((reserved_non_modifiers) (T_EXTENDS)) 55 ((reserved_non_modifiers) (T_IMPLEMENTS)) 56 ((reserved_non_modifiers) (T_NAMESPACE)) 57 ((reserved_non_modifiers) (T_TRAIT)) 58 ((reserved_non_modifiers) (T_INTERFACE)) 59 ((reserved_non_modifiers) (T_CLASS)) 60 ((reserved_non_modifiers) (T_CLASS_C)) 61 ((reserved_non_modifiers) (T_TRAIT_C)) 62 ((reserved_non_modifiers) (T_FUNC_C)) 63 ((reserved_non_modifiers) (T_METHOD_C)) 64 ((reserved_non_modifiers) (T_LINE)) 65 ((reserved_non_modifiers) (T_FILE)) 66 ((reserved_non_modifiers) (T_DIR)) 67 ((reserved_non_modifiers) (T_NS_C)) 68 ((reserved_non_modifiers) (T_FN)) 69 ((reserved_non_modifiers) (T_MATCH)) 70 ((reserved_non_modifiers) (T_ENUM)) 71 ((semi_reserved) (reserved_non_modifiers)) 72 ((semi_reserved) (T_STATIC)) 73 ((semi_reserved) (T_ABSTRACT)) 74 ((semi_reserved) (T_FINAL)) 75 ((semi_reserved) (T_PRIVATE)) 76 ((semi_reserved) (T_PROTECTED)) 77 ((semi_reserved) (T_PUBLIC)) 78 ((semi_reserved) (T_READONLY)) 79 ((ampersand) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG)) 80 ((ampersand) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG)) 81 ((identifier) (T_STRING)) 82 ((identifier) (semi_reserved)) 83 ((top_statement_list) (top_statement_list top_statement)) 84 ((top_statement_list) (%empty)) 85 ((namespace_declaration_name) (identifier)) 86 ((namespace_declaration_name) (T_NAME_QUALIFIED)) 87 ((namespace_name) (T_STRING)) 88 ((namespace_name) (T_NAME_QUALIFIED)) 89 ((legacy_namespace_name) (namespace_name)) 90 ((legacy_namespace_name) (T_NAME_FULLY_QUALIFIED)) 91 ((name) (T_STRING)) 92 ((name) (T_NAME_QUALIFIED)) 93 ((name) (T_NAME_FULLY_QUALIFIED)) 94 ((name) (T_NAME_RELATIVE)) 95 ((attribute_decl) (class_name)) 96 ((attribute_decl) (class_name argument_list)) 97 ((attribute_group) (attribute_decl)) 98 ((attribute_group) (attribute_group "," attribute_decl)) 99 ((attribute) (T_ATTRIBUTE attribute_group possible_comma "]")) 100 ((attributes) (attribute)) 101 ((attributes) (attributes attribute)) 102 ((attributed_statement) (function_declaration_statement)) 103 ((attributed_statement) (class_declaration_statement)) 104 ((attributed_statement) (trait_declaration_statement)) 105 ((attributed_statement) (interface_declaration_statement)) 106 ((attributed_statement) (enum_declaration_statement)) 107 ((top_statement) (statement)) 108 ((top_statement) (attributed_statement)) 109 ((top_statement) (attributes attributed_statement)) 110 ((top_statement) (T_HALT_COMPILER "(" ")" ";")) 111 ((top_statement) (T_NAMESPACE namespace_declaration_name ";")) 112 ((top_statement) (T_NAMESPACE namespace_declaration_name "{" top_statement_list "}")) 113 ((top_statement) (T_NAMESPACE "{" top_statement_list "}")) 114 ((top_statement) (T_USE mixed_group_use_declaration ";")) 115 ((top_statement) (T_USE use_type group_use_declaration ";")) 116 ((top_statement) (T_USE use_declarations ";")) 117 ((top_statement) (T_USE use_type use_declarations ";")) 118 ((top_statement) (T_CONST const_list ";")) 119 ((use_type) (T_FUNCTION)) 120 ((use_type) (T_CONST)) 121 ((group_use_declaration) (legacy_namespace_name T_NS_SEPARATOR "{" unprefixed_use_declarations possible_comma "}")) 122 ((mixed_group_use_declaration) (legacy_namespace_name T_NS_SEPARATOR "{" inline_use_declarations possible_comma "}")) 123 ((possible_comma) (%empty)) 124 ((possible_comma) (",")) 125 ((inline_use_declarations) (inline_use_declarations "," inline_use_declaration)) 126 ((inline_use_declarations) (inline_use_declaration)) 127 ((unprefixed_use_declarations) (unprefixed_use_declarations "," unprefixed_use_declaration)) 128 ((unprefixed_use_declarations) (unprefixed_use_declaration)) 129 ((use_declarations) (use_declarations "," use_declaration)) 130 ((use_declarations) (use_declaration)) 131 ((inline_use_declaration) (unprefixed_use_declaration)) 132 ((inline_use_declaration) (use_type unprefixed_use_declaration)) 133 ((unprefixed_use_declaration) (namespace_name)) 134 ((unprefixed_use_declaration) (namespace_name T_AS T_STRING)) 135 ((use_declaration) (legacy_namespace_name)) 136 ((use_declaration) (legacy_namespace_name T_AS T_STRING)) 137 ((const_list) (const_list "," const_decl)) 138 ((const_list) (const_decl)) 139 ((inner_statement_list) (inner_statement_list inner_statement)) 140 ((inner_statement_list) (%empty)) 141 ((inner_statement) (statement)) 142 ((inner_statement) (attributed_statement)) 143 ((inner_statement) (attributes attributed_statement)) 144 ((inner_statement) (T_HALT_COMPILER "(" ")" ";")) 145 ((statement) ("{" inner_statement_list "}")) 146 ((statement) (if_stmt)) 147 ((statement) (alt_if_stmt)) 148 ((statement) (T_WHILE "(" expr ")" while_statement)) 149 ((statement) (T_DO statement T_WHILE "(" expr ")" ";")) 150 ((statement) (T_FOR "(" for_exprs ";" for_exprs ";" for_exprs ")" for_statement)) 151 ((statement) (T_SWITCH "(" expr ")" switch_case_list)) 152 ((statement) (T_BREAK optional_expr ";")) 153 ((statement) (T_CONTINUE optional_expr ";")) 154 ((statement) (T_RETURN optional_expr ";")) 155 ((statement) (T_GLOBAL global_var_list ";")) 156 ((statement) (T_STATIC static_var_list ";")) 157 ((statement) (T_ECHO echo_expr_list ";")) 158 ((statement) (T_INLINE_HTML)) 159 ((statement) (expr ";")) 160 ((statement) (T_UNSET "(" unset_variables possible_comma ")" ";")) 161 ((statement) (T_FOREACH "(" expr T_AS foreach_variable ")" foreach_statement)) 162 ((statement) (T_FOREACH "(" expr T_AS foreach_variable T_DOUBLE_ARROW foreach_variable ")" foreach_statement)) 163 ((statement) (T_DECLARE "(" const_list ")" declare_statement)) 164 ((statement) (";")) 165 ((statement) (T_TRY "{" inner_statement_list "}" catch_list finally_statement)) 166 ((statement) (T_GOTO T_STRING ";")) 167 ((statement) (T_STRING ":")) 168 ((catch_list) (%empty)) 169 ((catch_list) (catch_list T_CATCH "(" catch_name_list optional_variable ")" "{" inner_statement_list "}")) 170 ((catch_name_list) (class_name)) 171 ((catch_name_list) (catch_name_list "|" class_name)) 172 ((optional_variable) (%empty)) 173 ((optional_variable) (T_VARIABLE)) 174 ((finally_statement) (%empty)) 175 ((finally_statement) (T_FINALLY "{" inner_statement_list "}")) 176 ((unset_variables) (unset_variable)) 177 ((unset_variables) (unset_variables "," unset_variable)) 178 ((unset_variable) (variable)) 179 ((function_declaration_statement) (function returns_ref T_STRING backup_doc_comment "(" parameter_list ")" return_type backup_fn_flags "{" inner_statement_list "}" backup_fn_flags)) 180 ((is_reference) (%empty)) 181 ((is_reference) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG)) 182 ((is_variadic) (%empty)) 183 ((is_variadic) (T_ELLIPSIS)) 184 ((class_declaration_statement) (class_modifiers T_CLASS T_STRING extends_from implements_list backup_doc_comment "{" class_statement_list "}")) 185 ((class_declaration_statement) (T_CLASS T_STRING extends_from implements_list backup_doc_comment "{" class_statement_list "}")) 186 ((class_modifiers) (class_modifier)) 187 ((class_modifiers) (class_modifiers class_modifier)) 188 ((class_modifier) (T_ABSTRACT)) 189 ((class_modifier) (T_FINAL)) 190 ((trait_declaration_statement) (T_TRAIT T_STRING backup_doc_comment "{" class_statement_list "}")) 191 ((interface_declaration_statement) (T_INTERFACE T_STRING interface_extends_list backup_doc_comment "{" class_statement_list "}")) 192 ((enum_declaration_statement) (T_ENUM T_STRING enum_backing_type implements_list backup_doc_comment "{" class_statement_list "}")) 193 ((enum_backing_type) (%empty)) 194 ((enum_backing_type) (":" type_expr)) 195 ((enum_case) (T_CASE backup_doc_comment identifier enum_case_expr ";")) 196 ((enum_case_expr) (%empty)) 197 ((enum_case_expr) ("=" expr)) 198 ((extends_from) (%empty)) 199 ((extends_from) (T_EXTENDS class_name)) 200 ((interface_extends_list) (%empty)) 201 ((interface_extends_list) (T_EXTENDS class_name_list)) 202 ((implements_list) (%empty)) 203 ((implements_list) (T_IMPLEMENTS class_name_list)) 204 ((foreach_variable) (variable)) 205 ((foreach_variable) (ampersand variable)) 206 ((foreach_variable) (T_LIST "(" array_pair_list ")")) 207 ((foreach_variable) ("[" array_pair_list "]")) 208 ((for_statement) (statement)) 209 ((for_statement) (":" inner_statement_list T_ENDFOR ";")) 210 ((foreach_statement) (statement)) 211 ((foreach_statement) (":" inner_statement_list T_ENDFOREACH ";")) 212 ((declare_statement) (statement)) 213 ((declare_statement) (":" inner_statement_list T_ENDDECLARE ";")) 214 ((switch_case_list) ("{" case_list "}")) 215 ((switch_case_list) ("{" ";" case_list "}")) 216 ((switch_case_list) (":" case_list T_ENDSWITCH ";")) 217 ((switch_case_list) (":" ";" case_list T_ENDSWITCH ";")) 218 ((case_list) (%empty)) 219 ((case_list) (case_list T_CASE expr case_separator inner_statement_list)) 220 ((case_list) (case_list T_DEFAULT case_separator inner_statement_list)) 221 ((case_separator) (":")) 222 ((case_separator) (";")) 223 ((match) (T_MATCH "(" expr ")" "{" match_arm_list "}")) 224 ((match_arm_list) (%empty)) 225 ((match_arm_list) (non_empty_match_arm_list possible_comma)) 226 ((non_empty_match_arm_list) (match_arm)) 227 ((non_empty_match_arm_list) (non_empty_match_arm_list "," match_arm)) 228 ((match_arm) (match_arm_cond_list possible_comma T_DOUBLE_ARROW expr)) 229 ((match_arm) (T_DEFAULT possible_comma T_DOUBLE_ARROW expr)) 230 ((match_arm_cond_list) (expr)) 231 ((match_arm_cond_list) (match_arm_cond_list "," expr)) 232 ((while_statement) (statement)) 233 ((while_statement) (":" inner_statement_list T_ENDWHILE ";")) 234 ((if_stmt_without_else) (T_IF "(" expr ")" statement)) 235 ((if_stmt_without_else) (if_stmt_without_else T_ELSEIF "(" expr ")" statement)) 236 ((if_stmt) (if_stmt_without_else)) 237 ((if_stmt) (if_stmt_without_else T_ELSE statement)) 238 ((alt_if_stmt_without_else) (T_IF "(" expr ")" ":" inner_statement_list)) 239 ((alt_if_stmt_without_else) (alt_if_stmt_without_else T_ELSEIF "(" expr ")" ":" inner_statement_list)) 240 ((alt_if_stmt) (alt_if_stmt_without_else T_ENDIF ";")) 241 ((alt_if_stmt) (alt_if_stmt_without_else T_ELSE ":" inner_statement_list T_ENDIF ";")) 242 ((parameter_list) (non_empty_parameter_list possible_comma)) 243 ((parameter_list) (%empty)) 244 ((non_empty_parameter_list) (attributed_parameter)) 245 ((non_empty_parameter_list) (non_empty_parameter_list "," attributed_parameter)) 246 ((attributed_parameter) (attributes parameter)) 247 ((attributed_parameter) (parameter)) 248 ((optional_property_modifiers) (%empty)) 249 ((optional_property_modifiers) (optional_property_modifiers property_modifier)) 250 ((property_modifier) (T_PUBLIC)) 251 ((property_modifier) (T_PROTECTED)) 252 ((property_modifier) (T_PRIVATE)) 253 ((property_modifier) (T_READONLY)) 254 ((parameter) (optional_property_modifiers optional_type_without_static is_reference is_variadic T_VARIABLE backup_doc_comment)) 255 ((parameter) (optional_property_modifiers optional_type_without_static is_reference is_variadic T_VARIABLE backup_doc_comment "=" expr)) 256 ((optional_type_without_static) (%empty)) 257 ((optional_type_without_static) (type_expr_without_static)) 258 ((type_expr) (type)) 259 ((type_expr) ("?" type)) 260 ((type_expr) (union_type)) 261 ((type_expr) (intersection_type)) 262 ((type) (type_without_static)) 263 ((type) (T_STATIC)) 264 ((union_type) (type "|" type)) 265 ((union_type) (union_type "|" type)) 266 ((intersection_type) (type T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG type)) 267 ((intersection_type) (intersection_type T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG type)) 268 ((type_expr_without_static) (type_without_static)) 269 ((type_expr_without_static) ("?" type_without_static)) 270 ((type_expr_without_static) (union_type_without_static)) 271 ((type_expr_without_static) (intersection_type_without_static)) 272 ((type_without_static) (T_ARRAY)) 273 ((type_without_static) (T_CALLABLE)) 274 ((type_without_static) (name)) 275 ((union_type_without_static) (type_without_static "|" type_without_static)) 276 ((union_type_without_static) (union_type_without_static "|" type_without_static)) 277 ((intersection_type_without_static) (type_without_static T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG type_without_static)) 278 ((intersection_type_without_static) (intersection_type_without_static T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG type_without_static)) 279 ((return_type) (%empty)) 280 ((return_type) (":" type_expr)) 281 ((argument_list) ("(" ")")) 282 ((argument_list) ("(" non_empty_argument_list possible_comma ")")) 283 ((argument_list) ("(" T_ELLIPSIS ")")) 284 ((non_empty_argument_list) (argument)) 285 ((non_empty_argument_list) (non_empty_argument_list "," argument)) 286 ((argument) (expr)) 287 ((argument) (identifier ":" expr)) 288 ((argument) (T_ELLIPSIS expr)) 289 ((global_var_list) (global_var_list "," global_var)) 290 ((global_var_list) (global_var)) 291 ((global_var) (simple_variable)) 292 ((static_var_list) (static_var_list "," static_var)) 293 ((static_var_list) (static_var)) 294 ((static_var) (T_VARIABLE)) 295 ((static_var) (T_VARIABLE "=" expr)) 296 ((class_statement_list) (class_statement_list class_statement)) 297 ((class_statement_list) (%empty)) 298 ((attributed_class_statement) (variable_modifiers optional_type_without_static property_list ";")) 299 ((attributed_class_statement) (method_modifiers T_CONST class_const_list ";")) 300 ((attributed_class_statement) (method_modifiers function returns_ref identifier backup_doc_comment "(" parameter_list ")" return_type backup_fn_flags method_body backup_fn_flags)) 301 ((attributed_class_statement) (enum_case)) 302 ((class_statement) (attributed_class_statement)) 303 ((class_statement) (attributes attributed_class_statement)) 304 ((class_statement) (T_USE class_name_list trait_adaptations)) 305 ((class_name_list) (class_name)) 306 ((class_name_list) (class_name_list "," class_name)) 307 ((trait_adaptations) (";")) 308 ((trait_adaptations) ("{" "}")) 309 ((trait_adaptations) ("{" trait_adaptation_list "}")) 310 ((trait_adaptation_list) (trait_adaptation)) 311 ((trait_adaptation_list) (trait_adaptation_list trait_adaptation)) 312 ((trait_adaptation) (trait_precedence ";")) 313 ((trait_adaptation) (trait_alias ";")) 314 ((trait_precedence) (absolute_trait_method_reference T_INSTEADOF class_name_list)) 315 ((trait_alias) (trait_method_reference T_AS T_STRING)) 316 ((trait_alias) (trait_method_reference T_AS reserved_non_modifiers)) 317 ((trait_alias) (trait_method_reference T_AS member_modifier identifier)) 318 ((trait_alias) (trait_method_reference T_AS member_modifier)) 319 ((trait_method_reference) (identifier)) 320 ((trait_method_reference) (absolute_trait_method_reference)) 321 ((absolute_trait_method_reference) (class_name T_PAAMAYIM_NEKUDOTAYIM identifier)) 322 ((method_body) (";")) 323 ((method_body) ("{" inner_statement_list "}")) 324 ((variable_modifiers) (non_empty_member_modifiers)) 325 ((variable_modifiers) (T_VAR)) 326 ((method_modifiers) (%empty)) 327 ((method_modifiers) (non_empty_member_modifiers)) 328 ((non_empty_member_modifiers) (member_modifier)) 329 ((non_empty_member_modifiers) (non_empty_member_modifiers member_modifier)) 330 ((member_modifier) (T_PUBLIC)) 331 ((member_modifier) (T_PROTECTED)) 332 ((member_modifier) (T_PRIVATE)) 333 ((member_modifier) (T_STATIC)) 334 ((member_modifier) (T_ABSTRACT)) 335 ((member_modifier) (T_FINAL)) 336 ((member_modifier) (T_READONLY)) 337 ((property_list) (property_list "," property)) 338 ((property_list) (property)) 339 ((property) (T_VARIABLE backup_doc_comment)) 340 ((property) (T_VARIABLE "=" expr backup_doc_comment)) 341 ((class_const_list) (class_const_list "," class_const_decl)) 342 ((class_const_list) (class_const_decl)) 343 ((class_const_decl) (identifier "=" expr backup_doc_comment)) 344 ((const_decl) (T_STRING "=" expr backup_doc_comment)) 345 ((echo_expr_list) (echo_expr_list "," echo_expr)) 346 ((echo_expr_list) (echo_expr)) 347 ((echo_expr) (expr)) 348 ((for_exprs) (%empty)) 349 ((for_exprs) (non_empty_for_exprs)) 350 ((non_empty_for_exprs) (non_empty_for_exprs "," expr)) 351 ((non_empty_for_exprs) (expr)) 352 ((anonymous_class) (T_CLASS ctor_arguments extends_from implements_list backup_doc_comment "{" class_statement_list "}")) 353 ((new_expr) (T_NEW class_name_reference ctor_arguments)) 354 ((new_expr) (T_NEW anonymous_class)) 355 ((new_expr) (T_NEW attributes anonymous_class)) 356 ((expr) (variable)) 357 ((expr) (T_LIST "(" array_pair_list ")" "=" expr)) 358 ((expr) ("[" array_pair_list "]" "=" expr)) 359 ((expr) (variable "=" expr)) 360 ((expr) (variable "=" ampersand variable)) 361 ((expr) (T_CLONE expr)) 362 ((expr) (variable T_PLUS_EQUAL expr)) 363 ((expr) (variable T_MINUS_EQUAL expr)) 364 ((expr) (variable T_MUL_EQUAL expr)) 365 ((expr) (variable T_POW_EQUAL expr)) 366 ((expr) (variable T_DIV_EQUAL expr)) 367 ((expr) (variable T_CONCAT_EQUAL expr)) 368 ((expr) (variable T_MOD_EQUAL expr)) 369 ((expr) (variable T_AND_EQUAL expr)) 370 ((expr) (variable T_OR_EQUAL expr)) 371 ((expr) (variable T_XOR_EQUAL expr)) 372 ((expr) (variable T_SL_EQUAL expr)) 373 ((expr) (variable T_SR_EQUAL expr)) 374 ((expr) (variable T_COALESCE_EQUAL expr)) 375 ((expr) (variable T_INC)) 376 ((expr) (T_INC variable)) 377 ((expr) (variable T_DEC)) 378 ((expr) (T_DEC variable)) 379 ((expr) (expr T_BOOLEAN_OR expr)) 380 ((expr) (expr T_BOOLEAN_AND expr)) 381 ((expr) (expr T_LOGICAL_OR expr)) 382 ((expr) (expr T_LOGICAL_AND expr)) 383 ((expr) (expr T_LOGICAL_XOR expr)) 384 ((expr) (expr "|" expr)) 385 ((expr) (expr T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG expr)) 386 ((expr) (expr T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG expr)) 387 ((expr) (expr "^" expr)) 388 ((expr) (expr "." expr)) 389 ((expr) (expr "+" expr)) 390 ((expr) (expr "-" expr)) 391 ((expr) (expr "*" expr)) 392 ((expr) (expr T_POW expr)) 393 ((expr) (expr "/" expr)) 394 ((expr) (expr "%" expr)) 395 ((expr) (expr T_SL expr)) 396 ((expr) (expr T_SR expr)) 397 ((expr) ("+" expr)) 398 ((expr) ("-" expr)) 399 ((expr) ("!" expr)) 400 ((expr) ("~" expr)) 401 ((expr) (expr T_IS_IDENTICAL expr)) 402 ((expr) (expr T_IS_NOT_IDENTICAL expr)) 403 ((expr) (expr T_IS_EQUAL expr)) 404 ((expr) (expr T_IS_NOT_EQUAL expr)) 405 ((expr) (expr "<" expr)) 406 ((expr) (expr T_IS_SMALLER_OR_EQUAL expr)) 407 ((expr) (expr ">" expr)) 408 ((expr) (expr T_IS_GREATER_OR_EQUAL expr)) 409 ((expr) (expr T_SPACESHIP expr)) 410 ((expr) (expr T_INSTANCEOF class_name_reference)) 411 ((expr) ("(" expr ")")) 412 ((expr) (new_expr)) 413 ((expr) (expr "?" expr ":" expr)) 414 ((expr) (expr "?" ":" expr)) 415 ((expr) (expr T_COALESCE expr)) 416 ((expr) (internal_functions_in_yacc)) 417 ((expr) (T_INT_CAST expr)) 418 ((expr) (T_DOUBLE_CAST expr)) 419 ((expr) (T_STRING_CAST expr)) 420 ((expr) (T_ARRAY_CAST expr)) 421 ((expr) (T_OBJECT_CAST expr)) 422 ((expr) (T_BOOL_CAST expr)) 423 ((expr) (T_UNSET_CAST expr)) 424 ((expr) (T_EXIT exit_expr)) 425 ((expr) ("@" expr)) 426 ((expr) (scalar)) 427 ((expr) ("`" backticks_expr "`")) 428 ((expr) (T_PRINT expr)) 429 ((expr) (T_YIELD)) 430 ((expr) (T_YIELD expr)) 431 ((expr) (T_YIELD expr T_DOUBLE_ARROW expr)) 432 ((expr) (T_YIELD_FROM expr)) 433 ((expr) (T_THROW expr)) 434 ((expr) (inline_function)) 435 ((expr) (attributes inline_function)) 436 ((expr) (T_STATIC inline_function)) 437 ((expr) (attributes T_STATIC inline_function)) 438 ((expr) (match)) 439 ((inline_function) (function returns_ref backup_doc_comment "(" parameter_list ")" lexical_vars return_type backup_fn_flags "{" inner_statement_list "}" backup_fn_flags)) 440 ((inline_function) (fn returns_ref backup_doc_comment "(" parameter_list ")" return_type T_DOUBLE_ARROW backup_fn_flags backup_lex_pos expr backup_fn_flags)) 441 ((fn) (T_FN)) 442 ((function) (T_FUNCTION)) 443 ((backup_doc_comment) (%empty)) 444 ((backup_fn_flags) (%empty)) 445 ((backup_lex_pos) (%empty)) 446 ((returns_ref) (%empty)) 447 ((returns_ref) (ampersand)) 448 ((lexical_vars) (%empty)) 449 ((lexical_vars) (T_USE "(" lexical_var_list possible_comma ")")) 450 ((lexical_var_list) (lexical_var_list "," lexical_var)) 451 ((lexical_var_list) (lexical_var)) 452 ((lexical_var) (T_VARIABLE)) 453 ((lexical_var) (ampersand T_VARIABLE)) 454 ((function_call) (name argument_list)) 455 ((function_call) (class_name T_PAAMAYIM_NEKUDOTAYIM member_name argument_list)) 456 ((function_call) (variable_class_name T_PAAMAYIM_NEKUDOTAYIM member_name argument_list)) 457 ((function_call) (callable_expr argument_list)) 458 ((class_name) (T_STATIC)) 459 ((class_name) (name)) 460 ((class_name_reference) (class_name)) 461 ((class_name_reference) (new_variable)) 462 ((class_name_reference) ("(" expr ")")) 463 ((exit_expr) (%empty)) 464 ((exit_expr) ("(" optional_expr ")")) 465 ((backticks_expr) (%empty)) 466 ((backticks_expr) (T_ENCAPSED_AND_WHITESPACE)) 467 ((backticks_expr) (encaps_list)) 468 ((ctor_arguments) (%empty)) 469 ((ctor_arguments) (argument_list)) 470 ((dereferenceable_scalar) (T_ARRAY "(" array_pair_list ")")) 471 ((dereferenceable_scalar) ("[" array_pair_list "]")) 472 ((dereferenceable_scalar) (T_CONSTANT_ENCAPSED_STRING)) 473 ((dereferenceable_scalar) ("\"" encaps_list "\"")) 474 ((scalar) (T_LNUMBER)) 475 ((scalar) (T_DNUMBER)) 476 ((scalar) (T_START_HEREDOC T_ENCAPSED_AND_WHITESPACE T_END_HEREDOC)) 477 ((scalar) (T_START_HEREDOC T_END_HEREDOC)) 478 ((scalar) (T_START_HEREDOC encaps_list T_END_HEREDOC)) 479 ((scalar) (dereferenceable_scalar)) 480 ((scalar) (constant)) 481 ((scalar) (class_constant)) 482 ((constant) (name)) 483 ((constant) (T_LINE)) 484 ((constant) (T_FILE)) 485 ((constant) (T_DIR)) 486 ((constant) (T_TRAIT_C)) 487 ((constant) (T_METHOD_C)) 488 ((constant) (T_FUNC_C)) 489 ((constant) (T_NS_C)) 490 ((constant) (T_CLASS_C)) 491 ((class_constant) (class_name T_PAAMAYIM_NEKUDOTAYIM identifier)) 492 ((class_constant) (variable_class_name T_PAAMAYIM_NEKUDOTAYIM identifier)) 493 ((optional_expr) (%empty)) 494 ((optional_expr) (expr)) 495 ((variable_class_name) (fully_dereferenceable)) 496 ((fully_dereferenceable) (variable)) 497 ((fully_dereferenceable) ("(" expr ")")) 498 ((fully_dereferenceable) (dereferenceable_scalar)) 499 ((fully_dereferenceable) (class_constant)) 500 ((array_object_dereferenceable) (fully_dereferenceable)) 501 ((array_object_dereferenceable) (constant)) 502 ((callable_expr) (callable_variable)) 503 ((callable_expr) ("(" expr ")")) 504 ((callable_expr) (dereferenceable_scalar)) 505 ((callable_variable) (simple_variable)) 506 ((callable_variable) (array_object_dereferenceable "[" optional_expr "]")) 507 ((callable_variable) (array_object_dereferenceable "{" expr "}")) 508 ((callable_variable) (array_object_dereferenceable T_OBJECT_OPERATOR property_name argument_list)) 509 ((callable_variable) (array_object_dereferenceable T_NULLSAFE_OBJECT_OPERATOR property_name argument_list)) 510 ((callable_variable) (function_call)) 511 ((variable) (callable_variable)) 512 ((variable) (static_member)) 513 ((variable) (array_object_dereferenceable T_OBJECT_OPERATOR property_name)) 514 ((variable) (array_object_dereferenceable T_NULLSAFE_OBJECT_OPERATOR property_name)) 515 ((simple_variable) (T_VARIABLE)) 516 ((simple_variable) ("$" "{" expr "}")) 517 ((simple_variable) ("$" simple_variable)) 518 ((static_member) (class_name T_PAAMAYIM_NEKUDOTAYIM simple_variable)) 519 ((static_member) (variable_class_name T_PAAMAYIM_NEKUDOTAYIM simple_variable)) 520 ((new_variable) (simple_variable)) 521 ((new_variable) (new_variable "[" optional_expr "]")) 522 ((new_variable) (new_variable "{" expr "}")) 523 ((new_variable) (new_variable T_OBJECT_OPERATOR property_name)) 524 ((new_variable) (new_variable T_NULLSAFE_OBJECT_OPERATOR property_name)) 525 ((new_variable) (class_name T_PAAMAYIM_NEKUDOTAYIM simple_variable)) 526 ((new_variable) (new_variable T_PAAMAYIM_NEKUDOTAYIM simple_variable)) 527 ((member_name) (identifier)) 528 ((member_name) ("{" expr "}")) 529 ((member_name) (simple_variable)) 530 ((property_name) (T_STRING)) 531 ((property_name) ("{" expr "}")) 532 ((property_name) (simple_variable)) 533 ((array_pair_list) (non_empty_array_pair_list)) 534 ((possible_array_pair) (%empty)) 535 ((possible_array_pair) (array_pair)) 536 ((non_empty_array_pair_list) (non_empty_array_pair_list "," possible_array_pair)) 537 ((non_empty_array_pair_list) (possible_array_pair)) 538 ((array_pair) (expr T_DOUBLE_ARROW expr)) 539 ((array_pair) (expr)) 540 ((array_pair) (expr T_DOUBLE_ARROW ampersand variable)) 541 ((array_pair) (ampersand variable)) 542 ((array_pair) (T_ELLIPSIS expr)) 543 ((array_pair) (expr T_DOUBLE_ARROW T_LIST "(" array_pair_list ")")) 544 ((array_pair) (T_LIST "(" array_pair_list ")")) 545 ((encaps_list) (encaps_list encaps_var)) 546 ((encaps_list) (encaps_list T_ENCAPSED_AND_WHITESPACE)) 547 ((encaps_list) (encaps_var)) 548 ((encaps_list) (T_ENCAPSED_AND_WHITESPACE encaps_var)) 549 ((encaps_var) (T_VARIABLE)) 550 ((encaps_var) (T_VARIABLE "[" encaps_var_offset "]")) 551 ((encaps_var) (T_VARIABLE T_OBJECT_OPERATOR T_STRING)) 552 ((encaps_var) (T_VARIABLE T_NULLSAFE_OBJECT_OPERATOR T_STRING)) 553 ((encaps_var) (T_DOLLAR_OPEN_CURLY_BRACES expr "}")) 554 ((encaps_var) (T_DOLLAR_OPEN_CURLY_BRACES T_STRING_VARNAME "}")) 555 ((encaps_var) (T_DOLLAR_OPEN_CURLY_BRACES T_STRING_VARNAME "[" expr "]" "}")) 556 ((encaps_var) (T_CURLY_OPEN variable "}")) 557 ((encaps_var_offset) (T_STRING)) 558 ((encaps_var_offset) (T_NUM_STRING)) 559 ((encaps_var_offset) ("-" T_NUM_STRING)) 560 ((encaps_var_offset) (T_VARIABLE)) 561 ((internal_functions_in_yacc) (T_ISSET "(" isset_variables possible_comma ")")) 562 ((internal_functions_in_yacc) (T_EMPTY "(" expr ")")) 563 ((internal_functions_in_yacc) (T_INCLUDE expr)) 564 ((internal_functions_in_yacc) (T_INCLUDE_ONCE expr)) 565 ((internal_functions_in_yacc) (T_EVAL "(" expr ")")) 566 ((internal_functions_in_yacc) (T_REQUIRE expr)) 567 ((internal_functions_in_yacc) (T_REQUIRE_ONCE expr)) 568 ((isset_variables) (isset_variable)) 569 ((isset_variables) (isset_variables "," isset_variable)) 570 ((isset_variable) (expr))))
  "The hash-table indexed by production-number and value is production.")

(defvar
  phps-mode-parser--table-look-aheads
  #s(hash-table size 217 test equal rehash-size 1.5 rehash-threshold 0.8125 data (("!") t ("\"") t ($) t ("$") t ("%") t ("(") t (")") t ("*") t ("+") t (",") t ("-") t (".") t ("/") t (":") t (";") t ("<") t ("=") t (">") t ("?") t ("@") t (PREC_ARROW_FUNCTION) t (T_ABSTRACT) t (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) t (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) t (T_AND_EQUAL) t (T_ARRAY) t (T_ARRAY_CAST) t (T_AS) t (T_ATTRIBUTE) t (T_BOOLEAN_AND) t (T_BOOLEAN_OR) t (T_BOOL_CAST) t (T_BREAK) t (T_CALLABLE) t (T_CASE) t (T_CATCH) t (T_CLASS) t (T_CLASS_C) t (T_CLONE) t (T_COALESCE) t (T_COALESCE_EQUAL) t (T_CONCAT_EQUAL) t (T_CONST) t (T_CONSTANT_ENCAPSED_STRING) t (T_CONTINUE) t (T_CURLY_OPEN) t (T_DEC) t (T_DECLARE) t (T_DEFAULT) t (T_DIR) t (T_DIV_EQUAL) t (T_DNUMBER) t (T_DO) t (T_DOLLAR_OPEN_CURLY_BRACES) t (T_DOUBLE_ARROW) t (T_DOUBLE_CAST) t (T_ECHO) t (T_ELLIPSIS) t (T_ELSE) t (T_ELSEIF) t (T_EMPTY) t (T_ENCAPSED_AND_WHITESPACE) t (T_ENDDECLARE) t (T_ENDFOR) t (T_ENDFOREACH) t (T_ENDIF) t (T_ENDSWITCH) t (T_ENDWHILE) t (T_END_HEREDOC) t (T_ENUM) t (T_EVAL) t (T_EXIT) t (T_EXTENDS) t (T_FILE) t (T_FINAL) t (T_FINALLY) t (T_FN) t (T_FOR) t (T_FOREACH) t (T_FUNCTION) t (T_FUNC_C) t (T_GLOBAL) t (T_GOTO) t (T_HALT_COMPILER) t (T_IF) t (T_IMPLEMENTS) t (T_INC) t (T_INCLUDE) t (T_INCLUDE_ONCE) t (T_INLINE_HTML) t (T_INSTANCEOF) t (T_INSTEADOF) t (T_INTERFACE) t (T_INT_CAST) t (T_ISSET) t (T_IS_EQUAL) t (T_IS_GREATER_OR_EQUAL) t (T_IS_IDENTICAL) t (T_IS_NOT_EQUAL) t (T_IS_NOT_IDENTICAL) t (T_IS_SMALLER_OR_EQUAL) t (T_LINE) t (T_LIST) t (T_LNUMBER) t (T_LOGICAL_AND) t (T_LOGICAL_OR) t (T_LOGICAL_XOR) t (T_MATCH) t (T_METHOD_C) t (T_MINUS_EQUAL) t (T_MOD_EQUAL) t (T_MUL_EQUAL) t (T_NAMESPACE) t (T_NAME_FULLY_QUALIFIED) t (T_NAME_QUALIFIED) t (T_NAME_RELATIVE) t (T_NEW) t (T_NOELSE) t (T_NS_C) t (T_NS_SEPARATOR) t (T_NULLSAFE_OBJECT_OPERATOR) t (T_NUM_STRING) t (T_OBJECT_CAST) t (T_OBJECT_OPERATOR) t (T_OR_EQUAL) t (T_PAAMAYIM_NEKUDOTAYIM) t (T_PLUS_EQUAL) t (T_POW) t (T_POW_EQUAL) t (T_PRINT) t (T_PRIVATE) t (T_PROTECTED) t (T_PUBLIC) t (T_READONLY) t (T_REQUIRE) t (T_REQUIRE_ONCE) t (T_RETURN) t (T_SL) t (T_SL_EQUAL) t (T_SPACESHIP) t (T_SR) t (T_SR_EQUAL) t (T_START_HEREDOC) t (T_STATIC) t (T_STRING) t (T_STRING_CAST) t (T_STRING_VARNAME) t (T_SWITCH) t (T_THROW) t (T_TRAIT) t (T_TRAIT_C) t (T_TRY) t (T_UNSET) t (T_UNSET_CAST) t (T_USE) t (T_VAR) t (T_VARIABLE) t (T_WHILE) t (T_XOR_EQUAL) t (T_YIELD) t (T_YIELD_FROM) t ("[") t ("]") t ("^") t ("`") t ("{") t ("|") t ("}") t ("~") t))
  "The hash-table of valid look-aheads.")

(defvar
  phps-mode-parser--table-terminal-p
  #s(hash-table size 217 test equal rehash-size 1.5 rehash-threshold 0.8125 data (T_NUM_STRING t T_CURLY_OPEN t T_STRING_VARNAME t T_DOLLAR_OPEN_CURLY_BRACES t "$" t T_NULLSAFE_OBJECT_OPERATOR t T_OBJECT_OPERATOR t T_END_HEREDOC t T_START_HEREDOC t T_DNUMBER t T_LNUMBER t "\"" t T_CONSTANT_ENCAPSED_STRING t T_ENCAPSED_AND_WHITESPACE t PREC_ARROW_FUNCTION t T_YIELD_FROM t "`" t "@" t T_UNSET_CAST t T_BOOL_CAST t T_OBJECT_CAST t T_ARRAY_CAST t T_STRING_CAST t T_DOUBLE_CAST t T_INT_CAST t T_COALESCE t T_SPACESHIP t T_IS_GREATER_OR_EQUAL t ">" t T_IS_SMALLER_OR_EQUAL t "<" t T_IS_NOT_EQUAL t T_IS_EQUAL t T_IS_NOT_IDENTICAL t T_IS_IDENTICAL t "!" t "~" t T_SR t T_SL t "%" t "/" t T_POW t "*" t "-" t "+" t "." t "^" t T_BOOLEAN_AND t T_BOOLEAN_OR t T_DEC t T_INC t T_COALESCE_EQUAL t T_SR_EQUAL t T_SL_EQUAL t T_XOR_EQUAL t T_OR_EQUAL t T_AND_EQUAL t T_MOD_EQUAL t T_CONCAT_EQUAL t T_DIV_EQUAL t T_POW_EQUAL t T_MUL_EQUAL t T_MINUS_EQUAL t T_PLUS_EQUAL t T_PAAMAYIM_NEKUDOTAYIM t "?" t T_NOELSE t "[" t "=" t T_ELLIPSIS t T_VARIABLE t "|" t ":" t T_DOUBLE_ARROW t T_INLINE_HTML t T_NS_SEPARATOR t "}" t "{" t ";" t ")" t "(" t T_HALT_COMPILER t "]" t T_ATTRIBUTE t "," t T_NAME_RELATIVE t T_NAME_FULLY_QUALIFIED t T_NAME_QUALIFIED t T_STRING t T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG t T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG t T_READONLY t T_PUBLIC t T_PROTECTED t T_PRIVATE t T_FINAL t T_ABSTRACT t T_STATIC t T_ENUM t T_MATCH t T_FN t T_NS_C t T_DIR t T_FILE t T_LINE t T_METHOD_C t T_FUNC_C t T_TRAIT_C t T_CLASS_C t T_CLASS t T_INTERFACE t T_TRAIT t T_NAMESPACE t T_IMPLEMENTS t T_EXTENDS t T_CALLABLE t T_ARRAY t T_BREAK t T_DEFAULT t T_CASE t T_ENDSWITCH t T_SWITCH t T_LIST t T_YIELD t T_PRINT t T_RETURN t T_CONST t T_FUNCTION t T_GOTO t T_CONTINUE t T_EMPTY t T_ISSET t T_UNSET t T_VAR t T_GLOBAL t T_INSTEADOF t T_USE t T_THROW t T_FINALLY t T_CATCH t T_TRY t T_AS t T_ENDDECLARE t T_DECLARE t T_ENDFOREACH t T_FOREACH t T_ENDFOR t T_FOR t T_ENDWHILE t T_WHILE t T_DO t T_ECHO t T_ENDIF t T_ELSE t T_ELSEIF t T_IF t T_EXIT t T_CLONE t T_NEW t T_INSTANCEOF t T_LOGICAL_AND t T_LOGICAL_XOR t T_LOGICAL_OR t T_REQUIRE_ONCE t T_REQUIRE t T_EVAL t T_INCLUDE_ONCE t T_INCLUDE t))
  "The hash-table of valid terminals.")

(defvar
  phps-mode-parser--table-non-terminal-p
  #s(hash-table size 217 test equal rehash-size 1.5 rehash-threshold 0.8125 data (isset_variable t isset_variables t encaps_var_offset t encaps_var t possible_array_pair t array_pair t non_empty_array_pair_list t static_member t function_call t property_name t array_object_dereferenceable t callable_variable t fully_dereferenceable t class_constant t constant t dereferenceable_scalar t encaps_list t new_variable t callable_expr t variable_class_name t member_name t lexical_var t lexical_var_list t backup_lex_pos t fn t lexical_vars t match t inline_function t backticks_expr t scalar t exit_expr t internal_functions_in_yacc t new_expr t anonymous_class t class_name_reference t ctor_arguments t non_empty_for_exprs t echo_expr t class_const_decl t property t non_empty_member_modifiers t member_modifier t trait_method_reference t absolute_trait_method_reference t trait_alias t trait_precedence t trait_adaptation t trait_adaptation_list t trait_adaptations t attributed_class_statement t enum_case t method_body t class_const_list t method_modifiers t property_list t variable_modifiers t class_statement t static_var t simple_variable t global_var t argument t non_empty_argument_list t name t intersection_type_without_static t union_type_without_static t type_without_static t intersection_type t union_type t type t type_expr_without_static t is_variadic t is_reference t optional_type_without_static t property_modifier t optional_property_modifiers t parameter t attributed_parameter t non_empty_parameter_list t alt_if_stmt_without_else t if_stmt_without_else t match_arm_cond_list t match_arm t non_empty_match_arm_list t match_arm_list t case_separator t case_list t array_pair_list t ampersand t class_name_list t enum_case_expr t type_expr t enum_backing_type t interface_extends_list t class_modifier t class_statement_list t implements_list t extends_from t class_modifiers t backup_fn_flags t return_type t parameter_list t backup_doc_comment t returns_ref t function t variable t unset_variable t optional_variable t catch_name_list t finally_statement t catch_list t declare_statement t foreach_statement t foreach_variable t unset_variables t echo_expr_list t static_var_list t global_var_list t optional_expr t switch_case_list t for_statement t for_exprs t while_statement t expr t alt_if_stmt t if_stmt t inner_statement t inner_statement_list t const_decl t use_declaration t unprefixed_use_declaration t inline_use_declaration t inline_use_declarations t unprefixed_use_declarations t legacy_namespace_name t const_list t use_declarations t group_use_declaration t use_type t mixed_group_use_declaration t namespace_declaration_name t attributed_statement t statement t enum_declaration_statement t interface_declaration_statement t trait_declaration_statement t class_declaration_statement t function_declaration_statement t attributes t attribute t possible_comma t attribute_group t attribute_decl t argument_list t class_name t namespace_name t identifier t top_statement t semi_reserved t reserved_non_modifiers t top_statement_list t))
  "The hash-table of valid non-terminals.")

(defvar
  phps-mode-parser--table-translations
  #s(hash-table size 65 test equal rehash-size 1.5 rehash-threshold 0.8125 data ())
  "The hash-table of translations.")

(defvar
  phps-mode-parser--e-identifier
  '%empty
  "The e-identifier.")

(defvar
  phps-mode-parser--eof-identifier
  '$
  "The end-of-file-identifier.")

(defvar
  phps-mode-parser--look-ahead-number
  1
  "The look-ahead number.")


;;; Local Variables:

(defvar-local
  phps-mode-parser-lex-analyzer--index
  0
  "The current index of the lex-analyzer.")

(defvar-local
  phps-mode-parser-lex-analyzer--move-to-index-flag
  nil
  "Non-nil means move index to value.")


;;; Variable Functions:

(defvar
  phps-mode-parser-lex-analyzer--get-function
  (lambda (token) (let ((start (car (cdr token))) (end (cdr (cdr token)))) (if (<= end (point-max)) (progn (buffer-substring-no-properties start end)))))
  "The lex-analyzer get function.")

(defvar
  phps-mode-parser-lex-analyzer--function
  (lambda (buffer-index) (let ((result (gethash buffer-index phps-mode-parser-tokens)) (token)) (if result (progn (cond ((numberp result) (setq phps-mode-parser-lex-analyzer--move-to-index-flag result)) ((listp result) (setq token result))))) token))
  "The lex-analyzer function.")

(defvar
  phps-mode-parser-lex-analyzer--reset-function
  (lambda nil (progn))
  "The lex-analyzer reset function.")


;;; Functions:


;;; Functions for Lex-Analyzer:

(defun
  phps-mode-parser-lex-analyzer--get-function (token)
  "Get information about TOKEN."
  (let ((meta-information))
    (condition-case
      error
      (progn
        (setq
          meta-information
          (funcall
            phps-mode-parser-lex-analyzer--get-function
            token)))
      (error (error
        "Lex-analyze failed to get token meta-data of %s, error: %s"
        token
        (car (cdr error)))))
    (unless meta-information
      (error "Could not find any token meta-information for: %s" token))
    meta-information))

(defun
  phps-mode-parser-lex-analyzer--reset
  ()
  "Reset Lex-Analyzer."
  (setq
    phps-mode-parser-lex-analyzer--index
    1)
  (when
    phps-mode-parser-lex-analyzer--reset-function
    (funcall
      phps-mode-parser-lex-analyzer--reset-function)))

(defun
  phps-mode-parser-lex-analyzer--peek-next-look-ahead
  ()
  "Peek next look-ahead number of tokens via lex-analyzer."
  (let ((look-ahead)
        (look-ahead-length 0)
        (index phps-mode-parser-lex-analyzer--index)
        (k (max
            1
            phps-mode-parser--look-ahead-number)))
    (while (<
            look-ahead-length
            k)
      (condition-case error
          (progn
            (setq-local
              phps-mode-parser-lex-analyzer--move-to-index-flag
              nil)
            (let ((next-look-ahead
                   (funcall
                    phps-mode-parser-lex-analyzer--function
                    index)))
                  (if phps-mode-parser-lex-analyzer--move-to-index-flag
                    (setq
                     index
                     phps-mode-parser-lex-analyzer--move-to-index-flag)
              (if next-look-ahead
                  (progn
                    (unless (listp (car next-look-ahead))
                      (setq next-look-ahead (list next-look-ahead)))
                    (dolist (next-look-ahead-item next-look-ahead)
                      (when (<
                             look-ahead-length
                             k)
                        (push next-look-ahead-item look-ahead)
                        (setq look-ahead-length (1+ look-ahead-length))
                        (setq index (cdr (cdr next-look-ahead-item))))))
                (push (list phps-mode-parser--eof-identifier) look-ahead)
                (setq look-ahead-length (1+ look-ahead-length))
                (setq index (1+ index))))))
        (error
         (error
          "Lex-analyze failed to peek next look-ahead at %s, error: %s"
          index
          error))))
    (nreverse look-ahead)))

(defun
  phps-mode-parser-lex-analyzer--pop-token ()
  "Pop next token via lex-analyzer."
  (let ((continue t)
        (tokens))
    (while continue
      (condition-case error
          (progn
            (setq-local
              phps-mode-parser-lex-analyzer--move-to-index-flag
              nil)
            (let ((token
                   (funcall
                    phps-mode-parser-lex-analyzer--function
                    phps-mode-parser-lex-analyzer--index)))
              (if phps-mode-parser-lex-analyzer--move-to-index-flag
                  (progn
                    (setq-local
                     phps-mode-parser-lex-analyzer--index
                     phps-mode-parser-lex-analyzer--move-to-index-flag))
              (when token
                (unless (listp (car token))
                  (setq token (list token)))
                (let ((first-token (car token)))
                  (setq
                   phps-mode-parser-lex-analyzer--index
                   (cdr (cdr first-token)))
                  (push 
                   first-token 
                   tokens)))
                (setq
                 continue
                 nil))))
        (error (error
                "Lex-analyze failed to pop token at %s, error: %s"
                phps-mode-parser-lex-analyzer--index
                (car (cdr error))))))
    (nreverse tokens)))


;;; Functions for Syntax-Analyzer / Parser:


(defun
  phps-mode-parser--get-grammar-production-by-number
  (production-number)
  "If PRODUCTION-NUMBER exist, return it's production."
  (gethash
   production-number
   phps-mode-parser--table-productions-number-reverse))

(defun
  phps-mode-parser--valid-symbol-p
  (symbol)
  "Return whether SYMBOL is valid or not."
  (let ((is-valid t))
    (unless (or
             (phps-mode-parser--valid-e-p symbol)
             (phps-mode-parser--valid-eof-p symbol)
             (phps-mode-parser--valid-non-terminal-p symbol)
             (phps-mode-parser--valid-terminal-p symbol))
      (setq is-valid nil))
    is-valid))

(defun
  phps-mode-parser--valid-e-p
  (symbol)
  "Return whether SYMBOL is the e identifier or not."
  (eq
   symbol
   phps-mode-parser--e-identifier))

(defun
  phps-mode-parser--valid-eof-p
  (symbol)
  "Return whether SYMBOL is the EOF identifier or not."
  (eq
    symbol
    phps-mode-parser--eof-identifier))

(defun
  phps-mode-parser--valid-non-terminal-p (symbol)
  "Return whether SYMBOL is a non-terminal in grammar or not."
  (gethash
   symbol
   phps-mode-parser--table-non-terminal-p))

(defun
  phps-mode-parser--valid-terminal-p (symbol)
  "Return whether SYMBOL is a terminal in grammar or not."
  (gethash
   symbol
   phps-mode-parser--table-terminal-p))

(defun
  phps-mode-parser--get-grammar-translation-by-number
  (production-number)
  "If translation for PRODUCTION-NUMBER exist, return it."
  (gethash
    production-number
    phps-mode-parser--table-translations))

(defun
  phps-mode-parser--parse
  (&optional
    perform-sdt
    input-tape-index
    pushdown-list
    output
    translation
    translation-symbol-table-list
    history)
  "Perform a LR-parse via lex-analyzer, optionally PERFORM-SDT means to
  perform syntax-directed translation and optioanlly start at INPUT-TAPE-INDEX
  with PUSHDOWN-LIST, OUTPUT, TRANSLATION, TRANSLATION-SYMBOL-TABLE-LIST
  and HISTORY."
  (unless input-tape-index
    (setq input-tape-index 1))
  (unless pushdown-list
    (push 0 pushdown-list))
  (let ((translation-symbol-table
         (make-hash-table :test 'equal)))
    (when translation-symbol-table-list
      (dolist
          (item translation-symbol-table-list)
        (puthash
         (nth 0 item)
         (nth 1 item)
         translation-symbol-table)))

    (if (and
         input-tape-index
         (> input-tape-index 1))
        (setq
         phps-mode-parser-lex-analyzer--index
         input-tape-index)
      (phps-mode-parser-lex-analyzer--reset))

    ;; Make sure tables exists
    (unless phps-mode-parser--action-tables
      (error "Missing action-tables for grammar!"))
    (unless phps-mode-parser--distinct-action-tables
      (error "Missing distinct GOTO-tables for grammar!"))
    (unless phps-mode-parser--goto-tables
      (error "Missing GOTO-tables for grammar!"))
    (unless phps-mode-parser--distinct-goto-tables
      (error "Missing distinct GOTO-tables for grammar!"))

    (let ((accept)
          (pre-index 0))
      (while (not accept)

        ;; Save history when index has changed to enable incremental parsing / translating
        (when
            (>
             phps-mode-parser-lex-analyzer--index
             pre-index)
          ;; We make a copy of the hash-table here to avoid passing same
          ;; hash-table every-time with pointer
          (let ((translation-symbol-table-list))
            (maphash
             (lambda (key value)
               (push
                `(,key ,value)
                translation-symbol-table-list))
             translation-symbol-table)
            (push
             `(,phps-mode-parser-lex-analyzer--index
               ,pushdown-list
               ,output
               ,translation
               ,translation-symbol-table-list)
             history)
            (setq
             pre-index
             phps-mode-parser-lex-analyzer--index)))

        ;; (1) The look-ahead string u, consisting of the next k input symbols, is determined.
        (let ((look-ahead
               (phps-mode-parser-lex-analyzer--peek-next-look-ahead))
              (look-ahead-full))

          ;; Save token stream indexes in separate variable if needed later
          (setq look-ahead-full look-ahead)

          ;; Create simplified look-ahead for logic below
          (setq look-ahead nil)
          (dolist (look-ahead-item look-ahead-full)
            (if (listp look-ahead-item)
                (push (car look-ahead-item) look-ahead)
              (push look-ahead-item look-ahead)))
          (setq look-ahead (nreverse look-ahead))

          (let ((table-index
                 (car pushdown-list)))
            (let ((action-table-distinct-index
                   (gethash
                    table-index
                    phps-mode-parser--action-tables)))
              (let ((action-table
                     (gethash
                      action-table-distinct-index
                      phps-mode-parser--distinct-action-tables)))
              (unless action-table
                (error
                 "Action-table with index %s is empty! Push-down-list: %s"
                 table-index
                 pushdown-list))
              (let ((action-match nil)
                    (action-table-length (length action-table))
                    (action-index 0)
                    (possible-look-aheads))

                ;; (2) The parsing action f of the table on top of the pushdown list is applied to the lookahead string u.
                (while (and
                        (not action-match)
                        (< action-index action-table-length))
                  (let ((action (nth action-index action-table)))
                    (let ((action-look-ahead (car action)))
                      (push
                       action-look-ahead
                       possible-look-aheads)
                      (when
                          (equal
                           action-look-ahead
                           look-ahead)
                        (setq
                         action-match
                         (cdr action)))
                      (when
                          (and
                           (=
                            phps-mode-parser--look-ahead-number
                            0)
                           (not
                            action-look-ahead))
                        ;; LR(0) reduce actions occupy entire row
                        ;; and is applied regardless of look-ahead
                        (setq
                         action-match
                         (cdr action))))
                    (setq
                     action-index
                     (1+ action-index))))

                (unless action-match
                  ;; (c) If f(u) = error, we halt parsing (and, in practice
                  ;; transfer to an error recovery routine).
                  (signal
                    'error
                    (list
                      (format
                        "Invalid syntax! Expected one of %s found %s at %s"

                   possible-look-aheads
                   look-ahead
                   phps-mode-parser-lex-analyzer--index)
                   possible-look-aheads
                   look-ahead
                   phps-mode-parser-lex-analyzer--index)))

                (cond

                 ((equal action-match '(shift))
                  ;; (a) If f(u) = shift, then the next input symbol, say a
                  ;; is removed from the input and shifted onto the pushdown list.
                  ;; The goto function g of the table on top of the pushdown list
                  ;; is applied to a to determine the new table to be placed on
                  ;; top of the pushdown list. We then return to step(1). If
                  ;; there is no next input symbol or g(a) is undefined, halt
                  ;; and declare error.

                  (let ((a (list (car look-ahead)))
                        (a-full (list (car look-ahead-full))))
                      (let ((goto-table-distinct-index
                             (gethash
                              table-index
                              phps-mode-parser--goto-tables)))
                        (let ((goto-table
                               (gethash
                                goto-table-distinct-index
                                phps-mode-parser--distinct-goto-tables)))
                      (let ((goto-table-length (length goto-table))
                            (goto-index 0)
                            (searching-match t)
                            (next-index)
                            (possible-look-aheads))

                        (while (and
                                searching-match
                                (< goto-index goto-table-length))
                          (let ((goto-item (nth goto-index goto-table)))
                            (let ((goto-item-symbol (list (car goto-item)))
                                  (goto-item-next-index (car (cdr goto-item))))
                              (push goto-item-symbol possible-look-aheads)

                              (when (equal
                                     goto-item-symbol
                                     a)
                                (setq next-index goto-item-next-index)
                                (setq searching-match nil))))

                          (setq goto-index (1+ goto-index)))
                        (unless next-index
                          (error
                           "In shift, found no GOTO-item for %s at %s, expected one of %s"
                           a
                           phps-mode-parser-lex-analyzer--index
                           possible-look-aheads))

                        (push (car a-full) pushdown-list)
                        (push next-index pushdown-list)
                        (phps-mode-parser-lex-analyzer--pop-token))))))

                 ((equal (car action-match) 'reduce)
                  ;; (b) If f(u) = reduce i and production i is A -> a,
                  ;; then 2|a| symbols are removed from the top of the pushdown
                  ;; list, and production number i is placed in the output
                  ;; buffer. A new table T' is then exposed as the top table
                  ;; of the pushdown list, and the goto function of T' is applied
                  ;; to A to determine the next table to be placed on top of the
                  ;; pushdown list. We place A and this new table on top of the
                  ;; the pushdown list and return to step (1)

                  (let ((production-number (car (cdr action-match))))

                    (let ((production
                           (phps-mode-parser--get-grammar-production-by-number
                            production-number)))
                      (let ((production-lhs (car production))
                            (production-rhs (car (cdr production)))
                            (popped-items-contents))
                        (unless (equal
                                 production-rhs
                                 (list phps-mode-parser--e-identifier))
                          (let ((pop-items (* 2 (length production-rhs)))
                                (popped-items 0)
                                (popped-item))
                            (while (< popped-items pop-items)
                              (setq popped-item (pop pushdown-list))
                              (when (and
                                     (listp popped-item)
                                     (phps-mode-parser--valid-symbol-p
                                      (car popped-item)))
                                (push
                                 popped-item
                                 popped-items-contents))
                              (setq popped-items (1+ popped-items)))))
                        (push production-number output)

                        (when perform-sdt
                          (let ((popped-items-meta-contents)
                                (popped-items-terminals))
                            (setq
                             popped-items-contents
                             (reverse popped-items-contents))
                            ;; Collect arguments for translation
                            (dolist (popped-item popped-items-contents)
                              (if (and
                                   (listp popped-item)
                                   (cdr popped-item))
                                  ;; If item is a terminal, use it's literal value
                                  (progn
                                    (push
                                      (phps-mode-parser-lex-analyzer--get-function
                                       popped-item)
                                       popped-items-meta-contents)
                                      (push
                                       popped-item
                                       popped-items-terminals))

                                ;; If item is a non-terminal
                                (let ((temp-hash-key
                                       (format
                                        "%S"
                                         popped-item)))

                                ;; If we have a translation for symbol, pop one
                                ;; otherwise push nil on translation argument stack
                                (if (gethash
                                         temp-hash-key
                                         translation-symbol-table)
                                        (let ((symbol-translations
                                               (gethash
                                                temp-hash-key
                                                translation-symbol-table)))
                                          (let ((symbol-translation
                                                 (pop symbol-translations)))
                                            (push
                                             (car symbol-translation)
                                             popped-items-meta-contents)
                                            (push
                                             (car (cdr symbol-translation))
                                             popped-items-terminals)
                                            (puthash
                                             temp-hash-key
                                             symbol-translations
                                             translation-symbol-table)))
                                      (push
                                       nil
                                       popped-items-meta-contents)
                                      (push
                                       nil
                                       popped-items-terminals)))))

                              ;; If we just have one argument, pass it as a instead of a list
                              (when (= (length popped-items-meta-contents) 1)
                                (setq
                                 popped-items-meta-contents
                                 (car popped-items-meta-contents))
                                (setq
                                 popped-items-terminals
                                 (car popped-items-terminals)))

                              ;; Perform translation at reduction if specified
                              (if
                                  (phps-mode-parser--get-grammar-translation-by-number
                                   production-number)
                                  (let ((partial-translation
                                         (funcall
                                          (phps-mode-parser--get-grammar-translation-by-number
                                           production-number)
                                          popped-items-meta-contents
                                          popped-items-terminals)))
                                    (let ((temp-hash-key
                                           (format
                                            "%S"
                                            production-lhs)))
                                      (let ((symbol-translations
                                             (gethash
                                              temp-hash-key
                                              translation-symbol-table)))
                                        (push
                                          (list
                                           partial-translation
                                           popped-items-terminals)
                                           symbol-translations)
                                        (puthash
                                         temp-hash-key
                                         symbol-translations
                                         translation-symbol-table)
                                        (setq
                                         translation
                                         partial-translation))))

                                ;; When no translation is specified just use popped contents as translation
                                (let ((partial-translation
                                       (list
                                        popped-items-meta-contents
                                        popped-items-terminals)))
                                  (let ((temp-hash-key
                                         (format
                                          "%S"
                                          production-lhs)))
                                    (let ((symbol-translations
                                           (gethash
                                            temp-hash-key
                                            translation-symbol-table)))
                                      (push
                                       partial-translation
                                       symbol-translations)
                                      (puthash
                                       temp-hash-key
                                       symbol-translations
                                       translation-symbol-table)
                                      (setq
                                       translation
                                       (car partial-translation))))))))

                          (let ((new-table-index (car pushdown-list)))
                            (let ((goto-table-distinct-index
                                   (gethash
                                    new-table-index
                                    phps-mode-parser--goto-tables)))
                              (let ((goto-table
                                     (gethash
                                      goto-table-distinct-index
                                      phps-mode-parser--distinct-goto-tables)))
                                (let ((goto-table-length
                                       (length goto-table))
                                      (goto-index 0)
                                      (searching-match t)
                                      (next-index))

                                  (while (and
                                          searching-match
                                          (< goto-index goto-table-length))
                                    (let ((goto-item (nth goto-index goto-table)))
                                      (let ((goto-item-symbol (list (car goto-item)))
                                            (goto-item-next-index (car (cdr goto-item))))

                                        (when (equal
                                               goto-item-symbol
                                               production-lhs)
                                          (setq next-index goto-item-next-index)
                                          (setq searching-match nil))))

                                    (setq goto-index (1+ goto-index)))

                                  (when next-index
                                    (push production-lhs pushdown-list)
                                    (push next-index pushdown-list))))))))))

                   ((equal action-match '(accept))
                    ;;    (d) If f(u) = accept, we halt and declare the string
                    ;;    in the output buffer to be the right parse of the original
                    ;;    input string.

                    (setq accept t))
                   (t (error
                       "Invalid action-match: %s!"
                       action-match)))))))))
      (unless accept
        (error
         "Parsed entire string without getting accepting! Output: %s"
         (reverse output)))
      (when history
        (setq history (reverse history)))
      (when output
        (setq output (reverse output)))
      (let ((translation-symbol-table-list))
        (when translation-symbol-table
          (maphash
           (lambda (key value)
             (push
              `(,key ,value)
              translation-symbol-table-list))
           translation-symbol-table))
        (list
         output
         translation
         translation-symbol-table-list
         history)))))

(defun phps-mode-parser-parse
    (&optional
     input-tape-index
     pushdown-list
     output
     translation
     history)
  "Perform a LR-parse via lex-analyzer, optionally at INPUT-TAPE-INDEX with
  PUSHDOWN-LIST, OUTPUT, TRANSLATION and HISTORY."
  (let ((result
         (phps-mode-parser--parse
          nil
          input-tape-index
          pushdown-list
          output
          translation
          history)))
    (nth 0 result)))

(defun phps-mode-parser-translate
    (&optional
     input-tape-index
     pushdown-list
     output
     translation
     history)
  "Perform a LR-parse via lex-analyzer, optionally at INPUT-TAPE-INDEX with
  PUSHDOWN-LIST, OUTPUT, TRANSLATION and HISTORY."
  (let ((result
         (phps-mode-parser--parse
          t
          input-tape-index
          pushdown-list
          output
          translation
          history)))
    (nth 1 result)))

(provide 'phps-mode-parser)

;;; phps-mode-parser.el ends here
